#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#define LANGUAGE_VERSION 14
#define STATE_COUNT 2742
#define LARGE_STATE_COUNT 593
#define SYMBOL_COUNT 430
#define ALIAS_COUNT 1
#define TOKEN_COUNT 195
#define EXTERNAL_TOKEN_COUNT 12
#define FIELD_COUNT 31
#define MAX_ALIAS_SEQUENCE_LENGTH 12
#define PRODUCTION_ID_COUNT 217

enum ts_symbol_identifiers {
  sym_name = 1,
  sym_php_tag = 2,
  anon_sym_QMARK_GT = 3,
  aux_sym_text_token1 = 4,
  aux_sym_text_token2 = 5,
  anon_sym_SEMI = 6,
  anon_sym_AMP = 7,
  aux_sym_function_static_declaration_token1 = 8,
  anon_sym_COMMA = 9,
  anon_sym_EQ = 10,
  aux_sym_global_declaration_token1 = 11,
  aux_sym_namespace_definition_token1 = 12,
  aux_sym_namespace_use_declaration_token1 = 13,
  aux_sym_namespace_use_clause_token1 = 14,
  aux_sym__namespace_use_type_token1 = 15,
  aux_sym__namespace_use_type_token2 = 16,
  anon_sym_BSLASH = 17,
  anon_sym_LBRACE = 18,
  anon_sym_RBRACE = 19,
  aux_sym_trait_declaration_token1 = 20,
  aux_sym_interface_declaration_token1 = 21,
  aux_sym_base_clause_token1 = 22,
  aux_sym_enum_declaration_token1 = 23,
  anon_sym_COLON = 24,
  anon_sym_string = 25,
  anon_sym_int = 26,
  aux_sym_enum_case_token1 = 27,
  aux_sym_class_declaration_token1 = 28,
  aux_sym_final_modifier_token1 = 29,
  aux_sym_abstract_modifier_token1 = 30,
  aux_sym_readonly_modifier_token1 = 31,
  aux_sym_class_interface_clause_token1 = 32,
  anon_sym_EQ_GT = 33,
  sym_var_modifier = 34,
  aux_sym_use_instead_of_clause_token1 = 35,
  aux_sym_visibility_modifier_token1 = 36,
  aux_sym_visibility_modifier_token2 = 37,
  aux_sym_visibility_modifier_token3 = 38,
  anon_sym_LPAREN = 39,
  anon_sym_RPAREN = 40,
  anon_sym_LPAREN2 = 41,
  anon_sym_RPAREN2 = 42,
  aux_sym__arrow_function_header_token1 = 43,
  anon_sym_DOT_DOT_DOT = 44,
  anon_sym_QMARK = 45,
  sym_bottom_type = 46,
  anon_sym_PIPE = 47,
  anon_sym_array = 48,
  aux_sym_primitive_type_token1 = 49,
  anon_sym_iterable = 50,
  anon_sym_bool = 51,
  anon_sym_float = 52,
  anon_sym_void = 53,
  anon_sym_mixed = 54,
  anon_sym_false = 55,
  anon_sym_null = 56,
  anon_sym_true = 57,
  aux_sym_cast_type_token1 = 58,
  aux_sym_cast_type_token2 = 59,
  aux_sym_cast_type_token3 = 60,
  aux_sym_cast_type_token4 = 61,
  aux_sym_cast_type_token5 = 62,
  aux_sym_cast_type_token6 = 63,
  aux_sym_cast_type_token7 = 64,
  aux_sym_cast_type_token8 = 65,
  aux_sym_cast_type_token9 = 66,
  aux_sym_cast_type_token10 = 67,
  aux_sym_cast_type_token11 = 68,
  aux_sym_cast_type_token12 = 69,
  aux_sym_echo_statement_token1 = 70,
  aux_sym_exit_statement_token1 = 71,
  anon_sym_unset = 72,
  aux_sym_declare_statement_token1 = 73,
  aux_sym_declare_statement_token2 = 74,
  anon_sym_ticks = 75,
  anon_sym_encoding = 76,
  anon_sym_strict_types = 77,
  sym_float = 78,
  aux_sym_try_statement_token1 = 79,
  aux_sym_catch_clause_token1 = 80,
  aux_sym_finally_clause_token1 = 81,
  aux_sym_goto_statement_token1 = 82,
  aux_sym_continue_statement_token1 = 83,
  aux_sym_break_statement_token1 = 84,
  sym_integer = 85,
  aux_sym_return_statement_token1 = 86,
  aux_sym_throw_expression_token1 = 87,
  aux_sym_while_statement_token1 = 88,
  aux_sym_while_statement_token2 = 89,
  aux_sym_do_statement_token1 = 90,
  aux_sym_for_statement_token1 = 91,
  aux_sym_for_statement_token2 = 92,
  aux_sym_foreach_statement_token1 = 93,
  aux_sym_foreach_statement_token2 = 94,
  aux_sym_if_statement_token1 = 95,
  aux_sym_if_statement_token2 = 96,
  aux_sym_else_if_clause_token1 = 97,
  aux_sym_else_clause_token1 = 98,
  aux_sym_match_expression_token1 = 99,
  aux_sym_match_default_expression_token1 = 100,
  aux_sym_switch_statement_token1 = 101,
  aux_sym_switch_block_token1 = 102,
  anon_sym_PLUS = 103,
  anon_sym_DASH = 104,
  anon_sym_TILDE = 105,
  anon_sym_BANG = 106,
  anon_sym_AT = 107,
  aux_sym_clone_expression_token1 = 108,
  anon_sym_COLON_COLON = 109,
  aux_sym_print_intrinsic_token1 = 110,
  aux_sym__new_non_dereferencable_expression_token1 = 111,
  anon_sym_DASH_DASH = 112,
  anon_sym_PLUS_PLUS = 113,
  anon_sym_STAR_STAR_EQ = 114,
  anon_sym_STAR_EQ = 115,
  anon_sym_SLASH_EQ = 116,
  anon_sym_PERCENT_EQ = 117,
  anon_sym_PLUS_EQ = 118,
  anon_sym_DASH_EQ = 119,
  anon_sym_DOT_EQ = 120,
  anon_sym_LT_LT_EQ = 121,
  anon_sym_GT_GT_EQ = 122,
  anon_sym_AMP_EQ = 123,
  anon_sym_CARET_EQ = 124,
  anon_sym_PIPE_EQ = 125,
  anon_sym_QMARK_QMARK_EQ = 126,
  anon_sym_DASH_GT = 127,
  anon_sym_QMARK_DASH_GT = 128,
  aux_sym__list_destructing_token1 = 129,
  anon_sym_LBRACK = 130,
  anon_sym_RBRACK = 131,
  anon_sym_self = 132,
  anon_sym_parent = 133,
  aux_sym__argument_name_token1 = 134,
  aux_sym__argument_name_token2 = 135,
  anon_sym_POUND_LBRACK = 136,
  sym_escape_sequence = 137,
  anon_sym_BSLASHu = 138,
  aux_sym_encapsed_string_token1 = 139,
  anon_sym_DQUOTE = 140,
  aux_sym_string_token1 = 141,
  anon_sym_SQUOTE = 142,
  aux_sym_string_token2 = 143,
  aux_sym_string_content_token1 = 144,
  anon_sym_LT_LT_LT = 145,
  anon_sym_DQUOTE2 = 146,
  sym__new_line = 147,
  anon_sym_SQUOTE2 = 148,
  anon_sym_BQUOTE = 149,
  anon_sym_DOLLAR = 150,
  aux_sym_yield_expression_token1 = 151,
  aux_sym_yield_expression_token2 = 152,
  aux_sym_binary_expression_token1 = 153,
  anon_sym_QMARK_QMARK = 154,
  anon_sym_STAR_STAR = 155,
  aux_sym_binary_expression_token2 = 156,
  aux_sym_binary_expression_token3 = 157,
  aux_sym_binary_expression_token4 = 158,
  anon_sym_PIPE_PIPE = 159,
  anon_sym_AMP_AMP = 160,
  anon_sym_CARET = 161,
  anon_sym_EQ_EQ = 162,
  anon_sym_BANG_EQ = 163,
  anon_sym_LT_GT = 164,
  anon_sym_EQ_EQ_EQ = 165,
  anon_sym_BANG_EQ_EQ = 166,
  anon_sym_LT = 167,
  anon_sym_GT = 168,
  anon_sym_LT_EQ = 169,
  anon_sym_GT_EQ = 170,
  anon_sym_LT_EQ_GT = 171,
  anon_sym_LT_LT = 172,
  anon_sym_GT_GT = 173,
  anon_sym_DOT = 174,
  anon_sym_STAR = 175,
  anon_sym_SLASH = 176,
  anon_sym_PERCENT = 177,
  aux_sym_include_expression_token1 = 178,
  aux_sym_include_once_expression_token1 = 179,
  aux_sym_require_expression_token1 = 180,
  aux_sym_require_once_expression_token1 = 181,
  sym_comment = 182,
  sym__automatic_semicolon = 183,
  sym_encapsed_string_chars = 184,
  sym_encapsed_string_chars_after_variable = 185,
  sym_execution_string_chars = 186,
  sym_execution_string_chars_after_variable = 187,
  sym_encapsed_string_chars_heredoc = 188,
  sym_encapsed_string_chars_after_variable_heredoc = 189,
  sym__eof = 190,
  sym_heredoc_start = 191,
  sym_heredoc_end = 192,
  sym_nowdoc_string = 193,
  sym_sentinel_error = 194,
  sym_program = 195,
  sym_text_interpolation = 196,
  sym_text = 197,
  sym_statement = 198,
  sym_empty_statement = 199,
  sym_reference_modifier = 200,
  sym_function_static_declaration = 201,
  sym_static_variable_declaration = 202,
  sym_global_declaration = 203,
  sym_namespace_definition = 204,
  sym_namespace_use_declaration = 205,
  sym_namespace_use_clause = 206,
  sym_qualified_name = 207,
  sym__name = 208,
  sym_namespace_name = 209,
  sym__namespace_use_group = 210,
  sym_namespace_use_group = 211,
  sym_trait_declaration = 212,
  sym_interface_declaration = 213,
  sym_base_clause = 214,
  sym_enum_declaration = 215,
  sym_enum_declaration_list = 216,
  sym__enum_member_declaration = 217,
  sym_enum_case = 218,
  sym_class_declaration = 219,
  sym_declaration_list = 220,
  sym_final_modifier = 221,
  sym_abstract_modifier = 222,
  sym_readonly_modifier = 223,
  sym_class_interface_clause = 224,
  sym__member_declaration = 225,
  sym_const_declaration = 226,
  sym__class_const_declaration = 227,
  sym__const_declaration = 228,
  sym_property_declaration = 229,
  sym__modifier = 230,
  sym_property_element = 231,
  sym_property_hook_list = 232,
  sym_property_hook = 233,
  sym__property_hook_body = 234,
  sym_method_declaration = 235,
  sym_static_modifier = 236,
  sym_use_declaration = 237,
  sym_use_list = 238,
  sym_use_instead_of_clause = 239,
  sym_use_as_clause = 240,
  sym_visibility_modifier = 241,
  sym_function_definition = 242,
  sym__function_definition_header = 243,
  sym_anonymous_function = 244,
  sym_anonymous_function_use_clause = 245,
  sym__anonymous_function_header = 246,
  sym__arrow_function_header = 247,
  sym_arrow_function = 248,
  sym_formal_parameters = 249,
  sym_property_promotion_parameter = 250,
  sym_simple_parameter = 251,
  sym_variadic_parameter = 252,
  sym_type = 253,
  sym__types = 254,
  sym_named_type = 255,
  sym_optional_type = 256,
  sym_union_type = 257,
  sym_intersection_type = 258,
  sym_disjunctive_normal_form_type = 259,
  sym_primitive_type = 260,
  sym_cast_type = 261,
  sym__return_type = 262,
  sym_const_element = 263,
  sym_echo_statement = 264,
  sym_exit_statement = 265,
  sym_unset_statement = 266,
  sym_declare_statement = 267,
  sym_declare_directive = 268,
  sym_literal = 269,
  sym_try_statement = 270,
  sym_catch_clause = 271,
  sym_type_list = 272,
  sym_finally_clause = 273,
  sym_goto_statement = 274,
  sym_continue_statement = 275,
  sym_break_statement = 276,
  sym_return_statement = 277,
  sym_throw_expression = 278,
  sym_while_statement = 279,
  sym_do_statement = 280,
  sym_for_statement = 281,
  sym__expressions = 282,
  sym_sequence_expression = 283,
  sym_foreach_statement = 284,
  sym_foreach_pair = 285,
  sym__foreach_value = 286,
  sym_if_statement = 287,
  sym_colon_block = 288,
  sym_else_if_clause = 289,
  sym_else_clause = 290,
  sym_else_if_clause_2 = 291,
  sym_else_clause_2 = 292,
  sym_match_expression = 293,
  sym_match_block = 294,
  sym_match_condition_list = 295,
  sym_match_conditional_expression = 296,
  sym_match_default_expression = 297,
  sym_switch_statement = 298,
  sym_switch_block = 299,
  sym_case_statement = 300,
  sym_default_statement = 301,
  sym_compound_statement = 302,
  sym_named_label_statement = 303,
  sym_expression_statement = 304,
  sym_expression = 305,
  sym__unary_expression = 306,
  sym_unary_op_expression = 307,
  sym_error_suppression_expression = 308,
  sym_clone_expression = 309,
  sym_primary_expression = 310,
  sym_parenthesized_expression = 311,
  sym_class_constant_access_expression = 312,
  sym_print_intrinsic = 313,
  sym_object_creation_expression = 314,
  sym__new_non_dereferencable_expression = 315,
  sym__new_dereferencable_expression = 316,
  sym__class_name_reference = 317,
  sym_anonymous_class = 318,
  sym_update_expression = 319,
  sym_cast_expression = 320,
  sym_cast_variable = 321,
  sym_assignment_expression = 322,
  sym_reference_assignment_expression = 323,
  sym_conditional_expression = 324,
  sym_augmented_assignment_expression = 325,
  sym__variable_member_access_expression = 326,
  sym_member_access_expression = 327,
  sym__variable_nullsafe_member_access_expression = 328,
  sym_nullsafe_member_access_expression = 329,
  sym__variable_scoped_property_access_expression = 330,
  sym_scoped_property_access_expression = 331,
  sym_list_literal = 332,
  sym__list_destructing = 333,
  sym__array_destructing = 334,
  sym__array_destructing_element = 335,
  sym_function_call_expression = 336,
  sym__callable_expression = 337,
  sym_scoped_call_expression = 338,
  sym__scope_resolution_qualifier = 339,
  sym_relative_scope = 340,
  sym_variadic_placeholder = 341,
  sym_arguments = 342,
  sym_argument = 343,
  sym__argument_name = 344,
  sym_member_call_expression = 345,
  sym_nullsafe_member_call_expression = 346,
  sym_variadic_unpacking = 347,
  sym__member_name = 348,
  sym__variable_subscript_expression = 349,
  sym__dereferencable_subscript_expression = 350,
  sym__dereferencable_expression = 351,
  sym__dereferencable_scalar = 352,
  sym_array_creation_expression = 353,
  sym_attribute_group = 354,
  sym_attribute_list = 355,
  sym_attribute = 356,
  sym__complex_string_part = 357,
  sym__simple_string_member_access_expression = 358,
  sym__simple_string_subscript_unary_expression = 359,
  sym__simple_string_array_access_argument = 360,
  sym__simple_string_subscript_expression = 361,
  sym__simple_string_part = 362,
  aux_sym__interpolated_string_body = 363,
  aux_sym__interpolated_string_body_heredoc = 364,
  sym_encapsed_string = 365,
  sym_string = 366,
  sym_string_content = 367,
  sym_heredoc_body = 368,
  sym_heredoc = 369,
  sym_nowdoc_body = 370,
  sym_nowdoc = 371,
  aux_sym__interpolated_execution_operator_body = 372,
  sym_shell_command_expression = 373,
  sym_boolean = 374,
  sym_null = 375,
  sym__string = 376,
  sym_dynamic_variable_name = 377,
  sym__simple_variable = 378,
  sym__new_variable = 379,
  sym__callable_variable = 380,
  sym_variable_name = 381,
  sym_by_ref = 382,
  sym_yield_expression = 383,
  sym_array_element_initializer = 384,
  sym_binary_expression = 385,
  sym_include_expression = 386,
  sym_include_once_expression = 387,
  sym_require_expression = 388,
  sym_require_once_expression = 389,
  sym__reserved_identifier = 390,
  sym__identifier = 391,
  sym__semicolon = 392,
  aux_sym_program_repeat1 = 393,
  aux_sym_text_repeat1 = 394,
  aux_sym_function_static_declaration_repeat1 = 395,
  aux_sym_global_declaration_repeat1 = 396,
  aux_sym_namespace_use_declaration_repeat1 = 397,
  aux_sym_namespace_name_repeat1 = 398,
  aux_sym_base_clause_repeat1 = 399,
  aux_sym_enum_declaration_list_repeat1 = 400,
  aux_sym_class_declaration_repeat1 = 401,
  aux_sym_declaration_list_repeat1 = 402,
  aux_sym__const_declaration_repeat1 = 403,
  aux_sym_property_declaration_repeat1 = 404,
  aux_sym_property_hook_list_repeat1 = 405,
  aux_sym_use_list_repeat1 = 406,
  aux_sym_anonymous_function_use_clause_repeat1 = 407,
  aux_sym_formal_parameters_repeat1 = 408,
  aux_sym_union_type_repeat1 = 409,
  aux_sym_intersection_type_repeat1 = 410,
  aux_sym_disjunctive_normal_form_type_repeat1 = 411,
  aux_sym_unset_statement_repeat1 = 412,
  aux_sym_try_statement_repeat1 = 413,
  aux_sym_type_list_repeat1 = 414,
  aux_sym_if_statement_repeat1 = 415,
  aux_sym_if_statement_repeat2 = 416,
  aux_sym_match_block_repeat1 = 417,
  aux_sym_match_condition_list_repeat1 = 418,
  aux_sym_switch_block_repeat1 = 419,
  aux_sym__list_destructing_repeat1 = 420,
  aux_sym__array_destructing_repeat1 = 421,
  aux_sym_arguments_repeat1 = 422,
  aux_sym_array_creation_expression_repeat1 = 423,
  aux_sym_attribute_group_repeat1 = 424,
  aux_sym_attribute_list_repeat1 = 425,
  aux_sym_string_repeat1 = 426,
  aux_sym_string_content_repeat1 = 427,
  aux_sym_heredoc_body_repeat1 = 428,
  aux_sym_nowdoc_body_repeat1 = 429,
  alias_sym_operation = 430,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym_name] = "name",
  [sym_php_tag] = "php_tag",
  [anon_sym_QMARK_GT] = "\?>",
  [aux_sym_text_token1] = "text_token1",
  [aux_sym_text_token2] = "text_token2",
  [anon_sym_SEMI] = ";",
  [anon_sym_AMP] = "&",
  [aux_sym_function_static_declaration_token1] = "static",
  [anon_sym_COMMA] = ",",
  [anon_sym_EQ] = "=",
  [aux_sym_global_declaration_token1] = "global",
  [aux_sym_namespace_definition_token1] = "namespace",
  [aux_sym_namespace_use_declaration_token1] = "use",
  [aux_sym_namespace_use_clause_token1] = "as",
  [aux_sym__namespace_use_type_token1] = "function",
  [aux_sym__namespace_use_type_token2] = "const",
  [anon_sym_BSLASH] = "\\",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [aux_sym_trait_declaration_token1] = "trait",
  [aux_sym_interface_declaration_token1] = "interface",
  [aux_sym_base_clause_token1] = "extends",
  [aux_sym_enum_declaration_token1] = "enum",
  [anon_sym_COLON] = ":",
  [anon_sym_string] = "string",
  [anon_sym_int] = "int",
  [aux_sym_enum_case_token1] = "case",
  [aux_sym_class_declaration_token1] = "class",
  [aux_sym_final_modifier_token1] = "final",
  [aux_sym_abstract_modifier_token1] = "abstract",
  [aux_sym_readonly_modifier_token1] = "readonly",
  [aux_sym_class_interface_clause_token1] = "implements",
  [anon_sym_EQ_GT] = "=>",
  [sym_var_modifier] = "var_modifier",
  [aux_sym_use_instead_of_clause_token1] = "insteadof",
  [aux_sym_visibility_modifier_token1] = "public",
  [aux_sym_visibility_modifier_token2] = "protected",
  [aux_sym_visibility_modifier_token3] = "private",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [anon_sym_LPAREN2] = "(",
  [anon_sym_RPAREN2] = ")",
  [aux_sym__arrow_function_header_token1] = "fn",
  [anon_sym_DOT_DOT_DOT] = "...",
  [anon_sym_QMARK] = "\?",
  [sym_bottom_type] = "bottom_type",
  [anon_sym_PIPE] = "|",
  [anon_sym_array] = "array",
  [aux_sym_primitive_type_token1] = "callable",
  [anon_sym_iterable] = "iterable",
  [anon_sym_bool] = "bool",
  [anon_sym_float] = "float",
  [anon_sym_void] = "void",
  [anon_sym_mixed] = "mixed",
  [anon_sym_false] = "false",
  [anon_sym_null] = "null",
  [anon_sym_true] = "true",
  [aux_sym_cast_type_token1] = "cast_type_token1",
  [aux_sym_cast_type_token2] = "cast_type_token2",
  [aux_sym_cast_type_token3] = "cast_type_token3",
  [aux_sym_cast_type_token4] = "cast_type_token4",
  [aux_sym_cast_type_token5] = "cast_type_token5",
  [aux_sym_cast_type_token6] = "cast_type_token6",
  [aux_sym_cast_type_token7] = "cast_type_token7",
  [aux_sym_cast_type_token8] = "cast_type_token8",
  [aux_sym_cast_type_token9] = "cast_type_token9",
  [aux_sym_cast_type_token10] = "cast_type_token10",
  [aux_sym_cast_type_token11] = "cast_type_token11",
  [aux_sym_cast_type_token12] = "cast_type_token12",
  [aux_sym_echo_statement_token1] = "echo",
  [aux_sym_exit_statement_token1] = "exit",
  [anon_sym_unset] = "unset",
  [aux_sym_declare_statement_token1] = "declare",
  [aux_sym_declare_statement_token2] = "enddeclare",
  [anon_sym_ticks] = "ticks",
  [anon_sym_encoding] = "encoding",
  [anon_sym_strict_types] = "strict_types",
  [sym_float] = "float",
  [aux_sym_try_statement_token1] = "try",
  [aux_sym_catch_clause_token1] = "catch",
  [aux_sym_finally_clause_token1] = "finally",
  [aux_sym_goto_statement_token1] = "goto",
  [aux_sym_continue_statement_token1] = "continue",
  [aux_sym_break_statement_token1] = "break",
  [sym_integer] = "integer",
  [aux_sym_return_statement_token1] = "return",
  [aux_sym_throw_expression_token1] = "throw",
  [aux_sym_while_statement_token1] = "while",
  [aux_sym_while_statement_token2] = "endwhile",
  [aux_sym_do_statement_token1] = "do",
  [aux_sym_for_statement_token1] = "for",
  [aux_sym_for_statement_token2] = "endfor",
  [aux_sym_foreach_statement_token1] = "foreach",
  [aux_sym_foreach_statement_token2] = "endforeach",
  [aux_sym_if_statement_token1] = "if",
  [aux_sym_if_statement_token2] = "endif",
  [aux_sym_else_if_clause_token1] = "elseif",
  [aux_sym_else_clause_token1] = "else",
  [aux_sym_match_expression_token1] = "match",
  [aux_sym_match_default_expression_token1] = "default",
  [aux_sym_switch_statement_token1] = "switch",
  [aux_sym_switch_block_token1] = "endswitch",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_TILDE] = "~",
  [anon_sym_BANG] = "!",
  [anon_sym_AT] = "@",
  [aux_sym_clone_expression_token1] = "clone",
  [anon_sym_COLON_COLON] = "::",
  [aux_sym_print_intrinsic_token1] = "print",
  [aux_sym__new_non_dereferencable_expression_token1] = "new",
  [anon_sym_DASH_DASH] = "--",
  [anon_sym_PLUS_PLUS] = "++",
  [anon_sym_STAR_STAR_EQ] = "**=",
  [anon_sym_STAR_EQ] = "*=",
  [anon_sym_SLASH_EQ] = "/=",
  [anon_sym_PERCENT_EQ] = "%=",
  [anon_sym_PLUS_EQ] = "+=",
  [anon_sym_DASH_EQ] = "-=",
  [anon_sym_DOT_EQ] = ".=",
  [anon_sym_LT_LT_EQ] = "<<=",
  [anon_sym_GT_GT_EQ] = ">>=",
  [anon_sym_AMP_EQ] = "&=",
  [anon_sym_CARET_EQ] = "^=",
  [anon_sym_PIPE_EQ] = "|=",
  [anon_sym_QMARK_QMARK_EQ] = "\?\?=",
  [anon_sym_DASH_GT] = "->",
  [anon_sym_QMARK_DASH_GT] = "\?->",
  [aux_sym__list_destructing_token1] = "list",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_self] = "self",
  [anon_sym_parent] = "parent",
  [aux_sym__argument_name_token1] = "_argument_name_token1",
  [aux_sym__argument_name_token2] = "_argument_name_token2",
  [anon_sym_POUND_LBRACK] = "#[",
  [sym_escape_sequence] = "escape_sequence",
  [anon_sym_BSLASHu] = "string_content",
  [aux_sym_encapsed_string_token1] = "encapsed_string_token1",
  [anon_sym_DQUOTE] = "\"",
  [aux_sym_string_token1] = "string_token1",
  [anon_sym_SQUOTE] = "'",
  [aux_sym_string_token2] = "escape_sequence",
  [aux_sym_string_content_token1] = "string_content_token1",
  [anon_sym_LT_LT_LT] = "<<<",
  [anon_sym_DQUOTE2] = "\"",
  [sym__new_line] = "_new_line",
  [anon_sym_SQUOTE2] = "'",
  [anon_sym_BQUOTE] = "`",
  [anon_sym_DOLLAR] = "$",
  [aux_sym_yield_expression_token1] = "yield",
  [aux_sym_yield_expression_token2] = "from",
  [aux_sym_binary_expression_token1] = "instanceof",
  [anon_sym_QMARK_QMARK] = "\?\?",
  [anon_sym_STAR_STAR] = "**",
  [aux_sym_binary_expression_token2] = "and",
  [aux_sym_binary_expression_token3] = "or",
  [aux_sym_binary_expression_token4] = "xor",
  [anon_sym_PIPE_PIPE] = "||",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_CARET] = "^",
  [anon_sym_EQ_EQ] = "==",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_LT_GT] = "<>",
  [anon_sym_EQ_EQ_EQ] = "===",
  [anon_sym_BANG_EQ_EQ] = "!==",
  [anon_sym_LT] = "<",
  [anon_sym_GT] = ">",
  [anon_sym_LT_EQ] = "<=",
  [anon_sym_GT_EQ] = ">=",
  [anon_sym_LT_EQ_GT] = "<=>",
  [anon_sym_LT_LT] = "<<",
  [anon_sym_GT_GT] = ">>",
  [anon_sym_DOT] = ".",
  [anon_sym_STAR] = "*",
  [anon_sym_SLASH] = "/",
  [anon_sym_PERCENT] = "%",
  [aux_sym_include_expression_token1] = "include",
  [aux_sym_include_once_expression_token1] = "include_once",
  [aux_sym_require_expression_token1] = "require",
  [aux_sym_require_once_expression_token1] = "require_once",
  [sym_comment] = "comment",
  [sym__automatic_semicolon] = "_automatic_semicolon",
  [sym_encapsed_string_chars] = "string_content",
  [sym_encapsed_string_chars_after_variable] = "string_content",
  [sym_execution_string_chars] = "string_content",
  [sym_execution_string_chars_after_variable] = "string_content",
  [sym_encapsed_string_chars_heredoc] = "string_content",
  [sym_encapsed_string_chars_after_variable_heredoc] = "string_content",
  [sym__eof] = "_eof",
  [sym_heredoc_start] = "heredoc_start",
  [sym_heredoc_end] = "heredoc_end",
  [sym_nowdoc_string] = "nowdoc_string",
  [sym_sentinel_error] = "sentinel_error",
  [sym_program] = "program",
  [sym_text_interpolation] = "text_interpolation",
  [sym_text] = "text",
  [sym_statement] = "statement",
  [sym_empty_statement] = "empty_statement",
  [sym_reference_modifier] = "reference_modifier",
  [sym_function_static_declaration] = "function_static_declaration",
  [sym_static_variable_declaration] = "static_variable_declaration",
  [sym_global_declaration] = "global_declaration",
  [sym_namespace_definition] = "namespace_definition",
  [sym_namespace_use_declaration] = "namespace_use_declaration",
  [sym_namespace_use_clause] = "namespace_use_clause",
  [sym_qualified_name] = "qualified_name",
  [sym__name] = "_name",
  [sym_namespace_name] = "namespace_name",
  [sym__namespace_use_group] = "_namespace_use_group",
  [sym_namespace_use_group] = "namespace_use_group",
  [sym_trait_declaration] = "trait_declaration",
  [sym_interface_declaration] = "interface_declaration",
  [sym_base_clause] = "base_clause",
  [sym_enum_declaration] = "enum_declaration",
  [sym_enum_declaration_list] = "enum_declaration_list",
  [sym__enum_member_declaration] = "_enum_member_declaration",
  [sym_enum_case] = "enum_case",
  [sym_class_declaration] = "class_declaration",
  [sym_declaration_list] = "declaration_list",
  [sym_final_modifier] = "final_modifier",
  [sym_abstract_modifier] = "abstract_modifier",
  [sym_readonly_modifier] = "readonly_modifier",
  [sym_class_interface_clause] = "class_interface_clause",
  [sym__member_declaration] = "_member_declaration",
  [sym_const_declaration] = "const_declaration",
  [sym__class_const_declaration] = "const_declaration",
  [sym__const_declaration] = "_const_declaration",
  [sym_property_declaration] = "property_declaration",
  [sym__modifier] = "_modifier",
  [sym_property_element] = "property_element",
  [sym_property_hook_list] = "property_hook_list",
  [sym_property_hook] = "property_hook",
  [sym__property_hook_body] = "_property_hook_body",
  [sym_method_declaration] = "method_declaration",
  [sym_static_modifier] = "static_modifier",
  [sym_use_declaration] = "use_declaration",
  [sym_use_list] = "use_list",
  [sym_use_instead_of_clause] = "use_instead_of_clause",
  [sym_use_as_clause] = "use_as_clause",
  [sym_visibility_modifier] = "visibility_modifier",
  [sym_function_definition] = "function_definition",
  [sym__function_definition_header] = "_function_definition_header",
  [sym_anonymous_function] = "anonymous_function",
  [sym_anonymous_function_use_clause] = "anonymous_function_use_clause",
  [sym__anonymous_function_header] = "_anonymous_function_header",
  [sym__arrow_function_header] = "_arrow_function_header",
  [sym_arrow_function] = "arrow_function",
  [sym_formal_parameters] = "formal_parameters",
  [sym_property_promotion_parameter] = "property_promotion_parameter",
  [sym_simple_parameter] = "simple_parameter",
  [sym_variadic_parameter] = "variadic_parameter",
  [sym_type] = "type",
  [sym__types] = "_types",
  [sym_named_type] = "named_type",
  [sym_optional_type] = "optional_type",
  [sym_union_type] = "union_type",
  [sym_intersection_type] = "intersection_type",
  [sym_disjunctive_normal_form_type] = "disjunctive_normal_form_type",
  [sym_primitive_type] = "primitive_type",
  [sym_cast_type] = "cast_type",
  [sym__return_type] = "_return_type",
  [sym_const_element] = "const_element",
  [sym_echo_statement] = "echo_statement",
  [sym_exit_statement] = "exit_statement",
  [sym_unset_statement] = "unset_statement",
  [sym_declare_statement] = "declare_statement",
  [sym_declare_directive] = "declare_directive",
  [sym_literal] = "literal",
  [sym_try_statement] = "try_statement",
  [sym_catch_clause] = "catch_clause",
  [sym_type_list] = "type_list",
  [sym_finally_clause] = "finally_clause",
  [sym_goto_statement] = "goto_statement",
  [sym_continue_statement] = "continue_statement",
  [sym_break_statement] = "break_statement",
  [sym_return_statement] = "return_statement",
  [sym_throw_expression] = "throw_expression",
  [sym_while_statement] = "while_statement",
  [sym_do_statement] = "do_statement",
  [sym_for_statement] = "for_statement",
  [sym__expressions] = "_expressions",
  [sym_sequence_expression] = "sequence_expression",
  [sym_foreach_statement] = "foreach_statement",
  [sym_foreach_pair] = "pair",
  [sym__foreach_value] = "_foreach_value",
  [sym_if_statement] = "if_statement",
  [sym_colon_block] = "colon_block",
  [sym_else_if_clause] = "else_if_clause",
  [sym_else_clause] = "else_clause",
  [sym_else_if_clause_2] = "else_if_clause",
  [sym_else_clause_2] = "else_clause",
  [sym_match_expression] = "match_expression",
  [sym_match_block] = "match_block",
  [sym_match_condition_list] = "match_condition_list",
  [sym_match_conditional_expression] = "match_conditional_expression",
  [sym_match_default_expression] = "match_default_expression",
  [sym_switch_statement] = "switch_statement",
  [sym_switch_block] = "switch_block",
  [sym_case_statement] = "case_statement",
  [sym_default_statement] = "default_statement",
  [sym_compound_statement] = "compound_statement",
  [sym_named_label_statement] = "named_label_statement",
  [sym_expression_statement] = "expression_statement",
  [sym_expression] = "expression",
  [sym__unary_expression] = "_unary_expression",
  [sym_unary_op_expression] = "unary_op_expression",
  [sym_error_suppression_expression] = "error_suppression_expression",
  [sym_clone_expression] = "clone_expression",
  [sym_primary_expression] = "primary_expression",
  [sym_parenthesized_expression] = "parenthesized_expression",
  [sym_class_constant_access_expression] = "class_constant_access_expression",
  [sym_print_intrinsic] = "print_intrinsic",
  [sym_object_creation_expression] = "object_creation_expression",
  [sym__new_non_dereferencable_expression] = "_new_non_dereferencable_expression",
  [sym__new_dereferencable_expression] = "_new_dereferencable_expression",
  [sym__class_name_reference] = "_class_name_reference",
  [sym_anonymous_class] = "anonymous_class",
  [sym_update_expression] = "update_expression",
  [sym_cast_expression] = "cast_expression",
  [sym_cast_variable] = "cast_expression",
  [sym_assignment_expression] = "assignment_expression",
  [sym_reference_assignment_expression] = "reference_assignment_expression",
  [sym_conditional_expression] = "conditional_expression",
  [sym_augmented_assignment_expression] = "augmented_assignment_expression",
  [sym__variable_member_access_expression] = "member_access_expression",
  [sym_member_access_expression] = "member_access_expression",
  [sym__variable_nullsafe_member_access_expression] = "nullsafe_member_access_expression",
  [sym_nullsafe_member_access_expression] = "nullsafe_member_access_expression",
  [sym__variable_scoped_property_access_expression] = "scoped_property_access_expression",
  [sym_scoped_property_access_expression] = "scoped_property_access_expression",
  [sym_list_literal] = "list_literal",
  [sym__list_destructing] = "_list_destructing",
  [sym__array_destructing] = "_array_destructing",
  [sym__array_destructing_element] = "_array_destructing_element",
  [sym_function_call_expression] = "function_call_expression",
  [sym__callable_expression] = "_callable_expression",
  [sym_scoped_call_expression] = "scoped_call_expression",
  [sym__scope_resolution_qualifier] = "_scope_resolution_qualifier",
  [sym_relative_scope] = "relative_scope",
  [sym_variadic_placeholder] = "variadic_placeholder",
  [sym_arguments] = "arguments",
  [sym_argument] = "argument",
  [sym__argument_name] = "_argument_name",
  [sym_member_call_expression] = "member_call_expression",
  [sym_nullsafe_member_call_expression] = "nullsafe_member_call_expression",
  [sym_variadic_unpacking] = "variadic_unpacking",
  [sym__member_name] = "_member_name",
  [sym__variable_subscript_expression] = "subscript_expression",
  [sym__dereferencable_subscript_expression] = "subscript_expression",
  [sym__dereferencable_expression] = "_dereferencable_expression",
  [sym__dereferencable_scalar] = "_dereferencable_scalar",
  [sym_array_creation_expression] = "array_creation_expression",
  [sym_attribute_group] = "attribute_group",
  [sym_attribute_list] = "attribute_list",
  [sym_attribute] = "attribute",
  [sym__complex_string_part] = "_complex_string_part",
  [sym__simple_string_member_access_expression] = "member_access_expression",
  [sym__simple_string_subscript_unary_expression] = "unary_op_expression",
  [sym__simple_string_array_access_argument] = "_simple_string_array_access_argument",
  [sym__simple_string_subscript_expression] = "subscript_expression",
  [sym__simple_string_part] = "_simple_string_part",
  [aux_sym__interpolated_string_body] = "_interpolated_string_body",
  [aux_sym__interpolated_string_body_heredoc] = "_interpolated_string_body_heredoc",
  [sym_encapsed_string] = "encapsed_string",
  [sym_string] = "string",
  [sym_string_content] = "string_content",
  [sym_heredoc_body] = "heredoc_body",
  [sym_heredoc] = "heredoc",
  [sym_nowdoc_body] = "nowdoc_body",
  [sym_nowdoc] = "nowdoc",
  [aux_sym__interpolated_execution_operator_body] = "_interpolated_execution_operator_body",
  [sym_shell_command_expression] = "shell_command_expression",
  [sym_boolean] = "boolean",
  [sym_null] = "null",
  [sym__string] = "_string",
  [sym_dynamic_variable_name] = "dynamic_variable_name",
  [sym__simple_variable] = "_simple_variable",
  [sym__new_variable] = "_new_variable",
  [sym__callable_variable] = "_callable_variable",
  [sym_variable_name] = "variable_name",
  [sym_by_ref] = "by_ref",
  [sym_yield_expression] = "yield_expression",
  [sym_array_element_initializer] = "array_element_initializer",
  [sym_binary_expression] = "binary_expression",
  [sym_include_expression] = "include_expression",
  [sym_include_once_expression] = "include_once_expression",
  [sym_require_expression] = "require_expression",
  [sym_require_once_expression] = "require_once_expression",
  [sym__reserved_identifier] = "name",
  [sym__identifier] = "_identifier",
  [sym__semicolon] = "_semicolon",
  [aux_sym_program_repeat1] = "program_repeat1",
  [aux_sym_text_repeat1] = "text_repeat1",
  [aux_sym_function_static_declaration_repeat1] = "function_static_declaration_repeat1",
  [aux_sym_global_declaration_repeat1] = "global_declaration_repeat1",
  [aux_sym_namespace_use_declaration_repeat1] = "namespace_use_declaration_repeat1",
  [aux_sym_namespace_name_repeat1] = "namespace_name_repeat1",
  [aux_sym_base_clause_repeat1] = "base_clause_repeat1",
  [aux_sym_enum_declaration_list_repeat1] = "enum_declaration_list_repeat1",
  [aux_sym_class_declaration_repeat1] = "class_declaration_repeat1",
  [aux_sym_declaration_list_repeat1] = "declaration_list_repeat1",
  [aux_sym__const_declaration_repeat1] = "_const_declaration_repeat1",
  [aux_sym_property_declaration_repeat1] = "property_declaration_repeat1",
  [aux_sym_property_hook_list_repeat1] = "property_hook_list_repeat1",
  [aux_sym_use_list_repeat1] = "use_list_repeat1",
  [aux_sym_anonymous_function_use_clause_repeat1] = "anonymous_function_use_clause_repeat1",
  [aux_sym_formal_parameters_repeat1] = "formal_parameters_repeat1",
  [aux_sym_union_type_repeat1] = "union_type_repeat1",
  [aux_sym_intersection_type_repeat1] = "intersection_type_repeat1",
  [aux_sym_disjunctive_normal_form_type_repeat1] = "disjunctive_normal_form_type_repeat1",
  [aux_sym_unset_statement_repeat1] = "unset_statement_repeat1",
  [aux_sym_try_statement_repeat1] = "try_statement_repeat1",
  [aux_sym_type_list_repeat1] = "type_list_repeat1",
  [aux_sym_if_statement_repeat1] = "if_statement_repeat1",
  [aux_sym_if_statement_repeat2] = "if_statement_repeat2",
  [aux_sym_match_block_repeat1] = "match_block_repeat1",
  [aux_sym_match_condition_list_repeat1] = "match_condition_list_repeat1",
  [aux_sym_switch_block_repeat1] = "switch_block_repeat1",
  [aux_sym__list_destructing_repeat1] = "_list_destructing_repeat1",
  [aux_sym__array_destructing_repeat1] = "_array_destructing_repeat1",
  [aux_sym_arguments_repeat1] = "arguments_repeat1",
  [aux_sym_array_creation_expression_repeat1] = "array_creation_expression_repeat1",
  [aux_sym_attribute_group_repeat1] = "attribute_group_repeat1",
  [aux_sym_attribute_list_repeat1] = "attribute_list_repeat1",
  [aux_sym_string_repeat1] = "string_repeat1",
  [aux_sym_string_content_repeat1] = "string_content_repeat1",
  [aux_sym_heredoc_body_repeat1] = "heredoc_body_repeat1",
  [aux_sym_nowdoc_body_repeat1] = "nowdoc_body_repeat1",
  [alias_sym_operation] = "operation",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym_name] = sym_name,
  [sym_php_tag] = sym_php_tag,
  [anon_sym_QMARK_GT] = anon_sym_QMARK_GT,
  [aux_sym_text_token1] = aux_sym_text_token1,
  [aux_sym_text_token2] = aux_sym_text_token2,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_AMP] = anon_sym_AMP,
  [aux_sym_function_static_declaration_token1] = aux_sym_function_static_declaration_token1,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_EQ] = anon_sym_EQ,
  [aux_sym_global_declaration_token1] = aux_sym_global_declaration_token1,
  [aux_sym_namespace_definition_token1] = aux_sym_namespace_definition_token1,
  [aux_sym_namespace_use_declaration_token1] = aux_sym_namespace_use_declaration_token1,
  [aux_sym_namespace_use_clause_token1] = aux_sym_namespace_use_clause_token1,
  [aux_sym__namespace_use_type_token1] = aux_sym__namespace_use_type_token1,
  [aux_sym__namespace_use_type_token2] = aux_sym__namespace_use_type_token2,
  [anon_sym_BSLASH] = anon_sym_BSLASH,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [aux_sym_trait_declaration_token1] = aux_sym_trait_declaration_token1,
  [aux_sym_interface_declaration_token1] = aux_sym_interface_declaration_token1,
  [aux_sym_base_clause_token1] = aux_sym_base_clause_token1,
  [aux_sym_enum_declaration_token1] = aux_sym_enum_declaration_token1,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_string] = anon_sym_string,
  [anon_sym_int] = anon_sym_int,
  [aux_sym_enum_case_token1] = aux_sym_enum_case_token1,
  [aux_sym_class_declaration_token1] = aux_sym_class_declaration_token1,
  [aux_sym_final_modifier_token1] = aux_sym_final_modifier_token1,
  [aux_sym_abstract_modifier_token1] = aux_sym_abstract_modifier_token1,
  [aux_sym_readonly_modifier_token1] = aux_sym_readonly_modifier_token1,
  [aux_sym_class_interface_clause_token1] = aux_sym_class_interface_clause_token1,
  [anon_sym_EQ_GT] = anon_sym_EQ_GT,
  [sym_var_modifier] = sym_var_modifier,
  [aux_sym_use_instead_of_clause_token1] = aux_sym_use_instead_of_clause_token1,
  [aux_sym_visibility_modifier_token1] = aux_sym_visibility_modifier_token1,
  [aux_sym_visibility_modifier_token2] = aux_sym_visibility_modifier_token2,
  [aux_sym_visibility_modifier_token3] = aux_sym_visibility_modifier_token3,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_LPAREN2] = anon_sym_LPAREN,
  [anon_sym_RPAREN2] = anon_sym_RPAREN,
  [aux_sym__arrow_function_header_token1] = aux_sym__arrow_function_header_token1,
  [anon_sym_DOT_DOT_DOT] = anon_sym_DOT_DOT_DOT,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [sym_bottom_type] = sym_bottom_type,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_array] = anon_sym_array,
  [aux_sym_primitive_type_token1] = aux_sym_primitive_type_token1,
  [anon_sym_iterable] = anon_sym_iterable,
  [anon_sym_bool] = anon_sym_bool,
  [anon_sym_float] = anon_sym_float,
  [anon_sym_void] = anon_sym_void,
  [anon_sym_mixed] = anon_sym_mixed,
  [anon_sym_false] = anon_sym_false,
  [anon_sym_null] = anon_sym_null,
  [anon_sym_true] = anon_sym_true,
  [aux_sym_cast_type_token1] = aux_sym_cast_type_token1,
  [aux_sym_cast_type_token2] = aux_sym_cast_type_token2,
  [aux_sym_cast_type_token3] = aux_sym_cast_type_token3,
  [aux_sym_cast_type_token4] = aux_sym_cast_type_token4,
  [aux_sym_cast_type_token5] = aux_sym_cast_type_token5,
  [aux_sym_cast_type_token6] = aux_sym_cast_type_token6,
  [aux_sym_cast_type_token7] = aux_sym_cast_type_token7,
  [aux_sym_cast_type_token8] = aux_sym_cast_type_token8,
  [aux_sym_cast_type_token9] = aux_sym_cast_type_token9,
  [aux_sym_cast_type_token10] = aux_sym_cast_type_token10,
  [aux_sym_cast_type_token11] = aux_sym_cast_type_token11,
  [aux_sym_cast_type_token12] = aux_sym_cast_type_token12,
  [aux_sym_echo_statement_token1] = aux_sym_echo_statement_token1,
  [aux_sym_exit_statement_token1] = aux_sym_exit_statement_token1,
  [anon_sym_unset] = anon_sym_unset,
  [aux_sym_declare_statement_token1] = aux_sym_declare_statement_token1,
  [aux_sym_declare_statement_token2] = aux_sym_declare_statement_token2,
  [anon_sym_ticks] = anon_sym_ticks,
  [anon_sym_encoding] = anon_sym_encoding,
  [anon_sym_strict_types] = anon_sym_strict_types,
  [sym_float] = sym_float,
  [aux_sym_try_statement_token1] = aux_sym_try_statement_token1,
  [aux_sym_catch_clause_token1] = aux_sym_catch_clause_token1,
  [aux_sym_finally_clause_token1] = aux_sym_finally_clause_token1,
  [aux_sym_goto_statement_token1] = aux_sym_goto_statement_token1,
  [aux_sym_continue_statement_token1] = aux_sym_continue_statement_token1,
  [aux_sym_break_statement_token1] = aux_sym_break_statement_token1,
  [sym_integer] = sym_integer,
  [aux_sym_return_statement_token1] = aux_sym_return_statement_token1,
  [aux_sym_throw_expression_token1] = aux_sym_throw_expression_token1,
  [aux_sym_while_statement_token1] = aux_sym_while_statement_token1,
  [aux_sym_while_statement_token2] = aux_sym_while_statement_token2,
  [aux_sym_do_statement_token1] = aux_sym_do_statement_token1,
  [aux_sym_for_statement_token1] = aux_sym_for_statement_token1,
  [aux_sym_for_statement_token2] = aux_sym_for_statement_token2,
  [aux_sym_foreach_statement_token1] = aux_sym_foreach_statement_token1,
  [aux_sym_foreach_statement_token2] = aux_sym_foreach_statement_token2,
  [aux_sym_if_statement_token1] = aux_sym_if_statement_token1,
  [aux_sym_if_statement_token2] = aux_sym_if_statement_token2,
  [aux_sym_else_if_clause_token1] = aux_sym_else_if_clause_token1,
  [aux_sym_else_clause_token1] = aux_sym_else_clause_token1,
  [aux_sym_match_expression_token1] = aux_sym_match_expression_token1,
  [aux_sym_match_default_expression_token1] = aux_sym_match_default_expression_token1,
  [aux_sym_switch_statement_token1] = aux_sym_switch_statement_token1,
  [aux_sym_switch_block_token1] = aux_sym_switch_block_token1,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym_AT] = anon_sym_AT,
  [aux_sym_clone_expression_token1] = aux_sym_clone_expression_token1,
  [anon_sym_COLON_COLON] = anon_sym_COLON_COLON,
  [aux_sym_print_intrinsic_token1] = aux_sym_print_intrinsic_token1,
  [aux_sym__new_non_dereferencable_expression_token1] = aux_sym__new_non_dereferencable_expression_token1,
  [anon_sym_DASH_DASH] = anon_sym_DASH_DASH,
  [anon_sym_PLUS_PLUS] = anon_sym_PLUS_PLUS,
  [anon_sym_STAR_STAR_EQ] = anon_sym_STAR_STAR_EQ,
  [anon_sym_STAR_EQ] = anon_sym_STAR_EQ,
  [anon_sym_SLASH_EQ] = anon_sym_SLASH_EQ,
  [anon_sym_PERCENT_EQ] = anon_sym_PERCENT_EQ,
  [anon_sym_PLUS_EQ] = anon_sym_PLUS_EQ,
  [anon_sym_DASH_EQ] = anon_sym_DASH_EQ,
  [anon_sym_DOT_EQ] = anon_sym_DOT_EQ,
  [anon_sym_LT_LT_EQ] = anon_sym_LT_LT_EQ,
  [anon_sym_GT_GT_EQ] = anon_sym_GT_GT_EQ,
  [anon_sym_AMP_EQ] = anon_sym_AMP_EQ,
  [anon_sym_CARET_EQ] = anon_sym_CARET_EQ,
  [anon_sym_PIPE_EQ] = anon_sym_PIPE_EQ,
  [anon_sym_QMARK_QMARK_EQ] = anon_sym_QMARK_QMARK_EQ,
  [anon_sym_DASH_GT] = anon_sym_DASH_GT,
  [anon_sym_QMARK_DASH_GT] = anon_sym_QMARK_DASH_GT,
  [aux_sym__list_destructing_token1] = aux_sym__list_destructing_token1,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_self] = anon_sym_self,
  [anon_sym_parent] = anon_sym_parent,
  [aux_sym__argument_name_token1] = aux_sym__argument_name_token1,
  [aux_sym__argument_name_token2] = aux_sym__argument_name_token2,
  [anon_sym_POUND_LBRACK] = anon_sym_POUND_LBRACK,
  [sym_escape_sequence] = sym_escape_sequence,
  [anon_sym_BSLASHu] = sym_string_content,
  [aux_sym_encapsed_string_token1] = aux_sym_encapsed_string_token1,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [aux_sym_string_token1] = aux_sym_string_token1,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [aux_sym_string_token2] = sym_escape_sequence,
  [aux_sym_string_content_token1] = aux_sym_string_content_token1,
  [anon_sym_LT_LT_LT] = anon_sym_LT_LT_LT,
  [anon_sym_DQUOTE2] = anon_sym_DQUOTE,
  [sym__new_line] = sym__new_line,
  [anon_sym_SQUOTE2] = anon_sym_SQUOTE,
  [anon_sym_BQUOTE] = anon_sym_BQUOTE,
  [anon_sym_DOLLAR] = anon_sym_DOLLAR,
  [aux_sym_yield_expression_token1] = aux_sym_yield_expression_token1,
  [aux_sym_yield_expression_token2] = aux_sym_yield_expression_token2,
  [aux_sym_binary_expression_token1] = aux_sym_binary_expression_token1,
  [anon_sym_QMARK_QMARK] = anon_sym_QMARK_QMARK,
  [anon_sym_STAR_STAR] = anon_sym_STAR_STAR,
  [aux_sym_binary_expression_token2] = aux_sym_binary_expression_token2,
  [aux_sym_binary_expression_token3] = aux_sym_binary_expression_token3,
  [aux_sym_binary_expression_token4] = aux_sym_binary_expression_token4,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym_EQ_EQ] = anon_sym_EQ_EQ,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_LT_GT] = anon_sym_LT_GT,
  [anon_sym_EQ_EQ_EQ] = anon_sym_EQ_EQ_EQ,
  [anon_sym_BANG_EQ_EQ] = anon_sym_BANG_EQ_EQ,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_LT_EQ] = anon_sym_LT_EQ,
  [anon_sym_GT_EQ] = anon_sym_GT_EQ,
  [anon_sym_LT_EQ_GT] = anon_sym_LT_EQ_GT,
  [anon_sym_LT_LT] = anon_sym_LT_LT,
  [anon_sym_GT_GT] = anon_sym_GT_GT,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [aux_sym_include_expression_token1] = aux_sym_include_expression_token1,
  [aux_sym_include_once_expression_token1] = aux_sym_include_once_expression_token1,
  [aux_sym_require_expression_token1] = aux_sym_require_expression_token1,
  [aux_sym_require_once_expression_token1] = aux_sym_require_once_expression_token1,
  [sym_comment] = sym_comment,
  [sym__automatic_semicolon] = sym__automatic_semicolon,
  [sym_encapsed_string_chars] = sym_string_content,
  [sym_encapsed_string_chars_after_variable] = sym_string_content,
  [sym_execution_string_chars] = sym_string_content,
  [sym_execution_string_chars_after_variable] = sym_string_content,
  [sym_encapsed_string_chars_heredoc] = sym_string_content,
  [sym_encapsed_string_chars_after_variable_heredoc] = sym_string_content,
  [sym__eof] = sym__eof,
  [sym_heredoc_start] = sym_heredoc_start,
  [sym_heredoc_end] = sym_heredoc_end,
  [sym_nowdoc_string] = sym_nowdoc_string,
  [sym_sentinel_error] = sym_sentinel_error,
  [sym_program] = sym_program,
  [sym_text_interpolation] = sym_text_interpolation,
  [sym_text] = sym_text,
  [sym_statement] = sym_statement,
  [sym_empty_statement] = sym_empty_statement,
  [sym_reference_modifier] = sym_reference_modifier,
  [sym_function_static_declaration] = sym_function_static_declaration,
  [sym_static_variable_declaration] = sym_static_variable_declaration,
  [sym_global_declaration] = sym_global_declaration,
  [sym_namespace_definition] = sym_namespace_definition,
  [sym_namespace_use_declaration] = sym_namespace_use_declaration,
  [sym_namespace_use_clause] = sym_namespace_use_clause,
  [sym_qualified_name] = sym_qualified_name,
  [sym__name] = sym__name,
  [sym_namespace_name] = sym_namespace_name,
  [sym__namespace_use_group] = sym__namespace_use_group,
  [sym_namespace_use_group] = sym_namespace_use_group,
  [sym_trait_declaration] = sym_trait_declaration,
  [sym_interface_declaration] = sym_interface_declaration,
  [sym_base_clause] = sym_base_clause,
  [sym_enum_declaration] = sym_enum_declaration,
  [sym_enum_declaration_list] = sym_enum_declaration_list,
  [sym__enum_member_declaration] = sym__enum_member_declaration,
  [sym_enum_case] = sym_enum_case,
  [sym_class_declaration] = sym_class_declaration,
  [sym_declaration_list] = sym_declaration_list,
  [sym_final_modifier] = sym_final_modifier,
  [sym_abstract_modifier] = sym_abstract_modifier,
  [sym_readonly_modifier] = sym_readonly_modifier,
  [sym_class_interface_clause] = sym_class_interface_clause,
  [sym__member_declaration] = sym__member_declaration,
  [sym_const_declaration] = sym_const_declaration,
  [sym__class_const_declaration] = sym_const_declaration,
  [sym__const_declaration] = sym__const_declaration,
  [sym_property_declaration] = sym_property_declaration,
  [sym__modifier] = sym__modifier,
  [sym_property_element] = sym_property_element,
  [sym_property_hook_list] = sym_property_hook_list,
  [sym_property_hook] = sym_property_hook,
  [sym__property_hook_body] = sym__property_hook_body,
  [sym_method_declaration] = sym_method_declaration,
  [sym_static_modifier] = sym_static_modifier,
  [sym_use_declaration] = sym_use_declaration,
  [sym_use_list] = sym_use_list,
  [sym_use_instead_of_clause] = sym_use_instead_of_clause,
  [sym_use_as_clause] = sym_use_as_clause,
  [sym_visibility_modifier] = sym_visibility_modifier,
  [sym_function_definition] = sym_function_definition,
  [sym__function_definition_header] = sym__function_definition_header,
  [sym_anonymous_function] = sym_anonymous_function,
  [sym_anonymous_function_use_clause] = sym_anonymous_function_use_clause,
  [sym__anonymous_function_header] = sym__anonymous_function_header,
  [sym__arrow_function_header] = sym__arrow_function_header,
  [sym_arrow_function] = sym_arrow_function,
  [sym_formal_parameters] = sym_formal_parameters,
  [sym_property_promotion_parameter] = sym_property_promotion_parameter,
  [sym_simple_parameter] = sym_simple_parameter,
  [sym_variadic_parameter] = sym_variadic_parameter,
  [sym_type] = sym_type,
  [sym__types] = sym__types,
  [sym_named_type] = sym_named_type,
  [sym_optional_type] = sym_optional_type,
  [sym_union_type] = sym_union_type,
  [sym_intersection_type] = sym_intersection_type,
  [sym_disjunctive_normal_form_type] = sym_disjunctive_normal_form_type,
  [sym_primitive_type] = sym_primitive_type,
  [sym_cast_type] = sym_cast_type,
  [sym__return_type] = sym__return_type,
  [sym_const_element] = sym_const_element,
  [sym_echo_statement] = sym_echo_statement,
  [sym_exit_statement] = sym_exit_statement,
  [sym_unset_statement] = sym_unset_statement,
  [sym_declare_statement] = sym_declare_statement,
  [sym_declare_directive] = sym_declare_directive,
  [sym_literal] = sym_literal,
  [sym_try_statement] = sym_try_statement,
  [sym_catch_clause] = sym_catch_clause,
  [sym_type_list] = sym_type_list,
  [sym_finally_clause] = sym_finally_clause,
  [sym_goto_statement] = sym_goto_statement,
  [sym_continue_statement] = sym_continue_statement,
  [sym_break_statement] = sym_break_statement,
  [sym_return_statement] = sym_return_statement,
  [sym_throw_expression] = sym_throw_expression,
  [sym_while_statement] = sym_while_statement,
  [sym_do_statement] = sym_do_statement,
  [sym_for_statement] = sym_for_statement,
  [sym__expressions] = sym__expressions,
  [sym_sequence_expression] = sym_sequence_expression,
  [sym_foreach_statement] = sym_foreach_statement,
  [sym_foreach_pair] = sym_foreach_pair,
  [sym__foreach_value] = sym__foreach_value,
  [sym_if_statement] = sym_if_statement,
  [sym_colon_block] = sym_colon_block,
  [sym_else_if_clause] = sym_else_if_clause,
  [sym_else_clause] = sym_else_clause,
  [sym_else_if_clause_2] = sym_else_if_clause,
  [sym_else_clause_2] = sym_else_clause,
  [sym_match_expression] = sym_match_expression,
  [sym_match_block] = sym_match_block,
  [sym_match_condition_list] = sym_match_condition_list,
  [sym_match_conditional_expression] = sym_match_conditional_expression,
  [sym_match_default_expression] = sym_match_default_expression,
  [sym_switch_statement] = sym_switch_statement,
  [sym_switch_block] = sym_switch_block,
  [sym_case_statement] = sym_case_statement,
  [sym_default_statement] = sym_default_statement,
  [sym_compound_statement] = sym_compound_statement,
  [sym_named_label_statement] = sym_named_label_statement,
  [sym_expression_statement] = sym_expression_statement,
  [sym_expression] = sym_expression,
  [sym__unary_expression] = sym__unary_expression,
  [sym_unary_op_expression] = sym_unary_op_expression,
  [sym_error_suppression_expression] = sym_error_suppression_expression,
  [sym_clone_expression] = sym_clone_expression,
  [sym_primary_expression] = sym_primary_expression,
  [sym_parenthesized_expression] = sym_parenthesized_expression,
  [sym_class_constant_access_expression] = sym_class_constant_access_expression,
  [sym_print_intrinsic] = sym_print_intrinsic,
  [sym_object_creation_expression] = sym_object_creation_expression,
  [sym__new_non_dereferencable_expression] = sym__new_non_dereferencable_expression,
  [sym__new_dereferencable_expression] = sym__new_dereferencable_expression,
  [sym__class_name_reference] = sym__class_name_reference,
  [sym_anonymous_class] = sym_anonymous_class,
  [sym_update_expression] = sym_update_expression,
  [sym_cast_expression] = sym_cast_expression,
  [sym_cast_variable] = sym_cast_expression,
  [sym_assignment_expression] = sym_assignment_expression,
  [sym_reference_assignment_expression] = sym_reference_assignment_expression,
  [sym_conditional_expression] = sym_conditional_expression,
  [sym_augmented_assignment_expression] = sym_augmented_assignment_expression,
  [sym__variable_member_access_expression] = sym_member_access_expression,
  [sym_member_access_expression] = sym_member_access_expression,
  [sym__variable_nullsafe_member_access_expression] = sym_nullsafe_member_access_expression,
  [sym_nullsafe_member_access_expression] = sym_nullsafe_member_access_expression,
  [sym__variable_scoped_property_access_expression] = sym_scoped_property_access_expression,
  [sym_scoped_property_access_expression] = sym_scoped_property_access_expression,
  [sym_list_literal] = sym_list_literal,
  [sym__list_destructing] = sym__list_destructing,
  [sym__array_destructing] = sym__array_destructing,
  [sym__array_destructing_element] = sym__array_destructing_element,
  [sym_function_call_expression] = sym_function_call_expression,
  [sym__callable_expression] = sym__callable_expression,
  [sym_scoped_call_expression] = sym_scoped_call_expression,
  [sym__scope_resolution_qualifier] = sym__scope_resolution_qualifier,
  [sym_relative_scope] = sym_relative_scope,
  [sym_variadic_placeholder] = sym_variadic_placeholder,
  [sym_arguments] = sym_arguments,
  [sym_argument] = sym_argument,
  [sym__argument_name] = sym__argument_name,
  [sym_member_call_expression] = sym_member_call_expression,
  [sym_nullsafe_member_call_expression] = sym_nullsafe_member_call_expression,
  [sym_variadic_unpacking] = sym_variadic_unpacking,
  [sym__member_name] = sym__member_name,
  [sym__variable_subscript_expression] = sym__variable_subscript_expression,
  [sym__dereferencable_subscript_expression] = sym__variable_subscript_expression,
  [sym__dereferencable_expression] = sym__dereferencable_expression,
  [sym__dereferencable_scalar] = sym__dereferencable_scalar,
  [sym_array_creation_expression] = sym_array_creation_expression,
  [sym_attribute_group] = sym_attribute_group,
  [sym_attribute_list] = sym_attribute_list,
  [sym_attribute] = sym_attribute,
  [sym__complex_string_part] = sym__complex_string_part,
  [sym__simple_string_member_access_expression] = sym_member_access_expression,
  [sym__simple_string_subscript_unary_expression] = sym_unary_op_expression,
  [sym__simple_string_array_access_argument] = sym__simple_string_array_access_argument,
  [sym__simple_string_subscript_expression] = sym__variable_subscript_expression,
  [sym__simple_string_part] = sym__simple_string_part,
  [aux_sym__interpolated_string_body] = aux_sym__interpolated_string_body,
  [aux_sym__interpolated_string_body_heredoc] = aux_sym__interpolated_string_body_heredoc,
  [sym_encapsed_string] = sym_encapsed_string,
  [sym_string] = sym_string,
  [sym_string_content] = sym_string_content,
  [sym_heredoc_body] = sym_heredoc_body,
  [sym_heredoc] = sym_heredoc,
  [sym_nowdoc_body] = sym_nowdoc_body,
  [sym_nowdoc] = sym_nowdoc,
  [aux_sym__interpolated_execution_operator_body] = aux_sym__interpolated_execution_operator_body,
  [sym_shell_command_expression] = sym_shell_command_expression,
  [sym_boolean] = sym_boolean,
  [sym_null] = sym_null,
  [sym__string] = sym__string,
  [sym_dynamic_variable_name] = sym_dynamic_variable_name,
  [sym__simple_variable] = sym__simple_variable,
  [sym__new_variable] = sym__new_variable,
  [sym__callable_variable] = sym__callable_variable,
  [sym_variable_name] = sym_variable_name,
  [sym_by_ref] = sym_by_ref,
  [sym_yield_expression] = sym_yield_expression,
  [sym_array_element_initializer] = sym_array_element_initializer,
  [sym_binary_expression] = sym_binary_expression,
  [sym_include_expression] = sym_include_expression,
  [sym_include_once_expression] = sym_include_once_expression,
  [sym_require_expression] = sym_require_expression,
  [sym_require_once_expression] = sym_require_once_expression,
  [sym__reserved_identifier] = sym_name,
  [sym__identifier] = sym__identifier,
  [sym__semicolon] = sym__semicolon,
  [aux_sym_program_repeat1] = aux_sym_program_repeat1,
  [aux_sym_text_repeat1] = aux_sym_text_repeat1,
  [aux_sym_function_static_declaration_repeat1] = aux_sym_function_static_declaration_repeat1,
  [aux_sym_global_declaration_repeat1] = aux_sym_global_declaration_repeat1,
  [aux_sym_namespace_use_declaration_repeat1] = aux_sym_namespace_use_declaration_repeat1,
  [aux_sym_namespace_name_repeat1] = aux_sym_namespace_name_repeat1,
  [aux_sym_base_clause_repeat1] = aux_sym_base_clause_repeat1,
  [aux_sym_enum_declaration_list_repeat1] = aux_sym_enum_declaration_list_repeat1,
  [aux_sym_class_declaration_repeat1] = aux_sym_class_declaration_repeat1,
  [aux_sym_declaration_list_repeat1] = aux_sym_declaration_list_repeat1,
  [aux_sym__const_declaration_repeat1] = aux_sym__const_declaration_repeat1,
  [aux_sym_property_declaration_repeat1] = aux_sym_property_declaration_repeat1,
  [aux_sym_property_hook_list_repeat1] = aux_sym_property_hook_list_repeat1,
  [aux_sym_use_list_repeat1] = aux_sym_use_list_repeat1,
  [aux_sym_anonymous_function_use_clause_repeat1] = aux_sym_anonymous_function_use_clause_repeat1,
  [aux_sym_formal_parameters_repeat1] = aux_sym_formal_parameters_repeat1,
  [aux_sym_union_type_repeat1] = aux_sym_union_type_repeat1,
  [aux_sym_intersection_type_repeat1] = aux_sym_intersection_type_repeat1,
  [aux_sym_disjunctive_normal_form_type_repeat1] = aux_sym_disjunctive_normal_form_type_repeat1,
  [aux_sym_unset_statement_repeat1] = aux_sym_unset_statement_repeat1,
  [aux_sym_try_statement_repeat1] = aux_sym_try_statement_repeat1,
  [aux_sym_type_list_repeat1] = aux_sym_type_list_repeat1,
  [aux_sym_if_statement_repeat1] = aux_sym_if_statement_repeat1,
  [aux_sym_if_statement_repeat2] = aux_sym_if_statement_repeat2,
  [aux_sym_match_block_repeat1] = aux_sym_match_block_repeat1,
  [aux_sym_match_condition_list_repeat1] = aux_sym_match_condition_list_repeat1,
  [aux_sym_switch_block_repeat1] = aux_sym_switch_block_repeat1,
  [aux_sym__list_destructing_repeat1] = aux_sym__list_destructing_repeat1,
  [aux_sym__array_destructing_repeat1] = aux_sym__array_destructing_repeat1,
  [aux_sym_arguments_repeat1] = aux_sym_arguments_repeat1,
  [aux_sym_array_creation_expression_repeat1] = aux_sym_array_creation_expression_repeat1,
  [aux_sym_attribute_group_repeat1] = aux_sym_attribute_group_repeat1,
  [aux_sym_attribute_list_repeat1] = aux_sym_attribute_list_repeat1,
  [aux_sym_string_repeat1] = aux_sym_string_repeat1,
  [aux_sym_string_content_repeat1] = aux_sym_string_content_repeat1,
  [aux_sym_heredoc_body_repeat1] = aux_sym_heredoc_body_repeat1,
  [aux_sym_nowdoc_body_repeat1] = aux_sym_nowdoc_body_repeat1,
  [alias_sym_operation] = alias_sym_operation,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym_name] = {
    .visible = true,
    .named = true,
  },
  [sym_php_tag] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_QMARK_GT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_text_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_text_token2] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_function_static_declaration_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_global_declaration_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_namespace_definition_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_namespace_use_declaration_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_namespace_use_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__namespace_use_type_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__namespace_use_type_token2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BSLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_trait_declaration_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_interface_declaration_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_base_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_enum_declaration_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_string] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_int] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_enum_case_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_class_declaration_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_final_modifier_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_abstract_modifier_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_readonly_modifier_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_class_interface_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_GT] = {
    .visible = true,
    .named = false,
  },
  [sym_var_modifier] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_use_instead_of_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_visibility_modifier_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_visibility_modifier_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_visibility_modifier_token3] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__arrow_function_header_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [sym_bottom_type] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_array] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_primitive_type_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_iterable] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_bool] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_float] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_void] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_mixed] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_false] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_null] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_true] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_cast_type_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token4] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token5] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token6] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token7] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token8] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token9] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token10] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token11] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token12] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_echo_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_exit_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unset] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_declare_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_declare_statement_token2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ticks] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_encoding] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_strict_types] = {
    .visible = true,
    .named = false,
  },
  [sym_float] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_try_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_catch_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_finally_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_goto_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_continue_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_break_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [sym_integer] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_return_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_throw_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_while_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_while_statement_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_do_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_for_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_for_statement_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_foreach_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_foreach_statement_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_if_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_if_statement_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_else_if_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_else_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_match_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_match_default_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_switch_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_switch_block_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_clone_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_COLON] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_print_intrinsic_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__new_non_dereferencable_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_QMARK_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__list_destructing_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_self] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_parent] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__argument_name_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__argument_name_token2] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_POUND_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [sym_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_BSLASHu] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_encapsed_string_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_string_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_string_token2] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_string_content_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_LT_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE2] = {
    .visible = true,
    .named = false,
  },
  [sym__new_line] = {
    .visible = false,
    .named = true,
  },
  [anon_sym_SQUOTE2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_yield_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_yield_expression_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_binary_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_STAR] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_binary_expression_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_binary_expression_token3] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_binary_expression_token4] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_include_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_include_once_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_require_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_require_once_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [sym__automatic_semicolon] = {
    .visible = false,
    .named = true,
  },
  [sym_encapsed_string_chars] = {
    .visible = true,
    .named = true,
  },
  [sym_encapsed_string_chars_after_variable] = {
    .visible = true,
    .named = true,
  },
  [sym_execution_string_chars] = {
    .visible = true,
    .named = true,
  },
  [sym_execution_string_chars_after_variable] = {
    .visible = true,
    .named = true,
  },
  [sym_encapsed_string_chars_heredoc] = {
    .visible = true,
    .named = true,
  },
  [sym_encapsed_string_chars_after_variable_heredoc] = {
    .visible = true,
    .named = true,
  },
  [sym__eof] = {
    .visible = false,
    .named = true,
  },
  [sym_heredoc_start] = {
    .visible = true,
    .named = true,
  },
  [sym_heredoc_end] = {
    .visible = true,
    .named = true,
  },
  [sym_nowdoc_string] = {
    .visible = true,
    .named = true,
  },
  [sym_sentinel_error] = {
    .visible = true,
    .named = true,
  },
  [sym_program] = {
    .visible = true,
    .named = true,
  },
  [sym_text_interpolation] = {
    .visible = true,
    .named = true,
  },
  [sym_text] = {
    .visible = true,
    .named = true,
  },
  [sym_statement] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_empty_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_reference_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_function_static_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_static_variable_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_global_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_namespace_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_namespace_use_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_namespace_use_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_qualified_name] = {
    .visible = true,
    .named = true,
  },
  [sym__name] = {
    .visible = false,
    .named = true,
  },
  [sym_namespace_name] = {
    .visible = true,
    .named = true,
  },
  [sym__namespace_use_group] = {
    .visible = false,
    .named = true,
  },
  [sym_namespace_use_group] = {
    .visible = true,
    .named = true,
  },
  [sym_trait_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_interface_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_base_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym__enum_member_declaration] = {
    .visible = false,
    .named = true,
  },
  [sym_enum_case] = {
    .visible = true,
    .named = true,
  },
  [sym_class_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym_final_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_abstract_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_readonly_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_class_interface_clause] = {
    .visible = true,
    .named = true,
  },
  [sym__member_declaration] = {
    .visible = false,
    .named = true,
  },
  [sym_const_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__class_const_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__const_declaration] = {
    .visible = false,
    .named = true,
  },
  [sym_property_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__modifier] = {
    .visible = false,
    .named = true,
  },
  [sym_property_element] = {
    .visible = true,
    .named = true,
  },
  [sym_property_hook_list] = {
    .visible = true,
    .named = true,
  },
  [sym_property_hook] = {
    .visible = true,
    .named = true,
  },
  [sym__property_hook_body] = {
    .visible = false,
    .named = true,
  },
  [sym_method_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_static_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_use_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_use_list] = {
    .visible = true,
    .named = true,
  },
  [sym_use_instead_of_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_use_as_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_visibility_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_function_definition] = {
    .visible = true,
    .named = true,
  },
  [sym__function_definition_header] = {
    .visible = false,
    .named = true,
  },
  [sym_anonymous_function] = {
    .visible = true,
    .named = true,
  },
  [sym_anonymous_function_use_clause] = {
    .visible = true,
    .named = true,
  },
  [sym__anonymous_function_header] = {
    .visible = false,
    .named = true,
  },
  [sym__arrow_function_header] = {
    .visible = false,
    .named = true,
  },
  [sym_arrow_function] = {
    .visible = true,
    .named = true,
  },
  [sym_formal_parameters] = {
    .visible = true,
    .named = true,
  },
  [sym_property_promotion_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_simple_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_variadic_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__types] = {
    .visible = false,
    .named = true,
  },
  [sym_named_type] = {
    .visible = true,
    .named = true,
  },
  [sym_optional_type] = {
    .visible = true,
    .named = true,
  },
  [sym_union_type] = {
    .visible = true,
    .named = true,
  },
  [sym_intersection_type] = {
    .visible = true,
    .named = true,
  },
  [sym_disjunctive_normal_form_type] = {
    .visible = true,
    .named = true,
  },
  [sym_primitive_type] = {
    .visible = true,
    .named = true,
  },
  [sym_cast_type] = {
    .visible = true,
    .named = true,
  },
  [sym__return_type] = {
    .visible = false,
    .named = true,
  },
  [sym_const_element] = {
    .visible = true,
    .named = true,
  },
  [sym_echo_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_exit_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_unset_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_declare_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_declare_directive] = {
    .visible = true,
    .named = true,
  },
  [sym_literal] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_try_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_catch_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_type_list] = {
    .visible = true,
    .named = true,
  },
  [sym_finally_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_goto_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_continue_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_break_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_return_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_throw_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_while_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_do_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_for_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__expressions] = {
    .visible = false,
    .named = true,
  },
  [sym_sequence_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_foreach_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_foreach_pair] = {
    .visible = true,
    .named = true,
  },
  [sym__foreach_value] = {
    .visible = false,
    .named = true,
  },
  [sym_if_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_colon_block] = {
    .visible = true,
    .named = true,
  },
  [sym_else_if_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_else_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_else_if_clause_2] = {
    .visible = true,
    .named = true,
  },
  [sym_else_clause_2] = {
    .visible = true,
    .named = true,
  },
  [sym_match_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_match_block] = {
    .visible = true,
    .named = true,
  },
  [sym_match_condition_list] = {
    .visible = true,
    .named = true,
  },
  [sym_match_conditional_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_match_default_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_block] = {
    .visible = true,
    .named = true,
  },
  [sym_case_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_default_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_compound_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_named_label_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_expression_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__unary_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_unary_op_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_error_suppression_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_clone_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_primary_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_class_constant_access_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_print_intrinsic] = {
    .visible = true,
    .named = true,
  },
  [sym_object_creation_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__new_non_dereferencable_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__new_dereferencable_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__class_name_reference] = {
    .visible = false,
    .named = true,
  },
  [sym_anonymous_class] = {
    .visible = true,
    .named = true,
  },
  [sym_update_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_cast_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_cast_variable] = {
    .visible = true,
    .named = true,
  },
  [sym_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_reference_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_conditional_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_augmented_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__variable_member_access_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_member_access_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__variable_nullsafe_member_access_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_nullsafe_member_access_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__variable_scoped_property_access_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_scoped_property_access_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_list_literal] = {
    .visible = true,
    .named = true,
  },
  [sym__list_destructing] = {
    .visible = false,
    .named = true,
  },
  [sym__array_destructing] = {
    .visible = false,
    .named = true,
  },
  [sym__array_destructing_element] = {
    .visible = false,
    .named = true,
  },
  [sym_function_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__callable_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_scoped_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__scope_resolution_qualifier] = {
    .visible = false,
    .named = true,
  },
  [sym_relative_scope] = {
    .visible = true,
    .named = true,
  },
  [sym_variadic_placeholder] = {
    .visible = true,
    .named = true,
  },
  [sym_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_argument] = {
    .visible = true,
    .named = true,
  },
  [sym__argument_name] = {
    .visible = false,
    .named = true,
  },
  [sym_member_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_nullsafe_member_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_variadic_unpacking] = {
    .visible = true,
    .named = true,
  },
  [sym__member_name] = {
    .visible = false,
    .named = true,
  },
  [sym__variable_subscript_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__dereferencable_subscript_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__dereferencable_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__dereferencable_scalar] = {
    .visible = false,
    .named = true,
  },
  [sym_array_creation_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_group] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_list] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym__complex_string_part] = {
    .visible = false,
    .named = true,
  },
  [sym__simple_string_member_access_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_string_subscript_unary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_string_array_access_argument] = {
    .visible = false,
    .named = true,
  },
  [sym__simple_string_subscript_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_string_part] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__interpolated_string_body] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__interpolated_string_body_heredoc] = {
    .visible = false,
    .named = false,
  },
  [sym_encapsed_string] = {
    .visible = true,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym_string_content] = {
    .visible = true,
    .named = true,
  },
  [sym_heredoc_body] = {
    .visible = true,
    .named = true,
  },
  [sym_heredoc] = {
    .visible = true,
    .named = true,
  },
  [sym_nowdoc_body] = {
    .visible = true,
    .named = true,
  },
  [sym_nowdoc] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__interpolated_execution_operator_body] = {
    .visible = false,
    .named = false,
  },
  [sym_shell_command_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_boolean] = {
    .visible = true,
    .named = true,
  },
  [sym_null] = {
    .visible = true,
    .named = true,
  },
  [sym__string] = {
    .visible = false,
    .named = true,
  },
  [sym_dynamic_variable_name] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_variable] = {
    .visible = false,
    .named = true,
  },
  [sym__new_variable] = {
    .visible = false,
    .named = true,
  },
  [sym__callable_variable] = {
    .visible = false,
    .named = true,
  },
  [sym_variable_name] = {
    .visible = true,
    .named = true,
  },
  [sym_by_ref] = {
    .visible = true,
    .named = true,
  },
  [sym_yield_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_array_element_initializer] = {
    .visible = true,
    .named = true,
  },
  [sym_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_include_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_include_once_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_require_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_require_once_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__reserved_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym__identifier] = {
    .visible = false,
    .named = true,
  },
  [sym__semicolon] = {
    .visible = false,
    .named = true,
  },
  [aux_sym_program_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_text_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_function_static_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_global_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_namespace_use_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_namespace_name_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_base_clause_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_enum_declaration_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_declaration_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__const_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_property_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_property_hook_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_use_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_anonymous_function_use_clause_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_formal_parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_union_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_intersection_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_disjunctive_normal_form_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_unset_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_try_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_if_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_if_statement_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_match_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_match_condition_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_switch_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__list_destructing_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__array_destructing_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_array_creation_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_group_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_heredoc_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_nowdoc_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_operation] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_alias = 1,
  field_alternative = 2,
  field_argument = 3,
  field_arguments = 4,
  field_attributes = 5,
  field_body = 6,
  field_condition = 7,
  field_conditional_expressions = 8,
  field_default_value = 9,
  field_end_tag = 10,
  field_final = 11,
  field_function = 12,
  field_identifier = 13,
  field_initialize = 14,
  field_left = 15,
  field_name = 16,
  field_object = 17,
  field_operator = 18,
  field_parameters = 19,
  field_prefix = 20,
  field_readonly = 21,
  field_reference_modifier = 22,
  field_return_expression = 23,
  field_return_type = 24,
  field_right = 25,
  field_scope = 26,
  field_static_modifier = 27,
  field_type = 28,
  field_update = 29,
  field_value = 30,
  field_visibility = 31,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_alias] = "alias",
  [field_alternative] = "alternative",
  [field_argument] = "argument",
  [field_arguments] = "arguments",
  [field_attributes] = "attributes",
  [field_body] = "body",
  [field_condition] = "condition",
  [field_conditional_expressions] = "conditional_expressions",
  [field_default_value] = "default_value",
  [field_end_tag] = "end_tag",
  [field_final] = "final",
  [field_function] = "function",
  [field_identifier] = "identifier",
  [field_initialize] = "initialize",
  [field_left] = "left",
  [field_name] = "name",
  [field_object] = "object",
  [field_operator] = "operator",
  [field_parameters] = "parameters",
  [field_prefix] = "prefix",
  [field_readonly] = "readonly",
  [field_reference_modifier] = "reference_modifier",
  [field_return_expression] = "return_expression",
  [field_return_type] = "return_type",
  [field_right] = "right",
  [field_scope] = "scope",
  [field_static_modifier] = "static_modifier",
  [field_type] = "type",
  [field_update] = "update",
  [field_value] = "value",
  [field_visibility] = "visibility",
};

static const TSFieldMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [1] = {.index = 0, .length = 1},
  [3] = {.index = 1, .length = 2},
  [4] = {.index = 3, .length = 2},
  [5] = {.index = 5, .length = 1},
  [6] = {.index = 6, .length = 1},
  [7] = {.index = 7, .length = 1},
  [8] = {.index = 8, .length = 1},
  [9] = {.index = 9, .length = 2},
  [11] = {.index = 11, .length = 2},
  [12] = {.index = 13, .length = 4},
  [13] = {.index = 17, .length = 6},
  [14] = {.index = 23, .length = 2},
  [15] = {.index = 25, .length = 2},
  [16] = {.index = 27, .length = 1},
  [17] = {.index = 28, .length = 1},
  [18] = {.index = 29, .length = 2},
  [19] = {.index = 31, .length = 2},
  [20] = {.index = 33, .length = 2},
  [21] = {.index = 35, .length = 2},
  [23] = {.index = 37, .length = 2},
  [24] = {.index = 39, .length = 2},
  [25] = {.index = 41, .length = 2},
  [26] = {.index = 43, .length = 2},
  [27] = {.index = 45, .length = 2},
  [28] = {.index = 47, .length = 6},
  [29] = {.index = 53, .length = 3},
  [30] = {.index = 56, .length = 2},
  [31] = {.index = 58, .length = 2},
  [32] = {.index = 60, .length = 2},
  [33] = {.index = 62, .length = 5},
  [34] = {.index = 67, .length = 2},
  [35] = {.index = 69, .length = 1},
  [36] = {.index = 70, .length = 1},
  [37] = {.index = 71, .length = 2},
  [38] = {.index = 73, .length = 2},
  [39] = {.index = 75, .length = 2},
  [40] = {.index = 77, .length = 2},
  [41] = {.index = 79, .length = 3},
  [42] = {.index = 82, .length = 2},
  [43] = {.index = 84, .length = 1},
  [44] = {.index = 85, .length = 2},
  [45] = {.index = 87, .length = 3},
  [46] = {.index = 90, .length = 1},
  [47] = {.index = 91, .length = 2},
  [48] = {.index = 93, .length = 2},
  [50] = {.index = 95, .length = 2},
  [51] = {.index = 97, .length = 1},
  [52] = {.index = 98, .length = 3},
  [53] = {.index = 101, .length = 3},
  [54] = {.index = 104, .length = 2},
  [57] = {.index = 106, .length = 2},
  [58] = {.index = 108, .length = 2},
  [59] = {.index = 110, .length = 3},
  [60] = {.index = 113, .length = 2},
  [61] = {.index = 5, .length = 1},
  [62] = {.index = 115, .length = 1},
  [63] = {.index = 116, .length = 1},
  [64] = {.index = 117, .length = 3},
  [65] = {.index = 120, .length = 3},
  [66] = {.index = 123, .length = 2},
  [67] = {.index = 125, .length = 2},
  [68] = {.index = 127, .length = 3},
  [69] = {.index = 130, .length = 3},
  [70] = {.index = 133, .length = 3},
  [71] = {.index = 136, .length = 3},
  [72] = {.index = 139, .length = 3},
  [73] = {.index = 142, .length = 3},
  [74] = {.index = 145, .length = 2},
  [75] = {.index = 147, .length = 2},
  [76] = {.index = 149, .length = 2},
  [77] = {.index = 151, .length = 2},
  [78] = {.index = 153, .length = 3},
  [79] = {.index = 156, .length = 3},
  [80] = {.index = 159, .length = 2},
  [81] = {.index = 161, .length = 3},
  [82] = {.index = 164, .length = 2},
  [83] = {.index = 166, .length = 3},
  [84] = {.index = 169, .length = 3},
  [85] = {.index = 172, .length = 3},
  [86] = {.index = 175, .length = 2},
  [87] = {.index = 177, .length = 3},
  [88] = {.index = 180, .length = 3},
  [89] = {.index = 183, .length = 1},
  [90] = {.index = 184, .length = 3},
  [91] = {.index = 187, .length = 2},
  [92] = {.index = 189, .length = 2},
  [93] = {.index = 189, .length = 2},
  [94] = {.index = 191, .length = 2},
  [95] = {.index = 193, .length = 4},
  [96] = {.index = 197, .length = 2},
  [97] = {.index = 199, .length = 1},
  [98] = {.index = 200, .length = 2},
  [99] = {.index = 202, .length = 3},
  [100] = {.index = 205, .length = 3},
  [101] = {.index = 208, .length = 2},
  [102] = {.index = 210, .length = 3},
  [103] = {.index = 213, .length = 2},
  [104] = {.index = 215, .length = 4},
  [105] = {.index = 219, .length = 3},
  [106] = {.index = 222, .length = 3},
  [108] = {.index = 225, .length = 4},
  [109] = {.index = 229, .length = 3},
  [110] = {.index = 232, .length = 3},
  [111] = {.index = 235, .length = 4},
  [112] = {.index = 239, .length = 4},
  [113] = {.index = 243, .length = 3},
  [114] = {.index = 246, .length = 1},
  [115] = {.index = 247, .length = 2},
  [116] = {.index = 249, .length = 3},
  [117] = {.index = 252, .length = 4},
  [118] = {.index = 256, .length = 3},
  [119] = {.index = 259, .length = 3},
  [120] = {.index = 262, .length = 3},
  [121] = {.index = 265, .length = 3},
  [122] = {.index = 268, .length = 4},
  [123] = {.index = 272, .length = 4},
  [124] = {.index = 276, .length = 3},
  [125] = {.index = 279, .length = 4},
  [126] = {.index = 283, .length = 3},
  [127] = {.index = 286, .length = 2},
  [128] = {.index = 288, .length = 4},
  [129] = {.index = 292, .length = 4},
  [130] = {.index = 296, .length = 3},
  [131] = {.index = 299, .length = 2},
  [132] = {.index = 301, .length = 1},
  [133] = {.index = 302, .length = 1},
  [134] = {.index = 303, .length = 2},
  [135] = {.index = 305, .length = 1},
  [136] = {.index = 306, .length = 1},
  [137] = {.index = 307, .length = 2},
  [140] = {.index = 309, .length = 3},
  [141] = {.index = 312, .length = 3},
  [142] = {.index = 315, .length = 3},
  [143] = {.index = 318, .length = 4},
  [144] = {.index = 322, .length = 4},
  [145] = {.index = 326, .length = 3},
  [146] = {.index = 326, .length = 3},
  [147] = {.index = 329, .length = 5},
  [148] = {.index = 334, .length = 4},
  [149] = {.index = 338, .length = 3},
  [150] = {.index = 341, .length = 2},
  [151] = {.index = 343, .length = 4},
  [152] = {.index = 347, .length = 4},
  [153] = {.index = 351, .length = 1},
  [154] = {.index = 352, .length = 1},
  [155] = {.index = 353, .length = 4},
  [156] = {.index = 357, .length = 4},
  [157] = {.index = 361, .length = 5},
  [158] = {.index = 366, .length = 4},
  [159] = {.index = 370, .length = 4},
  [160] = {.index = 374, .length = 4},
  [161] = {.index = 378, .length = 1},
  [162] = {.index = 379, .length = 5},
  [163] = {.index = 384, .length = 4},
  [164] = {.index = 388, .length = 2},
  [165] = {.index = 390, .length = 2},
  [166] = {.index = 392, .length = 1},
  [167] = {.index = 393, .length = 2},
  [168] = {.index = 395, .length = 1},
  [169] = {.index = 396, .length = 2},
  [170] = {.index = 398, .length = 1},
  [171] = {.index = 399, .length = 1},
  [172] = {.index = 400, .length = 2},
  [173] = {.index = 402, .length = 3},
  [174] = {.index = 405, .length = 3},
  [175] = {.index = 408, .length = 5},
  [176] = {.index = 413, .length = 3},
  [177] = {.index = 416, .length = 5},
  [178] = {.index = 421, .length = 2},
  [179] = {.index = 423, .length = 2},
  [180] = {.index = 425, .length = 2},
  [181] = {.index = 427, .length = 2},
  [182] = {.index = 429, .length = 5},
  [183] = {.index = 434, .length = 5},
  [184] = {.index = 439, .length = 5},
  [185] = {.index = 444, .length = 2},
  [186] = {.index = 446, .length = 2},
  [187] = {.index = 448, .length = 3},
  [188] = {.index = 451, .length = 3},
  [189] = {.index = 454, .length = 2},
  [190] = {.index = 456, .length = 3},
  [191] = {.index = 459, .length = 2},
  [192] = {.index = 461, .length = 3},
  [193] = {.index = 464, .length = 2},
  [194] = {.index = 466, .length = 2},
  [195] = {.index = 468, .length = 3},
  [196] = {.index = 471, .length = 3},
  [197] = {.index = 474, .length = 3},
  [198] = {.index = 477, .length = 3},
  [199] = {.index = 480, .length = 3},
  [200] = {.index = 483, .length = 3},
  [201] = {.index = 486, .length = 6},
  [202] = {.index = 492, .length = 3},
  [203] = {.index = 495, .length = 4},
  [204] = {.index = 499, .length = 3},
  [205] = {.index = 502, .length = 4},
  [206] = {.index = 506, .length = 4},
  [207] = {.index = 510, .length = 4},
  [208] = {.index = 514, .length = 4},
  [209] = {.index = 518, .length = 2},
  [210] = {.index = 520, .length = 2},
  [211] = {.index = 522, .length = 2},
  [212] = {.index = 524, .length = 5},
  [213] = {.index = 529, .length = 3},
  [214] = {.index = 532, .length = 3},
  [215] = {.index = 535, .length = 3},
  [216] = {.index = 538, .length = 4},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_type, 0, .inherited = true},
  [1] =
    {field_name, 0, .inherited = true},
    {field_object, 0, .inherited = true},
  [3] =
    {field_name, 0, .inherited = true},
    {field_scope, 0, .inherited = true},
  [5] =
    {field_name, 0},
  [6] =
    {field_body, 1},
  [7] =
    {field_parameters, 1},
  [8] =
    {field_prefix, 0},
  [9] =
    {field_argument, 1},
    {field_operator, 0},
  [11] =
    {field_arguments, 1},
    {field_function, 0},
  [13] =
    {field_body, 1},
    {field_name, 0, .inherited = true},
    {field_parameters, 0, .inherited = true},
    {field_return_type, 0, .inherited = true},
  [17] =
    {field_attributes, 0, .inherited = true},
    {field_body, 1},
    {field_parameters, 0, .inherited = true},
    {field_reference_modifier, 0, .inherited = true},
    {field_return_type, 0, .inherited = true},
    {field_static_modifier, 0, .inherited = true},
  [23] =
    {field_argument, 0},
    {field_operator, 1},
  [25] =
    {field_body, 2},
    {field_name, 1},
  [27] =
    {field_name, 1},
  [28] =
    {field_type, 0},
  [29] =
    {field_body, 1, .inherited = true},
    {field_type, 1, .inherited = true},
  [31] =
    {field_parameters, 2},
    {field_reference_modifier, 1},
  [33] =
    {field_parameters, 1},
    {field_return_type, 2, .inherited = true},
  [35] =
    {field_name, 1},
    {field_parameters, 2},
  [37] =
    {field_body, 2},
    {field_condition, 1},
  [39] =
    {field_end_tag, 2},
    {field_identifier, 1},
  [41] =
    {field_name, 2},
    {field_scope, 0},
  [43] =
    {field_prefix, 0},
    {field_prefix, 1},
  [45] =
    {field_parameters, 2},
    {field_static_modifier, 0},
  [47] =
    {field_attributes, 0, .inherited = true},
    {field_body, 2},
    {field_parameters, 0, .inherited = true},
    {field_reference_modifier, 0, .inherited = true},
    {field_return_type, 0, .inherited = true},
    {field_static_modifier, 0, .inherited = true},
  [53] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 2},
  [56] =
    {field_left, 0},
    {field_right, 2},
  [58] =
    {field_name, 2, .inherited = true},
    {field_object, 0},
  [60] =
    {field_attributes, 0},
    {field_parameters, 2},
  [62] =
    {field_attributes, 0},
    {field_body, 2},
    {field_name, 1, .inherited = true},
    {field_parameters, 1, .inherited = true},
    {field_return_type, 1, .inherited = true},
  [67] =
    {field_name, 0},
    {field_value, 2},
  [69] =
    {field_alias, 2},
  [70] =
    {field_body, 2},
  [71] =
    {field_name, 1},
    {field_reference_modifier, 0},
  [73] =
    {field_name, 1},
    {field_visibility, 0},
  [75] =
    {field_name, 1},
    {field_type, 0},
  [77] =
    {field_attributes, 0},
    {field_name, 1},
  [79] =
    {field_parameters, 2},
    {field_reference_modifier, 1},
    {field_return_type, 3, .inherited = true},
  [82] =
    {field_name, 2},
    {field_parameters, 3},
  [84] =
    {field_return_type, 1},
  [85] =
    {field_parameters, 1},
    {field_return_type, 3, .inherited = true},
  [87] =
    {field_name, 1},
    {field_parameters, 2},
    {field_return_type, 3, .inherited = true},
  [90] =
    {field_type, 1},
  [91] =
    {field_attributes, 0, .inherited = true},
    {field_type, 0, .inherited = true},
  [93] =
    {field_body, 3},
    {field_name, 1},
  [95] =
    {field_type, 1},
    {field_value, 3},
  [97] =
    {field_alternative, 0},
  [98] =
    {field_alternative, 3},
    {field_body, 2},
    {field_condition, 1},
  [101] =
    {field_alternative, 3, .inherited = true},
    {field_body, 2},
    {field_condition, 1},
  [104] =
    {field_attributes, 0},
    {field_body, 2},
  [106] =
    {field_name, 2},
    {field_object, 0},
  [108] =
    {field_end_tag, 3},
    {field_identifier, 1},
  [110] =
    {field_end_tag, 3},
    {field_identifier, 1},
    {field_value, 2},
  [113] =
    {field_end_tag, 3},
    {field_identifier, 2},
  [115] =
    {field_reference_modifier, 0},
  [116] =
    {field_name, 0, .inherited = true},
  [117] =
    {field_parameters, 3},
    {field_reference_modifier, 2},
    {field_static_modifier, 0},
  [120] =
    {field_parameters, 2},
    {field_return_type, 3, .inherited = true},
    {field_static_modifier, 0},
  [123] =
    {field_alternative, 3},
    {field_condition, 0},
  [125] =
    {field_left, 0},
    {field_right, 3},
  [127] =
    {field_arguments, 3},
    {field_name, 2, .inherited = true},
    {field_scope, 0},
  [130] =
    {field_arguments, 3},
    {field_name, 2, .inherited = true},
    {field_object, 0},
  [133] =
    {field_attributes, 0},
    {field_parameters, 3},
    {field_reference_modifier, 2},
  [136] =
    {field_attributes, 0},
    {field_parameters, 2},
    {field_return_type, 3, .inherited = true},
  [139] =
    {field_attributes, 0},
    {field_body, 3},
    {field_name, 2},
  [142] =
    {field_attributes, 0},
    {field_parameters, 3},
    {field_static_modifier, 1},
  [145] =
    {field_body, 3},
    {field_name, 2},
  [147] =
    {field_alias, 3},
    {field_type, 0},
  [149] =
    {field_body, 3},
    {field_type, 0},
  [151] =
    {field_name, 2},
    {field_reference_modifier, 0},
  [153] =
    {field_name, 2},
    {field_readonly, 1},
    {field_visibility, 0},
  [156] =
    {field_name, 2},
    {field_type, 1},
    {field_visibility, 0},
  [159] =
    {field_name, 2},
    {field_type, 0},
  [161] =
    {field_name, 2},
    {field_reference_modifier, 1},
    {field_type, 0},
  [164] =
    {field_attributes, 0},
    {field_name, 2},
  [166] =
    {field_attributes, 0},
    {field_name, 2},
    {field_reference_modifier, 1},
  [169] =
    {field_attributes, 0},
    {field_name, 2},
    {field_visibility, 1},
  [172] =
    {field_attributes, 0},
    {field_name, 2},
    {field_type, 1},
  [175] =
    {field_default_value, 2},
    {field_name, 0},
  [177] =
    {field_parameters, 2},
    {field_reference_modifier, 1},
    {field_return_type, 4, .inherited = true},
  [180] =
    {field_name, 2},
    {field_parameters, 3},
    {field_return_type, 4, .inherited = true},
  [183] =
    {field_type, 1, .inherited = true},
  [184] =
    {field_name, 0, .inherited = true},
    {field_parameters, 0, .inherited = true},
    {field_return_type, 0, .inherited = true},
  [187] =
    {field_attributes, 0},
    {field_type, 1, .inherited = true},
  [189] =
    {field_body, 4},
    {field_name, 1},
  [191] =
    {field_body, 1},
    {field_condition, 3},
  [193] =
    {field_alternative, 3, .inherited = true},
    {field_alternative, 4},
    {field_body, 2},
    {field_condition, 1},
  [197] =
    {field_alternative, 0, .inherited = true},
    {field_alternative, 1, .inherited = true},
  [199] =
    {field_body, 3},
  [200] =
    {field_attributes, 0},
    {field_body, 3},
  [202] =
    {field_end_tag, 4},
    {field_identifier, 1},
    {field_value, 3},
  [205] =
    {field_end_tag, 4},
    {field_identifier, 1},
    {field_value, 2},
  [208] =
    {field_end_tag, 4},
    {field_identifier, 2},
  [210] =
    {field_end_tag, 4},
    {field_identifier, 2},
    {field_value, 3},
  [213] =
    {field_name, 0, .inherited = true},
    {field_reference_modifier, 1},
  [215] =
    {field_parameters, 3},
    {field_reference_modifier, 2},
    {field_return_type, 4, .inherited = true},
    {field_static_modifier, 0},
  [219] =
    {field_parameters, 2},
    {field_return_type, 4, .inherited = true},
    {field_static_modifier, 0},
  [222] =
    {field_alternative, 4},
    {field_body, 2},
    {field_condition, 0},
  [225] =
    {field_attributes, 0},
    {field_parameters, 3},
    {field_reference_modifier, 2},
    {field_return_type, 4, .inherited = true},
  [229] =
    {field_attributes, 0},
    {field_parameters, 2},
    {field_return_type, 4, .inherited = true},
  [232] =
    {field_attributes, 0},
    {field_body, 4},
    {field_name, 2},
  [235] =
    {field_attributes, 0},
    {field_parameters, 4},
    {field_reference_modifier, 3},
    {field_static_modifier, 1},
  [239] =
    {field_attributes, 0},
    {field_parameters, 3},
    {field_return_type, 4, .inherited = true},
    {field_static_modifier, 1},
  [243] =
    {field_attributes, 0},
    {field_body, 4},
    {field_name, 3},
  [246] =
    {field_type, 2},
  [247] =
    {field_body, 4},
    {field_name, 2},
  [249] =
    {field_default_value, 3},
    {field_name, 1},
    {field_reference_modifier, 0},
  [252] =
    {field_name, 3},
    {field_readonly, 1},
    {field_type, 2},
    {field_visibility, 0},
  [256] =
    {field_default_value, 3},
    {field_name, 1},
    {field_visibility, 0},
  [259] =
    {field_name, 3},
    {field_reference_modifier, 1},
    {field_type, 0},
  [262] =
    {field_default_value, 3},
    {field_name, 1},
    {field_type, 0},
  [265] =
    {field_attributes, 0},
    {field_name, 3},
    {field_reference_modifier, 1},
  [268] =
    {field_attributes, 0},
    {field_name, 3},
    {field_readonly, 2},
    {field_visibility, 1},
  [272] =
    {field_attributes, 0},
    {field_name, 3},
    {field_type, 2},
    {field_visibility, 1},
  [276] =
    {field_attributes, 0},
    {field_name, 3},
    {field_type, 1},
  [279] =
    {field_attributes, 0},
    {field_name, 3},
    {field_reference_modifier, 2},
    {field_type, 1},
  [283] =
    {field_attributes, 0},
    {field_default_value, 3},
    {field_name, 1},
  [286] =
    {field_attributes, 0},
    {field_type, 2, .inherited = true},
  [288] =
    {field_attributes, 0},
    {field_name, 1, .inherited = true},
    {field_parameters, 1, .inherited = true},
    {field_return_type, 1, .inherited = true},
  [292] =
    {field_body, 2},
    {field_name, 1, .inherited = true},
    {field_parameters, 1, .inherited = true},
    {field_return_type, 1, .inherited = true},
  [296] =
    {field_name, 1, .inherited = true},
    {field_parameters, 1, .inherited = true},
    {field_return_type, 1, .inherited = true},
  [299] =
    {field_body, 5},
    {field_name, 1},
  [301] =
    {field_body, 5},
  [302] =
    {field_return_expression, 2},
  [303] =
    {field_conditional_expressions, 0},
    {field_return_expression, 2},
  [305] =
    {field_value, 1},
  [306] =
    {field_body, 4},
  [307] =
    {field_attributes, 0},
    {field_body, 4},
  [309] =
    {field_end_tag, 5},
    {field_identifier, 1},
    {field_value, 3},
  [312] =
    {field_end_tag, 5},
    {field_identifier, 2},
    {field_value, 4},
  [315] =
    {field_end_tag, 5},
    {field_identifier, 2},
    {field_value, 3},
  [318] =
    {field_parameters, 3},
    {field_reference_modifier, 2},
    {field_return_type, 5, .inherited = true},
    {field_static_modifier, 0},
  [322] =
    {field_attributes, 0},
    {field_parameters, 3},
    {field_reference_modifier, 2},
    {field_return_type, 5, .inherited = true},
  [326] =
    {field_attributes, 0},
    {field_body, 5},
    {field_name, 2},
  [329] =
    {field_attributes, 0},
    {field_parameters, 4},
    {field_reference_modifier, 3},
    {field_return_type, 5, .inherited = true},
    {field_static_modifier, 1},
  [334] =
    {field_attributes, 0},
    {field_parameters, 3},
    {field_return_type, 5, .inherited = true},
    {field_static_modifier, 1},
  [338] =
    {field_attributes, 0},
    {field_body, 5},
    {field_name, 3},
  [341] =
    {field_body, 5},
    {field_name, 2},
  [343] =
    {field_default_value, 4},
    {field_name, 2},
    {field_readonly, 1},
    {field_visibility, 0},
  [347] =
    {field_default_value, 4},
    {field_name, 2},
    {field_type, 1},
    {field_visibility, 0},
  [351] =
    {field_body, 1, .inherited = true},
  [352] =
    {field_body, 0},
  [353] =
    {field_default_value, 4},
    {field_name, 2},
    {field_reference_modifier, 1},
    {field_type, 0},
  [357] =
    {field_attributes, 0},
    {field_default_value, 4},
    {field_name, 2},
    {field_reference_modifier, 1},
  [361] =
    {field_attributes, 0},
    {field_name, 4},
    {field_readonly, 2},
    {field_type, 3},
    {field_visibility, 1},
  [366] =
    {field_attributes, 0},
    {field_default_value, 4},
    {field_name, 2},
    {field_visibility, 1},
  [370] =
    {field_attributes, 0},
    {field_name, 4},
    {field_reference_modifier, 2},
    {field_type, 1},
  [374] =
    {field_attributes, 0},
    {field_default_value, 4},
    {field_name, 2},
    {field_type, 1},
  [378] =
    {field_attributes, 0},
  [379] =
    {field_attributes, 0},
    {field_body, 3},
    {field_name, 2, .inherited = true},
    {field_parameters, 2, .inherited = true},
    {field_return_type, 2, .inherited = true},
  [384] =
    {field_attributes, 0},
    {field_name, 2, .inherited = true},
    {field_parameters, 2, .inherited = true},
    {field_return_type, 2, .inherited = true},
  [388] =
    {field_body, 4},
    {field_type, 2},
  [390] =
    {field_body, 6},
    {field_update, 4},
  [392] =
    {field_update, 4},
  [393] =
    {field_body, 6},
    {field_condition, 3},
  [395] =
    {field_condition, 3},
  [396] =
    {field_body, 6},
    {field_initialize, 2},
  [398] =
    {field_initialize, 2},
  [399] =
    {field_body, 6},
  [400] =
    {field_attributes, 0},
    {field_body, 5},
  [402] =
    {field_end_tag, 6},
    {field_identifier, 2},
    {field_value, 4},
  [405] =
    {field_attributes, 0},
    {field_body, 6},
    {field_name, 2},
  [408] =
    {field_attributes, 0},
    {field_parameters, 4},
    {field_reference_modifier, 3},
    {field_return_type, 6, .inherited = true},
    {field_static_modifier, 1},
  [413] =
    {field_attributes, 0},
    {field_body, 6},
    {field_name, 3},
  [416] =
    {field_default_value, 5},
    {field_name, 3},
    {field_readonly, 1},
    {field_type, 2},
    {field_visibility, 0},
  [421] =
    {field_body, 2, .inherited = true},
    {field_parameters, 1},
  [423] =
    {field_body, 2, .inherited = true},
    {field_reference_modifier, 0},
  [425] =
    {field_body, 2, .inherited = true},
    {field_final, 0},
  [427] =
    {field_attributes, 0},
    {field_body, 2, .inherited = true},
  [429] =
    {field_attributes, 0},
    {field_default_value, 5},
    {field_name, 3},
    {field_readonly, 2},
    {field_visibility, 1},
  [434] =
    {field_attributes, 0},
    {field_default_value, 5},
    {field_name, 3},
    {field_type, 2},
    {field_visibility, 1},
  [439] =
    {field_attributes, 0},
    {field_default_value, 5},
    {field_name, 3},
    {field_reference_modifier, 2},
    {field_type, 1},
  [444] =
    {field_attributes, 0},
    {field_type, 2},
  [446] =
    {field_name, 1},
    {field_value, 3},
  [448] =
    {field_body, 5},
    {field_name, 3},
    {field_type, 2},
  [451] =
    {field_body, 7},
    {field_condition, 3},
    {field_update, 5},
  [454] =
    {field_condition, 3},
    {field_update, 5},
  [456] =
    {field_body, 7},
    {field_initialize, 2},
    {field_update, 5},
  [459] =
    {field_initialize, 2},
    {field_update, 5},
  [461] =
    {field_body, 7},
    {field_condition, 4},
    {field_initialize, 2},
  [464] =
    {field_condition, 4},
    {field_initialize, 2},
  [466] =
    {field_attributes, 0},
    {field_body, 6},
  [468] =
    {field_body, 3, .inherited = true},
    {field_parameters, 2},
    {field_reference_modifier, 0},
  [471] =
    {field_body, 3, .inherited = true},
    {field_final, 0},
    {field_parameters, 2},
  [474] =
    {field_body, 3, .inherited = true},
    {field_final, 0},
    {field_reference_modifier, 1},
  [477] =
    {field_attributes, 0},
    {field_body, 3, .inherited = true},
    {field_parameters, 2},
  [480] =
    {field_attributes, 0},
    {field_body, 3, .inherited = true},
    {field_reference_modifier, 1},
  [483] =
    {field_attributes, 0},
    {field_body, 3, .inherited = true},
    {field_final, 1},
  [486] =
    {field_attributes, 0},
    {field_default_value, 6},
    {field_name, 4},
    {field_readonly, 2},
    {field_type, 3},
    {field_visibility, 1},
  [492] =
    {field_attributes, 0},
    {field_name, 2},
    {field_value, 4},
  [495] =
    {field_body, 8},
    {field_condition, 4},
    {field_initialize, 2},
    {field_update, 6},
  [499] =
    {field_condition, 4},
    {field_initialize, 2},
    {field_update, 6},
  [502] =
    {field_body, 4, .inherited = true},
    {field_final, 0},
    {field_parameters, 3},
    {field_reference_modifier, 1},
  [506] =
    {field_attributes, 0},
    {field_body, 4, .inherited = true},
    {field_parameters, 3},
    {field_reference_modifier, 1},
  [510] =
    {field_attributes, 0},
    {field_body, 4, .inherited = true},
    {field_final, 1},
    {field_parameters, 3},
  [514] =
    {field_attributes, 0},
    {field_body, 4, .inherited = true},
    {field_final, 1},
    {field_reference_modifier, 2},
  [518] =
    {field_body, 7},
    {field_update, 4},
  [520] =
    {field_body, 7},
    {field_condition, 3},
  [522] =
    {field_body, 7},
    {field_initialize, 2},
  [524] =
    {field_attributes, 0},
    {field_body, 5, .inherited = true},
    {field_final, 1},
    {field_parameters, 4},
    {field_reference_modifier, 2},
  [529] =
    {field_body, 8},
    {field_condition, 3},
    {field_update, 5},
  [532] =
    {field_body, 8},
    {field_initialize, 2},
    {field_update, 5},
  [535] =
    {field_body, 8},
    {field_condition, 4},
    {field_initialize, 2},
  [538] =
    {field_body, 9},
    {field_condition, 4},
    {field_initialize, 2},
    {field_update, 6},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [2] = {
    [0] = sym_object_creation_expression,
  },
  [10] = {
    [0] = sym_list_literal,
  },
  [22] = {
    [0] = anon_sym_array,
  },
  [49] = {
    [2] = alias_sym_operation,
  },
  [55] = {
    [1] = sym_list_literal,
  },
  [56] = {
    [2] = sym_list_literal,
  },
  [61] = {
    [0] = sym_name,
  },
  [92] = {
    [3] = sym_primitive_type,
  },
  [107] = {
    [3] = sym_name,
  },
  [131] = {
    [3] = sym_primitive_type,
  },
  [138] = {
    [3] = sym_list_literal,
  },
  [139] = {
    [4] = sym_list_literal,
  },
  [145] = {
    [4] = sym_primitive_type,
  },
  [174] = {
    [4] = sym_primitive_type,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym_expression, 2,
    sym_expression,
    sym_name,
  sym__new_dereferencable_expression, 2,
    sym__new_dereferencable_expression,
    sym_object_creation_expression,
  sym__list_destructing, 2,
    sym__list_destructing,
    sym_list_literal,
  sym__array_destructing, 2,
    sym__array_destructing,
    sym_list_literal,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 4,
  [5] = 5,
  [6] = 6,
  [7] = 7,
  [8] = 8,
  [9] = 9,
  [10] = 9,
  [11] = 2,
  [12] = 8,
  [13] = 7,
  [14] = 9,
  [15] = 9,
  [16] = 16,
  [17] = 17,
  [18] = 16,
  [19] = 19,
  [20] = 20,
  [21] = 21,
  [22] = 21,
  [23] = 23,
  [24] = 24,
  [25] = 25,
  [26] = 26,
  [27] = 16,
  [28] = 17,
  [29] = 17,
  [30] = 19,
  [31] = 16,
  [32] = 26,
  [33] = 20,
  [34] = 21,
  [35] = 23,
  [36] = 23,
  [37] = 24,
  [38] = 25,
  [39] = 26,
  [40] = 17,
  [41] = 24,
  [42] = 19,
  [43] = 20,
  [44] = 21,
  [45] = 25,
  [46] = 23,
  [47] = 24,
  [48] = 25,
  [49] = 26,
  [50] = 20,
  [51] = 19,
  [52] = 52,
  [53] = 53,
  [54] = 54,
  [55] = 55,
  [56] = 56,
  [57] = 57,
  [58] = 58,
  [59] = 53,
  [60] = 55,
  [61] = 61,
  [62] = 53,
  [63] = 54,
  [64] = 56,
  [65] = 57,
  [66] = 66,
  [67] = 67,
  [68] = 68,
  [69] = 69,
  [70] = 70,
  [71] = 71,
  [72] = 72,
  [73] = 73,
  [74] = 74,
  [75] = 75,
  [76] = 76,
  [77] = 77,
  [78] = 78,
  [79] = 79,
  [80] = 80,
  [81] = 53,
  [82] = 55,
  [83] = 61,
  [84] = 55,
  [85] = 61,
  [86] = 55,
  [87] = 61,
  [88] = 58,
  [89] = 89,
  [90] = 67,
  [91] = 68,
  [92] = 69,
  [93] = 70,
  [94] = 94,
  [95] = 95,
  [96] = 71,
  [97] = 52,
  [98] = 58,
  [99] = 72,
  [100] = 73,
  [101] = 74,
  [102] = 75,
  [103] = 76,
  [104] = 77,
  [105] = 78,
  [106] = 79,
  [107] = 80,
  [108] = 58,
  [109] = 61,
  [110] = 110,
  [111] = 110,
  [112] = 112,
  [113] = 113,
  [114] = 113,
  [115] = 110,
  [116] = 110,
  [117] = 112,
  [118] = 118,
  [119] = 118,
  [120] = 118,
  [121] = 118,
  [122] = 122,
  [123] = 123,
  [124] = 123,
  [125] = 122,
  [126] = 123,
  [127] = 123,
  [128] = 122,
  [129] = 122,
  [130] = 122,
  [131] = 123,
  [132] = 123,
  [133] = 133,
  [134] = 134,
  [135] = 134,
  [136] = 134,
  [137] = 134,
  [138] = 134,
  [139] = 133,
  [140] = 134,
  [141] = 133,
  [142] = 133,
  [143] = 133,
  [144] = 144,
  [145] = 144,
  [146] = 146,
  [147] = 146,
  [148] = 144,
  [149] = 146,
  [150] = 144,
  [151] = 146,
  [152] = 146,
  [153] = 146,
  [154] = 144,
  [155] = 144,
  [156] = 156,
  [157] = 157,
  [158] = 158,
  [159] = 156,
  [160] = 157,
  [161] = 161,
  [162] = 162,
  [163] = 163,
  [164] = 164,
  [165] = 165,
  [166] = 161,
  [167] = 162,
  [168] = 163,
  [169] = 165,
  [170] = 164,
  [171] = 171,
  [172] = 172,
  [173] = 173,
  [174] = 171,
  [175] = 172,
  [176] = 176,
  [177] = 177,
  [178] = 176,
  [179] = 179,
  [180] = 177,
  [181] = 181,
  [182] = 182,
  [183] = 183,
  [184] = 184,
  [185] = 183,
  [186] = 183,
  [187] = 183,
  [188] = 188,
  [189] = 189,
  [190] = 190,
  [191] = 191,
  [192] = 192,
  [193] = 193,
  [194] = 194,
  [195] = 195,
  [196] = 193,
  [197] = 197,
  [198] = 188,
  [199] = 192,
  [200] = 200,
  [201] = 194,
  [202] = 197,
  [203] = 203,
  [204] = 191,
  [205] = 197,
  [206] = 203,
  [207] = 190,
  [208] = 188,
  [209] = 191,
  [210] = 192,
  [211] = 203,
  [212] = 190,
  [213] = 189,
  [214] = 197,
  [215] = 203,
  [216] = 190,
  [217] = 188,
  [218] = 189,
  [219] = 195,
  [220] = 191,
  [221] = 192,
  [222] = 189,
  [223] = 223,
  [224] = 224,
  [225] = 224,
  [226] = 224,
  [227] = 224,
  [228] = 228,
  [229] = 224,
  [230] = 223,
  [231] = 231,
  [232] = 232,
  [233] = 224,
  [234] = 234,
  [235] = 235,
  [236] = 234,
  [237] = 237,
  [238] = 234,
  [239] = 239,
  [240] = 240,
  [241] = 235,
  [242] = 237,
  [243] = 243,
  [244] = 234,
  [245] = 245,
  [246] = 235,
  [247] = 247,
  [248] = 240,
  [249] = 239,
  [250] = 250,
  [251] = 235,
  [252] = 234,
  [253] = 239,
  [254] = 234,
  [255] = 240,
  [256] = 237,
  [257] = 257,
  [258] = 250,
  [259] = 239,
  [260] = 260,
  [261] = 261,
  [262] = 262,
  [263] = 263,
  [264] = 264,
  [265] = 265,
  [266] = 266,
  [267] = 267,
  [268] = 268,
  [269] = 269,
  [270] = 264,
  [271] = 271,
  [272] = 272,
  [273] = 273,
  [274] = 274,
  [275] = 275,
  [276] = 276,
  [277] = 271,
  [278] = 278,
  [279] = 279,
  [280] = 280,
  [281] = 281,
  [282] = 282,
  [283] = 283,
  [284] = 284,
  [285] = 285,
  [286] = 286,
  [287] = 287,
  [288] = 288,
  [289] = 289,
  [290] = 290,
  [291] = 291,
  [292] = 292,
  [293] = 293,
  [294] = 294,
  [295] = 295,
  [296] = 262,
  [297] = 263,
  [298] = 298,
  [299] = 299,
  [300] = 300,
  [301] = 272,
  [302] = 269,
  [303] = 303,
  [304] = 298,
  [305] = 303,
  [306] = 299,
  [307] = 307,
  [308] = 308,
  [309] = 309,
  [310] = 273,
  [311] = 260,
  [312] = 274,
  [313] = 275,
  [314] = 276,
  [315] = 315,
  [316] = 278,
  [317] = 317,
  [318] = 318,
  [319] = 300,
  [320] = 320,
  [321] = 321,
  [322] = 322,
  [323] = 323,
  [324] = 279,
  [325] = 303,
  [326] = 278,
  [327] = 317,
  [328] = 318,
  [329] = 307,
  [330] = 308,
  [331] = 331,
  [332] = 332,
  [333] = 280,
  [334] = 281,
  [335] = 282,
  [336] = 269,
  [337] = 264,
  [338] = 271,
  [339] = 272,
  [340] = 273,
  [341] = 274,
  [342] = 275,
  [343] = 276,
  [344] = 279,
  [345] = 280,
  [346] = 281,
  [347] = 282,
  [348] = 283,
  [349] = 284,
  [350] = 285,
  [351] = 286,
  [352] = 287,
  [353] = 260,
  [354] = 289,
  [355] = 290,
  [356] = 291,
  [357] = 292,
  [358] = 293,
  [359] = 294,
  [360] = 295,
  [361] = 262,
  [362] = 263,
  [363] = 298,
  [364] = 299,
  [365] = 365,
  [366] = 300,
  [367] = 367,
  [368] = 368,
  [369] = 369,
  [370] = 370,
  [371] = 371,
  [372] = 372,
  [373] = 373,
  [374] = 283,
  [375] = 284,
  [376] = 285,
  [377] = 286,
  [378] = 287,
  [379] = 303,
  [380] = 278,
  [381] = 317,
  [382] = 318,
  [383] = 288,
  [384] = 308,
  [385] = 289,
  [386] = 290,
  [387] = 269,
  [388] = 264,
  [389] = 271,
  [390] = 272,
  [391] = 273,
  [392] = 274,
  [393] = 275,
  [394] = 276,
  [395] = 279,
  [396] = 280,
  [397] = 281,
  [398] = 282,
  [399] = 283,
  [400] = 284,
  [401] = 285,
  [402] = 286,
  [403] = 287,
  [404] = 288,
  [405] = 289,
  [406] = 290,
  [407] = 291,
  [408] = 292,
  [409] = 293,
  [410] = 294,
  [411] = 295,
  [412] = 262,
  [413] = 263,
  [414] = 298,
  [415] = 299,
  [416] = 300,
  [417] = 291,
  [418] = 292,
  [419] = 303,
  [420] = 318,
  [421] = 308,
  [422] = 293,
  [423] = 318,
  [424] = 308,
  [425] = 318,
  [426] = 308,
  [427] = 318,
  [428] = 318,
  [429] = 318,
  [430] = 294,
  [431] = 295,
  [432] = 432,
  [433] = 433,
  [434] = 434,
  [435] = 435,
  [436] = 436,
  [437] = 260,
  [438] = 288,
  [439] = 439,
  [440] = 439,
  [441] = 439,
  [442] = 439,
  [443] = 439,
  [444] = 444,
  [445] = 445,
  [446] = 446,
  [447] = 447,
  [448] = 447,
  [449] = 446,
  [450] = 450,
  [451] = 451,
  [452] = 452,
  [453] = 453,
  [454] = 454,
  [455] = 455,
  [456] = 456,
  [457] = 457,
  [458] = 458,
  [459] = 459,
  [460] = 460,
  [461] = 461,
  [462] = 462,
  [463] = 463,
  [464] = 464,
  [465] = 465,
  [466] = 466,
  [467] = 467,
  [468] = 468,
  [469] = 469,
  [470] = 470,
  [471] = 471,
  [472] = 472,
  [473] = 473,
  [474] = 474,
  [475] = 475,
  [476] = 476,
  [477] = 477,
  [478] = 478,
  [479] = 479,
  [480] = 480,
  [481] = 481,
  [482] = 482,
  [483] = 483,
  [484] = 484,
  [485] = 485,
  [486] = 486,
  [487] = 487,
  [488] = 488,
  [489] = 489,
  [490] = 490,
  [491] = 491,
  [492] = 492,
  [493] = 493,
  [494] = 494,
  [495] = 495,
  [496] = 496,
  [497] = 497,
  [498] = 498,
  [499] = 499,
  [500] = 500,
  [501] = 501,
  [502] = 502,
  [503] = 503,
  [504] = 504,
  [505] = 505,
  [506] = 506,
  [507] = 507,
  [508] = 508,
  [509] = 509,
  [510] = 510,
  [511] = 511,
  [512] = 512,
  [513] = 513,
  [514] = 514,
  [515] = 515,
  [516] = 516,
  [517] = 517,
  [518] = 518,
  [519] = 519,
  [520] = 456,
  [521] = 521,
  [522] = 522,
  [523] = 523,
  [524] = 524,
  [525] = 525,
  [526] = 526,
  [527] = 527,
  [528] = 528,
  [529] = 529,
  [530] = 530,
  [531] = 531,
  [532] = 532,
  [533] = 533,
  [534] = 534,
  [535] = 452,
  [536] = 536,
  [537] = 537,
  [538] = 538,
  [539] = 539,
  [540] = 540,
  [541] = 541,
  [542] = 542,
  [543] = 543,
  [544] = 544,
  [545] = 545,
  [546] = 546,
  [547] = 547,
  [548] = 548,
  [549] = 549,
  [550] = 550,
  [551] = 551,
  [552] = 552,
  [553] = 553,
  [554] = 554,
  [555] = 555,
  [556] = 556,
  [557] = 557,
  [558] = 558,
  [559] = 559,
  [560] = 560,
  [561] = 561,
  [562] = 562,
  [563] = 563,
  [564] = 564,
  [565] = 565,
  [566] = 566,
  [567] = 567,
  [568] = 568,
  [569] = 569,
  [570] = 570,
  [571] = 571,
  [572] = 572,
  [573] = 573,
  [574] = 574,
  [575] = 575,
  [576] = 576,
  [577] = 577,
  [578] = 578,
  [579] = 579,
  [580] = 580,
  [581] = 581,
  [582] = 582,
  [583] = 583,
  [584] = 584,
  [585] = 584,
  [586] = 584,
  [587] = 584,
  [588] = 588,
  [589] = 589,
  [590] = 590,
  [591] = 591,
  [592] = 592,
  [593] = 593,
  [594] = 594,
  [595] = 595,
  [596] = 596,
  [597] = 597,
  [598] = 598,
  [599] = 599,
  [600] = 600,
  [601] = 601,
  [602] = 602,
  [603] = 603,
  [604] = 604,
  [605] = 605,
  [606] = 606,
  [607] = 607,
  [608] = 608,
  [609] = 609,
  [610] = 610,
  [611] = 611,
  [612] = 612,
  [613] = 613,
  [614] = 614,
  [615] = 615,
  [616] = 616,
  [617] = 617,
  [618] = 618,
  [619] = 619,
  [620] = 620,
  [621] = 621,
  [622] = 622,
  [623] = 623,
  [624] = 593,
  [625] = 594,
  [626] = 626,
  [627] = 627,
  [628] = 609,
  [629] = 629,
  [630] = 597,
  [631] = 606,
  [632] = 608,
  [633] = 591,
  [634] = 600,
  [635] = 620,
  [636] = 626,
  [637] = 621,
  [638] = 615,
  [639] = 639,
  [640] = 640,
  [641] = 590,
  [642] = 642,
  [643] = 601,
  [644] = 617,
  [645] = 645,
  [646] = 596,
  [647] = 599,
  [648] = 626,
  [649] = 613,
  [650] = 619,
  [651] = 651,
  [652] = 595,
  [653] = 589,
  [654] = 592,
  [655] = 607,
  [656] = 602,
  [657] = 598,
  [658] = 651,
  [659] = 623,
  [660] = 660,
  [661] = 604,
  [662] = 651,
  [663] = 614,
  [664] = 616,
  [665] = 618,
  [666] = 605,
  [667] = 613,
  [668] = 622,
  [669] = 629,
  [670] = 670,
  [671] = 603,
  [672] = 610,
  [673] = 611,
  [674] = 612,
  [675] = 626,
  [676] = 629,
  [677] = 622,
  [678] = 678,
  [679] = 623,
  [680] = 680,
  [681] = 678,
  [682] = 682,
  [683] = 683,
  [684] = 651,
  [685] = 629,
  [686] = 686,
  [687] = 687,
  [688] = 688,
  [689] = 689,
  [690] = 690,
  [691] = 691,
  [692] = 690,
  [693] = 693,
  [694] = 689,
  [695] = 686,
  [696] = 693,
  [697] = 697,
  [698] = 698,
  [699] = 699,
  [700] = 700,
  [701] = 697,
  [702] = 702,
  [703] = 700,
  [704] = 697,
  [705] = 697,
  [706] = 697,
  [707] = 707,
  [708] = 700,
  [709] = 699,
  [710] = 710,
  [711] = 711,
  [712] = 707,
  [713] = 713,
  [714] = 714,
  [715] = 566,
  [716] = 475,
  [717] = 598,
  [718] = 593,
  [719] = 719,
  [720] = 720,
  [721] = 594,
  [722] = 722,
  [723] = 599,
  [724] = 617,
  [725] = 725,
  [726] = 714,
  [727] = 727,
  [728] = 728,
  [729] = 596,
  [730] = 602,
  [731] = 598,
  [732] = 606,
  [733] = 591,
  [734] = 595,
  [735] = 607,
  [736] = 609,
  [737] = 592,
  [738] = 615,
  [739] = 620,
  [740] = 621,
  [741] = 619,
  [742] = 589,
  [743] = 608,
  [744] = 590,
  [745] = 599,
  [746] = 597,
  [747] = 600,
  [748] = 601,
  [749] = 749,
  [750] = 750,
  [751] = 614,
  [752] = 616,
  [753] = 603,
  [754] = 604,
  [755] = 618,
  [756] = 756,
  [757] = 605,
  [758] = 610,
  [759] = 593,
  [760] = 594,
  [761] = 611,
  [762] = 612,
  [763] = 763,
  [764] = 764,
  [765] = 765,
  [766] = 766,
  [767] = 767,
  [768] = 768,
  [769] = 769,
  [770] = 770,
  [771] = 771,
  [772] = 772,
  [773] = 773,
  [774] = 774,
  [775] = 775,
  [776] = 776,
  [777] = 777,
  [778] = 778,
  [779] = 779,
  [780] = 780,
  [781] = 781,
  [782] = 782,
  [783] = 783,
  [784] = 784,
  [785] = 785,
  [786] = 786,
  [787] = 787,
  [788] = 788,
  [789] = 722,
  [790] = 452,
  [791] = 791,
  [792] = 792,
  [793] = 793,
  [794] = 794,
  [795] = 795,
  [796] = 796,
  [797] = 797,
  [798] = 798,
  [799] = 456,
  [800] = 588,
  [801] = 801,
  [802] = 802,
  [803] = 803,
  [804] = 804,
  [805] = 805,
  [806] = 806,
  [807] = 807,
  [808] = 808,
  [809] = 614,
  [810] = 616,
  [811] = 618,
  [812] = 596,
  [813] = 605,
  [814] = 602,
  [815] = 815,
  [816] = 606,
  [817] = 591,
  [818] = 603,
  [819] = 595,
  [820] = 820,
  [821] = 607,
  [822] = 822,
  [823] = 609,
  [824] = 610,
  [825] = 611,
  [826] = 612,
  [827] = 592,
  [828] = 615,
  [829] = 620,
  [830] = 621,
  [831] = 619,
  [832] = 604,
  [833] = 589,
  [834] = 608,
  [835] = 590,
  [836] = 836,
  [837] = 837,
  [838] = 838,
  [839] = 839,
  [840] = 597,
  [841] = 600,
  [842] = 601,
  [843] = 617,
  [844] = 766,
  [845] = 845,
  [846] = 846,
  [847] = 847,
  [848] = 848,
  [849] = 849,
  [850] = 850,
  [851] = 727,
  [852] = 728,
  [853] = 853,
  [854] = 854,
  [855] = 711,
  [856] = 856,
  [857] = 857,
  [858] = 711,
  [859] = 593,
  [860] = 594,
  [861] = 850,
  [862] = 862,
  [863] = 599,
  [864] = 862,
  [865] = 595,
  [866] = 590,
  [867] = 606,
  [868] = 597,
  [869] = 591,
  [870] = 608,
  [871] = 796,
  [872] = 725,
  [873] = 607,
  [874] = 600,
  [875] = 609,
  [876] = 722,
  [877] = 617,
  [878] = 598,
  [879] = 592,
  [880] = 615,
  [881] = 620,
  [882] = 882,
  [883] = 621,
  [884] = 599,
  [885] = 885,
  [886] = 886,
  [887] = 749,
  [888] = 619,
  [889] = 602,
  [890] = 596,
  [891] = 882,
  [892] = 892,
  [893] = 589,
  [894] = 601,
  [895] = 796,
  [896] = 713,
  [897] = 593,
  [898] = 594,
  [899] = 713,
  [900] = 750,
  [901] = 756,
  [902] = 902,
  [903] = 614,
  [904] = 616,
  [905] = 618,
  [906] = 605,
  [907] = 598,
  [908] = 908,
  [909] = 603,
  [910] = 727,
  [911] = 728,
  [912] = 610,
  [913] = 611,
  [914] = 612,
  [915] = 604,
  [916] = 916,
  [917] = 917,
  [918] = 807,
  [919] = 621,
  [920] = 619,
  [921] = 604,
  [922] = 596,
  [923] = 475,
  [924] = 589,
  [925] = 608,
  [926] = 590,
  [927] = 722,
  [928] = 784,
  [929] = 785,
  [930] = 786,
  [931] = 848,
  [932] = 588,
  [933] = 837,
  [934] = 934,
  [935] = 602,
  [936] = 802,
  [937] = 772,
  [938] = 773,
  [939] = 792,
  [940] = 788,
  [941] = 796,
  [942] = 787,
  [943] = 815,
  [944] = 600,
  [945] = 601,
  [946] = 606,
  [947] = 591,
  [948] = 791,
  [949] = 598,
  [950] = 603,
  [951] = 595,
  [952] = 605,
  [953] = 599,
  [954] = 765,
  [955] = 775,
  [956] = 607,
  [957] = 597,
  [958] = 609,
  [959] = 610,
  [960] = 611,
  [961] = 612,
  [962] = 767,
  [963] = 768,
  [964] = 566,
  [965] = 769,
  [966] = 774,
  [967] = 588,
  [968] = 783,
  [969] = 764,
  [970] = 820,
  [971] = 934,
  [972] = 849,
  [973] = 803,
  [974] = 804,
  [975] = 793,
  [976] = 836,
  [977] = 794,
  [978] = 617,
  [979] = 770,
  [980] = 592,
  [981] = 795,
  [982] = 982,
  [983] = 771,
  [984] = 615,
  [985] = 805,
  [986] = 838,
  [987] = 806,
  [988] = 822,
  [989] = 808,
  [990] = 776,
  [991] = 839,
  [992] = 847,
  [993] = 993,
  [994] = 620,
  [995] = 763,
  [996] = 854,
  [997] = 614,
  [998] = 797,
  [999] = 798,
  [1000] = 616,
  [1001] = 618,
  [1002] = 777,
  [1003] = 778,
  [1004] = 779,
  [1005] = 780,
  [1006] = 781,
  [1007] = 782,
  [1008] = 801,
  [1009] = 1009,
  [1010] = 856,
  [1011] = 728,
  [1012] = 857,
  [1013] = 1013,
  [1014] = 1014,
  [1015] = 1015,
  [1016] = 1016,
  [1017] = 853,
  [1018] = 1018,
  [1019] = 1019,
  [1020] = 796,
  [1021] = 1021,
  [1022] = 1022,
  [1023] = 1023,
  [1024] = 1024,
  [1025] = 1025,
  [1026] = 1026,
  [1027] = 1027,
  [1028] = 1028,
  [1029] = 1029,
  [1030] = 1030,
  [1031] = 1031,
  [1032] = 845,
  [1033] = 1033,
  [1034] = 1034,
  [1035] = 727,
  [1036] = 1036,
  [1037] = 1037,
  [1038] = 846,
  [1039] = 711,
  [1040] = 1040,
  [1041] = 1041,
  [1042] = 1042,
  [1043] = 598,
  [1044] = 908,
  [1045] = 599,
  [1046] = 917,
  [1047] = 713,
  [1048] = 916,
  [1049] = 1049,
  [1050] = 1050,
  [1051] = 982,
  [1052] = 1052,
  [1053] = 1053,
  [1054] = 1054,
  [1055] = 1055,
  [1056] = 1056,
  [1057] = 1057,
  [1058] = 1058,
  [1059] = 1059,
  [1060] = 1060,
  [1061] = 1061,
  [1062] = 1062,
  [1063] = 1063,
  [1064] = 1064,
  [1065] = 1065,
  [1066] = 1066,
  [1067] = 1067,
  [1068] = 1068,
  [1069] = 1069,
  [1070] = 1070,
  [1071] = 1071,
  [1072] = 588,
  [1073] = 1073,
  [1074] = 1074,
  [1075] = 1075,
  [1076] = 1076,
  [1077] = 1077,
  [1078] = 1078,
  [1079] = 1079,
  [1080] = 1080,
  [1081] = 1081,
  [1082] = 1082,
  [1083] = 1055,
  [1084] = 1023,
  [1085] = 1037,
  [1086] = 1079,
  [1087] = 1036,
  [1088] = 1088,
  [1089] = 1009,
  [1090] = 1021,
  [1091] = 1025,
  [1092] = 1052,
  [1093] = 1024,
  [1094] = 1026,
  [1095] = 1053,
  [1096] = 1054,
  [1097] = 1055,
  [1098] = 1056,
  [1099] = 452,
  [1100] = 1027,
  [1101] = 1079,
  [1102] = 1031,
  [1103] = 1040,
  [1104] = 1104,
  [1105] = 456,
  [1106] = 1080,
  [1107] = 1033,
  [1108] = 1015,
  [1109] = 1058,
  [1110] = 1041,
  [1111] = 1111,
  [1112] = 1034,
  [1113] = 1028,
  [1114] = 1014,
  [1115] = 1042,
  [1116] = 1059,
  [1117] = 1060,
  [1118] = 1061,
  [1119] = 1049,
  [1120] = 1063,
  [1121] = 1064,
  [1122] = 1082,
  [1123] = 1066,
  [1124] = 1067,
  [1125] = 1068,
  [1126] = 1069,
  [1127] = 1070,
  [1128] = 1071,
  [1129] = 1073,
  [1130] = 1074,
  [1131] = 1019,
  [1132] = 1076,
  [1133] = 1016,
  [1134] = 1077,
  [1135] = 1078,
  [1136] = 1030,
  [1137] = 1013,
  [1138] = 1058,
  [1139] = 1081,
  [1140] = 1059,
  [1141] = 1060,
  [1142] = 1061,
  [1143] = 1049,
  [1144] = 1063,
  [1145] = 1064,
  [1146] = 1082,
  [1147] = 1066,
  [1148] = 1067,
  [1149] = 1068,
  [1150] = 1069,
  [1151] = 1070,
  [1152] = 1071,
  [1153] = 1073,
  [1154] = 1074,
  [1155] = 1075,
  [1156] = 1076,
  [1157] = 1077,
  [1158] = 1078,
  [1159] = 1081,
  [1160] = 1062,
  [1161] = 1065,
  [1162] = 1062,
  [1163] = 1163,
  [1164] = 1080,
  [1165] = 1065,
  [1166] = 1057,
  [1167] = 1057,
  [1168] = 1050,
  [1169] = 1052,
  [1170] = 1053,
  [1171] = 1054,
  [1172] = 1022,
  [1173] = 1056,
  [1174] = 1050,
  [1175] = 1029,
  [1176] = 1018,
  [1177] = 1075,
  [1178] = 1057,
  [1179] = 1055,
  [1180] = 1058,
  [1181] = 1059,
  [1182] = 1060,
  [1183] = 1061,
  [1184] = 1049,
  [1185] = 1063,
  [1186] = 1064,
  [1187] = 1082,
  [1188] = 1066,
  [1189] = 1056,
  [1190] = 1068,
  [1191] = 1069,
  [1192] = 1070,
  [1193] = 1071,
  [1194] = 1073,
  [1195] = 1074,
  [1196] = 1075,
  [1197] = 1076,
  [1198] = 1077,
  [1199] = 1078,
  [1200] = 1081,
  [1201] = 1062,
  [1202] = 1065,
  [1203] = 1079,
  [1204] = 1080,
  [1205] = 1050,
  [1206] = 1052,
  [1207] = 1053,
  [1208] = 1054,
  [1209] = 1067,
  [1210] = 1210,
  [1211] = 1211,
  [1212] = 1212,
  [1213] = 1213,
  [1214] = 1214,
  [1215] = 1215,
  [1216] = 1216,
  [1217] = 1217,
  [1218] = 1218,
  [1219] = 1219,
  [1220] = 1220,
  [1221] = 1221,
  [1222] = 1222,
  [1223] = 1223,
  [1224] = 1224,
  [1225] = 1052,
  [1226] = 1226,
  [1227] = 1227,
  [1228] = 1227,
  [1229] = 1229,
  [1230] = 1230,
  [1231] = 1231,
  [1232] = 1232,
  [1233] = 1224,
  [1234] = 1234,
  [1235] = 1229,
  [1236] = 1221,
  [1237] = 1223,
  [1238] = 1238,
  [1239] = 1239,
  [1240] = 1240,
  [1241] = 1241,
  [1242] = 1242,
  [1243] = 1052,
  [1244] = 1244,
  [1245] = 1234,
  [1246] = 1246,
  [1247] = 1247,
  [1248] = 1248,
  [1249] = 1249,
  [1250] = 1250,
  [1251] = 1247,
  [1252] = 1240,
  [1253] = 1222,
  [1254] = 1254,
  [1255] = 1255,
  [1256] = 1256,
  [1257] = 1257,
  [1258] = 1258,
  [1259] = 1259,
  [1260] = 1260,
  [1261] = 1261,
  [1262] = 1262,
  [1263] = 1247,
  [1264] = 1248,
  [1265] = 1265,
  [1266] = 1240,
  [1267] = 1267,
  [1268] = 1268,
  [1269] = 1269,
  [1270] = 1270,
  [1271] = 1248,
  [1272] = 1222,
  [1273] = 1273,
  [1274] = 1274,
  [1275] = 1275,
  [1276] = 1276,
  [1277] = 1277,
  [1278] = 1278,
  [1279] = 1275,
  [1280] = 1280,
  [1281] = 1281,
  [1282] = 1278,
  [1283] = 1283,
  [1284] = 1284,
  [1285] = 1285,
  [1286] = 1286,
  [1287] = 1274,
  [1288] = 1288,
  [1289] = 1289,
  [1290] = 1290,
  [1291] = 1291,
  [1292] = 1284,
  [1293] = 1281,
  [1294] = 1284,
  [1295] = 1285,
  [1296] = 1274,
  [1297] = 1288,
  [1298] = 1290,
  [1299] = 1285,
  [1300] = 1285,
  [1301] = 1274,
  [1302] = 1288,
  [1303] = 1290,
  [1304] = 1281,
  [1305] = 1285,
  [1306] = 1274,
  [1307] = 1288,
  [1308] = 1290,
  [1309] = 1285,
  [1310] = 1274,
  [1311] = 1288,
  [1312] = 1290,
  [1313] = 1285,
  [1314] = 1285,
  [1315] = 1285,
  [1316] = 1286,
  [1317] = 1289,
  [1318] = 1288,
  [1319] = 1290,
  [1320] = 1281,
  [1321] = 1286,
  [1322] = 1322,
  [1323] = 1323,
  [1324] = 1286,
  [1325] = 1275,
  [1326] = 1275,
  [1327] = 1281,
  [1328] = 1328,
  [1329] = 1329,
  [1330] = 1330,
  [1331] = 1331,
  [1332] = 1332,
  [1333] = 1333,
  [1334] = 1329,
  [1335] = 1335,
  [1336] = 1330,
  [1337] = 1337,
  [1338] = 1338,
  [1339] = 1339,
  [1340] = 1340,
  [1341] = 1340,
  [1342] = 1342,
  [1343] = 1343,
  [1344] = 1344,
  [1345] = 1345,
  [1346] = 1346,
  [1347] = 1347,
  [1348] = 1348,
  [1349] = 1349,
  [1350] = 1350,
  [1351] = 1351,
  [1352] = 1350,
  [1353] = 1348,
  [1354] = 1351,
  [1355] = 1355,
  [1356] = 1355,
  [1357] = 1349,
  [1358] = 1350,
  [1359] = 1359,
  [1360] = 1348,
  [1361] = 1359,
  [1362] = 1349,
  [1363] = 1363,
  [1364] = 1342,
  [1365] = 1365,
  [1366] = 1363,
  [1367] = 1363,
  [1368] = 1343,
  [1369] = 1365,
  [1370] = 1365,
  [1371] = 1371,
  [1372] = 1371,
  [1373] = 1345,
  [1374] = 1346,
  [1375] = 1344,
  [1376] = 1376,
  [1377] = 540,
  [1378] = 1347,
  [1379] = 1379,
  [1380] = 1379,
  [1381] = 1381,
  [1382] = 1379,
  [1383] = 1383,
  [1384] = 1384,
  [1385] = 1385,
  [1386] = 1385,
  [1387] = 1387,
  [1388] = 1388,
  [1389] = 1389,
  [1390] = 1390,
  [1391] = 1391,
  [1392] = 1391,
  [1393] = 1390,
  [1394] = 1394,
  [1395] = 1395,
  [1396] = 1396,
  [1397] = 1395,
  [1398] = 1394,
  [1399] = 1399,
  [1400] = 1400,
  [1401] = 1401,
  [1402] = 1402,
  [1403] = 1403,
  [1404] = 1404,
  [1405] = 1405,
  [1406] = 1406,
  [1407] = 1407,
  [1408] = 1408,
  [1409] = 1409,
  [1410] = 1410,
  [1411] = 1411,
  [1412] = 1411,
  [1413] = 1413,
  [1414] = 1414,
  [1415] = 1415,
  [1416] = 1416,
  [1417] = 1417,
  [1418] = 1418,
  [1419] = 1418,
  [1420] = 1420,
  [1421] = 1421,
  [1422] = 1422,
  [1423] = 1413,
  [1424] = 1424,
  [1425] = 1414,
  [1426] = 1415,
  [1427] = 1427,
  [1428] = 1428,
  [1429] = 1429,
  [1430] = 1430,
  [1431] = 1431,
  [1432] = 1432,
  [1433] = 1433,
  [1434] = 1434,
  [1435] = 1435,
  [1436] = 1436,
  [1437] = 1437,
  [1438] = 1438,
  [1439] = 1439,
  [1440] = 1440,
  [1441] = 1441,
  [1442] = 1442,
  [1443] = 1443,
  [1444] = 854,
  [1445] = 1445,
  [1446] = 1446,
  [1447] = 1447,
  [1448] = 1448,
  [1449] = 1449,
  [1450] = 1450,
  [1451] = 1451,
  [1452] = 1452,
  [1453] = 1453,
  [1454] = 1454,
  [1455] = 1431,
  [1456] = 1456,
  [1457] = 1457,
  [1458] = 1458,
  [1459] = 1459,
  [1460] = 1460,
  [1461] = 1461,
  [1462] = 1462,
  [1463] = 1458,
  [1464] = 1464,
  [1465] = 1465,
  [1466] = 1464,
  [1467] = 1461,
  [1468] = 1459,
  [1469] = 1460,
  [1470] = 1457,
  [1471] = 1462,
  [1472] = 1458,
  [1473] = 1464,
  [1474] = 1474,
  [1475] = 1475,
  [1476] = 1458,
  [1477] = 1464,
  [1478] = 1462,
  [1479] = 1462,
  [1480] = 1458,
  [1481] = 1464,
  [1482] = 1474,
  [1483] = 1475,
  [1484] = 1462,
  [1485] = 1458,
  [1486] = 1464,
  [1487] = 1462,
  [1488] = 1458,
  [1489] = 1489,
  [1490] = 1464,
  [1491] = 1462,
  [1492] = 1474,
  [1493] = 1475,
  [1494] = 1494,
  [1495] = 1495,
  [1496] = 1496,
  [1497] = 1497,
  [1498] = 1498,
  [1499] = 908,
  [1500] = 1500,
  [1501] = 1501,
  [1502] = 1500,
  [1503] = 1503,
  [1504] = 1504,
  [1505] = 1505,
  [1506] = 600,
  [1507] = 1507,
  [1508] = 1508,
  [1509] = 1509,
  [1510] = 1510,
  [1511] = 1511,
  [1512] = 1512,
  [1513] = 1513,
  [1514] = 1514,
  [1515] = 1515,
  [1516] = 1516,
  [1517] = 1517,
  [1518] = 1518,
  [1519] = 1519,
  [1520] = 1520,
  [1521] = 1521,
  [1522] = 1522,
  [1523] = 1523,
  [1524] = 1524,
  [1525] = 1525,
  [1526] = 1526,
  [1527] = 1527,
  [1528] = 1518,
  [1529] = 1505,
  [1530] = 1530,
  [1531] = 1531,
  [1532] = 1532,
  [1533] = 1533,
  [1534] = 600,
  [1535] = 1535,
  [1536] = 600,
  [1537] = 1537,
  [1538] = 1538,
  [1539] = 1539,
  [1540] = 1540,
  [1541] = 1541,
  [1542] = 1542,
  [1543] = 1543,
  [1544] = 1544,
  [1545] = 1545,
  [1546] = 1546,
  [1547] = 1494,
  [1548] = 1548,
  [1549] = 1549,
  [1550] = 444,
  [1551] = 1551,
  [1552] = 1552,
  [1553] = 1553,
  [1554] = 1554,
  [1555] = 1555,
  [1556] = 1555,
  [1557] = 1557,
  [1558] = 1558,
  [1559] = 1553,
  [1560] = 1552,
  [1561] = 1554,
  [1562] = 445,
  [1563] = 1557,
  [1564] = 1564,
  [1565] = 1565,
  [1566] = 1566,
  [1567] = 1567,
  [1568] = 1568,
  [1569] = 1569,
  [1570] = 1570,
  [1571] = 1571,
  [1572] = 1441,
  [1573] = 1573,
  [1574] = 1574,
  [1575] = 1575,
  [1576] = 1576,
  [1577] = 1577,
  [1578] = 1578,
  [1579] = 1579,
  [1580] = 1580,
  [1581] = 1581,
  [1582] = 1494,
  [1583] = 1583,
  [1584] = 1584,
  [1585] = 1585,
  [1586] = 1586,
  [1587] = 1587,
  [1588] = 1588,
  [1589] = 1589,
  [1590] = 1590,
  [1591] = 1591,
  [1592] = 1592,
  [1593] = 1578,
  [1594] = 1594,
  [1595] = 1578,
  [1596] = 902,
  [1597] = 1597,
  [1598] = 597,
  [1599] = 1599,
  [1600] = 1586,
  [1601] = 1601,
  [1602] = 1591,
  [1603] = 601,
  [1604] = 617,
  [1605] = 1605,
  [1606] = 1606,
  [1607] = 1607,
  [1608] = 1608,
  [1609] = 1609,
  [1610] = 1610,
  [1611] = 1611,
  [1612] = 600,
  [1613] = 1613,
  [1614] = 1614,
  [1615] = 1580,
  [1616] = 1616,
  [1617] = 1614,
  [1618] = 1618,
  [1619] = 1613,
  [1620] = 1620,
  [1621] = 1621,
  [1622] = 1622,
  [1623] = 1623,
  [1624] = 1624,
  [1625] = 1623,
  [1626] = 1626,
  [1627] = 1627,
  [1628] = 1628,
  [1629] = 1629,
  [1630] = 1622,
  [1631] = 1631,
  [1632] = 1632,
  [1633] = 1588,
  [1634] = 1634,
  [1635] = 1635,
  [1636] = 1636,
  [1637] = 1637,
  [1638] = 1638,
  [1639] = 1639,
  [1640] = 1640,
  [1641] = 1575,
  [1642] = 1599,
  [1643] = 1643,
  [1644] = 1644,
  [1645] = 1620,
  [1646] = 1646,
  [1647] = 711,
  [1648] = 1648,
  [1649] = 446,
  [1650] = 1650,
  [1651] = 1651,
  [1652] = 1631,
  [1653] = 447,
  [1654] = 1609,
  [1655] = 1613,
  [1656] = 1651,
  [1657] = 1631,
  [1658] = 1658,
  [1659] = 597,
  [1660] = 1651,
  [1661] = 1631,
  [1662] = 1621,
  [1663] = 601,
  [1664] = 1651,
  [1665] = 1631,
  [1666] = 1651,
  [1667] = 1609,
  [1668] = 1651,
  [1669] = 1651,
  [1670] = 1651,
  [1671] = 1651,
  [1672] = 617,
  [1673] = 1588,
  [1674] = 1674,
  [1675] = 597,
  [1676] = 1635,
  [1677] = 1677,
  [1678] = 601,
  [1679] = 446,
  [1680] = 1680,
  [1681] = 1681,
  [1682] = 1575,
  [1683] = 1494,
  [1684] = 1636,
  [1685] = 1599,
  [1686] = 1632,
  [1687] = 617,
  [1688] = 1637,
  [1689] = 1689,
  [1690] = 1690,
  [1691] = 1634,
  [1692] = 1692,
  [1693] = 1627,
  [1694] = 1643,
  [1695] = 600,
  [1696] = 600,
  [1697] = 1697,
  [1698] = 1651,
  [1699] = 1628,
  [1700] = 1629,
  [1701] = 1697,
  [1702] = 1624,
  [1703] = 1651,
  [1704] = 1631,
  [1705] = 447,
  [1706] = 1638,
  [1707] = 1639,
  [1708] = 1651,
  [1709] = 1504,
  [1710] = 1710,
  [1711] = 1711,
  [1712] = 1712,
  [1713] = 1713,
  [1714] = 1714,
  [1715] = 1715,
  [1716] = 1716,
  [1717] = 1717,
  [1718] = 1718,
  [1719] = 1507,
  [1720] = 1720,
  [1721] = 1721,
  [1722] = 1722,
  [1723] = 1723,
  [1724] = 1724,
  [1725] = 1722,
  [1726] = 1726,
  [1727] = 1727,
  [1728] = 1728,
  [1729] = 1729,
  [1730] = 1730,
  [1731] = 1731,
  [1732] = 1732,
  [1733] = 1496,
  [1734] = 1734,
  [1735] = 1735,
  [1736] = 1736,
  [1737] = 1737,
  [1738] = 1738,
  [1739] = 1739,
  [1740] = 1740,
  [1741] = 452,
  [1742] = 1742,
  [1743] = 1743,
  [1744] = 1744,
  [1745] = 456,
  [1746] = 1497,
  [1747] = 1747,
  [1748] = 540,
  [1749] = 1729,
  [1750] = 1750,
  [1751] = 1751,
  [1752] = 1752,
  [1753] = 1753,
  [1754] = 1754,
  [1755] = 1755,
  [1756] = 1756,
  [1757] = 1757,
  [1758] = 1758,
  [1759] = 1759,
  [1760] = 451,
  [1761] = 1761,
  [1762] = 1498,
  [1763] = 1726,
  [1764] = 1495,
  [1765] = 1765,
  [1766] = 1510,
  [1767] = 1513,
  [1768] = 1734,
  [1769] = 1769,
  [1770] = 1770,
  [1771] = 1504,
  [1772] = 1526,
  [1773] = 1773,
  [1774] = 1507,
  [1775] = 1775,
  [1776] = 1776,
  [1777] = 454,
  [1778] = 1717,
  [1779] = 1779,
  [1780] = 1780,
  [1781] = 1511,
  [1782] = 1782,
  [1783] = 794,
  [1784] = 1784,
  [1785] = 1785,
  [1786] = 1786,
  [1787] = 1787,
  [1788] = 1788,
  [1789] = 1747,
  [1790] = 1790,
  [1791] = 450,
  [1792] = 1792,
  [1793] = 1793,
  [1794] = 1794,
  [1795] = 1795,
  [1796] = 1714,
  [1797] = 1797,
  [1798] = 1798,
  [1799] = 1752,
  [1800] = 1800,
  [1801] = 1801,
  [1802] = 1802,
  [1803] = 1803,
  [1804] = 1804,
  [1805] = 1805,
  [1806] = 1806,
  [1807] = 1756,
  [1808] = 1808,
  [1809] = 1754,
  [1810] = 1810,
  [1811] = 455,
  [1812] = 1812,
  [1813] = 1813,
  [1814] = 1740,
  [1815] = 1815,
  [1816] = 1816,
  [1817] = 1817,
  [1818] = 1795,
  [1819] = 1819,
  [1820] = 1820,
  [1821] = 1520,
  [1822] = 1822,
  [1823] = 1804,
  [1824] = 1824,
  [1825] = 1825,
  [1826] = 1826,
  [1827] = 1510,
  [1828] = 1513,
  [1829] = 1727,
  [1830] = 1520,
  [1831] = 1525,
  [1832] = 1832,
  [1833] = 1770,
  [1834] = 1526,
  [1835] = 1835,
  [1836] = 1836,
  [1837] = 1711,
  [1838] = 453,
  [1839] = 1775,
  [1840] = 1713,
  [1841] = 1784,
  [1842] = 1723,
  [1843] = 1787,
  [1844] = 1844,
  [1845] = 1845,
  [1846] = 1846,
  [1847] = 1847,
  [1848] = 1848,
  [1849] = 1514,
  [1850] = 1850,
  [1851] = 1851,
  [1852] = 1852,
  [1853] = 1853,
  [1854] = 1854,
  [1855] = 1855,
  [1856] = 1846,
  [1857] = 1857,
  [1858] = 1858,
  [1859] = 1859,
  [1860] = 1860,
  [1861] = 1732,
  [1862] = 1862,
  [1863] = 1863,
  [1864] = 1864,
  [1865] = 1865,
  [1866] = 1866,
  [1867] = 1867,
  [1868] = 1868,
  [1869] = 1848,
  [1870] = 1866,
  [1871] = 1819,
  [1872] = 1848,
  [1873] = 1873,
  [1874] = 1874,
  [1875] = 1875,
  [1876] = 1511,
  [1877] = 1877,
  [1878] = 1878,
  [1879] = 1879,
  [1880] = 1865,
  [1881] = 1881,
  [1882] = 1508,
  [1883] = 1883,
  [1884] = 1884,
  [1885] = 1883,
  [1886] = 1845,
  [1887] = 1525,
  [1888] = 1888,
  [1889] = 1889,
  [1890] = 1890,
  [1891] = 1851,
  [1892] = 1892,
  [1893] = 1893,
  [1894] = 1852,
  [1895] = 1895,
  [1896] = 1896,
  [1897] = 1897,
  [1898] = 1898,
  [1899] = 1889,
  [1900] = 1900,
  [1901] = 1901,
  [1902] = 1848,
  [1903] = 1903,
  [1904] = 1878,
  [1905] = 1905,
  [1906] = 1906,
  [1907] = 1907,
  [1908] = 1512,
  [1909] = 1897,
  [1910] = 1848,
  [1911] = 1911,
  [1912] = 1912,
  [1913] = 1913,
  [1914] = 713,
  [1915] = 1848,
  [1916] = 1866,
  [1917] = 1917,
  [1918] = 1918,
  [1919] = 1919,
  [1920] = 1873,
  [1921] = 1521,
  [1922] = 1523,
  [1923] = 1923,
  [1924] = 1924,
  [1925] = 1925,
  [1926] = 1926,
  [1927] = 1927,
  [1928] = 1928,
  [1929] = 1758,
  [1930] = 1930,
  [1931] = 1931,
  [1932] = 1903,
  [1933] = 1918,
  [1934] = 1931,
  [1935] = 1923,
  [1936] = 1936,
  [1937] = 1937,
  [1938] = 1912,
  [1939] = 1884,
  [1940] = 1940,
  [1941] = 1941,
  [1942] = 1937,
  [1943] = 1848,
  [1944] = 1724,
  [1945] = 1945,
  [1946] = 1946,
  [1947] = 1858,
  [1948] = 1883,
  [1949] = 1866,
  [1950] = 1940,
  [1951] = 1855,
  [1952] = 1280,
  [1953] = 1718,
  [1954] = 1954,
  [1955] = 1955,
  [1956] = 1956,
  [1957] = 1957,
  [1958] = 1958,
  [1959] = 1959,
  [1960] = 1960,
  [1961] = 1961,
  [1962] = 1962,
  [1963] = 1963,
  [1964] = 1964,
  [1965] = 1965,
  [1966] = 1966,
  [1967] = 1967,
  [1968] = 1968,
  [1969] = 1969,
  [1970] = 1970,
  [1971] = 1971,
  [1972] = 559,
  [1973] = 1973,
  [1974] = 1974,
  [1975] = 1975,
  [1976] = 1976,
  [1977] = 560,
  [1978] = 561,
  [1979] = 1979,
  [1980] = 1980,
  [1981] = 1981,
  [1982] = 1982,
  [1983] = 1983,
  [1984] = 562,
  [1985] = 1867,
  [1986] = 1986,
  [1987] = 1987,
  [1988] = 564,
  [1989] = 1989,
  [1990] = 1990,
  [1991] = 1991,
  [1992] = 1992,
  [1993] = 551,
  [1994] = 1994,
  [1995] = 1995,
  [1996] = 1996,
  [1997] = 1997,
  [1998] = 565,
  [1999] = 567,
  [2000] = 568,
  [2001] = 569,
  [2002] = 2002,
  [2003] = 2003,
  [2004] = 571,
  [2005] = 1335,
  [2006] = 2006,
  [2007] = 2007,
  [2008] = 2008,
  [2009] = 2009,
  [2010] = 1960,
  [2011] = 2011,
  [2012] = 578,
  [2013] = 1956,
  [2014] = 2008,
  [2015] = 1963,
  [2016] = 1965,
  [2017] = 2017,
  [2018] = 2018,
  [2019] = 2019,
  [2020] = 2020,
  [2021] = 461,
  [2022] = 462,
  [2023] = 2023,
  [2024] = 2024,
  [2025] = 2025,
  [2026] = 2026,
  [2027] = 2027,
  [2028] = 463,
  [2029] = 2029,
  [2030] = 2030,
  [2031] = 2031,
  [2032] = 2032,
  [2033] = 2033,
  [2034] = 2034,
  [2035] = 2035,
  [2036] = 2036,
  [2037] = 2037,
  [2038] = 2038,
  [2039] = 2039,
  [2040] = 2030,
  [2041] = 2041,
  [2042] = 2042,
  [2043] = 469,
  [2044] = 470,
  [2045] = 471,
  [2046] = 472,
  [2047] = 2047,
  [2048] = 485,
  [2049] = 460,
  [2050] = 2050,
  [2051] = 2051,
  [2052] = 473,
  [2053] = 1996,
  [2054] = 492,
  [2055] = 2055,
  [2056] = 486,
  [2057] = 474,
  [2058] = 2058,
  [2059] = 1761,
  [2060] = 2060,
  [2061] = 1973,
  [2062] = 1974,
  [2063] = 2063,
  [2064] = 476,
  [2065] = 477,
  [2066] = 478,
  [2067] = 479,
  [2068] = 2068,
  [2069] = 2069,
  [2070] = 480,
  [2071] = 2071,
  [2072] = 481,
  [2073] = 2007,
  [2074] = 482,
  [2075] = 457,
  [2076] = 484,
  [2077] = 2077,
  [2078] = 487,
  [2079] = 489,
  [2080] = 491,
  [2081] = 1966,
  [2082] = 1967,
  [2083] = 2083,
  [2084] = 2084,
  [2085] = 2037,
  [2086] = 2086,
  [2087] = 494,
  [2088] = 2032,
  [2089] = 2038,
  [2090] = 2090,
  [2091] = 2091,
  [2092] = 2092,
  [2093] = 2093,
  [2094] = 2094,
  [2095] = 2095,
  [2096] = 495,
  [2097] = 499,
  [2098] = 2039,
  [2099] = 504,
  [2100] = 521,
  [2101] = 522,
  [2102] = 515,
  [2103] = 523,
  [2104] = 524,
  [2105] = 2041,
  [2106] = 2063,
  [2107] = 2107,
  [2108] = 2108,
  [2109] = 527,
  [2110] = 528,
  [2111] = 1957,
  [2112] = 1958,
  [2113] = 529,
  [2114] = 530,
  [2115] = 537,
  [2116] = 2116,
  [2117] = 554,
  [2118] = 2041,
  [2119] = 557,
  [2120] = 1969,
  [2121] = 2063,
  [2122] = 570,
  [2123] = 2123,
  [2124] = 468,
  [2125] = 572,
  [2126] = 573,
  [2127] = 574,
  [2128] = 2041,
  [2129] = 575,
  [2130] = 576,
  [2131] = 483,
  [2132] = 2063,
  [2133] = 583,
  [2134] = 2134,
  [2135] = 502,
  [2136] = 2136,
  [2137] = 464,
  [2138] = 490,
  [2139] = 2139,
  [2140] = 497,
  [2141] = 2041,
  [2142] = 500,
  [2143] = 2143,
  [2144] = 582,
  [2145] = 2063,
  [2146] = 458,
  [2147] = 503,
  [2148] = 505,
  [2149] = 465,
  [2150] = 466,
  [2151] = 467,
  [2152] = 1911,
  [2153] = 2153,
  [2154] = 2154,
  [2155] = 2155,
  [2156] = 493,
  [2157] = 496,
  [2158] = 2063,
  [2159] = 498,
  [2160] = 508,
  [2161] = 501,
  [2162] = 2162,
  [2163] = 2163,
  [2164] = 2143,
  [2165] = 1992,
  [2166] = 506,
  [2167] = 507,
  [2168] = 2108,
  [2169] = 510,
  [2170] = 511,
  [2171] = 513,
  [2172] = 2041,
  [2173] = 514,
  [2174] = 2026,
  [2175] = 2018,
  [2176] = 2176,
  [2177] = 516,
  [2178] = 2024,
  [2179] = 517,
  [2180] = 1991,
  [2181] = 519,
  [2182] = 1994,
  [2183] = 2006,
  [2184] = 531,
  [2185] = 2023,
  [2186] = 2186,
  [2187] = 2025,
  [2188] = 2034,
  [2189] = 2189,
  [2190] = 542,
  [2191] = 543,
  [2192] = 544,
  [2193] = 2193,
  [2194] = 545,
  [2195] = 550,
  [2196] = 2196,
  [2197] = 552,
  [2198] = 2198,
  [2199] = 553,
  [2200] = 555,
  [2201] = 2201,
  [2202] = 556,
  [2203] = 2203,
  [2204] = 2204,
  [2205] = 563,
  [2206] = 577,
  [2207] = 579,
  [2208] = 580,
  [2209] = 1989,
  [2210] = 581,
  [2211] = 2211,
  [2212] = 509,
  [2213] = 2050,
  [2214] = 2019,
  [2215] = 2215,
  [2216] = 2020,
  [2217] = 2217,
  [2218] = 1511,
  [2219] = 2219,
  [2220] = 2220,
  [2221] = 2134,
  [2222] = 2222,
  [2223] = 2223,
  [2224] = 2136,
  [2225] = 2225,
  [2226] = 533,
  [2227] = 2227,
  [2228] = 2228,
  [2229] = 2229,
  [2230] = 2230,
  [2231] = 1863,
  [2232] = 2232,
  [2233] = 2233,
  [2234] = 2234,
  [2235] = 2198,
  [2236] = 2236,
  [2237] = 2237,
  [2238] = 2238,
  [2239] = 2239,
  [2240] = 512,
  [2241] = 2201,
  [2242] = 2047,
  [2243] = 2243,
  [2244] = 2244,
  [2245] = 2245,
  [2246] = 2246,
  [2247] = 2002,
  [2248] = 2248,
  [2249] = 1820,
  [2250] = 536,
  [2251] = 539,
  [2252] = 541,
  [2253] = 2253,
  [2254] = 2254,
  [2255] = 558,
  [2256] = 1538,
  [2257] = 2257,
  [2258] = 2258,
  [2259] = 488,
  [2260] = 2162,
  [2261] = 2261,
  [2262] = 2262,
  [2263] = 2263,
  [2264] = 2264,
  [2265] = 2042,
  [2266] = 2266,
  [2267] = 2267,
  [2268] = 518,
  [2269] = 525,
  [2270] = 526,
  [2271] = 532,
  [2272] = 534,
  [2273] = 2273,
  [2274] = 2068,
  [2275] = 538,
  [2276] = 2203,
  [2277] = 2277,
  [2278] = 546,
  [2279] = 2092,
  [2280] = 2029,
  [2281] = 2281,
  [2282] = 2282,
  [2283] = 2204,
  [2284] = 547,
  [2285] = 2258,
  [2286] = 548,
  [2287] = 2189,
  [2288] = 2009,
  [2289] = 2289,
  [2290] = 2290,
  [2291] = 2291,
  [2292] = 549,
  [2293] = 2293,
  [2294] = 1875,
  [2295] = 2051,
  [2296] = 2163,
  [2297] = 2297,
  [2298] = 2298,
  [2299] = 2299,
  [2300] = 2300,
  [2301] = 2301,
  [2302] = 2302,
  [2303] = 2303,
  [2304] = 2304,
  [2305] = 2305,
  [2306] = 2306,
  [2307] = 2307,
  [2308] = 2308,
  [2309] = 2309,
  [2310] = 2306,
  [2311] = 2311,
  [2312] = 2312,
  [2313] = 1448,
  [2314] = 2314,
  [2315] = 2315,
  [2316] = 2316,
  [2317] = 1797,
  [2318] = 2318,
  [2319] = 2303,
  [2320] = 2320,
  [2321] = 2321,
  [2322] = 2322,
  [2323] = 2311,
  [2324] = 2300,
  [2325] = 2325,
  [2326] = 2326,
  [2327] = 2303,
  [2328] = 2328,
  [2329] = 2329,
  [2330] = 2299,
  [2331] = 2302,
  [2332] = 2332,
  [2333] = 2311,
  [2334] = 2303,
  [2335] = 2311,
  [2336] = 2336,
  [2337] = 2337,
  [2338] = 2338,
  [2339] = 2339,
  [2340] = 2340,
  [2341] = 2303,
  [2342] = 2342,
  [2343] = 2343,
  [2344] = 2315,
  [2345] = 2304,
  [2346] = 2346,
  [2347] = 2347,
  [2348] = 2348,
  [2349] = 2349,
  [2350] = 2036,
  [2351] = 2351,
  [2352] = 2352,
  [2353] = 2353,
  [2354] = 2354,
  [2355] = 2355,
  [2356] = 2356,
  [2357] = 2357,
  [2358] = 2358,
  [2359] = 2359,
  [2360] = 2360,
  [2361] = 2361,
  [2362] = 2362,
  [2363] = 2363,
  [2364] = 2364,
  [2365] = 2365,
  [2366] = 2366,
  [2367] = 2347,
  [2368] = 2368,
  [2369] = 2369,
  [2370] = 2370,
  [2371] = 2365,
  [2372] = 2356,
  [2373] = 2373,
  [2374] = 2374,
  [2375] = 2375,
  [2376] = 2376,
  [2377] = 2377,
  [2378] = 2378,
  [2379] = 2379,
  [2380] = 2308,
  [2381] = 2381,
  [2382] = 2382,
  [2383] = 2383,
  [2384] = 2384,
  [2385] = 2361,
  [2386] = 2386,
  [2387] = 2387,
  [2388] = 916,
  [2389] = 2389,
  [2390] = 2311,
  [2391] = 2391,
  [2392] = 2392,
  [2393] = 2393,
  [2394] = 2322,
  [2395] = 2326,
  [2396] = 2396,
  [2397] = 2332,
  [2398] = 2336,
  [2399] = 2337,
  [2400] = 2315,
  [2401] = 2401,
  [2402] = 2402,
  [2403] = 2403,
  [2404] = 2404,
  [2405] = 2340,
  [2406] = 2406,
  [2407] = 2393,
  [2408] = 2408,
  [2409] = 2409,
  [2410] = 2383,
  [2411] = 2411,
  [2412] = 2412,
  [2413] = 2413,
  [2414] = 2414,
  [2415] = 2415,
  [2416] = 1720,
  [2417] = 2417,
  [2418] = 2339,
  [2419] = 2419,
  [2420] = 2352,
  [2421] = 2421,
  [2422] = 2422,
  [2423] = 2309,
  [2424] = 2424,
  [2425] = 2411,
  [2426] = 2386,
  [2427] = 2427,
  [2428] = 2353,
  [2429] = 2429,
  [2430] = 2357,
  [2431] = 2431,
  [2432] = 2432,
  [2433] = 2312,
  [2434] = 2434,
  [2435] = 2419,
  [2436] = 2354,
  [2437] = 2437,
  [2438] = 2438,
  [2439] = 2439,
  [2440] = 2415,
  [2441] = 1955,
  [2442] = 2442,
  [2443] = 1986,
  [2444] = 2444,
  [2445] = 2417,
  [2446] = 2446,
  [2447] = 2447,
  [2448] = 2448,
  [2449] = 2236,
  [2450] = 2450,
  [2451] = 2451,
  [2452] = 2452,
  [2453] = 1454,
  [2454] = 2454,
  [2455] = 2455,
  [2456] = 2456,
  [2457] = 2442,
  [2458] = 2458,
  [2459] = 2343,
  [2460] = 2460,
  [2461] = 2338,
  [2462] = 2462,
  [2463] = 2463,
  [2464] = 2376,
  [2465] = 2349,
  [2466] = 2466,
  [2467] = 2467,
  [2468] = 2468,
  [2469] = 2315,
  [2470] = 2470,
  [2471] = 2471,
  [2472] = 1769,
  [2473] = 2473,
  [2474] = 2303,
  [2475] = 2346,
  [2476] = 2476,
  [2477] = 2477,
  [2478] = 2478,
  [2479] = 2362,
  [2480] = 2309,
  [2481] = 2481,
  [2482] = 2455,
  [2483] = 2483,
  [2484] = 2484,
  [2485] = 2369,
  [2486] = 2309,
  [2487] = 2487,
  [2488] = 2488,
  [2489] = 2370,
  [2490] = 2490,
  [2491] = 2314,
  [2492] = 2492,
  [2493] = 2493,
  [2494] = 2298,
  [2495] = 2487,
  [2496] = 2378,
  [2497] = 2342,
  [2498] = 2358,
  [2499] = 2311,
  [2500] = 2377,
  [2501] = 2501,
  [2502] = 2502,
  [2503] = 2503,
  [2504] = 2438,
  [2505] = 2505,
  [2506] = 2506,
  [2507] = 2507,
  [2508] = 2508,
  [2509] = 2509,
  [2510] = 2510,
  [2511] = 2511,
  [2512] = 2512,
  [2513] = 2513,
  [2514] = 2514,
  [2515] = 2515,
  [2516] = 2516,
  [2517] = 2517,
  [2518] = 2515,
  [2519] = 2519,
  [2520] = 2520,
  [2521] = 2521,
  [2522] = 2522,
  [2523] = 2523,
  [2524] = 2524,
  [2525] = 2525,
  [2526] = 2526,
  [2527] = 2527,
  [2528] = 2528,
  [2529] = 2529,
  [2530] = 2530,
  [2531] = 2519,
  [2532] = 2532,
  [2533] = 2516,
  [2534] = 2534,
  [2535] = 2535,
  [2536] = 2536,
  [2537] = 2537,
  [2538] = 2519,
  [2539] = 2539,
  [2540] = 2540,
  [2541] = 2541,
  [2542] = 2505,
  [2543] = 2543,
  [2544] = 2544,
  [2545] = 2545,
  [2546] = 2514,
  [2547] = 2547,
  [2548] = 2548,
  [2549] = 2527,
  [2550] = 2550,
  [2551] = 2551,
  [2552] = 2525,
  [2553] = 2553,
  [2554] = 2522,
  [2555] = 2555,
  [2556] = 2556,
  [2557] = 2557,
  [2558] = 2558,
  [2559] = 2559,
  [2560] = 2544,
  [2561] = 2543,
  [2562] = 2535,
  [2563] = 2523,
  [2564] = 2564,
  [2565] = 2506,
  [2566] = 2566,
  [2567] = 2567,
  [2568] = 2568,
  [2569] = 2055,
  [2570] = 2509,
  [2571] = 2571,
  [2572] = 2572,
  [2573] = 2567,
  [2574] = 2564,
  [2575] = 2575,
  [2576] = 2532,
  [2577] = 2505,
  [2578] = 2567,
  [2579] = 2579,
  [2580] = 2557,
  [2581] = 2581,
  [2582] = 2582,
  [2583] = 2583,
  [2584] = 2584,
  [2585] = 2585,
  [2586] = 2586,
  [2587] = 2534,
  [2588] = 2571,
  [2589] = 2507,
  [2590] = 2535,
  [2591] = 2591,
  [2592] = 2592,
  [2593] = 2541,
  [2594] = 2509,
  [2595] = 2540,
  [2596] = 2596,
  [2597] = 1987,
  [2598] = 2598,
  [2599] = 2550,
  [2600] = 2559,
  [2601] = 2539,
  [2602] = 2548,
  [2603] = 2603,
  [2604] = 2524,
  [2605] = 2605,
  [2606] = 2559,
  [2607] = 2532,
  [2608] = 2608,
  [2609] = 2609,
  [2610] = 2610,
  [2611] = 2551,
  [2612] = 2519,
  [2613] = 2528,
  [2614] = 2614,
  [2615] = 2615,
  [2616] = 2616,
  [2617] = 2519,
  [2618] = 2579,
  [2619] = 2581,
  [2620] = 1860,
  [2621] = 2621,
  [2622] = 2506,
  [2623] = 2623,
  [2624] = 2616,
  [2625] = 2625,
  [2626] = 2610,
  [2627] = 2572,
  [2628] = 2528,
  [2629] = 2610,
  [2630] = 2509,
  [2631] = 2631,
  [2632] = 2632,
  [2633] = 2603,
  [2634] = 2634,
  [2635] = 2610,
  [2636] = 2527,
  [2637] = 2637,
  [2638] = 2547,
  [2639] = 2548,
  [2640] = 2527,
  [2641] = 2641,
  [2642] = 2642,
  [2643] = 2643,
  [2644] = 2506,
  [2645] = 2645,
  [2646] = 2582,
  [2647] = 2647,
  [2648] = 2567,
  [2649] = 2649,
  [2650] = 2650,
  [2651] = 2506,
  [2652] = 2621,
  [2653] = 2547,
  [2654] = 2567,
  [2655] = 2544,
  [2656] = 2545,
  [2657] = 2657,
  [2658] = 2509,
  [2659] = 2555,
  [2660] = 2660,
  [2661] = 2585,
  [2662] = 2662,
  [2663] = 2548,
  [2664] = 2664,
  [2665] = 2525,
  [2666] = 2615,
  [2667] = 2527,
  [2668] = 2668,
  [2669] = 2558,
  [2670] = 2525,
  [2671] = 2671,
  [2672] = 2672,
  [2673] = 2673,
  [2674] = 2674,
  [2675] = 2610,
  [2676] = 2528,
  [2677] = 2547,
  [2678] = 2634,
  [2679] = 2555,
  [2680] = 2680,
  [2681] = 2509,
  [2682] = 2558,
  [2683] = 2505,
  [2684] = 2544,
  [2685] = 2685,
  [2686] = 2535,
  [2687] = 2687,
  [2688] = 2510,
  [2689] = 2689,
  [2690] = 2662,
  [2691] = 2509,
  [2692] = 2692,
  [2693] = 2693,
  [2694] = 2694,
  [2695] = 2641,
  [2696] = 2581,
  [2697] = 2643,
  [2698] = 2698,
  [2699] = 2662,
  [2700] = 2519,
  [2701] = 2671,
  [2702] = 2529,
  [2703] = 2703,
  [2704] = 2575,
  [2705] = 2705,
  [2706] = 2692,
  [2707] = 2631,
  [2708] = 2555,
  [2709] = 2709,
  [2710] = 2710,
  [2711] = 2528,
  [2712] = 2712,
  [2713] = 2521,
  [2714] = 2693,
  [2715] = 2579,
  [2716] = 2694,
  [2717] = 2709,
  [2718] = 2558,
  [2719] = 2674,
  [2720] = 2641,
  [2721] = 2632,
  [2722] = 2680,
  [2723] = 2532,
  [2724] = 2724,
  [2725] = 2605,
  [2726] = 2579,
  [2727] = 2709,
  [2728] = 2728,
  [2729] = 2662,
  [2730] = 2712,
  [2731] = 2731,
  [2732] = 2732,
  [2733] = 2671,
  [2734] = 2734,
  [2735] = 2712,
  [2736] = 2709,
  [2737] = 2737,
  [2738] = 2712,
  [2739] = 2671,
  [2740] = 2740,
  [2741] = 2741,
};

static TSCharacterRange sym_escape_sequence_character_set_1[] = {
  {'"', '"'}, {'$', '$'}, {'0', '7'}, {'\\', '\\'}, {'`', '`'}, {'e', 'f'}, {'n', 'n'}, {'r', 'r'},
  {'t', 'v'}, {'x', 'x'},
};

static TSCharacterRange sym_name_character_set_1[] = {
  {'0', '9'}, {'A', 'Z'}, {'_', '_'}, {'a', 'z'}, {0x80, 0x9f}, {0xa1, 0x200a}, {0x200c, 0x205f}, {0x2061, 0xfefe},
  {0xff00, 0xffff},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(67);
      ADVANCE_MAP(
        '!', 135,
        '"', 184,
        '#', 264,
        '$', 188,
        '%', 224,
        '&', 85,
        '\'', 186,
        '(', 95,
        ')', 96,
        '*', 219,
        '+', 127,
        ',', 86,
        '-', 130,
        '.', 217,
        '/', 222,
        '0', 117,
        ':', 93,
        ';', 82,
        '<', 205,
        '=', 87,
        '>', 206,
        '?', 100,
        '@', 136,
        '[', 155,
        '\\', 89,
        ']', 156,
        '^', 196,
        '_', 259,
        '`', 187,
        '{', 90,
        '|', 105,
        '}', 91,
        '~', 133,
        'A', 249,
        'a', 249,
        'B', 225,
        'b', 225,
        'E', 228,
        'e', 228,
        'F', 229,
        'f', 229,
        'I', 243,
        'i', 243,
        'N', 257,
        'n', 257,
        'S', 255,
        's', 255,
        'T', 248,
        't', 248,
        'U', 241,
        'u', 241,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(119);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(62);
      if (('C' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 1:
      if (lookahead == '\n') ADVANCE(262);
      END_STATE();
    case 2:
      if (lookahead == '\n') ADVANCE(262);
      if (lookahead == '\r') ADVANCE(1);
      if (lookahead != 0 &&
          lookahead != '>') ADVANCE(263);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        '\n', 185,
        '\r', 185,
        '"', 184,
        '#', 265,
        '$', 188,
        '-', 31,
        '/', 20,
        '?', 33,
        '[', 155,
        '\\', 35,
        '{', 90,
      );
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(4);
      END_STATE();
    case 4:
      ADVANCE_MAP(
        '\n', 185,
        '\r', 185,
        '#', 265,
        '$', 188,
        '-', 31,
        '/', 20,
        '?', 33,
        '[', 155,
        '\\', 36,
        '{', 90,
      );
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(4);
      END_STATE();
    case 5:
      ADVANCE_MAP(
        '!', 135,
        '"', 169,
        '#', 264,
        '$', 188,
        '%', 223,
        '&', 84,
        '\'', 171,
        '(', 97,
        ')', 98,
        '*', 220,
        '+', 126,
        ',', 86,
        '-', 129,
        '.', 217,
        '/', 221,
        '0', 117,
        ':', 92,
        ';', 82,
        '<', 202,
        '=', 30,
        '>', 207,
        '?', 103,
        '@', 136,
        '[', 155,
        '\\', 88,
        ']', 156,
        '^', 195,
        '_', 259,
        '`', 187,
        '{', 90,
        '|', 106,
        '}', 91,
        '~', 133,
        'A', 249,
        'a', 249,
        'B', 226,
        'b', 226,
        'E', 228,
        'e', 228,
        'F', 230,
        'f', 230,
        'N', 257,
        'n', 257,
        'T', 248,
        't', 248,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(119);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(5);
      if (('C' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 6:
      ADVANCE_MAP(
        '!', 134,
        '"', 169,
        '#', 264,
        '$', 188,
        '&', 83,
        '\'', 171,
        '(', 97,
        ')', 98,
        '+', 126,
        ',', 86,
        '-', 129,
        '.', 113,
        '/', 20,
        '0', 117,
        '<', 27,
        '?', 33,
        '@', 136,
        '[', 155,
        '\\', 88,
        ']', 156,
        '_', 259,
        '`', 187,
        '~', 133,
        'A', 249,
        'a', 249,
        'B', 226,
        'b', 226,
        'E', 228,
        'e', 228,
        'F', 230,
        'f', 230,
        'N', 257,
        'n', 257,
        'T', 248,
        't', 248,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(119);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(6);
      if (('C' <= lookahead && lookahead <= 'Z') ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 7:
      ADVANCE_MAP(
        '!', 134,
        '"', 169,
        '#', 264,
        '$', 188,
        '\'', 171,
        '(', 97,
        '+', 126,
        '-', 129,
        '.', 114,
        '/', 20,
        '0', 117,
        '<', 27,
        '?', 33,
        '@', 136,
        '[', 155,
        '\\', 88,
        '_', 259,
        '`', 187,
        '~', 133,
        'A', 249,
        'a', 249,
        'B', 225,
        'b', 225,
        'E', 228,
        'e', 228,
        'F', 229,
        'f', 229,
        'I', 243,
        'i', 243,
        'N', 257,
        'n', 257,
        'S', 255,
        's', 255,
        'T', 248,
        't', 248,
        'U', 241,
        'u', 241,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(119);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(7);
      if (('C' <= lookahead && lookahead <= 'Z') ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 8:
      ADVANCE_MAP(
        '!', 29,
        '"', 169,
        '#', 265,
        '$', 188,
        '%', 223,
        '&', 84,
        '(', 97,
        ')', 98,
        '*', 220,
        '+', 126,
        ',', 86,
        '-', 131,
        '.', 216,
        '/', 221,
        ':', 93,
        ';', 82,
        '<', 204,
        '=', 87,
        '>', 207,
        '?', 101,
        '[', 155,
        '\\', 35,
        ']', 156,
        '^', 195,
        '`', 187,
        '{', 90,
        '|', 106,
        '}', 91,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(9);
      if (('A' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 9:
      ADVANCE_MAP(
        '!', 29,
        '"', 169,
        '#', 265,
        '$', 188,
        '%', 223,
        '&', 84,
        '(', 97,
        ')', 98,
        '*', 220,
        '+', 126,
        ',', 86,
        '-', 131,
        '.', 216,
        '/', 221,
        ':', 93,
        ';', 82,
        '<', 204,
        '=', 87,
        '>', 207,
        '?', 101,
        '[', 155,
        '\\', 36,
        ']', 156,
        '^', 195,
        '`', 187,
        '{', 90,
        '|', 106,
        '}', 91,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(9);
      if (('A' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 10:
      ADVANCE_MAP(
        '!', 29,
        '#', 265,
        '$', 188,
        '%', 223,
        '&', 84,
        '\'', 186,
        '(', 97,
        ')', 98,
        '*', 220,
        '+', 125,
        ',', 86,
        '-', 132,
        '.', 216,
        '/', 221,
        '0', 122,
        ':', 93,
        ';', 82,
        '<', 204,
        '=', 87,
        '>', 207,
        '?', 101,
        '[', 155,
        '\\', 88,
        ']', 156,
        '^', 195,
        '{', 90,
        '|', 106,
        '}', 91,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(124);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(11);
      if (('A' <= lookahead && lookahead <= '_') ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 11:
      ADVANCE_MAP(
        '!', 29,
        '#', 265,
        '$', 188,
        '%', 223,
        '&', 84,
        '(', 97,
        ')', 98,
        '*', 220,
        '+', 125,
        ',', 86,
        '-', 132,
        '.', 216,
        '/', 221,
        '0', 122,
        ':', 93,
        ';', 82,
        '<', 204,
        '=', 87,
        '>', 207,
        '?', 101,
        '[', 155,
        '\\', 88,
        ']', 156,
        '^', 195,
        '{', 90,
        '|', 106,
        '}', 91,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(124);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(11);
      if (('A' <= lookahead && lookahead <= '_') ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 12:
      ADVANCE_MAP(
        '!', 29,
        '#', 265,
        '$', 188,
        '%', 223,
        '&', 84,
        '(', 97,
        ')', 98,
        '*', 220,
        '+', 125,
        ',', 86,
        '-', 128,
        '.', 216,
        '/', 221,
        ':', 93,
        ';', 82,
        '<', 204,
        '=', 87,
        '>', 207,
        '?', 103,
        '\\', 88,
        ']', 156,
        '^', 195,
        '{', 90,
        '|', 106,
        '}', 91,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(12);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 13:
      if (lookahead == '"') ADVANCE(168);
      if (lookahead == '\'') ADVANCE(170);
      END_STATE();
    case 14:
      ADVANCE_MAP(
        '"', 169,
        '#', 265,
        '\'', 171,
        '.', 114,
        '/', 20,
        '0', 117,
        '<', 27,
        '?', 33,
        '_', 56,
        'B', 13,
        'b', 13,
        'E', 39,
        'e', 39,
        'F', 40,
        'f', 40,
        'N', 50,
        'n', 50,
        'T', 47,
        't', 47,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(119);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(14);
      END_STATE();
    case 15:
      ADVANCE_MAP(
        '#', 264,
        '$', 188,
        '&', 83,
        '(', 97,
        ')', 98,
        ',', 86,
        '.', 23,
        '/', 20,
        '?', 102,
        '\\', 88,
        '}', 91,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(15);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 16:
      ADVANCE_MAP(
        '#', 265,
        '$', 188,
        '&', 83,
        '(', 95,
        ')', 96,
        '/', 20,
        ';', 82,
        '?', 102,
        '\\', 88,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(18);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 17:
      ADVANCE_MAP(
        '#', 265,
        '$', 188,
        '&', 83,
        '(', 97,
        ')', 98,
        '.', 23,
        '/', 20,
        '=', 32,
        '?', 102,
        '\\', 88,
        '{', 90,
        '|', 104,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(17);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 18:
      ADVANCE_MAP(
        '#', 265,
        '$', 188,
        '&', 83,
        '(', 97,
        '/', 20,
        ';', 82,
        '?', 102,
        '\\', 88,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(18);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 19:
      if (lookahead == '#') ADVANCE(181);
      if (lookahead == '\'') ADVANCE(171);
      if (lookahead == '/') ADVANCE(176);
      if (lookahead == '?') ADVANCE(179);
      if (lookahead == '\\') ADVANCE(38);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(175);
      if (lookahead != 0) ADVANCE(182);
      END_STATE();
    case 20:
      if (lookahead == '*') ADVANCE(22);
      if (lookahead == '/') ADVANCE(263);
      END_STATE();
    case 21:
      if (lookahead == '*') ADVANCE(21);
      if (lookahead == '/') ADVANCE(262);
      if (lookahead != 0) ADVANCE(22);
      END_STATE();
    case 22:
      if (lookahead == '*') ADVANCE(21);
      if (lookahead != 0) ADVANCE(22);
      END_STATE();
    case 23:
      if (lookahead == '.') ADVANCE(25);
      END_STATE();
    case 24:
      if (lookahead == '.') ADVANCE(114);
      if (lookahead == '_') ADVANCE(56);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(39);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(24);
      END_STATE();
    case 25:
      if (lookahead == '.') ADVANCE(99);
      END_STATE();
    case 26:
      if (lookahead == '<') ADVANCE(183);
      END_STATE();
    case 27:
      if (lookahead == '<') ADVANCE(26);
      END_STATE();
    case 28:
      if (lookahead == '<') ADVANCE(26);
      if (lookahead == '?') ADVANCE(69);
      END_STATE();
    case 29:
      if (lookahead == '=') ADVANCE(198);
      END_STATE();
    case 30:
      if (lookahead == '=') ADVANCE(197);
      if (lookahead == '>') ADVANCE(94);
      END_STATE();
    case 31:
      if (lookahead == '>') ADVANCE(153);
      END_STATE();
    case 32:
      if (lookahead == '>') ADVANCE(94);
      END_STATE();
    case 33:
      if (lookahead == '>') ADVANCE(70);
      END_STATE();
    case 34:
      if (lookahead == '>') ADVANCE(154);
      END_STATE();
    case 35:
      if (lookahead == 'u') ADVANCE(167);
      if (lookahead == 'x') ADVANCE(60);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(164);
      if (set_contains(sym_escape_sequence_character_set_1, 10, lookahead)) ADVANCE(162);
      END_STATE();
    case 36:
      if (lookahead == 'u') ADVANCE(166);
      END_STATE();
    case 37:
      if (lookahead == '}') ADVANCE(162);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(37);
      END_STATE();
    case 38:
      if (lookahead == '\'' ||
          lookahead == '\\') ADVANCE(172);
      if (lookahead != 0) ADVANCE(182);
      END_STATE();
    case 39:
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(57);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(116);
      END_STATE();
    case 40:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(43);
      END_STATE();
    case 41:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(159);
      END_STATE();
    case 42:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(46);
      END_STATE();
    case 43:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(48);
      END_STATE();
    case 44:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(157);
      END_STATE();
    case 45:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(44);
      END_STATE();
    case 46:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(68);
      END_STATE();
    case 47:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(49);
      END_STATE();
    case 48:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(41);
      END_STATE();
    case 49:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(41);
      END_STATE();
    case 50:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(45);
      END_STATE();
    case 51:
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(120);
      END_STATE();
    case 52:
      if (lookahead == '8' ||
          lookahead == '9') ADVANCE(24);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(118);
      END_STATE();
    case 53:
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(123);
      END_STATE();
    case 54:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(119);
      END_STATE();
    case 55:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(114);
      END_STATE();
    case 56:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(24);
      END_STATE();
    case 57:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(116);
      END_STATE();
    case 58:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(124);
      END_STATE();
    case 59:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(121);
      END_STATE();
    case 60:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(165);
      END_STATE();
    case 61:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(37);
      END_STATE();
    case 62:
      if (eof) ADVANCE(67);
      ADVANCE_MAP(
        '!', 135,
        '"', 169,
        '#', 264,
        '$', 188,
        '%', 224,
        '&', 85,
        '\'', 171,
        '(', 97,
        ')', 98,
        '*', 219,
        '+', 127,
        ',', 86,
        '-', 130,
        '.', 217,
        '/', 222,
        '0', 117,
        ':', 93,
        ';', 82,
        '<', 205,
        '=', 87,
        '>', 206,
        '?', 100,
        '@', 136,
        '[', 155,
        '\\', 89,
        ']', 156,
        '^', 196,
        '_', 259,
        '`', 187,
        '{', 90,
        '|', 105,
        '}', 91,
        '~', 133,
        'A', 249,
        'a', 249,
        'B', 225,
        'b', 225,
        'E', 228,
        'e', 228,
        'F', 229,
        'f', 229,
        'I', 243,
        'i', 243,
        'N', 257,
        'n', 257,
        'S', 255,
        's', 255,
        'T', 248,
        't', 248,
        'U', 241,
        'u', 241,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(119);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(62);
      if (('C' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 63:
      if (eof) ADVANCE(67);
      ADVANCE_MAP(
        '!', 134,
        '"', 169,
        '#', 264,
        '$', 188,
        '&', 83,
        '\'', 171,
        '(', 97,
        ')', 98,
        '+', 126,
        ',', 86,
        '-', 129,
        '.', 114,
        '/', 20,
        '0', 117,
        ':', 92,
        ';', 82,
        '<', 27,
        '=', 32,
        '?', 33,
        '@', 136,
        '[', 155,
        '\\', 88,
        ']', 156,
        '_', 259,
        '`', 187,
        '{', 90,
        '}', 91,
        '~', 133,
        'A', 249,
        'a', 249,
        'B', 226,
        'b', 226,
        'E', 228,
        'e', 228,
        'F', 230,
        'f', 230,
        'N', 257,
        'n', 257,
        'T', 248,
        't', 248,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(119);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(63);
      if (('C' <= lookahead && lookahead <= 'Z') ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 64:
      if (eof) ADVANCE(67);
      ADVANCE_MAP(
        '!', 29,
        '"', 169,
        '#', 265,
        '$', 188,
        '%', 224,
        '&', 85,
        '\'', 171,
        '(', 97,
        ')', 98,
        '*', 219,
        '+', 127,
        ',', 86,
        '-', 130,
        '.', 218,
        '/', 222,
        '0', 122,
        ':', 93,
        ';', 82,
        '<', 203,
        '=', 87,
        '>', 206,
        '?', 100,
        '[', 155,
        '\\', 88,
        ']', 156,
        '^', 196,
        '{', 90,
        '|', 105,
        '}', 91,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(124);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(64);
      if (('A' <= lookahead && lookahead <= '_') ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 65:
      if (eof) ADVANCE(67);
      ADVANCE_MAP(
        '"', 169,
        '#', 265,
        '$', 188,
        '&', 83,
        '\'', 171,
        '(', 97,
        ')', 98,
        '.', 23,
        '/', 20,
        '<', 28,
        '?', 33,
        '[', 155,
        '\\', 88,
        '|', 104,
        '}', 91,
        'A', 249,
        'a', 249,
        'B', 226,
        'b', 226,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(65);
      if (('C' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 66:
      if (eof) ADVANCE(67);
      ADVANCE_MAP(
        '#', 80,
        '/', 75,
        '<', 71,
        '?', 78,
        0xa0, 74,
        0x200b, 74,
        0x2060, 74,
        0xfeff, 74,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(66);
      if (lookahead != 0) ADVANCE(81);
      END_STATE();
    case 67:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 68:
      ACCEPT_TOKEN(sym_php_tag);
      END_STATE();
    case 69:
      ACCEPT_TOKEN(sym_php_tag);
      if (lookahead == '=') ADVANCE(68);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(42);
      END_STATE();
    case 70:
      ACCEPT_TOKEN(anon_sym_QMARK_GT);
      END_STATE();
    case 71:
      ACCEPT_TOKEN(aux_sym_text_token1);
      if (lookahead == '?') ADVANCE(69);
      END_STATE();
    case 72:
      ACCEPT_TOKEN(aux_sym_text_token2);
      if (lookahead == '\n') ADVANCE(81);
      if (lookahead == '\r') ADVANCE(73);
      if (lookahead == '>') ADVANCE(81);
      if (lookahead != 0 &&
          lookahead != '<') ADVANCE(79);
      END_STATE();
    case 73:
      ACCEPT_TOKEN(aux_sym_text_token2);
      if (lookahead == '\n') ADVANCE(81);
      if (lookahead != 0 &&
          lookahead != '<') ADVANCE(81);
      END_STATE();
    case 74:
      ACCEPT_TOKEN(aux_sym_text_token2);
      if (lookahead == '#') ADVANCE(80);
      if (lookahead == '/') ADVANCE(75);
      if (lookahead == '?') ADVANCE(78);
      if (lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(74);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(74);
      if (lookahead != 0 &&
          lookahead != '<') ADVANCE(81);
      END_STATE();
    case 75:
      ACCEPT_TOKEN(aux_sym_text_token2);
      if (lookahead == '*') ADVANCE(77);
      if (lookahead == '/') ADVANCE(79);
      if (lookahead != 0 &&
          lookahead != '<') ADVANCE(81);
      END_STATE();
    case 76:
      ACCEPT_TOKEN(aux_sym_text_token2);
      if (lookahead == '*') ADVANCE(76);
      if (lookahead == '/') ADVANCE(81);
      if (lookahead != 0 &&
          lookahead != '<') ADVANCE(77);
      END_STATE();
    case 77:
      ACCEPT_TOKEN(aux_sym_text_token2);
      if (lookahead == '*') ADVANCE(76);
      if (lookahead != 0 &&
          lookahead != '<') ADVANCE(77);
      END_STATE();
    case 78:
      ACCEPT_TOKEN(aux_sym_text_token2);
      if (lookahead == '>') ADVANCE(81);
      if (lookahead != 0 &&
          lookahead != '<') ADVANCE(81);
      END_STATE();
    case 79:
      ACCEPT_TOKEN(aux_sym_text_token2);
      if (lookahead == '?') ADVANCE(72);
      if (lookahead == '\n' ||
          lookahead == '\r') ADVANCE(81);
      if (lookahead != 0 &&
          lookahead != '<') ADVANCE(79);
      END_STATE();
    case 80:
      ACCEPT_TOKEN(aux_sym_text_token2);
      if (lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == '?' ||
          lookahead == '[') ADVANCE(81);
      if (lookahead != 0 &&
          lookahead != '<') ADVANCE(79);
      END_STATE();
    case 81:
      ACCEPT_TOKEN(aux_sym_text_token2);
      if (lookahead != 0 &&
          lookahead != '<') ADVANCE(81);
      END_STATE();
    case 82:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 83:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 84:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(194);
      END_STATE();
    case 85:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(194);
      if (lookahead == '=') ADVANCE(149);
      END_STATE();
    case 86:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 87:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(197);
      if (lookahead == '>') ADVANCE(94);
      END_STATE();
    case 88:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      END_STATE();
    case 89:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      if (lookahead == '\'' ||
          lookahead == '\\') ADVANCE(172);
      END_STATE();
    case 90:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 91:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 92:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 93:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(137);
      END_STATE();
    case 94:
      ACCEPT_TOKEN(anon_sym_EQ_GT);
      END_STATE();
    case 95:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 96:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 97:
      ACCEPT_TOKEN(anon_sym_LPAREN2);
      END_STATE();
    case 98:
      ACCEPT_TOKEN(anon_sym_RPAREN2);
      END_STATE();
    case 99:
      ACCEPT_TOKEN(anon_sym_DOT_DOT_DOT);
      END_STATE();
    case 100:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '-') ADVANCE(34);
      if (lookahead == '>') ADVANCE(70);
      if (lookahead == '?') ADVANCE(190);
      END_STATE();
    case 101:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '-') ADVANCE(34);
      if (lookahead == '>') ADVANCE(70);
      if (lookahead == '?') ADVANCE(189);
      END_STATE();
    case 102:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '>') ADVANCE(70);
      END_STATE();
    case 103:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '>') ADVANCE(70);
      if (lookahead == '?') ADVANCE(189);
      END_STATE();
    case 104:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    case 105:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '=') ADVANCE(151);
      if (lookahead == '|') ADVANCE(193);
      END_STATE();
    case 106:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '|') ADVANCE(193);
      END_STATE();
    case 107:
      ACCEPT_TOKEN(aux_sym_cast_type_token1);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 108:
      ACCEPT_TOKEN(aux_sym_cast_type_token3);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 109:
      ACCEPT_TOKEN(aux_sym_cast_type_token6);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 110:
      ACCEPT_TOKEN(aux_sym_cast_type_token8);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 111:
      ACCEPT_TOKEN(aux_sym_cast_type_token11);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 112:
      ACCEPT_TOKEN(aux_sym_cast_type_token12);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 113:
      ACCEPT_TOKEN(sym_float);
      if (lookahead == '.') ADVANCE(25);
      if (lookahead == '_') ADVANCE(55);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(39);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(114);
      END_STATE();
    case 114:
      ACCEPT_TOKEN(sym_float);
      if (lookahead == '_') ADVANCE(55);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(39);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(114);
      END_STATE();
    case 115:
      ACCEPT_TOKEN(sym_float);
      if (lookahead == '_') ADVANCE(260);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(115);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0x9f) ||
          (0xa1 <= lookahead && lookahead <= 0x200a) ||
          (0x200c <= lookahead && lookahead <= 0x205f) ||
          (0x2061 <= lookahead && lookahead <= 0xfefe) ||
          (0xff00 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 116:
      ACCEPT_TOKEN(sym_float);
      if (lookahead == '_') ADVANCE(57);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(116);
      END_STATE();
    case 117:
      ACCEPT_TOKEN(sym_integer);
      ADVANCE_MAP(
        '.', 114,
        '_', 52,
        'B', 51,
        'b', 51,
        'E', 39,
        'e', 39,
        'O', 123,
        'o', 123,
        'X', 59,
        'x', 59,
        '8', 24,
        '9', 24,
      );
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(118);
      END_STATE();
    case 118:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '.') ADVANCE(114);
      if (lookahead == '_') ADVANCE(52);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(39);
      if (lookahead == '8' ||
          lookahead == '9') ADVANCE(24);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(118);
      END_STATE();
    case 119:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '.') ADVANCE(114);
      if (lookahead == '_') ADVANCE(54);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(39);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(119);
      END_STATE();
    case 120:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '_') ADVANCE(51);
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(120);
      END_STATE();
    case 121:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '_') ADVANCE(59);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(121);
      END_STATE();
    case 122:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '_') ADVANCE(53);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(51);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(123);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(59);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(123);
      END_STATE();
    case 123:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '_') ADVANCE(53);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(123);
      END_STATE();
    case 124:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '_') ADVANCE(58);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(124);
      END_STATE();
    case 125:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 126:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(139);
      END_STATE();
    case 127:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(139);
      if (lookahead == '=') ADVANCE(144);
      END_STATE();
    case 128:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 129:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(138);
      END_STATE();
    case 130:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(138);
      if (lookahead == '=') ADVANCE(145);
      if (lookahead == '>') ADVANCE(153);
      END_STATE();
    case 131:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(138);
      if (lookahead == '>') ADVANCE(153);
      END_STATE();
    case 132:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '>') ADVANCE(153);
      END_STATE();
    case 133:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 134:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 135:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(198);
      END_STATE();
    case 136:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 137:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      END_STATE();
    case 138:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      END_STATE();
    case 139:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS);
      END_STATE();
    case 140:
      ACCEPT_TOKEN(anon_sym_STAR_STAR_EQ);
      END_STATE();
    case 141:
      ACCEPT_TOKEN(anon_sym_STAR_EQ);
      END_STATE();
    case 142:
      ACCEPT_TOKEN(anon_sym_SLASH_EQ);
      END_STATE();
    case 143:
      ACCEPT_TOKEN(anon_sym_PERCENT_EQ);
      END_STATE();
    case 144:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      END_STATE();
    case 145:
      ACCEPT_TOKEN(anon_sym_DASH_EQ);
      END_STATE();
    case 146:
      ACCEPT_TOKEN(anon_sym_DOT_EQ);
      END_STATE();
    case 147:
      ACCEPT_TOKEN(anon_sym_LT_LT_EQ);
      END_STATE();
    case 148:
      ACCEPT_TOKEN(anon_sym_GT_GT_EQ);
      END_STATE();
    case 149:
      ACCEPT_TOKEN(anon_sym_AMP_EQ);
      END_STATE();
    case 150:
      ACCEPT_TOKEN(anon_sym_CARET_EQ);
      END_STATE();
    case 151:
      ACCEPT_TOKEN(anon_sym_PIPE_EQ);
      END_STATE();
    case 152:
      ACCEPT_TOKEN(anon_sym_QMARK_QMARK_EQ);
      END_STATE();
    case 153:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      END_STATE();
    case 154:
      ACCEPT_TOKEN(anon_sym_QMARK_DASH_GT);
      END_STATE();
    case 155:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 157:
      ACCEPT_TOKEN(aux_sym__argument_name_token1);
      END_STATE();
    case 158:
      ACCEPT_TOKEN(aux_sym__argument_name_token1);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 159:
      ACCEPT_TOKEN(aux_sym__argument_name_token2);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(aux_sym__argument_name_token2);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(anon_sym_POUND_LBRACK);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(sym_escape_sequence);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(162);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(163);
      END_STATE();
    case 165:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(162);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(anon_sym_BSLASHu);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(anon_sym_BSLASHu);
      if (lookahead == '{') ADVANCE(61);
      END_STATE();
    case 168:
      ACCEPT_TOKEN(aux_sym_encapsed_string_token1);
      END_STATE();
    case 169:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 170:
      ACCEPT_TOKEN(aux_sym_string_token1);
      END_STATE();
    case 171:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(aux_sym_string_token2);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '\n') ADVANCE(182);
      if (lookahead == '\r') ADVANCE(174);
      if (lookahead == '>') ADVANCE(182);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(180);
      END_STATE();
    case 174:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '\n') ADVANCE(182);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(182);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '#') ADVANCE(181);
      if (lookahead == '/') ADVANCE(176);
      if (lookahead == '?') ADVANCE(179);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(175);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(182);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '*') ADVANCE(178);
      if (lookahead == '/') ADVANCE(180);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(182);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '*') ADVANCE(177);
      if (lookahead == '/') ADVANCE(182);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(178);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '*') ADVANCE(177);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(178);
      END_STATE();
    case 179:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '>') ADVANCE(182);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(182);
      END_STATE();
    case 180:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '?') ADVANCE(173);
      if (lookahead == '\n' ||
          lookahead == '\r') ADVANCE(182);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(180);
      END_STATE();
    case 181:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == '?' ||
          lookahead == '[') ADVANCE(182);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '[' &&
          lookahead != '\\') ADVANCE(180);
      END_STATE();
    case 182:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(182);
      END_STATE();
    case 183:
      ACCEPT_TOKEN(anon_sym_LT_LT_LT);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(anon_sym_DQUOTE2);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(sym__new_line);
      if (lookahead == '\n') ADVANCE(185);
      if (lookahead == '\r') ADVANCE(185);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(anon_sym_SQUOTE2);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(anon_sym_BQUOTE);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_QMARK_QMARK);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(anon_sym_QMARK_QMARK);
      if (lookahead == '=') ADVANCE(152);
      END_STATE();
    case 191:
      ACCEPT_TOKEN(anon_sym_STAR_STAR);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(anon_sym_STAR_STAR);
      if (lookahead == '=') ADVANCE(140);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 195:
      ACCEPT_TOKEN(anon_sym_CARET);
      END_STATE();
    case 196:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '=') ADVANCE(150);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      if (lookahead == '=') ADVANCE(200);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      if (lookahead == '=') ADVANCE(201);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_LT_GT);
      END_STATE();
    case 200:
      ACCEPT_TOKEN(anon_sym_EQ_EQ_EQ);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(anon_sym_BANG_EQ_EQ);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(212);
      if (lookahead == '=') ADVANCE(208);
      if (lookahead == '>') ADVANCE(199);
      END_STATE();
    case 203:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(213);
      if (lookahead == '=') ADVANCE(208);
      if (lookahead == '>') ADVANCE(199);
      END_STATE();
    case 204:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(211);
      if (lookahead == '=') ADVANCE(208);
      if (lookahead == '>') ADVANCE(199);
      END_STATE();
    case 205:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '?') ADVANCE(69);
      END_STATE();
    case 206:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(209);
      if (lookahead == '>') ADVANCE(215);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(209);
      if (lookahead == '>') ADVANCE(214);
      END_STATE();
    case 208:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      if (lookahead == '>') ADVANCE(210);
      END_STATE();
    case 209:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      END_STATE();
    case 210:
      ACCEPT_TOKEN(anon_sym_LT_EQ_GT);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '<') ADVANCE(183);
      END_STATE();
    case 213:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '=') ADVANCE(147);
      END_STATE();
    case 214:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      END_STATE();
    case 215:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '=') ADVANCE(148);
      END_STATE();
    case 216:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 217:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(25);
      if (lookahead == '_') ADVANCE(55);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(39);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(114);
      END_STATE();
    case 218:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '=') ADVANCE(146);
      END_STATE();
    case 219:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(192);
      if (lookahead == '=') ADVANCE(141);
      END_STATE();
    case 220:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(191);
      END_STATE();
    case 221:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '*') ADVANCE(22);
      if (lookahead == '/') ADVANCE(263);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '*') ADVANCE(22);
      if (lookahead == '/') ADVANCE(263);
      if (lookahead == '=') ADVANCE(142);
      END_STATE();
    case 223:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 224:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '=') ADVANCE(143);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == '"') ADVANCE(168);
      if (lookahead == '\'') ADVANCE(170);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(245);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 226:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == '"') ADVANCE(168);
      if (lookahead == '\'') ADVANCE(170);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 227:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == '.') ADVANCE(114);
      if (lookahead == '_') ADVANCE(259);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(228);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(227);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0x9f) ||
          (0xa1 <= lookahead && lookahead <= 0x200a) ||
          (0x200c <= lookahead && lookahead <= 0x205f) ||
          (0x2061 <= lookahead && lookahead <= 0xfefe) ||
          (0xff00 <= lookahead && lookahead <= 0xffff)) ADVANCE(261);
      END_STATE();
    case 228:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(57);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(115);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 229:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(240);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(244);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 230:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(240);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 231:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(258);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 232:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(253);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(160);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(254);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 235:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(111);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 236:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(242);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(108);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 238:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(158);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(238);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 240:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(250);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(251);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(235);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 243:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(252);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 244:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(232);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 245:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(237);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 246:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(231);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 247:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(236);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 248:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(256);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 249:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(246);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(233);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 251:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(234);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 252:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(109);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 253:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(110);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 254:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(112);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 255:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(247);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 256:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(233);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 257:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(239);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(107);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 259:
      ACCEPT_TOKEN(sym_name);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(227);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(sym_name);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(115);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 261:
      ACCEPT_TOKEN(sym_name);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(261);
      END_STATE();
    case 262:
      ACCEPT_TOKEN(sym_comment);
      END_STATE();
    case 263:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '?') ADVANCE(2);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(263);
      END_STATE();
    case 264:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '[') ADVANCE(161);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '?') ADVANCE(263);
      END_STATE();
    case 265:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '?' &&
          lookahead != '[') ADVANCE(263);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      ADVANCE_MAP(
        'A', 1,
        'B', 2,
        'E', 3,
        'F', 4,
        'I', 5,
        'M', 6,
        'N', 7,
        'P', 8,
        'S', 9,
        'T', 10,
        'U', 11,
        'V', 12,
        'a', 13,
        'b', 14,
        'e', 15,
        'f', 16,
        'i', 17,
        'm', 18,
        'n', 19,
        'p', 20,
        's', 21,
        't', 22,
        'u', 23,
        'v', 24,
        'C', 25,
        'c', 25,
        'D', 26,
        'd', 26,
        'G', 27,
        'g', 27,
        'L', 28,
        'l', 28,
        'O', 29,
        'o', 29,
        'R', 30,
        'r', 30,
        'W', 31,
        'w', 31,
        'X', 32,
        'x', 32,
        'Y', 33,
        'y', 33,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(0);
      END_STATE();
    case 1:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(34);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(35);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(36);
      END_STATE();
    case 2:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(37);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(38);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(39);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        'C', 40,
        'c', 40,
        'L', 41,
        'l', 41,
        'N', 42,
        'n', 42,
        'X', 43,
        'x', 43,
      );
      END_STATE();
    case 4:
      ADVANCE_MAP(
        'I', 44,
        'i', 44,
        'N', 45,
        'n', 45,
        'O', 46,
        'o', 46,
        'R', 47,
        'r', 47,
        'U', 48,
        'u', 48,
      );
      END_STATE();
    case 5:
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(49);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(50);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(51);
      END_STATE();
    case 6:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(52);
      END_STATE();
    case 7:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(53);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(54);
      END_STATE();
    case 8:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(55);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(56);
      END_STATE();
    case 9:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(57);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(58);
      END_STATE();
    case 10:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(59);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(60);
      END_STATE();
    case 11:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(61);
      END_STATE();
    case 12:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(62);
      END_STATE();
    case 13:
      if (lookahead == 'r') ADVANCE(63);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(34);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(35);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(36);
      END_STATE();
    case 14:
      if (lookahead == 'O') ADVANCE(38);
      if (lookahead == 'o') ADVANCE(64);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(37);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(39);
      END_STATE();
    case 15:
      ADVANCE_MAP(
        'N', 42,
        'n', 65,
        'C', 40,
        'c', 40,
        'L', 41,
        'l', 41,
        'X', 43,
        'x', 43,
      );
      END_STATE();
    case 16:
      ADVANCE_MAP(
        'a', 66,
        'l', 67,
        'I', 44,
        'i', 44,
        'N', 45,
        'n', 45,
        'O', 46,
        'o', 46,
        'R', 47,
        'r', 47,
        'U', 48,
        'u', 48,
      );
      END_STATE();
    case 17:
      if (lookahead == 'N') ADVANCE(51);
      if (lookahead == 'n') ADVANCE(68);
      if (lookahead == 't') ADVANCE(69);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(49);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(50);
      END_STATE();
    case 18:
      if (lookahead == 'i') ADVANCE(70);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(52);
      END_STATE();
    case 19:
      if (lookahead == 'E') ADVANCE(54);
      if (lookahead == 'e') ADVANCE(71);
      if (lookahead == 'u') ADVANCE(72);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(53);
      END_STATE();
    case 20:
      if (lookahead == 'a') ADVANCE(73);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(55);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(56);
      END_STATE();
    case 21:
      if (lookahead == 'T') ADVANCE(57);
      if (lookahead == 'e') ADVANCE(74);
      if (lookahead == 't') ADVANCE(75);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(58);
      END_STATE();
    case 22:
      if (lookahead == 'R') ADVANCE(60);
      if (lookahead == 'i') ADVANCE(76);
      if (lookahead == 'r') ADVANCE(77);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(59);
      END_STATE();
    case 23:
      if (lookahead == 'n') ADVANCE(78);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(61);
      END_STATE();
    case 24:
      if (lookahead == 'o') ADVANCE(79);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(62);
      END_STATE();
    case 25:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(80);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(81);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(82);
      END_STATE();
    case 26:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(83);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(84);
      END_STATE();
    case 27:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(85);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(86);
      END_STATE();
    case 28:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(87);
      END_STATE();
    case 29:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(88);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(89);
      END_STATE();
    case 30:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(90);
      END_STATE();
    case 31:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(91);
      END_STATE();
    case 32:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(92);
      END_STATE();
    case 33:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(93);
      END_STATE();
    case 34:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(94);
      END_STATE();
    case 35:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(95);
      END_STATE();
    case 36:
      ACCEPT_TOKEN(aux_sym_namespace_use_clause_token1);
      END_STATE();
    case 37:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(96);
      END_STATE();
    case 38:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(97);
      END_STATE();
    case 39:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(98);
      END_STATE();
    case 40:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(99);
      END_STATE();
    case 41:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(100);
      END_STATE();
    case 42:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(101);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(102);
      END_STATE();
    case 43:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(103);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(104);
      END_STATE();
    case 44:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(105);
      END_STATE();
    case 45:
      ACCEPT_TOKEN(aux_sym__arrow_function_header_token1);
      END_STATE();
    case 46:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(106);
      END_STATE();
    case 47:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(107);
      END_STATE();
    case 48:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(108);
      END_STATE();
    case 49:
      ACCEPT_TOKEN(aux_sym_if_statement_token1);
      END_STATE();
    case 50:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(109);
      END_STATE();
    case 51:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(110);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(111);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(112);
      END_STATE();
    case 52:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(113);
      END_STATE();
    case 53:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(114);
      END_STATE();
    case 54:
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(115);
      END_STATE();
    case 55:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(116);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(117);
      END_STATE();
    case 56:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(118);
      END_STATE();
    case 57:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(119);
      END_STATE();
    case 58:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(120);
      END_STATE();
    case 59:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(121);
      END_STATE();
    case 60:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(122);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(123);
      END_STATE();
    case 61:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(124);
      END_STATE();
    case 62:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(125);
      END_STATE();
    case 63:
      if (lookahead == 'r') ADVANCE(126);
      END_STATE();
    case 64:
      if (lookahead == 'O') ADVANCE(97);
      if (lookahead == 'o') ADVANCE(127);
      END_STATE();
    case 65:
      if (lookahead == 'c') ADVANCE(128);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(101);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(102);
      END_STATE();
    case 66:
      if (lookahead == 'l') ADVANCE(129);
      END_STATE();
    case 67:
      if (lookahead == 'o') ADVANCE(130);
      END_STATE();
    case 68:
      if (lookahead == 'T') ADVANCE(112);
      if (lookahead == 't') ADVANCE(131);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(110);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(111);
      END_STATE();
    case 69:
      if (lookahead == 'e') ADVANCE(132);
      END_STATE();
    case 70:
      if (lookahead == 'x') ADVANCE(133);
      END_STATE();
    case 71:
      if (lookahead == 'v') ADVANCE(134);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(115);
      END_STATE();
    case 72:
      if (lookahead == 'l') ADVANCE(135);
      END_STATE();
    case 73:
      if (lookahead == 'r') ADVANCE(136);
      END_STATE();
    case 74:
      if (lookahead == 'l') ADVANCE(137);
      END_STATE();
    case 75:
      if (lookahead == 'r') ADVANCE(138);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(119);
      END_STATE();
    case 76:
      if (lookahead == 'c') ADVANCE(139);
      END_STATE();
    case 77:
      if (lookahead == 'u') ADVANCE(140);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(122);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(123);
      END_STATE();
    case 78:
      if (lookahead == 's') ADVANCE(141);
      END_STATE();
    case 79:
      if (lookahead == 'i') ADVANCE(142);
      END_STATE();
    case 80:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(143);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(144);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(145);
      END_STATE();
    case 81:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(146);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(147);
      END_STATE();
    case 82:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(148);
      END_STATE();
    case 83:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(149);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(150);
      END_STATE();
    case 84:
      ACCEPT_TOKEN(aux_sym_do_statement_token1);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(151);
      END_STATE();
    case 85:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(152);
      END_STATE();
    case 86:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(153);
      END_STATE();
    case 87:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(154);
      END_STATE();
    case 88:
      if (lookahead == 'J' ||
          lookahead == 'j') ADVANCE(155);
      END_STATE();
    case 89:
      ACCEPT_TOKEN(aux_sym_binary_expression_token3);
      END_STATE();
    case 90:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(156);
      if (lookahead == 'Q' ||
          lookahead == 'q') ADVANCE(157);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(158);
      END_STATE();
    case 91:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(159);
      END_STATE();
    case 92:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(160);
      END_STATE();
    case 93:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(161);
      END_STATE();
    case 94:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(162);
      END_STATE();
    case 95:
      ACCEPT_TOKEN(aux_sym_binary_expression_token2);
      END_STATE();
    case 96:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(163);
      END_STATE();
    case 97:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(164);
      END_STATE();
    case 98:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(165);
      END_STATE();
    case 99:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(166);
      END_STATE();
    case 100:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(167);
      END_STATE();
    case 101:
      ADVANCE_MAP(
        'D', 168,
        'd', 168,
        'F', 169,
        'f', 169,
        'I', 170,
        'i', 170,
        'S', 171,
        's', 171,
        'W', 172,
        'w', 172,
      );
      END_STATE();
    case 102:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(173);
      END_STATE();
    case 103:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(174);
      END_STATE();
    case 104:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(175);
      END_STATE();
    case 105:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(176);
      END_STATE();
    case 106:
      ACCEPT_TOKEN(aux_sym_for_statement_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(177);
      END_STATE();
    case 107:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(178);
      END_STATE();
    case 108:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(179);
      END_STATE();
    case 109:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(180);
      END_STATE();
    case 110:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(181);
      END_STATE();
    case 111:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(182);
      END_STATE();
    case 112:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(183);
      END_STATE();
    case 113:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(184);
      END_STATE();
    case 114:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(185);
      END_STATE();
    case 115:
      ACCEPT_TOKEN(aux_sym__new_non_dereferencable_expression_token1);
      END_STATE();
    case 116:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(186);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(187);
      END_STATE();
    case 117:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(188);
      END_STATE();
    case 118:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(189);
      END_STATE();
    case 119:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(190);
      END_STATE();
    case 120:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(191);
      END_STATE();
    case 121:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(192);
      END_STATE();
    case 122:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(193);
      END_STATE();
    case 123:
      ACCEPT_TOKEN(aux_sym_try_statement_token1);
      END_STATE();
    case 124:
      ACCEPT_TOKEN(aux_sym_namespace_use_declaration_token1);
      END_STATE();
    case 125:
      ACCEPT_TOKEN(sym_var_modifier);
      END_STATE();
    case 126:
      if (lookahead == 'a') ADVANCE(194);
      END_STATE();
    case 127:
      if (lookahead == 'L') ADVANCE(164);
      if (lookahead == 'l') ADVANCE(195);
      END_STATE();
    case 128:
      if (lookahead == 'o') ADVANCE(196);
      END_STATE();
    case 129:
      if (lookahead == 's') ADVANCE(197);
      END_STATE();
    case 130:
      if (lookahead == 'a') ADVANCE(198);
      END_STATE();
    case 131:
      ACCEPT_TOKEN(anon_sym_int);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(183);
      END_STATE();
    case 132:
      if (lookahead == 'r') ADVANCE(199);
      END_STATE();
    case 133:
      if (lookahead == 'e') ADVANCE(200);
      END_STATE();
    case 134:
      if (lookahead == 'e') ADVANCE(201);
      END_STATE();
    case 135:
      if (lookahead == 'l') ADVANCE(202);
      END_STATE();
    case 136:
      if (lookahead == 'e') ADVANCE(203);
      END_STATE();
    case 137:
      if (lookahead == 'f') ADVANCE(204);
      END_STATE();
    case 138:
      if (lookahead == 'i') ADVANCE(205);
      END_STATE();
    case 139:
      if (lookahead == 'k') ADVANCE(206);
      END_STATE();
    case 140:
      if (lookahead == 'e') ADVANCE(207);
      END_STATE();
    case 141:
      if (lookahead == 'e') ADVANCE(208);
      END_STATE();
    case 142:
      if (lookahead == 'd') ADVANCE(209);
      END_STATE();
    case 143:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(210);
      END_STATE();
    case 144:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(211);
      END_STATE();
    case 145:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(212);
      END_STATE();
    case 146:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(213);
      END_STATE();
    case 147:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(214);
      END_STATE();
    case 148:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(215);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(216);
      END_STATE();
    case 149:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(217);
      END_STATE();
    case 150:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(218);
      END_STATE();
    case 151:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(219);
      END_STATE();
    case 152:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(220);
      END_STATE();
    case 153:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(221);
      END_STATE();
    case 154:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(222);
      END_STATE();
    case 155:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(223);
      END_STATE();
    case 156:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(224);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(225);
      END_STATE();
    case 157:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(226);
      END_STATE();
    case 158:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(227);
      END_STATE();
    case 159:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(228);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(aux_sym_binary_expression_token4);
      END_STATE();
    case 161:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(229);
      END_STATE();
    case 162:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(230);
      END_STATE();
    case 163:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(231);
      END_STATE();
    case 164:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(232);
      END_STATE();
    case 165:
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(233);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(aux_sym_echo_statement_token1);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(aux_sym_else_clause_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(234);
      END_STATE();
    case 168:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(235);
      END_STATE();
    case 169:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(236);
      END_STATE();
    case 170:
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(237);
      END_STATE();
    case 171:
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(238);
      END_STATE();
    case 172:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(239);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(aux_sym_enum_declaration_token1);
      END_STATE();
    case 174:
      ACCEPT_TOKEN(aux_sym_exit_statement_token1);
      END_STATE();
    case 175:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(240);
      END_STATE();
    case 176:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(241);
      END_STATE();
    case 177:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(242);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(aux_sym_yield_expression_token2);
      END_STATE();
    case 179:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(243);
      END_STATE();
    case 180:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(244);
      END_STATE();
    case 181:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(245);
      END_STATE();
    case 182:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(246);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(247);
      END_STATE();
    case 183:
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(248);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(249);
      END_STATE();
    case 184:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(250);
      END_STATE();
    case 185:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(251);
      END_STATE();
    case 186:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(252);
      END_STATE();
    case 187:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(253);
      END_STATE();
    case 188:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(254);
      END_STATE();
    case 189:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(255);
      END_STATE();
    case 190:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(256);
      END_STATE();
    case 191:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(257);
      END_STATE();
    case 192:
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(258);
      END_STATE();
    case 193:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(259);
      END_STATE();
    case 194:
      if (lookahead == 'y') ADVANCE(260);
      END_STATE();
    case 195:
      ACCEPT_TOKEN(anon_sym_bool);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(232);
      END_STATE();
    case 196:
      if (lookahead == 'd') ADVANCE(261);
      END_STATE();
    case 197:
      if (lookahead == 'e') ADVANCE(262);
      END_STATE();
    case 198:
      if (lookahead == 't') ADVANCE(263);
      END_STATE();
    case 199:
      if (lookahead == 'a') ADVANCE(264);
      END_STATE();
    case 200:
      if (lookahead == 'd') ADVANCE(265);
      END_STATE();
    case 201:
      if (lookahead == 'r') ADVANCE(266);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(anon_sym_null);
      END_STATE();
    case 203:
      if (lookahead == 'n') ADVANCE(267);
      END_STATE();
    case 204:
      ACCEPT_TOKEN(anon_sym_self);
      END_STATE();
    case 205:
      if (lookahead == 'c') ADVANCE(268);
      if (lookahead == 'n') ADVANCE(269);
      END_STATE();
    case 206:
      if (lookahead == 's') ADVANCE(270);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_true);
      END_STATE();
    case 208:
      if (lookahead == 't') ADVANCE(271);
      END_STATE();
    case 209:
      ACCEPT_TOKEN(anon_sym_void);
      END_STATE();
    case 210:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(272);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(aux_sym_enum_case_token1);
      END_STATE();
    case 212:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(273);
      END_STATE();
    case 213:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(274);
      END_STATE();
    case 214:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(275);
      END_STATE();
    case 215:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(276);
      END_STATE();
    case 216:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(277);
      END_STATE();
    case 217:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(278);
      END_STATE();
    case 218:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(279);
      END_STATE();
    case 219:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(280);
      END_STATE();
    case 220:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(281);
      END_STATE();
    case 221:
      ACCEPT_TOKEN(aux_sym_goto_statement_token1);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(aux_sym__list_destructing_token1);
      END_STATE();
    case 223:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(282);
      END_STATE();
    case 224:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(283);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(aux_sym_cast_type_token10);
      END_STATE();
    case 226:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(284);
      END_STATE();
    case 227:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(285);
      END_STATE();
    case 228:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(286);
      END_STATE();
    case 229:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(287);
      END_STATE();
    case 230:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(288);
      END_STATE();
    case 231:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(289);
      END_STATE();
    case 232:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(290);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(aux_sym_break_statement_token1);
      END_STATE();
    case 234:
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(291);
      END_STATE();
    case 235:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(292);
      END_STATE();
    case 236:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(293);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(aux_sym_if_statement_token2);
      END_STATE();
    case 238:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(294);
      END_STATE();
    case 239:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(295);
      END_STATE();
    case 240:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(296);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(aux_sym_final_modifier_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(297);
      END_STATE();
    case 242:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(298);
      END_STATE();
    case 243:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(299);
      END_STATE();
    case 244:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(300);
      END_STATE();
    case 245:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(301);
      END_STATE();
    case 246:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(302);
      END_STATE();
    case 247:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(303);
      END_STATE();
    case 248:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(304);
      END_STATE();
    case 249:
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(305);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(aux_sym_match_expression_token1);
      END_STATE();
    case 251:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(306);
      END_STATE();
    case 252:
      ACCEPT_TOKEN(aux_sym_print_intrinsic_token1);
      END_STATE();
    case 253:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(307);
      END_STATE();
    case 254:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(308);
      END_STATE();
    case 255:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(309);
      END_STATE();
    case 256:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(310);
      END_STATE();
    case 257:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(311);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(aux_sym_throw_expression_token1);
      END_STATE();
    case 259:
      ACCEPT_TOKEN(aux_sym_trait_declaration_token1);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(anon_sym_array);
      END_STATE();
    case 261:
      if (lookahead == 'i') ADVANCE(312);
      END_STATE();
    case 262:
      ACCEPT_TOKEN(anon_sym_false);
      END_STATE();
    case 263:
      ACCEPT_TOKEN(anon_sym_float);
      END_STATE();
    case 264:
      if (lookahead == 'b') ADVANCE(313);
      END_STATE();
    case 265:
      ACCEPT_TOKEN(anon_sym_mixed);
      END_STATE();
    case 266:
      ACCEPT_TOKEN(sym_bottom_type);
      END_STATE();
    case 267:
      if (lookahead == 't') ADVANCE(314);
      END_STATE();
    case 268:
      if (lookahead == 't') ADVANCE(315);
      END_STATE();
    case 269:
      if (lookahead == 'g') ADVANCE(316);
      END_STATE();
    case 270:
      ACCEPT_TOKEN(anon_sym_ticks);
      END_STATE();
    case 271:
      ACCEPT_TOKEN(anon_sym_unset);
      END_STATE();
    case 272:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(317);
      END_STATE();
    case 273:
      ACCEPT_TOKEN(aux_sym_catch_clause_token1);
      END_STATE();
    case 274:
      ACCEPT_TOKEN(aux_sym_class_declaration_token1);
      END_STATE();
    case 275:
      ACCEPT_TOKEN(aux_sym_clone_expression_token1);
      END_STATE();
    case 276:
      ACCEPT_TOKEN(aux_sym__namespace_use_type_token2);
      END_STATE();
    case 277:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(318);
      END_STATE();
    case 278:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(319);
      END_STATE();
    case 279:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(320);
      END_STATE();
    case 280:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(321);
      END_STATE();
    case 281:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(322);
      END_STATE();
    case 282:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(323);
      END_STATE();
    case 283:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(324);
      END_STATE();
    case 284:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(325);
      END_STATE();
    case 285:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(326);
      END_STATE();
    case 286:
      ACCEPT_TOKEN(aux_sym_while_statement_token1);
      END_STATE();
    case 287:
      ACCEPT_TOKEN(aux_sym_yield_expression_token1);
      END_STATE();
    case 288:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(327);
      END_STATE();
    case 289:
      ACCEPT_TOKEN(aux_sym_cast_type_token2);
      END_STATE();
    case 290:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(328);
      END_STATE();
    case 291:
      ACCEPT_TOKEN(aux_sym_else_if_clause_token1);
      END_STATE();
    case 292:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(329);
      END_STATE();
    case 293:
      ACCEPT_TOKEN(aux_sym_for_statement_token2);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(330);
      END_STATE();
    case 294:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(331);
      END_STATE();
    case 295:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(332);
      END_STATE();
    case 296:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(333);
      END_STATE();
    case 297:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(334);
      END_STATE();
    case 298:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(335);
      END_STATE();
    case 299:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(336);
      END_STATE();
    case 300:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(337);
      END_STATE();
    case 301:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(338);
      END_STATE();
    case 302:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(339);
      END_STATE();
    case 303:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(340);
      END_STATE();
    case 304:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(341);
      END_STATE();
    case 305:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(342);
      END_STATE();
    case 306:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(343);
      END_STATE();
    case 307:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(344);
      END_STATE();
    case 308:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(345);
      END_STATE();
    case 309:
      ACCEPT_TOKEN(aux_sym_visibility_modifier_token1);
      END_STATE();
    case 310:
      ACCEPT_TOKEN(aux_sym_function_static_declaration_token1);
      END_STATE();
    case 311:
      ACCEPT_TOKEN(aux_sym_switch_statement_token1);
      END_STATE();
    case 312:
      if (lookahead == 'n') ADVANCE(346);
      END_STATE();
    case 313:
      if (lookahead == 'l') ADVANCE(347);
      END_STATE();
    case 314:
      ACCEPT_TOKEN(anon_sym_parent);
      END_STATE();
    case 315:
      if (lookahead == '_') ADVANCE(348);
      END_STATE();
    case 316:
      ACCEPT_TOKEN(anon_sym_string);
      END_STATE();
    case 317:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(349);
      END_STATE();
    case 318:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(350);
      END_STATE();
    case 319:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(351);
      END_STATE();
    case 320:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(352);
      END_STATE();
    case 321:
      ACCEPT_TOKEN(aux_sym_cast_type_token5);
      END_STATE();
    case 322:
      ACCEPT_TOKEN(aux_sym_global_declaration_token1);
      END_STATE();
    case 323:
      ACCEPT_TOKEN(aux_sym_cast_type_token9);
      END_STATE();
    case 324:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(353);
      END_STATE();
    case 325:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(354);
      END_STATE();
    case 326:
      ACCEPT_TOKEN(aux_sym_return_statement_token1);
      END_STATE();
    case 327:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(355);
      END_STATE();
    case 328:
      ACCEPT_TOKEN(aux_sym_cast_type_token4);
      END_STATE();
    case 329:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(356);
      END_STATE();
    case 330:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(357);
      END_STATE();
    case 331:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(358);
      END_STATE();
    case 332:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(359);
      END_STATE();
    case 333:
      ACCEPT_TOKEN(aux_sym_base_clause_token1);
      END_STATE();
    case 334:
      ACCEPT_TOKEN(aux_sym_finally_clause_token1);
      END_STATE();
    case 335:
      ACCEPT_TOKEN(aux_sym_foreach_statement_token1);
      END_STATE();
    case 336:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(360);
      END_STATE();
    case 337:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(361);
      END_STATE();
    case 338:
      ACCEPT_TOKEN(aux_sym_include_expression_token1);
      if (lookahead == '_') ADVANCE(362);
      END_STATE();
    case 339:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(363);
      END_STATE();
    case 340:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(364);
      END_STATE();
    case 341:
      ACCEPT_TOKEN(aux_sym_cast_type_token7);
      END_STATE();
    case 342:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(365);
      END_STATE();
    case 343:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(366);
      END_STATE();
    case 344:
      ACCEPT_TOKEN(aux_sym_visibility_modifier_token3);
      END_STATE();
    case 345:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(367);
      END_STATE();
    case 346:
      if (lookahead == 'g') ADVANCE(368);
      END_STATE();
    case 347:
      if (lookahead == 'e') ADVANCE(369);
      END_STATE();
    case 348:
      if (lookahead == 't') ADVANCE(370);
      END_STATE();
    case 349:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(371);
      END_STATE();
    case 350:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(372);
      END_STATE();
    case 351:
      ACCEPT_TOKEN(aux_sym_declare_statement_token1);
      END_STATE();
    case 352:
      ACCEPT_TOKEN(aux_sym_match_default_expression_token1);
      END_STATE();
    case 353:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(373);
      END_STATE();
    case 354:
      ACCEPT_TOKEN(aux_sym_require_expression_token1);
      if (lookahead == '_') ADVANCE(374);
      END_STATE();
    case 355:
      ACCEPT_TOKEN(aux_sym_abstract_modifier_token1);
      END_STATE();
    case 356:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(375);
      END_STATE();
    case 357:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(376);
      END_STATE();
    case 358:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(377);
      END_STATE();
    case 359:
      ACCEPT_TOKEN(aux_sym_while_statement_token2);
      END_STATE();
    case 360:
      ACCEPT_TOKEN(aux_sym__namespace_use_type_token1);
      END_STATE();
    case 361:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(378);
      END_STATE();
    case 362:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(379);
      END_STATE();
    case 363:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(380);
      END_STATE();
    case 364:
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(381);
      END_STATE();
    case 365:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(382);
      END_STATE();
    case 366:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(383);
      END_STATE();
    case 367:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(384);
      END_STATE();
    case 368:
      ACCEPT_TOKEN(anon_sym_encoding);
      END_STATE();
    case 369:
      ACCEPT_TOKEN(anon_sym_iterable);
      END_STATE();
    case 370:
      if (lookahead == 'y') ADVANCE(385);
      END_STATE();
    case 371:
      ACCEPT_TOKEN(aux_sym_primitive_type_token1);
      END_STATE();
    case 372:
      ACCEPT_TOKEN(aux_sym_continue_statement_token1);
      END_STATE();
    case 373:
      ACCEPT_TOKEN(aux_sym_readonly_modifier_token1);
      END_STATE();
    case 374:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(386);
      END_STATE();
    case 375:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(387);
      END_STATE();
    case 376:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(388);
      END_STATE();
    case 377:
      ACCEPT_TOKEN(aux_sym_switch_block_token1);
      END_STATE();
    case 378:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(389);
      END_STATE();
    case 379:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(390);
      END_STATE();
    case 380:
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(391);
      END_STATE();
    case 381:
      ACCEPT_TOKEN(aux_sym_use_instead_of_clause_token1);
      END_STATE();
    case 382:
      ACCEPT_TOKEN(aux_sym_interface_declaration_token1);
      END_STATE();
    case 383:
      ACCEPT_TOKEN(aux_sym_namespace_definition_token1);
      END_STATE();
    case 384:
      ACCEPT_TOKEN(aux_sym_visibility_modifier_token2);
      END_STATE();
    case 385:
      if (lookahead == 'p') ADVANCE(392);
      END_STATE();
    case 386:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(393);
      END_STATE();
    case 387:
      ACCEPT_TOKEN(aux_sym_declare_statement_token2);
      END_STATE();
    case 388:
      ACCEPT_TOKEN(aux_sym_foreach_statement_token2);
      END_STATE();
    case 389:
      ACCEPT_TOKEN(aux_sym_class_interface_clause_token1);
      END_STATE();
    case 390:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(394);
      END_STATE();
    case 391:
      ACCEPT_TOKEN(aux_sym_binary_expression_token1);
      END_STATE();
    case 392:
      if (lookahead == 'e') ADVANCE(395);
      END_STATE();
    case 393:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(396);
      END_STATE();
    case 394:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(397);
      END_STATE();
    case 395:
      if (lookahead == 's') ADVANCE(398);
      END_STATE();
    case 396:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(399);
      END_STATE();
    case 397:
      ACCEPT_TOKEN(aux_sym_include_once_expression_token1);
      END_STATE();
    case 398:
      ACCEPT_TOKEN(anon_sym_strict_types);
      END_STATE();
    case 399:
      ACCEPT_TOKEN(aux_sym_require_once_expression_token1);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 66},
  [2] = {.lex_state = 63},
  [3] = {.lex_state = 63},
  [4] = {.lex_state = 63},
  [5] = {.lex_state = 63},
  [6] = {.lex_state = 63},
  [7] = {.lex_state = 63},
  [8] = {.lex_state = 63},
  [9] = {.lex_state = 63, .external_lex_state = 2},
  [10] = {.lex_state = 63, .external_lex_state = 2},
  [11] = {.lex_state = 63},
  [12] = {.lex_state = 63},
  [13] = {.lex_state = 63},
  [14] = {.lex_state = 63, .external_lex_state = 2},
  [15] = {.lex_state = 63, .external_lex_state = 2},
  [16] = {.lex_state = 63, .external_lex_state = 2},
  [17] = {.lex_state = 63, .external_lex_state = 2},
  [18] = {.lex_state = 63, .external_lex_state = 2},
  [19] = {.lex_state = 63, .external_lex_state = 2},
  [20] = {.lex_state = 63, .external_lex_state = 2},
  [21] = {.lex_state = 63, .external_lex_state = 2},
  [22] = {.lex_state = 63, .external_lex_state = 2},
  [23] = {.lex_state = 63, .external_lex_state = 2},
  [24] = {.lex_state = 63, .external_lex_state = 2},
  [25] = {.lex_state = 63, .external_lex_state = 2},
  [26] = {.lex_state = 63, .external_lex_state = 2},
  [27] = {.lex_state = 63, .external_lex_state = 2},
  [28] = {.lex_state = 63, .external_lex_state = 2},
  [29] = {.lex_state = 63, .external_lex_state = 2},
  [30] = {.lex_state = 63, .external_lex_state = 2},
  [31] = {.lex_state = 63, .external_lex_state = 2},
  [32] = {.lex_state = 63, .external_lex_state = 2},
  [33] = {.lex_state = 63, .external_lex_state = 2},
  [34] = {.lex_state = 63, .external_lex_state = 2},
  [35] = {.lex_state = 63, .external_lex_state = 2},
  [36] = {.lex_state = 63, .external_lex_state = 2},
  [37] = {.lex_state = 63, .external_lex_state = 2},
  [38] = {.lex_state = 63, .external_lex_state = 2},
  [39] = {.lex_state = 63, .external_lex_state = 2},
  [40] = {.lex_state = 63, .external_lex_state = 2},
  [41] = {.lex_state = 63, .external_lex_state = 2},
  [42] = {.lex_state = 63, .external_lex_state = 2},
  [43] = {.lex_state = 63, .external_lex_state = 2},
  [44] = {.lex_state = 63, .external_lex_state = 2},
  [45] = {.lex_state = 63, .external_lex_state = 2},
  [46] = {.lex_state = 63, .external_lex_state = 2},
  [47] = {.lex_state = 63, .external_lex_state = 2},
  [48] = {.lex_state = 63, .external_lex_state = 2},
  [49] = {.lex_state = 63, .external_lex_state = 2},
  [50] = {.lex_state = 63, .external_lex_state = 2},
  [51] = {.lex_state = 63, .external_lex_state = 2},
  [52] = {.lex_state = 63},
  [53] = {.lex_state = 63},
  [54] = {.lex_state = 63},
  [55] = {.lex_state = 63},
  [56] = {.lex_state = 63},
  [57] = {.lex_state = 63},
  [58] = {.lex_state = 63},
  [59] = {.lex_state = 63},
  [60] = {.lex_state = 63},
  [61] = {.lex_state = 63},
  [62] = {.lex_state = 63},
  [63] = {.lex_state = 63},
  [64] = {.lex_state = 63},
  [65] = {.lex_state = 63},
  [66] = {.lex_state = 63},
  [67] = {.lex_state = 63},
  [68] = {.lex_state = 63},
  [69] = {.lex_state = 63},
  [70] = {.lex_state = 63},
  [71] = {.lex_state = 63},
  [72] = {.lex_state = 63},
  [73] = {.lex_state = 63},
  [74] = {.lex_state = 63},
  [75] = {.lex_state = 63},
  [76] = {.lex_state = 63},
  [77] = {.lex_state = 63},
  [78] = {.lex_state = 63},
  [79] = {.lex_state = 63},
  [80] = {.lex_state = 63},
  [81] = {.lex_state = 63},
  [82] = {.lex_state = 63},
  [83] = {.lex_state = 63},
  [84] = {.lex_state = 63},
  [85] = {.lex_state = 63},
  [86] = {.lex_state = 63},
  [87] = {.lex_state = 63},
  [88] = {.lex_state = 63},
  [89] = {.lex_state = 63},
  [90] = {.lex_state = 63},
  [91] = {.lex_state = 63},
  [92] = {.lex_state = 63},
  [93] = {.lex_state = 63},
  [94] = {.lex_state = 63},
  [95] = {.lex_state = 63},
  [96] = {.lex_state = 63},
  [97] = {.lex_state = 63},
  [98] = {.lex_state = 63},
  [99] = {.lex_state = 63},
  [100] = {.lex_state = 63},
  [101] = {.lex_state = 63},
  [102] = {.lex_state = 63},
  [103] = {.lex_state = 63},
  [104] = {.lex_state = 63},
  [105] = {.lex_state = 63},
  [106] = {.lex_state = 63},
  [107] = {.lex_state = 63},
  [108] = {.lex_state = 63},
  [109] = {.lex_state = 63},
  [110] = {.lex_state = 63},
  [111] = {.lex_state = 63},
  [112] = {.lex_state = 63},
  [113] = {.lex_state = 63},
  [114] = {.lex_state = 63},
  [115] = {.lex_state = 63},
  [116] = {.lex_state = 63},
  [117] = {.lex_state = 63},
  [118] = {.lex_state = 5},
  [119] = {.lex_state = 5},
  [120] = {.lex_state = 5, .external_lex_state = 2},
  [121] = {.lex_state = 5},
  [122] = {.lex_state = 7},
  [123] = {.lex_state = 7},
  [124] = {.lex_state = 7},
  [125] = {.lex_state = 7},
  [126] = {.lex_state = 7},
  [127] = {.lex_state = 7},
  [128] = {.lex_state = 7},
  [129] = {.lex_state = 7},
  [130] = {.lex_state = 7},
  [131] = {.lex_state = 7},
  [132] = {.lex_state = 7},
  [133] = {.lex_state = 6},
  [134] = {.lex_state = 6},
  [135] = {.lex_state = 6},
  [136] = {.lex_state = 6},
  [137] = {.lex_state = 6},
  [138] = {.lex_state = 6},
  [139] = {.lex_state = 6},
  [140] = {.lex_state = 6},
  [141] = {.lex_state = 6},
  [142] = {.lex_state = 6},
  [143] = {.lex_state = 6},
  [144] = {.lex_state = 6},
  [145] = {.lex_state = 6},
  [146] = {.lex_state = 6},
  [147] = {.lex_state = 6},
  [148] = {.lex_state = 6},
  [149] = {.lex_state = 6},
  [150] = {.lex_state = 6},
  [151] = {.lex_state = 6},
  [152] = {.lex_state = 6},
  [153] = {.lex_state = 6},
  [154] = {.lex_state = 6},
  [155] = {.lex_state = 6},
  [156] = {.lex_state = 6},
  [157] = {.lex_state = 6},
  [158] = {.lex_state = 6},
  [159] = {.lex_state = 6},
  [160] = {.lex_state = 6},
  [161] = {.lex_state = 6},
  [162] = {.lex_state = 6},
  [163] = {.lex_state = 63},
  [164] = {.lex_state = 6},
  [165] = {.lex_state = 6},
  [166] = {.lex_state = 6},
  [167] = {.lex_state = 6},
  [168] = {.lex_state = 63},
  [169] = {.lex_state = 6},
  [170] = {.lex_state = 6},
  [171] = {.lex_state = 63},
  [172] = {.lex_state = 63},
  [173] = {.lex_state = 63},
  [174] = {.lex_state = 63},
  [175] = {.lex_state = 63},
  [176] = {.lex_state = 6},
  [177] = {.lex_state = 63},
  [178] = {.lex_state = 6},
  [179] = {.lex_state = 63},
  [180] = {.lex_state = 63},
  [181] = {.lex_state = 63},
  [182] = {.lex_state = 6},
  [183] = {.lex_state = 63},
  [184] = {.lex_state = 63},
  [185] = {.lex_state = 63},
  [186] = {.lex_state = 63},
  [187] = {.lex_state = 63},
  [188] = {.lex_state = 63},
  [189] = {.lex_state = 63},
  [190] = {.lex_state = 63},
  [191] = {.lex_state = 63},
  [192] = {.lex_state = 63},
  [193] = {.lex_state = 63, .external_lex_state = 2},
  [194] = {.lex_state = 63, .external_lex_state = 2},
  [195] = {.lex_state = 63, .external_lex_state = 2},
  [196] = {.lex_state = 63, .external_lex_state = 2},
  [197] = {.lex_state = 63},
  [198] = {.lex_state = 63},
  [199] = {.lex_state = 63},
  [200] = {.lex_state = 63},
  [201] = {.lex_state = 63, .external_lex_state = 2},
  [202] = {.lex_state = 63},
  [203] = {.lex_state = 63},
  [204] = {.lex_state = 63},
  [205] = {.lex_state = 63},
  [206] = {.lex_state = 63},
  [207] = {.lex_state = 63},
  [208] = {.lex_state = 63},
  [209] = {.lex_state = 63},
  [210] = {.lex_state = 63},
  [211] = {.lex_state = 63},
  [212] = {.lex_state = 63},
  [213] = {.lex_state = 63},
  [214] = {.lex_state = 63},
  [215] = {.lex_state = 63},
  [216] = {.lex_state = 63},
  [217] = {.lex_state = 63},
  [218] = {.lex_state = 63},
  [219] = {.lex_state = 63, .external_lex_state = 2},
  [220] = {.lex_state = 63},
  [221] = {.lex_state = 63},
  [222] = {.lex_state = 63},
  [223] = {.lex_state = 63},
  [224] = {.lex_state = 63},
  [225] = {.lex_state = 63},
  [226] = {.lex_state = 63},
  [227] = {.lex_state = 63},
  [228] = {.lex_state = 6},
  [229] = {.lex_state = 63},
  [230] = {.lex_state = 63},
  [231] = {.lex_state = 6},
  [232] = {.lex_state = 63},
  [233] = {.lex_state = 63},
  [234] = {.lex_state = 63},
  [235] = {.lex_state = 63},
  [236] = {.lex_state = 63},
  [237] = {.lex_state = 63},
  [238] = {.lex_state = 63},
  [239] = {.lex_state = 63},
  [240] = {.lex_state = 63},
  [241] = {.lex_state = 63},
  [242] = {.lex_state = 63},
  [243] = {.lex_state = 63},
  [244] = {.lex_state = 63},
  [245] = {.lex_state = 63},
  [246] = {.lex_state = 63},
  [247] = {.lex_state = 63},
  [248] = {.lex_state = 63},
  [249] = {.lex_state = 63},
  [250] = {.lex_state = 63},
  [251] = {.lex_state = 63},
  [252] = {.lex_state = 63},
  [253] = {.lex_state = 63},
  [254] = {.lex_state = 63},
  [255] = {.lex_state = 63},
  [256] = {.lex_state = 63},
  [257] = {.lex_state = 63},
  [258] = {.lex_state = 63},
  [259] = {.lex_state = 63},
  [260] = {.lex_state = 63},
  [261] = {.lex_state = 63},
  [262] = {.lex_state = 63},
  [263] = {.lex_state = 63},
  [264] = {.lex_state = 63},
  [265] = {.lex_state = 63},
  [266] = {.lex_state = 63},
  [267] = {.lex_state = 63},
  [268] = {.lex_state = 63},
  [269] = {.lex_state = 63},
  [270] = {.lex_state = 63},
  [271] = {.lex_state = 63},
  [272] = {.lex_state = 63},
  [273] = {.lex_state = 63},
  [274] = {.lex_state = 63},
  [275] = {.lex_state = 63},
  [276] = {.lex_state = 63},
  [277] = {.lex_state = 63},
  [278] = {.lex_state = 63},
  [279] = {.lex_state = 63},
  [280] = {.lex_state = 63},
  [281] = {.lex_state = 63},
  [282] = {.lex_state = 63},
  [283] = {.lex_state = 63},
  [284] = {.lex_state = 63},
  [285] = {.lex_state = 63},
  [286] = {.lex_state = 63},
  [287] = {.lex_state = 63},
  [288] = {.lex_state = 63},
  [289] = {.lex_state = 63},
  [290] = {.lex_state = 63},
  [291] = {.lex_state = 63},
  [292] = {.lex_state = 63},
  [293] = {.lex_state = 63},
  [294] = {.lex_state = 63},
  [295] = {.lex_state = 63},
  [296] = {.lex_state = 63},
  [297] = {.lex_state = 63},
  [298] = {.lex_state = 63},
  [299] = {.lex_state = 63},
  [300] = {.lex_state = 63},
  [301] = {.lex_state = 63},
  [302] = {.lex_state = 63},
  [303] = {.lex_state = 63},
  [304] = {.lex_state = 63},
  [305] = {.lex_state = 63},
  [306] = {.lex_state = 63},
  [307] = {.lex_state = 63},
  [308] = {.lex_state = 63},
  [309] = {.lex_state = 63},
  [310] = {.lex_state = 63},
  [311] = {.lex_state = 63},
  [312] = {.lex_state = 63},
  [313] = {.lex_state = 63},
  [314] = {.lex_state = 63},
  [315] = {.lex_state = 63},
  [316] = {.lex_state = 63},
  [317] = {.lex_state = 63},
  [318] = {.lex_state = 63},
  [319] = {.lex_state = 63},
  [320] = {.lex_state = 63},
  [321] = {.lex_state = 63},
  [322] = {.lex_state = 63},
  [323] = {.lex_state = 63},
  [324] = {.lex_state = 63},
  [325] = {.lex_state = 63},
  [326] = {.lex_state = 63},
  [327] = {.lex_state = 63},
  [328] = {.lex_state = 63},
  [329] = {.lex_state = 63},
  [330] = {.lex_state = 63},
  [331] = {.lex_state = 63},
  [332] = {.lex_state = 63},
  [333] = {.lex_state = 63},
  [334] = {.lex_state = 63},
  [335] = {.lex_state = 63},
  [336] = {.lex_state = 63},
  [337] = {.lex_state = 63},
  [338] = {.lex_state = 63},
  [339] = {.lex_state = 63},
  [340] = {.lex_state = 63},
  [341] = {.lex_state = 63},
  [342] = {.lex_state = 63},
  [343] = {.lex_state = 63},
  [344] = {.lex_state = 63},
  [345] = {.lex_state = 63},
  [346] = {.lex_state = 63},
  [347] = {.lex_state = 63},
  [348] = {.lex_state = 63},
  [349] = {.lex_state = 63},
  [350] = {.lex_state = 63},
  [351] = {.lex_state = 63},
  [352] = {.lex_state = 63},
  [353] = {.lex_state = 63},
  [354] = {.lex_state = 63},
  [355] = {.lex_state = 63},
  [356] = {.lex_state = 63},
  [357] = {.lex_state = 63},
  [358] = {.lex_state = 63},
  [359] = {.lex_state = 63},
  [360] = {.lex_state = 63},
  [361] = {.lex_state = 63},
  [362] = {.lex_state = 63},
  [363] = {.lex_state = 63},
  [364] = {.lex_state = 63},
  [365] = {.lex_state = 63},
  [366] = {.lex_state = 63},
  [367] = {.lex_state = 63},
  [368] = {.lex_state = 63},
  [369] = {.lex_state = 63},
  [370] = {.lex_state = 63},
  [371] = {.lex_state = 63},
  [372] = {.lex_state = 63},
  [373] = {.lex_state = 63},
  [374] = {.lex_state = 63},
  [375] = {.lex_state = 63},
  [376] = {.lex_state = 63},
  [377] = {.lex_state = 63},
  [378] = {.lex_state = 63},
  [379] = {.lex_state = 63},
  [380] = {.lex_state = 63},
  [381] = {.lex_state = 63},
  [382] = {.lex_state = 63},
  [383] = {.lex_state = 63},
  [384] = {.lex_state = 63},
  [385] = {.lex_state = 63},
  [386] = {.lex_state = 63},
  [387] = {.lex_state = 63},
  [388] = {.lex_state = 63},
  [389] = {.lex_state = 63},
  [390] = {.lex_state = 63},
  [391] = {.lex_state = 63},
  [392] = {.lex_state = 63},
  [393] = {.lex_state = 63},
  [394] = {.lex_state = 63},
  [395] = {.lex_state = 63},
  [396] = {.lex_state = 63},
  [397] = {.lex_state = 63},
  [398] = {.lex_state = 63},
  [399] = {.lex_state = 63},
  [400] = {.lex_state = 63},
  [401] = {.lex_state = 63},
  [402] = {.lex_state = 63},
  [403] = {.lex_state = 63},
  [404] = {.lex_state = 63},
  [405] = {.lex_state = 63},
  [406] = {.lex_state = 63},
  [407] = {.lex_state = 63},
  [408] = {.lex_state = 63},
  [409] = {.lex_state = 63},
  [410] = {.lex_state = 63},
  [411] = {.lex_state = 63},
  [412] = {.lex_state = 63},
  [413] = {.lex_state = 63},
  [414] = {.lex_state = 63},
  [415] = {.lex_state = 63},
  [416] = {.lex_state = 63},
  [417] = {.lex_state = 63},
  [418] = {.lex_state = 63},
  [419] = {.lex_state = 63},
  [420] = {.lex_state = 63},
  [421] = {.lex_state = 63},
  [422] = {.lex_state = 63},
  [423] = {.lex_state = 63},
  [424] = {.lex_state = 63},
  [425] = {.lex_state = 63},
  [426] = {.lex_state = 63},
  [427] = {.lex_state = 63},
  [428] = {.lex_state = 63},
  [429] = {.lex_state = 63},
  [430] = {.lex_state = 63},
  [431] = {.lex_state = 63},
  [432] = {.lex_state = 63},
  [433] = {.lex_state = 63},
  [434] = {.lex_state = 63},
  [435] = {.lex_state = 63},
  [436] = {.lex_state = 63},
  [437] = {.lex_state = 63},
  [438] = {.lex_state = 63},
  [439] = {.lex_state = 63},
  [440] = {.lex_state = 63},
  [441] = {.lex_state = 63},
  [442] = {.lex_state = 63},
  [443] = {.lex_state = 63},
  [444] = {.lex_state = 63},
  [445] = {.lex_state = 63},
  [446] = {.lex_state = 63},
  [447] = {.lex_state = 63},
  [448] = {.lex_state = 63},
  [449] = {.lex_state = 63},
  [450] = {.lex_state = 63},
  [451] = {.lex_state = 63},
  [452] = {.lex_state = 63},
  [453] = {.lex_state = 63},
  [454] = {.lex_state = 63},
  [455] = {.lex_state = 63},
  [456] = {.lex_state = 63},
  [457] = {.lex_state = 63},
  [458] = {.lex_state = 63},
  [459] = {.lex_state = 63},
  [460] = {.lex_state = 63},
  [461] = {.lex_state = 63},
  [462] = {.lex_state = 63},
  [463] = {.lex_state = 63},
  [464] = {.lex_state = 63},
  [465] = {.lex_state = 63},
  [466] = {.lex_state = 63},
  [467] = {.lex_state = 63},
  [468] = {.lex_state = 63},
  [469] = {.lex_state = 63},
  [470] = {.lex_state = 63},
  [471] = {.lex_state = 63},
  [472] = {.lex_state = 63},
  [473] = {.lex_state = 63},
  [474] = {.lex_state = 63},
  [475] = {.lex_state = 63},
  [476] = {.lex_state = 63},
  [477] = {.lex_state = 63},
  [478] = {.lex_state = 63},
  [479] = {.lex_state = 63},
  [480] = {.lex_state = 63},
  [481] = {.lex_state = 63},
  [482] = {.lex_state = 63},
  [483] = {.lex_state = 63},
  [484] = {.lex_state = 63},
  [485] = {.lex_state = 63},
  [486] = {.lex_state = 63},
  [487] = {.lex_state = 63},
  [488] = {.lex_state = 63},
  [489] = {.lex_state = 63},
  [490] = {.lex_state = 63},
  [491] = {.lex_state = 63},
  [492] = {.lex_state = 63},
  [493] = {.lex_state = 63},
  [494] = {.lex_state = 63},
  [495] = {.lex_state = 63},
  [496] = {.lex_state = 63},
  [497] = {.lex_state = 63},
  [498] = {.lex_state = 63},
  [499] = {.lex_state = 63},
  [500] = {.lex_state = 63},
  [501] = {.lex_state = 63},
  [502] = {.lex_state = 63},
  [503] = {.lex_state = 63},
  [504] = {.lex_state = 63},
  [505] = {.lex_state = 63},
  [506] = {.lex_state = 63},
  [507] = {.lex_state = 63},
  [508] = {.lex_state = 63},
  [509] = {.lex_state = 63},
  [510] = {.lex_state = 63},
  [511] = {.lex_state = 63},
  [512] = {.lex_state = 63},
  [513] = {.lex_state = 63},
  [514] = {.lex_state = 63},
  [515] = {.lex_state = 63},
  [516] = {.lex_state = 63},
  [517] = {.lex_state = 63},
  [518] = {.lex_state = 63},
  [519] = {.lex_state = 63},
  [520] = {.lex_state = 63},
  [521] = {.lex_state = 63},
  [522] = {.lex_state = 63},
  [523] = {.lex_state = 63},
  [524] = {.lex_state = 63},
  [525] = {.lex_state = 63},
  [526] = {.lex_state = 63},
  [527] = {.lex_state = 63},
  [528] = {.lex_state = 63},
  [529] = {.lex_state = 63},
  [530] = {.lex_state = 63},
  [531] = {.lex_state = 63},
  [532] = {.lex_state = 63},
  [533] = {.lex_state = 63},
  [534] = {.lex_state = 63},
  [535] = {.lex_state = 63},
  [536] = {.lex_state = 63},
  [537] = {.lex_state = 63},
  [538] = {.lex_state = 63},
  [539] = {.lex_state = 63},
  [540] = {.lex_state = 63},
  [541] = {.lex_state = 63},
  [542] = {.lex_state = 63},
  [543] = {.lex_state = 63},
  [544] = {.lex_state = 63},
  [545] = {.lex_state = 63},
  [546] = {.lex_state = 63},
  [547] = {.lex_state = 63},
  [548] = {.lex_state = 63},
  [549] = {.lex_state = 63},
  [550] = {.lex_state = 63},
  [551] = {.lex_state = 63},
  [552] = {.lex_state = 63},
  [553] = {.lex_state = 63},
  [554] = {.lex_state = 63},
  [555] = {.lex_state = 63},
  [556] = {.lex_state = 63},
  [557] = {.lex_state = 63},
  [558] = {.lex_state = 63},
  [559] = {.lex_state = 63},
  [560] = {.lex_state = 63},
  [561] = {.lex_state = 63},
  [562] = {.lex_state = 63},
  [563] = {.lex_state = 63},
  [564] = {.lex_state = 63},
  [565] = {.lex_state = 63},
  [566] = {.lex_state = 63},
  [567] = {.lex_state = 63},
  [568] = {.lex_state = 63},
  [569] = {.lex_state = 63},
  [570] = {.lex_state = 63},
  [571] = {.lex_state = 63},
  [572] = {.lex_state = 63},
  [573] = {.lex_state = 63},
  [574] = {.lex_state = 63},
  [575] = {.lex_state = 63},
  [576] = {.lex_state = 63},
  [577] = {.lex_state = 63},
  [578] = {.lex_state = 63},
  [579] = {.lex_state = 63},
  [580] = {.lex_state = 63},
  [581] = {.lex_state = 63},
  [582] = {.lex_state = 63},
  [583] = {.lex_state = 63},
  [584] = {.lex_state = 63},
  [585] = {.lex_state = 63},
  [586] = {.lex_state = 63},
  [587] = {.lex_state = 63},
  [588] = {.lex_state = 63},
  [589] = {.lex_state = 64},
  [590] = {.lex_state = 64},
  [591] = {.lex_state = 64},
  [592] = {.lex_state = 64},
  [593] = {.lex_state = 64},
  [594] = {.lex_state = 64},
  [595] = {.lex_state = 64},
  [596] = {.lex_state = 64},
  [597] = {.lex_state = 64},
  [598] = {.lex_state = 64},
  [599] = {.lex_state = 64},
  [600] = {.lex_state = 64},
  [601] = {.lex_state = 64},
  [602] = {.lex_state = 64},
  [603] = {.lex_state = 64},
  [604] = {.lex_state = 64},
  [605] = {.lex_state = 64},
  [606] = {.lex_state = 64},
  [607] = {.lex_state = 64},
  [608] = {.lex_state = 64},
  [609] = {.lex_state = 64},
  [610] = {.lex_state = 64},
  [611] = {.lex_state = 64},
  [612] = {.lex_state = 64},
  [613] = {.lex_state = 64},
  [614] = {.lex_state = 64},
  [615] = {.lex_state = 64},
  [616] = {.lex_state = 64},
  [617] = {.lex_state = 64},
  [618] = {.lex_state = 64},
  [619] = {.lex_state = 64},
  [620] = {.lex_state = 64},
  [621] = {.lex_state = 64},
  [622] = {.lex_state = 64},
  [623] = {.lex_state = 64},
  [624] = {.lex_state = 64, .external_lex_state = 2},
  [625] = {.lex_state = 64, .external_lex_state = 2},
  [626] = {.lex_state = 64},
  [627] = {.lex_state = 65},
  [628] = {.lex_state = 64, .external_lex_state = 2},
  [629] = {.lex_state = 64},
  [630] = {.lex_state = 64, .external_lex_state = 2},
  [631] = {.lex_state = 64, .external_lex_state = 2},
  [632] = {.lex_state = 64, .external_lex_state = 2},
  [633] = {.lex_state = 64, .external_lex_state = 2},
  [634] = {.lex_state = 64, .external_lex_state = 2},
  [635] = {.lex_state = 64, .external_lex_state = 2},
  [636] = {.lex_state = 64, .external_lex_state = 2},
  [637] = {.lex_state = 64, .external_lex_state = 2},
  [638] = {.lex_state = 64, .external_lex_state = 2},
  [639] = {.lex_state = 65},
  [640] = {.lex_state = 65},
  [641] = {.lex_state = 64, .external_lex_state = 2},
  [642] = {.lex_state = 65},
  [643] = {.lex_state = 64, .external_lex_state = 2},
  [644] = {.lex_state = 64, .external_lex_state = 2},
  [645] = {.lex_state = 65},
  [646] = {.lex_state = 64, .external_lex_state = 2},
  [647] = {.lex_state = 64, .external_lex_state = 2},
  [648] = {.lex_state = 64},
  [649] = {.lex_state = 64, .external_lex_state = 2},
  [650] = {.lex_state = 64, .external_lex_state = 2},
  [651] = {.lex_state = 64},
  [652] = {.lex_state = 64, .external_lex_state = 2},
  [653] = {.lex_state = 64, .external_lex_state = 2},
  [654] = {.lex_state = 64, .external_lex_state = 2},
  [655] = {.lex_state = 64, .external_lex_state = 2},
  [656] = {.lex_state = 64, .external_lex_state = 2},
  [657] = {.lex_state = 64, .external_lex_state = 2},
  [658] = {.lex_state = 64},
  [659] = {.lex_state = 64, .external_lex_state = 2},
  [660] = {.lex_state = 65},
  [661] = {.lex_state = 64, .external_lex_state = 2},
  [662] = {.lex_state = 64, .external_lex_state = 2},
  [663] = {.lex_state = 64, .external_lex_state = 2},
  [664] = {.lex_state = 64, .external_lex_state = 2},
  [665] = {.lex_state = 64, .external_lex_state = 2},
  [666] = {.lex_state = 64, .external_lex_state = 2},
  [667] = {.lex_state = 64},
  [668] = {.lex_state = 64, .external_lex_state = 2},
  [669] = {.lex_state = 64, .external_lex_state = 2},
  [670] = {.lex_state = 64},
  [671] = {.lex_state = 64, .external_lex_state = 2},
  [672] = {.lex_state = 64, .external_lex_state = 2},
  [673] = {.lex_state = 64, .external_lex_state = 2},
  [674] = {.lex_state = 64, .external_lex_state = 2},
  [675] = {.lex_state = 64},
  [676] = {.lex_state = 64},
  [677] = {.lex_state = 64},
  [678] = {.lex_state = 64},
  [679] = {.lex_state = 64},
  [680] = {.lex_state = 64},
  [681] = {.lex_state = 64},
  [682] = {.lex_state = 64},
  [683] = {.lex_state = 64},
  [684] = {.lex_state = 64},
  [685] = {.lex_state = 64},
  [686] = {.lex_state = 64},
  [687] = {.lex_state = 64},
  [688] = {.lex_state = 64},
  [689] = {.lex_state = 64},
  [690] = {.lex_state = 65},
  [691] = {.lex_state = 64},
  [692] = {.lex_state = 65},
  [693] = {.lex_state = 65},
  [694] = {.lex_state = 64},
  [695] = {.lex_state = 64},
  [696] = {.lex_state = 65},
  [697] = {.lex_state = 65},
  [698] = {.lex_state = 64},
  [699] = {.lex_state = 65},
  [700] = {.lex_state = 65},
  [701] = {.lex_state = 65},
  [702] = {.lex_state = 64},
  [703] = {.lex_state = 65},
  [704] = {.lex_state = 65},
  [705] = {.lex_state = 65},
  [706] = {.lex_state = 65},
  [707] = {.lex_state = 65},
  [708] = {.lex_state = 65},
  [709] = {.lex_state = 65},
  [710] = {.lex_state = 65},
  [711] = {.lex_state = 10},
  [712] = {.lex_state = 65},
  [713] = {.lex_state = 10},
  [714] = {.lex_state = 12},
  [715] = {.lex_state = 10},
  [716] = {.lex_state = 10},
  [717] = {.lex_state = 10},
  [718] = {.lex_state = 8},
  [719] = {.lex_state = 12},
  [720] = {.lex_state = 12},
  [721] = {.lex_state = 8},
  [722] = {.lex_state = 8},
  [723] = {.lex_state = 8},
  [724] = {.lex_state = 8},
  [725] = {.lex_state = 8},
  [726] = {.lex_state = 12, .external_lex_state = 2},
  [727] = {.lex_state = 8},
  [728] = {.lex_state = 8},
  [729] = {.lex_state = 8},
  [730] = {.lex_state = 8},
  [731] = {.lex_state = 8},
  [732] = {.lex_state = 8},
  [733] = {.lex_state = 8},
  [734] = {.lex_state = 8},
  [735] = {.lex_state = 8},
  [736] = {.lex_state = 8},
  [737] = {.lex_state = 8},
  [738] = {.lex_state = 8},
  [739] = {.lex_state = 8},
  [740] = {.lex_state = 8},
  [741] = {.lex_state = 8},
  [742] = {.lex_state = 8},
  [743] = {.lex_state = 8},
  [744] = {.lex_state = 8},
  [745] = {.lex_state = 10},
  [746] = {.lex_state = 8},
  [747] = {.lex_state = 8},
  [748] = {.lex_state = 8},
  [749] = {.lex_state = 10},
  [750] = {.lex_state = 8},
  [751] = {.lex_state = 8},
  [752] = {.lex_state = 8},
  [753] = {.lex_state = 8},
  [754] = {.lex_state = 8},
  [755] = {.lex_state = 8},
  [756] = {.lex_state = 8},
  [757] = {.lex_state = 8},
  [758] = {.lex_state = 8},
  [759] = {.lex_state = 10},
  [760] = {.lex_state = 10},
  [761] = {.lex_state = 8},
  [762] = {.lex_state = 8},
  [763] = {.lex_state = 10},
  [764] = {.lex_state = 10},
  [765] = {.lex_state = 10},
  [766] = {.lex_state = 15},
  [767] = {.lex_state = 10},
  [768] = {.lex_state = 10},
  [769] = {.lex_state = 10},
  [770] = {.lex_state = 10},
  [771] = {.lex_state = 10},
  [772] = {.lex_state = 10},
  [773] = {.lex_state = 10},
  [774] = {.lex_state = 10},
  [775] = {.lex_state = 10},
  [776] = {.lex_state = 10},
  [777] = {.lex_state = 10},
  [778] = {.lex_state = 10},
  [779] = {.lex_state = 10},
  [780] = {.lex_state = 10},
  [781] = {.lex_state = 10},
  [782] = {.lex_state = 10},
  [783] = {.lex_state = 10},
  [784] = {.lex_state = 10},
  [785] = {.lex_state = 10},
  [786] = {.lex_state = 10},
  [787] = {.lex_state = 10},
  [788] = {.lex_state = 10},
  [789] = {.lex_state = 10},
  [790] = {.lex_state = 12},
  [791] = {.lex_state = 10},
  [792] = {.lex_state = 10},
  [793] = {.lex_state = 10},
  [794] = {.lex_state = 10},
  [795] = {.lex_state = 10},
  [796] = {.lex_state = 10},
  [797] = {.lex_state = 10},
  [798] = {.lex_state = 10},
  [799] = {.lex_state = 12},
  [800] = {.lex_state = 10},
  [801] = {.lex_state = 10},
  [802] = {.lex_state = 10},
  [803] = {.lex_state = 10},
  [804] = {.lex_state = 10},
  [805] = {.lex_state = 10},
  [806] = {.lex_state = 10},
  [807] = {.lex_state = 10},
  [808] = {.lex_state = 10},
  [809] = {.lex_state = 10},
  [810] = {.lex_state = 10},
  [811] = {.lex_state = 10},
  [812] = {.lex_state = 10},
  [813] = {.lex_state = 10},
  [814] = {.lex_state = 10},
  [815] = {.lex_state = 10},
  [816] = {.lex_state = 10},
  [817] = {.lex_state = 10},
  [818] = {.lex_state = 10},
  [819] = {.lex_state = 10},
  [820] = {.lex_state = 10},
  [821] = {.lex_state = 10},
  [822] = {.lex_state = 10},
  [823] = {.lex_state = 10},
  [824] = {.lex_state = 10},
  [825] = {.lex_state = 10},
  [826] = {.lex_state = 10},
  [827] = {.lex_state = 10},
  [828] = {.lex_state = 10},
  [829] = {.lex_state = 10},
  [830] = {.lex_state = 10},
  [831] = {.lex_state = 10},
  [832] = {.lex_state = 10},
  [833] = {.lex_state = 10},
  [834] = {.lex_state = 10},
  [835] = {.lex_state = 10},
  [836] = {.lex_state = 10},
  [837] = {.lex_state = 10},
  [838] = {.lex_state = 10},
  [839] = {.lex_state = 10},
  [840] = {.lex_state = 10},
  [841] = {.lex_state = 10},
  [842] = {.lex_state = 10},
  [843] = {.lex_state = 10},
  [844] = {.lex_state = 15},
  [845] = {.lex_state = 10},
  [846] = {.lex_state = 10},
  [847] = {.lex_state = 10},
  [848] = {.lex_state = 10},
  [849] = {.lex_state = 10},
  [850] = {.lex_state = 15},
  [851] = {.lex_state = 10},
  [852] = {.lex_state = 10},
  [853] = {.lex_state = 10},
  [854] = {.lex_state = 12},
  [855] = {.lex_state = 12},
  [856] = {.lex_state = 10},
  [857] = {.lex_state = 10},
  [858] = {.lex_state = 10, .external_lex_state = 2},
  [859] = {.lex_state = 8, .external_lex_state = 2},
  [860] = {.lex_state = 8, .external_lex_state = 2},
  [861] = {.lex_state = 15},
  [862] = {.lex_state = 15},
  [863] = {.lex_state = 12},
  [864] = {.lex_state = 15},
  [865] = {.lex_state = 8, .external_lex_state = 2},
  [866] = {.lex_state = 8, .external_lex_state = 2},
  [867] = {.lex_state = 8, .external_lex_state = 2},
  [868] = {.lex_state = 8, .external_lex_state = 2},
  [869] = {.lex_state = 8, .external_lex_state = 2},
  [870] = {.lex_state = 8, .external_lex_state = 2},
  [871] = {.lex_state = 10},
  [872] = {.lex_state = 8, .external_lex_state = 2},
  [873] = {.lex_state = 8, .external_lex_state = 2},
  [874] = {.lex_state = 8, .external_lex_state = 2},
  [875] = {.lex_state = 8, .external_lex_state = 2},
  [876] = {.lex_state = 8, .external_lex_state = 2},
  [877] = {.lex_state = 8, .external_lex_state = 2},
  [878] = {.lex_state = 8, .external_lex_state = 2},
  [879] = {.lex_state = 8, .external_lex_state = 2},
  [880] = {.lex_state = 8, .external_lex_state = 2},
  [881] = {.lex_state = 8, .external_lex_state = 2},
  [882] = {.lex_state = 15},
  [883] = {.lex_state = 8, .external_lex_state = 2},
  [884] = {.lex_state = 8, .external_lex_state = 2},
  [885] = {.lex_state = 15},
  [886] = {.lex_state = 15},
  [887] = {.lex_state = 10, .external_lex_state = 2},
  [888] = {.lex_state = 8, .external_lex_state = 2},
  [889] = {.lex_state = 8, .external_lex_state = 2},
  [890] = {.lex_state = 8, .external_lex_state = 2},
  [891] = {.lex_state = 15},
  [892] = {.lex_state = 6},
  [893] = {.lex_state = 8, .external_lex_state = 2},
  [894] = {.lex_state = 8, .external_lex_state = 2},
  [895] = {.lex_state = 10, .external_lex_state = 2},
  [896] = {.lex_state = 10, .external_lex_state = 2},
  [897] = {.lex_state = 10, .external_lex_state = 2},
  [898] = {.lex_state = 10, .external_lex_state = 2},
  [899] = {.lex_state = 12},
  [900] = {.lex_state = 8, .external_lex_state = 2},
  [901] = {.lex_state = 8, .external_lex_state = 2},
  [902] = {.lex_state = 6},
  [903] = {.lex_state = 8, .external_lex_state = 2},
  [904] = {.lex_state = 8, .external_lex_state = 2},
  [905] = {.lex_state = 8, .external_lex_state = 2},
  [906] = {.lex_state = 8, .external_lex_state = 2},
  [907] = {.lex_state = 12},
  [908] = {.lex_state = 12},
  [909] = {.lex_state = 8, .external_lex_state = 2},
  [910] = {.lex_state = 8, .external_lex_state = 2},
  [911] = {.lex_state = 8, .external_lex_state = 2},
  [912] = {.lex_state = 8, .external_lex_state = 2},
  [913] = {.lex_state = 8, .external_lex_state = 2},
  [914] = {.lex_state = 8, .external_lex_state = 2},
  [915] = {.lex_state = 8, .external_lex_state = 2},
  [916] = {.lex_state = 12},
  [917] = {.lex_state = 12},
  [918] = {.lex_state = 10, .external_lex_state = 2},
  [919] = {.lex_state = 10, .external_lex_state = 2},
  [920] = {.lex_state = 10, .external_lex_state = 2},
  [921] = {.lex_state = 10, .external_lex_state = 2},
  [922] = {.lex_state = 10, .external_lex_state = 2},
  [923] = {.lex_state = 10, .external_lex_state = 2},
  [924] = {.lex_state = 10, .external_lex_state = 2},
  [925] = {.lex_state = 10, .external_lex_state = 2},
  [926] = {.lex_state = 10, .external_lex_state = 2},
  [927] = {.lex_state = 10, .external_lex_state = 2},
  [928] = {.lex_state = 10, .external_lex_state = 2},
  [929] = {.lex_state = 10, .external_lex_state = 2},
  [930] = {.lex_state = 10, .external_lex_state = 2},
  [931] = {.lex_state = 10, .external_lex_state = 2},
  [932] = {.lex_state = 12},
  [933] = {.lex_state = 10, .external_lex_state = 2},
  [934] = {.lex_state = 10, .external_lex_state = 2},
  [935] = {.lex_state = 10, .external_lex_state = 2},
  [936] = {.lex_state = 10, .external_lex_state = 2},
  [937] = {.lex_state = 10, .external_lex_state = 2},
  [938] = {.lex_state = 10, .external_lex_state = 2},
  [939] = {.lex_state = 10, .external_lex_state = 2},
  [940] = {.lex_state = 10, .external_lex_state = 2},
  [941] = {.lex_state = 10},
  [942] = {.lex_state = 10, .external_lex_state = 2},
  [943] = {.lex_state = 10, .external_lex_state = 2},
  [944] = {.lex_state = 10, .external_lex_state = 2},
  [945] = {.lex_state = 10, .external_lex_state = 2},
  [946] = {.lex_state = 10, .external_lex_state = 2},
  [947] = {.lex_state = 10, .external_lex_state = 2},
  [948] = {.lex_state = 10, .external_lex_state = 2},
  [949] = {.lex_state = 10, .external_lex_state = 2},
  [950] = {.lex_state = 10, .external_lex_state = 2},
  [951] = {.lex_state = 10, .external_lex_state = 2},
  [952] = {.lex_state = 10, .external_lex_state = 2},
  [953] = {.lex_state = 10, .external_lex_state = 2},
  [954] = {.lex_state = 10, .external_lex_state = 2},
  [955] = {.lex_state = 10, .external_lex_state = 2},
  [956] = {.lex_state = 10, .external_lex_state = 2},
  [957] = {.lex_state = 10, .external_lex_state = 2},
  [958] = {.lex_state = 10, .external_lex_state = 2},
  [959] = {.lex_state = 10, .external_lex_state = 2},
  [960] = {.lex_state = 10, .external_lex_state = 2},
  [961] = {.lex_state = 10, .external_lex_state = 2},
  [962] = {.lex_state = 10, .external_lex_state = 2},
  [963] = {.lex_state = 10, .external_lex_state = 2},
  [964] = {.lex_state = 10, .external_lex_state = 2},
  [965] = {.lex_state = 10, .external_lex_state = 2},
  [966] = {.lex_state = 10, .external_lex_state = 2},
  [967] = {.lex_state = 10, .external_lex_state = 2},
  [968] = {.lex_state = 10, .external_lex_state = 2},
  [969] = {.lex_state = 10, .external_lex_state = 2},
  [970] = {.lex_state = 10, .external_lex_state = 2},
  [971] = {.lex_state = 10, .external_lex_state = 2},
  [972] = {.lex_state = 10, .external_lex_state = 2},
  [973] = {.lex_state = 10, .external_lex_state = 2},
  [974] = {.lex_state = 10, .external_lex_state = 2},
  [975] = {.lex_state = 10, .external_lex_state = 2},
  [976] = {.lex_state = 10, .external_lex_state = 2},
  [977] = {.lex_state = 10, .external_lex_state = 2},
  [978] = {.lex_state = 10, .external_lex_state = 2},
  [979] = {.lex_state = 10, .external_lex_state = 2},
  [980] = {.lex_state = 10, .external_lex_state = 2},
  [981] = {.lex_state = 10, .external_lex_state = 2},
  [982] = {.lex_state = 12},
  [983] = {.lex_state = 10, .external_lex_state = 2},
  [984] = {.lex_state = 10, .external_lex_state = 2},
  [985] = {.lex_state = 10, .external_lex_state = 2},
  [986] = {.lex_state = 10, .external_lex_state = 2},
  [987] = {.lex_state = 10, .external_lex_state = 2},
  [988] = {.lex_state = 10, .external_lex_state = 2},
  [989] = {.lex_state = 10, .external_lex_state = 2},
  [990] = {.lex_state = 10, .external_lex_state = 2},
  [991] = {.lex_state = 10, .external_lex_state = 2},
  [992] = {.lex_state = 10, .external_lex_state = 2},
  [993] = {.lex_state = 10},
  [994] = {.lex_state = 10, .external_lex_state = 2},
  [995] = {.lex_state = 10, .external_lex_state = 2},
  [996] = {.lex_state = 12, .external_lex_state = 2},
  [997] = {.lex_state = 10, .external_lex_state = 2},
  [998] = {.lex_state = 10, .external_lex_state = 2},
  [999] = {.lex_state = 10, .external_lex_state = 2},
  [1000] = {.lex_state = 10, .external_lex_state = 2},
  [1001] = {.lex_state = 10, .external_lex_state = 2},
  [1002] = {.lex_state = 10, .external_lex_state = 2},
  [1003] = {.lex_state = 10, .external_lex_state = 2},
  [1004] = {.lex_state = 10, .external_lex_state = 2},
  [1005] = {.lex_state = 10, .external_lex_state = 2},
  [1006] = {.lex_state = 10, .external_lex_state = 2},
  [1007] = {.lex_state = 10, .external_lex_state = 2},
  [1008] = {.lex_state = 10, .external_lex_state = 2},
  [1009] = {.lex_state = 12},
  [1010] = {.lex_state = 10, .external_lex_state = 2},
  [1011] = {.lex_state = 10, .external_lex_state = 2},
  [1012] = {.lex_state = 10, .external_lex_state = 2},
  [1013] = {.lex_state = 12},
  [1014] = {.lex_state = 12},
  [1015] = {.lex_state = 12},
  [1016] = {.lex_state = 12},
  [1017] = {.lex_state = 10, .external_lex_state = 2},
  [1018] = {.lex_state = 12},
  [1019] = {.lex_state = 12},
  [1020] = {.lex_state = 10},
  [1021] = {.lex_state = 12},
  [1022] = {.lex_state = 12},
  [1023] = {.lex_state = 12},
  [1024] = {.lex_state = 12},
  [1025] = {.lex_state = 12},
  [1026] = {.lex_state = 12},
  [1027] = {.lex_state = 12},
  [1028] = {.lex_state = 12},
  [1029] = {.lex_state = 12},
  [1030] = {.lex_state = 12},
  [1031] = {.lex_state = 12},
  [1032] = {.lex_state = 10, .external_lex_state = 2},
  [1033] = {.lex_state = 12},
  [1034] = {.lex_state = 12},
  [1035] = {.lex_state = 10, .external_lex_state = 2},
  [1036] = {.lex_state = 12},
  [1037] = {.lex_state = 12},
  [1038] = {.lex_state = 10, .external_lex_state = 2},
  [1039] = {.lex_state = 12, .external_lex_state = 2},
  [1040] = {.lex_state = 12},
  [1041] = {.lex_state = 12},
  [1042] = {.lex_state = 12},
  [1043] = {.lex_state = 12, .external_lex_state = 2},
  [1044] = {.lex_state = 12, .external_lex_state = 2},
  [1045] = {.lex_state = 12, .external_lex_state = 2},
  [1046] = {.lex_state = 12, .external_lex_state = 2},
  [1047] = {.lex_state = 12, .external_lex_state = 2},
  [1048] = {.lex_state = 12, .external_lex_state = 2},
  [1049] = {.lex_state = 12},
  [1050] = {.lex_state = 12},
  [1051] = {.lex_state = 12, .external_lex_state = 2},
  [1052] = {.lex_state = 12},
  [1053] = {.lex_state = 12},
  [1054] = {.lex_state = 12},
  [1055] = {.lex_state = 12},
  [1056] = {.lex_state = 12},
  [1057] = {.lex_state = 12},
  [1058] = {.lex_state = 12},
  [1059] = {.lex_state = 12},
  [1060] = {.lex_state = 12},
  [1061] = {.lex_state = 12},
  [1062] = {.lex_state = 12},
  [1063] = {.lex_state = 12},
  [1064] = {.lex_state = 12},
  [1065] = {.lex_state = 12},
  [1066] = {.lex_state = 12},
  [1067] = {.lex_state = 12},
  [1068] = {.lex_state = 12},
  [1069] = {.lex_state = 12},
  [1070] = {.lex_state = 12},
  [1071] = {.lex_state = 12},
  [1072] = {.lex_state = 12, .external_lex_state = 2},
  [1073] = {.lex_state = 12},
  [1074] = {.lex_state = 12},
  [1075] = {.lex_state = 12},
  [1076] = {.lex_state = 12},
  [1077] = {.lex_state = 12},
  [1078] = {.lex_state = 12},
  [1079] = {.lex_state = 12},
  [1080] = {.lex_state = 12},
  [1081] = {.lex_state = 12},
  [1082] = {.lex_state = 12},
  [1083] = {.lex_state = 12},
  [1084] = {.lex_state = 12, .external_lex_state = 2},
  [1085] = {.lex_state = 12, .external_lex_state = 2},
  [1086] = {.lex_state = 12, .external_lex_state = 2},
  [1087] = {.lex_state = 12, .external_lex_state = 2},
  [1088] = {.lex_state = 12},
  [1089] = {.lex_state = 12, .external_lex_state = 2},
  [1090] = {.lex_state = 12, .external_lex_state = 2},
  [1091] = {.lex_state = 12, .external_lex_state = 2},
  [1092] = {.lex_state = 12, .external_lex_state = 2},
  [1093] = {.lex_state = 12, .external_lex_state = 2},
  [1094] = {.lex_state = 12, .external_lex_state = 2},
  [1095] = {.lex_state = 12, .external_lex_state = 2},
  [1096] = {.lex_state = 12, .external_lex_state = 2},
  [1097] = {.lex_state = 12, .external_lex_state = 2},
  [1098] = {.lex_state = 12, .external_lex_state = 2},
  [1099] = {.lex_state = 12, .external_lex_state = 2},
  [1100] = {.lex_state = 12, .external_lex_state = 2},
  [1101] = {.lex_state = 12},
  [1102] = {.lex_state = 12, .external_lex_state = 2},
  [1103] = {.lex_state = 12, .external_lex_state = 2},
  [1104] = {.lex_state = 17},
  [1105] = {.lex_state = 12, .external_lex_state = 2},
  [1106] = {.lex_state = 12},
  [1107] = {.lex_state = 12, .external_lex_state = 2},
  [1108] = {.lex_state = 12, .external_lex_state = 2},
  [1109] = {.lex_state = 12, .external_lex_state = 2},
  [1110] = {.lex_state = 12, .external_lex_state = 2},
  [1111] = {.lex_state = 12},
  [1112] = {.lex_state = 12, .external_lex_state = 2},
  [1113] = {.lex_state = 12, .external_lex_state = 2},
  [1114] = {.lex_state = 12, .external_lex_state = 2},
  [1115] = {.lex_state = 12, .external_lex_state = 2},
  [1116] = {.lex_state = 12, .external_lex_state = 2},
  [1117] = {.lex_state = 12, .external_lex_state = 2},
  [1118] = {.lex_state = 12, .external_lex_state = 2},
  [1119] = {.lex_state = 12, .external_lex_state = 2},
  [1120] = {.lex_state = 12, .external_lex_state = 2},
  [1121] = {.lex_state = 12, .external_lex_state = 2},
  [1122] = {.lex_state = 12, .external_lex_state = 2},
  [1123] = {.lex_state = 12, .external_lex_state = 2},
  [1124] = {.lex_state = 12, .external_lex_state = 2},
  [1125] = {.lex_state = 12, .external_lex_state = 2},
  [1126] = {.lex_state = 12, .external_lex_state = 2},
  [1127] = {.lex_state = 12, .external_lex_state = 2},
  [1128] = {.lex_state = 12, .external_lex_state = 2},
  [1129] = {.lex_state = 12, .external_lex_state = 2},
  [1130] = {.lex_state = 12, .external_lex_state = 2},
  [1131] = {.lex_state = 12, .external_lex_state = 2},
  [1132] = {.lex_state = 12, .external_lex_state = 2},
  [1133] = {.lex_state = 12, .external_lex_state = 2},
  [1134] = {.lex_state = 12, .external_lex_state = 2},
  [1135] = {.lex_state = 12, .external_lex_state = 2},
  [1136] = {.lex_state = 12, .external_lex_state = 2},
  [1137] = {.lex_state = 12, .external_lex_state = 2},
  [1138] = {.lex_state = 12},
  [1139] = {.lex_state = 12, .external_lex_state = 2},
  [1140] = {.lex_state = 12},
  [1141] = {.lex_state = 12},
  [1142] = {.lex_state = 12},
  [1143] = {.lex_state = 12},
  [1144] = {.lex_state = 12},
  [1145] = {.lex_state = 12},
  [1146] = {.lex_state = 12},
  [1147] = {.lex_state = 12},
  [1148] = {.lex_state = 12},
  [1149] = {.lex_state = 12},
  [1150] = {.lex_state = 12},
  [1151] = {.lex_state = 12},
  [1152] = {.lex_state = 12},
  [1153] = {.lex_state = 12},
  [1154] = {.lex_state = 12},
  [1155] = {.lex_state = 12},
  [1156] = {.lex_state = 12},
  [1157] = {.lex_state = 12},
  [1158] = {.lex_state = 12},
  [1159] = {.lex_state = 12},
  [1160] = {.lex_state = 12},
  [1161] = {.lex_state = 12},
  [1162] = {.lex_state = 12, .external_lex_state = 2},
  [1163] = {.lex_state = 12},
  [1164] = {.lex_state = 12, .external_lex_state = 2},
  [1165] = {.lex_state = 12, .external_lex_state = 2},
  [1166] = {.lex_state = 12, .external_lex_state = 2},
  [1167] = {.lex_state = 12},
  [1168] = {.lex_state = 12},
  [1169] = {.lex_state = 12},
  [1170] = {.lex_state = 12},
  [1171] = {.lex_state = 12},
  [1172] = {.lex_state = 12, .external_lex_state = 2},
  [1173] = {.lex_state = 12},
  [1174] = {.lex_state = 12, .external_lex_state = 2},
  [1175] = {.lex_state = 12, .external_lex_state = 2},
  [1176] = {.lex_state = 12, .external_lex_state = 2},
  [1177] = {.lex_state = 12, .external_lex_state = 2},
  [1178] = {.lex_state = 12},
  [1179] = {.lex_state = 12},
  [1180] = {.lex_state = 12},
  [1181] = {.lex_state = 12},
  [1182] = {.lex_state = 12},
  [1183] = {.lex_state = 12},
  [1184] = {.lex_state = 12},
  [1185] = {.lex_state = 12},
  [1186] = {.lex_state = 12},
  [1187] = {.lex_state = 12},
  [1188] = {.lex_state = 12},
  [1189] = {.lex_state = 12},
  [1190] = {.lex_state = 12},
  [1191] = {.lex_state = 12},
  [1192] = {.lex_state = 12},
  [1193] = {.lex_state = 12},
  [1194] = {.lex_state = 12},
  [1195] = {.lex_state = 12},
  [1196] = {.lex_state = 12},
  [1197] = {.lex_state = 12},
  [1198] = {.lex_state = 12},
  [1199] = {.lex_state = 12},
  [1200] = {.lex_state = 12},
  [1201] = {.lex_state = 12},
  [1202] = {.lex_state = 12},
  [1203] = {.lex_state = 12},
  [1204] = {.lex_state = 12},
  [1205] = {.lex_state = 12},
  [1206] = {.lex_state = 12},
  [1207] = {.lex_state = 12},
  [1208] = {.lex_state = 12},
  [1209] = {.lex_state = 12},
  [1210] = {.lex_state = 12},
  [1211] = {.lex_state = 12, .external_lex_state = 2},
  [1212] = {.lex_state = 12},
  [1213] = {.lex_state = 12},
  [1214] = {.lex_state = 12},
  [1215] = {.lex_state = 12},
  [1216] = {.lex_state = 12},
  [1217] = {.lex_state = 12},
  [1218] = {.lex_state = 12},
  [1219] = {.lex_state = 12},
  [1220] = {.lex_state = 12},
  [1221] = {.lex_state = 12},
  [1222] = {.lex_state = 12, .external_lex_state = 2},
  [1223] = {.lex_state = 12, .external_lex_state = 2},
  [1224] = {.lex_state = 12, .external_lex_state = 2},
  [1225] = {.lex_state = 12},
  [1226] = {.lex_state = 12, .external_lex_state = 2},
  [1227] = {.lex_state = 12},
  [1228] = {.lex_state = 12},
  [1229] = {.lex_state = 12, .external_lex_state = 2},
  [1230] = {.lex_state = 12},
  [1231] = {.lex_state = 12},
  [1232] = {.lex_state = 12},
  [1233] = {.lex_state = 12, .external_lex_state = 2},
  [1234] = {.lex_state = 12, .external_lex_state = 2},
  [1235] = {.lex_state = 12, .external_lex_state = 2},
  [1236] = {.lex_state = 12},
  [1237] = {.lex_state = 12, .external_lex_state = 2},
  [1238] = {.lex_state = 12, .external_lex_state = 2},
  [1239] = {.lex_state = 12, .external_lex_state = 2},
  [1240] = {.lex_state = 12, .external_lex_state = 2},
  [1241] = {.lex_state = 12, .external_lex_state = 2},
  [1242] = {.lex_state = 12},
  [1243] = {.lex_state = 12},
  [1244] = {.lex_state = 12, .external_lex_state = 2},
  [1245] = {.lex_state = 12, .external_lex_state = 2},
  [1246] = {.lex_state = 12},
  [1247] = {.lex_state = 15},
  [1248] = {.lex_state = 15},
  [1249] = {.lex_state = 12},
  [1250] = {.lex_state = 12},
  [1251] = {.lex_state = 15},
  [1252] = {.lex_state = 12},
  [1253] = {.lex_state = 12},
  [1254] = {.lex_state = 12},
  [1255] = {.lex_state = 12},
  [1256] = {.lex_state = 12},
  [1257] = {.lex_state = 12},
  [1258] = {.lex_state = 12},
  [1259] = {.lex_state = 12},
  [1260] = {.lex_state = 12},
  [1261] = {.lex_state = 12},
  [1262] = {.lex_state = 12},
  [1263] = {.lex_state = 15},
  [1264] = {.lex_state = 15},
  [1265] = {.lex_state = 12},
  [1266] = {.lex_state = 12},
  [1267] = {.lex_state = 12},
  [1268] = {.lex_state = 12},
  [1269] = {.lex_state = 15},
  [1270] = {.lex_state = 12},
  [1271] = {.lex_state = 15},
  [1272] = {.lex_state = 12},
  [1273] = {.lex_state = 12},
  [1274] = {.lex_state = 12},
  [1275] = {.lex_state = 12},
  [1276] = {.lex_state = 12},
  [1277] = {.lex_state = 12},
  [1278] = {.lex_state = 12},
  [1279] = {.lex_state = 12},
  [1280] = {.lex_state = 16},
  [1281] = {.lex_state = 12},
  [1282] = {.lex_state = 12},
  [1283] = {.lex_state = 12},
  [1284] = {.lex_state = 12},
  [1285] = {.lex_state = 12},
  [1286] = {.lex_state = 12},
  [1287] = {.lex_state = 12},
  [1288] = {.lex_state = 12},
  [1289] = {.lex_state = 12},
  [1290] = {.lex_state = 12},
  [1291] = {.lex_state = 12},
  [1292] = {.lex_state = 12},
  [1293] = {.lex_state = 12},
  [1294] = {.lex_state = 12},
  [1295] = {.lex_state = 12},
  [1296] = {.lex_state = 12},
  [1297] = {.lex_state = 12},
  [1298] = {.lex_state = 12},
  [1299] = {.lex_state = 12},
  [1300] = {.lex_state = 12},
  [1301] = {.lex_state = 12},
  [1302] = {.lex_state = 12},
  [1303] = {.lex_state = 12},
  [1304] = {.lex_state = 12},
  [1305] = {.lex_state = 12},
  [1306] = {.lex_state = 12},
  [1307] = {.lex_state = 12},
  [1308] = {.lex_state = 12},
  [1309] = {.lex_state = 12},
  [1310] = {.lex_state = 12},
  [1311] = {.lex_state = 12},
  [1312] = {.lex_state = 12},
  [1313] = {.lex_state = 12},
  [1314] = {.lex_state = 12},
  [1315] = {.lex_state = 12},
  [1316] = {.lex_state = 12},
  [1317] = {.lex_state = 12},
  [1318] = {.lex_state = 12},
  [1319] = {.lex_state = 12},
  [1320] = {.lex_state = 12},
  [1321] = {.lex_state = 12},
  [1322] = {.lex_state = 12},
  [1323] = {.lex_state = 12},
  [1324] = {.lex_state = 12},
  [1325] = {.lex_state = 12},
  [1326] = {.lex_state = 12},
  [1327] = {.lex_state = 12},
  [1328] = {.lex_state = 12},
  [1329] = {.lex_state = 15},
  [1330] = {.lex_state = 15},
  [1331] = {.lex_state = 12},
  [1332] = {.lex_state = 12},
  [1333] = {.lex_state = 15},
  [1334] = {.lex_state = 15},
  [1335] = {.lex_state = 12},
  [1336] = {.lex_state = 15},
  [1337] = {.lex_state = 12},
  [1338] = {.lex_state = 12},
  [1339] = {.lex_state = 12},
  [1340] = {.lex_state = 12},
  [1341] = {.lex_state = 12},
  [1342] = {.lex_state = 15},
  [1343] = {.lex_state = 15},
  [1344] = {.lex_state = 15},
  [1345] = {.lex_state = 15},
  [1346] = {.lex_state = 15},
  [1347] = {.lex_state = 15},
  [1348] = {.lex_state = 12},
  [1349] = {.lex_state = 12},
  [1350] = {.lex_state = 12},
  [1351] = {.lex_state = 64},
  [1352] = {.lex_state = 12},
  [1353] = {.lex_state = 12},
  [1354] = {.lex_state = 64},
  [1355] = {.lex_state = 64},
  [1356] = {.lex_state = 64},
  [1357] = {.lex_state = 12},
  [1358] = {.lex_state = 12},
  [1359] = {.lex_state = 12},
  [1360] = {.lex_state = 12},
  [1361] = {.lex_state = 12},
  [1362] = {.lex_state = 12},
  [1363] = {.lex_state = 12},
  [1364] = {.lex_state = 15},
  [1365] = {.lex_state = 12},
  [1366] = {.lex_state = 12},
  [1367] = {.lex_state = 12},
  [1368] = {.lex_state = 15},
  [1369] = {.lex_state = 12},
  [1370] = {.lex_state = 12},
  [1371] = {.lex_state = 64},
  [1372] = {.lex_state = 64},
  [1373] = {.lex_state = 15},
  [1374] = {.lex_state = 15},
  [1375] = {.lex_state = 15},
  [1376] = {.lex_state = 64},
  [1377] = {.lex_state = 65},
  [1378] = {.lex_state = 15},
  [1379] = {.lex_state = 64},
  [1380] = {.lex_state = 64},
  [1381] = {.lex_state = 64},
  [1382] = {.lex_state = 64},
  [1383] = {.lex_state = 64},
  [1384] = {.lex_state = 14},
  [1385] = {.lex_state = 64},
  [1386] = {.lex_state = 64},
  [1387] = {.lex_state = 64},
  [1388] = {.lex_state = 3, .external_lex_state = 3},
  [1389] = {.lex_state = 64},
  [1390] = {.lex_state = 64},
  [1391] = {.lex_state = 64},
  [1392] = {.lex_state = 64},
  [1393] = {.lex_state = 64},
  [1394] = {.lex_state = 64},
  [1395] = {.lex_state = 64},
  [1396] = {.lex_state = 3, .external_lex_state = 3},
  [1397] = {.lex_state = 64},
  [1398] = {.lex_state = 64},
  [1399] = {.lex_state = 65},
  [1400] = {.lex_state = 3, .external_lex_state = 4},
  [1401] = {.lex_state = 3, .external_lex_state = 3},
  [1402] = {.lex_state = 3, .external_lex_state = 3},
  [1403] = {.lex_state = 3, .external_lex_state = 3},
  [1404] = {.lex_state = 15},
  [1405] = {.lex_state = 15},
  [1406] = {.lex_state = 15},
  [1407] = {.lex_state = 15},
  [1408] = {.lex_state = 15},
  [1409] = {.lex_state = 15},
  [1410] = {.lex_state = 15},
  [1411] = {.lex_state = 64},
  [1412] = {.lex_state = 64},
  [1413] = {.lex_state = 8, .external_lex_state = 5},
  [1414] = {.lex_state = 8, .external_lex_state = 6},
  [1415] = {.lex_state = 8, .external_lex_state = 6},
  [1416] = {.lex_state = 8, .external_lex_state = 6},
  [1417] = {.lex_state = 15},
  [1418] = {.lex_state = 8, .external_lex_state = 5},
  [1419] = {.lex_state = 8, .external_lex_state = 5},
  [1420] = {.lex_state = 15},
  [1421] = {.lex_state = 15},
  [1422] = {.lex_state = 15},
  [1423] = {.lex_state = 8, .external_lex_state = 5},
  [1424] = {.lex_state = 15},
  [1425] = {.lex_state = 8, .external_lex_state = 6},
  [1426] = {.lex_state = 8, .external_lex_state = 6},
  [1427] = {.lex_state = 8, .external_lex_state = 5},
  [1428] = {.lex_state = 15},
  [1429] = {.lex_state = 15},
  [1430] = {.lex_state = 15},
  [1431] = {.lex_state = 64},
  [1432] = {.lex_state = 15},
  [1433] = {.lex_state = 15},
  [1434] = {.lex_state = 15},
  [1435] = {.lex_state = 15},
  [1436] = {.lex_state = 15},
  [1437] = {.lex_state = 15},
  [1438] = {.lex_state = 15},
  [1439] = {.lex_state = 15},
  [1440] = {.lex_state = 15},
  [1441] = {.lex_state = 17},
  [1442] = {.lex_state = 15},
  [1443] = {.lex_state = 15},
  [1444] = {.lex_state = 64},
  [1445] = {.lex_state = 15},
  [1446] = {.lex_state = 64},
  [1447] = {.lex_state = 8, .external_lex_state = 4},
  [1448] = {.lex_state = 15},
  [1449] = {.lex_state = 15},
  [1450] = {.lex_state = 15},
  [1451] = {.lex_state = 15},
  [1452] = {.lex_state = 15},
  [1453] = {.lex_state = 15},
  [1454] = {.lex_state = 15},
  [1455] = {.lex_state = 64},
  [1456] = {.lex_state = 15},
  [1457] = {.lex_state = 64},
  [1458] = {.lex_state = 64},
  [1459] = {.lex_state = 15},
  [1460] = {.lex_state = 15},
  [1461] = {.lex_state = 64},
  [1462] = {.lex_state = 64},
  [1463] = {.lex_state = 64},
  [1464] = {.lex_state = 64},
  [1465] = {.lex_state = 64},
  [1466] = {.lex_state = 64},
  [1467] = {.lex_state = 64},
  [1468] = {.lex_state = 15},
  [1469] = {.lex_state = 15},
  [1470] = {.lex_state = 64},
  [1471] = {.lex_state = 64},
  [1472] = {.lex_state = 64},
  [1473] = {.lex_state = 64},
  [1474] = {.lex_state = 64},
  [1475] = {.lex_state = 64},
  [1476] = {.lex_state = 64},
  [1477] = {.lex_state = 64},
  [1478] = {.lex_state = 64},
  [1479] = {.lex_state = 64},
  [1480] = {.lex_state = 64},
  [1481] = {.lex_state = 64},
  [1482] = {.lex_state = 64},
  [1483] = {.lex_state = 64},
  [1484] = {.lex_state = 64},
  [1485] = {.lex_state = 64},
  [1486] = {.lex_state = 64},
  [1487] = {.lex_state = 64},
  [1488] = {.lex_state = 64},
  [1489] = {.lex_state = 15},
  [1490] = {.lex_state = 64},
  [1491] = {.lex_state = 64},
  [1492] = {.lex_state = 64},
  [1493] = {.lex_state = 64},
  [1494] = {.lex_state = 64},
  [1495] = {.lex_state = 17},
  [1496] = {.lex_state = 17},
  [1497] = {.lex_state = 17},
  [1498] = {.lex_state = 17},
  [1499] = {.lex_state = 64},
  [1500] = {.lex_state = 64},
  [1501] = {.lex_state = 64},
  [1502] = {.lex_state = 64},
  [1503] = {.lex_state = 64},
  [1504] = {.lex_state = 64},
  [1505] = {.lex_state = 64},
  [1506] = {.lex_state = 3, .external_lex_state = 7},
  [1507] = {.lex_state = 64},
  [1508] = {.lex_state = 17},
  [1509] = {.lex_state = 64},
  [1510] = {.lex_state = 64},
  [1511] = {.lex_state = 64},
  [1512] = {.lex_state = 17},
  [1513] = {.lex_state = 64},
  [1514] = {.lex_state = 17},
  [1515] = {.lex_state = 64},
  [1516] = {.lex_state = 64},
  [1517] = {.lex_state = 64},
  [1518] = {.lex_state = 64},
  [1519] = {.lex_state = 64},
  [1520] = {.lex_state = 64},
  [1521] = {.lex_state = 17},
  [1522] = {.lex_state = 64},
  [1523] = {.lex_state = 17},
  [1524] = {.lex_state = 3, .external_lex_state = 7},
  [1525] = {.lex_state = 64},
  [1526] = {.lex_state = 64},
  [1527] = {.lex_state = 64},
  [1528] = {.lex_state = 64},
  [1529] = {.lex_state = 64},
  [1530] = {.lex_state = 64, .external_lex_state = 2},
  [1531] = {.lex_state = 64, .external_lex_state = 2},
  [1532] = {.lex_state = 64, .external_lex_state = 2},
  [1533] = {.lex_state = 64, .external_lex_state = 2},
  [1534] = {.lex_state = 8, .external_lex_state = 8},
  [1535] = {.lex_state = 64, .external_lex_state = 2},
  [1536] = {.lex_state = 8, .external_lex_state = 9},
  [1537] = {.lex_state = 8, .external_lex_state = 8},
  [1538] = {.lex_state = 17},
  [1539] = {.lex_state = 8, .external_lex_state = 9},
  [1540] = {.lex_state = 64, .external_lex_state = 2},
  [1541] = {.lex_state = 64},
  [1542] = {.lex_state = 64},
  [1543] = {.lex_state = 64, .external_lex_state = 2},
  [1544] = {.lex_state = 64, .external_lex_state = 2},
  [1545] = {.lex_state = 64},
  [1546] = {.lex_state = 64},
  [1547] = {.lex_state = 64, .external_lex_state = 2},
  [1548] = {.lex_state = 64},
  [1549] = {.lex_state = 64},
  [1550] = {.lex_state = 64},
  [1551] = {.lex_state = 64},
  [1552] = {.lex_state = 64},
  [1553] = {.lex_state = 64},
  [1554] = {.lex_state = 64},
  [1555] = {.lex_state = 64},
  [1556] = {.lex_state = 64},
  [1557] = {.lex_state = 64},
  [1558] = {.lex_state = 64},
  [1559] = {.lex_state = 64},
  [1560] = {.lex_state = 64},
  [1561] = {.lex_state = 64},
  [1562] = {.lex_state = 64},
  [1563] = {.lex_state = 64},
  [1564] = {.lex_state = 64},
  [1565] = {.lex_state = 64, .external_lex_state = 2},
  [1566] = {.lex_state = 3, .external_lex_state = 3},
  [1567] = {.lex_state = 64, .external_lex_state = 2},
  [1568] = {.lex_state = 64, .external_lex_state = 2},
  [1569] = {.lex_state = 64},
  [1570] = {.lex_state = 64},
  [1571] = {.lex_state = 64, .external_lex_state = 2},
  [1572] = {.lex_state = 64, .external_lex_state = 2},
  [1573] = {.lex_state = 64, .external_lex_state = 2},
  [1574] = {.lex_state = 64},
  [1575] = {.lex_state = 3, .external_lex_state = 3},
  [1576] = {.lex_state = 64, .external_lex_state = 2},
  [1577] = {.lex_state = 64},
  [1578] = {.lex_state = 10},
  [1579] = {.lex_state = 64, .external_lex_state = 2},
  [1580] = {.lex_state = 64},
  [1581] = {.lex_state = 3, .external_lex_state = 3},
  [1582] = {.lex_state = 65},
  [1583] = {.lex_state = 64, .external_lex_state = 2},
  [1584] = {.lex_state = 64, .external_lex_state = 2},
  [1585] = {.lex_state = 64},
  [1586] = {.lex_state = 64},
  [1587] = {.lex_state = 64, .external_lex_state = 2},
  [1588] = {.lex_state = 3, .external_lex_state = 3},
  [1589] = {.lex_state = 64},
  [1590] = {.lex_state = 64},
  [1591] = {.lex_state = 64},
  [1592] = {.lex_state = 64},
  [1593] = {.lex_state = 10},
  [1594] = {.lex_state = 64, .external_lex_state = 2},
  [1595] = {.lex_state = 10},
  [1596] = {.lex_state = 17},
  [1597] = {.lex_state = 64},
  [1598] = {.lex_state = 3, .external_lex_state = 3},
  [1599] = {.lex_state = 3, .external_lex_state = 3},
  [1600] = {.lex_state = 64},
  [1601] = {.lex_state = 64, .external_lex_state = 2},
  [1602] = {.lex_state = 64},
  [1603] = {.lex_state = 3, .external_lex_state = 3},
  [1604] = {.lex_state = 3, .external_lex_state = 3},
  [1605] = {.lex_state = 64, .external_lex_state = 2},
  [1606] = {.lex_state = 64},
  [1607] = {.lex_state = 64, .external_lex_state = 2},
  [1608] = {.lex_state = 64, .external_lex_state = 2},
  [1609] = {.lex_state = 3, .external_lex_state = 3},
  [1610] = {.lex_state = 64, .external_lex_state = 2},
  [1611] = {.lex_state = 64},
  [1612] = {.lex_state = 3, .external_lex_state = 3},
  [1613] = {.lex_state = 3, .external_lex_state = 3},
  [1614] = {.lex_state = 64},
  [1615] = {.lex_state = 64},
  [1616] = {.lex_state = 64, .external_lex_state = 2},
  [1617] = {.lex_state = 64},
  [1618] = {.lex_state = 64, .external_lex_state = 2},
  [1619] = {.lex_state = 8, .external_lex_state = 6},
  [1620] = {.lex_state = 64},
  [1621] = {.lex_state = 64},
  [1622] = {.lex_state = 64},
  [1623] = {.lex_state = 19},
  [1624] = {.lex_state = 64},
  [1625] = {.lex_state = 19},
  [1626] = {.lex_state = 64},
  [1627] = {.lex_state = 64},
  [1628] = {.lex_state = 64},
  [1629] = {.lex_state = 64},
  [1630] = {.lex_state = 64},
  [1631] = {.lex_state = 64},
  [1632] = {.lex_state = 64},
  [1633] = {.lex_state = 8, .external_lex_state = 6},
  [1634] = {.lex_state = 64},
  [1635] = {.lex_state = 64},
  [1636] = {.lex_state = 64},
  [1637] = {.lex_state = 64},
  [1638] = {.lex_state = 64},
  [1639] = {.lex_state = 64},
  [1640] = {.lex_state = 64},
  [1641] = {.lex_state = 8, .external_lex_state = 6},
  [1642] = {.lex_state = 8, .external_lex_state = 6},
  [1643] = {.lex_state = 64},
  [1644] = {.lex_state = 8, .external_lex_state = 6},
  [1645] = {.lex_state = 64},
  [1646] = {.lex_state = 64},
  [1647] = {.lex_state = 65},
  [1648] = {.lex_state = 8, .external_lex_state = 5},
  [1649] = {.lex_state = 64},
  [1650] = {.lex_state = 19},
  [1651] = {.lex_state = 64},
  [1652] = {.lex_state = 64},
  [1653] = {.lex_state = 64},
  [1654] = {.lex_state = 8, .external_lex_state = 5},
  [1655] = {.lex_state = 8, .external_lex_state = 5},
  [1656] = {.lex_state = 64},
  [1657] = {.lex_state = 64},
  [1658] = {.lex_state = 8, .external_lex_state = 6},
  [1659] = {.lex_state = 8, .external_lex_state = 5},
  [1660] = {.lex_state = 64},
  [1661] = {.lex_state = 64},
  [1662] = {.lex_state = 64},
  [1663] = {.lex_state = 8, .external_lex_state = 5},
  [1664] = {.lex_state = 64},
  [1665] = {.lex_state = 64},
  [1666] = {.lex_state = 64},
  [1667] = {.lex_state = 8, .external_lex_state = 6},
  [1668] = {.lex_state = 64},
  [1669] = {.lex_state = 64},
  [1670] = {.lex_state = 64},
  [1671] = {.lex_state = 64},
  [1672] = {.lex_state = 8, .external_lex_state = 5},
  [1673] = {.lex_state = 8, .external_lex_state = 5},
  [1674] = {.lex_state = 8, .external_lex_state = 5},
  [1675] = {.lex_state = 8, .external_lex_state = 6},
  [1676] = {.lex_state = 64},
  [1677] = {.lex_state = 64},
  [1678] = {.lex_state = 8, .external_lex_state = 6},
  [1679] = {.lex_state = 64},
  [1680] = {.lex_state = 64},
  [1681] = {.lex_state = 64},
  [1682] = {.lex_state = 8, .external_lex_state = 5},
  [1683] = {.lex_state = 64},
  [1684] = {.lex_state = 64},
  [1685] = {.lex_state = 8, .external_lex_state = 5},
  [1686] = {.lex_state = 64},
  [1687] = {.lex_state = 8, .external_lex_state = 6},
  [1688] = {.lex_state = 64},
  [1689] = {.lex_state = 66, .external_lex_state = 10},
  [1690] = {.lex_state = 64},
  [1691] = {.lex_state = 64},
  [1692] = {.lex_state = 64},
  [1693] = {.lex_state = 64},
  [1694] = {.lex_state = 64},
  [1695] = {.lex_state = 8, .external_lex_state = 5},
  [1696] = {.lex_state = 8, .external_lex_state = 6},
  [1697] = {.lex_state = 19},
  [1698] = {.lex_state = 64},
  [1699] = {.lex_state = 64},
  [1700] = {.lex_state = 64},
  [1701] = {.lex_state = 19},
  [1702] = {.lex_state = 64},
  [1703] = {.lex_state = 64},
  [1704] = {.lex_state = 64},
  [1705] = {.lex_state = 64},
  [1706] = {.lex_state = 64},
  [1707] = {.lex_state = 64},
  [1708] = {.lex_state = 64},
  [1709] = {.lex_state = 64, .external_lex_state = 2},
  [1710] = {.lex_state = 15},
  [1711] = {.lex_state = 64},
  [1712] = {.lex_state = 15},
  [1713] = {.lex_state = 64},
  [1714] = {.lex_state = 64, .external_lex_state = 2},
  [1715] = {.lex_state = 64},
  [1716] = {.lex_state = 64},
  [1717] = {.lex_state = 64, .external_lex_state = 2},
  [1718] = {.lex_state = 64, .external_lex_state = 2},
  [1719] = {.lex_state = 64, .external_lex_state = 2},
  [1720] = {.lex_state = 64, .external_lex_state = 2},
  [1721] = {.lex_state = 64},
  [1722] = {.lex_state = 64, .external_lex_state = 2},
  [1723] = {.lex_state = 64, .external_lex_state = 2},
  [1724] = {.lex_state = 64, .external_lex_state = 2},
  [1725] = {.lex_state = 64, .external_lex_state = 2},
  [1726] = {.lex_state = 64, .external_lex_state = 2},
  [1727] = {.lex_state = 64, .external_lex_state = 2},
  [1728] = {.lex_state = 15},
  [1729] = {.lex_state = 66, .external_lex_state = 10},
  [1730] = {.lex_state = 64, .external_lex_state = 2},
  [1731] = {.lex_state = 15},
  [1732] = {.lex_state = 64, .external_lex_state = 2},
  [1733] = {.lex_state = 64, .external_lex_state = 2},
  [1734] = {.lex_state = 64, .external_lex_state = 2},
  [1735] = {.lex_state = 15},
  [1736] = {.lex_state = 64},
  [1737] = {.lex_state = 64},
  [1738] = {.lex_state = 15},
  [1739] = {.lex_state = 64},
  [1740] = {.lex_state = 64},
  [1741] = {.lex_state = 15},
  [1742] = {.lex_state = 15},
  [1743] = {.lex_state = 64, .external_lex_state = 2},
  [1744] = {.lex_state = 64},
  [1745] = {.lex_state = 15},
  [1746] = {.lex_state = 64, .external_lex_state = 2},
  [1747] = {.lex_state = 64},
  [1748] = {.lex_state = 15},
  [1749] = {.lex_state = 66},
  [1750] = {.lex_state = 64},
  [1751] = {.lex_state = 15},
  [1752] = {.lex_state = 64, .external_lex_state = 2},
  [1753] = {.lex_state = 64, .external_lex_state = 2},
  [1754] = {.lex_state = 64, .external_lex_state = 2},
  [1755] = {.lex_state = 15},
  [1756] = {.lex_state = 64, .external_lex_state = 2},
  [1757] = {.lex_state = 64, .external_lex_state = 2},
  [1758] = {.lex_state = 64, .external_lex_state = 2},
  [1759] = {.lex_state = 64},
  [1760] = {.lex_state = 64},
  [1761] = {.lex_state = 64, .external_lex_state = 2},
  [1762] = {.lex_state = 64, .external_lex_state = 2},
  [1763] = {.lex_state = 64, .external_lex_state = 2},
  [1764] = {.lex_state = 64, .external_lex_state = 2},
  [1765] = {.lex_state = 64},
  [1766] = {.lex_state = 64, .external_lex_state = 2},
  [1767] = {.lex_state = 64, .external_lex_state = 2},
  [1768] = {.lex_state = 64, .external_lex_state = 2},
  [1769] = {.lex_state = 64, .external_lex_state = 2},
  [1770] = {.lex_state = 66},
  [1771] = {.lex_state = 65},
  [1772] = {.lex_state = 65},
  [1773] = {.lex_state = 64},
  [1774] = {.lex_state = 65},
  [1775] = {.lex_state = 64, .external_lex_state = 2},
  [1776] = {.lex_state = 64},
  [1777] = {.lex_state = 64},
  [1778] = {.lex_state = 64, .external_lex_state = 2},
  [1779] = {.lex_state = 64},
  [1780] = {.lex_state = 15},
  [1781] = {.lex_state = 64, .external_lex_state = 2},
  [1782] = {.lex_state = 64},
  [1783] = {.lex_state = 64},
  [1784] = {.lex_state = 64, .external_lex_state = 2},
  [1785] = {.lex_state = 64, .external_lex_state = 2},
  [1786] = {.lex_state = 64, .external_lex_state = 2},
  [1787] = {.lex_state = 64, .external_lex_state = 2},
  [1788] = {.lex_state = 64},
  [1789] = {.lex_state = 64},
  [1790] = {.lex_state = 15},
  [1791] = {.lex_state = 64},
  [1792] = {.lex_state = 64},
  [1793] = {.lex_state = 15},
  [1794] = {.lex_state = 64},
  [1795] = {.lex_state = 64, .external_lex_state = 2},
  [1796] = {.lex_state = 64, .external_lex_state = 2},
  [1797] = {.lex_state = 64, .external_lex_state = 2},
  [1798] = {.lex_state = 15},
  [1799] = {.lex_state = 64, .external_lex_state = 2},
  [1800] = {.lex_state = 64},
  [1801] = {.lex_state = 64},
  [1802] = {.lex_state = 65},
  [1803] = {.lex_state = 64},
  [1804] = {.lex_state = 64, .external_lex_state = 2},
  [1805] = {.lex_state = 15},
  [1806] = {.lex_state = 64},
  [1807] = {.lex_state = 64, .external_lex_state = 2},
  [1808] = {.lex_state = 64},
  [1809] = {.lex_state = 64, .external_lex_state = 2},
  [1810] = {.lex_state = 64, .external_lex_state = 2},
  [1811] = {.lex_state = 64},
  [1812] = {.lex_state = 64},
  [1813] = {.lex_state = 15},
  [1814] = {.lex_state = 64},
  [1815] = {.lex_state = 15},
  [1816] = {.lex_state = 15},
  [1817] = {.lex_state = 15},
  [1818] = {.lex_state = 64, .external_lex_state = 2},
  [1819] = {.lex_state = 64, .external_lex_state = 2},
  [1820] = {.lex_state = 64, .external_lex_state = 2},
  [1821] = {.lex_state = 64, .external_lex_state = 2},
  [1822] = {.lex_state = 64, .external_lex_state = 2},
  [1823] = {.lex_state = 64, .external_lex_state = 2},
  [1824] = {.lex_state = 15},
  [1825] = {.lex_state = 65},
  [1826] = {.lex_state = 64},
  [1827] = {.lex_state = 65},
  [1828] = {.lex_state = 65},
  [1829] = {.lex_state = 64, .external_lex_state = 2},
  [1830] = {.lex_state = 65},
  [1831] = {.lex_state = 64, .external_lex_state = 2},
  [1832] = {.lex_state = 64},
  [1833] = {.lex_state = 66, .external_lex_state = 10},
  [1834] = {.lex_state = 64, .external_lex_state = 2},
  [1835] = {.lex_state = 64},
  [1836] = {.lex_state = 64},
  [1837] = {.lex_state = 64},
  [1838] = {.lex_state = 64},
  [1839] = {.lex_state = 64, .external_lex_state = 2},
  [1840] = {.lex_state = 64},
  [1841] = {.lex_state = 64, .external_lex_state = 2},
  [1842] = {.lex_state = 64, .external_lex_state = 2},
  [1843] = {.lex_state = 64, .external_lex_state = 2},
  [1844] = {.lex_state = 15},
  [1845] = {.lex_state = 3, .external_lex_state = 11},
  [1846] = {.lex_state = 64},
  [1847] = {.lex_state = 64, .external_lex_state = 2},
  [1848] = {.lex_state = 64},
  [1849] = {.lex_state = 64, .external_lex_state = 2},
  [1850] = {.lex_state = 64},
  [1851] = {.lex_state = 64},
  [1852] = {.lex_state = 64},
  [1853] = {.lex_state = 64},
  [1854] = {.lex_state = 64},
  [1855] = {.lex_state = 64},
  [1856] = {.lex_state = 64},
  [1857] = {.lex_state = 64},
  [1858] = {.lex_state = 64},
  [1859] = {.lex_state = 64},
  [1860] = {.lex_state = 64, .external_lex_state = 2},
  [1861] = {.lex_state = 64},
  [1862] = {.lex_state = 64, .external_lex_state = 2},
  [1863] = {.lex_state = 64, .external_lex_state = 2},
  [1864] = {.lex_state = 64, .external_lex_state = 2},
  [1865] = {.lex_state = 64},
  [1866] = {.lex_state = 64},
  [1867] = {.lex_state = 64, .external_lex_state = 2},
  [1868] = {.lex_state = 64},
  [1869] = {.lex_state = 64},
  [1870] = {.lex_state = 64},
  [1871] = {.lex_state = 64},
  [1872] = {.lex_state = 64},
  [1873] = {.lex_state = 64},
  [1874] = {.lex_state = 64},
  [1875] = {.lex_state = 64, .external_lex_state = 2},
  [1876] = {.lex_state = 65},
  [1877] = {.lex_state = 64},
  [1878] = {.lex_state = 64},
  [1879] = {.lex_state = 64},
  [1880] = {.lex_state = 64},
  [1881] = {.lex_state = 64},
  [1882] = {.lex_state = 64, .external_lex_state = 2},
  [1883] = {.lex_state = 64},
  [1884] = {.lex_state = 64},
  [1885] = {.lex_state = 64},
  [1886] = {.lex_state = 3, .external_lex_state = 11},
  [1887] = {.lex_state = 65},
  [1888] = {.lex_state = 64},
  [1889] = {.lex_state = 64},
  [1890] = {.lex_state = 64},
  [1891] = {.lex_state = 64},
  [1892] = {.lex_state = 64, .external_lex_state = 2},
  [1893] = {.lex_state = 64},
  [1894] = {.lex_state = 64},
  [1895] = {.lex_state = 64},
  [1896] = {.lex_state = 64, .external_lex_state = 2},
  [1897] = {.lex_state = 64},
  [1898] = {.lex_state = 3, .external_lex_state = 12},
  [1899] = {.lex_state = 64},
  [1900] = {.lex_state = 64},
  [1901] = {.lex_state = 64},
  [1902] = {.lex_state = 64},
  [1903] = {.lex_state = 64, .external_lex_state = 2},
  [1904] = {.lex_state = 64},
  [1905] = {.lex_state = 19},
  [1906] = {.lex_state = 64},
  [1907] = {.lex_state = 3, .external_lex_state = 12},
  [1908] = {.lex_state = 64, .external_lex_state = 2},
  [1909] = {.lex_state = 64},
  [1910] = {.lex_state = 64},
  [1911] = {.lex_state = 64, .external_lex_state = 2},
  [1912] = {.lex_state = 64},
  [1913] = {.lex_state = 64},
  [1914] = {.lex_state = 65},
  [1915] = {.lex_state = 64},
  [1916] = {.lex_state = 64},
  [1917] = {.lex_state = 64},
  [1918] = {.lex_state = 3, .external_lex_state = 11},
  [1919] = {.lex_state = 64},
  [1920] = {.lex_state = 64},
  [1921] = {.lex_state = 64, .external_lex_state = 2},
  [1922] = {.lex_state = 64, .external_lex_state = 2},
  [1923] = {.lex_state = 66, .external_lex_state = 10},
  [1924] = {.lex_state = 64, .external_lex_state = 2},
  [1925] = {.lex_state = 64},
  [1926] = {.lex_state = 19},
  [1927] = {.lex_state = 64},
  [1928] = {.lex_state = 64, .external_lex_state = 2},
  [1929] = {.lex_state = 64},
  [1930] = {.lex_state = 64},
  [1931] = {.lex_state = 64},
  [1932] = {.lex_state = 64, .external_lex_state = 2},
  [1933] = {.lex_state = 3, .external_lex_state = 11},
  [1934] = {.lex_state = 64},
  [1935] = {.lex_state = 66},
  [1936] = {.lex_state = 64},
  [1937] = {.lex_state = 64},
  [1938] = {.lex_state = 64},
  [1939] = {.lex_state = 64},
  [1940] = {.lex_state = 64},
  [1941] = {.lex_state = 64},
  [1942] = {.lex_state = 64},
  [1943] = {.lex_state = 64},
  [1944] = {.lex_state = 64},
  [1945] = {.lex_state = 64},
  [1946] = {.lex_state = 64},
  [1947] = {.lex_state = 64},
  [1948] = {.lex_state = 64},
  [1949] = {.lex_state = 64},
  [1950] = {.lex_state = 64},
  [1951] = {.lex_state = 64},
  [1952] = {.lex_state = 16, .external_lex_state = 2},
  [1953] = {.lex_state = 64},
  [1954] = {.lex_state = 64},
  [1955] = {.lex_state = 64, .external_lex_state = 2},
  [1956] = {.lex_state = 64, .external_lex_state = 2},
  [1957] = {.lex_state = 64, .external_lex_state = 2},
  [1958] = {.lex_state = 64, .external_lex_state = 2},
  [1959] = {.lex_state = 64},
  [1960] = {.lex_state = 64},
  [1961] = {.lex_state = 64},
  [1962] = {.lex_state = 64},
  [1963] = {.lex_state = 64},
  [1964] = {.lex_state = 64},
  [1965] = {.lex_state = 64},
  [1966] = {.lex_state = 64, .external_lex_state = 2},
  [1967] = {.lex_state = 64, .external_lex_state = 2},
  [1968] = {.lex_state = 64},
  [1969] = {.lex_state = 64, .external_lex_state = 2},
  [1970] = {.lex_state = 64},
  [1971] = {.lex_state = 64},
  [1972] = {.lex_state = 64},
  [1973] = {.lex_state = 3, .external_lex_state = 11},
  [1974] = {.lex_state = 3, .external_lex_state = 11},
  [1975] = {.lex_state = 64},
  [1976] = {.lex_state = 64},
  [1977] = {.lex_state = 64},
  [1978] = {.lex_state = 64},
  [1979] = {.lex_state = 64},
  [1980] = {.lex_state = 64},
  [1981] = {.lex_state = 64},
  [1982] = {.lex_state = 64},
  [1983] = {.lex_state = 64},
  [1984] = {.lex_state = 64},
  [1985] = {.lex_state = 64},
  [1986] = {.lex_state = 64, .external_lex_state = 2},
  [1987] = {.lex_state = 64, .external_lex_state = 2},
  [1988] = {.lex_state = 64},
  [1989] = {.lex_state = 64, .external_lex_state = 2},
  [1990] = {.lex_state = 3, .external_lex_state = 12},
  [1991] = {.lex_state = 64, .external_lex_state = 2},
  [1992] = {.lex_state = 64, .external_lex_state = 2},
  [1993] = {.lex_state = 64},
  [1994] = {.lex_state = 64, .external_lex_state = 2},
  [1995] = {.lex_state = 64},
  [1996] = {.lex_state = 64},
  [1997] = {.lex_state = 19},
  [1998] = {.lex_state = 64},
  [1999] = {.lex_state = 64},
  [2000] = {.lex_state = 64},
  [2001] = {.lex_state = 64},
  [2002] = {.lex_state = 64, .external_lex_state = 2},
  [2003] = {.lex_state = 19},
  [2004] = {.lex_state = 64},
  [2005] = {.lex_state = 64, .external_lex_state = 2},
  [2006] = {.lex_state = 64, .external_lex_state = 2},
  [2007] = {.lex_state = 64},
  [2008] = {.lex_state = 64, .external_lex_state = 2},
  [2009] = {.lex_state = 64, .external_lex_state = 2},
  [2010] = {.lex_state = 64},
  [2011] = {.lex_state = 64},
  [2012] = {.lex_state = 64},
  [2013] = {.lex_state = 64, .external_lex_state = 2},
  [2014] = {.lex_state = 64, .external_lex_state = 2},
  [2015] = {.lex_state = 64},
  [2016] = {.lex_state = 64},
  [2017] = {.lex_state = 64},
  [2018] = {.lex_state = 64, .external_lex_state = 2},
  [2019] = {.lex_state = 64, .external_lex_state = 2},
  [2020] = {.lex_state = 64, .external_lex_state = 2},
  [2021] = {.lex_state = 64},
  [2022] = {.lex_state = 64},
  [2023] = {.lex_state = 64, .external_lex_state = 2},
  [2024] = {.lex_state = 64, .external_lex_state = 2},
  [2025] = {.lex_state = 64, .external_lex_state = 2},
  [2026] = {.lex_state = 64},
  [2027] = {.lex_state = 64},
  [2028] = {.lex_state = 64},
  [2029] = {.lex_state = 64, .external_lex_state = 2},
  [2030] = {.lex_state = 64},
  [2031] = {.lex_state = 64},
  [2032] = {.lex_state = 64, .external_lex_state = 2},
  [2033] = {.lex_state = 64},
  [2034] = {.lex_state = 64, .external_lex_state = 2},
  [2035] = {.lex_state = 64},
  [2036] = {.lex_state = 64, .external_lex_state = 2},
  [2037] = {.lex_state = 64},
  [2038] = {.lex_state = 64},
  [2039] = {.lex_state = 3, .external_lex_state = 11},
  [2040] = {.lex_state = 64},
  [2041] = {.lex_state = 64},
  [2042] = {.lex_state = 64, .external_lex_state = 2},
  [2043] = {.lex_state = 64},
  [2044] = {.lex_state = 64},
  [2045] = {.lex_state = 64},
  [2046] = {.lex_state = 64},
  [2047] = {.lex_state = 64},
  [2048] = {.lex_state = 64},
  [2049] = {.lex_state = 64},
  [2050] = {.lex_state = 64},
  [2051] = {.lex_state = 64},
  [2052] = {.lex_state = 64},
  [2053] = {.lex_state = 64},
  [2054] = {.lex_state = 64},
  [2055] = {.lex_state = 64, .external_lex_state = 2},
  [2056] = {.lex_state = 64},
  [2057] = {.lex_state = 64},
  [2058] = {.lex_state = 64},
  [2059] = {.lex_state = 64},
  [2060] = {.lex_state = 64},
  [2061] = {.lex_state = 3, .external_lex_state = 11},
  [2062] = {.lex_state = 3, .external_lex_state = 11},
  [2063] = {.lex_state = 64},
  [2064] = {.lex_state = 64},
  [2065] = {.lex_state = 64},
  [2066] = {.lex_state = 64},
  [2067] = {.lex_state = 64},
  [2068] = {.lex_state = 64, .external_lex_state = 2},
  [2069] = {.lex_state = 64},
  [2070] = {.lex_state = 64},
  [2071] = {.lex_state = 64},
  [2072] = {.lex_state = 64},
  [2073] = {.lex_state = 64},
  [2074] = {.lex_state = 64},
  [2075] = {.lex_state = 64},
  [2076] = {.lex_state = 64},
  [2077] = {.lex_state = 64},
  [2078] = {.lex_state = 64},
  [2079] = {.lex_state = 64},
  [2080] = {.lex_state = 64},
  [2081] = {.lex_state = 64, .external_lex_state = 2},
  [2082] = {.lex_state = 64, .external_lex_state = 2},
  [2083] = {.lex_state = 64},
  [2084] = {.lex_state = 64},
  [2085] = {.lex_state = 64},
  [2086] = {.lex_state = 64},
  [2087] = {.lex_state = 64},
  [2088] = {.lex_state = 64, .external_lex_state = 2},
  [2089] = {.lex_state = 64},
  [2090] = {.lex_state = 64},
  [2091] = {.lex_state = 64},
  [2092] = {.lex_state = 64, .external_lex_state = 2},
  [2093] = {.lex_state = 64},
  [2094] = {.lex_state = 64},
  [2095] = {.lex_state = 64},
  [2096] = {.lex_state = 64},
  [2097] = {.lex_state = 64},
  [2098] = {.lex_state = 3, .external_lex_state = 11},
  [2099] = {.lex_state = 64},
  [2100] = {.lex_state = 64},
  [2101] = {.lex_state = 64},
  [2102] = {.lex_state = 64},
  [2103] = {.lex_state = 64},
  [2104] = {.lex_state = 64},
  [2105] = {.lex_state = 64},
  [2106] = {.lex_state = 64},
  [2107] = {.lex_state = 64},
  [2108] = {.lex_state = 64, .external_lex_state = 2},
  [2109] = {.lex_state = 64},
  [2110] = {.lex_state = 64},
  [2111] = {.lex_state = 64, .external_lex_state = 2},
  [2112] = {.lex_state = 64, .external_lex_state = 2},
  [2113] = {.lex_state = 64},
  [2114] = {.lex_state = 64},
  [2115] = {.lex_state = 64},
  [2116] = {.lex_state = 64},
  [2117] = {.lex_state = 64},
  [2118] = {.lex_state = 64},
  [2119] = {.lex_state = 64},
  [2120] = {.lex_state = 64, .external_lex_state = 2},
  [2121] = {.lex_state = 64},
  [2122] = {.lex_state = 64},
  [2123] = {.lex_state = 64},
  [2124] = {.lex_state = 64},
  [2125] = {.lex_state = 64},
  [2126] = {.lex_state = 64},
  [2127] = {.lex_state = 64},
  [2128] = {.lex_state = 64},
  [2129] = {.lex_state = 64},
  [2130] = {.lex_state = 64},
  [2131] = {.lex_state = 64},
  [2132] = {.lex_state = 64},
  [2133] = {.lex_state = 64},
  [2134] = {.lex_state = 64, .external_lex_state = 2},
  [2135] = {.lex_state = 64},
  [2136] = {.lex_state = 64, .external_lex_state = 2},
  [2137] = {.lex_state = 64},
  [2138] = {.lex_state = 64},
  [2139] = {.lex_state = 64},
  [2140] = {.lex_state = 64},
  [2141] = {.lex_state = 64},
  [2142] = {.lex_state = 64},
  [2143] = {.lex_state = 64, .external_lex_state = 2},
  [2144] = {.lex_state = 64},
  [2145] = {.lex_state = 64},
  [2146] = {.lex_state = 64},
  [2147] = {.lex_state = 64},
  [2148] = {.lex_state = 64},
  [2149] = {.lex_state = 64},
  [2150] = {.lex_state = 64},
  [2151] = {.lex_state = 64},
  [2152] = {.lex_state = 64},
  [2153] = {.lex_state = 64},
  [2154] = {.lex_state = 64},
  [2155] = {.lex_state = 64},
  [2156] = {.lex_state = 64},
  [2157] = {.lex_state = 64},
  [2158] = {.lex_state = 64},
  [2159] = {.lex_state = 64},
  [2160] = {.lex_state = 64},
  [2161] = {.lex_state = 64},
  [2162] = {.lex_state = 64, .external_lex_state = 13},
  [2163] = {.lex_state = 64, .external_lex_state = 2},
  [2164] = {.lex_state = 64, .external_lex_state = 2},
  [2165] = {.lex_state = 64, .external_lex_state = 2},
  [2166] = {.lex_state = 64},
  [2167] = {.lex_state = 64},
  [2168] = {.lex_state = 64, .external_lex_state = 2},
  [2169] = {.lex_state = 64},
  [2170] = {.lex_state = 64},
  [2171] = {.lex_state = 64},
  [2172] = {.lex_state = 64},
  [2173] = {.lex_state = 64},
  [2174] = {.lex_state = 64},
  [2175] = {.lex_state = 64, .external_lex_state = 2},
  [2176] = {.lex_state = 64},
  [2177] = {.lex_state = 64},
  [2178] = {.lex_state = 64, .external_lex_state = 2},
  [2179] = {.lex_state = 64},
  [2180] = {.lex_state = 64, .external_lex_state = 2},
  [2181] = {.lex_state = 64},
  [2182] = {.lex_state = 64, .external_lex_state = 2},
  [2183] = {.lex_state = 64, .external_lex_state = 2},
  [2184] = {.lex_state = 64},
  [2185] = {.lex_state = 64, .external_lex_state = 2},
  [2186] = {.lex_state = 64},
  [2187] = {.lex_state = 64, .external_lex_state = 2},
  [2188] = {.lex_state = 64, .external_lex_state = 2},
  [2189] = {.lex_state = 64},
  [2190] = {.lex_state = 64},
  [2191] = {.lex_state = 64},
  [2192] = {.lex_state = 64},
  [2193] = {.lex_state = 64},
  [2194] = {.lex_state = 64},
  [2195] = {.lex_state = 64},
  [2196] = {.lex_state = 64, .external_lex_state = 2},
  [2197] = {.lex_state = 64},
  [2198] = {.lex_state = 64, .external_lex_state = 2},
  [2199] = {.lex_state = 64},
  [2200] = {.lex_state = 64},
  [2201] = {.lex_state = 64, .external_lex_state = 2},
  [2202] = {.lex_state = 64},
  [2203] = {.lex_state = 64, .external_lex_state = 2},
  [2204] = {.lex_state = 64, .external_lex_state = 2},
  [2205] = {.lex_state = 64},
  [2206] = {.lex_state = 64},
  [2207] = {.lex_state = 64},
  [2208] = {.lex_state = 64},
  [2209] = {.lex_state = 64, .external_lex_state = 2},
  [2210] = {.lex_state = 64},
  [2211] = {.lex_state = 64},
  [2212] = {.lex_state = 64},
  [2213] = {.lex_state = 64},
  [2214] = {.lex_state = 64, .external_lex_state = 2},
  [2215] = {.lex_state = 64},
  [2216] = {.lex_state = 64, .external_lex_state = 2},
  [2217] = {.lex_state = 64},
  [2218] = {.lex_state = 64},
  [2219] = {.lex_state = 64},
  [2220] = {.lex_state = 64},
  [2221] = {.lex_state = 64, .external_lex_state = 2},
  [2222] = {.lex_state = 64},
  [2223] = {.lex_state = 64},
  [2224] = {.lex_state = 64, .external_lex_state = 2},
  [2225] = {.lex_state = 64},
  [2226] = {.lex_state = 64},
  [2227] = {.lex_state = 64},
  [2228] = {.lex_state = 64},
  [2229] = {.lex_state = 64},
  [2230] = {.lex_state = 64, .external_lex_state = 2},
  [2231] = {.lex_state = 64},
  [2232] = {.lex_state = 64, .external_lex_state = 2},
  [2233] = {.lex_state = 64, .external_lex_state = 2},
  [2234] = {.lex_state = 64},
  [2235] = {.lex_state = 64, .external_lex_state = 2},
  [2236] = {.lex_state = 64, .external_lex_state = 2},
  [2237] = {.lex_state = 64},
  [2238] = {.lex_state = 64},
  [2239] = {.lex_state = 64},
  [2240] = {.lex_state = 64},
  [2241] = {.lex_state = 64, .external_lex_state = 2},
  [2242] = {.lex_state = 64},
  [2243] = {.lex_state = 65},
  [2244] = {.lex_state = 64},
  [2245] = {.lex_state = 65},
  [2246] = {.lex_state = 64},
  [2247] = {.lex_state = 64, .external_lex_state = 2},
  [2248] = {.lex_state = 64},
  [2249] = {.lex_state = 64},
  [2250] = {.lex_state = 64},
  [2251] = {.lex_state = 64},
  [2252] = {.lex_state = 64},
  [2253] = {.lex_state = 64},
  [2254] = {.lex_state = 64},
  [2255] = {.lex_state = 64},
  [2256] = {.lex_state = 64, .external_lex_state = 2},
  [2257] = {.lex_state = 64, .external_lex_state = 2},
  [2258] = {.lex_state = 64},
  [2259] = {.lex_state = 64},
  [2260] = {.lex_state = 64, .external_lex_state = 13},
  [2261] = {.lex_state = 64},
  [2262] = {.lex_state = 64},
  [2263] = {.lex_state = 64},
  [2264] = {.lex_state = 64},
  [2265] = {.lex_state = 64, .external_lex_state = 2},
  [2266] = {.lex_state = 64},
  [2267] = {.lex_state = 64},
  [2268] = {.lex_state = 64},
  [2269] = {.lex_state = 64},
  [2270] = {.lex_state = 64},
  [2271] = {.lex_state = 64},
  [2272] = {.lex_state = 64},
  [2273] = {.lex_state = 64},
  [2274] = {.lex_state = 64, .external_lex_state = 2},
  [2275] = {.lex_state = 64},
  [2276] = {.lex_state = 64, .external_lex_state = 2},
  [2277] = {.lex_state = 64},
  [2278] = {.lex_state = 64},
  [2279] = {.lex_state = 64, .external_lex_state = 2},
  [2280] = {.lex_state = 64, .external_lex_state = 2},
  [2281] = {.lex_state = 64},
  [2282] = {.lex_state = 64},
  [2283] = {.lex_state = 64, .external_lex_state = 2},
  [2284] = {.lex_state = 64},
  [2285] = {.lex_state = 64},
  [2286] = {.lex_state = 64},
  [2287] = {.lex_state = 64},
  [2288] = {.lex_state = 64, .external_lex_state = 2},
  [2289] = {.lex_state = 64},
  [2290] = {.lex_state = 64},
  [2291] = {.lex_state = 64},
  [2292] = {.lex_state = 64},
  [2293] = {.lex_state = 65},
  [2294] = {.lex_state = 64},
  [2295] = {.lex_state = 64},
  [2296] = {.lex_state = 64, .external_lex_state = 2},
  [2297] = {.lex_state = 65},
  [2298] = {.lex_state = 64},
  [2299] = {.lex_state = 3, .external_lex_state = 11},
  [2300] = {.lex_state = 64},
  [2301] = {.lex_state = 64},
  [2302] = {.lex_state = 64},
  [2303] = {.lex_state = 64},
  [2304] = {.lex_state = 64},
  [2305] = {.lex_state = 64},
  [2306] = {.lex_state = 64},
  [2307] = {.lex_state = 64},
  [2308] = {.lex_state = 64},
  [2309] = {.lex_state = 64},
  [2310] = {.lex_state = 64},
  [2311] = {.lex_state = 64},
  [2312] = {.lex_state = 64},
  [2313] = {.lex_state = 64},
  [2314] = {.lex_state = 64},
  [2315] = {.lex_state = 64},
  [2316] = {.lex_state = 64},
  [2317] = {.lex_state = 64},
  [2318] = {.lex_state = 64},
  [2319] = {.lex_state = 64},
  [2320] = {.lex_state = 64},
  [2321] = {.lex_state = 64},
  [2322] = {.lex_state = 64},
  [2323] = {.lex_state = 64},
  [2324] = {.lex_state = 64},
  [2325] = {.lex_state = 64},
  [2326] = {.lex_state = 64},
  [2327] = {.lex_state = 64},
  [2328] = {.lex_state = 64},
  [2329] = {.lex_state = 64},
  [2330] = {.lex_state = 3, .external_lex_state = 11},
  [2331] = {.lex_state = 64},
  [2332] = {.lex_state = 64},
  [2333] = {.lex_state = 64},
  [2334] = {.lex_state = 64},
  [2335] = {.lex_state = 64},
  [2336] = {.lex_state = 64},
  [2337] = {.lex_state = 64},
  [2338] = {.lex_state = 64},
  [2339] = {.lex_state = 64},
  [2340] = {.lex_state = 3, .external_lex_state = 11},
  [2341] = {.lex_state = 64},
  [2342] = {.lex_state = 64},
  [2343] = {.lex_state = 64},
  [2344] = {.lex_state = 64},
  [2345] = {.lex_state = 64},
  [2346] = {.lex_state = 64},
  [2347] = {.lex_state = 3, .external_lex_state = 11},
  [2348] = {.lex_state = 64, .external_lex_state = 2},
  [2349] = {.lex_state = 64},
  [2350] = {.lex_state = 64},
  [2351] = {.lex_state = 64},
  [2352] = {.lex_state = 64},
  [2353] = {.lex_state = 64},
  [2354] = {.lex_state = 64},
  [2355] = {.lex_state = 64},
  [2356] = {.lex_state = 64},
  [2357] = {.lex_state = 64},
  [2358] = {.lex_state = 64},
  [2359] = {.lex_state = 64},
  [2360] = {.lex_state = 64},
  [2361] = {.lex_state = 64},
  [2362] = {.lex_state = 64},
  [2363] = {.lex_state = 64},
  [2364] = {.lex_state = 64},
  [2365] = {.lex_state = 3, .external_lex_state = 11},
  [2366] = {.lex_state = 64},
  [2367] = {.lex_state = 3, .external_lex_state = 11},
  [2368] = {.lex_state = 64},
  [2369] = {.lex_state = 64},
  [2370] = {.lex_state = 64},
  [2371] = {.lex_state = 3, .external_lex_state = 11},
  [2372] = {.lex_state = 64},
  [2373] = {.lex_state = 64},
  [2374] = {.lex_state = 64},
  [2375] = {.lex_state = 64},
  [2376] = {.lex_state = 64},
  [2377] = {.lex_state = 64},
  [2378] = {.lex_state = 64},
  [2379] = {.lex_state = 64},
  [2380] = {.lex_state = 64},
  [2381] = {.lex_state = 64},
  [2382] = {.lex_state = 64, .external_lex_state = 2},
  [2383] = {.lex_state = 64},
  [2384] = {.lex_state = 64},
  [2385] = {.lex_state = 64},
  [2386] = {.lex_state = 64},
  [2387] = {.lex_state = 64},
  [2388] = {.lex_state = 64},
  [2389] = {.lex_state = 64},
  [2390] = {.lex_state = 64},
  [2391] = {.lex_state = 64, .external_lex_state = 14},
  [2392] = {.lex_state = 65},
  [2393] = {.lex_state = 3, .external_lex_state = 11},
  [2394] = {.lex_state = 64},
  [2395] = {.lex_state = 64},
  [2396] = {.lex_state = 64},
  [2397] = {.lex_state = 64},
  [2398] = {.lex_state = 64},
  [2399] = {.lex_state = 64},
  [2400] = {.lex_state = 64},
  [2401] = {.lex_state = 64},
  [2402] = {.lex_state = 64},
  [2403] = {.lex_state = 64},
  [2404] = {.lex_state = 64},
  [2405] = {.lex_state = 3, .external_lex_state = 11},
  [2406] = {.lex_state = 64, .external_lex_state = 2},
  [2407] = {.lex_state = 3, .external_lex_state = 11},
  [2408] = {.lex_state = 64},
  [2409] = {.lex_state = 64, .external_lex_state = 2},
  [2410] = {.lex_state = 64},
  [2411] = {.lex_state = 64},
  [2412] = {.lex_state = 64},
  [2413] = {.lex_state = 64},
  [2414] = {.lex_state = 64},
  [2415] = {.lex_state = 64},
  [2416] = {.lex_state = 64},
  [2417] = {.lex_state = 64},
  [2418] = {.lex_state = 64},
  [2419] = {.lex_state = 64},
  [2420] = {.lex_state = 64},
  [2421] = {.lex_state = 64},
  [2422] = {.lex_state = 64},
  [2423] = {.lex_state = 64},
  [2424] = {.lex_state = 64},
  [2425] = {.lex_state = 64},
  [2426] = {.lex_state = 64, .external_lex_state = 2},
  [2427] = {.lex_state = 64},
  [2428] = {.lex_state = 64},
  [2429] = {.lex_state = 64, .external_lex_state = 2},
  [2430] = {.lex_state = 64},
  [2431] = {.lex_state = 64},
  [2432] = {.lex_state = 64},
  [2433] = {.lex_state = 64},
  [2434] = {.lex_state = 64},
  [2435] = {.lex_state = 64},
  [2436] = {.lex_state = 64},
  [2437] = {.lex_state = 64},
  [2438] = {.lex_state = 64},
  [2439] = {.lex_state = 64},
  [2440] = {.lex_state = 64},
  [2441] = {.lex_state = 64},
  [2442] = {.lex_state = 64},
  [2443] = {.lex_state = 64},
  [2444] = {.lex_state = 64},
  [2445] = {.lex_state = 64},
  [2446] = {.lex_state = 64, .external_lex_state = 2},
  [2447] = {.lex_state = 64},
  [2448] = {.lex_state = 64},
  [2449] = {.lex_state = 64},
  [2450] = {.lex_state = 64},
  [2451] = {.lex_state = 64},
  [2452] = {.lex_state = 64},
  [2453] = {.lex_state = 64},
  [2454] = {.lex_state = 64},
  [2455] = {.lex_state = 64},
  [2456] = {.lex_state = 64},
  [2457] = {.lex_state = 64},
  [2458] = {.lex_state = 64},
  [2459] = {.lex_state = 64},
  [2460] = {.lex_state = 64},
  [2461] = {.lex_state = 64},
  [2462] = {.lex_state = 64},
  [2463] = {.lex_state = 64},
  [2464] = {.lex_state = 64},
  [2465] = {.lex_state = 64},
  [2466] = {.lex_state = 64},
  [2467] = {.lex_state = 64},
  [2468] = {.lex_state = 64},
  [2469] = {.lex_state = 64},
  [2470] = {.lex_state = 64},
  [2471] = {.lex_state = 64},
  [2472] = {.lex_state = 64},
  [2473] = {.lex_state = 64, .external_lex_state = 2},
  [2474] = {.lex_state = 64},
  [2475] = {.lex_state = 64},
  [2476] = {.lex_state = 64},
  [2477] = {.lex_state = 64},
  [2478] = {.lex_state = 64},
  [2479] = {.lex_state = 64},
  [2480] = {.lex_state = 64},
  [2481] = {.lex_state = 64},
  [2482] = {.lex_state = 64},
  [2483] = {.lex_state = 64},
  [2484] = {.lex_state = 65, .external_lex_state = 10},
  [2485] = {.lex_state = 64},
  [2486] = {.lex_state = 64},
  [2487] = {.lex_state = 64},
  [2488] = {.lex_state = 64},
  [2489] = {.lex_state = 64},
  [2490] = {.lex_state = 64},
  [2491] = {.lex_state = 64},
  [2492] = {.lex_state = 64},
  [2493] = {.lex_state = 64},
  [2494] = {.lex_state = 64, .external_lex_state = 2},
  [2495] = {.lex_state = 64},
  [2496] = {.lex_state = 64},
  [2497] = {.lex_state = 64},
  [2498] = {.lex_state = 64},
  [2499] = {.lex_state = 64},
  [2500] = {.lex_state = 64},
  [2501] = {.lex_state = 64},
  [2502] = {.lex_state = 64},
  [2503] = {.lex_state = 64},
  [2504] = {.lex_state = 64},
  [2505] = {.lex_state = 64},
  [2506] = {.lex_state = 64},
  [2507] = {.lex_state = 64},
  [2508] = {.lex_state = 64},
  [2509] = {.lex_state = 64},
  [2510] = {.lex_state = 64, .external_lex_state = 11},
  [2511] = {.lex_state = 64},
  [2512] = {.lex_state = 64},
  [2513] = {.lex_state = 64},
  [2514] = {.lex_state = 64},
  [2515] = {.lex_state = 64},
  [2516] = {.lex_state = 16},
  [2517] = {.lex_state = 64},
  [2518] = {.lex_state = 64},
  [2519] = {.lex_state = 64},
  [2520] = {.lex_state = 64},
  [2521] = {.lex_state = 64},
  [2522] = {.lex_state = 64},
  [2523] = {.lex_state = 64},
  [2524] = {.lex_state = 64},
  [2525] = {.lex_state = 64},
  [2526] = {.lex_state = 64},
  [2527] = {.lex_state = 64},
  [2528] = {.lex_state = 64},
  [2529] = {.lex_state = 64},
  [2530] = {.lex_state = 64},
  [2531] = {.lex_state = 64},
  [2532] = {.lex_state = 64},
  [2533] = {.lex_state = 16},
  [2534] = {.lex_state = 64},
  [2535] = {.lex_state = 64},
  [2536] = {.lex_state = 64},
  [2537] = {.lex_state = 64},
  [2538] = {.lex_state = 64},
  [2539] = {.lex_state = 64},
  [2540] = {.lex_state = 64, .external_lex_state = 13},
  [2541] = {.lex_state = 64},
  [2542] = {.lex_state = 64},
  [2543] = {.lex_state = 64},
  [2544] = {.lex_state = 64},
  [2545] = {.lex_state = 64},
  [2546] = {.lex_state = 64},
  [2547] = {.lex_state = 64},
  [2548] = {.lex_state = 64},
  [2549] = {.lex_state = 64},
  [2550] = {.lex_state = 64},
  [2551] = {.lex_state = 64},
  [2552] = {.lex_state = 64},
  [2553] = {.lex_state = 64},
  [2554] = {.lex_state = 64},
  [2555] = {.lex_state = 64},
  [2556] = {.lex_state = 64},
  [2557] = {.lex_state = 64},
  [2558] = {.lex_state = 64},
  [2559] = {.lex_state = 64},
  [2560] = {.lex_state = 64},
  [2561] = {.lex_state = 64},
  [2562] = {.lex_state = 64},
  [2563] = {.lex_state = 64},
  [2564] = {.lex_state = 64},
  [2565] = {.lex_state = 64},
  [2566] = {.lex_state = 64},
  [2567] = {.lex_state = 64},
  [2568] = {.lex_state = 64},
  [2569] = {.lex_state = 64},
  [2570] = {.lex_state = 64},
  [2571] = {.lex_state = 64, .external_lex_state = 11},
  [2572] = {.lex_state = 64, .external_lex_state = 11},
  [2573] = {.lex_state = 64},
  [2574] = {.lex_state = 64},
  [2575] = {.lex_state = 64},
  [2576] = {.lex_state = 64},
  [2577] = {.lex_state = 64},
  [2578] = {.lex_state = 64},
  [2579] = {.lex_state = 64},
  [2580] = {.lex_state = 64},
  [2581] = {.lex_state = 64},
  [2582] = {.lex_state = 10},
  [2583] = {.lex_state = 64},
  [2584] = {.lex_state = 64},
  [2585] = {.lex_state = 64},
  [2586] = {.lex_state = 64},
  [2587] = {.lex_state = 64},
  [2588] = {.lex_state = 64, .external_lex_state = 11},
  [2589] = {.lex_state = 64},
  [2590] = {.lex_state = 64},
  [2591] = {.lex_state = 64},
  [2592] = {.lex_state = 64},
  [2593] = {.lex_state = 64},
  [2594] = {.lex_state = 64},
  [2595] = {.lex_state = 64, .external_lex_state = 13},
  [2596] = {.lex_state = 64},
  [2597] = {.lex_state = 64},
  [2598] = {.lex_state = 64},
  [2599] = {.lex_state = 64},
  [2600] = {.lex_state = 64},
  [2601] = {.lex_state = 64},
  [2602] = {.lex_state = 64},
  [2603] = {.lex_state = 64, .external_lex_state = 11},
  [2604] = {.lex_state = 64},
  [2605] = {.lex_state = 64},
  [2606] = {.lex_state = 64},
  [2607] = {.lex_state = 64},
  [2608] = {.lex_state = 64},
  [2609] = {.lex_state = 64},
  [2610] = {.lex_state = 64},
  [2611] = {.lex_state = 64},
  [2612] = {.lex_state = 64},
  [2613] = {.lex_state = 64},
  [2614] = {.lex_state = 64},
  [2615] = {.lex_state = 64, .external_lex_state = 13},
  [2616] = {.lex_state = 64},
  [2617] = {.lex_state = 64},
  [2618] = {.lex_state = 64},
  [2619] = {.lex_state = 64},
  [2620] = {.lex_state = 64},
  [2621] = {.lex_state = 64},
  [2622] = {.lex_state = 64},
  [2623] = {.lex_state = 64},
  [2624] = {.lex_state = 64},
  [2625] = {.lex_state = 64},
  [2626] = {.lex_state = 64},
  [2627] = {.lex_state = 64, .external_lex_state = 11},
  [2628] = {.lex_state = 64},
  [2629] = {.lex_state = 64},
  [2630] = {.lex_state = 64},
  [2631] = {.lex_state = 64, .external_lex_state = 11},
  [2632] = {.lex_state = 64},
  [2633] = {.lex_state = 64, .external_lex_state = 11},
  [2634] = {.lex_state = 64},
  [2635] = {.lex_state = 64},
  [2636] = {.lex_state = 64},
  [2637] = {.lex_state = 64},
  [2638] = {.lex_state = 64},
  [2639] = {.lex_state = 64},
  [2640] = {.lex_state = 64},
  [2641] = {.lex_state = 64},
  [2642] = {.lex_state = 64},
  [2643] = {.lex_state = 64},
  [2644] = {.lex_state = 64},
  [2645] = {.lex_state = 64},
  [2646] = {.lex_state = 10},
  [2647] = {.lex_state = 64},
  [2648] = {.lex_state = 64},
  [2649] = {.lex_state = 64},
  [2650] = {.lex_state = 64},
  [2651] = {.lex_state = 64},
  [2652] = {.lex_state = 64},
  [2653] = {.lex_state = 64},
  [2654] = {.lex_state = 64},
  [2655] = {.lex_state = 64},
  [2656] = {.lex_state = 64},
  [2657] = {.lex_state = 64},
  [2658] = {.lex_state = 64},
  [2659] = {.lex_state = 64},
  [2660] = {.lex_state = 64},
  [2661] = {.lex_state = 64},
  [2662] = {.lex_state = 64},
  [2663] = {.lex_state = 64},
  [2664] = {.lex_state = 64},
  [2665] = {.lex_state = 64},
  [2666] = {.lex_state = 64, .external_lex_state = 13},
  [2667] = {.lex_state = 64},
  [2668] = {.lex_state = 64},
  [2669] = {.lex_state = 64},
  [2670] = {.lex_state = 64},
  [2671] = {.lex_state = 64},
  [2672] = {.lex_state = 64},
  [2673] = {.lex_state = 64},
  [2674] = {.lex_state = 64},
  [2675] = {.lex_state = 64},
  [2676] = {.lex_state = 64},
  [2677] = {.lex_state = 64},
  [2678] = {.lex_state = 64},
  [2679] = {.lex_state = 64},
  [2680] = {.lex_state = 64},
  [2681] = {.lex_state = 64},
  [2682] = {.lex_state = 64},
  [2683] = {.lex_state = 64},
  [2684] = {.lex_state = 64},
  [2685] = {.lex_state = 64},
  [2686] = {.lex_state = 64},
  [2687] = {.lex_state = 64},
  [2688] = {.lex_state = 64, .external_lex_state = 11},
  [2689] = {.lex_state = 64},
  [2690] = {.lex_state = 64},
  [2691] = {.lex_state = 64},
  [2692] = {.lex_state = 64},
  [2693] = {.lex_state = 64},
  [2694] = {.lex_state = 64},
  [2695] = {.lex_state = 64},
  [2696] = {.lex_state = 64},
  [2697] = {.lex_state = 64},
  [2698] = {.lex_state = 64},
  [2699] = {.lex_state = 64},
  [2700] = {.lex_state = 64},
  [2701] = {.lex_state = 64},
  [2702] = {.lex_state = 64},
  [2703] = {.lex_state = 64},
  [2704] = {.lex_state = 64},
  [2705] = {.lex_state = 64},
  [2706] = {.lex_state = 64},
  [2707] = {.lex_state = 64, .external_lex_state = 11},
  [2708] = {.lex_state = 64},
  [2709] = {.lex_state = 64},
  [2710] = {.lex_state = 64},
  [2711] = {.lex_state = 64},
  [2712] = {.lex_state = 64},
  [2713] = {.lex_state = 64},
  [2714] = {.lex_state = 64},
  [2715] = {.lex_state = 64},
  [2716] = {.lex_state = 64},
  [2717] = {.lex_state = 64},
  [2718] = {.lex_state = 64},
  [2719] = {.lex_state = 64},
  [2720] = {.lex_state = 64},
  [2721] = {.lex_state = 64},
  [2722] = {.lex_state = 64},
  [2723] = {.lex_state = 64},
  [2724] = {.lex_state = 64},
  [2725] = {.lex_state = 64},
  [2726] = {.lex_state = 64},
  [2727] = {.lex_state = 64},
  [2728] = {.lex_state = 64},
  [2729] = {.lex_state = 64},
  [2730] = {.lex_state = 64},
  [2731] = {.lex_state = 64},
  [2732] = {.lex_state = 64},
  [2733] = {.lex_state = 64},
  [2734] = {.lex_state = 64},
  [2735] = {.lex_state = 64},
  [2736] = {.lex_state = 64},
  [2737] = {.lex_state = 64},
  [2738] = {.lex_state = 64},
  [2739] = {.lex_state = 64},
  [2740] = {(TSStateId)(-1)},
  [2741] = {(TSStateId)(-1)},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [0] = {
    [sym_text_interpolation] = STATE(0),
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym_name] = ACTIONS(1),
    [sym_php_tag] = ACTIONS(1),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_text_token1] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [aux_sym_global_declaration_token1] = ACTIONS(1),
    [aux_sym_namespace_definition_token1] = ACTIONS(1),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1),
    [aux_sym_namespace_use_clause_token1] = ACTIONS(1),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1),
    [anon_sym_BSLASH] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [aux_sym_trait_declaration_token1] = ACTIONS(1),
    [aux_sym_interface_declaration_token1] = ACTIONS(1),
    [aux_sym_base_clause_token1] = ACTIONS(1),
    [aux_sym_enum_declaration_token1] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_string] = ACTIONS(1),
    [anon_sym_int] = ACTIONS(1),
    [aux_sym_enum_case_token1] = ACTIONS(1),
    [aux_sym_class_declaration_token1] = ACTIONS(1),
    [aux_sym_final_modifier_token1] = ACTIONS(1),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1),
    [aux_sym_class_interface_clause_token1] = ACTIONS(1),
    [anon_sym_EQ_GT] = ACTIONS(1),
    [sym_var_modifier] = ACTIONS(1),
    [aux_sym_use_instead_of_clause_token1] = ACTIONS(1),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_LPAREN2] = ACTIONS(1),
    [anon_sym_RPAREN2] = ACTIONS(1),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [sym_bottom_type] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_array] = ACTIONS(1),
    [aux_sym_primitive_type_token1] = ACTIONS(1),
    [anon_sym_iterable] = ACTIONS(1),
    [anon_sym_bool] = ACTIONS(1),
    [anon_sym_float] = ACTIONS(1),
    [anon_sym_void] = ACTIONS(1),
    [anon_sym_mixed] = ACTIONS(1),
    [anon_sym_false] = ACTIONS(1),
    [anon_sym_null] = ACTIONS(1),
    [anon_sym_true] = ACTIONS(1),
    [aux_sym_cast_type_token1] = ACTIONS(1),
    [aux_sym_cast_type_token2] = ACTIONS(1),
    [aux_sym_cast_type_token3] = ACTIONS(1),
    [aux_sym_cast_type_token4] = ACTIONS(1),
    [aux_sym_cast_type_token5] = ACTIONS(1),
    [aux_sym_cast_type_token6] = ACTIONS(1),
    [aux_sym_cast_type_token7] = ACTIONS(1),
    [aux_sym_cast_type_token8] = ACTIONS(1),
    [aux_sym_cast_type_token9] = ACTIONS(1),
    [aux_sym_cast_type_token10] = ACTIONS(1),
    [aux_sym_cast_type_token11] = ACTIONS(1),
    [aux_sym_cast_type_token12] = ACTIONS(1),
    [aux_sym_echo_statement_token1] = ACTIONS(1),
    [aux_sym_exit_statement_token1] = ACTIONS(1),
    [anon_sym_unset] = ACTIONS(1),
    [aux_sym_declare_statement_token1] = ACTIONS(1),
    [aux_sym_declare_statement_token2] = ACTIONS(1),
    [anon_sym_ticks] = ACTIONS(1),
    [anon_sym_encoding] = ACTIONS(1),
    [anon_sym_strict_types] = ACTIONS(1),
    [sym_float] = ACTIONS(1),
    [aux_sym_try_statement_token1] = ACTIONS(1),
    [aux_sym_catch_clause_token1] = ACTIONS(1),
    [aux_sym_finally_clause_token1] = ACTIONS(1),
    [aux_sym_goto_statement_token1] = ACTIONS(1),
    [aux_sym_continue_statement_token1] = ACTIONS(1),
    [aux_sym_break_statement_token1] = ACTIONS(1),
    [sym_integer] = ACTIONS(1),
    [aux_sym_return_statement_token1] = ACTIONS(1),
    [aux_sym_throw_expression_token1] = ACTIONS(1),
    [aux_sym_while_statement_token1] = ACTIONS(1),
    [aux_sym_while_statement_token2] = ACTIONS(1),
    [aux_sym_do_statement_token1] = ACTIONS(1),
    [aux_sym_for_statement_token1] = ACTIONS(1),
    [aux_sym_for_statement_token2] = ACTIONS(1),
    [aux_sym_foreach_statement_token1] = ACTIONS(1),
    [aux_sym_foreach_statement_token2] = ACTIONS(1),
    [aux_sym_if_statement_token1] = ACTIONS(1),
    [aux_sym_if_statement_token2] = ACTIONS(1),
    [aux_sym_else_if_clause_token1] = ACTIONS(1),
    [aux_sym_else_clause_token1] = ACTIONS(1),
    [aux_sym_match_expression_token1] = ACTIONS(1),
    [aux_sym_match_default_expression_token1] = ACTIONS(1),
    [aux_sym_switch_statement_token1] = ACTIONS(1),
    [aux_sym_switch_block_token1] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [aux_sym_clone_expression_token1] = ACTIONS(1),
    [anon_sym_COLON_COLON] = ACTIONS(1),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1),
    [anon_sym_DASH_DASH] = ACTIONS(1),
    [anon_sym_PLUS_PLUS] = ACTIONS(1),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1),
    [anon_sym_STAR_EQ] = ACTIONS(1),
    [anon_sym_SLASH_EQ] = ACTIONS(1),
    [anon_sym_PERCENT_EQ] = ACTIONS(1),
    [anon_sym_PLUS_EQ] = ACTIONS(1),
    [anon_sym_DASH_EQ] = ACTIONS(1),
    [anon_sym_GT_GT_EQ] = ACTIONS(1),
    [anon_sym_AMP_EQ] = ACTIONS(1),
    [anon_sym_CARET_EQ] = ACTIONS(1),
    [anon_sym_PIPE_EQ] = ACTIONS(1),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(1),
    [anon_sym_DASH_GT] = ACTIONS(1),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(1),
    [aux_sym__list_destructing_token1] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_self] = ACTIONS(1),
    [anon_sym_parent] = ACTIONS(1),
    [aux_sym__argument_name_token1] = ACTIONS(1),
    [aux_sym__argument_name_token2] = ACTIONS(1),
    [anon_sym_POUND_LBRACK] = ACTIONS(1),
    [aux_sym_encapsed_string_token1] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [aux_sym_string_token1] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [aux_sym_string_token2] = ACTIONS(1),
    [anon_sym_DQUOTE2] = ACTIONS(1),
    [anon_sym_SQUOTE2] = ACTIONS(1),
    [anon_sym_BQUOTE] = ACTIONS(1),
    [anon_sym_DOLLAR] = ACTIONS(1),
    [aux_sym_yield_expression_token1] = ACTIONS(1),
    [aux_sym_yield_expression_token2] = ACTIONS(1),
    [aux_sym_binary_expression_token1] = ACTIONS(1),
    [anon_sym_QMARK_QMARK] = ACTIONS(1),
    [anon_sym_STAR_STAR] = ACTIONS(1),
    [aux_sym_binary_expression_token2] = ACTIONS(1),
    [aux_sym_binary_expression_token3] = ACTIONS(1),
    [aux_sym_binary_expression_token4] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ] = ACTIONS(1),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_GT_EQ] = ACTIONS(1),
    [anon_sym_GT_GT] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [aux_sym_include_expression_token1] = ACTIONS(1),
    [aux_sym_include_once_expression_token1] = ACTIONS(1),
    [aux_sym_require_expression_token1] = ACTIONS(1),
    [aux_sym_require_once_expression_token1] = ACTIONS(1),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(1),
    [sym_encapsed_string_chars] = ACTIONS(1),
    [sym_encapsed_string_chars_after_variable] = ACTIONS(1),
    [sym_execution_string_chars] = ACTIONS(1),
    [sym_execution_string_chars_after_variable] = ACTIONS(1),
    [sym_encapsed_string_chars_heredoc] = ACTIONS(1),
    [sym_encapsed_string_chars_after_variable_heredoc] = ACTIONS(1),
    [sym__eof] = ACTIONS(1),
    [sym_heredoc_start] = ACTIONS(1),
    [sym_heredoc_end] = ACTIONS(1),
    [sym_nowdoc_string] = ACTIONS(1),
    [sym_sentinel_error] = ACTIONS(1),
  },
  [1] = {
    [sym_program] = STATE(2672),
    [sym_text_interpolation] = STATE(1),
    [sym_text] = STATE(2392),
    [aux_sym_text_repeat1] = STATE(1770),
    [ts_builtin_sym_end] = ACTIONS(7),
    [sym_php_tag] = ACTIONS(9),
    [anon_sym_QMARK_GT] = ACTIONS(11),
    [aux_sym_text_token1] = ACTIONS(13),
    [aux_sym_text_token2] = ACTIONS(15),
    [sym_comment] = ACTIONS(5),
  },
  [2] = {
    [sym_text_interpolation] = STATE(2),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [ts_builtin_sym_end] = ACTIONS(17),
    [sym_name] = ACTIONS(19),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(22),
    [aux_sym_function_static_declaration_token1] = ACTIONS(25),
    [aux_sym_global_declaration_token1] = ACTIONS(28),
    [aux_sym_namespace_definition_token1] = ACTIONS(31),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(34),
    [aux_sym__namespace_use_type_token1] = ACTIONS(37),
    [aux_sym__namespace_use_type_token2] = ACTIONS(40),
    [anon_sym_BSLASH] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(46),
    [anon_sym_RBRACE] = ACTIONS(17),
    [aux_sym_trait_declaration_token1] = ACTIONS(49),
    [aux_sym_interface_declaration_token1] = ACTIONS(52),
    [aux_sym_enum_declaration_token1] = ACTIONS(55),
    [aux_sym_enum_case_token1] = ACTIONS(58),
    [aux_sym_class_declaration_token1] = ACTIONS(60),
    [aux_sym_final_modifier_token1] = ACTIONS(63),
    [aux_sym_abstract_modifier_token1] = ACTIONS(66),
    [aux_sym_readonly_modifier_token1] = ACTIONS(69),
    [sym_var_modifier] = ACTIONS(72),
    [aux_sym_visibility_modifier_token1] = ACTIONS(75),
    [aux_sym_visibility_modifier_token2] = ACTIONS(75),
    [aux_sym_visibility_modifier_token3] = ACTIONS(75),
    [anon_sym_LPAREN2] = ACTIONS(78),
    [aux_sym__arrow_function_header_token1] = ACTIONS(81),
    [aux_sym_cast_type_token1] = ACTIONS(84),
    [aux_sym_echo_statement_token1] = ACTIONS(87),
    [aux_sym_exit_statement_token1] = ACTIONS(90),
    [anon_sym_unset] = ACTIONS(93),
    [aux_sym_declare_statement_token1] = ACTIONS(96),
    [aux_sym_declare_statement_token2] = ACTIONS(58),
    [sym_float] = ACTIONS(99),
    [aux_sym_try_statement_token1] = ACTIONS(102),
    [aux_sym_goto_statement_token1] = ACTIONS(105),
    [aux_sym_continue_statement_token1] = ACTIONS(108),
    [aux_sym_break_statement_token1] = ACTIONS(111),
    [sym_integer] = ACTIONS(99),
    [aux_sym_return_statement_token1] = ACTIONS(114),
    [aux_sym_throw_expression_token1] = ACTIONS(117),
    [aux_sym_while_statement_token1] = ACTIONS(120),
    [aux_sym_while_statement_token2] = ACTIONS(58),
    [aux_sym_do_statement_token1] = ACTIONS(123),
    [aux_sym_for_statement_token1] = ACTIONS(126),
    [aux_sym_for_statement_token2] = ACTIONS(58),
    [aux_sym_foreach_statement_token1] = ACTIONS(129),
    [aux_sym_foreach_statement_token2] = ACTIONS(58),
    [aux_sym_if_statement_token1] = ACTIONS(132),
    [aux_sym_if_statement_token2] = ACTIONS(58),
    [aux_sym_match_expression_token1] = ACTIONS(135),
    [aux_sym_match_default_expression_token1] = ACTIONS(58),
    [aux_sym_switch_statement_token1] = ACTIONS(138),
    [aux_sym_switch_block_token1] = ACTIONS(58),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [anon_sym_TILDE] = ACTIONS(144),
    [anon_sym_BANG] = ACTIONS(144),
    [anon_sym_AT] = ACTIONS(147),
    [aux_sym_clone_expression_token1] = ACTIONS(150),
    [aux_sym_print_intrinsic_token1] = ACTIONS(153),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(156),
    [anon_sym_DASH_DASH] = ACTIONS(159),
    [anon_sym_PLUS_PLUS] = ACTIONS(159),
    [aux_sym__list_destructing_token1] = ACTIONS(162),
    [anon_sym_LBRACK] = ACTIONS(165),
    [anon_sym_self] = ACTIONS(168),
    [anon_sym_parent] = ACTIONS(168),
    [aux_sym__argument_name_token1] = ACTIONS(171),
    [aux_sym__argument_name_token2] = ACTIONS(174),
    [anon_sym_POUND_LBRACK] = ACTIONS(177),
    [aux_sym_encapsed_string_token1] = ACTIONS(180),
    [anon_sym_DQUOTE] = ACTIONS(180),
    [aux_sym_string_token1] = ACTIONS(183),
    [anon_sym_SQUOTE] = ACTIONS(183),
    [anon_sym_LT_LT_LT] = ACTIONS(186),
    [anon_sym_BQUOTE] = ACTIONS(189),
    [anon_sym_DOLLAR] = ACTIONS(192),
    [aux_sym_yield_expression_token1] = ACTIONS(195),
    [aux_sym_include_expression_token1] = ACTIONS(198),
    [aux_sym_include_once_expression_token1] = ACTIONS(201),
    [aux_sym_require_expression_token1] = ACTIONS(204),
    [aux_sym_require_once_expression_token1] = ACTIONS(207),
    [sym_comment] = ACTIONS(5),
  },
  [3] = {
    [sym_text_interpolation] = STATE(3),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [anon_sym_RBRACE] = ACTIONS(230),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_enum_case_token1] = ACTIONS(238),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_match_default_expression_token1] = ACTIONS(238),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [aux_sym_switch_block_token1] = ACTIONS(238),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [4] = {
    [sym_text_interpolation] = STATE(4),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(3),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [anon_sym_RBRACE] = ACTIONS(340),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_enum_case_token1] = ACTIONS(342),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_match_default_expression_token1] = ACTIONS(342),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [aux_sym_switch_block_token1] = ACTIONS(342),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [5] = {
    [sym_text_interpolation] = STATE(5),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [anon_sym_RBRACE] = ACTIONS(344),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_enum_case_token1] = ACTIONS(346),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_match_default_expression_token1] = ACTIONS(346),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [aux_sym_switch_block_token1] = ACTIONS(346),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [6] = {
    [sym_text_interpolation] = STATE(6),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(5),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [anon_sym_RBRACE] = ACTIONS(348),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_enum_case_token1] = ACTIONS(350),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_match_default_expression_token1] = ACTIONS(350),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [aux_sym_switch_block_token1] = ACTIONS(350),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [7] = {
    [sym_text_interpolation] = STATE(7),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_while_statement_token2] = ACTIONS(352),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_foreach_statement_token2] = ACTIONS(352),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_if_statement_token2] = ACTIONS(352),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [8] = {
    [sym_text_interpolation] = STATE(8),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(7),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_while_statement_token2] = ACTIONS(354),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_foreach_statement_token2] = ACTIONS(354),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_if_statement_token2] = ACTIONS(354),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [9] = {
    [sym_text_interpolation] = STATE(9),
    [sym_statement] = STATE(465),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_colon_block] = STATE(2541),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(466),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(358),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [10] = {
    [sym_text_interpolation] = STATE(10),
    [sym_statement] = STATE(2149),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_colon_block] = STATE(2593),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2150),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(358),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(392),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(404),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(408),
    [aux_sym_foreach_statement_token1] = ACTIONS(410),
    [aux_sym_if_statement_token1] = ACTIONS(412),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [11] = {
    [sym_text_interpolation] = STATE(11),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(11),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(19),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(22),
    [aux_sym_function_static_declaration_token1] = ACTIONS(25),
    [aux_sym_global_declaration_token1] = ACTIONS(28),
    [aux_sym_namespace_definition_token1] = ACTIONS(31),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(34),
    [aux_sym__namespace_use_type_token1] = ACTIONS(37),
    [aux_sym__namespace_use_type_token2] = ACTIONS(40),
    [anon_sym_BSLASH] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(46),
    [aux_sym_trait_declaration_token1] = ACTIONS(49),
    [aux_sym_interface_declaration_token1] = ACTIONS(52),
    [aux_sym_enum_declaration_token1] = ACTIONS(55),
    [aux_sym_class_declaration_token1] = ACTIONS(60),
    [aux_sym_final_modifier_token1] = ACTIONS(63),
    [aux_sym_abstract_modifier_token1] = ACTIONS(66),
    [aux_sym_readonly_modifier_token1] = ACTIONS(69),
    [sym_var_modifier] = ACTIONS(72),
    [aux_sym_visibility_modifier_token1] = ACTIONS(75),
    [aux_sym_visibility_modifier_token2] = ACTIONS(75),
    [aux_sym_visibility_modifier_token3] = ACTIONS(75),
    [anon_sym_LPAREN2] = ACTIONS(78),
    [aux_sym__arrow_function_header_token1] = ACTIONS(81),
    [aux_sym_cast_type_token1] = ACTIONS(84),
    [aux_sym_echo_statement_token1] = ACTIONS(87),
    [aux_sym_exit_statement_token1] = ACTIONS(90),
    [anon_sym_unset] = ACTIONS(93),
    [aux_sym_declare_statement_token1] = ACTIONS(418),
    [sym_float] = ACTIONS(99),
    [aux_sym_try_statement_token1] = ACTIONS(102),
    [aux_sym_goto_statement_token1] = ACTIONS(105),
    [aux_sym_continue_statement_token1] = ACTIONS(108),
    [aux_sym_break_statement_token1] = ACTIONS(111),
    [sym_integer] = ACTIONS(99),
    [aux_sym_return_statement_token1] = ACTIONS(114),
    [aux_sym_throw_expression_token1] = ACTIONS(117),
    [aux_sym_while_statement_token1] = ACTIONS(421),
    [aux_sym_do_statement_token1] = ACTIONS(123),
    [aux_sym_for_statement_token1] = ACTIONS(424),
    [aux_sym_foreach_statement_token1] = ACTIONS(427),
    [aux_sym_if_statement_token1] = ACTIONS(430),
    [aux_sym_if_statement_token2] = ACTIONS(58),
    [aux_sym_else_if_clause_token1] = ACTIONS(58),
    [aux_sym_else_clause_token1] = ACTIONS(58),
    [aux_sym_match_expression_token1] = ACTIONS(135),
    [aux_sym_switch_statement_token1] = ACTIONS(138),
    [anon_sym_PLUS] = ACTIONS(141),
    [anon_sym_DASH] = ACTIONS(141),
    [anon_sym_TILDE] = ACTIONS(144),
    [anon_sym_BANG] = ACTIONS(144),
    [anon_sym_AT] = ACTIONS(147),
    [aux_sym_clone_expression_token1] = ACTIONS(150),
    [aux_sym_print_intrinsic_token1] = ACTIONS(153),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(156),
    [anon_sym_DASH_DASH] = ACTIONS(159),
    [anon_sym_PLUS_PLUS] = ACTIONS(159),
    [aux_sym__list_destructing_token1] = ACTIONS(162),
    [anon_sym_LBRACK] = ACTIONS(165),
    [anon_sym_self] = ACTIONS(168),
    [anon_sym_parent] = ACTIONS(168),
    [aux_sym__argument_name_token1] = ACTIONS(171),
    [aux_sym__argument_name_token2] = ACTIONS(174),
    [anon_sym_POUND_LBRACK] = ACTIONS(177),
    [aux_sym_encapsed_string_token1] = ACTIONS(180),
    [anon_sym_DQUOTE] = ACTIONS(180),
    [aux_sym_string_token1] = ACTIONS(183),
    [anon_sym_SQUOTE] = ACTIONS(183),
    [anon_sym_LT_LT_LT] = ACTIONS(186),
    [anon_sym_BQUOTE] = ACTIONS(189),
    [anon_sym_DOLLAR] = ACTIONS(192),
    [aux_sym_yield_expression_token1] = ACTIONS(195),
    [aux_sym_include_expression_token1] = ACTIONS(198),
    [aux_sym_include_once_expression_token1] = ACTIONS(201),
    [aux_sym_require_expression_token1] = ACTIONS(204),
    [aux_sym_require_once_expression_token1] = ACTIONS(207),
    [sym_comment] = ACTIONS(5),
  },
  [12] = {
    [sym_text_interpolation] = STATE(12),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(13),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token2] = ACTIONS(354),
    [aux_sym_else_if_clause_token1] = ACTIONS(354),
    [aux_sym_else_clause_token1] = ACTIONS(354),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [13] = {
    [sym_text_interpolation] = STATE(13),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(11),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token2] = ACTIONS(352),
    [aux_sym_else_if_clause_token1] = ACTIONS(352),
    [aux_sym_else_clause_token1] = ACTIONS(352),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [14] = {
    [sym_text_interpolation] = STATE(14),
    [sym_statement] = STATE(465),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_colon_block] = STATE(2541),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(466),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(358),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [15] = {
    [sym_text_interpolation] = STATE(15),
    [sym_statement] = STATE(2149),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_colon_block] = STATE(2593),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2150),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(358),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(443),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(445),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(447),
    [aux_sym_foreach_statement_token1] = ACTIONS(449),
    [aux_sym_if_statement_token1] = ACTIONS(451),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [16] = {
    [sym_text_interpolation] = STATE(16),
    [sym_statement] = STATE(2113),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2114),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(453),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(443),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(445),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(447),
    [aux_sym_foreach_statement_token1] = ACTIONS(449),
    [aux_sym_if_statement_token1] = ACTIONS(451),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [17] = {
    [sym_text_interpolation] = STATE(17),
    [sym_statement] = STATE(2074),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2074),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(455),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(392),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(404),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(408),
    [aux_sym_foreach_statement_token1] = ACTIONS(410),
    [aux_sym_if_statement_token1] = ACTIONS(412),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [18] = {
    [sym_text_interpolation] = STATE(18),
    [sym_statement] = STATE(2113),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2114),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(453),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(392),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(404),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(408),
    [aux_sym_foreach_statement_token1] = ACTIONS(410),
    [aux_sym_if_statement_token1] = ACTIONS(412),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [19] = {
    [sym_text_interpolation] = STATE(19),
    [sym_statement] = STATE(2137),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2138),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(457),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(392),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(404),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(408),
    [aux_sym_foreach_statement_token1] = ACTIONS(410),
    [aux_sym_if_statement_token1] = ACTIONS(412),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [20] = {
    [sym_text_interpolation] = STATE(20),
    [sym_statement] = STATE(2140),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2142),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(459),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(392),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(404),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(408),
    [aux_sym_foreach_statement_token1] = ACTIONS(410),
    [aux_sym_if_statement_token1] = ACTIONS(412),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [21] = {
    [sym_text_interpolation] = STATE(21),
    [sym_statement] = STATE(2144),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2146),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(461),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(392),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(404),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(408),
    [aux_sym_foreach_statement_token1] = ACTIONS(410),
    [aux_sym_if_statement_token1] = ACTIONS(412),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [22] = {
    [sym_text_interpolation] = STATE(22),
    [sym_statement] = STATE(582),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(458),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(463),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [23] = {
    [sym_text_interpolation] = STATE(23),
    [sym_statement] = STATE(2166),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2167),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(465),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(392),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(404),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(408),
    [aux_sym_foreach_statement_token1] = ACTIONS(410),
    [aux_sym_if_statement_token1] = ACTIONS(412),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [24] = {
    [sym_text_interpolation] = STATE(24),
    [sym_statement] = STATE(2169),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2170),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(467),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(392),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(404),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(408),
    [aux_sym_foreach_statement_token1] = ACTIONS(410),
    [aux_sym_if_statement_token1] = ACTIONS(412),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [25] = {
    [sym_text_interpolation] = STATE(25),
    [sym_statement] = STATE(2171),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2173),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(469),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(392),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(404),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(408),
    [aux_sym_foreach_statement_token1] = ACTIONS(410),
    [aux_sym_if_statement_token1] = ACTIONS(412),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [26] = {
    [sym_text_interpolation] = STATE(26),
    [sym_statement] = STATE(2190),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2191),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(471),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(392),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(404),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(408),
    [aux_sym_foreach_statement_token1] = ACTIONS(410),
    [aux_sym_if_statement_token1] = ACTIONS(412),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [27] = {
    [sym_text_interpolation] = STATE(27),
    [sym_statement] = STATE(529),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(530),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(473),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [28] = {
    [sym_text_interpolation] = STATE(28),
    [sym_statement] = STATE(482),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(482),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(475),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [29] = {
    [sym_text_interpolation] = STATE(29),
    [sym_statement] = STATE(482),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(482),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(475),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [30] = {
    [sym_text_interpolation] = STATE(30),
    [sym_statement] = STATE(464),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(490),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(477),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [31] = {
    [sym_text_interpolation] = STATE(31),
    [sym_statement] = STATE(529),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(530),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(473),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [32] = {
    [sym_text_interpolation] = STATE(32),
    [sym_statement] = STATE(542),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(543),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(479),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [33] = {
    [sym_text_interpolation] = STATE(33),
    [sym_statement] = STATE(497),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(500),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(481),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [34] = {
    [sym_text_interpolation] = STATE(34),
    [sym_statement] = STATE(582),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(458),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(463),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [35] = {
    [sym_text_interpolation] = STATE(35),
    [sym_statement] = STATE(506),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(507),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(483),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [36] = {
    [sym_text_interpolation] = STATE(36),
    [sym_statement] = STATE(506),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(507),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(483),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [37] = {
    [sym_text_interpolation] = STATE(37),
    [sym_statement] = STATE(510),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(511),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(485),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [38] = {
    [sym_text_interpolation] = STATE(38),
    [sym_statement] = STATE(513),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(514),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(487),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [39] = {
    [sym_text_interpolation] = STATE(39),
    [sym_statement] = STATE(542),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(543),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(479),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [40] = {
    [sym_text_interpolation] = STATE(40),
    [sym_statement] = STATE(2074),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2074),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(455),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(443),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(445),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(447),
    [aux_sym_foreach_statement_token1] = ACTIONS(449),
    [aux_sym_if_statement_token1] = ACTIONS(451),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [41] = {
    [sym_text_interpolation] = STATE(41),
    [sym_statement] = STATE(510),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(511),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(485),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [42] = {
    [sym_text_interpolation] = STATE(42),
    [sym_statement] = STATE(2137),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2138),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(457),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(443),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(445),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(447),
    [aux_sym_foreach_statement_token1] = ACTIONS(449),
    [aux_sym_if_statement_token1] = ACTIONS(451),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [43] = {
    [sym_text_interpolation] = STATE(43),
    [sym_statement] = STATE(2140),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2142),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(459),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(443),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(445),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(447),
    [aux_sym_foreach_statement_token1] = ACTIONS(449),
    [aux_sym_if_statement_token1] = ACTIONS(451),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [44] = {
    [sym_text_interpolation] = STATE(44),
    [sym_statement] = STATE(2144),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2146),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(461),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(443),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(445),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(447),
    [aux_sym_foreach_statement_token1] = ACTIONS(449),
    [aux_sym_if_statement_token1] = ACTIONS(451),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [45] = {
    [sym_text_interpolation] = STATE(45),
    [sym_statement] = STATE(513),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(514),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(487),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [46] = {
    [sym_text_interpolation] = STATE(46),
    [sym_statement] = STATE(2166),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2167),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(465),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(443),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(445),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(447),
    [aux_sym_foreach_statement_token1] = ACTIONS(449),
    [aux_sym_if_statement_token1] = ACTIONS(451),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [47] = {
    [sym_text_interpolation] = STATE(47),
    [sym_statement] = STATE(2169),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2170),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(467),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(443),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(445),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(447),
    [aux_sym_foreach_statement_token1] = ACTIONS(449),
    [aux_sym_if_statement_token1] = ACTIONS(451),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [48] = {
    [sym_text_interpolation] = STATE(48),
    [sym_statement] = STATE(2171),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2173),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(469),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(443),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(445),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(447),
    [aux_sym_foreach_statement_token1] = ACTIONS(449),
    [aux_sym_if_statement_token1] = ACTIONS(451),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [49] = {
    [sym_text_interpolation] = STATE(49),
    [sym_statement] = STATE(2190),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2191),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(471),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(443),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(445),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(447),
    [aux_sym_foreach_statement_token1] = ACTIONS(449),
    [aux_sym_if_statement_token1] = ACTIONS(451),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [50] = {
    [sym_text_interpolation] = STATE(50),
    [sym_statement] = STATE(497),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(500),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(481),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [51] = {
    [sym_text_interpolation] = STATE(51),
    [sym_statement] = STATE(464),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(490),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(477),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [52] = {
    [sym_text_interpolation] = STATE(52),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(489),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [53] = {
    [sym_text_interpolation] = STATE(53),
    [sym_statement] = STATE(1679),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_colon_block] = STATE(1627),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(491),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(493),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(443),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(445),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(447),
    [aux_sym_foreach_statement_token1] = ACTIONS(449),
    [aux_sym_if_statement_token1] = ACTIONS(451),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [54] = {
    [sym_text_interpolation] = STATE(54),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(56),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [aux_sym_declare_statement_token2] = ACTIONS(495),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [55] = {
    [sym_text_interpolation] = STATE(55),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(109),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [anon_sym_RBRACE] = ACTIONS(497),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [56] = {
    [sym_text_interpolation] = STATE(56),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [aux_sym_declare_statement_token2] = ACTIONS(499),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [57] = {
    [sym_text_interpolation] = STATE(57),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(90),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(501),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [58] = {
    [sym_text_interpolation] = STATE(58),
    [sym_statement] = STATE(534),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_colon_block] = STATE(2534),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(358),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [59] = {
    [sym_text_interpolation] = STATE(59),
    [sym_statement] = STATE(449),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_colon_block] = STATE(1693),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(493),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [60] = {
    [sym_text_interpolation] = STATE(60),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(61),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [anon_sym_RBRACE] = ACTIONS(503),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [61] = {
    [sym_text_interpolation] = STATE(61),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [anon_sym_RBRACE] = ACTIONS(505),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [62] = {
    [sym_text_interpolation] = STATE(62),
    [sym_statement] = STATE(446),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_colon_block] = STATE(1693),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(493),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [63] = {
    [sym_text_interpolation] = STATE(63),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(64),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [aux_sym_declare_statement_token2] = ACTIONS(507),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [64] = {
    [sym_text_interpolation] = STATE(64),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [aux_sym_declare_statement_token2] = ACTIONS(509),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [65] = {
    [sym_text_interpolation] = STATE(65),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(67),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(511),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [66] = {
    [sym_text_interpolation] = STATE(66),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(94),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [ts_builtin_sym_end] = ACTIONS(513),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [67] = {
    [sym_text_interpolation] = STATE(67),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(515),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [68] = {
    [sym_text_interpolation] = STATE(68),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(71),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(517),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [69] = {
    [sym_text_interpolation] = STATE(69),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(52),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(519),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [70] = {
    [sym_text_interpolation] = STATE(70),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(73),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(521),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [71] = {
    [sym_text_interpolation] = STATE(71),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(523),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [72] = {
    [sym_text_interpolation] = STATE(72),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(76),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(525),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [73] = {
    [sym_text_interpolation] = STATE(73),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(527),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [74] = {
    [sym_text_interpolation] = STATE(74),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(77),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(529),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [75] = {
    [sym_text_interpolation] = STATE(75),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(78),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(531),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [76] = {
    [sym_text_interpolation] = STATE(76),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(533),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [77] = {
    [sym_text_interpolation] = STATE(77),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(535),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [78] = {
    [sym_text_interpolation] = STATE(78),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(537),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [79] = {
    [sym_text_interpolation] = STATE(79),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(80),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(539),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [80] = {
    [sym_text_interpolation] = STATE(80),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(541),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [81] = {
    [sym_text_interpolation] = STATE(81),
    [sym_statement] = STATE(1649),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_colon_block] = STATE(1627),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(491),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(493),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(443),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(445),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(447),
    [aux_sym_foreach_statement_token1] = ACTIONS(449),
    [aux_sym_if_statement_token1] = ACTIONS(451),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [82] = {
    [sym_text_interpolation] = STATE(82),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(83),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [anon_sym_RBRACE] = ACTIONS(543),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [83] = {
    [sym_text_interpolation] = STATE(83),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [anon_sym_RBRACE] = ACTIONS(545),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [84] = {
    [sym_text_interpolation] = STATE(84),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(85),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [anon_sym_RBRACE] = ACTIONS(547),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [85] = {
    [sym_text_interpolation] = STATE(85),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [anon_sym_RBRACE] = ACTIONS(549),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [86] = {
    [sym_text_interpolation] = STATE(86),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(87),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [anon_sym_RBRACE] = ACTIONS(551),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [87] = {
    [sym_text_interpolation] = STATE(87),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [anon_sym_RBRACE] = ACTIONS(553),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [88] = {
    [sym_text_interpolation] = STATE(88),
    [sym_statement] = STATE(534),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_colon_block] = STATE(2534),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [anon_sym_COLON] = ACTIONS(358),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [89] = {
    [sym_text_interpolation] = STATE(89),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [ts_builtin_sym_end] = ACTIONS(555),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [90] = {
    [sym_text_interpolation] = STATE(90),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(557),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [91] = {
    [sym_text_interpolation] = STATE(91),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(96),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(559),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [92] = {
    [sym_text_interpolation] = STATE(92),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(97),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(561),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [93] = {
    [sym_text_interpolation] = STATE(93),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(100),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(563),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [94] = {
    [sym_text_interpolation] = STATE(94),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [ts_builtin_sym_end] = ACTIONS(565),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [95] = {
    [sym_text_interpolation] = STATE(95),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(89),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [ts_builtin_sym_end] = ACTIONS(565),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [96] = {
    [sym_text_interpolation] = STATE(96),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(567),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [97] = {
    [sym_text_interpolation] = STATE(97),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(569),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [98] = {
    [sym_text_interpolation] = STATE(98),
    [sym_statement] = STATE(2272),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_colon_block] = STATE(2587),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(491),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(358),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(443),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(445),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(447),
    [aux_sym_foreach_statement_token1] = ACTIONS(449),
    [aux_sym_if_statement_token1] = ACTIONS(451),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [99] = {
    [sym_text_interpolation] = STATE(99),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(103),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(571),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [100] = {
    [sym_text_interpolation] = STATE(100),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(573),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [101] = {
    [sym_text_interpolation] = STATE(101),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(104),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(575),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [102] = {
    [sym_text_interpolation] = STATE(102),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(105),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(577),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [103] = {
    [sym_text_interpolation] = STATE(103),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(579),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [104] = {
    [sym_text_interpolation] = STATE(104),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(581),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [105] = {
    [sym_text_interpolation] = STATE(105),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(583),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [106] = {
    [sym_text_interpolation] = STATE(106),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(107),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(585),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [107] = {
    [sym_text_interpolation] = STATE(107),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_for_statement_token2] = ACTIONS(587),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [108] = {
    [sym_text_interpolation] = STATE(108),
    [sym_statement] = STATE(2272),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_colon_block] = STATE(2587),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(491),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [anon_sym_COLON] = ACTIONS(358),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(392),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(404),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(408),
    [aux_sym_foreach_statement_token1] = ACTIONS(410),
    [aux_sym_if_statement_token1] = ACTIONS(412),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [109] = {
    [sym_text_interpolation] = STATE(109),
    [sym_statement] = STATE(459),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [anon_sym_RBRACE] = ACTIONS(589),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [110] = {
    [sym_text_interpolation] = STATE(110),
    [sym_statement] = STATE(2078),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(491),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(392),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(404),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(408),
    [aux_sym_foreach_statement_token1] = ACTIONS(410),
    [aux_sym_if_statement_token1] = ACTIONS(412),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [111] = {
    [sym_text_interpolation] = STATE(111),
    [sym_statement] = STATE(487),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [112] = {
    [sym_text_interpolation] = STATE(112),
    [sym_statement] = STATE(2702),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(491),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(392),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(404),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(408),
    [aux_sym_foreach_statement_token1] = ACTIONS(410),
    [aux_sym_if_statement_token1] = ACTIONS(412),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [113] = {
    [sym_text_interpolation] = STATE(113),
    [sym_statement] = STATE(2115),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(491),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(443),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(445),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(447),
    [aux_sym_foreach_statement_token1] = ACTIONS(449),
    [aux_sym_if_statement_token1] = ACTIONS(451),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [114] = {
    [sym_text_interpolation] = STATE(114),
    [sym_statement] = STATE(537),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(433),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(435),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(437),
    [aux_sym_foreach_statement_token1] = ACTIONS(439),
    [aux_sym_if_statement_token1] = ACTIONS(441),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [115] = {
    [sym_text_interpolation] = STATE(115),
    [sym_statement] = STATE(487),
    [sym_empty_statement] = STATE(460),
    [sym_function_static_declaration] = STATE(460),
    [sym_global_declaration] = STATE(460),
    [sym_namespace_definition] = STATE(460),
    [sym_namespace_use_declaration] = STATE(460),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(460),
    [sym_interface_declaration] = STATE(460),
    [sym_enum_declaration] = STATE(460),
    [sym_class_declaration] = STATE(460),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(460),
    [sym__const_declaration] = STATE(486),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(460),
    [sym__function_definition_header] = STATE(2461),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(460),
    [sym_exit_statement] = STATE(460),
    [sym_unset_statement] = STATE(460),
    [sym_declare_statement] = STATE(460),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(460),
    [sym_goto_statement] = STATE(460),
    [sym_continue_statement] = STATE(460),
    [sym_break_statement] = STATE(460),
    [sym_return_statement] = STATE(460),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(460),
    [sym_do_statement] = STATE(460),
    [sym_for_statement] = STATE(460),
    [sym_foreach_statement] = STATE(460),
    [sym_if_statement] = STATE(460),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(460),
    [sym_compound_statement] = STATE(460),
    [sym_named_label_statement] = STATE(460),
    [sym_expression_statement] = STATE(460),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1355),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_class_declaration_repeat1] = STATE(1386),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(214),
    [aux_sym_global_declaration_token1] = ACTIONS(216),
    [aux_sym_namespace_definition_token1] = ACTIONS(218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(224),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [aux_sym_trait_declaration_token1] = ACTIONS(232),
    [aux_sym_interface_declaration_token1] = ACTIONS(234),
    [aux_sym_enum_declaration_token1] = ACTIONS(236),
    [aux_sym_class_declaration_token1] = ACTIONS(240),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(258),
    [aux_sym_exit_statement_token1] = ACTIONS(260),
    [anon_sym_unset] = ACTIONS(262),
    [aux_sym_declare_statement_token1] = ACTIONS(264),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(268),
    [aux_sym_goto_statement_token1] = ACTIONS(270),
    [aux_sym_continue_statement_token1] = ACTIONS(272),
    [aux_sym_break_statement_token1] = ACTIONS(274),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(276),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(280),
    [aux_sym_do_statement_token1] = ACTIONS(282),
    [aux_sym_for_statement_token1] = ACTIONS(284),
    [aux_sym_foreach_statement_token1] = ACTIONS(286),
    [aux_sym_if_statement_token1] = ACTIONS(288),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(292),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [116] = {
    [sym_text_interpolation] = STATE(116),
    [sym_statement] = STATE(2078),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(491),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(443),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(445),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(447),
    [aux_sym_foreach_statement_token1] = ACTIONS(449),
    [aux_sym_if_statement_token1] = ACTIONS(451),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [117] = {
    [sym_text_interpolation] = STATE(117),
    [sym_statement] = STATE(2529),
    [sym_empty_statement] = STATE(2049),
    [sym_function_static_declaration] = STATE(2049),
    [sym_global_declaration] = STATE(2049),
    [sym_namespace_definition] = STATE(2049),
    [sym_namespace_use_declaration] = STATE(2049),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_trait_declaration] = STATE(2049),
    [sym_interface_declaration] = STATE(2049),
    [sym_enum_declaration] = STATE(2049),
    [sym_class_declaration] = STATE(2049),
    [sym_final_modifier] = STATE(1339),
    [sym_abstract_modifier] = STATE(1339),
    [sym_readonly_modifier] = STATE(1339),
    [sym_const_declaration] = STATE(2049),
    [sym__const_declaration] = STATE(2056),
    [sym__modifier] = STATE(1338),
    [sym_static_modifier] = STATE(1465),
    [sym_visibility_modifier] = STATE(1339),
    [sym_function_definition] = STATE(2049),
    [sym__function_definition_header] = STATE(2338),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_echo_statement] = STATE(2049),
    [sym_exit_statement] = STATE(2049),
    [sym_unset_statement] = STATE(2049),
    [sym_declare_statement] = STATE(2049),
    [sym_literal] = STATE(1087),
    [sym_try_statement] = STATE(2049),
    [sym_goto_statement] = STATE(2049),
    [sym_continue_statement] = STATE(2049),
    [sym_break_statement] = STATE(2049),
    [sym_return_statement] = STATE(2049),
    [sym_throw_expression] = STATE(1087),
    [sym_while_statement] = STATE(2049),
    [sym_do_statement] = STATE(2049),
    [sym_for_statement] = STATE(2049),
    [sym_foreach_statement] = STATE(2049),
    [sym_if_statement] = STATE(2049),
    [sym_match_expression] = STATE(1090),
    [sym_switch_statement] = STATE(2049),
    [sym_compound_statement] = STATE(2049),
    [sym_named_label_statement] = STATE(2049),
    [sym_expression_statement] = STATE(2049),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1356),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_class_declaration_repeat1] = STATE(1385),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(491),
    [aux_sym_function_static_declaration_token1] = ACTIONS(366),
    [aux_sym_global_declaration_token1] = ACTIONS(368),
    [aux_sym_namespace_definition_token1] = ACTIONS(370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(374),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(376),
    [aux_sym_trait_declaration_token1] = ACTIONS(378),
    [aux_sym_interface_declaration_token1] = ACTIONS(380),
    [aux_sym_enum_declaration_token1] = ACTIONS(382),
    [aux_sym_class_declaration_token1] = ACTIONS(384),
    [aux_sym_final_modifier_token1] = ACTIONS(242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(246),
    [sym_var_modifier] = ACTIONS(248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(250),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [aux_sym_echo_statement_token1] = ACTIONS(386),
    [aux_sym_exit_statement_token1] = ACTIONS(388),
    [anon_sym_unset] = ACTIONS(390),
    [aux_sym_declare_statement_token1] = ACTIONS(392),
    [sym_float] = ACTIONS(266),
    [aux_sym_try_statement_token1] = ACTIONS(394),
    [aux_sym_goto_statement_token1] = ACTIONS(396),
    [aux_sym_continue_statement_token1] = ACTIONS(398),
    [aux_sym_break_statement_token1] = ACTIONS(400),
    [sym_integer] = ACTIONS(266),
    [aux_sym_return_statement_token1] = ACTIONS(402),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_while_statement_token1] = ACTIONS(404),
    [aux_sym_do_statement_token1] = ACTIONS(406),
    [aux_sym_for_statement_token1] = ACTIONS(408),
    [aux_sym_foreach_statement_token1] = ACTIONS(410),
    [aux_sym_if_statement_token1] = ACTIONS(412),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [aux_sym_switch_statement_token1] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [118] = {
    [sym_text_interpolation] = STATE(118),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1052),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(1024),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(593),
    [anon_sym_AMP] = ACTIONS(595),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym_namespace_use_clause_token1] = ACTIONS(599),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_RBRACE] = ACTIONS(593),
    [anon_sym_COLON] = ACTIONS(593),
    [anon_sym_EQ_GT] = ACTIONS(593),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [anon_sym_RPAREN2] = ACTIONS(593),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(607),
    [anon_sym_QMARK] = ACTIONS(599),
    [anon_sym_PIPE] = ACTIONS(599),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(617),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_yield_expression_token2] = ACTIONS(649),
    [aux_sym_binary_expression_token1] = ACTIONS(599),
    [anon_sym_QMARK_QMARK] = ACTIONS(593),
    [anon_sym_STAR_STAR] = ACTIONS(593),
    [aux_sym_binary_expression_token2] = ACTIONS(599),
    [aux_sym_binary_expression_token3] = ACTIONS(599),
    [aux_sym_binary_expression_token4] = ACTIONS(599),
    [anon_sym_PIPE_PIPE] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(593),
    [anon_sym_CARET] = ACTIONS(593),
    [anon_sym_EQ_EQ] = ACTIONS(599),
    [anon_sym_BANG_EQ] = ACTIONS(599),
    [anon_sym_LT_GT] = ACTIONS(593),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(593),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(593),
    [anon_sym_LT] = ACTIONS(599),
    [anon_sym_GT] = ACTIONS(599),
    [anon_sym_LT_EQ] = ACTIONS(599),
    [anon_sym_GT_EQ] = ACTIONS(593),
    [anon_sym_LT_EQ_GT] = ACTIONS(593),
    [anon_sym_LT_LT] = ACTIONS(599),
    [anon_sym_GT_GT] = ACTIONS(593),
    [anon_sym_DOT] = ACTIONS(599),
    [anon_sym_STAR] = ACTIONS(599),
    [anon_sym_SLASH] = ACTIONS(599),
    [anon_sym_PERCENT] = ACTIONS(593),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [119] = {
    [sym_text_interpolation] = STATE(119),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1169),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(1024),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(593),
    [anon_sym_AMP] = ACTIONS(595),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(593),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_RBRACE] = ACTIONS(593),
    [anon_sym_EQ_GT] = ACTIONS(593),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(661),
    [anon_sym_QMARK] = ACTIONS(599),
    [anon_sym_PIPE] = ACTIONS(599),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(665),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_RBRACK] = ACTIONS(593),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_yield_expression_token2] = ACTIONS(679),
    [aux_sym_binary_expression_token1] = ACTIONS(599),
    [anon_sym_QMARK_QMARK] = ACTIONS(593),
    [anon_sym_STAR_STAR] = ACTIONS(593),
    [aux_sym_binary_expression_token2] = ACTIONS(599),
    [aux_sym_binary_expression_token3] = ACTIONS(599),
    [aux_sym_binary_expression_token4] = ACTIONS(599),
    [anon_sym_PIPE_PIPE] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(593),
    [anon_sym_CARET] = ACTIONS(593),
    [anon_sym_EQ_EQ] = ACTIONS(599),
    [anon_sym_BANG_EQ] = ACTIONS(599),
    [anon_sym_LT_GT] = ACTIONS(593),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(593),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(593),
    [anon_sym_LT] = ACTIONS(599),
    [anon_sym_GT] = ACTIONS(599),
    [anon_sym_LT_EQ] = ACTIONS(599),
    [anon_sym_GT_EQ] = ACTIONS(593),
    [anon_sym_LT_EQ_GT] = ACTIONS(593),
    [anon_sym_LT_LT] = ACTIONS(599),
    [anon_sym_GT_GT] = ACTIONS(593),
    [anon_sym_DOT] = ACTIONS(599),
    [anon_sym_STAR] = ACTIONS(599),
    [anon_sym_SLASH] = ACTIONS(599),
    [anon_sym_PERCENT] = ACTIONS(593),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [120] = {
    [sym_text_interpolation] = STATE(120),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1092),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym_variadic_unpacking] = STATE(1094),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_by_ref] = STATE(1094),
    [sym_yield_expression] = STATE(1090),
    [sym_array_element_initializer] = STATE(1093),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(593),
    [anon_sym_AMP] = ACTIONS(691),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(593),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(593),
    [anon_sym_EQ_GT] = ACTIONS(593),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(693),
    [anon_sym_QMARK] = ACTIONS(599),
    [anon_sym_PIPE] = ACTIONS(599),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(294),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_yield_expression_token2] = ACTIONS(695),
    [aux_sym_binary_expression_token1] = ACTIONS(599),
    [anon_sym_QMARK_QMARK] = ACTIONS(593),
    [anon_sym_STAR_STAR] = ACTIONS(593),
    [aux_sym_binary_expression_token2] = ACTIONS(599),
    [aux_sym_binary_expression_token3] = ACTIONS(599),
    [aux_sym_binary_expression_token4] = ACTIONS(599),
    [anon_sym_PIPE_PIPE] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(593),
    [anon_sym_CARET] = ACTIONS(593),
    [anon_sym_EQ_EQ] = ACTIONS(599),
    [anon_sym_BANG_EQ] = ACTIONS(599),
    [anon_sym_LT_GT] = ACTIONS(593),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(593),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(593),
    [anon_sym_LT] = ACTIONS(599),
    [anon_sym_GT] = ACTIONS(599),
    [anon_sym_LT_EQ] = ACTIONS(599),
    [anon_sym_GT_EQ] = ACTIONS(593),
    [anon_sym_LT_EQ_GT] = ACTIONS(593),
    [anon_sym_LT_LT] = ACTIONS(599),
    [anon_sym_GT_GT] = ACTIONS(593),
    [anon_sym_DOT] = ACTIONS(599),
    [anon_sym_STAR] = ACTIONS(599),
    [anon_sym_SLASH] = ACTIONS(599),
    [anon_sym_PERCENT] = ACTIONS(593),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(593),
  },
  [121] = {
    [sym_text_interpolation] = STATE(121),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1206),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(1024),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(595),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(593),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LBRACE] = ACTIONS(593),
    [anon_sym_EQ_GT] = ACTIONS(593),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(593),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [anon_sym_QMARK] = ACTIONS(599),
    [anon_sym_PIPE] = ACTIONS(599),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(703),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_yield_expression_token2] = ACTIONS(717),
    [aux_sym_binary_expression_token1] = ACTIONS(599),
    [anon_sym_QMARK_QMARK] = ACTIONS(593),
    [anon_sym_STAR_STAR] = ACTIONS(593),
    [aux_sym_binary_expression_token2] = ACTIONS(599),
    [aux_sym_binary_expression_token3] = ACTIONS(599),
    [aux_sym_binary_expression_token4] = ACTIONS(599),
    [anon_sym_PIPE_PIPE] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(593),
    [anon_sym_CARET] = ACTIONS(593),
    [anon_sym_EQ_EQ] = ACTIONS(599),
    [anon_sym_BANG_EQ] = ACTIONS(599),
    [anon_sym_LT_GT] = ACTIONS(593),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(593),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(593),
    [anon_sym_LT] = ACTIONS(599),
    [anon_sym_GT] = ACTIONS(599),
    [anon_sym_LT_EQ] = ACTIONS(599),
    [anon_sym_GT_EQ] = ACTIONS(593),
    [anon_sym_LT_EQ_GT] = ACTIONS(593),
    [anon_sym_LT_LT] = ACTIONS(599),
    [anon_sym_GT_GT] = ACTIONS(593),
    [anon_sym_DOT] = ACTIONS(599),
    [anon_sym_STAR] = ACTIONS(599),
    [anon_sym_SLASH] = ACTIONS(599),
    [anon_sym_PERCENT] = ACTIONS(593),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [122] = {
    [sym_text_interpolation] = STATE(122),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_cast_type] = STATE(2527),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1320),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(727),
    [aux_sym_cast_type_token2] = ACTIONS(729),
    [aux_sym_cast_type_token3] = ACTIONS(729),
    [aux_sym_cast_type_token4] = ACTIONS(729),
    [aux_sym_cast_type_token5] = ACTIONS(729),
    [aux_sym_cast_type_token6] = ACTIONS(729),
    [aux_sym_cast_type_token7] = ACTIONS(729),
    [aux_sym_cast_type_token8] = ACTIONS(729),
    [aux_sym_cast_type_token9] = ACTIONS(729),
    [aux_sym_cast_type_token10] = ACTIONS(729),
    [aux_sym_cast_type_token11] = ACTIONS(729),
    [aux_sym_cast_type_token12] = ACTIONS(729),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [123] = {
    [sym_text_interpolation] = STATE(123),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_cast_type] = STATE(2567),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1320),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(727),
    [aux_sym_cast_type_token2] = ACTIONS(729),
    [aux_sym_cast_type_token3] = ACTIONS(729),
    [aux_sym_cast_type_token4] = ACTIONS(729),
    [aux_sym_cast_type_token5] = ACTIONS(729),
    [aux_sym_cast_type_token6] = ACTIONS(729),
    [aux_sym_cast_type_token7] = ACTIONS(729),
    [aux_sym_cast_type_token8] = ACTIONS(729),
    [aux_sym_cast_type_token9] = ACTIONS(729),
    [aux_sym_cast_type_token10] = ACTIONS(729),
    [aux_sym_cast_type_token11] = ACTIONS(729),
    [aux_sym_cast_type_token12] = ACTIONS(729),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [124] = {
    [sym_text_interpolation] = STATE(124),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_cast_type] = STATE(2578),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1281),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(727),
    [aux_sym_cast_type_token2] = ACTIONS(729),
    [aux_sym_cast_type_token3] = ACTIONS(729),
    [aux_sym_cast_type_token4] = ACTIONS(729),
    [aux_sym_cast_type_token5] = ACTIONS(729),
    [aux_sym_cast_type_token6] = ACTIONS(729),
    [aux_sym_cast_type_token7] = ACTIONS(729),
    [aux_sym_cast_type_token8] = ACTIONS(729),
    [aux_sym_cast_type_token9] = ACTIONS(729),
    [aux_sym_cast_type_token10] = ACTIONS(729),
    [aux_sym_cast_type_token11] = ACTIONS(729),
    [aux_sym_cast_type_token12] = ACTIONS(729),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [125] = {
    [sym_text_interpolation] = STATE(125),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_cast_type] = STATE(2636),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1281),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(727),
    [aux_sym_cast_type_token2] = ACTIONS(729),
    [aux_sym_cast_type_token3] = ACTIONS(729),
    [aux_sym_cast_type_token4] = ACTIONS(729),
    [aux_sym_cast_type_token5] = ACTIONS(729),
    [aux_sym_cast_type_token6] = ACTIONS(729),
    [aux_sym_cast_type_token7] = ACTIONS(729),
    [aux_sym_cast_type_token8] = ACTIONS(729),
    [aux_sym_cast_type_token9] = ACTIONS(729),
    [aux_sym_cast_type_token10] = ACTIONS(729),
    [aux_sym_cast_type_token11] = ACTIONS(729),
    [aux_sym_cast_type_token12] = ACTIONS(729),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [126] = {
    [sym_text_interpolation] = STATE(126),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_cast_type] = STATE(2567),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1281),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(727),
    [aux_sym_cast_type_token2] = ACTIONS(729),
    [aux_sym_cast_type_token3] = ACTIONS(729),
    [aux_sym_cast_type_token4] = ACTIONS(729),
    [aux_sym_cast_type_token5] = ACTIONS(729),
    [aux_sym_cast_type_token6] = ACTIONS(729),
    [aux_sym_cast_type_token7] = ACTIONS(729),
    [aux_sym_cast_type_token8] = ACTIONS(729),
    [aux_sym_cast_type_token9] = ACTIONS(729),
    [aux_sym_cast_type_token10] = ACTIONS(729),
    [aux_sym_cast_type_token11] = ACTIONS(729),
    [aux_sym_cast_type_token12] = ACTIONS(729),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [127] = {
    [sym_text_interpolation] = STATE(127),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_cast_type] = STATE(2573),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1281),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(727),
    [aux_sym_cast_type_token2] = ACTIONS(729),
    [aux_sym_cast_type_token3] = ACTIONS(729),
    [aux_sym_cast_type_token4] = ACTIONS(729),
    [aux_sym_cast_type_token5] = ACTIONS(729),
    [aux_sym_cast_type_token6] = ACTIONS(729),
    [aux_sym_cast_type_token7] = ACTIONS(729),
    [aux_sym_cast_type_token8] = ACTIONS(729),
    [aux_sym_cast_type_token9] = ACTIONS(729),
    [aux_sym_cast_type_token10] = ACTIONS(729),
    [aux_sym_cast_type_token11] = ACTIONS(729),
    [aux_sym_cast_type_token12] = ACTIONS(729),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [128] = {
    [sym_text_interpolation] = STATE(128),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_cast_type] = STATE(2640),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1281),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(727),
    [aux_sym_cast_type_token2] = ACTIONS(729),
    [aux_sym_cast_type_token3] = ACTIONS(729),
    [aux_sym_cast_type_token4] = ACTIONS(729),
    [aux_sym_cast_type_token5] = ACTIONS(729),
    [aux_sym_cast_type_token6] = ACTIONS(729),
    [aux_sym_cast_type_token7] = ACTIONS(729),
    [aux_sym_cast_type_token8] = ACTIONS(729),
    [aux_sym_cast_type_token9] = ACTIONS(729),
    [aux_sym_cast_type_token10] = ACTIONS(729),
    [aux_sym_cast_type_token11] = ACTIONS(729),
    [aux_sym_cast_type_token12] = ACTIONS(729),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [129] = {
    [sym_text_interpolation] = STATE(129),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_cast_type] = STATE(2667),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1281),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(727),
    [aux_sym_cast_type_token2] = ACTIONS(729),
    [aux_sym_cast_type_token3] = ACTIONS(729),
    [aux_sym_cast_type_token4] = ACTIONS(729),
    [aux_sym_cast_type_token5] = ACTIONS(729),
    [aux_sym_cast_type_token6] = ACTIONS(729),
    [aux_sym_cast_type_token7] = ACTIONS(729),
    [aux_sym_cast_type_token8] = ACTIONS(729),
    [aux_sym_cast_type_token9] = ACTIONS(729),
    [aux_sym_cast_type_token10] = ACTIONS(729),
    [aux_sym_cast_type_token11] = ACTIONS(729),
    [aux_sym_cast_type_token12] = ACTIONS(729),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [130] = {
    [sym_text_interpolation] = STATE(130),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_cast_type] = STATE(2549),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1281),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(727),
    [aux_sym_cast_type_token2] = ACTIONS(729),
    [aux_sym_cast_type_token3] = ACTIONS(729),
    [aux_sym_cast_type_token4] = ACTIONS(729),
    [aux_sym_cast_type_token5] = ACTIONS(729),
    [aux_sym_cast_type_token6] = ACTIONS(729),
    [aux_sym_cast_type_token7] = ACTIONS(729),
    [aux_sym_cast_type_token8] = ACTIONS(729),
    [aux_sym_cast_type_token9] = ACTIONS(729),
    [aux_sym_cast_type_token10] = ACTIONS(729),
    [aux_sym_cast_type_token11] = ACTIONS(729),
    [aux_sym_cast_type_token12] = ACTIONS(729),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [131] = {
    [sym_text_interpolation] = STATE(131),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_cast_type] = STATE(2648),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1281),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(727),
    [aux_sym_cast_type_token2] = ACTIONS(729),
    [aux_sym_cast_type_token3] = ACTIONS(729),
    [aux_sym_cast_type_token4] = ACTIONS(729),
    [aux_sym_cast_type_token5] = ACTIONS(729),
    [aux_sym_cast_type_token6] = ACTIONS(729),
    [aux_sym_cast_type_token7] = ACTIONS(729),
    [aux_sym_cast_type_token8] = ACTIONS(729),
    [aux_sym_cast_type_token9] = ACTIONS(729),
    [aux_sym_cast_type_token10] = ACTIONS(729),
    [aux_sym_cast_type_token11] = ACTIONS(729),
    [aux_sym_cast_type_token12] = ACTIONS(729),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [132] = {
    [sym_text_interpolation] = STATE(132),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_cast_type] = STATE(2654),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1281),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(727),
    [aux_sym_cast_type_token2] = ACTIONS(729),
    [aux_sym_cast_type_token3] = ACTIONS(729),
    [aux_sym_cast_type_token4] = ACTIONS(729),
    [aux_sym_cast_type_token5] = ACTIONS(729),
    [aux_sym_cast_type_token6] = ACTIONS(729),
    [aux_sym_cast_type_token7] = ACTIONS(729),
    [aux_sym_cast_type_token8] = ACTIONS(729),
    [aux_sym_cast_type_token9] = ACTIONS(729),
    [aux_sym_cast_type_token10] = ACTIONS(729),
    [aux_sym_cast_type_token11] = ACTIONS(729),
    [aux_sym_cast_type_token12] = ACTIONS(729),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [133] = {
    [sym_text_interpolation] = STATE(133),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1232),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(694),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(694),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(694),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(694),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(1959),
    [sym__array_destructing_element] = STATE(1962),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(695),
    [sym__callable_variable] = STATE(678),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(2305),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2015),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym__array_destructing_repeat1] = STATE(1964),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(733),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(661),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_RBRACK] = ACTIONS(735),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [134] = {
    [sym_text_interpolation] = STATE(134),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_variadic_placeholder] = STATE(2617),
    [sym_argument] = STATE(2041),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(747),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(751),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [135] = {
    [sym_text_interpolation] = STATE(135),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_variadic_placeholder] = STATE(2519),
    [sym_argument] = STATE(2128),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(765),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(751),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [136] = {
    [sym_text_interpolation] = STATE(136),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_variadic_placeholder] = STATE(2700),
    [sym_argument] = STATE(2105),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(767),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(751),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [137] = {
    [sym_text_interpolation] = STATE(137),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_variadic_placeholder] = STATE(2612),
    [sym_argument] = STATE(2141),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(769),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(751),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [138] = {
    [sym_text_interpolation] = STATE(138),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_variadic_placeholder] = STATE(2538),
    [sym_argument] = STATE(2172),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(771),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(751),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [139] = {
    [sym_text_interpolation] = STATE(139),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1232),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(694),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(694),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(694),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(694),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(1959),
    [sym__array_destructing_element] = STATE(1962),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(695),
    [sym__callable_variable] = STATE(678),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(2305),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2015),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym__array_destructing_repeat1] = STATE(1964),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(733),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(661),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_RBRACK] = ACTIONS(773),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [140] = {
    [sym_text_interpolation] = STATE(140),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_variadic_placeholder] = STATE(2531),
    [sym_argument] = STATE(2118),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(775),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(751),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [141] = {
    [sym_text_interpolation] = STATE(141),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1232),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(694),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(694),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(694),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(694),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(1959),
    [sym__array_destructing_element] = STATE(1962),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(695),
    [sym__callable_variable] = STATE(678),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(2305),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2015),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym__array_destructing_repeat1] = STATE(1964),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(733),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(661),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_RBRACK] = ACTIONS(777),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [142] = {
    [sym_text_interpolation] = STATE(142),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1232),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(694),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(694),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(694),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(694),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(1959),
    [sym__array_destructing_element] = STATE(1962),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(695),
    [sym__callable_variable] = STATE(678),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(2305),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2015),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym__array_destructing_repeat1] = STATE(1964),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(733),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(661),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_RBRACK] = ACTIONS(779),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [143] = {
    [sym_text_interpolation] = STATE(143),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1232),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(694),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(694),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(694),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(694),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(1959),
    [sym__array_destructing_element] = STATE(1962),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(695),
    [sym__callable_variable] = STATE(678),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(2305),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(1963),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym__array_destructing_repeat1] = STATE(1964),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(781),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(661),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_RBRACK] = ACTIONS(783),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [144] = {
    [sym_text_interpolation] = STATE(144),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_argument] = STATE(2396),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(785),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [145] = {
    [sym_text_interpolation] = STATE(145),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_argument] = STATE(2396),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(787),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [146] = {
    [sym_text_interpolation] = STATE(146),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_argument] = STATE(2396),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(789),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [147] = {
    [sym_text_interpolation] = STATE(147),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_argument] = STATE(2396),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(791),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [148] = {
    [sym_text_interpolation] = STATE(148),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_argument] = STATE(2396),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(793),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [149] = {
    [sym_text_interpolation] = STATE(149),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_argument] = STATE(2396),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(795),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [150] = {
    [sym_text_interpolation] = STATE(150),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_argument] = STATE(2396),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(797),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [151] = {
    [sym_text_interpolation] = STATE(151),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_argument] = STATE(2396),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(799),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [152] = {
    [sym_text_interpolation] = STATE(152),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_argument] = STATE(2396),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(801),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [153] = {
    [sym_text_interpolation] = STATE(153),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_argument] = STATE(2396),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(803),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [154] = {
    [sym_text_interpolation] = STATE(154),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_argument] = STATE(2396),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(805),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [155] = {
    [sym_text_interpolation] = STATE(155),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_argument] = STATE(2396),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(807),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [156] = {
    [sym_text_interpolation] = STATE(156),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2040),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(809),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(811),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [157] = {
    [sym_text_interpolation] = STATE(157),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(1963),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(813),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(661),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_RBRACK] = ACTIONS(815),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [158] = {
    [sym_text_interpolation] = STATE(158),
    [sym_reference_modifier] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_argument] = STATE(2396),
    [sym__argument_name] = STATE(182),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2355),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1577),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(737),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(741),
    [aux_sym_namespace_definition_token1] = ACTIONS(743),
    [aux_sym__namespace_use_type_token1] = ACTIONS(745),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(749),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(753),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(755),
    [aux_sym_match_expression_token1] = ACTIONS(757),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(759),
    [anon_sym_parent] = ACTIONS(759),
    [aux_sym__argument_name_token1] = ACTIONS(761),
    [aux_sym__argument_name_token2] = ACTIONS(763),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [159] = {
    [sym_text_interpolation] = STATE(159),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2030),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(817),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(819),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [160] = {
    [sym_text_interpolation] = STATE(160),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2015),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(821),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(661),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_RBRACK] = ACTIONS(823),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [161] = {
    [sym_text_interpolation] = STATE(161),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2254),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(661),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_RBRACK] = ACTIONS(825),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [162] = {
    [sym_text_interpolation] = STATE(162),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2254),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(661),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_RBRACK] = ACTIONS(827),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [163] = {
    [sym_text_interpolation] = STATE(163),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_match_condition_list] = STATE(2530),
    [sym_match_conditional_expression] = STATE(1996),
    [sym_match_default_expression] = STATE(1996),
    [sym_expression] = STATE(1231),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(829),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_RBRACE] = ACTIONS(831),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [aux_sym_match_default_expression_token1] = ACTIONS(833),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [164] = {
    [sym_text_interpolation] = STATE(164),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2254),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(835),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [165] = {
    [sym_text_interpolation] = STATE(165),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2254),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(837),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [166] = {
    [sym_text_interpolation] = STATE(166),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2254),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(661),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_RBRACK] = ACTIONS(839),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [167] = {
    [sym_text_interpolation] = STATE(167),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2254),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(661),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_RBRACK] = ACTIONS(841),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [168] = {
    [sym_text_interpolation] = STATE(168),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_match_condition_list] = STATE(2530),
    [sym_match_conditional_expression] = STATE(2053),
    [sym_match_default_expression] = STATE(2053),
    [sym_expression] = STATE(1231),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(843),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_RBRACE] = ACTIONS(845),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [aux_sym_match_default_expression_token1] = ACTIONS(833),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [169] = {
    [sym_text_interpolation] = STATE(169),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2254),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(847),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [170] = {
    [sym_text_interpolation] = STATE(170),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2254),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(849),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [171] = {
    [sym_text_interpolation] = STATE(171),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_match_condition_list] = STATE(2530),
    [sym_match_conditional_expression] = STATE(2301),
    [sym_match_default_expression] = STATE(2301),
    [sym_expression] = STATE(1231),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_RBRACE] = ACTIONS(851),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [aux_sym_match_default_expression_token1] = ACTIONS(833),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [172] = {
    [sym_text_interpolation] = STATE(172),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_match_condition_list] = STATE(2530),
    [sym_match_conditional_expression] = STATE(2301),
    [sym_match_default_expression] = STATE(2301),
    [sym_expression] = STATE(1231),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_RBRACE] = ACTIONS(853),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [aux_sym_match_default_expression_token1] = ACTIONS(833),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [173] = {
    [sym_text_interpolation] = STATE(173),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1291),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(689),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(689),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(689),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(689),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(1959),
    [sym__array_destructing_element] = STATE(2422),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(686),
    [sym__callable_variable] = STATE(681),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(2424),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(855),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(857),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(859),
    [anon_sym_RBRACK] = ACTIONS(855),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [174] = {
    [sym_text_interpolation] = STATE(174),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_match_condition_list] = STATE(2530),
    [sym_match_conditional_expression] = STATE(2301),
    [sym_match_default_expression] = STATE(2301),
    [sym_expression] = STATE(1231),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_RBRACE] = ACTIONS(861),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [aux_sym_match_default_expression_token1] = ACTIONS(833),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [175] = {
    [sym_text_interpolation] = STATE(175),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_match_condition_list] = STATE(2530),
    [sym_match_conditional_expression] = STATE(2301),
    [sym_match_default_expression] = STATE(2301),
    [sym_expression] = STATE(1231),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_RBRACE] = ACTIONS(863),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [aux_sym_match_default_expression_token1] = ACTIONS(833),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [176] = {
    [sym_text_interpolation] = STATE(176),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2254),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(661),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [177] = {
    [sym_text_interpolation] = STATE(177),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1291),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(689),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(689),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(689),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(689),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(1959),
    [sym__array_destructing_element] = STATE(2422),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(686),
    [sym__callable_variable] = STATE(681),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(2424),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(855),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(857),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(859),
    [anon_sym_RBRACK] = ACTIONS(865),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [178] = {
    [sym_text_interpolation] = STATE(178),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(1026),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1026),
    [sym_yield_expression] = STATE(1021),
    [sym_array_element_initializer] = STATE(2254),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [179] = {
    [sym_text_interpolation] = STATE(179),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1283),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(683),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(683),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(683),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(683),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(1946),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(682),
    [sym__callable_variable] = STATE(670),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(2069),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym__list_destructing_repeat1] = STATE(2071),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(868),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(857),
    [anon_sym_RPAREN2] = ACTIONS(870),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [180] = {
    [sym_text_interpolation] = STATE(180),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1291),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(689),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(689),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(689),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(689),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(1959),
    [sym__array_destructing_element] = STATE(2422),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(686),
    [sym__callable_variable] = STATE(681),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(2424),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(855),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(857),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(859),
    [anon_sym_RBRACK] = ACTIONS(872),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [181] = {
    [sym_text_interpolation] = STATE(181),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1277),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(687),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(687),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(687),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(687),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2035),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(691),
    [sym__callable_variable] = STATE(680),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(2452),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [anon_sym_COMMA] = ACTIONS(875),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(857),
    [anon_sym_RPAREN2] = ACTIONS(875),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [182] = {
    [sym_text_interpolation] = STATE(182),
    [sym_reference_modifier] = STATE(228),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1261),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2381),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1589),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(739),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [183] = {
    [sym_text_interpolation] = STATE(183),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_foreach_pair] = STATE(2669),
    [sym__foreach_value] = STATE(2669),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1254),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2503),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(2734),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [184] = {
    [sym_text_interpolation] = STATE(184),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_match_condition_list] = STATE(2530),
    [sym_match_conditional_expression] = STATE(2301),
    [sym_match_default_expression] = STATE(2301),
    [sym_expression] = STATE(1231),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [aux_sym_match_default_expression_token1] = ACTIONS(833),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [185] = {
    [sym_text_interpolation] = STATE(185),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_foreach_pair] = STATE(2558),
    [sym__foreach_value] = STATE(2558),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1254),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2503),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(2734),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [186] = {
    [sym_text_interpolation] = STATE(186),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_foreach_pair] = STATE(2718),
    [sym__foreach_value] = STATE(2718),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1254),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2503),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(2734),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [187] = {
    [sym_text_interpolation] = STATE(187),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_foreach_pair] = STATE(2682),
    [sym__foreach_value] = STATE(2682),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1254),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2503),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(2734),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [188] = {
    [sym_text_interpolation] = STATE(188),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2590),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(877),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [189] = {
    [sym_text_interpolation] = STATE(189),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2733),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1272),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(879),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [190] = {
    [sym_text_interpolation] = STATE(190),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2655),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(881),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [191] = {
    [sym_text_interpolation] = STATE(191),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2607),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1272),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(883),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [192] = {
    [sym_text_interpolation] = STATE(192),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2618),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1272),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(885),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [193] = {
    [sym_text_interpolation] = STATE(193),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1233),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2284),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(416),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [194] = {
    [sym_text_interpolation] = STATE(194),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1235),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2286),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(416),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [195] = {
    [sym_text_interpolation] = STATE(195),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1237),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(2292),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(416),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(416),
  },
  [196] = {
    [sym_text_interpolation] = STATE(196),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1224),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(547),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(360),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [197] = {
    [sym_text_interpolation] = STATE(197),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2708),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(887),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [198] = {
    [sym_text_interpolation] = STATE(198),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2535),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(889),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [199] = {
    [sym_text_interpolation] = STATE(199),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2715),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1272),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(891),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [200] = {
    [sym_text_interpolation] = STATE(200),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__foreach_value] = STATE(2586),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1323),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2503),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(2734),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [201] = {
    [sym_text_interpolation] = STATE(201),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1229),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(548),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(360),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [202] = {
    [sym_text_interpolation] = STATE(202),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2659),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(893),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [203] = {
    [sym_text_interpolation] = STATE(203),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2577),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(895),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [204] = {
    [sym_text_interpolation] = STATE(204),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2532),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1272),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(897),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [205] = {
    [sym_text_interpolation] = STATE(205),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2555),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(899),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [206] = {
    [sym_text_interpolation] = STATE(206),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2505),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(901),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [207] = {
    [sym_text_interpolation] = STATE(207),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2560),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(903),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [208] = {
    [sym_text_interpolation] = STATE(208),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2562),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(905),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [209] = {
    [sym_text_interpolation] = STATE(209),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2576),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1272),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(907),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [210] = {
    [sym_text_interpolation] = STATE(210),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2579),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1272),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(909),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [211] = {
    [sym_text_interpolation] = STATE(211),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2542),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(911),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [212] = {
    [sym_text_interpolation] = STATE(212),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2544),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(913),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [213] = {
    [sym_text_interpolation] = STATE(213),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2671),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1272),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(915),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [214] = {
    [sym_text_interpolation] = STATE(214),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2679),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(917),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [215] = {
    [sym_text_interpolation] = STATE(215),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2683),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(919),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [216] = {
    [sym_text_interpolation] = STATE(216),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2684),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(921),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [217] = {
    [sym_text_interpolation] = STATE(217),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2686),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(923),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [218] = {
    [sym_text_interpolation] = STATE(218),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2701),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1272),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(925),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [219] = {
    [sym_text_interpolation] = STATE(219),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1223),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [sym__semicolon] = STATE(549),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(360),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
    [sym__automatic_semicolon] = ACTIONS(360),
  },
  [220] = {
    [sym_text_interpolation] = STATE(220),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2723),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1272),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(927),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [221] = {
    [sym_text_interpolation] = STATE(221),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2726),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1272),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(929),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [222] = {
    [sym_text_interpolation] = STATE(222),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__expressions] = STATE(2739),
    [sym_sequence_expression] = STATE(2298),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1272),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(931),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [223] = {
    [sym_text_interpolation] = STATE(223),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym__expressions] = STATE(2009),
    [sym_sequence_expression] = STATE(2494),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1222),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [224] = {
    [sym_text_interpolation] = STATE(224),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1106),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1027),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(933),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [225] = {
    [sym_text_interpolation] = STATE(225),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1164),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_by_ref] = STATE(1100),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(935),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [226] = {
    [sym_text_interpolation] = STATE(226),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1080),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1027),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(933),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [227] = {
    [sym_text_interpolation] = STATE(227),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1204),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1027),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(933),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [228] = {
    [sym_text_interpolation] = STATE(228),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1270),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2401),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1570),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [229] = {
    [sym_text_interpolation] = STATE(229),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1204),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1027),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [230] = {
    [sym_text_interpolation] = STATE(230),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym__expressions] = STATE(2288),
    [sym_sequence_expression] = STATE(2494),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1222),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [231] = {
    [sym_text_interpolation] = STATE(231),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1265),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym_variadic_unpacking] = STATE(2389),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(1611),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(699),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [232] = {
    [sym_text_interpolation] = STATE(232),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1106),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(698),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(698),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(698),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(698),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2227),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(702),
    [sym__callable_variable] = STATE(688),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(2421),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [233] = {
    [sym_text_interpolation] = STATE(233),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1106),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_by_ref] = STATE(1027),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(731),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [234] = {
    [sym_text_interpolation] = STATE(234),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1301),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_RBRACK] = ACTIONS(937),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [235] = {
    [sym_text_interpolation] = STATE(235),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1286),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_COLON] = ACTIONS(939),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [236] = {
    [sym_text_interpolation] = STATE(236),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1296),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_RBRACK] = ACTIONS(941),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [237] = {
    [sym_text_interpolation] = STATE(237),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_sequence_expression] = STATE(2386),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1252),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [238] = {
    [sym_text_interpolation] = STATE(238),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1287),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_RBRACK] = ACTIONS(943),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [239] = {
    [sym_text_interpolation] = STATE(239),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1198),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(945),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [240] = {
    [sym_text_interpolation] = STATE(240),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1288),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_RBRACK] = ACTIONS(947),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [241] = {
    [sym_text_interpolation] = STATE(241),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1324),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_COLON] = ACTIONS(949),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [242] = {
    [sym_text_interpolation] = STATE(242),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_sequence_expression] = STATE(2426),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1240),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [243] = {
    [sym_text_interpolation] = STATE(243),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1255),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_EQ_GT] = ACTIONS(951),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [244] = {
    [sym_text_interpolation] = STATE(244),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1306),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_RBRACK] = ACTIONS(953),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [245] = {
    [sym_text_interpolation] = STATE(245),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1203),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [anon_sym_RPAREN2] = ACTIONS(955),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [246] = {
    [sym_text_interpolation] = STATE(246),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1316),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_COLON] = ACTIONS(957),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [247] = {
    [sym_text_interpolation] = STATE(247),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1178),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_COLON] = ACTIONS(959),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [248] = {
    [sym_text_interpolation] = STATE(248),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1302),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_RBRACK] = ACTIONS(961),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [249] = {
    [sym_text_interpolation] = STATE(249),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1134),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(963),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [250] = {
    [sym_text_interpolation] = STATE(250),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1278),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [anon_sym_RPAREN2] = ACTIONS(965),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [251] = {
    [sym_text_interpolation] = STATE(251),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1321),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_COLON] = ACTIONS(967),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [252] = {
    [sym_text_interpolation] = STATE(252),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1274),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_RBRACK] = ACTIONS(969),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [253] = {
    [sym_text_interpolation] = STATE(253),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1157),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(971),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [254] = {
    [sym_text_interpolation] = STATE(254),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1310),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_RBRACK] = ACTIONS(973),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [255] = {
    [sym_text_interpolation] = STATE(255),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1311),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_RBRACK] = ACTIONS(975),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [256] = {
    [sym_text_interpolation] = STATE(256),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_sequence_expression] = STATE(2386),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1266),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [257] = {
    [sym_text_interpolation] = STATE(257),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1255),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_EQ_GT] = ACTIONS(977),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [258] = {
    [sym_text_interpolation] = STATE(258),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1282),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [anon_sym_RPAREN2] = ACTIONS(979),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [259] = {
    [sym_text_interpolation] = STATE(259),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1077),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_AMP] = ACTIONS(981),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [260] = {
    [sym_text_interpolation] = STATE(260),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1325),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [261] = {
    [sym_text_interpolation] = STATE(261),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1238),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [262] = {
    [sym_text_interpolation] = STATE(262),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1132),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [263] = {
    [sym_text_interpolation] = STATE(263),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1135),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [264] = {
    [sym_text_interpolation] = STATE(264),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1172),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [265] = {
    [sym_text_interpolation] = STATE(265),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1255),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [266] = {
    [sym_text_interpolation] = STATE(266),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1239),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [267] = {
    [sym_text_interpolation] = STATE(267),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1213),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [268] = {
    [sym_text_interpolation] = STATE(268),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1244),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [269] = {
    [sym_text_interpolation] = STATE(269),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1057),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [270] = {
    [sym_text_interpolation] = STATE(270),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1022),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [271] = {
    [sym_text_interpolation] = STATE(271),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1023),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [272] = {
    [sym_text_interpolation] = STATE(272),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1050),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [273] = {
    [sym_text_interpolation] = STATE(273),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1053),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [274] = {
    [sym_text_interpolation] = STATE(274),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1054),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [275] = {
    [sym_text_interpolation] = STATE(275),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1055),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [276] = {
    [sym_text_interpolation] = STATE(276),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1056),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [277] = {
    [sym_text_interpolation] = STATE(277),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1084),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [278] = {
    [sym_text_interpolation] = STATE(278),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1086),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [279] = {
    [sym_text_interpolation] = STATE(279),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1058),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [280] = {
    [sym_text_interpolation] = STATE(280),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1059),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [281] = {
    [sym_text_interpolation] = STATE(281),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1060),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [282] = {
    [sym_text_interpolation] = STATE(282),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1061),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [283] = {
    [sym_text_interpolation] = STATE(283),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1049),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [284] = {
    [sym_text_interpolation] = STATE(284),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1063),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [285] = {
    [sym_text_interpolation] = STATE(285),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1064),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [286] = {
    [sym_text_interpolation] = STATE(286),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1082),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [287] = {
    [sym_text_interpolation] = STATE(287),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1066),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [288] = {
    [sym_text_interpolation] = STATE(288),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1067),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [289] = {
    [sym_text_interpolation] = STATE(289),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1068),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [290] = {
    [sym_text_interpolation] = STATE(290),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1069),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [291] = {
    [sym_text_interpolation] = STATE(291),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1070),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [292] = {
    [sym_text_interpolation] = STATE(292),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1071),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [293] = {
    [sym_text_interpolation] = STATE(293),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1073),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [294] = {
    [sym_text_interpolation] = STATE(294),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1074),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [295] = {
    [sym_text_interpolation] = STATE(295),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1075),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [296] = {
    [sym_text_interpolation] = STATE(296),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1076),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [297] = {
    [sym_text_interpolation] = STATE(297),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1078),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [298] = {
    [sym_text_interpolation] = STATE(298),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1081),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [299] = {
    [sym_text_interpolation] = STATE(299),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1062),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [300] = {
    [sym_text_interpolation] = STATE(300),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1065),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [301] = {
    [sym_text_interpolation] = STATE(301),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1174),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [302] = {
    [sym_text_interpolation] = STATE(302),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1166),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [303] = {
    [sym_text_interpolation] = STATE(303),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1293),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [304] = {
    [sym_text_interpolation] = STATE(304),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1139),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [305] = {
    [sym_text_interpolation] = STATE(305),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1320),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [306] = {
    [sym_text_interpolation] = STATE(306),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1162),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [307] = {
    [sym_text_interpolation] = STATE(307),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1317),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [308] = {
    [sym_text_interpolation] = STATE(308),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1319),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [309] = {
    [sym_text_interpolation] = STATE(309),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1259),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [310] = {
    [sym_text_interpolation] = STATE(310),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1095),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [311] = {
    [sym_text_interpolation] = STATE(311),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1279),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [312] = {
    [sym_text_interpolation] = STATE(312),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1096),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [313] = {
    [sym_text_interpolation] = STATE(313),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1097),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [314] = {
    [sym_text_interpolation] = STATE(314),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1098),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [315] = {
    [sym_text_interpolation] = STATE(315),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1268),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [316] = {
    [sym_text_interpolation] = STATE(316),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1101),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [317] = {
    [sym_text_interpolation] = STATE(317),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1292),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [318] = {
    [sym_text_interpolation] = STATE(318),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1299),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [319] = {
    [sym_text_interpolation] = STATE(319),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1165),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [320] = {
    [sym_text_interpolation] = STATE(320),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1249),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [321] = {
    [sym_text_interpolation] = STATE(321),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1210),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [322] = {
    [sym_text_interpolation] = STATE(322),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1262),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [323] = {
    [sym_text_interpolation] = STATE(323),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1250),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [324] = {
    [sym_text_interpolation] = STATE(324),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1109),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [325] = {
    [sym_text_interpolation] = STATE(325),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1281),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [326] = {
    [sym_text_interpolation] = STATE(326),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1079),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [327] = {
    [sym_text_interpolation] = STATE(327),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1284),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [328] = {
    [sym_text_interpolation] = STATE(328),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1285),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [329] = {
    [sym_text_interpolation] = STATE(329),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1289),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [330] = {
    [sym_text_interpolation] = STATE(330),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1290),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [331] = {
    [sym_text_interpolation] = STATE(331),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1260),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [332] = {
    [sym_text_interpolation] = STATE(332),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1273),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [333] = {
    [sym_text_interpolation] = STATE(333),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1116),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [334] = {
    [sym_text_interpolation] = STATE(334),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1117),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [335] = {
    [sym_text_interpolation] = STATE(335),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1118),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [336] = {
    [sym_text_interpolation] = STATE(336),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1167),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [337] = {
    [sym_text_interpolation] = STATE(337),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1022),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [338] = {
    [sym_text_interpolation] = STATE(338),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1023),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [339] = {
    [sym_text_interpolation] = STATE(339),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1168),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [340] = {
    [sym_text_interpolation] = STATE(340),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1170),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [341] = {
    [sym_text_interpolation] = STATE(341),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1171),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [342] = {
    [sym_text_interpolation] = STATE(342),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1083),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [343] = {
    [sym_text_interpolation] = STATE(343),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1173),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [344] = {
    [sym_text_interpolation] = STATE(344),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1138),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [345] = {
    [sym_text_interpolation] = STATE(345),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1140),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [346] = {
    [sym_text_interpolation] = STATE(346),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1141),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [347] = {
    [sym_text_interpolation] = STATE(347),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1142),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [348] = {
    [sym_text_interpolation] = STATE(348),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1143),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [349] = {
    [sym_text_interpolation] = STATE(349),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1144),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [350] = {
    [sym_text_interpolation] = STATE(350),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1145),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [351] = {
    [sym_text_interpolation] = STATE(351),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1146),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [352] = {
    [sym_text_interpolation] = STATE(352),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1147),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [353] = {
    [sym_text_interpolation] = STATE(353),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1275),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [354] = {
    [sym_text_interpolation] = STATE(354),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1149),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [355] = {
    [sym_text_interpolation] = STATE(355),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1150),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [356] = {
    [sym_text_interpolation] = STATE(356),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1151),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [357] = {
    [sym_text_interpolation] = STATE(357),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1152),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [358] = {
    [sym_text_interpolation] = STATE(358),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1153),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [359] = {
    [sym_text_interpolation] = STATE(359),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1154),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [360] = {
    [sym_text_interpolation] = STATE(360),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1155),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [361] = {
    [sym_text_interpolation] = STATE(361),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1156),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [362] = {
    [sym_text_interpolation] = STATE(362),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1158),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [363] = {
    [sym_text_interpolation] = STATE(363),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1159),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [364] = {
    [sym_text_interpolation] = STATE(364),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1160),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [365] = {
    [sym_text_interpolation] = STATE(365),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1218),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [366] = {
    [sym_text_interpolation] = STATE(366),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1161),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [367] = {
    [sym_text_interpolation] = STATE(367),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1219),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [368] = {
    [sym_text_interpolation] = STATE(368),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1246),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [369] = {
    [sym_text_interpolation] = STATE(369),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1258),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [370] = {
    [sym_text_interpolation] = STATE(370),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1220),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [371] = {
    [sym_text_interpolation] = STATE(371),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1257),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [372] = {
    [sym_text_interpolation] = STATE(372),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1211),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [373] = {
    [sym_text_interpolation] = STATE(373),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1241),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [374] = {
    [sym_text_interpolation] = STATE(374),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1119),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [375] = {
    [sym_text_interpolation] = STATE(375),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1120),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [376] = {
    [sym_text_interpolation] = STATE(376),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1121),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [377] = {
    [sym_text_interpolation] = STATE(377),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1122),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [378] = {
    [sym_text_interpolation] = STATE(378),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1123),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [379] = {
    [sym_text_interpolation] = STATE(379),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1327),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [380] = {
    [sym_text_interpolation] = STATE(380),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1203),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [381] = {
    [sym_text_interpolation] = STATE(381),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1294),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [382] = {
    [sym_text_interpolation] = STATE(382),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1295),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [383] = {
    [sym_text_interpolation] = STATE(383),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1124),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [384] = {
    [sym_text_interpolation] = STATE(384),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1298),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [385] = {
    [sym_text_interpolation] = STATE(385),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1125),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [386] = {
    [sym_text_interpolation] = STATE(386),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1126),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [387] = {
    [sym_text_interpolation] = STATE(387),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1178),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [388] = {
    [sym_text_interpolation] = STATE(388),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1022),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [389] = {
    [sym_text_interpolation] = STATE(389),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1023),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [390] = {
    [sym_text_interpolation] = STATE(390),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1205),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [391] = {
    [sym_text_interpolation] = STATE(391),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1207),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [392] = {
    [sym_text_interpolation] = STATE(392),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1208),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [393] = {
    [sym_text_interpolation] = STATE(393),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1179),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [394] = {
    [sym_text_interpolation] = STATE(394),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1189),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [395] = {
    [sym_text_interpolation] = STATE(395),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1180),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [396] = {
    [sym_text_interpolation] = STATE(396),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1181),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [397] = {
    [sym_text_interpolation] = STATE(397),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1182),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [398] = {
    [sym_text_interpolation] = STATE(398),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1183),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [399] = {
    [sym_text_interpolation] = STATE(399),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1184),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [400] = {
    [sym_text_interpolation] = STATE(400),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1185),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [401] = {
    [sym_text_interpolation] = STATE(401),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [402] = {
    [sym_text_interpolation] = STATE(402),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1187),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [403] = {
    [sym_text_interpolation] = STATE(403),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1188),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [404] = {
    [sym_text_interpolation] = STATE(404),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1209),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [405] = {
    [sym_text_interpolation] = STATE(405),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1190),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [406] = {
    [sym_text_interpolation] = STATE(406),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1191),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [407] = {
    [sym_text_interpolation] = STATE(407),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1192),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [408] = {
    [sym_text_interpolation] = STATE(408),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1193),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [409] = {
    [sym_text_interpolation] = STATE(409),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1194),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [410] = {
    [sym_text_interpolation] = STATE(410),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [411] = {
    [sym_text_interpolation] = STATE(411),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1196),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [412] = {
    [sym_text_interpolation] = STATE(412),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1197),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [413] = {
    [sym_text_interpolation] = STATE(413),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1199),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [414] = {
    [sym_text_interpolation] = STATE(414),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1200),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [415] = {
    [sym_text_interpolation] = STATE(415),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1201),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [416] = {
    [sym_text_interpolation] = STATE(416),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1202),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [417] = {
    [sym_text_interpolation] = STATE(417),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1127),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [418] = {
    [sym_text_interpolation] = STATE(418),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1128),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [419] = {
    [sym_text_interpolation] = STATE(419),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1304),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [420] = {
    [sym_text_interpolation] = STATE(420),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1300),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [421] = {
    [sym_text_interpolation] = STATE(421),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1303),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [422] = {
    [sym_text_interpolation] = STATE(422),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1129),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [423] = {
    [sym_text_interpolation] = STATE(423),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1305),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [424] = {
    [sym_text_interpolation] = STATE(424),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1308),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [425] = {
    [sym_text_interpolation] = STATE(425),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1309),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [426] = {
    [sym_text_interpolation] = STATE(426),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1312),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [427] = {
    [sym_text_interpolation] = STATE(427),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1313),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [428] = {
    [sym_text_interpolation] = STATE(428),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1314),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [429] = {
    [sym_text_interpolation] = STATE(429),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1315),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [430] = {
    [sym_text_interpolation] = STATE(430),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1130),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [431] = {
    [sym_text_interpolation] = STATE(431),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1177),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [432] = {
    [sym_text_interpolation] = STATE(432),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1212),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [433] = {
    [sym_text_interpolation] = STATE(433),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_match_expression] = STATE(1090),
    [sym_expression] = STATE(1226),
    [sym__unary_expression] = STATE(1115),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1090),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(662),
    [sym_assignment_expression] = STATE(1090),
    [sym_reference_assignment_expression] = STATE(1090),
    [sym_conditional_expression] = STATE(1090),
    [sym_augmented_assignment_expression] = STATE(1090),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(662),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(662),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(662),
    [sym_list_literal] = STATE(2602),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(669),
    [sym__callable_variable] = STATE(636),
    [sym_variable_name] = STATE(630),
    [sym_yield_expression] = STATE(1090),
    [sym_binary_expression] = STATE(1090),
    [sym_include_expression] = STATE(1090),
    [sym_include_once_expression] = STATE(1090),
    [sym_require_expression] = STATE(1090),
    [sym_require_once_expression] = STATE(1090),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_match_expression_token1] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_yield_expression_token1] = ACTIONS(330),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [aux_sym_require_expression_token1] = ACTIONS(336),
    [aux_sym_require_once_expression_token1] = ACTIONS(338),
    [sym_comment] = ACTIONS(5),
  },
  [434] = {
    [sym_text_interpolation] = STATE(434),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1214),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [435] = {
    [sym_text_interpolation] = STATE(435),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1217),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [436] = {
    [sym_text_interpolation] = STATE(436),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1267),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(684),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(684),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(684),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(684),
    [sym_list_literal] = STATE(2639),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(685),
    [sym__callable_variable] = STATE(675),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(713),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(715),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [aux_sym_require_expression_token1] = ACTIONS(723),
    [aux_sym_require_once_expression_token1] = ACTIONS(725),
    [sym_comment] = ACTIONS(5),
  },
  [437] = {
    [sym_text_interpolation] = STATE(437),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1326),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(651),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(651),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(651),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(651),
    [sym_list_literal] = STATE(2663),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(629),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(631),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(647),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [aux_sym_require_expression_token1] = ACTIONS(655),
    [aux_sym_require_once_expression_token1] = ACTIONS(657),
    [sym_comment] = ACTIONS(5),
  },
  [438] = {
    [sym_text_interpolation] = STATE(438),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_match_expression] = STATE(1021),
    [sym_expression] = STATE(1148),
    [sym__unary_expression] = STATE(1042),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1021),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(658),
    [sym_assignment_expression] = STATE(1021),
    [sym_reference_assignment_expression] = STATE(1021),
    [sym_conditional_expression] = STATE(1021),
    [sym_augmented_assignment_expression] = STATE(1021),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(658),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(658),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(658),
    [sym_list_literal] = STATE(2548),
    [sym__list_destructing] = STATE(2329),
    [sym__array_destructing] = STATE(2329),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(648),
    [sym_variable_name] = STATE(597),
    [sym_yield_expression] = STATE(1021),
    [sym_binary_expression] = STATE(1021),
    [sym_include_expression] = STATE(1021),
    [sym_include_once_expression] = STATE(1021),
    [sym_require_expression] = STATE(1021),
    [sym_require_once_expression] = STATE(1021),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_match_expression_token1] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [aux_sym__list_destructing_token1] = ACTIONS(308),
    [anon_sym_LBRACK] = ACTIONS(675),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(5),
  },
  [439] = {
    [sym_text_interpolation] = STATE(439),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__unary_expression] = STATE(1029),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1029),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(623),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(623),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(623),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(623),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(622),
    [sym__callable_variable] = STATE(613),
    [sym_variable_name] = STATE(597),
    [sym_include_expression] = STATE(1029),
    [sym_include_once_expression] = STATE(1029),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(605),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [anon_sym_LBRACK] = ACTIONS(983),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [sym_comment] = ACTIONS(5),
  },
  [440] = {
    [sym_text_interpolation] = STATE(440),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__unary_expression] = STATE(1029),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1029),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(679),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(679),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(679),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(679),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(677),
    [sym__callable_variable] = STATE(667),
    [sym_variable_name] = STATE(597),
    [sym_include_expression] = STATE(1029),
    [sym_include_once_expression] = STATE(1029),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(857),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [anon_sym_PLUS] = ACTIONS(617),
    [anon_sym_DASH] = ACTIONS(617),
    [anon_sym_TILDE] = ACTIONS(619),
    [anon_sym_BANG] = ACTIONS(619),
    [anon_sym_AT] = ACTIONS(621),
    [aux_sym_clone_expression_token1] = ACTIONS(623),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [anon_sym_LBRACK] = ACTIONS(983),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_include_expression_token1] = ACTIONS(651),
    [aux_sym_include_once_expression_token1] = ACTIONS(653),
    [sym_comment] = ACTIONS(5),
  },
  [441] = {
    [sym_text_interpolation] = STATE(441),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__unary_expression] = STATE(1029),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1029),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(623),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(623),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(623),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(623),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(622),
    [sym__callable_variable] = STATE(613),
    [sym_variable_name] = STATE(597),
    [sym_include_expression] = STATE(1029),
    [sym_include_once_expression] = STATE(1029),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(659),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [anon_sym_PLUS] = ACTIONS(665),
    [anon_sym_DASH] = ACTIONS(665),
    [anon_sym_TILDE] = ACTIONS(667),
    [anon_sym_BANG] = ACTIONS(667),
    [anon_sym_AT] = ACTIONS(669),
    [aux_sym_clone_expression_token1] = ACTIONS(671),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [anon_sym_LBRACK] = ACTIONS(983),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [sym_comment] = ACTIONS(5),
  },
  [442] = {
    [sym_text_interpolation] = STATE(442),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1704),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym__unary_expression] = STATE(1029),
    [sym_unary_op_expression] = STATE(1030),
    [sym_error_suppression_expression] = STATE(1029),
    [sym_clone_expression] = STATE(1030),
    [sym_primary_expression] = STATE(1030),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_expression] = STATE(1030),
    [sym_cast_variable] = STATE(623),
    [sym__variable_member_access_expression] = STATE(614),
    [sym_member_access_expression] = STATE(623),
    [sym__variable_nullsafe_member_access_expression] = STATE(616),
    [sym_nullsafe_member_access_expression] = STATE(623),
    [sym__variable_scoped_property_access_expression] = STATE(618),
    [sym_scoped_property_access_expression] = STATE(623),
    [sym_function_call_expression] = STATE(596),
    [sym__callable_expression] = STATE(2341),
    [sym_scoped_call_expression] = STATE(596),
    [sym__scope_resolution_qualifier] = STATE(2691),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(596),
    [sym_nullsafe_member_call_expression] = STATE(596),
    [sym__variable_subscript_expression] = STATE(605),
    [sym__dereferencable_subscript_expression] = STATE(596),
    [sym__dereferencable_expression] = STATE(1848),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(597),
    [sym__simple_variable] = STATE(602),
    [sym__new_variable] = STATE(622),
    [sym__callable_variable] = STATE(613),
    [sym_variable_name] = STATE(597),
    [sym_include_expression] = STATE(1029),
    [sym_include_once_expression] = STATE(1029),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(697),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [anon_sym_PLUS] = ACTIONS(703),
    [anon_sym_DASH] = ACTIONS(703),
    [anon_sym_TILDE] = ACTIONS(705),
    [anon_sym_BANG] = ACTIONS(705),
    [anon_sym_AT] = ACTIONS(707),
    [aux_sym_clone_expression_token1] = ACTIONS(709),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [anon_sym_LBRACK] = ACTIONS(983),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(645),
    [aux_sym_include_expression_token1] = ACTIONS(719),
    [aux_sym_include_once_expression_token1] = ACTIONS(721),
    [sym_comment] = ACTIONS(5),
  },
  [443] = {
    [sym_text_interpolation] = STATE(443),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1631),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym__unary_expression] = STATE(1175),
    [sym_unary_op_expression] = STATE(1136),
    [sym_error_suppression_expression] = STATE(1175),
    [sym_clone_expression] = STATE(1136),
    [sym_primary_expression] = STATE(1136),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_expression] = STATE(1136),
    [sym_cast_variable] = STATE(659),
    [sym__variable_member_access_expression] = STATE(663),
    [sym_member_access_expression] = STATE(659),
    [sym__variable_nullsafe_member_access_expression] = STATE(664),
    [sym_nullsafe_member_access_expression] = STATE(659),
    [sym__variable_scoped_property_access_expression] = STATE(665),
    [sym_scoped_property_access_expression] = STATE(659),
    [sym_function_call_expression] = STATE(646),
    [sym__callable_expression] = STATE(2334),
    [sym_scoped_call_expression] = STATE(646),
    [sym__scope_resolution_qualifier] = STATE(2630),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(646),
    [sym_nullsafe_member_call_expression] = STATE(646),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(646),
    [sym__dereferencable_expression] = STATE(1902),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(630),
    [sym__simple_variable] = STATE(656),
    [sym__new_variable] = STATE(668),
    [sym__callable_variable] = STATE(649),
    [sym_variable_name] = STATE(630),
    [sym_include_expression] = STATE(1175),
    [sym_include_once_expression] = STATE(1175),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [anon_sym_PLUS] = ACTIONS(294),
    [anon_sym_DASH] = ACTIONS(294),
    [anon_sym_TILDE] = ACTIONS(296),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_AT] = ACTIONS(298),
    [aux_sym_clone_expression_token1] = ACTIONS(300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [anon_sym_LBRACK] = ACTIONS(985),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(328),
    [aux_sym_include_expression_token1] = ACTIONS(332),
    [aux_sym_include_once_expression_token1] = ACTIONS(334),
    [sym_comment] = ACTIONS(5),
  },
  [444] = {
    [sym_text_interpolation] = STATE(444),
    [sym_catch_clause] = STATE(454),
    [sym_finally_clause] = STATE(454),
    [aux_sym_try_statement_repeat1] = STATE(445),
    [ts_builtin_sym_end] = ACTIONS(987),
    [sym_name] = ACTIONS(989),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(987),
    [aux_sym_function_static_declaration_token1] = ACTIONS(989),
    [aux_sym_global_declaration_token1] = ACTIONS(989),
    [aux_sym_namespace_definition_token1] = ACTIONS(989),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(989),
    [aux_sym__namespace_use_type_token1] = ACTIONS(989),
    [aux_sym__namespace_use_type_token2] = ACTIONS(989),
    [anon_sym_BSLASH] = ACTIONS(987),
    [anon_sym_LBRACE] = ACTIONS(987),
    [anon_sym_RBRACE] = ACTIONS(987),
    [aux_sym_trait_declaration_token1] = ACTIONS(989),
    [aux_sym_interface_declaration_token1] = ACTIONS(989),
    [aux_sym_enum_declaration_token1] = ACTIONS(989),
    [aux_sym_enum_case_token1] = ACTIONS(989),
    [aux_sym_class_declaration_token1] = ACTIONS(989),
    [aux_sym_final_modifier_token1] = ACTIONS(989),
    [aux_sym_abstract_modifier_token1] = ACTIONS(989),
    [aux_sym_readonly_modifier_token1] = ACTIONS(989),
    [sym_var_modifier] = ACTIONS(989),
    [aux_sym_visibility_modifier_token1] = ACTIONS(989),
    [aux_sym_visibility_modifier_token2] = ACTIONS(989),
    [aux_sym_visibility_modifier_token3] = ACTIONS(989),
    [anon_sym_LPAREN2] = ACTIONS(987),
    [aux_sym__arrow_function_header_token1] = ACTIONS(989),
    [aux_sym_cast_type_token1] = ACTIONS(989),
    [aux_sym_echo_statement_token1] = ACTIONS(989),
    [aux_sym_exit_statement_token1] = ACTIONS(989),
    [anon_sym_unset] = ACTIONS(989),
    [aux_sym_declare_statement_token1] = ACTIONS(989),
    [aux_sym_declare_statement_token2] = ACTIONS(989),
    [sym_float] = ACTIONS(989),
    [aux_sym_try_statement_token1] = ACTIONS(989),
    [aux_sym_catch_clause_token1] = ACTIONS(991),
    [aux_sym_finally_clause_token1] = ACTIONS(993),
    [aux_sym_goto_statement_token1] = ACTIONS(989),
    [aux_sym_continue_statement_token1] = ACTIONS(989),
    [aux_sym_break_statement_token1] = ACTIONS(989),
    [sym_integer] = ACTIONS(989),
    [aux_sym_return_statement_token1] = ACTIONS(989),
    [aux_sym_throw_expression_token1] = ACTIONS(989),
    [aux_sym_while_statement_token1] = ACTIONS(989),
    [aux_sym_while_statement_token2] = ACTIONS(989),
    [aux_sym_do_statement_token1] = ACTIONS(989),
    [aux_sym_for_statement_token1] = ACTIONS(989),
    [aux_sym_for_statement_token2] = ACTIONS(989),
    [aux_sym_foreach_statement_token1] = ACTIONS(989),
    [aux_sym_foreach_statement_token2] = ACTIONS(989),
    [aux_sym_if_statement_token1] = ACTIONS(989),
    [aux_sym_if_statement_token2] = ACTIONS(989),
    [aux_sym_else_if_clause_token1] = ACTIONS(989),
    [aux_sym_else_clause_token1] = ACTIONS(989),
    [aux_sym_match_expression_token1] = ACTIONS(989),
    [aux_sym_match_default_expression_token1] = ACTIONS(989),
    [aux_sym_switch_statement_token1] = ACTIONS(989),
    [aux_sym_switch_block_token1] = ACTIONS(989),
    [anon_sym_PLUS] = ACTIONS(989),
    [anon_sym_DASH] = ACTIONS(989),
    [anon_sym_TILDE] = ACTIONS(987),
    [anon_sym_BANG] = ACTIONS(987),
    [anon_sym_AT] = ACTIONS(987),
    [aux_sym_clone_expression_token1] = ACTIONS(989),
    [aux_sym_print_intrinsic_token1] = ACTIONS(989),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(989),
    [anon_sym_DASH_DASH] = ACTIONS(987),
    [anon_sym_PLUS_PLUS] = ACTIONS(987),
    [aux_sym__list_destructing_token1] = ACTIONS(989),
    [anon_sym_LBRACK] = ACTIONS(987),
    [anon_sym_self] = ACTIONS(989),
    [anon_sym_parent] = ACTIONS(989),
    [aux_sym__argument_name_token1] = ACTIONS(989),
    [aux_sym__argument_name_token2] = ACTIONS(989),
    [anon_sym_POUND_LBRACK] = ACTIONS(987),
    [aux_sym_encapsed_string_token1] = ACTIONS(987),
    [anon_sym_DQUOTE] = ACTIONS(987),
    [aux_sym_string_token1] = ACTIONS(987),
    [anon_sym_SQUOTE] = ACTIONS(987),
    [anon_sym_LT_LT_LT] = ACTIONS(987),
    [anon_sym_BQUOTE] = ACTIONS(987),
    [anon_sym_DOLLAR] = ACTIONS(987),
    [aux_sym_yield_expression_token1] = ACTIONS(989),
    [aux_sym_include_expression_token1] = ACTIONS(989),
    [aux_sym_include_once_expression_token1] = ACTIONS(989),
    [aux_sym_require_expression_token1] = ACTIONS(989),
    [aux_sym_require_once_expression_token1] = ACTIONS(989),
    [sym_comment] = ACTIONS(5),
  },
  [445] = {
    [sym_text_interpolation] = STATE(445),
    [sym_catch_clause] = STATE(454),
    [sym_finally_clause] = STATE(454),
    [aux_sym_try_statement_repeat1] = STATE(445),
    [ts_builtin_sym_end] = ACTIONS(995),
    [sym_name] = ACTIONS(997),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(995),
    [aux_sym_function_static_declaration_token1] = ACTIONS(997),
    [aux_sym_global_declaration_token1] = ACTIONS(997),
    [aux_sym_namespace_definition_token1] = ACTIONS(997),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(997),
    [aux_sym__namespace_use_type_token1] = ACTIONS(997),
    [aux_sym__namespace_use_type_token2] = ACTIONS(997),
    [anon_sym_BSLASH] = ACTIONS(995),
    [anon_sym_LBRACE] = ACTIONS(995),
    [anon_sym_RBRACE] = ACTIONS(995),
    [aux_sym_trait_declaration_token1] = ACTIONS(997),
    [aux_sym_interface_declaration_token1] = ACTIONS(997),
    [aux_sym_enum_declaration_token1] = ACTIONS(997),
    [aux_sym_enum_case_token1] = ACTIONS(997),
    [aux_sym_class_declaration_token1] = ACTIONS(997),
    [aux_sym_final_modifier_token1] = ACTIONS(997),
    [aux_sym_abstract_modifier_token1] = ACTIONS(997),
    [aux_sym_readonly_modifier_token1] = ACTIONS(997),
    [sym_var_modifier] = ACTIONS(997),
    [aux_sym_visibility_modifier_token1] = ACTIONS(997),
    [aux_sym_visibility_modifier_token2] = ACTIONS(997),
    [aux_sym_visibility_modifier_token3] = ACTIONS(997),
    [anon_sym_LPAREN2] = ACTIONS(995),
    [aux_sym__arrow_function_header_token1] = ACTIONS(997),
    [aux_sym_cast_type_token1] = ACTIONS(997),
    [aux_sym_echo_statement_token1] = ACTIONS(997),
    [aux_sym_exit_statement_token1] = ACTIONS(997),
    [anon_sym_unset] = ACTIONS(997),
    [aux_sym_declare_statement_token1] = ACTIONS(997),
    [aux_sym_declare_statement_token2] = ACTIONS(997),
    [sym_float] = ACTIONS(997),
    [aux_sym_try_statement_token1] = ACTIONS(997),
    [aux_sym_catch_clause_token1] = ACTIONS(999),
    [aux_sym_finally_clause_token1] = ACTIONS(1002),
    [aux_sym_goto_statement_token1] = ACTIONS(997),
    [aux_sym_continue_statement_token1] = ACTIONS(997),
    [aux_sym_break_statement_token1] = ACTIONS(997),
    [sym_integer] = ACTIONS(997),
    [aux_sym_return_statement_token1] = ACTIONS(997),
    [aux_sym_throw_expression_token1] = ACTIONS(997),
    [aux_sym_while_statement_token1] = ACTIONS(997),
    [aux_sym_while_statement_token2] = ACTIONS(997),
    [aux_sym_do_statement_token1] = ACTIONS(997),
    [aux_sym_for_statement_token1] = ACTIONS(997),
    [aux_sym_for_statement_token2] = ACTIONS(997),
    [aux_sym_foreach_statement_token1] = ACTIONS(997),
    [aux_sym_foreach_statement_token2] = ACTIONS(997),
    [aux_sym_if_statement_token1] = ACTIONS(997),
    [aux_sym_if_statement_token2] = ACTIONS(997),
    [aux_sym_else_if_clause_token1] = ACTIONS(997),
    [aux_sym_else_clause_token1] = ACTIONS(997),
    [aux_sym_match_expression_token1] = ACTIONS(997),
    [aux_sym_match_default_expression_token1] = ACTIONS(997),
    [aux_sym_switch_statement_token1] = ACTIONS(997),
    [aux_sym_switch_block_token1] = ACTIONS(997),
    [anon_sym_PLUS] = ACTIONS(997),
    [anon_sym_DASH] = ACTIONS(997),
    [anon_sym_TILDE] = ACTIONS(995),
    [anon_sym_BANG] = ACTIONS(995),
    [anon_sym_AT] = ACTIONS(995),
    [aux_sym_clone_expression_token1] = ACTIONS(997),
    [aux_sym_print_intrinsic_token1] = ACTIONS(997),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(997),
    [anon_sym_DASH_DASH] = ACTIONS(995),
    [anon_sym_PLUS_PLUS] = ACTIONS(995),
    [aux_sym__list_destructing_token1] = ACTIONS(997),
    [anon_sym_LBRACK] = ACTIONS(995),
    [anon_sym_self] = ACTIONS(997),
    [anon_sym_parent] = ACTIONS(997),
    [aux_sym__argument_name_token1] = ACTIONS(997),
    [aux_sym__argument_name_token2] = ACTIONS(997),
    [anon_sym_POUND_LBRACK] = ACTIONS(995),
    [aux_sym_encapsed_string_token1] = ACTIONS(995),
    [anon_sym_DQUOTE] = ACTIONS(995),
    [aux_sym_string_token1] = ACTIONS(995),
    [anon_sym_SQUOTE] = ACTIONS(995),
    [anon_sym_LT_LT_LT] = ACTIONS(995),
    [anon_sym_BQUOTE] = ACTIONS(995),
    [anon_sym_DOLLAR] = ACTIONS(995),
    [aux_sym_yield_expression_token1] = ACTIONS(997),
    [aux_sym_include_expression_token1] = ACTIONS(997),
    [aux_sym_include_once_expression_token1] = ACTIONS(997),
    [aux_sym_require_expression_token1] = ACTIONS(997),
    [aux_sym_require_once_expression_token1] = ACTIONS(997),
    [sym_comment] = ACTIONS(5),
  },
  [446] = {
    [sym_text_interpolation] = STATE(446),
    [sym_else_if_clause] = STATE(461),
    [sym_else_clause] = STATE(462),
    [aux_sym_if_statement_repeat1] = STATE(447),
    [ts_builtin_sym_end] = ACTIONS(1005),
    [sym_name] = ACTIONS(1007),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1005),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1007),
    [aux_sym_global_declaration_token1] = ACTIONS(1007),
    [aux_sym_namespace_definition_token1] = ACTIONS(1007),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1007),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1007),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1007),
    [anon_sym_BSLASH] = ACTIONS(1005),
    [anon_sym_LBRACE] = ACTIONS(1005),
    [anon_sym_RBRACE] = ACTIONS(1005),
    [aux_sym_trait_declaration_token1] = ACTIONS(1007),
    [aux_sym_interface_declaration_token1] = ACTIONS(1007),
    [aux_sym_enum_declaration_token1] = ACTIONS(1007),
    [aux_sym_enum_case_token1] = ACTIONS(1007),
    [aux_sym_class_declaration_token1] = ACTIONS(1007),
    [aux_sym_final_modifier_token1] = ACTIONS(1007),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1007),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1007),
    [sym_var_modifier] = ACTIONS(1007),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1007),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1007),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1007),
    [anon_sym_LPAREN2] = ACTIONS(1005),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1007),
    [aux_sym_cast_type_token1] = ACTIONS(1007),
    [aux_sym_echo_statement_token1] = ACTIONS(1007),
    [aux_sym_exit_statement_token1] = ACTIONS(1007),
    [anon_sym_unset] = ACTIONS(1007),
    [aux_sym_declare_statement_token1] = ACTIONS(1007),
    [aux_sym_declare_statement_token2] = ACTIONS(1007),
    [sym_float] = ACTIONS(1007),
    [aux_sym_try_statement_token1] = ACTIONS(1007),
    [aux_sym_goto_statement_token1] = ACTIONS(1007),
    [aux_sym_continue_statement_token1] = ACTIONS(1007),
    [aux_sym_break_statement_token1] = ACTIONS(1007),
    [sym_integer] = ACTIONS(1007),
    [aux_sym_return_statement_token1] = ACTIONS(1007),
    [aux_sym_throw_expression_token1] = ACTIONS(1007),
    [aux_sym_while_statement_token1] = ACTIONS(1007),
    [aux_sym_while_statement_token2] = ACTIONS(1007),
    [aux_sym_do_statement_token1] = ACTIONS(1007),
    [aux_sym_for_statement_token1] = ACTIONS(1007),
    [aux_sym_for_statement_token2] = ACTIONS(1007),
    [aux_sym_foreach_statement_token1] = ACTIONS(1007),
    [aux_sym_foreach_statement_token2] = ACTIONS(1007),
    [aux_sym_if_statement_token1] = ACTIONS(1007),
    [aux_sym_if_statement_token2] = ACTIONS(1007),
    [aux_sym_else_if_clause_token1] = ACTIONS(1009),
    [aux_sym_else_clause_token1] = ACTIONS(1012),
    [aux_sym_match_expression_token1] = ACTIONS(1007),
    [aux_sym_match_default_expression_token1] = ACTIONS(1007),
    [aux_sym_switch_statement_token1] = ACTIONS(1007),
    [aux_sym_switch_block_token1] = ACTIONS(1007),
    [anon_sym_PLUS] = ACTIONS(1007),
    [anon_sym_DASH] = ACTIONS(1007),
    [anon_sym_TILDE] = ACTIONS(1005),
    [anon_sym_BANG] = ACTIONS(1005),
    [anon_sym_AT] = ACTIONS(1005),
    [aux_sym_clone_expression_token1] = ACTIONS(1007),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1007),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1005),
    [aux_sym__list_destructing_token1] = ACTIONS(1007),
    [anon_sym_LBRACK] = ACTIONS(1005),
    [anon_sym_self] = ACTIONS(1007),
    [anon_sym_parent] = ACTIONS(1007),
    [aux_sym__argument_name_token1] = ACTIONS(1007),
    [aux_sym__argument_name_token2] = ACTIONS(1007),
    [anon_sym_POUND_LBRACK] = ACTIONS(1005),
    [aux_sym_encapsed_string_token1] = ACTIONS(1005),
    [anon_sym_DQUOTE] = ACTIONS(1005),
    [aux_sym_string_token1] = ACTIONS(1005),
    [anon_sym_SQUOTE] = ACTIONS(1005),
    [anon_sym_LT_LT_LT] = ACTIONS(1005),
    [anon_sym_BQUOTE] = ACTIONS(1005),
    [anon_sym_DOLLAR] = ACTIONS(1005),
    [aux_sym_yield_expression_token1] = ACTIONS(1007),
    [aux_sym_include_expression_token1] = ACTIONS(1007),
    [aux_sym_include_once_expression_token1] = ACTIONS(1007),
    [aux_sym_require_expression_token1] = ACTIONS(1007),
    [aux_sym_require_once_expression_token1] = ACTIONS(1007),
    [sym_comment] = ACTIONS(5),
  },
  [447] = {
    [sym_text_interpolation] = STATE(447),
    [sym_else_if_clause] = STATE(461),
    [sym_else_clause] = STATE(489),
    [aux_sym_if_statement_repeat1] = STATE(451),
    [ts_builtin_sym_end] = ACTIONS(1015),
    [sym_name] = ACTIONS(1017),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1015),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1017),
    [aux_sym_global_declaration_token1] = ACTIONS(1017),
    [aux_sym_namespace_definition_token1] = ACTIONS(1017),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1017),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1017),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1017),
    [anon_sym_BSLASH] = ACTIONS(1015),
    [anon_sym_LBRACE] = ACTIONS(1015),
    [anon_sym_RBRACE] = ACTIONS(1015),
    [aux_sym_trait_declaration_token1] = ACTIONS(1017),
    [aux_sym_interface_declaration_token1] = ACTIONS(1017),
    [aux_sym_enum_declaration_token1] = ACTIONS(1017),
    [aux_sym_enum_case_token1] = ACTIONS(1017),
    [aux_sym_class_declaration_token1] = ACTIONS(1017),
    [aux_sym_final_modifier_token1] = ACTIONS(1017),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1017),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1017),
    [sym_var_modifier] = ACTIONS(1017),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1017),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1017),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1017),
    [anon_sym_LPAREN2] = ACTIONS(1015),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1017),
    [aux_sym_cast_type_token1] = ACTIONS(1017),
    [aux_sym_echo_statement_token1] = ACTIONS(1017),
    [aux_sym_exit_statement_token1] = ACTIONS(1017),
    [anon_sym_unset] = ACTIONS(1017),
    [aux_sym_declare_statement_token1] = ACTIONS(1017),
    [aux_sym_declare_statement_token2] = ACTIONS(1017),
    [sym_float] = ACTIONS(1017),
    [aux_sym_try_statement_token1] = ACTIONS(1017),
    [aux_sym_goto_statement_token1] = ACTIONS(1017),
    [aux_sym_continue_statement_token1] = ACTIONS(1017),
    [aux_sym_break_statement_token1] = ACTIONS(1017),
    [sym_integer] = ACTIONS(1017),
    [aux_sym_return_statement_token1] = ACTIONS(1017),
    [aux_sym_throw_expression_token1] = ACTIONS(1017),
    [aux_sym_while_statement_token1] = ACTIONS(1017),
    [aux_sym_while_statement_token2] = ACTIONS(1017),
    [aux_sym_do_statement_token1] = ACTIONS(1017),
    [aux_sym_for_statement_token1] = ACTIONS(1017),
    [aux_sym_for_statement_token2] = ACTIONS(1017),
    [aux_sym_foreach_statement_token1] = ACTIONS(1017),
    [aux_sym_foreach_statement_token2] = ACTIONS(1017),
    [aux_sym_if_statement_token1] = ACTIONS(1017),
    [aux_sym_if_statement_token2] = ACTIONS(1017),
    [aux_sym_else_if_clause_token1] = ACTIONS(1019),
    [aux_sym_else_clause_token1] = ACTIONS(1022),
    [aux_sym_match_expression_token1] = ACTIONS(1017),
    [aux_sym_match_default_expression_token1] = ACTIONS(1017),
    [aux_sym_switch_statement_token1] = ACTIONS(1017),
    [aux_sym_switch_block_token1] = ACTIONS(1017),
    [anon_sym_PLUS] = ACTIONS(1017),
    [anon_sym_DASH] = ACTIONS(1017),
    [anon_sym_TILDE] = ACTIONS(1015),
    [anon_sym_BANG] = ACTIONS(1015),
    [anon_sym_AT] = ACTIONS(1015),
    [aux_sym_clone_expression_token1] = ACTIONS(1017),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1017),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1017),
    [anon_sym_DASH_DASH] = ACTIONS(1015),
    [anon_sym_PLUS_PLUS] = ACTIONS(1015),
    [aux_sym__list_destructing_token1] = ACTIONS(1017),
    [anon_sym_LBRACK] = ACTIONS(1015),
    [anon_sym_self] = ACTIONS(1017),
    [anon_sym_parent] = ACTIONS(1017),
    [aux_sym__argument_name_token1] = ACTIONS(1017),
    [aux_sym__argument_name_token2] = ACTIONS(1017),
    [anon_sym_POUND_LBRACK] = ACTIONS(1015),
    [aux_sym_encapsed_string_token1] = ACTIONS(1015),
    [anon_sym_DQUOTE] = ACTIONS(1015),
    [aux_sym_string_token1] = ACTIONS(1015),
    [anon_sym_SQUOTE] = ACTIONS(1015),
    [anon_sym_LT_LT_LT] = ACTIONS(1015),
    [anon_sym_BQUOTE] = ACTIONS(1015),
    [anon_sym_DOLLAR] = ACTIONS(1015),
    [aux_sym_yield_expression_token1] = ACTIONS(1017),
    [aux_sym_include_expression_token1] = ACTIONS(1017),
    [aux_sym_include_once_expression_token1] = ACTIONS(1017),
    [aux_sym_require_expression_token1] = ACTIONS(1017),
    [aux_sym_require_once_expression_token1] = ACTIONS(1017),
    [sym_comment] = ACTIONS(5),
  },
  [448] = {
    [sym_text_interpolation] = STATE(448),
    [sym_else_if_clause] = STATE(461),
    [sym_else_clause] = STATE(489),
    [aux_sym_if_statement_repeat1] = STATE(451),
    [ts_builtin_sym_end] = ACTIONS(1015),
    [sym_name] = ACTIONS(1017),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1015),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1017),
    [aux_sym_global_declaration_token1] = ACTIONS(1017),
    [aux_sym_namespace_definition_token1] = ACTIONS(1017),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1017),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1017),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1017),
    [anon_sym_BSLASH] = ACTIONS(1015),
    [anon_sym_LBRACE] = ACTIONS(1015),
    [anon_sym_RBRACE] = ACTIONS(1015),
    [aux_sym_trait_declaration_token1] = ACTIONS(1017),
    [aux_sym_interface_declaration_token1] = ACTIONS(1017),
    [aux_sym_enum_declaration_token1] = ACTIONS(1017),
    [aux_sym_enum_case_token1] = ACTIONS(1017),
    [aux_sym_class_declaration_token1] = ACTIONS(1017),
    [aux_sym_final_modifier_token1] = ACTIONS(1017),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1017),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1017),
    [sym_var_modifier] = ACTIONS(1017),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1017),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1017),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1017),
    [anon_sym_LPAREN2] = ACTIONS(1015),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1017),
    [aux_sym_cast_type_token1] = ACTIONS(1017),
    [aux_sym_echo_statement_token1] = ACTIONS(1017),
    [aux_sym_exit_statement_token1] = ACTIONS(1017),
    [anon_sym_unset] = ACTIONS(1017),
    [aux_sym_declare_statement_token1] = ACTIONS(1017),
    [aux_sym_declare_statement_token2] = ACTIONS(1017),
    [sym_float] = ACTIONS(1017),
    [aux_sym_try_statement_token1] = ACTIONS(1017),
    [aux_sym_goto_statement_token1] = ACTIONS(1017),
    [aux_sym_continue_statement_token1] = ACTIONS(1017),
    [aux_sym_break_statement_token1] = ACTIONS(1017),
    [sym_integer] = ACTIONS(1017),
    [aux_sym_return_statement_token1] = ACTIONS(1017),
    [aux_sym_throw_expression_token1] = ACTIONS(1017),
    [aux_sym_while_statement_token1] = ACTIONS(1017),
    [aux_sym_while_statement_token2] = ACTIONS(1017),
    [aux_sym_do_statement_token1] = ACTIONS(1017),
    [aux_sym_for_statement_token1] = ACTIONS(1017),
    [aux_sym_for_statement_token2] = ACTIONS(1017),
    [aux_sym_foreach_statement_token1] = ACTIONS(1017),
    [aux_sym_foreach_statement_token2] = ACTIONS(1017),
    [aux_sym_if_statement_token1] = ACTIONS(1017),
    [aux_sym_if_statement_token2] = ACTIONS(1017),
    [aux_sym_else_if_clause_token1] = ACTIONS(1025),
    [aux_sym_else_clause_token1] = ACTIONS(1027),
    [aux_sym_match_expression_token1] = ACTIONS(1017),
    [aux_sym_match_default_expression_token1] = ACTIONS(1017),
    [aux_sym_switch_statement_token1] = ACTIONS(1017),
    [aux_sym_switch_block_token1] = ACTIONS(1017),
    [anon_sym_PLUS] = ACTIONS(1017),
    [anon_sym_DASH] = ACTIONS(1017),
    [anon_sym_TILDE] = ACTIONS(1015),
    [anon_sym_BANG] = ACTIONS(1015),
    [anon_sym_AT] = ACTIONS(1015),
    [aux_sym_clone_expression_token1] = ACTIONS(1017),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1017),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1017),
    [anon_sym_DASH_DASH] = ACTIONS(1015),
    [anon_sym_PLUS_PLUS] = ACTIONS(1015),
    [aux_sym__list_destructing_token1] = ACTIONS(1017),
    [anon_sym_LBRACK] = ACTIONS(1015),
    [anon_sym_self] = ACTIONS(1017),
    [anon_sym_parent] = ACTIONS(1017),
    [aux_sym__argument_name_token1] = ACTIONS(1017),
    [aux_sym__argument_name_token2] = ACTIONS(1017),
    [anon_sym_POUND_LBRACK] = ACTIONS(1015),
    [aux_sym_encapsed_string_token1] = ACTIONS(1015),
    [anon_sym_DQUOTE] = ACTIONS(1015),
    [aux_sym_string_token1] = ACTIONS(1015),
    [anon_sym_SQUOTE] = ACTIONS(1015),
    [anon_sym_LT_LT_LT] = ACTIONS(1015),
    [anon_sym_BQUOTE] = ACTIONS(1015),
    [anon_sym_DOLLAR] = ACTIONS(1015),
    [aux_sym_yield_expression_token1] = ACTIONS(1017),
    [aux_sym_include_expression_token1] = ACTIONS(1017),
    [aux_sym_include_once_expression_token1] = ACTIONS(1017),
    [aux_sym_require_expression_token1] = ACTIONS(1017),
    [aux_sym_require_once_expression_token1] = ACTIONS(1017),
    [sym_comment] = ACTIONS(5),
  },
  [449] = {
    [sym_text_interpolation] = STATE(449),
    [sym_else_if_clause] = STATE(461),
    [sym_else_clause] = STATE(462),
    [aux_sym_if_statement_repeat1] = STATE(448),
    [ts_builtin_sym_end] = ACTIONS(1005),
    [sym_name] = ACTIONS(1007),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1005),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1007),
    [aux_sym_global_declaration_token1] = ACTIONS(1007),
    [aux_sym_namespace_definition_token1] = ACTIONS(1007),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1007),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1007),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1007),
    [anon_sym_BSLASH] = ACTIONS(1005),
    [anon_sym_LBRACE] = ACTIONS(1005),
    [anon_sym_RBRACE] = ACTIONS(1005),
    [aux_sym_trait_declaration_token1] = ACTIONS(1007),
    [aux_sym_interface_declaration_token1] = ACTIONS(1007),
    [aux_sym_enum_declaration_token1] = ACTIONS(1007),
    [aux_sym_enum_case_token1] = ACTIONS(1007),
    [aux_sym_class_declaration_token1] = ACTIONS(1007),
    [aux_sym_final_modifier_token1] = ACTIONS(1007),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1007),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1007),
    [sym_var_modifier] = ACTIONS(1007),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1007),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1007),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1007),
    [anon_sym_LPAREN2] = ACTIONS(1005),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1007),
    [aux_sym_cast_type_token1] = ACTIONS(1007),
    [aux_sym_echo_statement_token1] = ACTIONS(1007),
    [aux_sym_exit_statement_token1] = ACTIONS(1007),
    [anon_sym_unset] = ACTIONS(1007),
    [aux_sym_declare_statement_token1] = ACTIONS(1007),
    [aux_sym_declare_statement_token2] = ACTIONS(1007),
    [sym_float] = ACTIONS(1007),
    [aux_sym_try_statement_token1] = ACTIONS(1007),
    [aux_sym_goto_statement_token1] = ACTIONS(1007),
    [aux_sym_continue_statement_token1] = ACTIONS(1007),
    [aux_sym_break_statement_token1] = ACTIONS(1007),
    [sym_integer] = ACTIONS(1007),
    [aux_sym_return_statement_token1] = ACTIONS(1007),
    [aux_sym_throw_expression_token1] = ACTIONS(1007),
    [aux_sym_while_statement_token1] = ACTIONS(1007),
    [aux_sym_while_statement_token2] = ACTIONS(1007),
    [aux_sym_do_statement_token1] = ACTIONS(1007),
    [aux_sym_for_statement_token1] = ACTIONS(1007),
    [aux_sym_for_statement_token2] = ACTIONS(1007),
    [aux_sym_foreach_statement_token1] = ACTIONS(1007),
    [aux_sym_foreach_statement_token2] = ACTIONS(1007),
    [aux_sym_if_statement_token1] = ACTIONS(1007),
    [aux_sym_if_statement_token2] = ACTIONS(1007),
    [aux_sym_else_if_clause_token1] = ACTIONS(1025),
    [aux_sym_else_clause_token1] = ACTIONS(1027),
    [aux_sym_match_expression_token1] = ACTIONS(1007),
    [aux_sym_match_default_expression_token1] = ACTIONS(1007),
    [aux_sym_switch_statement_token1] = ACTIONS(1007),
    [aux_sym_switch_block_token1] = ACTIONS(1007),
    [anon_sym_PLUS] = ACTIONS(1007),
    [anon_sym_DASH] = ACTIONS(1007),
    [anon_sym_TILDE] = ACTIONS(1005),
    [anon_sym_BANG] = ACTIONS(1005),
    [anon_sym_AT] = ACTIONS(1005),
    [aux_sym_clone_expression_token1] = ACTIONS(1007),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1007),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1005),
    [aux_sym__list_destructing_token1] = ACTIONS(1007),
    [anon_sym_LBRACK] = ACTIONS(1005),
    [anon_sym_self] = ACTIONS(1007),
    [anon_sym_parent] = ACTIONS(1007),
    [aux_sym__argument_name_token1] = ACTIONS(1007),
    [aux_sym__argument_name_token2] = ACTIONS(1007),
    [anon_sym_POUND_LBRACK] = ACTIONS(1005),
    [aux_sym_encapsed_string_token1] = ACTIONS(1005),
    [anon_sym_DQUOTE] = ACTIONS(1005),
    [aux_sym_string_token1] = ACTIONS(1005),
    [anon_sym_SQUOTE] = ACTIONS(1005),
    [anon_sym_LT_LT_LT] = ACTIONS(1005),
    [anon_sym_BQUOTE] = ACTIONS(1005),
    [anon_sym_DOLLAR] = ACTIONS(1005),
    [aux_sym_yield_expression_token1] = ACTIONS(1007),
    [aux_sym_include_expression_token1] = ACTIONS(1007),
    [aux_sym_include_once_expression_token1] = ACTIONS(1007),
    [aux_sym_require_expression_token1] = ACTIONS(1007),
    [aux_sym_require_once_expression_token1] = ACTIONS(1007),
    [sym_comment] = ACTIONS(5),
  },
  [450] = {
    [sym_text_interpolation] = STATE(450),
    [ts_builtin_sym_end] = ACTIONS(1029),
    [sym_name] = ACTIONS(1031),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1029),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1031),
    [aux_sym_global_declaration_token1] = ACTIONS(1031),
    [aux_sym_namespace_definition_token1] = ACTIONS(1031),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1031),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1031),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1031),
    [anon_sym_BSLASH] = ACTIONS(1029),
    [anon_sym_LBRACE] = ACTIONS(1029),
    [anon_sym_RBRACE] = ACTIONS(1029),
    [aux_sym_trait_declaration_token1] = ACTIONS(1031),
    [aux_sym_interface_declaration_token1] = ACTIONS(1031),
    [aux_sym_enum_declaration_token1] = ACTIONS(1031),
    [aux_sym_enum_case_token1] = ACTIONS(1031),
    [aux_sym_class_declaration_token1] = ACTIONS(1031),
    [aux_sym_final_modifier_token1] = ACTIONS(1031),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1031),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1031),
    [sym_var_modifier] = ACTIONS(1031),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1031),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1031),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1031),
    [anon_sym_LPAREN2] = ACTIONS(1029),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1031),
    [aux_sym_cast_type_token1] = ACTIONS(1031),
    [aux_sym_echo_statement_token1] = ACTIONS(1031),
    [aux_sym_exit_statement_token1] = ACTIONS(1031),
    [anon_sym_unset] = ACTIONS(1031),
    [aux_sym_declare_statement_token1] = ACTIONS(1031),
    [aux_sym_declare_statement_token2] = ACTIONS(1031),
    [sym_float] = ACTIONS(1031),
    [aux_sym_try_statement_token1] = ACTIONS(1031),
    [aux_sym_catch_clause_token1] = ACTIONS(1031),
    [aux_sym_finally_clause_token1] = ACTIONS(1031),
    [aux_sym_goto_statement_token1] = ACTIONS(1031),
    [aux_sym_continue_statement_token1] = ACTIONS(1031),
    [aux_sym_break_statement_token1] = ACTIONS(1031),
    [sym_integer] = ACTIONS(1031),
    [aux_sym_return_statement_token1] = ACTIONS(1031),
    [aux_sym_throw_expression_token1] = ACTIONS(1031),
    [aux_sym_while_statement_token1] = ACTIONS(1031),
    [aux_sym_while_statement_token2] = ACTIONS(1031),
    [aux_sym_do_statement_token1] = ACTIONS(1031),
    [aux_sym_for_statement_token1] = ACTIONS(1031),
    [aux_sym_for_statement_token2] = ACTIONS(1031),
    [aux_sym_foreach_statement_token1] = ACTIONS(1031),
    [aux_sym_foreach_statement_token2] = ACTIONS(1031),
    [aux_sym_if_statement_token1] = ACTIONS(1031),
    [aux_sym_if_statement_token2] = ACTIONS(1031),
    [aux_sym_else_if_clause_token1] = ACTIONS(1031),
    [aux_sym_else_clause_token1] = ACTIONS(1031),
    [aux_sym_match_expression_token1] = ACTIONS(1031),
    [aux_sym_match_default_expression_token1] = ACTIONS(1031),
    [aux_sym_switch_statement_token1] = ACTIONS(1031),
    [aux_sym_switch_block_token1] = ACTIONS(1031),
    [anon_sym_PLUS] = ACTIONS(1031),
    [anon_sym_DASH] = ACTIONS(1031),
    [anon_sym_TILDE] = ACTIONS(1029),
    [anon_sym_BANG] = ACTIONS(1029),
    [anon_sym_AT] = ACTIONS(1029),
    [aux_sym_clone_expression_token1] = ACTIONS(1031),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1031),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1031),
    [anon_sym_DASH_DASH] = ACTIONS(1029),
    [anon_sym_PLUS_PLUS] = ACTIONS(1029),
    [aux_sym__list_destructing_token1] = ACTIONS(1031),
    [anon_sym_LBRACK] = ACTIONS(1029),
    [anon_sym_self] = ACTIONS(1031),
    [anon_sym_parent] = ACTIONS(1031),
    [aux_sym__argument_name_token1] = ACTIONS(1031),
    [aux_sym__argument_name_token2] = ACTIONS(1031),
    [anon_sym_POUND_LBRACK] = ACTIONS(1029),
    [aux_sym_encapsed_string_token1] = ACTIONS(1029),
    [anon_sym_DQUOTE] = ACTIONS(1029),
    [aux_sym_string_token1] = ACTIONS(1029),
    [anon_sym_SQUOTE] = ACTIONS(1029),
    [anon_sym_LT_LT_LT] = ACTIONS(1029),
    [anon_sym_BQUOTE] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1029),
    [aux_sym_yield_expression_token1] = ACTIONS(1031),
    [aux_sym_include_expression_token1] = ACTIONS(1031),
    [aux_sym_include_once_expression_token1] = ACTIONS(1031),
    [aux_sym_require_expression_token1] = ACTIONS(1031),
    [aux_sym_require_once_expression_token1] = ACTIONS(1031),
    [sym_comment] = ACTIONS(5),
  },
  [451] = {
    [sym_text_interpolation] = STATE(451),
    [sym_else_if_clause] = STATE(461),
    [aux_sym_if_statement_repeat1] = STATE(451),
    [ts_builtin_sym_end] = ACTIONS(1033),
    [sym_name] = ACTIONS(1035),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1033),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1035),
    [aux_sym_global_declaration_token1] = ACTIONS(1035),
    [aux_sym_namespace_definition_token1] = ACTIONS(1035),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1035),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1035),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1035),
    [anon_sym_BSLASH] = ACTIONS(1033),
    [anon_sym_LBRACE] = ACTIONS(1033),
    [anon_sym_RBRACE] = ACTIONS(1033),
    [aux_sym_trait_declaration_token1] = ACTIONS(1035),
    [aux_sym_interface_declaration_token1] = ACTIONS(1035),
    [aux_sym_enum_declaration_token1] = ACTIONS(1035),
    [aux_sym_enum_case_token1] = ACTIONS(1035),
    [aux_sym_class_declaration_token1] = ACTIONS(1035),
    [aux_sym_final_modifier_token1] = ACTIONS(1035),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1035),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1035),
    [sym_var_modifier] = ACTIONS(1035),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1035),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1035),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1035),
    [anon_sym_LPAREN2] = ACTIONS(1033),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1035),
    [aux_sym_cast_type_token1] = ACTIONS(1035),
    [aux_sym_echo_statement_token1] = ACTIONS(1035),
    [aux_sym_exit_statement_token1] = ACTIONS(1035),
    [anon_sym_unset] = ACTIONS(1035),
    [aux_sym_declare_statement_token1] = ACTIONS(1035),
    [aux_sym_declare_statement_token2] = ACTIONS(1035),
    [sym_float] = ACTIONS(1035),
    [aux_sym_try_statement_token1] = ACTIONS(1035),
    [aux_sym_goto_statement_token1] = ACTIONS(1035),
    [aux_sym_continue_statement_token1] = ACTIONS(1035),
    [aux_sym_break_statement_token1] = ACTIONS(1035),
    [sym_integer] = ACTIONS(1035),
    [aux_sym_return_statement_token1] = ACTIONS(1035),
    [aux_sym_throw_expression_token1] = ACTIONS(1035),
    [aux_sym_while_statement_token1] = ACTIONS(1035),
    [aux_sym_while_statement_token2] = ACTIONS(1035),
    [aux_sym_do_statement_token1] = ACTIONS(1035),
    [aux_sym_for_statement_token1] = ACTIONS(1035),
    [aux_sym_for_statement_token2] = ACTIONS(1035),
    [aux_sym_foreach_statement_token1] = ACTIONS(1035),
    [aux_sym_foreach_statement_token2] = ACTIONS(1035),
    [aux_sym_if_statement_token1] = ACTIONS(1035),
    [aux_sym_if_statement_token2] = ACTIONS(1035),
    [aux_sym_else_if_clause_token1] = ACTIONS(1037),
    [aux_sym_else_clause_token1] = ACTIONS(1035),
    [aux_sym_match_expression_token1] = ACTIONS(1035),
    [aux_sym_match_default_expression_token1] = ACTIONS(1035),
    [aux_sym_switch_statement_token1] = ACTIONS(1035),
    [aux_sym_switch_block_token1] = ACTIONS(1035),
    [anon_sym_PLUS] = ACTIONS(1035),
    [anon_sym_DASH] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1033),
    [anon_sym_BANG] = ACTIONS(1033),
    [anon_sym_AT] = ACTIONS(1033),
    [aux_sym_clone_expression_token1] = ACTIONS(1035),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1035),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1035),
    [anon_sym_DASH_DASH] = ACTIONS(1033),
    [anon_sym_PLUS_PLUS] = ACTIONS(1033),
    [aux_sym__list_destructing_token1] = ACTIONS(1035),
    [anon_sym_LBRACK] = ACTIONS(1033),
    [anon_sym_self] = ACTIONS(1035),
    [anon_sym_parent] = ACTIONS(1035),
    [aux_sym__argument_name_token1] = ACTIONS(1035),
    [aux_sym__argument_name_token2] = ACTIONS(1035),
    [anon_sym_POUND_LBRACK] = ACTIONS(1033),
    [aux_sym_encapsed_string_token1] = ACTIONS(1033),
    [anon_sym_DQUOTE] = ACTIONS(1033),
    [aux_sym_string_token1] = ACTIONS(1033),
    [anon_sym_SQUOTE] = ACTIONS(1033),
    [anon_sym_LT_LT_LT] = ACTIONS(1033),
    [anon_sym_BQUOTE] = ACTIONS(1033),
    [anon_sym_DOLLAR] = ACTIONS(1033),
    [aux_sym_yield_expression_token1] = ACTIONS(1035),
    [aux_sym_include_expression_token1] = ACTIONS(1035),
    [aux_sym_include_once_expression_token1] = ACTIONS(1035),
    [aux_sym_require_expression_token1] = ACTIONS(1035),
    [aux_sym_require_once_expression_token1] = ACTIONS(1035),
    [sym_comment] = ACTIONS(5),
  },
  [452] = {
    [sym_text_interpolation] = STATE(452),
    [ts_builtin_sym_end] = ACTIONS(1040),
    [sym_name] = ACTIONS(1042),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1040),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1042),
    [aux_sym_global_declaration_token1] = ACTIONS(1042),
    [aux_sym_namespace_definition_token1] = ACTIONS(1042),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1042),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1042),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1042),
    [anon_sym_BSLASH] = ACTIONS(1040),
    [anon_sym_LBRACE] = ACTIONS(1040),
    [anon_sym_RBRACE] = ACTIONS(1040),
    [aux_sym_trait_declaration_token1] = ACTIONS(1042),
    [aux_sym_interface_declaration_token1] = ACTIONS(1042),
    [aux_sym_enum_declaration_token1] = ACTIONS(1042),
    [aux_sym_enum_case_token1] = ACTIONS(1042),
    [aux_sym_class_declaration_token1] = ACTIONS(1042),
    [aux_sym_final_modifier_token1] = ACTIONS(1042),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1042),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1042),
    [sym_var_modifier] = ACTIONS(1042),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1042),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1042),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1042),
    [anon_sym_LPAREN2] = ACTIONS(1040),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1042),
    [aux_sym_cast_type_token1] = ACTIONS(1042),
    [aux_sym_echo_statement_token1] = ACTIONS(1042),
    [aux_sym_exit_statement_token1] = ACTIONS(1042),
    [anon_sym_unset] = ACTIONS(1042),
    [aux_sym_declare_statement_token1] = ACTIONS(1042),
    [aux_sym_declare_statement_token2] = ACTIONS(1042),
    [sym_float] = ACTIONS(1042),
    [aux_sym_try_statement_token1] = ACTIONS(1042),
    [aux_sym_catch_clause_token1] = ACTIONS(1042),
    [aux_sym_finally_clause_token1] = ACTIONS(1042),
    [aux_sym_goto_statement_token1] = ACTIONS(1042),
    [aux_sym_continue_statement_token1] = ACTIONS(1042),
    [aux_sym_break_statement_token1] = ACTIONS(1042),
    [sym_integer] = ACTIONS(1042),
    [aux_sym_return_statement_token1] = ACTIONS(1042),
    [aux_sym_throw_expression_token1] = ACTIONS(1042),
    [aux_sym_while_statement_token1] = ACTIONS(1042),
    [aux_sym_while_statement_token2] = ACTIONS(1042),
    [aux_sym_do_statement_token1] = ACTIONS(1042),
    [aux_sym_for_statement_token1] = ACTIONS(1042),
    [aux_sym_for_statement_token2] = ACTIONS(1042),
    [aux_sym_foreach_statement_token1] = ACTIONS(1042),
    [aux_sym_foreach_statement_token2] = ACTIONS(1042),
    [aux_sym_if_statement_token1] = ACTIONS(1042),
    [aux_sym_if_statement_token2] = ACTIONS(1042),
    [aux_sym_else_if_clause_token1] = ACTIONS(1042),
    [aux_sym_else_clause_token1] = ACTIONS(1042),
    [aux_sym_match_expression_token1] = ACTIONS(1042),
    [aux_sym_match_default_expression_token1] = ACTIONS(1042),
    [aux_sym_switch_statement_token1] = ACTIONS(1042),
    [aux_sym_switch_block_token1] = ACTIONS(1042),
    [anon_sym_PLUS] = ACTIONS(1042),
    [anon_sym_DASH] = ACTIONS(1042),
    [anon_sym_TILDE] = ACTIONS(1040),
    [anon_sym_BANG] = ACTIONS(1040),
    [anon_sym_AT] = ACTIONS(1040),
    [aux_sym_clone_expression_token1] = ACTIONS(1042),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1042),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1042),
    [anon_sym_DASH_DASH] = ACTIONS(1040),
    [anon_sym_PLUS_PLUS] = ACTIONS(1040),
    [aux_sym__list_destructing_token1] = ACTIONS(1042),
    [anon_sym_LBRACK] = ACTIONS(1040),
    [anon_sym_self] = ACTIONS(1042),
    [anon_sym_parent] = ACTIONS(1042),
    [aux_sym__argument_name_token1] = ACTIONS(1042),
    [aux_sym__argument_name_token2] = ACTIONS(1042),
    [anon_sym_POUND_LBRACK] = ACTIONS(1040),
    [aux_sym_encapsed_string_token1] = ACTIONS(1040),
    [anon_sym_DQUOTE] = ACTIONS(1040),
    [aux_sym_string_token1] = ACTIONS(1040),
    [anon_sym_SQUOTE] = ACTIONS(1040),
    [anon_sym_LT_LT_LT] = ACTIONS(1040),
    [anon_sym_BQUOTE] = ACTIONS(1040),
    [anon_sym_DOLLAR] = ACTIONS(1040),
    [aux_sym_yield_expression_token1] = ACTIONS(1042),
    [aux_sym_include_expression_token1] = ACTIONS(1042),
    [aux_sym_include_once_expression_token1] = ACTIONS(1042),
    [aux_sym_require_expression_token1] = ACTIONS(1042),
    [aux_sym_require_once_expression_token1] = ACTIONS(1042),
    [sym_comment] = ACTIONS(5),
  },
  [453] = {
    [sym_text_interpolation] = STATE(453),
    [ts_builtin_sym_end] = ACTIONS(1044),
    [sym_name] = ACTIONS(1046),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1044),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1046),
    [aux_sym_global_declaration_token1] = ACTIONS(1046),
    [aux_sym_namespace_definition_token1] = ACTIONS(1046),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1046),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1046),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1046),
    [anon_sym_BSLASH] = ACTIONS(1044),
    [anon_sym_LBRACE] = ACTIONS(1044),
    [anon_sym_RBRACE] = ACTIONS(1044),
    [aux_sym_trait_declaration_token1] = ACTIONS(1046),
    [aux_sym_interface_declaration_token1] = ACTIONS(1046),
    [aux_sym_enum_declaration_token1] = ACTIONS(1046),
    [aux_sym_enum_case_token1] = ACTIONS(1046),
    [aux_sym_class_declaration_token1] = ACTIONS(1046),
    [aux_sym_final_modifier_token1] = ACTIONS(1046),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1046),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1046),
    [sym_var_modifier] = ACTIONS(1046),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1046),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1046),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1046),
    [anon_sym_LPAREN2] = ACTIONS(1044),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1046),
    [aux_sym_cast_type_token1] = ACTIONS(1046),
    [aux_sym_echo_statement_token1] = ACTIONS(1046),
    [aux_sym_exit_statement_token1] = ACTIONS(1046),
    [anon_sym_unset] = ACTIONS(1046),
    [aux_sym_declare_statement_token1] = ACTIONS(1046),
    [aux_sym_declare_statement_token2] = ACTIONS(1046),
    [sym_float] = ACTIONS(1046),
    [aux_sym_try_statement_token1] = ACTIONS(1046),
    [aux_sym_catch_clause_token1] = ACTIONS(1046),
    [aux_sym_finally_clause_token1] = ACTIONS(1046),
    [aux_sym_goto_statement_token1] = ACTIONS(1046),
    [aux_sym_continue_statement_token1] = ACTIONS(1046),
    [aux_sym_break_statement_token1] = ACTIONS(1046),
    [sym_integer] = ACTIONS(1046),
    [aux_sym_return_statement_token1] = ACTIONS(1046),
    [aux_sym_throw_expression_token1] = ACTIONS(1046),
    [aux_sym_while_statement_token1] = ACTIONS(1046),
    [aux_sym_while_statement_token2] = ACTIONS(1046),
    [aux_sym_do_statement_token1] = ACTIONS(1046),
    [aux_sym_for_statement_token1] = ACTIONS(1046),
    [aux_sym_for_statement_token2] = ACTIONS(1046),
    [aux_sym_foreach_statement_token1] = ACTIONS(1046),
    [aux_sym_foreach_statement_token2] = ACTIONS(1046),
    [aux_sym_if_statement_token1] = ACTIONS(1046),
    [aux_sym_if_statement_token2] = ACTIONS(1046),
    [aux_sym_else_if_clause_token1] = ACTIONS(1046),
    [aux_sym_else_clause_token1] = ACTIONS(1046),
    [aux_sym_match_expression_token1] = ACTIONS(1046),
    [aux_sym_match_default_expression_token1] = ACTIONS(1046),
    [aux_sym_switch_statement_token1] = ACTIONS(1046),
    [aux_sym_switch_block_token1] = ACTIONS(1046),
    [anon_sym_PLUS] = ACTIONS(1046),
    [anon_sym_DASH] = ACTIONS(1046),
    [anon_sym_TILDE] = ACTIONS(1044),
    [anon_sym_BANG] = ACTIONS(1044),
    [anon_sym_AT] = ACTIONS(1044),
    [aux_sym_clone_expression_token1] = ACTIONS(1046),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1046),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1046),
    [anon_sym_DASH_DASH] = ACTIONS(1044),
    [anon_sym_PLUS_PLUS] = ACTIONS(1044),
    [aux_sym__list_destructing_token1] = ACTIONS(1046),
    [anon_sym_LBRACK] = ACTIONS(1044),
    [anon_sym_self] = ACTIONS(1046),
    [anon_sym_parent] = ACTIONS(1046),
    [aux_sym__argument_name_token1] = ACTIONS(1046),
    [aux_sym__argument_name_token2] = ACTIONS(1046),
    [anon_sym_POUND_LBRACK] = ACTIONS(1044),
    [aux_sym_encapsed_string_token1] = ACTIONS(1044),
    [anon_sym_DQUOTE] = ACTIONS(1044),
    [aux_sym_string_token1] = ACTIONS(1044),
    [anon_sym_SQUOTE] = ACTIONS(1044),
    [anon_sym_LT_LT_LT] = ACTIONS(1044),
    [anon_sym_BQUOTE] = ACTIONS(1044),
    [anon_sym_DOLLAR] = ACTIONS(1044),
    [aux_sym_yield_expression_token1] = ACTIONS(1046),
    [aux_sym_include_expression_token1] = ACTIONS(1046),
    [aux_sym_include_once_expression_token1] = ACTIONS(1046),
    [aux_sym_require_expression_token1] = ACTIONS(1046),
    [aux_sym_require_once_expression_token1] = ACTIONS(1046),
    [sym_comment] = ACTIONS(5),
  },
  [454] = {
    [sym_text_interpolation] = STATE(454),
    [ts_builtin_sym_end] = ACTIONS(1048),
    [sym_name] = ACTIONS(1050),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1048),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1050),
    [aux_sym_global_declaration_token1] = ACTIONS(1050),
    [aux_sym_namespace_definition_token1] = ACTIONS(1050),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1050),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1050),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1050),
    [anon_sym_BSLASH] = ACTIONS(1048),
    [anon_sym_LBRACE] = ACTIONS(1048),
    [anon_sym_RBRACE] = ACTIONS(1048),
    [aux_sym_trait_declaration_token1] = ACTIONS(1050),
    [aux_sym_interface_declaration_token1] = ACTIONS(1050),
    [aux_sym_enum_declaration_token1] = ACTIONS(1050),
    [aux_sym_enum_case_token1] = ACTIONS(1050),
    [aux_sym_class_declaration_token1] = ACTIONS(1050),
    [aux_sym_final_modifier_token1] = ACTIONS(1050),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1050),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1050),
    [sym_var_modifier] = ACTIONS(1050),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1050),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1050),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1050),
    [anon_sym_LPAREN2] = ACTIONS(1048),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1050),
    [aux_sym_cast_type_token1] = ACTIONS(1050),
    [aux_sym_echo_statement_token1] = ACTIONS(1050),
    [aux_sym_exit_statement_token1] = ACTIONS(1050),
    [anon_sym_unset] = ACTIONS(1050),
    [aux_sym_declare_statement_token1] = ACTIONS(1050),
    [aux_sym_declare_statement_token2] = ACTIONS(1050),
    [sym_float] = ACTIONS(1050),
    [aux_sym_try_statement_token1] = ACTIONS(1050),
    [aux_sym_catch_clause_token1] = ACTIONS(1050),
    [aux_sym_finally_clause_token1] = ACTIONS(1050),
    [aux_sym_goto_statement_token1] = ACTIONS(1050),
    [aux_sym_continue_statement_token1] = ACTIONS(1050),
    [aux_sym_break_statement_token1] = ACTIONS(1050),
    [sym_integer] = ACTIONS(1050),
    [aux_sym_return_statement_token1] = ACTIONS(1050),
    [aux_sym_throw_expression_token1] = ACTIONS(1050),
    [aux_sym_while_statement_token1] = ACTIONS(1050),
    [aux_sym_while_statement_token2] = ACTIONS(1050),
    [aux_sym_do_statement_token1] = ACTIONS(1050),
    [aux_sym_for_statement_token1] = ACTIONS(1050),
    [aux_sym_for_statement_token2] = ACTIONS(1050),
    [aux_sym_foreach_statement_token1] = ACTIONS(1050),
    [aux_sym_foreach_statement_token2] = ACTIONS(1050),
    [aux_sym_if_statement_token1] = ACTIONS(1050),
    [aux_sym_if_statement_token2] = ACTIONS(1050),
    [aux_sym_else_if_clause_token1] = ACTIONS(1050),
    [aux_sym_else_clause_token1] = ACTIONS(1050),
    [aux_sym_match_expression_token1] = ACTIONS(1050),
    [aux_sym_match_default_expression_token1] = ACTIONS(1050),
    [aux_sym_switch_statement_token1] = ACTIONS(1050),
    [aux_sym_switch_block_token1] = ACTIONS(1050),
    [anon_sym_PLUS] = ACTIONS(1050),
    [anon_sym_DASH] = ACTIONS(1050),
    [anon_sym_TILDE] = ACTIONS(1048),
    [anon_sym_BANG] = ACTIONS(1048),
    [anon_sym_AT] = ACTIONS(1048),
    [aux_sym_clone_expression_token1] = ACTIONS(1050),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1050),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1050),
    [anon_sym_DASH_DASH] = ACTIONS(1048),
    [anon_sym_PLUS_PLUS] = ACTIONS(1048),
    [aux_sym__list_destructing_token1] = ACTIONS(1050),
    [anon_sym_LBRACK] = ACTIONS(1048),
    [anon_sym_self] = ACTIONS(1050),
    [anon_sym_parent] = ACTIONS(1050),
    [aux_sym__argument_name_token1] = ACTIONS(1050),
    [aux_sym__argument_name_token2] = ACTIONS(1050),
    [anon_sym_POUND_LBRACK] = ACTIONS(1048),
    [aux_sym_encapsed_string_token1] = ACTIONS(1048),
    [anon_sym_DQUOTE] = ACTIONS(1048),
    [aux_sym_string_token1] = ACTIONS(1048),
    [anon_sym_SQUOTE] = ACTIONS(1048),
    [anon_sym_LT_LT_LT] = ACTIONS(1048),
    [anon_sym_BQUOTE] = ACTIONS(1048),
    [anon_sym_DOLLAR] = ACTIONS(1048),
    [aux_sym_yield_expression_token1] = ACTIONS(1050),
    [aux_sym_include_expression_token1] = ACTIONS(1050),
    [aux_sym_include_once_expression_token1] = ACTIONS(1050),
    [aux_sym_require_expression_token1] = ACTIONS(1050),
    [aux_sym_require_once_expression_token1] = ACTIONS(1050),
    [sym_comment] = ACTIONS(5),
  },
  [455] = {
    [sym_text_interpolation] = STATE(455),
    [ts_builtin_sym_end] = ACTIONS(1052),
    [sym_name] = ACTIONS(1054),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1052),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1054),
    [aux_sym_global_declaration_token1] = ACTIONS(1054),
    [aux_sym_namespace_definition_token1] = ACTIONS(1054),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1054),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1054),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1054),
    [anon_sym_BSLASH] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1052),
    [anon_sym_RBRACE] = ACTIONS(1052),
    [aux_sym_trait_declaration_token1] = ACTIONS(1054),
    [aux_sym_interface_declaration_token1] = ACTIONS(1054),
    [aux_sym_enum_declaration_token1] = ACTIONS(1054),
    [aux_sym_enum_case_token1] = ACTIONS(1054),
    [aux_sym_class_declaration_token1] = ACTIONS(1054),
    [aux_sym_final_modifier_token1] = ACTIONS(1054),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1054),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1054),
    [sym_var_modifier] = ACTIONS(1054),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1054),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1054),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1054),
    [anon_sym_LPAREN2] = ACTIONS(1052),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1054),
    [aux_sym_cast_type_token1] = ACTIONS(1054),
    [aux_sym_echo_statement_token1] = ACTIONS(1054),
    [aux_sym_exit_statement_token1] = ACTIONS(1054),
    [anon_sym_unset] = ACTIONS(1054),
    [aux_sym_declare_statement_token1] = ACTIONS(1054),
    [aux_sym_declare_statement_token2] = ACTIONS(1054),
    [sym_float] = ACTIONS(1054),
    [aux_sym_try_statement_token1] = ACTIONS(1054),
    [aux_sym_catch_clause_token1] = ACTIONS(1054),
    [aux_sym_finally_clause_token1] = ACTIONS(1054),
    [aux_sym_goto_statement_token1] = ACTIONS(1054),
    [aux_sym_continue_statement_token1] = ACTIONS(1054),
    [aux_sym_break_statement_token1] = ACTIONS(1054),
    [sym_integer] = ACTIONS(1054),
    [aux_sym_return_statement_token1] = ACTIONS(1054),
    [aux_sym_throw_expression_token1] = ACTIONS(1054),
    [aux_sym_while_statement_token1] = ACTIONS(1054),
    [aux_sym_while_statement_token2] = ACTIONS(1054),
    [aux_sym_do_statement_token1] = ACTIONS(1054),
    [aux_sym_for_statement_token1] = ACTIONS(1054),
    [aux_sym_for_statement_token2] = ACTIONS(1054),
    [aux_sym_foreach_statement_token1] = ACTIONS(1054),
    [aux_sym_foreach_statement_token2] = ACTIONS(1054),
    [aux_sym_if_statement_token1] = ACTIONS(1054),
    [aux_sym_if_statement_token2] = ACTIONS(1054),
    [aux_sym_else_if_clause_token1] = ACTIONS(1054),
    [aux_sym_else_clause_token1] = ACTIONS(1054),
    [aux_sym_match_expression_token1] = ACTIONS(1054),
    [aux_sym_match_default_expression_token1] = ACTIONS(1054),
    [aux_sym_switch_statement_token1] = ACTIONS(1054),
    [aux_sym_switch_block_token1] = ACTIONS(1054),
    [anon_sym_PLUS] = ACTIONS(1054),
    [anon_sym_DASH] = ACTIONS(1054),
    [anon_sym_TILDE] = ACTIONS(1052),
    [anon_sym_BANG] = ACTIONS(1052),
    [anon_sym_AT] = ACTIONS(1052),
    [aux_sym_clone_expression_token1] = ACTIONS(1054),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1054),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1054),
    [anon_sym_DASH_DASH] = ACTIONS(1052),
    [anon_sym_PLUS_PLUS] = ACTIONS(1052),
    [aux_sym__list_destructing_token1] = ACTIONS(1054),
    [anon_sym_LBRACK] = ACTIONS(1052),
    [anon_sym_self] = ACTIONS(1054),
    [anon_sym_parent] = ACTIONS(1054),
    [aux_sym__argument_name_token1] = ACTIONS(1054),
    [aux_sym__argument_name_token2] = ACTIONS(1054),
    [anon_sym_POUND_LBRACK] = ACTIONS(1052),
    [aux_sym_encapsed_string_token1] = ACTIONS(1052),
    [anon_sym_DQUOTE] = ACTIONS(1052),
    [aux_sym_string_token1] = ACTIONS(1052),
    [anon_sym_SQUOTE] = ACTIONS(1052),
    [anon_sym_LT_LT_LT] = ACTIONS(1052),
    [anon_sym_BQUOTE] = ACTIONS(1052),
    [anon_sym_DOLLAR] = ACTIONS(1052),
    [aux_sym_yield_expression_token1] = ACTIONS(1054),
    [aux_sym_include_expression_token1] = ACTIONS(1054),
    [aux_sym_include_once_expression_token1] = ACTIONS(1054),
    [aux_sym_require_expression_token1] = ACTIONS(1054),
    [aux_sym_require_once_expression_token1] = ACTIONS(1054),
    [sym_comment] = ACTIONS(5),
  },
  [456] = {
    [sym_text_interpolation] = STATE(456),
    [ts_builtin_sym_end] = ACTIONS(1056),
    [sym_name] = ACTIONS(1058),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1056),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1058),
    [aux_sym_global_declaration_token1] = ACTIONS(1058),
    [aux_sym_namespace_definition_token1] = ACTIONS(1058),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1058),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1058),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1058),
    [anon_sym_BSLASH] = ACTIONS(1056),
    [anon_sym_LBRACE] = ACTIONS(1056),
    [anon_sym_RBRACE] = ACTIONS(1056),
    [aux_sym_trait_declaration_token1] = ACTIONS(1058),
    [aux_sym_interface_declaration_token1] = ACTIONS(1058),
    [aux_sym_enum_declaration_token1] = ACTIONS(1058),
    [aux_sym_enum_case_token1] = ACTIONS(1058),
    [aux_sym_class_declaration_token1] = ACTIONS(1058),
    [aux_sym_final_modifier_token1] = ACTIONS(1058),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1058),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1058),
    [sym_var_modifier] = ACTIONS(1058),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1058),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1058),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1058),
    [anon_sym_LPAREN2] = ACTIONS(1056),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1058),
    [aux_sym_cast_type_token1] = ACTIONS(1058),
    [aux_sym_echo_statement_token1] = ACTIONS(1058),
    [aux_sym_exit_statement_token1] = ACTIONS(1058),
    [anon_sym_unset] = ACTIONS(1058),
    [aux_sym_declare_statement_token1] = ACTIONS(1058),
    [aux_sym_declare_statement_token2] = ACTIONS(1058),
    [sym_float] = ACTIONS(1058),
    [aux_sym_try_statement_token1] = ACTIONS(1058),
    [aux_sym_catch_clause_token1] = ACTIONS(1058),
    [aux_sym_finally_clause_token1] = ACTIONS(1058),
    [aux_sym_goto_statement_token1] = ACTIONS(1058),
    [aux_sym_continue_statement_token1] = ACTIONS(1058),
    [aux_sym_break_statement_token1] = ACTIONS(1058),
    [sym_integer] = ACTIONS(1058),
    [aux_sym_return_statement_token1] = ACTIONS(1058),
    [aux_sym_throw_expression_token1] = ACTIONS(1058),
    [aux_sym_while_statement_token1] = ACTIONS(1058),
    [aux_sym_while_statement_token2] = ACTIONS(1058),
    [aux_sym_do_statement_token1] = ACTIONS(1058),
    [aux_sym_for_statement_token1] = ACTIONS(1058),
    [aux_sym_for_statement_token2] = ACTIONS(1058),
    [aux_sym_foreach_statement_token1] = ACTIONS(1058),
    [aux_sym_foreach_statement_token2] = ACTIONS(1058),
    [aux_sym_if_statement_token1] = ACTIONS(1058),
    [aux_sym_if_statement_token2] = ACTIONS(1058),
    [aux_sym_else_if_clause_token1] = ACTIONS(1058),
    [aux_sym_else_clause_token1] = ACTIONS(1058),
    [aux_sym_match_expression_token1] = ACTIONS(1058),
    [aux_sym_match_default_expression_token1] = ACTIONS(1058),
    [aux_sym_switch_statement_token1] = ACTIONS(1058),
    [aux_sym_switch_block_token1] = ACTIONS(1058),
    [anon_sym_PLUS] = ACTIONS(1058),
    [anon_sym_DASH] = ACTIONS(1058),
    [anon_sym_TILDE] = ACTIONS(1056),
    [anon_sym_BANG] = ACTIONS(1056),
    [anon_sym_AT] = ACTIONS(1056),
    [aux_sym_clone_expression_token1] = ACTIONS(1058),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1058),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1058),
    [anon_sym_DASH_DASH] = ACTIONS(1056),
    [anon_sym_PLUS_PLUS] = ACTIONS(1056),
    [aux_sym__list_destructing_token1] = ACTIONS(1058),
    [anon_sym_LBRACK] = ACTIONS(1056),
    [anon_sym_self] = ACTIONS(1058),
    [anon_sym_parent] = ACTIONS(1058),
    [aux_sym__argument_name_token1] = ACTIONS(1058),
    [aux_sym__argument_name_token2] = ACTIONS(1058),
    [anon_sym_POUND_LBRACK] = ACTIONS(1056),
    [aux_sym_encapsed_string_token1] = ACTIONS(1056),
    [anon_sym_DQUOTE] = ACTIONS(1056),
    [aux_sym_string_token1] = ACTIONS(1056),
    [anon_sym_SQUOTE] = ACTIONS(1056),
    [anon_sym_LT_LT_LT] = ACTIONS(1056),
    [anon_sym_BQUOTE] = ACTIONS(1056),
    [anon_sym_DOLLAR] = ACTIONS(1056),
    [aux_sym_yield_expression_token1] = ACTIONS(1058),
    [aux_sym_include_expression_token1] = ACTIONS(1058),
    [aux_sym_include_once_expression_token1] = ACTIONS(1058),
    [aux_sym_require_expression_token1] = ACTIONS(1058),
    [aux_sym_require_once_expression_token1] = ACTIONS(1058),
    [sym_comment] = ACTIONS(5),
  },
  [457] = {
    [sym_text_interpolation] = STATE(457),
    [ts_builtin_sym_end] = ACTIONS(1060),
    [sym_name] = ACTIONS(1062),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1060),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1062),
    [aux_sym_global_declaration_token1] = ACTIONS(1062),
    [aux_sym_namespace_definition_token1] = ACTIONS(1062),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1062),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1062),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1062),
    [anon_sym_BSLASH] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1060),
    [anon_sym_RBRACE] = ACTIONS(1060),
    [aux_sym_trait_declaration_token1] = ACTIONS(1062),
    [aux_sym_interface_declaration_token1] = ACTIONS(1062),
    [aux_sym_enum_declaration_token1] = ACTIONS(1062),
    [aux_sym_enum_case_token1] = ACTIONS(1062),
    [aux_sym_class_declaration_token1] = ACTIONS(1062),
    [aux_sym_final_modifier_token1] = ACTIONS(1062),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1062),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1062),
    [sym_var_modifier] = ACTIONS(1062),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1062),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1062),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1062),
    [anon_sym_LPAREN2] = ACTIONS(1060),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1062),
    [aux_sym_cast_type_token1] = ACTIONS(1062),
    [aux_sym_echo_statement_token1] = ACTIONS(1062),
    [aux_sym_exit_statement_token1] = ACTIONS(1062),
    [anon_sym_unset] = ACTIONS(1062),
    [aux_sym_declare_statement_token1] = ACTIONS(1062),
    [aux_sym_declare_statement_token2] = ACTIONS(1062),
    [sym_float] = ACTIONS(1062),
    [aux_sym_try_statement_token1] = ACTIONS(1062),
    [aux_sym_goto_statement_token1] = ACTIONS(1062),
    [aux_sym_continue_statement_token1] = ACTIONS(1062),
    [aux_sym_break_statement_token1] = ACTIONS(1062),
    [sym_integer] = ACTIONS(1062),
    [aux_sym_return_statement_token1] = ACTIONS(1062),
    [aux_sym_throw_expression_token1] = ACTIONS(1062),
    [aux_sym_while_statement_token1] = ACTIONS(1062),
    [aux_sym_while_statement_token2] = ACTIONS(1062),
    [aux_sym_do_statement_token1] = ACTIONS(1062),
    [aux_sym_for_statement_token1] = ACTIONS(1062),
    [aux_sym_for_statement_token2] = ACTIONS(1062),
    [aux_sym_foreach_statement_token1] = ACTIONS(1062),
    [aux_sym_foreach_statement_token2] = ACTIONS(1062),
    [aux_sym_if_statement_token1] = ACTIONS(1062),
    [aux_sym_if_statement_token2] = ACTIONS(1062),
    [aux_sym_else_if_clause_token1] = ACTIONS(1062),
    [aux_sym_else_clause_token1] = ACTIONS(1062),
    [aux_sym_match_expression_token1] = ACTIONS(1062),
    [aux_sym_match_default_expression_token1] = ACTIONS(1062),
    [aux_sym_switch_statement_token1] = ACTIONS(1062),
    [aux_sym_switch_block_token1] = ACTIONS(1062),
    [anon_sym_PLUS] = ACTIONS(1062),
    [anon_sym_DASH] = ACTIONS(1062),
    [anon_sym_TILDE] = ACTIONS(1060),
    [anon_sym_BANG] = ACTIONS(1060),
    [anon_sym_AT] = ACTIONS(1060),
    [aux_sym_clone_expression_token1] = ACTIONS(1062),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1062),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1062),
    [anon_sym_DASH_DASH] = ACTIONS(1060),
    [anon_sym_PLUS_PLUS] = ACTIONS(1060),
    [aux_sym__list_destructing_token1] = ACTIONS(1062),
    [anon_sym_LBRACK] = ACTIONS(1060),
    [anon_sym_self] = ACTIONS(1062),
    [anon_sym_parent] = ACTIONS(1062),
    [aux_sym__argument_name_token1] = ACTIONS(1062),
    [aux_sym__argument_name_token2] = ACTIONS(1062),
    [anon_sym_POUND_LBRACK] = ACTIONS(1060),
    [aux_sym_encapsed_string_token1] = ACTIONS(1060),
    [anon_sym_DQUOTE] = ACTIONS(1060),
    [aux_sym_string_token1] = ACTIONS(1060),
    [anon_sym_SQUOTE] = ACTIONS(1060),
    [anon_sym_LT_LT_LT] = ACTIONS(1060),
    [anon_sym_BQUOTE] = ACTIONS(1060),
    [anon_sym_DOLLAR] = ACTIONS(1060),
    [aux_sym_yield_expression_token1] = ACTIONS(1062),
    [aux_sym_include_expression_token1] = ACTIONS(1062),
    [aux_sym_include_once_expression_token1] = ACTIONS(1062),
    [aux_sym_require_expression_token1] = ACTIONS(1062),
    [aux_sym_require_once_expression_token1] = ACTIONS(1062),
    [sym_comment] = ACTIONS(5),
  },
  [458] = {
    [sym_text_interpolation] = STATE(458),
    [ts_builtin_sym_end] = ACTIONS(1064),
    [sym_name] = ACTIONS(1066),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1064),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1066),
    [aux_sym_global_declaration_token1] = ACTIONS(1066),
    [aux_sym_namespace_definition_token1] = ACTIONS(1066),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1066),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1066),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1066),
    [anon_sym_BSLASH] = ACTIONS(1064),
    [anon_sym_LBRACE] = ACTIONS(1064),
    [anon_sym_RBRACE] = ACTIONS(1064),
    [aux_sym_trait_declaration_token1] = ACTIONS(1066),
    [aux_sym_interface_declaration_token1] = ACTIONS(1066),
    [aux_sym_enum_declaration_token1] = ACTIONS(1066),
    [aux_sym_enum_case_token1] = ACTIONS(1066),
    [aux_sym_class_declaration_token1] = ACTIONS(1066),
    [aux_sym_final_modifier_token1] = ACTIONS(1066),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1066),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1066),
    [sym_var_modifier] = ACTIONS(1066),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1066),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1066),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1066),
    [anon_sym_LPAREN2] = ACTIONS(1064),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1066),
    [aux_sym_cast_type_token1] = ACTIONS(1066),
    [aux_sym_echo_statement_token1] = ACTIONS(1066),
    [aux_sym_exit_statement_token1] = ACTIONS(1066),
    [anon_sym_unset] = ACTIONS(1066),
    [aux_sym_declare_statement_token1] = ACTIONS(1066),
    [aux_sym_declare_statement_token2] = ACTIONS(1066),
    [sym_float] = ACTIONS(1066),
    [aux_sym_try_statement_token1] = ACTIONS(1066),
    [aux_sym_goto_statement_token1] = ACTIONS(1066),
    [aux_sym_continue_statement_token1] = ACTIONS(1066),
    [aux_sym_break_statement_token1] = ACTIONS(1066),
    [sym_integer] = ACTIONS(1066),
    [aux_sym_return_statement_token1] = ACTIONS(1066),
    [aux_sym_throw_expression_token1] = ACTIONS(1066),
    [aux_sym_while_statement_token1] = ACTIONS(1066),
    [aux_sym_while_statement_token2] = ACTIONS(1066),
    [aux_sym_do_statement_token1] = ACTIONS(1066),
    [aux_sym_for_statement_token1] = ACTIONS(1066),
    [aux_sym_for_statement_token2] = ACTIONS(1066),
    [aux_sym_foreach_statement_token1] = ACTIONS(1066),
    [aux_sym_foreach_statement_token2] = ACTIONS(1066),
    [aux_sym_if_statement_token1] = ACTIONS(1066),
    [aux_sym_if_statement_token2] = ACTIONS(1066),
    [aux_sym_else_if_clause_token1] = ACTIONS(1066),
    [aux_sym_else_clause_token1] = ACTIONS(1066),
    [aux_sym_match_expression_token1] = ACTIONS(1066),
    [aux_sym_match_default_expression_token1] = ACTIONS(1066),
    [aux_sym_switch_statement_token1] = ACTIONS(1066),
    [aux_sym_switch_block_token1] = ACTIONS(1066),
    [anon_sym_PLUS] = ACTIONS(1066),
    [anon_sym_DASH] = ACTIONS(1066),
    [anon_sym_TILDE] = ACTIONS(1064),
    [anon_sym_BANG] = ACTIONS(1064),
    [anon_sym_AT] = ACTIONS(1064),
    [aux_sym_clone_expression_token1] = ACTIONS(1066),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1066),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1066),
    [anon_sym_DASH_DASH] = ACTIONS(1064),
    [anon_sym_PLUS_PLUS] = ACTIONS(1064),
    [aux_sym__list_destructing_token1] = ACTIONS(1066),
    [anon_sym_LBRACK] = ACTIONS(1064),
    [anon_sym_self] = ACTIONS(1066),
    [anon_sym_parent] = ACTIONS(1066),
    [aux_sym__argument_name_token1] = ACTIONS(1066),
    [aux_sym__argument_name_token2] = ACTIONS(1066),
    [anon_sym_POUND_LBRACK] = ACTIONS(1064),
    [aux_sym_encapsed_string_token1] = ACTIONS(1064),
    [anon_sym_DQUOTE] = ACTIONS(1064),
    [aux_sym_string_token1] = ACTIONS(1064),
    [anon_sym_SQUOTE] = ACTIONS(1064),
    [anon_sym_LT_LT_LT] = ACTIONS(1064),
    [anon_sym_BQUOTE] = ACTIONS(1064),
    [anon_sym_DOLLAR] = ACTIONS(1064),
    [aux_sym_yield_expression_token1] = ACTIONS(1066),
    [aux_sym_include_expression_token1] = ACTIONS(1066),
    [aux_sym_include_once_expression_token1] = ACTIONS(1066),
    [aux_sym_require_expression_token1] = ACTIONS(1066),
    [aux_sym_require_once_expression_token1] = ACTIONS(1066),
    [sym_comment] = ACTIONS(5),
  },
  [459] = {
    [sym_text_interpolation] = STATE(459),
    [ts_builtin_sym_end] = ACTIONS(1068),
    [sym_name] = ACTIONS(1070),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1068),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1070),
    [aux_sym_global_declaration_token1] = ACTIONS(1070),
    [aux_sym_namespace_definition_token1] = ACTIONS(1070),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1070),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1070),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1070),
    [anon_sym_BSLASH] = ACTIONS(1068),
    [anon_sym_LBRACE] = ACTIONS(1068),
    [anon_sym_RBRACE] = ACTIONS(1068),
    [aux_sym_trait_declaration_token1] = ACTIONS(1070),
    [aux_sym_interface_declaration_token1] = ACTIONS(1070),
    [aux_sym_enum_declaration_token1] = ACTIONS(1070),
    [aux_sym_enum_case_token1] = ACTIONS(1070),
    [aux_sym_class_declaration_token1] = ACTIONS(1070),
    [aux_sym_final_modifier_token1] = ACTIONS(1070),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1070),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1070),
    [sym_var_modifier] = ACTIONS(1070),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1070),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1070),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1070),
    [anon_sym_LPAREN2] = ACTIONS(1068),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1070),
    [aux_sym_cast_type_token1] = ACTIONS(1070),
    [aux_sym_echo_statement_token1] = ACTIONS(1070),
    [aux_sym_exit_statement_token1] = ACTIONS(1070),
    [anon_sym_unset] = ACTIONS(1070),
    [aux_sym_declare_statement_token1] = ACTIONS(1070),
    [aux_sym_declare_statement_token2] = ACTIONS(1070),
    [sym_float] = ACTIONS(1070),
    [aux_sym_try_statement_token1] = ACTIONS(1070),
    [aux_sym_goto_statement_token1] = ACTIONS(1070),
    [aux_sym_continue_statement_token1] = ACTIONS(1070),
    [aux_sym_break_statement_token1] = ACTIONS(1070),
    [sym_integer] = ACTIONS(1070),
    [aux_sym_return_statement_token1] = ACTIONS(1070),
    [aux_sym_throw_expression_token1] = ACTIONS(1070),
    [aux_sym_while_statement_token1] = ACTIONS(1070),
    [aux_sym_while_statement_token2] = ACTIONS(1070),
    [aux_sym_do_statement_token1] = ACTIONS(1070),
    [aux_sym_for_statement_token1] = ACTIONS(1070),
    [aux_sym_for_statement_token2] = ACTIONS(1070),
    [aux_sym_foreach_statement_token1] = ACTIONS(1070),
    [aux_sym_foreach_statement_token2] = ACTIONS(1070),
    [aux_sym_if_statement_token1] = ACTIONS(1070),
    [aux_sym_if_statement_token2] = ACTIONS(1070),
    [aux_sym_else_if_clause_token1] = ACTIONS(1070),
    [aux_sym_else_clause_token1] = ACTIONS(1070),
    [aux_sym_match_expression_token1] = ACTIONS(1070),
    [aux_sym_match_default_expression_token1] = ACTIONS(1070),
    [aux_sym_switch_statement_token1] = ACTIONS(1070),
    [aux_sym_switch_block_token1] = ACTIONS(1070),
    [anon_sym_PLUS] = ACTIONS(1070),
    [anon_sym_DASH] = ACTIONS(1070),
    [anon_sym_TILDE] = ACTIONS(1068),
    [anon_sym_BANG] = ACTIONS(1068),
    [anon_sym_AT] = ACTIONS(1068),
    [aux_sym_clone_expression_token1] = ACTIONS(1070),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1070),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1070),
    [anon_sym_DASH_DASH] = ACTIONS(1068),
    [anon_sym_PLUS_PLUS] = ACTIONS(1068),
    [aux_sym__list_destructing_token1] = ACTIONS(1070),
    [anon_sym_LBRACK] = ACTIONS(1068),
    [anon_sym_self] = ACTIONS(1070),
    [anon_sym_parent] = ACTIONS(1070),
    [aux_sym__argument_name_token1] = ACTIONS(1070),
    [aux_sym__argument_name_token2] = ACTIONS(1070),
    [anon_sym_POUND_LBRACK] = ACTIONS(1068),
    [aux_sym_encapsed_string_token1] = ACTIONS(1068),
    [anon_sym_DQUOTE] = ACTIONS(1068),
    [aux_sym_string_token1] = ACTIONS(1068),
    [anon_sym_SQUOTE] = ACTIONS(1068),
    [anon_sym_LT_LT_LT] = ACTIONS(1068),
    [anon_sym_BQUOTE] = ACTIONS(1068),
    [anon_sym_DOLLAR] = ACTIONS(1068),
    [aux_sym_yield_expression_token1] = ACTIONS(1070),
    [aux_sym_include_expression_token1] = ACTIONS(1070),
    [aux_sym_include_once_expression_token1] = ACTIONS(1070),
    [aux_sym_require_expression_token1] = ACTIONS(1070),
    [aux_sym_require_once_expression_token1] = ACTIONS(1070),
    [sym_comment] = ACTIONS(5),
  },
  [460] = {
    [sym_text_interpolation] = STATE(460),
    [ts_builtin_sym_end] = ACTIONS(1072),
    [sym_name] = ACTIONS(1074),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1072),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1074),
    [aux_sym_global_declaration_token1] = ACTIONS(1074),
    [aux_sym_namespace_definition_token1] = ACTIONS(1074),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1074),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1074),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1074),
    [anon_sym_BSLASH] = ACTIONS(1072),
    [anon_sym_LBRACE] = ACTIONS(1072),
    [anon_sym_RBRACE] = ACTIONS(1072),
    [aux_sym_trait_declaration_token1] = ACTIONS(1074),
    [aux_sym_interface_declaration_token1] = ACTIONS(1074),
    [aux_sym_enum_declaration_token1] = ACTIONS(1074),
    [aux_sym_enum_case_token1] = ACTIONS(1074),
    [aux_sym_class_declaration_token1] = ACTIONS(1074),
    [aux_sym_final_modifier_token1] = ACTIONS(1074),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1074),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1074),
    [sym_var_modifier] = ACTIONS(1074),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1074),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1074),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1074),
    [anon_sym_LPAREN2] = ACTIONS(1072),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1074),
    [aux_sym_cast_type_token1] = ACTIONS(1074),
    [aux_sym_echo_statement_token1] = ACTIONS(1074),
    [aux_sym_exit_statement_token1] = ACTIONS(1074),
    [anon_sym_unset] = ACTIONS(1074),
    [aux_sym_declare_statement_token1] = ACTIONS(1074),
    [aux_sym_declare_statement_token2] = ACTIONS(1074),
    [sym_float] = ACTIONS(1074),
    [aux_sym_try_statement_token1] = ACTIONS(1074),
    [aux_sym_goto_statement_token1] = ACTIONS(1074),
    [aux_sym_continue_statement_token1] = ACTIONS(1074),
    [aux_sym_break_statement_token1] = ACTIONS(1074),
    [sym_integer] = ACTIONS(1074),
    [aux_sym_return_statement_token1] = ACTIONS(1074),
    [aux_sym_throw_expression_token1] = ACTIONS(1074),
    [aux_sym_while_statement_token1] = ACTIONS(1074),
    [aux_sym_while_statement_token2] = ACTIONS(1074),
    [aux_sym_do_statement_token1] = ACTIONS(1074),
    [aux_sym_for_statement_token1] = ACTIONS(1074),
    [aux_sym_for_statement_token2] = ACTIONS(1074),
    [aux_sym_foreach_statement_token1] = ACTIONS(1074),
    [aux_sym_foreach_statement_token2] = ACTIONS(1074),
    [aux_sym_if_statement_token1] = ACTIONS(1074),
    [aux_sym_if_statement_token2] = ACTIONS(1074),
    [aux_sym_else_if_clause_token1] = ACTIONS(1074),
    [aux_sym_else_clause_token1] = ACTIONS(1074),
    [aux_sym_match_expression_token1] = ACTIONS(1074),
    [aux_sym_match_default_expression_token1] = ACTIONS(1074),
    [aux_sym_switch_statement_token1] = ACTIONS(1074),
    [aux_sym_switch_block_token1] = ACTIONS(1074),
    [anon_sym_PLUS] = ACTIONS(1074),
    [anon_sym_DASH] = ACTIONS(1074),
    [anon_sym_TILDE] = ACTIONS(1072),
    [anon_sym_BANG] = ACTIONS(1072),
    [anon_sym_AT] = ACTIONS(1072),
    [aux_sym_clone_expression_token1] = ACTIONS(1074),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1074),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1074),
    [anon_sym_DASH_DASH] = ACTIONS(1072),
    [anon_sym_PLUS_PLUS] = ACTIONS(1072),
    [aux_sym__list_destructing_token1] = ACTIONS(1074),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_self] = ACTIONS(1074),
    [anon_sym_parent] = ACTIONS(1074),
    [aux_sym__argument_name_token1] = ACTIONS(1074),
    [aux_sym__argument_name_token2] = ACTIONS(1074),
    [anon_sym_POUND_LBRACK] = ACTIONS(1072),
    [aux_sym_encapsed_string_token1] = ACTIONS(1072),
    [anon_sym_DQUOTE] = ACTIONS(1072),
    [aux_sym_string_token1] = ACTIONS(1072),
    [anon_sym_SQUOTE] = ACTIONS(1072),
    [anon_sym_LT_LT_LT] = ACTIONS(1072),
    [anon_sym_BQUOTE] = ACTIONS(1072),
    [anon_sym_DOLLAR] = ACTIONS(1072),
    [aux_sym_yield_expression_token1] = ACTIONS(1074),
    [aux_sym_include_expression_token1] = ACTIONS(1074),
    [aux_sym_include_once_expression_token1] = ACTIONS(1074),
    [aux_sym_require_expression_token1] = ACTIONS(1074),
    [aux_sym_require_once_expression_token1] = ACTIONS(1074),
    [sym_comment] = ACTIONS(5),
  },
  [461] = {
    [sym_text_interpolation] = STATE(461),
    [ts_builtin_sym_end] = ACTIONS(1076),
    [sym_name] = ACTIONS(1078),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1076),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1078),
    [aux_sym_global_declaration_token1] = ACTIONS(1078),
    [aux_sym_namespace_definition_token1] = ACTIONS(1078),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1078),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1078),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1078),
    [anon_sym_BSLASH] = ACTIONS(1076),
    [anon_sym_LBRACE] = ACTIONS(1076),
    [anon_sym_RBRACE] = ACTIONS(1076),
    [aux_sym_trait_declaration_token1] = ACTIONS(1078),
    [aux_sym_interface_declaration_token1] = ACTIONS(1078),
    [aux_sym_enum_declaration_token1] = ACTIONS(1078),
    [aux_sym_enum_case_token1] = ACTIONS(1078),
    [aux_sym_class_declaration_token1] = ACTIONS(1078),
    [aux_sym_final_modifier_token1] = ACTIONS(1078),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1078),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1078),
    [sym_var_modifier] = ACTIONS(1078),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1078),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1078),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1078),
    [anon_sym_LPAREN2] = ACTIONS(1076),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1078),
    [aux_sym_cast_type_token1] = ACTIONS(1078),
    [aux_sym_echo_statement_token1] = ACTIONS(1078),
    [aux_sym_exit_statement_token1] = ACTIONS(1078),
    [anon_sym_unset] = ACTIONS(1078),
    [aux_sym_declare_statement_token1] = ACTIONS(1078),
    [aux_sym_declare_statement_token2] = ACTIONS(1078),
    [sym_float] = ACTIONS(1078),
    [aux_sym_try_statement_token1] = ACTIONS(1078),
    [aux_sym_goto_statement_token1] = ACTIONS(1078),
    [aux_sym_continue_statement_token1] = ACTIONS(1078),
    [aux_sym_break_statement_token1] = ACTIONS(1078),
    [sym_integer] = ACTIONS(1078),
    [aux_sym_return_statement_token1] = ACTIONS(1078),
    [aux_sym_throw_expression_token1] = ACTIONS(1078),
    [aux_sym_while_statement_token1] = ACTIONS(1078),
    [aux_sym_while_statement_token2] = ACTIONS(1078),
    [aux_sym_do_statement_token1] = ACTIONS(1078),
    [aux_sym_for_statement_token1] = ACTIONS(1078),
    [aux_sym_for_statement_token2] = ACTIONS(1078),
    [aux_sym_foreach_statement_token1] = ACTIONS(1078),
    [aux_sym_foreach_statement_token2] = ACTIONS(1078),
    [aux_sym_if_statement_token1] = ACTIONS(1078),
    [aux_sym_if_statement_token2] = ACTIONS(1078),
    [aux_sym_else_if_clause_token1] = ACTIONS(1078),
    [aux_sym_else_clause_token1] = ACTIONS(1078),
    [aux_sym_match_expression_token1] = ACTIONS(1078),
    [aux_sym_match_default_expression_token1] = ACTIONS(1078),
    [aux_sym_switch_statement_token1] = ACTIONS(1078),
    [aux_sym_switch_block_token1] = ACTIONS(1078),
    [anon_sym_PLUS] = ACTIONS(1078),
    [anon_sym_DASH] = ACTIONS(1078),
    [anon_sym_TILDE] = ACTIONS(1076),
    [anon_sym_BANG] = ACTIONS(1076),
    [anon_sym_AT] = ACTIONS(1076),
    [aux_sym_clone_expression_token1] = ACTIONS(1078),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1078),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1078),
    [anon_sym_DASH_DASH] = ACTIONS(1076),
    [anon_sym_PLUS_PLUS] = ACTIONS(1076),
    [aux_sym__list_destructing_token1] = ACTIONS(1078),
    [anon_sym_LBRACK] = ACTIONS(1076),
    [anon_sym_self] = ACTIONS(1078),
    [anon_sym_parent] = ACTIONS(1078),
    [aux_sym__argument_name_token1] = ACTIONS(1078),
    [aux_sym__argument_name_token2] = ACTIONS(1078),
    [anon_sym_POUND_LBRACK] = ACTIONS(1076),
    [aux_sym_encapsed_string_token1] = ACTIONS(1076),
    [anon_sym_DQUOTE] = ACTIONS(1076),
    [aux_sym_string_token1] = ACTIONS(1076),
    [anon_sym_SQUOTE] = ACTIONS(1076),
    [anon_sym_LT_LT_LT] = ACTIONS(1076),
    [anon_sym_BQUOTE] = ACTIONS(1076),
    [anon_sym_DOLLAR] = ACTIONS(1076),
    [aux_sym_yield_expression_token1] = ACTIONS(1078),
    [aux_sym_include_expression_token1] = ACTIONS(1078),
    [aux_sym_include_once_expression_token1] = ACTIONS(1078),
    [aux_sym_require_expression_token1] = ACTIONS(1078),
    [aux_sym_require_once_expression_token1] = ACTIONS(1078),
    [sym_comment] = ACTIONS(5),
  },
  [462] = {
    [sym_text_interpolation] = STATE(462),
    [ts_builtin_sym_end] = ACTIONS(1080),
    [sym_name] = ACTIONS(1082),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1080),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1082),
    [aux_sym_global_declaration_token1] = ACTIONS(1082),
    [aux_sym_namespace_definition_token1] = ACTIONS(1082),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1082),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1082),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1082),
    [anon_sym_BSLASH] = ACTIONS(1080),
    [anon_sym_LBRACE] = ACTIONS(1080),
    [anon_sym_RBRACE] = ACTIONS(1080),
    [aux_sym_trait_declaration_token1] = ACTIONS(1082),
    [aux_sym_interface_declaration_token1] = ACTIONS(1082),
    [aux_sym_enum_declaration_token1] = ACTIONS(1082),
    [aux_sym_enum_case_token1] = ACTIONS(1082),
    [aux_sym_class_declaration_token1] = ACTIONS(1082),
    [aux_sym_final_modifier_token1] = ACTIONS(1082),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1082),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1082),
    [sym_var_modifier] = ACTIONS(1082),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1082),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1082),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1082),
    [anon_sym_LPAREN2] = ACTIONS(1080),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1082),
    [aux_sym_cast_type_token1] = ACTIONS(1082),
    [aux_sym_echo_statement_token1] = ACTIONS(1082),
    [aux_sym_exit_statement_token1] = ACTIONS(1082),
    [anon_sym_unset] = ACTIONS(1082),
    [aux_sym_declare_statement_token1] = ACTIONS(1082),
    [aux_sym_declare_statement_token2] = ACTIONS(1082),
    [sym_float] = ACTIONS(1082),
    [aux_sym_try_statement_token1] = ACTIONS(1082),
    [aux_sym_goto_statement_token1] = ACTIONS(1082),
    [aux_sym_continue_statement_token1] = ACTIONS(1082),
    [aux_sym_break_statement_token1] = ACTIONS(1082),
    [sym_integer] = ACTIONS(1082),
    [aux_sym_return_statement_token1] = ACTIONS(1082),
    [aux_sym_throw_expression_token1] = ACTIONS(1082),
    [aux_sym_while_statement_token1] = ACTIONS(1082),
    [aux_sym_while_statement_token2] = ACTIONS(1082),
    [aux_sym_do_statement_token1] = ACTIONS(1082),
    [aux_sym_for_statement_token1] = ACTIONS(1082),
    [aux_sym_for_statement_token2] = ACTIONS(1082),
    [aux_sym_foreach_statement_token1] = ACTIONS(1082),
    [aux_sym_foreach_statement_token2] = ACTIONS(1082),
    [aux_sym_if_statement_token1] = ACTIONS(1082),
    [aux_sym_if_statement_token2] = ACTIONS(1082),
    [aux_sym_else_if_clause_token1] = ACTIONS(1082),
    [aux_sym_else_clause_token1] = ACTIONS(1082),
    [aux_sym_match_expression_token1] = ACTIONS(1082),
    [aux_sym_match_default_expression_token1] = ACTIONS(1082),
    [aux_sym_switch_statement_token1] = ACTIONS(1082),
    [aux_sym_switch_block_token1] = ACTIONS(1082),
    [anon_sym_PLUS] = ACTIONS(1082),
    [anon_sym_DASH] = ACTIONS(1082),
    [anon_sym_TILDE] = ACTIONS(1080),
    [anon_sym_BANG] = ACTIONS(1080),
    [anon_sym_AT] = ACTIONS(1080),
    [aux_sym_clone_expression_token1] = ACTIONS(1082),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1082),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1082),
    [anon_sym_DASH_DASH] = ACTIONS(1080),
    [anon_sym_PLUS_PLUS] = ACTIONS(1080),
    [aux_sym__list_destructing_token1] = ACTIONS(1082),
    [anon_sym_LBRACK] = ACTIONS(1080),
    [anon_sym_self] = ACTIONS(1082),
    [anon_sym_parent] = ACTIONS(1082),
    [aux_sym__argument_name_token1] = ACTIONS(1082),
    [aux_sym__argument_name_token2] = ACTIONS(1082),
    [anon_sym_POUND_LBRACK] = ACTIONS(1080),
    [aux_sym_encapsed_string_token1] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1080),
    [aux_sym_string_token1] = ACTIONS(1080),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_LT_LT_LT] = ACTIONS(1080),
    [anon_sym_BQUOTE] = ACTIONS(1080),
    [anon_sym_DOLLAR] = ACTIONS(1080),
    [aux_sym_yield_expression_token1] = ACTIONS(1082),
    [aux_sym_include_expression_token1] = ACTIONS(1082),
    [aux_sym_include_once_expression_token1] = ACTIONS(1082),
    [aux_sym_require_expression_token1] = ACTIONS(1082),
    [aux_sym_require_once_expression_token1] = ACTIONS(1082),
    [sym_comment] = ACTIONS(5),
  },
  [463] = {
    [sym_text_interpolation] = STATE(463),
    [ts_builtin_sym_end] = ACTIONS(1084),
    [sym_name] = ACTIONS(1086),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1084),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1086),
    [aux_sym_global_declaration_token1] = ACTIONS(1086),
    [aux_sym_namespace_definition_token1] = ACTIONS(1086),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1086),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1086),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1086),
    [anon_sym_BSLASH] = ACTIONS(1084),
    [anon_sym_LBRACE] = ACTIONS(1084),
    [anon_sym_RBRACE] = ACTIONS(1084),
    [aux_sym_trait_declaration_token1] = ACTIONS(1086),
    [aux_sym_interface_declaration_token1] = ACTIONS(1086),
    [aux_sym_enum_declaration_token1] = ACTIONS(1086),
    [aux_sym_enum_case_token1] = ACTIONS(1086),
    [aux_sym_class_declaration_token1] = ACTIONS(1086),
    [aux_sym_final_modifier_token1] = ACTIONS(1086),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1086),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1086),
    [sym_var_modifier] = ACTIONS(1086),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1086),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1086),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1086),
    [anon_sym_LPAREN2] = ACTIONS(1084),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1086),
    [aux_sym_cast_type_token1] = ACTIONS(1086),
    [aux_sym_echo_statement_token1] = ACTIONS(1086),
    [aux_sym_exit_statement_token1] = ACTIONS(1086),
    [anon_sym_unset] = ACTIONS(1086),
    [aux_sym_declare_statement_token1] = ACTIONS(1086),
    [aux_sym_declare_statement_token2] = ACTIONS(1086),
    [sym_float] = ACTIONS(1086),
    [aux_sym_try_statement_token1] = ACTIONS(1086),
    [aux_sym_goto_statement_token1] = ACTIONS(1086),
    [aux_sym_continue_statement_token1] = ACTIONS(1086),
    [aux_sym_break_statement_token1] = ACTIONS(1086),
    [sym_integer] = ACTIONS(1086),
    [aux_sym_return_statement_token1] = ACTIONS(1086),
    [aux_sym_throw_expression_token1] = ACTIONS(1086),
    [aux_sym_while_statement_token1] = ACTIONS(1086),
    [aux_sym_while_statement_token2] = ACTIONS(1086),
    [aux_sym_do_statement_token1] = ACTIONS(1086),
    [aux_sym_for_statement_token1] = ACTIONS(1086),
    [aux_sym_for_statement_token2] = ACTIONS(1086),
    [aux_sym_foreach_statement_token1] = ACTIONS(1086),
    [aux_sym_foreach_statement_token2] = ACTIONS(1086),
    [aux_sym_if_statement_token1] = ACTIONS(1086),
    [aux_sym_if_statement_token2] = ACTIONS(1086),
    [aux_sym_else_if_clause_token1] = ACTIONS(1086),
    [aux_sym_else_clause_token1] = ACTIONS(1086),
    [aux_sym_match_expression_token1] = ACTIONS(1086),
    [aux_sym_match_default_expression_token1] = ACTIONS(1086),
    [aux_sym_switch_statement_token1] = ACTIONS(1086),
    [aux_sym_switch_block_token1] = ACTIONS(1086),
    [anon_sym_PLUS] = ACTIONS(1086),
    [anon_sym_DASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1084),
    [anon_sym_BANG] = ACTIONS(1084),
    [anon_sym_AT] = ACTIONS(1084),
    [aux_sym_clone_expression_token1] = ACTIONS(1086),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1086),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1086),
    [anon_sym_DASH_DASH] = ACTIONS(1084),
    [anon_sym_PLUS_PLUS] = ACTIONS(1084),
    [aux_sym__list_destructing_token1] = ACTIONS(1086),
    [anon_sym_LBRACK] = ACTIONS(1084),
    [anon_sym_self] = ACTIONS(1086),
    [anon_sym_parent] = ACTIONS(1086),
    [aux_sym__argument_name_token1] = ACTIONS(1086),
    [aux_sym__argument_name_token2] = ACTIONS(1086),
    [anon_sym_POUND_LBRACK] = ACTIONS(1084),
    [aux_sym_encapsed_string_token1] = ACTIONS(1084),
    [anon_sym_DQUOTE] = ACTIONS(1084),
    [aux_sym_string_token1] = ACTIONS(1084),
    [anon_sym_SQUOTE] = ACTIONS(1084),
    [anon_sym_LT_LT_LT] = ACTIONS(1084),
    [anon_sym_BQUOTE] = ACTIONS(1084),
    [anon_sym_DOLLAR] = ACTIONS(1084),
    [aux_sym_yield_expression_token1] = ACTIONS(1086),
    [aux_sym_include_expression_token1] = ACTIONS(1086),
    [aux_sym_include_once_expression_token1] = ACTIONS(1086),
    [aux_sym_require_expression_token1] = ACTIONS(1086),
    [aux_sym_require_once_expression_token1] = ACTIONS(1086),
    [sym_comment] = ACTIONS(5),
  },
  [464] = {
    [sym_text_interpolation] = STATE(464),
    [ts_builtin_sym_end] = ACTIONS(1088),
    [sym_name] = ACTIONS(1090),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1088),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1090),
    [aux_sym_global_declaration_token1] = ACTIONS(1090),
    [aux_sym_namespace_definition_token1] = ACTIONS(1090),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1090),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1090),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1090),
    [anon_sym_BSLASH] = ACTIONS(1088),
    [anon_sym_LBRACE] = ACTIONS(1088),
    [anon_sym_RBRACE] = ACTIONS(1088),
    [aux_sym_trait_declaration_token1] = ACTIONS(1090),
    [aux_sym_interface_declaration_token1] = ACTIONS(1090),
    [aux_sym_enum_declaration_token1] = ACTIONS(1090),
    [aux_sym_enum_case_token1] = ACTIONS(1090),
    [aux_sym_class_declaration_token1] = ACTIONS(1090),
    [aux_sym_final_modifier_token1] = ACTIONS(1090),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1090),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1090),
    [sym_var_modifier] = ACTIONS(1090),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1090),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1090),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1090),
    [anon_sym_LPAREN2] = ACTIONS(1088),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1090),
    [aux_sym_cast_type_token1] = ACTIONS(1090),
    [aux_sym_echo_statement_token1] = ACTIONS(1090),
    [aux_sym_exit_statement_token1] = ACTIONS(1090),
    [anon_sym_unset] = ACTIONS(1090),
    [aux_sym_declare_statement_token1] = ACTIONS(1090),
    [aux_sym_declare_statement_token2] = ACTIONS(1090),
    [sym_float] = ACTIONS(1090),
    [aux_sym_try_statement_token1] = ACTIONS(1090),
    [aux_sym_goto_statement_token1] = ACTIONS(1090),
    [aux_sym_continue_statement_token1] = ACTIONS(1090),
    [aux_sym_break_statement_token1] = ACTIONS(1090),
    [sym_integer] = ACTIONS(1090),
    [aux_sym_return_statement_token1] = ACTIONS(1090),
    [aux_sym_throw_expression_token1] = ACTIONS(1090),
    [aux_sym_while_statement_token1] = ACTIONS(1090),
    [aux_sym_while_statement_token2] = ACTIONS(1090),
    [aux_sym_do_statement_token1] = ACTIONS(1090),
    [aux_sym_for_statement_token1] = ACTIONS(1090),
    [aux_sym_for_statement_token2] = ACTIONS(1090),
    [aux_sym_foreach_statement_token1] = ACTIONS(1090),
    [aux_sym_foreach_statement_token2] = ACTIONS(1090),
    [aux_sym_if_statement_token1] = ACTIONS(1090),
    [aux_sym_if_statement_token2] = ACTIONS(1090),
    [aux_sym_else_if_clause_token1] = ACTIONS(1090),
    [aux_sym_else_clause_token1] = ACTIONS(1090),
    [aux_sym_match_expression_token1] = ACTIONS(1090),
    [aux_sym_match_default_expression_token1] = ACTIONS(1090),
    [aux_sym_switch_statement_token1] = ACTIONS(1090),
    [aux_sym_switch_block_token1] = ACTIONS(1090),
    [anon_sym_PLUS] = ACTIONS(1090),
    [anon_sym_DASH] = ACTIONS(1090),
    [anon_sym_TILDE] = ACTIONS(1088),
    [anon_sym_BANG] = ACTIONS(1088),
    [anon_sym_AT] = ACTIONS(1088),
    [aux_sym_clone_expression_token1] = ACTIONS(1090),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1090),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1090),
    [anon_sym_DASH_DASH] = ACTIONS(1088),
    [anon_sym_PLUS_PLUS] = ACTIONS(1088),
    [aux_sym__list_destructing_token1] = ACTIONS(1090),
    [anon_sym_LBRACK] = ACTIONS(1088),
    [anon_sym_self] = ACTIONS(1090),
    [anon_sym_parent] = ACTIONS(1090),
    [aux_sym__argument_name_token1] = ACTIONS(1090),
    [aux_sym__argument_name_token2] = ACTIONS(1090),
    [anon_sym_POUND_LBRACK] = ACTIONS(1088),
    [aux_sym_encapsed_string_token1] = ACTIONS(1088),
    [anon_sym_DQUOTE] = ACTIONS(1088),
    [aux_sym_string_token1] = ACTIONS(1088),
    [anon_sym_SQUOTE] = ACTIONS(1088),
    [anon_sym_LT_LT_LT] = ACTIONS(1088),
    [anon_sym_BQUOTE] = ACTIONS(1088),
    [anon_sym_DOLLAR] = ACTIONS(1088),
    [aux_sym_yield_expression_token1] = ACTIONS(1090),
    [aux_sym_include_expression_token1] = ACTIONS(1090),
    [aux_sym_include_once_expression_token1] = ACTIONS(1090),
    [aux_sym_require_expression_token1] = ACTIONS(1090),
    [aux_sym_require_once_expression_token1] = ACTIONS(1090),
    [sym_comment] = ACTIONS(5),
  },
  [465] = {
    [sym_text_interpolation] = STATE(465),
    [ts_builtin_sym_end] = ACTIONS(1092),
    [sym_name] = ACTIONS(1094),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1092),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1094),
    [aux_sym_global_declaration_token1] = ACTIONS(1094),
    [aux_sym_namespace_definition_token1] = ACTIONS(1094),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1094),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1094),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1094),
    [anon_sym_BSLASH] = ACTIONS(1092),
    [anon_sym_LBRACE] = ACTIONS(1092),
    [anon_sym_RBRACE] = ACTIONS(1092),
    [aux_sym_trait_declaration_token1] = ACTIONS(1094),
    [aux_sym_interface_declaration_token1] = ACTIONS(1094),
    [aux_sym_enum_declaration_token1] = ACTIONS(1094),
    [aux_sym_enum_case_token1] = ACTIONS(1094),
    [aux_sym_class_declaration_token1] = ACTIONS(1094),
    [aux_sym_final_modifier_token1] = ACTIONS(1094),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1094),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1094),
    [sym_var_modifier] = ACTIONS(1094),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1094),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1094),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1094),
    [anon_sym_LPAREN2] = ACTIONS(1092),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1094),
    [aux_sym_cast_type_token1] = ACTIONS(1094),
    [aux_sym_echo_statement_token1] = ACTIONS(1094),
    [aux_sym_exit_statement_token1] = ACTIONS(1094),
    [anon_sym_unset] = ACTIONS(1094),
    [aux_sym_declare_statement_token1] = ACTIONS(1094),
    [aux_sym_declare_statement_token2] = ACTIONS(1094),
    [sym_float] = ACTIONS(1094),
    [aux_sym_try_statement_token1] = ACTIONS(1094),
    [aux_sym_goto_statement_token1] = ACTIONS(1094),
    [aux_sym_continue_statement_token1] = ACTIONS(1094),
    [aux_sym_break_statement_token1] = ACTIONS(1094),
    [sym_integer] = ACTIONS(1094),
    [aux_sym_return_statement_token1] = ACTIONS(1094),
    [aux_sym_throw_expression_token1] = ACTIONS(1094),
    [aux_sym_while_statement_token1] = ACTIONS(1094),
    [aux_sym_while_statement_token2] = ACTIONS(1094),
    [aux_sym_do_statement_token1] = ACTIONS(1094),
    [aux_sym_for_statement_token1] = ACTIONS(1094),
    [aux_sym_for_statement_token2] = ACTIONS(1094),
    [aux_sym_foreach_statement_token1] = ACTIONS(1094),
    [aux_sym_foreach_statement_token2] = ACTIONS(1094),
    [aux_sym_if_statement_token1] = ACTIONS(1094),
    [aux_sym_if_statement_token2] = ACTIONS(1094),
    [aux_sym_else_if_clause_token1] = ACTIONS(1094),
    [aux_sym_else_clause_token1] = ACTIONS(1094),
    [aux_sym_match_expression_token1] = ACTIONS(1094),
    [aux_sym_match_default_expression_token1] = ACTIONS(1094),
    [aux_sym_switch_statement_token1] = ACTIONS(1094),
    [aux_sym_switch_block_token1] = ACTIONS(1094),
    [anon_sym_PLUS] = ACTIONS(1094),
    [anon_sym_DASH] = ACTIONS(1094),
    [anon_sym_TILDE] = ACTIONS(1092),
    [anon_sym_BANG] = ACTIONS(1092),
    [anon_sym_AT] = ACTIONS(1092),
    [aux_sym_clone_expression_token1] = ACTIONS(1094),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1094),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1094),
    [anon_sym_DASH_DASH] = ACTIONS(1092),
    [anon_sym_PLUS_PLUS] = ACTIONS(1092),
    [aux_sym__list_destructing_token1] = ACTIONS(1094),
    [anon_sym_LBRACK] = ACTIONS(1092),
    [anon_sym_self] = ACTIONS(1094),
    [anon_sym_parent] = ACTIONS(1094),
    [aux_sym__argument_name_token1] = ACTIONS(1094),
    [aux_sym__argument_name_token2] = ACTIONS(1094),
    [anon_sym_POUND_LBRACK] = ACTIONS(1092),
    [aux_sym_encapsed_string_token1] = ACTIONS(1092),
    [anon_sym_DQUOTE] = ACTIONS(1092),
    [aux_sym_string_token1] = ACTIONS(1092),
    [anon_sym_SQUOTE] = ACTIONS(1092),
    [anon_sym_LT_LT_LT] = ACTIONS(1092),
    [anon_sym_BQUOTE] = ACTIONS(1092),
    [anon_sym_DOLLAR] = ACTIONS(1092),
    [aux_sym_yield_expression_token1] = ACTIONS(1094),
    [aux_sym_include_expression_token1] = ACTIONS(1094),
    [aux_sym_include_once_expression_token1] = ACTIONS(1094),
    [aux_sym_require_expression_token1] = ACTIONS(1094),
    [aux_sym_require_once_expression_token1] = ACTIONS(1094),
    [sym_comment] = ACTIONS(5),
  },
  [466] = {
    [sym_text_interpolation] = STATE(466),
    [ts_builtin_sym_end] = ACTIONS(1096),
    [sym_name] = ACTIONS(1098),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1096),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1098),
    [aux_sym_global_declaration_token1] = ACTIONS(1098),
    [aux_sym_namespace_definition_token1] = ACTIONS(1098),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1098),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1098),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1098),
    [anon_sym_BSLASH] = ACTIONS(1096),
    [anon_sym_LBRACE] = ACTIONS(1096),
    [anon_sym_RBRACE] = ACTIONS(1096),
    [aux_sym_trait_declaration_token1] = ACTIONS(1098),
    [aux_sym_interface_declaration_token1] = ACTIONS(1098),
    [aux_sym_enum_declaration_token1] = ACTIONS(1098),
    [aux_sym_enum_case_token1] = ACTIONS(1098),
    [aux_sym_class_declaration_token1] = ACTIONS(1098),
    [aux_sym_final_modifier_token1] = ACTIONS(1098),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1098),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1098),
    [sym_var_modifier] = ACTIONS(1098),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1098),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1098),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1098),
    [anon_sym_LPAREN2] = ACTIONS(1096),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1098),
    [aux_sym_cast_type_token1] = ACTIONS(1098),
    [aux_sym_echo_statement_token1] = ACTIONS(1098),
    [aux_sym_exit_statement_token1] = ACTIONS(1098),
    [anon_sym_unset] = ACTIONS(1098),
    [aux_sym_declare_statement_token1] = ACTIONS(1098),
    [aux_sym_declare_statement_token2] = ACTIONS(1098),
    [sym_float] = ACTIONS(1098),
    [aux_sym_try_statement_token1] = ACTIONS(1098),
    [aux_sym_goto_statement_token1] = ACTIONS(1098),
    [aux_sym_continue_statement_token1] = ACTIONS(1098),
    [aux_sym_break_statement_token1] = ACTIONS(1098),
    [sym_integer] = ACTIONS(1098),
    [aux_sym_return_statement_token1] = ACTIONS(1098),
    [aux_sym_throw_expression_token1] = ACTIONS(1098),
    [aux_sym_while_statement_token1] = ACTIONS(1098),
    [aux_sym_while_statement_token2] = ACTIONS(1098),
    [aux_sym_do_statement_token1] = ACTIONS(1098),
    [aux_sym_for_statement_token1] = ACTIONS(1098),
    [aux_sym_for_statement_token2] = ACTIONS(1098),
    [aux_sym_foreach_statement_token1] = ACTIONS(1098),
    [aux_sym_foreach_statement_token2] = ACTIONS(1098),
    [aux_sym_if_statement_token1] = ACTIONS(1098),
    [aux_sym_if_statement_token2] = ACTIONS(1098),
    [aux_sym_else_if_clause_token1] = ACTIONS(1098),
    [aux_sym_else_clause_token1] = ACTIONS(1098),
    [aux_sym_match_expression_token1] = ACTIONS(1098),
    [aux_sym_match_default_expression_token1] = ACTIONS(1098),
    [aux_sym_switch_statement_token1] = ACTIONS(1098),
    [aux_sym_switch_block_token1] = ACTIONS(1098),
    [anon_sym_PLUS] = ACTIONS(1098),
    [anon_sym_DASH] = ACTIONS(1098),
    [anon_sym_TILDE] = ACTIONS(1096),
    [anon_sym_BANG] = ACTIONS(1096),
    [anon_sym_AT] = ACTIONS(1096),
    [aux_sym_clone_expression_token1] = ACTIONS(1098),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1098),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1098),
    [anon_sym_DASH_DASH] = ACTIONS(1096),
    [anon_sym_PLUS_PLUS] = ACTIONS(1096),
    [aux_sym__list_destructing_token1] = ACTIONS(1098),
    [anon_sym_LBRACK] = ACTIONS(1096),
    [anon_sym_self] = ACTIONS(1098),
    [anon_sym_parent] = ACTIONS(1098),
    [aux_sym__argument_name_token1] = ACTIONS(1098),
    [aux_sym__argument_name_token2] = ACTIONS(1098),
    [anon_sym_POUND_LBRACK] = ACTIONS(1096),
    [aux_sym_encapsed_string_token1] = ACTIONS(1096),
    [anon_sym_DQUOTE] = ACTIONS(1096),
    [aux_sym_string_token1] = ACTIONS(1096),
    [anon_sym_SQUOTE] = ACTIONS(1096),
    [anon_sym_LT_LT_LT] = ACTIONS(1096),
    [anon_sym_BQUOTE] = ACTIONS(1096),
    [anon_sym_DOLLAR] = ACTIONS(1096),
    [aux_sym_yield_expression_token1] = ACTIONS(1098),
    [aux_sym_include_expression_token1] = ACTIONS(1098),
    [aux_sym_include_once_expression_token1] = ACTIONS(1098),
    [aux_sym_require_expression_token1] = ACTIONS(1098),
    [aux_sym_require_once_expression_token1] = ACTIONS(1098),
    [sym_comment] = ACTIONS(5),
  },
  [467] = {
    [sym_text_interpolation] = STATE(467),
    [ts_builtin_sym_end] = ACTIONS(1100),
    [sym_name] = ACTIONS(1102),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1100),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1102),
    [aux_sym_global_declaration_token1] = ACTIONS(1102),
    [aux_sym_namespace_definition_token1] = ACTIONS(1102),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1102),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1102),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1102),
    [anon_sym_BSLASH] = ACTIONS(1100),
    [anon_sym_LBRACE] = ACTIONS(1100),
    [anon_sym_RBRACE] = ACTIONS(1100),
    [aux_sym_trait_declaration_token1] = ACTIONS(1102),
    [aux_sym_interface_declaration_token1] = ACTIONS(1102),
    [aux_sym_enum_declaration_token1] = ACTIONS(1102),
    [aux_sym_enum_case_token1] = ACTIONS(1102),
    [aux_sym_class_declaration_token1] = ACTIONS(1102),
    [aux_sym_final_modifier_token1] = ACTIONS(1102),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1102),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1102),
    [sym_var_modifier] = ACTIONS(1102),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1102),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1102),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1102),
    [anon_sym_LPAREN2] = ACTIONS(1100),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1102),
    [aux_sym_cast_type_token1] = ACTIONS(1102),
    [aux_sym_echo_statement_token1] = ACTIONS(1102),
    [aux_sym_exit_statement_token1] = ACTIONS(1102),
    [anon_sym_unset] = ACTIONS(1102),
    [aux_sym_declare_statement_token1] = ACTIONS(1102),
    [aux_sym_declare_statement_token2] = ACTIONS(1102),
    [sym_float] = ACTIONS(1102),
    [aux_sym_try_statement_token1] = ACTIONS(1102),
    [aux_sym_goto_statement_token1] = ACTIONS(1102),
    [aux_sym_continue_statement_token1] = ACTIONS(1102),
    [aux_sym_break_statement_token1] = ACTIONS(1102),
    [sym_integer] = ACTIONS(1102),
    [aux_sym_return_statement_token1] = ACTIONS(1102),
    [aux_sym_throw_expression_token1] = ACTIONS(1102),
    [aux_sym_while_statement_token1] = ACTIONS(1102),
    [aux_sym_while_statement_token2] = ACTIONS(1102),
    [aux_sym_do_statement_token1] = ACTIONS(1102),
    [aux_sym_for_statement_token1] = ACTIONS(1102),
    [aux_sym_for_statement_token2] = ACTIONS(1102),
    [aux_sym_foreach_statement_token1] = ACTIONS(1102),
    [aux_sym_foreach_statement_token2] = ACTIONS(1102),
    [aux_sym_if_statement_token1] = ACTIONS(1102),
    [aux_sym_if_statement_token2] = ACTIONS(1102),
    [aux_sym_else_if_clause_token1] = ACTIONS(1102),
    [aux_sym_else_clause_token1] = ACTIONS(1102),
    [aux_sym_match_expression_token1] = ACTIONS(1102),
    [aux_sym_match_default_expression_token1] = ACTIONS(1102),
    [aux_sym_switch_statement_token1] = ACTIONS(1102),
    [aux_sym_switch_block_token1] = ACTIONS(1102),
    [anon_sym_PLUS] = ACTIONS(1102),
    [anon_sym_DASH] = ACTIONS(1102),
    [anon_sym_TILDE] = ACTIONS(1100),
    [anon_sym_BANG] = ACTIONS(1100),
    [anon_sym_AT] = ACTIONS(1100),
    [aux_sym_clone_expression_token1] = ACTIONS(1102),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1102),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1102),
    [anon_sym_DASH_DASH] = ACTIONS(1100),
    [anon_sym_PLUS_PLUS] = ACTIONS(1100),
    [aux_sym__list_destructing_token1] = ACTIONS(1102),
    [anon_sym_LBRACK] = ACTIONS(1100),
    [anon_sym_self] = ACTIONS(1102),
    [anon_sym_parent] = ACTIONS(1102),
    [aux_sym__argument_name_token1] = ACTIONS(1102),
    [aux_sym__argument_name_token2] = ACTIONS(1102),
    [anon_sym_POUND_LBRACK] = ACTIONS(1100),
    [aux_sym_encapsed_string_token1] = ACTIONS(1100),
    [anon_sym_DQUOTE] = ACTIONS(1100),
    [aux_sym_string_token1] = ACTIONS(1100),
    [anon_sym_SQUOTE] = ACTIONS(1100),
    [anon_sym_LT_LT_LT] = ACTIONS(1100),
    [anon_sym_BQUOTE] = ACTIONS(1100),
    [anon_sym_DOLLAR] = ACTIONS(1100),
    [aux_sym_yield_expression_token1] = ACTIONS(1102),
    [aux_sym_include_expression_token1] = ACTIONS(1102),
    [aux_sym_include_once_expression_token1] = ACTIONS(1102),
    [aux_sym_require_expression_token1] = ACTIONS(1102),
    [aux_sym_require_once_expression_token1] = ACTIONS(1102),
    [sym_comment] = ACTIONS(5),
  },
  [468] = {
    [sym_text_interpolation] = STATE(468),
    [ts_builtin_sym_end] = ACTIONS(1104),
    [sym_name] = ACTIONS(1106),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1104),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1106),
    [aux_sym_global_declaration_token1] = ACTIONS(1106),
    [aux_sym_namespace_definition_token1] = ACTIONS(1106),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1106),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1106),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1106),
    [anon_sym_BSLASH] = ACTIONS(1104),
    [anon_sym_LBRACE] = ACTIONS(1104),
    [anon_sym_RBRACE] = ACTIONS(1104),
    [aux_sym_trait_declaration_token1] = ACTIONS(1106),
    [aux_sym_interface_declaration_token1] = ACTIONS(1106),
    [aux_sym_enum_declaration_token1] = ACTIONS(1106),
    [aux_sym_enum_case_token1] = ACTIONS(1106),
    [aux_sym_class_declaration_token1] = ACTIONS(1106),
    [aux_sym_final_modifier_token1] = ACTIONS(1106),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1106),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1106),
    [sym_var_modifier] = ACTIONS(1106),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1106),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1106),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1106),
    [anon_sym_LPAREN2] = ACTIONS(1104),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1106),
    [aux_sym_cast_type_token1] = ACTIONS(1106),
    [aux_sym_echo_statement_token1] = ACTIONS(1106),
    [aux_sym_exit_statement_token1] = ACTIONS(1106),
    [anon_sym_unset] = ACTIONS(1106),
    [aux_sym_declare_statement_token1] = ACTIONS(1106),
    [aux_sym_declare_statement_token2] = ACTIONS(1106),
    [sym_float] = ACTIONS(1106),
    [aux_sym_try_statement_token1] = ACTIONS(1106),
    [aux_sym_goto_statement_token1] = ACTIONS(1106),
    [aux_sym_continue_statement_token1] = ACTIONS(1106),
    [aux_sym_break_statement_token1] = ACTIONS(1106),
    [sym_integer] = ACTIONS(1106),
    [aux_sym_return_statement_token1] = ACTIONS(1106),
    [aux_sym_throw_expression_token1] = ACTIONS(1106),
    [aux_sym_while_statement_token1] = ACTIONS(1106),
    [aux_sym_while_statement_token2] = ACTIONS(1106),
    [aux_sym_do_statement_token1] = ACTIONS(1106),
    [aux_sym_for_statement_token1] = ACTIONS(1106),
    [aux_sym_for_statement_token2] = ACTIONS(1106),
    [aux_sym_foreach_statement_token1] = ACTIONS(1106),
    [aux_sym_foreach_statement_token2] = ACTIONS(1106),
    [aux_sym_if_statement_token1] = ACTIONS(1106),
    [aux_sym_if_statement_token2] = ACTIONS(1106),
    [aux_sym_else_if_clause_token1] = ACTIONS(1106),
    [aux_sym_else_clause_token1] = ACTIONS(1106),
    [aux_sym_match_expression_token1] = ACTIONS(1106),
    [aux_sym_match_default_expression_token1] = ACTIONS(1106),
    [aux_sym_switch_statement_token1] = ACTIONS(1106),
    [aux_sym_switch_block_token1] = ACTIONS(1106),
    [anon_sym_PLUS] = ACTIONS(1106),
    [anon_sym_DASH] = ACTIONS(1106),
    [anon_sym_TILDE] = ACTIONS(1104),
    [anon_sym_BANG] = ACTIONS(1104),
    [anon_sym_AT] = ACTIONS(1104),
    [aux_sym_clone_expression_token1] = ACTIONS(1106),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1106),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1106),
    [anon_sym_DASH_DASH] = ACTIONS(1104),
    [anon_sym_PLUS_PLUS] = ACTIONS(1104),
    [aux_sym__list_destructing_token1] = ACTIONS(1106),
    [anon_sym_LBRACK] = ACTIONS(1104),
    [anon_sym_self] = ACTIONS(1106),
    [anon_sym_parent] = ACTIONS(1106),
    [aux_sym__argument_name_token1] = ACTIONS(1106),
    [aux_sym__argument_name_token2] = ACTIONS(1106),
    [anon_sym_POUND_LBRACK] = ACTIONS(1104),
    [aux_sym_encapsed_string_token1] = ACTIONS(1104),
    [anon_sym_DQUOTE] = ACTIONS(1104),
    [aux_sym_string_token1] = ACTIONS(1104),
    [anon_sym_SQUOTE] = ACTIONS(1104),
    [anon_sym_LT_LT_LT] = ACTIONS(1104),
    [anon_sym_BQUOTE] = ACTIONS(1104),
    [anon_sym_DOLLAR] = ACTIONS(1104),
    [aux_sym_yield_expression_token1] = ACTIONS(1106),
    [aux_sym_include_expression_token1] = ACTIONS(1106),
    [aux_sym_include_once_expression_token1] = ACTIONS(1106),
    [aux_sym_require_expression_token1] = ACTIONS(1106),
    [aux_sym_require_once_expression_token1] = ACTIONS(1106),
    [sym_comment] = ACTIONS(5),
  },
  [469] = {
    [sym_text_interpolation] = STATE(469),
    [ts_builtin_sym_end] = ACTIONS(1108),
    [sym_name] = ACTIONS(1110),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1108),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1110),
    [aux_sym_global_declaration_token1] = ACTIONS(1110),
    [aux_sym_namespace_definition_token1] = ACTIONS(1110),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1110),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1110),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1110),
    [anon_sym_BSLASH] = ACTIONS(1108),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_RBRACE] = ACTIONS(1108),
    [aux_sym_trait_declaration_token1] = ACTIONS(1110),
    [aux_sym_interface_declaration_token1] = ACTIONS(1110),
    [aux_sym_enum_declaration_token1] = ACTIONS(1110),
    [aux_sym_enum_case_token1] = ACTIONS(1110),
    [aux_sym_class_declaration_token1] = ACTIONS(1110),
    [aux_sym_final_modifier_token1] = ACTIONS(1110),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1110),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1110),
    [sym_var_modifier] = ACTIONS(1110),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1110),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1110),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1110),
    [anon_sym_LPAREN2] = ACTIONS(1108),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1110),
    [aux_sym_cast_type_token1] = ACTIONS(1110),
    [aux_sym_echo_statement_token1] = ACTIONS(1110),
    [aux_sym_exit_statement_token1] = ACTIONS(1110),
    [anon_sym_unset] = ACTIONS(1110),
    [aux_sym_declare_statement_token1] = ACTIONS(1110),
    [aux_sym_declare_statement_token2] = ACTIONS(1110),
    [sym_float] = ACTIONS(1110),
    [aux_sym_try_statement_token1] = ACTIONS(1110),
    [aux_sym_goto_statement_token1] = ACTIONS(1110),
    [aux_sym_continue_statement_token1] = ACTIONS(1110),
    [aux_sym_break_statement_token1] = ACTIONS(1110),
    [sym_integer] = ACTIONS(1110),
    [aux_sym_return_statement_token1] = ACTIONS(1110),
    [aux_sym_throw_expression_token1] = ACTIONS(1110),
    [aux_sym_while_statement_token1] = ACTIONS(1110),
    [aux_sym_while_statement_token2] = ACTIONS(1110),
    [aux_sym_do_statement_token1] = ACTIONS(1110),
    [aux_sym_for_statement_token1] = ACTIONS(1110),
    [aux_sym_for_statement_token2] = ACTIONS(1110),
    [aux_sym_foreach_statement_token1] = ACTIONS(1110),
    [aux_sym_foreach_statement_token2] = ACTIONS(1110),
    [aux_sym_if_statement_token1] = ACTIONS(1110),
    [aux_sym_if_statement_token2] = ACTIONS(1110),
    [aux_sym_else_if_clause_token1] = ACTIONS(1110),
    [aux_sym_else_clause_token1] = ACTIONS(1110),
    [aux_sym_match_expression_token1] = ACTIONS(1110),
    [aux_sym_match_default_expression_token1] = ACTIONS(1110),
    [aux_sym_switch_statement_token1] = ACTIONS(1110),
    [aux_sym_switch_block_token1] = ACTIONS(1110),
    [anon_sym_PLUS] = ACTIONS(1110),
    [anon_sym_DASH] = ACTIONS(1110),
    [anon_sym_TILDE] = ACTIONS(1108),
    [anon_sym_BANG] = ACTIONS(1108),
    [anon_sym_AT] = ACTIONS(1108),
    [aux_sym_clone_expression_token1] = ACTIONS(1110),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1110),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1110),
    [anon_sym_DASH_DASH] = ACTIONS(1108),
    [anon_sym_PLUS_PLUS] = ACTIONS(1108),
    [aux_sym__list_destructing_token1] = ACTIONS(1110),
    [anon_sym_LBRACK] = ACTIONS(1108),
    [anon_sym_self] = ACTIONS(1110),
    [anon_sym_parent] = ACTIONS(1110),
    [aux_sym__argument_name_token1] = ACTIONS(1110),
    [aux_sym__argument_name_token2] = ACTIONS(1110),
    [anon_sym_POUND_LBRACK] = ACTIONS(1108),
    [aux_sym_encapsed_string_token1] = ACTIONS(1108),
    [anon_sym_DQUOTE] = ACTIONS(1108),
    [aux_sym_string_token1] = ACTIONS(1108),
    [anon_sym_SQUOTE] = ACTIONS(1108),
    [anon_sym_LT_LT_LT] = ACTIONS(1108),
    [anon_sym_BQUOTE] = ACTIONS(1108),
    [anon_sym_DOLLAR] = ACTIONS(1108),
    [aux_sym_yield_expression_token1] = ACTIONS(1110),
    [aux_sym_include_expression_token1] = ACTIONS(1110),
    [aux_sym_include_once_expression_token1] = ACTIONS(1110),
    [aux_sym_require_expression_token1] = ACTIONS(1110),
    [aux_sym_require_once_expression_token1] = ACTIONS(1110),
    [sym_comment] = ACTIONS(5),
  },
  [470] = {
    [sym_text_interpolation] = STATE(470),
    [ts_builtin_sym_end] = ACTIONS(1112),
    [sym_name] = ACTIONS(1114),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1112),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1114),
    [aux_sym_global_declaration_token1] = ACTIONS(1114),
    [aux_sym_namespace_definition_token1] = ACTIONS(1114),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1114),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1114),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1114),
    [anon_sym_BSLASH] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [aux_sym_trait_declaration_token1] = ACTIONS(1114),
    [aux_sym_interface_declaration_token1] = ACTIONS(1114),
    [aux_sym_enum_declaration_token1] = ACTIONS(1114),
    [aux_sym_enum_case_token1] = ACTIONS(1114),
    [aux_sym_class_declaration_token1] = ACTIONS(1114),
    [aux_sym_final_modifier_token1] = ACTIONS(1114),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1114),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1114),
    [sym_var_modifier] = ACTIONS(1114),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1114),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1114),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1114),
    [anon_sym_LPAREN2] = ACTIONS(1112),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1114),
    [aux_sym_cast_type_token1] = ACTIONS(1114),
    [aux_sym_echo_statement_token1] = ACTIONS(1114),
    [aux_sym_exit_statement_token1] = ACTIONS(1114),
    [anon_sym_unset] = ACTIONS(1114),
    [aux_sym_declare_statement_token1] = ACTIONS(1114),
    [aux_sym_declare_statement_token2] = ACTIONS(1114),
    [sym_float] = ACTIONS(1114),
    [aux_sym_try_statement_token1] = ACTIONS(1114),
    [aux_sym_goto_statement_token1] = ACTIONS(1114),
    [aux_sym_continue_statement_token1] = ACTIONS(1114),
    [aux_sym_break_statement_token1] = ACTIONS(1114),
    [sym_integer] = ACTIONS(1114),
    [aux_sym_return_statement_token1] = ACTIONS(1114),
    [aux_sym_throw_expression_token1] = ACTIONS(1114),
    [aux_sym_while_statement_token1] = ACTIONS(1114),
    [aux_sym_while_statement_token2] = ACTIONS(1114),
    [aux_sym_do_statement_token1] = ACTIONS(1114),
    [aux_sym_for_statement_token1] = ACTIONS(1114),
    [aux_sym_for_statement_token2] = ACTIONS(1114),
    [aux_sym_foreach_statement_token1] = ACTIONS(1114),
    [aux_sym_foreach_statement_token2] = ACTIONS(1114),
    [aux_sym_if_statement_token1] = ACTIONS(1114),
    [aux_sym_if_statement_token2] = ACTIONS(1114),
    [aux_sym_else_if_clause_token1] = ACTIONS(1114),
    [aux_sym_else_clause_token1] = ACTIONS(1114),
    [aux_sym_match_expression_token1] = ACTIONS(1114),
    [aux_sym_match_default_expression_token1] = ACTIONS(1114),
    [aux_sym_switch_statement_token1] = ACTIONS(1114),
    [aux_sym_switch_block_token1] = ACTIONS(1114),
    [anon_sym_PLUS] = ACTIONS(1114),
    [anon_sym_DASH] = ACTIONS(1114),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [aux_sym_clone_expression_token1] = ACTIONS(1114),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1114),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1114),
    [anon_sym_DASH_DASH] = ACTIONS(1112),
    [anon_sym_PLUS_PLUS] = ACTIONS(1112),
    [aux_sym__list_destructing_token1] = ACTIONS(1114),
    [anon_sym_LBRACK] = ACTIONS(1112),
    [anon_sym_self] = ACTIONS(1114),
    [anon_sym_parent] = ACTIONS(1114),
    [aux_sym__argument_name_token1] = ACTIONS(1114),
    [aux_sym__argument_name_token2] = ACTIONS(1114),
    [anon_sym_POUND_LBRACK] = ACTIONS(1112),
    [aux_sym_encapsed_string_token1] = ACTIONS(1112),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [aux_sym_string_token1] = ACTIONS(1112),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_LT_LT_LT] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [aux_sym_yield_expression_token1] = ACTIONS(1114),
    [aux_sym_include_expression_token1] = ACTIONS(1114),
    [aux_sym_include_once_expression_token1] = ACTIONS(1114),
    [aux_sym_require_expression_token1] = ACTIONS(1114),
    [aux_sym_require_once_expression_token1] = ACTIONS(1114),
    [sym_comment] = ACTIONS(5),
  },
  [471] = {
    [sym_text_interpolation] = STATE(471),
    [ts_builtin_sym_end] = ACTIONS(1116),
    [sym_name] = ACTIONS(1118),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1116),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1118),
    [aux_sym_global_declaration_token1] = ACTIONS(1118),
    [aux_sym_namespace_definition_token1] = ACTIONS(1118),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1118),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1118),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1118),
    [anon_sym_BSLASH] = ACTIONS(1116),
    [anon_sym_LBRACE] = ACTIONS(1116),
    [anon_sym_RBRACE] = ACTIONS(1116),
    [aux_sym_trait_declaration_token1] = ACTIONS(1118),
    [aux_sym_interface_declaration_token1] = ACTIONS(1118),
    [aux_sym_enum_declaration_token1] = ACTIONS(1118),
    [aux_sym_enum_case_token1] = ACTIONS(1118),
    [aux_sym_class_declaration_token1] = ACTIONS(1118),
    [aux_sym_final_modifier_token1] = ACTIONS(1118),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1118),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1118),
    [sym_var_modifier] = ACTIONS(1118),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1118),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1118),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1118),
    [anon_sym_LPAREN2] = ACTIONS(1116),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1118),
    [aux_sym_cast_type_token1] = ACTIONS(1118),
    [aux_sym_echo_statement_token1] = ACTIONS(1118),
    [aux_sym_exit_statement_token1] = ACTIONS(1118),
    [anon_sym_unset] = ACTIONS(1118),
    [aux_sym_declare_statement_token1] = ACTIONS(1118),
    [aux_sym_declare_statement_token2] = ACTIONS(1118),
    [sym_float] = ACTIONS(1118),
    [aux_sym_try_statement_token1] = ACTIONS(1118),
    [aux_sym_goto_statement_token1] = ACTIONS(1118),
    [aux_sym_continue_statement_token1] = ACTIONS(1118),
    [aux_sym_break_statement_token1] = ACTIONS(1118),
    [sym_integer] = ACTIONS(1118),
    [aux_sym_return_statement_token1] = ACTIONS(1118),
    [aux_sym_throw_expression_token1] = ACTIONS(1118),
    [aux_sym_while_statement_token1] = ACTIONS(1118),
    [aux_sym_while_statement_token2] = ACTIONS(1118),
    [aux_sym_do_statement_token1] = ACTIONS(1118),
    [aux_sym_for_statement_token1] = ACTIONS(1118),
    [aux_sym_for_statement_token2] = ACTIONS(1118),
    [aux_sym_foreach_statement_token1] = ACTIONS(1118),
    [aux_sym_foreach_statement_token2] = ACTIONS(1118),
    [aux_sym_if_statement_token1] = ACTIONS(1118),
    [aux_sym_if_statement_token2] = ACTIONS(1118),
    [aux_sym_else_if_clause_token1] = ACTIONS(1118),
    [aux_sym_else_clause_token1] = ACTIONS(1118),
    [aux_sym_match_expression_token1] = ACTIONS(1118),
    [aux_sym_match_default_expression_token1] = ACTIONS(1118),
    [aux_sym_switch_statement_token1] = ACTIONS(1118),
    [aux_sym_switch_block_token1] = ACTIONS(1118),
    [anon_sym_PLUS] = ACTIONS(1118),
    [anon_sym_DASH] = ACTIONS(1118),
    [anon_sym_TILDE] = ACTIONS(1116),
    [anon_sym_BANG] = ACTIONS(1116),
    [anon_sym_AT] = ACTIONS(1116),
    [aux_sym_clone_expression_token1] = ACTIONS(1118),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1118),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1118),
    [anon_sym_DASH_DASH] = ACTIONS(1116),
    [anon_sym_PLUS_PLUS] = ACTIONS(1116),
    [aux_sym__list_destructing_token1] = ACTIONS(1118),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [anon_sym_self] = ACTIONS(1118),
    [anon_sym_parent] = ACTIONS(1118),
    [aux_sym__argument_name_token1] = ACTIONS(1118),
    [aux_sym__argument_name_token2] = ACTIONS(1118),
    [anon_sym_POUND_LBRACK] = ACTIONS(1116),
    [aux_sym_encapsed_string_token1] = ACTIONS(1116),
    [anon_sym_DQUOTE] = ACTIONS(1116),
    [aux_sym_string_token1] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1116),
    [anon_sym_LT_LT_LT] = ACTIONS(1116),
    [anon_sym_BQUOTE] = ACTIONS(1116),
    [anon_sym_DOLLAR] = ACTIONS(1116),
    [aux_sym_yield_expression_token1] = ACTIONS(1118),
    [aux_sym_include_expression_token1] = ACTIONS(1118),
    [aux_sym_include_once_expression_token1] = ACTIONS(1118),
    [aux_sym_require_expression_token1] = ACTIONS(1118),
    [aux_sym_require_once_expression_token1] = ACTIONS(1118),
    [sym_comment] = ACTIONS(5),
  },
  [472] = {
    [sym_text_interpolation] = STATE(472),
    [ts_builtin_sym_end] = ACTIONS(1120),
    [sym_name] = ACTIONS(1122),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1120),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1122),
    [aux_sym_global_declaration_token1] = ACTIONS(1122),
    [aux_sym_namespace_definition_token1] = ACTIONS(1122),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1122),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1122),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1122),
    [anon_sym_BSLASH] = ACTIONS(1120),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_RBRACE] = ACTIONS(1120),
    [aux_sym_trait_declaration_token1] = ACTIONS(1122),
    [aux_sym_interface_declaration_token1] = ACTIONS(1122),
    [aux_sym_enum_declaration_token1] = ACTIONS(1122),
    [aux_sym_enum_case_token1] = ACTIONS(1122),
    [aux_sym_class_declaration_token1] = ACTIONS(1122),
    [aux_sym_final_modifier_token1] = ACTIONS(1122),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1122),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1122),
    [sym_var_modifier] = ACTIONS(1122),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1122),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1122),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1122),
    [anon_sym_LPAREN2] = ACTIONS(1120),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1122),
    [aux_sym_cast_type_token1] = ACTIONS(1122),
    [aux_sym_echo_statement_token1] = ACTIONS(1122),
    [aux_sym_exit_statement_token1] = ACTIONS(1122),
    [anon_sym_unset] = ACTIONS(1122),
    [aux_sym_declare_statement_token1] = ACTIONS(1122),
    [aux_sym_declare_statement_token2] = ACTIONS(1122),
    [sym_float] = ACTIONS(1122),
    [aux_sym_try_statement_token1] = ACTIONS(1122),
    [aux_sym_goto_statement_token1] = ACTIONS(1122),
    [aux_sym_continue_statement_token1] = ACTIONS(1122),
    [aux_sym_break_statement_token1] = ACTIONS(1122),
    [sym_integer] = ACTIONS(1122),
    [aux_sym_return_statement_token1] = ACTIONS(1122),
    [aux_sym_throw_expression_token1] = ACTIONS(1122),
    [aux_sym_while_statement_token1] = ACTIONS(1122),
    [aux_sym_while_statement_token2] = ACTIONS(1122),
    [aux_sym_do_statement_token1] = ACTIONS(1122),
    [aux_sym_for_statement_token1] = ACTIONS(1122),
    [aux_sym_for_statement_token2] = ACTIONS(1122),
    [aux_sym_foreach_statement_token1] = ACTIONS(1122),
    [aux_sym_foreach_statement_token2] = ACTIONS(1122),
    [aux_sym_if_statement_token1] = ACTIONS(1122),
    [aux_sym_if_statement_token2] = ACTIONS(1122),
    [aux_sym_else_if_clause_token1] = ACTIONS(1122),
    [aux_sym_else_clause_token1] = ACTIONS(1122),
    [aux_sym_match_expression_token1] = ACTIONS(1122),
    [aux_sym_match_default_expression_token1] = ACTIONS(1122),
    [aux_sym_switch_statement_token1] = ACTIONS(1122),
    [aux_sym_switch_block_token1] = ACTIONS(1122),
    [anon_sym_PLUS] = ACTIONS(1122),
    [anon_sym_DASH] = ACTIONS(1122),
    [anon_sym_TILDE] = ACTIONS(1120),
    [anon_sym_BANG] = ACTIONS(1120),
    [anon_sym_AT] = ACTIONS(1120),
    [aux_sym_clone_expression_token1] = ACTIONS(1122),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1122),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1122),
    [anon_sym_DASH_DASH] = ACTIONS(1120),
    [anon_sym_PLUS_PLUS] = ACTIONS(1120),
    [aux_sym__list_destructing_token1] = ACTIONS(1122),
    [anon_sym_LBRACK] = ACTIONS(1120),
    [anon_sym_self] = ACTIONS(1122),
    [anon_sym_parent] = ACTIONS(1122),
    [aux_sym__argument_name_token1] = ACTIONS(1122),
    [aux_sym__argument_name_token2] = ACTIONS(1122),
    [anon_sym_POUND_LBRACK] = ACTIONS(1120),
    [aux_sym_encapsed_string_token1] = ACTIONS(1120),
    [anon_sym_DQUOTE] = ACTIONS(1120),
    [aux_sym_string_token1] = ACTIONS(1120),
    [anon_sym_SQUOTE] = ACTIONS(1120),
    [anon_sym_LT_LT_LT] = ACTIONS(1120),
    [anon_sym_BQUOTE] = ACTIONS(1120),
    [anon_sym_DOLLAR] = ACTIONS(1120),
    [aux_sym_yield_expression_token1] = ACTIONS(1122),
    [aux_sym_include_expression_token1] = ACTIONS(1122),
    [aux_sym_include_once_expression_token1] = ACTIONS(1122),
    [aux_sym_require_expression_token1] = ACTIONS(1122),
    [aux_sym_require_once_expression_token1] = ACTIONS(1122),
    [sym_comment] = ACTIONS(5),
  },
  [473] = {
    [sym_text_interpolation] = STATE(473),
    [ts_builtin_sym_end] = ACTIONS(1124),
    [sym_name] = ACTIONS(1126),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1124),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1126),
    [aux_sym_global_declaration_token1] = ACTIONS(1126),
    [aux_sym_namespace_definition_token1] = ACTIONS(1126),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1126),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1126),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1126),
    [anon_sym_BSLASH] = ACTIONS(1124),
    [anon_sym_LBRACE] = ACTIONS(1124),
    [anon_sym_RBRACE] = ACTIONS(1124),
    [aux_sym_trait_declaration_token1] = ACTIONS(1126),
    [aux_sym_interface_declaration_token1] = ACTIONS(1126),
    [aux_sym_enum_declaration_token1] = ACTIONS(1126),
    [aux_sym_enum_case_token1] = ACTIONS(1126),
    [aux_sym_class_declaration_token1] = ACTIONS(1126),
    [aux_sym_final_modifier_token1] = ACTIONS(1126),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1126),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1126),
    [sym_var_modifier] = ACTIONS(1126),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1126),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1126),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1126),
    [anon_sym_LPAREN2] = ACTIONS(1124),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1126),
    [aux_sym_cast_type_token1] = ACTIONS(1126),
    [aux_sym_echo_statement_token1] = ACTIONS(1126),
    [aux_sym_exit_statement_token1] = ACTIONS(1126),
    [anon_sym_unset] = ACTIONS(1126),
    [aux_sym_declare_statement_token1] = ACTIONS(1126),
    [aux_sym_declare_statement_token2] = ACTIONS(1126),
    [sym_float] = ACTIONS(1126),
    [aux_sym_try_statement_token1] = ACTIONS(1126),
    [aux_sym_goto_statement_token1] = ACTIONS(1126),
    [aux_sym_continue_statement_token1] = ACTIONS(1126),
    [aux_sym_break_statement_token1] = ACTIONS(1126),
    [sym_integer] = ACTIONS(1126),
    [aux_sym_return_statement_token1] = ACTIONS(1126),
    [aux_sym_throw_expression_token1] = ACTIONS(1126),
    [aux_sym_while_statement_token1] = ACTIONS(1126),
    [aux_sym_while_statement_token2] = ACTIONS(1126),
    [aux_sym_do_statement_token1] = ACTIONS(1126),
    [aux_sym_for_statement_token1] = ACTIONS(1126),
    [aux_sym_for_statement_token2] = ACTIONS(1126),
    [aux_sym_foreach_statement_token1] = ACTIONS(1126),
    [aux_sym_foreach_statement_token2] = ACTIONS(1126),
    [aux_sym_if_statement_token1] = ACTIONS(1126),
    [aux_sym_if_statement_token2] = ACTIONS(1126),
    [aux_sym_else_if_clause_token1] = ACTIONS(1126),
    [aux_sym_else_clause_token1] = ACTIONS(1126),
    [aux_sym_match_expression_token1] = ACTIONS(1126),
    [aux_sym_match_default_expression_token1] = ACTIONS(1126),
    [aux_sym_switch_statement_token1] = ACTIONS(1126),
    [aux_sym_switch_block_token1] = ACTIONS(1126),
    [anon_sym_PLUS] = ACTIONS(1126),
    [anon_sym_DASH] = ACTIONS(1126),
    [anon_sym_TILDE] = ACTIONS(1124),
    [anon_sym_BANG] = ACTIONS(1124),
    [anon_sym_AT] = ACTIONS(1124),
    [aux_sym_clone_expression_token1] = ACTIONS(1126),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1126),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1126),
    [anon_sym_DASH_DASH] = ACTIONS(1124),
    [anon_sym_PLUS_PLUS] = ACTIONS(1124),
    [aux_sym__list_destructing_token1] = ACTIONS(1126),
    [anon_sym_LBRACK] = ACTIONS(1124),
    [anon_sym_self] = ACTIONS(1126),
    [anon_sym_parent] = ACTIONS(1126),
    [aux_sym__argument_name_token1] = ACTIONS(1126),
    [aux_sym__argument_name_token2] = ACTIONS(1126),
    [anon_sym_POUND_LBRACK] = ACTIONS(1124),
    [aux_sym_encapsed_string_token1] = ACTIONS(1124),
    [anon_sym_DQUOTE] = ACTIONS(1124),
    [aux_sym_string_token1] = ACTIONS(1124),
    [anon_sym_SQUOTE] = ACTIONS(1124),
    [anon_sym_LT_LT_LT] = ACTIONS(1124),
    [anon_sym_BQUOTE] = ACTIONS(1124),
    [anon_sym_DOLLAR] = ACTIONS(1124),
    [aux_sym_yield_expression_token1] = ACTIONS(1126),
    [aux_sym_include_expression_token1] = ACTIONS(1126),
    [aux_sym_include_once_expression_token1] = ACTIONS(1126),
    [aux_sym_require_expression_token1] = ACTIONS(1126),
    [aux_sym_require_once_expression_token1] = ACTIONS(1126),
    [sym_comment] = ACTIONS(5),
  },
  [474] = {
    [sym_text_interpolation] = STATE(474),
    [ts_builtin_sym_end] = ACTIONS(1128),
    [sym_name] = ACTIONS(1130),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1128),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1130),
    [aux_sym_global_declaration_token1] = ACTIONS(1130),
    [aux_sym_namespace_definition_token1] = ACTIONS(1130),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1130),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1130),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1130),
    [anon_sym_BSLASH] = ACTIONS(1128),
    [anon_sym_LBRACE] = ACTIONS(1128),
    [anon_sym_RBRACE] = ACTIONS(1128),
    [aux_sym_trait_declaration_token1] = ACTIONS(1130),
    [aux_sym_interface_declaration_token1] = ACTIONS(1130),
    [aux_sym_enum_declaration_token1] = ACTIONS(1130),
    [aux_sym_enum_case_token1] = ACTIONS(1130),
    [aux_sym_class_declaration_token1] = ACTIONS(1130),
    [aux_sym_final_modifier_token1] = ACTIONS(1130),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1130),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1130),
    [sym_var_modifier] = ACTIONS(1130),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1130),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1130),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1130),
    [anon_sym_LPAREN2] = ACTIONS(1128),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1130),
    [aux_sym_cast_type_token1] = ACTIONS(1130),
    [aux_sym_echo_statement_token1] = ACTIONS(1130),
    [aux_sym_exit_statement_token1] = ACTIONS(1130),
    [anon_sym_unset] = ACTIONS(1130),
    [aux_sym_declare_statement_token1] = ACTIONS(1130),
    [aux_sym_declare_statement_token2] = ACTIONS(1130),
    [sym_float] = ACTIONS(1130),
    [aux_sym_try_statement_token1] = ACTIONS(1130),
    [aux_sym_goto_statement_token1] = ACTIONS(1130),
    [aux_sym_continue_statement_token1] = ACTIONS(1130),
    [aux_sym_break_statement_token1] = ACTIONS(1130),
    [sym_integer] = ACTIONS(1130),
    [aux_sym_return_statement_token1] = ACTIONS(1130),
    [aux_sym_throw_expression_token1] = ACTIONS(1130),
    [aux_sym_while_statement_token1] = ACTIONS(1130),
    [aux_sym_while_statement_token2] = ACTIONS(1130),
    [aux_sym_do_statement_token1] = ACTIONS(1130),
    [aux_sym_for_statement_token1] = ACTIONS(1130),
    [aux_sym_for_statement_token2] = ACTIONS(1130),
    [aux_sym_foreach_statement_token1] = ACTIONS(1130),
    [aux_sym_foreach_statement_token2] = ACTIONS(1130),
    [aux_sym_if_statement_token1] = ACTIONS(1130),
    [aux_sym_if_statement_token2] = ACTIONS(1130),
    [aux_sym_else_if_clause_token1] = ACTIONS(1130),
    [aux_sym_else_clause_token1] = ACTIONS(1130),
    [aux_sym_match_expression_token1] = ACTIONS(1130),
    [aux_sym_match_default_expression_token1] = ACTIONS(1130),
    [aux_sym_switch_statement_token1] = ACTIONS(1130),
    [aux_sym_switch_block_token1] = ACTIONS(1130),
    [anon_sym_PLUS] = ACTIONS(1130),
    [anon_sym_DASH] = ACTIONS(1130),
    [anon_sym_TILDE] = ACTIONS(1128),
    [anon_sym_BANG] = ACTIONS(1128),
    [anon_sym_AT] = ACTIONS(1128),
    [aux_sym_clone_expression_token1] = ACTIONS(1130),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1130),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1130),
    [anon_sym_DASH_DASH] = ACTIONS(1128),
    [anon_sym_PLUS_PLUS] = ACTIONS(1128),
    [aux_sym__list_destructing_token1] = ACTIONS(1130),
    [anon_sym_LBRACK] = ACTIONS(1128),
    [anon_sym_self] = ACTIONS(1130),
    [anon_sym_parent] = ACTIONS(1130),
    [aux_sym__argument_name_token1] = ACTIONS(1130),
    [aux_sym__argument_name_token2] = ACTIONS(1130),
    [anon_sym_POUND_LBRACK] = ACTIONS(1128),
    [aux_sym_encapsed_string_token1] = ACTIONS(1128),
    [anon_sym_DQUOTE] = ACTIONS(1128),
    [aux_sym_string_token1] = ACTIONS(1128),
    [anon_sym_SQUOTE] = ACTIONS(1128),
    [anon_sym_LT_LT_LT] = ACTIONS(1128),
    [anon_sym_BQUOTE] = ACTIONS(1128),
    [anon_sym_DOLLAR] = ACTIONS(1128),
    [aux_sym_yield_expression_token1] = ACTIONS(1130),
    [aux_sym_include_expression_token1] = ACTIONS(1130),
    [aux_sym_include_once_expression_token1] = ACTIONS(1130),
    [aux_sym_require_expression_token1] = ACTIONS(1130),
    [aux_sym_require_once_expression_token1] = ACTIONS(1130),
    [sym_comment] = ACTIONS(5),
  },
  [475] = {
    [sym_text_interpolation] = STATE(475),
    [ts_builtin_sym_end] = ACTIONS(1132),
    [sym_name] = ACTIONS(1134),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1132),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1134),
    [aux_sym_global_declaration_token1] = ACTIONS(1134),
    [aux_sym_namespace_definition_token1] = ACTIONS(1134),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1134),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1134),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1134),
    [anon_sym_BSLASH] = ACTIONS(1132),
    [anon_sym_LBRACE] = ACTIONS(1132),
    [anon_sym_RBRACE] = ACTIONS(1132),
    [aux_sym_trait_declaration_token1] = ACTIONS(1134),
    [aux_sym_interface_declaration_token1] = ACTIONS(1134),
    [aux_sym_enum_declaration_token1] = ACTIONS(1134),
    [aux_sym_enum_case_token1] = ACTIONS(1134),
    [aux_sym_class_declaration_token1] = ACTIONS(1134),
    [aux_sym_final_modifier_token1] = ACTIONS(1134),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1134),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1134),
    [sym_var_modifier] = ACTIONS(1134),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1134),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1134),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1134),
    [anon_sym_LPAREN2] = ACTIONS(1132),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1134),
    [aux_sym_cast_type_token1] = ACTIONS(1134),
    [aux_sym_echo_statement_token1] = ACTIONS(1134),
    [aux_sym_exit_statement_token1] = ACTIONS(1134),
    [anon_sym_unset] = ACTIONS(1134),
    [aux_sym_declare_statement_token1] = ACTIONS(1134),
    [aux_sym_declare_statement_token2] = ACTIONS(1134),
    [sym_float] = ACTIONS(1134),
    [aux_sym_try_statement_token1] = ACTIONS(1134),
    [aux_sym_goto_statement_token1] = ACTIONS(1134),
    [aux_sym_continue_statement_token1] = ACTIONS(1134),
    [aux_sym_break_statement_token1] = ACTIONS(1134),
    [sym_integer] = ACTIONS(1134),
    [aux_sym_return_statement_token1] = ACTIONS(1134),
    [aux_sym_throw_expression_token1] = ACTIONS(1134),
    [aux_sym_while_statement_token1] = ACTIONS(1134),
    [aux_sym_while_statement_token2] = ACTIONS(1134),
    [aux_sym_do_statement_token1] = ACTIONS(1134),
    [aux_sym_for_statement_token1] = ACTIONS(1134),
    [aux_sym_for_statement_token2] = ACTIONS(1134),
    [aux_sym_foreach_statement_token1] = ACTIONS(1134),
    [aux_sym_foreach_statement_token2] = ACTIONS(1134),
    [aux_sym_if_statement_token1] = ACTIONS(1134),
    [aux_sym_if_statement_token2] = ACTIONS(1134),
    [aux_sym_else_if_clause_token1] = ACTIONS(1134),
    [aux_sym_else_clause_token1] = ACTIONS(1134),
    [aux_sym_match_expression_token1] = ACTIONS(1134),
    [aux_sym_match_default_expression_token1] = ACTIONS(1134),
    [aux_sym_switch_statement_token1] = ACTIONS(1134),
    [aux_sym_switch_block_token1] = ACTIONS(1134),
    [anon_sym_PLUS] = ACTIONS(1134),
    [anon_sym_DASH] = ACTIONS(1134),
    [anon_sym_TILDE] = ACTIONS(1132),
    [anon_sym_BANG] = ACTIONS(1132),
    [anon_sym_AT] = ACTIONS(1132),
    [aux_sym_clone_expression_token1] = ACTIONS(1134),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1134),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1134),
    [anon_sym_DASH_DASH] = ACTIONS(1132),
    [anon_sym_PLUS_PLUS] = ACTIONS(1132),
    [aux_sym__list_destructing_token1] = ACTIONS(1134),
    [anon_sym_LBRACK] = ACTIONS(1132),
    [anon_sym_self] = ACTIONS(1134),
    [anon_sym_parent] = ACTIONS(1134),
    [aux_sym__argument_name_token1] = ACTIONS(1134),
    [aux_sym__argument_name_token2] = ACTIONS(1134),
    [anon_sym_POUND_LBRACK] = ACTIONS(1132),
    [aux_sym_encapsed_string_token1] = ACTIONS(1132),
    [anon_sym_DQUOTE] = ACTIONS(1132),
    [aux_sym_string_token1] = ACTIONS(1132),
    [anon_sym_SQUOTE] = ACTIONS(1132),
    [anon_sym_LT_LT_LT] = ACTIONS(1132),
    [anon_sym_BQUOTE] = ACTIONS(1132),
    [anon_sym_DOLLAR] = ACTIONS(1132),
    [aux_sym_yield_expression_token1] = ACTIONS(1134),
    [aux_sym_include_expression_token1] = ACTIONS(1134),
    [aux_sym_include_once_expression_token1] = ACTIONS(1134),
    [aux_sym_require_expression_token1] = ACTIONS(1134),
    [aux_sym_require_once_expression_token1] = ACTIONS(1134),
    [sym_comment] = ACTIONS(5),
  },
  [476] = {
    [sym_text_interpolation] = STATE(476),
    [ts_builtin_sym_end] = ACTIONS(1136),
    [sym_name] = ACTIONS(1138),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1136),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1138),
    [aux_sym_global_declaration_token1] = ACTIONS(1138),
    [aux_sym_namespace_definition_token1] = ACTIONS(1138),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1138),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1138),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1138),
    [anon_sym_BSLASH] = ACTIONS(1136),
    [anon_sym_LBRACE] = ACTIONS(1136),
    [anon_sym_RBRACE] = ACTIONS(1136),
    [aux_sym_trait_declaration_token1] = ACTIONS(1138),
    [aux_sym_interface_declaration_token1] = ACTIONS(1138),
    [aux_sym_enum_declaration_token1] = ACTIONS(1138),
    [aux_sym_enum_case_token1] = ACTIONS(1138),
    [aux_sym_class_declaration_token1] = ACTIONS(1138),
    [aux_sym_final_modifier_token1] = ACTIONS(1138),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1138),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1138),
    [sym_var_modifier] = ACTIONS(1138),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1138),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1138),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1138),
    [anon_sym_LPAREN2] = ACTIONS(1136),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1138),
    [aux_sym_cast_type_token1] = ACTIONS(1138),
    [aux_sym_echo_statement_token1] = ACTIONS(1138),
    [aux_sym_exit_statement_token1] = ACTIONS(1138),
    [anon_sym_unset] = ACTIONS(1138),
    [aux_sym_declare_statement_token1] = ACTIONS(1138),
    [aux_sym_declare_statement_token2] = ACTIONS(1138),
    [sym_float] = ACTIONS(1138),
    [aux_sym_try_statement_token1] = ACTIONS(1138),
    [aux_sym_goto_statement_token1] = ACTIONS(1138),
    [aux_sym_continue_statement_token1] = ACTIONS(1138),
    [aux_sym_break_statement_token1] = ACTIONS(1138),
    [sym_integer] = ACTIONS(1138),
    [aux_sym_return_statement_token1] = ACTIONS(1138),
    [aux_sym_throw_expression_token1] = ACTIONS(1138),
    [aux_sym_while_statement_token1] = ACTIONS(1138),
    [aux_sym_while_statement_token2] = ACTIONS(1138),
    [aux_sym_do_statement_token1] = ACTIONS(1138),
    [aux_sym_for_statement_token1] = ACTIONS(1138),
    [aux_sym_for_statement_token2] = ACTIONS(1138),
    [aux_sym_foreach_statement_token1] = ACTIONS(1138),
    [aux_sym_foreach_statement_token2] = ACTIONS(1138),
    [aux_sym_if_statement_token1] = ACTIONS(1138),
    [aux_sym_if_statement_token2] = ACTIONS(1138),
    [aux_sym_else_if_clause_token1] = ACTIONS(1138),
    [aux_sym_else_clause_token1] = ACTIONS(1138),
    [aux_sym_match_expression_token1] = ACTIONS(1138),
    [aux_sym_match_default_expression_token1] = ACTIONS(1138),
    [aux_sym_switch_statement_token1] = ACTIONS(1138),
    [aux_sym_switch_block_token1] = ACTIONS(1138),
    [anon_sym_PLUS] = ACTIONS(1138),
    [anon_sym_DASH] = ACTIONS(1138),
    [anon_sym_TILDE] = ACTIONS(1136),
    [anon_sym_BANG] = ACTIONS(1136),
    [anon_sym_AT] = ACTIONS(1136),
    [aux_sym_clone_expression_token1] = ACTIONS(1138),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1138),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1138),
    [anon_sym_DASH_DASH] = ACTIONS(1136),
    [anon_sym_PLUS_PLUS] = ACTIONS(1136),
    [aux_sym__list_destructing_token1] = ACTIONS(1138),
    [anon_sym_LBRACK] = ACTIONS(1136),
    [anon_sym_self] = ACTIONS(1138),
    [anon_sym_parent] = ACTIONS(1138),
    [aux_sym__argument_name_token1] = ACTIONS(1138),
    [aux_sym__argument_name_token2] = ACTIONS(1138),
    [anon_sym_POUND_LBRACK] = ACTIONS(1136),
    [aux_sym_encapsed_string_token1] = ACTIONS(1136),
    [anon_sym_DQUOTE] = ACTIONS(1136),
    [aux_sym_string_token1] = ACTIONS(1136),
    [anon_sym_SQUOTE] = ACTIONS(1136),
    [anon_sym_LT_LT_LT] = ACTIONS(1136),
    [anon_sym_BQUOTE] = ACTIONS(1136),
    [anon_sym_DOLLAR] = ACTIONS(1136),
    [aux_sym_yield_expression_token1] = ACTIONS(1138),
    [aux_sym_include_expression_token1] = ACTIONS(1138),
    [aux_sym_include_once_expression_token1] = ACTIONS(1138),
    [aux_sym_require_expression_token1] = ACTIONS(1138),
    [aux_sym_require_once_expression_token1] = ACTIONS(1138),
    [sym_comment] = ACTIONS(5),
  },
  [477] = {
    [sym_text_interpolation] = STATE(477),
    [ts_builtin_sym_end] = ACTIONS(1140),
    [sym_name] = ACTIONS(1142),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1140),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1142),
    [aux_sym_global_declaration_token1] = ACTIONS(1142),
    [aux_sym_namespace_definition_token1] = ACTIONS(1142),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1142),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1142),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1142),
    [anon_sym_BSLASH] = ACTIONS(1140),
    [anon_sym_LBRACE] = ACTIONS(1140),
    [anon_sym_RBRACE] = ACTIONS(1140),
    [aux_sym_trait_declaration_token1] = ACTIONS(1142),
    [aux_sym_interface_declaration_token1] = ACTIONS(1142),
    [aux_sym_enum_declaration_token1] = ACTIONS(1142),
    [aux_sym_enum_case_token1] = ACTIONS(1142),
    [aux_sym_class_declaration_token1] = ACTIONS(1142),
    [aux_sym_final_modifier_token1] = ACTIONS(1142),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1142),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1142),
    [sym_var_modifier] = ACTIONS(1142),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1142),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1142),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1142),
    [anon_sym_LPAREN2] = ACTIONS(1140),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1142),
    [aux_sym_cast_type_token1] = ACTIONS(1142),
    [aux_sym_echo_statement_token1] = ACTIONS(1142),
    [aux_sym_exit_statement_token1] = ACTIONS(1142),
    [anon_sym_unset] = ACTIONS(1142),
    [aux_sym_declare_statement_token1] = ACTIONS(1142),
    [aux_sym_declare_statement_token2] = ACTIONS(1142),
    [sym_float] = ACTIONS(1142),
    [aux_sym_try_statement_token1] = ACTIONS(1142),
    [aux_sym_goto_statement_token1] = ACTIONS(1142),
    [aux_sym_continue_statement_token1] = ACTIONS(1142),
    [aux_sym_break_statement_token1] = ACTIONS(1142),
    [sym_integer] = ACTIONS(1142),
    [aux_sym_return_statement_token1] = ACTIONS(1142),
    [aux_sym_throw_expression_token1] = ACTIONS(1142),
    [aux_sym_while_statement_token1] = ACTIONS(1142),
    [aux_sym_while_statement_token2] = ACTIONS(1142),
    [aux_sym_do_statement_token1] = ACTIONS(1142),
    [aux_sym_for_statement_token1] = ACTIONS(1142),
    [aux_sym_for_statement_token2] = ACTIONS(1142),
    [aux_sym_foreach_statement_token1] = ACTIONS(1142),
    [aux_sym_foreach_statement_token2] = ACTIONS(1142),
    [aux_sym_if_statement_token1] = ACTIONS(1142),
    [aux_sym_if_statement_token2] = ACTIONS(1142),
    [aux_sym_else_if_clause_token1] = ACTIONS(1142),
    [aux_sym_else_clause_token1] = ACTIONS(1142),
    [aux_sym_match_expression_token1] = ACTIONS(1142),
    [aux_sym_match_default_expression_token1] = ACTIONS(1142),
    [aux_sym_switch_statement_token1] = ACTIONS(1142),
    [aux_sym_switch_block_token1] = ACTIONS(1142),
    [anon_sym_PLUS] = ACTIONS(1142),
    [anon_sym_DASH] = ACTIONS(1142),
    [anon_sym_TILDE] = ACTIONS(1140),
    [anon_sym_BANG] = ACTIONS(1140),
    [anon_sym_AT] = ACTIONS(1140),
    [aux_sym_clone_expression_token1] = ACTIONS(1142),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1142),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1142),
    [anon_sym_DASH_DASH] = ACTIONS(1140),
    [anon_sym_PLUS_PLUS] = ACTIONS(1140),
    [aux_sym__list_destructing_token1] = ACTIONS(1142),
    [anon_sym_LBRACK] = ACTIONS(1140),
    [anon_sym_self] = ACTIONS(1142),
    [anon_sym_parent] = ACTIONS(1142),
    [aux_sym__argument_name_token1] = ACTIONS(1142),
    [aux_sym__argument_name_token2] = ACTIONS(1142),
    [anon_sym_POUND_LBRACK] = ACTIONS(1140),
    [aux_sym_encapsed_string_token1] = ACTIONS(1140),
    [anon_sym_DQUOTE] = ACTIONS(1140),
    [aux_sym_string_token1] = ACTIONS(1140),
    [anon_sym_SQUOTE] = ACTIONS(1140),
    [anon_sym_LT_LT_LT] = ACTIONS(1140),
    [anon_sym_BQUOTE] = ACTIONS(1140),
    [anon_sym_DOLLAR] = ACTIONS(1140),
    [aux_sym_yield_expression_token1] = ACTIONS(1142),
    [aux_sym_include_expression_token1] = ACTIONS(1142),
    [aux_sym_include_once_expression_token1] = ACTIONS(1142),
    [aux_sym_require_expression_token1] = ACTIONS(1142),
    [aux_sym_require_once_expression_token1] = ACTIONS(1142),
    [sym_comment] = ACTIONS(5),
  },
  [478] = {
    [sym_text_interpolation] = STATE(478),
    [ts_builtin_sym_end] = ACTIONS(1144),
    [sym_name] = ACTIONS(1146),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1144),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1146),
    [aux_sym_global_declaration_token1] = ACTIONS(1146),
    [aux_sym_namespace_definition_token1] = ACTIONS(1146),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1146),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1146),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1146),
    [anon_sym_BSLASH] = ACTIONS(1144),
    [anon_sym_LBRACE] = ACTIONS(1144),
    [anon_sym_RBRACE] = ACTIONS(1144),
    [aux_sym_trait_declaration_token1] = ACTIONS(1146),
    [aux_sym_interface_declaration_token1] = ACTIONS(1146),
    [aux_sym_enum_declaration_token1] = ACTIONS(1146),
    [aux_sym_enum_case_token1] = ACTIONS(1146),
    [aux_sym_class_declaration_token1] = ACTIONS(1146),
    [aux_sym_final_modifier_token1] = ACTIONS(1146),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1146),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1146),
    [sym_var_modifier] = ACTIONS(1146),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1146),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1146),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1146),
    [anon_sym_LPAREN2] = ACTIONS(1144),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1146),
    [aux_sym_cast_type_token1] = ACTIONS(1146),
    [aux_sym_echo_statement_token1] = ACTIONS(1146),
    [aux_sym_exit_statement_token1] = ACTIONS(1146),
    [anon_sym_unset] = ACTIONS(1146),
    [aux_sym_declare_statement_token1] = ACTIONS(1146),
    [aux_sym_declare_statement_token2] = ACTIONS(1146),
    [sym_float] = ACTIONS(1146),
    [aux_sym_try_statement_token1] = ACTIONS(1146),
    [aux_sym_goto_statement_token1] = ACTIONS(1146),
    [aux_sym_continue_statement_token1] = ACTIONS(1146),
    [aux_sym_break_statement_token1] = ACTIONS(1146),
    [sym_integer] = ACTIONS(1146),
    [aux_sym_return_statement_token1] = ACTIONS(1146),
    [aux_sym_throw_expression_token1] = ACTIONS(1146),
    [aux_sym_while_statement_token1] = ACTIONS(1146),
    [aux_sym_while_statement_token2] = ACTIONS(1146),
    [aux_sym_do_statement_token1] = ACTIONS(1146),
    [aux_sym_for_statement_token1] = ACTIONS(1146),
    [aux_sym_for_statement_token2] = ACTIONS(1146),
    [aux_sym_foreach_statement_token1] = ACTIONS(1146),
    [aux_sym_foreach_statement_token2] = ACTIONS(1146),
    [aux_sym_if_statement_token1] = ACTIONS(1146),
    [aux_sym_if_statement_token2] = ACTIONS(1146),
    [aux_sym_else_if_clause_token1] = ACTIONS(1146),
    [aux_sym_else_clause_token1] = ACTIONS(1146),
    [aux_sym_match_expression_token1] = ACTIONS(1146),
    [aux_sym_match_default_expression_token1] = ACTIONS(1146),
    [aux_sym_switch_statement_token1] = ACTIONS(1146),
    [aux_sym_switch_block_token1] = ACTIONS(1146),
    [anon_sym_PLUS] = ACTIONS(1146),
    [anon_sym_DASH] = ACTIONS(1146),
    [anon_sym_TILDE] = ACTIONS(1144),
    [anon_sym_BANG] = ACTIONS(1144),
    [anon_sym_AT] = ACTIONS(1144),
    [aux_sym_clone_expression_token1] = ACTIONS(1146),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1146),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1146),
    [anon_sym_DASH_DASH] = ACTIONS(1144),
    [anon_sym_PLUS_PLUS] = ACTIONS(1144),
    [aux_sym__list_destructing_token1] = ACTIONS(1146),
    [anon_sym_LBRACK] = ACTIONS(1144),
    [anon_sym_self] = ACTIONS(1146),
    [anon_sym_parent] = ACTIONS(1146),
    [aux_sym__argument_name_token1] = ACTIONS(1146),
    [aux_sym__argument_name_token2] = ACTIONS(1146),
    [anon_sym_POUND_LBRACK] = ACTIONS(1144),
    [aux_sym_encapsed_string_token1] = ACTIONS(1144),
    [anon_sym_DQUOTE] = ACTIONS(1144),
    [aux_sym_string_token1] = ACTIONS(1144),
    [anon_sym_SQUOTE] = ACTIONS(1144),
    [anon_sym_LT_LT_LT] = ACTIONS(1144),
    [anon_sym_BQUOTE] = ACTIONS(1144),
    [anon_sym_DOLLAR] = ACTIONS(1144),
    [aux_sym_yield_expression_token1] = ACTIONS(1146),
    [aux_sym_include_expression_token1] = ACTIONS(1146),
    [aux_sym_include_once_expression_token1] = ACTIONS(1146),
    [aux_sym_require_expression_token1] = ACTIONS(1146),
    [aux_sym_require_once_expression_token1] = ACTIONS(1146),
    [sym_comment] = ACTIONS(5),
  },
  [479] = {
    [sym_text_interpolation] = STATE(479),
    [ts_builtin_sym_end] = ACTIONS(1148),
    [sym_name] = ACTIONS(1150),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1148),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1150),
    [aux_sym_global_declaration_token1] = ACTIONS(1150),
    [aux_sym_namespace_definition_token1] = ACTIONS(1150),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1150),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1150),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1150),
    [anon_sym_BSLASH] = ACTIONS(1148),
    [anon_sym_LBRACE] = ACTIONS(1148),
    [anon_sym_RBRACE] = ACTIONS(1148),
    [aux_sym_trait_declaration_token1] = ACTIONS(1150),
    [aux_sym_interface_declaration_token1] = ACTIONS(1150),
    [aux_sym_enum_declaration_token1] = ACTIONS(1150),
    [aux_sym_enum_case_token1] = ACTIONS(1150),
    [aux_sym_class_declaration_token1] = ACTIONS(1150),
    [aux_sym_final_modifier_token1] = ACTIONS(1150),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1150),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1150),
    [sym_var_modifier] = ACTIONS(1150),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1150),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1150),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1150),
    [anon_sym_LPAREN2] = ACTIONS(1148),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1150),
    [aux_sym_cast_type_token1] = ACTIONS(1150),
    [aux_sym_echo_statement_token1] = ACTIONS(1150),
    [aux_sym_exit_statement_token1] = ACTIONS(1150),
    [anon_sym_unset] = ACTIONS(1150),
    [aux_sym_declare_statement_token1] = ACTIONS(1150),
    [aux_sym_declare_statement_token2] = ACTIONS(1150),
    [sym_float] = ACTIONS(1150),
    [aux_sym_try_statement_token1] = ACTIONS(1150),
    [aux_sym_goto_statement_token1] = ACTIONS(1150),
    [aux_sym_continue_statement_token1] = ACTIONS(1150),
    [aux_sym_break_statement_token1] = ACTIONS(1150),
    [sym_integer] = ACTIONS(1150),
    [aux_sym_return_statement_token1] = ACTIONS(1150),
    [aux_sym_throw_expression_token1] = ACTIONS(1150),
    [aux_sym_while_statement_token1] = ACTIONS(1150),
    [aux_sym_while_statement_token2] = ACTIONS(1150),
    [aux_sym_do_statement_token1] = ACTIONS(1150),
    [aux_sym_for_statement_token1] = ACTIONS(1150),
    [aux_sym_for_statement_token2] = ACTIONS(1150),
    [aux_sym_foreach_statement_token1] = ACTIONS(1150),
    [aux_sym_foreach_statement_token2] = ACTIONS(1150),
    [aux_sym_if_statement_token1] = ACTIONS(1150),
    [aux_sym_if_statement_token2] = ACTIONS(1150),
    [aux_sym_else_if_clause_token1] = ACTIONS(1150),
    [aux_sym_else_clause_token1] = ACTIONS(1150),
    [aux_sym_match_expression_token1] = ACTIONS(1150),
    [aux_sym_match_default_expression_token1] = ACTIONS(1150),
    [aux_sym_switch_statement_token1] = ACTIONS(1150),
    [aux_sym_switch_block_token1] = ACTIONS(1150),
    [anon_sym_PLUS] = ACTIONS(1150),
    [anon_sym_DASH] = ACTIONS(1150),
    [anon_sym_TILDE] = ACTIONS(1148),
    [anon_sym_BANG] = ACTIONS(1148),
    [anon_sym_AT] = ACTIONS(1148),
    [aux_sym_clone_expression_token1] = ACTIONS(1150),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1150),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1150),
    [anon_sym_DASH_DASH] = ACTIONS(1148),
    [anon_sym_PLUS_PLUS] = ACTIONS(1148),
    [aux_sym__list_destructing_token1] = ACTIONS(1150),
    [anon_sym_LBRACK] = ACTIONS(1148),
    [anon_sym_self] = ACTIONS(1150),
    [anon_sym_parent] = ACTIONS(1150),
    [aux_sym__argument_name_token1] = ACTIONS(1150),
    [aux_sym__argument_name_token2] = ACTIONS(1150),
    [anon_sym_POUND_LBRACK] = ACTIONS(1148),
    [aux_sym_encapsed_string_token1] = ACTIONS(1148),
    [anon_sym_DQUOTE] = ACTIONS(1148),
    [aux_sym_string_token1] = ACTIONS(1148),
    [anon_sym_SQUOTE] = ACTIONS(1148),
    [anon_sym_LT_LT_LT] = ACTIONS(1148),
    [anon_sym_BQUOTE] = ACTIONS(1148),
    [anon_sym_DOLLAR] = ACTIONS(1148),
    [aux_sym_yield_expression_token1] = ACTIONS(1150),
    [aux_sym_include_expression_token1] = ACTIONS(1150),
    [aux_sym_include_once_expression_token1] = ACTIONS(1150),
    [aux_sym_require_expression_token1] = ACTIONS(1150),
    [aux_sym_require_once_expression_token1] = ACTIONS(1150),
    [sym_comment] = ACTIONS(5),
  },
  [480] = {
    [sym_text_interpolation] = STATE(480),
    [ts_builtin_sym_end] = ACTIONS(1152),
    [sym_name] = ACTIONS(1154),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1152),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1154),
    [aux_sym_global_declaration_token1] = ACTIONS(1154),
    [aux_sym_namespace_definition_token1] = ACTIONS(1154),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1154),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1154),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1154),
    [anon_sym_BSLASH] = ACTIONS(1152),
    [anon_sym_LBRACE] = ACTIONS(1152),
    [anon_sym_RBRACE] = ACTIONS(1152),
    [aux_sym_trait_declaration_token1] = ACTIONS(1154),
    [aux_sym_interface_declaration_token1] = ACTIONS(1154),
    [aux_sym_enum_declaration_token1] = ACTIONS(1154),
    [aux_sym_enum_case_token1] = ACTIONS(1154),
    [aux_sym_class_declaration_token1] = ACTIONS(1154),
    [aux_sym_final_modifier_token1] = ACTIONS(1154),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1154),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1154),
    [sym_var_modifier] = ACTIONS(1154),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1154),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1154),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1154),
    [anon_sym_LPAREN2] = ACTIONS(1152),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1154),
    [aux_sym_cast_type_token1] = ACTIONS(1154),
    [aux_sym_echo_statement_token1] = ACTIONS(1154),
    [aux_sym_exit_statement_token1] = ACTIONS(1154),
    [anon_sym_unset] = ACTIONS(1154),
    [aux_sym_declare_statement_token1] = ACTIONS(1154),
    [aux_sym_declare_statement_token2] = ACTIONS(1154),
    [sym_float] = ACTIONS(1154),
    [aux_sym_try_statement_token1] = ACTIONS(1154),
    [aux_sym_goto_statement_token1] = ACTIONS(1154),
    [aux_sym_continue_statement_token1] = ACTIONS(1154),
    [aux_sym_break_statement_token1] = ACTIONS(1154),
    [sym_integer] = ACTIONS(1154),
    [aux_sym_return_statement_token1] = ACTIONS(1154),
    [aux_sym_throw_expression_token1] = ACTIONS(1154),
    [aux_sym_while_statement_token1] = ACTIONS(1154),
    [aux_sym_while_statement_token2] = ACTIONS(1154),
    [aux_sym_do_statement_token1] = ACTIONS(1154),
    [aux_sym_for_statement_token1] = ACTIONS(1154),
    [aux_sym_for_statement_token2] = ACTIONS(1154),
    [aux_sym_foreach_statement_token1] = ACTIONS(1154),
    [aux_sym_foreach_statement_token2] = ACTIONS(1154),
    [aux_sym_if_statement_token1] = ACTIONS(1154),
    [aux_sym_if_statement_token2] = ACTIONS(1154),
    [aux_sym_else_if_clause_token1] = ACTIONS(1154),
    [aux_sym_else_clause_token1] = ACTIONS(1154),
    [aux_sym_match_expression_token1] = ACTIONS(1154),
    [aux_sym_match_default_expression_token1] = ACTIONS(1154),
    [aux_sym_switch_statement_token1] = ACTIONS(1154),
    [aux_sym_switch_block_token1] = ACTIONS(1154),
    [anon_sym_PLUS] = ACTIONS(1154),
    [anon_sym_DASH] = ACTIONS(1154),
    [anon_sym_TILDE] = ACTIONS(1152),
    [anon_sym_BANG] = ACTIONS(1152),
    [anon_sym_AT] = ACTIONS(1152),
    [aux_sym_clone_expression_token1] = ACTIONS(1154),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1154),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1154),
    [anon_sym_DASH_DASH] = ACTIONS(1152),
    [anon_sym_PLUS_PLUS] = ACTIONS(1152),
    [aux_sym__list_destructing_token1] = ACTIONS(1154),
    [anon_sym_LBRACK] = ACTIONS(1152),
    [anon_sym_self] = ACTIONS(1154),
    [anon_sym_parent] = ACTIONS(1154),
    [aux_sym__argument_name_token1] = ACTIONS(1154),
    [aux_sym__argument_name_token2] = ACTIONS(1154),
    [anon_sym_POUND_LBRACK] = ACTIONS(1152),
    [aux_sym_encapsed_string_token1] = ACTIONS(1152),
    [anon_sym_DQUOTE] = ACTIONS(1152),
    [aux_sym_string_token1] = ACTIONS(1152),
    [anon_sym_SQUOTE] = ACTIONS(1152),
    [anon_sym_LT_LT_LT] = ACTIONS(1152),
    [anon_sym_BQUOTE] = ACTIONS(1152),
    [anon_sym_DOLLAR] = ACTIONS(1152),
    [aux_sym_yield_expression_token1] = ACTIONS(1154),
    [aux_sym_include_expression_token1] = ACTIONS(1154),
    [aux_sym_include_once_expression_token1] = ACTIONS(1154),
    [aux_sym_require_expression_token1] = ACTIONS(1154),
    [aux_sym_require_once_expression_token1] = ACTIONS(1154),
    [sym_comment] = ACTIONS(5),
  },
  [481] = {
    [sym_text_interpolation] = STATE(481),
    [ts_builtin_sym_end] = ACTIONS(1156),
    [sym_name] = ACTIONS(1158),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1156),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1158),
    [aux_sym_global_declaration_token1] = ACTIONS(1158),
    [aux_sym_namespace_definition_token1] = ACTIONS(1158),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1158),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1158),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1158),
    [anon_sym_BSLASH] = ACTIONS(1156),
    [anon_sym_LBRACE] = ACTIONS(1156),
    [anon_sym_RBRACE] = ACTIONS(1156),
    [aux_sym_trait_declaration_token1] = ACTIONS(1158),
    [aux_sym_interface_declaration_token1] = ACTIONS(1158),
    [aux_sym_enum_declaration_token1] = ACTIONS(1158),
    [aux_sym_enum_case_token1] = ACTIONS(1158),
    [aux_sym_class_declaration_token1] = ACTIONS(1158),
    [aux_sym_final_modifier_token1] = ACTIONS(1158),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1158),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1158),
    [sym_var_modifier] = ACTIONS(1158),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1158),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1158),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1158),
    [anon_sym_LPAREN2] = ACTIONS(1156),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1158),
    [aux_sym_cast_type_token1] = ACTIONS(1158),
    [aux_sym_echo_statement_token1] = ACTIONS(1158),
    [aux_sym_exit_statement_token1] = ACTIONS(1158),
    [anon_sym_unset] = ACTIONS(1158),
    [aux_sym_declare_statement_token1] = ACTIONS(1158),
    [aux_sym_declare_statement_token2] = ACTIONS(1158),
    [sym_float] = ACTIONS(1158),
    [aux_sym_try_statement_token1] = ACTIONS(1158),
    [aux_sym_goto_statement_token1] = ACTIONS(1158),
    [aux_sym_continue_statement_token1] = ACTIONS(1158),
    [aux_sym_break_statement_token1] = ACTIONS(1158),
    [sym_integer] = ACTIONS(1158),
    [aux_sym_return_statement_token1] = ACTIONS(1158),
    [aux_sym_throw_expression_token1] = ACTIONS(1158),
    [aux_sym_while_statement_token1] = ACTIONS(1158),
    [aux_sym_while_statement_token2] = ACTIONS(1158),
    [aux_sym_do_statement_token1] = ACTIONS(1158),
    [aux_sym_for_statement_token1] = ACTIONS(1158),
    [aux_sym_for_statement_token2] = ACTIONS(1158),
    [aux_sym_foreach_statement_token1] = ACTIONS(1158),
    [aux_sym_foreach_statement_token2] = ACTIONS(1158),
    [aux_sym_if_statement_token1] = ACTIONS(1158),
    [aux_sym_if_statement_token2] = ACTIONS(1158),
    [aux_sym_else_if_clause_token1] = ACTIONS(1158),
    [aux_sym_else_clause_token1] = ACTIONS(1158),
    [aux_sym_match_expression_token1] = ACTIONS(1158),
    [aux_sym_match_default_expression_token1] = ACTIONS(1158),
    [aux_sym_switch_statement_token1] = ACTIONS(1158),
    [aux_sym_switch_block_token1] = ACTIONS(1158),
    [anon_sym_PLUS] = ACTIONS(1158),
    [anon_sym_DASH] = ACTIONS(1158),
    [anon_sym_TILDE] = ACTIONS(1156),
    [anon_sym_BANG] = ACTIONS(1156),
    [anon_sym_AT] = ACTIONS(1156),
    [aux_sym_clone_expression_token1] = ACTIONS(1158),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1158),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1158),
    [anon_sym_DASH_DASH] = ACTIONS(1156),
    [anon_sym_PLUS_PLUS] = ACTIONS(1156),
    [aux_sym__list_destructing_token1] = ACTIONS(1158),
    [anon_sym_LBRACK] = ACTIONS(1156),
    [anon_sym_self] = ACTIONS(1158),
    [anon_sym_parent] = ACTIONS(1158),
    [aux_sym__argument_name_token1] = ACTIONS(1158),
    [aux_sym__argument_name_token2] = ACTIONS(1158),
    [anon_sym_POUND_LBRACK] = ACTIONS(1156),
    [aux_sym_encapsed_string_token1] = ACTIONS(1156),
    [anon_sym_DQUOTE] = ACTIONS(1156),
    [aux_sym_string_token1] = ACTIONS(1156),
    [anon_sym_SQUOTE] = ACTIONS(1156),
    [anon_sym_LT_LT_LT] = ACTIONS(1156),
    [anon_sym_BQUOTE] = ACTIONS(1156),
    [anon_sym_DOLLAR] = ACTIONS(1156),
    [aux_sym_yield_expression_token1] = ACTIONS(1158),
    [aux_sym_include_expression_token1] = ACTIONS(1158),
    [aux_sym_include_once_expression_token1] = ACTIONS(1158),
    [aux_sym_require_expression_token1] = ACTIONS(1158),
    [aux_sym_require_once_expression_token1] = ACTIONS(1158),
    [sym_comment] = ACTIONS(5),
  },
  [482] = {
    [sym_text_interpolation] = STATE(482),
    [ts_builtin_sym_end] = ACTIONS(1160),
    [sym_name] = ACTIONS(1162),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1160),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1162),
    [aux_sym_global_declaration_token1] = ACTIONS(1162),
    [aux_sym_namespace_definition_token1] = ACTIONS(1162),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1162),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1162),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1162),
    [anon_sym_BSLASH] = ACTIONS(1160),
    [anon_sym_LBRACE] = ACTIONS(1160),
    [anon_sym_RBRACE] = ACTIONS(1160),
    [aux_sym_trait_declaration_token1] = ACTIONS(1162),
    [aux_sym_interface_declaration_token1] = ACTIONS(1162),
    [aux_sym_enum_declaration_token1] = ACTIONS(1162),
    [aux_sym_enum_case_token1] = ACTIONS(1162),
    [aux_sym_class_declaration_token1] = ACTIONS(1162),
    [aux_sym_final_modifier_token1] = ACTIONS(1162),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1162),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1162),
    [sym_var_modifier] = ACTIONS(1162),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1162),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1162),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1162),
    [anon_sym_LPAREN2] = ACTIONS(1160),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1162),
    [aux_sym_cast_type_token1] = ACTIONS(1162),
    [aux_sym_echo_statement_token1] = ACTIONS(1162),
    [aux_sym_exit_statement_token1] = ACTIONS(1162),
    [anon_sym_unset] = ACTIONS(1162),
    [aux_sym_declare_statement_token1] = ACTIONS(1162),
    [aux_sym_declare_statement_token2] = ACTIONS(1162),
    [sym_float] = ACTIONS(1162),
    [aux_sym_try_statement_token1] = ACTIONS(1162),
    [aux_sym_goto_statement_token1] = ACTIONS(1162),
    [aux_sym_continue_statement_token1] = ACTIONS(1162),
    [aux_sym_break_statement_token1] = ACTIONS(1162),
    [sym_integer] = ACTIONS(1162),
    [aux_sym_return_statement_token1] = ACTIONS(1162),
    [aux_sym_throw_expression_token1] = ACTIONS(1162),
    [aux_sym_while_statement_token1] = ACTIONS(1162),
    [aux_sym_while_statement_token2] = ACTIONS(1162),
    [aux_sym_do_statement_token1] = ACTIONS(1162),
    [aux_sym_for_statement_token1] = ACTIONS(1162),
    [aux_sym_for_statement_token2] = ACTIONS(1162),
    [aux_sym_foreach_statement_token1] = ACTIONS(1162),
    [aux_sym_foreach_statement_token2] = ACTIONS(1162),
    [aux_sym_if_statement_token1] = ACTIONS(1162),
    [aux_sym_if_statement_token2] = ACTIONS(1162),
    [aux_sym_else_if_clause_token1] = ACTIONS(1162),
    [aux_sym_else_clause_token1] = ACTIONS(1162),
    [aux_sym_match_expression_token1] = ACTIONS(1162),
    [aux_sym_match_default_expression_token1] = ACTIONS(1162),
    [aux_sym_switch_statement_token1] = ACTIONS(1162),
    [aux_sym_switch_block_token1] = ACTIONS(1162),
    [anon_sym_PLUS] = ACTIONS(1162),
    [anon_sym_DASH] = ACTIONS(1162),
    [anon_sym_TILDE] = ACTIONS(1160),
    [anon_sym_BANG] = ACTIONS(1160),
    [anon_sym_AT] = ACTIONS(1160),
    [aux_sym_clone_expression_token1] = ACTIONS(1162),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1162),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1162),
    [anon_sym_DASH_DASH] = ACTIONS(1160),
    [anon_sym_PLUS_PLUS] = ACTIONS(1160),
    [aux_sym__list_destructing_token1] = ACTIONS(1162),
    [anon_sym_LBRACK] = ACTIONS(1160),
    [anon_sym_self] = ACTIONS(1162),
    [anon_sym_parent] = ACTIONS(1162),
    [aux_sym__argument_name_token1] = ACTIONS(1162),
    [aux_sym__argument_name_token2] = ACTIONS(1162),
    [anon_sym_POUND_LBRACK] = ACTIONS(1160),
    [aux_sym_encapsed_string_token1] = ACTIONS(1160),
    [anon_sym_DQUOTE] = ACTIONS(1160),
    [aux_sym_string_token1] = ACTIONS(1160),
    [anon_sym_SQUOTE] = ACTIONS(1160),
    [anon_sym_LT_LT_LT] = ACTIONS(1160),
    [anon_sym_BQUOTE] = ACTIONS(1160),
    [anon_sym_DOLLAR] = ACTIONS(1160),
    [aux_sym_yield_expression_token1] = ACTIONS(1162),
    [aux_sym_include_expression_token1] = ACTIONS(1162),
    [aux_sym_include_once_expression_token1] = ACTIONS(1162),
    [aux_sym_require_expression_token1] = ACTIONS(1162),
    [aux_sym_require_once_expression_token1] = ACTIONS(1162),
    [sym_comment] = ACTIONS(5),
  },
  [483] = {
    [sym_text_interpolation] = STATE(483),
    [ts_builtin_sym_end] = ACTIONS(1164),
    [sym_name] = ACTIONS(1166),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1164),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1166),
    [aux_sym_global_declaration_token1] = ACTIONS(1166),
    [aux_sym_namespace_definition_token1] = ACTIONS(1166),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1166),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1166),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1166),
    [anon_sym_BSLASH] = ACTIONS(1164),
    [anon_sym_LBRACE] = ACTIONS(1164),
    [anon_sym_RBRACE] = ACTIONS(1164),
    [aux_sym_trait_declaration_token1] = ACTIONS(1166),
    [aux_sym_interface_declaration_token1] = ACTIONS(1166),
    [aux_sym_enum_declaration_token1] = ACTIONS(1166),
    [aux_sym_enum_case_token1] = ACTIONS(1166),
    [aux_sym_class_declaration_token1] = ACTIONS(1166),
    [aux_sym_final_modifier_token1] = ACTIONS(1166),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1166),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1166),
    [sym_var_modifier] = ACTIONS(1166),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1166),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1166),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1166),
    [anon_sym_LPAREN2] = ACTIONS(1164),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1166),
    [aux_sym_cast_type_token1] = ACTIONS(1166),
    [aux_sym_echo_statement_token1] = ACTIONS(1166),
    [aux_sym_exit_statement_token1] = ACTIONS(1166),
    [anon_sym_unset] = ACTIONS(1166),
    [aux_sym_declare_statement_token1] = ACTIONS(1166),
    [aux_sym_declare_statement_token2] = ACTIONS(1166),
    [sym_float] = ACTIONS(1166),
    [aux_sym_try_statement_token1] = ACTIONS(1166),
    [aux_sym_goto_statement_token1] = ACTIONS(1166),
    [aux_sym_continue_statement_token1] = ACTIONS(1166),
    [aux_sym_break_statement_token1] = ACTIONS(1166),
    [sym_integer] = ACTIONS(1166),
    [aux_sym_return_statement_token1] = ACTIONS(1166),
    [aux_sym_throw_expression_token1] = ACTIONS(1166),
    [aux_sym_while_statement_token1] = ACTIONS(1166),
    [aux_sym_while_statement_token2] = ACTIONS(1166),
    [aux_sym_do_statement_token1] = ACTIONS(1166),
    [aux_sym_for_statement_token1] = ACTIONS(1166),
    [aux_sym_for_statement_token2] = ACTIONS(1166),
    [aux_sym_foreach_statement_token1] = ACTIONS(1166),
    [aux_sym_foreach_statement_token2] = ACTIONS(1166),
    [aux_sym_if_statement_token1] = ACTIONS(1166),
    [aux_sym_if_statement_token2] = ACTIONS(1166),
    [aux_sym_else_if_clause_token1] = ACTIONS(1166),
    [aux_sym_else_clause_token1] = ACTIONS(1166),
    [aux_sym_match_expression_token1] = ACTIONS(1166),
    [aux_sym_match_default_expression_token1] = ACTIONS(1166),
    [aux_sym_switch_statement_token1] = ACTIONS(1166),
    [aux_sym_switch_block_token1] = ACTIONS(1166),
    [anon_sym_PLUS] = ACTIONS(1166),
    [anon_sym_DASH] = ACTIONS(1166),
    [anon_sym_TILDE] = ACTIONS(1164),
    [anon_sym_BANG] = ACTIONS(1164),
    [anon_sym_AT] = ACTIONS(1164),
    [aux_sym_clone_expression_token1] = ACTIONS(1166),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1166),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1166),
    [anon_sym_DASH_DASH] = ACTIONS(1164),
    [anon_sym_PLUS_PLUS] = ACTIONS(1164),
    [aux_sym__list_destructing_token1] = ACTIONS(1166),
    [anon_sym_LBRACK] = ACTIONS(1164),
    [anon_sym_self] = ACTIONS(1166),
    [anon_sym_parent] = ACTIONS(1166),
    [aux_sym__argument_name_token1] = ACTIONS(1166),
    [aux_sym__argument_name_token2] = ACTIONS(1166),
    [anon_sym_POUND_LBRACK] = ACTIONS(1164),
    [aux_sym_encapsed_string_token1] = ACTIONS(1164),
    [anon_sym_DQUOTE] = ACTIONS(1164),
    [aux_sym_string_token1] = ACTIONS(1164),
    [anon_sym_SQUOTE] = ACTIONS(1164),
    [anon_sym_LT_LT_LT] = ACTIONS(1164),
    [anon_sym_BQUOTE] = ACTIONS(1164),
    [anon_sym_DOLLAR] = ACTIONS(1164),
    [aux_sym_yield_expression_token1] = ACTIONS(1166),
    [aux_sym_include_expression_token1] = ACTIONS(1166),
    [aux_sym_include_once_expression_token1] = ACTIONS(1166),
    [aux_sym_require_expression_token1] = ACTIONS(1166),
    [aux_sym_require_once_expression_token1] = ACTIONS(1166),
    [sym_comment] = ACTIONS(5),
  },
  [484] = {
    [sym_text_interpolation] = STATE(484),
    [ts_builtin_sym_end] = ACTIONS(1168),
    [sym_name] = ACTIONS(1170),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1168),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1170),
    [aux_sym_global_declaration_token1] = ACTIONS(1170),
    [aux_sym_namespace_definition_token1] = ACTIONS(1170),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1170),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1170),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1170),
    [anon_sym_BSLASH] = ACTIONS(1168),
    [anon_sym_LBRACE] = ACTIONS(1168),
    [anon_sym_RBRACE] = ACTIONS(1168),
    [aux_sym_trait_declaration_token1] = ACTIONS(1170),
    [aux_sym_interface_declaration_token1] = ACTIONS(1170),
    [aux_sym_enum_declaration_token1] = ACTIONS(1170),
    [aux_sym_enum_case_token1] = ACTIONS(1170),
    [aux_sym_class_declaration_token1] = ACTIONS(1170),
    [aux_sym_final_modifier_token1] = ACTIONS(1170),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1170),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1170),
    [sym_var_modifier] = ACTIONS(1170),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1170),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1170),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1170),
    [anon_sym_LPAREN2] = ACTIONS(1168),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1170),
    [aux_sym_cast_type_token1] = ACTIONS(1170),
    [aux_sym_echo_statement_token1] = ACTIONS(1170),
    [aux_sym_exit_statement_token1] = ACTIONS(1170),
    [anon_sym_unset] = ACTIONS(1170),
    [aux_sym_declare_statement_token1] = ACTIONS(1170),
    [aux_sym_declare_statement_token2] = ACTIONS(1170),
    [sym_float] = ACTIONS(1170),
    [aux_sym_try_statement_token1] = ACTIONS(1170),
    [aux_sym_goto_statement_token1] = ACTIONS(1170),
    [aux_sym_continue_statement_token1] = ACTIONS(1170),
    [aux_sym_break_statement_token1] = ACTIONS(1170),
    [sym_integer] = ACTIONS(1170),
    [aux_sym_return_statement_token1] = ACTIONS(1170),
    [aux_sym_throw_expression_token1] = ACTIONS(1170),
    [aux_sym_while_statement_token1] = ACTIONS(1170),
    [aux_sym_while_statement_token2] = ACTIONS(1170),
    [aux_sym_do_statement_token1] = ACTIONS(1170),
    [aux_sym_for_statement_token1] = ACTIONS(1170),
    [aux_sym_for_statement_token2] = ACTIONS(1170),
    [aux_sym_foreach_statement_token1] = ACTIONS(1170),
    [aux_sym_foreach_statement_token2] = ACTIONS(1170),
    [aux_sym_if_statement_token1] = ACTIONS(1170),
    [aux_sym_if_statement_token2] = ACTIONS(1170),
    [aux_sym_else_if_clause_token1] = ACTIONS(1170),
    [aux_sym_else_clause_token1] = ACTIONS(1170),
    [aux_sym_match_expression_token1] = ACTIONS(1170),
    [aux_sym_match_default_expression_token1] = ACTIONS(1170),
    [aux_sym_switch_statement_token1] = ACTIONS(1170),
    [aux_sym_switch_block_token1] = ACTIONS(1170),
    [anon_sym_PLUS] = ACTIONS(1170),
    [anon_sym_DASH] = ACTIONS(1170),
    [anon_sym_TILDE] = ACTIONS(1168),
    [anon_sym_BANG] = ACTIONS(1168),
    [anon_sym_AT] = ACTIONS(1168),
    [aux_sym_clone_expression_token1] = ACTIONS(1170),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1170),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1170),
    [anon_sym_DASH_DASH] = ACTIONS(1168),
    [anon_sym_PLUS_PLUS] = ACTIONS(1168),
    [aux_sym__list_destructing_token1] = ACTIONS(1170),
    [anon_sym_LBRACK] = ACTIONS(1168),
    [anon_sym_self] = ACTIONS(1170),
    [anon_sym_parent] = ACTIONS(1170),
    [aux_sym__argument_name_token1] = ACTIONS(1170),
    [aux_sym__argument_name_token2] = ACTIONS(1170),
    [anon_sym_POUND_LBRACK] = ACTIONS(1168),
    [aux_sym_encapsed_string_token1] = ACTIONS(1168),
    [anon_sym_DQUOTE] = ACTIONS(1168),
    [aux_sym_string_token1] = ACTIONS(1168),
    [anon_sym_SQUOTE] = ACTIONS(1168),
    [anon_sym_LT_LT_LT] = ACTIONS(1168),
    [anon_sym_BQUOTE] = ACTIONS(1168),
    [anon_sym_DOLLAR] = ACTIONS(1168),
    [aux_sym_yield_expression_token1] = ACTIONS(1170),
    [aux_sym_include_expression_token1] = ACTIONS(1170),
    [aux_sym_include_once_expression_token1] = ACTIONS(1170),
    [aux_sym_require_expression_token1] = ACTIONS(1170),
    [aux_sym_require_once_expression_token1] = ACTIONS(1170),
    [sym_comment] = ACTIONS(5),
  },
  [485] = {
    [sym_text_interpolation] = STATE(485),
    [ts_builtin_sym_end] = ACTIONS(1172),
    [sym_name] = ACTIONS(1174),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1172),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1174),
    [aux_sym_global_declaration_token1] = ACTIONS(1174),
    [aux_sym_namespace_definition_token1] = ACTIONS(1174),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1174),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1174),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1174),
    [anon_sym_BSLASH] = ACTIONS(1172),
    [anon_sym_LBRACE] = ACTIONS(1172),
    [anon_sym_RBRACE] = ACTIONS(1172),
    [aux_sym_trait_declaration_token1] = ACTIONS(1174),
    [aux_sym_interface_declaration_token1] = ACTIONS(1174),
    [aux_sym_enum_declaration_token1] = ACTIONS(1174),
    [aux_sym_enum_case_token1] = ACTIONS(1174),
    [aux_sym_class_declaration_token1] = ACTIONS(1174),
    [aux_sym_final_modifier_token1] = ACTIONS(1174),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1174),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1174),
    [sym_var_modifier] = ACTIONS(1174),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1174),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1174),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1174),
    [anon_sym_LPAREN2] = ACTIONS(1172),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1174),
    [aux_sym_cast_type_token1] = ACTIONS(1174),
    [aux_sym_echo_statement_token1] = ACTIONS(1174),
    [aux_sym_exit_statement_token1] = ACTIONS(1174),
    [anon_sym_unset] = ACTIONS(1174),
    [aux_sym_declare_statement_token1] = ACTIONS(1174),
    [aux_sym_declare_statement_token2] = ACTIONS(1174),
    [sym_float] = ACTIONS(1174),
    [aux_sym_try_statement_token1] = ACTIONS(1174),
    [aux_sym_goto_statement_token1] = ACTIONS(1174),
    [aux_sym_continue_statement_token1] = ACTIONS(1174),
    [aux_sym_break_statement_token1] = ACTIONS(1174),
    [sym_integer] = ACTIONS(1174),
    [aux_sym_return_statement_token1] = ACTIONS(1174),
    [aux_sym_throw_expression_token1] = ACTIONS(1174),
    [aux_sym_while_statement_token1] = ACTIONS(1174),
    [aux_sym_while_statement_token2] = ACTIONS(1174),
    [aux_sym_do_statement_token1] = ACTIONS(1174),
    [aux_sym_for_statement_token1] = ACTIONS(1174),
    [aux_sym_for_statement_token2] = ACTIONS(1174),
    [aux_sym_foreach_statement_token1] = ACTIONS(1174),
    [aux_sym_foreach_statement_token2] = ACTIONS(1174),
    [aux_sym_if_statement_token1] = ACTIONS(1174),
    [aux_sym_if_statement_token2] = ACTIONS(1174),
    [aux_sym_else_if_clause_token1] = ACTIONS(1174),
    [aux_sym_else_clause_token1] = ACTIONS(1174),
    [aux_sym_match_expression_token1] = ACTIONS(1174),
    [aux_sym_match_default_expression_token1] = ACTIONS(1174),
    [aux_sym_switch_statement_token1] = ACTIONS(1174),
    [aux_sym_switch_block_token1] = ACTIONS(1174),
    [anon_sym_PLUS] = ACTIONS(1174),
    [anon_sym_DASH] = ACTIONS(1174),
    [anon_sym_TILDE] = ACTIONS(1172),
    [anon_sym_BANG] = ACTIONS(1172),
    [anon_sym_AT] = ACTIONS(1172),
    [aux_sym_clone_expression_token1] = ACTIONS(1174),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1174),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1174),
    [anon_sym_DASH_DASH] = ACTIONS(1172),
    [anon_sym_PLUS_PLUS] = ACTIONS(1172),
    [aux_sym__list_destructing_token1] = ACTIONS(1174),
    [anon_sym_LBRACK] = ACTIONS(1172),
    [anon_sym_self] = ACTIONS(1174),
    [anon_sym_parent] = ACTIONS(1174),
    [aux_sym__argument_name_token1] = ACTIONS(1174),
    [aux_sym__argument_name_token2] = ACTIONS(1174),
    [anon_sym_POUND_LBRACK] = ACTIONS(1172),
    [aux_sym_encapsed_string_token1] = ACTIONS(1172),
    [anon_sym_DQUOTE] = ACTIONS(1172),
    [aux_sym_string_token1] = ACTIONS(1172),
    [anon_sym_SQUOTE] = ACTIONS(1172),
    [anon_sym_LT_LT_LT] = ACTIONS(1172),
    [anon_sym_BQUOTE] = ACTIONS(1172),
    [anon_sym_DOLLAR] = ACTIONS(1172),
    [aux_sym_yield_expression_token1] = ACTIONS(1174),
    [aux_sym_include_expression_token1] = ACTIONS(1174),
    [aux_sym_include_once_expression_token1] = ACTIONS(1174),
    [aux_sym_require_expression_token1] = ACTIONS(1174),
    [aux_sym_require_once_expression_token1] = ACTIONS(1174),
    [sym_comment] = ACTIONS(5),
  },
  [486] = {
    [sym_text_interpolation] = STATE(486),
    [ts_builtin_sym_end] = ACTIONS(1176),
    [sym_name] = ACTIONS(1178),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1176),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1178),
    [aux_sym_global_declaration_token1] = ACTIONS(1178),
    [aux_sym_namespace_definition_token1] = ACTIONS(1178),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1178),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1178),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1178),
    [anon_sym_BSLASH] = ACTIONS(1176),
    [anon_sym_LBRACE] = ACTIONS(1176),
    [anon_sym_RBRACE] = ACTIONS(1176),
    [aux_sym_trait_declaration_token1] = ACTIONS(1178),
    [aux_sym_interface_declaration_token1] = ACTIONS(1178),
    [aux_sym_enum_declaration_token1] = ACTIONS(1178),
    [aux_sym_enum_case_token1] = ACTIONS(1178),
    [aux_sym_class_declaration_token1] = ACTIONS(1178),
    [aux_sym_final_modifier_token1] = ACTIONS(1178),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1178),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1178),
    [sym_var_modifier] = ACTIONS(1178),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1178),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1178),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1178),
    [anon_sym_LPAREN2] = ACTIONS(1176),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1178),
    [aux_sym_cast_type_token1] = ACTIONS(1178),
    [aux_sym_echo_statement_token1] = ACTIONS(1178),
    [aux_sym_exit_statement_token1] = ACTIONS(1178),
    [anon_sym_unset] = ACTIONS(1178),
    [aux_sym_declare_statement_token1] = ACTIONS(1178),
    [aux_sym_declare_statement_token2] = ACTIONS(1178),
    [sym_float] = ACTIONS(1178),
    [aux_sym_try_statement_token1] = ACTIONS(1178),
    [aux_sym_goto_statement_token1] = ACTIONS(1178),
    [aux_sym_continue_statement_token1] = ACTIONS(1178),
    [aux_sym_break_statement_token1] = ACTIONS(1178),
    [sym_integer] = ACTIONS(1178),
    [aux_sym_return_statement_token1] = ACTIONS(1178),
    [aux_sym_throw_expression_token1] = ACTIONS(1178),
    [aux_sym_while_statement_token1] = ACTIONS(1178),
    [aux_sym_while_statement_token2] = ACTIONS(1178),
    [aux_sym_do_statement_token1] = ACTIONS(1178),
    [aux_sym_for_statement_token1] = ACTIONS(1178),
    [aux_sym_for_statement_token2] = ACTIONS(1178),
    [aux_sym_foreach_statement_token1] = ACTIONS(1178),
    [aux_sym_foreach_statement_token2] = ACTIONS(1178),
    [aux_sym_if_statement_token1] = ACTIONS(1178),
    [aux_sym_if_statement_token2] = ACTIONS(1178),
    [aux_sym_else_if_clause_token1] = ACTIONS(1178),
    [aux_sym_else_clause_token1] = ACTIONS(1178),
    [aux_sym_match_expression_token1] = ACTIONS(1178),
    [aux_sym_match_default_expression_token1] = ACTIONS(1178),
    [aux_sym_switch_statement_token1] = ACTIONS(1178),
    [aux_sym_switch_block_token1] = ACTIONS(1178),
    [anon_sym_PLUS] = ACTIONS(1178),
    [anon_sym_DASH] = ACTIONS(1178),
    [anon_sym_TILDE] = ACTIONS(1176),
    [anon_sym_BANG] = ACTIONS(1176),
    [anon_sym_AT] = ACTIONS(1176),
    [aux_sym_clone_expression_token1] = ACTIONS(1178),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1178),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1178),
    [anon_sym_DASH_DASH] = ACTIONS(1176),
    [anon_sym_PLUS_PLUS] = ACTIONS(1176),
    [aux_sym__list_destructing_token1] = ACTIONS(1178),
    [anon_sym_LBRACK] = ACTIONS(1176),
    [anon_sym_self] = ACTIONS(1178),
    [anon_sym_parent] = ACTIONS(1178),
    [aux_sym__argument_name_token1] = ACTIONS(1178),
    [aux_sym__argument_name_token2] = ACTIONS(1178),
    [anon_sym_POUND_LBRACK] = ACTIONS(1176),
    [aux_sym_encapsed_string_token1] = ACTIONS(1176),
    [anon_sym_DQUOTE] = ACTIONS(1176),
    [aux_sym_string_token1] = ACTIONS(1176),
    [anon_sym_SQUOTE] = ACTIONS(1176),
    [anon_sym_LT_LT_LT] = ACTIONS(1176),
    [anon_sym_BQUOTE] = ACTIONS(1176),
    [anon_sym_DOLLAR] = ACTIONS(1176),
    [aux_sym_yield_expression_token1] = ACTIONS(1178),
    [aux_sym_include_expression_token1] = ACTIONS(1178),
    [aux_sym_include_once_expression_token1] = ACTIONS(1178),
    [aux_sym_require_expression_token1] = ACTIONS(1178),
    [aux_sym_require_once_expression_token1] = ACTIONS(1178),
    [sym_comment] = ACTIONS(5),
  },
  [487] = {
    [sym_text_interpolation] = STATE(487),
    [ts_builtin_sym_end] = ACTIONS(1180),
    [sym_name] = ACTIONS(1182),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1180),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1182),
    [aux_sym_global_declaration_token1] = ACTIONS(1182),
    [aux_sym_namespace_definition_token1] = ACTIONS(1182),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1182),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1182),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1182),
    [anon_sym_BSLASH] = ACTIONS(1180),
    [anon_sym_LBRACE] = ACTIONS(1180),
    [anon_sym_RBRACE] = ACTIONS(1180),
    [aux_sym_trait_declaration_token1] = ACTIONS(1182),
    [aux_sym_interface_declaration_token1] = ACTIONS(1182),
    [aux_sym_enum_declaration_token1] = ACTIONS(1182),
    [aux_sym_enum_case_token1] = ACTIONS(1182),
    [aux_sym_class_declaration_token1] = ACTIONS(1182),
    [aux_sym_final_modifier_token1] = ACTIONS(1182),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1182),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1182),
    [sym_var_modifier] = ACTIONS(1182),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1182),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1182),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1182),
    [anon_sym_LPAREN2] = ACTIONS(1180),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1182),
    [aux_sym_cast_type_token1] = ACTIONS(1182),
    [aux_sym_echo_statement_token1] = ACTIONS(1182),
    [aux_sym_exit_statement_token1] = ACTIONS(1182),
    [anon_sym_unset] = ACTIONS(1182),
    [aux_sym_declare_statement_token1] = ACTIONS(1182),
    [aux_sym_declare_statement_token2] = ACTIONS(1182),
    [sym_float] = ACTIONS(1182),
    [aux_sym_try_statement_token1] = ACTIONS(1182),
    [aux_sym_goto_statement_token1] = ACTIONS(1182),
    [aux_sym_continue_statement_token1] = ACTIONS(1182),
    [aux_sym_break_statement_token1] = ACTIONS(1182),
    [sym_integer] = ACTIONS(1182),
    [aux_sym_return_statement_token1] = ACTIONS(1182),
    [aux_sym_throw_expression_token1] = ACTIONS(1182),
    [aux_sym_while_statement_token1] = ACTIONS(1182),
    [aux_sym_while_statement_token2] = ACTIONS(1182),
    [aux_sym_do_statement_token1] = ACTIONS(1182),
    [aux_sym_for_statement_token1] = ACTIONS(1182),
    [aux_sym_for_statement_token2] = ACTIONS(1182),
    [aux_sym_foreach_statement_token1] = ACTIONS(1182),
    [aux_sym_foreach_statement_token2] = ACTIONS(1182),
    [aux_sym_if_statement_token1] = ACTIONS(1182),
    [aux_sym_if_statement_token2] = ACTIONS(1182),
    [aux_sym_else_if_clause_token1] = ACTIONS(1182),
    [aux_sym_else_clause_token1] = ACTIONS(1182),
    [aux_sym_match_expression_token1] = ACTIONS(1182),
    [aux_sym_match_default_expression_token1] = ACTIONS(1182),
    [aux_sym_switch_statement_token1] = ACTIONS(1182),
    [aux_sym_switch_block_token1] = ACTIONS(1182),
    [anon_sym_PLUS] = ACTIONS(1182),
    [anon_sym_DASH] = ACTIONS(1182),
    [anon_sym_TILDE] = ACTIONS(1180),
    [anon_sym_BANG] = ACTIONS(1180),
    [anon_sym_AT] = ACTIONS(1180),
    [aux_sym_clone_expression_token1] = ACTIONS(1182),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1182),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1182),
    [anon_sym_DASH_DASH] = ACTIONS(1180),
    [anon_sym_PLUS_PLUS] = ACTIONS(1180),
    [aux_sym__list_destructing_token1] = ACTIONS(1182),
    [anon_sym_LBRACK] = ACTIONS(1180),
    [anon_sym_self] = ACTIONS(1182),
    [anon_sym_parent] = ACTIONS(1182),
    [aux_sym__argument_name_token1] = ACTIONS(1182),
    [aux_sym__argument_name_token2] = ACTIONS(1182),
    [anon_sym_POUND_LBRACK] = ACTIONS(1180),
    [aux_sym_encapsed_string_token1] = ACTIONS(1180),
    [anon_sym_DQUOTE] = ACTIONS(1180),
    [aux_sym_string_token1] = ACTIONS(1180),
    [anon_sym_SQUOTE] = ACTIONS(1180),
    [anon_sym_LT_LT_LT] = ACTIONS(1180),
    [anon_sym_BQUOTE] = ACTIONS(1180),
    [anon_sym_DOLLAR] = ACTIONS(1180),
    [aux_sym_yield_expression_token1] = ACTIONS(1182),
    [aux_sym_include_expression_token1] = ACTIONS(1182),
    [aux_sym_include_once_expression_token1] = ACTIONS(1182),
    [aux_sym_require_expression_token1] = ACTIONS(1182),
    [aux_sym_require_once_expression_token1] = ACTIONS(1182),
    [sym_comment] = ACTIONS(5),
  },
  [488] = {
    [sym_text_interpolation] = STATE(488),
    [ts_builtin_sym_end] = ACTIONS(1184),
    [sym_name] = ACTIONS(1186),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1184),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1186),
    [aux_sym_global_declaration_token1] = ACTIONS(1186),
    [aux_sym_namespace_definition_token1] = ACTIONS(1186),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1186),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1186),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1186),
    [anon_sym_BSLASH] = ACTIONS(1184),
    [anon_sym_LBRACE] = ACTIONS(1184),
    [anon_sym_RBRACE] = ACTIONS(1184),
    [aux_sym_trait_declaration_token1] = ACTIONS(1186),
    [aux_sym_interface_declaration_token1] = ACTIONS(1186),
    [aux_sym_enum_declaration_token1] = ACTIONS(1186),
    [aux_sym_enum_case_token1] = ACTIONS(1186),
    [aux_sym_class_declaration_token1] = ACTIONS(1186),
    [aux_sym_final_modifier_token1] = ACTIONS(1186),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1186),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1186),
    [sym_var_modifier] = ACTIONS(1186),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1186),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1186),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1186),
    [anon_sym_LPAREN2] = ACTIONS(1184),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1186),
    [aux_sym_cast_type_token1] = ACTIONS(1186),
    [aux_sym_echo_statement_token1] = ACTIONS(1186),
    [aux_sym_exit_statement_token1] = ACTIONS(1186),
    [anon_sym_unset] = ACTIONS(1186),
    [aux_sym_declare_statement_token1] = ACTIONS(1186),
    [aux_sym_declare_statement_token2] = ACTIONS(1186),
    [sym_float] = ACTIONS(1186),
    [aux_sym_try_statement_token1] = ACTIONS(1186),
    [aux_sym_goto_statement_token1] = ACTIONS(1186),
    [aux_sym_continue_statement_token1] = ACTIONS(1186),
    [aux_sym_break_statement_token1] = ACTIONS(1186),
    [sym_integer] = ACTIONS(1186),
    [aux_sym_return_statement_token1] = ACTIONS(1186),
    [aux_sym_throw_expression_token1] = ACTIONS(1186),
    [aux_sym_while_statement_token1] = ACTIONS(1186),
    [aux_sym_while_statement_token2] = ACTIONS(1186),
    [aux_sym_do_statement_token1] = ACTIONS(1186),
    [aux_sym_for_statement_token1] = ACTIONS(1186),
    [aux_sym_for_statement_token2] = ACTIONS(1186),
    [aux_sym_foreach_statement_token1] = ACTIONS(1186),
    [aux_sym_foreach_statement_token2] = ACTIONS(1186),
    [aux_sym_if_statement_token1] = ACTIONS(1186),
    [aux_sym_if_statement_token2] = ACTIONS(1186),
    [aux_sym_else_if_clause_token1] = ACTIONS(1186),
    [aux_sym_else_clause_token1] = ACTIONS(1186),
    [aux_sym_match_expression_token1] = ACTIONS(1186),
    [aux_sym_match_default_expression_token1] = ACTIONS(1186),
    [aux_sym_switch_statement_token1] = ACTIONS(1186),
    [aux_sym_switch_block_token1] = ACTIONS(1186),
    [anon_sym_PLUS] = ACTIONS(1186),
    [anon_sym_DASH] = ACTIONS(1186),
    [anon_sym_TILDE] = ACTIONS(1184),
    [anon_sym_BANG] = ACTIONS(1184),
    [anon_sym_AT] = ACTIONS(1184),
    [aux_sym_clone_expression_token1] = ACTIONS(1186),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1186),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1186),
    [anon_sym_DASH_DASH] = ACTIONS(1184),
    [anon_sym_PLUS_PLUS] = ACTIONS(1184),
    [aux_sym__list_destructing_token1] = ACTIONS(1186),
    [anon_sym_LBRACK] = ACTIONS(1184),
    [anon_sym_self] = ACTIONS(1186),
    [anon_sym_parent] = ACTIONS(1186),
    [aux_sym__argument_name_token1] = ACTIONS(1186),
    [aux_sym__argument_name_token2] = ACTIONS(1186),
    [anon_sym_POUND_LBRACK] = ACTIONS(1184),
    [aux_sym_encapsed_string_token1] = ACTIONS(1184),
    [anon_sym_DQUOTE] = ACTIONS(1184),
    [aux_sym_string_token1] = ACTIONS(1184),
    [anon_sym_SQUOTE] = ACTIONS(1184),
    [anon_sym_LT_LT_LT] = ACTIONS(1184),
    [anon_sym_BQUOTE] = ACTIONS(1184),
    [anon_sym_DOLLAR] = ACTIONS(1184),
    [aux_sym_yield_expression_token1] = ACTIONS(1186),
    [aux_sym_include_expression_token1] = ACTIONS(1186),
    [aux_sym_include_once_expression_token1] = ACTIONS(1186),
    [aux_sym_require_expression_token1] = ACTIONS(1186),
    [aux_sym_require_once_expression_token1] = ACTIONS(1186),
    [sym_comment] = ACTIONS(5),
  },
  [489] = {
    [sym_text_interpolation] = STATE(489),
    [ts_builtin_sym_end] = ACTIONS(1188),
    [sym_name] = ACTIONS(1190),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1188),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1190),
    [aux_sym_global_declaration_token1] = ACTIONS(1190),
    [aux_sym_namespace_definition_token1] = ACTIONS(1190),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1190),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1190),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1190),
    [anon_sym_BSLASH] = ACTIONS(1188),
    [anon_sym_LBRACE] = ACTIONS(1188),
    [anon_sym_RBRACE] = ACTIONS(1188),
    [aux_sym_trait_declaration_token1] = ACTIONS(1190),
    [aux_sym_interface_declaration_token1] = ACTIONS(1190),
    [aux_sym_enum_declaration_token1] = ACTIONS(1190),
    [aux_sym_enum_case_token1] = ACTIONS(1190),
    [aux_sym_class_declaration_token1] = ACTIONS(1190),
    [aux_sym_final_modifier_token1] = ACTIONS(1190),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1190),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1190),
    [sym_var_modifier] = ACTIONS(1190),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1190),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1190),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1190),
    [anon_sym_LPAREN2] = ACTIONS(1188),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1190),
    [aux_sym_cast_type_token1] = ACTIONS(1190),
    [aux_sym_echo_statement_token1] = ACTIONS(1190),
    [aux_sym_exit_statement_token1] = ACTIONS(1190),
    [anon_sym_unset] = ACTIONS(1190),
    [aux_sym_declare_statement_token1] = ACTIONS(1190),
    [aux_sym_declare_statement_token2] = ACTIONS(1190),
    [sym_float] = ACTIONS(1190),
    [aux_sym_try_statement_token1] = ACTIONS(1190),
    [aux_sym_goto_statement_token1] = ACTIONS(1190),
    [aux_sym_continue_statement_token1] = ACTIONS(1190),
    [aux_sym_break_statement_token1] = ACTIONS(1190),
    [sym_integer] = ACTIONS(1190),
    [aux_sym_return_statement_token1] = ACTIONS(1190),
    [aux_sym_throw_expression_token1] = ACTIONS(1190),
    [aux_sym_while_statement_token1] = ACTIONS(1190),
    [aux_sym_while_statement_token2] = ACTIONS(1190),
    [aux_sym_do_statement_token1] = ACTIONS(1190),
    [aux_sym_for_statement_token1] = ACTIONS(1190),
    [aux_sym_for_statement_token2] = ACTIONS(1190),
    [aux_sym_foreach_statement_token1] = ACTIONS(1190),
    [aux_sym_foreach_statement_token2] = ACTIONS(1190),
    [aux_sym_if_statement_token1] = ACTIONS(1190),
    [aux_sym_if_statement_token2] = ACTIONS(1190),
    [aux_sym_else_if_clause_token1] = ACTIONS(1190),
    [aux_sym_else_clause_token1] = ACTIONS(1190),
    [aux_sym_match_expression_token1] = ACTIONS(1190),
    [aux_sym_match_default_expression_token1] = ACTIONS(1190),
    [aux_sym_switch_statement_token1] = ACTIONS(1190),
    [aux_sym_switch_block_token1] = ACTIONS(1190),
    [anon_sym_PLUS] = ACTIONS(1190),
    [anon_sym_DASH] = ACTIONS(1190),
    [anon_sym_TILDE] = ACTIONS(1188),
    [anon_sym_BANG] = ACTIONS(1188),
    [anon_sym_AT] = ACTIONS(1188),
    [aux_sym_clone_expression_token1] = ACTIONS(1190),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1190),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1190),
    [anon_sym_DASH_DASH] = ACTIONS(1188),
    [anon_sym_PLUS_PLUS] = ACTIONS(1188),
    [aux_sym__list_destructing_token1] = ACTIONS(1190),
    [anon_sym_LBRACK] = ACTIONS(1188),
    [anon_sym_self] = ACTIONS(1190),
    [anon_sym_parent] = ACTIONS(1190),
    [aux_sym__argument_name_token1] = ACTIONS(1190),
    [aux_sym__argument_name_token2] = ACTIONS(1190),
    [anon_sym_POUND_LBRACK] = ACTIONS(1188),
    [aux_sym_encapsed_string_token1] = ACTIONS(1188),
    [anon_sym_DQUOTE] = ACTIONS(1188),
    [aux_sym_string_token1] = ACTIONS(1188),
    [anon_sym_SQUOTE] = ACTIONS(1188),
    [anon_sym_LT_LT_LT] = ACTIONS(1188),
    [anon_sym_BQUOTE] = ACTIONS(1188),
    [anon_sym_DOLLAR] = ACTIONS(1188),
    [aux_sym_yield_expression_token1] = ACTIONS(1190),
    [aux_sym_include_expression_token1] = ACTIONS(1190),
    [aux_sym_include_once_expression_token1] = ACTIONS(1190),
    [aux_sym_require_expression_token1] = ACTIONS(1190),
    [aux_sym_require_once_expression_token1] = ACTIONS(1190),
    [sym_comment] = ACTIONS(5),
  },
  [490] = {
    [sym_text_interpolation] = STATE(490),
    [ts_builtin_sym_end] = ACTIONS(1192),
    [sym_name] = ACTIONS(1194),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1192),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1194),
    [aux_sym_global_declaration_token1] = ACTIONS(1194),
    [aux_sym_namespace_definition_token1] = ACTIONS(1194),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1194),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1194),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1194),
    [anon_sym_BSLASH] = ACTIONS(1192),
    [anon_sym_LBRACE] = ACTIONS(1192),
    [anon_sym_RBRACE] = ACTIONS(1192),
    [aux_sym_trait_declaration_token1] = ACTIONS(1194),
    [aux_sym_interface_declaration_token1] = ACTIONS(1194),
    [aux_sym_enum_declaration_token1] = ACTIONS(1194),
    [aux_sym_enum_case_token1] = ACTIONS(1194),
    [aux_sym_class_declaration_token1] = ACTIONS(1194),
    [aux_sym_final_modifier_token1] = ACTIONS(1194),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1194),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1194),
    [sym_var_modifier] = ACTIONS(1194),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1194),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1194),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1194),
    [anon_sym_LPAREN2] = ACTIONS(1192),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1194),
    [aux_sym_cast_type_token1] = ACTIONS(1194),
    [aux_sym_echo_statement_token1] = ACTIONS(1194),
    [aux_sym_exit_statement_token1] = ACTIONS(1194),
    [anon_sym_unset] = ACTIONS(1194),
    [aux_sym_declare_statement_token1] = ACTIONS(1194),
    [aux_sym_declare_statement_token2] = ACTIONS(1194),
    [sym_float] = ACTIONS(1194),
    [aux_sym_try_statement_token1] = ACTIONS(1194),
    [aux_sym_goto_statement_token1] = ACTIONS(1194),
    [aux_sym_continue_statement_token1] = ACTIONS(1194),
    [aux_sym_break_statement_token1] = ACTIONS(1194),
    [sym_integer] = ACTIONS(1194),
    [aux_sym_return_statement_token1] = ACTIONS(1194),
    [aux_sym_throw_expression_token1] = ACTIONS(1194),
    [aux_sym_while_statement_token1] = ACTIONS(1194),
    [aux_sym_while_statement_token2] = ACTIONS(1194),
    [aux_sym_do_statement_token1] = ACTIONS(1194),
    [aux_sym_for_statement_token1] = ACTIONS(1194),
    [aux_sym_for_statement_token2] = ACTIONS(1194),
    [aux_sym_foreach_statement_token1] = ACTIONS(1194),
    [aux_sym_foreach_statement_token2] = ACTIONS(1194),
    [aux_sym_if_statement_token1] = ACTIONS(1194),
    [aux_sym_if_statement_token2] = ACTIONS(1194),
    [aux_sym_else_if_clause_token1] = ACTIONS(1194),
    [aux_sym_else_clause_token1] = ACTIONS(1194),
    [aux_sym_match_expression_token1] = ACTIONS(1194),
    [aux_sym_match_default_expression_token1] = ACTIONS(1194),
    [aux_sym_switch_statement_token1] = ACTIONS(1194),
    [aux_sym_switch_block_token1] = ACTIONS(1194),
    [anon_sym_PLUS] = ACTIONS(1194),
    [anon_sym_DASH] = ACTIONS(1194),
    [anon_sym_TILDE] = ACTIONS(1192),
    [anon_sym_BANG] = ACTIONS(1192),
    [anon_sym_AT] = ACTIONS(1192),
    [aux_sym_clone_expression_token1] = ACTIONS(1194),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1194),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1194),
    [anon_sym_DASH_DASH] = ACTIONS(1192),
    [anon_sym_PLUS_PLUS] = ACTIONS(1192),
    [aux_sym__list_destructing_token1] = ACTIONS(1194),
    [anon_sym_LBRACK] = ACTIONS(1192),
    [anon_sym_self] = ACTIONS(1194),
    [anon_sym_parent] = ACTIONS(1194),
    [aux_sym__argument_name_token1] = ACTIONS(1194),
    [aux_sym__argument_name_token2] = ACTIONS(1194),
    [anon_sym_POUND_LBRACK] = ACTIONS(1192),
    [aux_sym_encapsed_string_token1] = ACTIONS(1192),
    [anon_sym_DQUOTE] = ACTIONS(1192),
    [aux_sym_string_token1] = ACTIONS(1192),
    [anon_sym_SQUOTE] = ACTIONS(1192),
    [anon_sym_LT_LT_LT] = ACTIONS(1192),
    [anon_sym_BQUOTE] = ACTIONS(1192),
    [anon_sym_DOLLAR] = ACTIONS(1192),
    [aux_sym_yield_expression_token1] = ACTIONS(1194),
    [aux_sym_include_expression_token1] = ACTIONS(1194),
    [aux_sym_include_once_expression_token1] = ACTIONS(1194),
    [aux_sym_require_expression_token1] = ACTIONS(1194),
    [aux_sym_require_once_expression_token1] = ACTIONS(1194),
    [sym_comment] = ACTIONS(5),
  },
  [491] = {
    [sym_text_interpolation] = STATE(491),
    [ts_builtin_sym_end] = ACTIONS(1196),
    [sym_name] = ACTIONS(1198),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1196),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1198),
    [aux_sym_global_declaration_token1] = ACTIONS(1198),
    [aux_sym_namespace_definition_token1] = ACTIONS(1198),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1198),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1198),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1198),
    [anon_sym_BSLASH] = ACTIONS(1196),
    [anon_sym_LBRACE] = ACTIONS(1196),
    [anon_sym_RBRACE] = ACTIONS(1196),
    [aux_sym_trait_declaration_token1] = ACTIONS(1198),
    [aux_sym_interface_declaration_token1] = ACTIONS(1198),
    [aux_sym_enum_declaration_token1] = ACTIONS(1198),
    [aux_sym_enum_case_token1] = ACTIONS(1198),
    [aux_sym_class_declaration_token1] = ACTIONS(1198),
    [aux_sym_final_modifier_token1] = ACTIONS(1198),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1198),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1198),
    [sym_var_modifier] = ACTIONS(1198),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1198),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1198),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1198),
    [anon_sym_LPAREN2] = ACTIONS(1196),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1198),
    [aux_sym_cast_type_token1] = ACTIONS(1198),
    [aux_sym_echo_statement_token1] = ACTIONS(1198),
    [aux_sym_exit_statement_token1] = ACTIONS(1198),
    [anon_sym_unset] = ACTIONS(1198),
    [aux_sym_declare_statement_token1] = ACTIONS(1198),
    [aux_sym_declare_statement_token2] = ACTIONS(1198),
    [sym_float] = ACTIONS(1198),
    [aux_sym_try_statement_token1] = ACTIONS(1198),
    [aux_sym_goto_statement_token1] = ACTIONS(1198),
    [aux_sym_continue_statement_token1] = ACTIONS(1198),
    [aux_sym_break_statement_token1] = ACTIONS(1198),
    [sym_integer] = ACTIONS(1198),
    [aux_sym_return_statement_token1] = ACTIONS(1198),
    [aux_sym_throw_expression_token1] = ACTIONS(1198),
    [aux_sym_while_statement_token1] = ACTIONS(1198),
    [aux_sym_while_statement_token2] = ACTIONS(1198),
    [aux_sym_do_statement_token1] = ACTIONS(1198),
    [aux_sym_for_statement_token1] = ACTIONS(1198),
    [aux_sym_for_statement_token2] = ACTIONS(1198),
    [aux_sym_foreach_statement_token1] = ACTIONS(1198),
    [aux_sym_foreach_statement_token2] = ACTIONS(1198),
    [aux_sym_if_statement_token1] = ACTIONS(1198),
    [aux_sym_if_statement_token2] = ACTIONS(1198),
    [aux_sym_else_if_clause_token1] = ACTIONS(1198),
    [aux_sym_else_clause_token1] = ACTIONS(1198),
    [aux_sym_match_expression_token1] = ACTIONS(1198),
    [aux_sym_match_default_expression_token1] = ACTIONS(1198),
    [aux_sym_switch_statement_token1] = ACTIONS(1198),
    [aux_sym_switch_block_token1] = ACTIONS(1198),
    [anon_sym_PLUS] = ACTIONS(1198),
    [anon_sym_DASH] = ACTIONS(1198),
    [anon_sym_TILDE] = ACTIONS(1196),
    [anon_sym_BANG] = ACTIONS(1196),
    [anon_sym_AT] = ACTIONS(1196),
    [aux_sym_clone_expression_token1] = ACTIONS(1198),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1198),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1198),
    [anon_sym_DASH_DASH] = ACTIONS(1196),
    [anon_sym_PLUS_PLUS] = ACTIONS(1196),
    [aux_sym__list_destructing_token1] = ACTIONS(1198),
    [anon_sym_LBRACK] = ACTIONS(1196),
    [anon_sym_self] = ACTIONS(1198),
    [anon_sym_parent] = ACTIONS(1198),
    [aux_sym__argument_name_token1] = ACTIONS(1198),
    [aux_sym__argument_name_token2] = ACTIONS(1198),
    [anon_sym_POUND_LBRACK] = ACTIONS(1196),
    [aux_sym_encapsed_string_token1] = ACTIONS(1196),
    [anon_sym_DQUOTE] = ACTIONS(1196),
    [aux_sym_string_token1] = ACTIONS(1196),
    [anon_sym_SQUOTE] = ACTIONS(1196),
    [anon_sym_LT_LT_LT] = ACTIONS(1196),
    [anon_sym_BQUOTE] = ACTIONS(1196),
    [anon_sym_DOLLAR] = ACTIONS(1196),
    [aux_sym_yield_expression_token1] = ACTIONS(1198),
    [aux_sym_include_expression_token1] = ACTIONS(1198),
    [aux_sym_include_once_expression_token1] = ACTIONS(1198),
    [aux_sym_require_expression_token1] = ACTIONS(1198),
    [aux_sym_require_once_expression_token1] = ACTIONS(1198),
    [sym_comment] = ACTIONS(5),
  },
  [492] = {
    [sym_text_interpolation] = STATE(492),
    [ts_builtin_sym_end] = ACTIONS(1200),
    [sym_name] = ACTIONS(1202),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1200),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1202),
    [aux_sym_global_declaration_token1] = ACTIONS(1202),
    [aux_sym_namespace_definition_token1] = ACTIONS(1202),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1202),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1202),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1202),
    [anon_sym_BSLASH] = ACTIONS(1200),
    [anon_sym_LBRACE] = ACTIONS(1200),
    [anon_sym_RBRACE] = ACTIONS(1200),
    [aux_sym_trait_declaration_token1] = ACTIONS(1202),
    [aux_sym_interface_declaration_token1] = ACTIONS(1202),
    [aux_sym_enum_declaration_token1] = ACTIONS(1202),
    [aux_sym_enum_case_token1] = ACTIONS(1202),
    [aux_sym_class_declaration_token1] = ACTIONS(1202),
    [aux_sym_final_modifier_token1] = ACTIONS(1202),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1202),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1202),
    [sym_var_modifier] = ACTIONS(1202),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1202),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1202),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1202),
    [anon_sym_LPAREN2] = ACTIONS(1200),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1202),
    [aux_sym_cast_type_token1] = ACTIONS(1202),
    [aux_sym_echo_statement_token1] = ACTIONS(1202),
    [aux_sym_exit_statement_token1] = ACTIONS(1202),
    [anon_sym_unset] = ACTIONS(1202),
    [aux_sym_declare_statement_token1] = ACTIONS(1202),
    [aux_sym_declare_statement_token2] = ACTIONS(1202),
    [sym_float] = ACTIONS(1202),
    [aux_sym_try_statement_token1] = ACTIONS(1202),
    [aux_sym_goto_statement_token1] = ACTIONS(1202),
    [aux_sym_continue_statement_token1] = ACTIONS(1202),
    [aux_sym_break_statement_token1] = ACTIONS(1202),
    [sym_integer] = ACTIONS(1202),
    [aux_sym_return_statement_token1] = ACTIONS(1202),
    [aux_sym_throw_expression_token1] = ACTIONS(1202),
    [aux_sym_while_statement_token1] = ACTIONS(1202),
    [aux_sym_while_statement_token2] = ACTIONS(1202),
    [aux_sym_do_statement_token1] = ACTIONS(1202),
    [aux_sym_for_statement_token1] = ACTIONS(1202),
    [aux_sym_for_statement_token2] = ACTIONS(1202),
    [aux_sym_foreach_statement_token1] = ACTIONS(1202),
    [aux_sym_foreach_statement_token2] = ACTIONS(1202),
    [aux_sym_if_statement_token1] = ACTIONS(1202),
    [aux_sym_if_statement_token2] = ACTIONS(1202),
    [aux_sym_else_if_clause_token1] = ACTIONS(1202),
    [aux_sym_else_clause_token1] = ACTIONS(1202),
    [aux_sym_match_expression_token1] = ACTIONS(1202),
    [aux_sym_match_default_expression_token1] = ACTIONS(1202),
    [aux_sym_switch_statement_token1] = ACTIONS(1202),
    [aux_sym_switch_block_token1] = ACTIONS(1202),
    [anon_sym_PLUS] = ACTIONS(1202),
    [anon_sym_DASH] = ACTIONS(1202),
    [anon_sym_TILDE] = ACTIONS(1200),
    [anon_sym_BANG] = ACTIONS(1200),
    [anon_sym_AT] = ACTIONS(1200),
    [aux_sym_clone_expression_token1] = ACTIONS(1202),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1202),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1202),
    [anon_sym_DASH_DASH] = ACTIONS(1200),
    [anon_sym_PLUS_PLUS] = ACTIONS(1200),
    [aux_sym__list_destructing_token1] = ACTIONS(1202),
    [anon_sym_LBRACK] = ACTIONS(1200),
    [anon_sym_self] = ACTIONS(1202),
    [anon_sym_parent] = ACTIONS(1202),
    [aux_sym__argument_name_token1] = ACTIONS(1202),
    [aux_sym__argument_name_token2] = ACTIONS(1202),
    [anon_sym_POUND_LBRACK] = ACTIONS(1200),
    [aux_sym_encapsed_string_token1] = ACTIONS(1200),
    [anon_sym_DQUOTE] = ACTIONS(1200),
    [aux_sym_string_token1] = ACTIONS(1200),
    [anon_sym_SQUOTE] = ACTIONS(1200),
    [anon_sym_LT_LT_LT] = ACTIONS(1200),
    [anon_sym_BQUOTE] = ACTIONS(1200),
    [anon_sym_DOLLAR] = ACTIONS(1200),
    [aux_sym_yield_expression_token1] = ACTIONS(1202),
    [aux_sym_include_expression_token1] = ACTIONS(1202),
    [aux_sym_include_once_expression_token1] = ACTIONS(1202),
    [aux_sym_require_expression_token1] = ACTIONS(1202),
    [aux_sym_require_once_expression_token1] = ACTIONS(1202),
    [sym_comment] = ACTIONS(5),
  },
  [493] = {
    [sym_text_interpolation] = STATE(493),
    [ts_builtin_sym_end] = ACTIONS(1204),
    [sym_name] = ACTIONS(1206),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1204),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1206),
    [aux_sym_global_declaration_token1] = ACTIONS(1206),
    [aux_sym_namespace_definition_token1] = ACTIONS(1206),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1206),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1206),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1206),
    [anon_sym_BSLASH] = ACTIONS(1204),
    [anon_sym_LBRACE] = ACTIONS(1204),
    [anon_sym_RBRACE] = ACTIONS(1204),
    [aux_sym_trait_declaration_token1] = ACTIONS(1206),
    [aux_sym_interface_declaration_token1] = ACTIONS(1206),
    [aux_sym_enum_declaration_token1] = ACTIONS(1206),
    [aux_sym_enum_case_token1] = ACTIONS(1206),
    [aux_sym_class_declaration_token1] = ACTIONS(1206),
    [aux_sym_final_modifier_token1] = ACTIONS(1206),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1206),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1206),
    [sym_var_modifier] = ACTIONS(1206),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1206),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1206),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1206),
    [anon_sym_LPAREN2] = ACTIONS(1204),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1206),
    [aux_sym_cast_type_token1] = ACTIONS(1206),
    [aux_sym_echo_statement_token1] = ACTIONS(1206),
    [aux_sym_exit_statement_token1] = ACTIONS(1206),
    [anon_sym_unset] = ACTIONS(1206),
    [aux_sym_declare_statement_token1] = ACTIONS(1206),
    [aux_sym_declare_statement_token2] = ACTIONS(1206),
    [sym_float] = ACTIONS(1206),
    [aux_sym_try_statement_token1] = ACTIONS(1206),
    [aux_sym_goto_statement_token1] = ACTIONS(1206),
    [aux_sym_continue_statement_token1] = ACTIONS(1206),
    [aux_sym_break_statement_token1] = ACTIONS(1206),
    [sym_integer] = ACTIONS(1206),
    [aux_sym_return_statement_token1] = ACTIONS(1206),
    [aux_sym_throw_expression_token1] = ACTIONS(1206),
    [aux_sym_while_statement_token1] = ACTIONS(1206),
    [aux_sym_while_statement_token2] = ACTIONS(1206),
    [aux_sym_do_statement_token1] = ACTIONS(1206),
    [aux_sym_for_statement_token1] = ACTIONS(1206),
    [aux_sym_for_statement_token2] = ACTIONS(1206),
    [aux_sym_foreach_statement_token1] = ACTIONS(1206),
    [aux_sym_foreach_statement_token2] = ACTIONS(1206),
    [aux_sym_if_statement_token1] = ACTIONS(1206),
    [aux_sym_if_statement_token2] = ACTIONS(1206),
    [aux_sym_else_if_clause_token1] = ACTIONS(1206),
    [aux_sym_else_clause_token1] = ACTIONS(1206),
    [aux_sym_match_expression_token1] = ACTIONS(1206),
    [aux_sym_match_default_expression_token1] = ACTIONS(1206),
    [aux_sym_switch_statement_token1] = ACTIONS(1206),
    [aux_sym_switch_block_token1] = ACTIONS(1206),
    [anon_sym_PLUS] = ACTIONS(1206),
    [anon_sym_DASH] = ACTIONS(1206),
    [anon_sym_TILDE] = ACTIONS(1204),
    [anon_sym_BANG] = ACTIONS(1204),
    [anon_sym_AT] = ACTIONS(1204),
    [aux_sym_clone_expression_token1] = ACTIONS(1206),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1206),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1206),
    [anon_sym_DASH_DASH] = ACTIONS(1204),
    [anon_sym_PLUS_PLUS] = ACTIONS(1204),
    [aux_sym__list_destructing_token1] = ACTIONS(1206),
    [anon_sym_LBRACK] = ACTIONS(1204),
    [anon_sym_self] = ACTIONS(1206),
    [anon_sym_parent] = ACTIONS(1206),
    [aux_sym__argument_name_token1] = ACTIONS(1206),
    [aux_sym__argument_name_token2] = ACTIONS(1206),
    [anon_sym_POUND_LBRACK] = ACTIONS(1204),
    [aux_sym_encapsed_string_token1] = ACTIONS(1204),
    [anon_sym_DQUOTE] = ACTIONS(1204),
    [aux_sym_string_token1] = ACTIONS(1204),
    [anon_sym_SQUOTE] = ACTIONS(1204),
    [anon_sym_LT_LT_LT] = ACTIONS(1204),
    [anon_sym_BQUOTE] = ACTIONS(1204),
    [anon_sym_DOLLAR] = ACTIONS(1204),
    [aux_sym_yield_expression_token1] = ACTIONS(1206),
    [aux_sym_include_expression_token1] = ACTIONS(1206),
    [aux_sym_include_once_expression_token1] = ACTIONS(1206),
    [aux_sym_require_expression_token1] = ACTIONS(1206),
    [aux_sym_require_once_expression_token1] = ACTIONS(1206),
    [sym_comment] = ACTIONS(5),
  },
  [494] = {
    [sym_text_interpolation] = STATE(494),
    [ts_builtin_sym_end] = ACTIONS(1208),
    [sym_name] = ACTIONS(1210),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1208),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1210),
    [aux_sym_global_declaration_token1] = ACTIONS(1210),
    [aux_sym_namespace_definition_token1] = ACTIONS(1210),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1210),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1210),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1210),
    [anon_sym_BSLASH] = ACTIONS(1208),
    [anon_sym_LBRACE] = ACTIONS(1208),
    [anon_sym_RBRACE] = ACTIONS(1208),
    [aux_sym_trait_declaration_token1] = ACTIONS(1210),
    [aux_sym_interface_declaration_token1] = ACTIONS(1210),
    [aux_sym_enum_declaration_token1] = ACTIONS(1210),
    [aux_sym_enum_case_token1] = ACTIONS(1210),
    [aux_sym_class_declaration_token1] = ACTIONS(1210),
    [aux_sym_final_modifier_token1] = ACTIONS(1210),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1210),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1210),
    [sym_var_modifier] = ACTIONS(1210),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1210),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1210),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1210),
    [anon_sym_LPAREN2] = ACTIONS(1208),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1210),
    [aux_sym_cast_type_token1] = ACTIONS(1210),
    [aux_sym_echo_statement_token1] = ACTIONS(1210),
    [aux_sym_exit_statement_token1] = ACTIONS(1210),
    [anon_sym_unset] = ACTIONS(1210),
    [aux_sym_declare_statement_token1] = ACTIONS(1210),
    [aux_sym_declare_statement_token2] = ACTIONS(1210),
    [sym_float] = ACTIONS(1210),
    [aux_sym_try_statement_token1] = ACTIONS(1210),
    [aux_sym_goto_statement_token1] = ACTIONS(1210),
    [aux_sym_continue_statement_token1] = ACTIONS(1210),
    [aux_sym_break_statement_token1] = ACTIONS(1210),
    [sym_integer] = ACTIONS(1210),
    [aux_sym_return_statement_token1] = ACTIONS(1210),
    [aux_sym_throw_expression_token1] = ACTIONS(1210),
    [aux_sym_while_statement_token1] = ACTIONS(1210),
    [aux_sym_while_statement_token2] = ACTIONS(1210),
    [aux_sym_do_statement_token1] = ACTIONS(1210),
    [aux_sym_for_statement_token1] = ACTIONS(1210),
    [aux_sym_for_statement_token2] = ACTIONS(1210),
    [aux_sym_foreach_statement_token1] = ACTIONS(1210),
    [aux_sym_foreach_statement_token2] = ACTIONS(1210),
    [aux_sym_if_statement_token1] = ACTIONS(1210),
    [aux_sym_if_statement_token2] = ACTIONS(1210),
    [aux_sym_else_if_clause_token1] = ACTIONS(1210),
    [aux_sym_else_clause_token1] = ACTIONS(1210),
    [aux_sym_match_expression_token1] = ACTIONS(1210),
    [aux_sym_match_default_expression_token1] = ACTIONS(1210),
    [aux_sym_switch_statement_token1] = ACTIONS(1210),
    [aux_sym_switch_block_token1] = ACTIONS(1210),
    [anon_sym_PLUS] = ACTIONS(1210),
    [anon_sym_DASH] = ACTIONS(1210),
    [anon_sym_TILDE] = ACTIONS(1208),
    [anon_sym_BANG] = ACTIONS(1208),
    [anon_sym_AT] = ACTIONS(1208),
    [aux_sym_clone_expression_token1] = ACTIONS(1210),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1210),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1210),
    [anon_sym_DASH_DASH] = ACTIONS(1208),
    [anon_sym_PLUS_PLUS] = ACTIONS(1208),
    [aux_sym__list_destructing_token1] = ACTIONS(1210),
    [anon_sym_LBRACK] = ACTIONS(1208),
    [anon_sym_self] = ACTIONS(1210),
    [anon_sym_parent] = ACTIONS(1210),
    [aux_sym__argument_name_token1] = ACTIONS(1210),
    [aux_sym__argument_name_token2] = ACTIONS(1210),
    [anon_sym_POUND_LBRACK] = ACTIONS(1208),
    [aux_sym_encapsed_string_token1] = ACTIONS(1208),
    [anon_sym_DQUOTE] = ACTIONS(1208),
    [aux_sym_string_token1] = ACTIONS(1208),
    [anon_sym_SQUOTE] = ACTIONS(1208),
    [anon_sym_LT_LT_LT] = ACTIONS(1208),
    [anon_sym_BQUOTE] = ACTIONS(1208),
    [anon_sym_DOLLAR] = ACTIONS(1208),
    [aux_sym_yield_expression_token1] = ACTIONS(1210),
    [aux_sym_include_expression_token1] = ACTIONS(1210),
    [aux_sym_include_once_expression_token1] = ACTIONS(1210),
    [aux_sym_require_expression_token1] = ACTIONS(1210),
    [aux_sym_require_once_expression_token1] = ACTIONS(1210),
    [sym_comment] = ACTIONS(5),
  },
  [495] = {
    [sym_text_interpolation] = STATE(495),
    [ts_builtin_sym_end] = ACTIONS(1212),
    [sym_name] = ACTIONS(1214),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1212),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1214),
    [aux_sym_global_declaration_token1] = ACTIONS(1214),
    [aux_sym_namespace_definition_token1] = ACTIONS(1214),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1214),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1214),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1214),
    [anon_sym_BSLASH] = ACTIONS(1212),
    [anon_sym_LBRACE] = ACTIONS(1212),
    [anon_sym_RBRACE] = ACTIONS(1212),
    [aux_sym_trait_declaration_token1] = ACTIONS(1214),
    [aux_sym_interface_declaration_token1] = ACTIONS(1214),
    [aux_sym_enum_declaration_token1] = ACTIONS(1214),
    [aux_sym_enum_case_token1] = ACTIONS(1214),
    [aux_sym_class_declaration_token1] = ACTIONS(1214),
    [aux_sym_final_modifier_token1] = ACTIONS(1214),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1214),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1214),
    [sym_var_modifier] = ACTIONS(1214),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1214),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1214),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1214),
    [anon_sym_LPAREN2] = ACTIONS(1212),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1214),
    [aux_sym_cast_type_token1] = ACTIONS(1214),
    [aux_sym_echo_statement_token1] = ACTIONS(1214),
    [aux_sym_exit_statement_token1] = ACTIONS(1214),
    [anon_sym_unset] = ACTIONS(1214),
    [aux_sym_declare_statement_token1] = ACTIONS(1214),
    [aux_sym_declare_statement_token2] = ACTIONS(1214),
    [sym_float] = ACTIONS(1214),
    [aux_sym_try_statement_token1] = ACTIONS(1214),
    [aux_sym_goto_statement_token1] = ACTIONS(1214),
    [aux_sym_continue_statement_token1] = ACTIONS(1214),
    [aux_sym_break_statement_token1] = ACTIONS(1214),
    [sym_integer] = ACTIONS(1214),
    [aux_sym_return_statement_token1] = ACTIONS(1214),
    [aux_sym_throw_expression_token1] = ACTIONS(1214),
    [aux_sym_while_statement_token1] = ACTIONS(1214),
    [aux_sym_while_statement_token2] = ACTIONS(1214),
    [aux_sym_do_statement_token1] = ACTIONS(1214),
    [aux_sym_for_statement_token1] = ACTIONS(1214),
    [aux_sym_for_statement_token2] = ACTIONS(1214),
    [aux_sym_foreach_statement_token1] = ACTIONS(1214),
    [aux_sym_foreach_statement_token2] = ACTIONS(1214),
    [aux_sym_if_statement_token1] = ACTIONS(1214),
    [aux_sym_if_statement_token2] = ACTIONS(1214),
    [aux_sym_else_if_clause_token1] = ACTIONS(1214),
    [aux_sym_else_clause_token1] = ACTIONS(1214),
    [aux_sym_match_expression_token1] = ACTIONS(1214),
    [aux_sym_match_default_expression_token1] = ACTIONS(1214),
    [aux_sym_switch_statement_token1] = ACTIONS(1214),
    [aux_sym_switch_block_token1] = ACTIONS(1214),
    [anon_sym_PLUS] = ACTIONS(1214),
    [anon_sym_DASH] = ACTIONS(1214),
    [anon_sym_TILDE] = ACTIONS(1212),
    [anon_sym_BANG] = ACTIONS(1212),
    [anon_sym_AT] = ACTIONS(1212),
    [aux_sym_clone_expression_token1] = ACTIONS(1214),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1214),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1214),
    [anon_sym_DASH_DASH] = ACTIONS(1212),
    [anon_sym_PLUS_PLUS] = ACTIONS(1212),
    [aux_sym__list_destructing_token1] = ACTIONS(1214),
    [anon_sym_LBRACK] = ACTIONS(1212),
    [anon_sym_self] = ACTIONS(1214),
    [anon_sym_parent] = ACTIONS(1214),
    [aux_sym__argument_name_token1] = ACTIONS(1214),
    [aux_sym__argument_name_token2] = ACTIONS(1214),
    [anon_sym_POUND_LBRACK] = ACTIONS(1212),
    [aux_sym_encapsed_string_token1] = ACTIONS(1212),
    [anon_sym_DQUOTE] = ACTIONS(1212),
    [aux_sym_string_token1] = ACTIONS(1212),
    [anon_sym_SQUOTE] = ACTIONS(1212),
    [anon_sym_LT_LT_LT] = ACTIONS(1212),
    [anon_sym_BQUOTE] = ACTIONS(1212),
    [anon_sym_DOLLAR] = ACTIONS(1212),
    [aux_sym_yield_expression_token1] = ACTIONS(1214),
    [aux_sym_include_expression_token1] = ACTIONS(1214),
    [aux_sym_include_once_expression_token1] = ACTIONS(1214),
    [aux_sym_require_expression_token1] = ACTIONS(1214),
    [aux_sym_require_once_expression_token1] = ACTIONS(1214),
    [sym_comment] = ACTIONS(5),
  },
  [496] = {
    [sym_text_interpolation] = STATE(496),
    [ts_builtin_sym_end] = ACTIONS(1216),
    [sym_name] = ACTIONS(1218),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1216),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1218),
    [aux_sym_global_declaration_token1] = ACTIONS(1218),
    [aux_sym_namespace_definition_token1] = ACTIONS(1218),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1218),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1218),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1218),
    [anon_sym_BSLASH] = ACTIONS(1216),
    [anon_sym_LBRACE] = ACTIONS(1216),
    [anon_sym_RBRACE] = ACTIONS(1216),
    [aux_sym_trait_declaration_token1] = ACTIONS(1218),
    [aux_sym_interface_declaration_token1] = ACTIONS(1218),
    [aux_sym_enum_declaration_token1] = ACTIONS(1218),
    [aux_sym_enum_case_token1] = ACTIONS(1218),
    [aux_sym_class_declaration_token1] = ACTIONS(1218),
    [aux_sym_final_modifier_token1] = ACTIONS(1218),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1218),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1218),
    [sym_var_modifier] = ACTIONS(1218),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1218),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1218),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1218),
    [anon_sym_LPAREN2] = ACTIONS(1216),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1218),
    [aux_sym_cast_type_token1] = ACTIONS(1218),
    [aux_sym_echo_statement_token1] = ACTIONS(1218),
    [aux_sym_exit_statement_token1] = ACTIONS(1218),
    [anon_sym_unset] = ACTIONS(1218),
    [aux_sym_declare_statement_token1] = ACTIONS(1218),
    [aux_sym_declare_statement_token2] = ACTIONS(1218),
    [sym_float] = ACTIONS(1218),
    [aux_sym_try_statement_token1] = ACTIONS(1218),
    [aux_sym_goto_statement_token1] = ACTIONS(1218),
    [aux_sym_continue_statement_token1] = ACTIONS(1218),
    [aux_sym_break_statement_token1] = ACTIONS(1218),
    [sym_integer] = ACTIONS(1218),
    [aux_sym_return_statement_token1] = ACTIONS(1218),
    [aux_sym_throw_expression_token1] = ACTIONS(1218),
    [aux_sym_while_statement_token1] = ACTIONS(1218),
    [aux_sym_while_statement_token2] = ACTIONS(1218),
    [aux_sym_do_statement_token1] = ACTIONS(1218),
    [aux_sym_for_statement_token1] = ACTIONS(1218),
    [aux_sym_for_statement_token2] = ACTIONS(1218),
    [aux_sym_foreach_statement_token1] = ACTIONS(1218),
    [aux_sym_foreach_statement_token2] = ACTIONS(1218),
    [aux_sym_if_statement_token1] = ACTIONS(1218),
    [aux_sym_if_statement_token2] = ACTIONS(1218),
    [aux_sym_else_if_clause_token1] = ACTIONS(1218),
    [aux_sym_else_clause_token1] = ACTIONS(1218),
    [aux_sym_match_expression_token1] = ACTIONS(1218),
    [aux_sym_match_default_expression_token1] = ACTIONS(1218),
    [aux_sym_switch_statement_token1] = ACTIONS(1218),
    [aux_sym_switch_block_token1] = ACTIONS(1218),
    [anon_sym_PLUS] = ACTIONS(1218),
    [anon_sym_DASH] = ACTIONS(1218),
    [anon_sym_TILDE] = ACTIONS(1216),
    [anon_sym_BANG] = ACTIONS(1216),
    [anon_sym_AT] = ACTIONS(1216),
    [aux_sym_clone_expression_token1] = ACTIONS(1218),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1218),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1218),
    [anon_sym_DASH_DASH] = ACTIONS(1216),
    [anon_sym_PLUS_PLUS] = ACTIONS(1216),
    [aux_sym__list_destructing_token1] = ACTIONS(1218),
    [anon_sym_LBRACK] = ACTIONS(1216),
    [anon_sym_self] = ACTIONS(1218),
    [anon_sym_parent] = ACTIONS(1218),
    [aux_sym__argument_name_token1] = ACTIONS(1218),
    [aux_sym__argument_name_token2] = ACTIONS(1218),
    [anon_sym_POUND_LBRACK] = ACTIONS(1216),
    [aux_sym_encapsed_string_token1] = ACTIONS(1216),
    [anon_sym_DQUOTE] = ACTIONS(1216),
    [aux_sym_string_token1] = ACTIONS(1216),
    [anon_sym_SQUOTE] = ACTIONS(1216),
    [anon_sym_LT_LT_LT] = ACTIONS(1216),
    [anon_sym_BQUOTE] = ACTIONS(1216),
    [anon_sym_DOLLAR] = ACTIONS(1216),
    [aux_sym_yield_expression_token1] = ACTIONS(1218),
    [aux_sym_include_expression_token1] = ACTIONS(1218),
    [aux_sym_include_once_expression_token1] = ACTIONS(1218),
    [aux_sym_require_expression_token1] = ACTIONS(1218),
    [aux_sym_require_once_expression_token1] = ACTIONS(1218),
    [sym_comment] = ACTIONS(5),
  },
  [497] = {
    [sym_text_interpolation] = STATE(497),
    [ts_builtin_sym_end] = ACTIONS(1220),
    [sym_name] = ACTIONS(1222),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1220),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1222),
    [aux_sym_global_declaration_token1] = ACTIONS(1222),
    [aux_sym_namespace_definition_token1] = ACTIONS(1222),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1222),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1222),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1222),
    [anon_sym_BSLASH] = ACTIONS(1220),
    [anon_sym_LBRACE] = ACTIONS(1220),
    [anon_sym_RBRACE] = ACTIONS(1220),
    [aux_sym_trait_declaration_token1] = ACTIONS(1222),
    [aux_sym_interface_declaration_token1] = ACTIONS(1222),
    [aux_sym_enum_declaration_token1] = ACTIONS(1222),
    [aux_sym_enum_case_token1] = ACTIONS(1222),
    [aux_sym_class_declaration_token1] = ACTIONS(1222),
    [aux_sym_final_modifier_token1] = ACTIONS(1222),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1222),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1222),
    [sym_var_modifier] = ACTIONS(1222),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1222),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1222),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1222),
    [anon_sym_LPAREN2] = ACTIONS(1220),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1222),
    [aux_sym_cast_type_token1] = ACTIONS(1222),
    [aux_sym_echo_statement_token1] = ACTIONS(1222),
    [aux_sym_exit_statement_token1] = ACTIONS(1222),
    [anon_sym_unset] = ACTIONS(1222),
    [aux_sym_declare_statement_token1] = ACTIONS(1222),
    [aux_sym_declare_statement_token2] = ACTIONS(1222),
    [sym_float] = ACTIONS(1222),
    [aux_sym_try_statement_token1] = ACTIONS(1222),
    [aux_sym_goto_statement_token1] = ACTIONS(1222),
    [aux_sym_continue_statement_token1] = ACTIONS(1222),
    [aux_sym_break_statement_token1] = ACTIONS(1222),
    [sym_integer] = ACTIONS(1222),
    [aux_sym_return_statement_token1] = ACTIONS(1222),
    [aux_sym_throw_expression_token1] = ACTIONS(1222),
    [aux_sym_while_statement_token1] = ACTIONS(1222),
    [aux_sym_while_statement_token2] = ACTIONS(1222),
    [aux_sym_do_statement_token1] = ACTIONS(1222),
    [aux_sym_for_statement_token1] = ACTIONS(1222),
    [aux_sym_for_statement_token2] = ACTIONS(1222),
    [aux_sym_foreach_statement_token1] = ACTIONS(1222),
    [aux_sym_foreach_statement_token2] = ACTIONS(1222),
    [aux_sym_if_statement_token1] = ACTIONS(1222),
    [aux_sym_if_statement_token2] = ACTIONS(1222),
    [aux_sym_else_if_clause_token1] = ACTIONS(1222),
    [aux_sym_else_clause_token1] = ACTIONS(1222),
    [aux_sym_match_expression_token1] = ACTIONS(1222),
    [aux_sym_match_default_expression_token1] = ACTIONS(1222),
    [aux_sym_switch_statement_token1] = ACTIONS(1222),
    [aux_sym_switch_block_token1] = ACTIONS(1222),
    [anon_sym_PLUS] = ACTIONS(1222),
    [anon_sym_DASH] = ACTIONS(1222),
    [anon_sym_TILDE] = ACTIONS(1220),
    [anon_sym_BANG] = ACTIONS(1220),
    [anon_sym_AT] = ACTIONS(1220),
    [aux_sym_clone_expression_token1] = ACTIONS(1222),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1222),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1222),
    [anon_sym_DASH_DASH] = ACTIONS(1220),
    [anon_sym_PLUS_PLUS] = ACTIONS(1220),
    [aux_sym__list_destructing_token1] = ACTIONS(1222),
    [anon_sym_LBRACK] = ACTIONS(1220),
    [anon_sym_self] = ACTIONS(1222),
    [anon_sym_parent] = ACTIONS(1222),
    [aux_sym__argument_name_token1] = ACTIONS(1222),
    [aux_sym__argument_name_token2] = ACTIONS(1222),
    [anon_sym_POUND_LBRACK] = ACTIONS(1220),
    [aux_sym_encapsed_string_token1] = ACTIONS(1220),
    [anon_sym_DQUOTE] = ACTIONS(1220),
    [aux_sym_string_token1] = ACTIONS(1220),
    [anon_sym_SQUOTE] = ACTIONS(1220),
    [anon_sym_LT_LT_LT] = ACTIONS(1220),
    [anon_sym_BQUOTE] = ACTIONS(1220),
    [anon_sym_DOLLAR] = ACTIONS(1220),
    [aux_sym_yield_expression_token1] = ACTIONS(1222),
    [aux_sym_include_expression_token1] = ACTIONS(1222),
    [aux_sym_include_once_expression_token1] = ACTIONS(1222),
    [aux_sym_require_expression_token1] = ACTIONS(1222),
    [aux_sym_require_once_expression_token1] = ACTIONS(1222),
    [sym_comment] = ACTIONS(5),
  },
  [498] = {
    [sym_text_interpolation] = STATE(498),
    [ts_builtin_sym_end] = ACTIONS(1224),
    [sym_name] = ACTIONS(1226),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1224),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1226),
    [aux_sym_global_declaration_token1] = ACTIONS(1226),
    [aux_sym_namespace_definition_token1] = ACTIONS(1226),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1226),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1226),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1226),
    [anon_sym_BSLASH] = ACTIONS(1224),
    [anon_sym_LBRACE] = ACTIONS(1224),
    [anon_sym_RBRACE] = ACTIONS(1224),
    [aux_sym_trait_declaration_token1] = ACTIONS(1226),
    [aux_sym_interface_declaration_token1] = ACTIONS(1226),
    [aux_sym_enum_declaration_token1] = ACTIONS(1226),
    [aux_sym_enum_case_token1] = ACTIONS(1226),
    [aux_sym_class_declaration_token1] = ACTIONS(1226),
    [aux_sym_final_modifier_token1] = ACTIONS(1226),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1226),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1226),
    [sym_var_modifier] = ACTIONS(1226),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1226),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1226),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1226),
    [anon_sym_LPAREN2] = ACTIONS(1224),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1226),
    [aux_sym_cast_type_token1] = ACTIONS(1226),
    [aux_sym_echo_statement_token1] = ACTIONS(1226),
    [aux_sym_exit_statement_token1] = ACTIONS(1226),
    [anon_sym_unset] = ACTIONS(1226),
    [aux_sym_declare_statement_token1] = ACTIONS(1226),
    [aux_sym_declare_statement_token2] = ACTIONS(1226),
    [sym_float] = ACTIONS(1226),
    [aux_sym_try_statement_token1] = ACTIONS(1226),
    [aux_sym_goto_statement_token1] = ACTIONS(1226),
    [aux_sym_continue_statement_token1] = ACTIONS(1226),
    [aux_sym_break_statement_token1] = ACTIONS(1226),
    [sym_integer] = ACTIONS(1226),
    [aux_sym_return_statement_token1] = ACTIONS(1226),
    [aux_sym_throw_expression_token1] = ACTIONS(1226),
    [aux_sym_while_statement_token1] = ACTIONS(1226),
    [aux_sym_while_statement_token2] = ACTIONS(1226),
    [aux_sym_do_statement_token1] = ACTIONS(1226),
    [aux_sym_for_statement_token1] = ACTIONS(1226),
    [aux_sym_for_statement_token2] = ACTIONS(1226),
    [aux_sym_foreach_statement_token1] = ACTIONS(1226),
    [aux_sym_foreach_statement_token2] = ACTIONS(1226),
    [aux_sym_if_statement_token1] = ACTIONS(1226),
    [aux_sym_if_statement_token2] = ACTIONS(1226),
    [aux_sym_else_if_clause_token1] = ACTIONS(1226),
    [aux_sym_else_clause_token1] = ACTIONS(1226),
    [aux_sym_match_expression_token1] = ACTIONS(1226),
    [aux_sym_match_default_expression_token1] = ACTIONS(1226),
    [aux_sym_switch_statement_token1] = ACTIONS(1226),
    [aux_sym_switch_block_token1] = ACTIONS(1226),
    [anon_sym_PLUS] = ACTIONS(1226),
    [anon_sym_DASH] = ACTIONS(1226),
    [anon_sym_TILDE] = ACTIONS(1224),
    [anon_sym_BANG] = ACTIONS(1224),
    [anon_sym_AT] = ACTIONS(1224),
    [aux_sym_clone_expression_token1] = ACTIONS(1226),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1226),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1226),
    [anon_sym_DASH_DASH] = ACTIONS(1224),
    [anon_sym_PLUS_PLUS] = ACTIONS(1224),
    [aux_sym__list_destructing_token1] = ACTIONS(1226),
    [anon_sym_LBRACK] = ACTIONS(1224),
    [anon_sym_self] = ACTIONS(1226),
    [anon_sym_parent] = ACTIONS(1226),
    [aux_sym__argument_name_token1] = ACTIONS(1226),
    [aux_sym__argument_name_token2] = ACTIONS(1226),
    [anon_sym_POUND_LBRACK] = ACTIONS(1224),
    [aux_sym_encapsed_string_token1] = ACTIONS(1224),
    [anon_sym_DQUOTE] = ACTIONS(1224),
    [aux_sym_string_token1] = ACTIONS(1224),
    [anon_sym_SQUOTE] = ACTIONS(1224),
    [anon_sym_LT_LT_LT] = ACTIONS(1224),
    [anon_sym_BQUOTE] = ACTIONS(1224),
    [anon_sym_DOLLAR] = ACTIONS(1224),
    [aux_sym_yield_expression_token1] = ACTIONS(1226),
    [aux_sym_include_expression_token1] = ACTIONS(1226),
    [aux_sym_include_once_expression_token1] = ACTIONS(1226),
    [aux_sym_require_expression_token1] = ACTIONS(1226),
    [aux_sym_require_once_expression_token1] = ACTIONS(1226),
    [sym_comment] = ACTIONS(5),
  },
  [499] = {
    [sym_text_interpolation] = STATE(499),
    [ts_builtin_sym_end] = ACTIONS(1228),
    [sym_name] = ACTIONS(1230),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1228),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1230),
    [aux_sym_global_declaration_token1] = ACTIONS(1230),
    [aux_sym_namespace_definition_token1] = ACTIONS(1230),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1230),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1230),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1230),
    [anon_sym_BSLASH] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [anon_sym_RBRACE] = ACTIONS(1228),
    [aux_sym_trait_declaration_token1] = ACTIONS(1230),
    [aux_sym_interface_declaration_token1] = ACTIONS(1230),
    [aux_sym_enum_declaration_token1] = ACTIONS(1230),
    [aux_sym_enum_case_token1] = ACTIONS(1230),
    [aux_sym_class_declaration_token1] = ACTIONS(1230),
    [aux_sym_final_modifier_token1] = ACTIONS(1230),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1230),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1230),
    [sym_var_modifier] = ACTIONS(1230),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1230),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1230),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1230),
    [anon_sym_LPAREN2] = ACTIONS(1228),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1230),
    [aux_sym_cast_type_token1] = ACTIONS(1230),
    [aux_sym_echo_statement_token1] = ACTIONS(1230),
    [aux_sym_exit_statement_token1] = ACTIONS(1230),
    [anon_sym_unset] = ACTIONS(1230),
    [aux_sym_declare_statement_token1] = ACTIONS(1230),
    [aux_sym_declare_statement_token2] = ACTIONS(1230),
    [sym_float] = ACTIONS(1230),
    [aux_sym_try_statement_token1] = ACTIONS(1230),
    [aux_sym_goto_statement_token1] = ACTIONS(1230),
    [aux_sym_continue_statement_token1] = ACTIONS(1230),
    [aux_sym_break_statement_token1] = ACTIONS(1230),
    [sym_integer] = ACTIONS(1230),
    [aux_sym_return_statement_token1] = ACTIONS(1230),
    [aux_sym_throw_expression_token1] = ACTIONS(1230),
    [aux_sym_while_statement_token1] = ACTIONS(1230),
    [aux_sym_while_statement_token2] = ACTIONS(1230),
    [aux_sym_do_statement_token1] = ACTIONS(1230),
    [aux_sym_for_statement_token1] = ACTIONS(1230),
    [aux_sym_for_statement_token2] = ACTIONS(1230),
    [aux_sym_foreach_statement_token1] = ACTIONS(1230),
    [aux_sym_foreach_statement_token2] = ACTIONS(1230),
    [aux_sym_if_statement_token1] = ACTIONS(1230),
    [aux_sym_if_statement_token2] = ACTIONS(1230),
    [aux_sym_else_if_clause_token1] = ACTIONS(1230),
    [aux_sym_else_clause_token1] = ACTIONS(1230),
    [aux_sym_match_expression_token1] = ACTIONS(1230),
    [aux_sym_match_default_expression_token1] = ACTIONS(1230),
    [aux_sym_switch_statement_token1] = ACTIONS(1230),
    [aux_sym_switch_block_token1] = ACTIONS(1230),
    [anon_sym_PLUS] = ACTIONS(1230),
    [anon_sym_DASH] = ACTIONS(1230),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_BANG] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1228),
    [aux_sym_clone_expression_token1] = ACTIONS(1230),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1230),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1230),
    [anon_sym_DASH_DASH] = ACTIONS(1228),
    [anon_sym_PLUS_PLUS] = ACTIONS(1228),
    [aux_sym__list_destructing_token1] = ACTIONS(1230),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_self] = ACTIONS(1230),
    [anon_sym_parent] = ACTIONS(1230),
    [aux_sym__argument_name_token1] = ACTIONS(1230),
    [aux_sym__argument_name_token2] = ACTIONS(1230),
    [anon_sym_POUND_LBRACK] = ACTIONS(1228),
    [aux_sym_encapsed_string_token1] = ACTIONS(1228),
    [anon_sym_DQUOTE] = ACTIONS(1228),
    [aux_sym_string_token1] = ACTIONS(1228),
    [anon_sym_SQUOTE] = ACTIONS(1228),
    [anon_sym_LT_LT_LT] = ACTIONS(1228),
    [anon_sym_BQUOTE] = ACTIONS(1228),
    [anon_sym_DOLLAR] = ACTIONS(1228),
    [aux_sym_yield_expression_token1] = ACTIONS(1230),
    [aux_sym_include_expression_token1] = ACTIONS(1230),
    [aux_sym_include_once_expression_token1] = ACTIONS(1230),
    [aux_sym_require_expression_token1] = ACTIONS(1230),
    [aux_sym_require_once_expression_token1] = ACTIONS(1230),
    [sym_comment] = ACTIONS(5),
  },
  [500] = {
    [sym_text_interpolation] = STATE(500),
    [ts_builtin_sym_end] = ACTIONS(1232),
    [sym_name] = ACTIONS(1234),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1232),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1234),
    [aux_sym_global_declaration_token1] = ACTIONS(1234),
    [aux_sym_namespace_definition_token1] = ACTIONS(1234),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1234),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1234),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1234),
    [anon_sym_BSLASH] = ACTIONS(1232),
    [anon_sym_LBRACE] = ACTIONS(1232),
    [anon_sym_RBRACE] = ACTIONS(1232),
    [aux_sym_trait_declaration_token1] = ACTIONS(1234),
    [aux_sym_interface_declaration_token1] = ACTIONS(1234),
    [aux_sym_enum_declaration_token1] = ACTIONS(1234),
    [aux_sym_enum_case_token1] = ACTIONS(1234),
    [aux_sym_class_declaration_token1] = ACTIONS(1234),
    [aux_sym_final_modifier_token1] = ACTIONS(1234),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1234),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1234),
    [sym_var_modifier] = ACTIONS(1234),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1234),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1234),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1234),
    [anon_sym_LPAREN2] = ACTIONS(1232),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1234),
    [aux_sym_cast_type_token1] = ACTIONS(1234),
    [aux_sym_echo_statement_token1] = ACTIONS(1234),
    [aux_sym_exit_statement_token1] = ACTIONS(1234),
    [anon_sym_unset] = ACTIONS(1234),
    [aux_sym_declare_statement_token1] = ACTIONS(1234),
    [aux_sym_declare_statement_token2] = ACTIONS(1234),
    [sym_float] = ACTIONS(1234),
    [aux_sym_try_statement_token1] = ACTIONS(1234),
    [aux_sym_goto_statement_token1] = ACTIONS(1234),
    [aux_sym_continue_statement_token1] = ACTIONS(1234),
    [aux_sym_break_statement_token1] = ACTIONS(1234),
    [sym_integer] = ACTIONS(1234),
    [aux_sym_return_statement_token1] = ACTIONS(1234),
    [aux_sym_throw_expression_token1] = ACTIONS(1234),
    [aux_sym_while_statement_token1] = ACTIONS(1234),
    [aux_sym_while_statement_token2] = ACTIONS(1234),
    [aux_sym_do_statement_token1] = ACTIONS(1234),
    [aux_sym_for_statement_token1] = ACTIONS(1234),
    [aux_sym_for_statement_token2] = ACTIONS(1234),
    [aux_sym_foreach_statement_token1] = ACTIONS(1234),
    [aux_sym_foreach_statement_token2] = ACTIONS(1234),
    [aux_sym_if_statement_token1] = ACTIONS(1234),
    [aux_sym_if_statement_token2] = ACTIONS(1234),
    [aux_sym_else_if_clause_token1] = ACTIONS(1234),
    [aux_sym_else_clause_token1] = ACTIONS(1234),
    [aux_sym_match_expression_token1] = ACTIONS(1234),
    [aux_sym_match_default_expression_token1] = ACTIONS(1234),
    [aux_sym_switch_statement_token1] = ACTIONS(1234),
    [aux_sym_switch_block_token1] = ACTIONS(1234),
    [anon_sym_PLUS] = ACTIONS(1234),
    [anon_sym_DASH] = ACTIONS(1234),
    [anon_sym_TILDE] = ACTIONS(1232),
    [anon_sym_BANG] = ACTIONS(1232),
    [anon_sym_AT] = ACTIONS(1232),
    [aux_sym_clone_expression_token1] = ACTIONS(1234),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1234),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1234),
    [anon_sym_DASH_DASH] = ACTIONS(1232),
    [anon_sym_PLUS_PLUS] = ACTIONS(1232),
    [aux_sym__list_destructing_token1] = ACTIONS(1234),
    [anon_sym_LBRACK] = ACTIONS(1232),
    [anon_sym_self] = ACTIONS(1234),
    [anon_sym_parent] = ACTIONS(1234),
    [aux_sym__argument_name_token1] = ACTIONS(1234),
    [aux_sym__argument_name_token2] = ACTIONS(1234),
    [anon_sym_POUND_LBRACK] = ACTIONS(1232),
    [aux_sym_encapsed_string_token1] = ACTIONS(1232),
    [anon_sym_DQUOTE] = ACTIONS(1232),
    [aux_sym_string_token1] = ACTIONS(1232),
    [anon_sym_SQUOTE] = ACTIONS(1232),
    [anon_sym_LT_LT_LT] = ACTIONS(1232),
    [anon_sym_BQUOTE] = ACTIONS(1232),
    [anon_sym_DOLLAR] = ACTIONS(1232),
    [aux_sym_yield_expression_token1] = ACTIONS(1234),
    [aux_sym_include_expression_token1] = ACTIONS(1234),
    [aux_sym_include_once_expression_token1] = ACTIONS(1234),
    [aux_sym_require_expression_token1] = ACTIONS(1234),
    [aux_sym_require_once_expression_token1] = ACTIONS(1234),
    [sym_comment] = ACTIONS(5),
  },
  [501] = {
    [sym_text_interpolation] = STATE(501),
    [ts_builtin_sym_end] = ACTIONS(1236),
    [sym_name] = ACTIONS(1238),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1236),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1238),
    [aux_sym_global_declaration_token1] = ACTIONS(1238),
    [aux_sym_namespace_definition_token1] = ACTIONS(1238),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1238),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1238),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1238),
    [anon_sym_BSLASH] = ACTIONS(1236),
    [anon_sym_LBRACE] = ACTIONS(1236),
    [anon_sym_RBRACE] = ACTIONS(1236),
    [aux_sym_trait_declaration_token1] = ACTIONS(1238),
    [aux_sym_interface_declaration_token1] = ACTIONS(1238),
    [aux_sym_enum_declaration_token1] = ACTIONS(1238),
    [aux_sym_enum_case_token1] = ACTIONS(1238),
    [aux_sym_class_declaration_token1] = ACTIONS(1238),
    [aux_sym_final_modifier_token1] = ACTIONS(1238),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1238),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1238),
    [sym_var_modifier] = ACTIONS(1238),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1238),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1238),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1238),
    [anon_sym_LPAREN2] = ACTIONS(1236),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1238),
    [aux_sym_cast_type_token1] = ACTIONS(1238),
    [aux_sym_echo_statement_token1] = ACTIONS(1238),
    [aux_sym_exit_statement_token1] = ACTIONS(1238),
    [anon_sym_unset] = ACTIONS(1238),
    [aux_sym_declare_statement_token1] = ACTIONS(1238),
    [aux_sym_declare_statement_token2] = ACTIONS(1238),
    [sym_float] = ACTIONS(1238),
    [aux_sym_try_statement_token1] = ACTIONS(1238),
    [aux_sym_goto_statement_token1] = ACTIONS(1238),
    [aux_sym_continue_statement_token1] = ACTIONS(1238),
    [aux_sym_break_statement_token1] = ACTIONS(1238),
    [sym_integer] = ACTIONS(1238),
    [aux_sym_return_statement_token1] = ACTIONS(1238),
    [aux_sym_throw_expression_token1] = ACTIONS(1238),
    [aux_sym_while_statement_token1] = ACTIONS(1238),
    [aux_sym_while_statement_token2] = ACTIONS(1238),
    [aux_sym_do_statement_token1] = ACTIONS(1238),
    [aux_sym_for_statement_token1] = ACTIONS(1238),
    [aux_sym_for_statement_token2] = ACTIONS(1238),
    [aux_sym_foreach_statement_token1] = ACTIONS(1238),
    [aux_sym_foreach_statement_token2] = ACTIONS(1238),
    [aux_sym_if_statement_token1] = ACTIONS(1238),
    [aux_sym_if_statement_token2] = ACTIONS(1238),
    [aux_sym_else_if_clause_token1] = ACTIONS(1238),
    [aux_sym_else_clause_token1] = ACTIONS(1238),
    [aux_sym_match_expression_token1] = ACTIONS(1238),
    [aux_sym_match_default_expression_token1] = ACTIONS(1238),
    [aux_sym_switch_statement_token1] = ACTIONS(1238),
    [aux_sym_switch_block_token1] = ACTIONS(1238),
    [anon_sym_PLUS] = ACTIONS(1238),
    [anon_sym_DASH] = ACTIONS(1238),
    [anon_sym_TILDE] = ACTIONS(1236),
    [anon_sym_BANG] = ACTIONS(1236),
    [anon_sym_AT] = ACTIONS(1236),
    [aux_sym_clone_expression_token1] = ACTIONS(1238),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1238),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1238),
    [anon_sym_DASH_DASH] = ACTIONS(1236),
    [anon_sym_PLUS_PLUS] = ACTIONS(1236),
    [aux_sym__list_destructing_token1] = ACTIONS(1238),
    [anon_sym_LBRACK] = ACTIONS(1236),
    [anon_sym_self] = ACTIONS(1238),
    [anon_sym_parent] = ACTIONS(1238),
    [aux_sym__argument_name_token1] = ACTIONS(1238),
    [aux_sym__argument_name_token2] = ACTIONS(1238),
    [anon_sym_POUND_LBRACK] = ACTIONS(1236),
    [aux_sym_encapsed_string_token1] = ACTIONS(1236),
    [anon_sym_DQUOTE] = ACTIONS(1236),
    [aux_sym_string_token1] = ACTIONS(1236),
    [anon_sym_SQUOTE] = ACTIONS(1236),
    [anon_sym_LT_LT_LT] = ACTIONS(1236),
    [anon_sym_BQUOTE] = ACTIONS(1236),
    [anon_sym_DOLLAR] = ACTIONS(1236),
    [aux_sym_yield_expression_token1] = ACTIONS(1238),
    [aux_sym_include_expression_token1] = ACTIONS(1238),
    [aux_sym_include_once_expression_token1] = ACTIONS(1238),
    [aux_sym_require_expression_token1] = ACTIONS(1238),
    [aux_sym_require_once_expression_token1] = ACTIONS(1238),
    [sym_comment] = ACTIONS(5),
  },
  [502] = {
    [sym_text_interpolation] = STATE(502),
    [ts_builtin_sym_end] = ACTIONS(1240),
    [sym_name] = ACTIONS(1242),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1240),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1242),
    [aux_sym_global_declaration_token1] = ACTIONS(1242),
    [aux_sym_namespace_definition_token1] = ACTIONS(1242),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1242),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1242),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1242),
    [anon_sym_BSLASH] = ACTIONS(1240),
    [anon_sym_LBRACE] = ACTIONS(1240),
    [anon_sym_RBRACE] = ACTIONS(1240),
    [aux_sym_trait_declaration_token1] = ACTIONS(1242),
    [aux_sym_interface_declaration_token1] = ACTIONS(1242),
    [aux_sym_enum_declaration_token1] = ACTIONS(1242),
    [aux_sym_enum_case_token1] = ACTIONS(1242),
    [aux_sym_class_declaration_token1] = ACTIONS(1242),
    [aux_sym_final_modifier_token1] = ACTIONS(1242),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1242),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1242),
    [sym_var_modifier] = ACTIONS(1242),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1242),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1242),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1242),
    [anon_sym_LPAREN2] = ACTIONS(1240),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1242),
    [aux_sym_cast_type_token1] = ACTIONS(1242),
    [aux_sym_echo_statement_token1] = ACTIONS(1242),
    [aux_sym_exit_statement_token1] = ACTIONS(1242),
    [anon_sym_unset] = ACTIONS(1242),
    [aux_sym_declare_statement_token1] = ACTIONS(1242),
    [aux_sym_declare_statement_token2] = ACTIONS(1242),
    [sym_float] = ACTIONS(1242),
    [aux_sym_try_statement_token1] = ACTIONS(1242),
    [aux_sym_goto_statement_token1] = ACTIONS(1242),
    [aux_sym_continue_statement_token1] = ACTIONS(1242),
    [aux_sym_break_statement_token1] = ACTIONS(1242),
    [sym_integer] = ACTIONS(1242),
    [aux_sym_return_statement_token1] = ACTIONS(1242),
    [aux_sym_throw_expression_token1] = ACTIONS(1242),
    [aux_sym_while_statement_token1] = ACTIONS(1242),
    [aux_sym_while_statement_token2] = ACTIONS(1242),
    [aux_sym_do_statement_token1] = ACTIONS(1242),
    [aux_sym_for_statement_token1] = ACTIONS(1242),
    [aux_sym_for_statement_token2] = ACTIONS(1242),
    [aux_sym_foreach_statement_token1] = ACTIONS(1242),
    [aux_sym_foreach_statement_token2] = ACTIONS(1242),
    [aux_sym_if_statement_token1] = ACTIONS(1242),
    [aux_sym_if_statement_token2] = ACTIONS(1242),
    [aux_sym_else_if_clause_token1] = ACTIONS(1242),
    [aux_sym_else_clause_token1] = ACTIONS(1242),
    [aux_sym_match_expression_token1] = ACTIONS(1242),
    [aux_sym_match_default_expression_token1] = ACTIONS(1242),
    [aux_sym_switch_statement_token1] = ACTIONS(1242),
    [aux_sym_switch_block_token1] = ACTIONS(1242),
    [anon_sym_PLUS] = ACTIONS(1242),
    [anon_sym_DASH] = ACTIONS(1242),
    [anon_sym_TILDE] = ACTIONS(1240),
    [anon_sym_BANG] = ACTIONS(1240),
    [anon_sym_AT] = ACTIONS(1240),
    [aux_sym_clone_expression_token1] = ACTIONS(1242),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1242),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1242),
    [anon_sym_DASH_DASH] = ACTIONS(1240),
    [anon_sym_PLUS_PLUS] = ACTIONS(1240),
    [aux_sym__list_destructing_token1] = ACTIONS(1242),
    [anon_sym_LBRACK] = ACTIONS(1240),
    [anon_sym_self] = ACTIONS(1242),
    [anon_sym_parent] = ACTIONS(1242),
    [aux_sym__argument_name_token1] = ACTIONS(1242),
    [aux_sym__argument_name_token2] = ACTIONS(1242),
    [anon_sym_POUND_LBRACK] = ACTIONS(1240),
    [aux_sym_encapsed_string_token1] = ACTIONS(1240),
    [anon_sym_DQUOTE] = ACTIONS(1240),
    [aux_sym_string_token1] = ACTIONS(1240),
    [anon_sym_SQUOTE] = ACTIONS(1240),
    [anon_sym_LT_LT_LT] = ACTIONS(1240),
    [anon_sym_BQUOTE] = ACTIONS(1240),
    [anon_sym_DOLLAR] = ACTIONS(1240),
    [aux_sym_yield_expression_token1] = ACTIONS(1242),
    [aux_sym_include_expression_token1] = ACTIONS(1242),
    [aux_sym_include_once_expression_token1] = ACTIONS(1242),
    [aux_sym_require_expression_token1] = ACTIONS(1242),
    [aux_sym_require_once_expression_token1] = ACTIONS(1242),
    [sym_comment] = ACTIONS(5),
  },
  [503] = {
    [sym_text_interpolation] = STATE(503),
    [ts_builtin_sym_end] = ACTIONS(1244),
    [sym_name] = ACTIONS(1246),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1244),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1246),
    [aux_sym_global_declaration_token1] = ACTIONS(1246),
    [aux_sym_namespace_definition_token1] = ACTIONS(1246),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1246),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1246),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1246),
    [anon_sym_BSLASH] = ACTIONS(1244),
    [anon_sym_LBRACE] = ACTIONS(1244),
    [anon_sym_RBRACE] = ACTIONS(1244),
    [aux_sym_trait_declaration_token1] = ACTIONS(1246),
    [aux_sym_interface_declaration_token1] = ACTIONS(1246),
    [aux_sym_enum_declaration_token1] = ACTIONS(1246),
    [aux_sym_enum_case_token1] = ACTIONS(1246),
    [aux_sym_class_declaration_token1] = ACTIONS(1246),
    [aux_sym_final_modifier_token1] = ACTIONS(1246),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1246),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1246),
    [sym_var_modifier] = ACTIONS(1246),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1246),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1246),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1246),
    [anon_sym_LPAREN2] = ACTIONS(1244),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1246),
    [aux_sym_cast_type_token1] = ACTIONS(1246),
    [aux_sym_echo_statement_token1] = ACTIONS(1246),
    [aux_sym_exit_statement_token1] = ACTIONS(1246),
    [anon_sym_unset] = ACTIONS(1246),
    [aux_sym_declare_statement_token1] = ACTIONS(1246),
    [aux_sym_declare_statement_token2] = ACTIONS(1246),
    [sym_float] = ACTIONS(1246),
    [aux_sym_try_statement_token1] = ACTIONS(1246),
    [aux_sym_goto_statement_token1] = ACTIONS(1246),
    [aux_sym_continue_statement_token1] = ACTIONS(1246),
    [aux_sym_break_statement_token1] = ACTIONS(1246),
    [sym_integer] = ACTIONS(1246),
    [aux_sym_return_statement_token1] = ACTIONS(1246),
    [aux_sym_throw_expression_token1] = ACTIONS(1246),
    [aux_sym_while_statement_token1] = ACTIONS(1246),
    [aux_sym_while_statement_token2] = ACTIONS(1246),
    [aux_sym_do_statement_token1] = ACTIONS(1246),
    [aux_sym_for_statement_token1] = ACTIONS(1246),
    [aux_sym_for_statement_token2] = ACTIONS(1246),
    [aux_sym_foreach_statement_token1] = ACTIONS(1246),
    [aux_sym_foreach_statement_token2] = ACTIONS(1246),
    [aux_sym_if_statement_token1] = ACTIONS(1246),
    [aux_sym_if_statement_token2] = ACTIONS(1246),
    [aux_sym_else_if_clause_token1] = ACTIONS(1246),
    [aux_sym_else_clause_token1] = ACTIONS(1246),
    [aux_sym_match_expression_token1] = ACTIONS(1246),
    [aux_sym_match_default_expression_token1] = ACTIONS(1246),
    [aux_sym_switch_statement_token1] = ACTIONS(1246),
    [aux_sym_switch_block_token1] = ACTIONS(1246),
    [anon_sym_PLUS] = ACTIONS(1246),
    [anon_sym_DASH] = ACTIONS(1246),
    [anon_sym_TILDE] = ACTIONS(1244),
    [anon_sym_BANG] = ACTIONS(1244),
    [anon_sym_AT] = ACTIONS(1244),
    [aux_sym_clone_expression_token1] = ACTIONS(1246),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1246),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1246),
    [anon_sym_DASH_DASH] = ACTIONS(1244),
    [anon_sym_PLUS_PLUS] = ACTIONS(1244),
    [aux_sym__list_destructing_token1] = ACTIONS(1246),
    [anon_sym_LBRACK] = ACTIONS(1244),
    [anon_sym_self] = ACTIONS(1246),
    [anon_sym_parent] = ACTIONS(1246),
    [aux_sym__argument_name_token1] = ACTIONS(1246),
    [aux_sym__argument_name_token2] = ACTIONS(1246),
    [anon_sym_POUND_LBRACK] = ACTIONS(1244),
    [aux_sym_encapsed_string_token1] = ACTIONS(1244),
    [anon_sym_DQUOTE] = ACTIONS(1244),
    [aux_sym_string_token1] = ACTIONS(1244),
    [anon_sym_SQUOTE] = ACTIONS(1244),
    [anon_sym_LT_LT_LT] = ACTIONS(1244),
    [anon_sym_BQUOTE] = ACTIONS(1244),
    [anon_sym_DOLLAR] = ACTIONS(1244),
    [aux_sym_yield_expression_token1] = ACTIONS(1246),
    [aux_sym_include_expression_token1] = ACTIONS(1246),
    [aux_sym_include_once_expression_token1] = ACTIONS(1246),
    [aux_sym_require_expression_token1] = ACTIONS(1246),
    [aux_sym_require_once_expression_token1] = ACTIONS(1246),
    [sym_comment] = ACTIONS(5),
  },
  [504] = {
    [sym_text_interpolation] = STATE(504),
    [ts_builtin_sym_end] = ACTIONS(1248),
    [sym_name] = ACTIONS(1250),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1248),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1250),
    [aux_sym_global_declaration_token1] = ACTIONS(1250),
    [aux_sym_namespace_definition_token1] = ACTIONS(1250),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1250),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1250),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1250),
    [anon_sym_BSLASH] = ACTIONS(1248),
    [anon_sym_LBRACE] = ACTIONS(1248),
    [anon_sym_RBRACE] = ACTIONS(1248),
    [aux_sym_trait_declaration_token1] = ACTIONS(1250),
    [aux_sym_interface_declaration_token1] = ACTIONS(1250),
    [aux_sym_enum_declaration_token1] = ACTIONS(1250),
    [aux_sym_enum_case_token1] = ACTIONS(1250),
    [aux_sym_class_declaration_token1] = ACTIONS(1250),
    [aux_sym_final_modifier_token1] = ACTIONS(1250),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1250),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1250),
    [sym_var_modifier] = ACTIONS(1250),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1250),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1250),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1250),
    [anon_sym_LPAREN2] = ACTIONS(1248),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1250),
    [aux_sym_cast_type_token1] = ACTIONS(1250),
    [aux_sym_echo_statement_token1] = ACTIONS(1250),
    [aux_sym_exit_statement_token1] = ACTIONS(1250),
    [anon_sym_unset] = ACTIONS(1250),
    [aux_sym_declare_statement_token1] = ACTIONS(1250),
    [aux_sym_declare_statement_token2] = ACTIONS(1250),
    [sym_float] = ACTIONS(1250),
    [aux_sym_try_statement_token1] = ACTIONS(1250),
    [aux_sym_goto_statement_token1] = ACTIONS(1250),
    [aux_sym_continue_statement_token1] = ACTIONS(1250),
    [aux_sym_break_statement_token1] = ACTIONS(1250),
    [sym_integer] = ACTIONS(1250),
    [aux_sym_return_statement_token1] = ACTIONS(1250),
    [aux_sym_throw_expression_token1] = ACTIONS(1250),
    [aux_sym_while_statement_token1] = ACTIONS(1250),
    [aux_sym_while_statement_token2] = ACTIONS(1250),
    [aux_sym_do_statement_token1] = ACTIONS(1250),
    [aux_sym_for_statement_token1] = ACTIONS(1250),
    [aux_sym_for_statement_token2] = ACTIONS(1250),
    [aux_sym_foreach_statement_token1] = ACTIONS(1250),
    [aux_sym_foreach_statement_token2] = ACTIONS(1250),
    [aux_sym_if_statement_token1] = ACTIONS(1250),
    [aux_sym_if_statement_token2] = ACTIONS(1250),
    [aux_sym_else_if_clause_token1] = ACTIONS(1250),
    [aux_sym_else_clause_token1] = ACTIONS(1250),
    [aux_sym_match_expression_token1] = ACTIONS(1250),
    [aux_sym_match_default_expression_token1] = ACTIONS(1250),
    [aux_sym_switch_statement_token1] = ACTIONS(1250),
    [aux_sym_switch_block_token1] = ACTIONS(1250),
    [anon_sym_PLUS] = ACTIONS(1250),
    [anon_sym_DASH] = ACTIONS(1250),
    [anon_sym_TILDE] = ACTIONS(1248),
    [anon_sym_BANG] = ACTIONS(1248),
    [anon_sym_AT] = ACTIONS(1248),
    [aux_sym_clone_expression_token1] = ACTIONS(1250),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1250),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1250),
    [anon_sym_DASH_DASH] = ACTIONS(1248),
    [anon_sym_PLUS_PLUS] = ACTIONS(1248),
    [aux_sym__list_destructing_token1] = ACTIONS(1250),
    [anon_sym_LBRACK] = ACTIONS(1248),
    [anon_sym_self] = ACTIONS(1250),
    [anon_sym_parent] = ACTIONS(1250),
    [aux_sym__argument_name_token1] = ACTIONS(1250),
    [aux_sym__argument_name_token2] = ACTIONS(1250),
    [anon_sym_POUND_LBRACK] = ACTIONS(1248),
    [aux_sym_encapsed_string_token1] = ACTIONS(1248),
    [anon_sym_DQUOTE] = ACTIONS(1248),
    [aux_sym_string_token1] = ACTIONS(1248),
    [anon_sym_SQUOTE] = ACTIONS(1248),
    [anon_sym_LT_LT_LT] = ACTIONS(1248),
    [anon_sym_BQUOTE] = ACTIONS(1248),
    [anon_sym_DOLLAR] = ACTIONS(1248),
    [aux_sym_yield_expression_token1] = ACTIONS(1250),
    [aux_sym_include_expression_token1] = ACTIONS(1250),
    [aux_sym_include_once_expression_token1] = ACTIONS(1250),
    [aux_sym_require_expression_token1] = ACTIONS(1250),
    [aux_sym_require_once_expression_token1] = ACTIONS(1250),
    [sym_comment] = ACTIONS(5),
  },
  [505] = {
    [sym_text_interpolation] = STATE(505),
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_name] = ACTIONS(1254),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1252),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1254),
    [aux_sym_global_declaration_token1] = ACTIONS(1254),
    [aux_sym_namespace_definition_token1] = ACTIONS(1254),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1254),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1254),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1254),
    [anon_sym_BSLASH] = ACTIONS(1252),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_trait_declaration_token1] = ACTIONS(1254),
    [aux_sym_interface_declaration_token1] = ACTIONS(1254),
    [aux_sym_enum_declaration_token1] = ACTIONS(1254),
    [aux_sym_enum_case_token1] = ACTIONS(1254),
    [aux_sym_class_declaration_token1] = ACTIONS(1254),
    [aux_sym_final_modifier_token1] = ACTIONS(1254),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1254),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1254),
    [sym_var_modifier] = ACTIONS(1254),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1254),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1254),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1254),
    [anon_sym_LPAREN2] = ACTIONS(1252),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1254),
    [aux_sym_cast_type_token1] = ACTIONS(1254),
    [aux_sym_echo_statement_token1] = ACTIONS(1254),
    [aux_sym_exit_statement_token1] = ACTIONS(1254),
    [anon_sym_unset] = ACTIONS(1254),
    [aux_sym_declare_statement_token1] = ACTIONS(1254),
    [aux_sym_declare_statement_token2] = ACTIONS(1254),
    [sym_float] = ACTIONS(1254),
    [aux_sym_try_statement_token1] = ACTIONS(1254),
    [aux_sym_goto_statement_token1] = ACTIONS(1254),
    [aux_sym_continue_statement_token1] = ACTIONS(1254),
    [aux_sym_break_statement_token1] = ACTIONS(1254),
    [sym_integer] = ACTIONS(1254),
    [aux_sym_return_statement_token1] = ACTIONS(1254),
    [aux_sym_throw_expression_token1] = ACTIONS(1254),
    [aux_sym_while_statement_token1] = ACTIONS(1254),
    [aux_sym_while_statement_token2] = ACTIONS(1254),
    [aux_sym_do_statement_token1] = ACTIONS(1254),
    [aux_sym_for_statement_token1] = ACTIONS(1254),
    [aux_sym_for_statement_token2] = ACTIONS(1254),
    [aux_sym_foreach_statement_token1] = ACTIONS(1254),
    [aux_sym_foreach_statement_token2] = ACTIONS(1254),
    [aux_sym_if_statement_token1] = ACTIONS(1254),
    [aux_sym_if_statement_token2] = ACTIONS(1254),
    [aux_sym_else_if_clause_token1] = ACTIONS(1254),
    [aux_sym_else_clause_token1] = ACTIONS(1254),
    [aux_sym_match_expression_token1] = ACTIONS(1254),
    [aux_sym_match_default_expression_token1] = ACTIONS(1254),
    [aux_sym_switch_statement_token1] = ACTIONS(1254),
    [aux_sym_switch_block_token1] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1252),
    [anon_sym_AT] = ACTIONS(1252),
    [aux_sym_clone_expression_token1] = ACTIONS(1254),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1254),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1254),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [aux_sym__list_destructing_token1] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [anon_sym_self] = ACTIONS(1254),
    [anon_sym_parent] = ACTIONS(1254),
    [aux_sym__argument_name_token1] = ACTIONS(1254),
    [aux_sym__argument_name_token2] = ACTIONS(1254),
    [anon_sym_POUND_LBRACK] = ACTIONS(1252),
    [aux_sym_encapsed_string_token1] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1252),
    [aux_sym_string_token1] = ACTIONS(1252),
    [anon_sym_SQUOTE] = ACTIONS(1252),
    [anon_sym_LT_LT_LT] = ACTIONS(1252),
    [anon_sym_BQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR] = ACTIONS(1252),
    [aux_sym_yield_expression_token1] = ACTIONS(1254),
    [aux_sym_include_expression_token1] = ACTIONS(1254),
    [aux_sym_include_once_expression_token1] = ACTIONS(1254),
    [aux_sym_require_expression_token1] = ACTIONS(1254),
    [aux_sym_require_once_expression_token1] = ACTIONS(1254),
    [sym_comment] = ACTIONS(5),
  },
  [506] = {
    [sym_text_interpolation] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(1256),
    [sym_name] = ACTIONS(1258),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1256),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1258),
    [aux_sym_global_declaration_token1] = ACTIONS(1258),
    [aux_sym_namespace_definition_token1] = ACTIONS(1258),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1258),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1258),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1258),
    [anon_sym_BSLASH] = ACTIONS(1256),
    [anon_sym_LBRACE] = ACTIONS(1256),
    [anon_sym_RBRACE] = ACTIONS(1256),
    [aux_sym_trait_declaration_token1] = ACTIONS(1258),
    [aux_sym_interface_declaration_token1] = ACTIONS(1258),
    [aux_sym_enum_declaration_token1] = ACTIONS(1258),
    [aux_sym_enum_case_token1] = ACTIONS(1258),
    [aux_sym_class_declaration_token1] = ACTIONS(1258),
    [aux_sym_final_modifier_token1] = ACTIONS(1258),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1258),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1258),
    [sym_var_modifier] = ACTIONS(1258),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1258),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1258),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1258),
    [anon_sym_LPAREN2] = ACTIONS(1256),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1258),
    [aux_sym_cast_type_token1] = ACTIONS(1258),
    [aux_sym_echo_statement_token1] = ACTIONS(1258),
    [aux_sym_exit_statement_token1] = ACTIONS(1258),
    [anon_sym_unset] = ACTIONS(1258),
    [aux_sym_declare_statement_token1] = ACTIONS(1258),
    [aux_sym_declare_statement_token2] = ACTIONS(1258),
    [sym_float] = ACTIONS(1258),
    [aux_sym_try_statement_token1] = ACTIONS(1258),
    [aux_sym_goto_statement_token1] = ACTIONS(1258),
    [aux_sym_continue_statement_token1] = ACTIONS(1258),
    [aux_sym_break_statement_token1] = ACTIONS(1258),
    [sym_integer] = ACTIONS(1258),
    [aux_sym_return_statement_token1] = ACTIONS(1258),
    [aux_sym_throw_expression_token1] = ACTIONS(1258),
    [aux_sym_while_statement_token1] = ACTIONS(1258),
    [aux_sym_while_statement_token2] = ACTIONS(1258),
    [aux_sym_do_statement_token1] = ACTIONS(1258),
    [aux_sym_for_statement_token1] = ACTIONS(1258),
    [aux_sym_for_statement_token2] = ACTIONS(1258),
    [aux_sym_foreach_statement_token1] = ACTIONS(1258),
    [aux_sym_foreach_statement_token2] = ACTIONS(1258),
    [aux_sym_if_statement_token1] = ACTIONS(1258),
    [aux_sym_if_statement_token2] = ACTIONS(1258),
    [aux_sym_else_if_clause_token1] = ACTIONS(1258),
    [aux_sym_else_clause_token1] = ACTIONS(1258),
    [aux_sym_match_expression_token1] = ACTIONS(1258),
    [aux_sym_match_default_expression_token1] = ACTIONS(1258),
    [aux_sym_switch_statement_token1] = ACTIONS(1258),
    [aux_sym_switch_block_token1] = ACTIONS(1258),
    [anon_sym_PLUS] = ACTIONS(1258),
    [anon_sym_DASH] = ACTIONS(1258),
    [anon_sym_TILDE] = ACTIONS(1256),
    [anon_sym_BANG] = ACTIONS(1256),
    [anon_sym_AT] = ACTIONS(1256),
    [aux_sym_clone_expression_token1] = ACTIONS(1258),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1258),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1258),
    [anon_sym_DASH_DASH] = ACTIONS(1256),
    [anon_sym_PLUS_PLUS] = ACTIONS(1256),
    [aux_sym__list_destructing_token1] = ACTIONS(1258),
    [anon_sym_LBRACK] = ACTIONS(1256),
    [anon_sym_self] = ACTIONS(1258),
    [anon_sym_parent] = ACTIONS(1258),
    [aux_sym__argument_name_token1] = ACTIONS(1258),
    [aux_sym__argument_name_token2] = ACTIONS(1258),
    [anon_sym_POUND_LBRACK] = ACTIONS(1256),
    [aux_sym_encapsed_string_token1] = ACTIONS(1256),
    [anon_sym_DQUOTE] = ACTIONS(1256),
    [aux_sym_string_token1] = ACTIONS(1256),
    [anon_sym_SQUOTE] = ACTIONS(1256),
    [anon_sym_LT_LT_LT] = ACTIONS(1256),
    [anon_sym_BQUOTE] = ACTIONS(1256),
    [anon_sym_DOLLAR] = ACTIONS(1256),
    [aux_sym_yield_expression_token1] = ACTIONS(1258),
    [aux_sym_include_expression_token1] = ACTIONS(1258),
    [aux_sym_include_once_expression_token1] = ACTIONS(1258),
    [aux_sym_require_expression_token1] = ACTIONS(1258),
    [aux_sym_require_once_expression_token1] = ACTIONS(1258),
    [sym_comment] = ACTIONS(5),
  },
  [507] = {
    [sym_text_interpolation] = STATE(507),
    [ts_builtin_sym_end] = ACTIONS(1260),
    [sym_name] = ACTIONS(1262),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1260),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1262),
    [aux_sym_global_declaration_token1] = ACTIONS(1262),
    [aux_sym_namespace_definition_token1] = ACTIONS(1262),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1262),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1262),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1262),
    [anon_sym_BSLASH] = ACTIONS(1260),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [anon_sym_RBRACE] = ACTIONS(1260),
    [aux_sym_trait_declaration_token1] = ACTIONS(1262),
    [aux_sym_interface_declaration_token1] = ACTIONS(1262),
    [aux_sym_enum_declaration_token1] = ACTIONS(1262),
    [aux_sym_enum_case_token1] = ACTIONS(1262),
    [aux_sym_class_declaration_token1] = ACTIONS(1262),
    [aux_sym_final_modifier_token1] = ACTIONS(1262),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1262),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1262),
    [sym_var_modifier] = ACTIONS(1262),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1262),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1262),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1262),
    [anon_sym_LPAREN2] = ACTIONS(1260),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1262),
    [aux_sym_cast_type_token1] = ACTIONS(1262),
    [aux_sym_echo_statement_token1] = ACTIONS(1262),
    [aux_sym_exit_statement_token1] = ACTIONS(1262),
    [anon_sym_unset] = ACTIONS(1262),
    [aux_sym_declare_statement_token1] = ACTIONS(1262),
    [aux_sym_declare_statement_token2] = ACTIONS(1262),
    [sym_float] = ACTIONS(1262),
    [aux_sym_try_statement_token1] = ACTIONS(1262),
    [aux_sym_goto_statement_token1] = ACTIONS(1262),
    [aux_sym_continue_statement_token1] = ACTIONS(1262),
    [aux_sym_break_statement_token1] = ACTIONS(1262),
    [sym_integer] = ACTIONS(1262),
    [aux_sym_return_statement_token1] = ACTIONS(1262),
    [aux_sym_throw_expression_token1] = ACTIONS(1262),
    [aux_sym_while_statement_token1] = ACTIONS(1262),
    [aux_sym_while_statement_token2] = ACTIONS(1262),
    [aux_sym_do_statement_token1] = ACTIONS(1262),
    [aux_sym_for_statement_token1] = ACTIONS(1262),
    [aux_sym_for_statement_token2] = ACTIONS(1262),
    [aux_sym_foreach_statement_token1] = ACTIONS(1262),
    [aux_sym_foreach_statement_token2] = ACTIONS(1262),
    [aux_sym_if_statement_token1] = ACTIONS(1262),
    [aux_sym_if_statement_token2] = ACTIONS(1262),
    [aux_sym_else_if_clause_token1] = ACTIONS(1262),
    [aux_sym_else_clause_token1] = ACTIONS(1262),
    [aux_sym_match_expression_token1] = ACTIONS(1262),
    [aux_sym_match_default_expression_token1] = ACTIONS(1262),
    [aux_sym_switch_statement_token1] = ACTIONS(1262),
    [aux_sym_switch_block_token1] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_BANG] = ACTIONS(1260),
    [anon_sym_AT] = ACTIONS(1260),
    [aux_sym_clone_expression_token1] = ACTIONS(1262),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1262),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1262),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [aux_sym__list_destructing_token1] = ACTIONS(1262),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [anon_sym_self] = ACTIONS(1262),
    [anon_sym_parent] = ACTIONS(1262),
    [aux_sym__argument_name_token1] = ACTIONS(1262),
    [aux_sym__argument_name_token2] = ACTIONS(1262),
    [anon_sym_POUND_LBRACK] = ACTIONS(1260),
    [aux_sym_encapsed_string_token1] = ACTIONS(1260),
    [anon_sym_DQUOTE] = ACTIONS(1260),
    [aux_sym_string_token1] = ACTIONS(1260),
    [anon_sym_SQUOTE] = ACTIONS(1260),
    [anon_sym_LT_LT_LT] = ACTIONS(1260),
    [anon_sym_BQUOTE] = ACTIONS(1260),
    [anon_sym_DOLLAR] = ACTIONS(1260),
    [aux_sym_yield_expression_token1] = ACTIONS(1262),
    [aux_sym_include_expression_token1] = ACTIONS(1262),
    [aux_sym_include_once_expression_token1] = ACTIONS(1262),
    [aux_sym_require_expression_token1] = ACTIONS(1262),
    [aux_sym_require_once_expression_token1] = ACTIONS(1262),
    [sym_comment] = ACTIONS(5),
  },
  [508] = {
    [sym_text_interpolation] = STATE(508),
    [ts_builtin_sym_end] = ACTIONS(1264),
    [sym_name] = ACTIONS(1266),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1264),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1266),
    [aux_sym_global_declaration_token1] = ACTIONS(1266),
    [aux_sym_namespace_definition_token1] = ACTIONS(1266),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1266),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1266),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1266),
    [anon_sym_BSLASH] = ACTIONS(1264),
    [anon_sym_LBRACE] = ACTIONS(1264),
    [anon_sym_RBRACE] = ACTIONS(1264),
    [aux_sym_trait_declaration_token1] = ACTIONS(1266),
    [aux_sym_interface_declaration_token1] = ACTIONS(1266),
    [aux_sym_enum_declaration_token1] = ACTIONS(1266),
    [aux_sym_enum_case_token1] = ACTIONS(1266),
    [aux_sym_class_declaration_token1] = ACTIONS(1266),
    [aux_sym_final_modifier_token1] = ACTIONS(1266),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1266),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1266),
    [sym_var_modifier] = ACTIONS(1266),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1266),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1266),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1266),
    [anon_sym_LPAREN2] = ACTIONS(1264),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1266),
    [aux_sym_cast_type_token1] = ACTIONS(1266),
    [aux_sym_echo_statement_token1] = ACTIONS(1266),
    [aux_sym_exit_statement_token1] = ACTIONS(1266),
    [anon_sym_unset] = ACTIONS(1266),
    [aux_sym_declare_statement_token1] = ACTIONS(1266),
    [aux_sym_declare_statement_token2] = ACTIONS(1266),
    [sym_float] = ACTIONS(1266),
    [aux_sym_try_statement_token1] = ACTIONS(1266),
    [aux_sym_goto_statement_token1] = ACTIONS(1266),
    [aux_sym_continue_statement_token1] = ACTIONS(1266),
    [aux_sym_break_statement_token1] = ACTIONS(1266),
    [sym_integer] = ACTIONS(1266),
    [aux_sym_return_statement_token1] = ACTIONS(1266),
    [aux_sym_throw_expression_token1] = ACTIONS(1266),
    [aux_sym_while_statement_token1] = ACTIONS(1266),
    [aux_sym_while_statement_token2] = ACTIONS(1266),
    [aux_sym_do_statement_token1] = ACTIONS(1266),
    [aux_sym_for_statement_token1] = ACTIONS(1266),
    [aux_sym_for_statement_token2] = ACTIONS(1266),
    [aux_sym_foreach_statement_token1] = ACTIONS(1266),
    [aux_sym_foreach_statement_token2] = ACTIONS(1266),
    [aux_sym_if_statement_token1] = ACTIONS(1266),
    [aux_sym_if_statement_token2] = ACTIONS(1266),
    [aux_sym_else_if_clause_token1] = ACTIONS(1266),
    [aux_sym_else_clause_token1] = ACTIONS(1266),
    [aux_sym_match_expression_token1] = ACTIONS(1266),
    [aux_sym_match_default_expression_token1] = ACTIONS(1266),
    [aux_sym_switch_statement_token1] = ACTIONS(1266),
    [aux_sym_switch_block_token1] = ACTIONS(1266),
    [anon_sym_PLUS] = ACTIONS(1266),
    [anon_sym_DASH] = ACTIONS(1266),
    [anon_sym_TILDE] = ACTIONS(1264),
    [anon_sym_BANG] = ACTIONS(1264),
    [anon_sym_AT] = ACTIONS(1264),
    [aux_sym_clone_expression_token1] = ACTIONS(1266),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1266),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1266),
    [anon_sym_DASH_DASH] = ACTIONS(1264),
    [anon_sym_PLUS_PLUS] = ACTIONS(1264),
    [aux_sym__list_destructing_token1] = ACTIONS(1266),
    [anon_sym_LBRACK] = ACTIONS(1264),
    [anon_sym_self] = ACTIONS(1266),
    [anon_sym_parent] = ACTIONS(1266),
    [aux_sym__argument_name_token1] = ACTIONS(1266),
    [aux_sym__argument_name_token2] = ACTIONS(1266),
    [anon_sym_POUND_LBRACK] = ACTIONS(1264),
    [aux_sym_encapsed_string_token1] = ACTIONS(1264),
    [anon_sym_DQUOTE] = ACTIONS(1264),
    [aux_sym_string_token1] = ACTIONS(1264),
    [anon_sym_SQUOTE] = ACTIONS(1264),
    [anon_sym_LT_LT_LT] = ACTIONS(1264),
    [anon_sym_BQUOTE] = ACTIONS(1264),
    [anon_sym_DOLLAR] = ACTIONS(1264),
    [aux_sym_yield_expression_token1] = ACTIONS(1266),
    [aux_sym_include_expression_token1] = ACTIONS(1266),
    [aux_sym_include_once_expression_token1] = ACTIONS(1266),
    [aux_sym_require_expression_token1] = ACTIONS(1266),
    [aux_sym_require_once_expression_token1] = ACTIONS(1266),
    [sym_comment] = ACTIONS(5),
  },
  [509] = {
    [sym_text_interpolation] = STATE(509),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_name] = ACTIONS(1270),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1268),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1270),
    [aux_sym_global_declaration_token1] = ACTIONS(1270),
    [aux_sym_namespace_definition_token1] = ACTIONS(1270),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1270),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1270),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1270),
    [anon_sym_BSLASH] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_trait_declaration_token1] = ACTIONS(1270),
    [aux_sym_interface_declaration_token1] = ACTIONS(1270),
    [aux_sym_enum_declaration_token1] = ACTIONS(1270),
    [aux_sym_enum_case_token1] = ACTIONS(1270),
    [aux_sym_class_declaration_token1] = ACTIONS(1270),
    [aux_sym_final_modifier_token1] = ACTIONS(1270),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1270),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1270),
    [sym_var_modifier] = ACTIONS(1270),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1270),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1270),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1270),
    [anon_sym_LPAREN2] = ACTIONS(1268),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1270),
    [aux_sym_cast_type_token1] = ACTIONS(1270),
    [aux_sym_echo_statement_token1] = ACTIONS(1270),
    [aux_sym_exit_statement_token1] = ACTIONS(1270),
    [anon_sym_unset] = ACTIONS(1270),
    [aux_sym_declare_statement_token1] = ACTIONS(1270),
    [aux_sym_declare_statement_token2] = ACTIONS(1270),
    [sym_float] = ACTIONS(1270),
    [aux_sym_try_statement_token1] = ACTIONS(1270),
    [aux_sym_goto_statement_token1] = ACTIONS(1270),
    [aux_sym_continue_statement_token1] = ACTIONS(1270),
    [aux_sym_break_statement_token1] = ACTIONS(1270),
    [sym_integer] = ACTIONS(1270),
    [aux_sym_return_statement_token1] = ACTIONS(1270),
    [aux_sym_throw_expression_token1] = ACTIONS(1270),
    [aux_sym_while_statement_token1] = ACTIONS(1270),
    [aux_sym_while_statement_token2] = ACTIONS(1270),
    [aux_sym_do_statement_token1] = ACTIONS(1270),
    [aux_sym_for_statement_token1] = ACTIONS(1270),
    [aux_sym_for_statement_token2] = ACTIONS(1270),
    [aux_sym_foreach_statement_token1] = ACTIONS(1270),
    [aux_sym_foreach_statement_token2] = ACTIONS(1270),
    [aux_sym_if_statement_token1] = ACTIONS(1270),
    [aux_sym_if_statement_token2] = ACTIONS(1270),
    [aux_sym_else_if_clause_token1] = ACTIONS(1270),
    [aux_sym_else_clause_token1] = ACTIONS(1270),
    [aux_sym_match_expression_token1] = ACTIONS(1270),
    [aux_sym_match_default_expression_token1] = ACTIONS(1270),
    [aux_sym_switch_statement_token1] = ACTIONS(1270),
    [aux_sym_switch_block_token1] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1270),
    [anon_sym_DASH] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1268),
    [anon_sym_AT] = ACTIONS(1268),
    [aux_sym_clone_expression_token1] = ACTIONS(1270),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1270),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1270),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [aux_sym__list_destructing_token1] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1268),
    [anon_sym_self] = ACTIONS(1270),
    [anon_sym_parent] = ACTIONS(1270),
    [aux_sym__argument_name_token1] = ACTIONS(1270),
    [aux_sym__argument_name_token2] = ACTIONS(1270),
    [anon_sym_POUND_LBRACK] = ACTIONS(1268),
    [aux_sym_encapsed_string_token1] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1268),
    [aux_sym_string_token1] = ACTIONS(1268),
    [anon_sym_SQUOTE] = ACTIONS(1268),
    [anon_sym_LT_LT_LT] = ACTIONS(1268),
    [anon_sym_BQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR] = ACTIONS(1268),
    [aux_sym_yield_expression_token1] = ACTIONS(1270),
    [aux_sym_include_expression_token1] = ACTIONS(1270),
    [aux_sym_include_once_expression_token1] = ACTIONS(1270),
    [aux_sym_require_expression_token1] = ACTIONS(1270),
    [aux_sym_require_once_expression_token1] = ACTIONS(1270),
    [sym_comment] = ACTIONS(5),
  },
  [510] = {
    [sym_text_interpolation] = STATE(510),
    [ts_builtin_sym_end] = ACTIONS(1272),
    [sym_name] = ACTIONS(1274),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1272),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1274),
    [aux_sym_global_declaration_token1] = ACTIONS(1274),
    [aux_sym_namespace_definition_token1] = ACTIONS(1274),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1274),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1274),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1274),
    [anon_sym_BSLASH] = ACTIONS(1272),
    [anon_sym_LBRACE] = ACTIONS(1272),
    [anon_sym_RBRACE] = ACTIONS(1272),
    [aux_sym_trait_declaration_token1] = ACTIONS(1274),
    [aux_sym_interface_declaration_token1] = ACTIONS(1274),
    [aux_sym_enum_declaration_token1] = ACTIONS(1274),
    [aux_sym_enum_case_token1] = ACTIONS(1274),
    [aux_sym_class_declaration_token1] = ACTIONS(1274),
    [aux_sym_final_modifier_token1] = ACTIONS(1274),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1274),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1274),
    [sym_var_modifier] = ACTIONS(1274),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1274),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1274),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1274),
    [anon_sym_LPAREN2] = ACTIONS(1272),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1274),
    [aux_sym_cast_type_token1] = ACTIONS(1274),
    [aux_sym_echo_statement_token1] = ACTIONS(1274),
    [aux_sym_exit_statement_token1] = ACTIONS(1274),
    [anon_sym_unset] = ACTIONS(1274),
    [aux_sym_declare_statement_token1] = ACTIONS(1274),
    [aux_sym_declare_statement_token2] = ACTIONS(1274),
    [sym_float] = ACTIONS(1274),
    [aux_sym_try_statement_token1] = ACTIONS(1274),
    [aux_sym_goto_statement_token1] = ACTIONS(1274),
    [aux_sym_continue_statement_token1] = ACTIONS(1274),
    [aux_sym_break_statement_token1] = ACTIONS(1274),
    [sym_integer] = ACTIONS(1274),
    [aux_sym_return_statement_token1] = ACTIONS(1274),
    [aux_sym_throw_expression_token1] = ACTIONS(1274),
    [aux_sym_while_statement_token1] = ACTIONS(1274),
    [aux_sym_while_statement_token2] = ACTIONS(1274),
    [aux_sym_do_statement_token1] = ACTIONS(1274),
    [aux_sym_for_statement_token1] = ACTIONS(1274),
    [aux_sym_for_statement_token2] = ACTIONS(1274),
    [aux_sym_foreach_statement_token1] = ACTIONS(1274),
    [aux_sym_foreach_statement_token2] = ACTIONS(1274),
    [aux_sym_if_statement_token1] = ACTIONS(1274),
    [aux_sym_if_statement_token2] = ACTIONS(1274),
    [aux_sym_else_if_clause_token1] = ACTIONS(1274),
    [aux_sym_else_clause_token1] = ACTIONS(1274),
    [aux_sym_match_expression_token1] = ACTIONS(1274),
    [aux_sym_match_default_expression_token1] = ACTIONS(1274),
    [aux_sym_switch_statement_token1] = ACTIONS(1274),
    [aux_sym_switch_block_token1] = ACTIONS(1274),
    [anon_sym_PLUS] = ACTIONS(1274),
    [anon_sym_DASH] = ACTIONS(1274),
    [anon_sym_TILDE] = ACTIONS(1272),
    [anon_sym_BANG] = ACTIONS(1272),
    [anon_sym_AT] = ACTIONS(1272),
    [aux_sym_clone_expression_token1] = ACTIONS(1274),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1274),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1274),
    [anon_sym_DASH_DASH] = ACTIONS(1272),
    [anon_sym_PLUS_PLUS] = ACTIONS(1272),
    [aux_sym__list_destructing_token1] = ACTIONS(1274),
    [anon_sym_LBRACK] = ACTIONS(1272),
    [anon_sym_self] = ACTIONS(1274),
    [anon_sym_parent] = ACTIONS(1274),
    [aux_sym__argument_name_token1] = ACTIONS(1274),
    [aux_sym__argument_name_token2] = ACTIONS(1274),
    [anon_sym_POUND_LBRACK] = ACTIONS(1272),
    [aux_sym_encapsed_string_token1] = ACTIONS(1272),
    [anon_sym_DQUOTE] = ACTIONS(1272),
    [aux_sym_string_token1] = ACTIONS(1272),
    [anon_sym_SQUOTE] = ACTIONS(1272),
    [anon_sym_LT_LT_LT] = ACTIONS(1272),
    [anon_sym_BQUOTE] = ACTIONS(1272),
    [anon_sym_DOLLAR] = ACTIONS(1272),
    [aux_sym_yield_expression_token1] = ACTIONS(1274),
    [aux_sym_include_expression_token1] = ACTIONS(1274),
    [aux_sym_include_once_expression_token1] = ACTIONS(1274),
    [aux_sym_require_expression_token1] = ACTIONS(1274),
    [aux_sym_require_once_expression_token1] = ACTIONS(1274),
    [sym_comment] = ACTIONS(5),
  },
  [511] = {
    [sym_text_interpolation] = STATE(511),
    [ts_builtin_sym_end] = ACTIONS(1276),
    [sym_name] = ACTIONS(1278),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1276),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1278),
    [aux_sym_global_declaration_token1] = ACTIONS(1278),
    [aux_sym_namespace_definition_token1] = ACTIONS(1278),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1278),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1278),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1278),
    [anon_sym_BSLASH] = ACTIONS(1276),
    [anon_sym_LBRACE] = ACTIONS(1276),
    [anon_sym_RBRACE] = ACTIONS(1276),
    [aux_sym_trait_declaration_token1] = ACTIONS(1278),
    [aux_sym_interface_declaration_token1] = ACTIONS(1278),
    [aux_sym_enum_declaration_token1] = ACTIONS(1278),
    [aux_sym_enum_case_token1] = ACTIONS(1278),
    [aux_sym_class_declaration_token1] = ACTIONS(1278),
    [aux_sym_final_modifier_token1] = ACTIONS(1278),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1278),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1278),
    [sym_var_modifier] = ACTIONS(1278),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1278),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1278),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1278),
    [anon_sym_LPAREN2] = ACTIONS(1276),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1278),
    [aux_sym_cast_type_token1] = ACTIONS(1278),
    [aux_sym_echo_statement_token1] = ACTIONS(1278),
    [aux_sym_exit_statement_token1] = ACTIONS(1278),
    [anon_sym_unset] = ACTIONS(1278),
    [aux_sym_declare_statement_token1] = ACTIONS(1278),
    [aux_sym_declare_statement_token2] = ACTIONS(1278),
    [sym_float] = ACTIONS(1278),
    [aux_sym_try_statement_token1] = ACTIONS(1278),
    [aux_sym_goto_statement_token1] = ACTIONS(1278),
    [aux_sym_continue_statement_token1] = ACTIONS(1278),
    [aux_sym_break_statement_token1] = ACTIONS(1278),
    [sym_integer] = ACTIONS(1278),
    [aux_sym_return_statement_token1] = ACTIONS(1278),
    [aux_sym_throw_expression_token1] = ACTIONS(1278),
    [aux_sym_while_statement_token1] = ACTIONS(1278),
    [aux_sym_while_statement_token2] = ACTIONS(1278),
    [aux_sym_do_statement_token1] = ACTIONS(1278),
    [aux_sym_for_statement_token1] = ACTIONS(1278),
    [aux_sym_for_statement_token2] = ACTIONS(1278),
    [aux_sym_foreach_statement_token1] = ACTIONS(1278),
    [aux_sym_foreach_statement_token2] = ACTIONS(1278),
    [aux_sym_if_statement_token1] = ACTIONS(1278),
    [aux_sym_if_statement_token2] = ACTIONS(1278),
    [aux_sym_else_if_clause_token1] = ACTIONS(1278),
    [aux_sym_else_clause_token1] = ACTIONS(1278),
    [aux_sym_match_expression_token1] = ACTIONS(1278),
    [aux_sym_match_default_expression_token1] = ACTIONS(1278),
    [aux_sym_switch_statement_token1] = ACTIONS(1278),
    [aux_sym_switch_block_token1] = ACTIONS(1278),
    [anon_sym_PLUS] = ACTIONS(1278),
    [anon_sym_DASH] = ACTIONS(1278),
    [anon_sym_TILDE] = ACTIONS(1276),
    [anon_sym_BANG] = ACTIONS(1276),
    [anon_sym_AT] = ACTIONS(1276),
    [aux_sym_clone_expression_token1] = ACTIONS(1278),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1278),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1278),
    [anon_sym_DASH_DASH] = ACTIONS(1276),
    [anon_sym_PLUS_PLUS] = ACTIONS(1276),
    [aux_sym__list_destructing_token1] = ACTIONS(1278),
    [anon_sym_LBRACK] = ACTIONS(1276),
    [anon_sym_self] = ACTIONS(1278),
    [anon_sym_parent] = ACTIONS(1278),
    [aux_sym__argument_name_token1] = ACTIONS(1278),
    [aux_sym__argument_name_token2] = ACTIONS(1278),
    [anon_sym_POUND_LBRACK] = ACTIONS(1276),
    [aux_sym_encapsed_string_token1] = ACTIONS(1276),
    [anon_sym_DQUOTE] = ACTIONS(1276),
    [aux_sym_string_token1] = ACTIONS(1276),
    [anon_sym_SQUOTE] = ACTIONS(1276),
    [anon_sym_LT_LT_LT] = ACTIONS(1276),
    [anon_sym_BQUOTE] = ACTIONS(1276),
    [anon_sym_DOLLAR] = ACTIONS(1276),
    [aux_sym_yield_expression_token1] = ACTIONS(1278),
    [aux_sym_include_expression_token1] = ACTIONS(1278),
    [aux_sym_include_once_expression_token1] = ACTIONS(1278),
    [aux_sym_require_expression_token1] = ACTIONS(1278),
    [aux_sym_require_once_expression_token1] = ACTIONS(1278),
    [sym_comment] = ACTIONS(5),
  },
  [512] = {
    [sym_text_interpolation] = STATE(512),
    [ts_builtin_sym_end] = ACTIONS(1280),
    [sym_name] = ACTIONS(1282),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1280),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1282),
    [aux_sym_global_declaration_token1] = ACTIONS(1282),
    [aux_sym_namespace_definition_token1] = ACTIONS(1282),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1282),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1282),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1282),
    [anon_sym_BSLASH] = ACTIONS(1280),
    [anon_sym_LBRACE] = ACTIONS(1280),
    [anon_sym_RBRACE] = ACTIONS(1280),
    [aux_sym_trait_declaration_token1] = ACTIONS(1282),
    [aux_sym_interface_declaration_token1] = ACTIONS(1282),
    [aux_sym_enum_declaration_token1] = ACTIONS(1282),
    [aux_sym_enum_case_token1] = ACTIONS(1282),
    [aux_sym_class_declaration_token1] = ACTIONS(1282),
    [aux_sym_final_modifier_token1] = ACTIONS(1282),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1282),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1282),
    [sym_var_modifier] = ACTIONS(1282),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1282),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1282),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1282),
    [anon_sym_LPAREN2] = ACTIONS(1280),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1282),
    [aux_sym_cast_type_token1] = ACTIONS(1282),
    [aux_sym_echo_statement_token1] = ACTIONS(1282),
    [aux_sym_exit_statement_token1] = ACTIONS(1282),
    [anon_sym_unset] = ACTIONS(1282),
    [aux_sym_declare_statement_token1] = ACTIONS(1282),
    [aux_sym_declare_statement_token2] = ACTIONS(1282),
    [sym_float] = ACTIONS(1282),
    [aux_sym_try_statement_token1] = ACTIONS(1282),
    [aux_sym_goto_statement_token1] = ACTIONS(1282),
    [aux_sym_continue_statement_token1] = ACTIONS(1282),
    [aux_sym_break_statement_token1] = ACTIONS(1282),
    [sym_integer] = ACTIONS(1282),
    [aux_sym_return_statement_token1] = ACTIONS(1282),
    [aux_sym_throw_expression_token1] = ACTIONS(1282),
    [aux_sym_while_statement_token1] = ACTIONS(1282),
    [aux_sym_while_statement_token2] = ACTIONS(1282),
    [aux_sym_do_statement_token1] = ACTIONS(1282),
    [aux_sym_for_statement_token1] = ACTIONS(1282),
    [aux_sym_for_statement_token2] = ACTIONS(1282),
    [aux_sym_foreach_statement_token1] = ACTIONS(1282),
    [aux_sym_foreach_statement_token2] = ACTIONS(1282),
    [aux_sym_if_statement_token1] = ACTIONS(1282),
    [aux_sym_if_statement_token2] = ACTIONS(1282),
    [aux_sym_else_if_clause_token1] = ACTIONS(1282),
    [aux_sym_else_clause_token1] = ACTIONS(1282),
    [aux_sym_match_expression_token1] = ACTIONS(1282),
    [aux_sym_match_default_expression_token1] = ACTIONS(1282),
    [aux_sym_switch_statement_token1] = ACTIONS(1282),
    [aux_sym_switch_block_token1] = ACTIONS(1282),
    [anon_sym_PLUS] = ACTIONS(1282),
    [anon_sym_DASH] = ACTIONS(1282),
    [anon_sym_TILDE] = ACTIONS(1280),
    [anon_sym_BANG] = ACTIONS(1280),
    [anon_sym_AT] = ACTIONS(1280),
    [aux_sym_clone_expression_token1] = ACTIONS(1282),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1282),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1282),
    [anon_sym_DASH_DASH] = ACTIONS(1280),
    [anon_sym_PLUS_PLUS] = ACTIONS(1280),
    [aux_sym__list_destructing_token1] = ACTIONS(1282),
    [anon_sym_LBRACK] = ACTIONS(1280),
    [anon_sym_self] = ACTIONS(1282),
    [anon_sym_parent] = ACTIONS(1282),
    [aux_sym__argument_name_token1] = ACTIONS(1282),
    [aux_sym__argument_name_token2] = ACTIONS(1282),
    [anon_sym_POUND_LBRACK] = ACTIONS(1280),
    [aux_sym_encapsed_string_token1] = ACTIONS(1280),
    [anon_sym_DQUOTE] = ACTIONS(1280),
    [aux_sym_string_token1] = ACTIONS(1280),
    [anon_sym_SQUOTE] = ACTIONS(1280),
    [anon_sym_LT_LT_LT] = ACTIONS(1280),
    [anon_sym_BQUOTE] = ACTIONS(1280),
    [anon_sym_DOLLAR] = ACTIONS(1280),
    [aux_sym_yield_expression_token1] = ACTIONS(1282),
    [aux_sym_include_expression_token1] = ACTIONS(1282),
    [aux_sym_include_once_expression_token1] = ACTIONS(1282),
    [aux_sym_require_expression_token1] = ACTIONS(1282),
    [aux_sym_require_once_expression_token1] = ACTIONS(1282),
    [sym_comment] = ACTIONS(5),
  },
  [513] = {
    [sym_text_interpolation] = STATE(513),
    [ts_builtin_sym_end] = ACTIONS(1284),
    [sym_name] = ACTIONS(1286),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1284),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1286),
    [aux_sym_global_declaration_token1] = ACTIONS(1286),
    [aux_sym_namespace_definition_token1] = ACTIONS(1286),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1286),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1286),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1286),
    [anon_sym_BSLASH] = ACTIONS(1284),
    [anon_sym_LBRACE] = ACTIONS(1284),
    [anon_sym_RBRACE] = ACTIONS(1284),
    [aux_sym_trait_declaration_token1] = ACTIONS(1286),
    [aux_sym_interface_declaration_token1] = ACTIONS(1286),
    [aux_sym_enum_declaration_token1] = ACTIONS(1286),
    [aux_sym_enum_case_token1] = ACTIONS(1286),
    [aux_sym_class_declaration_token1] = ACTIONS(1286),
    [aux_sym_final_modifier_token1] = ACTIONS(1286),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1286),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1286),
    [sym_var_modifier] = ACTIONS(1286),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1286),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1286),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1286),
    [anon_sym_LPAREN2] = ACTIONS(1284),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1286),
    [aux_sym_cast_type_token1] = ACTIONS(1286),
    [aux_sym_echo_statement_token1] = ACTIONS(1286),
    [aux_sym_exit_statement_token1] = ACTIONS(1286),
    [anon_sym_unset] = ACTIONS(1286),
    [aux_sym_declare_statement_token1] = ACTIONS(1286),
    [aux_sym_declare_statement_token2] = ACTIONS(1286),
    [sym_float] = ACTIONS(1286),
    [aux_sym_try_statement_token1] = ACTIONS(1286),
    [aux_sym_goto_statement_token1] = ACTIONS(1286),
    [aux_sym_continue_statement_token1] = ACTIONS(1286),
    [aux_sym_break_statement_token1] = ACTIONS(1286),
    [sym_integer] = ACTIONS(1286),
    [aux_sym_return_statement_token1] = ACTIONS(1286),
    [aux_sym_throw_expression_token1] = ACTIONS(1286),
    [aux_sym_while_statement_token1] = ACTIONS(1286),
    [aux_sym_while_statement_token2] = ACTIONS(1286),
    [aux_sym_do_statement_token1] = ACTIONS(1286),
    [aux_sym_for_statement_token1] = ACTIONS(1286),
    [aux_sym_for_statement_token2] = ACTIONS(1286),
    [aux_sym_foreach_statement_token1] = ACTIONS(1286),
    [aux_sym_foreach_statement_token2] = ACTIONS(1286),
    [aux_sym_if_statement_token1] = ACTIONS(1286),
    [aux_sym_if_statement_token2] = ACTIONS(1286),
    [aux_sym_else_if_clause_token1] = ACTIONS(1286),
    [aux_sym_else_clause_token1] = ACTIONS(1286),
    [aux_sym_match_expression_token1] = ACTIONS(1286),
    [aux_sym_match_default_expression_token1] = ACTIONS(1286),
    [aux_sym_switch_statement_token1] = ACTIONS(1286),
    [aux_sym_switch_block_token1] = ACTIONS(1286),
    [anon_sym_PLUS] = ACTIONS(1286),
    [anon_sym_DASH] = ACTIONS(1286),
    [anon_sym_TILDE] = ACTIONS(1284),
    [anon_sym_BANG] = ACTIONS(1284),
    [anon_sym_AT] = ACTIONS(1284),
    [aux_sym_clone_expression_token1] = ACTIONS(1286),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1286),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1286),
    [anon_sym_DASH_DASH] = ACTIONS(1284),
    [anon_sym_PLUS_PLUS] = ACTIONS(1284),
    [aux_sym__list_destructing_token1] = ACTIONS(1286),
    [anon_sym_LBRACK] = ACTIONS(1284),
    [anon_sym_self] = ACTIONS(1286),
    [anon_sym_parent] = ACTIONS(1286),
    [aux_sym__argument_name_token1] = ACTIONS(1286),
    [aux_sym__argument_name_token2] = ACTIONS(1286),
    [anon_sym_POUND_LBRACK] = ACTIONS(1284),
    [aux_sym_encapsed_string_token1] = ACTIONS(1284),
    [anon_sym_DQUOTE] = ACTIONS(1284),
    [aux_sym_string_token1] = ACTIONS(1284),
    [anon_sym_SQUOTE] = ACTIONS(1284),
    [anon_sym_LT_LT_LT] = ACTIONS(1284),
    [anon_sym_BQUOTE] = ACTIONS(1284),
    [anon_sym_DOLLAR] = ACTIONS(1284),
    [aux_sym_yield_expression_token1] = ACTIONS(1286),
    [aux_sym_include_expression_token1] = ACTIONS(1286),
    [aux_sym_include_once_expression_token1] = ACTIONS(1286),
    [aux_sym_require_expression_token1] = ACTIONS(1286),
    [aux_sym_require_once_expression_token1] = ACTIONS(1286),
    [sym_comment] = ACTIONS(5),
  },
  [514] = {
    [sym_text_interpolation] = STATE(514),
    [ts_builtin_sym_end] = ACTIONS(1288),
    [sym_name] = ACTIONS(1290),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1288),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1290),
    [aux_sym_global_declaration_token1] = ACTIONS(1290),
    [aux_sym_namespace_definition_token1] = ACTIONS(1290),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1290),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1290),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1290),
    [anon_sym_BSLASH] = ACTIONS(1288),
    [anon_sym_LBRACE] = ACTIONS(1288),
    [anon_sym_RBRACE] = ACTIONS(1288),
    [aux_sym_trait_declaration_token1] = ACTIONS(1290),
    [aux_sym_interface_declaration_token1] = ACTIONS(1290),
    [aux_sym_enum_declaration_token1] = ACTIONS(1290),
    [aux_sym_enum_case_token1] = ACTIONS(1290),
    [aux_sym_class_declaration_token1] = ACTIONS(1290),
    [aux_sym_final_modifier_token1] = ACTIONS(1290),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1290),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1290),
    [sym_var_modifier] = ACTIONS(1290),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1290),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1290),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1290),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1290),
    [aux_sym_cast_type_token1] = ACTIONS(1290),
    [aux_sym_echo_statement_token1] = ACTIONS(1290),
    [aux_sym_exit_statement_token1] = ACTIONS(1290),
    [anon_sym_unset] = ACTIONS(1290),
    [aux_sym_declare_statement_token1] = ACTIONS(1290),
    [aux_sym_declare_statement_token2] = ACTIONS(1290),
    [sym_float] = ACTIONS(1290),
    [aux_sym_try_statement_token1] = ACTIONS(1290),
    [aux_sym_goto_statement_token1] = ACTIONS(1290),
    [aux_sym_continue_statement_token1] = ACTIONS(1290),
    [aux_sym_break_statement_token1] = ACTIONS(1290),
    [sym_integer] = ACTIONS(1290),
    [aux_sym_return_statement_token1] = ACTIONS(1290),
    [aux_sym_throw_expression_token1] = ACTIONS(1290),
    [aux_sym_while_statement_token1] = ACTIONS(1290),
    [aux_sym_while_statement_token2] = ACTIONS(1290),
    [aux_sym_do_statement_token1] = ACTIONS(1290),
    [aux_sym_for_statement_token1] = ACTIONS(1290),
    [aux_sym_for_statement_token2] = ACTIONS(1290),
    [aux_sym_foreach_statement_token1] = ACTIONS(1290),
    [aux_sym_foreach_statement_token2] = ACTIONS(1290),
    [aux_sym_if_statement_token1] = ACTIONS(1290),
    [aux_sym_if_statement_token2] = ACTIONS(1290),
    [aux_sym_else_if_clause_token1] = ACTIONS(1290),
    [aux_sym_else_clause_token1] = ACTIONS(1290),
    [aux_sym_match_expression_token1] = ACTIONS(1290),
    [aux_sym_match_default_expression_token1] = ACTIONS(1290),
    [aux_sym_switch_statement_token1] = ACTIONS(1290),
    [aux_sym_switch_block_token1] = ACTIONS(1290),
    [anon_sym_PLUS] = ACTIONS(1290),
    [anon_sym_DASH] = ACTIONS(1290),
    [anon_sym_TILDE] = ACTIONS(1288),
    [anon_sym_BANG] = ACTIONS(1288),
    [anon_sym_AT] = ACTIONS(1288),
    [aux_sym_clone_expression_token1] = ACTIONS(1290),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1290),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1290),
    [anon_sym_DASH_DASH] = ACTIONS(1288),
    [anon_sym_PLUS_PLUS] = ACTIONS(1288),
    [aux_sym__list_destructing_token1] = ACTIONS(1290),
    [anon_sym_LBRACK] = ACTIONS(1288),
    [anon_sym_self] = ACTIONS(1290),
    [anon_sym_parent] = ACTIONS(1290),
    [aux_sym__argument_name_token1] = ACTIONS(1290),
    [aux_sym__argument_name_token2] = ACTIONS(1290),
    [anon_sym_POUND_LBRACK] = ACTIONS(1288),
    [aux_sym_encapsed_string_token1] = ACTIONS(1288),
    [anon_sym_DQUOTE] = ACTIONS(1288),
    [aux_sym_string_token1] = ACTIONS(1288),
    [anon_sym_SQUOTE] = ACTIONS(1288),
    [anon_sym_LT_LT_LT] = ACTIONS(1288),
    [anon_sym_BQUOTE] = ACTIONS(1288),
    [anon_sym_DOLLAR] = ACTIONS(1288),
    [aux_sym_yield_expression_token1] = ACTIONS(1290),
    [aux_sym_include_expression_token1] = ACTIONS(1290),
    [aux_sym_include_once_expression_token1] = ACTIONS(1290),
    [aux_sym_require_expression_token1] = ACTIONS(1290),
    [aux_sym_require_once_expression_token1] = ACTIONS(1290),
    [sym_comment] = ACTIONS(5),
  },
  [515] = {
    [sym_text_interpolation] = STATE(515),
    [ts_builtin_sym_end] = ACTIONS(1292),
    [sym_name] = ACTIONS(1294),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1292),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1294),
    [aux_sym_global_declaration_token1] = ACTIONS(1294),
    [aux_sym_namespace_definition_token1] = ACTIONS(1294),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1294),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1294),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1294),
    [anon_sym_BSLASH] = ACTIONS(1292),
    [anon_sym_LBRACE] = ACTIONS(1292),
    [anon_sym_RBRACE] = ACTIONS(1292),
    [aux_sym_trait_declaration_token1] = ACTIONS(1294),
    [aux_sym_interface_declaration_token1] = ACTIONS(1294),
    [aux_sym_enum_declaration_token1] = ACTIONS(1294),
    [aux_sym_enum_case_token1] = ACTIONS(1294),
    [aux_sym_class_declaration_token1] = ACTIONS(1294),
    [aux_sym_final_modifier_token1] = ACTIONS(1294),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1294),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1294),
    [sym_var_modifier] = ACTIONS(1294),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1294),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1294),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1294),
    [anon_sym_LPAREN2] = ACTIONS(1292),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1294),
    [aux_sym_cast_type_token1] = ACTIONS(1294),
    [aux_sym_echo_statement_token1] = ACTIONS(1294),
    [aux_sym_exit_statement_token1] = ACTIONS(1294),
    [anon_sym_unset] = ACTIONS(1294),
    [aux_sym_declare_statement_token1] = ACTIONS(1294),
    [aux_sym_declare_statement_token2] = ACTIONS(1294),
    [sym_float] = ACTIONS(1294),
    [aux_sym_try_statement_token1] = ACTIONS(1294),
    [aux_sym_goto_statement_token1] = ACTIONS(1294),
    [aux_sym_continue_statement_token1] = ACTIONS(1294),
    [aux_sym_break_statement_token1] = ACTIONS(1294),
    [sym_integer] = ACTIONS(1294),
    [aux_sym_return_statement_token1] = ACTIONS(1294),
    [aux_sym_throw_expression_token1] = ACTIONS(1294),
    [aux_sym_while_statement_token1] = ACTIONS(1294),
    [aux_sym_while_statement_token2] = ACTIONS(1294),
    [aux_sym_do_statement_token1] = ACTIONS(1294),
    [aux_sym_for_statement_token1] = ACTIONS(1294),
    [aux_sym_for_statement_token2] = ACTIONS(1294),
    [aux_sym_foreach_statement_token1] = ACTIONS(1294),
    [aux_sym_foreach_statement_token2] = ACTIONS(1294),
    [aux_sym_if_statement_token1] = ACTIONS(1294),
    [aux_sym_if_statement_token2] = ACTIONS(1294),
    [aux_sym_else_if_clause_token1] = ACTIONS(1294),
    [aux_sym_else_clause_token1] = ACTIONS(1294),
    [aux_sym_match_expression_token1] = ACTIONS(1294),
    [aux_sym_match_default_expression_token1] = ACTIONS(1294),
    [aux_sym_switch_statement_token1] = ACTIONS(1294),
    [aux_sym_switch_block_token1] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(1294),
    [anon_sym_DASH] = ACTIONS(1294),
    [anon_sym_TILDE] = ACTIONS(1292),
    [anon_sym_BANG] = ACTIONS(1292),
    [anon_sym_AT] = ACTIONS(1292),
    [aux_sym_clone_expression_token1] = ACTIONS(1294),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1294),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1294),
    [anon_sym_DASH_DASH] = ACTIONS(1292),
    [anon_sym_PLUS_PLUS] = ACTIONS(1292),
    [aux_sym__list_destructing_token1] = ACTIONS(1294),
    [anon_sym_LBRACK] = ACTIONS(1292),
    [anon_sym_self] = ACTIONS(1294),
    [anon_sym_parent] = ACTIONS(1294),
    [aux_sym__argument_name_token1] = ACTIONS(1294),
    [aux_sym__argument_name_token2] = ACTIONS(1294),
    [anon_sym_POUND_LBRACK] = ACTIONS(1292),
    [aux_sym_encapsed_string_token1] = ACTIONS(1292),
    [anon_sym_DQUOTE] = ACTIONS(1292),
    [aux_sym_string_token1] = ACTIONS(1292),
    [anon_sym_SQUOTE] = ACTIONS(1292),
    [anon_sym_LT_LT_LT] = ACTIONS(1292),
    [anon_sym_BQUOTE] = ACTIONS(1292),
    [anon_sym_DOLLAR] = ACTIONS(1292),
    [aux_sym_yield_expression_token1] = ACTIONS(1294),
    [aux_sym_include_expression_token1] = ACTIONS(1294),
    [aux_sym_include_once_expression_token1] = ACTIONS(1294),
    [aux_sym_require_expression_token1] = ACTIONS(1294),
    [aux_sym_require_once_expression_token1] = ACTIONS(1294),
    [sym_comment] = ACTIONS(5),
  },
  [516] = {
    [sym_text_interpolation] = STATE(516),
    [ts_builtin_sym_end] = ACTIONS(1296),
    [sym_name] = ACTIONS(1298),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1296),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1298),
    [aux_sym_global_declaration_token1] = ACTIONS(1298),
    [aux_sym_namespace_definition_token1] = ACTIONS(1298),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1298),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1298),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1298),
    [anon_sym_BSLASH] = ACTIONS(1296),
    [anon_sym_LBRACE] = ACTIONS(1296),
    [anon_sym_RBRACE] = ACTIONS(1296),
    [aux_sym_trait_declaration_token1] = ACTIONS(1298),
    [aux_sym_interface_declaration_token1] = ACTIONS(1298),
    [aux_sym_enum_declaration_token1] = ACTIONS(1298),
    [aux_sym_enum_case_token1] = ACTIONS(1298),
    [aux_sym_class_declaration_token1] = ACTIONS(1298),
    [aux_sym_final_modifier_token1] = ACTIONS(1298),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1298),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1298),
    [sym_var_modifier] = ACTIONS(1298),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1298),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1298),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1298),
    [anon_sym_LPAREN2] = ACTIONS(1296),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1298),
    [aux_sym_cast_type_token1] = ACTIONS(1298),
    [aux_sym_echo_statement_token1] = ACTIONS(1298),
    [aux_sym_exit_statement_token1] = ACTIONS(1298),
    [anon_sym_unset] = ACTIONS(1298),
    [aux_sym_declare_statement_token1] = ACTIONS(1298),
    [aux_sym_declare_statement_token2] = ACTIONS(1298),
    [sym_float] = ACTIONS(1298),
    [aux_sym_try_statement_token1] = ACTIONS(1298),
    [aux_sym_goto_statement_token1] = ACTIONS(1298),
    [aux_sym_continue_statement_token1] = ACTIONS(1298),
    [aux_sym_break_statement_token1] = ACTIONS(1298),
    [sym_integer] = ACTIONS(1298),
    [aux_sym_return_statement_token1] = ACTIONS(1298),
    [aux_sym_throw_expression_token1] = ACTIONS(1298),
    [aux_sym_while_statement_token1] = ACTIONS(1298),
    [aux_sym_while_statement_token2] = ACTIONS(1298),
    [aux_sym_do_statement_token1] = ACTIONS(1298),
    [aux_sym_for_statement_token1] = ACTIONS(1298),
    [aux_sym_for_statement_token2] = ACTIONS(1298),
    [aux_sym_foreach_statement_token1] = ACTIONS(1298),
    [aux_sym_foreach_statement_token2] = ACTIONS(1298),
    [aux_sym_if_statement_token1] = ACTIONS(1298),
    [aux_sym_if_statement_token2] = ACTIONS(1298),
    [aux_sym_else_if_clause_token1] = ACTIONS(1298),
    [aux_sym_else_clause_token1] = ACTIONS(1298),
    [aux_sym_match_expression_token1] = ACTIONS(1298),
    [aux_sym_match_default_expression_token1] = ACTIONS(1298),
    [aux_sym_switch_statement_token1] = ACTIONS(1298),
    [aux_sym_switch_block_token1] = ACTIONS(1298),
    [anon_sym_PLUS] = ACTIONS(1298),
    [anon_sym_DASH] = ACTIONS(1298),
    [anon_sym_TILDE] = ACTIONS(1296),
    [anon_sym_BANG] = ACTIONS(1296),
    [anon_sym_AT] = ACTIONS(1296),
    [aux_sym_clone_expression_token1] = ACTIONS(1298),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1298),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1298),
    [anon_sym_DASH_DASH] = ACTIONS(1296),
    [anon_sym_PLUS_PLUS] = ACTIONS(1296),
    [aux_sym__list_destructing_token1] = ACTIONS(1298),
    [anon_sym_LBRACK] = ACTIONS(1296),
    [anon_sym_self] = ACTIONS(1298),
    [anon_sym_parent] = ACTIONS(1298),
    [aux_sym__argument_name_token1] = ACTIONS(1298),
    [aux_sym__argument_name_token2] = ACTIONS(1298),
    [anon_sym_POUND_LBRACK] = ACTIONS(1296),
    [aux_sym_encapsed_string_token1] = ACTIONS(1296),
    [anon_sym_DQUOTE] = ACTIONS(1296),
    [aux_sym_string_token1] = ACTIONS(1296),
    [anon_sym_SQUOTE] = ACTIONS(1296),
    [anon_sym_LT_LT_LT] = ACTIONS(1296),
    [anon_sym_BQUOTE] = ACTIONS(1296),
    [anon_sym_DOLLAR] = ACTIONS(1296),
    [aux_sym_yield_expression_token1] = ACTIONS(1298),
    [aux_sym_include_expression_token1] = ACTIONS(1298),
    [aux_sym_include_once_expression_token1] = ACTIONS(1298),
    [aux_sym_require_expression_token1] = ACTIONS(1298),
    [aux_sym_require_once_expression_token1] = ACTIONS(1298),
    [sym_comment] = ACTIONS(5),
  },
  [517] = {
    [sym_text_interpolation] = STATE(517),
    [ts_builtin_sym_end] = ACTIONS(1300),
    [sym_name] = ACTIONS(1302),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1300),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1302),
    [aux_sym_global_declaration_token1] = ACTIONS(1302),
    [aux_sym_namespace_definition_token1] = ACTIONS(1302),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1302),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1302),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1302),
    [anon_sym_BSLASH] = ACTIONS(1300),
    [anon_sym_LBRACE] = ACTIONS(1300),
    [anon_sym_RBRACE] = ACTIONS(1300),
    [aux_sym_trait_declaration_token1] = ACTIONS(1302),
    [aux_sym_interface_declaration_token1] = ACTIONS(1302),
    [aux_sym_enum_declaration_token1] = ACTIONS(1302),
    [aux_sym_enum_case_token1] = ACTIONS(1302),
    [aux_sym_class_declaration_token1] = ACTIONS(1302),
    [aux_sym_final_modifier_token1] = ACTIONS(1302),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1302),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1302),
    [sym_var_modifier] = ACTIONS(1302),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1302),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1302),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1302),
    [anon_sym_LPAREN2] = ACTIONS(1300),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1302),
    [aux_sym_cast_type_token1] = ACTIONS(1302),
    [aux_sym_echo_statement_token1] = ACTIONS(1302),
    [aux_sym_exit_statement_token1] = ACTIONS(1302),
    [anon_sym_unset] = ACTIONS(1302),
    [aux_sym_declare_statement_token1] = ACTIONS(1302),
    [aux_sym_declare_statement_token2] = ACTIONS(1302),
    [sym_float] = ACTIONS(1302),
    [aux_sym_try_statement_token1] = ACTIONS(1302),
    [aux_sym_goto_statement_token1] = ACTIONS(1302),
    [aux_sym_continue_statement_token1] = ACTIONS(1302),
    [aux_sym_break_statement_token1] = ACTIONS(1302),
    [sym_integer] = ACTIONS(1302),
    [aux_sym_return_statement_token1] = ACTIONS(1302),
    [aux_sym_throw_expression_token1] = ACTIONS(1302),
    [aux_sym_while_statement_token1] = ACTIONS(1302),
    [aux_sym_while_statement_token2] = ACTIONS(1302),
    [aux_sym_do_statement_token1] = ACTIONS(1302),
    [aux_sym_for_statement_token1] = ACTIONS(1302),
    [aux_sym_for_statement_token2] = ACTIONS(1302),
    [aux_sym_foreach_statement_token1] = ACTIONS(1302),
    [aux_sym_foreach_statement_token2] = ACTIONS(1302),
    [aux_sym_if_statement_token1] = ACTIONS(1302),
    [aux_sym_if_statement_token2] = ACTIONS(1302),
    [aux_sym_else_if_clause_token1] = ACTIONS(1302),
    [aux_sym_else_clause_token1] = ACTIONS(1302),
    [aux_sym_match_expression_token1] = ACTIONS(1302),
    [aux_sym_match_default_expression_token1] = ACTIONS(1302),
    [aux_sym_switch_statement_token1] = ACTIONS(1302),
    [aux_sym_switch_block_token1] = ACTIONS(1302),
    [anon_sym_PLUS] = ACTIONS(1302),
    [anon_sym_DASH] = ACTIONS(1302),
    [anon_sym_TILDE] = ACTIONS(1300),
    [anon_sym_BANG] = ACTIONS(1300),
    [anon_sym_AT] = ACTIONS(1300),
    [aux_sym_clone_expression_token1] = ACTIONS(1302),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1302),
    [anon_sym_DASH_DASH] = ACTIONS(1300),
    [anon_sym_PLUS_PLUS] = ACTIONS(1300),
    [aux_sym__list_destructing_token1] = ACTIONS(1302),
    [anon_sym_LBRACK] = ACTIONS(1300),
    [anon_sym_self] = ACTIONS(1302),
    [anon_sym_parent] = ACTIONS(1302),
    [aux_sym__argument_name_token1] = ACTIONS(1302),
    [aux_sym__argument_name_token2] = ACTIONS(1302),
    [anon_sym_POUND_LBRACK] = ACTIONS(1300),
    [aux_sym_encapsed_string_token1] = ACTIONS(1300),
    [anon_sym_DQUOTE] = ACTIONS(1300),
    [aux_sym_string_token1] = ACTIONS(1300),
    [anon_sym_SQUOTE] = ACTIONS(1300),
    [anon_sym_LT_LT_LT] = ACTIONS(1300),
    [anon_sym_BQUOTE] = ACTIONS(1300),
    [anon_sym_DOLLAR] = ACTIONS(1300),
    [aux_sym_yield_expression_token1] = ACTIONS(1302),
    [aux_sym_include_expression_token1] = ACTIONS(1302),
    [aux_sym_include_once_expression_token1] = ACTIONS(1302),
    [aux_sym_require_expression_token1] = ACTIONS(1302),
    [aux_sym_require_once_expression_token1] = ACTIONS(1302),
    [sym_comment] = ACTIONS(5),
  },
  [518] = {
    [sym_text_interpolation] = STATE(518),
    [ts_builtin_sym_end] = ACTIONS(1304),
    [sym_name] = ACTIONS(1306),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1304),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1306),
    [aux_sym_global_declaration_token1] = ACTIONS(1306),
    [aux_sym_namespace_definition_token1] = ACTIONS(1306),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1306),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1306),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1306),
    [anon_sym_BSLASH] = ACTIONS(1304),
    [anon_sym_LBRACE] = ACTIONS(1304),
    [anon_sym_RBRACE] = ACTIONS(1304),
    [aux_sym_trait_declaration_token1] = ACTIONS(1306),
    [aux_sym_interface_declaration_token1] = ACTIONS(1306),
    [aux_sym_enum_declaration_token1] = ACTIONS(1306),
    [aux_sym_enum_case_token1] = ACTIONS(1306),
    [aux_sym_class_declaration_token1] = ACTIONS(1306),
    [aux_sym_final_modifier_token1] = ACTIONS(1306),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1306),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1306),
    [sym_var_modifier] = ACTIONS(1306),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1306),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1306),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1306),
    [anon_sym_LPAREN2] = ACTIONS(1304),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1306),
    [aux_sym_cast_type_token1] = ACTIONS(1306),
    [aux_sym_echo_statement_token1] = ACTIONS(1306),
    [aux_sym_exit_statement_token1] = ACTIONS(1306),
    [anon_sym_unset] = ACTIONS(1306),
    [aux_sym_declare_statement_token1] = ACTIONS(1306),
    [aux_sym_declare_statement_token2] = ACTIONS(1306),
    [sym_float] = ACTIONS(1306),
    [aux_sym_try_statement_token1] = ACTIONS(1306),
    [aux_sym_goto_statement_token1] = ACTIONS(1306),
    [aux_sym_continue_statement_token1] = ACTIONS(1306),
    [aux_sym_break_statement_token1] = ACTIONS(1306),
    [sym_integer] = ACTIONS(1306),
    [aux_sym_return_statement_token1] = ACTIONS(1306),
    [aux_sym_throw_expression_token1] = ACTIONS(1306),
    [aux_sym_while_statement_token1] = ACTIONS(1306),
    [aux_sym_while_statement_token2] = ACTIONS(1306),
    [aux_sym_do_statement_token1] = ACTIONS(1306),
    [aux_sym_for_statement_token1] = ACTIONS(1306),
    [aux_sym_for_statement_token2] = ACTIONS(1306),
    [aux_sym_foreach_statement_token1] = ACTIONS(1306),
    [aux_sym_foreach_statement_token2] = ACTIONS(1306),
    [aux_sym_if_statement_token1] = ACTIONS(1306),
    [aux_sym_if_statement_token2] = ACTIONS(1306),
    [aux_sym_else_if_clause_token1] = ACTIONS(1306),
    [aux_sym_else_clause_token1] = ACTIONS(1306),
    [aux_sym_match_expression_token1] = ACTIONS(1306),
    [aux_sym_match_default_expression_token1] = ACTIONS(1306),
    [aux_sym_switch_statement_token1] = ACTIONS(1306),
    [aux_sym_switch_block_token1] = ACTIONS(1306),
    [anon_sym_PLUS] = ACTIONS(1306),
    [anon_sym_DASH] = ACTIONS(1306),
    [anon_sym_TILDE] = ACTIONS(1304),
    [anon_sym_BANG] = ACTIONS(1304),
    [anon_sym_AT] = ACTIONS(1304),
    [aux_sym_clone_expression_token1] = ACTIONS(1306),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1306),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1306),
    [anon_sym_DASH_DASH] = ACTIONS(1304),
    [anon_sym_PLUS_PLUS] = ACTIONS(1304),
    [aux_sym__list_destructing_token1] = ACTIONS(1306),
    [anon_sym_LBRACK] = ACTIONS(1304),
    [anon_sym_self] = ACTIONS(1306),
    [anon_sym_parent] = ACTIONS(1306),
    [aux_sym__argument_name_token1] = ACTIONS(1306),
    [aux_sym__argument_name_token2] = ACTIONS(1306),
    [anon_sym_POUND_LBRACK] = ACTIONS(1304),
    [aux_sym_encapsed_string_token1] = ACTIONS(1304),
    [anon_sym_DQUOTE] = ACTIONS(1304),
    [aux_sym_string_token1] = ACTIONS(1304),
    [anon_sym_SQUOTE] = ACTIONS(1304),
    [anon_sym_LT_LT_LT] = ACTIONS(1304),
    [anon_sym_BQUOTE] = ACTIONS(1304),
    [anon_sym_DOLLAR] = ACTIONS(1304),
    [aux_sym_yield_expression_token1] = ACTIONS(1306),
    [aux_sym_include_expression_token1] = ACTIONS(1306),
    [aux_sym_include_once_expression_token1] = ACTIONS(1306),
    [aux_sym_require_expression_token1] = ACTIONS(1306),
    [aux_sym_require_once_expression_token1] = ACTIONS(1306),
    [sym_comment] = ACTIONS(5),
  },
  [519] = {
    [sym_text_interpolation] = STATE(519),
    [ts_builtin_sym_end] = ACTIONS(1308),
    [sym_name] = ACTIONS(1310),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1308),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1310),
    [aux_sym_global_declaration_token1] = ACTIONS(1310),
    [aux_sym_namespace_definition_token1] = ACTIONS(1310),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1310),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1310),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1310),
    [anon_sym_BSLASH] = ACTIONS(1308),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_trait_declaration_token1] = ACTIONS(1310),
    [aux_sym_interface_declaration_token1] = ACTIONS(1310),
    [aux_sym_enum_declaration_token1] = ACTIONS(1310),
    [aux_sym_enum_case_token1] = ACTIONS(1310),
    [aux_sym_class_declaration_token1] = ACTIONS(1310),
    [aux_sym_final_modifier_token1] = ACTIONS(1310),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1310),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1310),
    [sym_var_modifier] = ACTIONS(1310),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1310),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1310),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1310),
    [anon_sym_LPAREN2] = ACTIONS(1308),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1310),
    [aux_sym_cast_type_token1] = ACTIONS(1310),
    [aux_sym_echo_statement_token1] = ACTIONS(1310),
    [aux_sym_exit_statement_token1] = ACTIONS(1310),
    [anon_sym_unset] = ACTIONS(1310),
    [aux_sym_declare_statement_token1] = ACTIONS(1310),
    [aux_sym_declare_statement_token2] = ACTIONS(1310),
    [sym_float] = ACTIONS(1310),
    [aux_sym_try_statement_token1] = ACTIONS(1310),
    [aux_sym_goto_statement_token1] = ACTIONS(1310),
    [aux_sym_continue_statement_token1] = ACTIONS(1310),
    [aux_sym_break_statement_token1] = ACTIONS(1310),
    [sym_integer] = ACTIONS(1310),
    [aux_sym_return_statement_token1] = ACTIONS(1310),
    [aux_sym_throw_expression_token1] = ACTIONS(1310),
    [aux_sym_while_statement_token1] = ACTIONS(1310),
    [aux_sym_while_statement_token2] = ACTIONS(1310),
    [aux_sym_do_statement_token1] = ACTIONS(1310),
    [aux_sym_for_statement_token1] = ACTIONS(1310),
    [aux_sym_for_statement_token2] = ACTIONS(1310),
    [aux_sym_foreach_statement_token1] = ACTIONS(1310),
    [aux_sym_foreach_statement_token2] = ACTIONS(1310),
    [aux_sym_if_statement_token1] = ACTIONS(1310),
    [aux_sym_if_statement_token2] = ACTIONS(1310),
    [aux_sym_else_if_clause_token1] = ACTIONS(1310),
    [aux_sym_else_clause_token1] = ACTIONS(1310),
    [aux_sym_match_expression_token1] = ACTIONS(1310),
    [aux_sym_match_default_expression_token1] = ACTIONS(1310),
    [aux_sym_switch_statement_token1] = ACTIONS(1310),
    [aux_sym_switch_block_token1] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1308),
    [anon_sym_AT] = ACTIONS(1308),
    [aux_sym_clone_expression_token1] = ACTIONS(1310),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1310),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1310),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [aux_sym__list_destructing_token1] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1308),
    [anon_sym_self] = ACTIONS(1310),
    [anon_sym_parent] = ACTIONS(1310),
    [aux_sym__argument_name_token1] = ACTIONS(1310),
    [aux_sym__argument_name_token2] = ACTIONS(1310),
    [anon_sym_POUND_LBRACK] = ACTIONS(1308),
    [aux_sym_encapsed_string_token1] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1308),
    [aux_sym_string_token1] = ACTIONS(1308),
    [anon_sym_SQUOTE] = ACTIONS(1308),
    [anon_sym_LT_LT_LT] = ACTIONS(1308),
    [anon_sym_BQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR] = ACTIONS(1308),
    [aux_sym_yield_expression_token1] = ACTIONS(1310),
    [aux_sym_include_expression_token1] = ACTIONS(1310),
    [aux_sym_include_once_expression_token1] = ACTIONS(1310),
    [aux_sym_require_expression_token1] = ACTIONS(1310),
    [aux_sym_require_once_expression_token1] = ACTIONS(1310),
    [sym_comment] = ACTIONS(5),
  },
  [520] = {
    [sym_text_interpolation] = STATE(520),
    [ts_builtin_sym_end] = ACTIONS(1056),
    [sym_name] = ACTIONS(1058),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1056),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1058),
    [aux_sym_global_declaration_token1] = ACTIONS(1058),
    [aux_sym_namespace_definition_token1] = ACTIONS(1058),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1058),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1058),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1058),
    [anon_sym_BSLASH] = ACTIONS(1056),
    [anon_sym_LBRACE] = ACTIONS(1056),
    [anon_sym_RBRACE] = ACTIONS(1056),
    [aux_sym_trait_declaration_token1] = ACTIONS(1058),
    [aux_sym_interface_declaration_token1] = ACTIONS(1058),
    [aux_sym_enum_declaration_token1] = ACTIONS(1058),
    [aux_sym_enum_case_token1] = ACTIONS(1058),
    [aux_sym_class_declaration_token1] = ACTIONS(1058),
    [aux_sym_final_modifier_token1] = ACTIONS(1058),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1058),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1058),
    [sym_var_modifier] = ACTIONS(1058),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1058),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1058),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1058),
    [anon_sym_LPAREN2] = ACTIONS(1056),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1058),
    [aux_sym_cast_type_token1] = ACTIONS(1058),
    [aux_sym_echo_statement_token1] = ACTIONS(1058),
    [aux_sym_exit_statement_token1] = ACTIONS(1058),
    [anon_sym_unset] = ACTIONS(1058),
    [aux_sym_declare_statement_token1] = ACTIONS(1058),
    [aux_sym_declare_statement_token2] = ACTIONS(1058),
    [sym_float] = ACTIONS(1058),
    [aux_sym_try_statement_token1] = ACTIONS(1058),
    [aux_sym_goto_statement_token1] = ACTIONS(1058),
    [aux_sym_continue_statement_token1] = ACTIONS(1058),
    [aux_sym_break_statement_token1] = ACTIONS(1058),
    [sym_integer] = ACTIONS(1058),
    [aux_sym_return_statement_token1] = ACTIONS(1058),
    [aux_sym_throw_expression_token1] = ACTIONS(1058),
    [aux_sym_while_statement_token1] = ACTIONS(1058),
    [aux_sym_while_statement_token2] = ACTIONS(1058),
    [aux_sym_do_statement_token1] = ACTIONS(1058),
    [aux_sym_for_statement_token1] = ACTIONS(1058),
    [aux_sym_for_statement_token2] = ACTIONS(1058),
    [aux_sym_foreach_statement_token1] = ACTIONS(1058),
    [aux_sym_foreach_statement_token2] = ACTIONS(1058),
    [aux_sym_if_statement_token1] = ACTIONS(1058),
    [aux_sym_if_statement_token2] = ACTIONS(1058),
    [aux_sym_else_if_clause_token1] = ACTIONS(1058),
    [aux_sym_else_clause_token1] = ACTIONS(1058),
    [aux_sym_match_expression_token1] = ACTIONS(1058),
    [aux_sym_match_default_expression_token1] = ACTIONS(1058),
    [aux_sym_switch_statement_token1] = ACTIONS(1058),
    [aux_sym_switch_block_token1] = ACTIONS(1058),
    [anon_sym_PLUS] = ACTIONS(1058),
    [anon_sym_DASH] = ACTIONS(1058),
    [anon_sym_TILDE] = ACTIONS(1056),
    [anon_sym_BANG] = ACTIONS(1056),
    [anon_sym_AT] = ACTIONS(1056),
    [aux_sym_clone_expression_token1] = ACTIONS(1058),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1058),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1058),
    [anon_sym_DASH_DASH] = ACTIONS(1056),
    [anon_sym_PLUS_PLUS] = ACTIONS(1056),
    [aux_sym__list_destructing_token1] = ACTIONS(1058),
    [anon_sym_LBRACK] = ACTIONS(1056),
    [anon_sym_self] = ACTIONS(1058),
    [anon_sym_parent] = ACTIONS(1058),
    [aux_sym__argument_name_token1] = ACTIONS(1058),
    [aux_sym__argument_name_token2] = ACTIONS(1058),
    [anon_sym_POUND_LBRACK] = ACTIONS(1056),
    [aux_sym_encapsed_string_token1] = ACTIONS(1056),
    [anon_sym_DQUOTE] = ACTIONS(1056),
    [aux_sym_string_token1] = ACTIONS(1056),
    [anon_sym_SQUOTE] = ACTIONS(1056),
    [anon_sym_LT_LT_LT] = ACTIONS(1056),
    [anon_sym_BQUOTE] = ACTIONS(1056),
    [anon_sym_DOLLAR] = ACTIONS(1056),
    [aux_sym_yield_expression_token1] = ACTIONS(1058),
    [aux_sym_include_expression_token1] = ACTIONS(1058),
    [aux_sym_include_once_expression_token1] = ACTIONS(1058),
    [aux_sym_require_expression_token1] = ACTIONS(1058),
    [aux_sym_require_once_expression_token1] = ACTIONS(1058),
    [sym_comment] = ACTIONS(5),
  },
  [521] = {
    [sym_text_interpolation] = STATE(521),
    [ts_builtin_sym_end] = ACTIONS(1312),
    [sym_name] = ACTIONS(1314),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1312),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1314),
    [aux_sym_global_declaration_token1] = ACTIONS(1314),
    [aux_sym_namespace_definition_token1] = ACTIONS(1314),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1314),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1314),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1314),
    [anon_sym_BSLASH] = ACTIONS(1312),
    [anon_sym_LBRACE] = ACTIONS(1312),
    [anon_sym_RBRACE] = ACTIONS(1312),
    [aux_sym_trait_declaration_token1] = ACTIONS(1314),
    [aux_sym_interface_declaration_token1] = ACTIONS(1314),
    [aux_sym_enum_declaration_token1] = ACTIONS(1314),
    [aux_sym_enum_case_token1] = ACTIONS(1314),
    [aux_sym_class_declaration_token1] = ACTIONS(1314),
    [aux_sym_final_modifier_token1] = ACTIONS(1314),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1314),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1314),
    [sym_var_modifier] = ACTIONS(1314),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1314),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1314),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1314),
    [anon_sym_LPAREN2] = ACTIONS(1312),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1314),
    [aux_sym_cast_type_token1] = ACTIONS(1314),
    [aux_sym_echo_statement_token1] = ACTIONS(1314),
    [aux_sym_exit_statement_token1] = ACTIONS(1314),
    [anon_sym_unset] = ACTIONS(1314),
    [aux_sym_declare_statement_token1] = ACTIONS(1314),
    [aux_sym_declare_statement_token2] = ACTIONS(1314),
    [sym_float] = ACTIONS(1314),
    [aux_sym_try_statement_token1] = ACTIONS(1314),
    [aux_sym_goto_statement_token1] = ACTIONS(1314),
    [aux_sym_continue_statement_token1] = ACTIONS(1314),
    [aux_sym_break_statement_token1] = ACTIONS(1314),
    [sym_integer] = ACTIONS(1314),
    [aux_sym_return_statement_token1] = ACTIONS(1314),
    [aux_sym_throw_expression_token1] = ACTIONS(1314),
    [aux_sym_while_statement_token1] = ACTIONS(1314),
    [aux_sym_while_statement_token2] = ACTIONS(1314),
    [aux_sym_do_statement_token1] = ACTIONS(1314),
    [aux_sym_for_statement_token1] = ACTIONS(1314),
    [aux_sym_for_statement_token2] = ACTIONS(1314),
    [aux_sym_foreach_statement_token1] = ACTIONS(1314),
    [aux_sym_foreach_statement_token2] = ACTIONS(1314),
    [aux_sym_if_statement_token1] = ACTIONS(1314),
    [aux_sym_if_statement_token2] = ACTIONS(1314),
    [aux_sym_else_if_clause_token1] = ACTIONS(1314),
    [aux_sym_else_clause_token1] = ACTIONS(1314),
    [aux_sym_match_expression_token1] = ACTIONS(1314),
    [aux_sym_match_default_expression_token1] = ACTIONS(1314),
    [aux_sym_switch_statement_token1] = ACTIONS(1314),
    [aux_sym_switch_block_token1] = ACTIONS(1314),
    [anon_sym_PLUS] = ACTIONS(1314),
    [anon_sym_DASH] = ACTIONS(1314),
    [anon_sym_TILDE] = ACTIONS(1312),
    [anon_sym_BANG] = ACTIONS(1312),
    [anon_sym_AT] = ACTIONS(1312),
    [aux_sym_clone_expression_token1] = ACTIONS(1314),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1314),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1314),
    [anon_sym_DASH_DASH] = ACTIONS(1312),
    [anon_sym_PLUS_PLUS] = ACTIONS(1312),
    [aux_sym__list_destructing_token1] = ACTIONS(1314),
    [anon_sym_LBRACK] = ACTIONS(1312),
    [anon_sym_self] = ACTIONS(1314),
    [anon_sym_parent] = ACTIONS(1314),
    [aux_sym__argument_name_token1] = ACTIONS(1314),
    [aux_sym__argument_name_token2] = ACTIONS(1314),
    [anon_sym_POUND_LBRACK] = ACTIONS(1312),
    [aux_sym_encapsed_string_token1] = ACTIONS(1312),
    [anon_sym_DQUOTE] = ACTIONS(1312),
    [aux_sym_string_token1] = ACTIONS(1312),
    [anon_sym_SQUOTE] = ACTIONS(1312),
    [anon_sym_LT_LT_LT] = ACTIONS(1312),
    [anon_sym_BQUOTE] = ACTIONS(1312),
    [anon_sym_DOLLAR] = ACTIONS(1312),
    [aux_sym_yield_expression_token1] = ACTIONS(1314),
    [aux_sym_include_expression_token1] = ACTIONS(1314),
    [aux_sym_include_once_expression_token1] = ACTIONS(1314),
    [aux_sym_require_expression_token1] = ACTIONS(1314),
    [aux_sym_require_once_expression_token1] = ACTIONS(1314),
    [sym_comment] = ACTIONS(5),
  },
  [522] = {
    [sym_text_interpolation] = STATE(522),
    [ts_builtin_sym_end] = ACTIONS(1316),
    [sym_name] = ACTIONS(1318),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1316),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1318),
    [aux_sym_global_declaration_token1] = ACTIONS(1318),
    [aux_sym_namespace_definition_token1] = ACTIONS(1318),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1318),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1318),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1318),
    [anon_sym_BSLASH] = ACTIONS(1316),
    [anon_sym_LBRACE] = ACTIONS(1316),
    [anon_sym_RBRACE] = ACTIONS(1316),
    [aux_sym_trait_declaration_token1] = ACTIONS(1318),
    [aux_sym_interface_declaration_token1] = ACTIONS(1318),
    [aux_sym_enum_declaration_token1] = ACTIONS(1318),
    [aux_sym_enum_case_token1] = ACTIONS(1318),
    [aux_sym_class_declaration_token1] = ACTIONS(1318),
    [aux_sym_final_modifier_token1] = ACTIONS(1318),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1318),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1318),
    [sym_var_modifier] = ACTIONS(1318),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1318),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1318),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1318),
    [anon_sym_LPAREN2] = ACTIONS(1316),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1318),
    [aux_sym_cast_type_token1] = ACTIONS(1318),
    [aux_sym_echo_statement_token1] = ACTIONS(1318),
    [aux_sym_exit_statement_token1] = ACTIONS(1318),
    [anon_sym_unset] = ACTIONS(1318),
    [aux_sym_declare_statement_token1] = ACTIONS(1318),
    [aux_sym_declare_statement_token2] = ACTIONS(1318),
    [sym_float] = ACTIONS(1318),
    [aux_sym_try_statement_token1] = ACTIONS(1318),
    [aux_sym_goto_statement_token1] = ACTIONS(1318),
    [aux_sym_continue_statement_token1] = ACTIONS(1318),
    [aux_sym_break_statement_token1] = ACTIONS(1318),
    [sym_integer] = ACTIONS(1318),
    [aux_sym_return_statement_token1] = ACTIONS(1318),
    [aux_sym_throw_expression_token1] = ACTIONS(1318),
    [aux_sym_while_statement_token1] = ACTIONS(1318),
    [aux_sym_while_statement_token2] = ACTIONS(1318),
    [aux_sym_do_statement_token1] = ACTIONS(1318),
    [aux_sym_for_statement_token1] = ACTIONS(1318),
    [aux_sym_for_statement_token2] = ACTIONS(1318),
    [aux_sym_foreach_statement_token1] = ACTIONS(1318),
    [aux_sym_foreach_statement_token2] = ACTIONS(1318),
    [aux_sym_if_statement_token1] = ACTIONS(1318),
    [aux_sym_if_statement_token2] = ACTIONS(1318),
    [aux_sym_else_if_clause_token1] = ACTIONS(1318),
    [aux_sym_else_clause_token1] = ACTIONS(1318),
    [aux_sym_match_expression_token1] = ACTIONS(1318),
    [aux_sym_match_default_expression_token1] = ACTIONS(1318),
    [aux_sym_switch_statement_token1] = ACTIONS(1318),
    [aux_sym_switch_block_token1] = ACTIONS(1318),
    [anon_sym_PLUS] = ACTIONS(1318),
    [anon_sym_DASH] = ACTIONS(1318),
    [anon_sym_TILDE] = ACTIONS(1316),
    [anon_sym_BANG] = ACTIONS(1316),
    [anon_sym_AT] = ACTIONS(1316),
    [aux_sym_clone_expression_token1] = ACTIONS(1318),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1318),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1318),
    [anon_sym_DASH_DASH] = ACTIONS(1316),
    [anon_sym_PLUS_PLUS] = ACTIONS(1316),
    [aux_sym__list_destructing_token1] = ACTIONS(1318),
    [anon_sym_LBRACK] = ACTIONS(1316),
    [anon_sym_self] = ACTIONS(1318),
    [anon_sym_parent] = ACTIONS(1318),
    [aux_sym__argument_name_token1] = ACTIONS(1318),
    [aux_sym__argument_name_token2] = ACTIONS(1318),
    [anon_sym_POUND_LBRACK] = ACTIONS(1316),
    [aux_sym_encapsed_string_token1] = ACTIONS(1316),
    [anon_sym_DQUOTE] = ACTIONS(1316),
    [aux_sym_string_token1] = ACTIONS(1316),
    [anon_sym_SQUOTE] = ACTIONS(1316),
    [anon_sym_LT_LT_LT] = ACTIONS(1316),
    [anon_sym_BQUOTE] = ACTIONS(1316),
    [anon_sym_DOLLAR] = ACTIONS(1316),
    [aux_sym_yield_expression_token1] = ACTIONS(1318),
    [aux_sym_include_expression_token1] = ACTIONS(1318),
    [aux_sym_include_once_expression_token1] = ACTIONS(1318),
    [aux_sym_require_expression_token1] = ACTIONS(1318),
    [aux_sym_require_once_expression_token1] = ACTIONS(1318),
    [sym_comment] = ACTIONS(5),
  },
  [523] = {
    [sym_text_interpolation] = STATE(523),
    [ts_builtin_sym_end] = ACTIONS(1320),
    [sym_name] = ACTIONS(1322),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1320),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1322),
    [aux_sym_global_declaration_token1] = ACTIONS(1322),
    [aux_sym_namespace_definition_token1] = ACTIONS(1322),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1322),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1322),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1322),
    [anon_sym_BSLASH] = ACTIONS(1320),
    [anon_sym_LBRACE] = ACTIONS(1320),
    [anon_sym_RBRACE] = ACTIONS(1320),
    [aux_sym_trait_declaration_token1] = ACTIONS(1322),
    [aux_sym_interface_declaration_token1] = ACTIONS(1322),
    [aux_sym_enum_declaration_token1] = ACTIONS(1322),
    [aux_sym_enum_case_token1] = ACTIONS(1322),
    [aux_sym_class_declaration_token1] = ACTIONS(1322),
    [aux_sym_final_modifier_token1] = ACTIONS(1322),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1322),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1322),
    [sym_var_modifier] = ACTIONS(1322),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1322),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1322),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1322),
    [anon_sym_LPAREN2] = ACTIONS(1320),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1322),
    [aux_sym_cast_type_token1] = ACTIONS(1322),
    [aux_sym_echo_statement_token1] = ACTIONS(1322),
    [aux_sym_exit_statement_token1] = ACTIONS(1322),
    [anon_sym_unset] = ACTIONS(1322),
    [aux_sym_declare_statement_token1] = ACTIONS(1322),
    [aux_sym_declare_statement_token2] = ACTIONS(1322),
    [sym_float] = ACTIONS(1322),
    [aux_sym_try_statement_token1] = ACTIONS(1322),
    [aux_sym_goto_statement_token1] = ACTIONS(1322),
    [aux_sym_continue_statement_token1] = ACTIONS(1322),
    [aux_sym_break_statement_token1] = ACTIONS(1322),
    [sym_integer] = ACTIONS(1322),
    [aux_sym_return_statement_token1] = ACTIONS(1322),
    [aux_sym_throw_expression_token1] = ACTIONS(1322),
    [aux_sym_while_statement_token1] = ACTIONS(1322),
    [aux_sym_while_statement_token2] = ACTIONS(1322),
    [aux_sym_do_statement_token1] = ACTIONS(1322),
    [aux_sym_for_statement_token1] = ACTIONS(1322),
    [aux_sym_for_statement_token2] = ACTIONS(1322),
    [aux_sym_foreach_statement_token1] = ACTIONS(1322),
    [aux_sym_foreach_statement_token2] = ACTIONS(1322),
    [aux_sym_if_statement_token1] = ACTIONS(1322),
    [aux_sym_if_statement_token2] = ACTIONS(1322),
    [aux_sym_else_if_clause_token1] = ACTIONS(1322),
    [aux_sym_else_clause_token1] = ACTIONS(1322),
    [aux_sym_match_expression_token1] = ACTIONS(1322),
    [aux_sym_match_default_expression_token1] = ACTIONS(1322),
    [aux_sym_switch_statement_token1] = ACTIONS(1322),
    [aux_sym_switch_block_token1] = ACTIONS(1322),
    [anon_sym_PLUS] = ACTIONS(1322),
    [anon_sym_DASH] = ACTIONS(1322),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_BANG] = ACTIONS(1320),
    [anon_sym_AT] = ACTIONS(1320),
    [aux_sym_clone_expression_token1] = ACTIONS(1322),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1322),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1322),
    [anon_sym_DASH_DASH] = ACTIONS(1320),
    [anon_sym_PLUS_PLUS] = ACTIONS(1320),
    [aux_sym__list_destructing_token1] = ACTIONS(1322),
    [anon_sym_LBRACK] = ACTIONS(1320),
    [anon_sym_self] = ACTIONS(1322),
    [anon_sym_parent] = ACTIONS(1322),
    [aux_sym__argument_name_token1] = ACTIONS(1322),
    [aux_sym__argument_name_token2] = ACTIONS(1322),
    [anon_sym_POUND_LBRACK] = ACTIONS(1320),
    [aux_sym_encapsed_string_token1] = ACTIONS(1320),
    [anon_sym_DQUOTE] = ACTIONS(1320),
    [aux_sym_string_token1] = ACTIONS(1320),
    [anon_sym_SQUOTE] = ACTIONS(1320),
    [anon_sym_LT_LT_LT] = ACTIONS(1320),
    [anon_sym_BQUOTE] = ACTIONS(1320),
    [anon_sym_DOLLAR] = ACTIONS(1320),
    [aux_sym_yield_expression_token1] = ACTIONS(1322),
    [aux_sym_include_expression_token1] = ACTIONS(1322),
    [aux_sym_include_once_expression_token1] = ACTIONS(1322),
    [aux_sym_require_expression_token1] = ACTIONS(1322),
    [aux_sym_require_once_expression_token1] = ACTIONS(1322),
    [sym_comment] = ACTIONS(5),
  },
  [524] = {
    [sym_text_interpolation] = STATE(524),
    [ts_builtin_sym_end] = ACTIONS(1324),
    [sym_name] = ACTIONS(1326),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1324),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1326),
    [aux_sym_global_declaration_token1] = ACTIONS(1326),
    [aux_sym_namespace_definition_token1] = ACTIONS(1326),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1326),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1326),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1326),
    [anon_sym_BSLASH] = ACTIONS(1324),
    [anon_sym_LBRACE] = ACTIONS(1324),
    [anon_sym_RBRACE] = ACTIONS(1324),
    [aux_sym_trait_declaration_token1] = ACTIONS(1326),
    [aux_sym_interface_declaration_token1] = ACTIONS(1326),
    [aux_sym_enum_declaration_token1] = ACTIONS(1326),
    [aux_sym_enum_case_token1] = ACTIONS(1326),
    [aux_sym_class_declaration_token1] = ACTIONS(1326),
    [aux_sym_final_modifier_token1] = ACTIONS(1326),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1326),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1326),
    [sym_var_modifier] = ACTIONS(1326),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1326),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1326),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1326),
    [anon_sym_LPAREN2] = ACTIONS(1324),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1326),
    [aux_sym_cast_type_token1] = ACTIONS(1326),
    [aux_sym_echo_statement_token1] = ACTIONS(1326),
    [aux_sym_exit_statement_token1] = ACTIONS(1326),
    [anon_sym_unset] = ACTIONS(1326),
    [aux_sym_declare_statement_token1] = ACTIONS(1326),
    [aux_sym_declare_statement_token2] = ACTIONS(1326),
    [sym_float] = ACTIONS(1326),
    [aux_sym_try_statement_token1] = ACTIONS(1326),
    [aux_sym_goto_statement_token1] = ACTIONS(1326),
    [aux_sym_continue_statement_token1] = ACTIONS(1326),
    [aux_sym_break_statement_token1] = ACTIONS(1326),
    [sym_integer] = ACTIONS(1326),
    [aux_sym_return_statement_token1] = ACTIONS(1326),
    [aux_sym_throw_expression_token1] = ACTIONS(1326),
    [aux_sym_while_statement_token1] = ACTIONS(1326),
    [aux_sym_while_statement_token2] = ACTIONS(1326),
    [aux_sym_do_statement_token1] = ACTIONS(1326),
    [aux_sym_for_statement_token1] = ACTIONS(1326),
    [aux_sym_for_statement_token2] = ACTIONS(1326),
    [aux_sym_foreach_statement_token1] = ACTIONS(1326),
    [aux_sym_foreach_statement_token2] = ACTIONS(1326),
    [aux_sym_if_statement_token1] = ACTIONS(1326),
    [aux_sym_if_statement_token2] = ACTIONS(1326),
    [aux_sym_else_if_clause_token1] = ACTIONS(1326),
    [aux_sym_else_clause_token1] = ACTIONS(1326),
    [aux_sym_match_expression_token1] = ACTIONS(1326),
    [aux_sym_match_default_expression_token1] = ACTIONS(1326),
    [aux_sym_switch_statement_token1] = ACTIONS(1326),
    [aux_sym_switch_block_token1] = ACTIONS(1326),
    [anon_sym_PLUS] = ACTIONS(1326),
    [anon_sym_DASH] = ACTIONS(1326),
    [anon_sym_TILDE] = ACTIONS(1324),
    [anon_sym_BANG] = ACTIONS(1324),
    [anon_sym_AT] = ACTIONS(1324),
    [aux_sym_clone_expression_token1] = ACTIONS(1326),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1326),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1326),
    [anon_sym_DASH_DASH] = ACTIONS(1324),
    [anon_sym_PLUS_PLUS] = ACTIONS(1324),
    [aux_sym__list_destructing_token1] = ACTIONS(1326),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_self] = ACTIONS(1326),
    [anon_sym_parent] = ACTIONS(1326),
    [aux_sym__argument_name_token1] = ACTIONS(1326),
    [aux_sym__argument_name_token2] = ACTIONS(1326),
    [anon_sym_POUND_LBRACK] = ACTIONS(1324),
    [aux_sym_encapsed_string_token1] = ACTIONS(1324),
    [anon_sym_DQUOTE] = ACTIONS(1324),
    [aux_sym_string_token1] = ACTIONS(1324),
    [anon_sym_SQUOTE] = ACTIONS(1324),
    [anon_sym_LT_LT_LT] = ACTIONS(1324),
    [anon_sym_BQUOTE] = ACTIONS(1324),
    [anon_sym_DOLLAR] = ACTIONS(1324),
    [aux_sym_yield_expression_token1] = ACTIONS(1326),
    [aux_sym_include_expression_token1] = ACTIONS(1326),
    [aux_sym_include_once_expression_token1] = ACTIONS(1326),
    [aux_sym_require_expression_token1] = ACTIONS(1326),
    [aux_sym_require_once_expression_token1] = ACTIONS(1326),
    [sym_comment] = ACTIONS(5),
  },
  [525] = {
    [sym_text_interpolation] = STATE(525),
    [ts_builtin_sym_end] = ACTIONS(1328),
    [sym_name] = ACTIONS(1330),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1328),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1330),
    [aux_sym_global_declaration_token1] = ACTIONS(1330),
    [aux_sym_namespace_definition_token1] = ACTIONS(1330),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1330),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1330),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1330),
    [anon_sym_BSLASH] = ACTIONS(1328),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_RBRACE] = ACTIONS(1328),
    [aux_sym_trait_declaration_token1] = ACTIONS(1330),
    [aux_sym_interface_declaration_token1] = ACTIONS(1330),
    [aux_sym_enum_declaration_token1] = ACTIONS(1330),
    [aux_sym_enum_case_token1] = ACTIONS(1330),
    [aux_sym_class_declaration_token1] = ACTIONS(1330),
    [aux_sym_final_modifier_token1] = ACTIONS(1330),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1330),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1330),
    [sym_var_modifier] = ACTIONS(1330),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1330),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1330),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1330),
    [anon_sym_LPAREN2] = ACTIONS(1328),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1330),
    [aux_sym_cast_type_token1] = ACTIONS(1330),
    [aux_sym_echo_statement_token1] = ACTIONS(1330),
    [aux_sym_exit_statement_token1] = ACTIONS(1330),
    [anon_sym_unset] = ACTIONS(1330),
    [aux_sym_declare_statement_token1] = ACTIONS(1330),
    [aux_sym_declare_statement_token2] = ACTIONS(1330),
    [sym_float] = ACTIONS(1330),
    [aux_sym_try_statement_token1] = ACTIONS(1330),
    [aux_sym_goto_statement_token1] = ACTIONS(1330),
    [aux_sym_continue_statement_token1] = ACTIONS(1330),
    [aux_sym_break_statement_token1] = ACTIONS(1330),
    [sym_integer] = ACTIONS(1330),
    [aux_sym_return_statement_token1] = ACTIONS(1330),
    [aux_sym_throw_expression_token1] = ACTIONS(1330),
    [aux_sym_while_statement_token1] = ACTIONS(1330),
    [aux_sym_while_statement_token2] = ACTIONS(1330),
    [aux_sym_do_statement_token1] = ACTIONS(1330),
    [aux_sym_for_statement_token1] = ACTIONS(1330),
    [aux_sym_for_statement_token2] = ACTIONS(1330),
    [aux_sym_foreach_statement_token1] = ACTIONS(1330),
    [aux_sym_foreach_statement_token2] = ACTIONS(1330),
    [aux_sym_if_statement_token1] = ACTIONS(1330),
    [aux_sym_if_statement_token2] = ACTIONS(1330),
    [aux_sym_else_if_clause_token1] = ACTIONS(1330),
    [aux_sym_else_clause_token1] = ACTIONS(1330),
    [aux_sym_match_expression_token1] = ACTIONS(1330),
    [aux_sym_match_default_expression_token1] = ACTIONS(1330),
    [aux_sym_switch_statement_token1] = ACTIONS(1330),
    [aux_sym_switch_block_token1] = ACTIONS(1330),
    [anon_sym_PLUS] = ACTIONS(1330),
    [anon_sym_DASH] = ACTIONS(1330),
    [anon_sym_TILDE] = ACTIONS(1328),
    [anon_sym_BANG] = ACTIONS(1328),
    [anon_sym_AT] = ACTIONS(1328),
    [aux_sym_clone_expression_token1] = ACTIONS(1330),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1330),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1330),
    [anon_sym_DASH_DASH] = ACTIONS(1328),
    [anon_sym_PLUS_PLUS] = ACTIONS(1328),
    [aux_sym__list_destructing_token1] = ACTIONS(1330),
    [anon_sym_LBRACK] = ACTIONS(1328),
    [anon_sym_self] = ACTIONS(1330),
    [anon_sym_parent] = ACTIONS(1330),
    [aux_sym__argument_name_token1] = ACTIONS(1330),
    [aux_sym__argument_name_token2] = ACTIONS(1330),
    [anon_sym_POUND_LBRACK] = ACTIONS(1328),
    [aux_sym_encapsed_string_token1] = ACTIONS(1328),
    [anon_sym_DQUOTE] = ACTIONS(1328),
    [aux_sym_string_token1] = ACTIONS(1328),
    [anon_sym_SQUOTE] = ACTIONS(1328),
    [anon_sym_LT_LT_LT] = ACTIONS(1328),
    [anon_sym_BQUOTE] = ACTIONS(1328),
    [anon_sym_DOLLAR] = ACTIONS(1328),
    [aux_sym_yield_expression_token1] = ACTIONS(1330),
    [aux_sym_include_expression_token1] = ACTIONS(1330),
    [aux_sym_include_once_expression_token1] = ACTIONS(1330),
    [aux_sym_require_expression_token1] = ACTIONS(1330),
    [aux_sym_require_once_expression_token1] = ACTIONS(1330),
    [sym_comment] = ACTIONS(5),
  },
  [526] = {
    [sym_text_interpolation] = STATE(526),
    [ts_builtin_sym_end] = ACTIONS(1332),
    [sym_name] = ACTIONS(1334),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1332),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1334),
    [aux_sym_global_declaration_token1] = ACTIONS(1334),
    [aux_sym_namespace_definition_token1] = ACTIONS(1334),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1334),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1334),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1334),
    [anon_sym_BSLASH] = ACTIONS(1332),
    [anon_sym_LBRACE] = ACTIONS(1332),
    [anon_sym_RBRACE] = ACTIONS(1332),
    [aux_sym_trait_declaration_token1] = ACTIONS(1334),
    [aux_sym_interface_declaration_token1] = ACTIONS(1334),
    [aux_sym_enum_declaration_token1] = ACTIONS(1334),
    [aux_sym_enum_case_token1] = ACTIONS(1334),
    [aux_sym_class_declaration_token1] = ACTIONS(1334),
    [aux_sym_final_modifier_token1] = ACTIONS(1334),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1334),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1334),
    [sym_var_modifier] = ACTIONS(1334),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1334),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1334),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1334),
    [anon_sym_LPAREN2] = ACTIONS(1332),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1334),
    [aux_sym_cast_type_token1] = ACTIONS(1334),
    [aux_sym_echo_statement_token1] = ACTIONS(1334),
    [aux_sym_exit_statement_token1] = ACTIONS(1334),
    [anon_sym_unset] = ACTIONS(1334),
    [aux_sym_declare_statement_token1] = ACTIONS(1334),
    [aux_sym_declare_statement_token2] = ACTIONS(1334),
    [sym_float] = ACTIONS(1334),
    [aux_sym_try_statement_token1] = ACTIONS(1334),
    [aux_sym_goto_statement_token1] = ACTIONS(1334),
    [aux_sym_continue_statement_token1] = ACTIONS(1334),
    [aux_sym_break_statement_token1] = ACTIONS(1334),
    [sym_integer] = ACTIONS(1334),
    [aux_sym_return_statement_token1] = ACTIONS(1334),
    [aux_sym_throw_expression_token1] = ACTIONS(1334),
    [aux_sym_while_statement_token1] = ACTIONS(1334),
    [aux_sym_while_statement_token2] = ACTIONS(1334),
    [aux_sym_do_statement_token1] = ACTIONS(1334),
    [aux_sym_for_statement_token1] = ACTIONS(1334),
    [aux_sym_for_statement_token2] = ACTIONS(1334),
    [aux_sym_foreach_statement_token1] = ACTIONS(1334),
    [aux_sym_foreach_statement_token2] = ACTIONS(1334),
    [aux_sym_if_statement_token1] = ACTIONS(1334),
    [aux_sym_if_statement_token2] = ACTIONS(1334),
    [aux_sym_else_if_clause_token1] = ACTIONS(1334),
    [aux_sym_else_clause_token1] = ACTIONS(1334),
    [aux_sym_match_expression_token1] = ACTIONS(1334),
    [aux_sym_match_default_expression_token1] = ACTIONS(1334),
    [aux_sym_switch_statement_token1] = ACTIONS(1334),
    [aux_sym_switch_block_token1] = ACTIONS(1334),
    [anon_sym_PLUS] = ACTIONS(1334),
    [anon_sym_DASH] = ACTIONS(1334),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1332),
    [anon_sym_AT] = ACTIONS(1332),
    [aux_sym_clone_expression_token1] = ACTIONS(1334),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1334),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1334),
    [anon_sym_DASH_DASH] = ACTIONS(1332),
    [anon_sym_PLUS_PLUS] = ACTIONS(1332),
    [aux_sym__list_destructing_token1] = ACTIONS(1334),
    [anon_sym_LBRACK] = ACTIONS(1332),
    [anon_sym_self] = ACTIONS(1334),
    [anon_sym_parent] = ACTIONS(1334),
    [aux_sym__argument_name_token1] = ACTIONS(1334),
    [aux_sym__argument_name_token2] = ACTIONS(1334),
    [anon_sym_POUND_LBRACK] = ACTIONS(1332),
    [aux_sym_encapsed_string_token1] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1332),
    [aux_sym_string_token1] = ACTIONS(1332),
    [anon_sym_SQUOTE] = ACTIONS(1332),
    [anon_sym_LT_LT_LT] = ACTIONS(1332),
    [anon_sym_BQUOTE] = ACTIONS(1332),
    [anon_sym_DOLLAR] = ACTIONS(1332),
    [aux_sym_yield_expression_token1] = ACTIONS(1334),
    [aux_sym_include_expression_token1] = ACTIONS(1334),
    [aux_sym_include_once_expression_token1] = ACTIONS(1334),
    [aux_sym_require_expression_token1] = ACTIONS(1334),
    [aux_sym_require_once_expression_token1] = ACTIONS(1334),
    [sym_comment] = ACTIONS(5),
  },
  [527] = {
    [sym_text_interpolation] = STATE(527),
    [ts_builtin_sym_end] = ACTIONS(1336),
    [sym_name] = ACTIONS(1338),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1336),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1338),
    [aux_sym_global_declaration_token1] = ACTIONS(1338),
    [aux_sym_namespace_definition_token1] = ACTIONS(1338),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1338),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1338),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1338),
    [anon_sym_BSLASH] = ACTIONS(1336),
    [anon_sym_LBRACE] = ACTIONS(1336),
    [anon_sym_RBRACE] = ACTIONS(1336),
    [aux_sym_trait_declaration_token1] = ACTIONS(1338),
    [aux_sym_interface_declaration_token1] = ACTIONS(1338),
    [aux_sym_enum_declaration_token1] = ACTIONS(1338),
    [aux_sym_enum_case_token1] = ACTIONS(1338),
    [aux_sym_class_declaration_token1] = ACTIONS(1338),
    [aux_sym_final_modifier_token1] = ACTIONS(1338),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1338),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1338),
    [sym_var_modifier] = ACTIONS(1338),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1338),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1338),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1338),
    [anon_sym_LPAREN2] = ACTIONS(1336),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1338),
    [aux_sym_cast_type_token1] = ACTIONS(1338),
    [aux_sym_echo_statement_token1] = ACTIONS(1338),
    [aux_sym_exit_statement_token1] = ACTIONS(1338),
    [anon_sym_unset] = ACTIONS(1338),
    [aux_sym_declare_statement_token1] = ACTIONS(1338),
    [aux_sym_declare_statement_token2] = ACTIONS(1338),
    [sym_float] = ACTIONS(1338),
    [aux_sym_try_statement_token1] = ACTIONS(1338),
    [aux_sym_goto_statement_token1] = ACTIONS(1338),
    [aux_sym_continue_statement_token1] = ACTIONS(1338),
    [aux_sym_break_statement_token1] = ACTIONS(1338),
    [sym_integer] = ACTIONS(1338),
    [aux_sym_return_statement_token1] = ACTIONS(1338),
    [aux_sym_throw_expression_token1] = ACTIONS(1338),
    [aux_sym_while_statement_token1] = ACTIONS(1338),
    [aux_sym_while_statement_token2] = ACTIONS(1338),
    [aux_sym_do_statement_token1] = ACTIONS(1338),
    [aux_sym_for_statement_token1] = ACTIONS(1338),
    [aux_sym_for_statement_token2] = ACTIONS(1338),
    [aux_sym_foreach_statement_token1] = ACTIONS(1338),
    [aux_sym_foreach_statement_token2] = ACTIONS(1338),
    [aux_sym_if_statement_token1] = ACTIONS(1338),
    [aux_sym_if_statement_token2] = ACTIONS(1338),
    [aux_sym_else_if_clause_token1] = ACTIONS(1338),
    [aux_sym_else_clause_token1] = ACTIONS(1338),
    [aux_sym_match_expression_token1] = ACTIONS(1338),
    [aux_sym_match_default_expression_token1] = ACTIONS(1338),
    [aux_sym_switch_statement_token1] = ACTIONS(1338),
    [aux_sym_switch_block_token1] = ACTIONS(1338),
    [anon_sym_PLUS] = ACTIONS(1338),
    [anon_sym_DASH] = ACTIONS(1338),
    [anon_sym_TILDE] = ACTIONS(1336),
    [anon_sym_BANG] = ACTIONS(1336),
    [anon_sym_AT] = ACTIONS(1336),
    [aux_sym_clone_expression_token1] = ACTIONS(1338),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1338),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1338),
    [anon_sym_DASH_DASH] = ACTIONS(1336),
    [anon_sym_PLUS_PLUS] = ACTIONS(1336),
    [aux_sym__list_destructing_token1] = ACTIONS(1338),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_self] = ACTIONS(1338),
    [anon_sym_parent] = ACTIONS(1338),
    [aux_sym__argument_name_token1] = ACTIONS(1338),
    [aux_sym__argument_name_token2] = ACTIONS(1338),
    [anon_sym_POUND_LBRACK] = ACTIONS(1336),
    [aux_sym_encapsed_string_token1] = ACTIONS(1336),
    [anon_sym_DQUOTE] = ACTIONS(1336),
    [aux_sym_string_token1] = ACTIONS(1336),
    [anon_sym_SQUOTE] = ACTIONS(1336),
    [anon_sym_LT_LT_LT] = ACTIONS(1336),
    [anon_sym_BQUOTE] = ACTIONS(1336),
    [anon_sym_DOLLAR] = ACTIONS(1336),
    [aux_sym_yield_expression_token1] = ACTIONS(1338),
    [aux_sym_include_expression_token1] = ACTIONS(1338),
    [aux_sym_include_once_expression_token1] = ACTIONS(1338),
    [aux_sym_require_expression_token1] = ACTIONS(1338),
    [aux_sym_require_once_expression_token1] = ACTIONS(1338),
    [sym_comment] = ACTIONS(5),
  },
  [528] = {
    [sym_text_interpolation] = STATE(528),
    [ts_builtin_sym_end] = ACTIONS(1340),
    [sym_name] = ACTIONS(1342),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1340),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1342),
    [aux_sym_global_declaration_token1] = ACTIONS(1342),
    [aux_sym_namespace_definition_token1] = ACTIONS(1342),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1342),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1342),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1342),
    [anon_sym_BSLASH] = ACTIONS(1340),
    [anon_sym_LBRACE] = ACTIONS(1340),
    [anon_sym_RBRACE] = ACTIONS(1340),
    [aux_sym_trait_declaration_token1] = ACTIONS(1342),
    [aux_sym_interface_declaration_token1] = ACTIONS(1342),
    [aux_sym_enum_declaration_token1] = ACTIONS(1342),
    [aux_sym_enum_case_token1] = ACTIONS(1342),
    [aux_sym_class_declaration_token1] = ACTIONS(1342),
    [aux_sym_final_modifier_token1] = ACTIONS(1342),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1342),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1342),
    [sym_var_modifier] = ACTIONS(1342),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1342),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1342),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1342),
    [anon_sym_LPAREN2] = ACTIONS(1340),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1342),
    [aux_sym_cast_type_token1] = ACTIONS(1342),
    [aux_sym_echo_statement_token1] = ACTIONS(1342),
    [aux_sym_exit_statement_token1] = ACTIONS(1342),
    [anon_sym_unset] = ACTIONS(1342),
    [aux_sym_declare_statement_token1] = ACTIONS(1342),
    [aux_sym_declare_statement_token2] = ACTIONS(1342),
    [sym_float] = ACTIONS(1342),
    [aux_sym_try_statement_token1] = ACTIONS(1342),
    [aux_sym_goto_statement_token1] = ACTIONS(1342),
    [aux_sym_continue_statement_token1] = ACTIONS(1342),
    [aux_sym_break_statement_token1] = ACTIONS(1342),
    [sym_integer] = ACTIONS(1342),
    [aux_sym_return_statement_token1] = ACTIONS(1342),
    [aux_sym_throw_expression_token1] = ACTIONS(1342),
    [aux_sym_while_statement_token1] = ACTIONS(1342),
    [aux_sym_while_statement_token2] = ACTIONS(1342),
    [aux_sym_do_statement_token1] = ACTIONS(1342),
    [aux_sym_for_statement_token1] = ACTIONS(1342),
    [aux_sym_for_statement_token2] = ACTIONS(1342),
    [aux_sym_foreach_statement_token1] = ACTIONS(1342),
    [aux_sym_foreach_statement_token2] = ACTIONS(1342),
    [aux_sym_if_statement_token1] = ACTIONS(1342),
    [aux_sym_if_statement_token2] = ACTIONS(1342),
    [aux_sym_else_if_clause_token1] = ACTIONS(1342),
    [aux_sym_else_clause_token1] = ACTIONS(1342),
    [aux_sym_match_expression_token1] = ACTIONS(1342),
    [aux_sym_match_default_expression_token1] = ACTIONS(1342),
    [aux_sym_switch_statement_token1] = ACTIONS(1342),
    [aux_sym_switch_block_token1] = ACTIONS(1342),
    [anon_sym_PLUS] = ACTIONS(1342),
    [anon_sym_DASH] = ACTIONS(1342),
    [anon_sym_TILDE] = ACTIONS(1340),
    [anon_sym_BANG] = ACTIONS(1340),
    [anon_sym_AT] = ACTIONS(1340),
    [aux_sym_clone_expression_token1] = ACTIONS(1342),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1342),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1342),
    [anon_sym_DASH_DASH] = ACTIONS(1340),
    [anon_sym_PLUS_PLUS] = ACTIONS(1340),
    [aux_sym__list_destructing_token1] = ACTIONS(1342),
    [anon_sym_LBRACK] = ACTIONS(1340),
    [anon_sym_self] = ACTIONS(1342),
    [anon_sym_parent] = ACTIONS(1342),
    [aux_sym__argument_name_token1] = ACTIONS(1342),
    [aux_sym__argument_name_token2] = ACTIONS(1342),
    [anon_sym_POUND_LBRACK] = ACTIONS(1340),
    [aux_sym_encapsed_string_token1] = ACTIONS(1340),
    [anon_sym_DQUOTE] = ACTIONS(1340),
    [aux_sym_string_token1] = ACTIONS(1340),
    [anon_sym_SQUOTE] = ACTIONS(1340),
    [anon_sym_LT_LT_LT] = ACTIONS(1340),
    [anon_sym_BQUOTE] = ACTIONS(1340),
    [anon_sym_DOLLAR] = ACTIONS(1340),
    [aux_sym_yield_expression_token1] = ACTIONS(1342),
    [aux_sym_include_expression_token1] = ACTIONS(1342),
    [aux_sym_include_once_expression_token1] = ACTIONS(1342),
    [aux_sym_require_expression_token1] = ACTIONS(1342),
    [aux_sym_require_once_expression_token1] = ACTIONS(1342),
    [sym_comment] = ACTIONS(5),
  },
  [529] = {
    [sym_text_interpolation] = STATE(529),
    [ts_builtin_sym_end] = ACTIONS(1344),
    [sym_name] = ACTIONS(1346),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1344),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1346),
    [aux_sym_global_declaration_token1] = ACTIONS(1346),
    [aux_sym_namespace_definition_token1] = ACTIONS(1346),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1346),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1346),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1346),
    [anon_sym_BSLASH] = ACTIONS(1344),
    [anon_sym_LBRACE] = ACTIONS(1344),
    [anon_sym_RBRACE] = ACTIONS(1344),
    [aux_sym_trait_declaration_token1] = ACTIONS(1346),
    [aux_sym_interface_declaration_token1] = ACTIONS(1346),
    [aux_sym_enum_declaration_token1] = ACTIONS(1346),
    [aux_sym_enum_case_token1] = ACTIONS(1346),
    [aux_sym_class_declaration_token1] = ACTIONS(1346),
    [aux_sym_final_modifier_token1] = ACTIONS(1346),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1346),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1346),
    [sym_var_modifier] = ACTIONS(1346),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1346),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1346),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1346),
    [anon_sym_LPAREN2] = ACTIONS(1344),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1346),
    [aux_sym_cast_type_token1] = ACTIONS(1346),
    [aux_sym_echo_statement_token1] = ACTIONS(1346),
    [aux_sym_exit_statement_token1] = ACTIONS(1346),
    [anon_sym_unset] = ACTIONS(1346),
    [aux_sym_declare_statement_token1] = ACTIONS(1346),
    [aux_sym_declare_statement_token2] = ACTIONS(1346),
    [sym_float] = ACTIONS(1346),
    [aux_sym_try_statement_token1] = ACTIONS(1346),
    [aux_sym_goto_statement_token1] = ACTIONS(1346),
    [aux_sym_continue_statement_token1] = ACTIONS(1346),
    [aux_sym_break_statement_token1] = ACTIONS(1346),
    [sym_integer] = ACTIONS(1346),
    [aux_sym_return_statement_token1] = ACTIONS(1346),
    [aux_sym_throw_expression_token1] = ACTIONS(1346),
    [aux_sym_while_statement_token1] = ACTIONS(1346),
    [aux_sym_while_statement_token2] = ACTIONS(1346),
    [aux_sym_do_statement_token1] = ACTIONS(1346),
    [aux_sym_for_statement_token1] = ACTIONS(1346),
    [aux_sym_for_statement_token2] = ACTIONS(1346),
    [aux_sym_foreach_statement_token1] = ACTIONS(1346),
    [aux_sym_foreach_statement_token2] = ACTIONS(1346),
    [aux_sym_if_statement_token1] = ACTIONS(1346),
    [aux_sym_if_statement_token2] = ACTIONS(1346),
    [aux_sym_else_if_clause_token1] = ACTIONS(1346),
    [aux_sym_else_clause_token1] = ACTIONS(1346),
    [aux_sym_match_expression_token1] = ACTIONS(1346),
    [aux_sym_match_default_expression_token1] = ACTIONS(1346),
    [aux_sym_switch_statement_token1] = ACTIONS(1346),
    [aux_sym_switch_block_token1] = ACTIONS(1346),
    [anon_sym_PLUS] = ACTIONS(1346),
    [anon_sym_DASH] = ACTIONS(1346),
    [anon_sym_TILDE] = ACTIONS(1344),
    [anon_sym_BANG] = ACTIONS(1344),
    [anon_sym_AT] = ACTIONS(1344),
    [aux_sym_clone_expression_token1] = ACTIONS(1346),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1346),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1346),
    [anon_sym_DASH_DASH] = ACTIONS(1344),
    [anon_sym_PLUS_PLUS] = ACTIONS(1344),
    [aux_sym__list_destructing_token1] = ACTIONS(1346),
    [anon_sym_LBRACK] = ACTIONS(1344),
    [anon_sym_self] = ACTIONS(1346),
    [anon_sym_parent] = ACTIONS(1346),
    [aux_sym__argument_name_token1] = ACTIONS(1346),
    [aux_sym__argument_name_token2] = ACTIONS(1346),
    [anon_sym_POUND_LBRACK] = ACTIONS(1344),
    [aux_sym_encapsed_string_token1] = ACTIONS(1344),
    [anon_sym_DQUOTE] = ACTIONS(1344),
    [aux_sym_string_token1] = ACTIONS(1344),
    [anon_sym_SQUOTE] = ACTIONS(1344),
    [anon_sym_LT_LT_LT] = ACTIONS(1344),
    [anon_sym_BQUOTE] = ACTIONS(1344),
    [anon_sym_DOLLAR] = ACTIONS(1344),
    [aux_sym_yield_expression_token1] = ACTIONS(1346),
    [aux_sym_include_expression_token1] = ACTIONS(1346),
    [aux_sym_include_once_expression_token1] = ACTIONS(1346),
    [aux_sym_require_expression_token1] = ACTIONS(1346),
    [aux_sym_require_once_expression_token1] = ACTIONS(1346),
    [sym_comment] = ACTIONS(5),
  },
  [530] = {
    [sym_text_interpolation] = STATE(530),
    [ts_builtin_sym_end] = ACTIONS(1348),
    [sym_name] = ACTIONS(1350),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1348),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1350),
    [aux_sym_global_declaration_token1] = ACTIONS(1350),
    [aux_sym_namespace_definition_token1] = ACTIONS(1350),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1350),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1350),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1350),
    [anon_sym_BSLASH] = ACTIONS(1348),
    [anon_sym_LBRACE] = ACTIONS(1348),
    [anon_sym_RBRACE] = ACTIONS(1348),
    [aux_sym_trait_declaration_token1] = ACTIONS(1350),
    [aux_sym_interface_declaration_token1] = ACTIONS(1350),
    [aux_sym_enum_declaration_token1] = ACTIONS(1350),
    [aux_sym_enum_case_token1] = ACTIONS(1350),
    [aux_sym_class_declaration_token1] = ACTIONS(1350),
    [aux_sym_final_modifier_token1] = ACTIONS(1350),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1350),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1350),
    [sym_var_modifier] = ACTIONS(1350),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1350),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1350),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1350),
    [anon_sym_LPAREN2] = ACTIONS(1348),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1350),
    [aux_sym_cast_type_token1] = ACTIONS(1350),
    [aux_sym_echo_statement_token1] = ACTIONS(1350),
    [aux_sym_exit_statement_token1] = ACTIONS(1350),
    [anon_sym_unset] = ACTIONS(1350),
    [aux_sym_declare_statement_token1] = ACTIONS(1350),
    [aux_sym_declare_statement_token2] = ACTIONS(1350),
    [sym_float] = ACTIONS(1350),
    [aux_sym_try_statement_token1] = ACTIONS(1350),
    [aux_sym_goto_statement_token1] = ACTIONS(1350),
    [aux_sym_continue_statement_token1] = ACTIONS(1350),
    [aux_sym_break_statement_token1] = ACTIONS(1350),
    [sym_integer] = ACTIONS(1350),
    [aux_sym_return_statement_token1] = ACTIONS(1350),
    [aux_sym_throw_expression_token1] = ACTIONS(1350),
    [aux_sym_while_statement_token1] = ACTIONS(1350),
    [aux_sym_while_statement_token2] = ACTIONS(1350),
    [aux_sym_do_statement_token1] = ACTIONS(1350),
    [aux_sym_for_statement_token1] = ACTIONS(1350),
    [aux_sym_for_statement_token2] = ACTIONS(1350),
    [aux_sym_foreach_statement_token1] = ACTIONS(1350),
    [aux_sym_foreach_statement_token2] = ACTIONS(1350),
    [aux_sym_if_statement_token1] = ACTIONS(1350),
    [aux_sym_if_statement_token2] = ACTIONS(1350),
    [aux_sym_else_if_clause_token1] = ACTIONS(1350),
    [aux_sym_else_clause_token1] = ACTIONS(1350),
    [aux_sym_match_expression_token1] = ACTIONS(1350),
    [aux_sym_match_default_expression_token1] = ACTIONS(1350),
    [aux_sym_switch_statement_token1] = ACTIONS(1350),
    [aux_sym_switch_block_token1] = ACTIONS(1350),
    [anon_sym_PLUS] = ACTIONS(1350),
    [anon_sym_DASH] = ACTIONS(1350),
    [anon_sym_TILDE] = ACTIONS(1348),
    [anon_sym_BANG] = ACTIONS(1348),
    [anon_sym_AT] = ACTIONS(1348),
    [aux_sym_clone_expression_token1] = ACTIONS(1350),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1350),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1350),
    [anon_sym_DASH_DASH] = ACTIONS(1348),
    [anon_sym_PLUS_PLUS] = ACTIONS(1348),
    [aux_sym__list_destructing_token1] = ACTIONS(1350),
    [anon_sym_LBRACK] = ACTIONS(1348),
    [anon_sym_self] = ACTIONS(1350),
    [anon_sym_parent] = ACTIONS(1350),
    [aux_sym__argument_name_token1] = ACTIONS(1350),
    [aux_sym__argument_name_token2] = ACTIONS(1350),
    [anon_sym_POUND_LBRACK] = ACTIONS(1348),
    [aux_sym_encapsed_string_token1] = ACTIONS(1348),
    [anon_sym_DQUOTE] = ACTIONS(1348),
    [aux_sym_string_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1348),
    [anon_sym_LT_LT_LT] = ACTIONS(1348),
    [anon_sym_BQUOTE] = ACTIONS(1348),
    [anon_sym_DOLLAR] = ACTIONS(1348),
    [aux_sym_yield_expression_token1] = ACTIONS(1350),
    [aux_sym_include_expression_token1] = ACTIONS(1350),
    [aux_sym_include_once_expression_token1] = ACTIONS(1350),
    [aux_sym_require_expression_token1] = ACTIONS(1350),
    [aux_sym_require_once_expression_token1] = ACTIONS(1350),
    [sym_comment] = ACTIONS(5),
  },
  [531] = {
    [sym_text_interpolation] = STATE(531),
    [ts_builtin_sym_end] = ACTIONS(1352),
    [sym_name] = ACTIONS(1354),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1352),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1354),
    [aux_sym_global_declaration_token1] = ACTIONS(1354),
    [aux_sym_namespace_definition_token1] = ACTIONS(1354),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1354),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1354),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1354),
    [anon_sym_BSLASH] = ACTIONS(1352),
    [anon_sym_LBRACE] = ACTIONS(1352),
    [anon_sym_RBRACE] = ACTIONS(1352),
    [aux_sym_trait_declaration_token1] = ACTIONS(1354),
    [aux_sym_interface_declaration_token1] = ACTIONS(1354),
    [aux_sym_enum_declaration_token1] = ACTIONS(1354),
    [aux_sym_enum_case_token1] = ACTIONS(1354),
    [aux_sym_class_declaration_token1] = ACTIONS(1354),
    [aux_sym_final_modifier_token1] = ACTIONS(1354),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1354),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1354),
    [sym_var_modifier] = ACTIONS(1354),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1354),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1354),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1354),
    [anon_sym_LPAREN2] = ACTIONS(1352),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1354),
    [aux_sym_cast_type_token1] = ACTIONS(1354),
    [aux_sym_echo_statement_token1] = ACTIONS(1354),
    [aux_sym_exit_statement_token1] = ACTIONS(1354),
    [anon_sym_unset] = ACTIONS(1354),
    [aux_sym_declare_statement_token1] = ACTIONS(1354),
    [aux_sym_declare_statement_token2] = ACTIONS(1354),
    [sym_float] = ACTIONS(1354),
    [aux_sym_try_statement_token1] = ACTIONS(1354),
    [aux_sym_goto_statement_token1] = ACTIONS(1354),
    [aux_sym_continue_statement_token1] = ACTIONS(1354),
    [aux_sym_break_statement_token1] = ACTIONS(1354),
    [sym_integer] = ACTIONS(1354),
    [aux_sym_return_statement_token1] = ACTIONS(1354),
    [aux_sym_throw_expression_token1] = ACTIONS(1354),
    [aux_sym_while_statement_token1] = ACTIONS(1354),
    [aux_sym_while_statement_token2] = ACTIONS(1354),
    [aux_sym_do_statement_token1] = ACTIONS(1354),
    [aux_sym_for_statement_token1] = ACTIONS(1354),
    [aux_sym_for_statement_token2] = ACTIONS(1354),
    [aux_sym_foreach_statement_token1] = ACTIONS(1354),
    [aux_sym_foreach_statement_token2] = ACTIONS(1354),
    [aux_sym_if_statement_token1] = ACTIONS(1354),
    [aux_sym_if_statement_token2] = ACTIONS(1354),
    [aux_sym_else_if_clause_token1] = ACTIONS(1354),
    [aux_sym_else_clause_token1] = ACTIONS(1354),
    [aux_sym_match_expression_token1] = ACTIONS(1354),
    [aux_sym_match_default_expression_token1] = ACTIONS(1354),
    [aux_sym_switch_statement_token1] = ACTIONS(1354),
    [aux_sym_switch_block_token1] = ACTIONS(1354),
    [anon_sym_PLUS] = ACTIONS(1354),
    [anon_sym_DASH] = ACTIONS(1354),
    [anon_sym_TILDE] = ACTIONS(1352),
    [anon_sym_BANG] = ACTIONS(1352),
    [anon_sym_AT] = ACTIONS(1352),
    [aux_sym_clone_expression_token1] = ACTIONS(1354),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1354),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1354),
    [anon_sym_DASH_DASH] = ACTIONS(1352),
    [anon_sym_PLUS_PLUS] = ACTIONS(1352),
    [aux_sym__list_destructing_token1] = ACTIONS(1354),
    [anon_sym_LBRACK] = ACTIONS(1352),
    [anon_sym_self] = ACTIONS(1354),
    [anon_sym_parent] = ACTIONS(1354),
    [aux_sym__argument_name_token1] = ACTIONS(1354),
    [aux_sym__argument_name_token2] = ACTIONS(1354),
    [anon_sym_POUND_LBRACK] = ACTIONS(1352),
    [aux_sym_encapsed_string_token1] = ACTIONS(1352),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [aux_sym_string_token1] = ACTIONS(1352),
    [anon_sym_SQUOTE] = ACTIONS(1352),
    [anon_sym_LT_LT_LT] = ACTIONS(1352),
    [anon_sym_BQUOTE] = ACTIONS(1352),
    [anon_sym_DOLLAR] = ACTIONS(1352),
    [aux_sym_yield_expression_token1] = ACTIONS(1354),
    [aux_sym_include_expression_token1] = ACTIONS(1354),
    [aux_sym_include_once_expression_token1] = ACTIONS(1354),
    [aux_sym_require_expression_token1] = ACTIONS(1354),
    [aux_sym_require_once_expression_token1] = ACTIONS(1354),
    [sym_comment] = ACTIONS(5),
  },
  [532] = {
    [sym_text_interpolation] = STATE(532),
    [ts_builtin_sym_end] = ACTIONS(1356),
    [sym_name] = ACTIONS(1358),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1358),
    [aux_sym_global_declaration_token1] = ACTIONS(1358),
    [aux_sym_namespace_definition_token1] = ACTIONS(1358),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1358),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1358),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1358),
    [anon_sym_BSLASH] = ACTIONS(1356),
    [anon_sym_LBRACE] = ACTIONS(1356),
    [anon_sym_RBRACE] = ACTIONS(1356),
    [aux_sym_trait_declaration_token1] = ACTIONS(1358),
    [aux_sym_interface_declaration_token1] = ACTIONS(1358),
    [aux_sym_enum_declaration_token1] = ACTIONS(1358),
    [aux_sym_enum_case_token1] = ACTIONS(1358),
    [aux_sym_class_declaration_token1] = ACTIONS(1358),
    [aux_sym_final_modifier_token1] = ACTIONS(1358),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1358),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1358),
    [sym_var_modifier] = ACTIONS(1358),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1358),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1358),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1358),
    [anon_sym_LPAREN2] = ACTIONS(1356),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1358),
    [aux_sym_cast_type_token1] = ACTIONS(1358),
    [aux_sym_echo_statement_token1] = ACTIONS(1358),
    [aux_sym_exit_statement_token1] = ACTIONS(1358),
    [anon_sym_unset] = ACTIONS(1358),
    [aux_sym_declare_statement_token1] = ACTIONS(1358),
    [aux_sym_declare_statement_token2] = ACTIONS(1358),
    [sym_float] = ACTIONS(1358),
    [aux_sym_try_statement_token1] = ACTIONS(1358),
    [aux_sym_goto_statement_token1] = ACTIONS(1358),
    [aux_sym_continue_statement_token1] = ACTIONS(1358),
    [aux_sym_break_statement_token1] = ACTIONS(1358),
    [sym_integer] = ACTIONS(1358),
    [aux_sym_return_statement_token1] = ACTIONS(1358),
    [aux_sym_throw_expression_token1] = ACTIONS(1358),
    [aux_sym_while_statement_token1] = ACTIONS(1358),
    [aux_sym_while_statement_token2] = ACTIONS(1358),
    [aux_sym_do_statement_token1] = ACTIONS(1358),
    [aux_sym_for_statement_token1] = ACTIONS(1358),
    [aux_sym_for_statement_token2] = ACTIONS(1358),
    [aux_sym_foreach_statement_token1] = ACTIONS(1358),
    [aux_sym_foreach_statement_token2] = ACTIONS(1358),
    [aux_sym_if_statement_token1] = ACTIONS(1358),
    [aux_sym_if_statement_token2] = ACTIONS(1358),
    [aux_sym_else_if_clause_token1] = ACTIONS(1358),
    [aux_sym_else_clause_token1] = ACTIONS(1358),
    [aux_sym_match_expression_token1] = ACTIONS(1358),
    [aux_sym_match_default_expression_token1] = ACTIONS(1358),
    [aux_sym_switch_statement_token1] = ACTIONS(1358),
    [aux_sym_switch_block_token1] = ACTIONS(1358),
    [anon_sym_PLUS] = ACTIONS(1358),
    [anon_sym_DASH] = ACTIONS(1358),
    [anon_sym_TILDE] = ACTIONS(1356),
    [anon_sym_BANG] = ACTIONS(1356),
    [anon_sym_AT] = ACTIONS(1356),
    [aux_sym_clone_expression_token1] = ACTIONS(1358),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1358),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1358),
    [anon_sym_DASH_DASH] = ACTIONS(1356),
    [anon_sym_PLUS_PLUS] = ACTIONS(1356),
    [aux_sym__list_destructing_token1] = ACTIONS(1358),
    [anon_sym_LBRACK] = ACTIONS(1356),
    [anon_sym_self] = ACTIONS(1358),
    [anon_sym_parent] = ACTIONS(1358),
    [aux_sym__argument_name_token1] = ACTIONS(1358),
    [aux_sym__argument_name_token2] = ACTIONS(1358),
    [anon_sym_POUND_LBRACK] = ACTIONS(1356),
    [aux_sym_encapsed_string_token1] = ACTIONS(1356),
    [anon_sym_DQUOTE] = ACTIONS(1356),
    [aux_sym_string_token1] = ACTIONS(1356),
    [anon_sym_SQUOTE] = ACTIONS(1356),
    [anon_sym_LT_LT_LT] = ACTIONS(1356),
    [anon_sym_BQUOTE] = ACTIONS(1356),
    [anon_sym_DOLLAR] = ACTIONS(1356),
    [aux_sym_yield_expression_token1] = ACTIONS(1358),
    [aux_sym_include_expression_token1] = ACTIONS(1358),
    [aux_sym_include_once_expression_token1] = ACTIONS(1358),
    [aux_sym_require_expression_token1] = ACTIONS(1358),
    [aux_sym_require_once_expression_token1] = ACTIONS(1358),
    [sym_comment] = ACTIONS(5),
  },
  [533] = {
    [sym_text_interpolation] = STATE(533),
    [ts_builtin_sym_end] = ACTIONS(1360),
    [sym_name] = ACTIONS(1362),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1360),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1362),
    [aux_sym_global_declaration_token1] = ACTIONS(1362),
    [aux_sym_namespace_definition_token1] = ACTIONS(1362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1362),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1362),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1362),
    [anon_sym_BSLASH] = ACTIONS(1360),
    [anon_sym_LBRACE] = ACTIONS(1360),
    [anon_sym_RBRACE] = ACTIONS(1360),
    [aux_sym_trait_declaration_token1] = ACTIONS(1362),
    [aux_sym_interface_declaration_token1] = ACTIONS(1362),
    [aux_sym_enum_declaration_token1] = ACTIONS(1362),
    [aux_sym_enum_case_token1] = ACTIONS(1362),
    [aux_sym_class_declaration_token1] = ACTIONS(1362),
    [aux_sym_final_modifier_token1] = ACTIONS(1362),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1362),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1362),
    [sym_var_modifier] = ACTIONS(1362),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1362),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1362),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1362),
    [anon_sym_LPAREN2] = ACTIONS(1360),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1362),
    [aux_sym_cast_type_token1] = ACTIONS(1362),
    [aux_sym_echo_statement_token1] = ACTIONS(1362),
    [aux_sym_exit_statement_token1] = ACTIONS(1362),
    [anon_sym_unset] = ACTIONS(1362),
    [aux_sym_declare_statement_token1] = ACTIONS(1362),
    [aux_sym_declare_statement_token2] = ACTIONS(1362),
    [sym_float] = ACTIONS(1362),
    [aux_sym_try_statement_token1] = ACTIONS(1362),
    [aux_sym_goto_statement_token1] = ACTIONS(1362),
    [aux_sym_continue_statement_token1] = ACTIONS(1362),
    [aux_sym_break_statement_token1] = ACTIONS(1362),
    [sym_integer] = ACTIONS(1362),
    [aux_sym_return_statement_token1] = ACTIONS(1362),
    [aux_sym_throw_expression_token1] = ACTIONS(1362),
    [aux_sym_while_statement_token1] = ACTIONS(1362),
    [aux_sym_while_statement_token2] = ACTIONS(1362),
    [aux_sym_do_statement_token1] = ACTIONS(1362),
    [aux_sym_for_statement_token1] = ACTIONS(1362),
    [aux_sym_for_statement_token2] = ACTIONS(1362),
    [aux_sym_foreach_statement_token1] = ACTIONS(1362),
    [aux_sym_foreach_statement_token2] = ACTIONS(1362),
    [aux_sym_if_statement_token1] = ACTIONS(1362),
    [aux_sym_if_statement_token2] = ACTIONS(1362),
    [aux_sym_else_if_clause_token1] = ACTIONS(1362),
    [aux_sym_else_clause_token1] = ACTIONS(1362),
    [aux_sym_match_expression_token1] = ACTIONS(1362),
    [aux_sym_match_default_expression_token1] = ACTIONS(1362),
    [aux_sym_switch_statement_token1] = ACTIONS(1362),
    [aux_sym_switch_block_token1] = ACTIONS(1362),
    [anon_sym_PLUS] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1360),
    [anon_sym_BANG] = ACTIONS(1360),
    [anon_sym_AT] = ACTIONS(1360),
    [aux_sym_clone_expression_token1] = ACTIONS(1362),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1362),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1362),
    [anon_sym_DASH_DASH] = ACTIONS(1360),
    [anon_sym_PLUS_PLUS] = ACTIONS(1360),
    [aux_sym__list_destructing_token1] = ACTIONS(1362),
    [anon_sym_LBRACK] = ACTIONS(1360),
    [anon_sym_self] = ACTIONS(1362),
    [anon_sym_parent] = ACTIONS(1362),
    [aux_sym__argument_name_token1] = ACTIONS(1362),
    [aux_sym__argument_name_token2] = ACTIONS(1362),
    [anon_sym_POUND_LBRACK] = ACTIONS(1360),
    [aux_sym_encapsed_string_token1] = ACTIONS(1360),
    [anon_sym_DQUOTE] = ACTIONS(1360),
    [aux_sym_string_token1] = ACTIONS(1360),
    [anon_sym_SQUOTE] = ACTIONS(1360),
    [anon_sym_LT_LT_LT] = ACTIONS(1360),
    [anon_sym_BQUOTE] = ACTIONS(1360),
    [anon_sym_DOLLAR] = ACTIONS(1360),
    [aux_sym_yield_expression_token1] = ACTIONS(1362),
    [aux_sym_include_expression_token1] = ACTIONS(1362),
    [aux_sym_include_once_expression_token1] = ACTIONS(1362),
    [aux_sym_require_expression_token1] = ACTIONS(1362),
    [aux_sym_require_once_expression_token1] = ACTIONS(1362),
    [sym_comment] = ACTIONS(5),
  },
  [534] = {
    [sym_text_interpolation] = STATE(534),
    [ts_builtin_sym_end] = ACTIONS(1364),
    [sym_name] = ACTIONS(1366),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1364),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1366),
    [aux_sym_global_declaration_token1] = ACTIONS(1366),
    [aux_sym_namespace_definition_token1] = ACTIONS(1366),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1366),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1366),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1366),
    [anon_sym_BSLASH] = ACTIONS(1364),
    [anon_sym_LBRACE] = ACTIONS(1364),
    [anon_sym_RBRACE] = ACTIONS(1364),
    [aux_sym_trait_declaration_token1] = ACTIONS(1366),
    [aux_sym_interface_declaration_token1] = ACTIONS(1366),
    [aux_sym_enum_declaration_token1] = ACTIONS(1366),
    [aux_sym_enum_case_token1] = ACTIONS(1366),
    [aux_sym_class_declaration_token1] = ACTIONS(1366),
    [aux_sym_final_modifier_token1] = ACTIONS(1366),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1366),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1366),
    [sym_var_modifier] = ACTIONS(1366),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1366),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1366),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1366),
    [anon_sym_LPAREN2] = ACTIONS(1364),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1366),
    [aux_sym_cast_type_token1] = ACTIONS(1366),
    [aux_sym_echo_statement_token1] = ACTIONS(1366),
    [aux_sym_exit_statement_token1] = ACTIONS(1366),
    [anon_sym_unset] = ACTIONS(1366),
    [aux_sym_declare_statement_token1] = ACTIONS(1366),
    [aux_sym_declare_statement_token2] = ACTIONS(1366),
    [sym_float] = ACTIONS(1366),
    [aux_sym_try_statement_token1] = ACTIONS(1366),
    [aux_sym_goto_statement_token1] = ACTIONS(1366),
    [aux_sym_continue_statement_token1] = ACTIONS(1366),
    [aux_sym_break_statement_token1] = ACTIONS(1366),
    [sym_integer] = ACTIONS(1366),
    [aux_sym_return_statement_token1] = ACTIONS(1366),
    [aux_sym_throw_expression_token1] = ACTIONS(1366),
    [aux_sym_while_statement_token1] = ACTIONS(1366),
    [aux_sym_while_statement_token2] = ACTIONS(1366),
    [aux_sym_do_statement_token1] = ACTIONS(1366),
    [aux_sym_for_statement_token1] = ACTIONS(1366),
    [aux_sym_for_statement_token2] = ACTIONS(1366),
    [aux_sym_foreach_statement_token1] = ACTIONS(1366),
    [aux_sym_foreach_statement_token2] = ACTIONS(1366),
    [aux_sym_if_statement_token1] = ACTIONS(1366),
    [aux_sym_if_statement_token2] = ACTIONS(1366),
    [aux_sym_else_if_clause_token1] = ACTIONS(1366),
    [aux_sym_else_clause_token1] = ACTIONS(1366),
    [aux_sym_match_expression_token1] = ACTIONS(1366),
    [aux_sym_match_default_expression_token1] = ACTIONS(1366),
    [aux_sym_switch_statement_token1] = ACTIONS(1366),
    [aux_sym_switch_block_token1] = ACTIONS(1366),
    [anon_sym_PLUS] = ACTIONS(1366),
    [anon_sym_DASH] = ACTIONS(1366),
    [anon_sym_TILDE] = ACTIONS(1364),
    [anon_sym_BANG] = ACTIONS(1364),
    [anon_sym_AT] = ACTIONS(1364),
    [aux_sym_clone_expression_token1] = ACTIONS(1366),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1366),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1366),
    [anon_sym_DASH_DASH] = ACTIONS(1364),
    [anon_sym_PLUS_PLUS] = ACTIONS(1364),
    [aux_sym__list_destructing_token1] = ACTIONS(1366),
    [anon_sym_LBRACK] = ACTIONS(1364),
    [anon_sym_self] = ACTIONS(1366),
    [anon_sym_parent] = ACTIONS(1366),
    [aux_sym__argument_name_token1] = ACTIONS(1366),
    [aux_sym__argument_name_token2] = ACTIONS(1366),
    [anon_sym_POUND_LBRACK] = ACTIONS(1364),
    [aux_sym_encapsed_string_token1] = ACTIONS(1364),
    [anon_sym_DQUOTE] = ACTIONS(1364),
    [aux_sym_string_token1] = ACTIONS(1364),
    [anon_sym_SQUOTE] = ACTIONS(1364),
    [anon_sym_LT_LT_LT] = ACTIONS(1364),
    [anon_sym_BQUOTE] = ACTIONS(1364),
    [anon_sym_DOLLAR] = ACTIONS(1364),
    [aux_sym_yield_expression_token1] = ACTIONS(1366),
    [aux_sym_include_expression_token1] = ACTIONS(1366),
    [aux_sym_include_once_expression_token1] = ACTIONS(1366),
    [aux_sym_require_expression_token1] = ACTIONS(1366),
    [aux_sym_require_once_expression_token1] = ACTIONS(1366),
    [sym_comment] = ACTIONS(5),
  },
  [535] = {
    [sym_text_interpolation] = STATE(535),
    [ts_builtin_sym_end] = ACTIONS(1040),
    [sym_name] = ACTIONS(1042),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1040),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1042),
    [aux_sym_global_declaration_token1] = ACTIONS(1042),
    [aux_sym_namespace_definition_token1] = ACTIONS(1042),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1042),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1042),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1042),
    [anon_sym_BSLASH] = ACTIONS(1040),
    [anon_sym_LBRACE] = ACTIONS(1040),
    [anon_sym_RBRACE] = ACTIONS(1040),
    [aux_sym_trait_declaration_token1] = ACTIONS(1042),
    [aux_sym_interface_declaration_token1] = ACTIONS(1042),
    [aux_sym_enum_declaration_token1] = ACTIONS(1042),
    [aux_sym_enum_case_token1] = ACTIONS(1042),
    [aux_sym_class_declaration_token1] = ACTIONS(1042),
    [aux_sym_final_modifier_token1] = ACTIONS(1042),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1042),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1042),
    [sym_var_modifier] = ACTIONS(1042),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1042),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1042),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1042),
    [anon_sym_LPAREN2] = ACTIONS(1040),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1042),
    [aux_sym_cast_type_token1] = ACTIONS(1042),
    [aux_sym_echo_statement_token1] = ACTIONS(1042),
    [aux_sym_exit_statement_token1] = ACTIONS(1042),
    [anon_sym_unset] = ACTIONS(1042),
    [aux_sym_declare_statement_token1] = ACTIONS(1042),
    [aux_sym_declare_statement_token2] = ACTIONS(1042),
    [sym_float] = ACTIONS(1042),
    [aux_sym_try_statement_token1] = ACTIONS(1042),
    [aux_sym_goto_statement_token1] = ACTIONS(1042),
    [aux_sym_continue_statement_token1] = ACTIONS(1042),
    [aux_sym_break_statement_token1] = ACTIONS(1042),
    [sym_integer] = ACTIONS(1042),
    [aux_sym_return_statement_token1] = ACTIONS(1042),
    [aux_sym_throw_expression_token1] = ACTIONS(1042),
    [aux_sym_while_statement_token1] = ACTIONS(1042),
    [aux_sym_while_statement_token2] = ACTIONS(1042),
    [aux_sym_do_statement_token1] = ACTIONS(1042),
    [aux_sym_for_statement_token1] = ACTIONS(1042),
    [aux_sym_for_statement_token2] = ACTIONS(1042),
    [aux_sym_foreach_statement_token1] = ACTIONS(1042),
    [aux_sym_foreach_statement_token2] = ACTIONS(1042),
    [aux_sym_if_statement_token1] = ACTIONS(1042),
    [aux_sym_if_statement_token2] = ACTIONS(1042),
    [aux_sym_else_if_clause_token1] = ACTIONS(1042),
    [aux_sym_else_clause_token1] = ACTIONS(1042),
    [aux_sym_match_expression_token1] = ACTIONS(1042),
    [aux_sym_match_default_expression_token1] = ACTIONS(1042),
    [aux_sym_switch_statement_token1] = ACTIONS(1042),
    [aux_sym_switch_block_token1] = ACTIONS(1042),
    [anon_sym_PLUS] = ACTIONS(1042),
    [anon_sym_DASH] = ACTIONS(1042),
    [anon_sym_TILDE] = ACTIONS(1040),
    [anon_sym_BANG] = ACTIONS(1040),
    [anon_sym_AT] = ACTIONS(1040),
    [aux_sym_clone_expression_token1] = ACTIONS(1042),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1042),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1042),
    [anon_sym_DASH_DASH] = ACTIONS(1040),
    [anon_sym_PLUS_PLUS] = ACTIONS(1040),
    [aux_sym__list_destructing_token1] = ACTIONS(1042),
    [anon_sym_LBRACK] = ACTIONS(1040),
    [anon_sym_self] = ACTIONS(1042),
    [anon_sym_parent] = ACTIONS(1042),
    [aux_sym__argument_name_token1] = ACTIONS(1042),
    [aux_sym__argument_name_token2] = ACTIONS(1042),
    [anon_sym_POUND_LBRACK] = ACTIONS(1040),
    [aux_sym_encapsed_string_token1] = ACTIONS(1040),
    [anon_sym_DQUOTE] = ACTIONS(1040),
    [aux_sym_string_token1] = ACTIONS(1040),
    [anon_sym_SQUOTE] = ACTIONS(1040),
    [anon_sym_LT_LT_LT] = ACTIONS(1040),
    [anon_sym_BQUOTE] = ACTIONS(1040),
    [anon_sym_DOLLAR] = ACTIONS(1040),
    [aux_sym_yield_expression_token1] = ACTIONS(1042),
    [aux_sym_include_expression_token1] = ACTIONS(1042),
    [aux_sym_include_once_expression_token1] = ACTIONS(1042),
    [aux_sym_require_expression_token1] = ACTIONS(1042),
    [aux_sym_require_once_expression_token1] = ACTIONS(1042),
    [sym_comment] = ACTIONS(5),
  },
  [536] = {
    [sym_text_interpolation] = STATE(536),
    [ts_builtin_sym_end] = ACTIONS(1368),
    [sym_name] = ACTIONS(1370),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1368),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1370),
    [aux_sym_global_declaration_token1] = ACTIONS(1370),
    [aux_sym_namespace_definition_token1] = ACTIONS(1370),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1370),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1370),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1370),
    [anon_sym_BSLASH] = ACTIONS(1368),
    [anon_sym_LBRACE] = ACTIONS(1368),
    [anon_sym_RBRACE] = ACTIONS(1368),
    [aux_sym_trait_declaration_token1] = ACTIONS(1370),
    [aux_sym_interface_declaration_token1] = ACTIONS(1370),
    [aux_sym_enum_declaration_token1] = ACTIONS(1370),
    [aux_sym_enum_case_token1] = ACTIONS(1370),
    [aux_sym_class_declaration_token1] = ACTIONS(1370),
    [aux_sym_final_modifier_token1] = ACTIONS(1370),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1370),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1370),
    [sym_var_modifier] = ACTIONS(1370),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1370),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1370),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1370),
    [anon_sym_LPAREN2] = ACTIONS(1368),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1370),
    [aux_sym_cast_type_token1] = ACTIONS(1370),
    [aux_sym_echo_statement_token1] = ACTIONS(1370),
    [aux_sym_exit_statement_token1] = ACTIONS(1370),
    [anon_sym_unset] = ACTIONS(1370),
    [aux_sym_declare_statement_token1] = ACTIONS(1370),
    [aux_sym_declare_statement_token2] = ACTIONS(1370),
    [sym_float] = ACTIONS(1370),
    [aux_sym_try_statement_token1] = ACTIONS(1370),
    [aux_sym_goto_statement_token1] = ACTIONS(1370),
    [aux_sym_continue_statement_token1] = ACTIONS(1370),
    [aux_sym_break_statement_token1] = ACTIONS(1370),
    [sym_integer] = ACTIONS(1370),
    [aux_sym_return_statement_token1] = ACTIONS(1370),
    [aux_sym_throw_expression_token1] = ACTIONS(1370),
    [aux_sym_while_statement_token1] = ACTIONS(1370),
    [aux_sym_while_statement_token2] = ACTIONS(1370),
    [aux_sym_do_statement_token1] = ACTIONS(1370),
    [aux_sym_for_statement_token1] = ACTIONS(1370),
    [aux_sym_for_statement_token2] = ACTIONS(1370),
    [aux_sym_foreach_statement_token1] = ACTIONS(1370),
    [aux_sym_foreach_statement_token2] = ACTIONS(1370),
    [aux_sym_if_statement_token1] = ACTIONS(1370),
    [aux_sym_if_statement_token2] = ACTIONS(1370),
    [aux_sym_else_if_clause_token1] = ACTIONS(1370),
    [aux_sym_else_clause_token1] = ACTIONS(1370),
    [aux_sym_match_expression_token1] = ACTIONS(1370),
    [aux_sym_match_default_expression_token1] = ACTIONS(1370),
    [aux_sym_switch_statement_token1] = ACTIONS(1370),
    [aux_sym_switch_block_token1] = ACTIONS(1370),
    [anon_sym_PLUS] = ACTIONS(1370),
    [anon_sym_DASH] = ACTIONS(1370),
    [anon_sym_TILDE] = ACTIONS(1368),
    [anon_sym_BANG] = ACTIONS(1368),
    [anon_sym_AT] = ACTIONS(1368),
    [aux_sym_clone_expression_token1] = ACTIONS(1370),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1370),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1370),
    [anon_sym_DASH_DASH] = ACTIONS(1368),
    [anon_sym_PLUS_PLUS] = ACTIONS(1368),
    [aux_sym__list_destructing_token1] = ACTIONS(1370),
    [anon_sym_LBRACK] = ACTIONS(1368),
    [anon_sym_self] = ACTIONS(1370),
    [anon_sym_parent] = ACTIONS(1370),
    [aux_sym__argument_name_token1] = ACTIONS(1370),
    [aux_sym__argument_name_token2] = ACTIONS(1370),
    [anon_sym_POUND_LBRACK] = ACTIONS(1368),
    [aux_sym_encapsed_string_token1] = ACTIONS(1368),
    [anon_sym_DQUOTE] = ACTIONS(1368),
    [aux_sym_string_token1] = ACTIONS(1368),
    [anon_sym_SQUOTE] = ACTIONS(1368),
    [anon_sym_LT_LT_LT] = ACTIONS(1368),
    [anon_sym_BQUOTE] = ACTIONS(1368),
    [anon_sym_DOLLAR] = ACTIONS(1368),
    [aux_sym_yield_expression_token1] = ACTIONS(1370),
    [aux_sym_include_expression_token1] = ACTIONS(1370),
    [aux_sym_include_once_expression_token1] = ACTIONS(1370),
    [aux_sym_require_expression_token1] = ACTIONS(1370),
    [aux_sym_require_once_expression_token1] = ACTIONS(1370),
    [sym_comment] = ACTIONS(5),
  },
  [537] = {
    [sym_text_interpolation] = STATE(537),
    [ts_builtin_sym_end] = ACTIONS(1372),
    [sym_name] = ACTIONS(1374),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1372),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1374),
    [aux_sym_global_declaration_token1] = ACTIONS(1374),
    [aux_sym_namespace_definition_token1] = ACTIONS(1374),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1374),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1374),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1374),
    [anon_sym_BSLASH] = ACTIONS(1372),
    [anon_sym_LBRACE] = ACTIONS(1372),
    [anon_sym_RBRACE] = ACTIONS(1372),
    [aux_sym_trait_declaration_token1] = ACTIONS(1374),
    [aux_sym_interface_declaration_token1] = ACTIONS(1374),
    [aux_sym_enum_declaration_token1] = ACTIONS(1374),
    [aux_sym_enum_case_token1] = ACTIONS(1374),
    [aux_sym_class_declaration_token1] = ACTIONS(1374),
    [aux_sym_final_modifier_token1] = ACTIONS(1374),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1374),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1374),
    [sym_var_modifier] = ACTIONS(1374),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1374),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1374),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1374),
    [anon_sym_LPAREN2] = ACTIONS(1372),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1374),
    [aux_sym_cast_type_token1] = ACTIONS(1374),
    [aux_sym_echo_statement_token1] = ACTIONS(1374),
    [aux_sym_exit_statement_token1] = ACTIONS(1374),
    [anon_sym_unset] = ACTIONS(1374),
    [aux_sym_declare_statement_token1] = ACTIONS(1374),
    [aux_sym_declare_statement_token2] = ACTIONS(1374),
    [sym_float] = ACTIONS(1374),
    [aux_sym_try_statement_token1] = ACTIONS(1374),
    [aux_sym_goto_statement_token1] = ACTIONS(1374),
    [aux_sym_continue_statement_token1] = ACTIONS(1374),
    [aux_sym_break_statement_token1] = ACTIONS(1374),
    [sym_integer] = ACTIONS(1374),
    [aux_sym_return_statement_token1] = ACTIONS(1374),
    [aux_sym_throw_expression_token1] = ACTIONS(1374),
    [aux_sym_while_statement_token1] = ACTIONS(1374),
    [aux_sym_while_statement_token2] = ACTIONS(1374),
    [aux_sym_do_statement_token1] = ACTIONS(1374),
    [aux_sym_for_statement_token1] = ACTIONS(1374),
    [aux_sym_for_statement_token2] = ACTIONS(1374),
    [aux_sym_foreach_statement_token1] = ACTIONS(1374),
    [aux_sym_foreach_statement_token2] = ACTIONS(1374),
    [aux_sym_if_statement_token1] = ACTIONS(1374),
    [aux_sym_if_statement_token2] = ACTIONS(1374),
    [aux_sym_else_if_clause_token1] = ACTIONS(1374),
    [aux_sym_else_clause_token1] = ACTIONS(1374),
    [aux_sym_match_expression_token1] = ACTIONS(1374),
    [aux_sym_match_default_expression_token1] = ACTIONS(1374),
    [aux_sym_switch_statement_token1] = ACTIONS(1374),
    [aux_sym_switch_block_token1] = ACTIONS(1374),
    [anon_sym_PLUS] = ACTIONS(1374),
    [anon_sym_DASH] = ACTIONS(1374),
    [anon_sym_TILDE] = ACTIONS(1372),
    [anon_sym_BANG] = ACTIONS(1372),
    [anon_sym_AT] = ACTIONS(1372),
    [aux_sym_clone_expression_token1] = ACTIONS(1374),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1374),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1374),
    [anon_sym_DASH_DASH] = ACTIONS(1372),
    [anon_sym_PLUS_PLUS] = ACTIONS(1372),
    [aux_sym__list_destructing_token1] = ACTIONS(1374),
    [anon_sym_LBRACK] = ACTIONS(1372),
    [anon_sym_self] = ACTIONS(1374),
    [anon_sym_parent] = ACTIONS(1374),
    [aux_sym__argument_name_token1] = ACTIONS(1374),
    [aux_sym__argument_name_token2] = ACTIONS(1374),
    [anon_sym_POUND_LBRACK] = ACTIONS(1372),
    [aux_sym_encapsed_string_token1] = ACTIONS(1372),
    [anon_sym_DQUOTE] = ACTIONS(1372),
    [aux_sym_string_token1] = ACTIONS(1372),
    [anon_sym_SQUOTE] = ACTIONS(1372),
    [anon_sym_LT_LT_LT] = ACTIONS(1372),
    [anon_sym_BQUOTE] = ACTIONS(1372),
    [anon_sym_DOLLAR] = ACTIONS(1372),
    [aux_sym_yield_expression_token1] = ACTIONS(1374),
    [aux_sym_include_expression_token1] = ACTIONS(1374),
    [aux_sym_include_once_expression_token1] = ACTIONS(1374),
    [aux_sym_require_expression_token1] = ACTIONS(1374),
    [aux_sym_require_once_expression_token1] = ACTIONS(1374),
    [sym_comment] = ACTIONS(5),
  },
  [538] = {
    [sym_text_interpolation] = STATE(538),
    [ts_builtin_sym_end] = ACTIONS(1376),
    [sym_name] = ACTIONS(1378),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1376),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1378),
    [aux_sym_global_declaration_token1] = ACTIONS(1378),
    [aux_sym_namespace_definition_token1] = ACTIONS(1378),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1378),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1378),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1378),
    [anon_sym_BSLASH] = ACTIONS(1376),
    [anon_sym_LBRACE] = ACTIONS(1376),
    [anon_sym_RBRACE] = ACTIONS(1376),
    [aux_sym_trait_declaration_token1] = ACTIONS(1378),
    [aux_sym_interface_declaration_token1] = ACTIONS(1378),
    [aux_sym_enum_declaration_token1] = ACTIONS(1378),
    [aux_sym_enum_case_token1] = ACTIONS(1378),
    [aux_sym_class_declaration_token1] = ACTIONS(1378),
    [aux_sym_final_modifier_token1] = ACTIONS(1378),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1378),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1378),
    [sym_var_modifier] = ACTIONS(1378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1378),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1378),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1378),
    [anon_sym_LPAREN2] = ACTIONS(1376),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1378),
    [aux_sym_cast_type_token1] = ACTIONS(1378),
    [aux_sym_echo_statement_token1] = ACTIONS(1378),
    [aux_sym_exit_statement_token1] = ACTIONS(1378),
    [anon_sym_unset] = ACTIONS(1378),
    [aux_sym_declare_statement_token1] = ACTIONS(1378),
    [aux_sym_declare_statement_token2] = ACTIONS(1378),
    [sym_float] = ACTIONS(1378),
    [aux_sym_try_statement_token1] = ACTIONS(1378),
    [aux_sym_goto_statement_token1] = ACTIONS(1378),
    [aux_sym_continue_statement_token1] = ACTIONS(1378),
    [aux_sym_break_statement_token1] = ACTIONS(1378),
    [sym_integer] = ACTIONS(1378),
    [aux_sym_return_statement_token1] = ACTIONS(1378),
    [aux_sym_throw_expression_token1] = ACTIONS(1378),
    [aux_sym_while_statement_token1] = ACTIONS(1378),
    [aux_sym_while_statement_token2] = ACTIONS(1378),
    [aux_sym_do_statement_token1] = ACTIONS(1378),
    [aux_sym_for_statement_token1] = ACTIONS(1378),
    [aux_sym_for_statement_token2] = ACTIONS(1378),
    [aux_sym_foreach_statement_token1] = ACTIONS(1378),
    [aux_sym_foreach_statement_token2] = ACTIONS(1378),
    [aux_sym_if_statement_token1] = ACTIONS(1378),
    [aux_sym_if_statement_token2] = ACTIONS(1378),
    [aux_sym_else_if_clause_token1] = ACTIONS(1378),
    [aux_sym_else_clause_token1] = ACTIONS(1378),
    [aux_sym_match_expression_token1] = ACTIONS(1378),
    [aux_sym_match_default_expression_token1] = ACTIONS(1378),
    [aux_sym_switch_statement_token1] = ACTIONS(1378),
    [aux_sym_switch_block_token1] = ACTIONS(1378),
    [anon_sym_PLUS] = ACTIONS(1378),
    [anon_sym_DASH] = ACTIONS(1378),
    [anon_sym_TILDE] = ACTIONS(1376),
    [anon_sym_BANG] = ACTIONS(1376),
    [anon_sym_AT] = ACTIONS(1376),
    [aux_sym_clone_expression_token1] = ACTIONS(1378),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1378),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1378),
    [anon_sym_DASH_DASH] = ACTIONS(1376),
    [anon_sym_PLUS_PLUS] = ACTIONS(1376),
    [aux_sym__list_destructing_token1] = ACTIONS(1378),
    [anon_sym_LBRACK] = ACTIONS(1376),
    [anon_sym_self] = ACTIONS(1378),
    [anon_sym_parent] = ACTIONS(1378),
    [aux_sym__argument_name_token1] = ACTIONS(1378),
    [aux_sym__argument_name_token2] = ACTIONS(1378),
    [anon_sym_POUND_LBRACK] = ACTIONS(1376),
    [aux_sym_encapsed_string_token1] = ACTIONS(1376),
    [anon_sym_DQUOTE] = ACTIONS(1376),
    [aux_sym_string_token1] = ACTIONS(1376),
    [anon_sym_SQUOTE] = ACTIONS(1376),
    [anon_sym_LT_LT_LT] = ACTIONS(1376),
    [anon_sym_BQUOTE] = ACTIONS(1376),
    [anon_sym_DOLLAR] = ACTIONS(1376),
    [aux_sym_yield_expression_token1] = ACTIONS(1378),
    [aux_sym_include_expression_token1] = ACTIONS(1378),
    [aux_sym_include_once_expression_token1] = ACTIONS(1378),
    [aux_sym_require_expression_token1] = ACTIONS(1378),
    [aux_sym_require_once_expression_token1] = ACTIONS(1378),
    [sym_comment] = ACTIONS(5),
  },
  [539] = {
    [sym_text_interpolation] = STATE(539),
    [ts_builtin_sym_end] = ACTIONS(1380),
    [sym_name] = ACTIONS(1382),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1380),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1382),
    [aux_sym_global_declaration_token1] = ACTIONS(1382),
    [aux_sym_namespace_definition_token1] = ACTIONS(1382),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1382),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1382),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1382),
    [anon_sym_BSLASH] = ACTIONS(1380),
    [anon_sym_LBRACE] = ACTIONS(1380),
    [anon_sym_RBRACE] = ACTIONS(1380),
    [aux_sym_trait_declaration_token1] = ACTIONS(1382),
    [aux_sym_interface_declaration_token1] = ACTIONS(1382),
    [aux_sym_enum_declaration_token1] = ACTIONS(1382),
    [aux_sym_enum_case_token1] = ACTIONS(1382),
    [aux_sym_class_declaration_token1] = ACTIONS(1382),
    [aux_sym_final_modifier_token1] = ACTIONS(1382),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1382),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1382),
    [sym_var_modifier] = ACTIONS(1382),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1382),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1382),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1382),
    [anon_sym_LPAREN2] = ACTIONS(1380),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1382),
    [aux_sym_cast_type_token1] = ACTIONS(1382),
    [aux_sym_echo_statement_token1] = ACTIONS(1382),
    [aux_sym_exit_statement_token1] = ACTIONS(1382),
    [anon_sym_unset] = ACTIONS(1382),
    [aux_sym_declare_statement_token1] = ACTIONS(1382),
    [aux_sym_declare_statement_token2] = ACTIONS(1382),
    [sym_float] = ACTIONS(1382),
    [aux_sym_try_statement_token1] = ACTIONS(1382),
    [aux_sym_goto_statement_token1] = ACTIONS(1382),
    [aux_sym_continue_statement_token1] = ACTIONS(1382),
    [aux_sym_break_statement_token1] = ACTIONS(1382),
    [sym_integer] = ACTIONS(1382),
    [aux_sym_return_statement_token1] = ACTIONS(1382),
    [aux_sym_throw_expression_token1] = ACTIONS(1382),
    [aux_sym_while_statement_token1] = ACTIONS(1382),
    [aux_sym_while_statement_token2] = ACTIONS(1382),
    [aux_sym_do_statement_token1] = ACTIONS(1382),
    [aux_sym_for_statement_token1] = ACTIONS(1382),
    [aux_sym_for_statement_token2] = ACTIONS(1382),
    [aux_sym_foreach_statement_token1] = ACTIONS(1382),
    [aux_sym_foreach_statement_token2] = ACTIONS(1382),
    [aux_sym_if_statement_token1] = ACTIONS(1382),
    [aux_sym_if_statement_token2] = ACTIONS(1382),
    [aux_sym_else_if_clause_token1] = ACTIONS(1382),
    [aux_sym_else_clause_token1] = ACTIONS(1382),
    [aux_sym_match_expression_token1] = ACTIONS(1382),
    [aux_sym_match_default_expression_token1] = ACTIONS(1382),
    [aux_sym_switch_statement_token1] = ACTIONS(1382),
    [aux_sym_switch_block_token1] = ACTIONS(1382),
    [anon_sym_PLUS] = ACTIONS(1382),
    [anon_sym_DASH] = ACTIONS(1382),
    [anon_sym_TILDE] = ACTIONS(1380),
    [anon_sym_BANG] = ACTIONS(1380),
    [anon_sym_AT] = ACTIONS(1380),
    [aux_sym_clone_expression_token1] = ACTIONS(1382),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1382),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1382),
    [anon_sym_DASH_DASH] = ACTIONS(1380),
    [anon_sym_PLUS_PLUS] = ACTIONS(1380),
    [aux_sym__list_destructing_token1] = ACTIONS(1382),
    [anon_sym_LBRACK] = ACTIONS(1380),
    [anon_sym_self] = ACTIONS(1382),
    [anon_sym_parent] = ACTIONS(1382),
    [aux_sym__argument_name_token1] = ACTIONS(1382),
    [aux_sym__argument_name_token2] = ACTIONS(1382),
    [anon_sym_POUND_LBRACK] = ACTIONS(1380),
    [aux_sym_encapsed_string_token1] = ACTIONS(1380),
    [anon_sym_DQUOTE] = ACTIONS(1380),
    [aux_sym_string_token1] = ACTIONS(1380),
    [anon_sym_SQUOTE] = ACTIONS(1380),
    [anon_sym_LT_LT_LT] = ACTIONS(1380),
    [anon_sym_BQUOTE] = ACTIONS(1380),
    [anon_sym_DOLLAR] = ACTIONS(1380),
    [aux_sym_yield_expression_token1] = ACTIONS(1382),
    [aux_sym_include_expression_token1] = ACTIONS(1382),
    [aux_sym_include_once_expression_token1] = ACTIONS(1382),
    [aux_sym_require_expression_token1] = ACTIONS(1382),
    [aux_sym_require_once_expression_token1] = ACTIONS(1382),
    [sym_comment] = ACTIONS(5),
  },
  [540] = {
    [sym_text_interpolation] = STATE(540),
    [ts_builtin_sym_end] = ACTIONS(1156),
    [sym_name] = ACTIONS(1158),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1156),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1158),
    [aux_sym_global_declaration_token1] = ACTIONS(1158),
    [aux_sym_namespace_definition_token1] = ACTIONS(1158),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1158),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1158),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1158),
    [anon_sym_BSLASH] = ACTIONS(1156),
    [anon_sym_LBRACE] = ACTIONS(1156),
    [anon_sym_RBRACE] = ACTIONS(1156),
    [aux_sym_trait_declaration_token1] = ACTIONS(1158),
    [aux_sym_interface_declaration_token1] = ACTIONS(1158),
    [aux_sym_enum_declaration_token1] = ACTIONS(1158),
    [aux_sym_enum_case_token1] = ACTIONS(1158),
    [aux_sym_class_declaration_token1] = ACTIONS(1158),
    [aux_sym_final_modifier_token1] = ACTIONS(1158),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1158),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1158),
    [sym_var_modifier] = ACTIONS(1158),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1158),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1158),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1158),
    [anon_sym_LPAREN2] = ACTIONS(1156),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1158),
    [aux_sym_cast_type_token1] = ACTIONS(1158),
    [aux_sym_echo_statement_token1] = ACTIONS(1158),
    [aux_sym_exit_statement_token1] = ACTIONS(1158),
    [anon_sym_unset] = ACTIONS(1158),
    [aux_sym_declare_statement_token1] = ACTIONS(1158),
    [aux_sym_declare_statement_token2] = ACTIONS(1158),
    [sym_float] = ACTIONS(1158),
    [aux_sym_try_statement_token1] = ACTIONS(1158),
    [aux_sym_goto_statement_token1] = ACTIONS(1158),
    [aux_sym_continue_statement_token1] = ACTIONS(1158),
    [aux_sym_break_statement_token1] = ACTIONS(1158),
    [sym_integer] = ACTIONS(1158),
    [aux_sym_return_statement_token1] = ACTIONS(1158),
    [aux_sym_throw_expression_token1] = ACTIONS(1158),
    [aux_sym_while_statement_token1] = ACTIONS(1158),
    [aux_sym_while_statement_token2] = ACTIONS(1158),
    [aux_sym_do_statement_token1] = ACTIONS(1158),
    [aux_sym_for_statement_token1] = ACTIONS(1158),
    [aux_sym_for_statement_token2] = ACTIONS(1158),
    [aux_sym_foreach_statement_token1] = ACTIONS(1158),
    [aux_sym_foreach_statement_token2] = ACTIONS(1158),
    [aux_sym_if_statement_token1] = ACTIONS(1158),
    [aux_sym_if_statement_token2] = ACTIONS(1158),
    [aux_sym_else_if_clause_token1] = ACTIONS(1158),
    [aux_sym_else_clause_token1] = ACTIONS(1158),
    [aux_sym_match_expression_token1] = ACTIONS(1158),
    [aux_sym_match_default_expression_token1] = ACTIONS(1158),
    [aux_sym_switch_statement_token1] = ACTIONS(1158),
    [aux_sym_switch_block_token1] = ACTIONS(1158),
    [anon_sym_PLUS] = ACTIONS(1158),
    [anon_sym_DASH] = ACTIONS(1158),
    [anon_sym_TILDE] = ACTIONS(1156),
    [anon_sym_BANG] = ACTIONS(1156),
    [anon_sym_AT] = ACTIONS(1156),
    [aux_sym_clone_expression_token1] = ACTIONS(1158),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1158),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1158),
    [anon_sym_DASH_DASH] = ACTIONS(1156),
    [anon_sym_PLUS_PLUS] = ACTIONS(1156),
    [aux_sym__list_destructing_token1] = ACTIONS(1158),
    [anon_sym_LBRACK] = ACTIONS(1156),
    [anon_sym_self] = ACTIONS(1158),
    [anon_sym_parent] = ACTIONS(1158),
    [aux_sym__argument_name_token1] = ACTIONS(1158),
    [aux_sym__argument_name_token2] = ACTIONS(1158),
    [anon_sym_POUND_LBRACK] = ACTIONS(1156),
    [aux_sym_encapsed_string_token1] = ACTIONS(1156),
    [anon_sym_DQUOTE] = ACTIONS(1156),
    [aux_sym_string_token1] = ACTIONS(1156),
    [anon_sym_SQUOTE] = ACTIONS(1156),
    [anon_sym_LT_LT_LT] = ACTIONS(1156),
    [anon_sym_BQUOTE] = ACTIONS(1156),
    [anon_sym_DOLLAR] = ACTIONS(1156),
    [aux_sym_yield_expression_token1] = ACTIONS(1158),
    [aux_sym_include_expression_token1] = ACTIONS(1158),
    [aux_sym_include_once_expression_token1] = ACTIONS(1158),
    [aux_sym_require_expression_token1] = ACTIONS(1158),
    [aux_sym_require_once_expression_token1] = ACTIONS(1158),
    [sym_comment] = ACTIONS(5),
  },
  [541] = {
    [sym_text_interpolation] = STATE(541),
    [ts_builtin_sym_end] = ACTIONS(1384),
    [sym_name] = ACTIONS(1386),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1384),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1386),
    [aux_sym_global_declaration_token1] = ACTIONS(1386),
    [aux_sym_namespace_definition_token1] = ACTIONS(1386),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1386),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1386),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1386),
    [anon_sym_BSLASH] = ACTIONS(1384),
    [anon_sym_LBRACE] = ACTIONS(1384),
    [anon_sym_RBRACE] = ACTIONS(1384),
    [aux_sym_trait_declaration_token1] = ACTIONS(1386),
    [aux_sym_interface_declaration_token1] = ACTIONS(1386),
    [aux_sym_enum_declaration_token1] = ACTIONS(1386),
    [aux_sym_enum_case_token1] = ACTIONS(1386),
    [aux_sym_class_declaration_token1] = ACTIONS(1386),
    [aux_sym_final_modifier_token1] = ACTIONS(1386),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1386),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1386),
    [sym_var_modifier] = ACTIONS(1386),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1386),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1386),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1386),
    [anon_sym_LPAREN2] = ACTIONS(1384),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1386),
    [aux_sym_cast_type_token1] = ACTIONS(1386),
    [aux_sym_echo_statement_token1] = ACTIONS(1386),
    [aux_sym_exit_statement_token1] = ACTIONS(1386),
    [anon_sym_unset] = ACTIONS(1386),
    [aux_sym_declare_statement_token1] = ACTIONS(1386),
    [aux_sym_declare_statement_token2] = ACTIONS(1386),
    [sym_float] = ACTIONS(1386),
    [aux_sym_try_statement_token1] = ACTIONS(1386),
    [aux_sym_goto_statement_token1] = ACTIONS(1386),
    [aux_sym_continue_statement_token1] = ACTIONS(1386),
    [aux_sym_break_statement_token1] = ACTIONS(1386),
    [sym_integer] = ACTIONS(1386),
    [aux_sym_return_statement_token1] = ACTIONS(1386),
    [aux_sym_throw_expression_token1] = ACTIONS(1386),
    [aux_sym_while_statement_token1] = ACTIONS(1386),
    [aux_sym_while_statement_token2] = ACTIONS(1386),
    [aux_sym_do_statement_token1] = ACTIONS(1386),
    [aux_sym_for_statement_token1] = ACTIONS(1386),
    [aux_sym_for_statement_token2] = ACTIONS(1386),
    [aux_sym_foreach_statement_token1] = ACTIONS(1386),
    [aux_sym_foreach_statement_token2] = ACTIONS(1386),
    [aux_sym_if_statement_token1] = ACTIONS(1386),
    [aux_sym_if_statement_token2] = ACTIONS(1386),
    [aux_sym_else_if_clause_token1] = ACTIONS(1386),
    [aux_sym_else_clause_token1] = ACTIONS(1386),
    [aux_sym_match_expression_token1] = ACTIONS(1386),
    [aux_sym_match_default_expression_token1] = ACTIONS(1386),
    [aux_sym_switch_statement_token1] = ACTIONS(1386),
    [aux_sym_switch_block_token1] = ACTIONS(1386),
    [anon_sym_PLUS] = ACTIONS(1386),
    [anon_sym_DASH] = ACTIONS(1386),
    [anon_sym_TILDE] = ACTIONS(1384),
    [anon_sym_BANG] = ACTIONS(1384),
    [anon_sym_AT] = ACTIONS(1384),
    [aux_sym_clone_expression_token1] = ACTIONS(1386),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1386),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1386),
    [anon_sym_DASH_DASH] = ACTIONS(1384),
    [anon_sym_PLUS_PLUS] = ACTIONS(1384),
    [aux_sym__list_destructing_token1] = ACTIONS(1386),
    [anon_sym_LBRACK] = ACTIONS(1384),
    [anon_sym_self] = ACTIONS(1386),
    [anon_sym_parent] = ACTIONS(1386),
    [aux_sym__argument_name_token1] = ACTIONS(1386),
    [aux_sym__argument_name_token2] = ACTIONS(1386),
    [anon_sym_POUND_LBRACK] = ACTIONS(1384),
    [aux_sym_encapsed_string_token1] = ACTIONS(1384),
    [anon_sym_DQUOTE] = ACTIONS(1384),
    [aux_sym_string_token1] = ACTIONS(1384),
    [anon_sym_SQUOTE] = ACTIONS(1384),
    [anon_sym_LT_LT_LT] = ACTIONS(1384),
    [anon_sym_BQUOTE] = ACTIONS(1384),
    [anon_sym_DOLLAR] = ACTIONS(1384),
    [aux_sym_yield_expression_token1] = ACTIONS(1386),
    [aux_sym_include_expression_token1] = ACTIONS(1386),
    [aux_sym_include_once_expression_token1] = ACTIONS(1386),
    [aux_sym_require_expression_token1] = ACTIONS(1386),
    [aux_sym_require_once_expression_token1] = ACTIONS(1386),
    [sym_comment] = ACTIONS(5),
  },
  [542] = {
    [sym_text_interpolation] = STATE(542),
    [ts_builtin_sym_end] = ACTIONS(1388),
    [sym_name] = ACTIONS(1390),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1388),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1390),
    [aux_sym_global_declaration_token1] = ACTIONS(1390),
    [aux_sym_namespace_definition_token1] = ACTIONS(1390),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1390),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1390),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1390),
    [anon_sym_BSLASH] = ACTIONS(1388),
    [anon_sym_LBRACE] = ACTIONS(1388),
    [anon_sym_RBRACE] = ACTIONS(1388),
    [aux_sym_trait_declaration_token1] = ACTIONS(1390),
    [aux_sym_interface_declaration_token1] = ACTIONS(1390),
    [aux_sym_enum_declaration_token1] = ACTIONS(1390),
    [aux_sym_enum_case_token1] = ACTIONS(1390),
    [aux_sym_class_declaration_token1] = ACTIONS(1390),
    [aux_sym_final_modifier_token1] = ACTIONS(1390),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1390),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1390),
    [sym_var_modifier] = ACTIONS(1390),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1390),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1390),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1390),
    [anon_sym_LPAREN2] = ACTIONS(1388),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1390),
    [aux_sym_cast_type_token1] = ACTIONS(1390),
    [aux_sym_echo_statement_token1] = ACTIONS(1390),
    [aux_sym_exit_statement_token1] = ACTIONS(1390),
    [anon_sym_unset] = ACTIONS(1390),
    [aux_sym_declare_statement_token1] = ACTIONS(1390),
    [aux_sym_declare_statement_token2] = ACTIONS(1390),
    [sym_float] = ACTIONS(1390),
    [aux_sym_try_statement_token1] = ACTIONS(1390),
    [aux_sym_goto_statement_token1] = ACTIONS(1390),
    [aux_sym_continue_statement_token1] = ACTIONS(1390),
    [aux_sym_break_statement_token1] = ACTIONS(1390),
    [sym_integer] = ACTIONS(1390),
    [aux_sym_return_statement_token1] = ACTIONS(1390),
    [aux_sym_throw_expression_token1] = ACTIONS(1390),
    [aux_sym_while_statement_token1] = ACTIONS(1390),
    [aux_sym_while_statement_token2] = ACTIONS(1390),
    [aux_sym_do_statement_token1] = ACTIONS(1390),
    [aux_sym_for_statement_token1] = ACTIONS(1390),
    [aux_sym_for_statement_token2] = ACTIONS(1390),
    [aux_sym_foreach_statement_token1] = ACTIONS(1390),
    [aux_sym_foreach_statement_token2] = ACTIONS(1390),
    [aux_sym_if_statement_token1] = ACTIONS(1390),
    [aux_sym_if_statement_token2] = ACTIONS(1390),
    [aux_sym_else_if_clause_token1] = ACTIONS(1390),
    [aux_sym_else_clause_token1] = ACTIONS(1390),
    [aux_sym_match_expression_token1] = ACTIONS(1390),
    [aux_sym_match_default_expression_token1] = ACTIONS(1390),
    [aux_sym_switch_statement_token1] = ACTIONS(1390),
    [aux_sym_switch_block_token1] = ACTIONS(1390),
    [anon_sym_PLUS] = ACTIONS(1390),
    [anon_sym_DASH] = ACTIONS(1390),
    [anon_sym_TILDE] = ACTIONS(1388),
    [anon_sym_BANG] = ACTIONS(1388),
    [anon_sym_AT] = ACTIONS(1388),
    [aux_sym_clone_expression_token1] = ACTIONS(1390),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1390),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1390),
    [anon_sym_DASH_DASH] = ACTIONS(1388),
    [anon_sym_PLUS_PLUS] = ACTIONS(1388),
    [aux_sym__list_destructing_token1] = ACTIONS(1390),
    [anon_sym_LBRACK] = ACTIONS(1388),
    [anon_sym_self] = ACTIONS(1390),
    [anon_sym_parent] = ACTIONS(1390),
    [aux_sym__argument_name_token1] = ACTIONS(1390),
    [aux_sym__argument_name_token2] = ACTIONS(1390),
    [anon_sym_POUND_LBRACK] = ACTIONS(1388),
    [aux_sym_encapsed_string_token1] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1388),
    [aux_sym_string_token1] = ACTIONS(1388),
    [anon_sym_SQUOTE] = ACTIONS(1388),
    [anon_sym_LT_LT_LT] = ACTIONS(1388),
    [anon_sym_BQUOTE] = ACTIONS(1388),
    [anon_sym_DOLLAR] = ACTIONS(1388),
    [aux_sym_yield_expression_token1] = ACTIONS(1390),
    [aux_sym_include_expression_token1] = ACTIONS(1390),
    [aux_sym_include_once_expression_token1] = ACTIONS(1390),
    [aux_sym_require_expression_token1] = ACTIONS(1390),
    [aux_sym_require_once_expression_token1] = ACTIONS(1390),
    [sym_comment] = ACTIONS(5),
  },
  [543] = {
    [sym_text_interpolation] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1392),
    [sym_name] = ACTIONS(1394),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1392),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1394),
    [aux_sym_global_declaration_token1] = ACTIONS(1394),
    [aux_sym_namespace_definition_token1] = ACTIONS(1394),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1394),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1394),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1394),
    [anon_sym_BSLASH] = ACTIONS(1392),
    [anon_sym_LBRACE] = ACTIONS(1392),
    [anon_sym_RBRACE] = ACTIONS(1392),
    [aux_sym_trait_declaration_token1] = ACTIONS(1394),
    [aux_sym_interface_declaration_token1] = ACTIONS(1394),
    [aux_sym_enum_declaration_token1] = ACTIONS(1394),
    [aux_sym_enum_case_token1] = ACTIONS(1394),
    [aux_sym_class_declaration_token1] = ACTIONS(1394),
    [aux_sym_final_modifier_token1] = ACTIONS(1394),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1394),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1394),
    [sym_var_modifier] = ACTIONS(1394),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1394),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1394),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1394),
    [anon_sym_LPAREN2] = ACTIONS(1392),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1394),
    [aux_sym_cast_type_token1] = ACTIONS(1394),
    [aux_sym_echo_statement_token1] = ACTIONS(1394),
    [aux_sym_exit_statement_token1] = ACTIONS(1394),
    [anon_sym_unset] = ACTIONS(1394),
    [aux_sym_declare_statement_token1] = ACTIONS(1394),
    [aux_sym_declare_statement_token2] = ACTIONS(1394),
    [sym_float] = ACTIONS(1394),
    [aux_sym_try_statement_token1] = ACTIONS(1394),
    [aux_sym_goto_statement_token1] = ACTIONS(1394),
    [aux_sym_continue_statement_token1] = ACTIONS(1394),
    [aux_sym_break_statement_token1] = ACTIONS(1394),
    [sym_integer] = ACTIONS(1394),
    [aux_sym_return_statement_token1] = ACTIONS(1394),
    [aux_sym_throw_expression_token1] = ACTIONS(1394),
    [aux_sym_while_statement_token1] = ACTIONS(1394),
    [aux_sym_while_statement_token2] = ACTIONS(1394),
    [aux_sym_do_statement_token1] = ACTIONS(1394),
    [aux_sym_for_statement_token1] = ACTIONS(1394),
    [aux_sym_for_statement_token2] = ACTIONS(1394),
    [aux_sym_foreach_statement_token1] = ACTIONS(1394),
    [aux_sym_foreach_statement_token2] = ACTIONS(1394),
    [aux_sym_if_statement_token1] = ACTIONS(1394),
    [aux_sym_if_statement_token2] = ACTIONS(1394),
    [aux_sym_else_if_clause_token1] = ACTIONS(1394),
    [aux_sym_else_clause_token1] = ACTIONS(1394),
    [aux_sym_match_expression_token1] = ACTIONS(1394),
    [aux_sym_match_default_expression_token1] = ACTIONS(1394),
    [aux_sym_switch_statement_token1] = ACTIONS(1394),
    [aux_sym_switch_block_token1] = ACTIONS(1394),
    [anon_sym_PLUS] = ACTIONS(1394),
    [anon_sym_DASH] = ACTIONS(1394),
    [anon_sym_TILDE] = ACTIONS(1392),
    [anon_sym_BANG] = ACTIONS(1392),
    [anon_sym_AT] = ACTIONS(1392),
    [aux_sym_clone_expression_token1] = ACTIONS(1394),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1394),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1394),
    [anon_sym_DASH_DASH] = ACTIONS(1392),
    [anon_sym_PLUS_PLUS] = ACTIONS(1392),
    [aux_sym__list_destructing_token1] = ACTIONS(1394),
    [anon_sym_LBRACK] = ACTIONS(1392),
    [anon_sym_self] = ACTIONS(1394),
    [anon_sym_parent] = ACTIONS(1394),
    [aux_sym__argument_name_token1] = ACTIONS(1394),
    [aux_sym__argument_name_token2] = ACTIONS(1394),
    [anon_sym_POUND_LBRACK] = ACTIONS(1392),
    [aux_sym_encapsed_string_token1] = ACTIONS(1392),
    [anon_sym_DQUOTE] = ACTIONS(1392),
    [aux_sym_string_token1] = ACTIONS(1392),
    [anon_sym_SQUOTE] = ACTIONS(1392),
    [anon_sym_LT_LT_LT] = ACTIONS(1392),
    [anon_sym_BQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR] = ACTIONS(1392),
    [aux_sym_yield_expression_token1] = ACTIONS(1394),
    [aux_sym_include_expression_token1] = ACTIONS(1394),
    [aux_sym_include_once_expression_token1] = ACTIONS(1394),
    [aux_sym_require_expression_token1] = ACTIONS(1394),
    [aux_sym_require_once_expression_token1] = ACTIONS(1394),
    [sym_comment] = ACTIONS(5),
  },
  [544] = {
    [sym_text_interpolation] = STATE(544),
    [ts_builtin_sym_end] = ACTIONS(1396),
    [sym_name] = ACTIONS(1398),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1396),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1398),
    [aux_sym_global_declaration_token1] = ACTIONS(1398),
    [aux_sym_namespace_definition_token1] = ACTIONS(1398),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1398),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1398),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1398),
    [anon_sym_BSLASH] = ACTIONS(1396),
    [anon_sym_LBRACE] = ACTIONS(1396),
    [anon_sym_RBRACE] = ACTIONS(1396),
    [aux_sym_trait_declaration_token1] = ACTIONS(1398),
    [aux_sym_interface_declaration_token1] = ACTIONS(1398),
    [aux_sym_enum_declaration_token1] = ACTIONS(1398),
    [aux_sym_enum_case_token1] = ACTIONS(1398),
    [aux_sym_class_declaration_token1] = ACTIONS(1398),
    [aux_sym_final_modifier_token1] = ACTIONS(1398),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1398),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1398),
    [sym_var_modifier] = ACTIONS(1398),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1398),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1398),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1398),
    [anon_sym_LPAREN2] = ACTIONS(1396),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1398),
    [aux_sym_cast_type_token1] = ACTIONS(1398),
    [aux_sym_echo_statement_token1] = ACTIONS(1398),
    [aux_sym_exit_statement_token1] = ACTIONS(1398),
    [anon_sym_unset] = ACTIONS(1398),
    [aux_sym_declare_statement_token1] = ACTIONS(1398),
    [aux_sym_declare_statement_token2] = ACTIONS(1398),
    [sym_float] = ACTIONS(1398),
    [aux_sym_try_statement_token1] = ACTIONS(1398),
    [aux_sym_goto_statement_token1] = ACTIONS(1398),
    [aux_sym_continue_statement_token1] = ACTIONS(1398),
    [aux_sym_break_statement_token1] = ACTIONS(1398),
    [sym_integer] = ACTIONS(1398),
    [aux_sym_return_statement_token1] = ACTIONS(1398),
    [aux_sym_throw_expression_token1] = ACTIONS(1398),
    [aux_sym_while_statement_token1] = ACTIONS(1398),
    [aux_sym_while_statement_token2] = ACTIONS(1398),
    [aux_sym_do_statement_token1] = ACTIONS(1398),
    [aux_sym_for_statement_token1] = ACTIONS(1398),
    [aux_sym_for_statement_token2] = ACTIONS(1398),
    [aux_sym_foreach_statement_token1] = ACTIONS(1398),
    [aux_sym_foreach_statement_token2] = ACTIONS(1398),
    [aux_sym_if_statement_token1] = ACTIONS(1398),
    [aux_sym_if_statement_token2] = ACTIONS(1398),
    [aux_sym_else_if_clause_token1] = ACTIONS(1398),
    [aux_sym_else_clause_token1] = ACTIONS(1398),
    [aux_sym_match_expression_token1] = ACTIONS(1398),
    [aux_sym_match_default_expression_token1] = ACTIONS(1398),
    [aux_sym_switch_statement_token1] = ACTIONS(1398),
    [aux_sym_switch_block_token1] = ACTIONS(1398),
    [anon_sym_PLUS] = ACTIONS(1398),
    [anon_sym_DASH] = ACTIONS(1398),
    [anon_sym_TILDE] = ACTIONS(1396),
    [anon_sym_BANG] = ACTIONS(1396),
    [anon_sym_AT] = ACTIONS(1396),
    [aux_sym_clone_expression_token1] = ACTIONS(1398),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1398),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1398),
    [anon_sym_DASH_DASH] = ACTIONS(1396),
    [anon_sym_PLUS_PLUS] = ACTIONS(1396),
    [aux_sym__list_destructing_token1] = ACTIONS(1398),
    [anon_sym_LBRACK] = ACTIONS(1396),
    [anon_sym_self] = ACTIONS(1398),
    [anon_sym_parent] = ACTIONS(1398),
    [aux_sym__argument_name_token1] = ACTIONS(1398),
    [aux_sym__argument_name_token2] = ACTIONS(1398),
    [anon_sym_POUND_LBRACK] = ACTIONS(1396),
    [aux_sym_encapsed_string_token1] = ACTIONS(1396),
    [anon_sym_DQUOTE] = ACTIONS(1396),
    [aux_sym_string_token1] = ACTIONS(1396),
    [anon_sym_SQUOTE] = ACTIONS(1396),
    [anon_sym_LT_LT_LT] = ACTIONS(1396),
    [anon_sym_BQUOTE] = ACTIONS(1396),
    [anon_sym_DOLLAR] = ACTIONS(1396),
    [aux_sym_yield_expression_token1] = ACTIONS(1398),
    [aux_sym_include_expression_token1] = ACTIONS(1398),
    [aux_sym_include_once_expression_token1] = ACTIONS(1398),
    [aux_sym_require_expression_token1] = ACTIONS(1398),
    [aux_sym_require_once_expression_token1] = ACTIONS(1398),
    [sym_comment] = ACTIONS(5),
  },
  [545] = {
    [sym_text_interpolation] = STATE(545),
    [ts_builtin_sym_end] = ACTIONS(1400),
    [sym_name] = ACTIONS(1402),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1400),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1402),
    [aux_sym_global_declaration_token1] = ACTIONS(1402),
    [aux_sym_namespace_definition_token1] = ACTIONS(1402),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1402),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1402),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1402),
    [anon_sym_BSLASH] = ACTIONS(1400),
    [anon_sym_LBRACE] = ACTIONS(1400),
    [anon_sym_RBRACE] = ACTIONS(1400),
    [aux_sym_trait_declaration_token1] = ACTIONS(1402),
    [aux_sym_interface_declaration_token1] = ACTIONS(1402),
    [aux_sym_enum_declaration_token1] = ACTIONS(1402),
    [aux_sym_enum_case_token1] = ACTIONS(1402),
    [aux_sym_class_declaration_token1] = ACTIONS(1402),
    [aux_sym_final_modifier_token1] = ACTIONS(1402),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1402),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1402),
    [sym_var_modifier] = ACTIONS(1402),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1402),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1402),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1402),
    [anon_sym_LPAREN2] = ACTIONS(1400),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1402),
    [aux_sym_cast_type_token1] = ACTIONS(1402),
    [aux_sym_echo_statement_token1] = ACTIONS(1402),
    [aux_sym_exit_statement_token1] = ACTIONS(1402),
    [anon_sym_unset] = ACTIONS(1402),
    [aux_sym_declare_statement_token1] = ACTIONS(1402),
    [aux_sym_declare_statement_token2] = ACTIONS(1402),
    [sym_float] = ACTIONS(1402),
    [aux_sym_try_statement_token1] = ACTIONS(1402),
    [aux_sym_goto_statement_token1] = ACTIONS(1402),
    [aux_sym_continue_statement_token1] = ACTIONS(1402),
    [aux_sym_break_statement_token1] = ACTIONS(1402),
    [sym_integer] = ACTIONS(1402),
    [aux_sym_return_statement_token1] = ACTIONS(1402),
    [aux_sym_throw_expression_token1] = ACTIONS(1402),
    [aux_sym_while_statement_token1] = ACTIONS(1402),
    [aux_sym_while_statement_token2] = ACTIONS(1402),
    [aux_sym_do_statement_token1] = ACTIONS(1402),
    [aux_sym_for_statement_token1] = ACTIONS(1402),
    [aux_sym_for_statement_token2] = ACTIONS(1402),
    [aux_sym_foreach_statement_token1] = ACTIONS(1402),
    [aux_sym_foreach_statement_token2] = ACTIONS(1402),
    [aux_sym_if_statement_token1] = ACTIONS(1402),
    [aux_sym_if_statement_token2] = ACTIONS(1402),
    [aux_sym_else_if_clause_token1] = ACTIONS(1402),
    [aux_sym_else_clause_token1] = ACTIONS(1402),
    [aux_sym_match_expression_token1] = ACTIONS(1402),
    [aux_sym_match_default_expression_token1] = ACTIONS(1402),
    [aux_sym_switch_statement_token1] = ACTIONS(1402),
    [aux_sym_switch_block_token1] = ACTIONS(1402),
    [anon_sym_PLUS] = ACTIONS(1402),
    [anon_sym_DASH] = ACTIONS(1402),
    [anon_sym_TILDE] = ACTIONS(1400),
    [anon_sym_BANG] = ACTIONS(1400),
    [anon_sym_AT] = ACTIONS(1400),
    [aux_sym_clone_expression_token1] = ACTIONS(1402),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1402),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1402),
    [anon_sym_DASH_DASH] = ACTIONS(1400),
    [anon_sym_PLUS_PLUS] = ACTIONS(1400),
    [aux_sym__list_destructing_token1] = ACTIONS(1402),
    [anon_sym_LBRACK] = ACTIONS(1400),
    [anon_sym_self] = ACTIONS(1402),
    [anon_sym_parent] = ACTIONS(1402),
    [aux_sym__argument_name_token1] = ACTIONS(1402),
    [aux_sym__argument_name_token2] = ACTIONS(1402),
    [anon_sym_POUND_LBRACK] = ACTIONS(1400),
    [aux_sym_encapsed_string_token1] = ACTIONS(1400),
    [anon_sym_DQUOTE] = ACTIONS(1400),
    [aux_sym_string_token1] = ACTIONS(1400),
    [anon_sym_SQUOTE] = ACTIONS(1400),
    [anon_sym_LT_LT_LT] = ACTIONS(1400),
    [anon_sym_BQUOTE] = ACTIONS(1400),
    [anon_sym_DOLLAR] = ACTIONS(1400),
    [aux_sym_yield_expression_token1] = ACTIONS(1402),
    [aux_sym_include_expression_token1] = ACTIONS(1402),
    [aux_sym_include_once_expression_token1] = ACTIONS(1402),
    [aux_sym_require_expression_token1] = ACTIONS(1402),
    [aux_sym_require_once_expression_token1] = ACTIONS(1402),
    [sym_comment] = ACTIONS(5),
  },
  [546] = {
    [sym_text_interpolation] = STATE(546),
    [ts_builtin_sym_end] = ACTIONS(1404),
    [sym_name] = ACTIONS(1406),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1404),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1406),
    [aux_sym_global_declaration_token1] = ACTIONS(1406),
    [aux_sym_namespace_definition_token1] = ACTIONS(1406),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1406),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1406),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1406),
    [anon_sym_BSLASH] = ACTIONS(1404),
    [anon_sym_LBRACE] = ACTIONS(1404),
    [anon_sym_RBRACE] = ACTIONS(1404),
    [aux_sym_trait_declaration_token1] = ACTIONS(1406),
    [aux_sym_interface_declaration_token1] = ACTIONS(1406),
    [aux_sym_enum_declaration_token1] = ACTIONS(1406),
    [aux_sym_enum_case_token1] = ACTIONS(1406),
    [aux_sym_class_declaration_token1] = ACTIONS(1406),
    [aux_sym_final_modifier_token1] = ACTIONS(1406),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1406),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1406),
    [sym_var_modifier] = ACTIONS(1406),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1406),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1406),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1406),
    [anon_sym_LPAREN2] = ACTIONS(1404),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1406),
    [aux_sym_cast_type_token1] = ACTIONS(1406),
    [aux_sym_echo_statement_token1] = ACTIONS(1406),
    [aux_sym_exit_statement_token1] = ACTIONS(1406),
    [anon_sym_unset] = ACTIONS(1406),
    [aux_sym_declare_statement_token1] = ACTIONS(1406),
    [aux_sym_declare_statement_token2] = ACTIONS(1406),
    [sym_float] = ACTIONS(1406),
    [aux_sym_try_statement_token1] = ACTIONS(1406),
    [aux_sym_goto_statement_token1] = ACTIONS(1406),
    [aux_sym_continue_statement_token1] = ACTIONS(1406),
    [aux_sym_break_statement_token1] = ACTIONS(1406),
    [sym_integer] = ACTIONS(1406),
    [aux_sym_return_statement_token1] = ACTIONS(1406),
    [aux_sym_throw_expression_token1] = ACTIONS(1406),
    [aux_sym_while_statement_token1] = ACTIONS(1406),
    [aux_sym_while_statement_token2] = ACTIONS(1406),
    [aux_sym_do_statement_token1] = ACTIONS(1406),
    [aux_sym_for_statement_token1] = ACTIONS(1406),
    [aux_sym_for_statement_token2] = ACTIONS(1406),
    [aux_sym_foreach_statement_token1] = ACTIONS(1406),
    [aux_sym_foreach_statement_token2] = ACTIONS(1406),
    [aux_sym_if_statement_token1] = ACTIONS(1406),
    [aux_sym_if_statement_token2] = ACTIONS(1406),
    [aux_sym_else_if_clause_token1] = ACTIONS(1406),
    [aux_sym_else_clause_token1] = ACTIONS(1406),
    [aux_sym_match_expression_token1] = ACTIONS(1406),
    [aux_sym_match_default_expression_token1] = ACTIONS(1406),
    [aux_sym_switch_statement_token1] = ACTIONS(1406),
    [aux_sym_switch_block_token1] = ACTIONS(1406),
    [anon_sym_PLUS] = ACTIONS(1406),
    [anon_sym_DASH] = ACTIONS(1406),
    [anon_sym_TILDE] = ACTIONS(1404),
    [anon_sym_BANG] = ACTIONS(1404),
    [anon_sym_AT] = ACTIONS(1404),
    [aux_sym_clone_expression_token1] = ACTIONS(1406),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1406),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1406),
    [anon_sym_DASH_DASH] = ACTIONS(1404),
    [anon_sym_PLUS_PLUS] = ACTIONS(1404),
    [aux_sym__list_destructing_token1] = ACTIONS(1406),
    [anon_sym_LBRACK] = ACTIONS(1404),
    [anon_sym_self] = ACTIONS(1406),
    [anon_sym_parent] = ACTIONS(1406),
    [aux_sym__argument_name_token1] = ACTIONS(1406),
    [aux_sym__argument_name_token2] = ACTIONS(1406),
    [anon_sym_POUND_LBRACK] = ACTIONS(1404),
    [aux_sym_encapsed_string_token1] = ACTIONS(1404),
    [anon_sym_DQUOTE] = ACTIONS(1404),
    [aux_sym_string_token1] = ACTIONS(1404),
    [anon_sym_SQUOTE] = ACTIONS(1404),
    [anon_sym_LT_LT_LT] = ACTIONS(1404),
    [anon_sym_BQUOTE] = ACTIONS(1404),
    [anon_sym_DOLLAR] = ACTIONS(1404),
    [aux_sym_yield_expression_token1] = ACTIONS(1406),
    [aux_sym_include_expression_token1] = ACTIONS(1406),
    [aux_sym_include_once_expression_token1] = ACTIONS(1406),
    [aux_sym_require_expression_token1] = ACTIONS(1406),
    [aux_sym_require_once_expression_token1] = ACTIONS(1406),
    [sym_comment] = ACTIONS(5),
  },
  [547] = {
    [sym_text_interpolation] = STATE(547),
    [ts_builtin_sym_end] = ACTIONS(1408),
    [sym_name] = ACTIONS(1410),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1408),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1410),
    [aux_sym_global_declaration_token1] = ACTIONS(1410),
    [aux_sym_namespace_definition_token1] = ACTIONS(1410),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1410),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1410),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1410),
    [anon_sym_BSLASH] = ACTIONS(1408),
    [anon_sym_LBRACE] = ACTIONS(1408),
    [anon_sym_RBRACE] = ACTIONS(1408),
    [aux_sym_trait_declaration_token1] = ACTIONS(1410),
    [aux_sym_interface_declaration_token1] = ACTIONS(1410),
    [aux_sym_enum_declaration_token1] = ACTIONS(1410),
    [aux_sym_enum_case_token1] = ACTIONS(1410),
    [aux_sym_class_declaration_token1] = ACTIONS(1410),
    [aux_sym_final_modifier_token1] = ACTIONS(1410),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1410),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1410),
    [sym_var_modifier] = ACTIONS(1410),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1410),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1410),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1410),
    [anon_sym_LPAREN2] = ACTIONS(1408),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1410),
    [aux_sym_cast_type_token1] = ACTIONS(1410),
    [aux_sym_echo_statement_token1] = ACTIONS(1410),
    [aux_sym_exit_statement_token1] = ACTIONS(1410),
    [anon_sym_unset] = ACTIONS(1410),
    [aux_sym_declare_statement_token1] = ACTIONS(1410),
    [aux_sym_declare_statement_token2] = ACTIONS(1410),
    [sym_float] = ACTIONS(1410),
    [aux_sym_try_statement_token1] = ACTIONS(1410),
    [aux_sym_goto_statement_token1] = ACTIONS(1410),
    [aux_sym_continue_statement_token1] = ACTIONS(1410),
    [aux_sym_break_statement_token1] = ACTIONS(1410),
    [sym_integer] = ACTIONS(1410),
    [aux_sym_return_statement_token1] = ACTIONS(1410),
    [aux_sym_throw_expression_token1] = ACTIONS(1410),
    [aux_sym_while_statement_token1] = ACTIONS(1410),
    [aux_sym_while_statement_token2] = ACTIONS(1410),
    [aux_sym_do_statement_token1] = ACTIONS(1410),
    [aux_sym_for_statement_token1] = ACTIONS(1410),
    [aux_sym_for_statement_token2] = ACTIONS(1410),
    [aux_sym_foreach_statement_token1] = ACTIONS(1410),
    [aux_sym_foreach_statement_token2] = ACTIONS(1410),
    [aux_sym_if_statement_token1] = ACTIONS(1410),
    [aux_sym_if_statement_token2] = ACTIONS(1410),
    [aux_sym_else_if_clause_token1] = ACTIONS(1410),
    [aux_sym_else_clause_token1] = ACTIONS(1410),
    [aux_sym_match_expression_token1] = ACTIONS(1410),
    [aux_sym_match_default_expression_token1] = ACTIONS(1410),
    [aux_sym_switch_statement_token1] = ACTIONS(1410),
    [aux_sym_switch_block_token1] = ACTIONS(1410),
    [anon_sym_PLUS] = ACTIONS(1410),
    [anon_sym_DASH] = ACTIONS(1410),
    [anon_sym_TILDE] = ACTIONS(1408),
    [anon_sym_BANG] = ACTIONS(1408),
    [anon_sym_AT] = ACTIONS(1408),
    [aux_sym_clone_expression_token1] = ACTIONS(1410),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1410),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1410),
    [anon_sym_DASH_DASH] = ACTIONS(1408),
    [anon_sym_PLUS_PLUS] = ACTIONS(1408),
    [aux_sym__list_destructing_token1] = ACTIONS(1410),
    [anon_sym_LBRACK] = ACTIONS(1408),
    [anon_sym_self] = ACTIONS(1410),
    [anon_sym_parent] = ACTIONS(1410),
    [aux_sym__argument_name_token1] = ACTIONS(1410),
    [aux_sym__argument_name_token2] = ACTIONS(1410),
    [anon_sym_POUND_LBRACK] = ACTIONS(1408),
    [aux_sym_encapsed_string_token1] = ACTIONS(1408),
    [anon_sym_DQUOTE] = ACTIONS(1408),
    [aux_sym_string_token1] = ACTIONS(1408),
    [anon_sym_SQUOTE] = ACTIONS(1408),
    [anon_sym_LT_LT_LT] = ACTIONS(1408),
    [anon_sym_BQUOTE] = ACTIONS(1408),
    [anon_sym_DOLLAR] = ACTIONS(1408),
    [aux_sym_yield_expression_token1] = ACTIONS(1410),
    [aux_sym_include_expression_token1] = ACTIONS(1410),
    [aux_sym_include_once_expression_token1] = ACTIONS(1410),
    [aux_sym_require_expression_token1] = ACTIONS(1410),
    [aux_sym_require_once_expression_token1] = ACTIONS(1410),
    [sym_comment] = ACTIONS(5),
  },
  [548] = {
    [sym_text_interpolation] = STATE(548),
    [ts_builtin_sym_end] = ACTIONS(1412),
    [sym_name] = ACTIONS(1414),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1412),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1414),
    [aux_sym_global_declaration_token1] = ACTIONS(1414),
    [aux_sym_namespace_definition_token1] = ACTIONS(1414),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1414),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1414),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1414),
    [anon_sym_BSLASH] = ACTIONS(1412),
    [anon_sym_LBRACE] = ACTIONS(1412),
    [anon_sym_RBRACE] = ACTIONS(1412),
    [aux_sym_trait_declaration_token1] = ACTIONS(1414),
    [aux_sym_interface_declaration_token1] = ACTIONS(1414),
    [aux_sym_enum_declaration_token1] = ACTIONS(1414),
    [aux_sym_enum_case_token1] = ACTIONS(1414),
    [aux_sym_class_declaration_token1] = ACTIONS(1414),
    [aux_sym_final_modifier_token1] = ACTIONS(1414),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1414),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1414),
    [sym_var_modifier] = ACTIONS(1414),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1414),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1414),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1414),
    [anon_sym_LPAREN2] = ACTIONS(1412),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1414),
    [aux_sym_cast_type_token1] = ACTIONS(1414),
    [aux_sym_echo_statement_token1] = ACTIONS(1414),
    [aux_sym_exit_statement_token1] = ACTIONS(1414),
    [anon_sym_unset] = ACTIONS(1414),
    [aux_sym_declare_statement_token1] = ACTIONS(1414),
    [aux_sym_declare_statement_token2] = ACTIONS(1414),
    [sym_float] = ACTIONS(1414),
    [aux_sym_try_statement_token1] = ACTIONS(1414),
    [aux_sym_goto_statement_token1] = ACTIONS(1414),
    [aux_sym_continue_statement_token1] = ACTIONS(1414),
    [aux_sym_break_statement_token1] = ACTIONS(1414),
    [sym_integer] = ACTIONS(1414),
    [aux_sym_return_statement_token1] = ACTIONS(1414),
    [aux_sym_throw_expression_token1] = ACTIONS(1414),
    [aux_sym_while_statement_token1] = ACTIONS(1414),
    [aux_sym_while_statement_token2] = ACTIONS(1414),
    [aux_sym_do_statement_token1] = ACTIONS(1414),
    [aux_sym_for_statement_token1] = ACTIONS(1414),
    [aux_sym_for_statement_token2] = ACTIONS(1414),
    [aux_sym_foreach_statement_token1] = ACTIONS(1414),
    [aux_sym_foreach_statement_token2] = ACTIONS(1414),
    [aux_sym_if_statement_token1] = ACTIONS(1414),
    [aux_sym_if_statement_token2] = ACTIONS(1414),
    [aux_sym_else_if_clause_token1] = ACTIONS(1414),
    [aux_sym_else_clause_token1] = ACTIONS(1414),
    [aux_sym_match_expression_token1] = ACTIONS(1414),
    [aux_sym_match_default_expression_token1] = ACTIONS(1414),
    [aux_sym_switch_statement_token1] = ACTIONS(1414),
    [aux_sym_switch_block_token1] = ACTIONS(1414),
    [anon_sym_PLUS] = ACTIONS(1414),
    [anon_sym_DASH] = ACTIONS(1414),
    [anon_sym_TILDE] = ACTIONS(1412),
    [anon_sym_BANG] = ACTIONS(1412),
    [anon_sym_AT] = ACTIONS(1412),
    [aux_sym_clone_expression_token1] = ACTIONS(1414),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1414),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1414),
    [anon_sym_DASH_DASH] = ACTIONS(1412),
    [anon_sym_PLUS_PLUS] = ACTIONS(1412),
    [aux_sym__list_destructing_token1] = ACTIONS(1414),
    [anon_sym_LBRACK] = ACTIONS(1412),
    [anon_sym_self] = ACTIONS(1414),
    [anon_sym_parent] = ACTIONS(1414),
    [aux_sym__argument_name_token1] = ACTIONS(1414),
    [aux_sym__argument_name_token2] = ACTIONS(1414),
    [anon_sym_POUND_LBRACK] = ACTIONS(1412),
    [aux_sym_encapsed_string_token1] = ACTIONS(1412),
    [anon_sym_DQUOTE] = ACTIONS(1412),
    [aux_sym_string_token1] = ACTIONS(1412),
    [anon_sym_SQUOTE] = ACTIONS(1412),
    [anon_sym_LT_LT_LT] = ACTIONS(1412),
    [anon_sym_BQUOTE] = ACTIONS(1412),
    [anon_sym_DOLLAR] = ACTIONS(1412),
    [aux_sym_yield_expression_token1] = ACTIONS(1414),
    [aux_sym_include_expression_token1] = ACTIONS(1414),
    [aux_sym_include_once_expression_token1] = ACTIONS(1414),
    [aux_sym_require_expression_token1] = ACTIONS(1414),
    [aux_sym_require_once_expression_token1] = ACTIONS(1414),
    [sym_comment] = ACTIONS(5),
  },
  [549] = {
    [sym_text_interpolation] = STATE(549),
    [ts_builtin_sym_end] = ACTIONS(1416),
    [sym_name] = ACTIONS(1418),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1416),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1418),
    [aux_sym_global_declaration_token1] = ACTIONS(1418),
    [aux_sym_namespace_definition_token1] = ACTIONS(1418),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1418),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1418),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1418),
    [anon_sym_BSLASH] = ACTIONS(1416),
    [anon_sym_LBRACE] = ACTIONS(1416),
    [anon_sym_RBRACE] = ACTIONS(1416),
    [aux_sym_trait_declaration_token1] = ACTIONS(1418),
    [aux_sym_interface_declaration_token1] = ACTIONS(1418),
    [aux_sym_enum_declaration_token1] = ACTIONS(1418),
    [aux_sym_enum_case_token1] = ACTIONS(1418),
    [aux_sym_class_declaration_token1] = ACTIONS(1418),
    [aux_sym_final_modifier_token1] = ACTIONS(1418),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1418),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1418),
    [sym_var_modifier] = ACTIONS(1418),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1418),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1418),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1418),
    [anon_sym_LPAREN2] = ACTIONS(1416),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1418),
    [aux_sym_cast_type_token1] = ACTIONS(1418),
    [aux_sym_echo_statement_token1] = ACTIONS(1418),
    [aux_sym_exit_statement_token1] = ACTIONS(1418),
    [anon_sym_unset] = ACTIONS(1418),
    [aux_sym_declare_statement_token1] = ACTIONS(1418),
    [aux_sym_declare_statement_token2] = ACTIONS(1418),
    [sym_float] = ACTIONS(1418),
    [aux_sym_try_statement_token1] = ACTIONS(1418),
    [aux_sym_goto_statement_token1] = ACTIONS(1418),
    [aux_sym_continue_statement_token1] = ACTIONS(1418),
    [aux_sym_break_statement_token1] = ACTIONS(1418),
    [sym_integer] = ACTIONS(1418),
    [aux_sym_return_statement_token1] = ACTIONS(1418),
    [aux_sym_throw_expression_token1] = ACTIONS(1418),
    [aux_sym_while_statement_token1] = ACTIONS(1418),
    [aux_sym_while_statement_token2] = ACTIONS(1418),
    [aux_sym_do_statement_token1] = ACTIONS(1418),
    [aux_sym_for_statement_token1] = ACTIONS(1418),
    [aux_sym_for_statement_token2] = ACTIONS(1418),
    [aux_sym_foreach_statement_token1] = ACTIONS(1418),
    [aux_sym_foreach_statement_token2] = ACTIONS(1418),
    [aux_sym_if_statement_token1] = ACTIONS(1418),
    [aux_sym_if_statement_token2] = ACTIONS(1418),
    [aux_sym_else_if_clause_token1] = ACTIONS(1418),
    [aux_sym_else_clause_token1] = ACTIONS(1418),
    [aux_sym_match_expression_token1] = ACTIONS(1418),
    [aux_sym_match_default_expression_token1] = ACTIONS(1418),
    [aux_sym_switch_statement_token1] = ACTIONS(1418),
    [aux_sym_switch_block_token1] = ACTIONS(1418),
    [anon_sym_PLUS] = ACTIONS(1418),
    [anon_sym_DASH] = ACTIONS(1418),
    [anon_sym_TILDE] = ACTIONS(1416),
    [anon_sym_BANG] = ACTIONS(1416),
    [anon_sym_AT] = ACTIONS(1416),
    [aux_sym_clone_expression_token1] = ACTIONS(1418),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1418),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1418),
    [anon_sym_DASH_DASH] = ACTIONS(1416),
    [anon_sym_PLUS_PLUS] = ACTIONS(1416),
    [aux_sym__list_destructing_token1] = ACTIONS(1418),
    [anon_sym_LBRACK] = ACTIONS(1416),
    [anon_sym_self] = ACTIONS(1418),
    [anon_sym_parent] = ACTIONS(1418),
    [aux_sym__argument_name_token1] = ACTIONS(1418),
    [aux_sym__argument_name_token2] = ACTIONS(1418),
    [anon_sym_POUND_LBRACK] = ACTIONS(1416),
    [aux_sym_encapsed_string_token1] = ACTIONS(1416),
    [anon_sym_DQUOTE] = ACTIONS(1416),
    [aux_sym_string_token1] = ACTIONS(1416),
    [anon_sym_SQUOTE] = ACTIONS(1416),
    [anon_sym_LT_LT_LT] = ACTIONS(1416),
    [anon_sym_BQUOTE] = ACTIONS(1416),
    [anon_sym_DOLLAR] = ACTIONS(1416),
    [aux_sym_yield_expression_token1] = ACTIONS(1418),
    [aux_sym_include_expression_token1] = ACTIONS(1418),
    [aux_sym_include_once_expression_token1] = ACTIONS(1418),
    [aux_sym_require_expression_token1] = ACTIONS(1418),
    [aux_sym_require_once_expression_token1] = ACTIONS(1418),
    [sym_comment] = ACTIONS(5),
  },
  [550] = {
    [sym_text_interpolation] = STATE(550),
    [ts_builtin_sym_end] = ACTIONS(1420),
    [sym_name] = ACTIONS(1422),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1420),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1422),
    [aux_sym_global_declaration_token1] = ACTIONS(1422),
    [aux_sym_namespace_definition_token1] = ACTIONS(1422),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1422),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1422),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1422),
    [anon_sym_BSLASH] = ACTIONS(1420),
    [anon_sym_LBRACE] = ACTIONS(1420),
    [anon_sym_RBRACE] = ACTIONS(1420),
    [aux_sym_trait_declaration_token1] = ACTIONS(1422),
    [aux_sym_interface_declaration_token1] = ACTIONS(1422),
    [aux_sym_enum_declaration_token1] = ACTIONS(1422),
    [aux_sym_enum_case_token1] = ACTIONS(1422),
    [aux_sym_class_declaration_token1] = ACTIONS(1422),
    [aux_sym_final_modifier_token1] = ACTIONS(1422),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1422),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1422),
    [sym_var_modifier] = ACTIONS(1422),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1422),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1422),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1422),
    [anon_sym_LPAREN2] = ACTIONS(1420),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1422),
    [aux_sym_cast_type_token1] = ACTIONS(1422),
    [aux_sym_echo_statement_token1] = ACTIONS(1422),
    [aux_sym_exit_statement_token1] = ACTIONS(1422),
    [anon_sym_unset] = ACTIONS(1422),
    [aux_sym_declare_statement_token1] = ACTIONS(1422),
    [aux_sym_declare_statement_token2] = ACTIONS(1422),
    [sym_float] = ACTIONS(1422),
    [aux_sym_try_statement_token1] = ACTIONS(1422),
    [aux_sym_goto_statement_token1] = ACTIONS(1422),
    [aux_sym_continue_statement_token1] = ACTIONS(1422),
    [aux_sym_break_statement_token1] = ACTIONS(1422),
    [sym_integer] = ACTIONS(1422),
    [aux_sym_return_statement_token1] = ACTIONS(1422),
    [aux_sym_throw_expression_token1] = ACTIONS(1422),
    [aux_sym_while_statement_token1] = ACTIONS(1422),
    [aux_sym_while_statement_token2] = ACTIONS(1422),
    [aux_sym_do_statement_token1] = ACTIONS(1422),
    [aux_sym_for_statement_token1] = ACTIONS(1422),
    [aux_sym_for_statement_token2] = ACTIONS(1422),
    [aux_sym_foreach_statement_token1] = ACTIONS(1422),
    [aux_sym_foreach_statement_token2] = ACTIONS(1422),
    [aux_sym_if_statement_token1] = ACTIONS(1422),
    [aux_sym_if_statement_token2] = ACTIONS(1422),
    [aux_sym_else_if_clause_token1] = ACTIONS(1422),
    [aux_sym_else_clause_token1] = ACTIONS(1422),
    [aux_sym_match_expression_token1] = ACTIONS(1422),
    [aux_sym_match_default_expression_token1] = ACTIONS(1422),
    [aux_sym_switch_statement_token1] = ACTIONS(1422),
    [aux_sym_switch_block_token1] = ACTIONS(1422),
    [anon_sym_PLUS] = ACTIONS(1422),
    [anon_sym_DASH] = ACTIONS(1422),
    [anon_sym_TILDE] = ACTIONS(1420),
    [anon_sym_BANG] = ACTIONS(1420),
    [anon_sym_AT] = ACTIONS(1420),
    [aux_sym_clone_expression_token1] = ACTIONS(1422),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1422),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1422),
    [anon_sym_DASH_DASH] = ACTIONS(1420),
    [anon_sym_PLUS_PLUS] = ACTIONS(1420),
    [aux_sym__list_destructing_token1] = ACTIONS(1422),
    [anon_sym_LBRACK] = ACTIONS(1420),
    [anon_sym_self] = ACTIONS(1422),
    [anon_sym_parent] = ACTIONS(1422),
    [aux_sym__argument_name_token1] = ACTIONS(1422),
    [aux_sym__argument_name_token2] = ACTIONS(1422),
    [anon_sym_POUND_LBRACK] = ACTIONS(1420),
    [aux_sym_encapsed_string_token1] = ACTIONS(1420),
    [anon_sym_DQUOTE] = ACTIONS(1420),
    [aux_sym_string_token1] = ACTIONS(1420),
    [anon_sym_SQUOTE] = ACTIONS(1420),
    [anon_sym_LT_LT_LT] = ACTIONS(1420),
    [anon_sym_BQUOTE] = ACTIONS(1420),
    [anon_sym_DOLLAR] = ACTIONS(1420),
    [aux_sym_yield_expression_token1] = ACTIONS(1422),
    [aux_sym_include_expression_token1] = ACTIONS(1422),
    [aux_sym_include_once_expression_token1] = ACTIONS(1422),
    [aux_sym_require_expression_token1] = ACTIONS(1422),
    [aux_sym_require_once_expression_token1] = ACTIONS(1422),
    [sym_comment] = ACTIONS(5),
  },
  [551] = {
    [sym_text_interpolation] = STATE(551),
    [ts_builtin_sym_end] = ACTIONS(1424),
    [sym_name] = ACTIONS(1426),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1424),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1426),
    [aux_sym_global_declaration_token1] = ACTIONS(1426),
    [aux_sym_namespace_definition_token1] = ACTIONS(1426),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1426),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1426),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1426),
    [anon_sym_BSLASH] = ACTIONS(1424),
    [anon_sym_LBRACE] = ACTIONS(1424),
    [anon_sym_RBRACE] = ACTIONS(1424),
    [aux_sym_trait_declaration_token1] = ACTIONS(1426),
    [aux_sym_interface_declaration_token1] = ACTIONS(1426),
    [aux_sym_enum_declaration_token1] = ACTIONS(1426),
    [aux_sym_enum_case_token1] = ACTIONS(1426),
    [aux_sym_class_declaration_token1] = ACTIONS(1426),
    [aux_sym_final_modifier_token1] = ACTIONS(1426),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1426),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1426),
    [sym_var_modifier] = ACTIONS(1426),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1426),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1426),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1426),
    [anon_sym_LPAREN2] = ACTIONS(1424),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1426),
    [aux_sym_cast_type_token1] = ACTIONS(1426),
    [aux_sym_echo_statement_token1] = ACTIONS(1426),
    [aux_sym_exit_statement_token1] = ACTIONS(1426),
    [anon_sym_unset] = ACTIONS(1426),
    [aux_sym_declare_statement_token1] = ACTIONS(1426),
    [aux_sym_declare_statement_token2] = ACTIONS(1426),
    [sym_float] = ACTIONS(1426),
    [aux_sym_try_statement_token1] = ACTIONS(1426),
    [aux_sym_goto_statement_token1] = ACTIONS(1426),
    [aux_sym_continue_statement_token1] = ACTIONS(1426),
    [aux_sym_break_statement_token1] = ACTIONS(1426),
    [sym_integer] = ACTIONS(1426),
    [aux_sym_return_statement_token1] = ACTIONS(1426),
    [aux_sym_throw_expression_token1] = ACTIONS(1426),
    [aux_sym_while_statement_token1] = ACTIONS(1426),
    [aux_sym_while_statement_token2] = ACTIONS(1426),
    [aux_sym_do_statement_token1] = ACTIONS(1426),
    [aux_sym_for_statement_token1] = ACTIONS(1426),
    [aux_sym_for_statement_token2] = ACTIONS(1426),
    [aux_sym_foreach_statement_token1] = ACTIONS(1426),
    [aux_sym_foreach_statement_token2] = ACTIONS(1426),
    [aux_sym_if_statement_token1] = ACTIONS(1426),
    [aux_sym_if_statement_token2] = ACTIONS(1426),
    [aux_sym_else_if_clause_token1] = ACTIONS(1426),
    [aux_sym_else_clause_token1] = ACTIONS(1426),
    [aux_sym_match_expression_token1] = ACTIONS(1426),
    [aux_sym_match_default_expression_token1] = ACTIONS(1426),
    [aux_sym_switch_statement_token1] = ACTIONS(1426),
    [aux_sym_switch_block_token1] = ACTIONS(1426),
    [anon_sym_PLUS] = ACTIONS(1426),
    [anon_sym_DASH] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1424),
    [anon_sym_BANG] = ACTIONS(1424),
    [anon_sym_AT] = ACTIONS(1424),
    [aux_sym_clone_expression_token1] = ACTIONS(1426),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1426),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1426),
    [anon_sym_DASH_DASH] = ACTIONS(1424),
    [anon_sym_PLUS_PLUS] = ACTIONS(1424),
    [aux_sym__list_destructing_token1] = ACTIONS(1426),
    [anon_sym_LBRACK] = ACTIONS(1424),
    [anon_sym_self] = ACTIONS(1426),
    [anon_sym_parent] = ACTIONS(1426),
    [aux_sym__argument_name_token1] = ACTIONS(1426),
    [aux_sym__argument_name_token2] = ACTIONS(1426),
    [anon_sym_POUND_LBRACK] = ACTIONS(1424),
    [aux_sym_encapsed_string_token1] = ACTIONS(1424),
    [anon_sym_DQUOTE] = ACTIONS(1424),
    [aux_sym_string_token1] = ACTIONS(1424),
    [anon_sym_SQUOTE] = ACTIONS(1424),
    [anon_sym_LT_LT_LT] = ACTIONS(1424),
    [anon_sym_BQUOTE] = ACTIONS(1424),
    [anon_sym_DOLLAR] = ACTIONS(1424),
    [aux_sym_yield_expression_token1] = ACTIONS(1426),
    [aux_sym_include_expression_token1] = ACTIONS(1426),
    [aux_sym_include_once_expression_token1] = ACTIONS(1426),
    [aux_sym_require_expression_token1] = ACTIONS(1426),
    [aux_sym_require_once_expression_token1] = ACTIONS(1426),
    [sym_comment] = ACTIONS(5),
  },
  [552] = {
    [sym_text_interpolation] = STATE(552),
    [ts_builtin_sym_end] = ACTIONS(1428),
    [sym_name] = ACTIONS(1430),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1428),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1430),
    [aux_sym_global_declaration_token1] = ACTIONS(1430),
    [aux_sym_namespace_definition_token1] = ACTIONS(1430),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1430),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1430),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1430),
    [anon_sym_BSLASH] = ACTIONS(1428),
    [anon_sym_LBRACE] = ACTIONS(1428),
    [anon_sym_RBRACE] = ACTIONS(1428),
    [aux_sym_trait_declaration_token1] = ACTIONS(1430),
    [aux_sym_interface_declaration_token1] = ACTIONS(1430),
    [aux_sym_enum_declaration_token1] = ACTIONS(1430),
    [aux_sym_enum_case_token1] = ACTIONS(1430),
    [aux_sym_class_declaration_token1] = ACTIONS(1430),
    [aux_sym_final_modifier_token1] = ACTIONS(1430),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1430),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1430),
    [sym_var_modifier] = ACTIONS(1430),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1430),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1430),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1430),
    [anon_sym_LPAREN2] = ACTIONS(1428),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1430),
    [aux_sym_cast_type_token1] = ACTIONS(1430),
    [aux_sym_echo_statement_token1] = ACTIONS(1430),
    [aux_sym_exit_statement_token1] = ACTIONS(1430),
    [anon_sym_unset] = ACTIONS(1430),
    [aux_sym_declare_statement_token1] = ACTIONS(1430),
    [aux_sym_declare_statement_token2] = ACTIONS(1430),
    [sym_float] = ACTIONS(1430),
    [aux_sym_try_statement_token1] = ACTIONS(1430),
    [aux_sym_goto_statement_token1] = ACTIONS(1430),
    [aux_sym_continue_statement_token1] = ACTIONS(1430),
    [aux_sym_break_statement_token1] = ACTIONS(1430),
    [sym_integer] = ACTIONS(1430),
    [aux_sym_return_statement_token1] = ACTIONS(1430),
    [aux_sym_throw_expression_token1] = ACTIONS(1430),
    [aux_sym_while_statement_token1] = ACTIONS(1430),
    [aux_sym_while_statement_token2] = ACTIONS(1430),
    [aux_sym_do_statement_token1] = ACTIONS(1430),
    [aux_sym_for_statement_token1] = ACTIONS(1430),
    [aux_sym_for_statement_token2] = ACTIONS(1430),
    [aux_sym_foreach_statement_token1] = ACTIONS(1430),
    [aux_sym_foreach_statement_token2] = ACTIONS(1430),
    [aux_sym_if_statement_token1] = ACTIONS(1430),
    [aux_sym_if_statement_token2] = ACTIONS(1430),
    [aux_sym_else_if_clause_token1] = ACTIONS(1430),
    [aux_sym_else_clause_token1] = ACTIONS(1430),
    [aux_sym_match_expression_token1] = ACTIONS(1430),
    [aux_sym_match_default_expression_token1] = ACTIONS(1430),
    [aux_sym_switch_statement_token1] = ACTIONS(1430),
    [aux_sym_switch_block_token1] = ACTIONS(1430),
    [anon_sym_PLUS] = ACTIONS(1430),
    [anon_sym_DASH] = ACTIONS(1430),
    [anon_sym_TILDE] = ACTIONS(1428),
    [anon_sym_BANG] = ACTIONS(1428),
    [anon_sym_AT] = ACTIONS(1428),
    [aux_sym_clone_expression_token1] = ACTIONS(1430),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1430),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1430),
    [anon_sym_DASH_DASH] = ACTIONS(1428),
    [anon_sym_PLUS_PLUS] = ACTIONS(1428),
    [aux_sym__list_destructing_token1] = ACTIONS(1430),
    [anon_sym_LBRACK] = ACTIONS(1428),
    [anon_sym_self] = ACTIONS(1430),
    [anon_sym_parent] = ACTIONS(1430),
    [aux_sym__argument_name_token1] = ACTIONS(1430),
    [aux_sym__argument_name_token2] = ACTIONS(1430),
    [anon_sym_POUND_LBRACK] = ACTIONS(1428),
    [aux_sym_encapsed_string_token1] = ACTIONS(1428),
    [anon_sym_DQUOTE] = ACTIONS(1428),
    [aux_sym_string_token1] = ACTIONS(1428),
    [anon_sym_SQUOTE] = ACTIONS(1428),
    [anon_sym_LT_LT_LT] = ACTIONS(1428),
    [anon_sym_BQUOTE] = ACTIONS(1428),
    [anon_sym_DOLLAR] = ACTIONS(1428),
    [aux_sym_yield_expression_token1] = ACTIONS(1430),
    [aux_sym_include_expression_token1] = ACTIONS(1430),
    [aux_sym_include_once_expression_token1] = ACTIONS(1430),
    [aux_sym_require_expression_token1] = ACTIONS(1430),
    [aux_sym_require_once_expression_token1] = ACTIONS(1430),
    [sym_comment] = ACTIONS(5),
  },
  [553] = {
    [sym_text_interpolation] = STATE(553),
    [ts_builtin_sym_end] = ACTIONS(1432),
    [sym_name] = ACTIONS(1434),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1432),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1434),
    [aux_sym_global_declaration_token1] = ACTIONS(1434),
    [aux_sym_namespace_definition_token1] = ACTIONS(1434),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1434),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1434),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1434),
    [anon_sym_BSLASH] = ACTIONS(1432),
    [anon_sym_LBRACE] = ACTIONS(1432),
    [anon_sym_RBRACE] = ACTIONS(1432),
    [aux_sym_trait_declaration_token1] = ACTIONS(1434),
    [aux_sym_interface_declaration_token1] = ACTIONS(1434),
    [aux_sym_enum_declaration_token1] = ACTIONS(1434),
    [aux_sym_enum_case_token1] = ACTIONS(1434),
    [aux_sym_class_declaration_token1] = ACTIONS(1434),
    [aux_sym_final_modifier_token1] = ACTIONS(1434),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1434),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1434),
    [sym_var_modifier] = ACTIONS(1434),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1434),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1434),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1434),
    [anon_sym_LPAREN2] = ACTIONS(1432),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1434),
    [aux_sym_cast_type_token1] = ACTIONS(1434),
    [aux_sym_echo_statement_token1] = ACTIONS(1434),
    [aux_sym_exit_statement_token1] = ACTIONS(1434),
    [anon_sym_unset] = ACTIONS(1434),
    [aux_sym_declare_statement_token1] = ACTIONS(1434),
    [aux_sym_declare_statement_token2] = ACTIONS(1434),
    [sym_float] = ACTIONS(1434),
    [aux_sym_try_statement_token1] = ACTIONS(1434),
    [aux_sym_goto_statement_token1] = ACTIONS(1434),
    [aux_sym_continue_statement_token1] = ACTIONS(1434),
    [aux_sym_break_statement_token1] = ACTIONS(1434),
    [sym_integer] = ACTIONS(1434),
    [aux_sym_return_statement_token1] = ACTIONS(1434),
    [aux_sym_throw_expression_token1] = ACTIONS(1434),
    [aux_sym_while_statement_token1] = ACTIONS(1434),
    [aux_sym_while_statement_token2] = ACTIONS(1434),
    [aux_sym_do_statement_token1] = ACTIONS(1434),
    [aux_sym_for_statement_token1] = ACTIONS(1434),
    [aux_sym_for_statement_token2] = ACTIONS(1434),
    [aux_sym_foreach_statement_token1] = ACTIONS(1434),
    [aux_sym_foreach_statement_token2] = ACTIONS(1434),
    [aux_sym_if_statement_token1] = ACTIONS(1434),
    [aux_sym_if_statement_token2] = ACTIONS(1434),
    [aux_sym_else_if_clause_token1] = ACTIONS(1434),
    [aux_sym_else_clause_token1] = ACTIONS(1434),
    [aux_sym_match_expression_token1] = ACTIONS(1434),
    [aux_sym_match_default_expression_token1] = ACTIONS(1434),
    [aux_sym_switch_statement_token1] = ACTIONS(1434),
    [aux_sym_switch_block_token1] = ACTIONS(1434),
    [anon_sym_PLUS] = ACTIONS(1434),
    [anon_sym_DASH] = ACTIONS(1434),
    [anon_sym_TILDE] = ACTIONS(1432),
    [anon_sym_BANG] = ACTIONS(1432),
    [anon_sym_AT] = ACTIONS(1432),
    [aux_sym_clone_expression_token1] = ACTIONS(1434),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1434),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1434),
    [anon_sym_DASH_DASH] = ACTIONS(1432),
    [anon_sym_PLUS_PLUS] = ACTIONS(1432),
    [aux_sym__list_destructing_token1] = ACTIONS(1434),
    [anon_sym_LBRACK] = ACTIONS(1432),
    [anon_sym_self] = ACTIONS(1434),
    [anon_sym_parent] = ACTIONS(1434),
    [aux_sym__argument_name_token1] = ACTIONS(1434),
    [aux_sym__argument_name_token2] = ACTIONS(1434),
    [anon_sym_POUND_LBRACK] = ACTIONS(1432),
    [aux_sym_encapsed_string_token1] = ACTIONS(1432),
    [anon_sym_DQUOTE] = ACTIONS(1432),
    [aux_sym_string_token1] = ACTIONS(1432),
    [anon_sym_SQUOTE] = ACTIONS(1432),
    [anon_sym_LT_LT_LT] = ACTIONS(1432),
    [anon_sym_BQUOTE] = ACTIONS(1432),
    [anon_sym_DOLLAR] = ACTIONS(1432),
    [aux_sym_yield_expression_token1] = ACTIONS(1434),
    [aux_sym_include_expression_token1] = ACTIONS(1434),
    [aux_sym_include_once_expression_token1] = ACTIONS(1434),
    [aux_sym_require_expression_token1] = ACTIONS(1434),
    [aux_sym_require_once_expression_token1] = ACTIONS(1434),
    [sym_comment] = ACTIONS(5),
  },
  [554] = {
    [sym_text_interpolation] = STATE(554),
    [ts_builtin_sym_end] = ACTIONS(1436),
    [sym_name] = ACTIONS(1438),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1436),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1438),
    [aux_sym_global_declaration_token1] = ACTIONS(1438),
    [aux_sym_namespace_definition_token1] = ACTIONS(1438),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1438),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1438),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1438),
    [anon_sym_BSLASH] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1436),
    [anon_sym_RBRACE] = ACTIONS(1436),
    [aux_sym_trait_declaration_token1] = ACTIONS(1438),
    [aux_sym_interface_declaration_token1] = ACTIONS(1438),
    [aux_sym_enum_declaration_token1] = ACTIONS(1438),
    [aux_sym_enum_case_token1] = ACTIONS(1438),
    [aux_sym_class_declaration_token1] = ACTIONS(1438),
    [aux_sym_final_modifier_token1] = ACTIONS(1438),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1438),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1438),
    [sym_var_modifier] = ACTIONS(1438),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1438),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1438),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1438),
    [anon_sym_LPAREN2] = ACTIONS(1436),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1438),
    [aux_sym_cast_type_token1] = ACTIONS(1438),
    [aux_sym_echo_statement_token1] = ACTIONS(1438),
    [aux_sym_exit_statement_token1] = ACTIONS(1438),
    [anon_sym_unset] = ACTIONS(1438),
    [aux_sym_declare_statement_token1] = ACTIONS(1438),
    [aux_sym_declare_statement_token2] = ACTIONS(1438),
    [sym_float] = ACTIONS(1438),
    [aux_sym_try_statement_token1] = ACTIONS(1438),
    [aux_sym_goto_statement_token1] = ACTIONS(1438),
    [aux_sym_continue_statement_token1] = ACTIONS(1438),
    [aux_sym_break_statement_token1] = ACTIONS(1438),
    [sym_integer] = ACTIONS(1438),
    [aux_sym_return_statement_token1] = ACTIONS(1438),
    [aux_sym_throw_expression_token1] = ACTIONS(1438),
    [aux_sym_while_statement_token1] = ACTIONS(1438),
    [aux_sym_while_statement_token2] = ACTIONS(1438),
    [aux_sym_do_statement_token1] = ACTIONS(1438),
    [aux_sym_for_statement_token1] = ACTIONS(1438),
    [aux_sym_for_statement_token2] = ACTIONS(1438),
    [aux_sym_foreach_statement_token1] = ACTIONS(1438),
    [aux_sym_foreach_statement_token2] = ACTIONS(1438),
    [aux_sym_if_statement_token1] = ACTIONS(1438),
    [aux_sym_if_statement_token2] = ACTIONS(1438),
    [aux_sym_else_if_clause_token1] = ACTIONS(1438),
    [aux_sym_else_clause_token1] = ACTIONS(1438),
    [aux_sym_match_expression_token1] = ACTIONS(1438),
    [aux_sym_match_default_expression_token1] = ACTIONS(1438),
    [aux_sym_switch_statement_token1] = ACTIONS(1438),
    [aux_sym_switch_block_token1] = ACTIONS(1438),
    [anon_sym_PLUS] = ACTIONS(1438),
    [anon_sym_DASH] = ACTIONS(1438),
    [anon_sym_TILDE] = ACTIONS(1436),
    [anon_sym_BANG] = ACTIONS(1436),
    [anon_sym_AT] = ACTIONS(1436),
    [aux_sym_clone_expression_token1] = ACTIONS(1438),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1438),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1438),
    [anon_sym_DASH_DASH] = ACTIONS(1436),
    [anon_sym_PLUS_PLUS] = ACTIONS(1436),
    [aux_sym__list_destructing_token1] = ACTIONS(1438),
    [anon_sym_LBRACK] = ACTIONS(1436),
    [anon_sym_self] = ACTIONS(1438),
    [anon_sym_parent] = ACTIONS(1438),
    [aux_sym__argument_name_token1] = ACTIONS(1438),
    [aux_sym__argument_name_token2] = ACTIONS(1438),
    [anon_sym_POUND_LBRACK] = ACTIONS(1436),
    [aux_sym_encapsed_string_token1] = ACTIONS(1436),
    [anon_sym_DQUOTE] = ACTIONS(1436),
    [aux_sym_string_token1] = ACTIONS(1436),
    [anon_sym_SQUOTE] = ACTIONS(1436),
    [anon_sym_LT_LT_LT] = ACTIONS(1436),
    [anon_sym_BQUOTE] = ACTIONS(1436),
    [anon_sym_DOLLAR] = ACTIONS(1436),
    [aux_sym_yield_expression_token1] = ACTIONS(1438),
    [aux_sym_include_expression_token1] = ACTIONS(1438),
    [aux_sym_include_once_expression_token1] = ACTIONS(1438),
    [aux_sym_require_expression_token1] = ACTIONS(1438),
    [aux_sym_require_once_expression_token1] = ACTIONS(1438),
    [sym_comment] = ACTIONS(5),
  },
  [555] = {
    [sym_text_interpolation] = STATE(555),
    [ts_builtin_sym_end] = ACTIONS(1440),
    [sym_name] = ACTIONS(1442),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1440),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1442),
    [aux_sym_global_declaration_token1] = ACTIONS(1442),
    [aux_sym_namespace_definition_token1] = ACTIONS(1442),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1442),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1442),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1442),
    [anon_sym_BSLASH] = ACTIONS(1440),
    [anon_sym_LBRACE] = ACTIONS(1440),
    [anon_sym_RBRACE] = ACTIONS(1440),
    [aux_sym_trait_declaration_token1] = ACTIONS(1442),
    [aux_sym_interface_declaration_token1] = ACTIONS(1442),
    [aux_sym_enum_declaration_token1] = ACTIONS(1442),
    [aux_sym_enum_case_token1] = ACTIONS(1442),
    [aux_sym_class_declaration_token1] = ACTIONS(1442),
    [aux_sym_final_modifier_token1] = ACTIONS(1442),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1442),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1442),
    [sym_var_modifier] = ACTIONS(1442),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1442),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1442),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1442),
    [anon_sym_LPAREN2] = ACTIONS(1440),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1442),
    [aux_sym_cast_type_token1] = ACTIONS(1442),
    [aux_sym_echo_statement_token1] = ACTIONS(1442),
    [aux_sym_exit_statement_token1] = ACTIONS(1442),
    [anon_sym_unset] = ACTIONS(1442),
    [aux_sym_declare_statement_token1] = ACTIONS(1442),
    [aux_sym_declare_statement_token2] = ACTIONS(1442),
    [sym_float] = ACTIONS(1442),
    [aux_sym_try_statement_token1] = ACTIONS(1442),
    [aux_sym_goto_statement_token1] = ACTIONS(1442),
    [aux_sym_continue_statement_token1] = ACTIONS(1442),
    [aux_sym_break_statement_token1] = ACTIONS(1442),
    [sym_integer] = ACTIONS(1442),
    [aux_sym_return_statement_token1] = ACTIONS(1442),
    [aux_sym_throw_expression_token1] = ACTIONS(1442),
    [aux_sym_while_statement_token1] = ACTIONS(1442),
    [aux_sym_while_statement_token2] = ACTIONS(1442),
    [aux_sym_do_statement_token1] = ACTIONS(1442),
    [aux_sym_for_statement_token1] = ACTIONS(1442),
    [aux_sym_for_statement_token2] = ACTIONS(1442),
    [aux_sym_foreach_statement_token1] = ACTIONS(1442),
    [aux_sym_foreach_statement_token2] = ACTIONS(1442),
    [aux_sym_if_statement_token1] = ACTIONS(1442),
    [aux_sym_if_statement_token2] = ACTIONS(1442),
    [aux_sym_else_if_clause_token1] = ACTIONS(1442),
    [aux_sym_else_clause_token1] = ACTIONS(1442),
    [aux_sym_match_expression_token1] = ACTIONS(1442),
    [aux_sym_match_default_expression_token1] = ACTIONS(1442),
    [aux_sym_switch_statement_token1] = ACTIONS(1442),
    [aux_sym_switch_block_token1] = ACTIONS(1442),
    [anon_sym_PLUS] = ACTIONS(1442),
    [anon_sym_DASH] = ACTIONS(1442),
    [anon_sym_TILDE] = ACTIONS(1440),
    [anon_sym_BANG] = ACTIONS(1440),
    [anon_sym_AT] = ACTIONS(1440),
    [aux_sym_clone_expression_token1] = ACTIONS(1442),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1442),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1442),
    [anon_sym_DASH_DASH] = ACTIONS(1440),
    [anon_sym_PLUS_PLUS] = ACTIONS(1440),
    [aux_sym__list_destructing_token1] = ACTIONS(1442),
    [anon_sym_LBRACK] = ACTIONS(1440),
    [anon_sym_self] = ACTIONS(1442),
    [anon_sym_parent] = ACTIONS(1442),
    [aux_sym__argument_name_token1] = ACTIONS(1442),
    [aux_sym__argument_name_token2] = ACTIONS(1442),
    [anon_sym_POUND_LBRACK] = ACTIONS(1440),
    [aux_sym_encapsed_string_token1] = ACTIONS(1440),
    [anon_sym_DQUOTE] = ACTIONS(1440),
    [aux_sym_string_token1] = ACTIONS(1440),
    [anon_sym_SQUOTE] = ACTIONS(1440),
    [anon_sym_LT_LT_LT] = ACTIONS(1440),
    [anon_sym_BQUOTE] = ACTIONS(1440),
    [anon_sym_DOLLAR] = ACTIONS(1440),
    [aux_sym_yield_expression_token1] = ACTIONS(1442),
    [aux_sym_include_expression_token1] = ACTIONS(1442),
    [aux_sym_include_once_expression_token1] = ACTIONS(1442),
    [aux_sym_require_expression_token1] = ACTIONS(1442),
    [aux_sym_require_once_expression_token1] = ACTIONS(1442),
    [sym_comment] = ACTIONS(5),
  },
  [556] = {
    [sym_text_interpolation] = STATE(556),
    [ts_builtin_sym_end] = ACTIONS(1444),
    [sym_name] = ACTIONS(1446),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1444),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1446),
    [aux_sym_global_declaration_token1] = ACTIONS(1446),
    [aux_sym_namespace_definition_token1] = ACTIONS(1446),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1446),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1446),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1446),
    [anon_sym_BSLASH] = ACTIONS(1444),
    [anon_sym_LBRACE] = ACTIONS(1444),
    [anon_sym_RBRACE] = ACTIONS(1444),
    [aux_sym_trait_declaration_token1] = ACTIONS(1446),
    [aux_sym_interface_declaration_token1] = ACTIONS(1446),
    [aux_sym_enum_declaration_token1] = ACTIONS(1446),
    [aux_sym_enum_case_token1] = ACTIONS(1446),
    [aux_sym_class_declaration_token1] = ACTIONS(1446),
    [aux_sym_final_modifier_token1] = ACTIONS(1446),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1446),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1446),
    [sym_var_modifier] = ACTIONS(1446),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1446),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1446),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1446),
    [anon_sym_LPAREN2] = ACTIONS(1444),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1446),
    [aux_sym_cast_type_token1] = ACTIONS(1446),
    [aux_sym_echo_statement_token1] = ACTIONS(1446),
    [aux_sym_exit_statement_token1] = ACTIONS(1446),
    [anon_sym_unset] = ACTIONS(1446),
    [aux_sym_declare_statement_token1] = ACTIONS(1446),
    [aux_sym_declare_statement_token2] = ACTIONS(1446),
    [sym_float] = ACTIONS(1446),
    [aux_sym_try_statement_token1] = ACTIONS(1446),
    [aux_sym_goto_statement_token1] = ACTIONS(1446),
    [aux_sym_continue_statement_token1] = ACTIONS(1446),
    [aux_sym_break_statement_token1] = ACTIONS(1446),
    [sym_integer] = ACTIONS(1446),
    [aux_sym_return_statement_token1] = ACTIONS(1446),
    [aux_sym_throw_expression_token1] = ACTIONS(1446),
    [aux_sym_while_statement_token1] = ACTIONS(1446),
    [aux_sym_while_statement_token2] = ACTIONS(1446),
    [aux_sym_do_statement_token1] = ACTIONS(1446),
    [aux_sym_for_statement_token1] = ACTIONS(1446),
    [aux_sym_for_statement_token2] = ACTIONS(1446),
    [aux_sym_foreach_statement_token1] = ACTIONS(1446),
    [aux_sym_foreach_statement_token2] = ACTIONS(1446),
    [aux_sym_if_statement_token1] = ACTIONS(1446),
    [aux_sym_if_statement_token2] = ACTIONS(1446),
    [aux_sym_else_if_clause_token1] = ACTIONS(1446),
    [aux_sym_else_clause_token1] = ACTIONS(1446),
    [aux_sym_match_expression_token1] = ACTIONS(1446),
    [aux_sym_match_default_expression_token1] = ACTIONS(1446),
    [aux_sym_switch_statement_token1] = ACTIONS(1446),
    [aux_sym_switch_block_token1] = ACTIONS(1446),
    [anon_sym_PLUS] = ACTIONS(1446),
    [anon_sym_DASH] = ACTIONS(1446),
    [anon_sym_TILDE] = ACTIONS(1444),
    [anon_sym_BANG] = ACTIONS(1444),
    [anon_sym_AT] = ACTIONS(1444),
    [aux_sym_clone_expression_token1] = ACTIONS(1446),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1446),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1446),
    [anon_sym_DASH_DASH] = ACTIONS(1444),
    [anon_sym_PLUS_PLUS] = ACTIONS(1444),
    [aux_sym__list_destructing_token1] = ACTIONS(1446),
    [anon_sym_LBRACK] = ACTIONS(1444),
    [anon_sym_self] = ACTIONS(1446),
    [anon_sym_parent] = ACTIONS(1446),
    [aux_sym__argument_name_token1] = ACTIONS(1446),
    [aux_sym__argument_name_token2] = ACTIONS(1446),
    [anon_sym_POUND_LBRACK] = ACTIONS(1444),
    [aux_sym_encapsed_string_token1] = ACTIONS(1444),
    [anon_sym_DQUOTE] = ACTIONS(1444),
    [aux_sym_string_token1] = ACTIONS(1444),
    [anon_sym_SQUOTE] = ACTIONS(1444),
    [anon_sym_LT_LT_LT] = ACTIONS(1444),
    [anon_sym_BQUOTE] = ACTIONS(1444),
    [anon_sym_DOLLAR] = ACTIONS(1444),
    [aux_sym_yield_expression_token1] = ACTIONS(1446),
    [aux_sym_include_expression_token1] = ACTIONS(1446),
    [aux_sym_include_once_expression_token1] = ACTIONS(1446),
    [aux_sym_require_expression_token1] = ACTIONS(1446),
    [aux_sym_require_once_expression_token1] = ACTIONS(1446),
    [sym_comment] = ACTIONS(5),
  },
  [557] = {
    [sym_text_interpolation] = STATE(557),
    [ts_builtin_sym_end] = ACTIONS(1448),
    [sym_name] = ACTIONS(1450),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1448),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1450),
    [aux_sym_global_declaration_token1] = ACTIONS(1450),
    [aux_sym_namespace_definition_token1] = ACTIONS(1450),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1450),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1450),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1450),
    [anon_sym_BSLASH] = ACTIONS(1448),
    [anon_sym_LBRACE] = ACTIONS(1448),
    [anon_sym_RBRACE] = ACTIONS(1448),
    [aux_sym_trait_declaration_token1] = ACTIONS(1450),
    [aux_sym_interface_declaration_token1] = ACTIONS(1450),
    [aux_sym_enum_declaration_token1] = ACTIONS(1450),
    [aux_sym_enum_case_token1] = ACTIONS(1450),
    [aux_sym_class_declaration_token1] = ACTIONS(1450),
    [aux_sym_final_modifier_token1] = ACTIONS(1450),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1450),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1450),
    [sym_var_modifier] = ACTIONS(1450),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1450),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1450),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1450),
    [anon_sym_LPAREN2] = ACTIONS(1448),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1450),
    [aux_sym_cast_type_token1] = ACTIONS(1450),
    [aux_sym_echo_statement_token1] = ACTIONS(1450),
    [aux_sym_exit_statement_token1] = ACTIONS(1450),
    [anon_sym_unset] = ACTIONS(1450),
    [aux_sym_declare_statement_token1] = ACTIONS(1450),
    [aux_sym_declare_statement_token2] = ACTIONS(1450),
    [sym_float] = ACTIONS(1450),
    [aux_sym_try_statement_token1] = ACTIONS(1450),
    [aux_sym_goto_statement_token1] = ACTIONS(1450),
    [aux_sym_continue_statement_token1] = ACTIONS(1450),
    [aux_sym_break_statement_token1] = ACTIONS(1450),
    [sym_integer] = ACTIONS(1450),
    [aux_sym_return_statement_token1] = ACTIONS(1450),
    [aux_sym_throw_expression_token1] = ACTIONS(1450),
    [aux_sym_while_statement_token1] = ACTIONS(1450),
    [aux_sym_while_statement_token2] = ACTIONS(1450),
    [aux_sym_do_statement_token1] = ACTIONS(1450),
    [aux_sym_for_statement_token1] = ACTIONS(1450),
    [aux_sym_for_statement_token2] = ACTIONS(1450),
    [aux_sym_foreach_statement_token1] = ACTIONS(1450),
    [aux_sym_foreach_statement_token2] = ACTIONS(1450),
    [aux_sym_if_statement_token1] = ACTIONS(1450),
    [aux_sym_if_statement_token2] = ACTIONS(1450),
    [aux_sym_else_if_clause_token1] = ACTIONS(1450),
    [aux_sym_else_clause_token1] = ACTIONS(1450),
    [aux_sym_match_expression_token1] = ACTIONS(1450),
    [aux_sym_match_default_expression_token1] = ACTIONS(1450),
    [aux_sym_switch_statement_token1] = ACTIONS(1450),
    [aux_sym_switch_block_token1] = ACTIONS(1450),
    [anon_sym_PLUS] = ACTIONS(1450),
    [anon_sym_DASH] = ACTIONS(1450),
    [anon_sym_TILDE] = ACTIONS(1448),
    [anon_sym_BANG] = ACTIONS(1448),
    [anon_sym_AT] = ACTIONS(1448),
    [aux_sym_clone_expression_token1] = ACTIONS(1450),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1450),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1450),
    [anon_sym_DASH_DASH] = ACTIONS(1448),
    [anon_sym_PLUS_PLUS] = ACTIONS(1448),
    [aux_sym__list_destructing_token1] = ACTIONS(1450),
    [anon_sym_LBRACK] = ACTIONS(1448),
    [anon_sym_self] = ACTIONS(1450),
    [anon_sym_parent] = ACTIONS(1450),
    [aux_sym__argument_name_token1] = ACTIONS(1450),
    [aux_sym__argument_name_token2] = ACTIONS(1450),
    [anon_sym_POUND_LBRACK] = ACTIONS(1448),
    [aux_sym_encapsed_string_token1] = ACTIONS(1448),
    [anon_sym_DQUOTE] = ACTIONS(1448),
    [aux_sym_string_token1] = ACTIONS(1448),
    [anon_sym_SQUOTE] = ACTIONS(1448),
    [anon_sym_LT_LT_LT] = ACTIONS(1448),
    [anon_sym_BQUOTE] = ACTIONS(1448),
    [anon_sym_DOLLAR] = ACTIONS(1448),
    [aux_sym_yield_expression_token1] = ACTIONS(1450),
    [aux_sym_include_expression_token1] = ACTIONS(1450),
    [aux_sym_include_once_expression_token1] = ACTIONS(1450),
    [aux_sym_require_expression_token1] = ACTIONS(1450),
    [aux_sym_require_once_expression_token1] = ACTIONS(1450),
    [sym_comment] = ACTIONS(5),
  },
  [558] = {
    [sym_text_interpolation] = STATE(558),
    [ts_builtin_sym_end] = ACTIONS(1452),
    [sym_name] = ACTIONS(1454),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1452),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1454),
    [aux_sym_global_declaration_token1] = ACTIONS(1454),
    [aux_sym_namespace_definition_token1] = ACTIONS(1454),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1454),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1454),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1454),
    [anon_sym_BSLASH] = ACTIONS(1452),
    [anon_sym_LBRACE] = ACTIONS(1452),
    [anon_sym_RBRACE] = ACTIONS(1452),
    [aux_sym_trait_declaration_token1] = ACTIONS(1454),
    [aux_sym_interface_declaration_token1] = ACTIONS(1454),
    [aux_sym_enum_declaration_token1] = ACTIONS(1454),
    [aux_sym_enum_case_token1] = ACTIONS(1454),
    [aux_sym_class_declaration_token1] = ACTIONS(1454),
    [aux_sym_final_modifier_token1] = ACTIONS(1454),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1454),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1454),
    [sym_var_modifier] = ACTIONS(1454),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1454),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1454),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1454),
    [anon_sym_LPAREN2] = ACTIONS(1452),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1454),
    [aux_sym_cast_type_token1] = ACTIONS(1454),
    [aux_sym_echo_statement_token1] = ACTIONS(1454),
    [aux_sym_exit_statement_token1] = ACTIONS(1454),
    [anon_sym_unset] = ACTIONS(1454),
    [aux_sym_declare_statement_token1] = ACTIONS(1454),
    [aux_sym_declare_statement_token2] = ACTIONS(1454),
    [sym_float] = ACTIONS(1454),
    [aux_sym_try_statement_token1] = ACTIONS(1454),
    [aux_sym_goto_statement_token1] = ACTIONS(1454),
    [aux_sym_continue_statement_token1] = ACTIONS(1454),
    [aux_sym_break_statement_token1] = ACTIONS(1454),
    [sym_integer] = ACTIONS(1454),
    [aux_sym_return_statement_token1] = ACTIONS(1454),
    [aux_sym_throw_expression_token1] = ACTIONS(1454),
    [aux_sym_while_statement_token1] = ACTIONS(1454),
    [aux_sym_while_statement_token2] = ACTIONS(1454),
    [aux_sym_do_statement_token1] = ACTIONS(1454),
    [aux_sym_for_statement_token1] = ACTIONS(1454),
    [aux_sym_for_statement_token2] = ACTIONS(1454),
    [aux_sym_foreach_statement_token1] = ACTIONS(1454),
    [aux_sym_foreach_statement_token2] = ACTIONS(1454),
    [aux_sym_if_statement_token1] = ACTIONS(1454),
    [aux_sym_if_statement_token2] = ACTIONS(1454),
    [aux_sym_else_if_clause_token1] = ACTIONS(1454),
    [aux_sym_else_clause_token1] = ACTIONS(1454),
    [aux_sym_match_expression_token1] = ACTIONS(1454),
    [aux_sym_match_default_expression_token1] = ACTIONS(1454),
    [aux_sym_switch_statement_token1] = ACTIONS(1454),
    [aux_sym_switch_block_token1] = ACTIONS(1454),
    [anon_sym_PLUS] = ACTIONS(1454),
    [anon_sym_DASH] = ACTIONS(1454),
    [anon_sym_TILDE] = ACTIONS(1452),
    [anon_sym_BANG] = ACTIONS(1452),
    [anon_sym_AT] = ACTIONS(1452),
    [aux_sym_clone_expression_token1] = ACTIONS(1454),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1454),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1454),
    [anon_sym_DASH_DASH] = ACTIONS(1452),
    [anon_sym_PLUS_PLUS] = ACTIONS(1452),
    [aux_sym__list_destructing_token1] = ACTIONS(1454),
    [anon_sym_LBRACK] = ACTIONS(1452),
    [anon_sym_self] = ACTIONS(1454),
    [anon_sym_parent] = ACTIONS(1454),
    [aux_sym__argument_name_token1] = ACTIONS(1454),
    [aux_sym__argument_name_token2] = ACTIONS(1454),
    [anon_sym_POUND_LBRACK] = ACTIONS(1452),
    [aux_sym_encapsed_string_token1] = ACTIONS(1452),
    [anon_sym_DQUOTE] = ACTIONS(1452),
    [aux_sym_string_token1] = ACTIONS(1452),
    [anon_sym_SQUOTE] = ACTIONS(1452),
    [anon_sym_LT_LT_LT] = ACTIONS(1452),
    [anon_sym_BQUOTE] = ACTIONS(1452),
    [anon_sym_DOLLAR] = ACTIONS(1452),
    [aux_sym_yield_expression_token1] = ACTIONS(1454),
    [aux_sym_include_expression_token1] = ACTIONS(1454),
    [aux_sym_include_once_expression_token1] = ACTIONS(1454),
    [aux_sym_require_expression_token1] = ACTIONS(1454),
    [aux_sym_require_once_expression_token1] = ACTIONS(1454),
    [sym_comment] = ACTIONS(5),
  },
  [559] = {
    [sym_text_interpolation] = STATE(559),
    [ts_builtin_sym_end] = ACTIONS(1456),
    [sym_name] = ACTIONS(1458),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1456),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1458),
    [aux_sym_global_declaration_token1] = ACTIONS(1458),
    [aux_sym_namespace_definition_token1] = ACTIONS(1458),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1458),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1458),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1458),
    [anon_sym_BSLASH] = ACTIONS(1456),
    [anon_sym_LBRACE] = ACTIONS(1456),
    [anon_sym_RBRACE] = ACTIONS(1456),
    [aux_sym_trait_declaration_token1] = ACTIONS(1458),
    [aux_sym_interface_declaration_token1] = ACTIONS(1458),
    [aux_sym_enum_declaration_token1] = ACTIONS(1458),
    [aux_sym_enum_case_token1] = ACTIONS(1458),
    [aux_sym_class_declaration_token1] = ACTIONS(1458),
    [aux_sym_final_modifier_token1] = ACTIONS(1458),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1458),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1458),
    [sym_var_modifier] = ACTIONS(1458),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1458),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1458),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1458),
    [anon_sym_LPAREN2] = ACTIONS(1456),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1458),
    [aux_sym_cast_type_token1] = ACTIONS(1458),
    [aux_sym_echo_statement_token1] = ACTIONS(1458),
    [aux_sym_exit_statement_token1] = ACTIONS(1458),
    [anon_sym_unset] = ACTIONS(1458),
    [aux_sym_declare_statement_token1] = ACTIONS(1458),
    [aux_sym_declare_statement_token2] = ACTIONS(1458),
    [sym_float] = ACTIONS(1458),
    [aux_sym_try_statement_token1] = ACTIONS(1458),
    [aux_sym_goto_statement_token1] = ACTIONS(1458),
    [aux_sym_continue_statement_token1] = ACTIONS(1458),
    [aux_sym_break_statement_token1] = ACTIONS(1458),
    [sym_integer] = ACTIONS(1458),
    [aux_sym_return_statement_token1] = ACTIONS(1458),
    [aux_sym_throw_expression_token1] = ACTIONS(1458),
    [aux_sym_while_statement_token1] = ACTIONS(1458),
    [aux_sym_while_statement_token2] = ACTIONS(1458),
    [aux_sym_do_statement_token1] = ACTIONS(1458),
    [aux_sym_for_statement_token1] = ACTIONS(1458),
    [aux_sym_for_statement_token2] = ACTIONS(1458),
    [aux_sym_foreach_statement_token1] = ACTIONS(1458),
    [aux_sym_foreach_statement_token2] = ACTIONS(1458),
    [aux_sym_if_statement_token1] = ACTIONS(1458),
    [aux_sym_if_statement_token2] = ACTIONS(1458),
    [aux_sym_else_if_clause_token1] = ACTIONS(1458),
    [aux_sym_else_clause_token1] = ACTIONS(1458),
    [aux_sym_match_expression_token1] = ACTIONS(1458),
    [aux_sym_match_default_expression_token1] = ACTIONS(1458),
    [aux_sym_switch_statement_token1] = ACTIONS(1458),
    [aux_sym_switch_block_token1] = ACTIONS(1458),
    [anon_sym_PLUS] = ACTIONS(1458),
    [anon_sym_DASH] = ACTIONS(1458),
    [anon_sym_TILDE] = ACTIONS(1456),
    [anon_sym_BANG] = ACTIONS(1456),
    [anon_sym_AT] = ACTIONS(1456),
    [aux_sym_clone_expression_token1] = ACTIONS(1458),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1458),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1458),
    [anon_sym_DASH_DASH] = ACTIONS(1456),
    [anon_sym_PLUS_PLUS] = ACTIONS(1456),
    [aux_sym__list_destructing_token1] = ACTIONS(1458),
    [anon_sym_LBRACK] = ACTIONS(1456),
    [anon_sym_self] = ACTIONS(1458),
    [anon_sym_parent] = ACTIONS(1458),
    [aux_sym__argument_name_token1] = ACTIONS(1458),
    [aux_sym__argument_name_token2] = ACTIONS(1458),
    [anon_sym_POUND_LBRACK] = ACTIONS(1456),
    [aux_sym_encapsed_string_token1] = ACTIONS(1456),
    [anon_sym_DQUOTE] = ACTIONS(1456),
    [aux_sym_string_token1] = ACTIONS(1456),
    [anon_sym_SQUOTE] = ACTIONS(1456),
    [anon_sym_LT_LT_LT] = ACTIONS(1456),
    [anon_sym_BQUOTE] = ACTIONS(1456),
    [anon_sym_DOLLAR] = ACTIONS(1456),
    [aux_sym_yield_expression_token1] = ACTIONS(1458),
    [aux_sym_include_expression_token1] = ACTIONS(1458),
    [aux_sym_include_once_expression_token1] = ACTIONS(1458),
    [aux_sym_require_expression_token1] = ACTIONS(1458),
    [aux_sym_require_once_expression_token1] = ACTIONS(1458),
    [sym_comment] = ACTIONS(5),
  },
  [560] = {
    [sym_text_interpolation] = STATE(560),
    [ts_builtin_sym_end] = ACTIONS(1460),
    [sym_name] = ACTIONS(1462),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1460),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1462),
    [aux_sym_global_declaration_token1] = ACTIONS(1462),
    [aux_sym_namespace_definition_token1] = ACTIONS(1462),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1462),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1462),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1462),
    [anon_sym_BSLASH] = ACTIONS(1460),
    [anon_sym_LBRACE] = ACTIONS(1460),
    [anon_sym_RBRACE] = ACTIONS(1460),
    [aux_sym_trait_declaration_token1] = ACTIONS(1462),
    [aux_sym_interface_declaration_token1] = ACTIONS(1462),
    [aux_sym_enum_declaration_token1] = ACTIONS(1462),
    [aux_sym_enum_case_token1] = ACTIONS(1462),
    [aux_sym_class_declaration_token1] = ACTIONS(1462),
    [aux_sym_final_modifier_token1] = ACTIONS(1462),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1462),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1462),
    [sym_var_modifier] = ACTIONS(1462),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1462),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1462),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1462),
    [anon_sym_LPAREN2] = ACTIONS(1460),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1462),
    [aux_sym_cast_type_token1] = ACTIONS(1462),
    [aux_sym_echo_statement_token1] = ACTIONS(1462),
    [aux_sym_exit_statement_token1] = ACTIONS(1462),
    [anon_sym_unset] = ACTIONS(1462),
    [aux_sym_declare_statement_token1] = ACTIONS(1462),
    [aux_sym_declare_statement_token2] = ACTIONS(1462),
    [sym_float] = ACTIONS(1462),
    [aux_sym_try_statement_token1] = ACTIONS(1462),
    [aux_sym_goto_statement_token1] = ACTIONS(1462),
    [aux_sym_continue_statement_token1] = ACTIONS(1462),
    [aux_sym_break_statement_token1] = ACTIONS(1462),
    [sym_integer] = ACTIONS(1462),
    [aux_sym_return_statement_token1] = ACTIONS(1462),
    [aux_sym_throw_expression_token1] = ACTIONS(1462),
    [aux_sym_while_statement_token1] = ACTIONS(1462),
    [aux_sym_while_statement_token2] = ACTIONS(1462),
    [aux_sym_do_statement_token1] = ACTIONS(1462),
    [aux_sym_for_statement_token1] = ACTIONS(1462),
    [aux_sym_for_statement_token2] = ACTIONS(1462),
    [aux_sym_foreach_statement_token1] = ACTIONS(1462),
    [aux_sym_foreach_statement_token2] = ACTIONS(1462),
    [aux_sym_if_statement_token1] = ACTIONS(1462),
    [aux_sym_if_statement_token2] = ACTIONS(1462),
    [aux_sym_else_if_clause_token1] = ACTIONS(1462),
    [aux_sym_else_clause_token1] = ACTIONS(1462),
    [aux_sym_match_expression_token1] = ACTIONS(1462),
    [aux_sym_match_default_expression_token1] = ACTIONS(1462),
    [aux_sym_switch_statement_token1] = ACTIONS(1462),
    [aux_sym_switch_block_token1] = ACTIONS(1462),
    [anon_sym_PLUS] = ACTIONS(1462),
    [anon_sym_DASH] = ACTIONS(1462),
    [anon_sym_TILDE] = ACTIONS(1460),
    [anon_sym_BANG] = ACTIONS(1460),
    [anon_sym_AT] = ACTIONS(1460),
    [aux_sym_clone_expression_token1] = ACTIONS(1462),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1462),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1462),
    [anon_sym_DASH_DASH] = ACTIONS(1460),
    [anon_sym_PLUS_PLUS] = ACTIONS(1460),
    [aux_sym__list_destructing_token1] = ACTIONS(1462),
    [anon_sym_LBRACK] = ACTIONS(1460),
    [anon_sym_self] = ACTIONS(1462),
    [anon_sym_parent] = ACTIONS(1462),
    [aux_sym__argument_name_token1] = ACTIONS(1462),
    [aux_sym__argument_name_token2] = ACTIONS(1462),
    [anon_sym_POUND_LBRACK] = ACTIONS(1460),
    [aux_sym_encapsed_string_token1] = ACTIONS(1460),
    [anon_sym_DQUOTE] = ACTIONS(1460),
    [aux_sym_string_token1] = ACTIONS(1460),
    [anon_sym_SQUOTE] = ACTIONS(1460),
    [anon_sym_LT_LT_LT] = ACTIONS(1460),
    [anon_sym_BQUOTE] = ACTIONS(1460),
    [anon_sym_DOLLAR] = ACTIONS(1460),
    [aux_sym_yield_expression_token1] = ACTIONS(1462),
    [aux_sym_include_expression_token1] = ACTIONS(1462),
    [aux_sym_include_once_expression_token1] = ACTIONS(1462),
    [aux_sym_require_expression_token1] = ACTIONS(1462),
    [aux_sym_require_once_expression_token1] = ACTIONS(1462),
    [sym_comment] = ACTIONS(5),
  },
  [561] = {
    [sym_text_interpolation] = STATE(561),
    [ts_builtin_sym_end] = ACTIONS(1464),
    [sym_name] = ACTIONS(1466),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1464),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1466),
    [aux_sym_global_declaration_token1] = ACTIONS(1466),
    [aux_sym_namespace_definition_token1] = ACTIONS(1466),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1466),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1466),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1466),
    [anon_sym_BSLASH] = ACTIONS(1464),
    [anon_sym_LBRACE] = ACTIONS(1464),
    [anon_sym_RBRACE] = ACTIONS(1464),
    [aux_sym_trait_declaration_token1] = ACTIONS(1466),
    [aux_sym_interface_declaration_token1] = ACTIONS(1466),
    [aux_sym_enum_declaration_token1] = ACTIONS(1466),
    [aux_sym_enum_case_token1] = ACTIONS(1466),
    [aux_sym_class_declaration_token1] = ACTIONS(1466),
    [aux_sym_final_modifier_token1] = ACTIONS(1466),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1466),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1466),
    [sym_var_modifier] = ACTIONS(1466),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1466),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1466),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1466),
    [anon_sym_LPAREN2] = ACTIONS(1464),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1466),
    [aux_sym_cast_type_token1] = ACTIONS(1466),
    [aux_sym_echo_statement_token1] = ACTIONS(1466),
    [aux_sym_exit_statement_token1] = ACTIONS(1466),
    [anon_sym_unset] = ACTIONS(1466),
    [aux_sym_declare_statement_token1] = ACTIONS(1466),
    [aux_sym_declare_statement_token2] = ACTIONS(1466),
    [sym_float] = ACTIONS(1466),
    [aux_sym_try_statement_token1] = ACTIONS(1466),
    [aux_sym_goto_statement_token1] = ACTIONS(1466),
    [aux_sym_continue_statement_token1] = ACTIONS(1466),
    [aux_sym_break_statement_token1] = ACTIONS(1466),
    [sym_integer] = ACTIONS(1466),
    [aux_sym_return_statement_token1] = ACTIONS(1466),
    [aux_sym_throw_expression_token1] = ACTIONS(1466),
    [aux_sym_while_statement_token1] = ACTIONS(1466),
    [aux_sym_while_statement_token2] = ACTIONS(1466),
    [aux_sym_do_statement_token1] = ACTIONS(1466),
    [aux_sym_for_statement_token1] = ACTIONS(1466),
    [aux_sym_for_statement_token2] = ACTIONS(1466),
    [aux_sym_foreach_statement_token1] = ACTIONS(1466),
    [aux_sym_foreach_statement_token2] = ACTIONS(1466),
    [aux_sym_if_statement_token1] = ACTIONS(1466),
    [aux_sym_if_statement_token2] = ACTIONS(1466),
    [aux_sym_else_if_clause_token1] = ACTIONS(1466),
    [aux_sym_else_clause_token1] = ACTIONS(1466),
    [aux_sym_match_expression_token1] = ACTIONS(1466),
    [aux_sym_match_default_expression_token1] = ACTIONS(1466),
    [aux_sym_switch_statement_token1] = ACTIONS(1466),
    [aux_sym_switch_block_token1] = ACTIONS(1466),
    [anon_sym_PLUS] = ACTIONS(1466),
    [anon_sym_DASH] = ACTIONS(1466),
    [anon_sym_TILDE] = ACTIONS(1464),
    [anon_sym_BANG] = ACTIONS(1464),
    [anon_sym_AT] = ACTIONS(1464),
    [aux_sym_clone_expression_token1] = ACTIONS(1466),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1466),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1466),
    [anon_sym_DASH_DASH] = ACTIONS(1464),
    [anon_sym_PLUS_PLUS] = ACTIONS(1464),
    [aux_sym__list_destructing_token1] = ACTIONS(1466),
    [anon_sym_LBRACK] = ACTIONS(1464),
    [anon_sym_self] = ACTIONS(1466),
    [anon_sym_parent] = ACTIONS(1466),
    [aux_sym__argument_name_token1] = ACTIONS(1466),
    [aux_sym__argument_name_token2] = ACTIONS(1466),
    [anon_sym_POUND_LBRACK] = ACTIONS(1464),
    [aux_sym_encapsed_string_token1] = ACTIONS(1464),
    [anon_sym_DQUOTE] = ACTIONS(1464),
    [aux_sym_string_token1] = ACTIONS(1464),
    [anon_sym_SQUOTE] = ACTIONS(1464),
    [anon_sym_LT_LT_LT] = ACTIONS(1464),
    [anon_sym_BQUOTE] = ACTIONS(1464),
    [anon_sym_DOLLAR] = ACTIONS(1464),
    [aux_sym_yield_expression_token1] = ACTIONS(1466),
    [aux_sym_include_expression_token1] = ACTIONS(1466),
    [aux_sym_include_once_expression_token1] = ACTIONS(1466),
    [aux_sym_require_expression_token1] = ACTIONS(1466),
    [aux_sym_require_once_expression_token1] = ACTIONS(1466),
    [sym_comment] = ACTIONS(5),
  },
  [562] = {
    [sym_text_interpolation] = STATE(562),
    [ts_builtin_sym_end] = ACTIONS(1468),
    [sym_name] = ACTIONS(1470),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1468),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1470),
    [aux_sym_global_declaration_token1] = ACTIONS(1470),
    [aux_sym_namespace_definition_token1] = ACTIONS(1470),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1470),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1470),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1470),
    [anon_sym_BSLASH] = ACTIONS(1468),
    [anon_sym_LBRACE] = ACTIONS(1468),
    [anon_sym_RBRACE] = ACTIONS(1468),
    [aux_sym_trait_declaration_token1] = ACTIONS(1470),
    [aux_sym_interface_declaration_token1] = ACTIONS(1470),
    [aux_sym_enum_declaration_token1] = ACTIONS(1470),
    [aux_sym_enum_case_token1] = ACTIONS(1470),
    [aux_sym_class_declaration_token1] = ACTIONS(1470),
    [aux_sym_final_modifier_token1] = ACTIONS(1470),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1470),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1470),
    [sym_var_modifier] = ACTIONS(1470),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1470),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1470),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1470),
    [anon_sym_LPAREN2] = ACTIONS(1468),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1470),
    [aux_sym_cast_type_token1] = ACTIONS(1470),
    [aux_sym_echo_statement_token1] = ACTIONS(1470),
    [aux_sym_exit_statement_token1] = ACTIONS(1470),
    [anon_sym_unset] = ACTIONS(1470),
    [aux_sym_declare_statement_token1] = ACTIONS(1470),
    [aux_sym_declare_statement_token2] = ACTIONS(1470),
    [sym_float] = ACTIONS(1470),
    [aux_sym_try_statement_token1] = ACTIONS(1470),
    [aux_sym_goto_statement_token1] = ACTIONS(1470),
    [aux_sym_continue_statement_token1] = ACTIONS(1470),
    [aux_sym_break_statement_token1] = ACTIONS(1470),
    [sym_integer] = ACTIONS(1470),
    [aux_sym_return_statement_token1] = ACTIONS(1470),
    [aux_sym_throw_expression_token1] = ACTIONS(1470),
    [aux_sym_while_statement_token1] = ACTIONS(1470),
    [aux_sym_while_statement_token2] = ACTIONS(1470),
    [aux_sym_do_statement_token1] = ACTIONS(1470),
    [aux_sym_for_statement_token1] = ACTIONS(1470),
    [aux_sym_for_statement_token2] = ACTIONS(1470),
    [aux_sym_foreach_statement_token1] = ACTIONS(1470),
    [aux_sym_foreach_statement_token2] = ACTIONS(1470),
    [aux_sym_if_statement_token1] = ACTIONS(1470),
    [aux_sym_if_statement_token2] = ACTIONS(1470),
    [aux_sym_else_if_clause_token1] = ACTIONS(1470),
    [aux_sym_else_clause_token1] = ACTIONS(1470),
    [aux_sym_match_expression_token1] = ACTIONS(1470),
    [aux_sym_match_default_expression_token1] = ACTIONS(1470),
    [aux_sym_switch_statement_token1] = ACTIONS(1470),
    [aux_sym_switch_block_token1] = ACTIONS(1470),
    [anon_sym_PLUS] = ACTIONS(1470),
    [anon_sym_DASH] = ACTIONS(1470),
    [anon_sym_TILDE] = ACTIONS(1468),
    [anon_sym_BANG] = ACTIONS(1468),
    [anon_sym_AT] = ACTIONS(1468),
    [aux_sym_clone_expression_token1] = ACTIONS(1470),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1470),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1470),
    [anon_sym_DASH_DASH] = ACTIONS(1468),
    [anon_sym_PLUS_PLUS] = ACTIONS(1468),
    [aux_sym__list_destructing_token1] = ACTIONS(1470),
    [anon_sym_LBRACK] = ACTIONS(1468),
    [anon_sym_self] = ACTIONS(1470),
    [anon_sym_parent] = ACTIONS(1470),
    [aux_sym__argument_name_token1] = ACTIONS(1470),
    [aux_sym__argument_name_token2] = ACTIONS(1470),
    [anon_sym_POUND_LBRACK] = ACTIONS(1468),
    [aux_sym_encapsed_string_token1] = ACTIONS(1468),
    [anon_sym_DQUOTE] = ACTIONS(1468),
    [aux_sym_string_token1] = ACTIONS(1468),
    [anon_sym_SQUOTE] = ACTIONS(1468),
    [anon_sym_LT_LT_LT] = ACTIONS(1468),
    [anon_sym_BQUOTE] = ACTIONS(1468),
    [anon_sym_DOLLAR] = ACTIONS(1468),
    [aux_sym_yield_expression_token1] = ACTIONS(1470),
    [aux_sym_include_expression_token1] = ACTIONS(1470),
    [aux_sym_include_once_expression_token1] = ACTIONS(1470),
    [aux_sym_require_expression_token1] = ACTIONS(1470),
    [aux_sym_require_once_expression_token1] = ACTIONS(1470),
    [sym_comment] = ACTIONS(5),
  },
  [563] = {
    [sym_text_interpolation] = STATE(563),
    [ts_builtin_sym_end] = ACTIONS(1472),
    [sym_name] = ACTIONS(1474),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1472),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1474),
    [aux_sym_global_declaration_token1] = ACTIONS(1474),
    [aux_sym_namespace_definition_token1] = ACTIONS(1474),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1474),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1474),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1474),
    [anon_sym_BSLASH] = ACTIONS(1472),
    [anon_sym_LBRACE] = ACTIONS(1472),
    [anon_sym_RBRACE] = ACTIONS(1472),
    [aux_sym_trait_declaration_token1] = ACTIONS(1474),
    [aux_sym_interface_declaration_token1] = ACTIONS(1474),
    [aux_sym_enum_declaration_token1] = ACTIONS(1474),
    [aux_sym_enum_case_token1] = ACTIONS(1474),
    [aux_sym_class_declaration_token1] = ACTIONS(1474),
    [aux_sym_final_modifier_token1] = ACTIONS(1474),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1474),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1474),
    [sym_var_modifier] = ACTIONS(1474),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1474),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1474),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1474),
    [anon_sym_LPAREN2] = ACTIONS(1472),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1474),
    [aux_sym_cast_type_token1] = ACTIONS(1474),
    [aux_sym_echo_statement_token1] = ACTIONS(1474),
    [aux_sym_exit_statement_token1] = ACTIONS(1474),
    [anon_sym_unset] = ACTIONS(1474),
    [aux_sym_declare_statement_token1] = ACTIONS(1474),
    [aux_sym_declare_statement_token2] = ACTIONS(1474),
    [sym_float] = ACTIONS(1474),
    [aux_sym_try_statement_token1] = ACTIONS(1474),
    [aux_sym_goto_statement_token1] = ACTIONS(1474),
    [aux_sym_continue_statement_token1] = ACTIONS(1474),
    [aux_sym_break_statement_token1] = ACTIONS(1474),
    [sym_integer] = ACTIONS(1474),
    [aux_sym_return_statement_token1] = ACTIONS(1474),
    [aux_sym_throw_expression_token1] = ACTIONS(1474),
    [aux_sym_while_statement_token1] = ACTIONS(1474),
    [aux_sym_while_statement_token2] = ACTIONS(1474),
    [aux_sym_do_statement_token1] = ACTIONS(1474),
    [aux_sym_for_statement_token1] = ACTIONS(1474),
    [aux_sym_for_statement_token2] = ACTIONS(1474),
    [aux_sym_foreach_statement_token1] = ACTIONS(1474),
    [aux_sym_foreach_statement_token2] = ACTIONS(1474),
    [aux_sym_if_statement_token1] = ACTIONS(1474),
    [aux_sym_if_statement_token2] = ACTIONS(1474),
    [aux_sym_else_if_clause_token1] = ACTIONS(1474),
    [aux_sym_else_clause_token1] = ACTIONS(1474),
    [aux_sym_match_expression_token1] = ACTIONS(1474),
    [aux_sym_match_default_expression_token1] = ACTIONS(1474),
    [aux_sym_switch_statement_token1] = ACTIONS(1474),
    [aux_sym_switch_block_token1] = ACTIONS(1474),
    [anon_sym_PLUS] = ACTIONS(1474),
    [anon_sym_DASH] = ACTIONS(1474),
    [anon_sym_TILDE] = ACTIONS(1472),
    [anon_sym_BANG] = ACTIONS(1472),
    [anon_sym_AT] = ACTIONS(1472),
    [aux_sym_clone_expression_token1] = ACTIONS(1474),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1474),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1474),
    [anon_sym_DASH_DASH] = ACTIONS(1472),
    [anon_sym_PLUS_PLUS] = ACTIONS(1472),
    [aux_sym__list_destructing_token1] = ACTIONS(1474),
    [anon_sym_LBRACK] = ACTIONS(1472),
    [anon_sym_self] = ACTIONS(1474),
    [anon_sym_parent] = ACTIONS(1474),
    [aux_sym__argument_name_token1] = ACTIONS(1474),
    [aux_sym__argument_name_token2] = ACTIONS(1474),
    [anon_sym_POUND_LBRACK] = ACTIONS(1472),
    [aux_sym_encapsed_string_token1] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1472),
    [aux_sym_string_token1] = ACTIONS(1472),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_LT_LT_LT] = ACTIONS(1472),
    [anon_sym_BQUOTE] = ACTIONS(1472),
    [anon_sym_DOLLAR] = ACTIONS(1472),
    [aux_sym_yield_expression_token1] = ACTIONS(1474),
    [aux_sym_include_expression_token1] = ACTIONS(1474),
    [aux_sym_include_once_expression_token1] = ACTIONS(1474),
    [aux_sym_require_expression_token1] = ACTIONS(1474),
    [aux_sym_require_once_expression_token1] = ACTIONS(1474),
    [sym_comment] = ACTIONS(5),
  },
  [564] = {
    [sym_text_interpolation] = STATE(564),
    [ts_builtin_sym_end] = ACTIONS(1476),
    [sym_name] = ACTIONS(1478),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1476),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1478),
    [aux_sym_global_declaration_token1] = ACTIONS(1478),
    [aux_sym_namespace_definition_token1] = ACTIONS(1478),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1478),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1478),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1478),
    [anon_sym_BSLASH] = ACTIONS(1476),
    [anon_sym_LBRACE] = ACTIONS(1476),
    [anon_sym_RBRACE] = ACTIONS(1476),
    [aux_sym_trait_declaration_token1] = ACTIONS(1478),
    [aux_sym_interface_declaration_token1] = ACTIONS(1478),
    [aux_sym_enum_declaration_token1] = ACTIONS(1478),
    [aux_sym_enum_case_token1] = ACTIONS(1478),
    [aux_sym_class_declaration_token1] = ACTIONS(1478),
    [aux_sym_final_modifier_token1] = ACTIONS(1478),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1478),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1478),
    [sym_var_modifier] = ACTIONS(1478),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1478),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1478),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1478),
    [anon_sym_LPAREN2] = ACTIONS(1476),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1478),
    [aux_sym_cast_type_token1] = ACTIONS(1478),
    [aux_sym_echo_statement_token1] = ACTIONS(1478),
    [aux_sym_exit_statement_token1] = ACTIONS(1478),
    [anon_sym_unset] = ACTIONS(1478),
    [aux_sym_declare_statement_token1] = ACTIONS(1478),
    [aux_sym_declare_statement_token2] = ACTIONS(1478),
    [sym_float] = ACTIONS(1478),
    [aux_sym_try_statement_token1] = ACTIONS(1478),
    [aux_sym_goto_statement_token1] = ACTIONS(1478),
    [aux_sym_continue_statement_token1] = ACTIONS(1478),
    [aux_sym_break_statement_token1] = ACTIONS(1478),
    [sym_integer] = ACTIONS(1478),
    [aux_sym_return_statement_token1] = ACTIONS(1478),
    [aux_sym_throw_expression_token1] = ACTIONS(1478),
    [aux_sym_while_statement_token1] = ACTIONS(1478),
    [aux_sym_while_statement_token2] = ACTIONS(1478),
    [aux_sym_do_statement_token1] = ACTIONS(1478),
    [aux_sym_for_statement_token1] = ACTIONS(1478),
    [aux_sym_for_statement_token2] = ACTIONS(1478),
    [aux_sym_foreach_statement_token1] = ACTIONS(1478),
    [aux_sym_foreach_statement_token2] = ACTIONS(1478),
    [aux_sym_if_statement_token1] = ACTIONS(1478),
    [aux_sym_if_statement_token2] = ACTIONS(1478),
    [aux_sym_else_if_clause_token1] = ACTIONS(1478),
    [aux_sym_else_clause_token1] = ACTIONS(1478),
    [aux_sym_match_expression_token1] = ACTIONS(1478),
    [aux_sym_match_default_expression_token1] = ACTIONS(1478),
    [aux_sym_switch_statement_token1] = ACTIONS(1478),
    [aux_sym_switch_block_token1] = ACTIONS(1478),
    [anon_sym_PLUS] = ACTIONS(1478),
    [anon_sym_DASH] = ACTIONS(1478),
    [anon_sym_TILDE] = ACTIONS(1476),
    [anon_sym_BANG] = ACTIONS(1476),
    [anon_sym_AT] = ACTIONS(1476),
    [aux_sym_clone_expression_token1] = ACTIONS(1478),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1478),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1478),
    [anon_sym_DASH_DASH] = ACTIONS(1476),
    [anon_sym_PLUS_PLUS] = ACTIONS(1476),
    [aux_sym__list_destructing_token1] = ACTIONS(1478),
    [anon_sym_LBRACK] = ACTIONS(1476),
    [anon_sym_self] = ACTIONS(1478),
    [anon_sym_parent] = ACTIONS(1478),
    [aux_sym__argument_name_token1] = ACTIONS(1478),
    [aux_sym__argument_name_token2] = ACTIONS(1478),
    [anon_sym_POUND_LBRACK] = ACTIONS(1476),
    [aux_sym_encapsed_string_token1] = ACTIONS(1476),
    [anon_sym_DQUOTE] = ACTIONS(1476),
    [aux_sym_string_token1] = ACTIONS(1476),
    [anon_sym_SQUOTE] = ACTIONS(1476),
    [anon_sym_LT_LT_LT] = ACTIONS(1476),
    [anon_sym_BQUOTE] = ACTIONS(1476),
    [anon_sym_DOLLAR] = ACTIONS(1476),
    [aux_sym_yield_expression_token1] = ACTIONS(1478),
    [aux_sym_include_expression_token1] = ACTIONS(1478),
    [aux_sym_include_once_expression_token1] = ACTIONS(1478),
    [aux_sym_require_expression_token1] = ACTIONS(1478),
    [aux_sym_require_once_expression_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
  },
  [565] = {
    [sym_text_interpolation] = STATE(565),
    [ts_builtin_sym_end] = ACTIONS(1480),
    [sym_name] = ACTIONS(1482),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1480),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1482),
    [aux_sym_global_declaration_token1] = ACTIONS(1482),
    [aux_sym_namespace_definition_token1] = ACTIONS(1482),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1482),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1482),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1482),
    [anon_sym_BSLASH] = ACTIONS(1480),
    [anon_sym_LBRACE] = ACTIONS(1480),
    [anon_sym_RBRACE] = ACTIONS(1480),
    [aux_sym_trait_declaration_token1] = ACTIONS(1482),
    [aux_sym_interface_declaration_token1] = ACTIONS(1482),
    [aux_sym_enum_declaration_token1] = ACTIONS(1482),
    [aux_sym_enum_case_token1] = ACTIONS(1482),
    [aux_sym_class_declaration_token1] = ACTIONS(1482),
    [aux_sym_final_modifier_token1] = ACTIONS(1482),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1482),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1482),
    [sym_var_modifier] = ACTIONS(1482),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1482),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1482),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1482),
    [anon_sym_LPAREN2] = ACTIONS(1480),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1482),
    [aux_sym_cast_type_token1] = ACTIONS(1482),
    [aux_sym_echo_statement_token1] = ACTIONS(1482),
    [aux_sym_exit_statement_token1] = ACTIONS(1482),
    [anon_sym_unset] = ACTIONS(1482),
    [aux_sym_declare_statement_token1] = ACTIONS(1482),
    [aux_sym_declare_statement_token2] = ACTIONS(1482),
    [sym_float] = ACTIONS(1482),
    [aux_sym_try_statement_token1] = ACTIONS(1482),
    [aux_sym_goto_statement_token1] = ACTIONS(1482),
    [aux_sym_continue_statement_token1] = ACTIONS(1482),
    [aux_sym_break_statement_token1] = ACTIONS(1482),
    [sym_integer] = ACTIONS(1482),
    [aux_sym_return_statement_token1] = ACTIONS(1482),
    [aux_sym_throw_expression_token1] = ACTIONS(1482),
    [aux_sym_while_statement_token1] = ACTIONS(1482),
    [aux_sym_while_statement_token2] = ACTIONS(1482),
    [aux_sym_do_statement_token1] = ACTIONS(1482),
    [aux_sym_for_statement_token1] = ACTIONS(1482),
    [aux_sym_for_statement_token2] = ACTIONS(1482),
    [aux_sym_foreach_statement_token1] = ACTIONS(1482),
    [aux_sym_foreach_statement_token2] = ACTIONS(1482),
    [aux_sym_if_statement_token1] = ACTIONS(1482),
    [aux_sym_if_statement_token2] = ACTIONS(1482),
    [aux_sym_else_if_clause_token1] = ACTIONS(1482),
    [aux_sym_else_clause_token1] = ACTIONS(1482),
    [aux_sym_match_expression_token1] = ACTIONS(1482),
    [aux_sym_match_default_expression_token1] = ACTIONS(1482),
    [aux_sym_switch_statement_token1] = ACTIONS(1482),
    [aux_sym_switch_block_token1] = ACTIONS(1482),
    [anon_sym_PLUS] = ACTIONS(1482),
    [anon_sym_DASH] = ACTIONS(1482),
    [anon_sym_TILDE] = ACTIONS(1480),
    [anon_sym_BANG] = ACTIONS(1480),
    [anon_sym_AT] = ACTIONS(1480),
    [aux_sym_clone_expression_token1] = ACTIONS(1482),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1482),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1482),
    [anon_sym_DASH_DASH] = ACTIONS(1480),
    [anon_sym_PLUS_PLUS] = ACTIONS(1480),
    [aux_sym__list_destructing_token1] = ACTIONS(1482),
    [anon_sym_LBRACK] = ACTIONS(1480),
    [anon_sym_self] = ACTIONS(1482),
    [anon_sym_parent] = ACTIONS(1482),
    [aux_sym__argument_name_token1] = ACTIONS(1482),
    [aux_sym__argument_name_token2] = ACTIONS(1482),
    [anon_sym_POUND_LBRACK] = ACTIONS(1480),
    [aux_sym_encapsed_string_token1] = ACTIONS(1480),
    [anon_sym_DQUOTE] = ACTIONS(1480),
    [aux_sym_string_token1] = ACTIONS(1480),
    [anon_sym_SQUOTE] = ACTIONS(1480),
    [anon_sym_LT_LT_LT] = ACTIONS(1480),
    [anon_sym_BQUOTE] = ACTIONS(1480),
    [anon_sym_DOLLAR] = ACTIONS(1480),
    [aux_sym_yield_expression_token1] = ACTIONS(1482),
    [aux_sym_include_expression_token1] = ACTIONS(1482),
    [aux_sym_include_once_expression_token1] = ACTIONS(1482),
    [aux_sym_require_expression_token1] = ACTIONS(1482),
    [aux_sym_require_once_expression_token1] = ACTIONS(1482),
    [sym_comment] = ACTIONS(5),
  },
  [566] = {
    [sym_text_interpolation] = STATE(566),
    [ts_builtin_sym_end] = ACTIONS(1484),
    [sym_name] = ACTIONS(1486),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1484),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1486),
    [aux_sym_global_declaration_token1] = ACTIONS(1486),
    [aux_sym_namespace_definition_token1] = ACTIONS(1486),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1486),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1486),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1486),
    [anon_sym_BSLASH] = ACTIONS(1484),
    [anon_sym_LBRACE] = ACTIONS(1484),
    [anon_sym_RBRACE] = ACTIONS(1484),
    [aux_sym_trait_declaration_token1] = ACTIONS(1486),
    [aux_sym_interface_declaration_token1] = ACTIONS(1486),
    [aux_sym_enum_declaration_token1] = ACTIONS(1486),
    [aux_sym_enum_case_token1] = ACTIONS(1486),
    [aux_sym_class_declaration_token1] = ACTIONS(1486),
    [aux_sym_final_modifier_token1] = ACTIONS(1486),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1486),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1486),
    [sym_var_modifier] = ACTIONS(1486),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1486),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1486),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1486),
    [anon_sym_LPAREN2] = ACTIONS(1484),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1486),
    [aux_sym_cast_type_token1] = ACTIONS(1486),
    [aux_sym_echo_statement_token1] = ACTIONS(1486),
    [aux_sym_exit_statement_token1] = ACTIONS(1486),
    [anon_sym_unset] = ACTIONS(1486),
    [aux_sym_declare_statement_token1] = ACTIONS(1486),
    [aux_sym_declare_statement_token2] = ACTIONS(1486),
    [sym_float] = ACTIONS(1486),
    [aux_sym_try_statement_token1] = ACTIONS(1486),
    [aux_sym_goto_statement_token1] = ACTIONS(1486),
    [aux_sym_continue_statement_token1] = ACTIONS(1486),
    [aux_sym_break_statement_token1] = ACTIONS(1486),
    [sym_integer] = ACTIONS(1486),
    [aux_sym_return_statement_token1] = ACTIONS(1486),
    [aux_sym_throw_expression_token1] = ACTIONS(1486),
    [aux_sym_while_statement_token1] = ACTIONS(1486),
    [aux_sym_while_statement_token2] = ACTIONS(1486),
    [aux_sym_do_statement_token1] = ACTIONS(1486),
    [aux_sym_for_statement_token1] = ACTIONS(1486),
    [aux_sym_for_statement_token2] = ACTIONS(1486),
    [aux_sym_foreach_statement_token1] = ACTIONS(1486),
    [aux_sym_foreach_statement_token2] = ACTIONS(1486),
    [aux_sym_if_statement_token1] = ACTIONS(1486),
    [aux_sym_if_statement_token2] = ACTIONS(1486),
    [aux_sym_else_if_clause_token1] = ACTIONS(1486),
    [aux_sym_else_clause_token1] = ACTIONS(1486),
    [aux_sym_match_expression_token1] = ACTIONS(1486),
    [aux_sym_match_default_expression_token1] = ACTIONS(1486),
    [aux_sym_switch_statement_token1] = ACTIONS(1486),
    [aux_sym_switch_block_token1] = ACTIONS(1486),
    [anon_sym_PLUS] = ACTIONS(1486),
    [anon_sym_DASH] = ACTIONS(1486),
    [anon_sym_TILDE] = ACTIONS(1484),
    [anon_sym_BANG] = ACTIONS(1484),
    [anon_sym_AT] = ACTIONS(1484),
    [aux_sym_clone_expression_token1] = ACTIONS(1486),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1486),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1486),
    [anon_sym_DASH_DASH] = ACTIONS(1484),
    [anon_sym_PLUS_PLUS] = ACTIONS(1484),
    [aux_sym__list_destructing_token1] = ACTIONS(1486),
    [anon_sym_LBRACK] = ACTIONS(1484),
    [anon_sym_self] = ACTIONS(1486),
    [anon_sym_parent] = ACTIONS(1486),
    [aux_sym__argument_name_token1] = ACTIONS(1486),
    [aux_sym__argument_name_token2] = ACTIONS(1486),
    [anon_sym_POUND_LBRACK] = ACTIONS(1484),
    [aux_sym_encapsed_string_token1] = ACTIONS(1484),
    [anon_sym_DQUOTE] = ACTIONS(1484),
    [aux_sym_string_token1] = ACTIONS(1484),
    [anon_sym_SQUOTE] = ACTIONS(1484),
    [anon_sym_LT_LT_LT] = ACTIONS(1484),
    [anon_sym_BQUOTE] = ACTIONS(1484),
    [anon_sym_DOLLAR] = ACTIONS(1484),
    [aux_sym_yield_expression_token1] = ACTIONS(1486),
    [aux_sym_include_expression_token1] = ACTIONS(1486),
    [aux_sym_include_once_expression_token1] = ACTIONS(1486),
    [aux_sym_require_expression_token1] = ACTIONS(1486),
    [aux_sym_require_once_expression_token1] = ACTIONS(1486),
    [sym_comment] = ACTIONS(5),
  },
  [567] = {
    [sym_text_interpolation] = STATE(567),
    [ts_builtin_sym_end] = ACTIONS(1488),
    [sym_name] = ACTIONS(1490),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1488),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1490),
    [aux_sym_global_declaration_token1] = ACTIONS(1490),
    [aux_sym_namespace_definition_token1] = ACTIONS(1490),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1490),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1490),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1490),
    [anon_sym_BSLASH] = ACTIONS(1488),
    [anon_sym_LBRACE] = ACTIONS(1488),
    [anon_sym_RBRACE] = ACTIONS(1488),
    [aux_sym_trait_declaration_token1] = ACTIONS(1490),
    [aux_sym_interface_declaration_token1] = ACTIONS(1490),
    [aux_sym_enum_declaration_token1] = ACTIONS(1490),
    [aux_sym_enum_case_token1] = ACTIONS(1490),
    [aux_sym_class_declaration_token1] = ACTIONS(1490),
    [aux_sym_final_modifier_token1] = ACTIONS(1490),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1490),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1490),
    [sym_var_modifier] = ACTIONS(1490),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1490),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1490),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1490),
    [anon_sym_LPAREN2] = ACTIONS(1488),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1490),
    [aux_sym_cast_type_token1] = ACTIONS(1490),
    [aux_sym_echo_statement_token1] = ACTIONS(1490),
    [aux_sym_exit_statement_token1] = ACTIONS(1490),
    [anon_sym_unset] = ACTIONS(1490),
    [aux_sym_declare_statement_token1] = ACTIONS(1490),
    [aux_sym_declare_statement_token2] = ACTIONS(1490),
    [sym_float] = ACTIONS(1490),
    [aux_sym_try_statement_token1] = ACTIONS(1490),
    [aux_sym_goto_statement_token1] = ACTIONS(1490),
    [aux_sym_continue_statement_token1] = ACTIONS(1490),
    [aux_sym_break_statement_token1] = ACTIONS(1490),
    [sym_integer] = ACTIONS(1490),
    [aux_sym_return_statement_token1] = ACTIONS(1490),
    [aux_sym_throw_expression_token1] = ACTIONS(1490),
    [aux_sym_while_statement_token1] = ACTIONS(1490),
    [aux_sym_while_statement_token2] = ACTIONS(1490),
    [aux_sym_do_statement_token1] = ACTIONS(1490),
    [aux_sym_for_statement_token1] = ACTIONS(1490),
    [aux_sym_for_statement_token2] = ACTIONS(1490),
    [aux_sym_foreach_statement_token1] = ACTIONS(1490),
    [aux_sym_foreach_statement_token2] = ACTIONS(1490),
    [aux_sym_if_statement_token1] = ACTIONS(1490),
    [aux_sym_if_statement_token2] = ACTIONS(1490),
    [aux_sym_else_if_clause_token1] = ACTIONS(1490),
    [aux_sym_else_clause_token1] = ACTIONS(1490),
    [aux_sym_match_expression_token1] = ACTIONS(1490),
    [aux_sym_match_default_expression_token1] = ACTIONS(1490),
    [aux_sym_switch_statement_token1] = ACTIONS(1490),
    [aux_sym_switch_block_token1] = ACTIONS(1490),
    [anon_sym_PLUS] = ACTIONS(1490),
    [anon_sym_DASH] = ACTIONS(1490),
    [anon_sym_TILDE] = ACTIONS(1488),
    [anon_sym_BANG] = ACTIONS(1488),
    [anon_sym_AT] = ACTIONS(1488),
    [aux_sym_clone_expression_token1] = ACTIONS(1490),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1490),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1490),
    [anon_sym_DASH_DASH] = ACTIONS(1488),
    [anon_sym_PLUS_PLUS] = ACTIONS(1488),
    [aux_sym__list_destructing_token1] = ACTIONS(1490),
    [anon_sym_LBRACK] = ACTIONS(1488),
    [anon_sym_self] = ACTIONS(1490),
    [anon_sym_parent] = ACTIONS(1490),
    [aux_sym__argument_name_token1] = ACTIONS(1490),
    [aux_sym__argument_name_token2] = ACTIONS(1490),
    [anon_sym_POUND_LBRACK] = ACTIONS(1488),
    [aux_sym_encapsed_string_token1] = ACTIONS(1488),
    [anon_sym_DQUOTE] = ACTIONS(1488),
    [aux_sym_string_token1] = ACTIONS(1488),
    [anon_sym_SQUOTE] = ACTIONS(1488),
    [anon_sym_LT_LT_LT] = ACTIONS(1488),
    [anon_sym_BQUOTE] = ACTIONS(1488),
    [anon_sym_DOLLAR] = ACTIONS(1488),
    [aux_sym_yield_expression_token1] = ACTIONS(1490),
    [aux_sym_include_expression_token1] = ACTIONS(1490),
    [aux_sym_include_once_expression_token1] = ACTIONS(1490),
    [aux_sym_require_expression_token1] = ACTIONS(1490),
    [aux_sym_require_once_expression_token1] = ACTIONS(1490),
    [sym_comment] = ACTIONS(5),
  },
  [568] = {
    [sym_text_interpolation] = STATE(568),
    [ts_builtin_sym_end] = ACTIONS(1492),
    [sym_name] = ACTIONS(1494),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1492),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1494),
    [aux_sym_global_declaration_token1] = ACTIONS(1494),
    [aux_sym_namespace_definition_token1] = ACTIONS(1494),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1494),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1494),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1494),
    [anon_sym_BSLASH] = ACTIONS(1492),
    [anon_sym_LBRACE] = ACTIONS(1492),
    [anon_sym_RBRACE] = ACTIONS(1492),
    [aux_sym_trait_declaration_token1] = ACTIONS(1494),
    [aux_sym_interface_declaration_token1] = ACTIONS(1494),
    [aux_sym_enum_declaration_token1] = ACTIONS(1494),
    [aux_sym_enum_case_token1] = ACTIONS(1494),
    [aux_sym_class_declaration_token1] = ACTIONS(1494),
    [aux_sym_final_modifier_token1] = ACTIONS(1494),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1494),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1494),
    [sym_var_modifier] = ACTIONS(1494),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1494),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1494),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1494),
    [anon_sym_LPAREN2] = ACTIONS(1492),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1494),
    [aux_sym_cast_type_token1] = ACTIONS(1494),
    [aux_sym_echo_statement_token1] = ACTIONS(1494),
    [aux_sym_exit_statement_token1] = ACTIONS(1494),
    [anon_sym_unset] = ACTIONS(1494),
    [aux_sym_declare_statement_token1] = ACTIONS(1494),
    [aux_sym_declare_statement_token2] = ACTIONS(1494),
    [sym_float] = ACTIONS(1494),
    [aux_sym_try_statement_token1] = ACTIONS(1494),
    [aux_sym_goto_statement_token1] = ACTIONS(1494),
    [aux_sym_continue_statement_token1] = ACTIONS(1494),
    [aux_sym_break_statement_token1] = ACTIONS(1494),
    [sym_integer] = ACTIONS(1494),
    [aux_sym_return_statement_token1] = ACTIONS(1494),
    [aux_sym_throw_expression_token1] = ACTIONS(1494),
    [aux_sym_while_statement_token1] = ACTIONS(1494),
    [aux_sym_while_statement_token2] = ACTIONS(1494),
    [aux_sym_do_statement_token1] = ACTIONS(1494),
    [aux_sym_for_statement_token1] = ACTIONS(1494),
    [aux_sym_for_statement_token2] = ACTIONS(1494),
    [aux_sym_foreach_statement_token1] = ACTIONS(1494),
    [aux_sym_foreach_statement_token2] = ACTIONS(1494),
    [aux_sym_if_statement_token1] = ACTIONS(1494),
    [aux_sym_if_statement_token2] = ACTIONS(1494),
    [aux_sym_else_if_clause_token1] = ACTIONS(1494),
    [aux_sym_else_clause_token1] = ACTIONS(1494),
    [aux_sym_match_expression_token1] = ACTIONS(1494),
    [aux_sym_match_default_expression_token1] = ACTIONS(1494),
    [aux_sym_switch_statement_token1] = ACTIONS(1494),
    [aux_sym_switch_block_token1] = ACTIONS(1494),
    [anon_sym_PLUS] = ACTIONS(1494),
    [anon_sym_DASH] = ACTIONS(1494),
    [anon_sym_TILDE] = ACTIONS(1492),
    [anon_sym_BANG] = ACTIONS(1492),
    [anon_sym_AT] = ACTIONS(1492),
    [aux_sym_clone_expression_token1] = ACTIONS(1494),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1494),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1494),
    [anon_sym_DASH_DASH] = ACTIONS(1492),
    [anon_sym_PLUS_PLUS] = ACTIONS(1492),
    [aux_sym__list_destructing_token1] = ACTIONS(1494),
    [anon_sym_LBRACK] = ACTIONS(1492),
    [anon_sym_self] = ACTIONS(1494),
    [anon_sym_parent] = ACTIONS(1494),
    [aux_sym__argument_name_token1] = ACTIONS(1494),
    [aux_sym__argument_name_token2] = ACTIONS(1494),
    [anon_sym_POUND_LBRACK] = ACTIONS(1492),
    [aux_sym_encapsed_string_token1] = ACTIONS(1492),
    [anon_sym_DQUOTE] = ACTIONS(1492),
    [aux_sym_string_token1] = ACTIONS(1492),
    [anon_sym_SQUOTE] = ACTIONS(1492),
    [anon_sym_LT_LT_LT] = ACTIONS(1492),
    [anon_sym_BQUOTE] = ACTIONS(1492),
    [anon_sym_DOLLAR] = ACTIONS(1492),
    [aux_sym_yield_expression_token1] = ACTIONS(1494),
    [aux_sym_include_expression_token1] = ACTIONS(1494),
    [aux_sym_include_once_expression_token1] = ACTIONS(1494),
    [aux_sym_require_expression_token1] = ACTIONS(1494),
    [aux_sym_require_once_expression_token1] = ACTIONS(1494),
    [sym_comment] = ACTIONS(5),
  },
  [569] = {
    [sym_text_interpolation] = STATE(569),
    [ts_builtin_sym_end] = ACTIONS(1496),
    [sym_name] = ACTIONS(1498),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1496),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1498),
    [aux_sym_global_declaration_token1] = ACTIONS(1498),
    [aux_sym_namespace_definition_token1] = ACTIONS(1498),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1498),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1498),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1498),
    [anon_sym_BSLASH] = ACTIONS(1496),
    [anon_sym_LBRACE] = ACTIONS(1496),
    [anon_sym_RBRACE] = ACTIONS(1496),
    [aux_sym_trait_declaration_token1] = ACTIONS(1498),
    [aux_sym_interface_declaration_token1] = ACTIONS(1498),
    [aux_sym_enum_declaration_token1] = ACTIONS(1498),
    [aux_sym_enum_case_token1] = ACTIONS(1498),
    [aux_sym_class_declaration_token1] = ACTIONS(1498),
    [aux_sym_final_modifier_token1] = ACTIONS(1498),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1498),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1498),
    [sym_var_modifier] = ACTIONS(1498),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1498),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1498),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1498),
    [anon_sym_LPAREN2] = ACTIONS(1496),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1498),
    [aux_sym_cast_type_token1] = ACTIONS(1498),
    [aux_sym_echo_statement_token1] = ACTIONS(1498),
    [aux_sym_exit_statement_token1] = ACTIONS(1498),
    [anon_sym_unset] = ACTIONS(1498),
    [aux_sym_declare_statement_token1] = ACTIONS(1498),
    [aux_sym_declare_statement_token2] = ACTIONS(1498),
    [sym_float] = ACTIONS(1498),
    [aux_sym_try_statement_token1] = ACTIONS(1498),
    [aux_sym_goto_statement_token1] = ACTIONS(1498),
    [aux_sym_continue_statement_token1] = ACTIONS(1498),
    [aux_sym_break_statement_token1] = ACTIONS(1498),
    [sym_integer] = ACTIONS(1498),
    [aux_sym_return_statement_token1] = ACTIONS(1498),
    [aux_sym_throw_expression_token1] = ACTIONS(1498),
    [aux_sym_while_statement_token1] = ACTIONS(1498),
    [aux_sym_while_statement_token2] = ACTIONS(1498),
    [aux_sym_do_statement_token1] = ACTIONS(1498),
    [aux_sym_for_statement_token1] = ACTIONS(1498),
    [aux_sym_for_statement_token2] = ACTIONS(1498),
    [aux_sym_foreach_statement_token1] = ACTIONS(1498),
    [aux_sym_foreach_statement_token2] = ACTIONS(1498),
    [aux_sym_if_statement_token1] = ACTIONS(1498),
    [aux_sym_if_statement_token2] = ACTIONS(1498),
    [aux_sym_else_if_clause_token1] = ACTIONS(1498),
    [aux_sym_else_clause_token1] = ACTIONS(1498),
    [aux_sym_match_expression_token1] = ACTIONS(1498),
    [aux_sym_match_default_expression_token1] = ACTIONS(1498),
    [aux_sym_switch_statement_token1] = ACTIONS(1498),
    [aux_sym_switch_block_token1] = ACTIONS(1498),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_TILDE] = ACTIONS(1496),
    [anon_sym_BANG] = ACTIONS(1496),
    [anon_sym_AT] = ACTIONS(1496),
    [aux_sym_clone_expression_token1] = ACTIONS(1498),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1498),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1498),
    [anon_sym_DASH_DASH] = ACTIONS(1496),
    [anon_sym_PLUS_PLUS] = ACTIONS(1496),
    [aux_sym__list_destructing_token1] = ACTIONS(1498),
    [anon_sym_LBRACK] = ACTIONS(1496),
    [anon_sym_self] = ACTIONS(1498),
    [anon_sym_parent] = ACTIONS(1498),
    [aux_sym__argument_name_token1] = ACTIONS(1498),
    [aux_sym__argument_name_token2] = ACTIONS(1498),
    [anon_sym_POUND_LBRACK] = ACTIONS(1496),
    [aux_sym_encapsed_string_token1] = ACTIONS(1496),
    [anon_sym_DQUOTE] = ACTIONS(1496),
    [aux_sym_string_token1] = ACTIONS(1496),
    [anon_sym_SQUOTE] = ACTIONS(1496),
    [anon_sym_LT_LT_LT] = ACTIONS(1496),
    [anon_sym_BQUOTE] = ACTIONS(1496),
    [anon_sym_DOLLAR] = ACTIONS(1496),
    [aux_sym_yield_expression_token1] = ACTIONS(1498),
    [aux_sym_include_expression_token1] = ACTIONS(1498),
    [aux_sym_include_once_expression_token1] = ACTIONS(1498),
    [aux_sym_require_expression_token1] = ACTIONS(1498),
    [aux_sym_require_once_expression_token1] = ACTIONS(1498),
    [sym_comment] = ACTIONS(5),
  },
  [570] = {
    [sym_text_interpolation] = STATE(570),
    [ts_builtin_sym_end] = ACTIONS(1500),
    [sym_name] = ACTIONS(1502),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1500),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1502),
    [aux_sym_global_declaration_token1] = ACTIONS(1502),
    [aux_sym_namespace_definition_token1] = ACTIONS(1502),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1502),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1502),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1502),
    [anon_sym_BSLASH] = ACTIONS(1500),
    [anon_sym_LBRACE] = ACTIONS(1500),
    [anon_sym_RBRACE] = ACTIONS(1500),
    [aux_sym_trait_declaration_token1] = ACTIONS(1502),
    [aux_sym_interface_declaration_token1] = ACTIONS(1502),
    [aux_sym_enum_declaration_token1] = ACTIONS(1502),
    [aux_sym_enum_case_token1] = ACTIONS(1502),
    [aux_sym_class_declaration_token1] = ACTIONS(1502),
    [aux_sym_final_modifier_token1] = ACTIONS(1502),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1502),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1502),
    [sym_var_modifier] = ACTIONS(1502),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1502),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1502),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1502),
    [anon_sym_LPAREN2] = ACTIONS(1500),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1502),
    [aux_sym_cast_type_token1] = ACTIONS(1502),
    [aux_sym_echo_statement_token1] = ACTIONS(1502),
    [aux_sym_exit_statement_token1] = ACTIONS(1502),
    [anon_sym_unset] = ACTIONS(1502),
    [aux_sym_declare_statement_token1] = ACTIONS(1502),
    [aux_sym_declare_statement_token2] = ACTIONS(1502),
    [sym_float] = ACTIONS(1502),
    [aux_sym_try_statement_token1] = ACTIONS(1502),
    [aux_sym_goto_statement_token1] = ACTIONS(1502),
    [aux_sym_continue_statement_token1] = ACTIONS(1502),
    [aux_sym_break_statement_token1] = ACTIONS(1502),
    [sym_integer] = ACTIONS(1502),
    [aux_sym_return_statement_token1] = ACTIONS(1502),
    [aux_sym_throw_expression_token1] = ACTIONS(1502),
    [aux_sym_while_statement_token1] = ACTIONS(1502),
    [aux_sym_while_statement_token2] = ACTIONS(1502),
    [aux_sym_do_statement_token1] = ACTIONS(1502),
    [aux_sym_for_statement_token1] = ACTIONS(1502),
    [aux_sym_for_statement_token2] = ACTIONS(1502),
    [aux_sym_foreach_statement_token1] = ACTIONS(1502),
    [aux_sym_foreach_statement_token2] = ACTIONS(1502),
    [aux_sym_if_statement_token1] = ACTIONS(1502),
    [aux_sym_if_statement_token2] = ACTIONS(1502),
    [aux_sym_else_if_clause_token1] = ACTIONS(1502),
    [aux_sym_else_clause_token1] = ACTIONS(1502),
    [aux_sym_match_expression_token1] = ACTIONS(1502),
    [aux_sym_match_default_expression_token1] = ACTIONS(1502),
    [aux_sym_switch_statement_token1] = ACTIONS(1502),
    [aux_sym_switch_block_token1] = ACTIONS(1502),
    [anon_sym_PLUS] = ACTIONS(1502),
    [anon_sym_DASH] = ACTIONS(1502),
    [anon_sym_TILDE] = ACTIONS(1500),
    [anon_sym_BANG] = ACTIONS(1500),
    [anon_sym_AT] = ACTIONS(1500),
    [aux_sym_clone_expression_token1] = ACTIONS(1502),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1502),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1502),
    [anon_sym_DASH_DASH] = ACTIONS(1500),
    [anon_sym_PLUS_PLUS] = ACTIONS(1500),
    [aux_sym__list_destructing_token1] = ACTIONS(1502),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_self] = ACTIONS(1502),
    [anon_sym_parent] = ACTIONS(1502),
    [aux_sym__argument_name_token1] = ACTIONS(1502),
    [aux_sym__argument_name_token2] = ACTIONS(1502),
    [anon_sym_POUND_LBRACK] = ACTIONS(1500),
    [aux_sym_encapsed_string_token1] = ACTIONS(1500),
    [anon_sym_DQUOTE] = ACTIONS(1500),
    [aux_sym_string_token1] = ACTIONS(1500),
    [anon_sym_SQUOTE] = ACTIONS(1500),
    [anon_sym_LT_LT_LT] = ACTIONS(1500),
    [anon_sym_BQUOTE] = ACTIONS(1500),
    [anon_sym_DOLLAR] = ACTIONS(1500),
    [aux_sym_yield_expression_token1] = ACTIONS(1502),
    [aux_sym_include_expression_token1] = ACTIONS(1502),
    [aux_sym_include_once_expression_token1] = ACTIONS(1502),
    [aux_sym_require_expression_token1] = ACTIONS(1502),
    [aux_sym_require_once_expression_token1] = ACTIONS(1502),
    [sym_comment] = ACTIONS(5),
  },
  [571] = {
    [sym_text_interpolation] = STATE(571),
    [ts_builtin_sym_end] = ACTIONS(1504),
    [sym_name] = ACTIONS(1506),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1504),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1506),
    [aux_sym_global_declaration_token1] = ACTIONS(1506),
    [aux_sym_namespace_definition_token1] = ACTIONS(1506),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1506),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1506),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1506),
    [anon_sym_BSLASH] = ACTIONS(1504),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_RBRACE] = ACTIONS(1504),
    [aux_sym_trait_declaration_token1] = ACTIONS(1506),
    [aux_sym_interface_declaration_token1] = ACTIONS(1506),
    [aux_sym_enum_declaration_token1] = ACTIONS(1506),
    [aux_sym_enum_case_token1] = ACTIONS(1506),
    [aux_sym_class_declaration_token1] = ACTIONS(1506),
    [aux_sym_final_modifier_token1] = ACTIONS(1506),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1506),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1506),
    [sym_var_modifier] = ACTIONS(1506),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1506),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1506),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1506),
    [anon_sym_LPAREN2] = ACTIONS(1504),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1506),
    [aux_sym_cast_type_token1] = ACTIONS(1506),
    [aux_sym_echo_statement_token1] = ACTIONS(1506),
    [aux_sym_exit_statement_token1] = ACTIONS(1506),
    [anon_sym_unset] = ACTIONS(1506),
    [aux_sym_declare_statement_token1] = ACTIONS(1506),
    [aux_sym_declare_statement_token2] = ACTIONS(1506),
    [sym_float] = ACTIONS(1506),
    [aux_sym_try_statement_token1] = ACTIONS(1506),
    [aux_sym_goto_statement_token1] = ACTIONS(1506),
    [aux_sym_continue_statement_token1] = ACTIONS(1506),
    [aux_sym_break_statement_token1] = ACTIONS(1506),
    [sym_integer] = ACTIONS(1506),
    [aux_sym_return_statement_token1] = ACTIONS(1506),
    [aux_sym_throw_expression_token1] = ACTIONS(1506),
    [aux_sym_while_statement_token1] = ACTIONS(1506),
    [aux_sym_while_statement_token2] = ACTIONS(1506),
    [aux_sym_do_statement_token1] = ACTIONS(1506),
    [aux_sym_for_statement_token1] = ACTIONS(1506),
    [aux_sym_for_statement_token2] = ACTIONS(1506),
    [aux_sym_foreach_statement_token1] = ACTIONS(1506),
    [aux_sym_foreach_statement_token2] = ACTIONS(1506),
    [aux_sym_if_statement_token1] = ACTIONS(1506),
    [aux_sym_if_statement_token2] = ACTIONS(1506),
    [aux_sym_else_if_clause_token1] = ACTIONS(1506),
    [aux_sym_else_clause_token1] = ACTIONS(1506),
    [aux_sym_match_expression_token1] = ACTIONS(1506),
    [aux_sym_match_default_expression_token1] = ACTIONS(1506),
    [aux_sym_switch_statement_token1] = ACTIONS(1506),
    [aux_sym_switch_block_token1] = ACTIONS(1506),
    [anon_sym_PLUS] = ACTIONS(1506),
    [anon_sym_DASH] = ACTIONS(1506),
    [anon_sym_TILDE] = ACTIONS(1504),
    [anon_sym_BANG] = ACTIONS(1504),
    [anon_sym_AT] = ACTIONS(1504),
    [aux_sym_clone_expression_token1] = ACTIONS(1506),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1506),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1506),
    [anon_sym_DASH_DASH] = ACTIONS(1504),
    [anon_sym_PLUS_PLUS] = ACTIONS(1504),
    [aux_sym__list_destructing_token1] = ACTIONS(1506),
    [anon_sym_LBRACK] = ACTIONS(1504),
    [anon_sym_self] = ACTIONS(1506),
    [anon_sym_parent] = ACTIONS(1506),
    [aux_sym__argument_name_token1] = ACTIONS(1506),
    [aux_sym__argument_name_token2] = ACTIONS(1506),
    [anon_sym_POUND_LBRACK] = ACTIONS(1504),
    [aux_sym_encapsed_string_token1] = ACTIONS(1504),
    [anon_sym_DQUOTE] = ACTIONS(1504),
    [aux_sym_string_token1] = ACTIONS(1504),
    [anon_sym_SQUOTE] = ACTIONS(1504),
    [anon_sym_LT_LT_LT] = ACTIONS(1504),
    [anon_sym_BQUOTE] = ACTIONS(1504),
    [anon_sym_DOLLAR] = ACTIONS(1504),
    [aux_sym_yield_expression_token1] = ACTIONS(1506),
    [aux_sym_include_expression_token1] = ACTIONS(1506),
    [aux_sym_include_once_expression_token1] = ACTIONS(1506),
    [aux_sym_require_expression_token1] = ACTIONS(1506),
    [aux_sym_require_once_expression_token1] = ACTIONS(1506),
    [sym_comment] = ACTIONS(5),
  },
  [572] = {
    [sym_text_interpolation] = STATE(572),
    [ts_builtin_sym_end] = ACTIONS(1508),
    [sym_name] = ACTIONS(1510),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1508),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1510),
    [aux_sym_global_declaration_token1] = ACTIONS(1510),
    [aux_sym_namespace_definition_token1] = ACTIONS(1510),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1510),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1510),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1510),
    [anon_sym_BSLASH] = ACTIONS(1508),
    [anon_sym_LBRACE] = ACTIONS(1508),
    [anon_sym_RBRACE] = ACTIONS(1508),
    [aux_sym_trait_declaration_token1] = ACTIONS(1510),
    [aux_sym_interface_declaration_token1] = ACTIONS(1510),
    [aux_sym_enum_declaration_token1] = ACTIONS(1510),
    [aux_sym_enum_case_token1] = ACTIONS(1510),
    [aux_sym_class_declaration_token1] = ACTIONS(1510),
    [aux_sym_final_modifier_token1] = ACTIONS(1510),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1510),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1510),
    [sym_var_modifier] = ACTIONS(1510),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1510),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1510),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1510),
    [anon_sym_LPAREN2] = ACTIONS(1508),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1510),
    [aux_sym_cast_type_token1] = ACTIONS(1510),
    [aux_sym_echo_statement_token1] = ACTIONS(1510),
    [aux_sym_exit_statement_token1] = ACTIONS(1510),
    [anon_sym_unset] = ACTIONS(1510),
    [aux_sym_declare_statement_token1] = ACTIONS(1510),
    [aux_sym_declare_statement_token2] = ACTIONS(1510),
    [sym_float] = ACTIONS(1510),
    [aux_sym_try_statement_token1] = ACTIONS(1510),
    [aux_sym_goto_statement_token1] = ACTIONS(1510),
    [aux_sym_continue_statement_token1] = ACTIONS(1510),
    [aux_sym_break_statement_token1] = ACTIONS(1510),
    [sym_integer] = ACTIONS(1510),
    [aux_sym_return_statement_token1] = ACTIONS(1510),
    [aux_sym_throw_expression_token1] = ACTIONS(1510),
    [aux_sym_while_statement_token1] = ACTIONS(1510),
    [aux_sym_while_statement_token2] = ACTIONS(1510),
    [aux_sym_do_statement_token1] = ACTIONS(1510),
    [aux_sym_for_statement_token1] = ACTIONS(1510),
    [aux_sym_for_statement_token2] = ACTIONS(1510),
    [aux_sym_foreach_statement_token1] = ACTIONS(1510),
    [aux_sym_foreach_statement_token2] = ACTIONS(1510),
    [aux_sym_if_statement_token1] = ACTIONS(1510),
    [aux_sym_if_statement_token2] = ACTIONS(1510),
    [aux_sym_else_if_clause_token1] = ACTIONS(1510),
    [aux_sym_else_clause_token1] = ACTIONS(1510),
    [aux_sym_match_expression_token1] = ACTIONS(1510),
    [aux_sym_match_default_expression_token1] = ACTIONS(1510),
    [aux_sym_switch_statement_token1] = ACTIONS(1510),
    [aux_sym_switch_block_token1] = ACTIONS(1510),
    [anon_sym_PLUS] = ACTIONS(1510),
    [anon_sym_DASH] = ACTIONS(1510),
    [anon_sym_TILDE] = ACTIONS(1508),
    [anon_sym_BANG] = ACTIONS(1508),
    [anon_sym_AT] = ACTIONS(1508),
    [aux_sym_clone_expression_token1] = ACTIONS(1510),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1510),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1510),
    [anon_sym_DASH_DASH] = ACTIONS(1508),
    [anon_sym_PLUS_PLUS] = ACTIONS(1508),
    [aux_sym__list_destructing_token1] = ACTIONS(1510),
    [anon_sym_LBRACK] = ACTIONS(1508),
    [anon_sym_self] = ACTIONS(1510),
    [anon_sym_parent] = ACTIONS(1510),
    [aux_sym__argument_name_token1] = ACTIONS(1510),
    [aux_sym__argument_name_token2] = ACTIONS(1510),
    [anon_sym_POUND_LBRACK] = ACTIONS(1508),
    [aux_sym_encapsed_string_token1] = ACTIONS(1508),
    [anon_sym_DQUOTE] = ACTIONS(1508),
    [aux_sym_string_token1] = ACTIONS(1508),
    [anon_sym_SQUOTE] = ACTIONS(1508),
    [anon_sym_LT_LT_LT] = ACTIONS(1508),
    [anon_sym_BQUOTE] = ACTIONS(1508),
    [anon_sym_DOLLAR] = ACTIONS(1508),
    [aux_sym_yield_expression_token1] = ACTIONS(1510),
    [aux_sym_include_expression_token1] = ACTIONS(1510),
    [aux_sym_include_once_expression_token1] = ACTIONS(1510),
    [aux_sym_require_expression_token1] = ACTIONS(1510),
    [aux_sym_require_once_expression_token1] = ACTIONS(1510),
    [sym_comment] = ACTIONS(5),
  },
  [573] = {
    [sym_text_interpolation] = STATE(573),
    [ts_builtin_sym_end] = ACTIONS(1512),
    [sym_name] = ACTIONS(1514),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1512),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1514),
    [aux_sym_global_declaration_token1] = ACTIONS(1514),
    [aux_sym_namespace_definition_token1] = ACTIONS(1514),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1514),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1514),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1514),
    [anon_sym_BSLASH] = ACTIONS(1512),
    [anon_sym_LBRACE] = ACTIONS(1512),
    [anon_sym_RBRACE] = ACTIONS(1512),
    [aux_sym_trait_declaration_token1] = ACTIONS(1514),
    [aux_sym_interface_declaration_token1] = ACTIONS(1514),
    [aux_sym_enum_declaration_token1] = ACTIONS(1514),
    [aux_sym_enum_case_token1] = ACTIONS(1514),
    [aux_sym_class_declaration_token1] = ACTIONS(1514),
    [aux_sym_final_modifier_token1] = ACTIONS(1514),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1514),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1514),
    [sym_var_modifier] = ACTIONS(1514),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1514),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1514),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1514),
    [anon_sym_LPAREN2] = ACTIONS(1512),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1514),
    [aux_sym_cast_type_token1] = ACTIONS(1514),
    [aux_sym_echo_statement_token1] = ACTIONS(1514),
    [aux_sym_exit_statement_token1] = ACTIONS(1514),
    [anon_sym_unset] = ACTIONS(1514),
    [aux_sym_declare_statement_token1] = ACTIONS(1514),
    [aux_sym_declare_statement_token2] = ACTIONS(1514),
    [sym_float] = ACTIONS(1514),
    [aux_sym_try_statement_token1] = ACTIONS(1514),
    [aux_sym_goto_statement_token1] = ACTIONS(1514),
    [aux_sym_continue_statement_token1] = ACTIONS(1514),
    [aux_sym_break_statement_token1] = ACTIONS(1514),
    [sym_integer] = ACTIONS(1514),
    [aux_sym_return_statement_token1] = ACTIONS(1514),
    [aux_sym_throw_expression_token1] = ACTIONS(1514),
    [aux_sym_while_statement_token1] = ACTIONS(1514),
    [aux_sym_while_statement_token2] = ACTIONS(1514),
    [aux_sym_do_statement_token1] = ACTIONS(1514),
    [aux_sym_for_statement_token1] = ACTIONS(1514),
    [aux_sym_for_statement_token2] = ACTIONS(1514),
    [aux_sym_foreach_statement_token1] = ACTIONS(1514),
    [aux_sym_foreach_statement_token2] = ACTIONS(1514),
    [aux_sym_if_statement_token1] = ACTIONS(1514),
    [aux_sym_if_statement_token2] = ACTIONS(1514),
    [aux_sym_else_if_clause_token1] = ACTIONS(1514),
    [aux_sym_else_clause_token1] = ACTIONS(1514),
    [aux_sym_match_expression_token1] = ACTIONS(1514),
    [aux_sym_match_default_expression_token1] = ACTIONS(1514),
    [aux_sym_switch_statement_token1] = ACTIONS(1514),
    [aux_sym_switch_block_token1] = ACTIONS(1514),
    [anon_sym_PLUS] = ACTIONS(1514),
    [anon_sym_DASH] = ACTIONS(1514),
    [anon_sym_TILDE] = ACTIONS(1512),
    [anon_sym_BANG] = ACTIONS(1512),
    [anon_sym_AT] = ACTIONS(1512),
    [aux_sym_clone_expression_token1] = ACTIONS(1514),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1514),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1514),
    [anon_sym_DASH_DASH] = ACTIONS(1512),
    [anon_sym_PLUS_PLUS] = ACTIONS(1512),
    [aux_sym__list_destructing_token1] = ACTIONS(1514),
    [anon_sym_LBRACK] = ACTIONS(1512),
    [anon_sym_self] = ACTIONS(1514),
    [anon_sym_parent] = ACTIONS(1514),
    [aux_sym__argument_name_token1] = ACTIONS(1514),
    [aux_sym__argument_name_token2] = ACTIONS(1514),
    [anon_sym_POUND_LBRACK] = ACTIONS(1512),
    [aux_sym_encapsed_string_token1] = ACTIONS(1512),
    [anon_sym_DQUOTE] = ACTIONS(1512),
    [aux_sym_string_token1] = ACTIONS(1512),
    [anon_sym_SQUOTE] = ACTIONS(1512),
    [anon_sym_LT_LT_LT] = ACTIONS(1512),
    [anon_sym_BQUOTE] = ACTIONS(1512),
    [anon_sym_DOLLAR] = ACTIONS(1512),
    [aux_sym_yield_expression_token1] = ACTIONS(1514),
    [aux_sym_include_expression_token1] = ACTIONS(1514),
    [aux_sym_include_once_expression_token1] = ACTIONS(1514),
    [aux_sym_require_expression_token1] = ACTIONS(1514),
    [aux_sym_require_once_expression_token1] = ACTIONS(1514),
    [sym_comment] = ACTIONS(5),
  },
  [574] = {
    [sym_text_interpolation] = STATE(574),
    [ts_builtin_sym_end] = ACTIONS(1516),
    [sym_name] = ACTIONS(1518),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1516),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1518),
    [aux_sym_global_declaration_token1] = ACTIONS(1518),
    [aux_sym_namespace_definition_token1] = ACTIONS(1518),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1518),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1518),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1518),
    [anon_sym_BSLASH] = ACTIONS(1516),
    [anon_sym_LBRACE] = ACTIONS(1516),
    [anon_sym_RBRACE] = ACTIONS(1516),
    [aux_sym_trait_declaration_token1] = ACTIONS(1518),
    [aux_sym_interface_declaration_token1] = ACTIONS(1518),
    [aux_sym_enum_declaration_token1] = ACTIONS(1518),
    [aux_sym_enum_case_token1] = ACTIONS(1518),
    [aux_sym_class_declaration_token1] = ACTIONS(1518),
    [aux_sym_final_modifier_token1] = ACTIONS(1518),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1518),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1518),
    [sym_var_modifier] = ACTIONS(1518),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1518),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1518),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1518),
    [anon_sym_LPAREN2] = ACTIONS(1516),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1518),
    [aux_sym_cast_type_token1] = ACTIONS(1518),
    [aux_sym_echo_statement_token1] = ACTIONS(1518),
    [aux_sym_exit_statement_token1] = ACTIONS(1518),
    [anon_sym_unset] = ACTIONS(1518),
    [aux_sym_declare_statement_token1] = ACTIONS(1518),
    [aux_sym_declare_statement_token2] = ACTIONS(1518),
    [sym_float] = ACTIONS(1518),
    [aux_sym_try_statement_token1] = ACTIONS(1518),
    [aux_sym_goto_statement_token1] = ACTIONS(1518),
    [aux_sym_continue_statement_token1] = ACTIONS(1518),
    [aux_sym_break_statement_token1] = ACTIONS(1518),
    [sym_integer] = ACTIONS(1518),
    [aux_sym_return_statement_token1] = ACTIONS(1518),
    [aux_sym_throw_expression_token1] = ACTIONS(1518),
    [aux_sym_while_statement_token1] = ACTIONS(1518),
    [aux_sym_while_statement_token2] = ACTIONS(1518),
    [aux_sym_do_statement_token1] = ACTIONS(1518),
    [aux_sym_for_statement_token1] = ACTIONS(1518),
    [aux_sym_for_statement_token2] = ACTIONS(1518),
    [aux_sym_foreach_statement_token1] = ACTIONS(1518),
    [aux_sym_foreach_statement_token2] = ACTIONS(1518),
    [aux_sym_if_statement_token1] = ACTIONS(1518),
    [aux_sym_if_statement_token2] = ACTIONS(1518),
    [aux_sym_else_if_clause_token1] = ACTIONS(1518),
    [aux_sym_else_clause_token1] = ACTIONS(1518),
    [aux_sym_match_expression_token1] = ACTIONS(1518),
    [aux_sym_match_default_expression_token1] = ACTIONS(1518),
    [aux_sym_switch_statement_token1] = ACTIONS(1518),
    [aux_sym_switch_block_token1] = ACTIONS(1518),
    [anon_sym_PLUS] = ACTIONS(1518),
    [anon_sym_DASH] = ACTIONS(1518),
    [anon_sym_TILDE] = ACTIONS(1516),
    [anon_sym_BANG] = ACTIONS(1516),
    [anon_sym_AT] = ACTIONS(1516),
    [aux_sym_clone_expression_token1] = ACTIONS(1518),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1518),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1518),
    [anon_sym_DASH_DASH] = ACTIONS(1516),
    [anon_sym_PLUS_PLUS] = ACTIONS(1516),
    [aux_sym__list_destructing_token1] = ACTIONS(1518),
    [anon_sym_LBRACK] = ACTIONS(1516),
    [anon_sym_self] = ACTIONS(1518),
    [anon_sym_parent] = ACTIONS(1518),
    [aux_sym__argument_name_token1] = ACTIONS(1518),
    [aux_sym__argument_name_token2] = ACTIONS(1518),
    [anon_sym_POUND_LBRACK] = ACTIONS(1516),
    [aux_sym_encapsed_string_token1] = ACTIONS(1516),
    [anon_sym_DQUOTE] = ACTIONS(1516),
    [aux_sym_string_token1] = ACTIONS(1516),
    [anon_sym_SQUOTE] = ACTIONS(1516),
    [anon_sym_LT_LT_LT] = ACTIONS(1516),
    [anon_sym_BQUOTE] = ACTIONS(1516),
    [anon_sym_DOLLAR] = ACTIONS(1516),
    [aux_sym_yield_expression_token1] = ACTIONS(1518),
    [aux_sym_include_expression_token1] = ACTIONS(1518),
    [aux_sym_include_once_expression_token1] = ACTIONS(1518),
    [aux_sym_require_expression_token1] = ACTIONS(1518),
    [aux_sym_require_once_expression_token1] = ACTIONS(1518),
    [sym_comment] = ACTIONS(5),
  },
  [575] = {
    [sym_text_interpolation] = STATE(575),
    [ts_builtin_sym_end] = ACTIONS(1520),
    [sym_name] = ACTIONS(1522),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1520),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1522),
    [aux_sym_global_declaration_token1] = ACTIONS(1522),
    [aux_sym_namespace_definition_token1] = ACTIONS(1522),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1522),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1522),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1522),
    [anon_sym_BSLASH] = ACTIONS(1520),
    [anon_sym_LBRACE] = ACTIONS(1520),
    [anon_sym_RBRACE] = ACTIONS(1520),
    [aux_sym_trait_declaration_token1] = ACTIONS(1522),
    [aux_sym_interface_declaration_token1] = ACTIONS(1522),
    [aux_sym_enum_declaration_token1] = ACTIONS(1522),
    [aux_sym_enum_case_token1] = ACTIONS(1522),
    [aux_sym_class_declaration_token1] = ACTIONS(1522),
    [aux_sym_final_modifier_token1] = ACTIONS(1522),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1522),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1522),
    [sym_var_modifier] = ACTIONS(1522),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1522),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1522),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1522),
    [anon_sym_LPAREN2] = ACTIONS(1520),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1522),
    [aux_sym_cast_type_token1] = ACTIONS(1522),
    [aux_sym_echo_statement_token1] = ACTIONS(1522),
    [aux_sym_exit_statement_token1] = ACTIONS(1522),
    [anon_sym_unset] = ACTIONS(1522),
    [aux_sym_declare_statement_token1] = ACTIONS(1522),
    [aux_sym_declare_statement_token2] = ACTIONS(1522),
    [sym_float] = ACTIONS(1522),
    [aux_sym_try_statement_token1] = ACTIONS(1522),
    [aux_sym_goto_statement_token1] = ACTIONS(1522),
    [aux_sym_continue_statement_token1] = ACTIONS(1522),
    [aux_sym_break_statement_token1] = ACTIONS(1522),
    [sym_integer] = ACTIONS(1522),
    [aux_sym_return_statement_token1] = ACTIONS(1522),
    [aux_sym_throw_expression_token1] = ACTIONS(1522),
    [aux_sym_while_statement_token1] = ACTIONS(1522),
    [aux_sym_while_statement_token2] = ACTIONS(1522),
    [aux_sym_do_statement_token1] = ACTIONS(1522),
    [aux_sym_for_statement_token1] = ACTIONS(1522),
    [aux_sym_for_statement_token2] = ACTIONS(1522),
    [aux_sym_foreach_statement_token1] = ACTIONS(1522),
    [aux_sym_foreach_statement_token2] = ACTIONS(1522),
    [aux_sym_if_statement_token1] = ACTIONS(1522),
    [aux_sym_if_statement_token2] = ACTIONS(1522),
    [aux_sym_else_if_clause_token1] = ACTIONS(1522),
    [aux_sym_else_clause_token1] = ACTIONS(1522),
    [aux_sym_match_expression_token1] = ACTIONS(1522),
    [aux_sym_match_default_expression_token1] = ACTIONS(1522),
    [aux_sym_switch_statement_token1] = ACTIONS(1522),
    [aux_sym_switch_block_token1] = ACTIONS(1522),
    [anon_sym_PLUS] = ACTIONS(1522),
    [anon_sym_DASH] = ACTIONS(1522),
    [anon_sym_TILDE] = ACTIONS(1520),
    [anon_sym_BANG] = ACTIONS(1520),
    [anon_sym_AT] = ACTIONS(1520),
    [aux_sym_clone_expression_token1] = ACTIONS(1522),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1522),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1522),
    [anon_sym_DASH_DASH] = ACTIONS(1520),
    [anon_sym_PLUS_PLUS] = ACTIONS(1520),
    [aux_sym__list_destructing_token1] = ACTIONS(1522),
    [anon_sym_LBRACK] = ACTIONS(1520),
    [anon_sym_self] = ACTIONS(1522),
    [anon_sym_parent] = ACTIONS(1522),
    [aux_sym__argument_name_token1] = ACTIONS(1522),
    [aux_sym__argument_name_token2] = ACTIONS(1522),
    [anon_sym_POUND_LBRACK] = ACTIONS(1520),
    [aux_sym_encapsed_string_token1] = ACTIONS(1520),
    [anon_sym_DQUOTE] = ACTIONS(1520),
    [aux_sym_string_token1] = ACTIONS(1520),
    [anon_sym_SQUOTE] = ACTIONS(1520),
    [anon_sym_LT_LT_LT] = ACTIONS(1520),
    [anon_sym_BQUOTE] = ACTIONS(1520),
    [anon_sym_DOLLAR] = ACTIONS(1520),
    [aux_sym_yield_expression_token1] = ACTIONS(1522),
    [aux_sym_include_expression_token1] = ACTIONS(1522),
    [aux_sym_include_once_expression_token1] = ACTIONS(1522),
    [aux_sym_require_expression_token1] = ACTIONS(1522),
    [aux_sym_require_once_expression_token1] = ACTIONS(1522),
    [sym_comment] = ACTIONS(5),
  },
  [576] = {
    [sym_text_interpolation] = STATE(576),
    [ts_builtin_sym_end] = ACTIONS(1524),
    [sym_name] = ACTIONS(1526),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1524),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1526),
    [aux_sym_global_declaration_token1] = ACTIONS(1526),
    [aux_sym_namespace_definition_token1] = ACTIONS(1526),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1526),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1526),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1526),
    [anon_sym_BSLASH] = ACTIONS(1524),
    [anon_sym_LBRACE] = ACTIONS(1524),
    [anon_sym_RBRACE] = ACTIONS(1524),
    [aux_sym_trait_declaration_token1] = ACTIONS(1526),
    [aux_sym_interface_declaration_token1] = ACTIONS(1526),
    [aux_sym_enum_declaration_token1] = ACTIONS(1526),
    [aux_sym_enum_case_token1] = ACTIONS(1526),
    [aux_sym_class_declaration_token1] = ACTIONS(1526),
    [aux_sym_final_modifier_token1] = ACTIONS(1526),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1526),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1526),
    [sym_var_modifier] = ACTIONS(1526),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1526),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1526),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1526),
    [anon_sym_LPAREN2] = ACTIONS(1524),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1526),
    [aux_sym_cast_type_token1] = ACTIONS(1526),
    [aux_sym_echo_statement_token1] = ACTIONS(1526),
    [aux_sym_exit_statement_token1] = ACTIONS(1526),
    [anon_sym_unset] = ACTIONS(1526),
    [aux_sym_declare_statement_token1] = ACTIONS(1526),
    [aux_sym_declare_statement_token2] = ACTIONS(1526),
    [sym_float] = ACTIONS(1526),
    [aux_sym_try_statement_token1] = ACTIONS(1526),
    [aux_sym_goto_statement_token1] = ACTIONS(1526),
    [aux_sym_continue_statement_token1] = ACTIONS(1526),
    [aux_sym_break_statement_token1] = ACTIONS(1526),
    [sym_integer] = ACTIONS(1526),
    [aux_sym_return_statement_token1] = ACTIONS(1526),
    [aux_sym_throw_expression_token1] = ACTIONS(1526),
    [aux_sym_while_statement_token1] = ACTIONS(1526),
    [aux_sym_while_statement_token2] = ACTIONS(1526),
    [aux_sym_do_statement_token1] = ACTIONS(1526),
    [aux_sym_for_statement_token1] = ACTIONS(1526),
    [aux_sym_for_statement_token2] = ACTIONS(1526),
    [aux_sym_foreach_statement_token1] = ACTIONS(1526),
    [aux_sym_foreach_statement_token2] = ACTIONS(1526),
    [aux_sym_if_statement_token1] = ACTIONS(1526),
    [aux_sym_if_statement_token2] = ACTIONS(1526),
    [aux_sym_else_if_clause_token1] = ACTIONS(1526),
    [aux_sym_else_clause_token1] = ACTIONS(1526),
    [aux_sym_match_expression_token1] = ACTIONS(1526),
    [aux_sym_match_default_expression_token1] = ACTIONS(1526),
    [aux_sym_switch_statement_token1] = ACTIONS(1526),
    [aux_sym_switch_block_token1] = ACTIONS(1526),
    [anon_sym_PLUS] = ACTIONS(1526),
    [anon_sym_DASH] = ACTIONS(1526),
    [anon_sym_TILDE] = ACTIONS(1524),
    [anon_sym_BANG] = ACTIONS(1524),
    [anon_sym_AT] = ACTIONS(1524),
    [aux_sym_clone_expression_token1] = ACTIONS(1526),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1526),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1526),
    [anon_sym_DASH_DASH] = ACTIONS(1524),
    [anon_sym_PLUS_PLUS] = ACTIONS(1524),
    [aux_sym__list_destructing_token1] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1524),
    [anon_sym_self] = ACTIONS(1526),
    [anon_sym_parent] = ACTIONS(1526),
    [aux_sym__argument_name_token1] = ACTIONS(1526),
    [aux_sym__argument_name_token2] = ACTIONS(1526),
    [anon_sym_POUND_LBRACK] = ACTIONS(1524),
    [aux_sym_encapsed_string_token1] = ACTIONS(1524),
    [anon_sym_DQUOTE] = ACTIONS(1524),
    [aux_sym_string_token1] = ACTIONS(1524),
    [anon_sym_SQUOTE] = ACTIONS(1524),
    [anon_sym_LT_LT_LT] = ACTIONS(1524),
    [anon_sym_BQUOTE] = ACTIONS(1524),
    [anon_sym_DOLLAR] = ACTIONS(1524),
    [aux_sym_yield_expression_token1] = ACTIONS(1526),
    [aux_sym_include_expression_token1] = ACTIONS(1526),
    [aux_sym_include_once_expression_token1] = ACTIONS(1526),
    [aux_sym_require_expression_token1] = ACTIONS(1526),
    [aux_sym_require_once_expression_token1] = ACTIONS(1526),
    [sym_comment] = ACTIONS(5),
  },
  [577] = {
    [sym_text_interpolation] = STATE(577),
    [ts_builtin_sym_end] = ACTIONS(1528),
    [sym_name] = ACTIONS(1530),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1528),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1530),
    [aux_sym_global_declaration_token1] = ACTIONS(1530),
    [aux_sym_namespace_definition_token1] = ACTIONS(1530),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1530),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1530),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1530),
    [anon_sym_BSLASH] = ACTIONS(1528),
    [anon_sym_LBRACE] = ACTIONS(1528),
    [anon_sym_RBRACE] = ACTIONS(1528),
    [aux_sym_trait_declaration_token1] = ACTIONS(1530),
    [aux_sym_interface_declaration_token1] = ACTIONS(1530),
    [aux_sym_enum_declaration_token1] = ACTIONS(1530),
    [aux_sym_enum_case_token1] = ACTIONS(1530),
    [aux_sym_class_declaration_token1] = ACTIONS(1530),
    [aux_sym_final_modifier_token1] = ACTIONS(1530),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1530),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1530),
    [sym_var_modifier] = ACTIONS(1530),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1530),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1530),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1530),
    [anon_sym_LPAREN2] = ACTIONS(1528),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1530),
    [aux_sym_cast_type_token1] = ACTIONS(1530),
    [aux_sym_echo_statement_token1] = ACTIONS(1530),
    [aux_sym_exit_statement_token1] = ACTIONS(1530),
    [anon_sym_unset] = ACTIONS(1530),
    [aux_sym_declare_statement_token1] = ACTIONS(1530),
    [aux_sym_declare_statement_token2] = ACTIONS(1530),
    [sym_float] = ACTIONS(1530),
    [aux_sym_try_statement_token1] = ACTIONS(1530),
    [aux_sym_goto_statement_token1] = ACTIONS(1530),
    [aux_sym_continue_statement_token1] = ACTIONS(1530),
    [aux_sym_break_statement_token1] = ACTIONS(1530),
    [sym_integer] = ACTIONS(1530),
    [aux_sym_return_statement_token1] = ACTIONS(1530),
    [aux_sym_throw_expression_token1] = ACTIONS(1530),
    [aux_sym_while_statement_token1] = ACTIONS(1530),
    [aux_sym_while_statement_token2] = ACTIONS(1530),
    [aux_sym_do_statement_token1] = ACTIONS(1530),
    [aux_sym_for_statement_token1] = ACTIONS(1530),
    [aux_sym_for_statement_token2] = ACTIONS(1530),
    [aux_sym_foreach_statement_token1] = ACTIONS(1530),
    [aux_sym_foreach_statement_token2] = ACTIONS(1530),
    [aux_sym_if_statement_token1] = ACTIONS(1530),
    [aux_sym_if_statement_token2] = ACTIONS(1530),
    [aux_sym_else_if_clause_token1] = ACTIONS(1530),
    [aux_sym_else_clause_token1] = ACTIONS(1530),
    [aux_sym_match_expression_token1] = ACTIONS(1530),
    [aux_sym_match_default_expression_token1] = ACTIONS(1530),
    [aux_sym_switch_statement_token1] = ACTIONS(1530),
    [aux_sym_switch_block_token1] = ACTIONS(1530),
    [anon_sym_PLUS] = ACTIONS(1530),
    [anon_sym_DASH] = ACTIONS(1530),
    [anon_sym_TILDE] = ACTIONS(1528),
    [anon_sym_BANG] = ACTIONS(1528),
    [anon_sym_AT] = ACTIONS(1528),
    [aux_sym_clone_expression_token1] = ACTIONS(1530),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1530),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1530),
    [anon_sym_DASH_DASH] = ACTIONS(1528),
    [anon_sym_PLUS_PLUS] = ACTIONS(1528),
    [aux_sym__list_destructing_token1] = ACTIONS(1530),
    [anon_sym_LBRACK] = ACTIONS(1528),
    [anon_sym_self] = ACTIONS(1530),
    [anon_sym_parent] = ACTIONS(1530),
    [aux_sym__argument_name_token1] = ACTIONS(1530),
    [aux_sym__argument_name_token2] = ACTIONS(1530),
    [anon_sym_POUND_LBRACK] = ACTIONS(1528),
    [aux_sym_encapsed_string_token1] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1528),
    [aux_sym_string_token1] = ACTIONS(1528),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_LT_LT_LT] = ACTIONS(1528),
    [anon_sym_BQUOTE] = ACTIONS(1528),
    [anon_sym_DOLLAR] = ACTIONS(1528),
    [aux_sym_yield_expression_token1] = ACTIONS(1530),
    [aux_sym_include_expression_token1] = ACTIONS(1530),
    [aux_sym_include_once_expression_token1] = ACTIONS(1530),
    [aux_sym_require_expression_token1] = ACTIONS(1530),
    [aux_sym_require_once_expression_token1] = ACTIONS(1530),
    [sym_comment] = ACTIONS(5),
  },
  [578] = {
    [sym_text_interpolation] = STATE(578),
    [ts_builtin_sym_end] = ACTIONS(1532),
    [sym_name] = ACTIONS(1534),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1532),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1534),
    [aux_sym_global_declaration_token1] = ACTIONS(1534),
    [aux_sym_namespace_definition_token1] = ACTIONS(1534),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1534),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1534),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1534),
    [anon_sym_BSLASH] = ACTIONS(1532),
    [anon_sym_LBRACE] = ACTIONS(1532),
    [anon_sym_RBRACE] = ACTIONS(1532),
    [aux_sym_trait_declaration_token1] = ACTIONS(1534),
    [aux_sym_interface_declaration_token1] = ACTIONS(1534),
    [aux_sym_enum_declaration_token1] = ACTIONS(1534),
    [aux_sym_enum_case_token1] = ACTIONS(1534),
    [aux_sym_class_declaration_token1] = ACTIONS(1534),
    [aux_sym_final_modifier_token1] = ACTIONS(1534),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1534),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1534),
    [sym_var_modifier] = ACTIONS(1534),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1534),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1534),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1534),
    [anon_sym_LPAREN2] = ACTIONS(1532),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1534),
    [aux_sym_cast_type_token1] = ACTIONS(1534),
    [aux_sym_echo_statement_token1] = ACTIONS(1534),
    [aux_sym_exit_statement_token1] = ACTIONS(1534),
    [anon_sym_unset] = ACTIONS(1534),
    [aux_sym_declare_statement_token1] = ACTIONS(1534),
    [aux_sym_declare_statement_token2] = ACTIONS(1534),
    [sym_float] = ACTIONS(1534),
    [aux_sym_try_statement_token1] = ACTIONS(1534),
    [aux_sym_goto_statement_token1] = ACTIONS(1534),
    [aux_sym_continue_statement_token1] = ACTIONS(1534),
    [aux_sym_break_statement_token1] = ACTIONS(1534),
    [sym_integer] = ACTIONS(1534),
    [aux_sym_return_statement_token1] = ACTIONS(1534),
    [aux_sym_throw_expression_token1] = ACTIONS(1534),
    [aux_sym_while_statement_token1] = ACTIONS(1534),
    [aux_sym_while_statement_token2] = ACTIONS(1534),
    [aux_sym_do_statement_token1] = ACTIONS(1534),
    [aux_sym_for_statement_token1] = ACTIONS(1534),
    [aux_sym_for_statement_token2] = ACTIONS(1534),
    [aux_sym_foreach_statement_token1] = ACTIONS(1534),
    [aux_sym_foreach_statement_token2] = ACTIONS(1534),
    [aux_sym_if_statement_token1] = ACTIONS(1534),
    [aux_sym_if_statement_token2] = ACTIONS(1534),
    [aux_sym_else_if_clause_token1] = ACTIONS(1534),
    [aux_sym_else_clause_token1] = ACTIONS(1534),
    [aux_sym_match_expression_token1] = ACTIONS(1534),
    [aux_sym_match_default_expression_token1] = ACTIONS(1534),
    [aux_sym_switch_statement_token1] = ACTIONS(1534),
    [aux_sym_switch_block_token1] = ACTIONS(1534),
    [anon_sym_PLUS] = ACTIONS(1534),
    [anon_sym_DASH] = ACTIONS(1534),
    [anon_sym_TILDE] = ACTIONS(1532),
    [anon_sym_BANG] = ACTIONS(1532),
    [anon_sym_AT] = ACTIONS(1532),
    [aux_sym_clone_expression_token1] = ACTIONS(1534),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1534),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1534),
    [anon_sym_DASH_DASH] = ACTIONS(1532),
    [anon_sym_PLUS_PLUS] = ACTIONS(1532),
    [aux_sym__list_destructing_token1] = ACTIONS(1534),
    [anon_sym_LBRACK] = ACTIONS(1532),
    [anon_sym_self] = ACTIONS(1534),
    [anon_sym_parent] = ACTIONS(1534),
    [aux_sym__argument_name_token1] = ACTIONS(1534),
    [aux_sym__argument_name_token2] = ACTIONS(1534),
    [anon_sym_POUND_LBRACK] = ACTIONS(1532),
    [aux_sym_encapsed_string_token1] = ACTIONS(1532),
    [anon_sym_DQUOTE] = ACTIONS(1532),
    [aux_sym_string_token1] = ACTIONS(1532),
    [anon_sym_SQUOTE] = ACTIONS(1532),
    [anon_sym_LT_LT_LT] = ACTIONS(1532),
    [anon_sym_BQUOTE] = ACTIONS(1532),
    [anon_sym_DOLLAR] = ACTIONS(1532),
    [aux_sym_yield_expression_token1] = ACTIONS(1534),
    [aux_sym_include_expression_token1] = ACTIONS(1534),
    [aux_sym_include_once_expression_token1] = ACTIONS(1534),
    [aux_sym_require_expression_token1] = ACTIONS(1534),
    [aux_sym_require_once_expression_token1] = ACTIONS(1534),
    [sym_comment] = ACTIONS(5),
  },
  [579] = {
    [sym_text_interpolation] = STATE(579),
    [ts_builtin_sym_end] = ACTIONS(1536),
    [sym_name] = ACTIONS(1538),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1536),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1538),
    [aux_sym_global_declaration_token1] = ACTIONS(1538),
    [aux_sym_namespace_definition_token1] = ACTIONS(1538),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1538),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1538),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1538),
    [anon_sym_BSLASH] = ACTIONS(1536),
    [anon_sym_LBRACE] = ACTIONS(1536),
    [anon_sym_RBRACE] = ACTIONS(1536),
    [aux_sym_trait_declaration_token1] = ACTIONS(1538),
    [aux_sym_interface_declaration_token1] = ACTIONS(1538),
    [aux_sym_enum_declaration_token1] = ACTIONS(1538),
    [aux_sym_enum_case_token1] = ACTIONS(1538),
    [aux_sym_class_declaration_token1] = ACTIONS(1538),
    [aux_sym_final_modifier_token1] = ACTIONS(1538),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1538),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1538),
    [sym_var_modifier] = ACTIONS(1538),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1538),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1538),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1538),
    [anon_sym_LPAREN2] = ACTIONS(1536),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1538),
    [aux_sym_cast_type_token1] = ACTIONS(1538),
    [aux_sym_echo_statement_token1] = ACTIONS(1538),
    [aux_sym_exit_statement_token1] = ACTIONS(1538),
    [anon_sym_unset] = ACTIONS(1538),
    [aux_sym_declare_statement_token1] = ACTIONS(1538),
    [aux_sym_declare_statement_token2] = ACTIONS(1538),
    [sym_float] = ACTIONS(1538),
    [aux_sym_try_statement_token1] = ACTIONS(1538),
    [aux_sym_goto_statement_token1] = ACTIONS(1538),
    [aux_sym_continue_statement_token1] = ACTIONS(1538),
    [aux_sym_break_statement_token1] = ACTIONS(1538),
    [sym_integer] = ACTIONS(1538),
    [aux_sym_return_statement_token1] = ACTIONS(1538),
    [aux_sym_throw_expression_token1] = ACTIONS(1538),
    [aux_sym_while_statement_token1] = ACTIONS(1538),
    [aux_sym_while_statement_token2] = ACTIONS(1538),
    [aux_sym_do_statement_token1] = ACTIONS(1538),
    [aux_sym_for_statement_token1] = ACTIONS(1538),
    [aux_sym_for_statement_token2] = ACTIONS(1538),
    [aux_sym_foreach_statement_token1] = ACTIONS(1538),
    [aux_sym_foreach_statement_token2] = ACTIONS(1538),
    [aux_sym_if_statement_token1] = ACTIONS(1538),
    [aux_sym_if_statement_token2] = ACTIONS(1538),
    [aux_sym_else_if_clause_token1] = ACTIONS(1538),
    [aux_sym_else_clause_token1] = ACTIONS(1538),
    [aux_sym_match_expression_token1] = ACTIONS(1538),
    [aux_sym_match_default_expression_token1] = ACTIONS(1538),
    [aux_sym_switch_statement_token1] = ACTIONS(1538),
    [aux_sym_switch_block_token1] = ACTIONS(1538),
    [anon_sym_PLUS] = ACTIONS(1538),
    [anon_sym_DASH] = ACTIONS(1538),
    [anon_sym_TILDE] = ACTIONS(1536),
    [anon_sym_BANG] = ACTIONS(1536),
    [anon_sym_AT] = ACTIONS(1536),
    [aux_sym_clone_expression_token1] = ACTIONS(1538),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1538),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1538),
    [anon_sym_DASH_DASH] = ACTIONS(1536),
    [anon_sym_PLUS_PLUS] = ACTIONS(1536),
    [aux_sym__list_destructing_token1] = ACTIONS(1538),
    [anon_sym_LBRACK] = ACTIONS(1536),
    [anon_sym_self] = ACTIONS(1538),
    [anon_sym_parent] = ACTIONS(1538),
    [aux_sym__argument_name_token1] = ACTIONS(1538),
    [aux_sym__argument_name_token2] = ACTIONS(1538),
    [anon_sym_POUND_LBRACK] = ACTIONS(1536),
    [aux_sym_encapsed_string_token1] = ACTIONS(1536),
    [anon_sym_DQUOTE] = ACTIONS(1536),
    [aux_sym_string_token1] = ACTIONS(1536),
    [anon_sym_SQUOTE] = ACTIONS(1536),
    [anon_sym_LT_LT_LT] = ACTIONS(1536),
    [anon_sym_BQUOTE] = ACTIONS(1536),
    [anon_sym_DOLLAR] = ACTIONS(1536),
    [aux_sym_yield_expression_token1] = ACTIONS(1538),
    [aux_sym_include_expression_token1] = ACTIONS(1538),
    [aux_sym_include_once_expression_token1] = ACTIONS(1538),
    [aux_sym_require_expression_token1] = ACTIONS(1538),
    [aux_sym_require_once_expression_token1] = ACTIONS(1538),
    [sym_comment] = ACTIONS(5),
  },
  [580] = {
    [sym_text_interpolation] = STATE(580),
    [ts_builtin_sym_end] = ACTIONS(1540),
    [sym_name] = ACTIONS(1542),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1540),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1542),
    [aux_sym_global_declaration_token1] = ACTIONS(1542),
    [aux_sym_namespace_definition_token1] = ACTIONS(1542),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1542),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1542),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1542),
    [anon_sym_BSLASH] = ACTIONS(1540),
    [anon_sym_LBRACE] = ACTIONS(1540),
    [anon_sym_RBRACE] = ACTIONS(1540),
    [aux_sym_trait_declaration_token1] = ACTIONS(1542),
    [aux_sym_interface_declaration_token1] = ACTIONS(1542),
    [aux_sym_enum_declaration_token1] = ACTIONS(1542),
    [aux_sym_enum_case_token1] = ACTIONS(1542),
    [aux_sym_class_declaration_token1] = ACTIONS(1542),
    [aux_sym_final_modifier_token1] = ACTIONS(1542),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1542),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1542),
    [sym_var_modifier] = ACTIONS(1542),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1542),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1542),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1542),
    [anon_sym_LPAREN2] = ACTIONS(1540),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1542),
    [aux_sym_cast_type_token1] = ACTIONS(1542),
    [aux_sym_echo_statement_token1] = ACTIONS(1542),
    [aux_sym_exit_statement_token1] = ACTIONS(1542),
    [anon_sym_unset] = ACTIONS(1542),
    [aux_sym_declare_statement_token1] = ACTIONS(1542),
    [aux_sym_declare_statement_token2] = ACTIONS(1542),
    [sym_float] = ACTIONS(1542),
    [aux_sym_try_statement_token1] = ACTIONS(1542),
    [aux_sym_goto_statement_token1] = ACTIONS(1542),
    [aux_sym_continue_statement_token1] = ACTIONS(1542),
    [aux_sym_break_statement_token1] = ACTIONS(1542),
    [sym_integer] = ACTIONS(1542),
    [aux_sym_return_statement_token1] = ACTIONS(1542),
    [aux_sym_throw_expression_token1] = ACTIONS(1542),
    [aux_sym_while_statement_token1] = ACTIONS(1542),
    [aux_sym_while_statement_token2] = ACTIONS(1542),
    [aux_sym_do_statement_token1] = ACTIONS(1542),
    [aux_sym_for_statement_token1] = ACTIONS(1542),
    [aux_sym_for_statement_token2] = ACTIONS(1542),
    [aux_sym_foreach_statement_token1] = ACTIONS(1542),
    [aux_sym_foreach_statement_token2] = ACTIONS(1542),
    [aux_sym_if_statement_token1] = ACTIONS(1542),
    [aux_sym_if_statement_token2] = ACTIONS(1542),
    [aux_sym_else_if_clause_token1] = ACTIONS(1542),
    [aux_sym_else_clause_token1] = ACTIONS(1542),
    [aux_sym_match_expression_token1] = ACTIONS(1542),
    [aux_sym_match_default_expression_token1] = ACTIONS(1542),
    [aux_sym_switch_statement_token1] = ACTIONS(1542),
    [aux_sym_switch_block_token1] = ACTIONS(1542),
    [anon_sym_PLUS] = ACTIONS(1542),
    [anon_sym_DASH] = ACTIONS(1542),
    [anon_sym_TILDE] = ACTIONS(1540),
    [anon_sym_BANG] = ACTIONS(1540),
    [anon_sym_AT] = ACTIONS(1540),
    [aux_sym_clone_expression_token1] = ACTIONS(1542),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1542),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1542),
    [anon_sym_DASH_DASH] = ACTIONS(1540),
    [anon_sym_PLUS_PLUS] = ACTIONS(1540),
    [aux_sym__list_destructing_token1] = ACTIONS(1542),
    [anon_sym_LBRACK] = ACTIONS(1540),
    [anon_sym_self] = ACTIONS(1542),
    [anon_sym_parent] = ACTIONS(1542),
    [aux_sym__argument_name_token1] = ACTIONS(1542),
    [aux_sym__argument_name_token2] = ACTIONS(1542),
    [anon_sym_POUND_LBRACK] = ACTIONS(1540),
    [aux_sym_encapsed_string_token1] = ACTIONS(1540),
    [anon_sym_DQUOTE] = ACTIONS(1540),
    [aux_sym_string_token1] = ACTIONS(1540),
    [anon_sym_SQUOTE] = ACTIONS(1540),
    [anon_sym_LT_LT_LT] = ACTIONS(1540),
    [anon_sym_BQUOTE] = ACTIONS(1540),
    [anon_sym_DOLLAR] = ACTIONS(1540),
    [aux_sym_yield_expression_token1] = ACTIONS(1542),
    [aux_sym_include_expression_token1] = ACTIONS(1542),
    [aux_sym_include_once_expression_token1] = ACTIONS(1542),
    [aux_sym_require_expression_token1] = ACTIONS(1542),
    [aux_sym_require_once_expression_token1] = ACTIONS(1542),
    [sym_comment] = ACTIONS(5),
  },
  [581] = {
    [sym_text_interpolation] = STATE(581),
    [ts_builtin_sym_end] = ACTIONS(1544),
    [sym_name] = ACTIONS(1546),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1544),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1546),
    [aux_sym_global_declaration_token1] = ACTIONS(1546),
    [aux_sym_namespace_definition_token1] = ACTIONS(1546),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1546),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1546),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1546),
    [anon_sym_BSLASH] = ACTIONS(1544),
    [anon_sym_LBRACE] = ACTIONS(1544),
    [anon_sym_RBRACE] = ACTIONS(1544),
    [aux_sym_trait_declaration_token1] = ACTIONS(1546),
    [aux_sym_interface_declaration_token1] = ACTIONS(1546),
    [aux_sym_enum_declaration_token1] = ACTIONS(1546),
    [aux_sym_enum_case_token1] = ACTIONS(1546),
    [aux_sym_class_declaration_token1] = ACTIONS(1546),
    [aux_sym_final_modifier_token1] = ACTIONS(1546),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1546),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1546),
    [sym_var_modifier] = ACTIONS(1546),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1546),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1546),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1546),
    [anon_sym_LPAREN2] = ACTIONS(1544),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1546),
    [aux_sym_cast_type_token1] = ACTIONS(1546),
    [aux_sym_echo_statement_token1] = ACTIONS(1546),
    [aux_sym_exit_statement_token1] = ACTIONS(1546),
    [anon_sym_unset] = ACTIONS(1546),
    [aux_sym_declare_statement_token1] = ACTIONS(1546),
    [aux_sym_declare_statement_token2] = ACTIONS(1546),
    [sym_float] = ACTIONS(1546),
    [aux_sym_try_statement_token1] = ACTIONS(1546),
    [aux_sym_goto_statement_token1] = ACTIONS(1546),
    [aux_sym_continue_statement_token1] = ACTIONS(1546),
    [aux_sym_break_statement_token1] = ACTIONS(1546),
    [sym_integer] = ACTIONS(1546),
    [aux_sym_return_statement_token1] = ACTIONS(1546),
    [aux_sym_throw_expression_token1] = ACTIONS(1546),
    [aux_sym_while_statement_token1] = ACTIONS(1546),
    [aux_sym_while_statement_token2] = ACTIONS(1546),
    [aux_sym_do_statement_token1] = ACTIONS(1546),
    [aux_sym_for_statement_token1] = ACTIONS(1546),
    [aux_sym_for_statement_token2] = ACTIONS(1546),
    [aux_sym_foreach_statement_token1] = ACTIONS(1546),
    [aux_sym_foreach_statement_token2] = ACTIONS(1546),
    [aux_sym_if_statement_token1] = ACTIONS(1546),
    [aux_sym_if_statement_token2] = ACTIONS(1546),
    [aux_sym_else_if_clause_token1] = ACTIONS(1546),
    [aux_sym_else_clause_token1] = ACTIONS(1546),
    [aux_sym_match_expression_token1] = ACTIONS(1546),
    [aux_sym_match_default_expression_token1] = ACTIONS(1546),
    [aux_sym_switch_statement_token1] = ACTIONS(1546),
    [aux_sym_switch_block_token1] = ACTIONS(1546),
    [anon_sym_PLUS] = ACTIONS(1546),
    [anon_sym_DASH] = ACTIONS(1546),
    [anon_sym_TILDE] = ACTIONS(1544),
    [anon_sym_BANG] = ACTIONS(1544),
    [anon_sym_AT] = ACTIONS(1544),
    [aux_sym_clone_expression_token1] = ACTIONS(1546),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1546),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1546),
    [anon_sym_DASH_DASH] = ACTIONS(1544),
    [anon_sym_PLUS_PLUS] = ACTIONS(1544),
    [aux_sym__list_destructing_token1] = ACTIONS(1546),
    [anon_sym_LBRACK] = ACTIONS(1544),
    [anon_sym_self] = ACTIONS(1546),
    [anon_sym_parent] = ACTIONS(1546),
    [aux_sym__argument_name_token1] = ACTIONS(1546),
    [aux_sym__argument_name_token2] = ACTIONS(1546),
    [anon_sym_POUND_LBRACK] = ACTIONS(1544),
    [aux_sym_encapsed_string_token1] = ACTIONS(1544),
    [anon_sym_DQUOTE] = ACTIONS(1544),
    [aux_sym_string_token1] = ACTIONS(1544),
    [anon_sym_SQUOTE] = ACTIONS(1544),
    [anon_sym_LT_LT_LT] = ACTIONS(1544),
    [anon_sym_BQUOTE] = ACTIONS(1544),
    [anon_sym_DOLLAR] = ACTIONS(1544),
    [aux_sym_yield_expression_token1] = ACTIONS(1546),
    [aux_sym_include_expression_token1] = ACTIONS(1546),
    [aux_sym_include_once_expression_token1] = ACTIONS(1546),
    [aux_sym_require_expression_token1] = ACTIONS(1546),
    [aux_sym_require_once_expression_token1] = ACTIONS(1546),
    [sym_comment] = ACTIONS(5),
  },
  [582] = {
    [sym_text_interpolation] = STATE(582),
    [ts_builtin_sym_end] = ACTIONS(1548),
    [sym_name] = ACTIONS(1550),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1548),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1550),
    [aux_sym_global_declaration_token1] = ACTIONS(1550),
    [aux_sym_namespace_definition_token1] = ACTIONS(1550),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1550),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1550),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1550),
    [anon_sym_BSLASH] = ACTIONS(1548),
    [anon_sym_LBRACE] = ACTIONS(1548),
    [anon_sym_RBRACE] = ACTIONS(1548),
    [aux_sym_trait_declaration_token1] = ACTIONS(1550),
    [aux_sym_interface_declaration_token1] = ACTIONS(1550),
    [aux_sym_enum_declaration_token1] = ACTIONS(1550),
    [aux_sym_enum_case_token1] = ACTIONS(1550),
    [aux_sym_class_declaration_token1] = ACTIONS(1550),
    [aux_sym_final_modifier_token1] = ACTIONS(1550),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1550),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1550),
    [sym_var_modifier] = ACTIONS(1550),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1550),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1550),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1550),
    [anon_sym_LPAREN2] = ACTIONS(1548),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1550),
    [aux_sym_cast_type_token1] = ACTIONS(1550),
    [aux_sym_echo_statement_token1] = ACTIONS(1550),
    [aux_sym_exit_statement_token1] = ACTIONS(1550),
    [anon_sym_unset] = ACTIONS(1550),
    [aux_sym_declare_statement_token1] = ACTIONS(1550),
    [aux_sym_declare_statement_token2] = ACTIONS(1550),
    [sym_float] = ACTIONS(1550),
    [aux_sym_try_statement_token1] = ACTIONS(1550),
    [aux_sym_goto_statement_token1] = ACTIONS(1550),
    [aux_sym_continue_statement_token1] = ACTIONS(1550),
    [aux_sym_break_statement_token1] = ACTIONS(1550),
    [sym_integer] = ACTIONS(1550),
    [aux_sym_return_statement_token1] = ACTIONS(1550),
    [aux_sym_throw_expression_token1] = ACTIONS(1550),
    [aux_sym_while_statement_token1] = ACTIONS(1550),
    [aux_sym_while_statement_token2] = ACTIONS(1550),
    [aux_sym_do_statement_token1] = ACTIONS(1550),
    [aux_sym_for_statement_token1] = ACTIONS(1550),
    [aux_sym_for_statement_token2] = ACTIONS(1550),
    [aux_sym_foreach_statement_token1] = ACTIONS(1550),
    [aux_sym_foreach_statement_token2] = ACTIONS(1550),
    [aux_sym_if_statement_token1] = ACTIONS(1550),
    [aux_sym_if_statement_token2] = ACTIONS(1550),
    [aux_sym_else_if_clause_token1] = ACTIONS(1550),
    [aux_sym_else_clause_token1] = ACTIONS(1550),
    [aux_sym_match_expression_token1] = ACTIONS(1550),
    [aux_sym_match_default_expression_token1] = ACTIONS(1550),
    [aux_sym_switch_statement_token1] = ACTIONS(1550),
    [aux_sym_switch_block_token1] = ACTIONS(1550),
    [anon_sym_PLUS] = ACTIONS(1550),
    [anon_sym_DASH] = ACTIONS(1550),
    [anon_sym_TILDE] = ACTIONS(1548),
    [anon_sym_BANG] = ACTIONS(1548),
    [anon_sym_AT] = ACTIONS(1548),
    [aux_sym_clone_expression_token1] = ACTIONS(1550),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1550),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1550),
    [anon_sym_DASH_DASH] = ACTIONS(1548),
    [anon_sym_PLUS_PLUS] = ACTIONS(1548),
    [aux_sym__list_destructing_token1] = ACTIONS(1550),
    [anon_sym_LBRACK] = ACTIONS(1548),
    [anon_sym_self] = ACTIONS(1550),
    [anon_sym_parent] = ACTIONS(1550),
    [aux_sym__argument_name_token1] = ACTIONS(1550),
    [aux_sym__argument_name_token2] = ACTIONS(1550),
    [anon_sym_POUND_LBRACK] = ACTIONS(1548),
    [aux_sym_encapsed_string_token1] = ACTIONS(1548),
    [anon_sym_DQUOTE] = ACTIONS(1548),
    [aux_sym_string_token1] = ACTIONS(1548),
    [anon_sym_SQUOTE] = ACTIONS(1548),
    [anon_sym_LT_LT_LT] = ACTIONS(1548),
    [anon_sym_BQUOTE] = ACTIONS(1548),
    [anon_sym_DOLLAR] = ACTIONS(1548),
    [aux_sym_yield_expression_token1] = ACTIONS(1550),
    [aux_sym_include_expression_token1] = ACTIONS(1550),
    [aux_sym_include_once_expression_token1] = ACTIONS(1550),
    [aux_sym_require_expression_token1] = ACTIONS(1550),
    [aux_sym_require_once_expression_token1] = ACTIONS(1550),
    [sym_comment] = ACTIONS(5),
  },
  [583] = {
    [sym_text_interpolation] = STATE(583),
    [ts_builtin_sym_end] = ACTIONS(1552),
    [sym_name] = ACTIONS(1554),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1552),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1554),
    [aux_sym_global_declaration_token1] = ACTIONS(1554),
    [aux_sym_namespace_definition_token1] = ACTIONS(1554),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1554),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1554),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1554),
    [anon_sym_BSLASH] = ACTIONS(1552),
    [anon_sym_LBRACE] = ACTIONS(1552),
    [anon_sym_RBRACE] = ACTIONS(1552),
    [aux_sym_trait_declaration_token1] = ACTIONS(1554),
    [aux_sym_interface_declaration_token1] = ACTIONS(1554),
    [aux_sym_enum_declaration_token1] = ACTIONS(1554),
    [aux_sym_enum_case_token1] = ACTIONS(1554),
    [aux_sym_class_declaration_token1] = ACTIONS(1554),
    [aux_sym_final_modifier_token1] = ACTIONS(1554),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1554),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1554),
    [sym_var_modifier] = ACTIONS(1554),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1554),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1554),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1554),
    [anon_sym_LPAREN2] = ACTIONS(1552),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1554),
    [aux_sym_cast_type_token1] = ACTIONS(1554),
    [aux_sym_echo_statement_token1] = ACTIONS(1554),
    [aux_sym_exit_statement_token1] = ACTIONS(1554),
    [anon_sym_unset] = ACTIONS(1554),
    [aux_sym_declare_statement_token1] = ACTIONS(1554),
    [aux_sym_declare_statement_token2] = ACTIONS(1554),
    [sym_float] = ACTIONS(1554),
    [aux_sym_try_statement_token1] = ACTIONS(1554),
    [aux_sym_goto_statement_token1] = ACTIONS(1554),
    [aux_sym_continue_statement_token1] = ACTIONS(1554),
    [aux_sym_break_statement_token1] = ACTIONS(1554),
    [sym_integer] = ACTIONS(1554),
    [aux_sym_return_statement_token1] = ACTIONS(1554),
    [aux_sym_throw_expression_token1] = ACTIONS(1554),
    [aux_sym_while_statement_token1] = ACTIONS(1554),
    [aux_sym_while_statement_token2] = ACTIONS(1554),
    [aux_sym_do_statement_token1] = ACTIONS(1554),
    [aux_sym_for_statement_token1] = ACTIONS(1554),
    [aux_sym_for_statement_token2] = ACTIONS(1554),
    [aux_sym_foreach_statement_token1] = ACTIONS(1554),
    [aux_sym_foreach_statement_token2] = ACTIONS(1554),
    [aux_sym_if_statement_token1] = ACTIONS(1554),
    [aux_sym_if_statement_token2] = ACTIONS(1554),
    [aux_sym_else_if_clause_token1] = ACTIONS(1554),
    [aux_sym_else_clause_token1] = ACTIONS(1554),
    [aux_sym_match_expression_token1] = ACTIONS(1554),
    [aux_sym_match_default_expression_token1] = ACTIONS(1554),
    [aux_sym_switch_statement_token1] = ACTIONS(1554),
    [aux_sym_switch_block_token1] = ACTIONS(1554),
    [anon_sym_PLUS] = ACTIONS(1554),
    [anon_sym_DASH] = ACTIONS(1554),
    [anon_sym_TILDE] = ACTIONS(1552),
    [anon_sym_BANG] = ACTIONS(1552),
    [anon_sym_AT] = ACTIONS(1552),
    [aux_sym_clone_expression_token1] = ACTIONS(1554),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1554),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1554),
    [anon_sym_DASH_DASH] = ACTIONS(1552),
    [anon_sym_PLUS_PLUS] = ACTIONS(1552),
    [aux_sym__list_destructing_token1] = ACTIONS(1554),
    [anon_sym_LBRACK] = ACTIONS(1552),
    [anon_sym_self] = ACTIONS(1554),
    [anon_sym_parent] = ACTIONS(1554),
    [aux_sym__argument_name_token1] = ACTIONS(1554),
    [aux_sym__argument_name_token2] = ACTIONS(1554),
    [anon_sym_POUND_LBRACK] = ACTIONS(1552),
    [aux_sym_encapsed_string_token1] = ACTIONS(1552),
    [anon_sym_DQUOTE] = ACTIONS(1552),
    [aux_sym_string_token1] = ACTIONS(1552),
    [anon_sym_SQUOTE] = ACTIONS(1552),
    [anon_sym_LT_LT_LT] = ACTIONS(1552),
    [anon_sym_BQUOTE] = ACTIONS(1552),
    [anon_sym_DOLLAR] = ACTIONS(1552),
    [aux_sym_yield_expression_token1] = ACTIONS(1554),
    [aux_sym_include_expression_token1] = ACTIONS(1554),
    [aux_sym_include_once_expression_token1] = ACTIONS(1554),
    [aux_sym_require_expression_token1] = ACTIONS(1554),
    [aux_sym_require_once_expression_token1] = ACTIONS(1554),
    [sym_comment] = ACTIONS(5),
  },
  [584] = {
    [sym_text_interpolation] = STATE(584),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1661),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2547),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_primary_expression] = STATE(1013),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_variable] = STATE(750),
    [sym__variable_member_access_expression] = STATE(751),
    [sym_member_access_expression] = STATE(750),
    [sym__variable_nullsafe_member_access_expression] = STATE(752),
    [sym_nullsafe_member_access_expression] = STATE(750),
    [sym__variable_scoped_property_access_expression] = STATE(755),
    [sym_scoped_property_access_expression] = STATE(750),
    [sym_function_call_expression] = STATE(729),
    [sym__callable_expression] = STATE(2319),
    [sym_scoped_call_expression] = STATE(729),
    [sym__scope_resolution_qualifier] = STATE(2570),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(729),
    [sym_nullsafe_member_call_expression] = STATE(729),
    [sym__variable_subscript_expression] = STATE(757),
    [sym__dereferencable_subscript_expression] = STATE(729),
    [sym__dereferencable_expression] = STATE(1915),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(746),
    [sym__simple_variable] = STATE(730),
    [sym__new_variable] = STATE(756),
    [sym__callable_variable] = STATE(725),
    [sym_variable_name] = STATE(746),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(1556),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(663),
    [aux_sym_print_intrinsic_token1] = ACTIONS(673),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [anon_sym_LBRACK] = ACTIONS(983),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(1558),
    [sym_comment] = ACTIONS(5),
  },
  [585] = {
    [sym_text_interpolation] = STATE(585),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1661),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2638),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_primary_expression] = STATE(1013),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_variable] = STATE(750),
    [sym__variable_member_access_expression] = STATE(751),
    [sym_member_access_expression] = STATE(750),
    [sym__variable_nullsafe_member_access_expression] = STATE(752),
    [sym_nullsafe_member_access_expression] = STATE(750),
    [sym__variable_scoped_property_access_expression] = STATE(755),
    [sym_scoped_property_access_expression] = STATE(750),
    [sym_function_call_expression] = STATE(729),
    [sym__callable_expression] = STATE(2319),
    [sym_scoped_call_expression] = STATE(729),
    [sym__scope_resolution_qualifier] = STATE(2570),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(729),
    [sym_nullsafe_member_call_expression] = STATE(729),
    [sym__variable_subscript_expression] = STATE(757),
    [sym__dereferencable_subscript_expression] = STATE(729),
    [sym__dereferencable_expression] = STATE(1915),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(746),
    [sym__simple_variable] = STATE(730),
    [sym__new_variable] = STATE(756),
    [sym__callable_variable] = STATE(725),
    [sym_variable_name] = STATE(746),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(1556),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(711),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [anon_sym_LBRACK] = ACTIONS(983),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(1558),
    [sym_comment] = ACTIONS(5),
  },
  [586] = {
    [sym_text_interpolation] = STATE(586),
    [sym_qualified_name] = STATE(936),
    [sym__name] = STATE(1652),
    [sym_namespace_name] = STATE(2622),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1087),
    [sym__anonymous_function_header] = STATE(2418),
    [sym__arrow_function_header] = STATE(2677),
    [sym_arrow_function] = STATE(1087),
    [sym_literal] = STATE(1087),
    [sym_throw_expression] = STATE(1087),
    [sym_primary_expression] = STATE(1137),
    [sym_parenthesized_expression] = STATE(948),
    [sym_class_constant_access_expression] = STATE(1017),
    [sym_print_intrinsic] = STATE(1087),
    [sym_object_creation_expression] = STATE(1087),
    [sym__new_non_dereferencable_expression] = STATE(1103),
    [sym__new_dereferencable_expression] = STATE(939),
    [sym_update_expression] = STATE(1087),
    [sym_cast_variable] = STATE(900),
    [sym__variable_member_access_expression] = STATE(903),
    [sym_member_access_expression] = STATE(900),
    [sym__variable_nullsafe_member_access_expression] = STATE(904),
    [sym_nullsafe_member_access_expression] = STATE(900),
    [sym__variable_scoped_property_access_expression] = STATE(905),
    [sym_scoped_property_access_expression] = STATE(900),
    [sym_function_call_expression] = STATE(890),
    [sym__callable_expression] = STATE(2474),
    [sym_scoped_call_expression] = STATE(890),
    [sym__scope_resolution_qualifier] = STATE(2509),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(890),
    [sym_nullsafe_member_call_expression] = STATE(890),
    [sym__variable_subscript_expression] = STATE(906),
    [sym__dereferencable_subscript_expression] = STATE(890),
    [sym__dereferencable_expression] = STATE(1869),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(943),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(955),
    [sym_string] = STATE(955),
    [sym_heredoc] = STATE(955),
    [sym_nowdoc] = STATE(955),
    [sym_shell_command_expression] = STATE(1087),
    [sym_boolean] = STATE(1176),
    [sym_null] = STATE(1176),
    [sym__string] = STATE(940),
    [sym_dynamic_variable_name] = STATE(868),
    [sym__simple_variable] = STATE(889),
    [sym__new_variable] = STATE(901),
    [sym__callable_variable] = STATE(872),
    [sym_variable_name] = STATE(868),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(689),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(226),
    [anon_sym_LPAREN2] = ACTIONS(1560),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(256),
    [sym_float] = ACTIONS(266),
    [sym_integer] = ACTIONS(266),
    [aux_sym_throw_expression_token1] = ACTIONS(278),
    [aux_sym_print_intrinsic_token1] = ACTIONS(302),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(304),
    [anon_sym_DASH_DASH] = ACTIONS(306),
    [anon_sym_PLUS_PLUS] = ACTIONS(306),
    [anon_sym_LBRACK] = ACTIONS(985),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(314),
    [aux_sym__argument_name_token2] = ACTIONS(316),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(320),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [aux_sym_string_token1] = ACTIONS(322),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_LT_LT_LT] = ACTIONS(324),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(1562),
    [sym_comment] = ACTIONS(5),
  },
  [587] = {
    [sym_text_interpolation] = STATE(587),
    [sym_qualified_name] = STATE(802),
    [sym__name] = STATE(1661),
    [sym_namespace_name] = STATE(2506),
    [sym_static_modifier] = STATE(2413),
    [sym_anonymous_function] = STATE(1036),
    [sym__anonymous_function_header] = STATE(2339),
    [sym__arrow_function_header] = STATE(2653),
    [sym_arrow_function] = STATE(1036),
    [sym_literal] = STATE(1036),
    [sym_throw_expression] = STATE(1036),
    [sym_primary_expression] = STATE(1013),
    [sym_parenthesized_expression] = STATE(791),
    [sym_class_constant_access_expression] = STATE(853),
    [sym_print_intrinsic] = STATE(1036),
    [sym_object_creation_expression] = STATE(1036),
    [sym__new_non_dereferencable_expression] = STATE(1040),
    [sym__new_dereferencable_expression] = STATE(792),
    [sym_update_expression] = STATE(1036),
    [sym_cast_variable] = STATE(750),
    [sym__variable_member_access_expression] = STATE(751),
    [sym_member_access_expression] = STATE(750),
    [sym__variable_nullsafe_member_access_expression] = STATE(752),
    [sym_nullsafe_member_access_expression] = STATE(750),
    [sym__variable_scoped_property_access_expression] = STATE(755),
    [sym_scoped_property_access_expression] = STATE(750),
    [sym_function_call_expression] = STATE(729),
    [sym__callable_expression] = STATE(2319),
    [sym_scoped_call_expression] = STATE(729),
    [sym__scope_resolution_qualifier] = STATE(2570),
    [sym_relative_scope] = STATE(2664),
    [sym_member_call_expression] = STATE(729),
    [sym_nullsafe_member_call_expression] = STATE(729),
    [sym__variable_subscript_expression] = STATE(757),
    [sym__dereferencable_subscript_expression] = STATE(729),
    [sym__dereferencable_expression] = STATE(1915),
    [sym__dereferencable_scalar] = STATE(1803),
    [sym_array_creation_expression] = STATE(815),
    [sym_attribute_group] = STATE(1374),
    [sym_attribute_list] = STATE(1893),
    [sym_encapsed_string] = STATE(775),
    [sym_string] = STATE(775),
    [sym_heredoc] = STATE(775),
    [sym_nowdoc] = STATE(775),
    [sym_shell_command_expression] = STATE(1036),
    [sym_boolean] = STATE(1018),
    [sym_null] = STATE(1018),
    [sym__string] = STATE(788),
    [sym_dynamic_variable_name] = STATE(746),
    [sym__simple_variable] = STATE(730),
    [sym__new_variable] = STATE(756),
    [sym__callable_variable] = STATE(725),
    [sym_variable_name] = STATE(746),
    [sym__reserved_identifier] = STATE(717),
    [sym__identifier] = STATE(1499),
    [aux_sym_attribute_list_repeat1] = STATE(1364),
    [sym_name] = ACTIONS(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [aux_sym_function_static_declaration_token1] = ACTIONS(597),
    [aux_sym__namespace_use_type_token1] = ACTIONS(601),
    [anon_sym_BSLASH] = ACTIONS(603),
    [anon_sym_LPAREN2] = ACTIONS(1556),
    [aux_sym__arrow_function_header_token1] = ACTIONS(254),
    [aux_sym_cast_type_token1] = ACTIONS(609),
    [sym_float] = ACTIONS(611),
    [sym_integer] = ACTIONS(611),
    [aux_sym_throw_expression_token1] = ACTIONS(613),
    [aux_sym_print_intrinsic_token1] = ACTIONS(625),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(627),
    [anon_sym_DASH_DASH] = ACTIONS(629),
    [anon_sym_PLUS_PLUS] = ACTIONS(629),
    [anon_sym_LBRACK] = ACTIONS(983),
    [anon_sym_self] = ACTIONS(312),
    [anon_sym_parent] = ACTIONS(312),
    [aux_sym__argument_name_token1] = ACTIONS(633),
    [aux_sym__argument_name_token2] = ACTIONS(635),
    [anon_sym_POUND_LBRACK] = ACTIONS(318),
    [aux_sym_encapsed_string_token1] = ACTIONS(637),
    [anon_sym_DQUOTE] = ACTIONS(637),
    [aux_sym_string_token1] = ACTIONS(639),
    [anon_sym_SQUOTE] = ACTIONS(639),
    [anon_sym_LT_LT_LT] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(643),
    [anon_sym_DOLLAR] = ACTIONS(1558),
    [sym_comment] = ACTIONS(5),
  },
  [588] = {
    [sym_text_interpolation] = STATE(588),
    [sym_name] = ACTIONS(1564),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1566),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1564),
    [aux_sym_global_declaration_token1] = ACTIONS(1564),
    [aux_sym_namespace_definition_token1] = ACTIONS(1564),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1564),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1564),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1564),
    [anon_sym_BSLASH] = ACTIONS(1566),
    [anon_sym_LBRACE] = ACTIONS(1566),
    [aux_sym_trait_declaration_token1] = ACTIONS(1564),
    [aux_sym_interface_declaration_token1] = ACTIONS(1564),
    [aux_sym_enum_declaration_token1] = ACTIONS(1564),
    [anon_sym_COLON] = ACTIONS(1566),
    [aux_sym_class_declaration_token1] = ACTIONS(1564),
    [aux_sym_final_modifier_token1] = ACTIONS(1564),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1564),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1564),
    [sym_var_modifier] = ACTIONS(1564),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1564),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1564),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1564),
    [anon_sym_LPAREN2] = ACTIONS(1566),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1564),
    [aux_sym_cast_type_token1] = ACTIONS(1564),
    [aux_sym_echo_statement_token1] = ACTIONS(1564),
    [aux_sym_exit_statement_token1] = ACTIONS(1564),
    [anon_sym_unset] = ACTIONS(1564),
    [aux_sym_declare_statement_token1] = ACTIONS(1564),
    [sym_float] = ACTIONS(1564),
    [aux_sym_try_statement_token1] = ACTIONS(1564),
    [aux_sym_goto_statement_token1] = ACTIONS(1564),
    [aux_sym_continue_statement_token1] = ACTIONS(1564),
    [aux_sym_break_statement_token1] = ACTIONS(1564),
    [sym_integer] = ACTIONS(1564),
    [aux_sym_return_statement_token1] = ACTIONS(1564),
    [aux_sym_throw_expression_token1] = ACTIONS(1564),
    [aux_sym_while_statement_token1] = ACTIONS(1564),
    [aux_sym_do_statement_token1] = ACTIONS(1564),
    [aux_sym_for_statement_token1] = ACTIONS(1564),
    [aux_sym_foreach_statement_token1] = ACTIONS(1564),
    [aux_sym_if_statement_token1] = ACTIONS(1564),
    [aux_sym_match_expression_token1] = ACTIONS(1564),
    [aux_sym_switch_statement_token1] = ACTIONS(1564),
    [anon_sym_PLUS] = ACTIONS(1564),
    [anon_sym_DASH] = ACTIONS(1564),
    [anon_sym_TILDE] = ACTIONS(1566),
    [anon_sym_BANG] = ACTIONS(1566),
    [anon_sym_AT] = ACTIONS(1566),
    [aux_sym_clone_expression_token1] = ACTIONS(1564),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1564),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1564),
    [anon_sym_DASH_DASH] = ACTIONS(1566),
    [anon_sym_PLUS_PLUS] = ACTIONS(1566),
    [aux_sym__list_destructing_token1] = ACTIONS(1564),
    [anon_sym_LBRACK] = ACTIONS(1566),
    [anon_sym_self] = ACTIONS(1564),
    [anon_sym_parent] = ACTIONS(1564),
    [aux_sym__argument_name_token1] = ACTIONS(1564),
    [aux_sym__argument_name_token2] = ACTIONS(1564),
    [anon_sym_POUND_LBRACK] = ACTIONS(1566),
    [aux_sym_encapsed_string_token1] = ACTIONS(1566),
    [anon_sym_DQUOTE] = ACTIONS(1566),
    [aux_sym_string_token1] = ACTIONS(1566),
    [anon_sym_SQUOTE] = ACTIONS(1566),
    [anon_sym_LT_LT_LT] = ACTIONS(1566),
    [anon_sym_BQUOTE] = ACTIONS(1566),
    [anon_sym_DOLLAR] = ACTIONS(1566),
    [aux_sym_yield_expression_token1] = ACTIONS(1564),
    [aux_sym_include_expression_token1] = ACTIONS(1564),
    [aux_sym_include_once_expression_token1] = ACTIONS(1564),
    [aux_sym_require_expression_token1] = ACTIONS(1564),
    [aux_sym_require_once_expression_token1] = ACTIONS(1564),
    [sym_comment] = ACTIONS(5),
  },
  [589] = {
    [sym_text_interpolation] = STATE(589),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1568),
    [anon_sym_AMP] = ACTIONS(1570),
    [anon_sym_COMMA] = ACTIONS(1568),
    [anon_sym_EQ] = ACTIONS(1570),
    [aux_sym_namespace_use_clause_token1] = ACTIONS(1568),
    [anon_sym_LBRACE] = ACTIONS(1568),
    [anon_sym_RBRACE] = ACTIONS(1568),
    [aux_sym_base_clause_token1] = ACTIONS(1568),
    [anon_sym_COLON] = ACTIONS(1570),
    [aux_sym_class_interface_clause_token1] = ACTIONS(1568),
    [anon_sym_EQ_GT] = ACTIONS(1568),
    [anon_sym_LPAREN2] = ACTIONS(1568),
    [anon_sym_RPAREN2] = ACTIONS(1568),
    [anon_sym_QMARK] = ACTIONS(1570),
    [anon_sym_PIPE] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1570),
    [anon_sym_DASH] = ACTIONS(1570),
    [anon_sym_COLON_COLON] = ACTIONS(1568),
    [anon_sym_DASH_DASH] = ACTIONS(1568),
    [anon_sym_PLUS_PLUS] = ACTIONS(1568),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1568),
    [anon_sym_STAR_EQ] = ACTIONS(1568),
    [anon_sym_SLASH_EQ] = ACTIONS(1568),
    [anon_sym_PERCENT_EQ] = ACTIONS(1568),
    [anon_sym_PLUS_EQ] = ACTIONS(1568),
    [anon_sym_DASH_EQ] = ACTIONS(1568),
    [anon_sym_DOT_EQ] = ACTIONS(1568),
    [anon_sym_LT_LT_EQ] = ACTIONS(1568),
    [anon_sym_GT_GT_EQ] = ACTIONS(1568),
    [anon_sym_AMP_EQ] = ACTIONS(1568),
    [anon_sym_CARET_EQ] = ACTIONS(1568),
    [anon_sym_PIPE_EQ] = ACTIONS(1568),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(1568),
    [anon_sym_DASH_GT] = ACTIONS(1568),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(1568),
    [anon_sym_LBRACK] = ACTIONS(1568),
    [anon_sym_RBRACK] = ACTIONS(1568),
    [aux_sym_binary_expression_token1] = ACTIONS(1568),
    [anon_sym_QMARK_QMARK] = ACTIONS(1570),
    [anon_sym_STAR_STAR] = ACTIONS(1570),
    [aux_sym_binary_expression_token2] = ACTIONS(1568),
    [aux_sym_binary_expression_token3] = ACTIONS(1568),
    [aux_sym_binary_expression_token4] = ACTIONS(1568),
    [anon_sym_PIPE_PIPE] = ACTIONS(1568),
    [anon_sym_AMP_AMP] = ACTIONS(1568),
    [anon_sym_CARET] = ACTIONS(1570),
    [anon_sym_EQ_EQ] = ACTIONS(1570),
    [anon_sym_BANG_EQ] = ACTIONS(1570),
    [anon_sym_LT_GT] = ACTIONS(1568),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1568),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1568),
    [anon_sym_LT] = ACTIONS(1570),
    [anon_sym_GT] = ACTIONS(1570),
    [anon_sym_LT_EQ] = ACTIONS(1570),
    [anon_sym_GT_EQ] = ACTIONS(1568),
    [anon_sym_LT_EQ_GT] = ACTIONS(1568),
    [anon_sym_LT_LT] = ACTIONS(1570),
    [anon_sym_GT_GT] = ACTIONS(1570),
    [anon_sym_DOT] = ACTIONS(1570),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PERCENT] = ACTIONS(1570),
    [sym_comment] = ACTIONS(1572),
  },
  [590] = {
    [sym_text_interpolation] = STATE(590),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1574),
    [anon_sym_AMP] = ACTIONS(1576),
    [anon_sym_COMMA] = ACTIONS(1574),
    [anon_sym_EQ] = ACTIONS(1576),
    [aux_sym_namespace_use_clause_token1] = ACTIONS(1574),
    [anon_sym_LBRACE] = ACTIONS(1574),
    [anon_sym_RBRACE] = ACTIONS(1574),
    [aux_sym_base_clause_token1] = ACTIONS(1574),
    [anon_sym_COLON] = ACTIONS(1576),
    [aux_sym_class_interface_clause_token1] = ACTIONS(1574),
    [anon_sym_EQ_GT] = ACTIONS(1574),
    [anon_sym_LPAREN2] = ACTIONS(1574),
    [anon_sym_RPAREN2] = ACTIONS(1574),
    [anon_sym_QMARK] = ACTIONS(1576),
    [anon_sym_PIPE] = ACTIONS(1576),
    [anon_sym_PLUS] = ACTIONS(1576),
    [anon_sym_DASH] = ACTIONS(1576),
    [anon_sym_COLON_COLON] = ACTIONS(1574),
    [anon_sym_DASH_DASH] = ACTIONS(1574),
    [anon_sym_PLUS_PLUS] = ACTIONS(1574),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1574),
    [anon_sym_STAR_EQ] = ACTIONS(1574),
    [anon_sym_SLASH_EQ] = ACTIONS(1574),
    [anon_sym_PERCENT_EQ] = ACTIONS(1574),
    [anon_sym_PLUS_EQ] = ACTIONS(1574),
    [anon_sym_DASH_EQ] = ACTIONS(1574),
    [anon_sym_DOT_EQ] = ACTIONS(1574),
    [anon_sym_LT_LT_EQ] = ACTIONS(1574),
    [anon_sym_GT_GT_EQ] = ACTIONS(1574),
    [anon_sym_AMP_EQ] = ACTIONS(1574),
    [anon_sym_CARET_EQ] = ACTIONS(1574),
    [anon_sym_PIPE_EQ] = ACTIONS(1574),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(1574),
    [anon_sym_DASH_GT] = ACTIONS(1574),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(1574),
    [anon_sym_LBRACK] = ACTIONS(1574),
    [anon_sym_RBRACK] = ACTIONS(1574),
    [aux_sym_binary_expression_token1] = ACTIONS(1574),
    [anon_sym_QMARK_QMARK] = ACTIONS(1576),
    [anon_sym_STAR_STAR] = ACTIONS(1576),
    [aux_sym_binary_expression_token2] = ACTIONS(1574),
    [aux_sym_binary_expression_token3] = ACTIONS(1574),
    [aux_sym_binary_expression_token4] = ACTIONS(1574),
    [anon_sym_PIPE_PIPE] = ACTIONS(1574),
    [anon_sym_AMP_AMP] = ACTIONS(1574),
    [anon_sym_CARET] = ACTIONS(1576),
    [anon_sym_EQ_EQ] = ACTIONS(1576),
    [anon_sym_BANG_EQ] = ACTIONS(1576),
    [anon_sym_LT_GT] = ACTIONS(1574),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1574),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1574),
    [anon_sym_LT] = ACTIONS(1576),
    [anon_sym_GT] = ACTIONS(1576),
    [anon_sym_LT_EQ] = ACTIONS(1576),
    [anon_sym_GT_EQ] = ACTIONS(1574),
    [anon_sym_LT_EQ_GT] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT] = ACTIONS(1576),
    [anon_sym_STAR] = ACTIONS(1576),
    [anon_sym_SLASH] = ACTIONS(1576),
    [anon_sym_PERCENT] = ACTIONS(1576),
    [sym_comment] = ACTIONS(1572),
  },
  [591] = {
    [sym_text_interpolation] = STATE(591),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1578),
    [anon_sym_AMP] = ACTIONS(1580),
    [anon_sym_COMMA] = ACTIONS(1578),
    [anon_sym_EQ] = ACTIONS(1580),
    [aux_sym_namespace_use_clause_token1] = ACTIONS(1578),
    [anon_sym_LBRACE] = ACTIONS(1578),
    [anon_sym_RBRACE] = ACTIONS(1578),
    [aux_sym_base_clause_token1] = ACTIONS(1578),
    [anon_sym_COLON] = ACTIONS(1580),
    [aux_sym_class_interface_clause_token1] = ACTIONS(1578),
    [anon_sym_EQ_GT] = ACTIONS(1578),
    [anon_sym_LPAREN2] = ACTIONS(1578),
    [anon_sym_RPAREN2] = ACTIONS(1578),
    [anon_sym_QMARK] = ACTIONS(1580),
    [anon_sym_PIPE] = ACTIONS(1580),
    [anon_sym_PLUS] = ACTIONS(1580),
    [anon_sym_DASH] = ACTIONS(1580),
    [anon_sym_COLON_COLON] = ACTIONS(1578),
    [anon_sym_DASH_DASH] = ACTIONS(1578),
    [anon_sym_PLUS_PLUS] = ACTIONS(1578),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1578),
    [anon_sym_STAR_EQ] = ACTIONS(1578),
    [anon_sym_SLASH_EQ] = ACTIONS(1578),
    [anon_sym_PERCENT_EQ] = ACTIONS(1578),
    [anon_sym_PLUS_EQ] = ACTIONS(1578),
    [anon_sym_DASH_EQ] = ACTIONS(1578),
    [anon_sym_DOT_EQ] = ACTIONS(1578),
    [anon_sym_LT_LT_EQ] = ACTIONS(1578),
    [anon_sym_GT_GT_EQ] = ACTIONS(1578),
    [anon_sym_AMP_EQ] = ACTIONS(1578),
    [anon_sym_CARET_EQ] = ACTIONS(1578),
    [anon_sym_PIPE_EQ] = ACTIONS(1578),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(1578),
    [anon_sym_DASH_GT] = ACTIONS(1578),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(1578),
    [anon_sym_LBRACK] = ACTIONS(1578),
    [anon_sym_RBRACK] = ACTIONS(1578),
    [aux_sym_binary_expression_token1] = ACTIONS(1578),
    [anon_sym_QMARK_QMARK] = ACTIONS(1580),
    [anon_sym_STAR_STAR] = ACTIONS(1580),
    [aux_sym_binary_expression_token2] = ACTIONS(1578),
    [aux_sym_binary_expression_token3] = ACTIONS(1578),
    [aux_sym_binary_expression_token4] = ACTIONS(1578),
    [anon_sym_PIPE_PIPE] = ACTIONS(1578),
    [anon_sym_AMP_AMP] = ACTIONS(1578),
    [anon_sym_CARET] = ACTIONS(1580),
    [anon_sym_EQ_EQ] = ACTIONS(1580),
    [anon_sym_BANG_EQ] = ACTIONS(1580),
    [anon_sym_LT_GT] = ACTIONS(1578),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1578),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1578),
    [anon_sym_LT_EQ_GT] = ACTIONS(1578),
    [anon_sym_LT_LT] = ACTIONS(1580),
    [anon_sym_GT_GT] = ACTIONS(1580),
    [anon_sym_DOT] = ACTIONS(1580),
    [anon_sym_STAR] = ACTIONS(1580),
    [anon_sym_SLASH] = ACTIONS(1580),
    [anon_sym_PERCENT] = ACTIONS(1580),
    [sym_comment] = ACTIONS(1572),
  },
  [592] = {
    [sym_text_interpolation] = STATE(592),
    [anon_sym_QMARK_GT] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1582),
    [anon_sym_AMP] = ACTIONS(1584),
    [anon_sym_COMMA] = ACTIONS(1582),
    [anon_sym_EQ] = ACTIONS(1584),
    [aux_sym_namespace_use_clause_token1] = ACTIONS(1582),
    [anon_sym_LBRACE] = ACTIONS(1582),
    [anon_sym_RBRACE] = ACTIONS(1582),
    [aux_sym_base_clause_token1] = ACTIONS(1582),
    [anon_sym_COLON] = ACTIONS(1584),
    [aux_sym_class_interface_clause_token1] = ACTIONS(1582),
    [anon_sym_EQ_GT] = ACTIONS(1582),
    [anon_sym_LPAREN2] = ACTIONS(1582),
    [anon_sym_RPAREN2] = ACTIONS(1582),
    [anon_sym_QMARK] = ACTIONS(1584),
    [anon_sym_PIPE] = ACTIONS(1584),
    [anon_sym_PLUS] = ACTIONS(1584),
    [anon_sym_DASH] = ACTIONS(1584),
    [anon_sym_COLON_COLON] = ACTIONS(1582),
    [anon_sym_DASH_DASH] = ACTIONS(1582),
    [anon_sym_PLUS_PLUS] = ACTIONS(1582),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1582),
    [anon_sym_STAR_EQ] = ACTIONS(1582),
    [anon_sym_SLASH_EQ] = ACTIONS(1582),
    [anon_sym_PERCENT_EQ] = ACTIONS(1582),
    [anon_sym_PLUS_EQ] = ACTIONS(1582),
    [anon_sym_DASH_EQ] = ACTIONS(1582),
    [anon_sym_DOT_EQ] = ACTIONS(1582),
    [anon_sym_LT_LT_EQ] = ACTIONS(1582),
    [anon_sym_GT_GT_EQ] = ACTIONS(1582),
    [anon_sym_AMP_EQ] = ACTIONS(1582),
    [anon_sym_CARET_EQ] = ACTIONS(1582),
    [anon_sym_PIPE_EQ] = ACTIONS(1582),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(1582),
    [anon_sym_DASH_GT] = ACTIONS(1582),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(1582),
    [anon_sym_LBRACK] = ACTIONS(1582),
    [anon_sym_RBRACK] = ACTIONS(1582),
    [aux_sym_binary_expression_token1] = ACTIONS(1582),
    [anon_sym_QMARK_QMARK] = ACTIONS(1584),
    [anon_sym_STAR_STAR] = ACTIONS(1584),
    [aux_sym_binary_expression_token2] = ACTIONS(1582),
    [aux_sym_binary_expression_token3] = ACTIONS(1582),
    [aux_sym_binary_expression_token4] = ACTIONS(1582),
    [anon_sym_PIPE_PIPE] = ACTIONS(1582),
    [anon_sym_AMP_AMP] = ACTIONS(1582),
    [anon_sym_CARET] = ACTIONS(1584),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_GT] = ACTIONS(1582),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1582),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1582),
    [anon_sym_LT] = ACTIONS(1584),
    [anon_sym_GT] = ACTIONS(1584),
    [anon_sym_LT_EQ] = ACTIONS(1584),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_LT_EQ_GT] = ACTIONS(1582),
    [anon_sym_LT_LT] = ACTIONS(1584),
    [anon_sym_GT_GT] = ACTIONS(1584),
    [anon_sym_DOT] = ACTIONS(1584),
    [anon_sym_STAR] = ACTIONS(1584),
    [anon_sym_SLASH] = ACTIONS(1584),
    [anon_sym_PERCENT] = ACTIONS(1584),
    [sym_comment] = ACTIONS(1572),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1590), 1,
      anon_sym_LPAREN2,
    STATE(593), 1,
      sym_text_interpolation,
    STATE(620), 1,
      sym_arguments,
    ACTIONS(1588), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1586), 38,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [79] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1590), 1,
      anon_sym_LPAREN2,
    STATE(594), 1,
      sym_text_interpolation,
    STATE(621), 1,
      sym_arguments,
    ACTIONS(1594), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1592), 38,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [158] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1600), 1,
      anon_sym_LPAREN2,
    STATE(595), 1,
      sym_text_interpolation,
    ACTIONS(1598), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1596), 38,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [234] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(596), 1,
      sym_text_interpolation,
    ACTIONS(1604), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1602), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [308] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(597), 1,
      sym_text_interpolation,
    ACTIONS(1608), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1606), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [382] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(598), 1,
      sym_text_interpolation,
    ACTIONS(1612), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1610), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [456] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(599), 1,
      sym_text_interpolation,
    ACTIONS(1616), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1614), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [530] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(600), 1,
      sym_text_interpolation,
    ACTIONS(1620), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1618), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [604] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(601), 1,
      sym_text_interpolation,
    ACTIONS(1624), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1622), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [678] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1602), 1,
      anon_sym_LPAREN2,
    STATE(602), 1,
      sym_text_interpolation,
    ACTIONS(1628), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1626), 38,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [754] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(603), 1,
      sym_text_interpolation,
    ACTIONS(1632), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1630), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [828] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(604), 1,
      sym_text_interpolation,
    ACTIONS(1636), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1634), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [902] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(605), 1,
      sym_text_interpolation,
    ACTIONS(1628), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1626), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [976] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(606), 1,
      sym_text_interpolation,
    ACTIONS(1640), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1638), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1050] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(607), 1,
      sym_text_interpolation,
    ACTIONS(1642), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1600), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1124] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(608), 1,
      sym_text_interpolation,
    ACTIONS(1646), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1644), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1198] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(609), 1,
      sym_text_interpolation,
    ACTIONS(1650), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1648), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1272] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(610), 1,
      sym_text_interpolation,
    ACTIONS(1654), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1652), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1346] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(611), 1,
      sym_text_interpolation,
    ACTIONS(1658), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1656), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1420] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(612), 1,
      sym_text_interpolation,
    ACTIONS(1662), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1660), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1494] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1668), 1,
      anon_sym_EQ,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(613), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1676), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 19,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 20,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [1578] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(614), 1,
      sym_text_interpolation,
    ACTIONS(1680), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1678), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1652] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(615), 1,
      sym_text_interpolation,
    ACTIONS(1684), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1682), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1726] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(616), 1,
      sym_text_interpolation,
    ACTIONS(1680), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1678), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1800] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(617), 1,
      sym_text_interpolation,
    ACTIONS(1688), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1686), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1874] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(618), 1,
      sym_text_interpolation,
    ACTIONS(1692), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1690), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1948] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(619), 1,
      sym_text_interpolation,
    ACTIONS(1696), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1694), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2022] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(620), 1,
      sym_text_interpolation,
    ACTIONS(1700), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1698), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2096] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(621), 1,
      sym_text_interpolation,
    ACTIONS(1704), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1702), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2170] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1668), 1,
      anon_sym_EQ,
    STATE(622), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1676), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 19,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 20,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [2251] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1668), 1,
      anon_sym_EQ,
    STATE(623), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1676), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 19,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 20,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [2332] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1706), 1,
      anon_sym_LPAREN2,
    STATE(624), 1,
      sym_text_interpolation,
    STATE(635), 1,
      sym_arguments,
    ACTIONS(1588), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1586), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2407] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1706), 1,
      anon_sym_LPAREN2,
    STATE(625), 1,
      sym_text_interpolation,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(1594), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1592), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2482] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    ACTIONS(1708), 1,
      anon_sym_EQ,
    STATE(626), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1710), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 16,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 20,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [2563] = 39,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1712), 1,
      sym_name,
    ACTIONS(1714), 1,
      anon_sym_RBRACE,
    ACTIONS(1716), 1,
      anon_sym_LPAREN2,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    STATE(602), 1,
      sym__simple_variable,
    STATE(605), 1,
      sym__variable_subscript_expression,
    STATE(614), 1,
      sym__variable_member_access_expression,
    STATE(616), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(618), 1,
      sym__variable_scoped_property_access_expression,
    STATE(627), 1,
      sym_text_interpolation,
    STATE(645), 1,
      aux_sym_use_list_repeat1,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1677), 1,
      sym_class_constant_access_expression,
    STATE(1704), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1872), 1,
      sym__dereferencable_expression,
    STATE(1913), 1,
      sym__new_variable,
    STATE(2341), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2691), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(2196), 2,
      sym_use_instead_of_clause,
      sym_use_as_clause,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1803), 3,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym__callable_variable,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(1917), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(596), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [2701] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(628), 1,
      sym_text_interpolation,
    ACTIONS(1650), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1648), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2771] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1708), 1,
      anon_sym_EQ,
    STATE(629), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1710), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 16,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 20,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [2849] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(630), 1,
      sym_text_interpolation,
    ACTIONS(1608), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1606), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2919] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(631), 1,
      sym_text_interpolation,
    ACTIONS(1640), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1638), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2989] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(632), 1,
      sym_text_interpolation,
    ACTIONS(1646), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1644), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [3059] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(633), 1,
      sym_text_interpolation,
    ACTIONS(1580), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1578), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [3129] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(634), 1,
      sym_text_interpolation,
    ACTIONS(1620), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1618), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [3199] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(635), 1,
      sym_text_interpolation,
    ACTIONS(1700), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1698), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [3269] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    ACTIONS(1720), 1,
      anon_sym_EQ,
    STATE(636), 1,
      sym_text_interpolation,
    ACTIONS(1722), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1724), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 16,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [3349] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(637), 1,
      sym_text_interpolation,
    ACTIONS(1704), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1702), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [3419] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(638), 1,
      sym_text_interpolation,
    ACTIONS(1684), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1682), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [3489] = 41,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(308), 1,
      aux_sym__list_destructing_token1,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(731), 1,
      anon_sym_AMP,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1716), 1,
      anon_sym_LPAREN2,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    STATE(602), 1,
      sym__simple_variable,
    STATE(605), 1,
      sym__variable_subscript_expression,
    STATE(614), 1,
      sym__variable_member_access_expression,
    STATE(616), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(618), 1,
      sym__variable_scoped_property_access_expression,
    STATE(639), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1548), 1,
      sym__callable_variable,
    STATE(1574), 1,
      sym__new_variable,
    STATE(1704), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1872), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2090), 1,
      sym__list_destructing,
    STATE(2091), 1,
      sym_by_ref,
    STATE(2341), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2691), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(1569), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(596), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [3631] = 39,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1712), 1,
      sym_name,
    ACTIONS(1716), 1,
      anon_sym_LPAREN2,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1728), 1,
      anon_sym_RBRACE,
    STATE(602), 1,
      sym__simple_variable,
    STATE(605), 1,
      sym__variable_subscript_expression,
    STATE(614), 1,
      sym__variable_member_access_expression,
    STATE(616), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(618), 1,
      sym__variable_scoped_property_access_expression,
    STATE(627), 1,
      aux_sym_use_list_repeat1,
    STATE(640), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1677), 1,
      sym_class_constant_access_expression,
    STATE(1704), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1872), 1,
      sym__dereferencable_expression,
    STATE(1913), 1,
      sym__new_variable,
    STATE(2341), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2691), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(2196), 2,
      sym_use_instead_of_clause,
      sym_use_as_clause,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1803), 3,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym__callable_variable,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(1917), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(596), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [3769] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(641), 1,
      sym_text_interpolation,
    ACTIONS(1576), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1574), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [3839] = 41,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(308), 1,
      aux_sym__list_destructing_token1,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(731), 1,
      anon_sym_AMP,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1716), 1,
      anon_sym_LPAREN2,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    STATE(602), 1,
      sym__simple_variable,
    STATE(605), 1,
      sym__variable_subscript_expression,
    STATE(614), 1,
      sym__variable_member_access_expression,
    STATE(616), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(618), 1,
      sym__variable_scoped_property_access_expression,
    STATE(642), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1597), 1,
      sym__callable_variable,
    STATE(1692), 1,
      sym__new_variable,
    STATE(1704), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1872), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2341), 1,
      sym__callable_expression,
    STATE(2364), 1,
      sym__list_destructing,
    STATE(2384), 1,
      sym_by_ref,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2691), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(1690), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(596), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [3981] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(643), 1,
      sym_text_interpolation,
    ACTIONS(1624), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1622), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4051] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(644), 1,
      sym_text_interpolation,
    ACTIONS(1688), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1686), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4121] = 38,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1730), 1,
      sym_name,
    ACTIONS(1736), 1,
      anon_sym_BSLASH,
    ACTIONS(1739), 1,
      anon_sym_RBRACE,
    ACTIONS(1741), 1,
      anon_sym_LPAREN2,
    ACTIONS(1744), 1,
      aux_sym_cast_type_token1,
    ACTIONS(1747), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1750), 1,
      anon_sym_LBRACK,
    ACTIONS(1759), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1762), 1,
      anon_sym_DOLLAR,
    STATE(602), 1,
      sym__simple_variable,
    STATE(605), 1,
      sym__variable_subscript_expression,
    STATE(614), 1,
      sym__variable_member_access_expression,
    STATE(616), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(618), 1,
      sym__variable_scoped_property_access_expression,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1677), 1,
      sym_class_constant_access_expression,
    STATE(1704), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1872), 1,
      sym__dereferencable_expression,
    STATE(1913), 1,
      sym__new_variable,
    STATE(2341), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2691), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(1753), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(1756), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(645), 2,
      sym_text_interpolation,
      aux_sym_use_list_repeat1,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(2196), 2,
      sym_use_instead_of_clause,
      sym_use_as_clause,
    ACTIONS(1733), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1803), 3,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym__callable_variable,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(1917), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(596), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [4257] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(646), 1,
      sym_text_interpolation,
    ACTIONS(1604), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1602), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4327] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(647), 1,
      sym_text_interpolation,
    ACTIONS(1616), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1614), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4397] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    ACTIONS(1765), 1,
      anon_sym_EQ,
    STATE(648), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1767), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 16,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [4477] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1668), 1,
      anon_sym_EQ,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(649), 1,
      sym_text_interpolation,
    ACTIONS(1722), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1676), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 16,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [4557] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(650), 1,
      sym_text_interpolation,
    ACTIONS(1696), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1694), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4627] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1708), 1,
      anon_sym_EQ,
    STATE(651), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1710), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 16,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 20,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [4705] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1600), 1,
      anon_sym_LPAREN2,
    STATE(652), 1,
      sym_text_interpolation,
    ACTIONS(1598), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1596), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4777] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(653), 1,
      sym_text_interpolation,
    ACTIONS(1570), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1568), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4847] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(654), 1,
      sym_text_interpolation,
    ACTIONS(1584), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1582), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4917] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(655), 1,
      sym_text_interpolation,
    ACTIONS(1642), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1600), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4987] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1602), 1,
      anon_sym_LPAREN2,
    STATE(656), 1,
      sym_text_interpolation,
    ACTIONS(1628), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1626), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [5059] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(657), 1,
      sym_text_interpolation,
    ACTIONS(1612), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1610), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [5129] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1765), 1,
      anon_sym_EQ,
    STATE(658), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1767), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 16,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [5206] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1668), 1,
      anon_sym_EQ,
    STATE(659), 1,
      sym_text_interpolation,
    ACTIONS(1722), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1676), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 16,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [5283] = 40,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(731), 1,
      anon_sym_AMP,
    ACTIONS(859), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1716), 1,
      anon_sym_LPAREN2,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    STATE(602), 1,
      sym__simple_variable,
    STATE(605), 1,
      sym__variable_subscript_expression,
    STATE(614), 1,
      sym__variable_member_access_expression,
    STATE(616), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(618), 1,
      sym__variable_scoped_property_access_expression,
    STATE(660), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1592), 1,
      sym__callable_variable,
    STATE(1646), 1,
      sym__new_variable,
    STATE(1704), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1872), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2341), 1,
      sym__callable_expression,
    STATE(2468), 1,
      sym__array_destructing,
    STATE(2481), 1,
      sym_by_ref,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2691), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(1640), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(596), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [5422] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(661), 1,
      sym_text_interpolation,
    ACTIONS(1636), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1634), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [5491] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1720), 1,
      anon_sym_EQ,
    STATE(662), 1,
      sym_text_interpolation,
    ACTIONS(1722), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1724), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 16,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [5568] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(663), 1,
      sym_text_interpolation,
    ACTIONS(1680), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1678), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [5637] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(664), 1,
      sym_text_interpolation,
    ACTIONS(1680), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1678), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [5706] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(665), 1,
      sym_text_interpolation,
    ACTIONS(1692), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1690), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [5775] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(666), 1,
      sym_text_interpolation,
    ACTIONS(1628), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1626), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [5844] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1668), 1,
      anon_sym_EQ,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(667), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1676), 16,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_RBRACK,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [5923] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1668), 1,
      anon_sym_EQ,
    STATE(668), 1,
      sym_text_interpolation,
    ACTIONS(1722), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1676), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 16,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6000] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1720), 1,
      anon_sym_EQ,
    STATE(669), 1,
      sym_text_interpolation,
    ACTIONS(1722), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1724), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 16,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6077] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(868), 1,
      anon_sym_COMMA,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    ACTIONS(1708), 1,
      anon_sym_EQ,
    ACTIONS(1769), 1,
      anon_sym_RPAREN2,
    STATE(670), 1,
      sym_text_interpolation,
    STATE(2154), 1,
      aux_sym__list_destructing_repeat1,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1710), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6162] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(671), 1,
      sym_text_interpolation,
    ACTIONS(1632), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1630), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [6231] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(672), 1,
      sym_text_interpolation,
    ACTIONS(1654), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1652), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [6300] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(673), 1,
      sym_text_interpolation,
    ACTIONS(1658), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1656), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [6369] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(674), 1,
      sym_text_interpolation,
    ACTIONS(1662), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1660), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [6438] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    ACTIONS(1771), 1,
      anon_sym_EQ,
    STATE(675), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1773), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 15,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6517] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1765), 1,
      anon_sym_EQ,
    STATE(676), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1767), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 16,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6594] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1668), 1,
      anon_sym_EQ,
    STATE(677), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1676), 16,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_RBRACK,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6670] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    ACTIONS(1765), 1,
      anon_sym_EQ,
    STATE(678), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1775), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1767), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6750] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1668), 1,
      anon_sym_EQ,
    STATE(679), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1676), 16,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_RBRACK,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6826] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    ACTIONS(1708), 1,
      anon_sym_EQ,
    STATE(680), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1778), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1710), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6906] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    ACTIONS(1708), 1,
      anon_sym_EQ,
    STATE(681), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1780), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1710), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6986] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(868), 1,
      anon_sym_COMMA,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1708), 1,
      anon_sym_EQ,
    ACTIONS(1769), 1,
      anon_sym_RPAREN2,
    STATE(682), 1,
      sym_text_interpolation,
    STATE(2154), 1,
      aux_sym__list_destructing_repeat1,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1710), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [7068] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(868), 1,
      anon_sym_COMMA,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1708), 1,
      anon_sym_EQ,
    ACTIONS(1769), 1,
      anon_sym_RPAREN2,
    STATE(683), 1,
      sym_text_interpolation,
    STATE(2154), 1,
      aux_sym__list_destructing_repeat1,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1710), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [7150] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1771), 1,
      anon_sym_EQ,
    STATE(684), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1773), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 15,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [7226] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1771), 1,
      anon_sym_EQ,
    STATE(685), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1773), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1664), 15,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [7302] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1708), 1,
      anon_sym_EQ,
    STATE(686), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1780), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1710), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [7379] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1708), 1,
      anon_sym_EQ,
    STATE(687), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1778), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1710), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [7456] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    ACTIONS(1765), 1,
      anon_sym_EQ,
    STATE(688), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1782), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 11,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1767), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [7535] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1708), 1,
      anon_sym_EQ,
    STATE(689), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1780), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1710), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [7612] = 38,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1716), 1,
      anon_sym_LPAREN2,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    ACTIONS(1785), 1,
      anon_sym_RPAREN2,
    STATE(602), 1,
      sym__simple_variable,
    STATE(605), 1,
      sym__variable_subscript_expression,
    STATE(614), 1,
      sym__variable_member_access_expression,
    STATE(616), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(618), 1,
      sym__variable_scoped_property_access_expression,
    STATE(690), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1585), 1,
      sym__callable_variable,
    STATE(1681), 1,
      sym__new_variable,
    STATE(1704), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1872), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2341), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2691), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(1680), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(596), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [7745] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1708), 1,
      anon_sym_EQ,
    STATE(691), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1778), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1710), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [7822] = 38,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1716), 1,
      anon_sym_LPAREN2,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    ACTIONS(1787), 1,
      anon_sym_RPAREN2,
    STATE(602), 1,
      sym__simple_variable,
    STATE(605), 1,
      sym__variable_subscript_expression,
    STATE(614), 1,
      sym__variable_member_access_expression,
    STATE(616), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(618), 1,
      sym__variable_scoped_property_access_expression,
    STATE(692), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1585), 1,
      sym__callable_variable,
    STATE(1681), 1,
      sym__new_variable,
    STATE(1704), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1872), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2341), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2691), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(1680), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(596), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [7955] = 38,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1716), 1,
      anon_sym_LPAREN2,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    ACTIONS(1789), 1,
      anon_sym_RPAREN2,
    STATE(602), 1,
      sym__simple_variable,
    STATE(605), 1,
      sym__variable_subscript_expression,
    STATE(614), 1,
      sym__variable_member_access_expression,
    STATE(616), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(618), 1,
      sym__variable_scoped_property_access_expression,
    STATE(693), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1585), 1,
      sym__callable_variable,
    STATE(1681), 1,
      sym__new_variable,
    STATE(1704), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1872), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2341), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2691), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(1680), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(596), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [8088] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1765), 1,
      anon_sym_EQ,
    STATE(694), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1775), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1767), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [8165] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1765), 1,
      anon_sym_EQ,
    STATE(695), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1775), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1767), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [8242] = 38,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1716), 1,
      anon_sym_LPAREN2,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    ACTIONS(1791), 1,
      anon_sym_RPAREN2,
    STATE(602), 1,
      sym__simple_variable,
    STATE(605), 1,
      sym__variable_subscript_expression,
    STATE(614), 1,
      sym__variable_member_access_expression,
    STATE(616), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(618), 1,
      sym__variable_scoped_property_access_expression,
    STATE(696), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1585), 1,
      sym__callable_variable,
    STATE(1681), 1,
      sym__new_variable,
    STATE(1704), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1872), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2341), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2691), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(1680), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(596), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [8375] = 37,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1562), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    ACTIONS(1793), 1,
      anon_sym_LPAREN2,
    STATE(697), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(876), 1,
      sym__callable_variable,
    STATE(889), 1,
      sym__simple_variable,
    STATE(903), 1,
      sym__variable_member_access_expression,
    STATE(904), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(905), 1,
      sym__variable_scoped_property_access_expression,
    STATE(906), 1,
      sym__variable_subscript_expression,
    STATE(911), 1,
      sym__new_variable,
    STATE(1652), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1869), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2474), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2594), 1,
      sym__scope_resolution_qualifier,
    STATE(2664), 1,
      sym_relative_scope,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(868), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(910), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(890), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [8505] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1765), 1,
      anon_sym_EQ,
    STATE(698), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1782), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 11,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1767), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [8581] = 37,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    ACTIONS(1795), 1,
      anon_sym_LPAREN2,
    ACTIONS(1797), 1,
      anon_sym_DOLLAR,
    STATE(699), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(795), 1,
      sym__callable_variable,
    STATE(809), 1,
      sym__variable_member_access_expression,
    STATE(810), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(811), 1,
      sym__variable_scoped_property_access_expression,
    STATE(813), 1,
      sym__variable_subscript_expression,
    STATE(814), 1,
      sym__simple_variable,
    STATE(857), 1,
      sym__new_variable,
    STATE(1665), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(1943), 1,
      sym__dereferencable_expression,
    STATE(2327), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2681), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(840), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(856), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(812), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [8711] = 37,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    ACTIONS(1799), 1,
      anon_sym_LPAREN2,
    ACTIONS(1801), 1,
      anon_sym_DOLLAR,
    STATE(700), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(935), 1,
      sym__simple_variable,
    STATE(952), 1,
      sym__variable_subscript_expression,
    STATE(954), 1,
      sym__callable_variable,
    STATE(997), 1,
      sym__variable_member_access_expression,
    STATE(1000), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1001), 1,
      sym__variable_scoped_property_access_expression,
    STATE(1038), 1,
      sym__new_variable,
    STATE(1657), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1910), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2303), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2658), 1,
      sym__scope_resolution_qualifier,
    STATE(2664), 1,
      sym_relative_scope,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(957), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(1032), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(922), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [8841] = 37,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    ACTIONS(1799), 1,
      anon_sym_LPAREN2,
    ACTIONS(1801), 1,
      anon_sym_DOLLAR,
    STATE(701), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(927), 1,
      sym__callable_variable,
    STATE(935), 1,
      sym__simple_variable,
    STATE(952), 1,
      sym__variable_subscript_expression,
    STATE(997), 1,
      sym__variable_member_access_expression,
    STATE(1000), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1001), 1,
      sym__variable_scoped_property_access_expression,
    STATE(1011), 1,
      sym__new_variable,
    STATE(1657), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1910), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2303), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2658), 1,
      sym__scope_resolution_qualifier,
    STATE(2664), 1,
      sym_relative_scope,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(957), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(1035), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(922), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [8971] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1765), 1,
      anon_sym_EQ,
    STATE(702), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1782), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1664), 11,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1767), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1666), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [9047] = 37,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    ACTIONS(1795), 1,
      anon_sym_LPAREN2,
    ACTIONS(1797), 1,
      anon_sym_DOLLAR,
    STATE(703), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(765), 1,
      sym__callable_variable,
    STATE(809), 1,
      sym__variable_member_access_expression,
    STATE(810), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(811), 1,
      sym__variable_scoped_property_access_expression,
    STATE(813), 1,
      sym__variable_subscript_expression,
    STATE(814), 1,
      sym__simple_variable,
    STATE(846), 1,
      sym__new_variable,
    STATE(1665), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(1943), 1,
      sym__dereferencable_expression,
    STATE(2327), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2681), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(840), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(845), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(812), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [9177] = 37,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1556), 1,
      anon_sym_LPAREN2,
    ACTIONS(1558), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    STATE(704), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(722), 1,
      sym__callable_variable,
    STATE(728), 1,
      sym__new_variable,
    STATE(730), 1,
      sym__simple_variable,
    STATE(751), 1,
      sym__variable_member_access_expression,
    STATE(752), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(755), 1,
      sym__variable_scoped_property_access_expression,
    STATE(757), 1,
      sym__variable_subscript_expression,
    STATE(1661), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1915), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2319), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2570), 1,
      sym__scope_resolution_qualifier,
    STATE(2664), 1,
      sym_relative_scope,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(746), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(727), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(729), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [9307] = 37,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    ACTIONS(1795), 1,
      anon_sym_LPAREN2,
    ACTIONS(1797), 1,
      anon_sym_DOLLAR,
    STATE(705), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(789), 1,
      sym__callable_variable,
    STATE(809), 1,
      sym__variable_member_access_expression,
    STATE(810), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(811), 1,
      sym__variable_scoped_property_access_expression,
    STATE(813), 1,
      sym__variable_subscript_expression,
    STATE(814), 1,
      sym__simple_variable,
    STATE(852), 1,
      sym__new_variable,
    STATE(1665), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(1943), 1,
      sym__dereferencable_expression,
    STATE(2327), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2681), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(840), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(851), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(812), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [9437] = 37,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1716), 1,
      anon_sym_LPAREN2,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    STATE(602), 1,
      sym__simple_variable,
    STATE(605), 1,
      sym__variable_subscript_expression,
    STATE(614), 1,
      sym__variable_member_access_expression,
    STATE(616), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(618), 1,
      sym__variable_scoped_property_access_expression,
    STATE(706), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(722), 1,
      sym__callable_variable,
    STATE(728), 1,
      sym__new_variable,
    STATE(1704), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1872), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2341), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2691), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(727), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(596), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [9567] = 37,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1716), 1,
      anon_sym_LPAREN2,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    STATE(602), 1,
      sym__simple_variable,
    STATE(605), 1,
      sym__variable_subscript_expression,
    STATE(614), 1,
      sym__variable_member_access_expression,
    STATE(616), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(618), 1,
      sym__variable_scoped_property_access_expression,
    STATE(707), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1552), 1,
      sym__callable_variable,
    STATE(1586), 1,
      sym__new_variable,
    STATE(1704), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1872), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2341), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2691), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(1580), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(596), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [9697] = 37,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1716), 1,
      anon_sym_LPAREN2,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    STATE(602), 1,
      sym__simple_variable,
    STATE(605), 1,
      sym__variable_subscript_expression,
    STATE(614), 1,
      sym__variable_member_access_expression,
    STATE(616), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(618), 1,
      sym__variable_scoped_property_access_expression,
    STATE(708), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(765), 1,
      sym__callable_variable,
    STATE(846), 1,
      sym__new_variable,
    STATE(1704), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1872), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2341), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2691), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(845), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(596), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [9827] = 37,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    ACTIONS(1799), 1,
      anon_sym_LPAREN2,
    ACTIONS(1801), 1,
      anon_sym_DOLLAR,
    STATE(709), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(935), 1,
      sym__simple_variable,
    STATE(952), 1,
      sym__variable_subscript_expression,
    STATE(981), 1,
      sym__callable_variable,
    STATE(997), 1,
      sym__variable_member_access_expression,
    STATE(1000), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1001), 1,
      sym__variable_scoped_property_access_expression,
    STATE(1012), 1,
      sym__new_variable,
    STATE(1657), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1910), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2303), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2658), 1,
      sym__scope_resolution_qualifier,
    STATE(2664), 1,
      sym_relative_scope,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(957), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(1010), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(922), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [9957] = 37,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1716), 1,
      anon_sym_LPAREN2,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    STATE(602), 1,
      sym__simple_variable,
    STATE(605), 1,
      sym__variable_subscript_expression,
    STATE(614), 1,
      sym__variable_member_access_expression,
    STATE(616), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(618), 1,
      sym__variable_scoped_property_access_expression,
    STATE(710), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1585), 1,
      sym__callable_variable,
    STATE(1681), 1,
      sym__new_variable,
    STATE(1704), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1872), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2341), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2691), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(1680), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(596), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [10087] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1807), 1,
      anon_sym_BSLASH,
    STATE(711), 1,
      sym_text_interpolation,
    STATE(2472), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1803), 22,
      anon_sym_AMP,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_clause_token1,
      anon_sym_COLON,
      aux_sym_class_interface_clause_token1,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_self,
      anon_sym_parent,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
      sym_name,
    ACTIONS(1805), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOLLAR,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10157] = 37,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(609), 1,
      aux_sym_cast_type_token1,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(983), 1,
      anon_sym_LBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1716), 1,
      anon_sym_LPAREN2,
    ACTIONS(1718), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1726), 1,
      sym_name,
    STATE(602), 1,
      sym__simple_variable,
    STATE(605), 1,
      sym__variable_subscript_expression,
    STATE(614), 1,
      sym__variable_member_access_expression,
    STATE(616), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(618), 1,
      sym__variable_scoped_property_access_expression,
    STATE(712), 1,
      sym_text_interpolation,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1560), 1,
      sym__callable_variable,
    STATE(1600), 1,
      sym__new_variable,
    STATE(1704), 1,
      sym__name,
    STATE(1736), 1,
      sym__new_dereferencable_expression,
    STATE(1872), 1,
      sym__dereferencable_expression,
    STATE(1917), 1,
      sym_class_constant_access_expression,
    STATE(2341), 1,
      sym__callable_expression,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2664), 1,
      sym_relative_scope,
    STATE(2691), 1,
      sym__scope_resolution_qualifier,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    STATE(1779), 2,
      sym_array_creation_expression,
      sym__string,
    STATE(1803), 2,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
    ACTIONS(312), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
    STATE(1615), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(596), 5,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
  [10287] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(713), 1,
      sym_text_interpolation,
    ACTIONS(1810), 22,
      anon_sym_AMP,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_clause_token1,
      anon_sym_COLON,
      aux_sym_class_interface_clause_token1,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_self,
      anon_sym_parent,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
      sym_name,
    ACTIONS(1812), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOLLAR,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10351] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(714), 1,
      sym_text_interpolation,
    ACTIONS(1814), 9,
      aux_sym_function_static_declaration_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    ACTIONS(1616), 11,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1614), 30,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10417] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(715), 1,
      sym_text_interpolation,
    ACTIONS(1486), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      aux_sym_else_clause_token1,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1484), 34,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10478] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(716), 1,
      sym_text_interpolation,
    ACTIONS(1134), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      aux_sym_else_clause_token1,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1132), 34,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10539] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(717), 1,
      sym_text_interpolation,
    ACTIONS(1612), 13,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1610), 34,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      aux_sym_class_interface_clause_token1,
      anon_sym_EQ_GT,
      aux_sym_use_instead_of_clause_token1,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10600] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1816), 1,
      anon_sym_LPAREN2,
    STATE(718), 1,
      sym_text_interpolation,
    STATE(739), 1,
      sym_arguments,
    ACTIONS(1588), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1586), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10665] = 29,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(242), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(244), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(246), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(248), 1,
      sym_var_modifier,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1820), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(1822), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(1824), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(1828), 1,
      anon_sym_LPAREN2,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    ACTIONS(1832), 1,
      anon_sym_DOLLAR,
    STATE(719), 1,
      sym_text_interpolation,
    STATE(1088), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1494), 1,
      sym__types,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1601), 1,
      sym_property_element,
    STATE(1757), 1,
      sym__function_definition_header,
    STATE(1810), 1,
      sym_variable_name,
    STATE(2031), 1,
      sym_type,
    STATE(2506), 1,
      sym_namespace_name,
    ACTIONS(250), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [10774] = 29,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(242), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(244), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(246), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(248), 1,
      sym_var_modifier,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1820), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(1822), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(1824), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(1828), 1,
      anon_sym_LPAREN2,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    ACTIONS(1832), 1,
      anon_sym_DOLLAR,
    STATE(720), 1,
      sym_text_interpolation,
    STATE(1088), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1494), 1,
      sym__types,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1565), 1,
      sym_property_element,
    STATE(1730), 1,
      sym__function_definition_header,
    STATE(1810), 1,
      sym_variable_name,
    STATE(2289), 1,
      sym_type,
    STATE(2506), 1,
      sym_namespace_name,
    ACTIONS(250), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [10883] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1816), 1,
      anon_sym_LPAREN2,
    STATE(721), 1,
      sym_text_interpolation,
    STATE(740), 1,
      sym_arguments,
    ACTIONS(1594), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1592), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10948] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(722), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1668), 14,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1676), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11013] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(723), 1,
      sym_text_interpolation,
    ACTIONS(1616), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1614), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11073] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(724), 1,
      sym_text_interpolation,
    ACTIONS(1688), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1686), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11133] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(725), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 26,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11199] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(726), 1,
      sym_text_interpolation,
    ACTIONS(1814), 9,
      aux_sym_function_static_declaration_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    ACTIONS(1616), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1614), 27,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11261] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(727), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1668), 14,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1676), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11323] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(728), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1668), 14,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1676), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11385] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(729), 1,
      sym_text_interpolation,
    ACTIONS(1604), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1602), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11445] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1602), 1,
      anon_sym_LPAREN2,
    STATE(730), 1,
      sym_text_interpolation,
    ACTIONS(1628), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1626), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11507] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(731), 1,
      sym_text_interpolation,
    ACTIONS(1612), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1610), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11567] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(732), 1,
      sym_text_interpolation,
    ACTIONS(1640), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1638), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11627] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(733), 1,
      sym_text_interpolation,
    ACTIONS(1580), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1578), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11687] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1600), 1,
      anon_sym_LPAREN2,
    STATE(734), 1,
      sym_text_interpolation,
    ACTIONS(1598), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1596), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11749] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(735), 1,
      sym_text_interpolation,
    ACTIONS(1642), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1600), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11809] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(736), 1,
      sym_text_interpolation,
    ACTIONS(1650), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1648), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11869] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(737), 1,
      sym_text_interpolation,
    ACTIONS(1584), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1582), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11929] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(738), 1,
      sym_text_interpolation,
    ACTIONS(1684), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1682), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11989] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(739), 1,
      sym_text_interpolation,
    ACTIONS(1700), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1698), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12049] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(740), 1,
      sym_text_interpolation,
    ACTIONS(1704), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1702), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12109] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(741), 1,
      sym_text_interpolation,
    ACTIONS(1696), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1694), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12169] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(742), 1,
      sym_text_interpolation,
    ACTIONS(1570), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1568), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12229] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(743), 1,
      sym_text_interpolation,
    ACTIONS(1646), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1644), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12289] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(744), 1,
      sym_text_interpolation,
    ACTIONS(1576), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1574), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12349] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(745), 1,
      sym_text_interpolation,
    ACTIONS(1616), 13,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1614), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      aux_sym_use_instead_of_clause_token1,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12409] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(746), 1,
      sym_text_interpolation,
    ACTIONS(1608), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1606), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12469] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(747), 1,
      sym_text_interpolation,
    ACTIONS(1620), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1618), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12529] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(748), 1,
      sym_text_interpolation,
    ACTIONS(1624), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1622), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12589] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1834), 1,
      anon_sym_BSLASH,
    STATE(749), 1,
      sym_text_interpolation,
    STATE(2416), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1610), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12655] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(750), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 26,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12718] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(751), 1,
      sym_text_interpolation,
    ACTIONS(1680), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1678), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12777] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(752), 1,
      sym_text_interpolation,
    ACTIONS(1680), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1678), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12836] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(753), 1,
      sym_text_interpolation,
    ACTIONS(1632), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1630), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12895] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(754), 1,
      sym_text_interpolation,
    ACTIONS(1636), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1634), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12954] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(755), 1,
      sym_text_interpolation,
    ACTIONS(1692), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1690), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13013] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(756), 1,
      sym_text_interpolation,
    ACTIONS(1674), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 26,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13076] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(757), 1,
      sym_text_interpolation,
    ACTIONS(1628), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1626), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13135] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(758), 1,
      sym_text_interpolation,
    ACTIONS(1654), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1652), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13194] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1837), 1,
      anon_sym_LPAREN2,
    STATE(759), 1,
      sym_text_interpolation,
    STATE(829), 1,
      sym_arguments,
    ACTIONS(1588), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1586), 31,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13257] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1837), 1,
      anon_sym_LPAREN2,
    STATE(760), 1,
      sym_text_interpolation,
    STATE(830), 1,
      sym_arguments,
    ACTIONS(1594), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1592), 31,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13320] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(761), 1,
      sym_text_interpolation,
    ACTIONS(1658), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1656), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13379] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(762), 1,
      sym_text_interpolation,
    ACTIONS(1662), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1660), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13438] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1600), 1,
      anon_sym_LPAREN2,
    STATE(763), 1,
      sym_text_interpolation,
    ACTIONS(1841), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1839), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      aux_sym_use_instead_of_clause_token1,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13499] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1644), 1,
      anon_sym_LPAREN2,
    STATE(764), 1,
      sym_text_interpolation,
    ACTIONS(1845), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1843), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      aux_sym_use_instead_of_clause_token1,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13560] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(765), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1849), 13,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1847), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13623] = 28,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(1853), 1,
      anon_sym_COMMA,
    ACTIONS(1855), 1,
      anon_sym_BSLASH,
    ACTIONS(1857), 1,
      anon_sym_LPAREN2,
    ACTIONS(1859), 1,
      anon_sym_RPAREN2,
    ACTIONS(1861), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1863), 1,
      anon_sym_QMARK,
    ACTIONS(1865), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(766), 1,
      sym_text_interpolation,
    STATE(1104), 1,
      sym_attribute_list,
    STATE(1230), 1,
      sym_visibility_modifier,
    STATE(1342), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1346), 1,
      sym_attribute_group,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1582), 1,
      sym__types,
    STATE(1825), 1,
      sym_type,
    STATE(1995), 1,
      sym_variable_name,
    STATE(2297), 1,
      sym_reference_modifier,
    STATE(2644), 1,
      sym_namespace_name,
    ACTIONS(250), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(2174), 3,
      sym_property_promotion_parameter,
      sym_simple_parameter,
      sym_variadic_parameter,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [13727] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(767), 1,
      sym_text_interpolation,
    ACTIONS(1871), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1869), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13785] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(768), 1,
      sym_text_interpolation,
    ACTIONS(1875), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1873), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13843] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(769), 1,
      sym_text_interpolation,
    ACTIONS(1879), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1877), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13901] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(770), 1,
      sym_text_interpolation,
    ACTIONS(1883), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1881), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13959] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(771), 1,
      sym_text_interpolation,
    ACTIONS(1887), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1885), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14017] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(772), 1,
      sym_text_interpolation,
    ACTIONS(1891), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1889), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14075] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(773), 1,
      sym_text_interpolation,
    ACTIONS(1895), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1893), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14133] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(774), 1,
      sym_text_interpolation,
    ACTIONS(1899), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1897), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14191] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(775), 1,
      sym_text_interpolation,
    ACTIONS(1903), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1901), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14249] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(776), 1,
      sym_text_interpolation,
    ACTIONS(1907), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1905), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14307] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(777), 1,
      sym_text_interpolation,
    ACTIONS(1911), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1909), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14365] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(778), 1,
      sym_text_interpolation,
    ACTIONS(1915), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1913), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14423] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(779), 1,
      sym_text_interpolation,
    ACTIONS(1919), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1917), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14481] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(780), 1,
      sym_text_interpolation,
    ACTIONS(1923), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1921), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14539] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(781), 1,
      sym_text_interpolation,
    ACTIONS(1927), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1925), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14597] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(782), 1,
      sym_text_interpolation,
    ACTIONS(1931), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1929), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14655] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(783), 1,
      sym_text_interpolation,
    ACTIONS(1935), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1933), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14713] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(784), 1,
      sym_text_interpolation,
    ACTIONS(1939), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1937), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14771] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(785), 1,
      sym_text_interpolation,
    ACTIONS(1943), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1941), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14829] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(786), 1,
      sym_text_interpolation,
    ACTIONS(1947), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1945), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14887] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(787), 1,
      sym_text_interpolation,
    ACTIONS(1951), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1949), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14945] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(788), 1,
      sym_text_interpolation,
    ACTIONS(1957), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1955), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1953), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15005] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(789), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1668), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1676), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15067] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(790), 1,
      sym_text_interpolation,
    ACTIONS(1042), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      aux_sym_else_clause_token1,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1040), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15125] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(791), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15187] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1963), 1,
      anon_sym_LPAREN2,
    STATE(792), 1,
      sym_text_interpolation,
    ACTIONS(1965), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1961), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1959), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15249] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(793), 1,
      sym_text_interpolation,
    ACTIONS(1969), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1967), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15307] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1975), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1977), 1,
      anon_sym_DASH_GT,
    ACTIONS(1979), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(1981), 1,
      anon_sym_LBRACK,
    STATE(794), 1,
      sym_text_interpolation,
    ACTIONS(1973), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1971), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15373] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(795), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1985), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1983), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15435] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1994), 1,
      anon_sym_EQ,
    STATE(796), 1,
      sym_text_interpolation,
    ACTIONS(1991), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1989), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1987), 29,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15497] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(797), 1,
      sym_text_interpolation,
    ACTIONS(1998), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1996), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15555] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(798), 1,
      sym_text_interpolation,
    ACTIONS(2002), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2000), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15613] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(799), 1,
      sym_text_interpolation,
    ACTIONS(1058), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      aux_sym_else_clause_token1,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1056), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15671] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(800), 1,
      sym_text_interpolation,
    ACTIONS(1564), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1566), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15729] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(801), 1,
      sym_text_interpolation,
    ACTIONS(2006), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2004), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15787] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(802), 1,
      sym_text_interpolation,
    ACTIONS(2008), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15847] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(803), 1,
      sym_text_interpolation,
    ACTIONS(2012), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2010), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15905] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(804), 1,
      sym_text_interpolation,
    ACTIONS(2016), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2014), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15963] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(805), 1,
      sym_text_interpolation,
    ACTIONS(2020), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2018), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16021] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(806), 1,
      sym_text_interpolation,
    ACTIONS(2024), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2022), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16079] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(807), 1,
      sym_text_interpolation,
    ACTIONS(2028), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2026), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16137] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(808), 1,
      sym_text_interpolation,
    ACTIONS(2032), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2030), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16195] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(809), 1,
      sym_text_interpolation,
    ACTIONS(1680), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1678), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16253] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(810), 1,
      sym_text_interpolation,
    ACTIONS(1680), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1678), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16311] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(811), 1,
      sym_text_interpolation,
    ACTIONS(1692), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1690), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16369] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(812), 1,
      sym_text_interpolation,
    ACTIONS(1604), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1602), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16427] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(813), 1,
      sym_text_interpolation,
    ACTIONS(1628), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1626), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16485] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1602), 1,
      anon_sym_LPAREN2,
    STATE(814), 1,
      sym_text_interpolation,
    ACTIONS(1628), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1626), 31,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16545] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(815), 1,
      sym_text_interpolation,
    ACTIONS(1957), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16605] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(816), 1,
      sym_text_interpolation,
    ACTIONS(1640), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1638), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16663] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(817), 1,
      sym_text_interpolation,
    ACTIONS(1580), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1578), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16721] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(818), 1,
      sym_text_interpolation,
    ACTIONS(1632), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1630), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16779] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1600), 1,
      anon_sym_LPAREN2,
    STATE(819), 1,
      sym_text_interpolation,
    ACTIONS(1598), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1596), 31,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16839] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(820), 1,
      sym_text_interpolation,
    ACTIONS(2036), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2034), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16897] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(821), 1,
      sym_text_interpolation,
    ACTIONS(1642), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1600), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16955] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(822), 1,
      sym_text_interpolation,
    ACTIONS(2040), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2038), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17013] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(823), 1,
      sym_text_interpolation,
    ACTIONS(1650), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1648), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17071] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(824), 1,
      sym_text_interpolation,
    ACTIONS(1654), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1652), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17129] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(825), 1,
      sym_text_interpolation,
    ACTIONS(1658), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1656), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17187] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(826), 1,
      sym_text_interpolation,
    ACTIONS(1662), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1660), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17245] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(827), 1,
      sym_text_interpolation,
    ACTIONS(1584), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1582), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17303] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(828), 1,
      sym_text_interpolation,
    ACTIONS(1684), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1682), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17361] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(829), 1,
      sym_text_interpolation,
    ACTIONS(1700), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1698), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17419] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(830), 1,
      sym_text_interpolation,
    ACTIONS(1704), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1702), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17477] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(831), 1,
      sym_text_interpolation,
    ACTIONS(1696), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1694), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17535] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(832), 1,
      sym_text_interpolation,
    ACTIONS(1636), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1634), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17593] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(833), 1,
      sym_text_interpolation,
    ACTIONS(1570), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1568), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17651] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(834), 1,
      sym_text_interpolation,
    ACTIONS(1646), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1644), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17709] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(835), 1,
      sym_text_interpolation,
    ACTIONS(1576), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1574), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17767] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(836), 1,
      sym_text_interpolation,
    ACTIONS(2044), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2042), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17825] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(837), 1,
      sym_text_interpolation,
    ACTIONS(2048), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2046), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17883] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(838), 1,
      sym_text_interpolation,
    ACTIONS(2052), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2050), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17941] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(839), 1,
      sym_text_interpolation,
    ACTIONS(2056), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2054), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17999] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(840), 1,
      sym_text_interpolation,
    ACTIONS(1608), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1606), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18057] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(841), 1,
      sym_text_interpolation,
    ACTIONS(1620), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1618), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18115] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(842), 1,
      sym_text_interpolation,
    ACTIONS(1624), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1622), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18173] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(843), 1,
      sym_text_interpolation,
    ACTIONS(1688), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1686), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18231] = 28,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(1855), 1,
      anon_sym_BSLASH,
    ACTIONS(1857), 1,
      anon_sym_LPAREN2,
    ACTIONS(1861), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1863), 1,
      anon_sym_QMARK,
    ACTIONS(1865), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(2058), 1,
      anon_sym_COMMA,
    ACTIONS(2060), 1,
      anon_sym_RPAREN2,
    STATE(844), 1,
      sym_text_interpolation,
    STATE(1104), 1,
      sym_attribute_list,
    STATE(1230), 1,
      sym_visibility_modifier,
    STATE(1342), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1346), 1,
      sym_attribute_group,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1582), 1,
      sym__types,
    STATE(1825), 1,
      sym_type,
    STATE(1995), 1,
      sym_variable_name,
    STATE(2297), 1,
      sym_reference_modifier,
    STATE(2644), 1,
      sym_namespace_name,
    ACTIONS(250), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(2026), 3,
      sym_property_promotion_parameter,
      sym_simple_parameter,
      sym_variadic_parameter,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [18335] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(845), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1849), 13,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1847), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18395] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(846), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1849), 13,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1847), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18455] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(847), 1,
      sym_text_interpolation,
    ACTIONS(2064), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2062), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18513] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(848), 1,
      sym_text_interpolation,
    ACTIONS(2068), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2066), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18571] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(849), 1,
      sym_text_interpolation,
    ACTIONS(1989), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1987), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18629] = 27,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(1855), 1,
      anon_sym_BSLASH,
    ACTIONS(1857), 1,
      anon_sym_LPAREN2,
    ACTIONS(1861), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1863), 1,
      anon_sym_QMARK,
    ACTIONS(1865), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(2070), 1,
      anon_sym_RPAREN2,
    STATE(850), 1,
      sym_text_interpolation,
    STATE(1104), 1,
      sym_attribute_list,
    STATE(1230), 1,
      sym_visibility_modifier,
    STATE(1342), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1346), 1,
      sym_attribute_group,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1582), 1,
      sym__types,
    STATE(1825), 1,
      sym_type,
    STATE(1995), 1,
      sym_variable_name,
    STATE(2297), 1,
      sym_reference_modifier,
    STATE(2644), 1,
      sym_namespace_name,
    ACTIONS(250), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(2318), 3,
      sym_property_promotion_parameter,
      sym_simple_parameter,
      sym_variadic_parameter,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [18730] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(851), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1668), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1676), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18789] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(852), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1668), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1676), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18848] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(853), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18907] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1834), 1,
      anon_sym_BSLASH,
    STATE(854), 1,
      sym_text_interpolation,
    STATE(2416), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1612), 11,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1610), 30,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18968] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1807), 1,
      anon_sym_BSLASH,
    STATE(855), 1,
      sym_text_interpolation,
    STATE(2472), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1803), 11,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1805), 30,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19029] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(856), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1985), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1983), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19088] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(857), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1985), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1983), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19147] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1807), 1,
      anon_sym_BSLASH,
    STATE(858), 1,
      sym_text_interpolation,
    STATE(2472), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1803), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1805), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19208] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2072), 1,
      anon_sym_LPAREN2,
    STATE(859), 1,
      sym_text_interpolation,
    STATE(881), 1,
      sym_arguments,
    ACTIONS(1588), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1586), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19269] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2072), 1,
      anon_sym_LPAREN2,
    STATE(860), 1,
      sym_text_interpolation,
    STATE(883), 1,
      sym_arguments,
    ACTIONS(1594), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1592), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19330] = 27,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(1855), 1,
      anon_sym_BSLASH,
    ACTIONS(1857), 1,
      anon_sym_LPAREN2,
    ACTIONS(1861), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1863), 1,
      anon_sym_QMARK,
    ACTIONS(1865), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(2074), 1,
      anon_sym_RPAREN2,
    STATE(861), 1,
      sym_text_interpolation,
    STATE(1104), 1,
      sym_attribute_list,
    STATE(1230), 1,
      sym_visibility_modifier,
    STATE(1342), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1346), 1,
      sym_attribute_group,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1582), 1,
      sym__types,
    STATE(1825), 1,
      sym_type,
    STATE(1995), 1,
      sym_variable_name,
    STATE(2297), 1,
      sym_reference_modifier,
    STATE(2644), 1,
      sym_namespace_name,
    ACTIONS(250), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(2318), 3,
      sym_property_promotion_parameter,
      sym_simple_parameter,
      sym_variadic_parameter,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [19431] = 27,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(1855), 1,
      anon_sym_BSLASH,
    ACTIONS(1857), 1,
      anon_sym_LPAREN2,
    ACTIONS(1861), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1863), 1,
      anon_sym_QMARK,
    ACTIONS(1865), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(2076), 1,
      anon_sym_RPAREN2,
    STATE(862), 1,
      sym_text_interpolation,
    STATE(1104), 1,
      sym_attribute_list,
    STATE(1230), 1,
      sym_visibility_modifier,
    STATE(1342), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1346), 1,
      sym_attribute_group,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1582), 1,
      sym__types,
    STATE(1825), 1,
      sym_type,
    STATE(1995), 1,
      sym_variable_name,
    STATE(2297), 1,
      sym_reference_modifier,
    STATE(2644), 1,
      sym_namespace_name,
    ACTIONS(250), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(2318), 3,
      sym_property_promotion_parameter,
      sym_simple_parameter,
      sym_variadic_parameter,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [19532] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(863), 1,
      sym_text_interpolation,
    ACTIONS(1616), 12,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1614), 31,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      aux_sym_class_interface_clause_token1,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19589] = 27,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(1855), 1,
      anon_sym_BSLASH,
    ACTIONS(1857), 1,
      anon_sym_LPAREN2,
    ACTIONS(1861), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1863), 1,
      anon_sym_QMARK,
    ACTIONS(1865), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(2078), 1,
      anon_sym_RPAREN2,
    STATE(864), 1,
      sym_text_interpolation,
    STATE(1104), 1,
      sym_attribute_list,
    STATE(1230), 1,
      sym_visibility_modifier,
    STATE(1342), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1346), 1,
      sym_attribute_group,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1582), 1,
      sym__types,
    STATE(1825), 1,
      sym_type,
    STATE(1995), 1,
      sym_variable_name,
    STATE(2297), 1,
      sym_reference_modifier,
    STATE(2644), 1,
      sym_namespace_name,
    ACTIONS(250), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(2318), 3,
      sym_property_promotion_parameter,
      sym_simple_parameter,
      sym_variadic_parameter,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [19690] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1600), 1,
      anon_sym_LPAREN2,
    STATE(865), 1,
      sym_text_interpolation,
    ACTIONS(1598), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1596), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19748] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(866), 1,
      sym_text_interpolation,
    ACTIONS(1576), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1574), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19804] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(867), 1,
      sym_text_interpolation,
    ACTIONS(1640), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1638), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19860] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(868), 1,
      sym_text_interpolation,
    ACTIONS(1608), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1606), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19916] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(869), 1,
      sym_text_interpolation,
    ACTIONS(1580), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1578), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19972] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(870), 1,
      sym_text_interpolation,
    ACTIONS(1646), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1644), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20028] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1991), 1,
      anon_sym_RPAREN2,
    ACTIONS(1994), 1,
      anon_sym_EQ,
    STATE(871), 1,
      sym_text_interpolation,
    ACTIONS(1989), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1987), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20088] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(872), 1,
      sym_text_interpolation,
    ACTIONS(1722), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 23,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20150] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(873), 1,
      sym_text_interpolation,
    ACTIONS(1642), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1600), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20206] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(874), 1,
      sym_text_interpolation,
    ACTIONS(1620), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1618), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20262] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(875), 1,
      sym_text_interpolation,
    ACTIONS(1650), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1648), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20318] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(876), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1668), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1676), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20378] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(877), 1,
      sym_text_interpolation,
    ACTIONS(1688), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1686), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20434] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(878), 1,
      sym_text_interpolation,
    ACTIONS(1612), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1610), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20490] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(879), 1,
      sym_text_interpolation,
    ACTIONS(1584), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1582), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20546] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(880), 1,
      sym_text_interpolation,
    ACTIONS(1684), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1682), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20602] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(881), 1,
      sym_text_interpolation,
    ACTIONS(1700), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1698), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20658] = 35,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(242), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(244), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(246), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(248), 1,
      sym_var_modifier,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1797), 1,
      anon_sym_DOLLAR,
    ACTIONS(2080), 1,
      sym_name,
    ACTIONS(2082), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2084), 1,
      anon_sym_BSLASH,
    ACTIONS(2086), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2088), 1,
      anon_sym_LPAREN2,
    STATE(793), 1,
      sym_anonymous_class,
    STATE(794), 1,
      sym__new_variable,
    STATE(809), 1,
      sym__variable_member_access_expression,
    STATE(810), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(811), 1,
      sym__variable_scoped_property_access_expression,
    STATE(882), 1,
      sym_text_interpolation,
    STATE(907), 1,
      sym__reserved_identifier,
    STATE(916), 1,
      sym__name,
    STATE(917), 1,
      sym__class_name_reference,
    STATE(982), 1,
      sym_parenthesized_expression,
    STATE(1338), 1,
      sym__modifier,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1391), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1393), 1,
      sym_attribute_list,
    STATE(2651), 1,
      sym_namespace_name,
    ACTIONS(2090), 2,
      anon_sym_self,
      anon_sym_parent,
    STATE(813), 2,
      sym__variable_subscript_expression,
      sym__simple_variable,
    STATE(840), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(908), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(250), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [20774] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(883), 1,
      sym_text_interpolation,
    ACTIONS(1704), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1702), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20830] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(884), 1,
      sym_text_interpolation,
    ACTIONS(1616), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1614), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20886] = 35,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(242), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(244), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(246), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(248), 1,
      sym_var_modifier,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(1726), 1,
      sym_name,
    ACTIONS(2082), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2086), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2092), 1,
      anon_sym_LPAREN2,
    STATE(614), 1,
      sym__variable_member_access_expression,
    STATE(616), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(618), 1,
      sym__variable_scoped_property_access_expression,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(793), 1,
      sym_anonymous_class,
    STATE(885), 1,
      sym_text_interpolation,
    STATE(982), 1,
      sym_parenthesized_expression,
    STATE(1338), 1,
      sym__modifier,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1391), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1393), 1,
      sym_attribute_list,
    STATE(1783), 1,
      sym__new_variable,
    STATE(2388), 1,
      sym__name,
    STATE(2471), 1,
      sym__class_name_reference,
    STATE(2506), 1,
      sym_namespace_name,
    ACTIONS(2090), 2,
      anon_sym_self,
      anon_sym_parent,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(605), 2,
      sym__variable_subscript_expression,
      sym__simple_variable,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(250), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [21002] = 26,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(1855), 1,
      anon_sym_BSLASH,
    ACTIONS(1857), 1,
      anon_sym_LPAREN2,
    ACTIONS(1861), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1863), 1,
      anon_sym_QMARK,
    ACTIONS(1865), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(886), 1,
      sym_text_interpolation,
    STATE(1104), 1,
      sym_attribute_list,
    STATE(1230), 1,
      sym_visibility_modifier,
    STATE(1342), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1346), 1,
      sym_attribute_group,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1582), 1,
      sym__types,
    STATE(1825), 1,
      sym_type,
    STATE(1995), 1,
      sym_variable_name,
    STATE(2297), 1,
      sym_reference_modifier,
    STATE(2644), 1,
      sym_namespace_name,
    ACTIONS(250), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(2318), 3,
      sym_property_promotion_parameter,
      sym_simple_parameter,
      sym_variadic_parameter,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [21100] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1834), 1,
      anon_sym_BSLASH,
    STATE(887), 1,
      sym_text_interpolation,
    STATE(2416), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1610), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21162] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(888), 1,
      sym_text_interpolation,
    ACTIONS(1696), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1694), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21218] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1602), 1,
      anon_sym_LPAREN2,
    STATE(889), 1,
      sym_text_interpolation,
    ACTIONS(1628), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1626), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21276] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(890), 1,
      sym_text_interpolation,
    ACTIONS(1604), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1602), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21332] = 35,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(242), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(244), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(246), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(248), 1,
      sym_var_modifier,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1801), 1,
      anon_sym_DOLLAR,
    ACTIONS(2094), 1,
      sym_name,
    ACTIONS(2096), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2098), 1,
      anon_sym_BSLASH,
    ACTIONS(2100), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2102), 1,
      anon_sym_LPAREN2,
    STATE(891), 1,
      sym_text_interpolation,
    STATE(975), 1,
      sym_anonymous_class,
    STATE(977), 1,
      sym__new_variable,
    STATE(997), 1,
      sym__variable_member_access_expression,
    STATE(1000), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1001), 1,
      sym__variable_scoped_property_access_expression,
    STATE(1043), 1,
      sym__reserved_identifier,
    STATE(1046), 1,
      sym__class_name_reference,
    STATE(1048), 1,
      sym__name,
    STATE(1051), 1,
      sym_parenthesized_expression,
    STATE(1338), 1,
      sym__modifier,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1390), 1,
      sym_attribute_list,
    STATE(1392), 1,
      aux_sym_class_declaration_repeat1,
    STATE(2565), 1,
      sym_namespace_name,
    ACTIONS(2104), 2,
      anon_sym_self,
      anon_sym_parent,
    STATE(952), 2,
      sym__variable_subscript_expression,
      sym__simple_variable,
    STATE(957), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1044), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(250), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [21448] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(892), 1,
      sym_text_interpolation,
    ACTIONS(2108), 18,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_AT,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_LBRACK,
      anon_sym_POUND_LBRACK,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
    ACTIONS(2106), 24,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__arrow_function_header_token1,
      aux_sym_cast_type_token1,
      sym_float,
      sym_integer,
      aux_sym_throw_expression_token1,
      aux_sym_match_expression_token1,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_clone_expression_token1,
      aux_sym_print_intrinsic_token1,
      aux_sym__new_non_dereferencable_expression_token1,
      aux_sym__list_destructing_token1,
      anon_sym_self,
      anon_sym_parent,
      aux_sym__argument_name_token1,
      aux_sym__argument_name_token2,
      aux_sym_yield_expression_token1,
      aux_sym_include_expression_token1,
      aux_sym_include_once_expression_token1,
      aux_sym_require_expression_token1,
      aux_sym_require_once_expression_token1,
      sym_name,
  [21504] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(893), 1,
      sym_text_interpolation,
    ACTIONS(1570), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1568), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21560] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(894), 1,
      sym_text_interpolation,
    ACTIONS(1624), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1622), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21616] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1994), 1,
      anon_sym_EQ,
    STATE(895), 1,
      sym_text_interpolation,
    ACTIONS(1989), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1987), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21673] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(896), 1,
      sym_text_interpolation,
    ACTIONS(1810), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1812), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21728] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2110), 1,
      anon_sym_LPAREN2,
    STATE(897), 1,
      sym_text_interpolation,
    STATE(994), 1,
      sym_arguments,
    ACTIONS(1588), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1586), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21787] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2110), 1,
      anon_sym_LPAREN2,
    STATE(898), 1,
      sym_text_interpolation,
    STATE(919), 1,
      sym_arguments,
    ACTIONS(1594), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1592), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21846] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(899), 1,
      sym_text_interpolation,
    ACTIONS(1810), 11,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1812), 30,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21901] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(900), 1,
      sym_text_interpolation,
    ACTIONS(1722), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 23,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21960] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(901), 1,
      sym_text_interpolation,
    ACTIONS(1722), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 23,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22019] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(902), 1,
      sym_text_interpolation,
    ACTIONS(2114), 17,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_AT,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_LBRACK,
      anon_sym_POUND_LBRACK,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
    ACTIONS(2112), 24,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__arrow_function_header_token1,
      aux_sym_cast_type_token1,
      sym_float,
      sym_integer,
      aux_sym_throw_expression_token1,
      aux_sym_match_expression_token1,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_clone_expression_token1,
      aux_sym_print_intrinsic_token1,
      aux_sym__new_non_dereferencable_expression_token1,
      aux_sym__list_destructing_token1,
      anon_sym_self,
      anon_sym_parent,
      aux_sym__argument_name_token1,
      aux_sym__argument_name_token2,
      aux_sym_yield_expression_token1,
      aux_sym_include_expression_token1,
      aux_sym_include_once_expression_token1,
      aux_sym_require_expression_token1,
      aux_sym_require_once_expression_token1,
      sym_name,
  [22074] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(903), 1,
      sym_text_interpolation,
    ACTIONS(1680), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1678), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22129] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(904), 1,
      sym_text_interpolation,
    ACTIONS(1680), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1678), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22184] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(905), 1,
      sym_text_interpolation,
    ACTIONS(1692), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1690), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22239] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(906), 1,
      sym_text_interpolation,
    ACTIONS(1628), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1626), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22294] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(907), 1,
      sym_text_interpolation,
    ACTIONS(1612), 11,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1610), 30,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22349] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(908), 1,
      sym_text_interpolation,
    ACTIONS(2116), 11,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2008), 30,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22404] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(909), 1,
      sym_text_interpolation,
    ACTIONS(1632), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1630), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22459] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(910), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1668), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1676), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22516] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(911), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1668), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1676), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22573] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(912), 1,
      sym_text_interpolation,
    ACTIONS(1654), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1652), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22628] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(913), 1,
      sym_text_interpolation,
    ACTIONS(1658), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1656), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22683] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(914), 1,
      sym_text_interpolation,
    ACTIONS(1662), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1660), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22738] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(915), 1,
      sym_text_interpolation,
    ACTIONS(1636), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1634), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22793] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1975), 1,
      anon_sym_COLON_COLON,
    STATE(916), 1,
      sym_text_interpolation,
    ACTIONS(1973), 11,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1971), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22850] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1837), 1,
      anon_sym_LPAREN2,
    STATE(804), 1,
      sym_arguments,
    STATE(917), 1,
      sym_text_interpolation,
    ACTIONS(2120), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2118), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22909] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(918), 1,
      sym_text_interpolation,
    ACTIONS(2028), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2026), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22963] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(919), 1,
      sym_text_interpolation,
    ACTIONS(1704), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1702), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23017] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(920), 1,
      sym_text_interpolation,
    ACTIONS(1696), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1694), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23071] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(921), 1,
      sym_text_interpolation,
    ACTIONS(1636), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1634), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23125] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(922), 1,
      sym_text_interpolation,
    ACTIONS(1604), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1602), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23179] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(923), 1,
      sym_text_interpolation,
    ACTIONS(1134), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1132), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23233] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(924), 1,
      sym_text_interpolation,
    ACTIONS(1570), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1568), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23287] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(925), 1,
      sym_text_interpolation,
    ACTIONS(1646), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1644), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23341] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(926), 1,
      sym_text_interpolation,
    ACTIONS(1576), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1574), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23395] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(927), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1668), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1676), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23453] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(928), 1,
      sym_text_interpolation,
    ACTIONS(1939), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1937), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23507] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(929), 1,
      sym_text_interpolation,
    ACTIONS(1943), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1941), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23561] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(930), 1,
      sym_text_interpolation,
    ACTIONS(1947), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1945), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23615] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(931), 1,
      sym_text_interpolation,
    ACTIONS(2068), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2066), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23669] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(932), 1,
      sym_text_interpolation,
    ACTIONS(1564), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1566), 30,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23723] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(933), 1,
      sym_text_interpolation,
    ACTIONS(2048), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2046), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23777] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1834), 1,
      anon_sym_BSLASH,
    ACTIONS(2122), 1,
      anon_sym_COLON,
    STATE(934), 1,
      sym_text_interpolation,
    STATE(2416), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1610), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 21,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23839] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1602), 1,
      anon_sym_LPAREN2,
    STATE(935), 1,
      sym_text_interpolation,
    ACTIONS(1628), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1626), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23895] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(936), 1,
      sym_text_interpolation,
    ACTIONS(2008), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23951] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(937), 1,
      sym_text_interpolation,
    ACTIONS(1891), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1889), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24005] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(938), 1,
      sym_text_interpolation,
    ACTIONS(1895), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1893), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24059] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1963), 1,
      anon_sym_LPAREN2,
    STATE(939), 1,
      sym_text_interpolation,
    ACTIONS(1965), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1961), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1959), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24117] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(940), 1,
      sym_text_interpolation,
    ACTIONS(1957), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1955), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1953), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24173] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1994), 1,
      anon_sym_EQ,
    STATE(941), 1,
      sym_text_interpolation,
    ACTIONS(1989), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1987), 28,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24229] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(942), 1,
      sym_text_interpolation,
    ACTIONS(1951), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1949), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24283] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(943), 1,
      sym_text_interpolation,
    ACTIONS(1957), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24339] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(944), 1,
      sym_text_interpolation,
    ACTIONS(1620), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1618), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24393] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(945), 1,
      sym_text_interpolation,
    ACTIONS(1624), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1622), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24447] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(946), 1,
      sym_text_interpolation,
    ACTIONS(1640), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1638), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24501] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(947), 1,
      sym_text_interpolation,
    ACTIONS(1580), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1578), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24555] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(948), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24613] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(949), 1,
      sym_text_interpolation,
    ACTIONS(1612), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1610), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24667] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(950), 1,
      sym_text_interpolation,
    ACTIONS(1632), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1630), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24721] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1600), 1,
      anon_sym_LPAREN2,
    STATE(951), 1,
      sym_text_interpolation,
    ACTIONS(1598), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1596), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24777] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(952), 1,
      sym_text_interpolation,
    ACTIONS(1628), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1626), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24831] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(953), 1,
      sym_text_interpolation,
    ACTIONS(1616), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1614), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24885] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(954), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1849), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1847), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24943] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(955), 1,
      sym_text_interpolation,
    ACTIONS(1903), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1901), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24997] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(956), 1,
      sym_text_interpolation,
    ACTIONS(1642), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1600), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25051] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(957), 1,
      sym_text_interpolation,
    ACTIONS(1608), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1606), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25105] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(958), 1,
      sym_text_interpolation,
    ACTIONS(1650), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1648), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25159] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(959), 1,
      sym_text_interpolation,
    ACTIONS(1654), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1652), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25213] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(960), 1,
      sym_text_interpolation,
    ACTIONS(1658), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1656), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25267] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(961), 1,
      sym_text_interpolation,
    ACTIONS(1662), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1660), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25321] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(962), 1,
      sym_text_interpolation,
    ACTIONS(1871), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1869), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25375] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(963), 1,
      sym_text_interpolation,
    ACTIONS(1875), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1873), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25429] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(964), 1,
      sym_text_interpolation,
    ACTIONS(1486), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1484), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25483] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(965), 1,
      sym_text_interpolation,
    ACTIONS(1879), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1877), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25537] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(966), 1,
      sym_text_interpolation,
    ACTIONS(1899), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1897), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25591] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(967), 1,
      sym_text_interpolation,
    ACTIONS(1564), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1566), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25645] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(968), 1,
      sym_text_interpolation,
    ACTIONS(1935), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1933), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25699] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1644), 1,
      anon_sym_LPAREN2,
    STATE(969), 1,
      sym_text_interpolation,
    ACTIONS(1845), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1843), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25755] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(970), 1,
      sym_text_interpolation,
    ACTIONS(2036), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2034), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25809] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1834), 1,
      anon_sym_BSLASH,
    ACTIONS(2124), 1,
      anon_sym_COLON,
    STATE(971), 1,
      sym_text_interpolation,
    STATE(2416), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1610), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 21,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25871] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(972), 1,
      sym_text_interpolation,
    ACTIONS(1989), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1987), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25925] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(973), 1,
      sym_text_interpolation,
    ACTIONS(2012), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2010), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25979] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(974), 1,
      sym_text_interpolation,
    ACTIONS(2016), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2014), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26033] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(975), 1,
      sym_text_interpolation,
    ACTIONS(1969), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1967), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26087] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(976), 1,
      sym_text_interpolation,
    ACTIONS(2044), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2042), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26141] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2126), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2128), 1,
      anon_sym_DASH_GT,
    ACTIONS(2130), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(2132), 1,
      anon_sym_LBRACK,
    STATE(977), 1,
      sym_text_interpolation,
    ACTIONS(1973), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1971), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26203] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(978), 1,
      sym_text_interpolation,
    ACTIONS(1688), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1686), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26257] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(979), 1,
      sym_text_interpolation,
    ACTIONS(1883), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1881), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26311] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(980), 1,
      sym_text_interpolation,
    ACTIONS(1584), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1582), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26365] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(981), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1985), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1983), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26423] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(982), 1,
      sym_text_interpolation,
    ACTIONS(1973), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1971), 30,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26477] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(983), 1,
      sym_text_interpolation,
    ACTIONS(1887), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1885), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26531] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(984), 1,
      sym_text_interpolation,
    ACTIONS(1684), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1682), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26585] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(985), 1,
      sym_text_interpolation,
    ACTIONS(2020), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2018), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26639] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(986), 1,
      sym_text_interpolation,
    ACTIONS(2052), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2050), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26693] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(987), 1,
      sym_text_interpolation,
    ACTIONS(2024), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2022), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26747] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(988), 1,
      sym_text_interpolation,
    ACTIONS(2040), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2038), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26801] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(989), 1,
      sym_text_interpolation,
    ACTIONS(2032), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2030), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26855] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(990), 1,
      sym_text_interpolation,
    ACTIONS(1907), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1905), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26909] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(991), 1,
      sym_text_interpolation,
    ACTIONS(2056), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2054), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26963] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(992), 1,
      sym_text_interpolation,
    ACTIONS(2064), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2062), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27017] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1834), 1,
      anon_sym_BSLASH,
    ACTIONS(2134), 1,
      anon_sym_COLON,
    STATE(993), 1,
      sym_text_interpolation,
    STATE(2416), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1610), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 21,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27079] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(994), 1,
      sym_text_interpolation,
    ACTIONS(1700), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1698), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27133] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1600), 1,
      anon_sym_LPAREN2,
    STATE(995), 1,
      sym_text_interpolation,
    ACTIONS(1841), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1839), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27189] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1834), 1,
      anon_sym_BSLASH,
    STATE(996), 1,
      sym_text_interpolation,
    STATE(2416), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1612), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1610), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27247] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(997), 1,
      sym_text_interpolation,
    ACTIONS(1680), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1678), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27301] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(998), 1,
      sym_text_interpolation,
    ACTIONS(1998), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1996), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27355] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(999), 1,
      sym_text_interpolation,
    ACTIONS(2002), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2000), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27409] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1000), 1,
      sym_text_interpolation,
    ACTIONS(1680), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1678), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27463] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1001), 1,
      sym_text_interpolation,
    ACTIONS(1692), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1690), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27517] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1002), 1,
      sym_text_interpolation,
    ACTIONS(1911), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1909), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27571] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1003), 1,
      sym_text_interpolation,
    ACTIONS(1915), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1913), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27625] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1004), 1,
      sym_text_interpolation,
    ACTIONS(1919), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1917), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27679] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1005), 1,
      sym_text_interpolation,
    ACTIONS(1923), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1921), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27733] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1006), 1,
      sym_text_interpolation,
    ACTIONS(1927), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1925), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27787] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1007), 1,
      sym_text_interpolation,
    ACTIONS(1931), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1929), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27841] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1008), 1,
      sym_text_interpolation,
    ACTIONS(2006), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2004), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27895] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1009), 1,
      sym_text_interpolation,
    ACTIONS(2138), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2136), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27948] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1010), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1985), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1983), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28003] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1011), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1668), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1676), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28058] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1012), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1985), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1983), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28113] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1013), 1,
      sym_text_interpolation,
    ACTIONS(2142), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2140), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28166] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1014), 1,
      sym_text_interpolation,
    ACTIONS(2146), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2144), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28219] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1015), 1,
      sym_text_interpolation,
    ACTIONS(2150), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2148), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28272] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1016), 1,
      sym_text_interpolation,
    ACTIONS(2154), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2152), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28325] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1017), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1666), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28380] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1018), 1,
      sym_text_interpolation,
    ACTIONS(1955), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1953), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28433] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1019), 1,
      sym_text_interpolation,
    ACTIONS(2158), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2156), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28486] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1994), 1,
      anon_sym_EQ,
    STATE(1020), 1,
      sym_text_interpolation,
    ACTIONS(2160), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1989), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1987), 25,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28543] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1021), 1,
      sym_text_interpolation,
    ACTIONS(2164), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28596] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1022), 1,
      sym_text_interpolation,
    ACTIONS(2168), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2166), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28649] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1023), 1,
      sym_text_interpolation,
    ACTIONS(2172), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2170), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28702] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1024), 1,
      sym_text_interpolation,
    ACTIONS(2176), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2174), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28755] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1025), 1,
      sym_text_interpolation,
    ACTIONS(2180), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2178), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28808] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1026), 1,
      sym_text_interpolation,
    ACTIONS(2184), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2182), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28861] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1027), 1,
      sym_text_interpolation,
    ACTIONS(2188), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2186), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28914] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1028), 1,
      sym_text_interpolation,
    ACTIONS(2192), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2190), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28967] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1029), 1,
      sym_text_interpolation,
    ACTIONS(2196), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2194), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29020] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1030), 1,
      sym_text_interpolation,
    ACTIONS(2200), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2198), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29073] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1031), 1,
      sym_text_interpolation,
    ACTIONS(2204), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2202), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29126] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1032), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1849), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1847), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29181] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1033), 1,
      sym_text_interpolation,
    ACTIONS(2208), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2206), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29234] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1034), 1,
      sym_text_interpolation,
    ACTIONS(2212), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2210), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29287] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1035), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1668), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1676), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29342] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1036), 1,
      sym_text_interpolation,
    ACTIONS(1666), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29395] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1037), 1,
      sym_text_interpolation,
    ACTIONS(2216), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2214), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29448] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1038), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1849), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1847), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29503] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1807), 1,
      anon_sym_BSLASH,
    STATE(1039), 1,
      sym_text_interpolation,
    STATE(2472), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1803), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1805), 27,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29560] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1040), 1,
      sym_text_interpolation,
    ACTIONS(1961), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1959), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29613] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1041), 1,
      sym_text_interpolation,
    ACTIONS(2220), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2218), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29666] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2222), 1,
      aux_sym_binary_expression_token1,
    STATE(1042), 1,
      sym_text_interpolation,
    ACTIONS(2164), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29721] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1043), 1,
      sym_text_interpolation,
    ACTIONS(1612), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1610), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29773] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1044), 1,
      sym_text_interpolation,
    ACTIONS(2116), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2008), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29825] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1045), 1,
      sym_text_interpolation,
    ACTIONS(1616), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1614), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29877] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2110), 1,
      anon_sym_LPAREN2,
    STATE(974), 1,
      sym_arguments,
    STATE(1046), 1,
      sym_text_interpolation,
    ACTIONS(2120), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2118), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29932] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1047), 1,
      sym_text_interpolation,
    ACTIONS(1810), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1812), 27,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29983] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2126), 1,
      anon_sym_COLON_COLON,
    STATE(1048), 1,
      sym_text_interpolation,
    ACTIONS(1973), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1971), 26,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [30036] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    STATE(1049), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2154), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2152), 24,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
  [30092] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1050), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2230), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [30180] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1051), 1,
      sym_text_interpolation,
    ACTIONS(1973), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1971), 26,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [30230] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2266), 1,
      anon_sym_EQ_GT,
    STATE(1052), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2182), 6,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [30320] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1053), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2268), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [30408] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1054), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2270), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [30496] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1055), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2272), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [30584] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1056), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2274), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [30672] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1057), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2276), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [30760] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1058), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2278), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [30848] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1059), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2280), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [30936] = 15,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1060), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2154), 3,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 14,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [31006] = 17,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1061), 1,
      sym_text_interpolation,
    ACTIONS(2154), 2,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 13,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [31080] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1062), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2282), 10,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [31162] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2154), 1,
      anon_sym_QMARK,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1063), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 10,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [31244] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    STATE(1064), 1,
      sym_text_interpolation,
    ACTIONS(2154), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2152), 25,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [31296] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2286), 1,
      anon_sym_QMARK,
    STATE(1065), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2284), 10,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [31378] = 23,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1066), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 8,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token3,
  [31464] = 22,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1067), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 9,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [31548] = 19,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2154), 1,
      anon_sym_QMARK,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1068), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 12,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
  [31626] = 18,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2154), 1,
      anon_sym_QMARK,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1069), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 13,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [31702] = 16,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1070), 1,
      sym_text_interpolation,
    ACTIONS(2154), 2,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 14,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [31774] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1071), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2154), 5,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2152), 18,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [31840] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1072), 1,
      sym_text_interpolation,
    ACTIONS(1564), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1566), 26,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [31890] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1073), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2154), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2152), 19,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [31952] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    STATE(1074), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2154), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2152), 22,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
  [32010] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    STATE(1075), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2154), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2152), 20,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_DOT,
  [32070] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    STATE(1076), 1,
      sym_text_interpolation,
    ACTIONS(2154), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2152), 25,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [32122] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1077), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2288), 10,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [32204] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1078), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2290), 10,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [32286] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1079), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2292), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [32374] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1080), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2186), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [32462] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2296), 1,
      anon_sym_QMARK,
    STATE(1081), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2294), 10,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [32544] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1082), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 10,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [32626] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1083), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2272), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [32713] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1084), 1,
      sym_text_interpolation,
    ACTIONS(2172), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2170), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [32762] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1085), 1,
      sym_text_interpolation,
    ACTIONS(2216), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2214), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [32811] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1086), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2292), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [32898] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1087), 1,
      sym_text_interpolation,
    ACTIONS(1666), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [32947] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2380), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2385), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2388), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2391), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2394), 1,
      sym_var_modifier,
    STATE(1338), 1,
      sym__modifier,
    STATE(1088), 2,
      sym_text_interpolation,
      aux_sym_class_declaration_repeat1,
    ACTIONS(2383), 3,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOLLAR,
    ACTIONS(2397), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
    ACTIONS(2378), 17,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_class_declaration_token1,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [33012] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1089), 1,
      sym_text_interpolation,
    ACTIONS(2138), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2136), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [33061] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1090), 1,
      sym_text_interpolation,
    ACTIONS(2164), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [33110] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1091), 1,
      sym_text_interpolation,
    ACTIONS(2180), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2178), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [33159] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    ACTIONS(2400), 1,
      anon_sym_EQ_GT,
    STATE(1092), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2182), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      aux_sym_binary_expression_token1,
  [33248] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1093), 1,
      sym_text_interpolation,
    ACTIONS(2176), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2174), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [33297] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1094), 1,
      sym_text_interpolation,
    ACTIONS(2184), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2182), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [33346] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1095), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2268), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [33433] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1096), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2270), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [33520] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1097), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2272), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [33607] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1098), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2274), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [33694] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1099), 1,
      sym_text_interpolation,
    ACTIONS(1042), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1040), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [33743] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1100), 1,
      sym_text_interpolation,
    ACTIONS(2188), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2186), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [33792] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1101), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2292), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [33879] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1102), 1,
      sym_text_interpolation,
    ACTIONS(2204), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2202), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [33928] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1103), 1,
      sym_text_interpolation,
    ACTIONS(1961), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1959), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [33977] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(1855), 1,
      anon_sym_BSLASH,
    ACTIONS(1857), 1,
      anon_sym_LPAREN2,
    ACTIONS(1863), 1,
      anon_sym_QMARK,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(2402), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(1104), 1,
      sym_text_interpolation,
    STATE(1242), 1,
      sym_visibility_modifier,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1582), 1,
      sym__types,
    STATE(1802), 1,
      sym_type,
    STATE(2245), 1,
      sym_reference_modifier,
    STATE(2246), 1,
      sym_variable_name,
    STATE(2644), 1,
      sym_namespace_name,
    ACTIONS(250), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [34058] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1105), 1,
      sym_text_interpolation,
    ACTIONS(1058), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1056), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [34107] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1106), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2186), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [34194] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1107), 1,
      sym_text_interpolation,
    ACTIONS(2208), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2206), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [34243] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1108), 1,
      sym_text_interpolation,
    ACTIONS(2150), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2148), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [34292] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1109), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2278), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [34379] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1110), 1,
      sym_text_interpolation,
    ACTIONS(2220), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2218), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [34428] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1111), 1,
      sym_text_interpolation,
    STATE(1338), 1,
      sym__modifier,
    STATE(1389), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1453), 1,
      sym__const_declaration,
    ACTIONS(2406), 3,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOLLAR,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
    ACTIONS(2404), 24,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [34485] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1112), 1,
      sym_text_interpolation,
    ACTIONS(2212), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2210), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [34534] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1113), 1,
      sym_text_interpolation,
    ACTIONS(2192), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2190), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [34583] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1114), 1,
      sym_text_interpolation,
    ACTIONS(2146), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2144), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [34632] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2408), 1,
      aux_sym_binary_expression_token1,
    STATE(1115), 1,
      sym_text_interpolation,
    ACTIONS(2164), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [34683] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1116), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2280), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [34770] = 15,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1117), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2154), 3,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 13,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [34839] = 17,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1118), 1,
      sym_text_interpolation,
    ACTIONS(2154), 2,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 12,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [34912] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1119), 1,
      sym_text_interpolation,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2154), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2152), 23,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
  [34967] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2154), 1,
      anon_sym_QMARK,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1120), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 9,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [35048] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    STATE(1121), 1,
      sym_text_interpolation,
    ACTIONS(2154), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2152), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [35099] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1122), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 9,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [35180] = 23,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1123), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 7,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token3,
  [35265] = 22,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1124), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 8,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [35348] = 19,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2154), 1,
      anon_sym_QMARK,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1125), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 11,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
  [35425] = 18,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2154), 1,
      anon_sym_QMARK,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1126), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 12,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [35500] = 16,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1127), 1,
      sym_text_interpolation,
    ACTIONS(2154), 2,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 13,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [35571] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1128), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2154), 5,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2152), 17,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [35636] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1129), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2154), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2152), 18,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [35697] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1130), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2154), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2152), 21,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
  [35754] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1131), 1,
      sym_text_interpolation,
    ACTIONS(2158), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2156), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [35803] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    STATE(1132), 1,
      sym_text_interpolation,
    ACTIONS(2154), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2152), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [35854] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1133), 1,
      sym_text_interpolation,
    ACTIONS(2154), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2152), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [35903] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1134), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2288), 9,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [35984] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1135), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2290), 9,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [36065] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1136), 1,
      sym_text_interpolation,
    ACTIONS(2200), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2198), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [36114] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1137), 1,
      sym_text_interpolation,
    ACTIONS(2142), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2140), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [36163] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1138), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2278), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [36250] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2296), 1,
      anon_sym_QMARK,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1139), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2294), 9,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [36331] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1140), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2280), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [36418] = 15,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1141), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2154), 3,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 13,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [36487] = 17,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1142), 1,
      sym_text_interpolation,
    ACTIONS(2154), 2,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 12,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [36560] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1143), 1,
      sym_text_interpolation,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2154), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2152), 23,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
  [36615] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2154), 1,
      anon_sym_QMARK,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1144), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [36696] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    STATE(1145), 1,
      sym_text_interpolation,
    ACTIONS(2154), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2152), 24,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [36747] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1146), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [36828] = 23,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1147), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token3,
  [36913] = 22,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1148), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [36996] = 19,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2154), 1,
      anon_sym_QMARK,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1149), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 11,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
  [37073] = 18,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2154), 1,
      anon_sym_QMARK,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1150), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 12,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [37148] = 16,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1151), 1,
      sym_text_interpolation,
    ACTIONS(2154), 2,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 13,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [37219] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1152), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2154), 5,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2152), 17,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [37284] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1153), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2154), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2152), 18,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [37345] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1154), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2154), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2152), 21,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
  [37402] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1155), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2154), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2152), 19,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_DOT,
  [37461] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    STATE(1156), 1,
      sym_text_interpolation,
    ACTIONS(2154), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2152), 24,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [37512] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1157), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2288), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [37593] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1158), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2290), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [37674] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2296), 1,
      anon_sym_QMARK,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1159), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2294), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [37755] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1160), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2282), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [37836] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2286), 1,
      anon_sym_QMARK,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1161), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2284), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [37917] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1162), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2282), 9,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [37998] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1163), 1,
      sym_text_interpolation,
    STATE(1338), 1,
      sym__modifier,
    STATE(1389), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1443), 1,
      sym__const_declaration,
    ACTIONS(2406), 3,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOLLAR,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
    ACTIONS(2404), 24,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [38055] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1164), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2186), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [38142] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2286), 1,
      anon_sym_QMARK,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1165), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2284), 9,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [38223] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1166), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2276), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [38310] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1167), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2276), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [38397] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1168), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2230), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [38484] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2410), 1,
      anon_sym_EQ_GT,
    STATE(1169), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2182), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [38573] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1170), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2268), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [38660] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1171), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2270), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [38747] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1172), 1,
      sym_text_interpolation,
    ACTIONS(2168), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2166), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [38796] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1173), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2274), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [38883] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1174), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2230), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [38970] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1175), 1,
      sym_text_interpolation,
    ACTIONS(2196), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2194), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [39019] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1176), 1,
      sym_text_interpolation,
    ACTIONS(1955), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1953), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [39068] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1177), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2154), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2152), 19,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_DOT,
  [39127] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1178), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2276), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [39213] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1179), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2272), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [39299] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1180), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2278), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [39385] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1181), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2280), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [39471] = 15,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1182), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2154), 3,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 12,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [39539] = 17,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1183), 1,
      sym_text_interpolation,
    ACTIONS(2154), 2,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 11,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [39611] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1184), 1,
      sym_text_interpolation,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2154), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2152), 22,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
  [39665] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2154), 1,
      anon_sym_QMARK,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1185), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 8,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [39745] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    STATE(1186), 1,
      sym_text_interpolation,
    ACTIONS(2154), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2152), 23,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [39795] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1187), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 8,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [39875] = 23,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1188), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 6,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token3,
  [39959] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1189), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2274), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [40045] = 19,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2154), 1,
      anon_sym_QMARK,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1190), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 10,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
  [40121] = 18,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2154), 1,
      anon_sym_QMARK,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1191), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 11,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [40195] = 16,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1192), 1,
      sym_text_interpolation,
    ACTIONS(2154), 2,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 12,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [40265] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1193), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2154), 5,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2152), 16,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [40329] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1194), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2154), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2152), 17,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [40389] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1195), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2154), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2152), 20,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
  [40445] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1196), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2154), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2152), 18,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_DOT,
  [40503] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    STATE(1197), 1,
      sym_text_interpolation,
    ACTIONS(2154), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2152), 23,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [40553] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1198), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2288), 8,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [40633] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1199), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2290), 8,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [40713] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2296), 1,
      anon_sym_QMARK,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1200), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2294), 8,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [40793] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1201), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2282), 8,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [40873] = 21,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2286), 1,
      anon_sym_QMARK,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1202), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2284), 8,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [40953] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1203), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2292), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [41039] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1204), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2186), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [41125] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1205), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2230), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [41211] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    ACTIONS(2452), 1,
      anon_sym_EQ_GT,
    STATE(1206), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2182), 4,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41299] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1207), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2268), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [41385] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1208), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2270), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [41471] = 22,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1209), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2152), 7,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [41553] = 26,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    ACTIONS(2456), 1,
      anon_sym_LBRACE,
    STATE(1210), 1,
      sym_text_interpolation,
    STATE(2408), 1,
      sym_property_hook_list,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2454), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41642] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1211), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2458), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [41727] = 26,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    ACTIONS(2456), 1,
      anon_sym_LBRACE,
    STATE(1212), 1,
      sym_text_interpolation,
    STATE(2454), 1,
      sym_property_hook_list,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2460), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41816] = 26,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    ACTIONS(2456), 1,
      anon_sym_LBRACE,
    STATE(1213), 1,
      sym_text_interpolation,
    STATE(2325), 1,
      sym_property_hook_list,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2462), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41905] = 26,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    ACTIONS(2456), 1,
      anon_sym_LBRACE,
    STATE(1214), 1,
      sym_text_interpolation,
    STATE(2316), 1,
      sym_property_hook_list,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2464), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41994] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(959), 1,
      anon_sym_COLON,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1215), 1,
      sym_text_interpolation,
    ACTIONS(2204), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2202), 22,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [42043] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(959), 1,
      anon_sym_COLON,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1216), 1,
      sym_text_interpolation,
    ACTIONS(2220), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2218), 22,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [42092] = 26,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    ACTIONS(2456), 1,
      anon_sym_LBRACE,
    STATE(1217), 1,
      sym_text_interpolation,
    STATE(2320), 1,
      sym_property_hook_list,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2466), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42181] = 26,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    ACTIONS(2456), 1,
      anon_sym_LBRACE,
    STATE(1218), 1,
      sym_text_interpolation,
    STATE(2462), 1,
      sym_property_hook_list,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2468), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42270] = 26,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    ACTIONS(2456), 1,
      anon_sym_LBRACE,
    STATE(1219), 1,
      sym_text_interpolation,
    STATE(2463), 1,
      sym_property_hook_list,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2470), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42359] = 26,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    ACTIONS(2456), 1,
      anon_sym_LBRACE,
    STATE(1220), 1,
      sym_text_interpolation,
    STATE(2490), 1,
      sym_property_hook_list,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2472), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42448] = 18,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1828), 1,
      anon_sym_LPAREN2,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    ACTIONS(2474), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1221), 1,
      sym_text_interpolation,
    STATE(1494), 1,
      sym__types,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1602), 1,
      sym_type,
    STATE(1818), 1,
      sym_const_element,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2591), 1,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [42520] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    ACTIONS(2478), 1,
      anon_sym_COMMA,
    STATE(1222), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2476), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42606] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(532), 1,
      sym__semicolon,
    STATE(1223), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42692] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(525), 1,
      sym__semicolon,
    STATE(1224), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42778] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    ACTIONS(2480), 1,
      anon_sym_EQ_GT,
    STATE(1225), 1,
      sym_text_interpolation,
    ACTIONS(2182), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42864] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1226), 1,
      sym_text_interpolation,
    STATE(1793), 1,
      sym__semicolon,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42950] = 18,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1828), 1,
      anon_sym_LPAREN2,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    ACTIONS(2474), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1227), 1,
      sym_text_interpolation,
    STATE(1494), 1,
      sym__types,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1614), 1,
      sym_type,
    STATE(1778), 1,
      sym_const_element,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2591), 1,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [43022] = 18,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1828), 1,
      anon_sym_LPAREN2,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    ACTIONS(2474), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1228), 1,
      sym_text_interpolation,
    STATE(1494), 1,
      sym__types,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1617), 1,
      sym_type,
    STATE(1717), 1,
      sym_const_element,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2591), 1,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [43094] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(526), 1,
      sym__semicolon,
    STATE(1229), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43180] = 19,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(246), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(731), 1,
      anon_sym_AMP,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1828), 1,
      anon_sym_LPAREN2,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(1230), 1,
      sym_text_interpolation,
    STATE(1322), 1,
      sym_readonly_modifier,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1683), 1,
      sym__types,
    STATE(1874), 1,
      sym_type,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1800), 2,
      sym_variable_name,
      sym_by_ref,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [43254] = 26,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2484), 1,
      anon_sym_COMMA,
    ACTIONS(2486), 1,
      anon_sym_EQ_GT,
    STATE(1231), 1,
      sym_text_interpolation,
    STATE(2011), 1,
      aux_sym_match_condition_list_repeat1,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43342] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2488), 1,
      anon_sym_EQ_GT,
    STATE(1232), 1,
      sym_text_interpolation,
    ACTIONS(2182), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43428] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1233), 1,
      sym_text_interpolation,
    STATE(2269), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43514] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1234), 1,
      sym_text_interpolation,
    STATE(2102), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43600] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1235), 1,
      sym_text_interpolation,
    STATE(2270), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43686] = 18,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1828), 1,
      anon_sym_LPAREN2,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    ACTIONS(2474), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1236), 1,
      sym_text_interpolation,
    STATE(1494), 1,
      sym__types,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1591), 1,
      sym_type,
    STATE(1795), 1,
      sym_const_element,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2591), 1,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [43758] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1237), 1,
      sym_text_interpolation,
    STATE(2271), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43844] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1238), 1,
      sym_text_interpolation,
    STATE(1428), 1,
      sym__semicolon,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43930] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1239), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2490), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44014] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    ACTIONS(2478), 1,
      anon_sym_COMMA,
    STATE(1240), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44100] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1241), 1,
      sym_text_interpolation,
    STATE(1452), 1,
      sym__semicolon,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44186] = 19,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(246), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(731), 1,
      anon_sym_AMP,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1828), 1,
      anon_sym_LPAREN2,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(1242), 1,
      sym_text_interpolation,
    STATE(1276), 1,
      sym_readonly_modifier,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1683), 1,
      sym__types,
    STATE(1945), 1,
      sym_type,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1801), 2,
      sym_variable_name,
      sym_by_ref,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [44260] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2494), 1,
      anon_sym_EQ_GT,
    STATE(1243), 1,
      sym_text_interpolation,
    ACTIONS(2182), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44346] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(1244), 1,
      sym_text_interpolation,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2496), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44430] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2338), 1,
      anon_sym_AMP,
    ACTIONS(2340), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_PIPE,
    ACTIONS(2346), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2348), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2352), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2354), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2356), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2358), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2360), 1,
      anon_sym_CARET,
    ACTIONS(2368), 1,
      anon_sym_GT_EQ,
    ACTIONS(2372), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_PERCENT,
    STATE(515), 1,
      sym__semicolon,
    STATE(1245), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2362), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2370), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2374), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2366), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2364), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44516] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1246), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2498), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44599] = 27,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2502), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2504), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2506), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2508), 1,
      anon_sym_RBRACE,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    STATE(719), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1163), 1,
      sym_final_modifier,
    STATE(1247), 1,
      sym_text_interpolation,
    STATE(1269), 1,
      aux_sym_declaration_list_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1376), 1,
      sym_attribute_list,
    STATE(1433), 1,
      sym__member_declaration,
    STATE(1434), 1,
      sym__class_const_declaration,
    STATE(1435), 1,
      sym__const_declaration,
    STATE(1822), 1,
      sym__function_definition_header,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1436), 3,
      sym_property_declaration,
      sym_method_declaration,
      sym_use_declaration,
    STATE(1339), 4,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [44688] = 27,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2502), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2504), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2506), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2520), 1,
      anon_sym_RBRACE,
    STATE(719), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1163), 1,
      sym_final_modifier,
    STATE(1248), 1,
      sym_text_interpolation,
    STATE(1251), 1,
      aux_sym_declaration_list_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1376), 1,
      sym_attribute_list,
    STATE(1433), 1,
      sym__member_declaration,
    STATE(1434), 1,
      sym__class_const_declaration,
    STATE(1435), 1,
      sym__const_declaration,
    STATE(1822), 1,
      sym__function_definition_header,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1436), 3,
      sym_property_declaration,
      sym_method_declaration,
      sym_use_declaration,
    STATE(1339), 4,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [44777] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1249), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2522), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44860] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1250), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44943] = 27,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2502), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2504), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2506), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2526), 1,
      anon_sym_RBRACE,
    STATE(719), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1163), 1,
      sym_final_modifier,
    STATE(1251), 1,
      sym_text_interpolation,
    STATE(1269), 1,
      aux_sym_declaration_list_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1376), 1,
      sym_attribute_list,
    STATE(1433), 1,
      sym__member_declaration,
    STATE(1434), 1,
      sym__class_const_declaration,
    STATE(1435), 1,
      sym__const_declaration,
    STATE(1822), 1,
      sym__function_definition_header,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1436), 3,
      sym_property_declaration,
      sym_method_declaration,
      sym_use_declaration,
    STATE(1339), 4,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [45032] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2492), 1,
      anon_sym_SEMI,
    ACTIONS(2528), 1,
      anon_sym_COMMA,
    STATE(1252), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [45117] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    ACTIONS(2476), 1,
      anon_sym_RPAREN2,
    ACTIONS(2530), 1,
      anon_sym_COMMA,
    STATE(1253), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [45202] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2532), 1,
      anon_sym_EQ_GT,
    ACTIONS(2534), 1,
      anon_sym_RPAREN2,
    STATE(1254), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [45287] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1255), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2536), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [45370] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1256), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2538), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [45453] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1257), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2540), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [45536] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1258), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2542), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [45619] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1259), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2544), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [45702] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1260), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2546), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [45785] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1261), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2548), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [45868] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1262), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2550), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [45951] = 27,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2502), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2504), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2506), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2552), 1,
      anon_sym_RBRACE,
    STATE(719), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1163), 1,
      sym_final_modifier,
    STATE(1263), 1,
      sym_text_interpolation,
    STATE(1269), 1,
      aux_sym_declaration_list_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1376), 1,
      sym_attribute_list,
    STATE(1433), 1,
      sym__member_declaration,
    STATE(1434), 1,
      sym__class_const_declaration,
    STATE(1435), 1,
      sym__const_declaration,
    STATE(1822), 1,
      sym__function_definition_header,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1436), 3,
      sym_property_declaration,
      sym_method_declaration,
      sym_use_declaration,
    STATE(1339), 4,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [46040] = 27,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2502), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2504), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2506), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2554), 1,
      anon_sym_RBRACE,
    STATE(719), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1163), 1,
      sym_final_modifier,
    STATE(1247), 1,
      aux_sym_declaration_list_repeat1,
    STATE(1264), 1,
      sym_text_interpolation,
    STATE(1338), 1,
      sym__modifier,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1376), 1,
      sym_attribute_list,
    STATE(1433), 1,
      sym__member_declaration,
    STATE(1434), 1,
      sym__class_const_declaration,
    STATE(1435), 1,
      sym__const_declaration,
    STATE(1822), 1,
      sym__function_definition_header,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1436), 3,
      sym_property_declaration,
      sym_method_declaration,
      sym_use_declaration,
    STATE(1339), 4,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [46129] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1265), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [46212] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    ACTIONS(2492), 1,
      anon_sym_RPAREN2,
    ACTIONS(2530), 1,
      anon_sym_COMMA,
    STATE(1266), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [46297] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1267), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2558), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [46380] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    STATE(1268), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2560), 2,
      anon_sym_SEMI,
      anon_sym_COLON,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [46463] = 26,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2562), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2565), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2568), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2571), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2574), 1,
      anon_sym_RBRACE,
    ACTIONS(2576), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2579), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2582), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2585), 1,
      sym_var_modifier,
    ACTIONS(2591), 1,
      anon_sym_POUND_LBRACK,
    STATE(719), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1163), 1,
      sym_final_modifier,
    STATE(1338), 1,
      sym__modifier,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1376), 1,
      sym_attribute_list,
    STATE(1433), 1,
      sym__member_declaration,
    STATE(1434), 1,
      sym__class_const_declaration,
    STATE(1435), 1,
      sym__const_declaration,
    STATE(1822), 1,
      sym__function_definition_header,
    STATE(1269), 2,
      sym_text_interpolation,
      aux_sym_declaration_list_repeat1,
    ACTIONS(2588), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1436), 3,
      sym_property_declaration,
      sym_method_declaration,
      sym_use_declaration,
    STATE(1339), 4,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [46550] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2412), 1,
      anon_sym_AMP,
    ACTIONS(2414), 1,
      anon_sym_QMARK,
    ACTIONS(2416), 1,
      anon_sym_PIPE,
    ACTIONS(2420), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2422), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2424), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2426), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2428), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2430), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2432), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2434), 1,
      anon_sym_CARET,
    ACTIONS(2442), 1,
      anon_sym_GT_EQ,
    ACTIONS(2446), 1,
      anon_sym_DOT,
    ACTIONS(2450), 1,
      anon_sym_PERCENT,
    STATE(1270), 1,
      sym_text_interpolation,
    ACTIONS(2418), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2444), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2448), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2594), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2440), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2438), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [46633] = 27,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2502), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2504), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2506), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2596), 1,
      anon_sym_RBRACE,
    STATE(719), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1163), 1,
      sym_final_modifier,
    STATE(1263), 1,
      aux_sym_declaration_list_repeat1,
    STATE(1271), 1,
      sym_text_interpolation,
    STATE(1338), 1,
      sym__modifier,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1376), 1,
      sym_attribute_list,
    STATE(1433), 1,
      sym__member_declaration,
    STATE(1434), 1,
      sym__class_const_declaration,
    STATE(1435), 1,
      sym__const_declaration,
    STATE(1822), 1,
      sym__function_definition_header,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1436), 3,
      sym_property_declaration,
      sym_method_declaration,
      sym_use_declaration,
    STATE(1339), 4,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [46722] = 25,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2476), 1,
      anon_sym_SEMI,
    ACTIONS(2528), 1,
      anon_sym_COMMA,
    STATE(1272), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [46807] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    STATE(1273), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2598), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [46890] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2600), 1,
      anon_sym_RBRACK,
    STATE(1274), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [46972] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2602), 1,
      aux_sym_namespace_use_clause_token1,
    STATE(1275), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [47054] = 17,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(731), 1,
      anon_sym_AMP,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1828), 1,
      anon_sym_LPAREN2,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(1276), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1683), 1,
      sym__types,
    STATE(1877), 1,
      sym_type,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1715), 2,
      sym_variable_name,
      sym_by_ref,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [47122] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2604), 1,
      anon_sym_EQ_GT,
    STATE(1277), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [47204] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2606), 1,
      anon_sym_RPAREN2,
    STATE(1278), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [47286] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2608), 1,
      aux_sym_namespace_use_clause_token1,
    STATE(1279), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [47368] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2614), 1,
      anon_sym_LPAREN,
    STATE(1280), 1,
      sym_text_interpolation,
    ACTIONS(2612), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_DOLLAR,
    ACTIONS(2610), 26,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_LPAREN2,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [47414] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2616), 1,
      anon_sym_RPAREN2,
    STATE(1281), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [47496] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2618), 1,
      anon_sym_RPAREN2,
    STATE(1282), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [47578] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2620), 1,
      anon_sym_EQ_GT,
    STATE(1283), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [47660] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2622), 1,
      anon_sym_RBRACE,
    STATE(1284), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [47742] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2624), 1,
      anon_sym_RBRACE,
    STATE(1285), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [47824] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2626), 1,
      anon_sym_COLON,
    STATE(1286), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [47906] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2628), 1,
      anon_sym_RBRACK,
    STATE(1287), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [47988] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2630), 1,
      anon_sym_RBRACK,
    STATE(1288), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [48070] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2632), 1,
      anon_sym_RBRACE,
    STATE(1289), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [48152] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2634), 1,
      anon_sym_RBRACE,
    STATE(1290), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [48234] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2636), 1,
      anon_sym_EQ_GT,
    STATE(1291), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [48316] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2638), 1,
      anon_sym_RBRACE,
    STATE(1292), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [48398] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2640), 1,
      anon_sym_RPAREN2,
    STATE(1293), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [48480] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2642), 1,
      anon_sym_RBRACE,
    STATE(1294), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [48562] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2644), 1,
      anon_sym_RBRACE,
    STATE(1295), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [48644] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2646), 1,
      anon_sym_RBRACK,
    STATE(1296), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [48726] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2648), 1,
      anon_sym_RBRACK,
    STATE(1297), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [48808] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2650), 1,
      anon_sym_RBRACE,
    STATE(1298), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [48890] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2652), 1,
      anon_sym_RBRACE,
    STATE(1299), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [48972] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2654), 1,
      anon_sym_RBRACE,
    STATE(1300), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [49054] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2656), 1,
      anon_sym_RBRACK,
    STATE(1301), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [49136] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2658), 1,
      anon_sym_RBRACK,
    STATE(1302), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [49218] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2660), 1,
      anon_sym_RBRACE,
    STATE(1303), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [49300] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2662), 1,
      anon_sym_RPAREN2,
    STATE(1304), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [49382] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2664), 1,
      anon_sym_RBRACE,
    STATE(1305), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [49464] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2666), 1,
      anon_sym_RBRACK,
    STATE(1306), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [49546] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2668), 1,
      anon_sym_RBRACK,
    STATE(1307), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [49628] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2670), 1,
      anon_sym_RBRACE,
    STATE(1308), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [49710] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2672), 1,
      anon_sym_RBRACE,
    STATE(1309), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [49792] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2674), 1,
      anon_sym_RBRACK,
    STATE(1310), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [49874] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2676), 1,
      anon_sym_RBRACK,
    STATE(1311), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [49956] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2678), 1,
      anon_sym_RBRACE,
    STATE(1312), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [50038] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2680), 1,
      anon_sym_RBRACE,
    STATE(1313), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [50120] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2682), 1,
      anon_sym_RBRACE,
    STATE(1314), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [50202] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2684), 1,
      anon_sym_RBRACE,
    STATE(1315), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [50284] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2686), 1,
      anon_sym_COLON,
    STATE(1316), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [50366] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2688), 1,
      anon_sym_RBRACE,
    STATE(1317), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [50448] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2298), 1,
      anon_sym_AMP,
    ACTIONS(2300), 1,
      anon_sym_QMARK,
    ACTIONS(2302), 1,
      anon_sym_PIPE,
    ACTIONS(2306), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2308), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2310), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2312), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2314), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2316), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2318), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2320), 1,
      anon_sym_CARET,
    ACTIONS(2328), 1,
      anon_sym_GT_EQ,
    ACTIONS(2332), 1,
      anon_sym_DOT,
    ACTIONS(2336), 1,
      anon_sym_PERCENT,
    ACTIONS(2690), 1,
      anon_sym_RBRACK,
    STATE(1318), 1,
      sym_text_interpolation,
    ACTIONS(2304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2322), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2330), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2334), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2326), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2324), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [50530] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2692), 1,
      anon_sym_RBRACE,
    STATE(1319), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [50612] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2694), 1,
      anon_sym_RPAREN2,
    STATE(1320), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [50694] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2696), 1,
      anon_sym_COLON,
    STATE(1321), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [50776] = 17,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(731), 1,
      anon_sym_AMP,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1828), 1,
      anon_sym_LPAREN2,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(1322), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1683), 1,
      sym__types,
    STATE(1890), 1,
      sym_type,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1832), 2,
      sym_variable_name,
      sym_by_ref,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [50844] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2534), 1,
      anon_sym_RPAREN2,
    STATE(1323), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [50926] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2698), 1,
      anon_sym_COLON,
    STATE(1324), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [51008] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2700), 1,
      aux_sym_namespace_use_clause_token1,
    STATE(1325), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [51090] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2702), 1,
      aux_sym_namespace_use_clause_token1,
    STATE(1326), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [51172] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2224), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2228), 1,
      anon_sym_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_AMP,
    ACTIONS(2234), 1,
      anon_sym_QMARK,
    ACTIONS(2236), 1,
      anon_sym_PIPE,
    ACTIONS(2240), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2242), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2244), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2246), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2252), 1,
      anon_sym_CARET,
    ACTIONS(2260), 1,
      anon_sym_GT_EQ,
    ACTIONS(2264), 1,
      anon_sym_DOT,
    ACTIONS(2704), 1,
      anon_sym_RPAREN2,
    STATE(1327), 1,
      sym_text_interpolation,
    ACTIONS(2226), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2238), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2254), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2262), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2258), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2256), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [51254] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1328), 1,
      sym_text_interpolation,
    ACTIONS(2708), 4,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOLLAR,
    ACTIONS(2706), 25,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [51297] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2502), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2504), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2710), 1,
      anon_sym_RBRACE,
    ACTIONS(2712), 1,
      aux_sym_enum_case_token1,
    STATE(1329), 1,
      sym_text_interpolation,
    STATE(1333), 1,
      aux_sym_enum_declaration_list_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1381), 1,
      sym_attribute_list,
    STATE(1387), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1438), 1,
      sym__enum_member_declaration,
    STATE(1822), 1,
      sym__function_definition_header,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1439), 3,
      sym_enum_case,
      sym_method_declaration,
      sym_use_declaration,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [51378] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2502), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2504), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2712), 1,
      aux_sym_enum_case_token1,
    ACTIONS(2714), 1,
      anon_sym_RBRACE,
    STATE(1330), 1,
      sym_text_interpolation,
    STATE(1334), 1,
      aux_sym_enum_declaration_list_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1381), 1,
      sym_attribute_list,
    STATE(1387), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1438), 1,
      sym__enum_member_declaration,
    STATE(1822), 1,
      sym__function_definition_header,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1439), 3,
      sym_enum_case,
      sym_method_declaration,
      sym_use_declaration,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [51459] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1331), 1,
      sym_text_interpolation,
    ACTIONS(2718), 4,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOLLAR,
    ACTIONS(2716), 25,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [51502] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1332), 1,
      sym_text_interpolation,
    ACTIONS(1814), 3,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOLLAR,
    ACTIONS(2720), 26,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [51545] = 23,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2722), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2725), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2728), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2731), 1,
      anon_sym_RBRACE,
    ACTIONS(2733), 1,
      aux_sym_enum_case_token1,
    ACTIONS(2736), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2739), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2742), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2745), 1,
      sym_var_modifier,
    ACTIONS(2751), 1,
      anon_sym_POUND_LBRACK,
    STATE(1338), 1,
      sym__modifier,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1381), 1,
      sym_attribute_list,
    STATE(1387), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1438), 1,
      sym__enum_member_declaration,
    STATE(1822), 1,
      sym__function_definition_header,
    STATE(1333), 2,
      sym_text_interpolation,
      aux_sym_enum_declaration_list_repeat1,
    ACTIONS(2748), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1439), 3,
      sym_enum_case,
      sym_method_declaration,
      sym_use_declaration,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [51624] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2502), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2504), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2712), 1,
      aux_sym_enum_case_token1,
    ACTIONS(2754), 1,
      anon_sym_RBRACE,
    STATE(1333), 1,
      aux_sym_enum_declaration_list_repeat1,
    STATE(1334), 1,
      sym_text_interpolation,
    STATE(1338), 1,
      sym__modifier,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1381), 1,
      sym_attribute_list,
    STATE(1387), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1438), 1,
      sym__enum_member_declaration,
    STATE(1822), 1,
      sym__function_definition_header,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1439), 3,
      sym_enum_case,
      sym_method_declaration,
      sym_use_declaration,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [51705] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1335), 1,
      sym_text_interpolation,
    ACTIONS(2758), 4,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOLLAR,
    ACTIONS(2756), 25,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [51748] = 24,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2502), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2504), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2712), 1,
      aux_sym_enum_case_token1,
    ACTIONS(2760), 1,
      anon_sym_RBRACE,
    STATE(1329), 1,
      aux_sym_enum_declaration_list_repeat1,
    STATE(1336), 1,
      sym_text_interpolation,
    STATE(1338), 1,
      sym__modifier,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1381), 1,
      sym_attribute_list,
    STATE(1387), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1438), 1,
      sym__enum_member_declaration,
    STATE(1822), 1,
      sym__function_definition_header,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1439), 3,
      sym_enum_case,
      sym_method_declaration,
      sym_use_declaration,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [51829] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1337), 1,
      sym_text_interpolation,
    ACTIONS(2764), 3,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOLLAR,
    ACTIONS(2762), 25,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [51871] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1338), 1,
      sym_text_interpolation,
    ACTIONS(2768), 3,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOLLAR,
    ACTIONS(2766), 25,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [51913] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1339), 1,
      sym_text_interpolation,
    ACTIONS(2406), 3,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOLLAR,
    ACTIONS(2404), 25,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [51955] = 15,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(226), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2770), 1,
      sym_name,
    ACTIONS(2774), 1,
      anon_sym_LPAREN2,
    ACTIONS(2776), 1,
      anon_sym_QMARK,
    ACTIONS(2778), 1,
      sym_bottom_type,
    STATE(1340), 1,
      sym_text_interpolation,
    STATE(1547), 1,
      sym__types,
    STATE(1746), 1,
      sym_qualified_name,
    STATE(1986), 1,
      sym_type,
    STATE(2622), 1,
      sym_namespace_name,
    STATE(1764), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(2256), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    ACTIONS(2772), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [52016] = 15,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1828), 1,
      anon_sym_LPAREN2,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    ACTIONS(2780), 1,
      sym_bottom_type,
    STATE(1341), 1,
      sym_text_interpolation,
    STATE(1494), 1,
      sym__types,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(2443), 1,
      sym_type,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1538), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [52077] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1865), 1,
      anon_sym_POUND_LBRACK,
    STATE(1342), 1,
      sym_text_interpolation,
    STATE(1343), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1346), 1,
      sym_attribute_group,
    ACTIONS(2784), 5,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
    ACTIONS(2782), 17,
      anon_sym_string,
      anon_sym_int,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [52122] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2790), 1,
      anon_sym_POUND_LBRACK,
    STATE(1346), 1,
      sym_attribute_group,
    STATE(1343), 2,
      sym_text_interpolation,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2788), 5,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
    ACTIONS(2786), 17,
      anon_sym_string,
      anon_sym_int,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [52165] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1344), 1,
      sym_text_interpolation,
    ACTIONS(2795), 6,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_POUND_LBRACK,
      anon_sym_DOLLAR,
    ACTIONS(2793), 17,
      anon_sym_string,
      anon_sym_int,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [52202] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1345), 1,
      sym_text_interpolation,
    ACTIONS(2799), 6,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_POUND_LBRACK,
      anon_sym_DOLLAR,
    ACTIONS(2797), 17,
      anon_sym_string,
      anon_sym_int,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [52239] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1346), 1,
      sym_text_interpolation,
    ACTIONS(2803), 6,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_POUND_LBRACK,
      anon_sym_DOLLAR,
    ACTIONS(2801), 17,
      anon_sym_string,
      anon_sym_int,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [52276] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1347), 1,
      sym_text_interpolation,
    ACTIONS(2807), 6,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_POUND_LBRACK,
      anon_sym_DOLLAR,
    ACTIONS(2805), 17,
      anon_sym_string,
      anon_sym_int,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_QMARK,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [52313] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    ACTIONS(2809), 1,
      anon_sym_LPAREN2,
    STATE(1348), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1512), 1,
      sym__types,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [52363] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    STATE(1349), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(2060), 1,
      sym__types,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2641), 1,
      sym_intersection_type,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [52413] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    ACTIONS(2809), 1,
      anon_sym_LPAREN2,
    STATE(1350), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1523), 1,
      sym__types,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [52463] = 20,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1797), 1,
      anon_sym_DOLLAR,
    ACTIONS(2080), 1,
      sym_name,
    ACTIONS(2084), 1,
      anon_sym_BSLASH,
    ACTIONS(2088), 1,
      anon_sym_LPAREN2,
    STATE(794), 1,
      sym__new_variable,
    STATE(809), 1,
      sym__variable_member_access_expression,
    STATE(810), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(811), 1,
      sym__variable_scoped_property_access_expression,
    STATE(907), 1,
      sym__reserved_identifier,
    STATE(916), 1,
      sym__name,
    STATE(982), 1,
      sym_parenthesized_expression,
    STATE(1016), 1,
      sym__class_name_reference,
    STATE(1351), 1,
      sym_text_interpolation,
    STATE(2651), 1,
      sym_namespace_name,
    STATE(813), 2,
      sym__variable_subscript_expression,
      sym__simple_variable,
    STATE(840), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(908), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [52529] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(226), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2770), 1,
      sym_name,
    ACTIONS(2776), 1,
      anon_sym_QMARK,
    ACTIONS(2811), 1,
      anon_sym_LPAREN2,
    STATE(1352), 1,
      sym_text_interpolation,
    STATE(1746), 1,
      sym_qualified_name,
    STATE(1922), 1,
      sym__types,
    STATE(2622), 1,
      sym_namespace_name,
    STATE(1764), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(2772), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [52579] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(226), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2770), 1,
      sym_name,
    ACTIONS(2776), 1,
      anon_sym_QMARK,
    ACTIONS(2811), 1,
      anon_sym_LPAREN2,
    STATE(1353), 1,
      sym_text_interpolation,
    STATE(1746), 1,
      sym_qualified_name,
    STATE(1908), 1,
      sym__types,
    STATE(2622), 1,
      sym_namespace_name,
    STATE(1764), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(2772), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [52629] = 20,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1801), 1,
      anon_sym_DOLLAR,
    ACTIONS(2094), 1,
      sym_name,
    ACTIONS(2098), 1,
      anon_sym_BSLASH,
    ACTIONS(2102), 1,
      anon_sym_LPAREN2,
    STATE(977), 1,
      sym__new_variable,
    STATE(997), 1,
      sym__variable_member_access_expression,
    STATE(1000), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1001), 1,
      sym__variable_scoped_property_access_expression,
    STATE(1043), 1,
      sym__reserved_identifier,
    STATE(1048), 1,
      sym__name,
    STATE(1051), 1,
      sym_parenthesized_expression,
    STATE(1133), 1,
      sym__class_name_reference,
    STATE(1354), 1,
      sym_text_interpolation,
    STATE(2565), 1,
      sym_namespace_name,
    STATE(952), 2,
      sym__variable_subscript_expression,
      sym__simple_variable,
    STATE(957), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(1044), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2104), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [52695] = 20,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2813), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2815), 1,
      aux_sym_trait_declaration_token1,
    ACTIONS(2817), 1,
      aux_sym_interface_declaration_token1,
    ACTIONS(2819), 1,
      aux_sym_enum_declaration_token1,
    ACTIONS(2821), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2823), 1,
      aux_sym__arrow_function_header_token1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1355), 1,
      sym_text_interpolation,
    STATE(1394), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1503), 1,
      sym_static_modifier,
    STATE(2455), 1,
      sym__function_definition_header,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 4,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_visibility_modifier,
  [52761] = 20,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2813), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2823), 1,
      aux_sym__arrow_function_header_token1,
    ACTIONS(2825), 1,
      aux_sym_trait_declaration_token1,
    ACTIONS(2827), 1,
      aux_sym_interface_declaration_token1,
    ACTIONS(2829), 1,
      aux_sym_enum_declaration_token1,
    ACTIONS(2831), 1,
      aux_sym_class_declaration_token1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1356), 1,
      sym_text_interpolation,
    STATE(1398), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1503), 1,
      sym_static_modifier,
    STATE(2482), 1,
      sym__function_definition_header,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 4,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_visibility_modifier,
  [52827] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    STATE(1357), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(2060), 1,
      sym__types,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2695), 1,
      sym_intersection_type,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [52877] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1855), 1,
      anon_sym_BSLASH,
    ACTIONS(1863), 1,
      anon_sym_QMARK,
    ACTIONS(2809), 1,
      anon_sym_LPAREN2,
    STATE(1358), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1523), 1,
      sym__types,
    STATE(2644), 1,
      sym_namespace_name,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [52927] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    STATE(1359), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(2060), 1,
      sym__types,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2652), 1,
      sym_intersection_type,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [52977] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1855), 1,
      anon_sym_BSLASH,
    ACTIONS(1863), 1,
      anon_sym_QMARK,
    ACTIONS(2809), 1,
      anon_sym_LPAREN2,
    STATE(1360), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1512), 1,
      sym__types,
    STATE(2644), 1,
      sym_namespace_name,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [53027] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    STATE(1361), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(2060), 1,
      sym__types,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2621), 1,
      sym_intersection_type,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [53077] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    STATE(1362), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(2060), 1,
      sym__types,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(2720), 1,
      sym_intersection_type,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [53127] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    STATE(1363), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1521), 1,
      sym__types,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [53174] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2784), 1,
      anon_sym_AMP,
    STATE(1364), 1,
      sym_text_interpolation,
    STATE(1368), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    ACTIONS(2782), 17,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_trait_declaration_token1,
      aux_sym_interface_declaration_token1,
      aux_sym_enum_declaration_token1,
      aux_sym_enum_case_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
      sym_name,
  [53215] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(226), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2770), 1,
      sym_name,
    ACTIONS(2776), 1,
      anon_sym_QMARK,
    STATE(1365), 1,
      sym_text_interpolation,
    STATE(1746), 1,
      sym_qualified_name,
    STATE(1882), 1,
      sym__types,
    STATE(2622), 1,
      sym_namespace_name,
    STATE(1764), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(2772), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [53262] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1855), 1,
      anon_sym_BSLASH,
    ACTIONS(1863), 1,
      anon_sym_QMARK,
    STATE(1366), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1521), 1,
      sym__types,
    STATE(2644), 1,
      sym_namespace_name,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [53309] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(226), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2770), 1,
      sym_name,
    ACTIONS(2776), 1,
      anon_sym_QMARK,
    STATE(1367), 1,
      sym_text_interpolation,
    STATE(1746), 1,
      sym_qualified_name,
    STATE(1921), 1,
      sym__types,
    STATE(2622), 1,
      sym_namespace_name,
    STATE(1764), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(2772), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [53356] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2788), 1,
      anon_sym_AMP,
    ACTIONS(2833), 1,
      anon_sym_POUND_LBRACK,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1368), 2,
      sym_text_interpolation,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2786), 17,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_trait_declaration_token1,
      aux_sym_interface_declaration_token1,
      aux_sym_enum_declaration_token1,
      aux_sym_enum_case_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
      sym_name,
  [53395] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1855), 1,
      anon_sym_BSLASH,
    ACTIONS(1863), 1,
      anon_sym_QMARK,
    STATE(1369), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1508), 1,
      sym__types,
    STATE(2644), 1,
      sym_namespace_name,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [53442] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1830), 1,
      anon_sym_QMARK,
    STATE(1370), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1508), 1,
      sym__types,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1495), 3,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [53489] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1832), 1,
      anon_sym_DOLLAR,
    ACTIONS(2836), 1,
      anon_sym_COLON_COLON,
    STATE(1371), 1,
      sym_text_interpolation,
    STATE(1839), 1,
      sym_static_variable_declaration,
    STATE(1924), 1,
      sym_variable_name,
    ACTIONS(1614), 4,
      anon_sym_LPAREN2,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1814), 12,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
  [53531] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1832), 1,
      anon_sym_DOLLAR,
    ACTIONS(2836), 1,
      anon_sym_COLON_COLON,
    STATE(1372), 1,
      sym_text_interpolation,
    STATE(1775), 1,
      sym_static_variable_declaration,
    STATE(1924), 1,
      sym_variable_name,
    ACTIONS(1614), 4,
      anon_sym_LPAREN2,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1814), 12,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
  [53573] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1373), 1,
      sym_text_interpolation,
    ACTIONS(2799), 2,
      anon_sym_AMP,
      anon_sym_POUND_LBRACK,
    ACTIONS(2797), 17,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_trait_declaration_token1,
      aux_sym_interface_declaration_token1,
      aux_sym_enum_declaration_token1,
      aux_sym_enum_case_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
      sym_name,
  [53606] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1374), 1,
      sym_text_interpolation,
    ACTIONS(2803), 2,
      anon_sym_AMP,
      anon_sym_POUND_LBRACK,
    ACTIONS(2801), 17,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_trait_declaration_token1,
      aux_sym_interface_declaration_token1,
      aux_sym_enum_declaration_token1,
      aux_sym_enum_case_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
      sym_name,
  [53639] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1375), 1,
      sym_text_interpolation,
    ACTIONS(2795), 2,
      anon_sym_AMP,
      anon_sym_POUND_LBRACK,
    ACTIONS(2793), 17,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_trait_declaration_token1,
      aux_sym_interface_declaration_token1,
      aux_sym_enum_declaration_token1,
      aux_sym_enum_case_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
      sym_name,
  [53672] = 17,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2504), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2506), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    STATE(720), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1111), 1,
      sym_final_modifier,
    STATE(1338), 1,
      sym__modifier,
    STATE(1376), 1,
      sym_text_interpolation,
    STATE(1456), 1,
      sym__const_declaration,
    STATE(1743), 1,
      sym__function_definition_header,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 4,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [53729] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1377), 1,
      sym_text_interpolation,
    ACTIONS(1158), 9,
      aux_sym_function_static_declaration_token1,
      aux_sym_cast_type_token1,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
      aux_sym_else_clause_token1,
      aux_sym__new_non_dereferencable_expression_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(1156), 10,
      anon_sym_BSLASH,
      anon_sym_RBRACE,
      anon_sym_LPAREN2,
      anon_sym_LBRACK,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR,
  [53762] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1378), 1,
      sym_text_interpolation,
    ACTIONS(2807), 2,
      anon_sym_AMP,
      anon_sym_POUND_LBRACK,
    ACTIONS(2805), 17,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_trait_declaration_token1,
      aux_sym_interface_declaration_token1,
      aux_sym_enum_declaration_token1,
      aux_sym_enum_case_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
      sym_name,
  [53795] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    ACTIONS(1855), 1,
      anon_sym_BSLASH,
    STATE(1379), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(2644), 1,
      sym_namespace_name,
    STATE(1498), 2,
      sym_named_type,
      sym_primitive_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [53835] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(226), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2770), 1,
      sym_name,
    STATE(1380), 1,
      sym_text_interpolation,
    STATE(1746), 1,
      sym_qualified_name,
    STATE(2622), 1,
      sym_namespace_name,
    STATE(1762), 2,
      sym_named_type,
      sym_primitive_type,
    ACTIONS(2772), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [53875] = 15,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2504), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2838), 1,
      aux_sym_enum_case_token1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1381), 1,
      sym_text_interpolation,
    STATE(1383), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1743), 1,
      sym__function_definition_header,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [53927] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1818), 1,
      sym_name,
    STATE(1382), 1,
      sym_text_interpolation,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1498), 2,
      sym_named_type,
      sym_primitive_type,
    ACTIONS(1826), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [53967] = 14,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2504), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    STATE(1088), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1383), 1,
      sym_text_interpolation,
    STATE(1730), 1,
      sym__function_definition_header,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [54016] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(611), 1,
      sym_integer,
    ACTIONS(641), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2840), 1,
      sym_float,
    ACTIONS(2842), 1,
      aux_sym__argument_name_token1,
    ACTIONS(2844), 1,
      aux_sym__argument_name_token2,
    STATE(1384), 1,
      sym_text_interpolation,
    STATE(2673), 1,
      sym_literal,
    ACTIONS(637), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(639), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(1018), 3,
      sym_boolean,
      sym_null,
      sym__string,
    STATE(775), 4,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
  [54063] = 14,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2846), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2848), 1,
      aux_sym_class_declaration_token1,
    STATE(1088), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1385), 1,
      sym_text_interpolation,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [54112] = 14,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2850), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2852), 1,
      aux_sym_class_declaration_token1,
    STATE(1088), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1386), 1,
      sym_text_interpolation,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [54161] = 14,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2504), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    STATE(1088), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1387), 1,
      sym_text_interpolation,
    STATE(1757), 1,
      sym__function_definition_header,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [54210] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2854), 1,
      anon_sym_LBRACE,
    ACTIONS(2860), 1,
      anon_sym_BSLASHu,
    ACTIONS(2863), 1,
      sym__new_line,
    ACTIONS(2866), 1,
      anon_sym_DOLLAR,
    ACTIONS(2869), 1,
      sym_heredoc_end,
    STATE(1402), 1,
      aux_sym__interpolated_string_body_heredoc,
    STATE(1524), 1,
      sym_variable_name,
    STATE(1575), 1,
      sym__simple_string_member_access_expression,
    STATE(1598), 1,
      sym_dynamic_variable_name,
    ACTIONS(2857), 2,
      sym_encapsed_string_chars_heredoc,
      sym_escape_sequence,
    STATE(1388), 2,
      sym_text_interpolation,
      aux_sym_heredoc_body_repeat1,
    STATE(1581), 2,
      sym__complex_string_part,
      sym__simple_string_part,
    STATE(1599), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
  [54260] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2850), 1,
      aux_sym__namespace_use_type_token2,
    STATE(1088), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1389), 1,
      sym_text_interpolation,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [54306] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2871), 1,
      aux_sym_class_declaration_token1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1390), 1,
      sym_text_interpolation,
    STATE(1397), 1,
      aux_sym_class_declaration_repeat1,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [54352] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2873), 1,
      aux_sym_class_declaration_token1,
    STATE(1088), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1391), 1,
      sym_text_interpolation,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [54398] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2875), 1,
      aux_sym_class_declaration_token1,
    STATE(1088), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1392), 1,
      sym_text_interpolation,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [54444] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2877), 1,
      aux_sym_class_declaration_token1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1393), 1,
      sym_text_interpolation,
    STATE(1395), 1,
      aux_sym_class_declaration_repeat1,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [54490] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2879), 1,
      aux_sym_class_declaration_token1,
    STATE(1088), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1394), 1,
      sym_text_interpolation,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [54536] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2881), 1,
      aux_sym_class_declaration_token1,
    STATE(1088), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1395), 1,
      sym_text_interpolation,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [54582] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2883), 1,
      anon_sym_LBRACE,
    ACTIONS(2887), 1,
      anon_sym_BSLASHu,
    ACTIONS(2889), 1,
      sym__new_line,
    ACTIONS(2892), 1,
      anon_sym_DOLLAR,
    ACTIONS(2894), 1,
      sym_heredoc_end,
    STATE(1388), 1,
      aux_sym_heredoc_body_repeat1,
    STATE(1396), 1,
      sym_text_interpolation,
    STATE(1402), 1,
      aux_sym__interpolated_string_body_heredoc,
    STATE(1524), 1,
      sym_variable_name,
    STATE(1575), 1,
      sym__simple_string_member_access_expression,
    STATE(1598), 1,
      sym_dynamic_variable_name,
    ACTIONS(2885), 2,
      sym_encapsed_string_chars_heredoc,
      sym_escape_sequence,
    STATE(1581), 2,
      sym__complex_string_part,
      sym__simple_string_part,
    STATE(1599), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
  [54634] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2896), 1,
      aux_sym_class_declaration_token1,
    STATE(1088), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1397), 1,
      sym_text_interpolation,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [54680] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2510), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2512), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2514), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2516), 1,
      sym_var_modifier,
    ACTIONS(2898), 1,
      aux_sym_class_declaration_token1,
    STATE(1088), 1,
      aux_sym_class_declaration_repeat1,
    STATE(1338), 1,
      sym__modifier,
    STATE(1398), 1,
      sym_text_interpolation,
    ACTIONS(2518), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1339), 5,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym_static_modifier,
      sym_visibility_modifier,
  [54726] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1399), 1,
      sym_text_interpolation,
    ACTIONS(2900), 6,
      aux_sym_function_static_declaration_token1,
      aux_sym_cast_type_token1,
      aux_sym__new_non_dereferencable_expression_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(1739), 10,
      anon_sym_BSLASH,
      anon_sym_RBRACE,
      anon_sym_LPAREN2,
      anon_sym_LBRACK,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR,
  [54756] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2883), 1,
      anon_sym_LBRACE,
    ACTIONS(2887), 1,
      anon_sym_BSLASHu,
    ACTIONS(2892), 1,
      anon_sym_DOLLAR,
    ACTIONS(2902), 1,
      sym__new_line,
    STATE(1396), 1,
      aux_sym_heredoc_body_repeat1,
    STATE(1400), 1,
      sym_text_interpolation,
    STATE(1402), 1,
      aux_sym__interpolated_string_body_heredoc,
    STATE(1524), 1,
      sym_variable_name,
    STATE(1575), 1,
      sym__simple_string_member_access_expression,
    STATE(1598), 1,
      sym_dynamic_variable_name,
    ACTIONS(2885), 2,
      sym_encapsed_string_chars_heredoc,
      sym_escape_sequence,
    STATE(1581), 2,
      sym__complex_string_part,
      sym__simple_string_part,
    STATE(1599), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
  [54805] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2883), 1,
      anon_sym_LBRACE,
    ACTIONS(2887), 1,
      anon_sym_BSLASHu,
    ACTIONS(2892), 1,
      anon_sym_DOLLAR,
    STATE(1401), 1,
      sym_text_interpolation,
    STATE(1403), 1,
      aux_sym__interpolated_string_body_heredoc,
    STATE(1524), 1,
      sym_variable_name,
    STATE(1575), 1,
      sym__simple_string_member_access_expression,
    STATE(1598), 1,
      sym_dynamic_variable_name,
    ACTIONS(2869), 2,
      sym_heredoc_end,
      sym__new_line,
    ACTIONS(2885), 2,
      sym_encapsed_string_chars_heredoc,
      sym_escape_sequence,
    STATE(1581), 2,
      sym__complex_string_part,
      sym__simple_string_part,
    STATE(1599), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
  [54852] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2883), 1,
      anon_sym_LBRACE,
    ACTIONS(2887), 1,
      anon_sym_BSLASHu,
    ACTIONS(2892), 1,
      anon_sym_DOLLAR,
    STATE(1402), 1,
      sym_text_interpolation,
    STATE(1403), 1,
      aux_sym__interpolated_string_body_heredoc,
    STATE(1524), 1,
      sym_variable_name,
    STATE(1575), 1,
      sym__simple_string_member_access_expression,
    STATE(1598), 1,
      sym_dynamic_variable_name,
    ACTIONS(2885), 2,
      sym_encapsed_string_chars_heredoc,
      sym_escape_sequence,
    ACTIONS(2904), 2,
      sym_heredoc_end,
      sym__new_line,
    STATE(1581), 2,
      sym__complex_string_part,
      sym__simple_string_part,
    STATE(1599), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
  [54899] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2906), 1,
      anon_sym_LBRACE,
    ACTIONS(2912), 1,
      anon_sym_BSLASHu,
    ACTIONS(2917), 1,
      anon_sym_DOLLAR,
    STATE(1524), 1,
      sym_variable_name,
    STATE(1575), 1,
      sym__simple_string_member_access_expression,
    STATE(1598), 1,
      sym_dynamic_variable_name,
    ACTIONS(2909), 2,
      sym_encapsed_string_chars_heredoc,
      sym_escape_sequence,
    ACTIONS(2915), 2,
      sym_heredoc_end,
      sym__new_line,
    STATE(1403), 2,
      sym_text_interpolation,
      aux_sym__interpolated_string_body_heredoc,
    STATE(1581), 2,
      sym__complex_string_part,
      sym__simple_string_part,
    STATE(1599), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
  [54944] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1404), 1,
      sym_text_interpolation,
    ACTIONS(2920), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [54970] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1405), 1,
      sym_text_interpolation,
    ACTIONS(2922), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [54996] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1406), 1,
      sym_text_interpolation,
    ACTIONS(2924), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [55022] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1407), 1,
      sym_text_interpolation,
    ACTIONS(2926), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [55048] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1408), 1,
      sym_text_interpolation,
    ACTIONS(2928), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [55074] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1409), 1,
      sym_text_interpolation,
    ACTIONS(2930), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [55100] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1410), 1,
      sym_text_interpolation,
    ACTIONS(2932), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [55126] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(226), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2094), 1,
      sym_name,
    STATE(1043), 1,
      sym__reserved_identifier,
    STATE(1411), 1,
      sym_text_interpolation,
    STATE(1714), 1,
      sym_namespace_use_clause,
    STATE(1863), 1,
      sym__name,
    STATE(2002), 1,
      sym__namespace_use_group,
    STATE(2556), 1,
      sym_namespace_name,
    ACTIONS(2934), 2,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
    STATE(1044), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2104), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [55170] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(226), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2094), 1,
      sym_name,
    STATE(1043), 1,
      sym__reserved_identifier,
    STATE(1412), 1,
      sym_text_interpolation,
    STATE(1796), 1,
      sym_namespace_use_clause,
    STATE(1863), 1,
      sym__name,
    STATE(2247), 1,
      sym__namespace_use_group,
    STATE(2556), 1,
      sym_namespace_name,
    ACTIONS(2934), 2,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
    STATE(1044), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2104), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [55214] = 14,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2936), 1,
      anon_sym_LBRACE,
    ACTIONS(2940), 1,
      anon_sym_BSLASHu,
    ACTIONS(2942), 1,
      anon_sym_DQUOTE,
    ACTIONS(2944), 1,
      anon_sym_DOLLAR,
    STATE(1413), 1,
      sym_text_interpolation,
    STATE(1418), 1,
      aux_sym__interpolated_string_body,
    STATE(1537), 1,
      sym_variable_name,
    STATE(1659), 1,
      sym_dynamic_variable_name,
    STATE(1682), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2938), 2,
      sym_encapsed_string_chars,
      sym_escape_sequence,
    STATE(1674), 2,
      sym__complex_string_part,
      sym__simple_string_part,
    STATE(1685), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
  [55260] = 14,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2946), 1,
      anon_sym_LBRACE,
    ACTIONS(2950), 1,
      anon_sym_BSLASHu,
    ACTIONS(2952), 1,
      anon_sym_BQUOTE,
    ACTIONS(2954), 1,
      anon_sym_DOLLAR,
    STATE(1414), 1,
      sym_text_interpolation,
    STATE(1415), 1,
      aux_sym__interpolated_execution_operator_body,
    STATE(1539), 1,
      sym_variable_name,
    STATE(1641), 1,
      sym__simple_string_member_access_expression,
    STATE(1675), 1,
      sym_dynamic_variable_name,
    ACTIONS(2948), 2,
      sym_execution_string_chars,
      sym_escape_sequence,
    STATE(1642), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
    STATE(1644), 2,
      sym__complex_string_part,
      sym__simple_string_part,
  [55306] = 14,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2946), 1,
      anon_sym_LBRACE,
    ACTIONS(2950), 1,
      anon_sym_BSLASHu,
    ACTIONS(2954), 1,
      anon_sym_DOLLAR,
    ACTIONS(2956), 1,
      anon_sym_BQUOTE,
    STATE(1415), 1,
      sym_text_interpolation,
    STATE(1416), 1,
      aux_sym__interpolated_execution_operator_body,
    STATE(1539), 1,
      sym_variable_name,
    STATE(1641), 1,
      sym__simple_string_member_access_expression,
    STATE(1675), 1,
      sym_dynamic_variable_name,
    ACTIONS(2948), 2,
      sym_execution_string_chars,
      sym_escape_sequence,
    STATE(1642), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
    STATE(1644), 2,
      sym__complex_string_part,
      sym__simple_string_part,
  [55352] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2958), 1,
      anon_sym_LBRACE,
    ACTIONS(2964), 1,
      anon_sym_BSLASHu,
    ACTIONS(2967), 1,
      anon_sym_BQUOTE,
    ACTIONS(2969), 1,
      anon_sym_DOLLAR,
    STATE(1539), 1,
      sym_variable_name,
    STATE(1641), 1,
      sym__simple_string_member_access_expression,
    STATE(1675), 1,
      sym_dynamic_variable_name,
    ACTIONS(2961), 2,
      sym_execution_string_chars,
      sym_escape_sequence,
    STATE(1416), 2,
      sym_text_interpolation,
      aux_sym__interpolated_execution_operator_body,
    STATE(1642), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
    STATE(1644), 2,
      sym__complex_string_part,
      sym__simple_string_part,
  [55396] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1417), 1,
      sym_text_interpolation,
    ACTIONS(2972), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [55422] = 14,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2936), 1,
      anon_sym_LBRACE,
    ACTIONS(2940), 1,
      anon_sym_BSLASHu,
    ACTIONS(2944), 1,
      anon_sym_DOLLAR,
    ACTIONS(2974), 1,
      anon_sym_DQUOTE,
    STATE(1418), 1,
      sym_text_interpolation,
    STATE(1427), 1,
      aux_sym__interpolated_string_body,
    STATE(1537), 1,
      sym_variable_name,
    STATE(1659), 1,
      sym_dynamic_variable_name,
    STATE(1682), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2938), 2,
      sym_encapsed_string_chars,
      sym_escape_sequence,
    STATE(1674), 2,
      sym__complex_string_part,
      sym__simple_string_part,
    STATE(1685), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
  [55468] = 14,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2936), 1,
      anon_sym_LBRACE,
    ACTIONS(2940), 1,
      anon_sym_BSLASHu,
    ACTIONS(2944), 1,
      anon_sym_DOLLAR,
    ACTIONS(2976), 1,
      anon_sym_DQUOTE,
    STATE(1419), 1,
      sym_text_interpolation,
    STATE(1427), 1,
      aux_sym__interpolated_string_body,
    STATE(1537), 1,
      sym_variable_name,
    STATE(1659), 1,
      sym_dynamic_variable_name,
    STATE(1682), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2938), 2,
      sym_encapsed_string_chars,
      sym_escape_sequence,
    STATE(1674), 2,
      sym__complex_string_part,
      sym__simple_string_part,
    STATE(1685), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
  [55514] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1420), 1,
      sym_text_interpolation,
    ACTIONS(2978), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [55540] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1421), 1,
      sym_text_interpolation,
    ACTIONS(2980), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [55566] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1422), 1,
      sym_text_interpolation,
    ACTIONS(2982), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [55592] = 14,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2936), 1,
      anon_sym_LBRACE,
    ACTIONS(2940), 1,
      anon_sym_BSLASHu,
    ACTIONS(2944), 1,
      anon_sym_DOLLAR,
    ACTIONS(2984), 1,
      anon_sym_DQUOTE,
    STATE(1419), 1,
      aux_sym__interpolated_string_body,
    STATE(1423), 1,
      sym_text_interpolation,
    STATE(1537), 1,
      sym_variable_name,
    STATE(1659), 1,
      sym_dynamic_variable_name,
    STATE(1682), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2938), 2,
      sym_encapsed_string_chars,
      sym_escape_sequence,
    STATE(1674), 2,
      sym__complex_string_part,
      sym__simple_string_part,
    STATE(1685), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
  [55638] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1424), 1,
      sym_text_interpolation,
    ACTIONS(2986), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [55664] = 14,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2946), 1,
      anon_sym_LBRACE,
    ACTIONS(2950), 1,
      anon_sym_BSLASHu,
    ACTIONS(2954), 1,
      anon_sym_DOLLAR,
    ACTIONS(2988), 1,
      anon_sym_BQUOTE,
    STATE(1425), 1,
      sym_text_interpolation,
    STATE(1426), 1,
      aux_sym__interpolated_execution_operator_body,
    STATE(1539), 1,
      sym_variable_name,
    STATE(1641), 1,
      sym__simple_string_member_access_expression,
    STATE(1675), 1,
      sym_dynamic_variable_name,
    ACTIONS(2948), 2,
      sym_execution_string_chars,
      sym_escape_sequence,
    STATE(1642), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
    STATE(1644), 2,
      sym__complex_string_part,
      sym__simple_string_part,
  [55710] = 14,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2946), 1,
      anon_sym_LBRACE,
    ACTIONS(2950), 1,
      anon_sym_BSLASHu,
    ACTIONS(2954), 1,
      anon_sym_DOLLAR,
    ACTIONS(2990), 1,
      anon_sym_BQUOTE,
    STATE(1416), 1,
      aux_sym__interpolated_execution_operator_body,
    STATE(1426), 1,
      sym_text_interpolation,
    STATE(1539), 1,
      sym_variable_name,
    STATE(1641), 1,
      sym__simple_string_member_access_expression,
    STATE(1675), 1,
      sym_dynamic_variable_name,
    ACTIONS(2948), 2,
      sym_execution_string_chars,
      sym_escape_sequence,
    STATE(1642), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
    STATE(1644), 2,
      sym__complex_string_part,
      sym__simple_string_part,
  [55756] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2992), 1,
      anon_sym_LBRACE,
    ACTIONS(2998), 1,
      anon_sym_BSLASHu,
    ACTIONS(3001), 1,
      anon_sym_DQUOTE,
    ACTIONS(3003), 1,
      anon_sym_DOLLAR,
    STATE(1537), 1,
      sym_variable_name,
    STATE(1659), 1,
      sym_dynamic_variable_name,
    STATE(1682), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2995), 2,
      sym_encapsed_string_chars,
      sym_escape_sequence,
    STATE(1427), 2,
      sym_text_interpolation,
      aux_sym__interpolated_string_body,
    STATE(1674), 2,
      sym__complex_string_part,
      sym__simple_string_part,
    STATE(1685), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
  [55800] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1428), 1,
      sym_text_interpolation,
    ACTIONS(3006), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [55825] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1429), 1,
      sym_text_interpolation,
    ACTIONS(3008), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [55850] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1430), 1,
      sym_text_interpolation,
    ACTIONS(3010), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [55875] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(226), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2094), 1,
      sym_name,
    STATE(1043), 1,
      sym__reserved_identifier,
    STATE(1431), 1,
      sym_text_interpolation,
    STATE(1863), 1,
      sym__name,
    STATE(1955), 1,
      sym_namespace_use_clause,
    STATE(2622), 1,
      sym_namespace_name,
    ACTIONS(3012), 2,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
    STATE(1044), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2104), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [55916] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1432), 1,
      sym_text_interpolation,
    ACTIONS(3014), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [55941] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1433), 1,
      sym_text_interpolation,
    ACTIONS(3016), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [55966] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1434), 1,
      sym_text_interpolation,
    ACTIONS(3018), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [55991] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1435), 1,
      sym_text_interpolation,
    ACTIONS(3020), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56016] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1436), 1,
      sym_text_interpolation,
    ACTIONS(3022), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56041] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1437), 1,
      sym_text_interpolation,
    ACTIONS(3024), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56066] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1438), 1,
      sym_text_interpolation,
    ACTIONS(3026), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56091] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1439), 1,
      sym_text_interpolation,
    ACTIONS(3028), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56116] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1440), 1,
      sym_text_interpolation,
    ACTIONS(3030), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56141] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1834), 1,
      anon_sym_BSLASH,
    STATE(1441), 1,
      sym_text_interpolation,
    STATE(2416), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(3032), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3034), 7,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [56172] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1442), 1,
      sym_text_interpolation,
    ACTIONS(3036), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56197] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1443), 1,
      sym_text_interpolation,
    ACTIONS(3038), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56222] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1834), 1,
      anon_sym_BSLASH,
    STATE(1444), 1,
      sym_text_interpolation,
    STATE(2416), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1610), 11,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      aux_sym_class_interface_clause_token1,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [56251] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1445), 1,
      sym_text_interpolation,
    ACTIONS(3040), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56276] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1726), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1446), 1,
      sym_text_interpolation,
    STATE(2077), 1,
      sym_namespace_use_clause,
    STATE(2231), 1,
      sym__name,
    STATE(2506), 1,
      sym_namespace_name,
    ACTIONS(3042), 2,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [56317] = 13,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2887), 1,
      anon_sym_BSLASHu,
    ACTIONS(3044), 1,
      anon_sym_LBRACE,
    ACTIONS(3046), 1,
      anon_sym_DOLLAR,
    STATE(1401), 1,
      aux_sym__interpolated_string_body_heredoc,
    STATE(1447), 1,
      sym_text_interpolation,
    STATE(1524), 1,
      sym_variable_name,
    STATE(1575), 1,
      sym__simple_string_member_access_expression,
    STATE(1598), 1,
      sym_dynamic_variable_name,
    ACTIONS(2885), 2,
      sym_encapsed_string_chars_heredoc,
      sym_escape_sequence,
    STATE(1581), 2,
      sym__complex_string_part,
      sym__simple_string_part,
    STATE(1599), 2,
      sym__simple_string_subscript_expression,
      sym__simple_variable,
  [56360] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1448), 1,
      sym_text_interpolation,
    ACTIONS(3048), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56385] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1449), 1,
      sym_text_interpolation,
    ACTIONS(3050), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56410] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1450), 1,
      sym_text_interpolation,
    ACTIONS(3052), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56435] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1451), 1,
      sym_text_interpolation,
    ACTIONS(3054), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56460] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1452), 1,
      sym_text_interpolation,
    ACTIONS(3056), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56485] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1453), 1,
      sym_text_interpolation,
    ACTIONS(3058), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56510] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1454), 1,
      sym_text_interpolation,
    ACTIONS(3060), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56535] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1726), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1455), 1,
      sym_text_interpolation,
    STATE(2231), 1,
      sym__name,
    STATE(2441), 1,
      sym_namespace_use_clause,
    STATE(2506), 1,
      sym_namespace_name,
    ACTIONS(3042), 2,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [56576] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1456), 1,
      sym_text_interpolation,
    ACTIONS(3062), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [56601] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1726), 1,
      sym_name,
    ACTIONS(3064), 1,
      anon_sym_RBRACK,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1457), 1,
      sym_text_interpolation,
    STATE(1868), 1,
      sym__name,
    STATE(2467), 1,
      sym_attribute,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [56641] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(328), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3066), 1,
      sym_name,
    ACTIONS(3070), 1,
      anon_sym_LBRACE,
    STATE(624), 1,
      sym__member_name,
    STATE(657), 1,
      sym__reserved_identifier,
    STATE(1458), 1,
      sym_text_interpolation,
    STATE(630), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(655), 2,
      sym__simple_variable,
      sym__identifier,
    ACTIONS(3068), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [56679] = 15,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(242), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3072), 1,
      sym_name,
    ACTIONS(3074), 1,
      anon_sym_RBRACE,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1459), 1,
      sym_text_interpolation,
    STATE(1460), 1,
      aux_sym_property_hook_list_repeat1,
    STATE(1790), 1,
      sym_property_hook,
    STATE(1794), 1,
      sym_attribute_list,
    STATE(2261), 1,
      sym_final_modifier,
    STATE(2614), 1,
      sym_reference_modifier,
  [56725] = 15,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(242), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3072), 1,
      sym_name,
    ACTIONS(3076), 1,
      anon_sym_RBRACE,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1460), 1,
      sym_text_interpolation,
    STATE(1489), 1,
      aux_sym_property_hook_list_repeat1,
    STATE(1790), 1,
      sym_property_hook,
    STATE(1794), 1,
      sym_attribute_list,
    STATE(2261), 1,
      sym_final_modifier,
    STATE(2614), 1,
      sym_reference_modifier,
  [56771] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1726), 1,
      sym_name,
    ACTIONS(3078), 1,
      anon_sym_RBRACK,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1461), 1,
      sym_text_interpolation,
    STATE(1868), 1,
      sym__name,
    STATE(2467), 1,
      sym_attribute,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [56811] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1562), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3080), 1,
      sym_name,
    ACTIONS(3084), 1,
      anon_sym_LBRACE,
    STATE(865), 1,
      sym__simple_variable,
    STATE(949), 1,
      sym__reserved_identifier,
    STATE(995), 1,
      sym__identifier,
    STATE(1462), 1,
      sym_text_interpolation,
    STATE(2499), 1,
      sym__member_name,
    STATE(868), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3082), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [56851] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3086), 1,
      sym_name,
    ACTIONS(3090), 1,
      anon_sym_LBRACE,
    STATE(593), 1,
      sym__member_name,
    STATE(598), 1,
      sym__reserved_identifier,
    STATE(1463), 1,
      sym_text_interpolation,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(607), 2,
      sym__simple_variable,
      sym__identifier,
    ACTIONS(3088), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [56889] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3086), 1,
      sym_name,
    ACTIONS(3090), 1,
      anon_sym_LBRACE,
    STATE(594), 1,
      sym__member_name,
    STATE(598), 1,
      sym__reserved_identifier,
    STATE(1464), 1,
      sym_text_interpolation,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(607), 2,
      sym__simple_variable,
      sym__identifier,
    ACTIONS(3088), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [56927] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3092), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(3094), 1,
      aux_sym__arrow_function_header_token1,
    STATE(1465), 1,
      sym_text_interpolation,
    ACTIONS(2406), 10,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
  [56955] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(328), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3066), 1,
      sym_name,
    ACTIONS(3070), 1,
      anon_sym_LBRACE,
    STATE(625), 1,
      sym__member_name,
    STATE(657), 1,
      sym__reserved_identifier,
    STATE(1466), 1,
      sym_text_interpolation,
    STATE(630), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(655), 2,
      sym__simple_variable,
      sym__identifier,
    ACTIONS(3068), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [56993] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1726), 1,
      sym_name,
    ACTIONS(3096), 1,
      anon_sym_RBRACK,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1467), 1,
      sym_text_interpolation,
    STATE(1868), 1,
      sym__name,
    STATE(2467), 1,
      sym_attribute,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57033] = 15,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(242), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3072), 1,
      sym_name,
    ACTIONS(3098), 1,
      anon_sym_RBRACE,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1468), 1,
      sym_text_interpolation,
    STATE(1469), 1,
      aux_sym_property_hook_list_repeat1,
    STATE(1790), 1,
      sym_property_hook,
    STATE(1794), 1,
      sym_attribute_list,
    STATE(2261), 1,
      sym_final_modifier,
    STATE(2614), 1,
      sym_reference_modifier,
  [57079] = 15,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(242), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(318), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3072), 1,
      sym_name,
    ACTIONS(3100), 1,
      anon_sym_RBRACE,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1469), 1,
      sym_text_interpolation,
    STATE(1489), 1,
      aux_sym_property_hook_list_repeat1,
    STATE(1790), 1,
      sym_property_hook,
    STATE(1794), 1,
      sym_attribute_list,
    STATE(2261), 1,
      sym_final_modifier,
    STATE(2614), 1,
      sym_reference_modifier,
  [57125] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1726), 1,
      sym_name,
    ACTIONS(3102), 1,
      anon_sym_RBRACK,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1470), 1,
      sym_text_interpolation,
    STATE(1868), 1,
      sym__name,
    STATE(2467), 1,
      sym_attribute,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57165] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1801), 1,
      anon_sym_DOLLAR,
    ACTIONS(3104), 1,
      sym_name,
    ACTIONS(3108), 1,
      anon_sym_LBRACE,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(763), 1,
      sym__identifier,
    STATE(951), 1,
      sym__simple_variable,
    STATE(1471), 1,
      sym_text_interpolation,
    STATE(2311), 1,
      sym__member_name,
    STATE(957), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3106), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57205] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3090), 1,
      anon_sym_LBRACE,
    ACTIONS(3104), 1,
      sym_name,
    STATE(593), 1,
      sym__member_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1472), 1,
      sym_text_interpolation,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(607), 2,
      sym__simple_variable,
      sym__identifier,
    ACTIONS(3106), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57243] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3090), 1,
      anon_sym_LBRACE,
    ACTIONS(3104), 1,
      sym_name,
    STATE(594), 1,
      sym__member_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1473), 1,
      sym_text_interpolation,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(607), 2,
      sym__simple_variable,
      sym__identifier,
    ACTIONS(3106), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57281] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3090), 1,
      anon_sym_LBRACE,
    ACTIONS(3104), 1,
      sym_name,
    STATE(610), 1,
      sym__member_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1474), 1,
      sym_text_interpolation,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(607), 2,
      sym__simple_variable,
      sym__identifier,
    ACTIONS(3106), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57319] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3090), 1,
      anon_sym_LBRACE,
    ACTIONS(3104), 1,
      sym_name,
    STATE(611), 1,
      sym__member_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1475), 1,
      sym_text_interpolation,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(607), 2,
      sym__simple_variable,
      sym__identifier,
    ACTIONS(3106), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57357] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1562), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3110), 1,
      sym_name,
    ACTIONS(3114), 1,
      anon_sym_LBRACE,
    STATE(859), 1,
      sym__member_name,
    STATE(878), 1,
      sym__reserved_identifier,
    STATE(1476), 1,
      sym_text_interpolation,
    STATE(868), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(873), 2,
      sym__simple_variable,
      sym__identifier,
    ACTIONS(3112), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57395] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1562), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3110), 1,
      sym_name,
    ACTIONS(3114), 1,
      anon_sym_LBRACE,
    STATE(860), 1,
      sym__member_name,
    STATE(878), 1,
      sym__reserved_identifier,
    STATE(1477), 1,
      sym_text_interpolation,
    STATE(868), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    STATE(873), 2,
      sym__simple_variable,
      sym__identifier,
    ACTIONS(3112), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57433] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(328), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3080), 1,
      sym_name,
    ACTIONS(3084), 1,
      anon_sym_LBRACE,
    STATE(652), 1,
      sym__simple_variable,
    STATE(949), 1,
      sym__reserved_identifier,
    STATE(995), 1,
      sym__identifier,
    STATE(1478), 1,
      sym_text_interpolation,
    STATE(2390), 1,
      sym__member_name,
    STATE(630), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3082), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57473] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1558), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3104), 1,
      sym_name,
    ACTIONS(3108), 1,
      anon_sym_LBRACE,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(734), 1,
      sym__simple_variable,
    STATE(763), 1,
      sym__identifier,
    STATE(1479), 1,
      sym_text_interpolation,
    STATE(2323), 1,
      sym__member_name,
    STATE(746), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3106), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57513] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1801), 1,
      anon_sym_DOLLAR,
    ACTIONS(3080), 1,
      sym_name,
    ACTIONS(3116), 1,
      anon_sym_LBRACE,
    STATE(897), 1,
      sym__member_name,
    STATE(949), 1,
      sym__reserved_identifier,
    STATE(1480), 1,
      sym_text_interpolation,
    STATE(956), 2,
      sym__simple_variable,
      sym__identifier,
    STATE(957), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3082), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57551] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1801), 1,
      anon_sym_DOLLAR,
    ACTIONS(3080), 1,
      sym_name,
    ACTIONS(3116), 1,
      anon_sym_LBRACE,
    STATE(898), 1,
      sym__member_name,
    STATE(949), 1,
      sym__reserved_identifier,
    STATE(1481), 1,
      sym_text_interpolation,
    STATE(956), 2,
      sym__simple_variable,
      sym__identifier,
    STATE(957), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3082), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57589] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1801), 1,
      anon_sym_DOLLAR,
    ACTIONS(3080), 1,
      sym_name,
    ACTIONS(3116), 1,
      anon_sym_LBRACE,
    STATE(949), 1,
      sym__reserved_identifier,
    STATE(959), 1,
      sym__member_name,
    STATE(1482), 1,
      sym_text_interpolation,
    STATE(956), 2,
      sym__simple_variable,
      sym__identifier,
    STATE(957), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3082), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57627] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1801), 1,
      anon_sym_DOLLAR,
    ACTIONS(3080), 1,
      sym_name,
    ACTIONS(3116), 1,
      anon_sym_LBRACE,
    STATE(949), 1,
      sym__reserved_identifier,
    STATE(960), 1,
      sym__member_name,
    STATE(1483), 1,
      sym_text_interpolation,
    STATE(956), 2,
      sym__simple_variable,
      sym__identifier,
    STATE(957), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3082), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57665] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1797), 1,
      anon_sym_DOLLAR,
    ACTIONS(3104), 1,
      sym_name,
    ACTIONS(3108), 1,
      anon_sym_LBRACE,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(763), 1,
      sym__identifier,
    STATE(819), 1,
      sym__simple_variable,
    STATE(1484), 1,
      sym_text_interpolation,
    STATE(2333), 1,
      sym__member_name,
    STATE(840), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3106), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57705] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1558), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3118), 1,
      sym_name,
    ACTIONS(3122), 1,
      anon_sym_LBRACE,
    STATE(718), 1,
      sym__member_name,
    STATE(731), 1,
      sym__reserved_identifier,
    STATE(1485), 1,
      sym_text_interpolation,
    STATE(735), 2,
      sym__simple_variable,
      sym__identifier,
    STATE(746), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3120), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57743] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1558), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3118), 1,
      sym_name,
    ACTIONS(3122), 1,
      anon_sym_LBRACE,
    STATE(721), 1,
      sym__member_name,
    STATE(731), 1,
      sym__reserved_identifier,
    STATE(1486), 1,
      sym_text_interpolation,
    STATE(735), 2,
      sym__simple_variable,
      sym__identifier,
    STATE(746), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3120), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57781] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3104), 1,
      sym_name,
    ACTIONS(3108), 1,
      anon_sym_LBRACE,
    STATE(595), 1,
      sym__simple_variable,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(763), 1,
      sym__identifier,
    STATE(1487), 1,
      sym_text_interpolation,
    STATE(2335), 1,
      sym__member_name,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3106), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57821] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1797), 1,
      anon_sym_DOLLAR,
    ACTIONS(3104), 1,
      sym_name,
    ACTIONS(3124), 1,
      anon_sym_LBRACE,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(759), 1,
      sym__member_name,
    STATE(1488), 1,
      sym_text_interpolation,
    STATE(821), 2,
      sym__simple_variable,
      sym__identifier,
    STATE(840), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3106), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57859] = 14,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3126), 1,
      sym_name,
    ACTIONS(3129), 1,
      anon_sym_AMP,
    ACTIONS(3132), 1,
      anon_sym_RBRACE,
    ACTIONS(3134), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(3137), 1,
      anon_sym_POUND_LBRACK,
    STATE(1364), 1,
      aux_sym_attribute_list_repeat1,
    STATE(1374), 1,
      sym_attribute_group,
    STATE(1790), 1,
      sym_property_hook,
    STATE(1794), 1,
      sym_attribute_list,
    STATE(2261), 1,
      sym_final_modifier,
    STATE(2614), 1,
      sym_reference_modifier,
    STATE(1489), 2,
      sym_text_interpolation,
      aux_sym_property_hook_list_repeat1,
  [57903] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1797), 1,
      anon_sym_DOLLAR,
    ACTIONS(3104), 1,
      sym_name,
    ACTIONS(3124), 1,
      anon_sym_LBRACE,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(760), 1,
      sym__member_name,
    STATE(1490), 1,
      sym_text_interpolation,
    STATE(821), 2,
      sym__simple_variable,
      sym__identifier,
    STATE(840), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3106), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57941] = 12,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1562), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3104), 1,
      sym_name,
    ACTIONS(3108), 1,
      anon_sym_LBRACE,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(763), 1,
      sym__identifier,
    STATE(865), 1,
      sym__simple_variable,
    STATE(1491), 1,
      sym_text_interpolation,
    STATE(2499), 1,
      sym__member_name,
    STATE(868), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3106), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [57981] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1797), 1,
      anon_sym_DOLLAR,
    ACTIONS(3104), 1,
      sym_name,
    ACTIONS(3124), 1,
      anon_sym_LBRACE,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(824), 1,
      sym__member_name,
    STATE(1492), 1,
      sym_text_interpolation,
    STATE(821), 2,
      sym__simple_variable,
      sym__identifier,
    STATE(840), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3106), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [58019] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1797), 1,
      anon_sym_DOLLAR,
    ACTIONS(3104), 1,
      sym_name,
    ACTIONS(3124), 1,
      anon_sym_LBRACE,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(825), 1,
      sym__member_name,
    STATE(1493), 1,
      sym_text_interpolation,
    STATE(821), 2,
      sym__simple_variable,
      sym__identifier,
    STATE(840), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
    ACTIONS(3106), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [58057] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3145), 1,
      anon_sym_AMP,
    ACTIONS(3152), 1,
      anon_sym_PIPE,
    STATE(1494), 1,
      sym_text_interpolation,
    STATE(1510), 1,
      aux_sym_union_type_repeat1,
    STATE(1511), 1,
      aux_sym_intersection_type_repeat1,
    STATE(1513), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3147), 3,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOLLAR,
    ACTIONS(3140), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
  [58093] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1495), 1,
      sym_text_interpolation,
    ACTIONS(3154), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3156), 7,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [58118] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1496), 1,
      sym_text_interpolation,
    ACTIONS(3158), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3160), 7,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [58143] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1497), 1,
      sym_text_interpolation,
    ACTIONS(3032), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3034), 7,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [58168] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1498), 1,
      sym_text_interpolation,
    ACTIONS(3162), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3164), 7,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [58193] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1499), 1,
      sym_text_interpolation,
    ACTIONS(2008), 11,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      aux_sym_class_interface_clause_token1,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [58216] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1726), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1500), 1,
      sym_text_interpolation,
    STATE(1868), 1,
      sym__name,
    STATE(2016), 1,
      sym_attribute,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [58253] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1726), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1501), 1,
      sym_text_interpolation,
    STATE(1868), 1,
      sym__name,
    STATE(2467), 1,
      sym_attribute,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [58290] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1726), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1502), 1,
      sym_text_interpolation,
    STATE(1868), 1,
      sym__name,
    STATE(1965), 1,
      sym_attribute,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [58327] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3166), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(3168), 1,
      aux_sym__arrow_function_header_token1,
    STATE(1503), 1,
      sym_text_interpolation,
    ACTIONS(2406), 9,
      aux_sym_function_static_declaration_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
  [58354] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3174), 1,
      anon_sym_PIPE,
    STATE(1504), 2,
      sym_text_interpolation,
      aux_sym_union_type_repeat1,
    ACTIONS(3170), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3172), 4,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOLLAR,
  [58380] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(226), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2094), 1,
      sym_name,
    STATE(1043), 1,
      sym__reserved_identifier,
    STATE(1505), 1,
      sym_text_interpolation,
    STATE(1875), 1,
      sym__name,
    STATE(2622), 1,
      sym_namespace_name,
    STATE(1044), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2104), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [58414] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    STATE(1506), 1,
      sym_text_interpolation,
    ACTIONS(1618), 5,
      sym_encapsed_string_chars_heredoc,
      sym_encapsed_string_chars_after_variable_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
    ACTIONS(1620), 5,
      anon_sym_LBRACE,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
  [58438] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3181), 1,
      anon_sym_PIPE,
    STATE(1507), 1,
      sym_text_interpolation,
    STATE(1526), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3177), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3179), 4,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOLLAR,
  [58466] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1508), 1,
      sym_text_interpolation,
    ACTIONS(3170), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3172), 6,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [58490] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3104), 1,
      sym_name,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1509), 1,
      sym_text_interpolation,
    STATE(1551), 1,
      sym_reference_modifier,
    STATE(1788), 1,
      sym_formal_parameters,
    STATE(2479), 1,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [58526] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3189), 1,
      anon_sym_PIPE,
    STATE(1504), 1,
      aux_sym_union_type_repeat1,
    STATE(1510), 1,
      sym_text_interpolation,
    ACTIONS(3185), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3187), 4,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOLLAR,
  [58554] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3145), 1,
      anon_sym_AMP,
    STATE(1511), 1,
      sym_text_interpolation,
    STATE(1525), 1,
      aux_sym_intersection_type_repeat1,
    ACTIONS(3191), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3193), 4,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DOLLAR,
  [58582] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1512), 1,
      sym_text_interpolation,
    ACTIONS(3195), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3197), 6,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [58606] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3181), 1,
      anon_sym_PIPE,
    STATE(1513), 1,
      sym_text_interpolation,
    STATE(1526), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3199), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3201), 4,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOLLAR,
  [58634] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1514), 1,
      sym_text_interpolation,
    ACTIONS(3203), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3205), 6,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [58658] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3104), 1,
      sym_name,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1515), 1,
      sym_text_interpolation,
    STATE(1545), 1,
      sym_reference_modifier,
    STATE(1826), 1,
      sym_formal_parameters,
    STATE(2479), 1,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [58694] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1726), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1516), 1,
      sym_text_interpolation,
    STATE(1919), 1,
      sym__name,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [58728] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(226), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2094), 1,
      sym_name,
    STATE(1043), 1,
      sym__reserved_identifier,
    STATE(1517), 1,
      sym_text_interpolation,
    STATE(1911), 1,
      sym__name,
    STATE(2731), 1,
      sym_namespace_name,
    STATE(1044), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2104), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [58762] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(226), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2094), 1,
      sym_name,
    STATE(1043), 1,
      sym__reserved_identifier,
    STATE(1518), 1,
      sym_text_interpolation,
    STATE(1911), 1,
      sym__name,
    STATE(2622), 1,
      sym_namespace_name,
    STATE(1044), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2104), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [58796] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2134), 1,
      anon_sym_COLON,
    ACTIONS(2836), 1,
      anon_sym_COLON_COLON,
    STATE(1519), 1,
      sym_text_interpolation,
    ACTIONS(1814), 2,
      aux_sym__namespace_use_type_token1,
      aux_sym__arrow_function_header_token1,
    ACTIONS(1614), 6,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [58824] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3181), 1,
      anon_sym_PIPE,
    STATE(1507), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    STATE(1520), 1,
      sym_text_interpolation,
    ACTIONS(3207), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3209), 4,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOLLAR,
  [58852] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1521), 1,
      sym_text_interpolation,
    ACTIONS(3211), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3213), 6,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [58876] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1726), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1522), 1,
      sym_text_interpolation,
    STATE(2291), 1,
      sym__name,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [58910] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1523), 1,
      sym_text_interpolation,
    ACTIONS(3215), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3218), 6,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [58934] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3221), 1,
      anon_sym_DASH_GT,
    ACTIONS(3223), 1,
      anon_sym_LBRACK,
    ACTIONS(3225), 1,
      sym_encapsed_string_chars_after_variable_heredoc,
    STATE(1524), 1,
      sym_text_interpolation,
    ACTIONS(1608), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(1606), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [58964] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3227), 1,
      anon_sym_AMP,
    STATE(1525), 2,
      sym_text_interpolation,
      aux_sym_intersection_type_repeat1,
    ACTIONS(3211), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3213), 4,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DOLLAR,
  [58990] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3230), 1,
      anon_sym_PIPE,
    STATE(1526), 2,
      sym_text_interpolation,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3195), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3197), 4,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOLLAR,
  [59016] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(226), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2094), 1,
      sym_name,
    STATE(1043), 1,
      sym__reserved_identifier,
    STATE(1527), 1,
      sym_text_interpolation,
    STATE(1568), 1,
      sym__name,
    STATE(2622), 1,
      sym_namespace_name,
    STATE(1044), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2104), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [59050] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1726), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1528), 1,
      sym_text_interpolation,
    STATE(2152), 1,
      sym__name,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [59084] = 10,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1726), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1529), 1,
      sym_text_interpolation,
    STATE(2294), 1,
      sym__name,
    STATE(2506), 1,
      sym_namespace_name,
    STATE(1499), 2,
      sym_qualified_name,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [59118] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_LBRACE,
    ACTIONS(3235), 1,
      anon_sym_EQ_GT,
    ACTIONS(3237), 1,
      anon_sym_LPAREN2,
    STATE(1530), 1,
      sym_text_interpolation,
    STATE(1587), 1,
      sym_formal_parameters,
    STATE(1712), 1,
      sym_compound_statement,
    STATE(1735), 1,
      sym__property_hook_body,
    STATE(1844), 1,
      sym__semicolon,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59153] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_LBRACE,
    ACTIONS(3235), 1,
      anon_sym_EQ_GT,
    ACTIONS(3237), 1,
      anon_sym_LPAREN2,
    STATE(1531), 1,
      sym_text_interpolation,
    STATE(1576), 1,
      sym_formal_parameters,
    STATE(1712), 1,
      sym_compound_statement,
    STATE(1813), 1,
      sym__property_hook_body,
    STATE(1844), 1,
      sym__semicolon,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59188] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_LBRACE,
    ACTIONS(3235), 1,
      anon_sym_EQ_GT,
    ACTIONS(3237), 1,
      anon_sym_LPAREN2,
    STATE(1532), 1,
      sym_text_interpolation,
    STATE(1605), 1,
      sym_formal_parameters,
    STATE(1712), 1,
      sym_compound_statement,
    STATE(1816), 1,
      sym__property_hook_body,
    STATE(1844), 1,
      sym__semicolon,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59223] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_LBRACE,
    ACTIONS(3235), 1,
      anon_sym_EQ_GT,
    ACTIONS(3237), 1,
      anon_sym_LPAREN2,
    STATE(1533), 1,
      sym_text_interpolation,
    STATE(1607), 1,
      sym_formal_parameters,
    STATE(1712), 1,
      sym_compound_statement,
    STATE(1817), 1,
      sym__property_hook_body,
    STATE(1844), 1,
      sym__semicolon,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59258] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_BSLASHu,
    STATE(1534), 1,
      sym_text_interpolation,
    ACTIONS(1618), 8,
      sym_encapsed_string_chars,
      sym_encapsed_string_chars_after_variable,
      anon_sym_LBRACE,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [59281] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_LBRACE,
    ACTIONS(3235), 1,
      anon_sym_EQ_GT,
    ACTIONS(3237), 1,
      anon_sym_LPAREN2,
    STATE(1535), 1,
      sym_text_interpolation,
    STATE(1610), 1,
      sym_formal_parameters,
    STATE(1712), 1,
      sym_compound_statement,
    STATE(1824), 1,
      sym__property_hook_body,
    STATE(1844), 1,
      sym__semicolon,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59316] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_BSLASHu,
    STATE(1536), 1,
      sym_text_interpolation,
    ACTIONS(1618), 8,
      sym_execution_string_chars,
      sym_execution_string_chars_after_variable,
      anon_sym_LBRACE,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [59339] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1608), 1,
      anon_sym_BSLASHu,
    ACTIONS(3239), 1,
      anon_sym_DASH_GT,
    ACTIONS(3241), 1,
      anon_sym_LBRACK,
    ACTIONS(3243), 1,
      sym_encapsed_string_chars_after_variable,
    STATE(1537), 1,
      sym_text_interpolation,
    ACTIONS(1606), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [59368] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1538), 1,
      sym_text_interpolation,
    ACTIONS(3245), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3247), 5,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [59391] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1608), 1,
      anon_sym_BSLASHu,
    ACTIONS(3249), 1,
      anon_sym_DASH_GT,
    ACTIONS(3251), 1,
      anon_sym_LBRACK,
    ACTIONS(3253), 1,
      sym_execution_string_chars_after_variable,
    STATE(1539), 1,
      sym_text_interpolation,
    ACTIONS(1606), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [59420] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_LBRACE,
    ACTIONS(3235), 1,
      anon_sym_EQ_GT,
    ACTIONS(3237), 1,
      anon_sym_LPAREN2,
    STATE(1540), 1,
      sym_text_interpolation,
    STATE(1573), 1,
      sym_formal_parameters,
    STATE(1712), 1,
      sym_compound_statement,
    STATE(1755), 1,
      sym__property_hook_body,
    STATE(1844), 1,
      sym__semicolon,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59455] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1610), 1,
      anon_sym_EQ,
    ACTIONS(1834), 1,
      anon_sym_BSLASH,
    STATE(1541), 1,
      sym_text_interpolation,
    STATE(2416), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(3034), 2,
      anon_sym_AMP,
      anon_sym_PIPE,
    ACTIONS(3032), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
  [59484] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2836), 1,
      anon_sym_COLON_COLON,
    STATE(1542), 1,
      sym_text_interpolation,
    ACTIONS(1814), 2,
      aux_sym__namespace_use_type_token1,
      aux_sym__arrow_function_header_token1,
    ACTIONS(1614), 6,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [59509] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_LBRACE,
    ACTIONS(3235), 1,
      anon_sym_EQ_GT,
    ACTIONS(3237), 1,
      anon_sym_LPAREN2,
    STATE(1543), 1,
      sym_text_interpolation,
    STATE(1608), 1,
      sym_formal_parameters,
    STATE(1712), 1,
      sym_compound_statement,
    STATE(1728), 1,
      sym__property_hook_body,
    STATE(1844), 1,
      sym__semicolon,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59544] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_LBRACE,
    ACTIONS(3235), 1,
      anon_sym_EQ_GT,
    ACTIONS(3237), 1,
      anon_sym_LPAREN2,
    STATE(1544), 1,
      sym_text_interpolation,
    STATE(1579), 1,
      sym_formal_parameters,
    STATE(1712), 1,
      sym_compound_statement,
    STATE(1780), 1,
      sym__property_hook_body,
    STATE(1844), 1,
      sym__semicolon,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59579] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3104), 1,
      sym_name,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1545), 1,
      sym_text_interpolation,
    STATE(1836), 1,
      sym_formal_parameters,
    STATE(2314), 1,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [59609] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2134), 1,
      anon_sym_COLON,
    ACTIONS(2836), 1,
      anon_sym_COLON_COLON,
    STATE(1546), 1,
      sym_text_interpolation,
    ACTIONS(1614), 6,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [59633] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3255), 1,
      anon_sym_AMP,
    ACTIONS(3257), 1,
      anon_sym_PIPE,
    STATE(1547), 1,
      sym_text_interpolation,
    STATE(1766), 1,
      aux_sym_union_type_repeat1,
    STATE(1767), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    STATE(1781), 1,
      aux_sym_intersection_type_repeat1,
    ACTIONS(3147), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [59663] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(868), 1,
      anon_sym_COMMA,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    ACTIONS(3259), 1,
      anon_sym_RPAREN2,
    STATE(1548), 1,
      sym_text_interpolation,
    STATE(1980), 1,
      aux_sym__list_destructing_repeat1,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [59691] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1834), 1,
      anon_sym_BSLASH,
    ACTIONS(3261), 1,
      aux_sym_namespace_use_clause_token1,
    STATE(1549), 1,
      sym_text_interpolation,
    STATE(2416), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1610), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [59717] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(989), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3263), 1,
      aux_sym_catch_clause_token1,
    ACTIONS(3265), 1,
      aux_sym_finally_clause_token1,
    STATE(1550), 1,
      sym_text_interpolation,
    STATE(1562), 1,
      aux_sym_try_statement_repeat1,
    ACTIONS(987), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
    STATE(1777), 2,
      sym_catch_clause,
      sym_finally_clause,
  [59747] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3104), 1,
      sym_name,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1551), 1,
      sym_text_interpolation,
    STATE(1812), 1,
      sym_formal_parameters,
    STATE(2314), 1,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [59777] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    ACTIONS(3267), 1,
      anon_sym_COMMA,
    ACTIONS(3269), 1,
      anon_sym_RPAREN2,
    STATE(1552), 1,
      sym_text_interpolation,
    STATE(2051), 1,
      aux_sym_unset_statement_repeat1,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [59805] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1590), 1,
      anon_sym_LPAREN2,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(836), 1,
      sym_declaration_list,
    STATE(1553), 1,
      sym_text_interpolation,
    STATE(1638), 1,
      sym_arguments,
    STATE(1912), 1,
      sym_base_clause,
    STATE(2361), 1,
      sym_class_interface_clause,
  [59839] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1590), 1,
      anon_sym_LPAREN2,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(822), 1,
      sym_declaration_list,
    STATE(1554), 1,
      sym_text_interpolation,
    STATE(1639), 1,
      sym_arguments,
    STATE(1899), 1,
      sym_base_clause,
    STATE(2356), 1,
      sym_class_interface_clause,
  [59873] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1590), 1,
      anon_sym_LPAREN2,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(973), 1,
      sym_declaration_list,
    STATE(1555), 1,
      sym_text_interpolation,
    STATE(1622), 1,
      sym_arguments,
    STATE(1942), 1,
      sym_base_clause,
    STATE(2394), 1,
      sym_class_interface_clause,
  [59907] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1590), 1,
      anon_sym_LPAREN2,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(803), 1,
      sym_declaration_list,
    STATE(1556), 1,
      sym_text_interpolation,
    STATE(1630), 1,
      sym_arguments,
    STATE(1937), 1,
      sym_base_clause,
    STATE(2322), 1,
      sym_class_interface_clause,
  [59941] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1590), 1,
      anon_sym_LPAREN2,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(768), 1,
      sym_declaration_list,
    STATE(1557), 1,
      sym_text_interpolation,
    STATE(1645), 1,
      sym_arguments,
    STATE(1856), 1,
      sym_base_clause,
    STATE(2397), 1,
      sym_class_interface_clause,
  [59975] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3104), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1558), 1,
      sym_text_interpolation,
    STATE(1626), 1,
      sym_reference_modifier,
    STATE(2362), 1,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [60005] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1590), 1,
      anon_sym_LPAREN2,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(976), 1,
      sym_declaration_list,
    STATE(1559), 1,
      sym_text_interpolation,
    STATE(1706), 1,
      sym_arguments,
    STATE(1938), 1,
      sym_base_clause,
    STATE(2385), 1,
      sym_class_interface_clause,
  [60039] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    ACTIONS(3279), 1,
      anon_sym_COMMA,
    ACTIONS(3281), 1,
      anon_sym_RPAREN2,
    STATE(1560), 1,
      sym_text_interpolation,
    STATE(2295), 1,
      aux_sym_unset_statement_repeat1,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [60067] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1590), 1,
      anon_sym_LPAREN2,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(988), 1,
      sym_declaration_list,
    STATE(1561), 1,
      sym_text_interpolation,
    STATE(1707), 1,
      sym_arguments,
    STATE(1889), 1,
      sym_base_clause,
    STATE(2372), 1,
      sym_class_interface_clause,
  [60101] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(997), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3283), 1,
      aux_sym_catch_clause_token1,
    ACTIONS(3286), 1,
      aux_sym_finally_clause_token1,
    ACTIONS(995), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
    STATE(1562), 2,
      sym_text_interpolation,
      aux_sym_try_statement_repeat1,
    STATE(1777), 2,
      sym_catch_clause,
      sym_finally_clause,
  [60129] = 11,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1590), 1,
      anon_sym_LPAREN2,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(963), 1,
      sym_declaration_list,
    STATE(1563), 1,
      sym_text_interpolation,
    STATE(1620), 1,
      sym_arguments,
    STATE(1846), 1,
      sym_base_clause,
    STATE(2332), 1,
      sym_class_interface_clause,
  [60163] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3291), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3294), 1,
      aux_sym_match_default_expression_token1,
    ACTIONS(3289), 2,
      anon_sym_RBRACE,
      aux_sym_switch_block_token1,
    STATE(1564), 2,
      sym_text_interpolation,
      aux_sym_switch_block_repeat1,
    STATE(1888), 2,
      sym_case_statement,
      sym_default_statement,
  [60188] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3297), 1,
      anon_sym_COMMA,
    ACTIONS(3299), 1,
      anon_sym_LBRACE,
    STATE(1565), 1,
      sym_text_interpolation,
    STATE(1583), 1,
      aux_sym_property_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1429), 2,
      sym_property_hook_list,
      sym__semicolon,
  [60215] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    STATE(1566), 1,
      sym_text_interpolation,
    ACTIONS(3301), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(2915), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [60236] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3297), 1,
      anon_sym_COMMA,
    ACTIONS(3299), 1,
      anon_sym_LBRACE,
    STATE(1567), 1,
      sym_text_interpolation,
    STATE(1785), 1,
      aux_sym_property_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1432), 2,
      sym_property_hook_list,
      sym__semicolon,
  [60263] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_COMMA,
    ACTIONS(3305), 1,
      anon_sym_LBRACE,
    STATE(1568), 1,
      sym_text_interpolation,
    STATE(1594), 1,
      aux_sym_base_clause_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1417), 2,
      sym_use_list,
      sym__semicolon,
  [60290] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(868), 1,
      anon_sym_COMMA,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3259), 1,
      anon_sym_RPAREN2,
    STATE(1569), 1,
      sym_text_interpolation,
    STATE(1980), 1,
      aux_sym__list_destructing_repeat1,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [60315] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1570), 1,
      sym_text_interpolation,
    ACTIONS(2594), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1610), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [60336] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3297), 1,
      anon_sym_COMMA,
    ACTIONS(3299), 1,
      anon_sym_LBRACE,
    STATE(1571), 1,
      sym_text_interpolation,
    STATE(1618), 1,
      aux_sym_property_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1437), 2,
      sym_property_hook_list,
      sym__semicolon,
  [60363] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1834), 1,
      anon_sym_BSLASH,
    STATE(1572), 1,
      sym_text_interpolation,
    STATE(2416), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(3034), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [60386] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_LBRACE,
    ACTIONS(3235), 1,
      anon_sym_EQ_GT,
    STATE(1573), 1,
      sym_text_interpolation,
    STATE(1712), 1,
      sym_compound_statement,
    STATE(1798), 1,
      sym__property_hook_body,
    STATE(1844), 1,
      sym__semicolon,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60415] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(868), 1,
      anon_sym_COMMA,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3259), 1,
      anon_sym_RPAREN2,
    STATE(1574), 1,
      sym_text_interpolation,
    STATE(1980), 1,
      aux_sym__list_destructing_repeat1,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [60440] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    STATE(1575), 1,
      sym_text_interpolation,
    ACTIONS(3307), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(3309), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [60461] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_LBRACE,
    ACTIONS(3235), 1,
      anon_sym_EQ_GT,
    STATE(1576), 1,
      sym_text_interpolation,
    STATE(1712), 1,
      sym_compound_statement,
    STATE(1815), 1,
      sym__property_hook_body,
    STATE(1844), 1,
      sym__semicolon,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60490] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1577), 1,
      sym_text_interpolation,
    ACTIONS(2538), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1610), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [60511] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(3313), 1,
      anon_sym_DASH,
    STATE(1578), 1,
      sym_text_interpolation,
    STATE(2600), 1,
      sym__simple_string_array_access_argument,
    ACTIONS(3311), 2,
      sym_integer,
      sym_name,
    STATE(2724), 2,
      sym__simple_string_subscript_unary_expression,
      sym_variable_name,
  [60538] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_LBRACE,
    ACTIONS(3235), 1,
      anon_sym_EQ_GT,
    STATE(1579), 1,
      sym_text_interpolation,
    STATE(1712), 1,
      sym_compound_statement,
    STATE(1751), 1,
      sym__property_hook_body,
    STATE(1844), 1,
      sym__semicolon,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60567] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3267), 1,
      anon_sym_COMMA,
    ACTIONS(3269), 1,
      anon_sym_RPAREN2,
    STATE(1580), 1,
      sym_text_interpolation,
    STATE(2051), 1,
      aux_sym_unset_statement_repeat1,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [60592] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    STATE(1581), 1,
      sym_text_interpolation,
    ACTIONS(3315), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(3317), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [60613] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3319), 1,
      anon_sym_AMP,
    ACTIONS(3325), 1,
      anon_sym_PIPE,
    STATE(1582), 1,
      sym_text_interpolation,
    STATE(1827), 1,
      aux_sym_union_type_repeat1,
    STATE(1828), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    STATE(1876), 1,
      aux_sym_intersection_type_repeat1,
    ACTIONS(3147), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [60642] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3297), 1,
      anon_sym_COMMA,
    ACTIONS(3299), 1,
      anon_sym_LBRACE,
    STATE(1583), 1,
      sym_text_interpolation,
    STATE(1785), 1,
      aux_sym_property_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1449), 2,
      sym_property_hook_list,
      sym__semicolon,
  [60669] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3297), 1,
      anon_sym_COMMA,
    ACTIONS(3299), 1,
      anon_sym_LBRACE,
    STATE(1584), 1,
      sym_text_interpolation,
    STATE(1616), 1,
      aux_sym_property_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1450), 2,
      sym_property_hook_list,
      sym__semicolon,
  [60696] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(1585), 1,
      sym_text_interpolation,
    ACTIONS(3327), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [60719] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3267), 1,
      anon_sym_COMMA,
    ACTIONS(3269), 1,
      anon_sym_RPAREN2,
    STATE(1586), 1,
      sym_text_interpolation,
    STATE(2051), 1,
      aux_sym_unset_statement_repeat1,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [60744] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_LBRACE,
    ACTIONS(3235), 1,
      anon_sym_EQ_GT,
    STATE(1587), 1,
      sym_text_interpolation,
    STATE(1712), 1,
      sym_compound_statement,
    STATE(1742), 1,
      sym__property_hook_body,
    STATE(1844), 1,
      sym__semicolon,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60773] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    STATE(1588), 1,
      sym_text_interpolation,
    ACTIONS(3329), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(3331), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [60794] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1589), 1,
      sym_text_interpolation,
    ACTIONS(2548), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1610), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [60815] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2836), 1,
      anon_sym_COLON_COLON,
    STATE(1590), 1,
      sym_text_interpolation,
    ACTIONS(1614), 6,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [60836] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3104), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1591), 1,
      sym_text_interpolation,
    STATE(1829), 1,
      sym_const_element,
    STATE(2591), 1,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [60863] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(1592), 1,
      sym_text_interpolation,
    ACTIONS(3333), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [60886] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(3313), 1,
      anon_sym_DASH,
    STATE(1593), 1,
      sym_text_interpolation,
    STATE(2606), 1,
      sym__simple_string_array_access_argument,
    ACTIONS(3311), 2,
      sym_integer,
      sym_name,
    STATE(2724), 2,
      sym__simple_string_subscript_unary_expression,
      sym_variable_name,
  [60913] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_COMMA,
    ACTIONS(3305), 1,
      anon_sym_LBRACE,
    STATE(1594), 1,
      sym_text_interpolation,
    STATE(1732), 1,
      aux_sym_base_clause_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1404), 2,
      sym_use_list,
      sym__semicolon,
  [60940] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(3313), 1,
      anon_sym_DASH,
    STATE(1595), 1,
      sym_text_interpolation,
    STATE(2559), 1,
      sym__simple_string_array_access_argument,
    ACTIONS(3311), 2,
      sym_integer,
      sym_name,
    STATE(2724), 2,
      sym__simple_string_subscript_unary_expression,
      sym_variable_name,
  [60967] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1596), 1,
      sym_text_interpolation,
    ACTIONS(2114), 3,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
    ACTIONS(2112), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
  [60988] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(1597), 1,
      sym_text_interpolation,
    ACTIONS(3335), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [61011] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    STATE(1598), 1,
      sym_text_interpolation,
    ACTIONS(1608), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(1606), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [61032] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    STATE(1599), 1,
      sym_text_interpolation,
    ACTIONS(3337), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(3339), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [61053] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_COMMA,
    ACTIONS(3281), 1,
      anon_sym_RPAREN2,
    STATE(1600), 1,
      sym_text_interpolation,
    STATE(2295), 1,
      aux_sym_unset_statement_repeat1,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [61078] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3297), 1,
      anon_sym_COMMA,
    ACTIONS(3299), 1,
      anon_sym_LBRACE,
    STATE(1567), 1,
      aux_sym_property_declaration_repeat1,
    STATE(1601), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1442), 2,
      sym_property_hook_list,
      sym__semicolon,
  [61105] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3104), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1602), 1,
      sym_text_interpolation,
    STATE(1727), 1,
      sym_const_element,
    STATE(2591), 1,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [61132] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    STATE(1603), 1,
      sym_text_interpolation,
    ACTIONS(1624), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(1622), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [61153] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    STATE(1604), 1,
      sym_text_interpolation,
    ACTIONS(1688), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(1686), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [61174] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_LBRACE,
    ACTIONS(3235), 1,
      anon_sym_EQ_GT,
    STATE(1605), 1,
      sym_text_interpolation,
    STATE(1712), 1,
      sym_compound_statement,
    STATE(1805), 1,
      sym__property_hook_body,
    STATE(1844), 1,
      sym__semicolon,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [61203] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3104), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1606), 1,
      sym_text_interpolation,
    STATE(2257), 1,
      sym_const_element,
    STATE(2591), 1,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [61230] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_LBRACE,
    ACTIONS(3235), 1,
      anon_sym_EQ_GT,
    STATE(1607), 1,
      sym_text_interpolation,
    STATE(1710), 1,
      sym__property_hook_body,
    STATE(1712), 1,
      sym_compound_statement,
    STATE(1844), 1,
      sym__semicolon,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [61259] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_LBRACE,
    ACTIONS(3235), 1,
      anon_sym_EQ_GT,
    STATE(1608), 1,
      sym_text_interpolation,
    STATE(1712), 1,
      sym_compound_statement,
    STATE(1738), 1,
      sym__property_hook_body,
    STATE(1844), 1,
      sym__semicolon,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [61288] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    STATE(1609), 1,
      sym_text_interpolation,
    ACTIONS(3341), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(3343), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [61309] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_LBRACE,
    ACTIONS(3235), 1,
      anon_sym_EQ_GT,
    STATE(1610), 1,
      sym_text_interpolation,
    STATE(1712), 1,
      sym_compound_statement,
    STATE(1731), 1,
      sym__property_hook_body,
    STATE(1844), 1,
      sym__semicolon,
    ACTIONS(2482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [61338] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1611), 1,
      sym_text_interpolation,
    ACTIONS(2556), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1610), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [61359] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    STATE(1612), 1,
      sym_text_interpolation,
    ACTIONS(1620), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(1618), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [61380] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    STATE(1613), 1,
      sym_text_interpolation,
    ACTIONS(3345), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(3347), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [61401] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3104), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1614), 1,
      sym_text_interpolation,
    STATE(1726), 1,
      sym_const_element,
    STATE(2591), 1,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [61428] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_COMMA,
    ACTIONS(3281), 1,
      anon_sym_RPAREN2,
    STATE(1615), 1,
      sym_text_interpolation,
    STATE(2295), 1,
      aux_sym_unset_statement_repeat1,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [61453] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3297), 1,
      anon_sym_COMMA,
    ACTIONS(3299), 1,
      anon_sym_LBRACE,
    STATE(1616), 1,
      sym_text_interpolation,
    STATE(1785), 1,
      aux_sym_property_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1430), 2,
      sym_property_hook_list,
      sym__semicolon,
  [61480] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3104), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1617), 1,
      sym_text_interpolation,
    STATE(1763), 1,
      sym_const_element,
    STATE(2591), 1,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [61507] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3297), 1,
      anon_sym_COMMA,
    ACTIONS(3299), 1,
      anon_sym_LBRACE,
    STATE(1618), 1,
      sym_text_interpolation,
    STATE(1785), 1,
      aux_sym_property_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1451), 2,
      sym_property_hook_list,
      sym__semicolon,
  [61534] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3345), 1,
      anon_sym_BSLASHu,
    STATE(1619), 1,
      sym_text_interpolation,
    ACTIONS(3347), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [61554] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(1003), 1,
      sym_declaration_list,
    STATE(1620), 1,
      sym_text_interpolation,
    STATE(1852), 1,
      sym_base_clause,
    STATE(2353), 1,
      sym_class_interface_clause,
  [61582] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(1621), 1,
      sym_text_interpolation,
    STATE(1851), 1,
      sym_base_clause,
    STATE(2255), 1,
      sym_declaration_list,
    STATE(2302), 1,
      sym_class_interface_clause,
  [61610] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(988), 1,
      sym_declaration_list,
    STATE(1622), 1,
      sym_text_interpolation,
    STATE(1889), 1,
      sym_base_clause,
    STATE(2372), 1,
      sym_class_interface_clause,
  [61638] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3349), 1,
      anon_sym_SQUOTE,
    ACTIONS(3351), 1,
      aux_sym_string_token2,
    ACTIONS(3353), 1,
      aux_sym_string_content_token1,
    STATE(1623), 1,
      sym_text_interpolation,
    STATE(1650), 1,
      aux_sym_string_repeat1,
    STATE(1926), 1,
      aux_sym_string_content_repeat1,
    STATE(1997), 1,
      sym_string_content,
  [61666] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3355), 1,
      sym_name,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1624), 1,
      sym_text_interpolation,
    STATE(1857), 1,
      sym_named_type,
    STATE(2287), 1,
      sym_type_list,
    STATE(2506), 1,
      sym_namespace_name,
  [61694] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3351), 1,
      aux_sym_string_token2,
    ACTIONS(3353), 1,
      aux_sym_string_content_token1,
    ACTIONS(3357), 1,
      anon_sym_SQUOTE,
    STATE(1625), 1,
      sym_text_interpolation,
    STATE(1650), 1,
      aux_sym_string_repeat1,
    STATE(1926), 1,
      aux_sym_string_content_repeat1,
    STATE(1997), 1,
      sym_string_content,
  [61722] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3104), 1,
      sym_name,
    STATE(717), 1,
      sym__reserved_identifier,
    STATE(1626), 1,
      sym_text_interpolation,
    STATE(2491), 1,
      sym__identifier,
    ACTIONS(2090), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [61746] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3359), 1,
      aux_sym_if_statement_token2,
    ACTIONS(3361), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3363), 1,
      aux_sym_else_clause_token1,
    STATE(1627), 1,
      sym_text_interpolation,
    STATE(1635), 1,
      aux_sym_if_statement_repeat2,
    STATE(2225), 1,
      sym_else_if_clause_2,
    STATE(2518), 1,
      sym_else_clause_2,
  [61774] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3365), 1,
      anon_sym_RBRACE,
    ACTIONS(3367), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3369), 1,
      aux_sym_match_default_expression_token1,
    STATE(1628), 1,
      sym_text_interpolation,
    STATE(1636), 1,
      aux_sym_switch_block_repeat1,
    STATE(1888), 2,
      sym_case_statement,
      sym_default_statement,
  [61800] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3369), 1,
      aux_sym_match_default_expression_token1,
    ACTIONS(3371), 1,
      aux_sym_switch_block_token1,
    STATE(1629), 1,
      sym_text_interpolation,
    STATE(1637), 1,
      aux_sym_switch_block_repeat1,
    STATE(1888), 2,
      sym_case_statement,
      sym_default_statement,
  [61826] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(822), 1,
      sym_declaration_list,
    STATE(1630), 1,
      sym_text_interpolation,
    STATE(1899), 1,
      sym_base_clause,
    STATE(2356), 1,
      sym_class_interface_clause,
  [61854] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1706), 1,
      anon_sym_LPAREN2,
    STATE(631), 1,
      sym_arguments,
    STATE(1631), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [61876] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(1632), 1,
      sym_text_interpolation,
    STATE(1940), 1,
      sym_base_clause,
    STATE(2046), 1,
      sym_declaration_list,
    STATE(2378), 1,
      sym_class_interface_clause,
  [61904] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3329), 1,
      anon_sym_BSLASHu,
    STATE(1633), 1,
      sym_text_interpolation,
    ACTIONS(3331), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [61924] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(1634), 1,
      sym_text_interpolation,
    STATE(1951), 1,
      sym_base_clause,
    STATE(2052), 1,
      sym_declaration_list,
    STATE(2380), 1,
      sym_class_interface_clause,
  [61952] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3361), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3363), 1,
      aux_sym_else_clause_token1,
    ACTIONS(3373), 1,
      aux_sym_if_statement_token2,
    STATE(1635), 1,
      sym_text_interpolation,
    STATE(1808), 1,
      aux_sym_if_statement_repeat2,
    STATE(2225), 1,
      sym_else_if_clause_2,
    STATE(2674), 1,
      sym_else_clause_2,
  [61980] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3369), 1,
      aux_sym_match_default_expression_token1,
    ACTIONS(3375), 1,
      anon_sym_RBRACE,
    STATE(1564), 1,
      aux_sym_switch_block_repeat1,
    STATE(1636), 1,
      sym_text_interpolation,
    STATE(1888), 2,
      sym_case_statement,
      sym_default_statement,
  [62006] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3369), 1,
      aux_sym_match_default_expression_token1,
    ACTIONS(3377), 1,
      aux_sym_switch_block_token1,
    STATE(1564), 1,
      aux_sym_switch_block_repeat1,
    STATE(1637), 1,
      sym_text_interpolation,
    STATE(1888), 2,
      sym_case_statement,
      sym_default_statement,
  [62032] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(768), 1,
      sym_declaration_list,
    STATE(1638), 1,
      sym_text_interpolation,
    STATE(1856), 1,
      sym_base_clause,
    STATE(2397), 1,
      sym_class_interface_clause,
  [62060] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(767), 1,
      sym_declaration_list,
    STATE(1639), 1,
      sym_text_interpolation,
    STATE(1865), 1,
      sym_base_clause,
    STATE(2395), 1,
      sym_class_interface_clause,
  [62088] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1640), 1,
      sym_text_interpolation,
    ACTIONS(3333), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [62108] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3307), 1,
      anon_sym_BSLASHu,
    STATE(1641), 1,
      sym_text_interpolation,
    ACTIONS(3309), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [62128] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3337), 1,
      anon_sym_BSLASHu,
    STATE(1642), 1,
      sym_text_interpolation,
    ACTIONS(3339), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [62148] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(1643), 1,
      sym_text_interpolation,
    STATE(1878), 1,
      sym_base_clause,
    STATE(2100), 1,
      sym_declaration_list,
    STATE(2415), 1,
      sym_class_interface_clause,
  [62176] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3381), 1,
      anon_sym_BSLASHu,
    STATE(1644), 1,
      sym_text_interpolation,
    ACTIONS(3379), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [62196] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(778), 1,
      sym_declaration_list,
    STATE(1645), 1,
      sym_text_interpolation,
    STATE(1894), 1,
      sym_base_clause,
    STATE(2428), 1,
      sym_class_interface_clause,
  [62224] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1646), 1,
      sym_text_interpolation,
    ACTIONS(3333), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [62244] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1807), 1,
      anon_sym_BSLASH,
    STATE(1647), 1,
      sym_text_interpolation,
    STATE(2472), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1805), 4,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [62266] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3383), 1,
      anon_sym_BSLASHu,
    STATE(1648), 1,
      sym_text_interpolation,
    ACTIONS(3001), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [62286] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1005), 1,
      aux_sym_while_statement_token1,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3385), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3388), 1,
      aux_sym_else_clause_token1,
    STATE(1649), 1,
      sym_text_interpolation,
    STATE(1653), 1,
      aux_sym_if_statement_repeat1,
    STATE(2021), 1,
      sym_else_if_clause,
    STATE(2022), 1,
      sym_else_clause,
  [62314] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3391), 1,
      anon_sym_SQUOTE,
    ACTIONS(3393), 1,
      aux_sym_string_token2,
    ACTIONS(3396), 1,
      aux_sym_string_content_token1,
    STATE(1926), 1,
      aux_sym_string_content_repeat1,
    STATE(1997), 1,
      sym_string_content,
    STATE(1650), 2,
      sym_text_interpolation,
      aux_sym_string_repeat1,
  [62340] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1562), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3399), 1,
      sym_name,
    ACTIONS(3401), 1,
      anon_sym_LBRACE,
    STATE(894), 1,
      sym__simple_variable,
    STATE(1651), 1,
      sym_text_interpolation,
    STATE(868), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [62366] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2072), 1,
      anon_sym_LPAREN2,
    STATE(867), 1,
      sym_arguments,
    STATE(1652), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [62388] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1015), 1,
      aux_sym_while_statement_token1,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3403), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3406), 1,
      aux_sym_else_clause_token1,
    STATE(1653), 1,
      sym_text_interpolation,
    STATE(1760), 1,
      aux_sym_if_statement_repeat1,
    STATE(2021), 1,
      sym_else_if_clause,
    STATE(2079), 1,
      sym_else_clause,
  [62416] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3341), 1,
      anon_sym_BSLASHu,
    STATE(1654), 1,
      sym_text_interpolation,
    ACTIONS(3343), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [62436] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3345), 1,
      anon_sym_BSLASHu,
    STATE(1655), 1,
      sym_text_interpolation,
    ACTIONS(3347), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [62456] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1801), 1,
      anon_sym_DOLLAR,
    ACTIONS(3409), 1,
      sym_name,
    ACTIONS(3411), 1,
      anon_sym_LBRACE,
    STATE(945), 1,
      sym__simple_variable,
    STATE(1656), 1,
      sym_text_interpolation,
    STATE(957), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [62482] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2110), 1,
      anon_sym_LPAREN2,
    STATE(946), 1,
      sym_arguments,
    STATE(1657), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [62504] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3413), 1,
      anon_sym_BSLASHu,
    STATE(1658), 1,
      sym_text_interpolation,
    ACTIONS(2967), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [62524] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1608), 1,
      anon_sym_BSLASHu,
    STATE(1659), 1,
      sym_text_interpolation,
    ACTIONS(1606), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [62544] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3415), 1,
      sym_name,
    ACTIONS(3417), 1,
      anon_sym_LBRACE,
    ACTIONS(3419), 1,
      anon_sym_DOLLAR,
    STATE(1660), 1,
      sym_text_interpolation,
    STATE(1663), 1,
      sym__simple_variable,
    STATE(1659), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [62570] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1816), 1,
      anon_sym_LPAREN2,
    STATE(732), 1,
      sym_arguments,
    STATE(1661), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [62592] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(558), 1,
      sym_declaration_list,
    STATE(1662), 1,
      sym_text_interpolation,
    STATE(1891), 1,
      sym_base_clause,
    STATE(2331), 1,
      sym_class_interface_clause,
  [62620] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1624), 1,
      anon_sym_BSLASHu,
    STATE(1663), 1,
      sym_text_interpolation,
    ACTIONS(1622), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [62640] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3423), 1,
      sym_name,
    ACTIONS(3425), 1,
      anon_sym_LBRACE,
    ACTIONS(3427), 1,
      anon_sym_DOLLAR,
    STATE(1664), 1,
      sym_text_interpolation,
    STATE(1678), 1,
      sym__simple_variable,
    STATE(1675), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [62666] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1837), 1,
      anon_sym_LPAREN2,
    STATE(816), 1,
      sym_arguments,
    STATE(1665), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [62688] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1558), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3429), 1,
      sym_name,
    ACTIONS(3431), 1,
      anon_sym_LBRACE,
    STATE(748), 1,
      sym__simple_variable,
    STATE(1666), 1,
      sym_text_interpolation,
    STATE(746), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [62714] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3341), 1,
      anon_sym_BSLASHu,
    STATE(1667), 1,
      sym_text_interpolation,
    ACTIONS(3343), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [62734] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3417), 1,
      anon_sym_LBRACE,
    ACTIONS(3419), 1,
      anon_sym_DOLLAR,
    ACTIONS(3433), 1,
      sym_name,
    STATE(1663), 1,
      sym__simple_variable,
    STATE(1668), 1,
      sym_text_interpolation,
    STATE(1659), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [62760] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3425), 1,
      anon_sym_LBRACE,
    ACTIONS(3427), 1,
      anon_sym_DOLLAR,
    ACTIONS(3435), 1,
      sym_name,
    STATE(1669), 1,
      sym_text_interpolation,
    STATE(1678), 1,
      sym__simple_variable,
    STATE(1675), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [62786] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3437), 1,
      sym_name,
    ACTIONS(3439), 1,
      anon_sym_LBRACE,
    ACTIONS(3441), 1,
      anon_sym_DOLLAR,
    STATE(1603), 1,
      sym__simple_variable,
    STATE(1670), 1,
      sym_text_interpolation,
    STATE(1598), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [62812] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3439), 1,
      anon_sym_LBRACE,
    ACTIONS(3441), 1,
      anon_sym_DOLLAR,
    ACTIONS(3443), 1,
      sym_name,
    STATE(1603), 1,
      sym__simple_variable,
    STATE(1671), 1,
      sym_text_interpolation,
    STATE(1598), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [62838] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1688), 1,
      anon_sym_BSLASHu,
    STATE(1672), 1,
      sym_text_interpolation,
    ACTIONS(1686), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [62858] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3329), 1,
      anon_sym_BSLASHu,
    STATE(1673), 1,
      sym_text_interpolation,
    ACTIONS(3331), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [62878] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3447), 1,
      anon_sym_BSLASHu,
    STATE(1674), 1,
      sym_text_interpolation,
    ACTIONS(3445), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [62898] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1608), 1,
      anon_sym_BSLASHu,
    STATE(1675), 1,
      sym_text_interpolation,
    ACTIONS(1606), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [62918] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3361), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3363), 1,
      aux_sym_else_clause_token1,
    ACTIONS(3449), 1,
      aux_sym_if_statement_token2,
    STATE(1676), 1,
      sym_text_interpolation,
    STATE(1808), 1,
      aux_sym_if_statement_repeat2,
    STATE(2225), 1,
      sym_else_if_clause_2,
    STATE(2719), 1,
      sym_else_clause_2,
  [62946] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3261), 1,
      aux_sym_namespace_use_clause_token1,
    ACTIONS(3451), 1,
      aux_sym_use_instead_of_clause_token1,
    STATE(1677), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [62968] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1624), 1,
      anon_sym_BSLASHu,
    STATE(1678), 1,
      sym_text_interpolation,
    ACTIONS(1622), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [62988] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1005), 1,
      aux_sym_while_statement_token1,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3453), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3455), 1,
      aux_sym_else_clause_token1,
    STATE(1679), 1,
      sym_text_interpolation,
    STATE(1705), 1,
      aux_sym_if_statement_repeat1,
    STATE(2021), 1,
      sym_else_if_clause,
    STATE(2022), 1,
      sym_else_clause,
  [63016] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1680), 1,
      sym_text_interpolation,
    ACTIONS(3327), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [63036] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1681), 1,
      sym_text_interpolation,
    ACTIONS(3327), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [63056] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3307), 1,
      anon_sym_BSLASHu,
    STATE(1682), 1,
      sym_text_interpolation,
    ACTIONS(3309), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [63076] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3147), 1,
      anon_sym_DOLLAR,
    ACTIONS(3152), 1,
      anon_sym_PIPE,
    ACTIONS(3457), 1,
      anon_sym_AMP,
    STATE(1510), 1,
      aux_sym_union_type_repeat1,
    STATE(1513), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    STATE(1683), 1,
      sym_text_interpolation,
    STATE(2218), 1,
      aux_sym_intersection_type_repeat1,
  [63104] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3369), 1,
      aux_sym_match_default_expression_token1,
    ACTIONS(3463), 1,
      anon_sym_RBRACE,
    STATE(1564), 1,
      aux_sym_switch_block_repeat1,
    STATE(1684), 1,
      sym_text_interpolation,
    STATE(1888), 2,
      sym_case_statement,
      sym_default_statement,
  [63130] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3337), 1,
      anon_sym_BSLASHu,
    STATE(1685), 1,
      sym_text_interpolation,
    ACTIONS(3339), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [63150] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(472), 1,
      sym_declaration_list,
    STATE(1686), 1,
      sym_text_interpolation,
    STATE(1950), 1,
      sym_base_clause,
    STATE(2496), 1,
      sym_class_interface_clause,
  [63178] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1688), 1,
      anon_sym_BSLASHu,
    STATE(1687), 1,
      sym_text_interpolation,
    ACTIONS(1686), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [63198] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3369), 1,
      aux_sym_match_default_expression_token1,
    ACTIONS(3465), 1,
      aux_sym_switch_block_token1,
    STATE(1564), 1,
      aux_sym_switch_block_repeat1,
    STATE(1688), 1,
      sym_text_interpolation,
    STATE(1888), 2,
      sym_case_statement,
      sym_default_statement,
  [63224] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3467), 1,
      sym_php_tag,
    ACTIONS(3469), 1,
      aux_sym_text_token1,
    ACTIONS(3471), 1,
      aux_sym_text_token2,
    ACTIONS(3473), 1,
      sym__eof,
    STATE(1689), 1,
      sym_text_interpolation,
    STATE(1833), 1,
      aux_sym_text_repeat1,
    STATE(2484), 1,
      sym_text,
  [63252] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1690), 1,
      sym_text_interpolation,
    ACTIONS(3335), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [63272] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(473), 1,
      sym_declaration_list,
    STATE(1691), 1,
      sym_text_interpolation,
    STATE(1855), 1,
      sym_base_clause,
    STATE(2308), 1,
      sym_class_interface_clause,
  [63300] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1692), 1,
      sym_text_interpolation,
    ACTIONS(3335), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [63320] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3361), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3363), 1,
      aux_sym_else_clause_token1,
    ACTIONS(3475), 1,
      aux_sym_if_statement_token2,
    STATE(1676), 1,
      aux_sym_if_statement_repeat2,
    STATE(1693), 1,
      sym_text_interpolation,
    STATE(2225), 1,
      sym_else_if_clause_2,
    STATE(2515), 1,
      sym_else_clause_2,
  [63348] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(521), 1,
      sym_declaration_list,
    STATE(1694), 1,
      sym_text_interpolation,
    STATE(1904), 1,
      sym_base_clause,
    STATE(2440), 1,
      sym_class_interface_clause,
  [63376] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_BSLASHu,
    STATE(1695), 1,
      sym_text_interpolation,
    ACTIONS(1618), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [63396] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_BSLASHu,
    STATE(1696), 1,
      sym_text_interpolation,
    ACTIONS(1618), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [63416] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3351), 1,
      aux_sym_string_token2,
    ACTIONS(3353), 1,
      aux_sym_string_content_token1,
    ACTIONS(3477), 1,
      anon_sym_SQUOTE,
    STATE(1623), 1,
      aux_sym_string_repeat1,
    STATE(1697), 1,
      sym_text_interpolation,
    STATE(1926), 1,
      aux_sym_string_content_repeat1,
    STATE(1997), 1,
      sym_string_content,
  [63444] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(328), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3479), 1,
      sym_name,
    ACTIONS(3481), 1,
      anon_sym_LBRACE,
    STATE(643), 1,
      sym__simple_variable,
    STATE(1698), 1,
      sym_text_interpolation,
    STATE(630), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [63470] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3369), 1,
      aux_sym_match_default_expression_token1,
    ACTIONS(3483), 1,
      anon_sym_RBRACE,
    STATE(1684), 1,
      aux_sym_switch_block_repeat1,
    STATE(1699), 1,
      sym_text_interpolation,
    STATE(1888), 2,
      sym_case_statement,
      sym_default_statement,
  [63496] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3369), 1,
      aux_sym_match_default_expression_token1,
    ACTIONS(3485), 1,
      aux_sym_switch_block_token1,
    STATE(1688), 1,
      aux_sym_switch_block_repeat1,
    STATE(1700), 1,
      sym_text_interpolation,
    STATE(1888), 2,
      sym_case_statement,
      sym_default_statement,
  [63522] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3351), 1,
      aux_sym_string_token2,
    ACTIONS(3353), 1,
      aux_sym_string_content_token1,
    ACTIONS(3487), 1,
      anon_sym_SQUOTE,
    STATE(1625), 1,
      aux_sym_string_repeat1,
    STATE(1701), 1,
      sym_text_interpolation,
    STATE(1926), 1,
      aux_sym_string_content_repeat1,
    STATE(1997), 1,
      sym_string_content,
  [63550] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3355), 1,
      sym_name,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1702), 1,
      sym_text_interpolation,
    STATE(1857), 1,
      sym_named_type,
    STATE(2189), 1,
      sym_type_list,
    STATE(2506), 1,
      sym_namespace_name,
  [63578] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3489), 1,
      sym_name,
    ACTIONS(3491), 1,
      anon_sym_LBRACE,
    STATE(601), 1,
      sym__simple_variable,
    STATE(1703), 1,
      sym_text_interpolation,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [63604] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1590), 1,
      anon_sym_LPAREN2,
    STATE(606), 1,
      sym_arguments,
    STATE(1704), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [63626] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1015), 1,
      aux_sym_while_statement_token1,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3453), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3455), 1,
      aux_sym_else_clause_token1,
    STATE(1705), 1,
      sym_text_interpolation,
    STATE(1760), 1,
      aux_sym_if_statement_repeat1,
    STATE(2021), 1,
      sym_else_if_clause,
    STATE(2079), 1,
      sym_else_clause,
  [63654] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(963), 1,
      sym_declaration_list,
    STATE(1706), 1,
      sym_text_interpolation,
    STATE(1846), 1,
      sym_base_clause,
    STATE(2332), 1,
      sym_class_interface_clause,
  [63682] = 9,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(962), 1,
      sym_declaration_list,
    STATE(1707), 1,
      sym_text_interpolation,
    STATE(1880), 1,
      sym_base_clause,
    STATE(2326), 1,
      sym_class_interface_clause,
  [63710] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1797), 1,
      anon_sym_DOLLAR,
    ACTIONS(3493), 1,
      sym_name,
    ACTIONS(3495), 1,
      anon_sym_LBRACE,
    STATE(842), 1,
      sym__simple_variable,
    STATE(1708), 1,
      sym_text_interpolation,
    STATE(840), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [63736] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3497), 1,
      anon_sym_PIPE,
    STATE(1709), 2,
      sym_text_interpolation,
      aux_sym_union_type_repeat1,
    ACTIONS(3172), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [63755] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1710), 1,
      sym_text_interpolation,
    ACTIONS(3500), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3502), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [63774] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3263), 1,
      aux_sym_catch_clause_token1,
    ACTIONS(3265), 1,
      aux_sym_finally_clause_token1,
    STATE(1550), 1,
      aux_sym_try_statement_repeat1,
    STATE(1711), 1,
      sym_text_interpolation,
    STATE(1777), 2,
      sym_catch_clause,
      sym_finally_clause,
  [63797] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1712), 1,
      sym_text_interpolation,
    ACTIONS(3504), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3506), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [63816] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3508), 1,
      anon_sym_LBRACE,
    ACTIONS(3510), 1,
      anon_sym_COLON,
    STATE(1713), 1,
      sym_text_interpolation,
    STATE(2252), 1,
      sym_enum_declaration_list,
    STATE(2300), 1,
      sym_class_interface_clause,
  [63841] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3512), 1,
      anon_sym_COMMA,
    STATE(1714), 1,
      sym_text_interpolation,
    STATE(1756), 1,
      aux_sym_namespace_use_declaration_repeat1,
    STATE(2160), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [63864] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2456), 1,
      anon_sym_LBRACE,
    ACTIONS(3516), 1,
      anon_sym_EQ,
    STATE(1715), 1,
      sym_text_interpolation,
    STATE(2351), 1,
      sym_property_hook_list,
    ACTIONS(3514), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [63887] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2456), 1,
      anon_sym_LBRACE,
    ACTIONS(3520), 1,
      anon_sym_EQ,
    STATE(1716), 1,
      sym_text_interpolation,
    STATE(2360), 1,
      sym_property_hook_list,
    ACTIONS(3518), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [63910] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3522), 1,
      anon_sym_COMMA,
    STATE(1717), 1,
      sym_text_interpolation,
    STATE(1768), 1,
      aux_sym__const_declaration_repeat1,
    STATE(2240), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [63933] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1718), 1,
      sym_text_interpolation,
    ACTIONS(3524), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
  [63950] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3526), 1,
      anon_sym_PIPE,
    STATE(1719), 1,
      sym_text_interpolation,
    STATE(1834), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3179), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [63971] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3530), 1,
      anon_sym_BSLASH,
    STATE(1720), 1,
      sym_text_interpolation,
    STATE(1797), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(3528), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [63992] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(731), 1,
      anon_sym_AMP,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(3532), 1,
      anon_sym_RPAREN2,
    STATE(1721), 1,
      sym_text_interpolation,
    STATE(2375), 2,
      sym_variable_name,
      sym_by_ref,
  [64015] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3522), 1,
      anon_sym_COMMA,
    STATE(1722), 1,
      sym_text_interpolation,
    STATE(1892), 1,
      aux_sym__const_declaration_repeat1,
    STATE(2129), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [64038] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3522), 1,
      anon_sym_COMMA,
    STATE(523), 1,
      sym__semicolon,
    STATE(1723), 1,
      sym_text_interpolation,
    STATE(1892), 1,
      aux_sym__const_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [64061] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1724), 1,
      sym_text_interpolation,
    ACTIONS(3534), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
  [64078] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3522), 1,
      anon_sym_COMMA,
    STATE(575), 1,
      sym__semicolon,
    STATE(1725), 1,
      sym_text_interpolation,
    STATE(1892), 1,
      aux_sym__const_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [64101] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3522), 1,
      anon_sym_COMMA,
    STATE(564), 1,
      sym__semicolon,
    STATE(1726), 1,
      sym_text_interpolation,
    STATE(1804), 1,
      aux_sym__const_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [64124] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3522), 1,
      anon_sym_COMMA,
    STATE(1722), 1,
      aux_sym__const_declaration_repeat1,
    STATE(1727), 1,
      sym_text_interpolation,
    STATE(2101), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [64147] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1728), 1,
      sym_text_interpolation,
    ACTIONS(3536), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3538), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [64166] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3540), 1,
      sym_php_tag,
    ACTIONS(3542), 1,
      aux_sym_text_token1,
    ACTIONS(3545), 1,
      aux_sym_text_token2,
    ACTIONS(3548), 1,
      sym__eof,
    STATE(1729), 2,
      sym_text_interpolation,
      aux_sym_text_repeat1,
  [64189] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(228), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1405), 1,
      sym_compound_statement,
    STATE(1406), 1,
      sym__semicolon,
    STATE(1730), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [64212] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1731), 1,
      sym_text_interpolation,
    ACTIONS(3550), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3552), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [64231] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3556), 1,
      anon_sym_COMMA,
    STATE(1732), 2,
      sym_text_interpolation,
      aux_sym_base_clause_repeat1,
    ACTIONS(3554), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [64250] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1733), 1,
      sym_text_interpolation,
    ACTIONS(3160), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [64267] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3522), 1,
      anon_sym_COMMA,
    STATE(565), 1,
      sym__semicolon,
    STATE(1734), 1,
      sym_text_interpolation,
    STATE(1892), 1,
      aux_sym__const_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [64290] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1735), 1,
      sym_text_interpolation,
    ACTIONS(3559), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3561), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [64309] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1963), 1,
      anon_sym_LPAREN2,
    STATE(1736), 1,
      sym_text_interpolation,
    ACTIONS(1965), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [64328] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2456), 1,
      anon_sym_LBRACE,
    ACTIONS(3565), 1,
      anon_sym_EQ,
    STATE(1737), 1,
      sym_text_interpolation,
    STATE(2387), 1,
      sym_property_hook_list,
    ACTIONS(3563), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [64351] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1738), 1,
      sym_text_interpolation,
    ACTIONS(3567), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3569), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [64370] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(959), 1,
      anon_sym_COLON,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(1739), 1,
      sym_text_interpolation,
    STATE(1826), 1,
      sym_formal_parameters,
    STATE(2359), 1,
      sym_reference_modifier,
  [64395] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(228), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3571), 1,
      sym_name,
    ACTIONS(3573), 1,
      anon_sym_BSLASH,
    STATE(533), 1,
      sym_compound_statement,
    STATE(1740), 1,
      sym_text_interpolation,
    STATE(1787), 1,
      sym_namespace_name,
  [64420] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1741), 1,
      sym_text_interpolation,
    ACTIONS(1042), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(1040), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [64439] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1742), 1,
      sym_text_interpolation,
    ACTIONS(3575), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3577), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [64458] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(228), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1407), 1,
      sym__semicolon,
    STATE(1421), 1,
      sym_compound_statement,
    STATE(1743), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [64481] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(959), 1,
      anon_sym_COLON,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(1744), 1,
      sym_text_interpolation,
    STATE(2266), 1,
      sym_formal_parameters,
    STATE(2456), 1,
      sym_reference_modifier,
  [64506] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1745), 1,
      sym_text_interpolation,
    ACTIONS(1058), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(1056), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [64525] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1746), 1,
      sym_text_interpolation,
    ACTIONS(3034), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [64542] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3508), 1,
      anon_sym_LBRACE,
    ACTIONS(3579), 1,
      anon_sym_COLON,
    STATE(1747), 1,
      sym_text_interpolation,
    STATE(2045), 1,
      sym_enum_declaration_list,
    STATE(2377), 1,
      sym_class_interface_clause,
  [64567] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1748), 1,
      sym_text_interpolation,
    ACTIONS(1158), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(1156), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [64586] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3540), 1,
      sym_php_tag,
    ACTIONS(3548), 1,
      ts_builtin_sym_end,
    ACTIONS(3581), 1,
      aux_sym_text_token1,
    ACTIONS(3584), 1,
      aux_sym_text_token2,
    STATE(1749), 2,
      sym_text_interpolation,
      aux_sym_text_repeat1,
  [64609] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(603), 1,
      anon_sym_BSLASH,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3355), 1,
      sym_name,
    STATE(1497), 1,
      sym_qualified_name,
    STATE(1750), 1,
      sym_text_interpolation,
    STATE(2222), 1,
      sym_named_type,
    STATE(2506), 1,
      sym_namespace_name,
  [64634] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1751), 1,
      sym_text_interpolation,
    ACTIONS(3587), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3589), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [64653] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3591), 1,
      anon_sym_COMMA,
    STATE(1752), 1,
      sym_text_interpolation,
    STATE(1862), 1,
      aux_sym_function_static_declaration_repeat1,
    STATE(1977), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [64676] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3530), 1,
      anon_sym_BSLASH,
    STATE(1753), 1,
      sym_text_interpolation,
    STATE(1769), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(3528), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [64697] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3593), 1,
      anon_sym_COMMA,
    STATE(1754), 1,
      sym_text_interpolation,
    STATE(1864), 1,
      aux_sym_global_declaration_repeat1,
    STATE(1978), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [64720] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1755), 1,
      sym_text_interpolation,
    ACTIONS(3595), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3597), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [64739] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3512), 1,
      anon_sym_COMMA,
    STATE(1756), 1,
      sym_text_interpolation,
    STATE(1867), 1,
      aux_sym_namespace_use_declaration_repeat1,
    STATE(1984), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [64762] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(228), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1408), 1,
      sym_compound_statement,
    STATE(1409), 1,
      sym__semicolon,
    STATE(1757), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [64785] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1758), 1,
      sym_text_interpolation,
    ACTIONS(3599), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
  [64802] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2456), 1,
      anon_sym_LBRACE,
    ACTIONS(3603), 1,
      anon_sym_EQ,
    STATE(1759), 1,
      sym_text_interpolation,
    STATE(2488), 1,
      sym_property_hook_list,
    ACTIONS(3601), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [64825] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1033), 1,
      aux_sym_while_statement_token1,
    ACTIONS(1035), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3605), 1,
      aux_sym_else_if_clause_token1,
    STATE(2021), 1,
      sym_else_if_clause,
    STATE(1760), 2,
      sym_text_interpolation,
      aux_sym_if_statement_repeat1,
  [64848] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3610), 1,
      anon_sym_COLON,
    STATE(1761), 1,
      sym_text_interpolation,
    STATE(1987), 1,
      sym__return_type,
    ACTIONS(3608), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [64869] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1762), 1,
      sym_text_interpolation,
    ACTIONS(3164), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [64886] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3522), 1,
      anon_sym_COMMA,
    STATE(1763), 1,
      sym_text_interpolation,
    STATE(1823), 1,
      aux_sym__const_declaration_repeat1,
    STATE(1988), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [64909] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1764), 1,
      sym_text_interpolation,
    ACTIONS(3156), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [64926] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3612), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(3614), 1,
      anon_sym_LBRACE,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    STATE(1765), 1,
      sym_text_interpolation,
    STATE(2237), 1,
      sym_anonymous_function_use_clause,
    STATE(2566), 1,
      sym__return_type,
  [64951] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3618), 1,
      anon_sym_PIPE,
    STATE(1709), 1,
      aux_sym_union_type_repeat1,
    STATE(1766), 1,
      sym_text_interpolation,
    ACTIONS(3187), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [64972] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3526), 1,
      anon_sym_PIPE,
    STATE(1767), 1,
      sym_text_interpolation,
    STATE(1834), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3201), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [64993] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3522), 1,
      anon_sym_COMMA,
    STATE(1768), 1,
      sym_text_interpolation,
    STATE(1892), 1,
      aux_sym__const_declaration_repeat1,
    STATE(1998), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [65016] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3530), 1,
      anon_sym_BSLASH,
    STATE(1769), 1,
      sym_text_interpolation,
    STATE(1797), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(3620), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [65037] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(13), 1,
      aux_sym_text_token1,
    ACTIONS(15), 1,
      aux_sym_text_token2,
    ACTIONS(3622), 1,
      ts_builtin_sym_end,
    ACTIONS(3624), 1,
      sym_php_tag,
    STATE(1749), 1,
      aux_sym_text_repeat1,
    STATE(1770), 1,
      sym_text_interpolation,
  [65062] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3626), 1,
      anon_sym_PIPE,
    STATE(1771), 2,
      sym_text_interpolation,
      aux_sym_union_type_repeat1,
    ACTIONS(3172), 3,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [65081] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3629), 1,
      anon_sym_PIPE,
    STATE(1772), 2,
      sym_text_interpolation,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3197), 3,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [65100] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3612), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(3632), 1,
      anon_sym_LBRACE,
    STATE(1773), 1,
      sym_text_interpolation,
    STATE(2220), 1,
      sym_anonymous_function_use_clause,
    STATE(2513), 1,
      sym__return_type,
  [65125] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3634), 1,
      anon_sym_PIPE,
    STATE(1772), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    STATE(1774), 1,
      sym_text_interpolation,
    ACTIONS(3179), 3,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [65146] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3591), 1,
      anon_sym_COMMA,
    STATE(468), 1,
      sym__semicolon,
    STATE(1775), 1,
      sym_text_interpolation,
    STATE(1799), 1,
      aux_sym_function_static_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [65169] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(731), 1,
      anon_sym_AMP,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(3636), 1,
      anon_sym_RPAREN2,
    STATE(1776), 1,
      sym_text_interpolation,
    STATE(2375), 2,
      sym_variable_name,
      sym_by_ref,
  [65192] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1050), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1777), 1,
      sym_text_interpolation,
    ACTIONS(1048), 4,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [65211] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3522), 1,
      anon_sym_COMMA,
    STATE(512), 1,
      sym__semicolon,
    STATE(1734), 1,
      aux_sym__const_declaration_repeat1,
    STATE(1778), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [65234] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1779), 1,
      sym_text_interpolation,
    ACTIONS(1957), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [65251] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1780), 1,
      sym_text_interpolation,
    ACTIONS(3638), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3640), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [65270] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3255), 1,
      anon_sym_AMP,
    STATE(1781), 1,
      sym_text_interpolation,
    STATE(1831), 1,
      aux_sym_intersection_type_repeat1,
    ACTIONS(3193), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [65291] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2456), 1,
      anon_sym_LBRACE,
    ACTIONS(3644), 1,
      anon_sym_EQ,
    STATE(1782), 1,
      sym_text_interpolation,
    STATE(2451), 1,
      sym_property_hook_list,
    ACTIONS(3642), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [65314] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1971), 1,
      anon_sym_LPAREN2,
    ACTIONS(3646), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3648), 1,
      anon_sym_DASH_GT,
    ACTIONS(3650), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(3652), 1,
      anon_sym_LBRACK,
    STATE(1783), 1,
      sym_text_interpolation,
  [65339] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3593), 1,
      anon_sym_COMMA,
    STATE(502), 1,
      sym__semicolon,
    STATE(1784), 1,
      sym_text_interpolation,
    STATE(1809), 1,
      aux_sym_global_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [65362] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3656), 1,
      anon_sym_COMMA,
    STATE(1785), 2,
      sym_text_interpolation,
      aux_sym_property_declaration_repeat1,
    ACTIONS(3654), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [65381] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3530), 1,
      anon_sym_BSLASH,
    STATE(1720), 1,
      aux_sym_namespace_name_repeat1,
    STATE(1786), 1,
      sym_text_interpolation,
    ACTIONS(3659), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [65402] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(228), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(503), 1,
      sym_compound_statement,
    STATE(505), 1,
      sym__semicolon,
    STATE(1787), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [65425] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3612), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(3661), 1,
      anon_sym_LBRACE,
    STATE(1788), 1,
      sym_text_interpolation,
    STATE(2264), 1,
      sym_anonymous_function_use_clause,
    STATE(2623), 1,
      sym__return_type,
  [65450] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3663), 1,
      anon_sym_LBRACE,
    ACTIONS(3665), 1,
      anon_sym_COLON,
    STATE(471), 1,
      sym_enum_declaration_list,
    STATE(1789), 1,
      sym_text_interpolation,
    STATE(2500), 1,
      sym_class_interface_clause,
  [65475] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1790), 1,
      sym_text_interpolation,
    ACTIONS(3667), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3669), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [65494] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1031), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1791), 1,
      sym_text_interpolation,
    ACTIONS(1029), 4,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [65513] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3612), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(3671), 1,
      anon_sym_LBRACE,
    STATE(1792), 1,
      sym_text_interpolation,
    STATE(2219), 1,
      sym_anonymous_function_use_clause,
    STATE(2508), 1,
      sym__return_type,
  [65538] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1793), 1,
      sym_text_interpolation,
    ACTIONS(3673), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3675), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [65557] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(242), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3677), 1,
      sym_name,
    STATE(1794), 1,
      sym_text_interpolation,
    STATE(2116), 1,
      sym_final_modifier,
    STATE(2710), 1,
      sym_reference_modifier,
  [65582] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3522), 1,
      anon_sym_COMMA,
    STATE(565), 1,
      sym__semicolon,
    STATE(1723), 1,
      aux_sym__const_declaration_repeat1,
    STATE(1795), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [65605] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3512), 1,
      anon_sym_COMMA,
    STATE(508), 1,
      sym__semicolon,
    STATE(1796), 1,
      sym_text_interpolation,
    STATE(1807), 1,
      aux_sym_namespace_use_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [65628] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3681), 1,
      anon_sym_BSLASH,
    STATE(1797), 2,
      sym_text_interpolation,
      aux_sym_namespace_name_repeat1,
    ACTIONS(3679), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [65647] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1798), 1,
      sym_text_interpolation,
    ACTIONS(3684), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3686), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [65666] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3591), 1,
      anon_sym_COMMA,
    STATE(560), 1,
      sym__semicolon,
    STATE(1799), 1,
      sym_text_interpolation,
    STATE(1862), 1,
      aux_sym_function_static_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [65689] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2456), 1,
      anon_sym_LBRACE,
    ACTIONS(3690), 1,
      anon_sym_EQ,
    STATE(1800), 1,
      sym_text_interpolation,
    STATE(2501), 1,
      sym_property_hook_list,
    ACTIONS(3688), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [65712] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2456), 1,
      anon_sym_LBRACE,
    ACTIONS(3694), 1,
      anon_sym_EQ,
    STATE(1801), 1,
      sym_text_interpolation,
    STATE(2476), 1,
      sym_property_hook_list,
    ACTIONS(3692), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [65735] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(3696), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(1802), 1,
      sym_text_interpolation,
    STATE(1961), 1,
      sym_variable_name,
    STATE(2293), 1,
      sym_reference_modifier,
  [65760] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1670), 1,
      anon_sym_LPAREN2,
    STATE(1803), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [65779] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3522), 1,
      anon_sym_COMMA,
    STATE(474), 1,
      sym__semicolon,
    STATE(1804), 1,
      sym_text_interpolation,
    STATE(1892), 1,
      aux_sym__const_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [65802] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1805), 1,
      sym_text_interpolation,
    ACTIONS(3698), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3700), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [65821] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3612), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(3702), 1,
      anon_sym_LBRACE,
    STATE(1806), 1,
      sym_text_interpolation,
    STATE(2027), 1,
      sym_anonymous_function_use_clause,
    STATE(2657), 1,
      sym__return_type,
  [65846] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3512), 1,
      anon_sym_COMMA,
    STATE(562), 1,
      sym__semicolon,
    STATE(1807), 1,
      sym_text_interpolation,
    STATE(1867), 1,
      aux_sym_namespace_use_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [65869] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3704), 1,
      aux_sym_if_statement_token2,
    ACTIONS(3706), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3709), 1,
      aux_sym_else_clause_token1,
    STATE(2225), 1,
      sym_else_if_clause_2,
    STATE(1808), 2,
      sym_text_interpolation,
      aux_sym_if_statement_repeat2,
  [65892] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3593), 1,
      anon_sym_COMMA,
    STATE(561), 1,
      sym__semicolon,
    STATE(1809), 1,
      sym_text_interpolation,
    STATE(1864), 1,
      aux_sym_global_declaration_repeat1,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [65915] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3713), 1,
      anon_sym_EQ,
    STATE(1810), 1,
      sym_text_interpolation,
    ACTIONS(3711), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [65934] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1054), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1811), 1,
      sym_text_interpolation,
    ACTIONS(1052), 4,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [65953] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3612), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(3715), 1,
      anon_sym_LBRACE,
    STATE(1812), 1,
      sym_text_interpolation,
    STATE(2229), 1,
      sym_anonymous_function_use_clause,
    STATE(2536), 1,
      sym__return_type,
  [65978] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1813), 1,
      sym_text_interpolation,
    ACTIONS(3717), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3719), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [65997] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(376), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3571), 1,
      sym_name,
    ACTIONS(3573), 1,
      anon_sym_BSLASH,
    STATE(1814), 1,
      sym_text_interpolation,
    STATE(1843), 1,
      sym_namespace_name,
    STATE(2226), 1,
      sym_compound_statement,
  [66022] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1815), 1,
      sym_text_interpolation,
    ACTIONS(3721), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3723), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [66041] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1816), 1,
      sym_text_interpolation,
    ACTIONS(3725), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3727), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [66060] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1817), 1,
      sym_text_interpolation,
    ACTIONS(3729), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3731), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [66079] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3522), 1,
      anon_sym_COMMA,
    STATE(1818), 1,
      sym_text_interpolation,
    STATE(1842), 1,
      aux_sym__const_declaration_repeat1,
    STATE(1998), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [66102] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1819), 1,
      sym_text_interpolation,
    ACTIONS(3733), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
  [66119] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3610), 1,
      anon_sym_COLON,
    STATE(1820), 1,
      sym_text_interpolation,
    STATE(2055), 1,
      sym__return_type,
    ACTIONS(3735), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [66140] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3526), 1,
      anon_sym_PIPE,
    STATE(1719), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    STATE(1821), 1,
      sym_text_interpolation,
    ACTIONS(3209), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [66161] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(228), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1410), 1,
      sym_compound_statement,
    STATE(1420), 1,
      sym__semicolon,
    STATE(1822), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [66184] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3522), 1,
      anon_sym_COMMA,
    STATE(1823), 1,
      sym_text_interpolation,
    STATE(1892), 1,
      aux_sym__const_declaration_repeat1,
    STATE(2057), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [66207] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1824), 1,
      sym_text_interpolation,
    ACTIONS(3737), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3739), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [66226] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(3741), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(1825), 1,
      sym_text_interpolation,
    STATE(2243), 1,
      sym_reference_modifier,
    STATE(2244), 1,
      sym_variable_name,
  [66251] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3612), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(3743), 1,
      anon_sym_LBRACE,
    STATE(1826), 1,
      sym_text_interpolation,
    STATE(2058), 1,
      sym_anonymous_function_use_clause,
    STATE(2584), 1,
      sym__return_type,
  [66276] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3745), 1,
      anon_sym_PIPE,
    STATE(1771), 1,
      aux_sym_union_type_repeat1,
    STATE(1827), 1,
      sym_text_interpolation,
    ACTIONS(3187), 3,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [66297] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3634), 1,
      anon_sym_PIPE,
    STATE(1772), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    STATE(1828), 1,
      sym_text_interpolation,
    ACTIONS(3201), 3,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [66318] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3522), 1,
      anon_sym_COMMA,
    STATE(522), 1,
      sym__semicolon,
    STATE(1725), 1,
      aux_sym__const_declaration_repeat1,
    STATE(1829), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [66341] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3634), 1,
      anon_sym_PIPE,
    STATE(1774), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    STATE(1830), 1,
      sym_text_interpolation,
    ACTIONS(3209), 3,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [66362] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3747), 1,
      anon_sym_AMP,
    STATE(1831), 2,
      sym_text_interpolation,
      aux_sym_intersection_type_repeat1,
    ACTIONS(3213), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [66381] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2456), 1,
      anon_sym_LBRACE,
    ACTIONS(3752), 1,
      anon_sym_EQ,
    STATE(1832), 1,
      sym_text_interpolation,
    STATE(2450), 1,
      sym_property_hook_list,
    ACTIONS(3750), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [66404] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3469), 1,
      aux_sym_text_token1,
    ACTIONS(3471), 1,
      aux_sym_text_token2,
    ACTIONS(3622), 1,
      sym__eof,
    ACTIONS(3624), 1,
      sym_php_tag,
    STATE(1729), 1,
      aux_sym_text_repeat1,
    STATE(1833), 1,
      sym_text_interpolation,
  [66429] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3754), 1,
      anon_sym_PIPE,
    STATE(1834), 2,
      sym_text_interpolation,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3197), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [66448] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3757), 1,
      sym_name,
    STATE(1835), 1,
      sym_text_interpolation,
    STATE(2233), 1,
      sym_visibility_modifier,
    ACTIONS(3759), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
  [66469] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3612), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(3761), 1,
      anon_sym_LBRACE,
    STATE(1836), 1,
      sym_text_interpolation,
    STATE(2248), 1,
      sym_anonymous_function_use_clause,
    STATE(2592), 1,
      sym__return_type,
  [66494] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3763), 1,
      aux_sym_catch_clause_token1,
    ACTIONS(3765), 1,
      aux_sym_finally_clause_token1,
    STATE(444), 1,
      aux_sym_try_statement_repeat1,
    STATE(1837), 1,
      sym_text_interpolation,
    STATE(454), 2,
      sym_catch_clause,
      sym_finally_clause,
  [66517] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1046), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1838), 1,
      sym_text_interpolation,
    ACTIONS(1044), 4,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [66536] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3591), 1,
      anon_sym_COMMA,
    STATE(1752), 1,
      aux_sym_function_static_declaration_repeat1,
    STATE(1839), 1,
      sym_text_interpolation,
    STATE(2124), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [66559] = 8,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3663), 1,
      anon_sym_LBRACE,
    ACTIONS(3767), 1,
      anon_sym_COLON,
    STATE(541), 1,
      sym_enum_declaration_list,
    STATE(1840), 1,
      sym_text_interpolation,
    STATE(2324), 1,
      sym_class_interface_clause,
  [66584] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3593), 1,
      anon_sym_COMMA,
    STATE(1754), 1,
      aux_sym_global_declaration_repeat1,
    STATE(1841), 1,
      sym_text_interpolation,
    STATE(2135), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [66607] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3522), 1,
      anon_sym_COMMA,
    STATE(1842), 1,
      sym_text_interpolation,
    STATE(1892), 1,
      aux_sym__const_declaration_repeat1,
    STATE(2103), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [66630] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(376), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1843), 1,
      sym_text_interpolation,
    STATE(2147), 1,
      sym_compound_statement,
    STATE(2148), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [66653] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1844), 1,
      sym_text_interpolation,
    ACTIONS(3769), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3771), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [66672] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3773), 1,
      anon_sym_DQUOTE2,
    ACTIONS(3775), 1,
      sym__new_line,
    ACTIONS(3777), 1,
      sym_heredoc_end,
    STATE(1845), 1,
      sym_text_interpolation,
    STATE(2347), 1,
      sym_heredoc_body,
  [66694] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(1003), 1,
      sym_declaration_list,
    STATE(1846), 1,
      sym_text_interpolation,
    STATE(2353), 1,
      sym_class_interface_clause,
  [66716] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3779), 1,
      anon_sym_EQ,
    STATE(1445), 1,
      sym__semicolon,
    STATE(1847), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [66736] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3781), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3783), 1,
      anon_sym_DASH_GT,
    ACTIONS(3785), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(3787), 1,
      anon_sym_LBRACK,
    STATE(1848), 1,
      sym_text_interpolation,
  [66758] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1849), 1,
      sym_text_interpolation,
    ACTIONS(3205), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [66774] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3791), 1,
      anon_sym_PIPE,
    ACTIONS(3789), 2,
      anon_sym_RPAREN2,
      anon_sym_DOLLAR,
    STATE(1850), 2,
      sym_text_interpolation,
      aux_sym_type_list_repeat1,
  [66792] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(1851), 1,
      sym_text_interpolation,
    STATE(2004), 1,
      sym_declaration_list,
    STATE(2342), 1,
      sym_class_interface_clause,
  [66814] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(968), 1,
      sym_declaration_list,
    STATE(1852), 1,
      sym_text_interpolation,
    STATE(2457), 1,
      sym_class_interface_clause,
  [66836] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(1773), 1,
      sym_formal_parameters,
    STATE(1853), 1,
      sym_text_interpolation,
    STATE(2363), 1,
      sym_reference_modifier,
  [66858] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(1854), 1,
      sym_text_interpolation,
    STATE(2186), 1,
      sym_formal_parameters,
    STATE(2366), 1,
      sym_reference_modifier,
  [66880] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(524), 1,
      sym_declaration_list,
    STATE(1855), 1,
      sym_text_interpolation,
    STATE(2445), 1,
      sym_class_interface_clause,
  [66902] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(778), 1,
      sym_declaration_list,
    STATE(1856), 1,
      sym_text_interpolation,
    STATE(2428), 1,
      sym_class_interface_clause,
  [66924] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3796), 1,
      anon_sym_PIPE,
    STATE(1857), 1,
      sym_text_interpolation,
    STATE(1954), 1,
      aux_sym_type_list_repeat1,
    ACTIONS(3794), 2,
      anon_sym_RPAREN2,
      anon_sym_DOLLAR,
  [66944] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    STATE(1858), 1,
      sym_text_interpolation,
    STATE(2044), 1,
      sym_declaration_list,
    STATE(2376), 1,
      sym_base_clause,
  [66966] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1859), 1,
      sym_text_interpolation,
    ACTIONS(3798), 4,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
  [66982] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1860), 1,
      sym_text_interpolation,
    ACTIONS(3679), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      anon_sym_LBRACE,
  [66998] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3800), 1,
      anon_sym_COMMA,
    ACTIONS(3554), 2,
      anon_sym_LBRACE,
      aux_sym_class_interface_clause_token1,
    STATE(1861), 2,
      sym_text_interpolation,
      aux_sym_base_clause_repeat1,
  [67016] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3805), 1,
      anon_sym_COMMA,
    ACTIONS(3803), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1862), 2,
      sym_text_interpolation,
      aux_sym_function_static_declaration_repeat1,
  [67034] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3810), 1,
      aux_sym_namespace_use_clause_token1,
    STATE(1863), 1,
      sym_text_interpolation,
    ACTIONS(3808), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [67052] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3814), 1,
      anon_sym_COMMA,
    ACTIONS(3812), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1864), 2,
      sym_text_interpolation,
      aux_sym_global_declaration_repeat1,
  [67070] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(777), 1,
      sym_declaration_list,
    STATE(1865), 1,
      sym_text_interpolation,
    STATE(2430), 1,
      sym_class_interface_clause,
  [67092] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1866), 1,
      sym_text_interpolation,
    STATE(2525), 1,
      sym_declare_directive,
    ACTIONS(3817), 3,
      anon_sym_ticks,
      anon_sym_encoding,
      anon_sym_strict_types,
  [67110] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3821), 1,
      anon_sym_COMMA,
    ACTIONS(3819), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1867), 2,
      sym_text_interpolation,
      aux_sym_namespace_use_declaration_repeat1,
  [67128] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1590), 1,
      anon_sym_LPAREN2,
    STATE(1868), 1,
      sym_text_interpolation,
    STATE(2447), 1,
      sym_arguments,
    ACTIONS(3824), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [67148] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3781), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3826), 1,
      anon_sym_DASH_GT,
    ACTIONS(3828), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(3830), 1,
      anon_sym_LBRACK,
    STATE(1869), 1,
      sym_text_interpolation,
  [67170] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1870), 1,
      sym_text_interpolation,
    STATE(2552), 1,
      sym_declare_directive,
    ACTIONS(3817), 3,
      anon_sym_ticks,
      anon_sym_encoding,
      anon_sym_strict_types,
  [67188] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1871), 1,
      sym_text_interpolation,
    ACTIONS(3733), 4,
      aux_sym_namespace_use_declaration_token1,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
  [67204] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3781), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3787), 1,
      anon_sym_LBRACK,
    ACTIONS(3832), 1,
      anon_sym_DASH_GT,
    ACTIONS(3834), 1,
      anon_sym_QMARK_DASH_GT,
    STATE(1872), 1,
      sym_text_interpolation,
  [67226] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3508), 1,
      anon_sym_LBRACE,
    STATE(1873), 1,
      sym_text_interpolation,
    STATE(2125), 1,
      sym_enum_declaration_list,
    STATE(2435), 1,
      sym_class_interface_clause,
  [67248] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(731), 1,
      anon_sym_AMP,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(1874), 1,
      sym_text_interpolation,
    STATE(1782), 2,
      sym_variable_name,
      sym_by_ref,
  [67268] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1875), 1,
      sym_text_interpolation,
    ACTIONS(3554), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [67284] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3836), 1,
      anon_sym_AMP,
    STATE(1876), 1,
      sym_text_interpolation,
    STATE(1887), 1,
      aux_sym_intersection_type_repeat1,
    ACTIONS(3193), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [67304] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(731), 1,
      anon_sym_AMP,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(1877), 1,
      sym_text_interpolation,
    STATE(1759), 2,
      sym_variable_name,
      sym_by_ref,
  [67324] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(1878), 1,
      sym_text_interpolation,
    STATE(2127), 1,
      sym_declaration_list,
    STATE(2436), 1,
      sym_class_interface_clause,
  [67346] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(731), 1,
      anon_sym_AMP,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(1879), 1,
      sym_text_interpolation,
    STATE(1975), 2,
      sym_variable_name,
      sym_by_ref,
  [67366] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(1002), 1,
      sym_declaration_list,
    STATE(1880), 1,
      sym_text_interpolation,
    STATE(2357), 1,
      sym_class_interface_clause,
  [67388] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(1826), 1,
      sym_formal_parameters,
    STATE(1881), 1,
      sym_text_interpolation,
    STATE(2359), 1,
      sym_reference_modifier,
  [67410] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1882), 1,
      sym_text_interpolation,
    ACTIONS(3172), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [67426] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(645), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(603), 1,
      sym__simple_variable,
    STATE(1883), 1,
      sym_text_interpolation,
    STATE(597), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [67446] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(328), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1784), 1,
      sym__simple_variable,
    STATE(1884), 1,
      sym_text_interpolation,
    STATE(630), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [67466] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1797), 1,
      anon_sym_DOLLAR,
    STATE(818), 1,
      sym__simple_variable,
    STATE(1885), 1,
      sym_text_interpolation,
    STATE(840), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [67486] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3775), 1,
      sym__new_line,
    ACTIONS(3839), 1,
      anon_sym_DQUOTE2,
    ACTIONS(3841), 1,
      sym_heredoc_end,
    STATE(1886), 1,
      sym_text_interpolation,
    STATE(2367), 1,
      sym_heredoc_body,
  [67508] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3843), 1,
      anon_sym_AMP,
    ACTIONS(3213), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
    STATE(1887), 2,
      sym_text_interpolation,
      aux_sym_intersection_type_repeat1,
  [67526] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1888), 1,
      sym_text_interpolation,
    ACTIONS(3846), 4,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_match_default_expression_token1,
      aux_sym_switch_block_token1,
  [67542] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(962), 1,
      sym_declaration_list,
    STATE(1889), 1,
      sym_text_interpolation,
    STATE(2326), 1,
      sym_class_interface_clause,
  [67564] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(731), 1,
      anon_sym_AMP,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(1890), 1,
      sym_text_interpolation,
    STATE(1737), 2,
      sym_variable_name,
      sym_by_ref,
  [67584] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(571), 1,
      sym_declaration_list,
    STATE(1891), 1,
      sym_text_interpolation,
    STATE(2497), 1,
      sym_class_interface_clause,
  [67606] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3850), 1,
      anon_sym_COMMA,
    ACTIONS(3848), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1892), 2,
      sym_text_interpolation,
      aux_sym__const_declaration_repeat1,
  [67624] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2500), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2823), 1,
      aux_sym__arrow_function_header_token1,
    ACTIONS(3853), 1,
      aux_sym__namespace_use_type_token1,
    STATE(1893), 1,
      sym_text_interpolation,
    STATE(2379), 1,
      sym_static_modifier,
  [67646] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(783), 1,
      sym_declaration_list,
    STATE(1894), 1,
      sym_text_interpolation,
    STATE(2442), 1,
      sym_class_interface_clause,
  [67668] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(328), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1895), 1,
      sym_text_interpolation,
    STATE(2232), 1,
      sym__simple_variable,
    STATE(630), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [67688] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3855), 1,
      anon_sym_EQ,
    STATE(1440), 1,
      sym__semicolon,
    STATE(1896), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [67708] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    STATE(1897), 1,
      sym_text_interpolation,
    STATE(2251), 1,
      sym_declaration_list,
    STATE(2495), 1,
      sym_base_clause,
  [67730] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3859), 1,
      sym_nowdoc_string,
    ACTIONS(3857), 2,
      sym_heredoc_end,
      sym__new_line,
    STATE(1898), 2,
      sym_text_interpolation,
      aux_sym_nowdoc_body_repeat1,
  [67748] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(767), 1,
      sym_declaration_list,
    STATE(1899), 1,
      sym_text_interpolation,
    STATE(2395), 1,
      sym_class_interface_clause,
  [67770] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(1765), 1,
      sym_formal_parameters,
    STATE(1900), 1,
      sym_text_interpolation,
    STATE(2321), 1,
      sym_reference_modifier,
  [67792] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(1901), 1,
      sym_text_interpolation,
    STATE(1983), 1,
      sym_formal_parameters,
    STATE(2434), 1,
      sym_reference_modifier,
  [67814] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3781), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3862), 1,
      anon_sym_DASH_GT,
    ACTIONS(3864), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(3866), 1,
      anon_sym_LBRACK,
    STATE(1902), 1,
      sym_text_interpolation,
  [67836] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3868), 1,
      anon_sym_LPAREN2,
    STATE(1903), 1,
      sym_text_interpolation,
    STATE(2278), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [67856] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(574), 1,
      sym_declaration_list,
    STATE(1904), 1,
      sym_text_interpolation,
    STATE(2354), 1,
      sym_class_interface_clause,
  [67878] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3872), 1,
      aux_sym_string_content_token1,
    ACTIONS(3870), 2,
      anon_sym_SQUOTE,
      aux_sym_string_token2,
    STATE(1905), 2,
      sym_text_interpolation,
      aux_sym_string_content_repeat1,
  [67896] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1906), 1,
      sym_text_interpolation,
    ACTIONS(3875), 4,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
  [67912] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3879), 1,
      sym_nowdoc_string,
    STATE(1898), 1,
      aux_sym_nowdoc_body_repeat1,
    STATE(1907), 1,
      sym_text_interpolation,
    ACTIONS(3877), 2,
      sym_heredoc_end,
      sym__new_line,
  [67932] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1908), 1,
      sym_text_interpolation,
    ACTIONS(3197), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [67948] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(539), 1,
      sym_declaration_list,
    STATE(1909), 1,
      sym_text_interpolation,
    STATE(2487), 1,
      sym_base_clause,
  [67970] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3781), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3881), 1,
      anon_sym_DASH_GT,
    ACTIONS(3883), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(3885), 1,
      anon_sym_LBRACK,
    STATE(1910), 1,
      sym_text_interpolation,
  [67992] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3889), 1,
      aux_sym_namespace_use_clause_token1,
    STATE(1911), 1,
      sym_text_interpolation,
    ACTIONS(3887), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [68010] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(768), 1,
      sym_declaration_list,
    STATE(1912), 1,
      sym_text_interpolation,
    STATE(2397), 1,
      sym_class_interface_clause,
  [68032] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1913), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [68048] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1914), 1,
      sym_text_interpolation,
    ACTIONS(1812), 4,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [68064] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3781), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3891), 1,
      anon_sym_DASH_GT,
    ACTIONS(3893), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(3895), 1,
      anon_sym_LBRACK,
    STATE(1915), 1,
      sym_text_interpolation,
  [68086] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1916), 1,
      sym_text_interpolation,
    STATE(2665), 1,
      sym_declare_directive,
    ACTIONS(3817), 3,
      anon_sym_ticks,
      anon_sym_encoding,
      anon_sym_strict_types,
  [68104] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1917), 1,
      sym_text_interpolation,
    ACTIONS(1672), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [68120] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3775), 1,
      sym__new_line,
    ACTIONS(3897), 1,
      anon_sym_DQUOTE2,
    ACTIONS(3899), 1,
      sym_heredoc_end,
    STATE(1918), 1,
      sym_text_interpolation,
    STATE(2330), 1,
      sym_heredoc_body,
  [68142] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3901), 1,
      anon_sym_COMMA,
    STATE(1919), 1,
      sym_text_interpolation,
    STATE(1925), 1,
      aux_sym_base_clause_repeat1,
    ACTIONS(3903), 2,
      anon_sym_LBRACE,
      aux_sym_class_interface_clause_token1,
  [68162] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3663), 1,
      anon_sym_LBRACE,
    STATE(572), 1,
      sym_enum_declaration_list,
    STATE(1920), 1,
      sym_text_interpolation,
    STATE(2419), 1,
      sym_class_interface_clause,
  [68184] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1921), 1,
      sym_text_interpolation,
    ACTIONS(3213), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACE,
  [68200] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1922), 1,
      sym_text_interpolation,
    ACTIONS(3218), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [68216] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    STATE(1923), 1,
      sym_text_interpolation,
    ACTIONS(3905), 2,
      sym_php_tag,
      aux_sym_text_token1,
    ACTIONS(3907), 2,
      sym__eof,
      aux_sym_text_token2,
  [68234] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3911), 1,
      anon_sym_EQ,
    STATE(1924), 1,
      sym_text_interpolation,
    ACTIONS(3909), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [68252] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3901), 1,
      anon_sym_COMMA,
    STATE(1861), 1,
      aux_sym_base_clause_repeat1,
    STATE(1925), 1,
      sym_text_interpolation,
    ACTIONS(3913), 2,
      anon_sym_LBRACE,
      aux_sym_class_interface_clause_token1,
  [68272] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3353), 1,
      aux_sym_string_content_token1,
    STATE(1905), 1,
      aux_sym_string_content_repeat1,
    STATE(1926), 1,
      sym_text_interpolation,
    ACTIONS(3915), 2,
      anon_sym_SQUOTE,
      aux_sym_string_token2,
  [68292] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(1927), 1,
      sym_text_interpolation,
    STATE(2266), 1,
      sym_formal_parameters,
    STATE(2456), 1,
      sym_reference_modifier,
  [68314] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1928), 1,
      sym_text_interpolation,
    ACTIONS(3654), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [68330] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1929), 1,
      sym_text_interpolation,
    ACTIONS(3599), 4,
      aux_sym_namespace_use_declaration_token1,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
  [68346] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(1930), 1,
      sym_text_interpolation,
    STATE(2217), 1,
      sym_formal_parameters,
    STATE(2402), 1,
      sym_reference_modifier,
  [68368] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3508), 1,
      anon_sym_LBRACE,
    STATE(1931), 1,
      sym_text_interpolation,
    STATE(2065), 1,
      sym_enum_declaration_list,
    STATE(2383), 1,
      sym_class_interface_clause,
  [68390] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3917), 1,
      anon_sym_LPAREN2,
    STATE(546), 1,
      sym__semicolon,
    STATE(1932), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [68410] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3775), 1,
      sym__new_line,
    ACTIONS(3919), 1,
      anon_sym_DQUOTE2,
    ACTIONS(3921), 1,
      sym_heredoc_end,
    STATE(1933), 1,
      sym_text_interpolation,
    STATE(2299), 1,
      sym_heredoc_body,
  [68432] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3663), 1,
      anon_sym_LBRACE,
    STATE(477), 1,
      sym_enum_declaration_list,
    STATE(1934), 1,
      sym_text_interpolation,
    STATE(2410), 1,
      sym_class_interface_clause,
  [68454] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    STATE(1935), 1,
      sym_text_interpolation,
    ACTIONS(3905), 2,
      sym_php_tag,
      aux_sym_text_token1,
    ACTIONS(3907), 2,
      ts_builtin_sym_end,
      aux_sym_text_token2,
  [68472] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(731), 1,
      anon_sym_AMP,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(1936), 1,
      sym_text_interpolation,
    STATE(2375), 2,
      sym_variable_name,
      sym_by_ref,
  [68492] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(822), 1,
      sym_declaration_list,
    STATE(1937), 1,
      sym_text_interpolation,
    STATE(2356), 1,
      sym_class_interface_clause,
  [68514] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(963), 1,
      sym_declaration_list,
    STATE(1938), 1,
      sym_text_interpolation,
    STATE(2332), 1,
      sym_class_interface_clause,
  [68536] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(328), 1,
      anon_sym_DOLLAR,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1841), 1,
      sym__simple_variable,
    STATE(1939), 1,
      sym_text_interpolation,
    STATE(630), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [68556] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(1940), 1,
      sym_text_interpolation,
    STATE(2099), 1,
      sym_declaration_list,
    STATE(2411), 1,
      sym_class_interface_clause,
  [68578] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(1788), 1,
      sym_formal_parameters,
    STATE(1941), 1,
      sym_text_interpolation,
    STATE(2502), 1,
      sym_reference_modifier,
  [68600] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(988), 1,
      sym_declaration_list,
    STATE(1942), 1,
      sym_text_interpolation,
    STATE(2372), 1,
      sym_class_interface_clause,
  [68622] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3781), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3923), 1,
      anon_sym_DASH_GT,
    ACTIONS(3925), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(3927), 1,
      anon_sym_LBRACK,
    STATE(1943), 1,
      sym_text_interpolation,
  [68644] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1944), 1,
      sym_text_interpolation,
    ACTIONS(3534), 4,
      aux_sym_namespace_use_declaration_token1,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
  [68660] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(731), 1,
      anon_sym_AMP,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(1945), 1,
      sym_text_interpolation,
    STATE(1716), 2,
      sym_variable_name,
      sym_by_ref,
  [68680] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(868), 1,
      anon_sym_COMMA,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3929), 1,
      anon_sym_EQ,
    ACTIONS(3931), 1,
      anon_sym_RPAREN2,
    STATE(1946), 1,
      sym_text_interpolation,
    STATE(2176), 1,
      aux_sym__list_destructing_repeat1,
  [68702] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3273), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(470), 1,
      sym_declaration_list,
    STATE(1947), 1,
      sym_text_interpolation,
    STATE(2464), 1,
      sym_base_clause,
  [68724] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1801), 1,
      anon_sym_DOLLAR,
    STATE(950), 1,
      sym__simple_variable,
    STATE(1948), 1,
      sym_text_interpolation,
    STATE(957), 2,
      sym_dynamic_variable_name,
      sym_variable_name,
  [68744] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1949), 1,
      sym_text_interpolation,
    STATE(2670), 1,
      sym_declare_directive,
    ACTIONS(3817), 3,
      anon_sym_ticks,
      anon_sym_encoding,
      anon_sym_strict_types,
  [68762] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(504), 1,
      sym_declaration_list,
    STATE(1950), 1,
      sym_text_interpolation,
    STATE(2425), 1,
      sym_class_interface_clause,
  [68784] = 7,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    ACTIONS(3275), 1,
      aux_sym_class_interface_clause_token1,
    STATE(1951), 1,
      sym_text_interpolation,
    STATE(2104), 1,
      sym_declaration_list,
    STATE(2417), 1,
      sym_class_interface_clause,
  [68806] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3933), 1,
      anon_sym_LPAREN,
    STATE(1952), 1,
      sym_text_interpolation,
    ACTIONS(2612), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      sym_name,
  [68824] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1953), 1,
      sym_text_interpolation,
    ACTIONS(3524), 4,
      aux_sym_namespace_use_declaration_token1,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
  [68840] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3796), 1,
      anon_sym_PIPE,
    STATE(1850), 1,
      aux_sym_type_list_repeat1,
    STATE(1954), 1,
      sym_text_interpolation,
    ACTIONS(3935), 2,
      anon_sym_RPAREN2,
      anon_sym_DOLLAR,
  [68860] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1955), 1,
      sym_text_interpolation,
    ACTIONS(3819), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [68875] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(479), 1,
      sym__semicolon,
    STATE(1956), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [68892] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(483), 1,
      sym__semicolon,
    STATE(1957), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [68909] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(583), 1,
      sym__semicolon,
    STATE(1958), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [68926] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3929), 1,
      anon_sym_EQ,
    STATE(1959), 1,
      sym_text_interpolation,
    ACTIONS(3937), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [68943] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3939), 1,
      anon_sym_LBRACE,
    ACTIONS(3941), 1,
      anon_sym_COLON,
    STATE(538), 1,
      sym_switch_block,
    STATE(1960), 1,
      sym_text_interpolation,
  [68962] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3945), 1,
      anon_sym_EQ,
    STATE(1961), 1,
      sym_text_interpolation,
    ACTIONS(3943), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [68979] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3947), 1,
      anon_sym_COMMA,
    ACTIONS(3949), 1,
      anon_sym_RBRACK,
    STATE(1962), 1,
      sym_text_interpolation,
    STATE(2083), 1,
      aux_sym__array_destructing_repeat1,
  [68998] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3951), 1,
      anon_sym_COMMA,
    ACTIONS(3953), 1,
      anon_sym_RBRACK,
    STATE(1963), 1,
      sym_text_interpolation,
    STATE(2085), 1,
      aux_sym_array_creation_expression_repeat1,
  [69017] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3947), 1,
      anon_sym_COMMA,
    ACTIONS(3949), 1,
      anon_sym_RBRACK,
    STATE(1964), 1,
      sym_text_interpolation,
    STATE(2086), 1,
      aux_sym__array_destructing_repeat1,
  [69036] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3955), 1,
      anon_sym_COMMA,
    ACTIONS(3957), 1,
      anon_sym_RBRACK,
    STATE(1965), 1,
      sym_text_interpolation,
    STATE(2089), 1,
      aux_sym_attribute_group_repeat1,
  [69055] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(554), 1,
      sym__semicolon,
    STATE(1966), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [69072] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(557), 1,
      sym__semicolon,
    STATE(1967), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [69089] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3961), 1,
      aux_sym_else_clause_token1,
    STATE(1968), 1,
      sym_text_interpolation,
    ACTIONS(3959), 2,
      aux_sym_if_statement_token2,
      aux_sym_else_if_clause_token1,
  [69106] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(467), 1,
      sym__semicolon,
    STATE(1969), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [69123] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3963), 1,
      anon_sym_COMMA,
    ACTIONS(3966), 1,
      anon_sym_RBRACE,
    STATE(1970), 2,
      sym_text_interpolation,
      aux_sym_match_block_repeat1,
  [69140] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2536), 1,
      anon_sym_EQ_GT,
    ACTIONS(3968), 1,
      anon_sym_COMMA,
    STATE(1971), 2,
      sym_text_interpolation,
      aux_sym_match_condition_list_repeat1,
  [69157] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1458), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1972), 1,
      sym_text_interpolation,
    ACTIONS(1456), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [69174] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3775), 1,
      sym__new_line,
    ACTIONS(3971), 1,
      sym_heredoc_end,
    STATE(1973), 1,
      sym_text_interpolation,
    STATE(2340), 1,
      sym_heredoc_body,
  [69193] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3973), 1,
      sym__new_line,
    ACTIONS(3975), 1,
      sym_heredoc_end,
    STATE(1974), 1,
      sym_text_interpolation,
    STATE(2393), 1,
      sym_nowdoc_body,
  [69212] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3977), 1,
      anon_sym_COMMA,
    ACTIONS(3979), 1,
      anon_sym_RPAREN2,
    STATE(1975), 1,
      sym_text_interpolation,
    STATE(2282), 1,
      aux_sym_anonymous_function_use_clause_repeat1,
  [69231] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1832), 1,
      anon_sym_DOLLAR,
    STATE(1924), 1,
      sym_variable_name,
    STATE(1976), 1,
      sym_text_interpolation,
    STATE(2230), 1,
      sym_static_variable_declaration,
  [69250] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1462), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1977), 1,
      sym_text_interpolation,
    ACTIONS(1460), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [69267] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1466), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1978), 1,
      sym_text_interpolation,
    ACTIONS(1464), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [69284] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1979), 1,
      sym_text_interpolation,
    ACTIONS(3981), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [69299] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(868), 1,
      anon_sym_COMMA,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3983), 1,
      anon_sym_RPAREN2,
    STATE(1980), 1,
      sym_text_interpolation,
    STATE(2211), 1,
      aux_sym__list_destructing_repeat1,
  [69318] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1981), 1,
      sym_text_interpolation,
    ACTIONS(3985), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [69333] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(868), 1,
      anon_sym_COMMA,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3987), 1,
      anon_sym_RPAREN2,
    STATE(1982), 1,
      sym_text_interpolation,
    STATE(2211), 1,
      aux_sym__list_destructing_repeat1,
  [69352] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(3989), 1,
      anon_sym_EQ_GT,
    STATE(1983), 1,
      sym_text_interpolation,
    STATE(2568), 1,
      sym__return_type,
  [69371] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1470), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1984), 1,
      sym_text_interpolation,
    ACTIONS(1468), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [69388] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3819), 1,
      anon_sym_RBRACE,
    ACTIONS(3991), 1,
      anon_sym_COMMA,
    STATE(1985), 2,
      sym_text_interpolation,
      aux_sym_namespace_use_declaration_repeat1,
  [69405] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1986), 1,
      sym_text_interpolation,
    ACTIONS(3994), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [69420] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1987), 1,
      sym_text_interpolation,
    ACTIONS(3996), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [69435] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1478), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1988), 1,
      sym_text_interpolation,
    ACTIONS(1476), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [69452] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(581), 1,
      sym__semicolon,
    STATE(1989), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [69469] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    STATE(1990), 1,
      sym_text_interpolation,
    ACTIONS(3998), 3,
      sym_heredoc_end,
      sym_nowdoc_string,
      sym__new_line,
  [69484] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(545), 1,
      sym__semicolon,
    STATE(1991), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [69501] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(519), 1,
      sym__semicolon,
    STATE(1992), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [69518] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1426), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1993), 1,
      sym_text_interpolation,
    ACTIONS(1424), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [69535] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(550), 1,
      sym__semicolon,
    STATE(1994), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [69552] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4002), 1,
      anon_sym_EQ,
    STATE(1995), 1,
      sym_text_interpolation,
    ACTIONS(4000), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [69569] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4004), 1,
      anon_sym_COMMA,
    ACTIONS(4006), 1,
      anon_sym_RBRACE,
    STATE(1996), 1,
      sym_text_interpolation,
    STATE(2007), 1,
      aux_sym_match_block_repeat1,
  [69588] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4010), 1,
      aux_sym_string_content_token1,
    STATE(1997), 1,
      sym_text_interpolation,
    ACTIONS(4008), 2,
      anon_sym_SQUOTE,
      aux_sym_string_token2,
  [69605] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1482), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1998), 1,
      sym_text_interpolation,
    ACTIONS(1480), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [69622] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1490), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1999), 1,
      sym_text_interpolation,
    ACTIONS(1488), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [69639] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1494), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2000), 1,
      sym_text_interpolation,
    ACTIONS(1492), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [69656] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1498), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2001), 1,
      sym_text_interpolation,
    ACTIONS(1496), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [69673] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2002), 1,
      sym_text_interpolation,
    STATE(2212), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [69690] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4014), 1,
      aux_sym_string_content_token1,
    STATE(2003), 1,
      sym_text_interpolation,
    ACTIONS(4012), 2,
      anon_sym_SQUOTE,
      aux_sym_string_token2,
  [69707] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1506), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2004), 1,
      sym_text_interpolation,
    ACTIONS(1504), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [69724] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2005), 1,
      sym_text_interpolation,
    ACTIONS(2758), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      sym_name,
  [69739] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(552), 1,
      sym__semicolon,
    STATE(2006), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [69756] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(863), 1,
      anon_sym_RBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4016), 1,
      anon_sym_COMMA,
    STATE(1970), 1,
      aux_sym_match_block_repeat1,
    STATE(2007), 1,
      sym_text_interpolation,
  [69775] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(480), 1,
      sym__semicolon,
    STATE(2008), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [69792] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2009), 1,
      sym_text_interpolation,
    STATE(2259), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [69809] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4018), 1,
      anon_sym_LBRACE,
    ACTIONS(4020), 1,
      anon_sym_COLON,
    STATE(2010), 1,
      sym_text_interpolation,
    STATE(2275), 1,
      sym_switch_block,
  [69828] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(977), 1,
      anon_sym_EQ_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4022), 1,
      anon_sym_COMMA,
    STATE(1971), 1,
      aux_sym_match_condition_list_repeat1,
    STATE(2011), 1,
      sym_text_interpolation,
  [69847] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1534), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2012), 1,
      sym_text_interpolation,
    ACTIONS(1532), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [69864] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2013), 1,
      sym_text_interpolation,
    STATE(2067), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [69881] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2014), 1,
      sym_text_interpolation,
    STATE(2070), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [69898] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4024), 1,
      anon_sym_COMMA,
    ACTIONS(4026), 1,
      anon_sym_RBRACK,
    STATE(2015), 1,
      sym_text_interpolation,
    STATE(2037), 1,
      aux_sym_array_creation_expression_repeat1,
  [69917] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4028), 1,
      anon_sym_COMMA,
    ACTIONS(4030), 1,
      anon_sym_RBRACK,
    STATE(2016), 1,
      sym_text_interpolation,
    STATE(2038), 1,
      aux_sym_attribute_group_repeat1,
  [69936] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(4032), 1,
      anon_sym_EQ_GT,
    STATE(2017), 1,
      sym_text_interpolation,
    STATE(2728), 1,
      sym__return_type,
  [69955] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(544), 1,
      sym__semicolon,
    STATE(2018), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [69972] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2019), 1,
      sym_text_interpolation,
    STATE(2075), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [69989] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2020), 1,
      sym_text_interpolation,
    STATE(2076), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [70006] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1078), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2021), 1,
      sym_text_interpolation,
    ACTIONS(1076), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70023] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1082), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2022), 1,
      sym_text_interpolation,
    ACTIONS(1080), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70040] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(553), 1,
      sym__semicolon,
    STATE(2023), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [70057] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2024), 1,
      sym_text_interpolation,
    STATE(2080), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [70074] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(555), 1,
      sym__semicolon,
    STATE(2025), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [70091] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4034), 1,
      anon_sym_COMMA,
    ACTIONS(4036), 1,
      anon_sym_RPAREN2,
    STATE(2026), 1,
      sym_text_interpolation,
    STATE(2047), 1,
      aux_sym_formal_parameters_repeat1,
  [70110] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(4038), 1,
      anon_sym_LBRACE,
    STATE(2027), 1,
      sym_text_interpolation,
    STATE(2625), 1,
      sym__return_type,
  [70129] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1086), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2028), 1,
      sym_text_interpolation,
    ACTIONS(1084), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70146] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2029), 1,
      sym_text_interpolation,
    STATE(2087), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [70163] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4040), 1,
      anon_sym_COMMA,
    ACTIONS(4042), 1,
      anon_sym_RPAREN2,
    STATE(2030), 1,
      sym_text_interpolation,
    STATE(2050), 1,
      aux_sym_array_creation_expression_repeat1,
  [70182] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1832), 1,
      anon_sym_DOLLAR,
    STATE(1571), 1,
      sym_property_element,
    STATE(1810), 1,
      sym_variable_name,
    STATE(2031), 1,
      sym_text_interpolation,
  [70201] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(570), 1,
      sym__semicolon,
    STATE(2032), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [70218] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2033), 1,
      sym_text_interpolation,
    ACTIONS(4044), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [70233] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(556), 1,
      sym__semicolon,
    STATE(2034), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [70250] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3929), 1,
      anon_sym_EQ,
    STATE(2035), 1,
      sym_text_interpolation,
    ACTIONS(4046), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [70267] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2036), 1,
      sym_text_interpolation,
    ACTIONS(4048), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [70282] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(841), 1,
      anon_sym_RBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4050), 1,
      anon_sym_COMMA,
    STATE(2037), 1,
      sym_text_interpolation,
    STATE(2258), 1,
      aux_sym_array_creation_expression_repeat1,
  [70301] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3064), 1,
      anon_sym_RBRACK,
    ACTIONS(4052), 1,
      anon_sym_COMMA,
    STATE(2038), 1,
      sym_text_interpolation,
    STATE(2273), 1,
      aux_sym_attribute_group_repeat1,
  [70320] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3775), 1,
      sym__new_line,
    ACTIONS(4054), 1,
      sym_heredoc_end,
    STATE(2039), 1,
      sym_text_interpolation,
    STATE(2365), 1,
      sym_heredoc_body,
  [70339] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4056), 1,
      anon_sym_COMMA,
    ACTIONS(4058), 1,
      anon_sym_RPAREN2,
    STATE(2040), 1,
      sym_text_interpolation,
    STATE(2213), 1,
      aux_sym_array_creation_expression_repeat1,
  [70358] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4060), 1,
      anon_sym_COMMA,
    ACTIONS(4062), 1,
      anon_sym_RPAREN2,
    STATE(2041), 1,
      sym_text_interpolation,
    STATE(2063), 1,
      aux_sym_arguments_repeat1,
  [70377] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(578), 1,
      sym__semicolon,
    STATE(2042), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [70394] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1110), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2043), 1,
      sym_text_interpolation,
    ACTIONS(1108), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70411] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1114), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2044), 1,
      sym_text_interpolation,
    ACTIONS(1112), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70428] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1118), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2045), 1,
      sym_text_interpolation,
    ACTIONS(1116), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70445] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1122), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2046), 1,
      sym_text_interpolation,
    ACTIONS(1120), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70462] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2074), 1,
      anon_sym_RPAREN2,
    ACTIONS(4064), 1,
      anon_sym_COMMA,
    STATE(2047), 1,
      sym_text_interpolation,
    STATE(2153), 1,
      aux_sym_formal_parameters_repeat1,
  [70481] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1174), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2048), 1,
      sym_text_interpolation,
    ACTIONS(1172), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70498] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1074), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2049), 1,
      sym_text_interpolation,
    ACTIONS(1072), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70515] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(847), 1,
      anon_sym_RPAREN2,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4066), 1,
      anon_sym_COMMA,
    STATE(2050), 1,
      sym_text_interpolation,
    STATE(2285), 1,
      aux_sym_array_creation_expression_repeat1,
  [70534] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1787), 1,
      anon_sym_RPAREN2,
    ACTIONS(4068), 1,
      anon_sym_COMMA,
    STATE(2051), 1,
      sym_text_interpolation,
    STATE(2281), 1,
      aux_sym_unset_statement_repeat1,
  [70553] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1126), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2052), 1,
      sym_text_interpolation,
    ACTIONS(1124), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70570] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4070), 1,
      anon_sym_COMMA,
    ACTIONS(4072), 1,
      anon_sym_RBRACE,
    STATE(2053), 1,
      sym_text_interpolation,
    STATE(2073), 1,
      aux_sym_match_block_repeat1,
  [70589] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1202), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2054), 1,
      sym_text_interpolation,
    ACTIONS(1200), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70606] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2055), 1,
      sym_text_interpolation,
    ACTIONS(4074), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [70621] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1178), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2056), 1,
      sym_text_interpolation,
    ACTIONS(1176), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70638] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1130), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2057), 1,
      sym_text_interpolation,
    ACTIONS(1128), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70655] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(4076), 1,
      anon_sym_LBRACE,
    STATE(2058), 1,
      sym_text_interpolation,
    STATE(2596), 1,
      sym__return_type,
  [70674] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3608), 1,
      anon_sym_LBRACE,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    STATE(2059), 1,
      sym_text_interpolation,
    STATE(2597), 1,
      sym__return_type,
  [70693] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3145), 1,
      anon_sym_AMP,
    ACTIONS(4078), 1,
      anon_sym_RPAREN2,
    STATE(1511), 1,
      aux_sym_intersection_type_repeat1,
    STATE(2060), 1,
      sym_text_interpolation,
  [70712] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3775), 1,
      sym__new_line,
    ACTIONS(4080), 1,
      sym_heredoc_end,
    STATE(2061), 1,
      sym_text_interpolation,
    STATE(2405), 1,
      sym_heredoc_body,
  [70731] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3973), 1,
      sym__new_line,
    ACTIONS(4082), 1,
      sym_heredoc_end,
    STATE(2062), 1,
      sym_text_interpolation,
    STATE(2407), 1,
      sym_nowdoc_body,
  [70750] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(805), 1,
      anon_sym_RPAREN2,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4084), 1,
      anon_sym_COMMA,
    STATE(2063), 1,
      sym_text_interpolation,
    STATE(2215), 1,
      aux_sym_arguments_repeat1,
  [70769] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1138), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2064), 1,
      sym_text_interpolation,
    ACTIONS(1136), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70786] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1142), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2065), 1,
      sym_text_interpolation,
    ACTIONS(1140), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70803] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1146), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2066), 1,
      sym_text_interpolation,
    ACTIONS(1144), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70820] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1150), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2067), 1,
      sym_text_interpolation,
    ACTIONS(1148), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70837] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2068), 1,
      sym_text_interpolation,
    STATE(2110), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [70854] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(868), 1,
      anon_sym_COMMA,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1769), 1,
      anon_sym_RPAREN2,
    STATE(2069), 1,
      sym_text_interpolation,
    STATE(2154), 1,
      aux_sym__list_destructing_repeat1,
  [70873] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1154), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2070), 1,
      sym_text_interpolation,
    ACTIONS(1152), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70890] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(868), 1,
      anon_sym_COMMA,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1769), 1,
      anon_sym_RPAREN2,
    STATE(2071), 1,
      sym_text_interpolation,
    STATE(2211), 1,
      aux_sym__list_destructing_repeat1,
  [70909] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1158), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2072), 1,
      sym_text_interpolation,
    ACTIONS(1156), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70926] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(853), 1,
      anon_sym_RBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4086), 1,
      anon_sym_COMMA,
    STATE(1970), 1,
      aux_sym_match_block_repeat1,
    STATE(2073), 1,
      sym_text_interpolation,
  [70945] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1162), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2074), 1,
      sym_text_interpolation,
    ACTIONS(1160), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70962] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1062), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2075), 1,
      sym_text_interpolation,
    ACTIONS(1060), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70979] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1170), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2076), 1,
      sym_text_interpolation,
    ACTIONS(1168), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [70996] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4088), 1,
      anon_sym_COMMA,
    ACTIONS(4090), 1,
      anon_sym_RBRACE,
    STATE(2077), 1,
      sym_text_interpolation,
    STATE(2253), 1,
      aux_sym_namespace_use_declaration_repeat1,
  [71015] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1182), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2078), 1,
      sym_text_interpolation,
    ACTIONS(1180), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71032] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1190), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2079), 1,
      sym_text_interpolation,
    ACTIONS(1188), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71049] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1198), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2080), 1,
      sym_text_interpolation,
    ACTIONS(1196), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71066] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2081), 1,
      sym_text_interpolation,
    STATE(2117), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [71083] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2082), 1,
      sym_text_interpolation,
    STATE(2119), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [71100] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3947), 1,
      anon_sym_COMMA,
    ACTIONS(4092), 1,
      anon_sym_RBRACK,
    STATE(2083), 1,
      sym_text_interpolation,
    STATE(2086), 1,
      aux_sym__array_destructing_repeat1,
  [71119] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(4094), 1,
      anon_sym_EQ_GT,
    STATE(2084), 1,
      sym_text_interpolation,
    STATE(2553), 1,
      sym__return_type,
  [71138] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(827), 1,
      anon_sym_RBRACK,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4096), 1,
      anon_sym_COMMA,
    STATE(2085), 1,
      sym_text_interpolation,
    STATE(2258), 1,
      aux_sym_array_creation_expression_repeat1,
  [71157] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4098), 1,
      anon_sym_COMMA,
    ACTIONS(4101), 1,
      anon_sym_RBRACK,
    STATE(2086), 2,
      sym_text_interpolation,
      aux_sym__array_destructing_repeat1,
  [71174] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1210), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2087), 1,
      sym_text_interpolation,
    ACTIONS(1208), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71191] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2088), 1,
      sym_text_interpolation,
    STATE(2122), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [71208] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3102), 1,
      anon_sym_RBRACK,
    ACTIONS(4103), 1,
      anon_sym_COMMA,
    STATE(2089), 1,
      sym_text_interpolation,
    STATE(2273), 1,
      aux_sym_attribute_group_repeat1,
  [71227] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(868), 1,
      anon_sym_COMMA,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4105), 1,
      anon_sym_RPAREN2,
    STATE(1982), 1,
      aux_sym__list_destructing_repeat1,
    STATE(2090), 1,
      sym_text_interpolation,
  [71246] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(868), 1,
      anon_sym_COMMA,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3259), 1,
      anon_sym_RPAREN2,
    STATE(1980), 1,
      aux_sym__list_destructing_repeat1,
    STATE(2091), 1,
      sym_text_interpolation,
  [71265] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(518), 1,
      sym__semicolon,
    STATE(2092), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [71282] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2093), 1,
      sym_text_interpolation,
    ACTIONS(4107), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [71297] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2094), 1,
      sym_text_interpolation,
    ACTIONS(4109), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [71312] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4111), 1,
      sym_name,
    ACTIONS(4113), 1,
      anon_sym_LBRACE,
    STATE(2095), 1,
      sym_text_interpolation,
    STATE(2429), 1,
      sym_namespace_use_group,
  [71331] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1214), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2096), 1,
      sym_text_interpolation,
    ACTIONS(1212), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71348] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1230), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2097), 1,
      sym_text_interpolation,
    ACTIONS(1228), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71365] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3775), 1,
      sym__new_line,
    ACTIONS(4115), 1,
      sym_heredoc_end,
    STATE(2098), 1,
      sym_text_interpolation,
    STATE(2371), 1,
      sym_heredoc_body,
  [71384] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1250), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2099), 1,
      sym_text_interpolation,
    ACTIONS(1248), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71401] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1314), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2100), 1,
      sym_text_interpolation,
    ACTIONS(1312), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71418] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1318), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2101), 1,
      sym_text_interpolation,
    ACTIONS(1316), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71435] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1294), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2102), 1,
      sym_text_interpolation,
    ACTIONS(1292), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71452] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1322), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2103), 1,
      sym_text_interpolation,
    ACTIONS(1320), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71469] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1326), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2104), 1,
      sym_text_interpolation,
    ACTIONS(1324), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71486] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4117), 1,
      anon_sym_COMMA,
    ACTIONS(4119), 1,
      anon_sym_RPAREN2,
    STATE(2105), 1,
      sym_text_interpolation,
    STATE(2106), 1,
      aux_sym_arguments_repeat1,
  [71505] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(807), 1,
      anon_sym_RPAREN2,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4121), 1,
      anon_sym_COMMA,
    STATE(2106), 1,
      sym_text_interpolation,
    STATE(2215), 1,
      aux_sym_arguments_repeat1,
  [71524] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3901), 1,
      anon_sym_COMMA,
    ACTIONS(4123), 1,
      anon_sym_LBRACE,
    STATE(1861), 1,
      aux_sym_base_clause_repeat1,
    STATE(2107), 1,
      sym_text_interpolation,
  [71543] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(531), 1,
      sym__semicolon,
    STATE(2108), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [71560] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1338), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2109), 1,
      sym_text_interpolation,
    ACTIONS(1336), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71577] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1342), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2110), 1,
      sym_text_interpolation,
    ACTIONS(1340), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71594] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2111), 1,
      sym_text_interpolation,
    STATE(2131), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [71611] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2112), 1,
      sym_text_interpolation,
    STATE(2133), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [71628] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1346), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2113), 1,
      sym_text_interpolation,
    ACTIONS(1344), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71645] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1350), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2114), 1,
      sym_text_interpolation,
    ACTIONS(1348), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71662] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1374), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2115), 1,
      sym_text_interpolation,
    ACTIONS(1372), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71679] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(4125), 1,
      sym_name,
    STATE(2116), 1,
      sym_text_interpolation,
    STATE(2687), 1,
      sym_reference_modifier,
  [71698] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1438), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2117), 1,
      sym_text_interpolation,
    ACTIONS(1436), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71715] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4127), 1,
      anon_sym_COMMA,
    ACTIONS(4129), 1,
      anon_sym_RPAREN2,
    STATE(2118), 1,
      sym_text_interpolation,
    STATE(2121), 1,
      aux_sym_arguments_repeat1,
  [71734] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1450), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2119), 1,
      sym_text_interpolation,
    ACTIONS(1448), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71751] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2120), 1,
      sym_text_interpolation,
    STATE(2151), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [71768] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(797), 1,
      anon_sym_RPAREN2,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4131), 1,
      anon_sym_COMMA,
    STATE(2121), 1,
      sym_text_interpolation,
    STATE(2215), 1,
      aux_sym_arguments_repeat1,
  [71787] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1502), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2122), 1,
      sym_text_interpolation,
    ACTIONS(1500), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71804] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2123), 1,
      sym_text_interpolation,
    ACTIONS(4133), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [71819] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1106), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2124), 1,
      sym_text_interpolation,
    ACTIONS(1104), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71836] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1510), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2125), 1,
      sym_text_interpolation,
    ACTIONS(1508), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71853] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1514), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2126), 1,
      sym_text_interpolation,
    ACTIONS(1512), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71870] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1518), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2127), 1,
      sym_text_interpolation,
    ACTIONS(1516), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71887] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4135), 1,
      anon_sym_COMMA,
    ACTIONS(4137), 1,
      anon_sym_RPAREN2,
    STATE(2128), 1,
      sym_text_interpolation,
    STATE(2132), 1,
      aux_sym_arguments_repeat1,
  [71906] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1522), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2129), 1,
      sym_text_interpolation,
    ACTIONS(1520), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71923] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1526), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2130), 1,
      sym_text_interpolation,
    ACTIONS(1524), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71940] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1166), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2131), 1,
      sym_text_interpolation,
    ACTIONS(1164), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71957] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(793), 1,
      anon_sym_RPAREN2,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4139), 1,
      anon_sym_COMMA,
    STATE(2132), 1,
      sym_text_interpolation,
    STATE(2215), 1,
      aux_sym_arguments_repeat1,
  [71976] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1554), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2133), 1,
      sym_text_interpolation,
    ACTIONS(1552), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [71993] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2134), 1,
      sym_text_interpolation,
    STATE(2159), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [72010] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1242), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2135), 1,
      sym_text_interpolation,
    ACTIONS(1240), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72027] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2136), 1,
      sym_text_interpolation,
    STATE(2161), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [72044] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1090), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2137), 1,
      sym_text_interpolation,
    ACTIONS(1088), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72061] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1194), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2138), 1,
      sym_text_interpolation,
    ACTIONS(1192), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72078] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(959), 1,
      anon_sym_COLON,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2092), 1,
      anon_sym_LPAREN2,
    STATE(2139), 1,
      sym_text_interpolation,
    STATE(2399), 1,
      sym_parenthesized_expression,
  [72097] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1222), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2140), 1,
      sym_text_interpolation,
    ACTIONS(1220), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72114] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4141), 1,
      anon_sym_COMMA,
    ACTIONS(4143), 1,
      anon_sym_RPAREN2,
    STATE(2141), 1,
      sym_text_interpolation,
    STATE(2145), 1,
      aux_sym_arguments_repeat1,
  [72133] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1234), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2142), 1,
      sym_text_interpolation,
    ACTIONS(1232), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72150] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(517), 1,
      sym__semicolon,
    STATE(2143), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [72167] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1550), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2144), 1,
      sym_text_interpolation,
    ACTIONS(1548), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72184] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(787), 1,
      anon_sym_RPAREN2,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4145), 1,
      anon_sym_COMMA,
    STATE(2145), 1,
      sym_text_interpolation,
    STATE(2215), 1,
      aux_sym_arguments_repeat1,
  [72203] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1066), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2146), 1,
      sym_text_interpolation,
    ACTIONS(1064), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72220] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1246), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2147), 1,
      sym_text_interpolation,
    ACTIONS(1244), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72237] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1254), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2148), 1,
      sym_text_interpolation,
    ACTIONS(1252), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72254] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1094), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2149), 1,
      sym_text_interpolation,
    ACTIONS(1092), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72271] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1098), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2150), 1,
      sym_text_interpolation,
    ACTIONS(1096), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72288] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1102), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2151), 1,
      sym_text_interpolation,
    ACTIONS(1100), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72305] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4147), 1,
      aux_sym_namespace_use_clause_token1,
    STATE(2152), 1,
      sym_text_interpolation,
    ACTIONS(3887), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [72322] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4149), 1,
      anon_sym_COMMA,
    ACTIONS(4152), 1,
      anon_sym_RPAREN2,
    STATE(2153), 2,
      sym_text_interpolation,
      aux_sym_formal_parameters_repeat1,
  [72339] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(868), 1,
      anon_sym_COMMA,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4154), 1,
      anon_sym_RPAREN2,
    STATE(2154), 1,
      sym_text_interpolation,
    STATE(2211), 1,
      aux_sym__list_destructing_repeat1,
  [72358] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2155), 1,
      sym_text_interpolation,
    ACTIONS(4156), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [72373] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1206), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2156), 1,
      sym_text_interpolation,
    ACTIONS(1204), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72390] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1218), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2157), 1,
      sym_text_interpolation,
    ACTIONS(1216), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72407] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(785), 1,
      anon_sym_RPAREN2,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4158), 1,
      anon_sym_COMMA,
    STATE(2158), 1,
      sym_text_interpolation,
    STATE(2215), 1,
      aux_sym_arguments_repeat1,
  [72426] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1226), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2159), 1,
      sym_text_interpolation,
    ACTIONS(1224), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72443] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1266), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2160), 1,
      sym_text_interpolation,
    ACTIONS(1264), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72460] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1238), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2161), 1,
      sym_text_interpolation,
    ACTIONS(1236), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72477] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4160), 1,
      anon_sym_DQUOTE,
    ACTIONS(4162), 1,
      anon_sym_SQUOTE,
    ACTIONS(4164), 1,
      sym_heredoc_start,
    STATE(2162), 1,
      sym_text_interpolation,
  [72496] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2163), 1,
      sym_text_interpolation,
    STATE(2177), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [72513] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2164), 1,
      sym_text_interpolation,
    STATE(2179), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [72530] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2165), 1,
      sym_text_interpolation,
    STATE(2181), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [72547] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1258), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2166), 1,
      sym_text_interpolation,
    ACTIONS(1256), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72564] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1262), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2167), 1,
      sym_text_interpolation,
    ACTIONS(1260), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72581] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2168), 1,
      sym_text_interpolation,
    STATE(2184), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [72598] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1274), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2169), 1,
      sym_text_interpolation,
    ACTIONS(1272), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72615] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1278), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2170), 1,
      sym_text_interpolation,
    ACTIONS(1276), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72632] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1286), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2171), 1,
      sym_text_interpolation,
    ACTIONS(1284), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72649] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4166), 1,
      anon_sym_COMMA,
    ACTIONS(4168), 1,
      anon_sym_RPAREN2,
    STATE(2158), 1,
      aux_sym_arguments_repeat1,
    STATE(2172), 1,
      sym_text_interpolation,
  [72668] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1290), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2173), 1,
      sym_text_interpolation,
    ACTIONS(1288), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72685] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4170), 1,
      anon_sym_COMMA,
    ACTIONS(4172), 1,
      anon_sym_RPAREN2,
    STATE(2174), 1,
      sym_text_interpolation,
    STATE(2242), 1,
      aux_sym_formal_parameters_repeat1,
  [72704] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2175), 1,
      sym_text_interpolation,
    STATE(2192), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [72721] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(868), 1,
      anon_sym_COMMA,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4174), 1,
      anon_sym_RPAREN2,
    STATE(2176), 1,
      sym_text_interpolation,
    STATE(2211), 1,
      aux_sym__list_destructing_repeat1,
  [72740] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1298), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2177), 1,
      sym_text_interpolation,
    ACTIONS(1296), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72757] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(491), 1,
      sym__semicolon,
    STATE(2178), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [72774] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1302), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2179), 1,
      sym_text_interpolation,
    ACTIONS(1300), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72791] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2180), 1,
      sym_text_interpolation,
    STATE(2194), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [72808] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1310), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2181), 1,
      sym_text_interpolation,
    ACTIONS(1308), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72825] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2182), 1,
      sym_text_interpolation,
    STATE(2195), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [72842] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2183), 1,
      sym_text_interpolation,
    STATE(2197), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [72859] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1354), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2184), 1,
      sym_text_interpolation,
    ACTIONS(1352), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72876] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2185), 1,
      sym_text_interpolation,
    STATE(2199), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [72893] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(4176), 1,
      anon_sym_EQ_GT,
    STATE(2186), 1,
      sym_text_interpolation,
    STATE(2520), 1,
      sym__return_type,
  [72912] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2187), 1,
      sym_text_interpolation,
    STATE(2200), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [72929] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2188), 1,
      sym_text_interpolation,
    STATE(2202), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [72946] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(4178), 1,
      anon_sym_RPAREN2,
    STATE(2189), 1,
      sym_text_interpolation,
    STATE(2624), 1,
      sym_variable_name,
  [72965] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1390), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2190), 1,
      sym_text_interpolation,
    ACTIONS(1388), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72982] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1394), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2191), 1,
      sym_text_interpolation,
    ACTIONS(1392), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [72999] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1398), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2192), 1,
      sym_text_interpolation,
    ACTIONS(1396), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [73016] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4180), 1,
      anon_sym_COMMA,
    ACTIONS(4183), 1,
      anon_sym_RPAREN2,
    STATE(2193), 2,
      sym_text_interpolation,
      aux_sym_anonymous_function_use_clause_repeat1,
  [73033] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1402), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2194), 1,
      sym_text_interpolation,
    ACTIONS(1400), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [73050] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1422), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2195), 1,
      sym_text_interpolation,
    ACTIONS(1420), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [73067] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1399), 1,
      sym__semicolon,
    STATE(2196), 1,
      sym_text_interpolation,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [73084] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1430), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2197), 1,
      sym_text_interpolation,
    ACTIONS(1428), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [73101] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2198), 1,
      sym_text_interpolation,
    STATE(2205), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [73118] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1434), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2199), 1,
      sym_text_interpolation,
    ACTIONS(1432), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [73135] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1442), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2200), 1,
      sym_text_interpolation,
    ACTIONS(1440), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [73152] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2201), 1,
      sym_text_interpolation,
    STATE(2206), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [73169] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1446), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2202), 1,
      sym_text_interpolation,
    ACTIONS(1444), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [73186] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2203), 1,
      sym_text_interpolation,
    STATE(2207), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [73203] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2204), 1,
      sym_text_interpolation,
    STATE(2208), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [73220] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1474), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2205), 1,
      sym_text_interpolation,
    ACTIONS(1472), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [73237] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1530), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2206), 1,
      sym_text_interpolation,
    ACTIONS(1528), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [73254] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1538), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2207), 1,
      sym_text_interpolation,
    ACTIONS(1536), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [73271] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1542), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2208), 1,
      sym_text_interpolation,
    ACTIONS(1540), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [73288] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2209), 1,
      sym_text_interpolation,
    STATE(2210), 1,
      sym__semicolon,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [73305] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1546), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2210), 1,
      sym_text_interpolation,
    ACTIONS(1544), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [73322] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1778), 1,
      anon_sym_RPAREN2,
    ACTIONS(4185), 1,
      anon_sym_COMMA,
    STATE(2211), 2,
      sym_text_interpolation,
      aux_sym__list_destructing_repeat1,
  [73339] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1270), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2212), 1,
      sym_text_interpolation,
    ACTIONS(1268), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [73356] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(837), 1,
      anon_sym_RPAREN2,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4188), 1,
      anon_sym_COMMA,
    STATE(2213), 1,
      sym_text_interpolation,
    STATE(2285), 1,
      aux_sym_array_creation_expression_repeat1,
  [73375] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(457), 1,
      sym__semicolon,
    STATE(2214), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [73392] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4193), 1,
      anon_sym_RPAREN2,
    STATE(2215), 2,
      sym_text_interpolation,
      aux_sym_arguments_repeat1,
  [73409] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(484), 1,
      sym__semicolon,
    STATE(2216), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [73426] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(4195), 1,
      anon_sym_EQ_GT,
    STATE(2217), 1,
      sym_text_interpolation,
    STATE(2609), 1,
      sym__return_type,
  [73445] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3193), 1,
      anon_sym_DOLLAR,
    ACTIONS(4197), 1,
      anon_sym_AMP,
    STATE(1525), 1,
      aux_sym_intersection_type_repeat1,
    STATE(2218), 1,
      sym_text_interpolation,
  [73464] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(4200), 1,
      anon_sym_LBRACE,
    STATE(2219), 1,
      sym_text_interpolation,
    STATE(2649), 1,
      sym__return_type,
  [73483] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(4202), 1,
      anon_sym_LBRACE,
    STATE(2220), 1,
      sym_text_interpolation,
    STATE(2511), 1,
      sym__return_type,
  [73502] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(498), 1,
      sym__semicolon,
    STATE(2221), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [73519] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2222), 1,
      sym_text_interpolation,
    ACTIONS(3789), 3,
      anon_sym_RPAREN2,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [73534] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(4204), 1,
      anon_sym_EQ_GT,
    STATE(2223), 1,
      sym_text_interpolation,
    STATE(2526), 1,
      sym__return_type,
  [73553] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(501), 1,
      sym__semicolon,
    STATE(2224), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [73570] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4208), 1,
      aux_sym_else_clause_token1,
    STATE(2225), 1,
      sym_text_interpolation,
    ACTIONS(4206), 2,
      aux_sym_if_statement_token2,
      aux_sym_else_if_clause_token1,
  [73587] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1362), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2226), 1,
      sym_text_interpolation,
    ACTIONS(1360), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [73604] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3929), 1,
      anon_sym_EQ,
    STATE(2227), 1,
      sym_text_interpolation,
    ACTIONS(4210), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [73621] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4214), 1,
      anon_sym_EQ,
    STATE(2228), 1,
      sym_text_interpolation,
    ACTIONS(4212), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [73638] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(4216), 1,
      anon_sym_LBRACE,
    STATE(2229), 1,
      sym_text_interpolation,
    STATE(2689), 1,
      sym__return_type,
  [73657] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2230), 1,
      sym_text_interpolation,
    ACTIONS(3803), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [73672] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4218), 1,
      aux_sym_namespace_use_clause_token1,
    STATE(2231), 1,
      sym_text_interpolation,
    ACTIONS(3808), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [73689] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2232), 1,
      sym_text_interpolation,
    ACTIONS(3812), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [73704] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4220), 1,
      sym_name,
    STATE(2233), 1,
      sym_text_interpolation,
    ACTIONS(4222), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [73721] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4111), 1,
      sym_name,
    ACTIONS(4113), 1,
      anon_sym_LBRACE,
    STATE(2234), 1,
      sym_text_interpolation,
    STATE(2348), 1,
      sym_namespace_use_group,
  [73740] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(563), 1,
      sym__semicolon,
    STATE(2235), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [73757] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2236), 1,
      sym_text_interpolation,
    ACTIONS(4224), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [73772] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(4226), 1,
      anon_sym_LBRACE,
    STATE(2237), 1,
      sym_text_interpolation,
    STATE(2668), 1,
      sym__return_type,
  [73791] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(4228), 1,
      anon_sym_EQ_GT,
    STATE(2238), 1,
      sym_text_interpolation,
    STATE(2705), 1,
      sym__return_type,
  [73810] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4232), 1,
      anon_sym_EQ,
    STATE(2239), 1,
      sym_text_interpolation,
    ACTIONS(4230), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [73827] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1282), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2240), 1,
      sym_text_interpolation,
    ACTIONS(1280), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [73844] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(577), 1,
      sym__semicolon,
    STATE(2241), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [73861] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2070), 1,
      anon_sym_RPAREN2,
    ACTIONS(4234), 1,
      anon_sym_COMMA,
    STATE(2153), 1,
      aux_sym_formal_parameters_repeat1,
    STATE(2242), 1,
      sym_text_interpolation,
  [73880] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(4236), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(2228), 1,
      sym_variable_name,
    STATE(2243), 1,
      sym_text_interpolation,
  [73899] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4240), 1,
      anon_sym_EQ,
    STATE(2244), 1,
      sym_text_interpolation,
    ACTIONS(4238), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [73916] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(4242), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(2245), 1,
      sym_text_interpolation,
    STATE(2267), 1,
      sym_variable_name,
  [73935] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4246), 1,
      anon_sym_EQ,
    STATE(2246), 1,
      sym_text_interpolation,
    ACTIONS(4244), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [73952] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(509), 1,
      sym__semicolon,
    STATE(2247), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [73969] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(4248), 1,
      anon_sym_LBRACE,
    STATE(2248), 1,
      sym_text_interpolation,
    STATE(2583), 1,
      sym__return_type,
  [73988] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(3735), 1,
      anon_sym_LBRACE,
    STATE(2249), 1,
      sym_text_interpolation,
    STATE(2569), 1,
      sym__return_type,
  [74007] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1370), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2250), 1,
      sym_text_interpolation,
    ACTIONS(1368), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [74024] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1382), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2251), 1,
      sym_text_interpolation,
    ACTIONS(1380), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [74041] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1386), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2252), 1,
      sym_text_interpolation,
    ACTIONS(1384), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [74058] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4088), 1,
      anon_sym_COMMA,
    ACTIONS(4250), 1,
      anon_sym_RBRACE,
    STATE(1985), 1,
      aux_sym_namespace_use_declaration_repeat1,
    STATE(2253), 1,
      sym_text_interpolation,
  [74077] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2254), 1,
      sym_text_interpolation,
    ACTIONS(4252), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
  [74092] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1454), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2255), 1,
      sym_text_interpolation,
    ACTIONS(1452), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [74109] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2256), 1,
      sym_text_interpolation,
    ACTIONS(3247), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [74124] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2257), 1,
      sym_text_interpolation,
    ACTIONS(3848), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [74139] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4252), 1,
      anon_sym_RBRACK,
    ACTIONS(4254), 1,
      anon_sym_COMMA,
    STATE(2258), 2,
      sym_text_interpolation,
      aux_sym_array_creation_expression_repeat1,
  [74156] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1186), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2259), 1,
      sym_text_interpolation,
    ACTIONS(1184), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [74173] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4257), 1,
      anon_sym_DQUOTE,
    ACTIONS(4259), 1,
      anon_sym_SQUOTE,
    ACTIONS(4261), 1,
      sym_heredoc_start,
    STATE(2260), 1,
      sym_text_interpolation,
  [74192] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1851), 1,
      anon_sym_AMP,
    ACTIONS(4263), 1,
      sym_name,
    STATE(2261), 1,
      sym_text_interpolation,
    STATE(2732), 1,
      sym_reference_modifier,
  [74211] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2262), 1,
      sym_text_interpolation,
    ACTIONS(4265), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [74226] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2263), 1,
      sym_text_interpolation,
    ACTIONS(4267), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [74241] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(4269), 1,
      anon_sym_LBRACE,
    STATE(2264), 1,
      sym_text_interpolation,
    STATE(2537), 1,
      sym__return_type,
  [74260] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2012), 1,
      sym__semicolon,
    STATE(2265), 1,
      sym_text_interpolation,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [74277] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_COLON,
    ACTIONS(4271), 1,
      anon_sym_EQ_GT,
    STATE(2266), 1,
      sym_text_interpolation,
    STATE(2608), 1,
      sym__return_type,
  [74296] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4275), 1,
      anon_sym_EQ,
    STATE(2267), 1,
      sym_text_interpolation,
    ACTIONS(4273), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [74313] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1306), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2268), 1,
      sym_text_interpolation,
    ACTIONS(1304), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [74330] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1330), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2269), 1,
      sym_text_interpolation,
    ACTIONS(1328), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [74347] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1334), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2270), 1,
      sym_text_interpolation,
    ACTIONS(1332), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [74364] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1358), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2271), 1,
      sym_text_interpolation,
    ACTIONS(1356), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [74381] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1366), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2272), 1,
      sym_text_interpolation,
    ACTIONS(1364), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [74398] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4277), 1,
      anon_sym_COMMA,
    ACTIONS(4280), 1,
      anon_sym_RBRACK,
    STATE(2273), 2,
      sym_text_interpolation,
      aux_sym_attribute_group_repeat1,
  [74415] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(528), 1,
      sym__semicolon,
    STATE(2274), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [74432] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1378), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2275), 1,
      sym_text_interpolation,
    ACTIONS(1376), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [74449] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(579), 1,
      sym__semicolon,
    STATE(2276), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [74466] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4284), 1,
      anon_sym_EQ,
    STATE(2277), 1,
      sym_text_interpolation,
    ACTIONS(4282), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [74483] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1406), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2278), 1,
      sym_text_interpolation,
    ACTIONS(1404), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [74500] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2268), 1,
      sym__semicolon,
    STATE(2279), 1,
      sym_text_interpolation,
    ACTIONS(416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [74517] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(494), 1,
      sym__semicolon,
    STATE(2280), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [74534] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3327), 1,
      anon_sym_RPAREN2,
    ACTIONS(4286), 1,
      anon_sym_COMMA,
    STATE(2281), 2,
      sym_text_interpolation,
      aux_sym_unset_statement_repeat1,
  [74551] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3532), 1,
      anon_sym_RPAREN2,
    ACTIONS(4289), 1,
      anon_sym_COMMA,
    STATE(2193), 1,
      aux_sym_anonymous_function_use_clause_repeat1,
    STATE(2282), 1,
      sym_text_interpolation,
  [74570] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(580), 1,
      sym__semicolon,
    STATE(2283), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [74587] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1410), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2284), 1,
      sym_text_interpolation,
    ACTIONS(1408), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [74604] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4252), 1,
      anon_sym_RPAREN2,
    ACTIONS(4291), 1,
      anon_sym_COMMA,
    STATE(2285), 2,
      sym_text_interpolation,
      aux_sym_array_creation_expression_repeat1,
  [74621] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1414), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2286), 1,
      sym_text_interpolation,
    ACTIONS(1412), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [74638] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(4294), 1,
      anon_sym_RPAREN2,
    STATE(2287), 1,
      sym_text_interpolation,
    STATE(2616), 1,
      sym_variable_name,
  [74657] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(488), 1,
      sym__semicolon,
    STATE(2288), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [74674] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1832), 1,
      anon_sym_DOLLAR,
    STATE(1584), 1,
      sym_property_element,
    STATE(1810), 1,
      sym_variable_name,
    STATE(2289), 1,
      sym_text_interpolation,
  [74693] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1832), 1,
      anon_sym_DOLLAR,
    STATE(1810), 1,
      sym_variable_name,
    STATE(1928), 1,
      sym_property_element,
    STATE(2290), 1,
      sym_text_interpolation,
  [74712] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3901), 1,
      anon_sym_COMMA,
    ACTIONS(4296), 1,
      anon_sym_LBRACE,
    STATE(2107), 1,
      aux_sym_base_clause_repeat1,
    STATE(2291), 1,
      sym_text_interpolation,
  [74731] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1418), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2292), 1,
      sym_text_interpolation,
    ACTIONS(1416), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [74748] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(4298), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(2277), 1,
      sym_variable_name,
    STATE(2293), 1,
      sym_text_interpolation,
  [74767] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2294), 1,
      sym_text_interpolation,
    ACTIONS(3554), 3,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      aux_sym_class_interface_clause_token1,
  [74782] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1785), 1,
      anon_sym_RPAREN2,
    ACTIONS(4300), 1,
      anon_sym_COMMA,
    STATE(2281), 1,
      aux_sym_unset_statement_repeat1,
    STATE(2295), 1,
      sym_text_interpolation,
  [74801] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(516), 1,
      sym__semicolon,
    STATE(2296), 1,
      sym_text_interpolation,
    ACTIONS(360), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [74818] = 6,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    ACTIONS(4302), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(2239), 1,
      sym_variable_name,
    STATE(2297), 1,
      sym_text_interpolation,
  [74837] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2298), 1,
      sym_text_interpolation,
    ACTIONS(2476), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN2,
  [74851] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4304), 1,
      sym__new_line,
    ACTIONS(4306), 1,
      sym_heredoc_end,
    STATE(2299), 1,
      sym_text_interpolation,
  [74867] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3508), 1,
      anon_sym_LBRACE,
    STATE(2001), 1,
      sym_enum_declaration_list,
    STATE(2300), 1,
      sym_text_interpolation,
  [74883] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2301), 1,
      sym_text_interpolation,
    ACTIONS(3966), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [74897] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(2004), 1,
      sym_declaration_list,
    STATE(2302), 1,
      sym_text_interpolation,
  [74913] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2110), 1,
      anon_sym_LPAREN2,
    STATE(946), 1,
      sym_arguments,
    STATE(2303), 1,
      sym_text_interpolation,
  [74929] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2092), 1,
      anon_sym_LPAREN2,
    STATE(2304), 1,
      sym_text_interpolation,
    STATE(2337), 1,
      sym_parenthesized_expression,
  [74945] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2305), 1,
      sym_text_interpolation,
    ACTIONS(4308), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [74959] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4311), 1,
      anon_sym_LBRACE,
    STATE(453), 1,
      sym_compound_statement,
    STATE(2306), 1,
      sym_text_interpolation,
  [74975] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(2307), 1,
      sym_text_interpolation,
    STATE(2431), 1,
      sym_variable_name,
  [74991] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(524), 1,
      sym_declaration_list,
    STATE(2308), 1,
      sym_text_interpolation,
  [75007] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4313), 1,
      anon_sym_LPAREN2,
    STATE(108), 1,
      sym_parenthesized_expression,
    STATE(2309), 1,
      sym_text_interpolation,
  [75023] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(376), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1838), 1,
      sym_compound_statement,
    STATE(2310), 1,
      sym_text_interpolation,
  [75039] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2110), 1,
      anon_sym_LPAREN2,
    STATE(984), 1,
      sym_arguments,
    STATE(2311), 1,
      sym_text_interpolation,
  [75055] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4315), 1,
      anon_sym_LPAREN2,
    STATE(2020), 1,
      sym_parenthesized_expression,
    STATE(2312), 1,
      sym_text_interpolation,
  [75071] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2313), 1,
      sym_text_interpolation,
    ACTIONS(3048), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [75085] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(2249), 1,
      sym_formal_parameters,
    STATE(2314), 1,
      sym_text_interpolation,
  [75101] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4313), 1,
      anon_sym_LPAREN2,
    STATE(53), 1,
      sym_parenthesized_expression,
    STATE(2315), 1,
      sym_text_interpolation,
  [75117] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2316), 1,
      sym_text_interpolation,
    ACTIONS(4317), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [75131] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4319), 1,
      anon_sym_BSLASH,
    STATE(2317), 2,
      sym_text_interpolation,
      aux_sym_namespace_name_repeat1,
  [75145] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2318), 1,
      sym_text_interpolation,
    ACTIONS(4152), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [75159] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1816), 1,
      anon_sym_LPAREN2,
    STATE(732), 1,
      sym_arguments,
    STATE(2319), 1,
      sym_text_interpolation,
  [75175] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2320), 1,
      sym_text_interpolation,
    ACTIONS(4322), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [75189] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(1806), 1,
      sym_formal_parameters,
    STATE(2321), 1,
      sym_text_interpolation,
  [75205] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(822), 1,
      sym_declaration_list,
    STATE(2322), 1,
      sym_text_interpolation,
  [75221] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1816), 1,
      anon_sym_LPAREN2,
    STATE(738), 1,
      sym_arguments,
    STATE(2323), 1,
      sym_text_interpolation,
  [75237] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3663), 1,
      anon_sym_LBRACE,
    STATE(569), 1,
      sym_enum_declaration_list,
    STATE(2324), 1,
      sym_text_interpolation,
  [75253] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2325), 1,
      sym_text_interpolation,
    ACTIONS(4324), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [75267] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(1002), 1,
      sym_declaration_list,
    STATE(2326), 1,
      sym_text_interpolation,
  [75283] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1837), 1,
      anon_sym_LPAREN2,
    STATE(816), 1,
      sym_arguments,
    STATE(2327), 1,
      sym_text_interpolation,
  [75299] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(959), 1,
      anon_sym_COLON,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4326), 1,
      anon_sym_LPAREN2,
    STATE(2328), 1,
      sym_text_interpolation,
  [75315] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2329), 1,
      sym_text_interpolation,
    ACTIONS(3929), 2,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [75329] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4328), 1,
      sym__new_line,
    ACTIONS(4330), 1,
      sym_heredoc_end,
    STATE(2330), 1,
      sym_text_interpolation,
  [75345] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(571), 1,
      sym_declaration_list,
    STATE(2331), 1,
      sym_text_interpolation,
  [75361] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(1003), 1,
      sym_declaration_list,
    STATE(2332), 1,
      sym_text_interpolation,
  [75377] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1837), 1,
      anon_sym_LPAREN2,
    STATE(828), 1,
      sym_arguments,
    STATE(2333), 1,
      sym_text_interpolation,
  [75393] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1706), 1,
      anon_sym_LPAREN2,
    STATE(631), 1,
      sym_arguments,
    STATE(2334), 1,
      sym_text_interpolation,
  [75409] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1590), 1,
      anon_sym_LPAREN2,
    STATE(615), 1,
      sym_arguments,
    STATE(2335), 1,
      sym_text_interpolation,
  [75425] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(2043), 1,
      sym_declaration_list,
    STATE(2336), 1,
      sym_text_interpolation,
  [75441] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4332), 1,
      anon_sym_LBRACE,
    STATE(1131), 1,
      sym_match_block,
    STATE(2337), 1,
      sym_text_interpolation,
  [75457] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(376), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2054), 1,
      sym_compound_statement,
    STATE(2338), 1,
      sym_text_interpolation,
  [75473] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(376), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1028), 1,
      sym_compound_statement,
    STATE(2339), 1,
      sym_text_interpolation,
  [75489] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4334), 1,
      sym__new_line,
    ACTIONS(4336), 1,
      sym_heredoc_end,
    STATE(2340), 1,
      sym_text_interpolation,
  [75505] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1590), 1,
      anon_sym_LPAREN2,
    STATE(606), 1,
      sym_arguments,
    STATE(2341), 1,
      sym_text_interpolation,
  [75521] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(2066), 1,
      sym_declaration_list,
    STATE(2342), 1,
      sym_text_interpolation,
  [75537] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(376), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1711), 1,
      sym_compound_statement,
    STATE(2343), 1,
      sym_text_interpolation,
  [75553] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4313), 1,
      anon_sym_LPAREN2,
    STATE(62), 1,
      sym_parenthesized_expression,
    STATE(2344), 1,
      sym_text_interpolation,
  [75569] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2092), 1,
      anon_sym_LPAREN2,
    STATE(2345), 1,
      sym_text_interpolation,
    STATE(2399), 1,
      sym_parenthesized_expression,
  [75585] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2092), 1,
      anon_sym_LPAREN2,
    STATE(2010), 1,
      sym_parenthesized_expression,
    STATE(2346), 1,
      sym_text_interpolation,
  [75601] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4338), 1,
      sym__new_line,
    ACTIONS(4340), 1,
      sym_heredoc_end,
    STATE(2347), 1,
      sym_text_interpolation,
  [75617] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2348), 1,
      sym_text_interpolation,
    ACTIONS(4342), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [75631] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4313), 1,
      anon_sym_LPAREN2,
    STATE(113), 1,
      sym_parenthesized_expression,
    STATE(2349), 1,
      sym_text_interpolation,
  [75647] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2350), 1,
      sym_text_interpolation,
    ACTIONS(4048), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [75661] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2351), 1,
      sym_text_interpolation,
    ACTIONS(4344), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [75675] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4311), 1,
      anon_sym_LBRACE,
    STATE(450), 1,
      sym_compound_statement,
    STATE(2352), 1,
      sym_text_interpolation,
  [75691] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(968), 1,
      sym_declaration_list,
    STATE(2353), 1,
      sym_text_interpolation,
  [75707] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(496), 1,
      sym_declaration_list,
    STATE(2354), 1,
      sym_text_interpolation,
  [75723] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2355), 1,
      sym_text_interpolation,
    ACTIONS(2538), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [75737] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(767), 1,
      sym_declaration_list,
    STATE(2356), 1,
      sym_text_interpolation,
  [75753] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(928), 1,
      sym_declaration_list,
    STATE(2357), 1,
      sym_text_interpolation,
  [75769] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2358), 1,
      sym_text_interpolation,
    ACTIONS(4346), 2,
      anon_sym_string,
      anon_sym_int,
  [75783] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(1836), 1,
      sym_formal_parameters,
    STATE(2359), 1,
      sym_text_interpolation,
  [75799] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2360), 1,
      sym_text_interpolation,
    ACTIONS(4348), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [75813] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(768), 1,
      sym_declaration_list,
    STATE(2361), 1,
      sym_text_interpolation,
  [75829] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3237), 1,
      anon_sym_LPAREN2,
    STATE(1761), 1,
      sym_formal_parameters,
    STATE(2362), 1,
      sym_text_interpolation,
  [75845] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(1792), 1,
      sym_formal_parameters,
    STATE(2363), 1,
      sym_text_interpolation,
  [75861] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2364), 1,
      sym_text_interpolation,
    ACTIONS(4350), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [75875] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4352), 1,
      sym__new_line,
    ACTIONS(4354), 1,
      sym_heredoc_end,
    STATE(2365), 1,
      sym_text_interpolation,
  [75891] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(2223), 1,
      sym_formal_parameters,
    STATE(2366), 1,
      sym_text_interpolation,
  [75907] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4356), 1,
      sym__new_line,
    ACTIONS(4358), 1,
      sym_heredoc_end,
    STATE(2367), 1,
      sym_text_interpolation,
  [75923] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2368), 1,
      sym_text_interpolation,
    ACTIONS(4360), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [75937] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(2250), 1,
      sym_declaration_list,
    STATE(2369), 1,
      sym_text_interpolation,
  [75953] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2370), 1,
      sym_text_interpolation,
    ACTIONS(4362), 2,
      anon_sym_string,
      anon_sym_int,
  [75967] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4364), 1,
      sym__new_line,
    ACTIONS(4366), 1,
      sym_heredoc_end,
    STATE(2371), 1,
      sym_text_interpolation,
  [75983] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(962), 1,
      sym_declaration_list,
    STATE(2372), 1,
      sym_text_interpolation,
  [75999] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2092), 1,
      anon_sym_LPAREN2,
    STATE(2373), 1,
      sym_text_interpolation,
    STATE(2444), 1,
      sym_parenthesized_expression,
  [76015] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2374), 1,
      sym_text_interpolation,
    ACTIONS(4368), 2,
      anon_sym_LBRACE,
      anon_sym_COLON,
  [76029] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2375), 1,
      sym_text_interpolation,
    ACTIONS(4183), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [76043] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(2096), 1,
      sym_declaration_list,
    STATE(2376), 1,
      sym_text_interpolation,
  [76059] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3508), 1,
      anon_sym_LBRACE,
    STATE(2097), 1,
      sym_enum_declaration_list,
    STATE(2377), 1,
      sym_text_interpolation,
  [76075] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(2099), 1,
      sym_declaration_list,
    STATE(2378), 1,
      sym_text_interpolation,
  [76091] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3166), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(3168), 1,
      aux_sym__arrow_function_header_token1,
    STATE(2379), 1,
      sym_text_interpolation,
  [76107] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(2104), 1,
      sym_declaration_list,
    STATE(2380), 1,
      sym_text_interpolation,
  [76123] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2381), 1,
      sym_text_interpolation,
    ACTIONS(2548), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [76137] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2382), 1,
      sym_text_interpolation,
    ACTIONS(4370), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [76151] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3508), 1,
      anon_sym_LBRACE,
    STATE(2109), 1,
      sym_enum_declaration_list,
    STATE(2383), 1,
      sym_text_interpolation,
  [76167] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2384), 1,
      sym_text_interpolation,
    ACTIONS(3335), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [76181] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(963), 1,
      sym_declaration_list,
    STATE(2385), 1,
      sym_text_interpolation,
  [76197] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2386), 1,
      sym_text_interpolation,
    ACTIONS(2492), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN2,
  [76211] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2387), 1,
      sym_text_interpolation,
    ACTIONS(4372), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [76225] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1971), 1,
      anon_sym_LPAREN2,
    ACTIONS(3646), 1,
      anon_sym_COLON_COLON,
    STATE(2388), 1,
      sym_text_interpolation,
  [76241] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2389), 1,
      sym_text_interpolation,
    ACTIONS(2556), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [76255] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1706), 1,
      anon_sym_LPAREN2,
    STATE(638), 1,
      sym_arguments,
    STATE(2390), 1,
      sym_text_interpolation,
  [76271] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3879), 1,
      sym_nowdoc_string,
    STATE(1907), 1,
      aux_sym_nowdoc_body_repeat1,
    STATE(2391), 1,
      sym_text_interpolation,
  [76287] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(513), 1,
      ts_builtin_sym_end,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4374), 1,
      sym_php_tag,
    STATE(2392), 1,
      sym_text_interpolation,
  [76303] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4376), 1,
      sym__new_line,
    ACTIONS(4378), 1,
      sym_heredoc_end,
    STATE(2393), 1,
      sym_text_interpolation,
  [76319] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(988), 1,
      sym_declaration_list,
    STATE(2394), 1,
      sym_text_interpolation,
  [76335] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(777), 1,
      sym_declaration_list,
    STATE(2395), 1,
      sym_text_interpolation,
  [76351] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2396), 1,
      sym_text_interpolation,
    ACTIONS(4193), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [76365] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(778), 1,
      sym_declaration_list,
    STATE(2397), 1,
      sym_text_interpolation,
  [76381] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(469), 1,
      sym_declaration_list,
    STATE(2398), 1,
      sym_text_interpolation,
  [76397] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4380), 1,
      anon_sym_LBRACE,
    STATE(1019), 1,
      sym_match_block,
    STATE(2399), 1,
      sym_text_interpolation,
  [76413] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4313), 1,
      anon_sym_LPAREN2,
    STATE(59), 1,
      sym_parenthesized_expression,
    STATE(2400), 1,
      sym_text_interpolation,
  [76429] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2401), 1,
      sym_text_interpolation,
    ACTIONS(2594), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [76443] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(2084), 1,
      sym_formal_parameters,
    STATE(2402), 1,
      sym_text_interpolation,
  [76459] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2403), 1,
      sym_text_interpolation,
    ACTIONS(4382), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [76473] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(2404), 1,
      sym_text_interpolation,
    STATE(2466), 1,
      sym_variable_name,
  [76489] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4384), 1,
      sym__new_line,
    ACTIONS(4386), 1,
      sym_heredoc_end,
    STATE(2405), 1,
      sym_text_interpolation,
  [76505] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2406), 1,
      sym_text_interpolation,
    ACTIONS(4388), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [76519] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(11), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4390), 1,
      sym__new_line,
    ACTIONS(4392), 1,
      sym_heredoc_end,
    STATE(2407), 1,
      sym_text_interpolation,
  [76535] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2408), 1,
      sym_text_interpolation,
    ACTIONS(4394), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [76549] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2409), 1,
      sym_text_interpolation,
    ACTIONS(4222), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [76563] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3663), 1,
      anon_sym_LBRACE,
    STATE(527), 1,
      sym_enum_declaration_list,
    STATE(2410), 1,
      sym_text_interpolation,
  [76579] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(2126), 1,
      sym_declaration_list,
    STATE(2411), 1,
      sym_text_interpolation,
  [76595] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(358), 1,
      anon_sym_COLON,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2412), 1,
      sym_text_interpolation,
    STATE(2598), 1,
      sym_colon_block,
  [76611] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3092), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(3094), 1,
      aux_sym__arrow_function_header_token1,
    STATE(2413), 1,
      sym_text_interpolation,
  [76627] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2414), 1,
      sym_text_interpolation,
    ACTIONS(4396), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [76641] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(2127), 1,
      sym_declaration_list,
    STATE(2415), 1,
      sym_text_interpolation,
  [76657] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1807), 1,
      anon_sym_BSLASH,
    STATE(2317), 1,
      aux_sym_namespace_name_repeat1,
    STATE(2416), 1,
      sym_text_interpolation,
  [76673] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(2130), 1,
      sym_declaration_list,
    STATE(2417), 1,
      sym_text_interpolation,
  [76689] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4398), 1,
      anon_sym_LBRACE,
    STATE(1113), 1,
      sym_compound_statement,
    STATE(2418), 1,
      sym_text_interpolation,
  [76705] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3663), 1,
      anon_sym_LBRACE,
    STATE(493), 1,
      sym_enum_declaration_list,
    STATE(2419), 1,
      sym_text_interpolation,
  [76721] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(376), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1791), 1,
      sym_compound_statement,
    STATE(2420), 1,
      sym_text_interpolation,
  [76737] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2421), 1,
      sym_text_interpolation,
    ACTIONS(4400), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [76751] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2422), 1,
      sym_text_interpolation,
    ACTIONS(4101), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [76765] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4313), 1,
      anon_sym_LPAREN2,
    STATE(98), 1,
      sym_parenthesized_expression,
    STATE(2423), 1,
      sym_text_interpolation,
  [76781] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2424), 1,
      sym_text_interpolation,
    ACTIONS(1780), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [76795] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(573), 1,
      sym_declaration_list,
    STATE(2425), 1,
      sym_text_interpolation,
  [76811] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2426), 1,
      sym_text_interpolation,
    ACTIONS(2492), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [76825] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4326), 1,
      anon_sym_LPAREN2,
    ACTIONS(4403), 1,
      anon_sym_RPAREN2,
    STATE(2427), 1,
      sym_text_interpolation,
  [76841] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(783), 1,
      sym_declaration_list,
    STATE(2428), 1,
      sym_text_interpolation,
  [76857] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2429), 1,
      sym_text_interpolation,
    ACTIONS(4405), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [76871] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(784), 1,
      sym_declaration_list,
    STATE(2430), 1,
      sym_text_interpolation,
  [76887] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2431), 1,
      sym_text_interpolation,
    ACTIONS(4407), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [76901] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(2414), 1,
      sym_variable_name,
    STATE(2432), 1,
      sym_text_interpolation,
  [76917] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4315), 1,
      anon_sym_LPAREN2,
    STATE(2216), 1,
      sym_parenthesized_expression,
    STATE(2433), 1,
      sym_text_interpolation,
  [76933] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(2238), 1,
      sym_formal_parameters,
    STATE(2434), 1,
      sym_text_interpolation,
  [76949] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3508), 1,
      anon_sym_LBRACE,
    STATE(2156), 1,
      sym_enum_declaration_list,
    STATE(2435), 1,
      sym_text_interpolation,
  [76965] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(2157), 1,
      sym_declaration_list,
    STATE(2436), 1,
      sym_text_interpolation,
  [76981] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(2403), 1,
      sym_variable_name,
    STATE(2437), 1,
      sym_text_interpolation,
  [76997] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(376), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1811), 1,
      sym_compound_statement,
    STATE(2438), 1,
      sym_text_interpolation,
  [77013] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(2439), 1,
      sym_text_interpolation,
    STATE(2458), 1,
      sym_variable_name,
  [77029] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(574), 1,
      sym_declaration_list,
    STATE(2440), 1,
      sym_text_interpolation,
  [77045] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2441), 1,
      sym_text_interpolation,
    ACTIONS(3819), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [77059] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(787), 1,
      sym_declaration_list,
    STATE(2442), 1,
      sym_text_interpolation,
  [77075] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2443), 1,
      sym_text_interpolation,
    ACTIONS(3994), 2,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
  [77089] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(493), 1,
      anon_sym_COLON,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1968), 1,
      sym_colon_block,
    STATE(2444), 1,
      sym_text_interpolation,
  [77105] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(576), 1,
      sym_declaration_list,
    STATE(2445), 1,
      sym_text_interpolation,
  [77121] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2446), 1,
      sym_text_interpolation,
    ACTIONS(4409), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [77135] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2447), 1,
      sym_text_interpolation,
    ACTIONS(4411), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [77149] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2448), 1,
      sym_text_interpolation,
    ACTIONS(4413), 2,
      anon_sym_SEMI,
      anon_sym_COLON,
  [77163] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2449), 1,
      sym_text_interpolation,
    ACTIONS(4224), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [77177] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2450), 1,
      sym_text_interpolation,
    ACTIONS(4415), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [77191] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2451), 1,
      sym_text_interpolation,
    ACTIONS(4417), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [77205] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2452), 1,
      sym_text_interpolation,
    ACTIONS(1778), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [77219] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2453), 1,
      sym_text_interpolation,
    ACTIONS(3060), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [77233] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2454), 1,
      sym_text_interpolation,
    ACTIONS(4419), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [77247] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(228), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(559), 1,
      sym_compound_statement,
    STATE(2455), 1,
      sym_text_interpolation,
  [77263] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(2017), 1,
      sym_formal_parameters,
    STATE(2456), 1,
      sym_text_interpolation,
  [77279] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(942), 1,
      sym_declaration_list,
    STATE(2457), 1,
      sym_text_interpolation,
  [77295] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2458), 1,
      sym_text_interpolation,
    ACTIONS(4421), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [77309] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(376), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1837), 1,
      sym_compound_statement,
    STATE(2459), 1,
      sym_text_interpolation,
  [77325] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(2460), 1,
      sym_text_interpolation,
    STATE(2470), 1,
      sym_variable_name,
  [77341] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(228), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(492), 1,
      sym_compound_statement,
    STATE(2461), 1,
      sym_text_interpolation,
  [77357] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2462), 1,
      sym_text_interpolation,
    ACTIONS(4423), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [77371] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2463), 1,
      sym_text_interpolation,
    ACTIONS(4425), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [77385] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(495), 1,
      sym_declaration_list,
    STATE(2464), 1,
      sym_text_interpolation,
  [77401] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4313), 1,
      anon_sym_LPAREN2,
    STATE(114), 1,
      sym_parenthesized_expression,
    STATE(2465), 1,
      sym_text_interpolation,
  [77417] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2466), 1,
      sym_text_interpolation,
    ACTIONS(4427), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [77431] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2467), 1,
      sym_text_interpolation,
    ACTIONS(4280), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [77445] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2468), 1,
      sym_text_interpolation,
    ACTIONS(4210), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [77459] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4313), 1,
      anon_sym_LPAREN2,
    STATE(81), 1,
      sym_parenthesized_expression,
    STATE(2469), 1,
      sym_text_interpolation,
  [77475] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2470), 1,
      sym_text_interpolation,
    ACTIONS(4429), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [77489] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1590), 1,
      anon_sym_LPAREN2,
    STATE(804), 1,
      sym_arguments,
    STATE(2471), 1,
      sym_text_interpolation,
  [77505] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4431), 1,
      anon_sym_BSLASH,
    STATE(2317), 1,
      aux_sym_namespace_name_repeat1,
    STATE(2472), 1,
      sym_text_interpolation,
  [77521] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2473), 1,
      sym_text_interpolation,
    ACTIONS(4434), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [77535] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2072), 1,
      anon_sym_LPAREN2,
    STATE(867), 1,
      sym_arguments,
    STATE(2474), 1,
      sym_text_interpolation,
  [77551] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2092), 1,
      anon_sym_LPAREN2,
    STATE(1960), 1,
      sym_parenthesized_expression,
    STATE(2475), 1,
      sym_text_interpolation,
  [77567] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2476), 1,
      sym_text_interpolation,
    ACTIONS(4436), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [77581] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2477), 1,
      sym_text_interpolation,
    ACTIONS(4438), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [77595] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(2368), 1,
      sym_variable_name,
    STATE(2478), 1,
      sym_text_interpolation,
  [77611] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(2059), 1,
      sym_formal_parameters,
    STATE(2479), 1,
      sym_text_interpolation,
  [77627] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4313), 1,
      anon_sym_LPAREN2,
    STATE(88), 1,
      sym_parenthesized_expression,
    STATE(2480), 1,
      sym_text_interpolation,
  [77643] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2481), 1,
      sym_text_interpolation,
    ACTIONS(3333), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [77657] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(376), 1,
      anon_sym_LBRACE,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(1972), 1,
      sym_compound_statement,
    STATE(2482), 1,
      sym_text_interpolation,
  [77673] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2483), 1,
      sym_text_interpolation,
    ACTIONS(4440), 2,
      anon_sym_LBRACE,
      anon_sym_COLON,
  [77687] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2484), 1,
      sym_text_interpolation,
    ACTIONS(4442), 2,
      sym__eof,
      sym_php_tag,
  [77701] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(536), 1,
      sym_declaration_list,
    STATE(2485), 1,
      sym_text_interpolation,
  [77717] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4313), 1,
      anon_sym_LPAREN2,
    STATE(58), 1,
      sym_parenthesized_expression,
    STATE(2486), 1,
      sym_text_interpolation,
  [77733] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(567), 1,
      sym_declaration_list,
    STATE(2487), 1,
      sym_text_interpolation,
  [77749] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2488), 1,
      sym_text_interpolation,
    ACTIONS(4444), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [77763] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2489), 1,
      sym_text_interpolation,
    ACTIONS(4446), 2,
      anon_sym_string,
      anon_sym_int,
  [77777] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2490), 1,
      sym_text_interpolation,
    ACTIONS(4448), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [77791] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3237), 1,
      anon_sym_LPAREN2,
    STATE(1820), 1,
      sym_formal_parameters,
    STATE(2491), 1,
      sym_text_interpolation,
  [77807] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2492), 1,
      sym_text_interpolation,
    ACTIONS(4450), 2,
      anon_sym_LBRACE,
      anon_sym_COLON,
  [77821] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(1867), 1,
      anon_sym_DOLLAR,
    STATE(2477), 1,
      sym_variable_name,
    STATE(2493), 1,
      sym_text_interpolation,
  [77837] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2494), 1,
      sym_text_interpolation,
    ACTIONS(2476), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [77851] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_LBRACE,
    STATE(1999), 1,
      sym_declaration_list,
    STATE(2495), 1,
      sym_text_interpolation,
  [77867] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(504), 1,
      sym_declaration_list,
    STATE(2496), 1,
      sym_text_interpolation,
  [77883] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3421), 1,
      anon_sym_LBRACE,
    STATE(478), 1,
      sym_declaration_list,
    STATE(2497), 1,
      sym_text_interpolation,
  [77899] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2498), 1,
      sym_text_interpolation,
    ACTIONS(4452), 2,
      anon_sym_string,
      anon_sym_int,
  [77913] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2072), 1,
      anon_sym_LPAREN2,
    STATE(880), 1,
      sym_arguments,
    STATE(2499), 1,
      sym_text_interpolation,
  [77929] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3663), 1,
      anon_sym_LBRACE,
    STATE(499), 1,
      sym_enum_declaration_list,
    STATE(2500), 1,
      sym_text_interpolation,
  [77945] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2501), 1,
      sym_text_interpolation,
    ACTIONS(4454), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [77959] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_LPAREN2,
    STATE(1812), 1,
      sym_formal_parameters,
    STATE(2502), 1,
      sym_text_interpolation,
  [77975] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2534), 1,
      anon_sym_RPAREN2,
    ACTIONS(4456), 1,
      anon_sym_EQ,
    STATE(2503), 1,
      sym_text_interpolation,
  [77991] = 5,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4311), 1,
      anon_sym_LBRACE,
    STATE(455), 1,
      sym_compound_statement,
    STATE(2504), 1,
      sym_text_interpolation,
  [78007] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4458), 1,
      anon_sym_RPAREN2,
    STATE(2505), 1,
      sym_text_interpolation,
  [78020] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4460), 1,
      anon_sym_BSLASH,
    STATE(2506), 1,
      sym_text_interpolation,
  [78033] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4462), 1,
      sym_name,
    STATE(2507), 1,
      sym_text_interpolation,
  [78046] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4464), 1,
      anon_sym_LBRACE,
    STATE(2508), 1,
      sym_text_interpolation,
  [78059] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4466), 1,
      anon_sym_COLON_COLON,
    STATE(2509), 1,
      sym_text_interpolation,
  [78072] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4468), 1,
      sym_heredoc_end,
    STATE(2510), 1,
      sym_text_interpolation,
  [78085] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4470), 1,
      anon_sym_LBRACE,
    STATE(2511), 1,
      sym_text_interpolation,
  [78098] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4472), 1,
      anon_sym_LPAREN2,
    STATE(2512), 1,
      sym_text_interpolation,
  [78111] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4474), 1,
      anon_sym_LBRACE,
    STATE(2513), 1,
      sym_text_interpolation,
  [78124] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3489), 1,
      sym_name,
    STATE(2514), 1,
      sym_text_interpolation,
  [78137] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4476), 1,
      aux_sym_if_statement_token2,
    STATE(2515), 1,
      sym_text_interpolation,
  [78150] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4478), 1,
      anon_sym_RPAREN,
    STATE(2516), 1,
      sym_text_interpolation,
  [78163] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4480), 1,
      anon_sym_EQ_GT,
    STATE(2517), 1,
      sym_text_interpolation,
  [78176] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4482), 1,
      aux_sym_if_statement_token2,
    STATE(2518), 1,
      sym_text_interpolation,
  [78189] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4137), 1,
      anon_sym_RPAREN2,
    STATE(2519), 1,
      sym_text_interpolation,
  [78202] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4484), 1,
      anon_sym_EQ_GT,
    STATE(2520), 1,
      sym_text_interpolation,
  [78215] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4486), 1,
      sym_name,
    STATE(2521), 1,
      sym_text_interpolation,
  [78228] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4488), 1,
      sym_name,
    STATE(2522), 1,
      sym_text_interpolation,
  [78241] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4072), 1,
      anon_sym_RBRACE,
    STATE(2523), 1,
      sym_text_interpolation,
  [78254] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4490), 1,
      sym_name,
    STATE(2524), 1,
      sym_text_interpolation,
  [78267] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4492), 1,
      anon_sym_RPAREN2,
    STATE(2525), 1,
      sym_text_interpolation,
  [78280] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4494), 1,
      anon_sym_EQ_GT,
    STATE(2526), 1,
      sym_text_interpolation,
  [78293] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4496), 1,
      anon_sym_RPAREN2,
    STATE(2527), 1,
      sym_text_interpolation,
  [78306] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4111), 1,
      sym_name,
    STATE(2528), 1,
      sym_text_interpolation,
  [78319] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4498), 1,
      aux_sym_while_statement_token1,
    STATE(2529), 1,
      sym_text_interpolation,
  [78332] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4500), 1,
      anon_sym_EQ_GT,
    STATE(2530), 1,
      sym_text_interpolation,
  [78345] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4129), 1,
      anon_sym_RPAREN2,
    STATE(2531), 1,
      sym_text_interpolation,
  [78358] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4502), 1,
      anon_sym_SEMI,
    STATE(2532), 1,
      sym_text_interpolation,
  [78371] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4504), 1,
      anon_sym_RPAREN,
    STATE(2533), 1,
      sym_text_interpolation,
  [78384] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4506), 1,
      aux_sym_while_statement_token2,
    STATE(2534), 1,
      sym_text_interpolation,
  [78397] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4508), 1,
      anon_sym_RPAREN2,
    STATE(2535), 1,
      sym_text_interpolation,
  [78410] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4510), 1,
      anon_sym_LBRACE,
    STATE(2536), 1,
      sym_text_interpolation,
  [78423] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4512), 1,
      anon_sym_LBRACE,
    STATE(2537), 1,
      sym_text_interpolation,
  [78436] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4168), 1,
      anon_sym_RPAREN2,
    STATE(2538), 1,
      sym_text_interpolation,
  [78449] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4058), 1,
      anon_sym_RPAREN2,
    STATE(2539), 1,
      sym_text_interpolation,
  [78462] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4514), 1,
      sym_heredoc_start,
    STATE(2540), 1,
      sym_text_interpolation,
  [78475] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4516), 1,
      aux_sym_foreach_statement_token2,
    STATE(2541), 1,
      sym_text_interpolation,
  [78488] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4518), 1,
      anon_sym_RPAREN2,
    STATE(2542), 1,
      sym_text_interpolation,
  [78501] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4520), 1,
      sym_name,
    STATE(2543), 1,
      sym_text_interpolation,
  [78514] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4522), 1,
      anon_sym_RPAREN2,
    STATE(2544), 1,
      sym_text_interpolation,
  [78527] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4524), 1,
      sym_name,
    STATE(2545), 1,
      sym_text_interpolation,
  [78540] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3479), 1,
      sym_name,
    STATE(2546), 1,
      sym_text_interpolation,
  [78553] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4526), 1,
      anon_sym_EQ_GT,
    STATE(2547), 1,
      sym_text_interpolation,
  [78566] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4528), 1,
      anon_sym_EQ,
    STATE(2548), 1,
      sym_text_interpolation,
  [78579] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4530), 1,
      anon_sym_RPAREN2,
    STATE(2549), 1,
      sym_text_interpolation,
  [78592] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4532), 1,
      sym_name,
    STATE(2550), 1,
      sym_text_interpolation,
  [78605] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4534), 1,
      sym_name,
    STATE(2551), 1,
      sym_text_interpolation,
  [78618] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4536), 1,
      anon_sym_RPAREN2,
    STATE(2552), 1,
      sym_text_interpolation,
  [78631] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4538), 1,
      anon_sym_EQ_GT,
    STATE(2553), 1,
      sym_text_interpolation,
  [78644] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4540), 1,
      sym_name,
    STATE(2554), 1,
      sym_text_interpolation,
  [78657] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4542), 1,
      anon_sym_RPAREN2,
    STATE(2555), 1,
      sym_text_interpolation,
  [78670] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4544), 1,
      anon_sym_BSLASH,
    STATE(2556), 1,
      sym_text_interpolation,
  [78683] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4546), 1,
      sym_name,
    STATE(2557), 1,
      sym_text_interpolation,
  [78696] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4548), 1,
      anon_sym_RPAREN2,
    STATE(2558), 1,
      sym_text_interpolation,
  [78709] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4550), 1,
      anon_sym_RBRACK,
    STATE(2559), 1,
      sym_text_interpolation,
  [78722] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4552), 1,
      anon_sym_RPAREN2,
    STATE(2560), 1,
      sym_text_interpolation,
  [78735] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4554), 1,
      sym_name,
    STATE(2561), 1,
      sym_text_interpolation,
  [78748] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4556), 1,
      anon_sym_RPAREN2,
    STATE(2562), 1,
      sym_text_interpolation,
  [78761] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4006), 1,
      anon_sym_RBRACE,
    STATE(2563), 1,
      sym_text_interpolation,
  [78774] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4558), 1,
      sym_name,
    STATE(2564), 1,
      sym_text_interpolation,
  [78787] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4560), 1,
      anon_sym_BSLASH,
    STATE(2565), 1,
      sym_text_interpolation,
  [78800] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4562), 1,
      anon_sym_LBRACE,
    STATE(2566), 1,
      sym_text_interpolation,
  [78813] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4564), 1,
      anon_sym_RPAREN2,
    STATE(2567), 1,
      sym_text_interpolation,
  [78826] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4566), 1,
      anon_sym_EQ_GT,
    STATE(2568), 1,
      sym_text_interpolation,
  [78839] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4074), 1,
      anon_sym_LBRACE,
    STATE(2569), 1,
      sym_text_interpolation,
  [78852] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4568), 1,
      anon_sym_COLON_COLON,
    STATE(2570), 1,
      sym_text_interpolation,
  [78865] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4570), 1,
      sym_heredoc_end,
    STATE(2571), 1,
      sym_text_interpolation,
  [78878] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4572), 1,
      sym_heredoc_end,
    STATE(2572), 1,
      sym_text_interpolation,
  [78891] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4574), 1,
      anon_sym_RPAREN2,
    STATE(2573), 1,
      sym_text_interpolation,
  [78904] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4576), 1,
      sym_name,
    STATE(2574), 1,
      sym_text_interpolation,
  [78917] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4578), 1,
      anon_sym_LPAREN2,
    STATE(2575), 1,
      sym_text_interpolation,
  [78930] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4580), 1,
      anon_sym_SEMI,
    STATE(2576), 1,
      sym_text_interpolation,
  [78943] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4582), 1,
      anon_sym_RPAREN2,
    STATE(2577), 1,
      sym_text_interpolation,
  [78956] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_RPAREN2,
    STATE(2578), 1,
      sym_text_interpolation,
  [78969] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4586), 1,
      anon_sym_SEMI,
    STATE(2579), 1,
      sym_text_interpolation,
  [78982] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4588), 1,
      sym_name,
    STATE(2580), 1,
      sym_text_interpolation,
  [78995] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4590), 1,
      sym_name,
    STATE(2581), 1,
      sym_text_interpolation,
  [79008] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4592), 1,
      anon_sym_SQUOTE2,
    STATE(2582), 1,
      sym_text_interpolation,
  [79021] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4594), 1,
      anon_sym_LBRACE,
    STATE(2583), 1,
      sym_text_interpolation,
  [79034] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4596), 1,
      anon_sym_LBRACE,
    STATE(2584), 1,
      sym_text_interpolation,
  [79047] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4598), 1,
      anon_sym_LPAREN2,
    STATE(2585), 1,
      sym_text_interpolation,
  [79060] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4600), 1,
      anon_sym_RPAREN2,
    STATE(2586), 1,
      sym_text_interpolation,
  [79073] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4602), 1,
      aux_sym_while_statement_token2,
    STATE(2587), 1,
      sym_text_interpolation,
  [79086] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4604), 1,
      sym_heredoc_end,
    STATE(2588), 1,
      sym_text_interpolation,
  [79099] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4606), 1,
      sym_name,
    STATE(2589), 1,
      sym_text_interpolation,
  [79112] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4608), 1,
      anon_sym_RPAREN2,
    STATE(2590), 1,
      sym_text_interpolation,
  [79125] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4610), 1,
      anon_sym_EQ,
    STATE(2591), 1,
      sym_text_interpolation,
  [79138] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4612), 1,
      anon_sym_LBRACE,
    STATE(2592), 1,
      sym_text_interpolation,
  [79151] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4614), 1,
      aux_sym_foreach_statement_token2,
    STATE(2593), 1,
      sym_text_interpolation,
  [79164] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4616), 1,
      anon_sym_COLON_COLON,
    STATE(2594), 1,
      sym_text_interpolation,
  [79177] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4618), 1,
      sym_heredoc_start,
    STATE(2595), 1,
      sym_text_interpolation,
  [79190] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4620), 1,
      anon_sym_LBRACE,
    STATE(2596), 1,
      sym_text_interpolation,
  [79203] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3996), 1,
      anon_sym_LBRACE,
    STATE(2597), 1,
      sym_text_interpolation,
  [79216] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4622), 1,
      aux_sym_if_statement_token2,
    STATE(2598), 1,
      sym_text_interpolation,
  [79229] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4624), 1,
      sym_name,
    STATE(2599), 1,
      sym_text_interpolation,
  [79242] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4626), 1,
      anon_sym_RBRACK,
    STATE(2600), 1,
      sym_text_interpolation,
  [79255] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4042), 1,
      anon_sym_RPAREN2,
    STATE(2601), 1,
      sym_text_interpolation,
  [79268] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4628), 1,
      anon_sym_EQ,
    STATE(2602), 1,
      sym_text_interpolation,
  [79281] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4630), 1,
      sym_heredoc_end,
    STATE(2603), 1,
      sym_text_interpolation,
  [79294] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4632), 1,
      sym_name,
    STATE(2604), 1,
      sym_text_interpolation,
  [79307] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3953), 1,
      anon_sym_RBRACK,
    STATE(2605), 1,
      sym_text_interpolation,
  [79320] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4634), 1,
      anon_sym_RBRACK,
    STATE(2606), 1,
      sym_text_interpolation,
  [79333] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4636), 1,
      anon_sym_SEMI,
    STATE(2607), 1,
      sym_text_interpolation,
  [79346] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4638), 1,
      anon_sym_EQ_GT,
    STATE(2608), 1,
      sym_text_interpolation,
  [79359] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4640), 1,
      anon_sym_EQ_GT,
    STATE(2609), 1,
      sym_text_interpolation,
  [79372] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4642), 1,
      sym_name,
    STATE(2610), 1,
      sym_text_interpolation,
  [79385] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4644), 1,
      sym_name,
    STATE(2611), 1,
      sym_text_interpolation,
  [79398] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4143), 1,
      anon_sym_RPAREN2,
    STATE(2612), 1,
      sym_text_interpolation,
  [79411] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4646), 1,
      sym_name,
    STATE(2613), 1,
      sym_text_interpolation,
  [79424] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4648), 1,
      sym_name,
    STATE(2614), 1,
      sym_text_interpolation,
  [79437] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4650), 1,
      sym_heredoc_start,
    STATE(2615), 1,
      sym_text_interpolation,
  [79450] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4652), 1,
      anon_sym_RPAREN2,
    STATE(2616), 1,
      sym_text_interpolation,
  [79463] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4062), 1,
      anon_sym_RPAREN2,
    STATE(2617), 1,
      sym_text_interpolation,
  [79476] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4654), 1,
      anon_sym_SEMI,
    STATE(2618), 1,
      sym_text_interpolation,
  [79489] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4656), 1,
      sym_name,
    STATE(2619), 1,
      sym_text_interpolation,
  [79502] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3679), 1,
      anon_sym_BSLASH,
    STATE(2620), 1,
      sym_text_interpolation,
  [79515] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4658), 1,
      anon_sym_RPAREN2,
    STATE(2621), 1,
      sym_text_interpolation,
  [79528] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4660), 1,
      anon_sym_BSLASH,
    STATE(2622), 1,
      sym_text_interpolation,
  [79541] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4662), 1,
      anon_sym_LBRACE,
    STATE(2623), 1,
      sym_text_interpolation,
  [79554] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4664), 1,
      anon_sym_RPAREN2,
    STATE(2624), 1,
      sym_text_interpolation,
  [79567] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4666), 1,
      anon_sym_LBRACE,
    STATE(2625), 1,
      sym_text_interpolation,
  [79580] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4668), 1,
      sym_name,
    STATE(2626), 1,
      sym_text_interpolation,
  [79593] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4670), 1,
      sym_heredoc_end,
    STATE(2627), 1,
      sym_text_interpolation,
  [79606] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4672), 1,
      sym_name,
    STATE(2628), 1,
      sym_text_interpolation,
  [79619] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4674), 1,
      sym_name,
    STATE(2629), 1,
      sym_text_interpolation,
  [79632] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4676), 1,
      anon_sym_COLON_COLON,
    STATE(2630), 1,
      sym_text_interpolation,
  [79645] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4678), 1,
      sym_heredoc_end,
    STATE(2631), 1,
      sym_text_interpolation,
  [79658] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4680), 1,
      anon_sym_LPAREN2,
    STATE(2632), 1,
      sym_text_interpolation,
  [79671] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4682), 1,
      sym_heredoc_end,
    STATE(2633), 1,
      sym_text_interpolation,
  [79684] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4684), 1,
      sym_name,
    STATE(2634), 1,
      sym_text_interpolation,
  [79697] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4686), 1,
      sym_name,
    STATE(2635), 1,
      sym_text_interpolation,
  [79710] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4688), 1,
      anon_sym_RPAREN2,
    STATE(2636), 1,
      sym_text_interpolation,
  [79723] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4403), 1,
      anon_sym_RPAREN2,
    STATE(2637), 1,
      sym_text_interpolation,
  [79736] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4690), 1,
      anon_sym_EQ_GT,
    STATE(2638), 1,
      sym_text_interpolation,
  [79749] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4692), 1,
      anon_sym_EQ,
    STATE(2639), 1,
      sym_text_interpolation,
  [79762] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4694), 1,
      anon_sym_RPAREN2,
    STATE(2640), 1,
      sym_text_interpolation,
  [79775] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4696), 1,
      anon_sym_RPAREN2,
    STATE(2641), 1,
      sym_text_interpolation,
  [79788] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4698), 1,
      anon_sym_RBRACK,
    STATE(2642), 1,
      sym_text_interpolation,
  [79801] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4700), 1,
      sym_name,
    STATE(2643), 1,
      sym_text_interpolation,
  [79814] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4702), 1,
      anon_sym_BSLASH,
    STATE(2644), 1,
      sym_text_interpolation,
  [79827] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4704), 1,
      sym_name,
    STATE(2645), 1,
      sym_text_interpolation,
  [79840] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4706), 1,
      anon_sym_SQUOTE2,
    STATE(2646), 1,
      sym_text_interpolation,
  [79853] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4708), 1,
      anon_sym_EQ,
    STATE(2647), 1,
      sym_text_interpolation,
  [79866] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4710), 1,
      anon_sym_RPAREN2,
    STATE(2648), 1,
      sym_text_interpolation,
  [79879] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4712), 1,
      anon_sym_LBRACE,
    STATE(2649), 1,
      sym_text_interpolation,
  [79892] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4714), 1,
      anon_sym_LPAREN2,
    STATE(2650), 1,
      sym_text_interpolation,
  [79905] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4716), 1,
      anon_sym_BSLASH,
    STATE(2651), 1,
      sym_text_interpolation,
  [79918] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4718), 1,
      anon_sym_RPAREN2,
    STATE(2652), 1,
      sym_text_interpolation,
  [79931] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4720), 1,
      anon_sym_EQ_GT,
    STATE(2653), 1,
      sym_text_interpolation,
  [79944] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4722), 1,
      anon_sym_RPAREN2,
    STATE(2654), 1,
      sym_text_interpolation,
  [79957] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4724), 1,
      anon_sym_RPAREN2,
    STATE(2655), 1,
      sym_text_interpolation,
  [79970] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4726), 1,
      sym_name,
    STATE(2656), 1,
      sym_text_interpolation,
  [79983] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4728), 1,
      anon_sym_LBRACE,
    STATE(2657), 1,
      sym_text_interpolation,
  [79996] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4730), 1,
      anon_sym_COLON_COLON,
    STATE(2658), 1,
      sym_text_interpolation,
  [80009] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4732), 1,
      anon_sym_RPAREN2,
    STATE(2659), 1,
      sym_text_interpolation,
  [80022] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4734), 1,
      sym_name,
    STATE(2660), 1,
      sym_text_interpolation,
  [80035] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4736), 1,
      anon_sym_LPAREN2,
    STATE(2661), 1,
      sym_text_interpolation,
  [80048] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4738), 1,
      anon_sym_LPAREN2,
    STATE(2662), 1,
      sym_text_interpolation,
  [80061] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4456), 1,
      anon_sym_EQ,
    STATE(2663), 1,
      sym_text_interpolation,
  [80074] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(3781), 1,
      anon_sym_COLON_COLON,
    STATE(2664), 1,
      sym_text_interpolation,
  [80087] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4740), 1,
      anon_sym_RPAREN2,
    STATE(2665), 1,
      sym_text_interpolation,
  [80100] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4742), 1,
      sym_heredoc_start,
    STATE(2666), 1,
      sym_text_interpolation,
  [80113] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4744), 1,
      anon_sym_RPAREN2,
    STATE(2667), 1,
      sym_text_interpolation,
  [80126] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4746), 1,
      anon_sym_LBRACE,
    STATE(2668), 1,
      sym_text_interpolation,
  [80139] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4748), 1,
      anon_sym_RPAREN2,
    STATE(2669), 1,
      sym_text_interpolation,
  [80152] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4750), 1,
      anon_sym_RPAREN2,
    STATE(2670), 1,
      sym_text_interpolation,
  [80165] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4752), 1,
      anon_sym_SEMI,
    STATE(2671), 1,
      sym_text_interpolation,
  [80178] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4754), 1,
      ts_builtin_sym_end,
    STATE(2672), 1,
      sym_text_interpolation,
  [80191] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4756), 1,
      anon_sym_RPAREN2,
    STATE(2673), 1,
      sym_text_interpolation,
  [80204] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4758), 1,
      aux_sym_if_statement_token2,
    STATE(2674), 1,
      sym_text_interpolation,
  [80217] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4760), 1,
      sym_name,
    STATE(2675), 1,
      sym_text_interpolation,
  [80230] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4762), 1,
      sym_name,
    STATE(2676), 1,
      sym_text_interpolation,
  [80243] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4764), 1,
      anon_sym_EQ_GT,
    STATE(2677), 1,
      sym_text_interpolation,
  [80256] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4766), 1,
      sym_name,
    STATE(2678), 1,
      sym_text_interpolation,
  [80269] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4768), 1,
      anon_sym_RPAREN2,
    STATE(2679), 1,
      sym_text_interpolation,
  [80282] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4036), 1,
      anon_sym_RPAREN2,
    STATE(2680), 1,
      sym_text_interpolation,
  [80295] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4770), 1,
      anon_sym_COLON_COLON,
    STATE(2681), 1,
      sym_text_interpolation,
  [80308] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4772), 1,
      anon_sym_RPAREN2,
    STATE(2682), 1,
      sym_text_interpolation,
  [80321] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4774), 1,
      anon_sym_RPAREN2,
    STATE(2683), 1,
      sym_text_interpolation,
  [80334] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4776), 1,
      anon_sym_RPAREN2,
    STATE(2684), 1,
      sym_text_interpolation,
  [80347] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4778), 1,
      sym_name,
    STATE(2685), 1,
      sym_text_interpolation,
  [80360] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4780), 1,
      anon_sym_RPAREN2,
    STATE(2686), 1,
      sym_text_interpolation,
  [80373] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4782), 1,
      sym_name,
    STATE(2687), 1,
      sym_text_interpolation,
  [80386] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4784), 1,
      sym_heredoc_end,
    STATE(2688), 1,
      sym_text_interpolation,
  [80399] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4786), 1,
      anon_sym_LBRACE,
    STATE(2689), 1,
      sym_text_interpolation,
  [80412] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4788), 1,
      anon_sym_LPAREN2,
    STATE(2690), 1,
      sym_text_interpolation,
  [80425] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4790), 1,
      anon_sym_COLON_COLON,
    STATE(2691), 1,
      sym_text_interpolation,
  [80438] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4792), 1,
      sym_name,
    STATE(2692), 1,
      sym_text_interpolation,
  [80451] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4794), 1,
      sym_name,
    STATE(2693), 1,
      sym_text_interpolation,
  [80464] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4796), 1,
      sym_name,
    STATE(2694), 1,
      sym_text_interpolation,
  [80477] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4798), 1,
      anon_sym_RPAREN2,
    STATE(2695), 1,
      sym_text_interpolation,
  [80490] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4800), 1,
      sym_name,
    STATE(2696), 1,
      sym_text_interpolation,
  [80503] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4802), 1,
      sym_name,
    STATE(2697), 1,
      sym_text_interpolation,
  [80516] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(959), 1,
      anon_sym_COLON,
    ACTIONS(1572), 1,
      sym_comment,
    STATE(2698), 1,
      sym_text_interpolation,
  [80529] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4804), 1,
      anon_sym_LPAREN2,
    STATE(2699), 1,
      sym_text_interpolation,
  [80542] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4119), 1,
      anon_sym_RPAREN2,
    STATE(2700), 1,
      sym_text_interpolation,
  [80555] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4806), 1,
      anon_sym_SEMI,
    STATE(2701), 1,
      sym_text_interpolation,
  [80568] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4808), 1,
      aux_sym_while_statement_token1,
    STATE(2702), 1,
      sym_text_interpolation,
  [80581] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4810), 1,
      sym_name,
    STATE(2703), 1,
      sym_text_interpolation,
  [80594] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4326), 1,
      anon_sym_LPAREN2,
    STATE(2704), 1,
      sym_text_interpolation,
  [80607] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4812), 1,
      anon_sym_EQ_GT,
    STATE(2705), 1,
      sym_text_interpolation,
  [80620] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4814), 1,
      sym_name,
    STATE(2706), 1,
      sym_text_interpolation,
  [80633] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4816), 1,
      sym_heredoc_end,
    STATE(2707), 1,
      sym_text_interpolation,
  [80646] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4818), 1,
      anon_sym_RPAREN2,
    STATE(2708), 1,
      sym_text_interpolation,
  [80659] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4820), 1,
      anon_sym_LPAREN2,
    STATE(2709), 1,
      sym_text_interpolation,
  [80672] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4822), 1,
      sym_name,
    STATE(2710), 1,
      sym_text_interpolation,
  [80685] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4824), 1,
      sym_name,
    STATE(2711), 1,
      sym_text_interpolation,
  [80698] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4826), 1,
      anon_sym_LPAREN2,
    STATE(2712), 1,
      sym_text_interpolation,
  [80711] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4828), 1,
      sym_name,
    STATE(2713), 1,
      sym_text_interpolation,
  [80724] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4830), 1,
      sym_name,
    STATE(2714), 1,
      sym_text_interpolation,
  [80737] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4832), 1,
      anon_sym_SEMI,
    STATE(2715), 1,
      sym_text_interpolation,
  [80750] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4834), 1,
      sym_name,
    STATE(2716), 1,
      sym_text_interpolation,
  [80763] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4836), 1,
      anon_sym_LPAREN2,
    STATE(2717), 1,
      sym_text_interpolation,
  [80776] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4838), 1,
      anon_sym_RPAREN2,
    STATE(2718), 1,
      sym_text_interpolation,
  [80789] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4840), 1,
      aux_sym_if_statement_token2,
    STATE(2719), 1,
      sym_text_interpolation,
  [80802] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4842), 1,
      anon_sym_RPAREN2,
    STATE(2720), 1,
      sym_text_interpolation,
  [80815] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4844), 1,
      anon_sym_LPAREN2,
    STATE(2721), 1,
      sym_text_interpolation,
  [80828] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4172), 1,
      anon_sym_RPAREN2,
    STATE(2722), 1,
      sym_text_interpolation,
  [80841] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4846), 1,
      anon_sym_SEMI,
    STATE(2723), 1,
      sym_text_interpolation,
  [80854] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4848), 1,
      anon_sym_RBRACK,
    STATE(2724), 1,
      sym_text_interpolation,
  [80867] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4026), 1,
      anon_sym_RBRACK,
    STATE(2725), 1,
      sym_text_interpolation,
  [80880] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4850), 1,
      anon_sym_SEMI,
    STATE(2726), 1,
      sym_text_interpolation,
  [80893] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4852), 1,
      anon_sym_LPAREN2,
    STATE(2727), 1,
      sym_text_interpolation,
  [80906] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4854), 1,
      anon_sym_EQ_GT,
    STATE(2728), 1,
      sym_text_interpolation,
  [80919] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4856), 1,
      anon_sym_LPAREN2,
    STATE(2729), 1,
      sym_text_interpolation,
  [80932] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4858), 1,
      anon_sym_LPAREN2,
    STATE(2730), 1,
      sym_text_interpolation,
  [80945] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4860), 1,
      anon_sym_BSLASH,
    STATE(2731), 1,
      sym_text_interpolation,
  [80958] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4862), 1,
      sym_name,
    STATE(2732), 1,
      sym_text_interpolation,
  [80971] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4864), 1,
      anon_sym_SEMI,
    STATE(2733), 1,
      sym_text_interpolation,
  [80984] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(2534), 1,
      anon_sym_RPAREN2,
    STATE(2734), 1,
      sym_text_interpolation,
  [80997] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4866), 1,
      anon_sym_LPAREN2,
    STATE(2735), 1,
      sym_text_interpolation,
  [81010] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4868), 1,
      anon_sym_LPAREN2,
    STATE(2736), 1,
      sym_text_interpolation,
  [81023] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4870), 1,
      sym_integer,
    STATE(2737), 1,
      sym_text_interpolation,
  [81036] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4872), 1,
      anon_sym_LPAREN2,
    STATE(2738), 1,
      sym_text_interpolation,
  [81049] = 4,
    ACTIONS(3), 1,
      anon_sym_QMARK_GT,
    ACTIONS(1572), 1,
      sym_comment,
    ACTIONS(4874), 1,
      anon_sym_SEMI,
    STATE(2739), 1,
      sym_text_interpolation,
  [81062] = 1,
    ACTIONS(4876), 1,
      ts_builtin_sym_end,
  [81066] = 1,
    ACTIONS(4878), 1,
      ts_builtin_sym_end,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(593)] = 0,
  [SMALL_STATE(594)] = 79,
  [SMALL_STATE(595)] = 158,
  [SMALL_STATE(596)] = 234,
  [SMALL_STATE(597)] = 308,
  [SMALL_STATE(598)] = 382,
  [SMALL_STATE(599)] = 456,
  [SMALL_STATE(600)] = 530,
  [SMALL_STATE(601)] = 604,
  [SMALL_STATE(602)] = 678,
  [SMALL_STATE(603)] = 754,
  [SMALL_STATE(604)] = 828,
  [SMALL_STATE(605)] = 902,
  [SMALL_STATE(606)] = 976,
  [SMALL_STATE(607)] = 1050,
  [SMALL_STATE(608)] = 1124,
  [SMALL_STATE(609)] = 1198,
  [SMALL_STATE(610)] = 1272,
  [SMALL_STATE(611)] = 1346,
  [SMALL_STATE(612)] = 1420,
  [SMALL_STATE(613)] = 1494,
  [SMALL_STATE(614)] = 1578,
  [SMALL_STATE(615)] = 1652,
  [SMALL_STATE(616)] = 1726,
  [SMALL_STATE(617)] = 1800,
  [SMALL_STATE(618)] = 1874,
  [SMALL_STATE(619)] = 1948,
  [SMALL_STATE(620)] = 2022,
  [SMALL_STATE(621)] = 2096,
  [SMALL_STATE(622)] = 2170,
  [SMALL_STATE(623)] = 2251,
  [SMALL_STATE(624)] = 2332,
  [SMALL_STATE(625)] = 2407,
  [SMALL_STATE(626)] = 2482,
  [SMALL_STATE(627)] = 2563,
  [SMALL_STATE(628)] = 2701,
  [SMALL_STATE(629)] = 2771,
  [SMALL_STATE(630)] = 2849,
  [SMALL_STATE(631)] = 2919,
  [SMALL_STATE(632)] = 2989,
  [SMALL_STATE(633)] = 3059,
  [SMALL_STATE(634)] = 3129,
  [SMALL_STATE(635)] = 3199,
  [SMALL_STATE(636)] = 3269,
  [SMALL_STATE(637)] = 3349,
  [SMALL_STATE(638)] = 3419,
  [SMALL_STATE(639)] = 3489,
  [SMALL_STATE(640)] = 3631,
  [SMALL_STATE(641)] = 3769,
  [SMALL_STATE(642)] = 3839,
  [SMALL_STATE(643)] = 3981,
  [SMALL_STATE(644)] = 4051,
  [SMALL_STATE(645)] = 4121,
  [SMALL_STATE(646)] = 4257,
  [SMALL_STATE(647)] = 4327,
  [SMALL_STATE(648)] = 4397,
  [SMALL_STATE(649)] = 4477,
  [SMALL_STATE(650)] = 4557,
  [SMALL_STATE(651)] = 4627,
  [SMALL_STATE(652)] = 4705,
  [SMALL_STATE(653)] = 4777,
  [SMALL_STATE(654)] = 4847,
  [SMALL_STATE(655)] = 4917,
  [SMALL_STATE(656)] = 4987,
  [SMALL_STATE(657)] = 5059,
  [SMALL_STATE(658)] = 5129,
  [SMALL_STATE(659)] = 5206,
  [SMALL_STATE(660)] = 5283,
  [SMALL_STATE(661)] = 5422,
  [SMALL_STATE(662)] = 5491,
  [SMALL_STATE(663)] = 5568,
  [SMALL_STATE(664)] = 5637,
  [SMALL_STATE(665)] = 5706,
  [SMALL_STATE(666)] = 5775,
  [SMALL_STATE(667)] = 5844,
  [SMALL_STATE(668)] = 5923,
  [SMALL_STATE(669)] = 6000,
  [SMALL_STATE(670)] = 6077,
  [SMALL_STATE(671)] = 6162,
  [SMALL_STATE(672)] = 6231,
  [SMALL_STATE(673)] = 6300,
  [SMALL_STATE(674)] = 6369,
  [SMALL_STATE(675)] = 6438,
  [SMALL_STATE(676)] = 6517,
  [SMALL_STATE(677)] = 6594,
  [SMALL_STATE(678)] = 6670,
  [SMALL_STATE(679)] = 6750,
  [SMALL_STATE(680)] = 6826,
  [SMALL_STATE(681)] = 6906,
  [SMALL_STATE(682)] = 6986,
  [SMALL_STATE(683)] = 7068,
  [SMALL_STATE(684)] = 7150,
  [SMALL_STATE(685)] = 7226,
  [SMALL_STATE(686)] = 7302,
  [SMALL_STATE(687)] = 7379,
  [SMALL_STATE(688)] = 7456,
  [SMALL_STATE(689)] = 7535,
  [SMALL_STATE(690)] = 7612,
  [SMALL_STATE(691)] = 7745,
  [SMALL_STATE(692)] = 7822,
  [SMALL_STATE(693)] = 7955,
  [SMALL_STATE(694)] = 8088,
  [SMALL_STATE(695)] = 8165,
  [SMALL_STATE(696)] = 8242,
  [SMALL_STATE(697)] = 8375,
  [SMALL_STATE(698)] = 8505,
  [SMALL_STATE(699)] = 8581,
  [SMALL_STATE(700)] = 8711,
  [SMALL_STATE(701)] = 8841,
  [SMALL_STATE(702)] = 8971,
  [SMALL_STATE(703)] = 9047,
  [SMALL_STATE(704)] = 9177,
  [SMALL_STATE(705)] = 9307,
  [SMALL_STATE(706)] = 9437,
  [SMALL_STATE(707)] = 9567,
  [SMALL_STATE(708)] = 9697,
  [SMALL_STATE(709)] = 9827,
  [SMALL_STATE(710)] = 9957,
  [SMALL_STATE(711)] = 10087,
  [SMALL_STATE(712)] = 10157,
  [SMALL_STATE(713)] = 10287,
  [SMALL_STATE(714)] = 10351,
  [SMALL_STATE(715)] = 10417,
  [SMALL_STATE(716)] = 10478,
  [SMALL_STATE(717)] = 10539,
  [SMALL_STATE(718)] = 10600,
  [SMALL_STATE(719)] = 10665,
  [SMALL_STATE(720)] = 10774,
  [SMALL_STATE(721)] = 10883,
  [SMALL_STATE(722)] = 10948,
  [SMALL_STATE(723)] = 11013,
  [SMALL_STATE(724)] = 11073,
  [SMALL_STATE(725)] = 11133,
  [SMALL_STATE(726)] = 11199,
  [SMALL_STATE(727)] = 11261,
  [SMALL_STATE(728)] = 11323,
  [SMALL_STATE(729)] = 11385,
  [SMALL_STATE(730)] = 11445,
  [SMALL_STATE(731)] = 11507,
  [SMALL_STATE(732)] = 11567,
  [SMALL_STATE(733)] = 11627,
  [SMALL_STATE(734)] = 11687,
  [SMALL_STATE(735)] = 11749,
  [SMALL_STATE(736)] = 11809,
  [SMALL_STATE(737)] = 11869,
  [SMALL_STATE(738)] = 11929,
  [SMALL_STATE(739)] = 11989,
  [SMALL_STATE(740)] = 12049,
  [SMALL_STATE(741)] = 12109,
  [SMALL_STATE(742)] = 12169,
  [SMALL_STATE(743)] = 12229,
  [SMALL_STATE(744)] = 12289,
  [SMALL_STATE(745)] = 12349,
  [SMALL_STATE(746)] = 12409,
  [SMALL_STATE(747)] = 12469,
  [SMALL_STATE(748)] = 12529,
  [SMALL_STATE(749)] = 12589,
  [SMALL_STATE(750)] = 12655,
  [SMALL_STATE(751)] = 12718,
  [SMALL_STATE(752)] = 12777,
  [SMALL_STATE(753)] = 12836,
  [SMALL_STATE(754)] = 12895,
  [SMALL_STATE(755)] = 12954,
  [SMALL_STATE(756)] = 13013,
  [SMALL_STATE(757)] = 13076,
  [SMALL_STATE(758)] = 13135,
  [SMALL_STATE(759)] = 13194,
  [SMALL_STATE(760)] = 13257,
  [SMALL_STATE(761)] = 13320,
  [SMALL_STATE(762)] = 13379,
  [SMALL_STATE(763)] = 13438,
  [SMALL_STATE(764)] = 13499,
  [SMALL_STATE(765)] = 13560,
  [SMALL_STATE(766)] = 13623,
  [SMALL_STATE(767)] = 13727,
  [SMALL_STATE(768)] = 13785,
  [SMALL_STATE(769)] = 13843,
  [SMALL_STATE(770)] = 13901,
  [SMALL_STATE(771)] = 13959,
  [SMALL_STATE(772)] = 14017,
  [SMALL_STATE(773)] = 14075,
  [SMALL_STATE(774)] = 14133,
  [SMALL_STATE(775)] = 14191,
  [SMALL_STATE(776)] = 14249,
  [SMALL_STATE(777)] = 14307,
  [SMALL_STATE(778)] = 14365,
  [SMALL_STATE(779)] = 14423,
  [SMALL_STATE(780)] = 14481,
  [SMALL_STATE(781)] = 14539,
  [SMALL_STATE(782)] = 14597,
  [SMALL_STATE(783)] = 14655,
  [SMALL_STATE(784)] = 14713,
  [SMALL_STATE(785)] = 14771,
  [SMALL_STATE(786)] = 14829,
  [SMALL_STATE(787)] = 14887,
  [SMALL_STATE(788)] = 14945,
  [SMALL_STATE(789)] = 15005,
  [SMALL_STATE(790)] = 15067,
  [SMALL_STATE(791)] = 15125,
  [SMALL_STATE(792)] = 15187,
  [SMALL_STATE(793)] = 15249,
  [SMALL_STATE(794)] = 15307,
  [SMALL_STATE(795)] = 15373,
  [SMALL_STATE(796)] = 15435,
  [SMALL_STATE(797)] = 15497,
  [SMALL_STATE(798)] = 15555,
  [SMALL_STATE(799)] = 15613,
  [SMALL_STATE(800)] = 15671,
  [SMALL_STATE(801)] = 15729,
  [SMALL_STATE(802)] = 15787,
  [SMALL_STATE(803)] = 15847,
  [SMALL_STATE(804)] = 15905,
  [SMALL_STATE(805)] = 15963,
  [SMALL_STATE(806)] = 16021,
  [SMALL_STATE(807)] = 16079,
  [SMALL_STATE(808)] = 16137,
  [SMALL_STATE(809)] = 16195,
  [SMALL_STATE(810)] = 16253,
  [SMALL_STATE(811)] = 16311,
  [SMALL_STATE(812)] = 16369,
  [SMALL_STATE(813)] = 16427,
  [SMALL_STATE(814)] = 16485,
  [SMALL_STATE(815)] = 16545,
  [SMALL_STATE(816)] = 16605,
  [SMALL_STATE(817)] = 16663,
  [SMALL_STATE(818)] = 16721,
  [SMALL_STATE(819)] = 16779,
  [SMALL_STATE(820)] = 16839,
  [SMALL_STATE(821)] = 16897,
  [SMALL_STATE(822)] = 16955,
  [SMALL_STATE(823)] = 17013,
  [SMALL_STATE(824)] = 17071,
  [SMALL_STATE(825)] = 17129,
  [SMALL_STATE(826)] = 17187,
  [SMALL_STATE(827)] = 17245,
  [SMALL_STATE(828)] = 17303,
  [SMALL_STATE(829)] = 17361,
  [SMALL_STATE(830)] = 17419,
  [SMALL_STATE(831)] = 17477,
  [SMALL_STATE(832)] = 17535,
  [SMALL_STATE(833)] = 17593,
  [SMALL_STATE(834)] = 17651,
  [SMALL_STATE(835)] = 17709,
  [SMALL_STATE(836)] = 17767,
  [SMALL_STATE(837)] = 17825,
  [SMALL_STATE(838)] = 17883,
  [SMALL_STATE(839)] = 17941,
  [SMALL_STATE(840)] = 17999,
  [SMALL_STATE(841)] = 18057,
  [SMALL_STATE(842)] = 18115,
  [SMALL_STATE(843)] = 18173,
  [SMALL_STATE(844)] = 18231,
  [SMALL_STATE(845)] = 18335,
  [SMALL_STATE(846)] = 18395,
  [SMALL_STATE(847)] = 18455,
  [SMALL_STATE(848)] = 18513,
  [SMALL_STATE(849)] = 18571,
  [SMALL_STATE(850)] = 18629,
  [SMALL_STATE(851)] = 18730,
  [SMALL_STATE(852)] = 18789,
  [SMALL_STATE(853)] = 18848,
  [SMALL_STATE(854)] = 18907,
  [SMALL_STATE(855)] = 18968,
  [SMALL_STATE(856)] = 19029,
  [SMALL_STATE(857)] = 19088,
  [SMALL_STATE(858)] = 19147,
  [SMALL_STATE(859)] = 19208,
  [SMALL_STATE(860)] = 19269,
  [SMALL_STATE(861)] = 19330,
  [SMALL_STATE(862)] = 19431,
  [SMALL_STATE(863)] = 19532,
  [SMALL_STATE(864)] = 19589,
  [SMALL_STATE(865)] = 19690,
  [SMALL_STATE(866)] = 19748,
  [SMALL_STATE(867)] = 19804,
  [SMALL_STATE(868)] = 19860,
  [SMALL_STATE(869)] = 19916,
  [SMALL_STATE(870)] = 19972,
  [SMALL_STATE(871)] = 20028,
  [SMALL_STATE(872)] = 20088,
  [SMALL_STATE(873)] = 20150,
  [SMALL_STATE(874)] = 20206,
  [SMALL_STATE(875)] = 20262,
  [SMALL_STATE(876)] = 20318,
  [SMALL_STATE(877)] = 20378,
  [SMALL_STATE(878)] = 20434,
  [SMALL_STATE(879)] = 20490,
  [SMALL_STATE(880)] = 20546,
  [SMALL_STATE(881)] = 20602,
  [SMALL_STATE(882)] = 20658,
  [SMALL_STATE(883)] = 20774,
  [SMALL_STATE(884)] = 20830,
  [SMALL_STATE(885)] = 20886,
  [SMALL_STATE(886)] = 21002,
  [SMALL_STATE(887)] = 21100,
  [SMALL_STATE(888)] = 21162,
  [SMALL_STATE(889)] = 21218,
  [SMALL_STATE(890)] = 21276,
  [SMALL_STATE(891)] = 21332,
  [SMALL_STATE(892)] = 21448,
  [SMALL_STATE(893)] = 21504,
  [SMALL_STATE(894)] = 21560,
  [SMALL_STATE(895)] = 21616,
  [SMALL_STATE(896)] = 21673,
  [SMALL_STATE(897)] = 21728,
  [SMALL_STATE(898)] = 21787,
  [SMALL_STATE(899)] = 21846,
  [SMALL_STATE(900)] = 21901,
  [SMALL_STATE(901)] = 21960,
  [SMALL_STATE(902)] = 22019,
  [SMALL_STATE(903)] = 22074,
  [SMALL_STATE(904)] = 22129,
  [SMALL_STATE(905)] = 22184,
  [SMALL_STATE(906)] = 22239,
  [SMALL_STATE(907)] = 22294,
  [SMALL_STATE(908)] = 22349,
  [SMALL_STATE(909)] = 22404,
  [SMALL_STATE(910)] = 22459,
  [SMALL_STATE(911)] = 22516,
  [SMALL_STATE(912)] = 22573,
  [SMALL_STATE(913)] = 22628,
  [SMALL_STATE(914)] = 22683,
  [SMALL_STATE(915)] = 22738,
  [SMALL_STATE(916)] = 22793,
  [SMALL_STATE(917)] = 22850,
  [SMALL_STATE(918)] = 22909,
  [SMALL_STATE(919)] = 22963,
  [SMALL_STATE(920)] = 23017,
  [SMALL_STATE(921)] = 23071,
  [SMALL_STATE(922)] = 23125,
  [SMALL_STATE(923)] = 23179,
  [SMALL_STATE(924)] = 23233,
  [SMALL_STATE(925)] = 23287,
  [SMALL_STATE(926)] = 23341,
  [SMALL_STATE(927)] = 23395,
  [SMALL_STATE(928)] = 23453,
  [SMALL_STATE(929)] = 23507,
  [SMALL_STATE(930)] = 23561,
  [SMALL_STATE(931)] = 23615,
  [SMALL_STATE(932)] = 23669,
  [SMALL_STATE(933)] = 23723,
  [SMALL_STATE(934)] = 23777,
  [SMALL_STATE(935)] = 23839,
  [SMALL_STATE(936)] = 23895,
  [SMALL_STATE(937)] = 23951,
  [SMALL_STATE(938)] = 24005,
  [SMALL_STATE(939)] = 24059,
  [SMALL_STATE(940)] = 24117,
  [SMALL_STATE(941)] = 24173,
  [SMALL_STATE(942)] = 24229,
  [SMALL_STATE(943)] = 24283,
  [SMALL_STATE(944)] = 24339,
  [SMALL_STATE(945)] = 24393,
  [SMALL_STATE(946)] = 24447,
  [SMALL_STATE(947)] = 24501,
  [SMALL_STATE(948)] = 24555,
  [SMALL_STATE(949)] = 24613,
  [SMALL_STATE(950)] = 24667,
  [SMALL_STATE(951)] = 24721,
  [SMALL_STATE(952)] = 24777,
  [SMALL_STATE(953)] = 24831,
  [SMALL_STATE(954)] = 24885,
  [SMALL_STATE(955)] = 24943,
  [SMALL_STATE(956)] = 24997,
  [SMALL_STATE(957)] = 25051,
  [SMALL_STATE(958)] = 25105,
  [SMALL_STATE(959)] = 25159,
  [SMALL_STATE(960)] = 25213,
  [SMALL_STATE(961)] = 25267,
  [SMALL_STATE(962)] = 25321,
  [SMALL_STATE(963)] = 25375,
  [SMALL_STATE(964)] = 25429,
  [SMALL_STATE(965)] = 25483,
  [SMALL_STATE(966)] = 25537,
  [SMALL_STATE(967)] = 25591,
  [SMALL_STATE(968)] = 25645,
  [SMALL_STATE(969)] = 25699,
  [SMALL_STATE(970)] = 25755,
  [SMALL_STATE(971)] = 25809,
  [SMALL_STATE(972)] = 25871,
  [SMALL_STATE(973)] = 25925,
  [SMALL_STATE(974)] = 25979,
  [SMALL_STATE(975)] = 26033,
  [SMALL_STATE(976)] = 26087,
  [SMALL_STATE(977)] = 26141,
  [SMALL_STATE(978)] = 26203,
  [SMALL_STATE(979)] = 26257,
  [SMALL_STATE(980)] = 26311,
  [SMALL_STATE(981)] = 26365,
  [SMALL_STATE(982)] = 26423,
  [SMALL_STATE(983)] = 26477,
  [SMALL_STATE(984)] = 26531,
  [SMALL_STATE(985)] = 26585,
  [SMALL_STATE(986)] = 26639,
  [SMALL_STATE(987)] = 26693,
  [SMALL_STATE(988)] = 26747,
  [SMALL_STATE(989)] = 26801,
  [SMALL_STATE(990)] = 26855,
  [SMALL_STATE(991)] = 26909,
  [SMALL_STATE(992)] = 26963,
  [SMALL_STATE(993)] = 27017,
  [SMALL_STATE(994)] = 27079,
  [SMALL_STATE(995)] = 27133,
  [SMALL_STATE(996)] = 27189,
  [SMALL_STATE(997)] = 27247,
  [SMALL_STATE(998)] = 27301,
  [SMALL_STATE(999)] = 27355,
  [SMALL_STATE(1000)] = 27409,
  [SMALL_STATE(1001)] = 27463,
  [SMALL_STATE(1002)] = 27517,
  [SMALL_STATE(1003)] = 27571,
  [SMALL_STATE(1004)] = 27625,
  [SMALL_STATE(1005)] = 27679,
  [SMALL_STATE(1006)] = 27733,
  [SMALL_STATE(1007)] = 27787,
  [SMALL_STATE(1008)] = 27841,
  [SMALL_STATE(1009)] = 27895,
  [SMALL_STATE(1010)] = 27948,
  [SMALL_STATE(1011)] = 28003,
  [SMALL_STATE(1012)] = 28058,
  [SMALL_STATE(1013)] = 28113,
  [SMALL_STATE(1014)] = 28166,
  [SMALL_STATE(1015)] = 28219,
  [SMALL_STATE(1016)] = 28272,
  [SMALL_STATE(1017)] = 28325,
  [SMALL_STATE(1018)] = 28380,
  [SMALL_STATE(1019)] = 28433,
  [SMALL_STATE(1020)] = 28486,
  [SMALL_STATE(1021)] = 28543,
  [SMALL_STATE(1022)] = 28596,
  [SMALL_STATE(1023)] = 28649,
  [SMALL_STATE(1024)] = 28702,
  [SMALL_STATE(1025)] = 28755,
  [SMALL_STATE(1026)] = 28808,
  [SMALL_STATE(1027)] = 28861,
  [SMALL_STATE(1028)] = 28914,
  [SMALL_STATE(1029)] = 28967,
  [SMALL_STATE(1030)] = 29020,
  [SMALL_STATE(1031)] = 29073,
  [SMALL_STATE(1032)] = 29126,
  [SMALL_STATE(1033)] = 29181,
  [SMALL_STATE(1034)] = 29234,
  [SMALL_STATE(1035)] = 29287,
  [SMALL_STATE(1036)] = 29342,
  [SMALL_STATE(1037)] = 29395,
  [SMALL_STATE(1038)] = 29448,
  [SMALL_STATE(1039)] = 29503,
  [SMALL_STATE(1040)] = 29560,
  [SMALL_STATE(1041)] = 29613,
  [SMALL_STATE(1042)] = 29666,
  [SMALL_STATE(1043)] = 29721,
  [SMALL_STATE(1044)] = 29773,
  [SMALL_STATE(1045)] = 29825,
  [SMALL_STATE(1046)] = 29877,
  [SMALL_STATE(1047)] = 29932,
  [SMALL_STATE(1048)] = 29983,
  [SMALL_STATE(1049)] = 30036,
  [SMALL_STATE(1050)] = 30092,
  [SMALL_STATE(1051)] = 30180,
  [SMALL_STATE(1052)] = 30230,
  [SMALL_STATE(1053)] = 30320,
  [SMALL_STATE(1054)] = 30408,
  [SMALL_STATE(1055)] = 30496,
  [SMALL_STATE(1056)] = 30584,
  [SMALL_STATE(1057)] = 30672,
  [SMALL_STATE(1058)] = 30760,
  [SMALL_STATE(1059)] = 30848,
  [SMALL_STATE(1060)] = 30936,
  [SMALL_STATE(1061)] = 31006,
  [SMALL_STATE(1062)] = 31080,
  [SMALL_STATE(1063)] = 31162,
  [SMALL_STATE(1064)] = 31244,
  [SMALL_STATE(1065)] = 31296,
  [SMALL_STATE(1066)] = 31378,
  [SMALL_STATE(1067)] = 31464,
  [SMALL_STATE(1068)] = 31548,
  [SMALL_STATE(1069)] = 31626,
  [SMALL_STATE(1070)] = 31702,
  [SMALL_STATE(1071)] = 31774,
  [SMALL_STATE(1072)] = 31840,
  [SMALL_STATE(1073)] = 31890,
  [SMALL_STATE(1074)] = 31952,
  [SMALL_STATE(1075)] = 32010,
  [SMALL_STATE(1076)] = 32070,
  [SMALL_STATE(1077)] = 32122,
  [SMALL_STATE(1078)] = 32204,
  [SMALL_STATE(1079)] = 32286,
  [SMALL_STATE(1080)] = 32374,
  [SMALL_STATE(1081)] = 32462,
  [SMALL_STATE(1082)] = 32544,
  [SMALL_STATE(1083)] = 32626,
  [SMALL_STATE(1084)] = 32713,
  [SMALL_STATE(1085)] = 32762,
  [SMALL_STATE(1086)] = 32811,
  [SMALL_STATE(1087)] = 32898,
  [SMALL_STATE(1088)] = 32947,
  [SMALL_STATE(1089)] = 33012,
  [SMALL_STATE(1090)] = 33061,
  [SMALL_STATE(1091)] = 33110,
  [SMALL_STATE(1092)] = 33159,
  [SMALL_STATE(1093)] = 33248,
  [SMALL_STATE(1094)] = 33297,
  [SMALL_STATE(1095)] = 33346,
  [SMALL_STATE(1096)] = 33433,
  [SMALL_STATE(1097)] = 33520,
  [SMALL_STATE(1098)] = 33607,
  [SMALL_STATE(1099)] = 33694,
  [SMALL_STATE(1100)] = 33743,
  [SMALL_STATE(1101)] = 33792,
  [SMALL_STATE(1102)] = 33879,
  [SMALL_STATE(1103)] = 33928,
  [SMALL_STATE(1104)] = 33977,
  [SMALL_STATE(1105)] = 34058,
  [SMALL_STATE(1106)] = 34107,
  [SMALL_STATE(1107)] = 34194,
  [SMALL_STATE(1108)] = 34243,
  [SMALL_STATE(1109)] = 34292,
  [SMALL_STATE(1110)] = 34379,
  [SMALL_STATE(1111)] = 34428,
  [SMALL_STATE(1112)] = 34485,
  [SMALL_STATE(1113)] = 34534,
  [SMALL_STATE(1114)] = 34583,
  [SMALL_STATE(1115)] = 34632,
  [SMALL_STATE(1116)] = 34683,
  [SMALL_STATE(1117)] = 34770,
  [SMALL_STATE(1118)] = 34839,
  [SMALL_STATE(1119)] = 34912,
  [SMALL_STATE(1120)] = 34967,
  [SMALL_STATE(1121)] = 35048,
  [SMALL_STATE(1122)] = 35099,
  [SMALL_STATE(1123)] = 35180,
  [SMALL_STATE(1124)] = 35265,
  [SMALL_STATE(1125)] = 35348,
  [SMALL_STATE(1126)] = 35425,
  [SMALL_STATE(1127)] = 35500,
  [SMALL_STATE(1128)] = 35571,
  [SMALL_STATE(1129)] = 35636,
  [SMALL_STATE(1130)] = 35697,
  [SMALL_STATE(1131)] = 35754,
  [SMALL_STATE(1132)] = 35803,
  [SMALL_STATE(1133)] = 35854,
  [SMALL_STATE(1134)] = 35903,
  [SMALL_STATE(1135)] = 35984,
  [SMALL_STATE(1136)] = 36065,
  [SMALL_STATE(1137)] = 36114,
  [SMALL_STATE(1138)] = 36163,
  [SMALL_STATE(1139)] = 36250,
  [SMALL_STATE(1140)] = 36331,
  [SMALL_STATE(1141)] = 36418,
  [SMALL_STATE(1142)] = 36487,
  [SMALL_STATE(1143)] = 36560,
  [SMALL_STATE(1144)] = 36615,
  [SMALL_STATE(1145)] = 36696,
  [SMALL_STATE(1146)] = 36747,
  [SMALL_STATE(1147)] = 36828,
  [SMALL_STATE(1148)] = 36913,
  [SMALL_STATE(1149)] = 36996,
  [SMALL_STATE(1150)] = 37073,
  [SMALL_STATE(1151)] = 37148,
  [SMALL_STATE(1152)] = 37219,
  [SMALL_STATE(1153)] = 37284,
  [SMALL_STATE(1154)] = 37345,
  [SMALL_STATE(1155)] = 37402,
  [SMALL_STATE(1156)] = 37461,
  [SMALL_STATE(1157)] = 37512,
  [SMALL_STATE(1158)] = 37593,
  [SMALL_STATE(1159)] = 37674,
  [SMALL_STATE(1160)] = 37755,
  [SMALL_STATE(1161)] = 37836,
  [SMALL_STATE(1162)] = 37917,
  [SMALL_STATE(1163)] = 37998,
  [SMALL_STATE(1164)] = 38055,
  [SMALL_STATE(1165)] = 38142,
  [SMALL_STATE(1166)] = 38223,
  [SMALL_STATE(1167)] = 38310,
  [SMALL_STATE(1168)] = 38397,
  [SMALL_STATE(1169)] = 38484,
  [SMALL_STATE(1170)] = 38573,
  [SMALL_STATE(1171)] = 38660,
  [SMALL_STATE(1172)] = 38747,
  [SMALL_STATE(1173)] = 38796,
  [SMALL_STATE(1174)] = 38883,
  [SMALL_STATE(1175)] = 38970,
  [SMALL_STATE(1176)] = 39019,
  [SMALL_STATE(1177)] = 39068,
  [SMALL_STATE(1178)] = 39127,
  [SMALL_STATE(1179)] = 39213,
  [SMALL_STATE(1180)] = 39299,
  [SMALL_STATE(1181)] = 39385,
  [SMALL_STATE(1182)] = 39471,
  [SMALL_STATE(1183)] = 39539,
  [SMALL_STATE(1184)] = 39611,
  [SMALL_STATE(1185)] = 39665,
  [SMALL_STATE(1186)] = 39745,
  [SMALL_STATE(1187)] = 39795,
  [SMALL_STATE(1188)] = 39875,
  [SMALL_STATE(1189)] = 39959,
  [SMALL_STATE(1190)] = 40045,
  [SMALL_STATE(1191)] = 40121,
  [SMALL_STATE(1192)] = 40195,
  [SMALL_STATE(1193)] = 40265,
  [SMALL_STATE(1194)] = 40329,
  [SMALL_STATE(1195)] = 40389,
  [SMALL_STATE(1196)] = 40445,
  [SMALL_STATE(1197)] = 40503,
  [SMALL_STATE(1198)] = 40553,
  [SMALL_STATE(1199)] = 40633,
  [SMALL_STATE(1200)] = 40713,
  [SMALL_STATE(1201)] = 40793,
  [SMALL_STATE(1202)] = 40873,
  [SMALL_STATE(1203)] = 40953,
  [SMALL_STATE(1204)] = 41039,
  [SMALL_STATE(1205)] = 41125,
  [SMALL_STATE(1206)] = 41211,
  [SMALL_STATE(1207)] = 41299,
  [SMALL_STATE(1208)] = 41385,
  [SMALL_STATE(1209)] = 41471,
  [SMALL_STATE(1210)] = 41553,
  [SMALL_STATE(1211)] = 41642,
  [SMALL_STATE(1212)] = 41727,
  [SMALL_STATE(1213)] = 41816,
  [SMALL_STATE(1214)] = 41905,
  [SMALL_STATE(1215)] = 41994,
  [SMALL_STATE(1216)] = 42043,
  [SMALL_STATE(1217)] = 42092,
  [SMALL_STATE(1218)] = 42181,
  [SMALL_STATE(1219)] = 42270,
  [SMALL_STATE(1220)] = 42359,
  [SMALL_STATE(1221)] = 42448,
  [SMALL_STATE(1222)] = 42520,
  [SMALL_STATE(1223)] = 42606,
  [SMALL_STATE(1224)] = 42692,
  [SMALL_STATE(1225)] = 42778,
  [SMALL_STATE(1226)] = 42864,
  [SMALL_STATE(1227)] = 42950,
  [SMALL_STATE(1228)] = 43022,
  [SMALL_STATE(1229)] = 43094,
  [SMALL_STATE(1230)] = 43180,
  [SMALL_STATE(1231)] = 43254,
  [SMALL_STATE(1232)] = 43342,
  [SMALL_STATE(1233)] = 43428,
  [SMALL_STATE(1234)] = 43514,
  [SMALL_STATE(1235)] = 43600,
  [SMALL_STATE(1236)] = 43686,
  [SMALL_STATE(1237)] = 43758,
  [SMALL_STATE(1238)] = 43844,
  [SMALL_STATE(1239)] = 43930,
  [SMALL_STATE(1240)] = 44014,
  [SMALL_STATE(1241)] = 44100,
  [SMALL_STATE(1242)] = 44186,
  [SMALL_STATE(1243)] = 44260,
  [SMALL_STATE(1244)] = 44346,
  [SMALL_STATE(1245)] = 44430,
  [SMALL_STATE(1246)] = 44516,
  [SMALL_STATE(1247)] = 44599,
  [SMALL_STATE(1248)] = 44688,
  [SMALL_STATE(1249)] = 44777,
  [SMALL_STATE(1250)] = 44860,
  [SMALL_STATE(1251)] = 44943,
  [SMALL_STATE(1252)] = 45032,
  [SMALL_STATE(1253)] = 45117,
  [SMALL_STATE(1254)] = 45202,
  [SMALL_STATE(1255)] = 45287,
  [SMALL_STATE(1256)] = 45370,
  [SMALL_STATE(1257)] = 45453,
  [SMALL_STATE(1258)] = 45536,
  [SMALL_STATE(1259)] = 45619,
  [SMALL_STATE(1260)] = 45702,
  [SMALL_STATE(1261)] = 45785,
  [SMALL_STATE(1262)] = 45868,
  [SMALL_STATE(1263)] = 45951,
  [SMALL_STATE(1264)] = 46040,
  [SMALL_STATE(1265)] = 46129,
  [SMALL_STATE(1266)] = 46212,
  [SMALL_STATE(1267)] = 46297,
  [SMALL_STATE(1268)] = 46380,
  [SMALL_STATE(1269)] = 46463,
  [SMALL_STATE(1270)] = 46550,
  [SMALL_STATE(1271)] = 46633,
  [SMALL_STATE(1272)] = 46722,
  [SMALL_STATE(1273)] = 46807,
  [SMALL_STATE(1274)] = 46890,
  [SMALL_STATE(1275)] = 46972,
  [SMALL_STATE(1276)] = 47054,
  [SMALL_STATE(1277)] = 47122,
  [SMALL_STATE(1278)] = 47204,
  [SMALL_STATE(1279)] = 47286,
  [SMALL_STATE(1280)] = 47368,
  [SMALL_STATE(1281)] = 47414,
  [SMALL_STATE(1282)] = 47496,
  [SMALL_STATE(1283)] = 47578,
  [SMALL_STATE(1284)] = 47660,
  [SMALL_STATE(1285)] = 47742,
  [SMALL_STATE(1286)] = 47824,
  [SMALL_STATE(1287)] = 47906,
  [SMALL_STATE(1288)] = 47988,
  [SMALL_STATE(1289)] = 48070,
  [SMALL_STATE(1290)] = 48152,
  [SMALL_STATE(1291)] = 48234,
  [SMALL_STATE(1292)] = 48316,
  [SMALL_STATE(1293)] = 48398,
  [SMALL_STATE(1294)] = 48480,
  [SMALL_STATE(1295)] = 48562,
  [SMALL_STATE(1296)] = 48644,
  [SMALL_STATE(1297)] = 48726,
  [SMALL_STATE(1298)] = 48808,
  [SMALL_STATE(1299)] = 48890,
  [SMALL_STATE(1300)] = 48972,
  [SMALL_STATE(1301)] = 49054,
  [SMALL_STATE(1302)] = 49136,
  [SMALL_STATE(1303)] = 49218,
  [SMALL_STATE(1304)] = 49300,
  [SMALL_STATE(1305)] = 49382,
  [SMALL_STATE(1306)] = 49464,
  [SMALL_STATE(1307)] = 49546,
  [SMALL_STATE(1308)] = 49628,
  [SMALL_STATE(1309)] = 49710,
  [SMALL_STATE(1310)] = 49792,
  [SMALL_STATE(1311)] = 49874,
  [SMALL_STATE(1312)] = 49956,
  [SMALL_STATE(1313)] = 50038,
  [SMALL_STATE(1314)] = 50120,
  [SMALL_STATE(1315)] = 50202,
  [SMALL_STATE(1316)] = 50284,
  [SMALL_STATE(1317)] = 50366,
  [SMALL_STATE(1318)] = 50448,
  [SMALL_STATE(1319)] = 50530,
  [SMALL_STATE(1320)] = 50612,
  [SMALL_STATE(1321)] = 50694,
  [SMALL_STATE(1322)] = 50776,
  [SMALL_STATE(1323)] = 50844,
  [SMALL_STATE(1324)] = 50926,
  [SMALL_STATE(1325)] = 51008,
  [SMALL_STATE(1326)] = 51090,
  [SMALL_STATE(1327)] = 51172,
  [SMALL_STATE(1328)] = 51254,
  [SMALL_STATE(1329)] = 51297,
  [SMALL_STATE(1330)] = 51378,
  [SMALL_STATE(1331)] = 51459,
  [SMALL_STATE(1332)] = 51502,
  [SMALL_STATE(1333)] = 51545,
  [SMALL_STATE(1334)] = 51624,
  [SMALL_STATE(1335)] = 51705,
  [SMALL_STATE(1336)] = 51748,
  [SMALL_STATE(1337)] = 51829,
  [SMALL_STATE(1338)] = 51871,
  [SMALL_STATE(1339)] = 51913,
  [SMALL_STATE(1340)] = 51955,
  [SMALL_STATE(1341)] = 52016,
  [SMALL_STATE(1342)] = 52077,
  [SMALL_STATE(1343)] = 52122,
  [SMALL_STATE(1344)] = 52165,
  [SMALL_STATE(1345)] = 52202,
  [SMALL_STATE(1346)] = 52239,
  [SMALL_STATE(1347)] = 52276,
  [SMALL_STATE(1348)] = 52313,
  [SMALL_STATE(1349)] = 52363,
  [SMALL_STATE(1350)] = 52413,
  [SMALL_STATE(1351)] = 52463,
  [SMALL_STATE(1352)] = 52529,
  [SMALL_STATE(1353)] = 52579,
  [SMALL_STATE(1354)] = 52629,
  [SMALL_STATE(1355)] = 52695,
  [SMALL_STATE(1356)] = 52761,
  [SMALL_STATE(1357)] = 52827,
  [SMALL_STATE(1358)] = 52877,
  [SMALL_STATE(1359)] = 52927,
  [SMALL_STATE(1360)] = 52977,
  [SMALL_STATE(1361)] = 53027,
  [SMALL_STATE(1362)] = 53077,
  [SMALL_STATE(1363)] = 53127,
  [SMALL_STATE(1364)] = 53174,
  [SMALL_STATE(1365)] = 53215,
  [SMALL_STATE(1366)] = 53262,
  [SMALL_STATE(1367)] = 53309,
  [SMALL_STATE(1368)] = 53356,
  [SMALL_STATE(1369)] = 53395,
  [SMALL_STATE(1370)] = 53442,
  [SMALL_STATE(1371)] = 53489,
  [SMALL_STATE(1372)] = 53531,
  [SMALL_STATE(1373)] = 53573,
  [SMALL_STATE(1374)] = 53606,
  [SMALL_STATE(1375)] = 53639,
  [SMALL_STATE(1376)] = 53672,
  [SMALL_STATE(1377)] = 53729,
  [SMALL_STATE(1378)] = 53762,
  [SMALL_STATE(1379)] = 53795,
  [SMALL_STATE(1380)] = 53835,
  [SMALL_STATE(1381)] = 53875,
  [SMALL_STATE(1382)] = 53927,
  [SMALL_STATE(1383)] = 53967,
  [SMALL_STATE(1384)] = 54016,
  [SMALL_STATE(1385)] = 54063,
  [SMALL_STATE(1386)] = 54112,
  [SMALL_STATE(1387)] = 54161,
  [SMALL_STATE(1388)] = 54210,
  [SMALL_STATE(1389)] = 54260,
  [SMALL_STATE(1390)] = 54306,
  [SMALL_STATE(1391)] = 54352,
  [SMALL_STATE(1392)] = 54398,
  [SMALL_STATE(1393)] = 54444,
  [SMALL_STATE(1394)] = 54490,
  [SMALL_STATE(1395)] = 54536,
  [SMALL_STATE(1396)] = 54582,
  [SMALL_STATE(1397)] = 54634,
  [SMALL_STATE(1398)] = 54680,
  [SMALL_STATE(1399)] = 54726,
  [SMALL_STATE(1400)] = 54756,
  [SMALL_STATE(1401)] = 54805,
  [SMALL_STATE(1402)] = 54852,
  [SMALL_STATE(1403)] = 54899,
  [SMALL_STATE(1404)] = 54944,
  [SMALL_STATE(1405)] = 54970,
  [SMALL_STATE(1406)] = 54996,
  [SMALL_STATE(1407)] = 55022,
  [SMALL_STATE(1408)] = 55048,
  [SMALL_STATE(1409)] = 55074,
  [SMALL_STATE(1410)] = 55100,
  [SMALL_STATE(1411)] = 55126,
  [SMALL_STATE(1412)] = 55170,
  [SMALL_STATE(1413)] = 55214,
  [SMALL_STATE(1414)] = 55260,
  [SMALL_STATE(1415)] = 55306,
  [SMALL_STATE(1416)] = 55352,
  [SMALL_STATE(1417)] = 55396,
  [SMALL_STATE(1418)] = 55422,
  [SMALL_STATE(1419)] = 55468,
  [SMALL_STATE(1420)] = 55514,
  [SMALL_STATE(1421)] = 55540,
  [SMALL_STATE(1422)] = 55566,
  [SMALL_STATE(1423)] = 55592,
  [SMALL_STATE(1424)] = 55638,
  [SMALL_STATE(1425)] = 55664,
  [SMALL_STATE(1426)] = 55710,
  [SMALL_STATE(1427)] = 55756,
  [SMALL_STATE(1428)] = 55800,
  [SMALL_STATE(1429)] = 55825,
  [SMALL_STATE(1430)] = 55850,
  [SMALL_STATE(1431)] = 55875,
  [SMALL_STATE(1432)] = 55916,
  [SMALL_STATE(1433)] = 55941,
  [SMALL_STATE(1434)] = 55966,
  [SMALL_STATE(1435)] = 55991,
  [SMALL_STATE(1436)] = 56016,
  [SMALL_STATE(1437)] = 56041,
  [SMALL_STATE(1438)] = 56066,
  [SMALL_STATE(1439)] = 56091,
  [SMALL_STATE(1440)] = 56116,
  [SMALL_STATE(1441)] = 56141,
  [SMALL_STATE(1442)] = 56172,
  [SMALL_STATE(1443)] = 56197,
  [SMALL_STATE(1444)] = 56222,
  [SMALL_STATE(1445)] = 56251,
  [SMALL_STATE(1446)] = 56276,
  [SMALL_STATE(1447)] = 56317,
  [SMALL_STATE(1448)] = 56360,
  [SMALL_STATE(1449)] = 56385,
  [SMALL_STATE(1450)] = 56410,
  [SMALL_STATE(1451)] = 56435,
  [SMALL_STATE(1452)] = 56460,
  [SMALL_STATE(1453)] = 56485,
  [SMALL_STATE(1454)] = 56510,
  [SMALL_STATE(1455)] = 56535,
  [SMALL_STATE(1456)] = 56576,
  [SMALL_STATE(1457)] = 56601,
  [SMALL_STATE(1458)] = 56641,
  [SMALL_STATE(1459)] = 56679,
  [SMALL_STATE(1460)] = 56725,
  [SMALL_STATE(1461)] = 56771,
  [SMALL_STATE(1462)] = 56811,
  [SMALL_STATE(1463)] = 56851,
  [SMALL_STATE(1464)] = 56889,
  [SMALL_STATE(1465)] = 56927,
  [SMALL_STATE(1466)] = 56955,
  [SMALL_STATE(1467)] = 56993,
  [SMALL_STATE(1468)] = 57033,
  [SMALL_STATE(1469)] = 57079,
  [SMALL_STATE(1470)] = 57125,
  [SMALL_STATE(1471)] = 57165,
  [SMALL_STATE(1472)] = 57205,
  [SMALL_STATE(1473)] = 57243,
  [SMALL_STATE(1474)] = 57281,
  [SMALL_STATE(1475)] = 57319,
  [SMALL_STATE(1476)] = 57357,
  [SMALL_STATE(1477)] = 57395,
  [SMALL_STATE(1478)] = 57433,
  [SMALL_STATE(1479)] = 57473,
  [SMALL_STATE(1480)] = 57513,
  [SMALL_STATE(1481)] = 57551,
  [SMALL_STATE(1482)] = 57589,
  [SMALL_STATE(1483)] = 57627,
  [SMALL_STATE(1484)] = 57665,
  [SMALL_STATE(1485)] = 57705,
  [SMALL_STATE(1486)] = 57743,
  [SMALL_STATE(1487)] = 57781,
  [SMALL_STATE(1488)] = 57821,
  [SMALL_STATE(1489)] = 57859,
  [SMALL_STATE(1490)] = 57903,
  [SMALL_STATE(1491)] = 57941,
  [SMALL_STATE(1492)] = 57981,
  [SMALL_STATE(1493)] = 58019,
  [SMALL_STATE(1494)] = 58057,
  [SMALL_STATE(1495)] = 58093,
  [SMALL_STATE(1496)] = 58118,
  [SMALL_STATE(1497)] = 58143,
  [SMALL_STATE(1498)] = 58168,
  [SMALL_STATE(1499)] = 58193,
  [SMALL_STATE(1500)] = 58216,
  [SMALL_STATE(1501)] = 58253,
  [SMALL_STATE(1502)] = 58290,
  [SMALL_STATE(1503)] = 58327,
  [SMALL_STATE(1504)] = 58354,
  [SMALL_STATE(1505)] = 58380,
  [SMALL_STATE(1506)] = 58414,
  [SMALL_STATE(1507)] = 58438,
  [SMALL_STATE(1508)] = 58466,
  [SMALL_STATE(1509)] = 58490,
  [SMALL_STATE(1510)] = 58526,
  [SMALL_STATE(1511)] = 58554,
  [SMALL_STATE(1512)] = 58582,
  [SMALL_STATE(1513)] = 58606,
  [SMALL_STATE(1514)] = 58634,
  [SMALL_STATE(1515)] = 58658,
  [SMALL_STATE(1516)] = 58694,
  [SMALL_STATE(1517)] = 58728,
  [SMALL_STATE(1518)] = 58762,
  [SMALL_STATE(1519)] = 58796,
  [SMALL_STATE(1520)] = 58824,
  [SMALL_STATE(1521)] = 58852,
  [SMALL_STATE(1522)] = 58876,
  [SMALL_STATE(1523)] = 58910,
  [SMALL_STATE(1524)] = 58934,
  [SMALL_STATE(1525)] = 58964,
  [SMALL_STATE(1526)] = 58990,
  [SMALL_STATE(1527)] = 59016,
  [SMALL_STATE(1528)] = 59050,
  [SMALL_STATE(1529)] = 59084,
  [SMALL_STATE(1530)] = 59118,
  [SMALL_STATE(1531)] = 59153,
  [SMALL_STATE(1532)] = 59188,
  [SMALL_STATE(1533)] = 59223,
  [SMALL_STATE(1534)] = 59258,
  [SMALL_STATE(1535)] = 59281,
  [SMALL_STATE(1536)] = 59316,
  [SMALL_STATE(1537)] = 59339,
  [SMALL_STATE(1538)] = 59368,
  [SMALL_STATE(1539)] = 59391,
  [SMALL_STATE(1540)] = 59420,
  [SMALL_STATE(1541)] = 59455,
  [SMALL_STATE(1542)] = 59484,
  [SMALL_STATE(1543)] = 59509,
  [SMALL_STATE(1544)] = 59544,
  [SMALL_STATE(1545)] = 59579,
  [SMALL_STATE(1546)] = 59609,
  [SMALL_STATE(1547)] = 59633,
  [SMALL_STATE(1548)] = 59663,
  [SMALL_STATE(1549)] = 59691,
  [SMALL_STATE(1550)] = 59717,
  [SMALL_STATE(1551)] = 59747,
  [SMALL_STATE(1552)] = 59777,
  [SMALL_STATE(1553)] = 59805,
  [SMALL_STATE(1554)] = 59839,
  [SMALL_STATE(1555)] = 59873,
  [SMALL_STATE(1556)] = 59907,
  [SMALL_STATE(1557)] = 59941,
  [SMALL_STATE(1558)] = 59975,
  [SMALL_STATE(1559)] = 60005,
  [SMALL_STATE(1560)] = 60039,
  [SMALL_STATE(1561)] = 60067,
  [SMALL_STATE(1562)] = 60101,
  [SMALL_STATE(1563)] = 60129,
  [SMALL_STATE(1564)] = 60163,
  [SMALL_STATE(1565)] = 60188,
  [SMALL_STATE(1566)] = 60215,
  [SMALL_STATE(1567)] = 60236,
  [SMALL_STATE(1568)] = 60263,
  [SMALL_STATE(1569)] = 60290,
  [SMALL_STATE(1570)] = 60315,
  [SMALL_STATE(1571)] = 60336,
  [SMALL_STATE(1572)] = 60363,
  [SMALL_STATE(1573)] = 60386,
  [SMALL_STATE(1574)] = 60415,
  [SMALL_STATE(1575)] = 60440,
  [SMALL_STATE(1576)] = 60461,
  [SMALL_STATE(1577)] = 60490,
  [SMALL_STATE(1578)] = 60511,
  [SMALL_STATE(1579)] = 60538,
  [SMALL_STATE(1580)] = 60567,
  [SMALL_STATE(1581)] = 60592,
  [SMALL_STATE(1582)] = 60613,
  [SMALL_STATE(1583)] = 60642,
  [SMALL_STATE(1584)] = 60669,
  [SMALL_STATE(1585)] = 60696,
  [SMALL_STATE(1586)] = 60719,
  [SMALL_STATE(1587)] = 60744,
  [SMALL_STATE(1588)] = 60773,
  [SMALL_STATE(1589)] = 60794,
  [SMALL_STATE(1590)] = 60815,
  [SMALL_STATE(1591)] = 60836,
  [SMALL_STATE(1592)] = 60863,
  [SMALL_STATE(1593)] = 60886,
  [SMALL_STATE(1594)] = 60913,
  [SMALL_STATE(1595)] = 60940,
  [SMALL_STATE(1596)] = 60967,
  [SMALL_STATE(1597)] = 60988,
  [SMALL_STATE(1598)] = 61011,
  [SMALL_STATE(1599)] = 61032,
  [SMALL_STATE(1600)] = 61053,
  [SMALL_STATE(1601)] = 61078,
  [SMALL_STATE(1602)] = 61105,
  [SMALL_STATE(1603)] = 61132,
  [SMALL_STATE(1604)] = 61153,
  [SMALL_STATE(1605)] = 61174,
  [SMALL_STATE(1606)] = 61203,
  [SMALL_STATE(1607)] = 61230,
  [SMALL_STATE(1608)] = 61259,
  [SMALL_STATE(1609)] = 61288,
  [SMALL_STATE(1610)] = 61309,
  [SMALL_STATE(1611)] = 61338,
  [SMALL_STATE(1612)] = 61359,
  [SMALL_STATE(1613)] = 61380,
  [SMALL_STATE(1614)] = 61401,
  [SMALL_STATE(1615)] = 61428,
  [SMALL_STATE(1616)] = 61453,
  [SMALL_STATE(1617)] = 61480,
  [SMALL_STATE(1618)] = 61507,
  [SMALL_STATE(1619)] = 61534,
  [SMALL_STATE(1620)] = 61554,
  [SMALL_STATE(1621)] = 61582,
  [SMALL_STATE(1622)] = 61610,
  [SMALL_STATE(1623)] = 61638,
  [SMALL_STATE(1624)] = 61666,
  [SMALL_STATE(1625)] = 61694,
  [SMALL_STATE(1626)] = 61722,
  [SMALL_STATE(1627)] = 61746,
  [SMALL_STATE(1628)] = 61774,
  [SMALL_STATE(1629)] = 61800,
  [SMALL_STATE(1630)] = 61826,
  [SMALL_STATE(1631)] = 61854,
  [SMALL_STATE(1632)] = 61876,
  [SMALL_STATE(1633)] = 61904,
  [SMALL_STATE(1634)] = 61924,
  [SMALL_STATE(1635)] = 61952,
  [SMALL_STATE(1636)] = 61980,
  [SMALL_STATE(1637)] = 62006,
  [SMALL_STATE(1638)] = 62032,
  [SMALL_STATE(1639)] = 62060,
  [SMALL_STATE(1640)] = 62088,
  [SMALL_STATE(1641)] = 62108,
  [SMALL_STATE(1642)] = 62128,
  [SMALL_STATE(1643)] = 62148,
  [SMALL_STATE(1644)] = 62176,
  [SMALL_STATE(1645)] = 62196,
  [SMALL_STATE(1646)] = 62224,
  [SMALL_STATE(1647)] = 62244,
  [SMALL_STATE(1648)] = 62266,
  [SMALL_STATE(1649)] = 62286,
  [SMALL_STATE(1650)] = 62314,
  [SMALL_STATE(1651)] = 62340,
  [SMALL_STATE(1652)] = 62366,
  [SMALL_STATE(1653)] = 62388,
  [SMALL_STATE(1654)] = 62416,
  [SMALL_STATE(1655)] = 62436,
  [SMALL_STATE(1656)] = 62456,
  [SMALL_STATE(1657)] = 62482,
  [SMALL_STATE(1658)] = 62504,
  [SMALL_STATE(1659)] = 62524,
  [SMALL_STATE(1660)] = 62544,
  [SMALL_STATE(1661)] = 62570,
  [SMALL_STATE(1662)] = 62592,
  [SMALL_STATE(1663)] = 62620,
  [SMALL_STATE(1664)] = 62640,
  [SMALL_STATE(1665)] = 62666,
  [SMALL_STATE(1666)] = 62688,
  [SMALL_STATE(1667)] = 62714,
  [SMALL_STATE(1668)] = 62734,
  [SMALL_STATE(1669)] = 62760,
  [SMALL_STATE(1670)] = 62786,
  [SMALL_STATE(1671)] = 62812,
  [SMALL_STATE(1672)] = 62838,
  [SMALL_STATE(1673)] = 62858,
  [SMALL_STATE(1674)] = 62878,
  [SMALL_STATE(1675)] = 62898,
  [SMALL_STATE(1676)] = 62918,
  [SMALL_STATE(1677)] = 62946,
  [SMALL_STATE(1678)] = 62968,
  [SMALL_STATE(1679)] = 62988,
  [SMALL_STATE(1680)] = 63016,
  [SMALL_STATE(1681)] = 63036,
  [SMALL_STATE(1682)] = 63056,
  [SMALL_STATE(1683)] = 63076,
  [SMALL_STATE(1684)] = 63104,
  [SMALL_STATE(1685)] = 63130,
  [SMALL_STATE(1686)] = 63150,
  [SMALL_STATE(1687)] = 63178,
  [SMALL_STATE(1688)] = 63198,
  [SMALL_STATE(1689)] = 63224,
  [SMALL_STATE(1690)] = 63252,
  [SMALL_STATE(1691)] = 63272,
  [SMALL_STATE(1692)] = 63300,
  [SMALL_STATE(1693)] = 63320,
  [SMALL_STATE(1694)] = 63348,
  [SMALL_STATE(1695)] = 63376,
  [SMALL_STATE(1696)] = 63396,
  [SMALL_STATE(1697)] = 63416,
  [SMALL_STATE(1698)] = 63444,
  [SMALL_STATE(1699)] = 63470,
  [SMALL_STATE(1700)] = 63496,
  [SMALL_STATE(1701)] = 63522,
  [SMALL_STATE(1702)] = 63550,
  [SMALL_STATE(1703)] = 63578,
  [SMALL_STATE(1704)] = 63604,
  [SMALL_STATE(1705)] = 63626,
  [SMALL_STATE(1706)] = 63654,
  [SMALL_STATE(1707)] = 63682,
  [SMALL_STATE(1708)] = 63710,
  [SMALL_STATE(1709)] = 63736,
  [SMALL_STATE(1710)] = 63755,
  [SMALL_STATE(1711)] = 63774,
  [SMALL_STATE(1712)] = 63797,
  [SMALL_STATE(1713)] = 63816,
  [SMALL_STATE(1714)] = 63841,
  [SMALL_STATE(1715)] = 63864,
  [SMALL_STATE(1716)] = 63887,
  [SMALL_STATE(1717)] = 63910,
  [SMALL_STATE(1718)] = 63933,
  [SMALL_STATE(1719)] = 63950,
  [SMALL_STATE(1720)] = 63971,
  [SMALL_STATE(1721)] = 63992,
  [SMALL_STATE(1722)] = 64015,
  [SMALL_STATE(1723)] = 64038,
  [SMALL_STATE(1724)] = 64061,
  [SMALL_STATE(1725)] = 64078,
  [SMALL_STATE(1726)] = 64101,
  [SMALL_STATE(1727)] = 64124,
  [SMALL_STATE(1728)] = 64147,
  [SMALL_STATE(1729)] = 64166,
  [SMALL_STATE(1730)] = 64189,
  [SMALL_STATE(1731)] = 64212,
  [SMALL_STATE(1732)] = 64231,
  [SMALL_STATE(1733)] = 64250,
  [SMALL_STATE(1734)] = 64267,
  [SMALL_STATE(1735)] = 64290,
  [SMALL_STATE(1736)] = 64309,
  [SMALL_STATE(1737)] = 64328,
  [SMALL_STATE(1738)] = 64351,
  [SMALL_STATE(1739)] = 64370,
  [SMALL_STATE(1740)] = 64395,
  [SMALL_STATE(1741)] = 64420,
  [SMALL_STATE(1742)] = 64439,
  [SMALL_STATE(1743)] = 64458,
  [SMALL_STATE(1744)] = 64481,
  [SMALL_STATE(1745)] = 64506,
  [SMALL_STATE(1746)] = 64525,
  [SMALL_STATE(1747)] = 64542,
  [SMALL_STATE(1748)] = 64567,
  [SMALL_STATE(1749)] = 64586,
  [SMALL_STATE(1750)] = 64609,
  [SMALL_STATE(1751)] = 64634,
  [SMALL_STATE(1752)] = 64653,
  [SMALL_STATE(1753)] = 64676,
  [SMALL_STATE(1754)] = 64697,
  [SMALL_STATE(1755)] = 64720,
  [SMALL_STATE(1756)] = 64739,
  [SMALL_STATE(1757)] = 64762,
  [SMALL_STATE(1758)] = 64785,
  [SMALL_STATE(1759)] = 64802,
  [SMALL_STATE(1760)] = 64825,
  [SMALL_STATE(1761)] = 64848,
  [SMALL_STATE(1762)] = 64869,
  [SMALL_STATE(1763)] = 64886,
  [SMALL_STATE(1764)] = 64909,
  [SMALL_STATE(1765)] = 64926,
  [SMALL_STATE(1766)] = 64951,
  [SMALL_STATE(1767)] = 64972,
  [SMALL_STATE(1768)] = 64993,
  [SMALL_STATE(1769)] = 65016,
  [SMALL_STATE(1770)] = 65037,
  [SMALL_STATE(1771)] = 65062,
  [SMALL_STATE(1772)] = 65081,
  [SMALL_STATE(1773)] = 65100,
  [SMALL_STATE(1774)] = 65125,
  [SMALL_STATE(1775)] = 65146,
  [SMALL_STATE(1776)] = 65169,
  [SMALL_STATE(1777)] = 65192,
  [SMALL_STATE(1778)] = 65211,
  [SMALL_STATE(1779)] = 65234,
  [SMALL_STATE(1780)] = 65251,
  [SMALL_STATE(1781)] = 65270,
  [SMALL_STATE(1782)] = 65291,
  [SMALL_STATE(1783)] = 65314,
  [SMALL_STATE(1784)] = 65339,
  [SMALL_STATE(1785)] = 65362,
  [SMALL_STATE(1786)] = 65381,
  [SMALL_STATE(1787)] = 65402,
  [SMALL_STATE(1788)] = 65425,
  [SMALL_STATE(1789)] = 65450,
  [SMALL_STATE(1790)] = 65475,
  [SMALL_STATE(1791)] = 65494,
  [SMALL_STATE(1792)] = 65513,
  [SMALL_STATE(1793)] = 65538,
  [SMALL_STATE(1794)] = 65557,
  [SMALL_STATE(1795)] = 65582,
  [SMALL_STATE(1796)] = 65605,
  [SMALL_STATE(1797)] = 65628,
  [SMALL_STATE(1798)] = 65647,
  [SMALL_STATE(1799)] = 65666,
  [SMALL_STATE(1800)] = 65689,
  [SMALL_STATE(1801)] = 65712,
  [SMALL_STATE(1802)] = 65735,
  [SMALL_STATE(1803)] = 65760,
  [SMALL_STATE(1804)] = 65779,
  [SMALL_STATE(1805)] = 65802,
  [SMALL_STATE(1806)] = 65821,
  [SMALL_STATE(1807)] = 65846,
  [SMALL_STATE(1808)] = 65869,
  [SMALL_STATE(1809)] = 65892,
  [SMALL_STATE(1810)] = 65915,
  [SMALL_STATE(1811)] = 65934,
  [SMALL_STATE(1812)] = 65953,
  [SMALL_STATE(1813)] = 65978,
  [SMALL_STATE(1814)] = 65997,
  [SMALL_STATE(1815)] = 66022,
  [SMALL_STATE(1816)] = 66041,
  [SMALL_STATE(1817)] = 66060,
  [SMALL_STATE(1818)] = 66079,
  [SMALL_STATE(1819)] = 66102,
  [SMALL_STATE(1820)] = 66119,
  [SMALL_STATE(1821)] = 66140,
  [SMALL_STATE(1822)] = 66161,
  [SMALL_STATE(1823)] = 66184,
  [SMALL_STATE(1824)] = 66207,
  [SMALL_STATE(1825)] = 66226,
  [SMALL_STATE(1826)] = 66251,
  [SMALL_STATE(1827)] = 66276,
  [SMALL_STATE(1828)] = 66297,
  [SMALL_STATE(1829)] = 66318,
  [SMALL_STATE(1830)] = 66341,
  [SMALL_STATE(1831)] = 66362,
  [SMALL_STATE(1832)] = 66381,
  [SMALL_STATE(1833)] = 66404,
  [SMALL_STATE(1834)] = 66429,
  [SMALL_STATE(1835)] = 66448,
  [SMALL_STATE(1836)] = 66469,
  [SMALL_STATE(1837)] = 66494,
  [SMALL_STATE(1838)] = 66517,
  [SMALL_STATE(1839)] = 66536,
  [SMALL_STATE(1840)] = 66559,
  [SMALL_STATE(1841)] = 66584,
  [SMALL_STATE(1842)] = 66607,
  [SMALL_STATE(1843)] = 66630,
  [SMALL_STATE(1844)] = 66653,
  [SMALL_STATE(1845)] = 66672,
  [SMALL_STATE(1846)] = 66694,
  [SMALL_STATE(1847)] = 66716,
  [SMALL_STATE(1848)] = 66736,
  [SMALL_STATE(1849)] = 66758,
  [SMALL_STATE(1850)] = 66774,
  [SMALL_STATE(1851)] = 66792,
  [SMALL_STATE(1852)] = 66814,
  [SMALL_STATE(1853)] = 66836,
  [SMALL_STATE(1854)] = 66858,
  [SMALL_STATE(1855)] = 66880,
  [SMALL_STATE(1856)] = 66902,
  [SMALL_STATE(1857)] = 66924,
  [SMALL_STATE(1858)] = 66944,
  [SMALL_STATE(1859)] = 66966,
  [SMALL_STATE(1860)] = 66982,
  [SMALL_STATE(1861)] = 66998,
  [SMALL_STATE(1862)] = 67016,
  [SMALL_STATE(1863)] = 67034,
  [SMALL_STATE(1864)] = 67052,
  [SMALL_STATE(1865)] = 67070,
  [SMALL_STATE(1866)] = 67092,
  [SMALL_STATE(1867)] = 67110,
  [SMALL_STATE(1868)] = 67128,
  [SMALL_STATE(1869)] = 67148,
  [SMALL_STATE(1870)] = 67170,
  [SMALL_STATE(1871)] = 67188,
  [SMALL_STATE(1872)] = 67204,
  [SMALL_STATE(1873)] = 67226,
  [SMALL_STATE(1874)] = 67248,
  [SMALL_STATE(1875)] = 67268,
  [SMALL_STATE(1876)] = 67284,
  [SMALL_STATE(1877)] = 67304,
  [SMALL_STATE(1878)] = 67324,
  [SMALL_STATE(1879)] = 67346,
  [SMALL_STATE(1880)] = 67366,
  [SMALL_STATE(1881)] = 67388,
  [SMALL_STATE(1882)] = 67410,
  [SMALL_STATE(1883)] = 67426,
  [SMALL_STATE(1884)] = 67446,
  [SMALL_STATE(1885)] = 67466,
  [SMALL_STATE(1886)] = 67486,
  [SMALL_STATE(1887)] = 67508,
  [SMALL_STATE(1888)] = 67526,
  [SMALL_STATE(1889)] = 67542,
  [SMALL_STATE(1890)] = 67564,
  [SMALL_STATE(1891)] = 67584,
  [SMALL_STATE(1892)] = 67606,
  [SMALL_STATE(1893)] = 67624,
  [SMALL_STATE(1894)] = 67646,
  [SMALL_STATE(1895)] = 67668,
  [SMALL_STATE(1896)] = 67688,
  [SMALL_STATE(1897)] = 67708,
  [SMALL_STATE(1898)] = 67730,
  [SMALL_STATE(1899)] = 67748,
  [SMALL_STATE(1900)] = 67770,
  [SMALL_STATE(1901)] = 67792,
  [SMALL_STATE(1902)] = 67814,
  [SMALL_STATE(1903)] = 67836,
  [SMALL_STATE(1904)] = 67856,
  [SMALL_STATE(1905)] = 67878,
  [SMALL_STATE(1906)] = 67896,
  [SMALL_STATE(1907)] = 67912,
  [SMALL_STATE(1908)] = 67932,
  [SMALL_STATE(1909)] = 67948,
  [SMALL_STATE(1910)] = 67970,
  [SMALL_STATE(1911)] = 67992,
  [SMALL_STATE(1912)] = 68010,
  [SMALL_STATE(1913)] = 68032,
  [SMALL_STATE(1914)] = 68048,
  [SMALL_STATE(1915)] = 68064,
  [SMALL_STATE(1916)] = 68086,
  [SMALL_STATE(1917)] = 68104,
  [SMALL_STATE(1918)] = 68120,
  [SMALL_STATE(1919)] = 68142,
  [SMALL_STATE(1920)] = 68162,
  [SMALL_STATE(1921)] = 68184,
  [SMALL_STATE(1922)] = 68200,
  [SMALL_STATE(1923)] = 68216,
  [SMALL_STATE(1924)] = 68234,
  [SMALL_STATE(1925)] = 68252,
  [SMALL_STATE(1926)] = 68272,
  [SMALL_STATE(1927)] = 68292,
  [SMALL_STATE(1928)] = 68314,
  [SMALL_STATE(1929)] = 68330,
  [SMALL_STATE(1930)] = 68346,
  [SMALL_STATE(1931)] = 68368,
  [SMALL_STATE(1932)] = 68390,
  [SMALL_STATE(1933)] = 68410,
  [SMALL_STATE(1934)] = 68432,
  [SMALL_STATE(1935)] = 68454,
  [SMALL_STATE(1936)] = 68472,
  [SMALL_STATE(1937)] = 68492,
  [SMALL_STATE(1938)] = 68514,
  [SMALL_STATE(1939)] = 68536,
  [SMALL_STATE(1940)] = 68556,
  [SMALL_STATE(1941)] = 68578,
  [SMALL_STATE(1942)] = 68600,
  [SMALL_STATE(1943)] = 68622,
  [SMALL_STATE(1944)] = 68644,
  [SMALL_STATE(1945)] = 68660,
  [SMALL_STATE(1946)] = 68680,
  [SMALL_STATE(1947)] = 68702,
  [SMALL_STATE(1948)] = 68724,
  [SMALL_STATE(1949)] = 68744,
  [SMALL_STATE(1950)] = 68762,
  [SMALL_STATE(1951)] = 68784,
  [SMALL_STATE(1952)] = 68806,
  [SMALL_STATE(1953)] = 68824,
  [SMALL_STATE(1954)] = 68840,
  [SMALL_STATE(1955)] = 68860,
  [SMALL_STATE(1956)] = 68875,
  [SMALL_STATE(1957)] = 68892,
  [SMALL_STATE(1958)] = 68909,
  [SMALL_STATE(1959)] = 68926,
  [SMALL_STATE(1960)] = 68943,
  [SMALL_STATE(1961)] = 68962,
  [SMALL_STATE(1962)] = 68979,
  [SMALL_STATE(1963)] = 68998,
  [SMALL_STATE(1964)] = 69017,
  [SMALL_STATE(1965)] = 69036,
  [SMALL_STATE(1966)] = 69055,
  [SMALL_STATE(1967)] = 69072,
  [SMALL_STATE(1968)] = 69089,
  [SMALL_STATE(1969)] = 69106,
  [SMALL_STATE(1970)] = 69123,
  [SMALL_STATE(1971)] = 69140,
  [SMALL_STATE(1972)] = 69157,
  [SMALL_STATE(1973)] = 69174,
  [SMALL_STATE(1974)] = 69193,
  [SMALL_STATE(1975)] = 69212,
  [SMALL_STATE(1976)] = 69231,
  [SMALL_STATE(1977)] = 69250,
  [SMALL_STATE(1978)] = 69267,
  [SMALL_STATE(1979)] = 69284,
  [SMALL_STATE(1980)] = 69299,
  [SMALL_STATE(1981)] = 69318,
  [SMALL_STATE(1982)] = 69333,
  [SMALL_STATE(1983)] = 69352,
  [SMALL_STATE(1984)] = 69371,
  [SMALL_STATE(1985)] = 69388,
  [SMALL_STATE(1986)] = 69405,
  [SMALL_STATE(1987)] = 69420,
  [SMALL_STATE(1988)] = 69435,
  [SMALL_STATE(1989)] = 69452,
  [SMALL_STATE(1990)] = 69469,
  [SMALL_STATE(1991)] = 69484,
  [SMALL_STATE(1992)] = 69501,
  [SMALL_STATE(1993)] = 69518,
  [SMALL_STATE(1994)] = 69535,
  [SMALL_STATE(1995)] = 69552,
  [SMALL_STATE(1996)] = 69569,
  [SMALL_STATE(1997)] = 69588,
  [SMALL_STATE(1998)] = 69605,
  [SMALL_STATE(1999)] = 69622,
  [SMALL_STATE(2000)] = 69639,
  [SMALL_STATE(2001)] = 69656,
  [SMALL_STATE(2002)] = 69673,
  [SMALL_STATE(2003)] = 69690,
  [SMALL_STATE(2004)] = 69707,
  [SMALL_STATE(2005)] = 69724,
  [SMALL_STATE(2006)] = 69739,
  [SMALL_STATE(2007)] = 69756,
  [SMALL_STATE(2008)] = 69775,
  [SMALL_STATE(2009)] = 69792,
  [SMALL_STATE(2010)] = 69809,
  [SMALL_STATE(2011)] = 69828,
  [SMALL_STATE(2012)] = 69847,
  [SMALL_STATE(2013)] = 69864,
  [SMALL_STATE(2014)] = 69881,
  [SMALL_STATE(2015)] = 69898,
  [SMALL_STATE(2016)] = 69917,
  [SMALL_STATE(2017)] = 69936,
  [SMALL_STATE(2018)] = 69955,
  [SMALL_STATE(2019)] = 69972,
  [SMALL_STATE(2020)] = 69989,
  [SMALL_STATE(2021)] = 70006,
  [SMALL_STATE(2022)] = 70023,
  [SMALL_STATE(2023)] = 70040,
  [SMALL_STATE(2024)] = 70057,
  [SMALL_STATE(2025)] = 70074,
  [SMALL_STATE(2026)] = 70091,
  [SMALL_STATE(2027)] = 70110,
  [SMALL_STATE(2028)] = 70129,
  [SMALL_STATE(2029)] = 70146,
  [SMALL_STATE(2030)] = 70163,
  [SMALL_STATE(2031)] = 70182,
  [SMALL_STATE(2032)] = 70201,
  [SMALL_STATE(2033)] = 70218,
  [SMALL_STATE(2034)] = 70233,
  [SMALL_STATE(2035)] = 70250,
  [SMALL_STATE(2036)] = 70267,
  [SMALL_STATE(2037)] = 70282,
  [SMALL_STATE(2038)] = 70301,
  [SMALL_STATE(2039)] = 70320,
  [SMALL_STATE(2040)] = 70339,
  [SMALL_STATE(2041)] = 70358,
  [SMALL_STATE(2042)] = 70377,
  [SMALL_STATE(2043)] = 70394,
  [SMALL_STATE(2044)] = 70411,
  [SMALL_STATE(2045)] = 70428,
  [SMALL_STATE(2046)] = 70445,
  [SMALL_STATE(2047)] = 70462,
  [SMALL_STATE(2048)] = 70481,
  [SMALL_STATE(2049)] = 70498,
  [SMALL_STATE(2050)] = 70515,
  [SMALL_STATE(2051)] = 70534,
  [SMALL_STATE(2052)] = 70553,
  [SMALL_STATE(2053)] = 70570,
  [SMALL_STATE(2054)] = 70589,
  [SMALL_STATE(2055)] = 70606,
  [SMALL_STATE(2056)] = 70621,
  [SMALL_STATE(2057)] = 70638,
  [SMALL_STATE(2058)] = 70655,
  [SMALL_STATE(2059)] = 70674,
  [SMALL_STATE(2060)] = 70693,
  [SMALL_STATE(2061)] = 70712,
  [SMALL_STATE(2062)] = 70731,
  [SMALL_STATE(2063)] = 70750,
  [SMALL_STATE(2064)] = 70769,
  [SMALL_STATE(2065)] = 70786,
  [SMALL_STATE(2066)] = 70803,
  [SMALL_STATE(2067)] = 70820,
  [SMALL_STATE(2068)] = 70837,
  [SMALL_STATE(2069)] = 70854,
  [SMALL_STATE(2070)] = 70873,
  [SMALL_STATE(2071)] = 70890,
  [SMALL_STATE(2072)] = 70909,
  [SMALL_STATE(2073)] = 70926,
  [SMALL_STATE(2074)] = 70945,
  [SMALL_STATE(2075)] = 70962,
  [SMALL_STATE(2076)] = 70979,
  [SMALL_STATE(2077)] = 70996,
  [SMALL_STATE(2078)] = 71015,
  [SMALL_STATE(2079)] = 71032,
  [SMALL_STATE(2080)] = 71049,
  [SMALL_STATE(2081)] = 71066,
  [SMALL_STATE(2082)] = 71083,
  [SMALL_STATE(2083)] = 71100,
  [SMALL_STATE(2084)] = 71119,
  [SMALL_STATE(2085)] = 71138,
  [SMALL_STATE(2086)] = 71157,
  [SMALL_STATE(2087)] = 71174,
  [SMALL_STATE(2088)] = 71191,
  [SMALL_STATE(2089)] = 71208,
  [SMALL_STATE(2090)] = 71227,
  [SMALL_STATE(2091)] = 71246,
  [SMALL_STATE(2092)] = 71265,
  [SMALL_STATE(2093)] = 71282,
  [SMALL_STATE(2094)] = 71297,
  [SMALL_STATE(2095)] = 71312,
  [SMALL_STATE(2096)] = 71331,
  [SMALL_STATE(2097)] = 71348,
  [SMALL_STATE(2098)] = 71365,
  [SMALL_STATE(2099)] = 71384,
  [SMALL_STATE(2100)] = 71401,
  [SMALL_STATE(2101)] = 71418,
  [SMALL_STATE(2102)] = 71435,
  [SMALL_STATE(2103)] = 71452,
  [SMALL_STATE(2104)] = 71469,
  [SMALL_STATE(2105)] = 71486,
  [SMALL_STATE(2106)] = 71505,
  [SMALL_STATE(2107)] = 71524,
  [SMALL_STATE(2108)] = 71543,
  [SMALL_STATE(2109)] = 71560,
  [SMALL_STATE(2110)] = 71577,
  [SMALL_STATE(2111)] = 71594,
  [SMALL_STATE(2112)] = 71611,
  [SMALL_STATE(2113)] = 71628,
  [SMALL_STATE(2114)] = 71645,
  [SMALL_STATE(2115)] = 71662,
  [SMALL_STATE(2116)] = 71679,
  [SMALL_STATE(2117)] = 71698,
  [SMALL_STATE(2118)] = 71715,
  [SMALL_STATE(2119)] = 71734,
  [SMALL_STATE(2120)] = 71751,
  [SMALL_STATE(2121)] = 71768,
  [SMALL_STATE(2122)] = 71787,
  [SMALL_STATE(2123)] = 71804,
  [SMALL_STATE(2124)] = 71819,
  [SMALL_STATE(2125)] = 71836,
  [SMALL_STATE(2126)] = 71853,
  [SMALL_STATE(2127)] = 71870,
  [SMALL_STATE(2128)] = 71887,
  [SMALL_STATE(2129)] = 71906,
  [SMALL_STATE(2130)] = 71923,
  [SMALL_STATE(2131)] = 71940,
  [SMALL_STATE(2132)] = 71957,
  [SMALL_STATE(2133)] = 71976,
  [SMALL_STATE(2134)] = 71993,
  [SMALL_STATE(2135)] = 72010,
  [SMALL_STATE(2136)] = 72027,
  [SMALL_STATE(2137)] = 72044,
  [SMALL_STATE(2138)] = 72061,
  [SMALL_STATE(2139)] = 72078,
  [SMALL_STATE(2140)] = 72097,
  [SMALL_STATE(2141)] = 72114,
  [SMALL_STATE(2142)] = 72133,
  [SMALL_STATE(2143)] = 72150,
  [SMALL_STATE(2144)] = 72167,
  [SMALL_STATE(2145)] = 72184,
  [SMALL_STATE(2146)] = 72203,
  [SMALL_STATE(2147)] = 72220,
  [SMALL_STATE(2148)] = 72237,
  [SMALL_STATE(2149)] = 72254,
  [SMALL_STATE(2150)] = 72271,
  [SMALL_STATE(2151)] = 72288,
  [SMALL_STATE(2152)] = 72305,
  [SMALL_STATE(2153)] = 72322,
  [SMALL_STATE(2154)] = 72339,
  [SMALL_STATE(2155)] = 72358,
  [SMALL_STATE(2156)] = 72373,
  [SMALL_STATE(2157)] = 72390,
  [SMALL_STATE(2158)] = 72407,
  [SMALL_STATE(2159)] = 72426,
  [SMALL_STATE(2160)] = 72443,
  [SMALL_STATE(2161)] = 72460,
  [SMALL_STATE(2162)] = 72477,
  [SMALL_STATE(2163)] = 72496,
  [SMALL_STATE(2164)] = 72513,
  [SMALL_STATE(2165)] = 72530,
  [SMALL_STATE(2166)] = 72547,
  [SMALL_STATE(2167)] = 72564,
  [SMALL_STATE(2168)] = 72581,
  [SMALL_STATE(2169)] = 72598,
  [SMALL_STATE(2170)] = 72615,
  [SMALL_STATE(2171)] = 72632,
  [SMALL_STATE(2172)] = 72649,
  [SMALL_STATE(2173)] = 72668,
  [SMALL_STATE(2174)] = 72685,
  [SMALL_STATE(2175)] = 72704,
  [SMALL_STATE(2176)] = 72721,
  [SMALL_STATE(2177)] = 72740,
  [SMALL_STATE(2178)] = 72757,
  [SMALL_STATE(2179)] = 72774,
  [SMALL_STATE(2180)] = 72791,
  [SMALL_STATE(2181)] = 72808,
  [SMALL_STATE(2182)] = 72825,
  [SMALL_STATE(2183)] = 72842,
  [SMALL_STATE(2184)] = 72859,
  [SMALL_STATE(2185)] = 72876,
  [SMALL_STATE(2186)] = 72893,
  [SMALL_STATE(2187)] = 72912,
  [SMALL_STATE(2188)] = 72929,
  [SMALL_STATE(2189)] = 72946,
  [SMALL_STATE(2190)] = 72965,
  [SMALL_STATE(2191)] = 72982,
  [SMALL_STATE(2192)] = 72999,
  [SMALL_STATE(2193)] = 73016,
  [SMALL_STATE(2194)] = 73033,
  [SMALL_STATE(2195)] = 73050,
  [SMALL_STATE(2196)] = 73067,
  [SMALL_STATE(2197)] = 73084,
  [SMALL_STATE(2198)] = 73101,
  [SMALL_STATE(2199)] = 73118,
  [SMALL_STATE(2200)] = 73135,
  [SMALL_STATE(2201)] = 73152,
  [SMALL_STATE(2202)] = 73169,
  [SMALL_STATE(2203)] = 73186,
  [SMALL_STATE(2204)] = 73203,
  [SMALL_STATE(2205)] = 73220,
  [SMALL_STATE(2206)] = 73237,
  [SMALL_STATE(2207)] = 73254,
  [SMALL_STATE(2208)] = 73271,
  [SMALL_STATE(2209)] = 73288,
  [SMALL_STATE(2210)] = 73305,
  [SMALL_STATE(2211)] = 73322,
  [SMALL_STATE(2212)] = 73339,
  [SMALL_STATE(2213)] = 73356,
  [SMALL_STATE(2214)] = 73375,
  [SMALL_STATE(2215)] = 73392,
  [SMALL_STATE(2216)] = 73409,
  [SMALL_STATE(2217)] = 73426,
  [SMALL_STATE(2218)] = 73445,
  [SMALL_STATE(2219)] = 73464,
  [SMALL_STATE(2220)] = 73483,
  [SMALL_STATE(2221)] = 73502,
  [SMALL_STATE(2222)] = 73519,
  [SMALL_STATE(2223)] = 73534,
  [SMALL_STATE(2224)] = 73553,
  [SMALL_STATE(2225)] = 73570,
  [SMALL_STATE(2226)] = 73587,
  [SMALL_STATE(2227)] = 73604,
  [SMALL_STATE(2228)] = 73621,
  [SMALL_STATE(2229)] = 73638,
  [SMALL_STATE(2230)] = 73657,
  [SMALL_STATE(2231)] = 73672,
  [SMALL_STATE(2232)] = 73689,
  [SMALL_STATE(2233)] = 73704,
  [SMALL_STATE(2234)] = 73721,
  [SMALL_STATE(2235)] = 73740,
  [SMALL_STATE(2236)] = 73757,
  [SMALL_STATE(2237)] = 73772,
  [SMALL_STATE(2238)] = 73791,
  [SMALL_STATE(2239)] = 73810,
  [SMALL_STATE(2240)] = 73827,
  [SMALL_STATE(2241)] = 73844,
  [SMALL_STATE(2242)] = 73861,
  [SMALL_STATE(2243)] = 73880,
  [SMALL_STATE(2244)] = 73899,
  [SMALL_STATE(2245)] = 73916,
  [SMALL_STATE(2246)] = 73935,
  [SMALL_STATE(2247)] = 73952,
  [SMALL_STATE(2248)] = 73969,
  [SMALL_STATE(2249)] = 73988,
  [SMALL_STATE(2250)] = 74007,
  [SMALL_STATE(2251)] = 74024,
  [SMALL_STATE(2252)] = 74041,
  [SMALL_STATE(2253)] = 74058,
  [SMALL_STATE(2254)] = 74077,
  [SMALL_STATE(2255)] = 74092,
  [SMALL_STATE(2256)] = 74109,
  [SMALL_STATE(2257)] = 74124,
  [SMALL_STATE(2258)] = 74139,
  [SMALL_STATE(2259)] = 74156,
  [SMALL_STATE(2260)] = 74173,
  [SMALL_STATE(2261)] = 74192,
  [SMALL_STATE(2262)] = 74211,
  [SMALL_STATE(2263)] = 74226,
  [SMALL_STATE(2264)] = 74241,
  [SMALL_STATE(2265)] = 74260,
  [SMALL_STATE(2266)] = 74277,
  [SMALL_STATE(2267)] = 74296,
  [SMALL_STATE(2268)] = 74313,
  [SMALL_STATE(2269)] = 74330,
  [SMALL_STATE(2270)] = 74347,
  [SMALL_STATE(2271)] = 74364,
  [SMALL_STATE(2272)] = 74381,
  [SMALL_STATE(2273)] = 74398,
  [SMALL_STATE(2274)] = 74415,
  [SMALL_STATE(2275)] = 74432,
  [SMALL_STATE(2276)] = 74449,
  [SMALL_STATE(2277)] = 74466,
  [SMALL_STATE(2278)] = 74483,
  [SMALL_STATE(2279)] = 74500,
  [SMALL_STATE(2280)] = 74517,
  [SMALL_STATE(2281)] = 74534,
  [SMALL_STATE(2282)] = 74551,
  [SMALL_STATE(2283)] = 74570,
  [SMALL_STATE(2284)] = 74587,
  [SMALL_STATE(2285)] = 74604,
  [SMALL_STATE(2286)] = 74621,
  [SMALL_STATE(2287)] = 74638,
  [SMALL_STATE(2288)] = 74657,
  [SMALL_STATE(2289)] = 74674,
  [SMALL_STATE(2290)] = 74693,
  [SMALL_STATE(2291)] = 74712,
  [SMALL_STATE(2292)] = 74731,
  [SMALL_STATE(2293)] = 74748,
  [SMALL_STATE(2294)] = 74767,
  [SMALL_STATE(2295)] = 74782,
  [SMALL_STATE(2296)] = 74801,
  [SMALL_STATE(2297)] = 74818,
  [SMALL_STATE(2298)] = 74837,
  [SMALL_STATE(2299)] = 74851,
  [SMALL_STATE(2300)] = 74867,
  [SMALL_STATE(2301)] = 74883,
  [SMALL_STATE(2302)] = 74897,
  [SMALL_STATE(2303)] = 74913,
  [SMALL_STATE(2304)] = 74929,
  [SMALL_STATE(2305)] = 74945,
  [SMALL_STATE(2306)] = 74959,
  [SMALL_STATE(2307)] = 74975,
  [SMALL_STATE(2308)] = 74991,
  [SMALL_STATE(2309)] = 75007,
  [SMALL_STATE(2310)] = 75023,
  [SMALL_STATE(2311)] = 75039,
  [SMALL_STATE(2312)] = 75055,
  [SMALL_STATE(2313)] = 75071,
  [SMALL_STATE(2314)] = 75085,
  [SMALL_STATE(2315)] = 75101,
  [SMALL_STATE(2316)] = 75117,
  [SMALL_STATE(2317)] = 75131,
  [SMALL_STATE(2318)] = 75145,
  [SMALL_STATE(2319)] = 75159,
  [SMALL_STATE(2320)] = 75175,
  [SMALL_STATE(2321)] = 75189,
  [SMALL_STATE(2322)] = 75205,
  [SMALL_STATE(2323)] = 75221,
  [SMALL_STATE(2324)] = 75237,
  [SMALL_STATE(2325)] = 75253,
  [SMALL_STATE(2326)] = 75267,
  [SMALL_STATE(2327)] = 75283,
  [SMALL_STATE(2328)] = 75299,
  [SMALL_STATE(2329)] = 75315,
  [SMALL_STATE(2330)] = 75329,
  [SMALL_STATE(2331)] = 75345,
  [SMALL_STATE(2332)] = 75361,
  [SMALL_STATE(2333)] = 75377,
  [SMALL_STATE(2334)] = 75393,
  [SMALL_STATE(2335)] = 75409,
  [SMALL_STATE(2336)] = 75425,
  [SMALL_STATE(2337)] = 75441,
  [SMALL_STATE(2338)] = 75457,
  [SMALL_STATE(2339)] = 75473,
  [SMALL_STATE(2340)] = 75489,
  [SMALL_STATE(2341)] = 75505,
  [SMALL_STATE(2342)] = 75521,
  [SMALL_STATE(2343)] = 75537,
  [SMALL_STATE(2344)] = 75553,
  [SMALL_STATE(2345)] = 75569,
  [SMALL_STATE(2346)] = 75585,
  [SMALL_STATE(2347)] = 75601,
  [SMALL_STATE(2348)] = 75617,
  [SMALL_STATE(2349)] = 75631,
  [SMALL_STATE(2350)] = 75647,
  [SMALL_STATE(2351)] = 75661,
  [SMALL_STATE(2352)] = 75675,
  [SMALL_STATE(2353)] = 75691,
  [SMALL_STATE(2354)] = 75707,
  [SMALL_STATE(2355)] = 75723,
  [SMALL_STATE(2356)] = 75737,
  [SMALL_STATE(2357)] = 75753,
  [SMALL_STATE(2358)] = 75769,
  [SMALL_STATE(2359)] = 75783,
  [SMALL_STATE(2360)] = 75799,
  [SMALL_STATE(2361)] = 75813,
  [SMALL_STATE(2362)] = 75829,
  [SMALL_STATE(2363)] = 75845,
  [SMALL_STATE(2364)] = 75861,
  [SMALL_STATE(2365)] = 75875,
  [SMALL_STATE(2366)] = 75891,
  [SMALL_STATE(2367)] = 75907,
  [SMALL_STATE(2368)] = 75923,
  [SMALL_STATE(2369)] = 75937,
  [SMALL_STATE(2370)] = 75953,
  [SMALL_STATE(2371)] = 75967,
  [SMALL_STATE(2372)] = 75983,
  [SMALL_STATE(2373)] = 75999,
  [SMALL_STATE(2374)] = 76015,
  [SMALL_STATE(2375)] = 76029,
  [SMALL_STATE(2376)] = 76043,
  [SMALL_STATE(2377)] = 76059,
  [SMALL_STATE(2378)] = 76075,
  [SMALL_STATE(2379)] = 76091,
  [SMALL_STATE(2380)] = 76107,
  [SMALL_STATE(2381)] = 76123,
  [SMALL_STATE(2382)] = 76137,
  [SMALL_STATE(2383)] = 76151,
  [SMALL_STATE(2384)] = 76167,
  [SMALL_STATE(2385)] = 76181,
  [SMALL_STATE(2386)] = 76197,
  [SMALL_STATE(2387)] = 76211,
  [SMALL_STATE(2388)] = 76225,
  [SMALL_STATE(2389)] = 76241,
  [SMALL_STATE(2390)] = 76255,
  [SMALL_STATE(2391)] = 76271,
  [SMALL_STATE(2392)] = 76287,
  [SMALL_STATE(2393)] = 76303,
  [SMALL_STATE(2394)] = 76319,
  [SMALL_STATE(2395)] = 76335,
  [SMALL_STATE(2396)] = 76351,
  [SMALL_STATE(2397)] = 76365,
  [SMALL_STATE(2398)] = 76381,
  [SMALL_STATE(2399)] = 76397,
  [SMALL_STATE(2400)] = 76413,
  [SMALL_STATE(2401)] = 76429,
  [SMALL_STATE(2402)] = 76443,
  [SMALL_STATE(2403)] = 76459,
  [SMALL_STATE(2404)] = 76473,
  [SMALL_STATE(2405)] = 76489,
  [SMALL_STATE(2406)] = 76505,
  [SMALL_STATE(2407)] = 76519,
  [SMALL_STATE(2408)] = 76535,
  [SMALL_STATE(2409)] = 76549,
  [SMALL_STATE(2410)] = 76563,
  [SMALL_STATE(2411)] = 76579,
  [SMALL_STATE(2412)] = 76595,
  [SMALL_STATE(2413)] = 76611,
  [SMALL_STATE(2414)] = 76627,
  [SMALL_STATE(2415)] = 76641,
  [SMALL_STATE(2416)] = 76657,
  [SMALL_STATE(2417)] = 76673,
  [SMALL_STATE(2418)] = 76689,
  [SMALL_STATE(2419)] = 76705,
  [SMALL_STATE(2420)] = 76721,
  [SMALL_STATE(2421)] = 76737,
  [SMALL_STATE(2422)] = 76751,
  [SMALL_STATE(2423)] = 76765,
  [SMALL_STATE(2424)] = 76781,
  [SMALL_STATE(2425)] = 76795,
  [SMALL_STATE(2426)] = 76811,
  [SMALL_STATE(2427)] = 76825,
  [SMALL_STATE(2428)] = 76841,
  [SMALL_STATE(2429)] = 76857,
  [SMALL_STATE(2430)] = 76871,
  [SMALL_STATE(2431)] = 76887,
  [SMALL_STATE(2432)] = 76901,
  [SMALL_STATE(2433)] = 76917,
  [SMALL_STATE(2434)] = 76933,
  [SMALL_STATE(2435)] = 76949,
  [SMALL_STATE(2436)] = 76965,
  [SMALL_STATE(2437)] = 76981,
  [SMALL_STATE(2438)] = 76997,
  [SMALL_STATE(2439)] = 77013,
  [SMALL_STATE(2440)] = 77029,
  [SMALL_STATE(2441)] = 77045,
  [SMALL_STATE(2442)] = 77059,
  [SMALL_STATE(2443)] = 77075,
  [SMALL_STATE(2444)] = 77089,
  [SMALL_STATE(2445)] = 77105,
  [SMALL_STATE(2446)] = 77121,
  [SMALL_STATE(2447)] = 77135,
  [SMALL_STATE(2448)] = 77149,
  [SMALL_STATE(2449)] = 77163,
  [SMALL_STATE(2450)] = 77177,
  [SMALL_STATE(2451)] = 77191,
  [SMALL_STATE(2452)] = 77205,
  [SMALL_STATE(2453)] = 77219,
  [SMALL_STATE(2454)] = 77233,
  [SMALL_STATE(2455)] = 77247,
  [SMALL_STATE(2456)] = 77263,
  [SMALL_STATE(2457)] = 77279,
  [SMALL_STATE(2458)] = 77295,
  [SMALL_STATE(2459)] = 77309,
  [SMALL_STATE(2460)] = 77325,
  [SMALL_STATE(2461)] = 77341,
  [SMALL_STATE(2462)] = 77357,
  [SMALL_STATE(2463)] = 77371,
  [SMALL_STATE(2464)] = 77385,
  [SMALL_STATE(2465)] = 77401,
  [SMALL_STATE(2466)] = 77417,
  [SMALL_STATE(2467)] = 77431,
  [SMALL_STATE(2468)] = 77445,
  [SMALL_STATE(2469)] = 77459,
  [SMALL_STATE(2470)] = 77475,
  [SMALL_STATE(2471)] = 77489,
  [SMALL_STATE(2472)] = 77505,
  [SMALL_STATE(2473)] = 77521,
  [SMALL_STATE(2474)] = 77535,
  [SMALL_STATE(2475)] = 77551,
  [SMALL_STATE(2476)] = 77567,
  [SMALL_STATE(2477)] = 77581,
  [SMALL_STATE(2478)] = 77595,
  [SMALL_STATE(2479)] = 77611,
  [SMALL_STATE(2480)] = 77627,
  [SMALL_STATE(2481)] = 77643,
  [SMALL_STATE(2482)] = 77657,
  [SMALL_STATE(2483)] = 77673,
  [SMALL_STATE(2484)] = 77687,
  [SMALL_STATE(2485)] = 77701,
  [SMALL_STATE(2486)] = 77717,
  [SMALL_STATE(2487)] = 77733,
  [SMALL_STATE(2488)] = 77749,
  [SMALL_STATE(2489)] = 77763,
  [SMALL_STATE(2490)] = 77777,
  [SMALL_STATE(2491)] = 77791,
  [SMALL_STATE(2492)] = 77807,
  [SMALL_STATE(2493)] = 77821,
  [SMALL_STATE(2494)] = 77837,
  [SMALL_STATE(2495)] = 77851,
  [SMALL_STATE(2496)] = 77867,
  [SMALL_STATE(2497)] = 77883,
  [SMALL_STATE(2498)] = 77899,
  [SMALL_STATE(2499)] = 77913,
  [SMALL_STATE(2500)] = 77929,
  [SMALL_STATE(2501)] = 77945,
  [SMALL_STATE(2502)] = 77959,
  [SMALL_STATE(2503)] = 77975,
  [SMALL_STATE(2504)] = 77991,
  [SMALL_STATE(2505)] = 78007,
  [SMALL_STATE(2506)] = 78020,
  [SMALL_STATE(2507)] = 78033,
  [SMALL_STATE(2508)] = 78046,
  [SMALL_STATE(2509)] = 78059,
  [SMALL_STATE(2510)] = 78072,
  [SMALL_STATE(2511)] = 78085,
  [SMALL_STATE(2512)] = 78098,
  [SMALL_STATE(2513)] = 78111,
  [SMALL_STATE(2514)] = 78124,
  [SMALL_STATE(2515)] = 78137,
  [SMALL_STATE(2516)] = 78150,
  [SMALL_STATE(2517)] = 78163,
  [SMALL_STATE(2518)] = 78176,
  [SMALL_STATE(2519)] = 78189,
  [SMALL_STATE(2520)] = 78202,
  [SMALL_STATE(2521)] = 78215,
  [SMALL_STATE(2522)] = 78228,
  [SMALL_STATE(2523)] = 78241,
  [SMALL_STATE(2524)] = 78254,
  [SMALL_STATE(2525)] = 78267,
  [SMALL_STATE(2526)] = 78280,
  [SMALL_STATE(2527)] = 78293,
  [SMALL_STATE(2528)] = 78306,
  [SMALL_STATE(2529)] = 78319,
  [SMALL_STATE(2530)] = 78332,
  [SMALL_STATE(2531)] = 78345,
  [SMALL_STATE(2532)] = 78358,
  [SMALL_STATE(2533)] = 78371,
  [SMALL_STATE(2534)] = 78384,
  [SMALL_STATE(2535)] = 78397,
  [SMALL_STATE(2536)] = 78410,
  [SMALL_STATE(2537)] = 78423,
  [SMALL_STATE(2538)] = 78436,
  [SMALL_STATE(2539)] = 78449,
  [SMALL_STATE(2540)] = 78462,
  [SMALL_STATE(2541)] = 78475,
  [SMALL_STATE(2542)] = 78488,
  [SMALL_STATE(2543)] = 78501,
  [SMALL_STATE(2544)] = 78514,
  [SMALL_STATE(2545)] = 78527,
  [SMALL_STATE(2546)] = 78540,
  [SMALL_STATE(2547)] = 78553,
  [SMALL_STATE(2548)] = 78566,
  [SMALL_STATE(2549)] = 78579,
  [SMALL_STATE(2550)] = 78592,
  [SMALL_STATE(2551)] = 78605,
  [SMALL_STATE(2552)] = 78618,
  [SMALL_STATE(2553)] = 78631,
  [SMALL_STATE(2554)] = 78644,
  [SMALL_STATE(2555)] = 78657,
  [SMALL_STATE(2556)] = 78670,
  [SMALL_STATE(2557)] = 78683,
  [SMALL_STATE(2558)] = 78696,
  [SMALL_STATE(2559)] = 78709,
  [SMALL_STATE(2560)] = 78722,
  [SMALL_STATE(2561)] = 78735,
  [SMALL_STATE(2562)] = 78748,
  [SMALL_STATE(2563)] = 78761,
  [SMALL_STATE(2564)] = 78774,
  [SMALL_STATE(2565)] = 78787,
  [SMALL_STATE(2566)] = 78800,
  [SMALL_STATE(2567)] = 78813,
  [SMALL_STATE(2568)] = 78826,
  [SMALL_STATE(2569)] = 78839,
  [SMALL_STATE(2570)] = 78852,
  [SMALL_STATE(2571)] = 78865,
  [SMALL_STATE(2572)] = 78878,
  [SMALL_STATE(2573)] = 78891,
  [SMALL_STATE(2574)] = 78904,
  [SMALL_STATE(2575)] = 78917,
  [SMALL_STATE(2576)] = 78930,
  [SMALL_STATE(2577)] = 78943,
  [SMALL_STATE(2578)] = 78956,
  [SMALL_STATE(2579)] = 78969,
  [SMALL_STATE(2580)] = 78982,
  [SMALL_STATE(2581)] = 78995,
  [SMALL_STATE(2582)] = 79008,
  [SMALL_STATE(2583)] = 79021,
  [SMALL_STATE(2584)] = 79034,
  [SMALL_STATE(2585)] = 79047,
  [SMALL_STATE(2586)] = 79060,
  [SMALL_STATE(2587)] = 79073,
  [SMALL_STATE(2588)] = 79086,
  [SMALL_STATE(2589)] = 79099,
  [SMALL_STATE(2590)] = 79112,
  [SMALL_STATE(2591)] = 79125,
  [SMALL_STATE(2592)] = 79138,
  [SMALL_STATE(2593)] = 79151,
  [SMALL_STATE(2594)] = 79164,
  [SMALL_STATE(2595)] = 79177,
  [SMALL_STATE(2596)] = 79190,
  [SMALL_STATE(2597)] = 79203,
  [SMALL_STATE(2598)] = 79216,
  [SMALL_STATE(2599)] = 79229,
  [SMALL_STATE(2600)] = 79242,
  [SMALL_STATE(2601)] = 79255,
  [SMALL_STATE(2602)] = 79268,
  [SMALL_STATE(2603)] = 79281,
  [SMALL_STATE(2604)] = 79294,
  [SMALL_STATE(2605)] = 79307,
  [SMALL_STATE(2606)] = 79320,
  [SMALL_STATE(2607)] = 79333,
  [SMALL_STATE(2608)] = 79346,
  [SMALL_STATE(2609)] = 79359,
  [SMALL_STATE(2610)] = 79372,
  [SMALL_STATE(2611)] = 79385,
  [SMALL_STATE(2612)] = 79398,
  [SMALL_STATE(2613)] = 79411,
  [SMALL_STATE(2614)] = 79424,
  [SMALL_STATE(2615)] = 79437,
  [SMALL_STATE(2616)] = 79450,
  [SMALL_STATE(2617)] = 79463,
  [SMALL_STATE(2618)] = 79476,
  [SMALL_STATE(2619)] = 79489,
  [SMALL_STATE(2620)] = 79502,
  [SMALL_STATE(2621)] = 79515,
  [SMALL_STATE(2622)] = 79528,
  [SMALL_STATE(2623)] = 79541,
  [SMALL_STATE(2624)] = 79554,
  [SMALL_STATE(2625)] = 79567,
  [SMALL_STATE(2626)] = 79580,
  [SMALL_STATE(2627)] = 79593,
  [SMALL_STATE(2628)] = 79606,
  [SMALL_STATE(2629)] = 79619,
  [SMALL_STATE(2630)] = 79632,
  [SMALL_STATE(2631)] = 79645,
  [SMALL_STATE(2632)] = 79658,
  [SMALL_STATE(2633)] = 79671,
  [SMALL_STATE(2634)] = 79684,
  [SMALL_STATE(2635)] = 79697,
  [SMALL_STATE(2636)] = 79710,
  [SMALL_STATE(2637)] = 79723,
  [SMALL_STATE(2638)] = 79736,
  [SMALL_STATE(2639)] = 79749,
  [SMALL_STATE(2640)] = 79762,
  [SMALL_STATE(2641)] = 79775,
  [SMALL_STATE(2642)] = 79788,
  [SMALL_STATE(2643)] = 79801,
  [SMALL_STATE(2644)] = 79814,
  [SMALL_STATE(2645)] = 79827,
  [SMALL_STATE(2646)] = 79840,
  [SMALL_STATE(2647)] = 79853,
  [SMALL_STATE(2648)] = 79866,
  [SMALL_STATE(2649)] = 79879,
  [SMALL_STATE(2650)] = 79892,
  [SMALL_STATE(2651)] = 79905,
  [SMALL_STATE(2652)] = 79918,
  [SMALL_STATE(2653)] = 79931,
  [SMALL_STATE(2654)] = 79944,
  [SMALL_STATE(2655)] = 79957,
  [SMALL_STATE(2656)] = 79970,
  [SMALL_STATE(2657)] = 79983,
  [SMALL_STATE(2658)] = 79996,
  [SMALL_STATE(2659)] = 80009,
  [SMALL_STATE(2660)] = 80022,
  [SMALL_STATE(2661)] = 80035,
  [SMALL_STATE(2662)] = 80048,
  [SMALL_STATE(2663)] = 80061,
  [SMALL_STATE(2664)] = 80074,
  [SMALL_STATE(2665)] = 80087,
  [SMALL_STATE(2666)] = 80100,
  [SMALL_STATE(2667)] = 80113,
  [SMALL_STATE(2668)] = 80126,
  [SMALL_STATE(2669)] = 80139,
  [SMALL_STATE(2670)] = 80152,
  [SMALL_STATE(2671)] = 80165,
  [SMALL_STATE(2672)] = 80178,
  [SMALL_STATE(2673)] = 80191,
  [SMALL_STATE(2674)] = 80204,
  [SMALL_STATE(2675)] = 80217,
  [SMALL_STATE(2676)] = 80230,
  [SMALL_STATE(2677)] = 80243,
  [SMALL_STATE(2678)] = 80256,
  [SMALL_STATE(2679)] = 80269,
  [SMALL_STATE(2680)] = 80282,
  [SMALL_STATE(2681)] = 80295,
  [SMALL_STATE(2682)] = 80308,
  [SMALL_STATE(2683)] = 80321,
  [SMALL_STATE(2684)] = 80334,
  [SMALL_STATE(2685)] = 80347,
  [SMALL_STATE(2686)] = 80360,
  [SMALL_STATE(2687)] = 80373,
  [SMALL_STATE(2688)] = 80386,
  [SMALL_STATE(2689)] = 80399,
  [SMALL_STATE(2690)] = 80412,
  [SMALL_STATE(2691)] = 80425,
  [SMALL_STATE(2692)] = 80438,
  [SMALL_STATE(2693)] = 80451,
  [SMALL_STATE(2694)] = 80464,
  [SMALL_STATE(2695)] = 80477,
  [SMALL_STATE(2696)] = 80490,
  [SMALL_STATE(2697)] = 80503,
  [SMALL_STATE(2698)] = 80516,
  [SMALL_STATE(2699)] = 80529,
  [SMALL_STATE(2700)] = 80542,
  [SMALL_STATE(2701)] = 80555,
  [SMALL_STATE(2702)] = 80568,
  [SMALL_STATE(2703)] = 80581,
  [SMALL_STATE(2704)] = 80594,
  [SMALL_STATE(2705)] = 80607,
  [SMALL_STATE(2706)] = 80620,
  [SMALL_STATE(2707)] = 80633,
  [SMALL_STATE(2708)] = 80646,
  [SMALL_STATE(2709)] = 80659,
  [SMALL_STATE(2710)] = 80672,
  [SMALL_STATE(2711)] = 80685,
  [SMALL_STATE(2712)] = 80698,
  [SMALL_STATE(2713)] = 80711,
  [SMALL_STATE(2714)] = 80724,
  [SMALL_STATE(2715)] = 80737,
  [SMALL_STATE(2716)] = 80750,
  [SMALL_STATE(2717)] = 80763,
  [SMALL_STATE(2718)] = 80776,
  [SMALL_STATE(2719)] = 80789,
  [SMALL_STATE(2720)] = 80802,
  [SMALL_STATE(2721)] = 80815,
  [SMALL_STATE(2722)] = 80828,
  [SMALL_STATE(2723)] = 80841,
  [SMALL_STATE(2724)] = 80854,
  [SMALL_STATE(2725)] = 80867,
  [SMALL_STATE(2726)] = 80880,
  [SMALL_STATE(2727)] = 80893,
  [SMALL_STATE(2728)] = 80906,
  [SMALL_STATE(2729)] = 80919,
  [SMALL_STATE(2730)] = 80932,
  [SMALL_STATE(2731)] = 80945,
  [SMALL_STATE(2732)] = 80958,
  [SMALL_STATE(2733)] = 80971,
  [SMALL_STATE(2734)] = 80984,
  [SMALL_STATE(2735)] = 80997,
  [SMALL_STATE(2736)] = 81010,
  [SMALL_STATE(2737)] = 81023,
  [SMALL_STATE(2738)] = 81036,
  [SMALL_STATE(2739)] = 81049,
  [SMALL_STATE(2740)] = 81062,
  [SMALL_STATE(2741)] = 81066,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1689),
  [5] = {.entry = {.count = 1, .reusable = false}}, SHIFT_EXTRA(),
  [7] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_program, 0, 0, 0),
  [9] = {.entry = {.count = 1, .reusable = false}}, SHIFT(66),
  [11] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1689),
  [13] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1935),
  [15] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1935),
  [17] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0),
  [19] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(934),
  [22] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(551),
  [25] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1372),
  [28] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1884),
  [31] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1740),
  [34] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1412),
  [37] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1515),
  [40] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1227),
  [43] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2626),
  [46] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(55),
  [49] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2706),
  [52] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2714),
  [55] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2716),
  [58] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0),
  [60] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2643),
  [63] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1331),
  [66] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1337),
  [69] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1328),
  [72] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1339),
  [75] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1280),
  [78] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(122),
  [81] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1927),
  [84] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2575),
  [87] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(230),
  [90] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1932),
  [93] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2585),
  [96] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2690),
  [99] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1176),
  [102] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2459),
  [105] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2543),
  [108] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(196),
  [111] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(201),
  [114] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(219),
  [117] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(302),
  [120] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2486),
  [123] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(112),
  [126] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2709),
  [129] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2712),
  [132] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2400),
  [135] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2304),
  [138] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2475),
  [141] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(264),
  [144] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(264),
  [147] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(277),
  [150] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(586),
  [153] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(301),
  [156] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(891),
  [159] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(709),
  [162] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2512),
  [165] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(143),
  [168] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1590),
  [171] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1102),
  [174] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1110),
  [177] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1502),
  [180] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1413),
  [183] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1697),
  [186] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2260),
  [189] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1414),
  [192] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1698),
  [195] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(120),
  [198] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(310),
  [201] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(312),
  [204] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(313),
  [207] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(314),
  [210] = {.entry = {.count = 1, .reusable = false}}, SHIFT(934),
  [212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(551),
  [214] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1372),
  [216] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1884),
  [218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1740),
  [220] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1412),
  [222] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1515),
  [224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1227),
  [226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2626),
  [228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(55),
  [230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_default_statement, 3, 0, 0),
  [232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2706),
  [234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2714),
  [236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2716),
  [238] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_default_statement, 3, 0, 0),
  [240] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2643),
  [242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1331),
  [244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1337),
  [246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1328),
  [248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1339),
  [250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1280),
  [252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1927),
  [256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2575),
  [258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(230),
  [260] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1932),
  [262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2585),
  [264] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2690),
  [266] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1176),
  [268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2459),
  [270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2543),
  [272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(196),
  [274] = {.entry = {.count = 1, .reusable = false}}, SHIFT(201),
  [276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(219),
  [278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(302),
  [280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2486),
  [282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(112),
  [284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2709),
  [286] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2712),
  [288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2400),
  [290] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2304),
  [292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2475),
  [294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(264),
  [296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(586),
  [302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(301),
  [304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(891),
  [306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(709),
  [308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2512),
  [310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [312] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1590),
  [314] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1102),
  [316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1110),
  [318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1502),
  [320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1413),
  [322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1697),
  [324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2260),
  [326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1414),
  [328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1698),
  [330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(120),
  [332] = {.entry = {.count = 1, .reusable = false}}, SHIFT(310),
  [334] = {.entry = {.count = 1, .reusable = false}}, SHIFT(312),
  [336] = {.entry = {.count = 1, .reusable = false}}, SHIFT(313),
  [338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(314),
  [340] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_default_statement, 2, 0, 0),
  [342] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_default_statement, 2, 0, 0),
  [344] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 4, 0, 135),
  [346] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 4, 0, 135),
  [348] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 3, 0, 135),
  [350] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 3, 0, 135),
  [352] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_colon_block, 2, 0, 0),
  [354] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_colon_block, 1, 0, 0),
  [356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(481),
  [358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8),
  [360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(540),
  [362] = {.entry = {.count = 1, .reusable = false}}, SHIFT(971),
  [364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2072),
  [366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1371),
  [368] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1939),
  [370] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1814),
  [372] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1411),
  [374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1228),
  [376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(60),
  [378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2692),
  [380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2693),
  [382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2694),
  [384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2697),
  [386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(223),
  [388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1903),
  [390] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2661),
  [392] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2662),
  [394] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2343),
  [396] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2561),
  [398] = {.entry = {.count = 1, .reusable = false}}, SHIFT(193),
  [400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(194),
  [402] = {.entry = {.count = 1, .reusable = false}}, SHIFT(195),
  [404] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2309),
  [406] = {.entry = {.count = 1, .reusable = false}}, SHIFT(117),
  [408] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2717),
  [410] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2730),
  [412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2315),
  [414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2346),
  [416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1377),
  [418] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2729),
  [421] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2480),
  [424] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2736),
  [427] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2738),
  [430] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2344),
  [433] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2729),
  [435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2480),
  [437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2736),
  [439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2738),
  [441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2344),
  [443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2699),
  [445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2423),
  [447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2727),
  [449] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2735),
  [451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2469),
  [453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(63),
  [457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(69),
  [461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(70),
  [463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(72),
  [467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(74),
  [469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(75),
  [471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(79),
  [473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(57),
  [475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(54),
  [477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(106),
  [481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(92),
  [483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(99),
  [485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(102),
  [489] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2182),
  [491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1993),
  [493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(12),
  [495] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1958),
  [497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(535),
  [499] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2221),
  [501] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2224),
  [503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(790),
  [505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(799),
  [507] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2112),
  [509] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2134),
  [511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2136),
  [513] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_program, 1, 0, 0),
  [515] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2163),
  [517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2164),
  [519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2165),
  [521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2168),
  [523] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2180),
  [525] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2183),
  [527] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2185),
  [529] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2187),
  [531] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2188),
  [533] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2198),
  [535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2201),
  [537] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2203),
  [539] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2204),
  [541] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2209),
  [543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1099),
  [545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1105),
  [547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(452),
  [549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(456),
  [551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1741),
  [553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1745),
  [555] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_program, 3, 0, 0),
  [557] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2296),
  [559] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2143),
  [561] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1992),
  [563] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2108),
  [565] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_program, 2, 0, 0),
  [567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1991),
  [569] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1994),
  [571] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2006),
  [573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2023),
  [575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2025),
  [577] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2034),
  [579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2235),
  [581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2241),
  [583] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2276),
  [585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2283),
  [587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1989),
  [589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(520),
  [591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(749),
  [593] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_yield_expression, 1, 0, 0),
  [595] = {.entry = {.count = 1, .reusable = false}}, SHIFT(703),
  [597] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1542),
  [599] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_yield_expression, 1, 0, 0),
  [601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1881),
  [603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2675),
  [605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(125),
  [607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [609] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2704),
  [611] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1018),
  [613] = {.entry = {.count = 1, .reusable = false}}, SHIFT(269),
  [615] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2345),
  [617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(270),
  [619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(587),
  [625] = {.entry = {.count = 1, .reusable = false}}, SHIFT(272),
  [627] = {.entry = {.count = 1, .reusable = false}}, SHIFT(882),
  [629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(699),
  [631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(142),
  [633] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1031),
  [635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1041),
  [637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1423),
  [639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1701),
  [641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2162),
  [643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1425),
  [645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1703),
  [647] = {.entry = {.count = 1, .reusable = false}}, SHIFT(118),
  [649] = {.entry = {.count = 1, .reusable = false}}, SHIFT(279),
  [651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(273),
  [653] = {.entry = {.count = 1, .reusable = false}}, SHIFT(274),
  [655] = {.entry = {.count = 1, .reusable = false}}, SHIFT(275),
  [657] = {.entry = {.count = 1, .reusable = false}}, SHIFT(276),
  [659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(130),
  [661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(336),
  [665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(337),
  [667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(337),
  [669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [671] = {.entry = {.count = 1, .reusable = false}}, SHIFT(584),
  [673] = {.entry = {.count = 1, .reusable = false}}, SHIFT(339),
  [675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(139),
  [677] = {.entry = {.count = 1, .reusable = false}}, SHIFT(119),
  [679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(344),
  [681] = {.entry = {.count = 1, .reusable = false}}, SHIFT(340),
  [683] = {.entry = {.count = 1, .reusable = false}}, SHIFT(341),
  [685] = {.entry = {.count = 1, .reusable = false}}, SHIFT(342),
  [687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(343),
  [689] = {.entry = {.count = 1, .reusable = false}}, SHIFT(887),
  [691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(700),
  [693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(324),
  [697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(387),
  [703] = {.entry = {.count = 1, .reusable = false}}, SHIFT(388),
  [705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(388),
  [707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [709] = {.entry = {.count = 1, .reusable = false}}, SHIFT(585),
  [711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(390),
  [713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(133),
  [715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(121),
  [717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(395),
  [719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(391),
  [721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(392),
  [723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(393),
  [725] = {.entry = {.count = 1, .reusable = false}}, SHIFT(394),
  [727] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2427),
  [729] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2637),
  [731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(708),
  [733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(941),
  [737] = {.entry = {.count = 1, .reusable = false}}, SHIFT(993),
  [739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(902),
  [741] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1519),
  [743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2698),
  [745] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1739),
  [747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(591),
  [749] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1744),
  [751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2328),
  [755] = {.entry = {.count = 1, .reusable = false}}, SHIFT(247),
  [757] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2139),
  [759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1546),
  [761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1215),
  [763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1216),
  [765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(733),
  [767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(869),
  [769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(817),
  [771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(633),
  [773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(871),
  [775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(947),
  [777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1020),
  [779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(796),
  [781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(895),
  [785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(653),
  [787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(833),
  [789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(866),
  [791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(641),
  [793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(742),
  [795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(835),
  [797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(924),
  [799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(926),
  [801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(744),
  [803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(590),
  [805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(589),
  [807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(893),
  [809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2539),
  [811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1008),
  [813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2605),
  [815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(972),
  [817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2601),
  [819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(801),
  [821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2725),
  [823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(849),
  [825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(965),
  [827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(933),
  [829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2563),
  [831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1107),
  [833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2517),
  [835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(990),
  [837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(931),
  [839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(769),
  [841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(837),
  [843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2523),
  [845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1033),
  [847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(848),
  [849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(776),
  [851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1114),
  [853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1025),
  [855] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__array_destructing_repeat1, 1, 0, 0),
  [857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1014),
  [863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1091),
  [865] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__array_destructing_repeat1, 1, 0, 0), SHIFT(985),
  [868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2033),
  [872] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__array_destructing_repeat1, 1, 0, 0), SHIFT(805),
  [875] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__list_destructing_repeat1, 1, 0, 0),
  [877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(25),
  [879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(22),
  [883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(202),
  [885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(27),
  [889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(190),
  [893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(18),
  [895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(50),
  [897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(197),
  [899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(16),
  [901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(44),
  [905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(48),
  [907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(20),
  [913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(21),
  [915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(31),
  [919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(33),
  [921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(34),
  [923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(38),
  [925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(214),
  [929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(216),
  [931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(204),
  [933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(703),
  [935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(700),
  [937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(958),
  [939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(298),
  [941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(875),
  [943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(609),
  [945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(415),
  [947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(612),
  [949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(414),
  [951] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_condition_list, 3, 0, 0),
  [953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(736),
  [955] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_placeholder, 1, 0, 0),
  [957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(892),
  [961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(961),
  [963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2042),
  [967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(628),
  [971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(364),
  [973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(823),
  [975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(826),
  [977] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_condition_list, 2, 0, 0),
  [979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2265),
  [981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [987] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 3, 0, 6),
  [989] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 3, 0, 6),
  [991] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2632),
  [993] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2306),
  [995] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0),
  [997] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0),
  [999] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2632),
  [1002] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2306),
  [1005] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 3, 0, 23),
  [1007] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 3, 0, 23),
  [1009] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_statement, 3, 0, 23), SHIFT(2465),
  [1012] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_statement, 3, 0, 23), SHIFT(111),
  [1015] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 4, 0, 53),
  [1017] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 4, 0, 53),
  [1019] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_statement, 4, 0, 53), SHIFT(2465),
  [1022] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_statement, 4, 0, 53), SHIFT(111),
  [1025] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2465),
  [1027] = {.entry = {.count = 1, .reusable = false}}, SHIFT(115),
  [1029] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_catch_clause, 5, 0, 164),
  [1031] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_catch_clause, 5, 0, 164),
  [1033] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_if_statement_repeat1, 2, 0, 96),
  [1035] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_if_statement_repeat1, 2, 0, 96),
  [1037] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_if_statement_repeat1, 2, 0, 96), SHIFT_REPEAT(2465),
  [1040] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compound_statement, 2, 0, 0),
  [1042] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_compound_statement, 2, 0, 0),
  [1044] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_finally_clause, 2, 0, 6),
  [1046] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_finally_clause, 2, 0, 6),
  [1048] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 1, 0, 0),
  [1050] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 1, 0, 0),
  [1052] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_catch_clause, 6, 0, 187),
  [1054] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_catch_clause, 6, 0, 187),
  [1056] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compound_statement, 3, 0, 0),
  [1058] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_compound_statement, 3, 0, 0),
  [1060] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_statement, 5, 0, 23),
  [1062] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_statement, 5, 0, 23),
  [1064] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 170),
  [1066] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 170),
  [1068] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 1, 0, 0),
  [1070] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 1, 0, 0),
  [1072] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_statement, 1, 0, 0),
  [1074] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_statement, 1, 0, 0),
  [1076] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_if_statement_repeat1, 1, 0, 51),
  [1078] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_if_statement_repeat1, 1, 0, 51),
  [1080] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 4, 0, 52),
  [1082] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 4, 0, 52),
  [1084] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_block, 2, 0, 0),
  [1086] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_block, 2, 0, 0),
  [1088] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 165),
  [1090] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 165),
  [1092] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_foreach_statement, 7, 0, 171),
  [1094] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_foreach_statement, 7, 0, 171),
  [1096] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_foreach_statement, 7, 0, 0),
  [1098] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_foreach_statement, 7, 0, 0),
  [1100] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 7, 0, 95),
  [1102] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 7, 0, 95),
  [1104] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_static_declaration, 3, 0, 0),
  [1106] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_static_declaration, 3, 0, 0),
  [1108] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_declaration, 4, 0, 72),
  [1110] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_declaration, 4, 0, 72),
  [1112] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_declaration, 4, 0, 72),
  [1114] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interface_declaration, 4, 0, 72),
  [1116] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration, 4, 0, 72),
  [1118] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration, 4, 0, 72),
  [1120] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 4, 0, 72),
  [1122] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 4, 0, 72),
  [1124] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 4, 0, 74),
  [1126] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 4, 0, 74),
  [1128] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__const_declaration, 5, 0, 46),
  [1130] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__const_declaration, 5, 0, 46),
  [1132] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration_list, 3, 0, 0),
  [1134] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration_list, 3, 0, 0),
  [1136] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration_list, 3, 0, 0),
  [1138] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration_list, 3, 0, 0),
  [1140] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration, 5, 0, 92),
  [1142] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration, 5, 0, 92),
  [1144] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 5, 0, 93),
  [1146] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 5, 0, 93),
  [1148] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exit_statement, 5, 0, 0),
  [1150] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exit_statement, 5, 0, 0),
  [1152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unset_statement, 5, 0, 0),
  [1154] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unset_statement, 5, 0, 0),
  [1156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__semicolon, 1, 0, 0),
  [1158] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__semicolon, 1, 0, 0),
  [1160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declare_statement, 5, 0, 0),
  [1162] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declare_statement, 5, 0, 0),
  [1164] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unset_statement, 7, 0, 0),
  [1166] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unset_statement, 7, 0, 0),
  [1168] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_statement, 5, 0, 94),
  [1170] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_statement, 5, 0, 94),
  [1172] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_named_label_statement, 2, 0, 0),
  [1174] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_named_label_statement, 2, 0, 0),
  [1176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_declaration, 1, 0, 1),
  [1178] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_const_declaration, 1, 0, 1),
  [1180] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_clause, 2, 0, 6),
  [1182] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_clause, 2, 0, 6),
  [1184] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_echo_statement, 3, 0, 0),
  [1186] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_echo_statement, 3, 0, 0),
  [1188] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 5, 0, 95),
  [1190] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 5, 0, 95),
  [1192] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 166),
  [1194] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 166),
  [1196] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 5, 0, 23),
  [1198] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 5, 0, 23),
  [1200] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 2, 0, 12),
  [1202] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 2, 0, 12),
  [1204] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration, 7, 0, 174),
  [1206] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration, 7, 0, 174),
  [1208] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_block, 3, 0, 0),
  [1210] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_block, 3, 0, 0),
  [1212] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_declaration, 5, 0, 110),
  [1214] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interface_declaration, 5, 0, 110),
  [1216] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 7, 0, 176),
  [1218] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 7, 0, 176),
  [1220] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 167),
  [1222] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 167),
  [1224] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declare_statement, 8, 0, 0),
  [1226] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declare_statement, 8, 0, 0),
  [1228] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration, 5, 0, 110),
  [1230] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration, 5, 0, 110),
  [1232] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 168),
  [1234] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 168),
  [1236] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 0),
  [1238] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 0),
  [1240] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_global_declaration, 3, 0, 0),
  [1242] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_global_declaration, 3, 0, 0),
  [1244] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_definition, 3, 0, 15),
  [1246] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_namespace_definition, 3, 0, 15),
  [1248] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 5, 0, 110),
  [1250] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 5, 0, 110),
  [1252] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_definition, 3, 0, 16),
  [1254] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_namespace_definition, 3, 0, 16),
  [1256] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 188),
  [1258] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 188),
  [1260] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 189),
  [1262] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 189),
  [1264] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_declaration, 3, 0, 0),
  [1266] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_namespace_use_declaration, 3, 0, 0),
  [1268] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_declaration, 3, 0, 18),
  [1270] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_namespace_use_declaration, 3, 0, 18),
  [1272] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 190),
  [1274] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 190),
  [1276] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 191),
  [1278] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 191),
  [1280] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__const_declaration, 3, 0, 0),
  [1282] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__const_declaration, 3, 0, 0),
  [1284] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 192),
  [1286] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 192),
  [1288] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 193),
  [1290] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 193),
  [1292] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_statement, 2, 0, 0),
  [1294] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_statement, 2, 0, 0),
  [1296] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 9, 0, 171),
  [1298] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 9, 0, 171),
  [1300] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 9, 0, 166),
  [1302] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 9, 0, 166),
  [1304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_goto_statement, 3, 0, 0),
  [1306] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_goto_statement, 3, 0, 0),
  [1308] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 9, 0, 168),
  [1310] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 9, 0, 168),
  [1312] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 5, 0, 113),
  [1314] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 5, 0, 113),
  [1316] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__const_declaration, 5, 0, 114),
  [1318] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__const_declaration, 5, 0, 114),
  [1320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__const_declaration, 5, 0, 0),
  [1322] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__const_declaration, 5, 0, 0),
  [1324] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 5, 0, 115),
  [1326] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 5, 0, 115),
  [1328] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_continue_statement, 3, 0, 0),
  [1330] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_continue_statement, 3, 0, 0),
  [1332] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_break_statement, 3, 0, 0),
  [1334] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_break_statement, 3, 0, 0),
  [1336] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration, 6, 0, 131),
  [1338] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration, 6, 0, 131),
  [1340] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unset_statement, 6, 0, 0),
  [1342] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unset_statement, 6, 0, 0),
  [1344] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 6, 0, 132),
  [1346] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 6, 0, 132),
  [1348] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 6, 0, 0),
  [1350] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 6, 0, 0),
  [1352] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 9, 0, 170),
  [1354] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 9, 0, 170),
  [1356] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 3, 0, 0),
  [1358] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_statement, 3, 0, 0),
  [1360] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_definition, 2, 0, 6),
  [1362] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_namespace_definition, 2, 0, 6),
  [1364] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_statement, 3, 0, 23),
  [1366] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_statement, 3, 0, 23),
  [1368] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_declaration, 3, 0, 15),
  [1370] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_declaration, 3, 0, 15),
  [1372] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_if_clause, 3, 0, 23),
  [1374] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_if_clause, 3, 0, 23),
  [1376] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_statement, 3, 0, 23),
  [1378] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_statement, 3, 0, 23),
  [1380] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_declaration, 3, 0, 15),
  [1382] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interface_declaration, 3, 0, 15),
  [1384] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration, 3, 0, 15),
  [1386] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration, 3, 0, 15),
  [1388] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 9, 0, 203),
  [1390] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 9, 0, 203),
  [1392] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 9, 0, 204),
  [1394] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 9, 0, 204),
  [1396] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_foreach_statement, 9, 0, 171),
  [1398] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_foreach_statement, 9, 0, 171),
  [1400] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 10, 0, 209),
  [1402] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 10, 0, 209),
  [1404] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exit_statement, 2, 0, 0),
  [1406] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exit_statement, 2, 0, 0),
  [1408] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_continue_statement, 2, 0, 0),
  [1410] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_continue_statement, 2, 0, 0),
  [1412] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_break_statement, 2, 0, 0),
  [1414] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_break_statement, 2, 0, 0),
  [1416] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 2, 0, 0),
  [1418] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_statement, 2, 0, 0),
  [1420] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 10, 0, 210),
  [1422] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 10, 0, 210),
  [1424] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_empty_statement, 1, 0, 0),
  [1426] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_empty_statement, 1, 0, 0),
  [1428] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 10, 0, 189),
  [1430] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 10, 0, 189),
  [1432] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 10, 0, 211),
  [1434] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 10, 0, 211),
  [1436] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 6, 0, 52),
  [1438] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 6, 0, 52),
  [1440] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 10, 0, 191),
  [1442] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 10, 0, 191),
  [1444] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 10, 0, 193),
  [1446] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 10, 0, 193),
  [1448] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 6, 0, 53),
  [1450] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 6, 0, 53),
  [1452] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 3, 0, 15),
  [1454] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 3, 0, 15),
  [1456] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 3, 0, 33),
  [1458] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 3, 0, 33),
  [1460] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_static_declaration, 4, 0, 0),
  [1462] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_static_declaration, 4, 0, 0),
  [1464] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_global_declaration, 4, 0, 0),
  [1466] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_global_declaration, 4, 0, 0),
  [1468] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_declaration, 4, 0, 0),
  [1470] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_namespace_use_declaration, 4, 0, 0),
  [1472] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 11, 0, 213),
  [1474] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 11, 0, 213),
  [1476] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__const_declaration, 4, 0, 46),
  [1478] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__const_declaration, 4, 0, 46),
  [1480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__const_declaration, 4, 0, 0),
  [1482] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__const_declaration, 4, 0, 0),
  [1484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration_list, 2, 0, 0),
  [1486] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration_list, 2, 0, 0),
  [1488] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_declaration, 4, 0, 48),
  [1490] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interface_declaration, 4, 0, 48),
  [1492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration_list, 2, 0, 0),
  [1494] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration_list, 2, 0, 0),
  [1496] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration, 4, 0, 48),
  [1498] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration, 4, 0, 48),
  [1500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_block, 4, 0, 0),
  [1502] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_block, 4, 0, 0),
  [1504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 4, 0, 48),
  [1506] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 4, 0, 48),
  [1508] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration, 6, 0, 145),
  [1510] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration, 6, 0, 145),
  [1512] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 6, 0, 146),
  [1514] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 6, 0, 146),
  [1516] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 6, 0, 149),
  [1518] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 6, 0, 149),
  [1520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__const_declaration, 6, 0, 114),
  [1522] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__const_declaration, 6, 0, 114),
  [1524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 6, 0, 150),
  [1526] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 6, 0, 150),
  [1528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 11, 0, 214),
  [1530] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 11, 0, 214),
  [1532] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exit_statement, 4, 0, 0),
  [1534] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exit_statement, 4, 0, 0),
  [1536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 11, 0, 215),
  [1538] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 11, 0, 215),
  [1540] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 11, 0, 204),
  [1542] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 11, 0, 204),
  [1544] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 12, 0, 216),
  [1546] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 12, 0, 216),
  [1548] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 169),
  [1550] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 169),
  [1552] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declare_statement, 7, 0, 0),
  [1554] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declare_statement, 7, 0, 0),
  [1556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [1558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1666),
  [1560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(123),
  [1562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1651),
  [1564] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [1566] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [1568] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 4, 0, 0),
  [1570] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 4, 0, 0),
  [1572] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [1574] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 5, 0, 0),
  [1576] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 5, 0, 0),
  [1578] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 2, 0, 0),
  [1580] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 2, 0, 0),
  [1582] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 3, 0, 0),
  [1584] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 3, 0, 0),
  [1586] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_member_access_expression, 3, 0, 31),
  [1588] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_member_access_expression, 3, 0, 31),
  [1590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(134),
  [1592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_nullsafe_member_access_expression, 3, 0, 31),
  [1594] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_nullsafe_member_access_expression, 3, 0, 31),
  [1596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_property_access_expression, 3, 0, 25),
  [1598] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scoped_property_access_expression, 3, 0, 25),
  [1600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__member_name, 1, 0, 5),
  [1602] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__callable_variable, 1, 0, 0),
  [1604] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__callable_variable, 1, 0, 0),
  [1606] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_variable, 1, 0, 0),
  [1608] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_variable, 1, 0, 0),
  [1610] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__identifier, 1, 0, 0),
  [1612] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__identifier, 1, 0, 0),
  [1614] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__reserved_identifier, 1, 0, 0),
  [1616] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__reserved_identifier, 1, 0, 0),
  [1618] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_name, 2, 0, 0),
  [1620] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variable_name, 2, 0, 0),
  [1622] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dynamic_variable_name, 2, 0, 0),
  [1624] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dynamic_variable_name, 2, 0, 0),
  [1626] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__new_variable, 1, 0, 0),
  [1628] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__new_variable, 1, 0, 0),
  [1630] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__variable_scoped_property_access_expression, 3, 0, 25),
  [1632] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__variable_scoped_property_access_expression, 3, 0, 25),
  [1634] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__variable_subscript_expression, 4, 0, 0),
  [1636] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__variable_subscript_expression, 4, 0, 0),
  [1638] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_call_expression, 2, 0, 11),
  [1640] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_call_expression, 2, 0, 11),
  [1642] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__member_name, 1, 0, 5),
  [1644] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__member_name, 3, 0, 16),
  [1646] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__member_name, 3, 0, 16),
  [1648] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__dereferencable_subscript_expression, 3, 0, 0),
  [1650] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__dereferencable_subscript_expression, 3, 0, 0),
  [1652] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__variable_member_access_expression, 3, 0, 31),
  [1654] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__variable_member_access_expression, 3, 0, 31),
  [1656] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__variable_nullsafe_member_access_expression, 3, 0, 31),
  [1658] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__variable_nullsafe_member_access_expression, 3, 0, 31),
  [1660] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__variable_subscript_expression, 3, 0, 0),
  [1662] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__variable_subscript_expression, 3, 0, 0),
  [1664] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 0),
  [1666] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 0),
  [1668] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cast_variable, 4, 0, 50),
  [1670] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0),
  [1672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__dereferencable_expression, 1, 0, 0),
  [1674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1037),
  [1676] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cast_variable, 4, 0, 50),
  [1678] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__new_variable, 1, 0, 3),
  [1680] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__new_variable, 1, 0, 3),
  [1682] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_call_expression, 4, 0, 68),
  [1684] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scoped_call_expression, 4, 0, 68),
  [1686] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dynamic_variable_name, 4, 0, 0),
  [1688] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dynamic_variable_name, 4, 0, 0),
  [1690] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__new_variable, 1, 0, 4),
  [1692] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__new_variable, 1, 0, 4),
  [1694] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__dereferencable_subscript_expression, 4, 0, 0),
  [1696] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__dereferencable_subscript_expression, 4, 0, 0),
  [1698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_member_call_expression, 4, 0, 69),
  [1700] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_member_call_expression, 4, 0, 69),
  [1702] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_nullsafe_member_call_expression, 4, 0, 69),
  [1704] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_nullsafe_member_call_expression, 4, 0, 69),
  [1706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [1708] = {.entry = {.count = 1, .reusable = false}}, SHIFT(259),
  [1710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [1712] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1549),
  [1714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1424),
  [1716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [1718] = {.entry = {.count = 1, .reusable = false}}, SHIFT(885),
  [1720] = {.entry = {.count = 1, .reusable = false}}, SHIFT(249),
  [1722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1085),
  [1724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [1726] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1444),
  [1728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1422),
  [1730] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1549),
  [1733] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1590),
  [1736] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2675),
  [1739] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0),
  [1741] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(132),
  [1744] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2704),
  [1747] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(885),
  [1750] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(160),
  [1753] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1423),
  [1756] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1701),
  [1759] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2162),
  [1762] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1703),
  [1765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(253),
  [1767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(362),
  [1769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2123),
  [1771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(239),
  [1773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(413),
  [1775] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 0), REDUCE(sym__array_destructing_element, 1, 0, 0),
  [1778] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__list_destructing_repeat1, 2, 0, 0),
  [1780] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_destructing_element, 1, 0, 0),
  [1782] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 0), REDUCE(sym__array_destructing_element, 3, 0, 0),
  [1785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2274),
  [1787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2068),
  [1789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2111),
  [1791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1957),
  [1793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [1795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(131),
  [1797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1708),
  [1799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [1801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1656),
  [1803] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_qualified_name, 2, 0, 8),
  [1805] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_qualified_name, 2, 0, 8),
  [1807] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_namespace_name, 2, 0, 0), SHIFT(2524),
  [1810] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_qualified_name, 3, 0, 26),
  [1812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_qualified_name, 3, 0, 26),
  [1814] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_static_modifier, 1, 0, 0),
  [1816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(135),
  [1818] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1441),
  [1820] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1332),
  [1822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1558),
  [1824] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1236),
  [1826] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1496),
  [1828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1349),
  [1830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1382),
  [1832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2546),
  [1834] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_namespace_name, 1, 0, 0), SHIFT(2524),
  [1837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(137),
  [1839] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_constant_access_expression, 3, 0, 0),
  [1841] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_constant_access_expression, 3, 0, 0),
  [1843] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_constant_access_expression, 5, 0, 107),
  [1845] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_constant_access_expression, 5, 0, 107),
  [1847] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_by_ref, 2, 0, 0),
  [1849] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_by_ref, 2, 0, 0),
  [1851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1596),
  [1853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2722),
  [1855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2610),
  [1857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1362),
  [1859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1929),
  [1861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2307),
  [1863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1379),
  [1865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1500),
  [1867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2514),
  [1869] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 4, 0, 97),
  [1871] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 4, 0, 97),
  [1873] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 4, 0, 98),
  [1875] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 4, 0, 98),
  [1877] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_creation_expression, 5, 0, 0),
  [1879] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_creation_expression, 5, 0, 0),
  [1881] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 5, 0, 99),
  [1883] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 5, 0, 99),
  [1885] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 5, 0, 100),
  [1887] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 5, 0, 100),
  [1889] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 5, 0, 101),
  [1891] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 5, 0, 101),
  [1893] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 5, 0, 102),
  [1895] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 5, 0, 102),
  [1897] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_nowdoc, 5, 0, 101),
  [1899] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_nowdoc, 5, 0, 101),
  [1901] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__string, 1, 0, 0),
  [1903] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__string, 1, 0, 0),
  [1905] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_creation_expression, 6, 0, 22),
  [1907] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_creation_expression, 6, 0, 22),
  [1909] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 5, 0, 136),
  [1911] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 5, 0, 136),
  [1913] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 5, 0, 137),
  [1915] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 5, 0, 137),
  [1917] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 6, 0, 140),
  [1919] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 6, 0, 140),
  [1921] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 6, 0, 141),
  [1923] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 6, 0, 141),
  [1925] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 6, 0, 142),
  [1927] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 6, 0, 142),
  [1929] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_nowdoc, 6, 0, 141),
  [1931] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_nowdoc, 6, 0, 141),
  [1933] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 6, 0, 172),
  [1935] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 6, 0, 172),
  [1937] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 6, 0, 132),
  [1939] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 6, 0, 132),
  [1941] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 7, 0, 173),
  [1943] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 7, 0, 173),
  [1945] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_nowdoc, 7, 0, 173),
  [1947] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_nowdoc, 7, 0, 173),
  [1949] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 7, 0, 194),
  [1951] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 7, 0, 194),
  [1953] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_literal, 1, 0, 0),
  [1955] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_literal, 1, 0, 0),
  [1957] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__dereferencable_scalar, 1, 0, 0),
  [1959] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_creation_expression, 1, 0, 0),
  [1961] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_creation_expression, 1, 0, 0),
  [1963] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 2),
  [1965] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__dereferencable_expression, 1, 0, 2),
  [1967] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__new_dereferencable_expression, 2, 0, 0),
  [1969] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__new_dereferencable_expression, 2, 0, 0),
  [1971] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_name_reference, 1, 0, 0),
  [1973] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_name_reference, 1, 0, 0),
  [1975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1885),
  [1977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1492),
  [1979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1493),
  [1981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [1983] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_update_expression, 2, 0, 9),
  [1985] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_update_expression, 2, 0, 9),
  [1987] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_creation_expression, 2, 0, 0),
  [1989] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_creation_expression, 2, 0, 0),
  [1991] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__array_destructing, 2, 0, 0), REDUCE(sym_array_creation_expression, 2, 0, 0),
  [1994] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__array_destructing, 2, 0, 0),
  [1996] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_encapsed_string, 2, 0, 0),
  [1998] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_encapsed_string, 2, 0, 0),
  [2000] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 2, 0, 0),
  [2002] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 2, 0, 0),
  [2004] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_creation_expression, 3, 0, 22),
  [2006] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_creation_expression, 3, 0, 22),
  [2008] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__name, 1, 0, 0),
  [2010] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 2, 0, 6),
  [2012] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 2, 0, 6),
  [2014] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__new_dereferencable_expression, 3, 0, 0),
  [2016] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__new_dereferencable_expression, 3, 0, 0),
  [2018] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_creation_expression, 3, 0, 0),
  [2020] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_creation_expression, 3, 0, 0),
  [2022] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_encapsed_string, 3, 0, 0),
  [2024] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_encapsed_string, 3, 0, 0),
  [2026] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 3, 0, 0),
  [2028] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 3, 0, 0),
  [2030] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 3, 0, 24),
  [2032] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 3, 0, 24),
  [2034] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_creation_expression, 4, 0, 22),
  [2036] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_creation_expression, 4, 0, 22),
  [2038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 3, 0, 36),
  [2040] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 3, 0, 36),
  [2042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 3, 0, 54),
  [2044] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 3, 0, 54),
  [2046] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_creation_expression, 4, 0, 0),
  [2048] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_creation_expression, 4, 0, 0),
  [2050] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 4, 0, 58),
  [2052] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 4, 0, 58),
  [2054] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 4, 0, 59),
  [2056] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 4, 0, 59),
  [2058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2680),
  [2060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1758),
  [2062] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 4, 0, 60),
  [2064] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 4, 0, 60),
  [2066] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_creation_expression, 5, 0, 22),
  [2068] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_creation_expression, 5, 0, 22),
  [2070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1953),
  [2072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [2074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1718),
  [2076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1724),
  [2078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1944),
  [2080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(854),
  [2082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(714),
  [2084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2629),
  [2086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1556),
  [2088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(419),
  [2090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(863),
  [2092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(325),
  [2094] = {.entry = {.count = 1, .reusable = false}}, SHIFT(996),
  [2096] = {.entry = {.count = 1, .reusable = false}}, SHIFT(726),
  [2098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2635),
  [2100] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1555),
  [2102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [2104] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1045),
  [2106] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__argument_name, 2, 0, 61),
  [2108] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__argument_name, 2, 0, 61),
  [2110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [2112] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_reference_modifier, 1, 0, 0),
  [2114] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_reference_modifier, 1, 0, 0),
  [2116] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__name, 1, 0, 0),
  [2118] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__new_non_dereferencable_expression, 2, 0, 0),
  [2120] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__new_non_dereferencable_expression, 2, 0, 0),
  [2122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(485),
  [2124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2048),
  [2126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1948),
  [2128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1482),
  [2130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1483),
  [2132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [2134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(892),
  [2136] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_shell_command_expression, 2, 0, 0),
  [2138] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_shell_command_expression, 2, 0, 0),
  [2140] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_clone_expression, 2, 0, 0),
  [2142] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_clone_expression, 2, 0, 0),
  [2144] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_block, 5, 0, 0),
  [2146] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_block, 5, 0, 0),
  [2148] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_shell_command_expression, 3, 0, 0),
  [2150] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_shell_command_expression, 3, 0, 0),
  [2152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_expression, 3, 0, 29),
  [2154] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary_expression, 3, 0, 29),
  [2156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_expression, 3, 0, 23),
  [2158] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_expression, 3, 0, 23),
  [2160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_destructing, 2, 0, 0),
  [2162] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression, 1, 0, 0),
  [2164] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression, 1, 0, 0),
  [2166] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_op_expression, 2, 0, 9),
  [2168] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unary_op_expression, 2, 0, 9),
  [2170] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_error_suppression_expression, 2, 0, 0),
  [2172] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_error_suppression_expression, 2, 0, 0),
  [2174] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_yield_expression, 2, 0, 0),
  [2176] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_yield_expression, 2, 0, 0),
  [2178] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_block, 4, 0, 0),
  [2180] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_block, 4, 0, 0),
  [2182] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_element_initializer, 1, 0, 0),
  [2184] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_element_initializer, 1, 0, 0),
  [2186] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_element_initializer, 3, 0, 0),
  [2188] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_element_initializer, 3, 0, 0),
  [2190] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function, 2, 0, 13),
  [2192] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_function, 2, 0, 13),
  [2194] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cast_expression, 4, 0, 50),
  [2196] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cast_expression, 4, 0, 50),
  [2198] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__unary_expression, 1, 0, 0),
  [2200] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__unary_expression, 1, 0, 0),
  [2202] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_null, 1, 0, 0),
  [2204] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_null, 1, 0, 0),
  [2206] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_block, 2, 0, 0),
  [2208] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_block, 2, 0, 0),
  [2210] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_block, 3, 0, 0),
  [2212] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_block, 3, 0, 0),
  [2214] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_update_expression, 2, 0, 14),
  [2216] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_update_expression, 2, 0, 14),
  [2218] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean, 1, 0, 0),
  [2220] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_boolean, 1, 0, 0),
  [2222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1351),
  [2224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [2226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(296),
  [2228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(296),
  [2230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_print_intrinsic, 2, 0, 0),
  [2232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(281),
  [2234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(235),
  [2236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(282),
  [2238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [2240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [2242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [2244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(287),
  [2246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [2248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [2250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(290),
  [2252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [2254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(292),
  [2256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [2258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(293),
  [2260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(293),
  [2262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [2264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [2266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [2268] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_expression, 2, 0, 0),
  [2270] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_once_expression, 2, 0, 0),
  [2272] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_require_expression, 2, 0, 0),
  [2274] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_require_once_expression, 2, 0, 0),
  [2276] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_throw_expression, 2, 0, 0),
  [2278] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_yield_expression, 3, 0, 0),
  [2280] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arrow_function, 3, 0, 28),
  [2282] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_reference_assignment_expression, 4, 0, 67),
  [2284] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_expression, 5, 0, 106),
  [2286] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_conditional_expression, 5, 0, 106),
  [2288] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment_expression, 3, 0, 30),
  [2290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_augmented_assignment_expression, 3, 0, 29),
  [2292] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_unpacking, 2, 0, 0),
  [2294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_expression, 4, 0, 66),
  [2296] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_conditional_expression, 4, 0, 66),
  [2298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(346),
  [2300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(251),
  [2302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(347),
  [2304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(348),
  [2306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(349),
  [2308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [2310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [2312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(352),
  [2314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(438),
  [2316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(354),
  [2318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(355),
  [2320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(356),
  [2322] = {.entry = {.count = 1, .reusable = false}}, SHIFT(357),
  [2324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [2326] = {.entry = {.count = 1, .reusable = false}}, SHIFT(358),
  [2328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(358),
  [2330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(359),
  [2332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(360),
  [2334] = {.entry = {.count = 1, .reusable = false}}, SHIFT(361),
  [2336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(361),
  [2338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(334),
  [2340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(246),
  [2342] = {.entry = {.count = 1, .reusable = false}}, SHIFT(335),
  [2344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(374),
  [2346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(375),
  [2348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(376),
  [2350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(377),
  [2352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(378),
  [2354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [2356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(385),
  [2358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [2360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(417),
  [2362] = {.entry = {.count = 1, .reusable = false}}, SHIFT(418),
  [2364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(418),
  [2366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(422),
  [2368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(422),
  [2370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(430),
  [2372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(431),
  [2374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(262),
  [2376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [2378] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0),
  [2380] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1332),
  [2383] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0),
  [2385] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1331),
  [2388] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1337),
  [2391] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1328),
  [2394] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1339),
  [2397] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1280),
  [2400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [2402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2439),
  [2404] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__modifier, 1, 0, 0),
  [2406] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__modifier, 1, 0, 0),
  [2408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1354),
  [2410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [2412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(397),
  [2414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(241),
  [2416] = {.entry = {.count = 1, .reusable = false}}, SHIFT(398),
  [2418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(399),
  [2420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(400),
  [2422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(401),
  [2424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(402),
  [2426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(403),
  [2428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(404),
  [2430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(405),
  [2432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(406),
  [2434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(407),
  [2436] = {.entry = {.count = 1, .reusable = false}}, SHIFT(408),
  [2438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(408),
  [2440] = {.entry = {.count = 1, .reusable = false}}, SHIFT(409),
  [2442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [2444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(410),
  [2446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(411),
  [2448] = {.entry = {.count = 1, .reusable = false}}, SHIFT(412),
  [2450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(412),
  [2452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [2454] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 4, 0, 118),
  [2456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1459),
  [2458] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_element, 3, 0, 86),
  [2460] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 6, 0, 177),
  [2462] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 7, 0, 201),
  [2464] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 6, 0, 182),
  [2466] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 6, 0, 183),
  [2468] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 5, 0, 151),
  [2470] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 5, 0, 152),
  [2472] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 5, 0, 158),
  [2474] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1541),
  [2476] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expressions, 1, 0, 0),
  [2478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [2480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(229),
  [2482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1748),
  [2484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [2486] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_condition_list, 1, 0, 0),
  [2488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [2490] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_static_variable_declaration, 3, 0, 34),
  [2492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_expression, 3, 0, 0),
  [2494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [2496] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_element, 3, 0, 0),
  [2498] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 5, 0, 155),
  [2500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1332),
  [2502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1527),
  [2504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1558),
  [2506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1227),
  [2508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(475),
  [2510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1331),
  [2512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1337),
  [2514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1328),
  [2516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1339),
  [2518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1280),
  [2520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(715),
  [2522] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 4, 0, 116),
  [2524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 4, 0, 126),
  [2526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(716),
  [2528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(237),
  [2530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [2532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [2534] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__foreach_value, 1, 0, 0),
  [2536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_match_condition_list_repeat1, 2, 0, 0),
  [2538] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument, 1, 0, 0),
  [2540] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 5, 0, 160),
  [2542] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 5, 0, 156),
  [2544] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 3, 0, 86),
  [2546] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_default_expression, 3, 0, 133),
  [2548] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument, 2, 0, 63),
  [2550] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 4, 0, 120),
  [2552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(923),
  [2554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(566),
  [2556] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument, 2, 0, 62),
  [2558] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 6, 0, 184),
  [2560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6),
  [2562] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1332),
  [2565] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1527),
  [2568] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1558),
  [2571] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1227),
  [2574] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0),
  [2576] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1331),
  [2579] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1337),
  [2582] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1328),
  [2585] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1339),
  [2588] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1280),
  [2591] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1502),
  [2594] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument, 3, 0, 103),
  [2596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(964),
  [2598] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_conditional_expression, 3, 0, 134),
  [2600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(650),
  [2602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [2604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(642),
  [2606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1956),
  [2608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(186),
  [2610] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_visibility_modifier, 1, 0, 0),
  [2612] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_visibility_modifier, 1, 0, 0),
  [2614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2713),
  [2616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(800),
  [2618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2013),
  [2620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(639),
  [2622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1667),
  [2624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(617),
  [2626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [2628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(619),
  [2630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(604),
  [2632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(764),
  [2634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(608),
  [2636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(660),
  [2638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1654),
  [2640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(588),
  [2642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1609),
  [2644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(877),
  [2646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(888),
  [2648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(915),
  [2650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(870),
  [2652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(644),
  [2654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(978),
  [2656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(920),
  [2658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(921),
  [2660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(925),
  [2662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(932),
  [2664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1672),
  [2666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(741),
  [2668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(754),
  [2670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(743),
  [2672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1687),
  [2674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(831),
  [2676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(832),
  [2678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(834),
  [2680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(724),
  [2682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(843),
  [2684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1604),
  [2686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [2688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(969),
  [2690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(661),
  [2692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(632),
  [2694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(967),
  [2696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(366),
  [2698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [2700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(183),
  [2702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [2704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1072),
  [2706] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_readonly_modifier, 1, 0, 0),
  [2708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_readonly_modifier, 1, 0, 0),
  [2710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2064),
  [2712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2660),
  [2714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(568),
  [2716] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_final_modifier, 1, 0, 0),
  [2718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_final_modifier, 1, 0, 0),
  [2720] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_static_modifier, 1, 0, 0),
  [2722] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1332),
  [2725] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1527),
  [2728] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1558),
  [2731] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0),
  [2733] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2660),
  [2736] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1331),
  [2739] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1337),
  [2742] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1328),
  [2745] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1339),
  [2748] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1280),
  [2751] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1502),
  [2754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(476),
  [2756] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_visibility_modifier, 4, 0, 49),
  [2758] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_visibility_modifier, 4, 0, 49),
  [2760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2000),
  [2762] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_modifier, 1, 0, 0),
  [2764] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_modifier, 1, 0, 0),
  [2766] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_declaration_repeat1, 1, 0, 0),
  [2768] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_declaration_repeat1, 1, 0, 0),
  [2770] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1572),
  [2772] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1733),
  [2774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1357),
  [2776] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1380),
  [2778] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1986),
  [2780] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2443),
  [2782] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_list, 1, 0, 0),
  [2784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_list, 1, 0, 0),
  [2786] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_attribute_list_repeat1, 2, 0, 0),
  [2788] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_list_repeat1, 2, 0, 0),
  [2790] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1500),
  [2793] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_group, 4, 0, 0),
  [2795] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_group, 4, 0, 0),
  [2797] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_group, 3, 0, 0),
  [2799] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_group, 3, 0, 0),
  [2801] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_attribute_list_repeat1, 1, 0, 0),
  [2803] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_list_repeat1, 1, 0, 0),
  [2805] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_group, 5, 0, 0),
  [2807] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_group, 5, 0, 0),
  [2809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1361),
  [2811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1359),
  [2813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1509),
  [2815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2599),
  [2817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2611),
  [2819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2522),
  [2821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2580),
  [2823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1930),
  [2825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2550),
  [2827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2551),
  [2829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2554),
  [2831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2557),
  [2833] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1502),
  [2836] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_relative_scope, 1, 0, 0),
  [2838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2645),
  [2840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1018),
  [2842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1031),
  [2844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1041),
  [2846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1221),
  [2848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2564),
  [2850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1236),
  [2852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2574),
  [2854] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_heredoc_body_repeat1, 2, 0, 0), SHIFT_REPEAT(381),
  [2857] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_heredoc_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1581),
  [2860] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_heredoc_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1581),
  [2863] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_heredoc_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1447),
  [2866] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_heredoc_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1670),
  [2869] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_heredoc_body_repeat1, 2, 0, 0),
  [2871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1559),
  [2873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1554),
  [2875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1561),
  [2877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1553),
  [2879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2507),
  [2881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1557),
  [2883] = {.entry = {.count = 1, .reusable = false}}, SHIFT(381),
  [2885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1581),
  [2887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1581),
  [2889] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_heredoc_body, 2, 0, 0), SHIFT(1447),
  [2892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1670),
  [2894] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_body, 2, 0, 0),
  [2896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1563),
  [2898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2589),
  [2900] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0),
  [2902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1447),
  [2904] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_heredoc_body_repeat1, 1, 0, 0),
  [2906] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolated_string_body_heredoc, 2, 0, 0), SHIFT_REPEAT(381),
  [2909] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_string_body_heredoc, 2, 0, 0), SHIFT_REPEAT(1581),
  [2912] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolated_string_body_heredoc, 2, 0, 0), SHIFT_REPEAT(1581),
  [2915] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolated_string_body_heredoc, 2, 0, 0),
  [2917] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolated_string_body_heredoc, 2, 0, 0), SHIFT_REPEAT(1670),
  [2920] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_declaration, 4, 0, 0),
  [2922] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 4, 0, 162),
  [2924] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 4, 0, 163),
  [2926] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 3, 0, 128),
  [2928] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 3, 0, 129),
  [2930] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 3, 0, 130),
  [2932] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 2, 0, 12),
  [2934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1517),
  [2936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(317),
  [2938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1674),
  [2940] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1674),
  [2942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(998),
  [2944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1660),
  [2946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [2948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1644),
  [2950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1644),
  [2952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1089),
  [2954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1664),
  [2956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1108),
  [2958] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_execution_operator_body, 2, 0, 0), SHIFT_REPEAT(327),
  [2961] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_execution_operator_body, 2, 0, 0), SHIFT_REPEAT(1644),
  [2964] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolated_execution_operator_body, 2, 0, 0), SHIFT_REPEAT(1644),
  [2967] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolated_execution_operator_body, 2, 0, 0),
  [2969] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_execution_operator_body, 2, 0, 0), SHIFT_REPEAT(1664),
  [2972] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_declaration, 3, 0, 0),
  [2974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(987),
  [2976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(806),
  [2978] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 2, 0, 90),
  [2980] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 3, 0, 33),
  [2982] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_list, 2, 0, 0),
  [2984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(797),
  [2986] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_list, 3, 0, 0),
  [2988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1009),
  [2990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1015),
  [2992] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_string_body, 2, 0, 0), SHIFT_REPEAT(317),
  [2995] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_string_body, 2, 0, 0), SHIFT_REPEAT(1674),
  [2998] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolated_string_body, 2, 0, 0), SHIFT_REPEAT(1674),
  [3001] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolated_string_body, 2, 0, 0),
  [3003] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_string_body, 2, 0, 0), SHIFT_REPEAT(1660),
  [3006] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_case, 6, 0, 202),
  [3008] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 4, 0, 161),
  [3010] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 6, 0, 185),
  [3012] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1518),
  [3014] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 4, 0, 0),
  [3016] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 1, 0, 0),
  [3018] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__member_declaration, 1, 0, 47),
  [3020] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_const_declaration, 1, 0, 1),
  [3022] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__member_declaration, 1, 0, 0),
  [3024] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 4, 0, 46),
  [3026] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 1, 0, 0),
  [3028] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__enum_member_declaration, 1, 0, 0),
  [3030] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_case, 4, 0, 82),
  [3032] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_named_type, 1, 0, 0),
  [3034] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_named_type, 1, 0, 0),
  [3036] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 3, 0, 0),
  [3038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_const_declaration, 2, 0, 89),
  [3040] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_case, 3, 0, 16),
  [3042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1528),
  [3044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [3046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1670),
  [3048] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook_list, 3, 0, 0),
  [3050] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 5, 0, 161),
  [3052] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 5, 0, 185),
  [3054] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 5, 0, 46),
  [3056] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_case, 5, 0, 186),
  [3058] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_const_declaration, 3, 0, 127),
  [3060] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook_list, 2, 0, 0),
  [3062] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_const_declaration, 2, 0, 91),
  [3064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1344),
  [3066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(657),
  [3068] = {.entry = {.count = 1, .reusable = false}}, SHIFT(647),
  [3070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [3072] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1531),
  [3074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2453),
  [3076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2313),
  [3078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1378),
  [3080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(949),
  [3082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(953),
  [3084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [3086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(598),
  [3088] = {.entry = {.count = 1, .reusable = false}}, SHIFT(599),
  [3090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [3092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1853),
  [3094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1854),
  [3096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1347),
  [3098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1454),
  [3100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1448),
  [3102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1375),
  [3104] = {.entry = {.count = 1, .reusable = false}}, SHIFT(717),
  [3106] = {.entry = {.count = 1, .reusable = false}}, SHIFT(745),
  [3108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(329),
  [3110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(878),
  [3112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(884),
  [3114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(384),
  [3116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(421),
  [3118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(731),
  [3120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(723),
  [3122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(424),
  [3124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(426),
  [3126] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_property_hook_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1531),
  [3129] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_property_hook_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1596),
  [3132] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_property_hook_list_repeat1, 2, 0, 0),
  [3134] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_property_hook_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1331),
  [3137] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_property_hook_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1502),
  [3140] = {.entry = {.count = 4, .reusable = false}}, REDUCE(sym_type, 1, 0, 0), REDUCE(sym_union_type, 1, 0, 0), REDUCE(sym_intersection_type, 1, 0, 0), REDUCE(sym_disjunctive_normal_form_type, 1, -1, 0),
  [3145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1363),
  [3147] = {.entry = {.count = 4, .reusable = true}}, REDUCE(sym_type, 1, 0, 0), REDUCE(sym_union_type, 1, 0, 0), REDUCE(sym_intersection_type, 1, 0, 0), REDUCE(sym_disjunctive_normal_form_type, 1, -1, 0),
  [3152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1350),
  [3154] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__types, 1, 0, 0),
  [3156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__types, 1, 0, 0),
  [3158] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_primitive_type, 1, 0, 0),
  [3160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_primitive_type, 1, 0, 0),
  [3162] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_optional_type, 2, 0, 0),
  [3164] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_optional_type, 2, 0, 0),
  [3166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1900),
  [3168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1901),
  [3170] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_union_type_repeat1, 2, 0, 0),
  [3172] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_union_type_repeat1, 2, 0, 0),
  [3174] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_union_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1370),
  [3177] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_disjunctive_normal_form_type, 4, -1, 0),
  [3179] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_disjunctive_normal_form_type, 4, -1, 0),
  [3181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1348),
  [3183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(766),
  [3185] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_type, 2, 0, 0),
  [3187] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_type, 2, 0, 0),
  [3189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1370),
  [3191] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_intersection_type, 2, 0, 0),
  [3193] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_intersection_type, 2, 0, 0),
  [3195] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 2, 0, 0),
  [3197] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 2, 0, 0),
  [3199] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_disjunctive_normal_form_type, 2, -1, 0),
  [3201] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_disjunctive_normal_form_type, 2, -1, 0),
  [3203] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 4, 0, 0),
  [3205] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 4, 0, 0),
  [3207] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_disjunctive_normal_form_type, 3, -1, 0),
  [3209] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_disjunctive_normal_form_type, 3, -1, 0),
  [3211] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_intersection_type_repeat1, 2, 0, 0),
  [3213] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_intersection_type_repeat1, 2, 0, 0),
  [3215] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_union_type_repeat1, 2, 0, 0), REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 2, 0, 0),
  [3218] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_union_type_repeat1, 2, 0, 0), REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 2, 0, 0),
  [3221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2696),
  [3223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1593),
  [3225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1566),
  [3227] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_intersection_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1363),
  [3230] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1348),
  [3233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(86),
  [3235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(433),
  [3237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(844),
  [3239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2619),
  [3241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1595),
  [3243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1648),
  [3245] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type, 1, 0, 0),
  [3247] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type, 1, 0, 0),
  [3249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2581),
  [3251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1578),
  [3253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1658),
  [3255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1367),
  [3257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1352),
  [3259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1979),
  [3261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1835),
  [3263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2721),
  [3265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2310),
  [3267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(692),
  [3269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2014),
  [3271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1248),
  [3273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1516),
  [3275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1522),
  [3277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1271),
  [3279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(690),
  [3281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2008),
  [3283] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2721),
  [3286] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2310),
  [3289] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_switch_block_repeat1, 2, 0, 0),
  [3291] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_switch_block_repeat1, 2, 0, 0), SHIFT_REPEAT(315),
  [3294] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_switch_block_repeat1, 2, 0, 0), SHIFT_REPEAT(2448),
  [3297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2290),
  [3299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1468),
  [3301] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolated_string_body_heredoc, 2, 0, 0),
  [3303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1505),
  [3305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(640),
  [3307] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_string_part, 1, 0, 3),
  [3309] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_string_part, 1, 0, 3),
  [3311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2724),
  [3313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2737),
  [3315] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolated_string_body_heredoc, 1, 0, 0),
  [3317] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolated_string_body_heredoc, 1, 0, 0),
  [3319] = {.entry = {.count = 5, .reusable = true}}, REDUCE(sym_type, 1, 0, 0), REDUCE(sym_union_type, 1, 0, 0), REDUCE(sym_intersection_type, 1, 0, 0), REDUCE(sym_disjunctive_normal_form_type, 1, -1, 0), SHIFT(1366),
  [3325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1358),
  [3327] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_unset_statement_repeat1, 2, 0, 0),
  [3329] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_string_subscript_expression, 4, 0, 0),
  [3331] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_string_subscript_expression, 4, 0, 0),
  [3333] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_destructing_element, 3, 0, 0),
  [3335] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__list_destructing_repeat1, 4, 0, 0),
  [3337] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_string_part, 1, 0, 0),
  [3339] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_string_part, 1, 0, 0),
  [3341] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__complex_string_part, 3, 0, 0),
  [3343] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__complex_string_part, 3, 0, 0),
  [3345] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_string_member_access_expression, 3, 0, 57),
  [3347] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_string_member_access_expression, 3, 0, 57),
  [3349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(918),
  [3351] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1997),
  [3353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2003),
  [3355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1441),
  [3357] = {.entry = {.count = 1, .reusable = false}}, SHIFT(807),
  [3359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2024),
  [3361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2373),
  [3363] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2412),
  [3365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2028),
  [3367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [3369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2448),
  [3371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2029),
  [3373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2082),
  [3375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2087),
  [3377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2088),
  [3379] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolated_execution_operator_body, 1, 0, 0),
  [3381] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolated_execution_operator_body, 1, 0, 0),
  [3383] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolated_string_body, 2, 0, 0),
  [3385] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_if_statement, 3, 0, 23), SHIFT(2349),
  [3388] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_statement, 3, 0, 23), SHIFT(116),
  [3391] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 2, 0, 0),
  [3393] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1997),
  [3396] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_string_repeat1, 2, 0, 0), SHIFT_REPEAT(2003),
  [3399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(874),
  [3401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(382),
  [3403] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_if_statement, 4, 0, 53), SHIFT(2349),
  [3406] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_statement, 4, 0, 53), SHIFT(116),
  [3409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(944),
  [3411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(420),
  [3413] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolated_execution_operator_body, 2, 0, 0),
  [3415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1534),
  [3417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [3419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1668),
  [3421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1264),
  [3423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1536),
  [3425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(425),
  [3427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1669),
  [3429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(747),
  [3431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(427),
  [3433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1695),
  [3435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1696),
  [3437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1506),
  [3439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(429),
  [3441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1671),
  [3443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1612),
  [3445] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolated_string_body, 1, 0, 0),
  [3447] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolated_string_body, 1, 0, 0),
  [3449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1967),
  [3451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2703),
  [3453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2349),
  [3455] = {.entry = {.count = 1, .reusable = false}}, SHIFT(110),
  [3457] = {.entry = {.count = 5, .reusable = true}}, REDUCE(sym_type, 1, 0, 0), REDUCE(sym_union_type, 1, 0, 0), REDUCE(sym_intersection_type, 1, 0, 0), REDUCE(sym_disjunctive_normal_form_type, 1, -1, 0), SHIFT(1363),
  [3463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(494),
  [3465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2032),
  [3467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2740),
  [3469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1923),
  [3471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1923),
  [3473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2740),
  [3475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2178),
  [3477] = {.entry = {.count = 1, .reusable = false}}, SHIFT(999),
  [3479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(634),
  [3481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [3483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(463),
  [3485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2280),
  [3487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(798),
  [3489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(600),
  [3491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(328),
  [3493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(841),
  [3495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(428),
  [3497] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_union_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1365),
  [3500] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 4, 0, 196),
  [3502] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 4, 0, 196),
  [3504] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__property_hook_body, 1, 0, 154),
  [3506] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__property_hook_body, 1, 0, 154),
  [3508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1336),
  [3510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2358),
  [3512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1431),
  [3514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 4, 0, 122),
  [3516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(434),
  [3518] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 4, 0, 123),
  [3520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(435),
  [3522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1606),
  [3524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_formal_parameters, 4, 0, 0),
  [3526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1353),
  [3528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_name, 2, 0, 0),
  [3530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2604),
  [3532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2374),
  [3534] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_formal_parameters, 5, 0, 0),
  [3536] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 4, 0, 197),
  [3538] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 4, 0, 197),
  [3540] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_text_repeat1, 2, 0, 0),
  [3542] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_text_repeat1, 2, 0, 0), SHIFT_REPEAT(1923),
  [3545] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_text_repeat1, 2, 0, 0), SHIFT_REPEAT(1923),
  [3548] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_text_repeat1, 2, 0, 0),
  [3550] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 4, 0, 198),
  [3552] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 4, 0, 198),
  [3554] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_base_clause_repeat1, 2, 0, 0),
  [3556] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_base_clause_repeat1, 2, 0, 0), SHIFT_REPEAT(1505),
  [3559] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 4, 0, 199),
  [3561] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 4, 0, 199),
  [3563] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 4, 0, 117),
  [3565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(432),
  [3567] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 5, 0, 205),
  [3569] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 5, 0, 205),
  [3571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1786),
  [3573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2685),
  [3575] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 5, 0, 206),
  [3577] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 5, 0, 206),
  [3579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2370),
  [3581] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_text_repeat1, 2, 0, 0), SHIFT_REPEAT(1935),
  [3584] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_text_repeat1, 2, 0, 0), SHIFT_REPEAT(1935),
  [3587] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 5, 0, 207),
  [3589] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 5, 0, 207),
  [3591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1976),
  [3593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1895),
  [3595] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 5, 0, 208),
  [3597] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 5, 0, 208),
  [3599] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_formal_parameters, 2, 0, 0),
  [3601] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 5, 0, 157),
  [3603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [3605] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_statement_repeat1, 2, 0, 96), SHIFT_REPEAT(2349),
  [3608] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__function_definition_header, 3, 0, 21),
  [3610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1340),
  [3612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2650),
  [3614] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 73),
  [3616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1341),
  [3618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1365),
  [3620] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_name, 3, 0, 0),
  [3622] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_text, 1, 0, 0),
  [3624] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_text, 1, 0, 0),
  [3626] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_union_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1369),
  [3629] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1360),
  [3632] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 3, 0, 27),
  [3634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1360),
  [3636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2492),
  [3638] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 4, 0, 200),
  [3640] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 4, 0, 200),
  [3642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 3, 0, 79),
  [3644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(367),
  [3646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1883),
  [3648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1474),
  [3650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1475),
  [3652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(240),
  [3654] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_property_declaration_repeat1, 2, 0, 0),
  [3656] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_property_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(2290),
  [3659] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_name, 1, 0, 0),
  [3661] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 3, 0, 32),
  [3663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1330),
  [3665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2489),
  [3667] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_property_hook_list_repeat1, 1, 0, 0),
  [3669] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_property_hook_list_repeat1, 1, 0, 0),
  [3671] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 64),
  [3673] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__property_hook_body, 3, 0, 6),
  [3675] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__property_hook_body, 3, 0, 6),
  [3677] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1535),
  [3679] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_namespace_name_repeat1, 2, 0, 0),
  [3681] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_namespace_name_repeat1, 2, 0, 0), SHIFT_REPEAT(2604),
  [3684] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 6, 0, 212),
  [3686] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 6, 0, 212),
  [3688] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 2, 0, 38),
  [3690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [3692] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 3, 0, 84),
  [3694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [3696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2493),
  [3698] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 4, 0, 195),
  [3700] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 4, 0, 195),
  [3702] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 111),
  [3704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_if_statement_repeat2, 2, 0, 96),
  [3706] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_statement_repeat2, 2, 0, 96), SHIFT_REPEAT(2373),
  [3709] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_if_statement_repeat2, 2, 0, 96),
  [3711] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_element, 1, 0, 5),
  [3713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(372),
  [3715] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 70),
  [3717] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 2, 0, 153),
  [3719] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 2, 0, 153),
  [3721] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 3, 0, 178),
  [3723] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 3, 0, 178),
  [3725] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 3, 0, 179),
  [3727] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 3, 0, 179),
  [3729] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 3, 0, 180),
  [3731] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 3, 0, 180),
  [3733] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_formal_parameters, 3, 0, 0),
  [3735] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__function_definition_header, 4, 0, 42),
  [3737] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 3, 0, 181),
  [3739] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 3, 0, 181),
  [3741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2437),
  [3743] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 2, 0, 7),
  [3745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1369),
  [3747] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_intersection_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1367),
  [3750] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 3, 0, 78),
  [3752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(365),
  [3754] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1353),
  [3757] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2409),
  [3759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1952),
  [3761] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 3, 0, 19),
  [3763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2632),
  [3765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2306),
  [3767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2498),
  [3769] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__property_hook_body, 1, 0, 0),
  [3771] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__property_hook_body, 1, 0, 0),
  [3773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1973),
  [3775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1400),
  [3777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(992),
  [3779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(373),
  [3781] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__scope_resolution_qualifier, 1, 0, 0),
  [3783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1463),
  [3785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1464),
  [3787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(238),
  [3789] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_list_repeat1, 2, 0, 0),
  [3791] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1750),
  [3794] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_list, 1, 0, 0),
  [3796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1750),
  [3798] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_destructing, 4, 0, 0),
  [3800] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_base_clause_repeat1, 2, 0, 0), SHIFT_REPEAT(1529),
  [3803] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_function_static_declaration_repeat1, 2, 0, 0),
  [3805] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_function_static_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1976),
  [3808] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_clause, 1, 0, 0),
  [3810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2634),
  [3812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_global_declaration_repeat1, 2, 0, 0),
  [3814] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_global_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1895),
  [3817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2647),
  [3819] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_namespace_use_declaration_repeat1, 2, 0, 0),
  [3821] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_namespace_use_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1431),
  [3824] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 1, 0, 0),
  [3826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1476),
  [3828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1477),
  [3830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [3832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1472),
  [3834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1473),
  [3836] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_intersection_type, 2, 0, 0), SHIFT(1366),
  [3839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2061),
  [3841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(847),
  [3843] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_intersection_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1366),
  [3846] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_switch_block_repeat1, 1, 0, 0),
  [3848] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__const_declaration_repeat1, 2, 0, 0),
  [3850] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__const_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1606),
  [3853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1941),
  [3855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [3857] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_nowdoc_body_repeat1, 2, 0, 0),
  [3859] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_nowdoc_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1990),
  [3862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1458),
  [3864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1466),
  [3866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [3868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(258),
  [3870] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_content_repeat1, 2, 0, 0),
  [3872] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(2003),
  [3875] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_destructing, 3, 0, 0),
  [3877] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_nowdoc_body, 2, 0, 0),
  [3879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1990),
  [3881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1480),
  [3883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1481),
  [3885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [3887] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_clause, 2, 0, 17),
  [3889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2545),
  [3891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1485),
  [3893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1486),
  [3895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [3897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2039),
  [3899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(808),
  [3901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1529),
  [3903] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_base_clause, 2, 0, 0),
  [3905] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_text_repeat1, 1, 0, 0),
  [3907] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_text_repeat1, 1, 0, 0),
  [3909] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_static_variable_declaration, 1, 0, 5),
  [3911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(266),
  [3913] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_base_clause, 3, 0, 0),
  [3915] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_content, 1, 0, 0),
  [3917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [3919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2098),
  [3921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(989),
  [3923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1488),
  [3925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1490),
  [3927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [3929] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_literal, 1, 0, 0),
  [3931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2155),
  [3933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2521),
  [3935] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_list, 2, 0, 0),
  [3937] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_destructing_element, 1, 0, 10),
  [3939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1699),
  [3941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1700),
  [3943] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 3, 0, 85),
  [3945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(371),
  [3947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(173),
  [3949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1906),
  [3951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [3953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(985),
  [3955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1470),
  [3957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1373),
  [3959] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_if_clause_2, 3, 0, 23),
  [3961] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_if_clause_2, 3, 0, 23),
  [3963] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_match_block_repeat1, 2, 0, 0), SHIFT_REPEAT(184),
  [3966] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_match_block_repeat1, 2, 0, 0),
  [3968] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_match_condition_list_repeat1, 2, 0, 0), SHIFT_REPEAT(265),
  [3971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(937),
  [3973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2391),
  [3975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(966),
  [3977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1721),
  [3979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2483),
  [3981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 6, 0, 0),
  [3983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2262),
  [3985] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 6, 0, 139),
  [3987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2263),
  [3989] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 4, 0, 73),
  [3991] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_namespace_use_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1455),
  [3994] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__return_type, 2, 0, 43),
  [3996] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__function_definition_header, 4, 0, 45),
  [3998] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_nowdoc_body_repeat1, 1, 0, 0),
  [4000] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 1, 0, 5),
  [4002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [4004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [4006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1112),
  [4008] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 1, 0, 0),
  [4010] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_string_repeat1, 1, 0, 0),
  [4012] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_content_repeat1, 1, 0, 0),
  [4014] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_string_content_repeat1, 1, 0, 0),
  [4016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [4018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1628),
  [4020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1629),
  [4022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(243),
  [4024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(167),
  [4026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(805),
  [4028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1457),
  [4030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1345),
  [4032] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 3, 0, 19),
  [4034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(861),
  [4036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1819),
  [4038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 6, 0, 111),
  [4040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(169),
  [4042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(820),
  [4044] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 3, 0, 0),
  [4046] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__list_destructing_repeat1, 2, 0, 55),
  [4048] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_clause, 4, 0, 75),
  [4050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(166),
  [4052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1467),
  [4054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(838),
  [4056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [4058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(970),
  [4060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(154),
  [4062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(592),
  [4064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(862),
  [4066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [4068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(693),
  [4070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(172),
  [4072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1034),
  [4074] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__function_definition_header, 5, 0, 88),
  [4076] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 3, 0, 7),
  [4078] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_intersection_type, 1, 0, 0),
  [4080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(772),
  [4082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(774),
  [4084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(153),
  [4086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(174),
  [4088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1455),
  [4090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2446),
  [4092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1859),
  [4094] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 4, 0, 70),
  [4096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [4098] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__array_destructing_repeat1, 2, 0, 0), SHIFT_REPEAT(173),
  [4101] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__array_destructing_repeat1, 2, 0, 0),
  [4103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1461),
  [4105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1981),
  [4107] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 5, 0, 0),
  [4109] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 5, 0, 56),
  [4111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(896),
  [4113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1446),
  [4115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(986),
  [4117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(155),
  [4119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(879),
  [4121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [4123] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface_clause, 3, 0, 0),
  [4125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1544),
  [4127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(150),
  [4129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(980),
  [4131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [4133] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 4, 0, 0),
  [4135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [4137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(737),
  [4139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [4141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [4143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(827),
  [4145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(149),
  [4147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2656),
  [4149] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_formal_parameters_repeat1, 2, 0, 0), SHIFT_REPEAT(886),
  [4152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_formal_parameters_repeat1, 2, 0, 0),
  [4154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2093),
  [4156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 4, 0, 56),
  [4158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(147),
  [4160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2540),
  [4162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2666),
  [4164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1918),
  [4166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(144),
  [4168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(654),
  [4170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(850),
  [4172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1871),
  [4174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2094),
  [4176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 3, 0, 27),
  [4178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2420),
  [4180] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_anonymous_function_use_clause_repeat1, 2, 0, 0), SHIFT_REPEAT(1936),
  [4183] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_anonymous_function_use_clause_repeat1, 2, 0, 0),
  [4185] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__list_destructing_repeat1, 2, 0, 0), SHIFT_REPEAT(181),
  [4188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [4190] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(158),
  [4193] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_arguments_repeat1, 2, 0, 0),
  [4195] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 3, 0, 32),
  [4197] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_intersection_type, 2, 0, 0), SHIFT(1363),
  [4200] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 64),
  [4202] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 27),
  [4204] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 4, 0, 64),
  [4206] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_if_statement_repeat2, 1, 0, 51),
  [4208] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_if_statement_repeat2, 1, 0, 51),
  [4210] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_destructing_element, 3, 0, 56),
  [4212] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 3, 0, 81),
  [4214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(368),
  [4216] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 70),
  [4218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2678),
  [4220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2406),
  [4222] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_as_clause, 3, 0, 0),
  [4224] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_clause, 3, 0, 35),
  [4226] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 73),
  [4228] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 5, 0, 111),
  [4230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 2, 0, 37),
  [4232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [4234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(864),
  [4236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2404),
  [4238] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 2, 0, 39),
  [4240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [4242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2460),
  [4244] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 2, 0, 40),
  [4246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [4248] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 19),
  [4250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2382),
  [4252] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_array_creation_expression_repeat1, 2, 0, 0),
  [4254] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_array_creation_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(176),
  [4257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2595),
  [4259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2615),
  [4261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1933),
  [4263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1533),
  [4265] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 7, 0, 0),
  [4267] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 7, 0, 139),
  [4269] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 32),
  [4271] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 2, 0, 7),
  [4273] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 3, 0, 83),
  [4275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(369),
  [4277] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_group_repeat1, 2, 0, 0), SHIFT_REPEAT(1501),
  [4280] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_group_repeat1, 2, 0, 0),
  [4282] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 4, 0, 125),
  [4284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(436),
  [4286] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(710),
  [4289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1776),
  [4291] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_array_creation_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(178),
  [4294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2352),
  [4296] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface_clause, 2, 0, 0),
  [4298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2478),
  [4300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(696),
  [4302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2432),
  [4304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2633),
  [4306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(991),
  [4308] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__array_destructing_element, 1, 0, 0), REDUCE(sym_array_element_initializer, 1, 0, 0),
  [4311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(84),
  [4313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [4315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [4317] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 7, 0, 182),
  [4319] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_namespace_name_repeat1, 2, 0, 0), SHIFT_REPEAT(2524),
  [4322] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 7, 0, 183),
  [4324] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 8, 0, 201),
  [4326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [4328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2603),
  [4330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(839),
  [4332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [4334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2588),
  [4336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1005),
  [4338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2688),
  [4340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(938),
  [4342] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__namespace_use_group, 4, 0, 76),
  [4344] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 5, 0, 122),
  [4346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1931),
  [4348] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 5, 0, 123),
  [4350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__list_destructing_repeat1, 4, 0, 138),
  [4352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2631),
  [4354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(770),
  [4356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2510),
  [4358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(773),
  [4360] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 5, 0, 159),
  [4362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1873),
  [4364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2707),
  [4366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(979),
  [4368] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function_use_clause, 5, 0, 0),
  [4370] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_group, 4, 0, 0),
  [4372] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 5, 0, 117),
  [4374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(95),
  [4376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2627),
  [4378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1007),
  [4380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [4382] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 3, 0, 80),
  [4384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2571),
  [4386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(780),
  [4388] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_as_clause, 4, 0, 0),
  [4390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2572),
  [4392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(782),
  [4394] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 5, 0, 118),
  [4396] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 3, 0, 77),
  [4398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(82),
  [4400] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__array_destructing_element, 3, 0, 0), REDUCE(sym_array_element_initializer, 3, 0, 0),
  [4403] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cast_type, 1, 0, 0),
  [4405] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__namespace_use_group, 3, 0, 36),
  [4407] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 2, 0, 16),
  [4409] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_group, 3, 0, 0),
  [4411] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 2, 0, 7),
  [4413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4),
  [4415] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 4, 0, 78),
  [4417] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 4, 0, 79),
  [4419] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 7, 0, 177),
  [4421] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 3, 0, 82),
  [4423] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 6, 0, 151),
  [4425] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 6, 0, 152),
  [4427] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 4, 0, 119),
  [4429] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 4, 0, 121),
  [4431] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_namespace_name, 3, 0, 0), SHIFT(2524),
  [4434] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_instead_of_clause, 3, 0, 0),
  [4436] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 4, 0, 84),
  [4438] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 4, 0, 124),
  [4440] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function_use_clause, 4, 0, 0),
  [4442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2741),
  [4444] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 6, 0, 157),
  [4446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1920),
  [4448] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 6, 0, 158),
  [4450] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function_use_clause, 6, 0, 0),
  [4452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1934),
  [4454] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 3, 0, 38),
  [4456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [4458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [4460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2613),
  [4462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1694),
  [4464] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 104),
  [4466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1462),
  [4468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(781),
  [4470] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 105),
  [4472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(179),
  [4474] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 65),
  [4476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1966),
  [4478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2005),
  [4480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(331),
  [4482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2081),
  [4484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 4, 0, 65),
  [4486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2516),
  [4488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1789),
  [4490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2620),
  [4492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(17),
  [4494] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 5, 0, 104),
  [4496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(443),
  [4498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2312),
  [4500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(332),
  [4502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(203),
  [4504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1335),
  [4506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2214),
  [4508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(32),
  [4510] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 108),
  [4512] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 109),
  [4514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1886),
  [4516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2018),
  [4518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(23),
  [4520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2092),
  [4522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(24),
  [4524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2036),
  [4526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(345),
  [4528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [4530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(441),
  [4532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2336),
  [4534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1858),
  [4536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(40),
  [4538] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 5, 0, 108),
  [4540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1747),
  [4542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(42),
  [4544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2095),
  [4546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1632),
  [4548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(15),
  [4550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1673),
  [4552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [4554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2279),
  [4556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [4558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1634),
  [4560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2676),
  [4562] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 112),
  [4564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(697),
  [4566] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 5, 0, 112),
  [4568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1479),
  [4570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(785),
  [4572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(786),
  [4574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(704),
  [4576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1691),
  [4578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [4580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(206),
  [4582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(35),
  [4584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(701),
  [4586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [4588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1686),
  [4590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1619),
  [4592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2062),
  [4594] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 87),
  [4596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 3, 0, 20),
  [4598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(712),
  [4600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_foreach_pair, 3, 0, 0),
  [4602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2019),
  [4604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(929),
  [4606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1643),
  [4608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(26),
  [4610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(268),
  [4612] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 41),
  [4614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2175),
  [4616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1491),
  [4618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1845),
  [4620] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 44),
  [4622] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_clause_2, 2, 0, 6),
  [4624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2398),
  [4626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1633),
  [4628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [4630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(771),
  [4632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1860),
  [4634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1588),
  [4636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(211),
  [4638] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 3, 0, 20),
  [4640] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 4, 0, 71),
  [4642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1647),
  [4644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1947),
  [4646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(713),
  [4648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1532),
  [4650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2646),
  [4652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2504),
  [4654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(188),
  [4656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1655),
  [4658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1514),
  [4660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2528),
  [4662] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 71),
  [4664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2438),
  [4666] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 7, 0, 175),
  [4668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(858),
  [4670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(930),
  [4672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1914),
  [4674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(855),
  [4676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1478),
  [4678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(779),
  [4680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1624),
  [4682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(983),
  [4684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2236),
  [4686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1039),
  [4688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(439),
  [4690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(396),
  [4692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [4694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(442),
  [4696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1520),
  [4698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_string_subscript_unary_expression, 2, 0, 0),
  [4700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1662),
  [4702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2628),
  [4704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1896),
  [4706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1974),
  [4708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1384),
  [4710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(705),
  [4712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 6, 0, 143),
  [4714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1879),
  [4716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2711),
  [4718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1849),
  [4720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [4722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(706),
  [4724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(41),
  [4726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2350),
  [4728] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 6, 0, 147),
  [4730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1471),
  [4732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(19),
  [4734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1847),
  [4736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(707),
  [4738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1866),
  [4740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(28),
  [4742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2582),
  [4744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(440),
  [4746] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 6, 0, 148),
  [4748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(10),
  [4750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(29),
  [4752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(192),
  [4754] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [4756] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declare_directive, 3, 0, 0),
  [4758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2120),
  [4760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(711),
  [4762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1047),
  [4764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [4766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2449),
  [4768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(51),
  [4770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1484),
  [4772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(14),
  [4774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [4776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(37),
  [4778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1753),
  [4780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(39),
  [4782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1540),
  [4784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1006),
  [4786] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 6, 0, 144),
  [4788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1916),
  [4790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1487),
  [4792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2369),
  [4794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1897),
  [4796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1713),
  [4798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1821),
  [4800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1613),
  [4802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1621),
  [4804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1870),
  [4806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [4808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2433),
  [4810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2473),
  [4812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 6, 0, 147),
  [4814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2485),
  [4816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1004),
  [4818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(30),
  [4820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(222),
  [4822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1530),
  [4824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(899),
  [4826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [4828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2533),
  [4830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1909),
  [4832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(198),
  [4834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1840),
  [4836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(213),
  [4838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(9),
  [4840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1969),
  [4842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1830),
  [4844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1702),
  [4846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [4848] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_string_array_access_argument, 1, 0, 0),
  [4850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(217),
  [4852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(218),
  [4854] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 4, 0, 41),
  [4856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1949),
  [4858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [4860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2234),
  [4862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1543),
  [4864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [4866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [4868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [4870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2642),
  [4872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(353),
  [4874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(199),
  [4876] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_text_interpolation, 2, 0, 0),
  [4878] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_text_interpolation, 3, 0, 0),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token__automatic_semicolon = 0,
  ts_external_token_encapsed_string_chars = 1,
  ts_external_token_encapsed_string_chars_after_variable = 2,
  ts_external_token_execution_string_chars = 3,
  ts_external_token_execution_string_chars_after_variable = 4,
  ts_external_token_encapsed_string_chars_heredoc = 5,
  ts_external_token_encapsed_string_chars_after_variable_heredoc = 6,
  ts_external_token__eof = 7,
  ts_external_token_heredoc_start = 8,
  ts_external_token_heredoc_end = 9,
  ts_external_token_nowdoc_string = 10,
  ts_external_token_sentinel_error = 11,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token__automatic_semicolon] = sym__automatic_semicolon,
  [ts_external_token_encapsed_string_chars] = sym_encapsed_string_chars,
  [ts_external_token_encapsed_string_chars_after_variable] = sym_encapsed_string_chars_after_variable,
  [ts_external_token_execution_string_chars] = sym_execution_string_chars,
  [ts_external_token_execution_string_chars_after_variable] = sym_execution_string_chars_after_variable,
  [ts_external_token_encapsed_string_chars_heredoc] = sym_encapsed_string_chars_heredoc,
  [ts_external_token_encapsed_string_chars_after_variable_heredoc] = sym_encapsed_string_chars_after_variable_heredoc,
  [ts_external_token__eof] = sym__eof,
  [ts_external_token_heredoc_start] = sym_heredoc_start,
  [ts_external_token_heredoc_end] = sym_heredoc_end,
  [ts_external_token_nowdoc_string] = sym_nowdoc_string,
  [ts_external_token_sentinel_error] = sym_sentinel_error,
};

static const bool ts_external_scanner_states[15][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token__automatic_semicolon] = true,
    [ts_external_token_encapsed_string_chars] = true,
    [ts_external_token_encapsed_string_chars_after_variable] = true,
    [ts_external_token_execution_string_chars] = true,
    [ts_external_token_execution_string_chars_after_variable] = true,
    [ts_external_token_encapsed_string_chars_heredoc] = true,
    [ts_external_token_encapsed_string_chars_after_variable_heredoc] = true,
    [ts_external_token__eof] = true,
    [ts_external_token_heredoc_start] = true,
    [ts_external_token_heredoc_end] = true,
    [ts_external_token_nowdoc_string] = true,
    [ts_external_token_sentinel_error] = true,
  },
  [2] = {
    [ts_external_token__automatic_semicolon] = true,
  },
  [3] = {
    [ts_external_token_encapsed_string_chars_heredoc] = true,
    [ts_external_token_heredoc_end] = true,
  },
  [4] = {
    [ts_external_token_encapsed_string_chars_heredoc] = true,
  },
  [5] = {
    [ts_external_token_encapsed_string_chars] = true,
  },
  [6] = {
    [ts_external_token_execution_string_chars] = true,
  },
  [7] = {
    [ts_external_token_encapsed_string_chars_heredoc] = true,
    [ts_external_token_encapsed_string_chars_after_variable_heredoc] = true,
    [ts_external_token_heredoc_end] = true,
  },
  [8] = {
    [ts_external_token_encapsed_string_chars] = true,
    [ts_external_token_encapsed_string_chars_after_variable] = true,
  },
  [9] = {
    [ts_external_token_execution_string_chars] = true,
    [ts_external_token_execution_string_chars_after_variable] = true,
  },
  [10] = {
    [ts_external_token__eof] = true,
  },
  [11] = {
    [ts_external_token_heredoc_end] = true,
  },
  [12] = {
    [ts_external_token_heredoc_end] = true,
    [ts_external_token_nowdoc_string] = true,
  },
  [13] = {
    [ts_external_token_heredoc_start] = true,
  },
  [14] = {
    [ts_external_token_nowdoc_string] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_php_external_scanner_create(void);
void tree_sitter_php_external_scanner_destroy(void *);
bool tree_sitter_php_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_php_external_scanner_serialize(void *, char *);
void tree_sitter_php_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_php(void) {
  static const TSLanguage language = {
    .version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym_name,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_php_external_scanner_create,
      tree_sitter_php_external_scanner_destroy,
      tree_sitter_php_external_scanner_scan,
      tree_sitter_php_external_scanner_serialize,
      tree_sitter_php_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
